/**

  This file fills the remainder lookup table. The bit mask is a short way of encoding
  the allowed remainders for a given base modulo some value. The basic derivation
  can be described with the example of base 10 repunits:

  Similar to Mersenne factors we find for base 10 repunits that Legendre(10,p)===1 (10 is a quadratic residue mod p)
  Legendre(10,p) = Legendre(2,p)*Legendre(5,p)
                                     / +1 if p = 1 or 7 mod 8
  Legendre(2,p) = (-1)^((p^2-1)/8) = |
                                     \ -1 if p = 3 or 5 mod 8

                                        / +1 if p = 1 or 4 mod 5
  Legendre(5,p) = (-1)^Floor((p+2)/5) = |
                                        \ -1 if p = 2 or 3 mod 5

  The product must be +1, so either both products of (-1)(-1) and (+1)(+1) are possible.
  Filtering the possible values mod 40=gcd(8,5) gives the following allowed values:
  {1,3,9,13,27,31,37,39} or {±1, ±3, ±9, ±13} for 2kp+1.
  See also: https://math.stackexchange.com/questions/1767306/find-all-prime-p-such-that-legendre-symbol-of-left-frac10p-right-1

  Encoding this as a bit map gives 2^1+2^3+2^9+2^13+2^27+2^31+2^37+2^39 = 689476477450.
  In base 2 this is 1010000010001000000000000010001000001010 (each 1 marks the allowed remainder)
  or in hex 0xa08800220a. Splitting this into 32 bit blocks we get 0xa0 0x8800220a. Storing these in increasing
  block order this is then [0x8800220a,0xa0] as defined for remainders_pos[10].

  The Mathematica program to generate this file:
  Some note: This program is certainly not optimal (code wise).

    mapCR[n_, a_, r_] := ChineseRemainder[{a, #}, {4, n}] & /@ r;
    mapCRJoin[n_, r_, s_] := Sort[mapCR[n, 1, r[s]]~Join~mapCR[n, 3, r[-s]]];
    (* Return remainders for prime parameters plus special case of -1 *)

    remainders[n_] := remainders[n] = Block[{r},
        If[n > 2,
          If[PrimeQ[n],
            r = GroupBy[Range[n - 1], JacobiSymbol[#, n] &];
            If[Mod[n, 4] == 1, Return[{#, n} & /@ r]];
            If[Mod[n, 4] == 3, Return[{#, LCM[4, n]} & /@ AssociationMap[mapCRJoin[n, r, #] &, {1, -1}]]];
            ];
          ];
        If[n == 2, Return[{#, 8} & /@ AssociationThread[{1, -1} -> {{1, 7}, {3, 5}}]]];
        If[n == -1, Return[{#, 4} & /@ AssociationThread[{1, -1} -> {{1}, {3}}]]];
        Abort[];
        ];
    lazyOrMod[mod_, list_] := lazyOrMod[mod, list] = With[{m = mod, l = list}, Apply[Or, m == # & /@ l]]
    lazyAnd[l_, q_] := Hold[Apply[And, lazyOrMod[Mod[l, #[[2]]], #[[1]]] & /@ q]];
    lazyOr[l_, p_] := Apply[Or, lazyAnd[l, #] & /@ p];

    (* Combine remainder classes *)
    combineRemainders[rel__] := With[{lcm = Fold[LCM, 1, Apply[LCM, #] & /@ ((#\[Transpose][[2]]) & /@ {rel})]},
       {Select[Table[i, {i, lcm}], ReleaseHold[lazyOr[#, {rel}]] &], lcm}
       ];
    createTuples[n_] := createTuples[n] = Select[Tuples[{-1, 1}, n], Fold[Times, 1, #] == 1 &];
    findSolutions[n_] := findSolutions[n] = With[
        (* Split into prime factors, also -1 is here a factor *)
        {factors = Flatten[ConstantArray @@@ FactorInteger[n]]},
        (* Create possible tuples *)
        Map[Inner[remainders[#1][#2] &, factors, #, {##} &] &, createTuples[Length[factors]]]
        ];

    (* Entry point to calculate the remainder classes for a general \
    repunit of base n, all numbers except 0 and 1 are allowed *)

    calcRemainders[n_] := calcRemainders[n] = If[n == 0 || n == 1, Return[{}], combineRemainders @@ findSolutions[n]];

    (* Create the file needed for gr-mfaktc *)
    limit = 1000;
    convertToHexList[r_] := {Total /@ Map[2^Mod[#, 32] &, SplitBy[r[[1]], Floor[#/32] &]], r[[2]]}
    Monitor[remaindersPositiveBases = Reap[Do[Sow[i -> convertToHexList[calcRemainders[i]]], {i, 2, limit}]][[2]][[1]];, i];
    Monitor[remaindersNegativeBases = Reap[Do[Sow[-i -> convertToHexList[calcRemainders[-i]]], {i, 2, limit}]][[2]][[1]];, i];
    arpos = Association[remaindersPositiveBases];
    arneg = Association[remaindersNegativeBases];
    keyspos = Keys[arpos];
    keysneg = Keys[arneg];
    valuespos = Values[arpos];
    valuesneg = Values[arneg];
    writeRemainderLine[posNegString_, values_, stream_] := Module[{r},
      r = "remainders_" <> posNegString <> "[" <> ToString[Abs[keyspos[[i - 1]]]] <> "]";
      WriteLine[stream, r <> ".modulo_value=" <> ToString[values[[2]]] <> "; " <> r <> ".bit_mask=create_bit_mask(" <> ToString[Length[values[[1]]]] <> "," <> ToString[Row["0x" <> IntegerString[#, 16] & /@ values[[1]], ","]] <> ");"];
      ]
    stream = OpenWrite["allowed-remainders-data.c"];
    For[i = 2, i <= limit, i++,
      writeRemainderLine["pos", valuespos[[i - 1]], stream];
      writeRemainderLine["neg", valuesneg[[i - 1]], stream]
    ];
    Close[stream];
*/

remainders_pos[2].modulo_value=8; remainders_pos[2].bit_mask=create_bit_mask(1,0x82);
remainders_neg[2].modulo_value=8; remainders_neg[2].bit_mask=create_bit_mask(1,0xa);
remainders_pos[3].modulo_value=12; remainders_pos[3].bit_mask=create_bit_mask(1,0x802);
remainders_neg[3].modulo_value=12; remainders_neg[3].bit_mask=create_bit_mask(1,0x82);
remainders_pos[4].modulo_value=8; remainders_pos[4].bit_mask=create_bit_mask(1,0xaa);
remainders_neg[4].modulo_value=8; remainders_neg[4].bit_mask=create_bit_mask(1,0x22);
remainders_pos[5].modulo_value=5; remainders_pos[5].bit_mask=create_bit_mask(1,0x12);
remainders_neg[5].modulo_value=20; remainders_neg[5].bit_mask=create_bit_mask(1,0x28a);
remainders_pos[6].modulo_value=24; remainders_pos[6].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[6].modulo_value=24; remainders_neg[6].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[7].modulo_value=28; remainders_pos[7].bit_mask=create_bit_mask(1,0xa08020a);
remainders_neg[7].modulo_value=28; remainders_neg[7].bit_mask=create_bit_mask(1,0x2808a02);
remainders_pos[8].modulo_value=8; remainders_pos[8].bit_mask=create_bit_mask(1,0x82);
remainders_neg[8].modulo_value=8; remainders_neg[8].bit_mask=create_bit_mask(1,0xa);
remainders_pos[9].modulo_value=12; remainders_pos[9].bit_mask=create_bit_mask(1,0x8a2);
remainders_neg[9].modulo_value=12; remainders_neg[9].bit_mask=create_bit_mask(1,0x22);
remainders_pos[10].modulo_value=40; remainders_pos[10].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[10].modulo_value=40; remainders_neg[10].bit_mask=create_bit_mask(2,0x882a82,0x20);
remainders_pos[11].modulo_value=44; remainders_pos[11].bit_mask=create_bit_mask(2,0x20802a2,0x8a8);
remainders_neg[11].modulo_value=44; remainders_neg[11].bit_mask=create_bit_mask(2,0x8a80822a,0x20);
remainders_pos[12].modulo_value=24; remainders_pos[12].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[12].modulo_value=24; remainders_neg[12].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[13].modulo_value=13; remainders_pos[13].bit_mask=create_bit_mask(1,0x161a);
remainders_neg[13].modulo_value=52; remainders_neg[13].bit_mask=create_bit_mask(2,0xa20a8a82,0x28000);
remainders_pos[14].modulo_value=56; remainders_pos[14].bit_mask=create_bit_mask(2,0x82002a22,0x88a800);
remainders_neg[14].modulo_value=56; remainders_neg[14].bit_mask=create_bit_mask(2,0xa88a22a,0x2080);
remainders_pos[15].modulo_value=60; remainders_pos[15].bit_mask=create_bit_mask(2,0x20882,0x8220800);
remainders_neg[15].modulo_value=60; remainders_neg[15].bit_mask=create_bit_mask(2,0x808a0002,0x228000);
remainders_pos[16].modulo_value=8; remainders_pos[16].bit_mask=create_bit_mask(1,0xaa);
remainders_neg[16].modulo_value=8; remainders_neg[16].bit_mask=create_bit_mask(1,0x22);
remainders_pos[17].modulo_value=17; remainders_pos[17].bit_mask=create_bit_mask(1,0x1a316);
remainders_neg[17].modulo_value=68; remainders_neg[17].bit_mask=create_bit_mask(2,0x8aa02a8a,0x80220082);
remainders_pos[18].modulo_value=24; remainders_pos[18].bit_mask=create_bit_mask(1,0x820082);
remainders_neg[18].modulo_value=24; remainders_neg[18].bit_mask=create_bit_mask(1,0xa0802);
remainders_pos[19].modulo_value=76; remainders_pos[19].bit_mask=create_bit_mask(3,0x8a02822a,0x280a2000,0xa88);
remainders_neg[19].modulo_value=76; remainders_neg[19].bit_mask=create_bit_mask(3,0x2820aa2,0xa082a888,0x200);
remainders_pos[20].modulo_value=40; remainders_pos[20].bit_mask=create_bit_mask(2,0xa0280a02,0x80);
remainders_neg[20].modulo_value=40; remainders_neg[20].bit_mask=create_bit_mask(1,0x28a0028a);
remainders_pos[21].modulo_value=84; remainders_pos[21].bit_mask=create_bit_mask(3,0x2020022,0x8008a20,0x88008);
remainders_neg[21].modulo_value=84; remainders_neg[21].bit_mask=create_bit_mask(3,0x828a0822,0x800220,0x80);
remainders_pos[22].modulo_value=88; remainders_pos[22].bit_mask=create_bit_mask(3,0x2a20228a,0xa8020080,0xa28808);
remainders_neg[22].modulo_value=88; remainders_neg[22].bit_mask=create_bit_mask(3,0xa2a8a202,0x200a8808,0x2a0080);
remainders_pos[23].modulo_value=92; remainders_pos[23].bit_mask=create_bit_mask(3,0x2208aa82,0x800a0a00,0x82aa208);
remainders_neg[23].modulo_value=92; remainders_neg[23].bit_mask=create_bit_mask(3,0xaa00220a,0x8828288,0xa22a80);
remainders_pos[24].modulo_value=24; remainders_pos[24].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[24].modulo_value=24; remainders_neg[24].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[25].modulo_value=5; remainders_pos[25].bit_mask=create_bit_mask(1,0x1e);
remainders_neg[25].modulo_value=20; remainders_neg[25].bit_mask=create_bit_mask(1,0x22202);
remainders_pos[26].modulo_value=104; remainders_pos[26].bit_mask=create_bit_mask(4,0x2aa0a22,0x8822020,0xa0aa8008,0x88);
remainders_neg[26].modulo_value=104; remainders_neg[26].bit_mask=create_bit_mask(3,0x8a2282aa,0x800aa828,0x20220880);
remainders_pos[27].modulo_value=12; remainders_pos[27].bit_mask=create_bit_mask(1,0x802);
remainders_neg[27].modulo_value=12; remainders_neg[27].bit_mask=create_bit_mask(1,0x82);
remainders_pos[28].modulo_value=56; remainders_pos[28].bit_mask=create_bit_mask(2,0xaa08020a,0xa08020);
remainders_neg[28].modulo_value=56; remainders_neg[28].bit_mask=create_bit_mask(2,0x22808a02,0x2808a0);
remainders_pos[29].modulo_value=29; remainders_pos[29].bit_mask=create_bit_mask(1,0x13d122f2);
remainders_neg[29].modulo_value=116; remainders_neg[29].bit_mask=create_bit_mask(4,0x8a08aa2a,0x2a2a882,0xa0028802,0x2008);
remainders_pos[30].modulo_value=120; remainders_pos[30].bit_mask=create_bit_mask(4,0x200a2082,0x20020,0x8080080,0x8208a0);
remainders_neg[30].modulo_value=120; remainders_neg[30].bit_mask=create_bit_mask(4,0xa0822802,0x8028820,0x8008,0x20020);
remainders_pos[31].modulo_value=124; remainders_pos[31].bit_mask=create_bit_mask(4,0xa808a2a,0x822a02,0x80a8820,0xa8a202a);
remainders_neg[31].modulo_value=124; remainders_neg[31].bit_mask=create_bit_mask(4,0x20802a2,0x880aa28a,0x808200a8,0x202a8a2);
remainders_pos[32].modulo_value=8; remainders_pos[32].bit_mask=create_bit_mask(1,0x82);
remainders_neg[32].modulo_value=8; remainders_neg[32].bit_mask=create_bit_mask(1,0xa);
remainders_pos[33].modulo_value=132; remainders_pos[33].bit_mask=create_bit_mask(5,0xa2020002,0x20228,0x8808000a,0x808a2,0x8);
remainders_neg[33].modulo_value=132; remainders_neg[33].bit_mask=create_bit_mask(4,0x228a0082,0x8028a20,0x8082,0x80800022);
remainders_pos[34].modulo_value=136; remainders_pos[34].bit_mask=create_bit_mask(5,0x2a008a2a,0x2082a022,0xa820800,0xa200a888,0xa8);
remainders_neg[34].modulo_value=136; remainders_neg[34].bit_mask=create_bit_mask(5,0xa28802a2,0xa80228aa,0x820a8088,0x2a082000,0x20);
remainders_pos[35].modulo_value=140; remainders_pos[35].bit_mask=create_bit_mask(5,0xa08a2202,0x8000802,0x20208,0x8a20a802,0x808);
remainders_neg[35].modulo_value=140; remainders_neg[35].bit_mask=create_bit_mask(4,0x28022a0a,0x88082,0x8a8280,0x220208a);
remainders_pos[36].modulo_value=24; remainders_pos[36].bit_mask=create_bit_mask(1,0x8a28a2);
remainders_neg[36].modulo_value=24; remainders_neg[36].bit_mask=create_bit_mask(1,0x22022);
remainders_pos[37].modulo_value=37; remainders_pos[37].bit_mask=create_bit_mask(2,0x5e211e9a,0x16);
remainders_neg[37].modulo_value=148; remainders_neg[37].bit_mask=create_bit_mask(5,0x82a88202,0x8aa0a8a,0x8a2a202,0x28000a0,0x2a288);
remainders_pos[38].modulo_value=152; remainders_pos[38].bit_mask=create_bit_mask(5,0xa222aa02,0x220828,0xa82a0,0x8a282088,0x80aa88);
remainders_neg[38].modulo_value=152; remainders_neg[38].bit_mask=create_bit_mask(5,0x2aa2228a,0x88aa80a0,0x8820a28,0x2a0a800,0x82200);
remainders_pos[39].modulo_value=156; remainders_pos[39].bit_mask=create_bit_mask(5,0x828800a2,0x20020208,0x82000008,0x22080800,0x8a00228);
remainders_neg[39].modulo_value=156; remainders_neg[39].bit_mask=create_bit_mask(5,0x2000822,0x28828a00,0x2088080,0xa2800080,0x200a20);
remainders_pos[40].modulo_value=40; remainders_pos[40].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[40].modulo_value=40; remainders_neg[40].bit_mask=create_bit_mask(2,0x882a82,0x20);
remainders_pos[41].modulo_value=41; remainders_pos[41].bit_mask=create_bit_mask(2,0x82b50736,0x1b3);
remainders_neg[41].modulo_value=164; remainders_neg[41].bit_mask=create_bit_mask(5,0xa288aaa,0xa282a02a,0x8002aa88,0x22028208,0x8820a0);
remainders_pos[42].modulo_value=168; remainders_pos[42].bit_mask=create_bit_mask(6,0x220a2802,0x20208200,0x2008000,0x82080800,0x28a08800,0x80);
remainders_neg[42].modulo_value=168; remainders_neg[42].bit_mask=create_bit_mask(6,0xa2822002,0x28a00a00,0x82080088,0x2000080,0x20200008,0x8);
remainders_pos[43].modulo_value=172; remainders_pos[43].bit_mask=create_bit_mask(6,0xa2a228a,0x82aa0280,0x8020880,0xaa82022,0x8a8a0028,0xa28);
remainders_neg[43].modulo_value=172; remainders_neg[43].bit_mask=create_bit_mask(6,0x82a2aa02,0xa228208,0x808a8008,0x8220a8aa,0x20288a0,0x2a0);
remainders_pos[44].modulo_value=88; remainders_pos[44].bit_mask=create_bit_mask(3,0x20802a2,0x802a28a8,0x8a8020);
remainders_neg[44].modulo_value=88; remainders_neg[44].bit_mask=create_bit_mask(3,0x8a80822a,0x822a020,0x208a8);
remainders_pos[45].modulo_value=60; remainders_pos[45].bit_mask=create_bit_mask(2,0xa0080802,0x8020200);
remainders_neg[45].modulo_value=60; remainders_neg[45].bit_mask=create_bit_mask(2,0x20800082,0x28a00);
remainders_pos[46].modulo_value=184; remainders_pos[46].bit_mask=create_bit_mask(6,0xa2082aa,0xa8222228,0x28a00,0x2a00a280,0xa0288888,0xaa8208);
remainders_neg[46].modulo_value=184; remainders_neg[46].bit_mask=create_bit_mask(6,0x82280a22,0x20aaaaa0,0x888a0288,0xa2802a08,0x28a00000,0x220a80);
remainders_pos[47].modulo_value=188; remainders_pos[47].bit_mask=create_bit_mask(6,0x82aa8a02,0x202208a8,0xa82000a,0x8a00082a,0x22a20880,0x80a2aa8);
remainders_neg[47].modulo_value=188; remainders_neg[47].bit_mask=create_bit_mask(6,0xa22028a,0xa8aa0020,0x820a8882,0x28880a2,0xaa2a8008,0x82a220);
remainders_pos[48].modulo_value=24; remainders_pos[48].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[48].modulo_value=24; remainders_neg[48].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[49].modulo_value=28; remainders_pos[49].bit_mask=create_bit_mask(1,0xa8aaa2a);
remainders_neg[49].modulo_value=28; remainders_neg[49].bit_mask=create_bit_mask(1,0x2022222);
remainders_pos[50].modulo_value=40; remainders_pos[50].bit_mask=create_bit_mask(2,0x80820282,0x82);
remainders_neg[50].modulo_value=40; remainders_neg[50].bit_mask=create_bit_mask(2,0x80a0a0a,0x2);
remainders_pos[51].modulo_value=204; remainders_pos[51].bit_mask=create_bit_mask(7,0xa20020a2,0x8028208,0x8088002,0x22020000,0x28020800,0x8008a208,0x8a0);
remainders_neg[51].modulo_value=204; remainders_neg[51].bit_mask=create_bit_mask(7,0x22882822,0x820a00,0x8000008a,0xa20a0880,0x20828008,0x2280,0x20);
remainders_pos[52].modulo_value=104; remainders_pos[52].bit_mask=create_bit_mask(4,0x2a82020a,0x20aa0808,0x8082a820,0xa0);
remainders_neg[52].modulo_value=104; remainders_neg[52].bit_mask=create_bit_mask(4,0xa20a8a82,0xa8228000,0xa20a8,0x28);
remainders_pos[53].modulo_value=53; remainders_pos[53].bit_mask=create_bit_mask(2,0x3303aed2,0x12dd70);
remainders_neg[53].modulo_value=212; remainders_neg[53].bit_mask=create_bit_mask(7,0xaa8a220a,0x28a00a8,0x228aa8a8,0x82228282,0x2a80800,0x80080aa0,0xa220);
remainders_pos[54].modulo_value=24; remainders_pos[54].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[54].modulo_value=24; remainders_neg[54].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[55].modulo_value=220; remainders_pos[55].bit_mask=create_bit_mask(7,0x88a220a,0x20a8080,0x2028228,0x200080,0x2882808,0x202a08,0xa088a22);
remainders_neg[55].modulo_value=220; remainders_neg[55].bit_mask=create_bit_mask(7,0x80022282,0x8a020800,0xa8a02a0,0x88a08800,0x82002000,0x80a82280,0x20002a2);
remainders_pos[56].modulo_value=56; remainders_pos[56].bit_mask=create_bit_mask(2,0x82002a22,0x88a800);
remainders_neg[56].modulo_value=56; remainders_neg[56].bit_mask=create_bit_mask(2,0xa88a22a,0x2080);
remainders_pos[57].modulo_value=228; remainders_pos[57].bit_mask=create_bit_mask(8,0x22000082,0x28a20a00,0x2200282,0x20a0800,0x28008800,0xa08a288,0x20000880,0x8);
remainders_neg[57].modulo_value=228; remainders_neg[57].bit_mask=create_bit_mask(7,0xa2800802,0x20228208,0xa28820a,0x82820080,0x20800008,0x82002200,0xa0880000);
remainders_pos[58].modulo_value=232; remainders_pos[58].bit_mask=create_bit_mask(8,0xaa80a8a,0xa2020822,0x2228a2,0x20080a8,0x2888002a,0x20808a8a,0xa02aa088,0xa2);
remainders_neg[58].modulo_value=232; remainders_neg[58].bit_mask=create_bit_mask(8,0x82208202,0x2a8a80aa,0x882aa02a,0x8a880820,0xa00088a2,0xa8080202,0x28a22000,0x2a);
remainders_pos[59].modulo_value=236; remainders_pos[59].bit_mask=create_bit_mask(8,0xa2a20a22,0x2a2aa80,0x82a0008,0x22088288,0xa820228,0xa8a80200,0x8a8a02a,0x88a);
remainders_neg[59].modulo_value=236; remainders_neg[59].bit_mask=create_bit_mask(8,0x2a2a82aa,0x822a2208,0x80a28880,0xaa800a00,0x820a8aa0,0x20208a88,0x802028a2,0x2);
remainders_pos[60].modulo_value=120; remainders_pos[60].bit_mask=create_bit_mask(4,0x20882,0x28220800,0x2088,0x822080);
remainders_neg[60].modulo_value=120; remainders_neg[60].bit_mask=create_bit_mask(4,0x808a0002,0x20228000,0x808a000,0x22800);
remainders_pos[61].modulo_value=61; remainders_pos[61].bit_mask=create_bit_mask(2,0xa59f23a,0x1713e694);
remainders_neg[61].modulo_value=244; remainders_neg[61].bit_mask=create_bit_mask(8,0x82802aa2,0x8a8a2a08,0x882a28a,0x222aa00a,0x88a0aa00,0x80008202,0x8288a020,0x2002a);
remainders_pos[62].modulo_value=248; remainders_pos[62].bit_mask=create_bit_mask(8,0x22a8a202,0x28aa022a,0x22a008,0x82a20802,0x80208a,0x28200a88,0x88a880aa,0x808a2a);
remainders_neg[62].modulo_value=248; remainders_neg[62].bit_mask=create_bit_mask(8,0x2a202a8a,0xa0228aa2,0x88aa2880,0xa2a808a,0x8008a802,0xa0a88200,0x200822,0x802a2);
remainders_pos[63].modulo_value=84; remainders_pos[63].bit_mask=create_bit_mask(3,0xa2080002,0x8a08020,0x80002);
remainders_neg[63].modulo_value=84; remainders_neg[63].bit_mask=create_bit_mask(3,0x22800802,0x200820,0x808a);
remainders_pos[64].modulo_value=8; remainders_pos[64].bit_mask=create_bit_mask(1,0xaa);
remainders_neg[64].modulo_value=8; remainders_neg[64].bit_mask=create_bit_mask(1,0x22);
remainders_pos[65].modulo_value=65; remainders_pos[65].bit_mask=create_bit_mask(3,0x30054396,0xa70a8033,0x1);
remainders_neg[65].modulo_value=260; remainders_neg[65].bit_mask=create_bit_mask(8,0xa8880a0a,0x2a020822,0x208202a0,0x828088aa,0x880202,0xa20800,0x8228022,0x20088000);
remainders_pos[66].modulo_value=264; remainders_pos[66].bit_mask=create_bit_mask(9,0x820a2022,0x28220a00,0x80200002,0x20002082,0x8000800,0x80282,0xa0882880,0x8a08200,0x88);
remainders_neg[66].modulo_value=264; remainders_neg[66].bit_mask=create_bit_mask(8,0x28220a2,0x20228208,0x828808a,0xa0882802,0x800008,0x8000820a,0x20002000,0x200a08);
remainders_pos[67].modulo_value=268; remainders_pos[67].bit_mask=create_bit_mask(9,0xaa220a8a,0x800a0822,0xa282aa02,0x2888008,0x22280802,0xa828a200,0xa00280a,0x88aa882,0xa2a);
remainders_neg[67].modulo_value=268; remainders_neg[67].bit_mask=create_bit_mask(9,0x22aa8202,0x88280aa,0x2a0a2282,0x8a000880,0xaaa0808a,0x20a02a88,0x8288a082,0x8002200a,0x2a2);
remainders_pos[68].modulo_value=136; remainders_pos[68].bit_mask=create_bit_mask(5,0x228a202,0x8a2880a,0x228a2028,0x8a2880a0,0x80);
remainders_neg[68].modulo_value=136; remainders_neg[68].bit_mask=create_bit_mask(5,0x8aa02a8a,0x80220082,0xaa02a8a0,0x2200828,0x8);
remainders_pos[69].modulo_value=276; remainders_pos[69].bit_mask=create_bit_mask(9,0x82022822,0xa20000,0x2280202,0xa2020800,0x8a08a20,0x88000208,0xa0080802,0x8200008,0x88288);
remainders_neg[69].modulo_value=276; remainders_neg[69].bit_mask=create_bit_mask(9,0x20a20a2,0x8228808,0x8a20828a,0x22820080,0x200228,0x88280,0x20800082,0xa08800,0x200);
remainders_pos[70].modulo_value=280; remainders_pos[70].bit_mask=create_bit_mask(9,0x88820a0a,0x20280022,0x200a0220,0x8200802a,0x82000000,0x8a80200,0x80880a0,0x22880028,0xa0a082);
remainders_neg[70].modulo_value=280; remainders_neg[70].bit_mask=create_bit_mask(9,0xa0202,0x282088a2,0x208282a8,0xa0008a2,0x2808808,0x80200a88,0x80000020,0x2a008020,0x202802);
remainders_pos[71].modulo_value=284; remainders_pos[71].bit_mask=create_bit_mask(9,0xa2800aa2,0x8a8aa0a8,0x2022208,0xaa082028,0xa8020802,0x280820a,0x22088808,0xa2a0aa2a,0x8aa0028);
remainders_neg[71].modulo_value=284; remainders_neg[71].bit_mask=create_bit_mask(9,0x2a08822a,0x2022820,0x8a8aaa00,0x2280a8a0,0x208a808a,0x8a080a82,0xaa800080,0x2a2822a2,0x2288a0);
remainders_pos[72].modulo_value=24; remainders_pos[72].bit_mask=create_bit_mask(1,0x820082);
remainders_neg[72].modulo_value=24; remainders_neg[72].bit_mask=create_bit_mask(1,0xa0802);
remainders_pos[73].modulo_value=73; remainders_pos[73].bit_mask=create_bit_mask(3,0xb8d135e,0x22c74279,0x1eb);
remainders_neg[73].modulo_value=292; remainders_neg[73].bit_mask=create_bit_mask(10,0x82008a82,0xaa0a8aa0,0x82aa2022,0x2082a8a,0x80a20a88,0x8028a2a8,0x22200280,0x80a002,0x808aa28a,0x2);
remainders_pos[74].modulo_value=296; remainders_pos[74].bit_mask=create_bit_mask(10,0x220822a2,0xa80aaa0a,0xa80202a2,0xa2202000,0x80820228,0x8088a28,0x80802a00,0xaaa02a8a,0x882088a0,0x8a);
remainders_neg[74].modulo_value=296; remainders_neg[74].bit_mask=create_bit_mask(10,0xaa80aa2a,0x20822282,0x208a8a2a,0x2aa82888,0x80a8aa0,0x808002a0,0x808a288,0x2228a202,0xa80020,0x2);
remainders_pos[75].modulo_value=60; remainders_pos[75].bit_mask=create_bit_mask(2,0x802802,0x8028020);
remainders_neg[75].modulo_value=60; remainders_neg[75].bit_mask=create_bit_mask(2,0x80082082,0x20820);
remainders_pos[76].modulo_value=152; remainders_pos[76].bit_mask=create_bit_mask(5,0x8a02822a,0x280a2000,0x2822aa88,0xa20008a0,0xa88280);
remainders_neg[76].modulo_value=152; remainders_neg[76].bit_mask=create_bit_mask(5,0x2820aa2,0xa082a888,0x20aa2200,0x2a888028,0x200a08);
remainders_pos[77].modulo_value=308; remainders_pos[77].bit_mask=create_bit_mask(10,0x28aa202,0x20200220,0x208a0288,0x220020,0xa022a8a,0x800a2a88,0x20808008,0x8082280a,0x28008800,0x808aa);
remainders_neg[77].modulo_value=308; remainders_neg[77].bit_mask=create_bit_mask(10,0x8a02220a,0x28288aa0,0xa0028a00,0x882a88a0,0x82822202,0x822200,0xa8080080,0x80aa002,0x20800080,0x8022);
remainders_pos[78].modulo_value=312; remainders_pos[78].bit_mask=create_bit_mask(10,0xa2800882,0x8220a20,0x82282000,0x2002020,0x800a00,0x8000a002,0x82080800,0x20000828,0x8a08a088,0x822002);
remainders_neg[78].modulo_value=312; remainders_neg[78].bit_mask=create_bit_mask(10,0x22080002,0xa28228,0x220a088,0x828828a0,0x8000208,0x808208a,0x2800080,0x28008020,0x2002000,0xa2082);
remainders_pos[79].modulo_value=316; remainders_pos[79].bit_mask=create_bit_mask(10,0xa20a2aa,0x8802a888,0xa020a82,0xa2200aa2,0x82882,0xa8228200,0x8aa0088,0x282a080a,0x222a802,0xaa8a08a);
remainders_neg[79].modulo_value=316; remainders_neg[79].bit_mask=create_bit_mask(10,0x82a82a22,0x8a2000,0x828a020a,0x2aa8822a,0x8880a00a,0x20aa0a88,0x80228800,0xa0a28082,0x8aaa208a,0x2202802);
remainders_pos[80].modulo_value=40; remainders_pos[80].bit_mask=create_bit_mask(2,0xa0280a02,0x80);
remainders_neg[80].modulo_value=40; remainders_neg[80].bit_mask=create_bit_mask(1,0x28a0028a);
remainders_pos[81].modulo_value=12; remainders_pos[81].bit_mask=create_bit_mask(1,0x8a2);
remainders_neg[81].modulo_value=12; remainders_neg[81].bit_mask=create_bit_mask(1,0x22);
remainders_pos[82].modulo_value=328; remainders_pos[82].bit_mask=create_bit_mask(11,0xaa882a0a,0x222008a,0x20a20a28,0x82a222a8,0xa0288000,0x2280a00,0x888a8200,0xa08a082a,0x888028,0xa822aaa2,0xa0);
remainders_neg[82].modulo_value=328; remainders_neg[82].bit_mask=create_bit_mask(11,0x2200a282,0x8aaa8802,0xa82a82a0,0x22aaa20,0x28a00888,0x8aa08288,0x20a88,0x280280a2,0x80008a0,0x20aa222a,0x28);
remainders_pos[83].modulo_value=332; remainders_pos[83].bit_mask=create_bit_mask(11,0x222a8202,0x20828aaa,0x2802a0aa,0x20a2880,0x8a008880,0x200a2a0a,0x8a080022,0xa8028022,0x28208aa0,0x2a888aaa,0x808);
remainders_neg[83].modulo_value=332; remainders_neg[83].bit_mask=create_bit_mask(11,0xaaa20a8a,0xa80a0222,0xa0822822,0x8a82a008,0x2880008,0xa882a282,0x28088aa,0x208a08aa,0xa0a80228,0xa2000222,0x80);
remainders_pos[84].modulo_value=168; remainders_pos[84].bit_mask=create_bit_mask(6,0x2020022,0x8008a20,0x2288008,0xa2002020,0x808008,0x88);
remainders_neg[84].modulo_value=168; remainders_neg[84].bit_mask=create_bit_mask(5,0x828a0822,0x800220,0x82200080,0x220828a0,0x8000800);
remainders_pos[85].modulo_value=85; remainders_pos[85].bit_mask=create_bit_mask(3,0x1ce9129a,0xce030030,0x165225);
remainders_neg[85].modulo_value=340; remainders_neg[85].bit_mask=create_bit_mask(11,0x80200a02,0x20288a0,0xa8a82a8,0x8a0200aa,0x80200828,0x20822202,0x800808a2,0x8002a0,0x82a00080,0x22880800,0x22082);
remainders_pos[86].modulo_value=344; remainders_pos[86].bit_mask=create_bit_mask(11,0x22020aa2,0xaa8222a8,0x202a20a8,0x2280080a,0xa2a22800,0x8a008200,0x8800288a,0x8a02002,0xaa2a08a8,0x882a8882,0x8aa080);
remainders_neg[86].modulo_value=344; remainders_neg[86].bit_mask=create_bit_mask(11,0xaa8a822a,0x220aa220,0xa8a2a820,0xaa088082,0x2a2aa088,0x2880a88,0x8a002,0x8028a88a,0x22a28020,0xa2000a,0x22808);
remainders_pos[87].modulo_value=348; remainders_pos[87].bit_mask=create_bit_mask(11,0x828a2002,0x8820a08,0x208a080,0x82022820,0x200,0x2280288,0x20080000,0x828808,0x20a208,0x220a0822,0x8008a28);
remainders_neg[87].modulo_value=348; remainders_neg[87].bit_mask=create_bit_mask(11,0x2022882,0x28200,0x8a002008,0x28a20a0,0x8808a08,0x8a208200,0xa0800080,0x8020000,0x88282280,0x228200a2,0x800220);
remainders_pos[88].modulo_value=88; remainders_pos[88].bit_mask=create_bit_mask(3,0x2a20228a,0xa8020080,0xa28808);
remainders_neg[88].modulo_value=88; remainders_neg[88].bit_mask=create_bit_mask(3,0xa2a8a202,0x200a8808,0x2a0080);
remainders_pos[89].modulo_value=89; remainders_pos[89].bit_mask=create_bit_mask(3,0x2770f36,0x2a6b595,0x1b3c3b9);
remainders_neg[89].modulo_value=356; remainders_neg[89].bit_mask=create_bit_mask(11,0x8aaa82aa,0x8a2a2808,0xa02a0a20,0xa2882282,0xaa8880a2,0xa002aa0a,0xa888000,0x82288202,0x20a02202,0xa820208a,0x2800088);
remainders_pos[90].modulo_value=120; remainders_pos[90].bit_mask=create_bit_mask(4,0x80002002,0x220a20,0x208a088,0x800800);
remainders_neg[90].modulo_value=120; remainders_neg[90].bit_mask=create_bit_mask(4,0x882882,0x8228220,0xa002000,0x80);
remainders_pos[91].modulo_value=364; remainders_pos[91].bit_mask=create_bit_mask(12,0x2a008a2a,0xa02202,0x2820a88,0x2a02008a,0x80820888,0x28,0x8202280,0x80a8222,0x8280a20,0x80a0022a,0x200a8808,0xa8a);
remainders_neg[91].modulo_value=364; remainders_neg[91].bit_mask=create_bit_mask(12,0xa2880222,0x828aa02,0x820a8200,0xa20a8802,0x8020000,0x888808a0,0x80a8a200,0x80020a2a,0xa082a0,0x2882a2,0x28020080,0x202);
remainders_pos[92].modulo_value=184; remainders_pos[92].bit_mask=create_bit_mask(6,0x2208aa82,0x800a0a00,0x282aa208,0x2208aa8,0x8800a0a0,0x82aa20);
remainders_neg[92].modulo_value=184; remainders_neg[92].bit_mask=create_bit_mask(6,0xaa00220a,0x8828288,0xa0a22a80,0x8aa00220,0x882828,0xa22a8);
remainders_pos[93].modulo_value=372; remainders_pos[93].bit_mask=create_bit_mask(12,0x228a0882,0x220000,0x208200a,0x2802082,0x20000220,0xa08828a,0x8a2822,0x28008800,0x8082080,0x800a0082,0x28800008,0x8220a);
remainders_neg[93].modulo_value=372; remainders_neg[93].bit_mask=create_bit_mask(12,0x22020002,0x8a28808,0x8a00a082,0x82082802,0x20808a28,0x82000202,0x800220a2,0x20800008,0x8000a008,0x820802,0x20008800,0xa282);
remainders_pos[94].modulo_value=376; remainders_pos[94].bit_mask=create_bit_mask(12,0xaa82a22a,0x80a2080,0x22aa2822,0xa2282002,0xa8a00a8,0x220280,0xa0028088,0x8a2a00a2,0x88800828,0x208828aa,0xaa0208a0,0xa88a82);
remainders_neg[94].modulo_value=376; remainders_neg[94].bit_mask=create_bit_mask(12,0x220a2aa2,0x80822808,0xaa22a0aa,0x2aa0a88a,0x82028820,0x88aa8a08,0x288a0800,0x2a2802a,0x880a0,0xa800a022,0x228a8028,0x20020a);
remainders_pos[95].modulo_value=380; remainders_pos[95].bit_mask=create_bit_mask(12,0x80802282,0xa82a8822,0x88a8080,0xa220022,0xa00000,0x808220a,0xa088202,0xa0,0x880088a,0xa0202a22,0x28822a82,0x8288020);
remainders_neg[95].modulo_value=380; remainders_neg[95].bit_mask=create_bit_mask(12,0x8002a0a,0x202200a2,0x20008,0x82a288aa,0x80288808,0x80802282,0x82000a82,0x8088028,0x80008002,0x20a8a22a,0xa00a2202,0x2008a0);
remainders_pos[96].modulo_value=24; remainders_pos[96].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[96].modulo_value=24; remainders_neg[96].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[97].modulo_value=97; remainders_pos[97].bit_mask=create_bit_mask(4,0x8b451b5e,0x6067981b,0xeb628b47,0x1);
remainders_neg[97].modulo_value=388; remainders_neg[97].bit_mask=create_bit_mask(13,0x2888282,0xa8aa0082,0x22aa028a,0x8a02aa20,0x802a8aa,0xa088a02,0xa088a0a2,0x802a8a2,0x2002a080,0x82a00222,0x8aa00800,0x828882a2,0x2);
remainders_pos[98].modulo_value=56; remainders_pos[98].bit_mask=create_bit_mask(2,0x82828202,0x808282);
remainders_neg[98].modulo_value=56; remainders_neg[98].bit_mask=create_bit_mask(2,0xa0a0a0a,0x80a02);
remainders_pos[99].modulo_value=132; remainders_pos[99].bit_mask=create_bit_mask(5,0x20800a2,0x220828,0x82088000,0xa0020802,0x8);
remainders_neg[99].modulo_value=132; remainders_neg[99].bit_mask=create_bit_mask(4,0x82800022,0x8228020,0xa000088,0x208a0082);
remainders_pos[100].modulo_value=40; remainders_pos[100].bit_mask=create_bit_mask(2,0xa8aa2a8a,0xa2);
remainders_neg[100].modulo_value=40; remainders_neg[100].bit_mask=create_bit_mask(2,0x20222202,0x22);
remainders_pos[101].modulo_value=101; remainders_pos[101].bit_mask=create_bit_mask(4,0xc3fb6272,0x552a832,0x91b7f0d3,0x13);
remainders_neg[101].modulo_value=404; remainders_neg[101].bit_mask=create_bit_mask(13,0xa22aaaa,0x888a20aa,0x82a280a,0xa2a0828a,0x22888a80,0xaa2282a8,0x200002aa,0x82280282,0x20a8088,0x28a0028a,0x8880a820,0x20a00a20,0x2);
remainders_pos[102].modulo_value=408; remainders_pos[102].bit_mask=create_bit_mask(13,0x82080882,0x20228a20,0xa00a,0x20a2820,0x220028,0x80208200,0x2020080,0x8208,0x80280088,0x828a0,0x8a00a00,0x8208a288,0x822020);
remainders_neg[102].modulo_value=408; remainders_neg[102].bit_mask=create_bit_mask(13,0x2800002,0x28a20228,0x88082082,0x820220a0,0x8a28820,0x280288,0x828a0800,0x8800a00,0x8208000,0x80802020,0x208200,0xa002200,0xa2820);
remainders_pos[103].modulo_value=412; remainders_pos[103].bit_mask=create_bit_mask(13,0xaa022a0a,0x200a8a8a,0xa0820888,0x8a28020a,0x2a8a222,0x8a800202,0x20280280,0x808002a,0x2888a2a8,0xaa08028a,0x82220820,0xaa2a2a00,0xa0a880a);
remainders_neg[103].modulo_value=412; remainders_neg[103].bit_mask=create_bit_mask(13,0x228aa282,0xa8820202,0x280a8000,0x2a08a02,0x8a202aaa,0x2088a8a,0xa8a08a08,0x808088a2,0xa0002a20,0x22808a02,0xaaa80a8,0x22a2a288,0x2820082);
remainders_pos[104].modulo_value=104; remainders_pos[104].bit_mask=create_bit_mask(4,0x2aa0a22,0x8822020,0xa0aa8008,0x88);
remainders_neg[104].modulo_value=104; remainders_neg[104].bit_mask=create_bit_mask(3,0x8a2282aa,0x800aa828,0x20220880);
remainders_pos[105].modulo_value=420; remainders_pos[105].bit_mask=create_bit_mask(14,0x802002,0x8200200,0x2008200,0x20228a2,0x20800208,0x8000200,0x800a0020,0x8000200,0x8002080,0xa2880802,0x8200808,0x8008200,0x802000,0x8);
remainders_neg[105].modulo_value=420; remainders_neg[105].bit_mask=create_bit_mask(13,0x80082802,0x208a00,0x2080288,0x82022022,0x20008a00,0x80080288,0x200a0,0x808208,0xa000,0x22000002,0x200000,0x80000280,0x82080);
remainders_pos[106].modulo_value=424; remainders_pos[106].bit_mask=create_bit_mask(14,0xa2a82aa,0xa20aa008,0x822a0808,0x22822222,0x2208a8a0,0x20a8aa00,0x80000280,0xaa2a0802,0x2a208800,0x8882880a,0x20a88288,0xaa08a20,0x82a8a020,0xaa);
remainders_neg[106].modulo_value=424; remainders_neg[106].bit_mask=create_bit_mask(14,0x82a20a22,0x2a822880,0xaa28080,0xaa0aaaaa,0x2a802028,0xa8202288,0x8888a08,0x22a2808a,0xa2a80088,0xa0082,0xa8200a00,0x822002a8,0xa2028a8,0x22);
remainders_pos[107].modulo_value=428; remainders_pos[107].bit_mask=create_bit_mask(14,0xa200a282,0xaaaa0a22,0x8a2200a8,0x822802a0,0x202a88,0x88088280,0xa0a082a,0x22022a82,0x80800028,0x288222a,0x88a20a8,0xaaaa2a0,0xa008a88a,0x828);
remainders_neg[107].modulo_value=428; remainders_neg[107].bit_mask=create_bit_mask(14,0x2a882a0a,0x222282aa,0x2aa8820,0xaa08228,0x88a8a200,0x800a08,0x828280a2,0xaa8aa20a,0x80888a0,0x8a00aaa2,0x8002a820,0x82222a28,0x28802002,0xa0);
remainders_pos[108].modulo_value=24; remainders_pos[108].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[108].modulo_value=24; remainders_neg[108].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[109].modulo_value=109; remainders_pos[109].bit_mask=create_bit_mask(4,0xbe7192ba,0xb003685c,0x639f4e85,0x1752);
remainders_neg[109].modulo_value=436; remainders_neg[109].bit_mask=create_bit_mask(14,0x22a80a22,0x288aa080,0xaa028208,0xaa8a8a8a,0xa0820a02,0x20082a88,0x222280aa,0xaa200a82,0x8a088028,0x8002a0a2,0xa0008000,0x8828a282,0x22a8a00,0x220a8);
remainders_pos[110].modulo_value=440; remainders_pos[110].bit_mask=create_bit_mask(14,0x20a20202,0x2a2288a0,0x2a0a8200,0x80028a0,0x22a00020,0x20080200,0x82208202,0x8808208,0x88008020,0x2008000a,0xa80a2800,0xa80082a0,0x80a2288,0x80808a);
remainders_neg[110].modulo_value=440; remainders_neg[110].bit_mask=create_bit_mask(14,0xa82a028a,0xa22a0020,0x22820288,0x8080a020,0xa2280828,0xa0800a88,0xa280a82,0x80200,0x8008a8,0x20808802,0x20822088,0x20080220,0x2a200,0x802);
remainders_pos[111].modulo_value=444; remainders_pos[111].bit_mask=create_bit_mask(14,0xa20a0822,0x828800,0x8a288280,0x22020880,0x28020000,0x820020a,0x820080,0x200820,0x8a080082,0x80000802,0x20220808,0x28228a,0xa0022820,0x8820a08);
remainders_neg[111].modulo_value=444; remainders_neg[111].bit_mask=create_bit_mask(14,0x228200a2,0x8020008,0x2200208,0xa28a0000,0x20828808,0x80288282,0x800a0800,0x8a08028,0x2008002,0x880082,0x28a28000,0x8820a202,0x200a20a0,0x28200);
remainders_pos[112].modulo_value=56; remainders_pos[112].bit_mask=create_bit_mask(2,0xaa08020a,0xa08020);
remainders_neg[112].modulo_value=56; remainders_neg[112].bit_mask=create_bit_mask(2,0x22808a02,0x2808a0);
remainders_pos[113].modulo_value=113; remainders_pos[113].bit_mask=create_bit_mask(4,0xd645eb96,0xf33e1211,0x89ae2121,0x1a75e);
remainders_neg[113].modulo_value=452; remainders_neg[113].bit_mask=create_bit_mask(14,0xa88220a,0x2aa28a88,0x22a8a8,0xaa02a82,0xa8aa2002,0x8a8a2aa0,0x82208a80,0x808a2282,0x20808a,0xaa20080a,0x802200a2,0x8022aa2,0x80820028,0xa22880a8);
remainders_pos[114].modulo_value=456; remainders_pos[114].bit_mask=create_bit_mask(15,0x20028a2,0x820228,0xa08228a,0x22022020,0x208008,0x2208280,0x820a0,0x8200000,0x8208800a,0x2080002,0x8808820,0x8820a008,0x808200a2,0x28008028,0x8a);
remainders_neg[114].modulo_value=456; remainders_neg[114].bit_mask=create_bit_mask(15,0x82802022,0x8028a20,0x200a202,0xa28a28a0,0x8a00800,0x8a280208,0x80802820,0x208808,0xa000082,0x82800882,0x20,0x282080,0xa0822,0x20800820,0x2);
remainders_pos[115].modulo_value=460; remainders_pos[115].bit_mask=create_bit_mask(15,0x280a0a0a,0x22a8222,0x8828000,0x8a02802a,0xa2082200,0x88000a88,0x82020080,0x8082000,0x22020,0x8208a22a,0x280a2008,0x28220a80,0x2a880000,0xa028888,0xa0a);
remainders_neg[115].modulo_value=460; remainders_neg[115].bit_mask=create_bit_mask(15,0xa0020282,0x8a220aa2,0xa0088,0x28208a2,0x2280aa08,0x880200,0xa0a0800,0x80a088,0x8882a8a8,0x2802222,0xa0822880,0xa02a8200,0x22008008,0x82820000,0x202);
remainders_pos[116].modulo_value=232; remainders_pos[116].bit_mask=create_bit_mask(8,0x28022a2,0x8a2a200a,0x280a008a,0x2a28a880,0xa02802,0x8a8a2a2,0x880280a0,0x8a);
remainders_neg[116].modulo_value=232; remainders_neg[116].bit_mask=create_bit_mask(8,0x8a08aa2a,0x2a2a882,0xa0028802,0xa2a02008,0x8828a08a,0x80202a2a,0x8a0028,0x2);
remainders_pos[117].modulo_value=156; remainders_pos[117].bit_mask=create_bit_mask(5,0x22820002,0x20a20808,0x8000a000,0x820208a0,0x8000828);
remainders_neg[117].modulo_value=156; remainders_neg[117].bit_mask=create_bit_mask(5,0xa20a0882,0x28228000,0x82088,0x28a0020,0x800020);
remainders_pos[118].modulo_value=472; remainders_pos[118].bit_mask=create_bit_mask(15,0x8a8a220a,0x228a82a8,0x20022820,0xa20aaa0,0x22aa2a00,0x80802a28,0x20808802,0x88a00a2,0x2802202,0x8828a802,0xa000a8aa,0x80aaa08,0x88082880,0xa22a82a2,0xa088a2);
remainders_neg[118].modulo_value=472; remainders_neg[118].bit_mask=create_bit_mask(15,0x202aa82,0xa2020a20,0xa88aa0a8,0x82a82228,0xaa22a288,0x808a2a0,0xa808008a,0x8002882a,0x8a08aa8a,0xa0200a,0x28882022,0x80822280,0x80a008,0x2aa20a2a,0x28002a);
remainders_pos[119].modulo_value=476; remainders_pos[119].bit_mask=create_bit_mask(15,0x2880a22,0x28a0a280,0xa08a8280,0x2208888a,0x200202,0x202a08,0x800aa208,0x28020802,0x208aa00,0x20a8080,0x88080080,0xaa222208,0x80282a20,0x28a0a2,0x88a0228);
remainders_neg[119].modulo_value=476; remainders_neg[119].bit_mask=create_bit_mask(15,0x8a00822a,0x20202a00,0x20020a08,0xaa000002,0x88a08a8a,0x80a82280,0x8822280,0x208a800a,0xa802280,0x82820808,0x8800,0x22a2aa80,0x8a02228,0x88a0202a,0x28220);
remainders_pos[120].modulo_value=120; remainders_pos[120].bit_mask=create_bit_mask(4,0x200a2082,0x20020,0x8080080,0x8208a0);
remainders_neg[120].modulo_value=120; remainders_neg[120].bit_mask=create_bit_mask(4,0xa0822802,0x8028820,0x8008,0x20020);
remainders_pos[121].modulo_value=44; remainders_pos[121].bit_mask=create_bit_mask(2,0xaaaaa2aa,0xaa8);
remainders_neg[121].modulo_value=44; remainders_neg[121].bit_mask=create_bit_mask(2,0x22222222,0x220);
remainders_pos[122].modulo_value=488; remainders_pos[122].bit_mask=create_bit_mask(16,0x22208a02,0xa2a8aa8,0xa822022a,0x828a00aa,0x28000aa0,0x202022a2,0x22280080,0xa82a08a,0x2888a2,0x88080802,0xa000288a,0xa2820a,0x80882aaa,0xa2a8a0a8,0xa208882a,0x80);
remainders_neg[122].modulo_value=488; remainders_neg[122].bit_mask=create_bit_mask(16,0xaaa8028a,0x82a20220,0x20aa8aa2,0xa028822,0xa0888228,0xa828aa2a,0xaaa08808,0x820a2802,0x88a0002a,0x80808a,0x2888a002,0x882a0a82,0x800a222,0x2a202020,0x2a8000a2,0x8);
remainders_pos[123].modulo_value=492; remainders_pos[123].bit_mask=create_bit_mask(16,0x228a0082,0x28a20020,0x208820a,0x2800820,0x8a08228,0x8202,0x880880,0x8020a00,0x220000a,0x20000022,0x20a20808,0x280288,0x22080082,0x8a28a08,0xa288080,0x820);
remainders_neg[123].modulo_value=492; remainders_neg[123].bit_mask=create_bit_mask(16,0xa2020802,0x20228828,0x8a000282,0x820800a0,0x200a20,0x8808028a,0x80000000,0x828208,0xa288082,0xa08808a2,0x28228000,0x88208200,0xa2800802,0x220200,0x82200000,0xa0);
remainders_pos[124].modulo_value=248; remainders_pos[124].bit_mask=create_bit_mask(8,0xa808a2a,0x822a02,0x80a8820,0xaa8a202a,0x20a808a2,0x822a0,0xa080a882,0xa8a202);
remainders_neg[124].modulo_value=248; remainders_neg[124].bit_mask=create_bit_mask(8,0x20802a2,0x880aa28a,0x808200a8,0x2202a8a2,0xa020802a,0x8880aa28,0x2808200a,0x202a8a);
remainders_pos[125].modulo_value=5; remainders_pos[125].bit_mask=create_bit_mask(1,0x12);
remainders_neg[125].modulo_value=20; remainders_neg[125].bit_mask=create_bit_mask(1,0x28a);
remainders_pos[126].modulo_value=168; remainders_pos[126].bit_mask=create_bit_mask(6,0x82002822,0x20808800,0xa,0x220208a0,0x28008200,0x88);
remainders_neg[126].modulo_value=168; remainders_neg[126].bit_mask=create_bit_mask(5,0x2882022,0x28000000,0x80088082,0xa20a0020,0x20800a08);
remainders_pos[127].modulo_value=508; remainders_pos[127].bit_mask=create_bit_mask(16,0xaa2228a,0xa88a0aa0,0x880a0a28,0xaa28000,0x2aaa0802,0x208282a2,0xa8028220,0x2088802,0xa8022208,0x80882802,0x88a82820,0x8020aaa,0x200028aa,0xa28a0a02,0xaa0a22,0xa2888aa);
remainders_neg[127].modulo_value=508; remainders_neg[127].bit_mask=create_bit_mask(16,0x822aaa02,0x20028228,0x8282a0,0x22a0888,0xa222808a,0xa80a0a2a,0x208a0aa8,0x8a80008a,0x208aaa80,0x800a08a,0x20a0a8,0x808a8222,0xa888a022,0x2a02828a,0x882282aa,0x2a00022);
remainders_pos[128].modulo_value=8; remainders_pos[128].bit_mask=create_bit_mask(1,0x82);
remainders_neg[128].modulo_value=8; remainders_neg[128].bit_mask=create_bit_mask(1,0xa);
remainders_pos[129].modulo_value=516; remainders_pos[129].bit_mask=create_bit_mask(17,0xa2002022,0x20000,0x200a08a,0xa2822082,0x8220a28,0x88280202,0x20002,0x20200228,0x8800808a,0x80002,0x8028228,0x8a088208,0x208828a2,0x20a00808,0x8000a,0x808008a0,0x8);
remainders_neg[129].modulo_value=516; remainders_neg[129].bit_mask=create_bit_mask(16,0x228828a2,0x8828008,0x8a082002,0x220a2802,0xa28220,0x20828a,0x800a0882,0x28a08a20,0x80002,0x80800882,0x820a20,0x2000280,0xa0002022,0x28208000,0x88008082,0x80020);
remainders_pos[130].modulo_value=520; remainders_pos[130].bit_mask=create_bit_mask(17,0x8280a8a,0x8a228802,0x2a200,0x200280a,0xa0282222,0xa0022880,0x88020082,0x820a0,0x8200000,0x80220a,0x28800a82,0x88280a02,0x28008088,0x8a8000a0,0x2288a200,0xa0282080,0xa2);
remainders_neg[130].modulo_value=520; remainders_neg[130].bit_mask=create_bit_mask(17,0x80a00202,0x22a0002,0x8a2288,0x8a80a082,0x20a02a2a,0x288a2008,0xa8802,0x800a028,0x80a08888,0x8008a202,0x2008020a,0x208282,0x20800000,0x2008828,0xa2002a08,0x20a02808,0x22);
remainders_pos[131].modulo_value=524; remainders_pos[131].bit_mask=create_bit_mask(17,0x82a82222,0x202aa202,0x828aa08a,0xa2aaa2a0,0x8000802,0x28020a28,0x8800a882,0x88a00a0a,0xa22208,0xa0022a0a,0x8028822,0x2028a,0xaaa8a802,0xaa2820a8,0xaa808a20,0x82a82808,0x888);
remainders_neg[131].modulo_value=524; remainders_neg[131].bit_mask=create_bit_mask(16,0xa20aaaa,0xa8a22a8a,0xa022802,0x2a222a28,0x80888082,0xa08a82a0,0x88200a,0x288282,0x882aaa80,0x288aa282,0x808a00aa,0x888a8a02,0x2220208a,0x22a0a820,0x220802a8,0xa20a080);
remainders_pos[132].modulo_value=264; remainders_pos[132].bit_mask=create_bit_mask(9,0xa2020002,0x20228,0x8808000a,0x808a2,0x20200028,0x20228a,0x808000a0,0x808a28,0x80);
remainders_neg[132].modulo_value=264; remainders_neg[132].bit_mask=create_bit_mask(9,0x228a0082,0x8028a20,0x8082,0x80800022,0x28a00820,0x8028a202,0x80820,0x8000220,0x8);
remainders_pos[133].modulo_value=532; remainders_pos[133].bit_mask=create_bit_mask(17,0x8a802a0a,0x8000a82,0x22220820,0xa20008a,0x228202,0x82222288,0x80008020,0xa00a8a0a,0xa82a82,0x2a0a2a,0x88208020,0x80022888,0x8a080828,0x888a2000,0x2008208,0x2a282a00,0xa0a80);
remainders_neg[133].modulo_value=532; remainders_neg[133].bit_mask=create_bit_mask(17,0x200a202,0x888202,0x22aa80a8,0x82288802,0x88a20a8a,0xaaa2200,0x88800a8,0x28020202,0x820a202,0x80a282a2,0x2808a0,0x882a000,0x28000a0,0x2a088,0x82880a00,0x22a0a280,0x28208);
remainders_pos[134].modulo_value=536; remainders_pos[134].bit_mask=create_bit_mask(17,0x820a22a2,0xa822200a,0x8aaa8222,0x2aa0a820,0xa00202a,0x80008a28,0x22280022,0x20a280aa,0x8808202,0x88aa028a,0x2880028,0xa028a200,0xa8a80800,0xa2082a0a,0x2a8882aa,0x82a00888,0x8a88a0);
remainders_neg[134].modulo_value=536; remainders_neg[134].bit_mask=create_bit_mask(17,0xa82aa2a,0x20aaa882,0x2220aa2,0xa22820a8,0x8288a8a2,0x88802a0,0xaaa088aa,0xa82a0822,0x80080a8a,0x228a02,0x8a0008a0,0x28a02a88,0x20208088,0x2a80a282,0xa2000a22,0xa288000,0x20028);
remainders_pos[135].modulo_value=60; remainders_pos[135].bit_mask=create_bit_mask(2,0x20882,0x8220800);
remainders_neg[135].modulo_value=60; remainders_neg[135].bit_mask=create_bit_mask(2,0x808a0002,0x228000);
remainders_pos[136].modulo_value=136; remainders_pos[136].bit_mask=create_bit_mask(5,0x2a008a2a,0x2082a022,0xa820800,0xa200a888,0xa8);
remainders_neg[136].modulo_value=136; remainders_neg[136].bit_mask=create_bit_mask(5,0xa28802a2,0xa80228aa,0x820a8088,0x2a082000,0x20);
remainders_pos[137].modulo_value=137; remainders_pos[137].bit_mask=create_bit_mask(5,0x524fcb96,0xb90610f5,0x21827733,0x4fc92abc,0x1a7);
remainders_neg[137].modulo_value=548; remainders_neg[137].bit_mask=create_bit_mask(17,0x8a82020a,0x208a8828,0xa80aaaaa,0x8200a220,0x20aa02a,0x80a82228,0x8822aaaa,0xa88a08a,0xa2828202,0x8a0880a2,0x22880,0x22280a80,0x2200a2a8,0x220aa280,0xa80a,0x28808a20,0xa2a28088);
remainders_pos[138].modulo_value=552; remainders_pos[138].bit_mask=create_bit_mask(18,0xa20a0002,0x28820828,0xa00220a,0x82022020,0x20808208,0x80282200,0x80002022,0x820,0x8a00a280,0x20000002,0x8000208,0x88280288,0x82020800,0x8808220,0x8800a008,0x208228a0,0xa08a28,0x80);
remainders_neg[138].modulo_value=552; remainders_neg[138].bit_mask=create_bit_mask(18,0x22020882,0x20028020,0x8208a282,0x28228a0,0x28000a00,0x820a288,0x8828a2,0x8808028,0x2082208,0xa0880082,0x808a00,0x208200,0x28a0000,0xa28,0x82080,0xa00a2020,0x8200220,0x8);
remainders_pos[139].modulo_value=556; remainders_pos[139].bit_mask=create_bit_mask(18,0x2a88a22a,0xa022aa0,0x82822822,0x2aaa8080,0x8a0282,0xa2a2a00,0x8288a08a,0xa8008a,0x20008822,0x2a00882,0xa2282a20,0x8a8a0a20,0xa20000a,0x2aaa8828,0x88282020,0x880a0882,0xa22a80aa,0xa88);
remainders_neg[139].modulo_value=556; remainders_neg[139].bit_mask=create_bit_mask(18,0xa2002aa2,0x828aa228,0xa0aa0aa,0xa2220808,0x8802820a,0x82a2a288,0xa002802,0x88208802,0xa88800aa,0x8a28800a,0x2aa0a2a8,0x20282a8,0x82a88882,0xa22200a0,0xa0a8a8,0x82800a,0x2aa20822,0x200);
remainders_pos[140].modulo_value=280; remainders_pos[140].bit_mask=create_bit_mask(9,0xa08a2202,0x8000802,0x20208,0x8a20a802,0xa2202808,0x802a08,0x20208080,0xa802000,0x8088a2);
remainders_neg[140].modulo_value=280; remainders_neg[140].bit_mask=create_bit_mask(9,0x28022a0a,0x88082,0x8a8280,0x220208a,0x22a0a000,0x88082280,0xa8280000,0x208a008,0x22);
remainders_pos[141].modulo_value=564; remainders_pos[141].bit_mask=create_bit_mask(18,0x228008a2,0x20820228,0xa000,0x220a0882,0x20020200,0x2088282,0x20002800,0x20a20208,0x8a28800a,0xa08a0022,0x820808,0x8000280,0x882822,0x8800808,0x8220a,0x208000a0,0x28828808,0x8a200);
remainders_neg[141].modulo_value=564; remainders_neg[141].bit_mask=create_bit_mask(18,0xa2080022,0x28020a20,0x88082088,0xa2820002,0x28828a08,0x8a00020a,0xa0882080,0x28228200,0x2200082,0x200208a2,0x8028000,0x80088208,0x800020a2,0x8000,0x8800a282,0xa0080820,0x20020000,0x2288);
remainders_pos[142].modulo_value=568; remainders_pos[142].bit_mask=create_bit_mask(18,0x8aa8228a,0xa2a28880,0x2a2a0a20,0x82200800,0x802a202a,0x2aa8aa22,0xa00a020,0x8a888202,0x822800,0xa2002882,0xa0808222,0xa8080a00,0x288aa2a,0x82a808a8,0xa8002008,0x8a08a0a8,0xa202228a,0xa2882a);
remainders_neg[142].modulo_value=568; remainders_neg[142].bit_mask=create_bit_mask(18,0x220aa02,0x2a2a0008,0xa2a28228,0xaa88088,0x8a2a8a2,0xa22022aa,0x828828a8,0x2000a8a,0x880aa088,0x2a88a00a,0x28080aaa,0x20808280,0x8a0022a2,0xa208020,0x2088a880,0x2802820,0x2a8aaa02,0x2a00a2);
remainders_pos[143].modulo_value=572; remainders_pos[143].bit_mask=create_bit_mask(18,0x82208202,0xa2a8a08,0x80a282a8,0x808aa800,0x202208a8,0x22a0000a,0x80080822,0x2020002,0xa80288a,0xa22802a,0x28000808,0x88820200,0x8a0000a8,0x22a20880,0x2002aa20,0x2a828a0,0x220a2a8a,0x8082088);
remainders_neg[143].modulo_value=572; remainders_neg[143].bit_mask=create_bit_mask(18,0xaa8028a,0x82220200,0x2a0a20,0x8022080,0xa8aa0020,0xa2288882,0x88080a2,0x8a0a880a,0x8200a002,0x2aa00a2,0x20888080,0xa8288,0x2008820,0xaa2a8008,0xa88a2220,0x82202028,0xa282a202,0x802800);
remainders_pos[144].modulo_value=24; remainders_pos[144].bit_mask=create_bit_mask(1,0x8a28a2);
remainders_neg[144].modulo_value=24; remainders_neg[144].bit_mask=create_bit_mask(1,0x22022);
remainders_pos[145].modulo_value=145; remainders_pos[145].bit_mask=create_bit_mask(5,0x907135e,0x84b8835,0x48422791,0x8242b047,0x1eb23);
remainders_neg[145].modulo_value=580; remainders_neg[145].bit_mask=create_bit_mask(19,0x808a0a82,0x800200a0,0xaa208,0xa82288a,0xa220222a,0x88a20a82,0x822a8002,0x28020008,0xa828808,0x20808022,0xa8a22808,0xa8000280,0x820080,0x2222002,0x80208020,0xa0000a80,0xa228a80,0x808208,0x2);
remainders_pos[146].modulo_value=584; remainders_pos[146].bit_mask=create_bit_mask(19,0x22a02a22,0xaaa2a00,0x220a8082,0xa2a88a2a,0x2002aa28,0x20880208,0x8280a220,0xa02000a2,0x202a022a,0x80a80882,0x80aa8,0x8a02828a,0x80220808,0xaa800820,0xa22a8a28,0x2a088a8,0xa8aaa082,0xa80a8800,0x88);
remainders_neg[146].modulo_value=584; remainders_neg[146].bit_mask=create_bit_mask(18,0xaa28a2aa,0x8222a288,0xaa82080a,0x2a2002a2,0xa88a22a0,0xa8008a80,0x2082aa8,0x28a8882a,0xa8a28aa2,0x820800a,0x88808220,0x28a0a02,0x8aa8080,0x220880a8,0x2aa202a0,0xa280020,0x2022280a,0x20820088);
remainders_pos[147].modulo_value=84; remainders_pos[147].bit_mask=create_bit_mask(3,0x2802802,0x28008020,0x80280);
remainders_neg[147].modulo_value=84; remainders_neg[147].bit_mask=create_bit_mask(3,0x82082002,0x20800820,0x8208);
remainders_pos[148].modulo_value=296; remainders_pos[148].bit_mask=create_bit_mask(10,0xa200a8a,0x80228202,0x802a2a8a,0x8a080828,0xa8aa2a00,0x2020a200,0xa8a80228,0x828802a2,0xa008a080,0xa2);
remainders_neg[148].modulo_value=296; remainders_neg[148].bit_mask=create_bit_mask(10,0x82a88202,0x8aa0a8a,0x8a2a202,0x28000a0,0x2022a288,0xa8a82a88,0x20208aa0,0xa008a2a,0x28802800,0x2a);
remainders_pos[149].modulo_value=149; remainders_pos[149].bit_mask=create_bit_mask(5,0xf75b02f2,0xa062e4ba,0x81671239,0xbbd749d1,0x13d036);
remainders_neg[149].modulo_value=596; remainders_neg[149].bit_mask=create_bit_mask(19,0x2a828a2a,0x28aa2822,0x8aa8aa0,0x220a8008,0x828a08aa,0xa8222a8,0x8aa88080,0xa2a8a02a,0x8a0aa220,0xa08202a2,0x20a2200,0x802a08,0x80aa8a88,0x2880222,0xa22008a0,0x8a8aa80,0x82a0080,0x80222820,0x2082);
remainders_pos[150].modulo_value=120; remainders_pos[150].bit_mask=create_bit_mask(4,0x20880002,0x228800,0x8002288,0x800022);
remainders_neg[150].modulo_value=120; remainders_neg[150].bit_mask=create_bit_mask(4,0xa0000882,0x8220000,0x8a200,0x8a2);
remainders_pos[151].modulo_value=604; remainders_pos[151].bit_mask=create_bit_mask(19,0x2aa282aa,0x800a2028,0xaa88a8,0x22888a02,0x20a8288,0x2208a20a,0x2a8a82,0x82aa2022,0x800a0a8,0xa00000a,0x22a0a002,0x8808aa8,0x882a2a80,0xa08a208,0x82282a08,0x80a2228,0x22a22aa0,0x82808a00,0xaa828aa);
remainders_neg[151].modulo_value=604; remainders_neg[151].bit_mask=create_bit_mask(19,0xa22a0a22,0x882a8a0,0x88220020,0xaa00028a,0x8a020a00,0xaa802a82,0x88a2020a,0xa22a8aa,0x80882820,0x82888882,0xaa28288a,0x80080220,0xa2a208,0x82802a80,0xaa0a280,0x8082aaa0,0xaa2aa228,0xa080288,0x220a022);
remainders_pos[152].modulo_value=152; remainders_pos[152].bit_mask=create_bit_mask(5,0xa222aa02,0x220828,0xa82a0,0x8a282088,0x80aa88);
remainders_neg[152].modulo_value=152; remainders_neg[152].bit_mask=create_bit_mask(5,0x2aa2228a,0x88aa80a0,0x8820a28,0x2a0a800,0x82200);
remainders_pos[153].modulo_value=204; remainders_pos[153].bit_mask=create_bit_mask(7,0x2082002,0x8a28808,0x2082008,0x820800a0,0x28a20200,0x82080202,0x800);
remainders_neg[153].modulo_value=204; remainders_neg[153].bit_mask=create_bit_mask(7,0x82802882,0x220000,0x8a00a080,0x2000820,0x20228a08,0x200828a,0x80);
remainders_pos[154].modulo_value=616; remainders_pos[154].bit_mask=create_bit_mask(20,0x2a82822a,0x8082a00,0xa20aa0,0x280a2800,0x22220282,0x80220280,0x8a8a020,0x888a0022,0x20a020,0x82,0xa080082,0xa22208,0xa2a2088,0x80880208,0x80888802,0x28a02882,0xa08a0000,0xa820200a,0x8282a800,0xa8);
remainders_neg[154].modulo_value=616; remainders_neg[154].bit_mask=create_bit_mask(20,0xa20a0222,0xa280,0x802a8228,0xa002a080,0xaaa20a0a,0xaa0a08,0x802020a8,0x2882a,0x8a028a0,0x88880a,0x8200880a,0x8822a280,0x82222800,0x8008280,0x8000002,0x20282002,0x20028888,0x2028a882,0xa022008,0x20);
remainders_pos[155].modulo_value=620; remainders_pos[155].bit_mask=create_bit_mask(20,0x22a82,0x82228220,0x882a228,0x2a028a8,0x2a08a02,0x88a0208,0x202000a0,0xa008008,0x8a208808,0x208000a,0x208a2a00,0x200a0202,0x808200,0xa2200a0,0x20a80208,0x80a8080a,0xa82202a2,0x28882288,0x88000088,0x82a);
remainders_neg[155].modulo_value=620; remainders_neg[155].bit_mask=create_bit_mask(20,0x88a220a,0xa2a0aa0,0xa22a0,0x8a20a020,0x8228020a,0x80020a80,0xa8288820,0x2880080,0x2200080,0x82808802,0xa8022288,0xa8028a82,0x8000288,0x82a28828,0xa0208a00,0x8200082,0x202a0a22,0x2000a200,0x808800,0x22);
remainders_pos[156].modulo_value=312; remainders_pos[156].bit_mask=create_bit_mask(10,0x828800a2,0x20020208,0x82000008,0x22080800,0x28a00228,0x8828800a,0x82002020,0x8200000,0x82208080,0x8a0022);
remainders_neg[156].modulo_value=312; remainders_neg[156].bit_mask=create_bit_mask(10,0x2000822,0x28828a00,0x2088080,0xa2800080,0x20200a20,0x200082,0x28828a0,0x208808,0xa280008,0x200a2);
remainders_pos[157].modulo_value=157; remainders_pos[157].bit_mask=create_bit_mask(5,0xca0b7e1a,0x71bd5ba,0x26461899,0xd76af638,0x161fb414);
remainders_neg[157].modulo_value=628; remainders_neg[157].bit_mask=create_bit_mask(20,0x282a282,0x8a820822,0xaa8a8000,0xaa22aa0,0x8a822888,0x8088a20a,0xa82aa22a,0x88000a8a,0x220820a,0xa20a2a0a,0x22a0a020,0xa880a28a,0x280080aa,0x8a802200,0x8000a208,0xa88822,0x800a0022,0x808aaa80,0x82a228a2,0x28202);
remainders_pos[158].modulo_value=632; remainders_pos[158].bit_mask=create_bit_mask(20,0x22088a82,0xa02a80a0,0x222a22aa,0x8a08228a,0x282000aa,0x800aaa28,0x208228a0,0x20022,0x2a0a802a,0x828088a2,0xa88a2202,0xa802a0,0x8880080,0x20a2882,0x2828aaa0,0xa2aa0008,0x88a28820,0xaaa88a8,0x880a02a8,0x82a220);
remainders_neg[158].modulo_value=632; remainders_neg[158].bit_mask=create_bit_mask(20,0xaa80020a,0x28a20828,0xaaa22a22,0x280aa02,0xa0a88822,0x88222a0,0xa80aa028,0x888a88aa,0xa28208a2,0xa08002a,0x2002aa8a,0x88208a28,0x80008008,0x8a82a00a,0xa0a02228,0x2a228880,0x2a00a8,0x82220020,0x828a20,0xa2aa8);
remainders_pos[159].modulo_value=636; remainders_pos[159].bit_mask=create_bit_mask(20,0x82082822,0x8828220,0x8000800a,0xa28008a2,0x8808a08,0x2002202,0x22800,0xa20a28,0x8080002,0x802082,0x8208020,0x8000202,0x28a08a0,0x28800,0x8088008,0xa20a2022,0x28a20028,0xa002000,0x20882822,0x8828208);
remainders_neg[159].modulo_value=636; remainders_neg[159].bit_mask=create_bit_mask(20,0x28020a2,0x20a28,0x8080082,0x22080022,0x200,0x8a08a28a,0x808a2080,0x8228220,0x8000808a,0x80082802,0xa00828,0x8000828a,0x82020020,0x8820008,0x80000080,0x228228a2,0x20228820,0x8208a088,0xa0002022,0x20a00);
remainders_pos[160].modulo_value=40; remainders_pos[160].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[160].modulo_value=40; remainders_neg[160].bit_mask=create_bit_mask(2,0x882a82,0x20);
remainders_pos[161].modulo_value=644; remainders_pos[161].bit_mask=create_bit_mask(21,0x220a0222,0x20002082,0xa22a0080,0xaa008082,0xa282a002,0x2a0a28,0x80802aa,0x802a22a,0x8200020,0x22028222,0x8828088a,0x80008208,0x88a80200,0xa802020a,0x8a0a800a,0xa828a2,0x20200aa8,0x200a88a8,0x20800080,0x880a0888,0x8);
remainders_neg[161].modulo_value=644; remainders_neg[161].bit_mask=create_bit_mask(20,0xaa028a2a,0x2888a802,0x22a28808,0x2200080a,0x2a02288a,0x88a202a0,0x80800222,0x808a2a22,0xa888a0,0xa28a02aa,0x208002,0x8800a80,0x200a08,0x208a8282,0x2820802,0x820a022,0xa0a88220,0x28020020,0xa8008808,0x820000);
remainders_pos[162].modulo_value=24; remainders_pos[162].bit_mask=create_bit_mask(1,0x820082);
remainders_neg[162].modulo_value=24; remainders_neg[162].bit_mask=create_bit_mask(1,0xa0802);
remainders_pos[163].modulo_value=652; remainders_pos[163].bit_mask=create_bit_mask(21,0x8aa80a8a,0xaa220202,0x2022a82a,0x8a02088a,0x800a0820,0xa282aa02,0x228a8228,0xa008a882,0x8a888000,0x2a0020a,0xa80802,0x222a2a08,0xa200a000,0x2a288822,0xa828a288,0xa00280a,0x80a2082,0xa8808a22,0x88aaa82,0x2aa2808,0xa2a);
remainders_neg[163].modulo_value=652; remainders_neg[163].bit_mask=create_bit_mask(21,0x2208202,0x22aa8a8a,0xa8aa20a2,0x28a8002,0x88280a8,0x2a0a2282,0xaa020aa0,0x2880200a,0x2000888,0x8a288a82,0x8820808a,0xaaa2a280,0x2a882888,0xa2a000aa,0x20a02a00,0x8288a082,0x8082a80a,0x200802aa,0x8002220a,0x8a22a080,0x2a2);
remainders_pos[164].modulo_value=328; remainders_pos[164].bit_mask=create_bit_mask(11,0x82a00222,0x2a0a28a2,0x88a2200,0xa28a0a80,0x8800a828,0x2a002228,0xa0a28a28,0x88a22002,0x28a0a800,0x800a828a,0x88);
remainders_neg[164].modulo_value=328; remainders_neg[164].bit_mask=create_bit_mask(10,0xa288aaa,0xa282a02a,0x8002aa88,0x22028208,0x8820a0,0xa288aaa0,0x282a02a0,0x2aa88a,0x20282088,0x8820a02);
remainders_pos[165].modulo_value=660; remainders_pos[165].bit_mask=create_bit_mask(21,0xa0802082,0x228a00,0x8000200,0x80020020,0x200208,0x200202,0x2000882,0x8020,0x8a200002,0x20082802,0x820a08,0x8a280282,0x80000,0x8008020,0x80082200,0x80080800,0x220800,0x2008008,0x80000800,0x20a00a28,0x82080);
remainders_neg[165].modulo_value=660; remainders_neg[165].bit_mask=create_bit_mask(21,0x20082002,0x8220200,0x88288,0x8208a0,0xa00a00,0x8028028a,0x82080002,0x8800028,0x220808a,0x20802002,0x8020200,0x2208202,0x80,0x800828,0xa200,0x800080,0x228000,0xa080080,0x800000,0x20208220,0x2008);
remainders_pos[166].modulo_value=664; remainders_pos[166].bit_mask=create_bit_mask(21,0xa02aa2a,0x8aaa282,0x228882,0x2a2200a8,0xa228a0a8,0x8220222,0xa220280a,0x802aa80a,0x8a288,0x2a0a282,0x80080020,0x828a0a,0x2228a20,0x8aa02aa8,0x20a02808,0x8a888088,0xa82a0a28,0x2a0088,0x20822288,0xa0828aaa,0xa8aa80);
remainders_neg[166].modulo_value=664; remainders_neg[166].bit_mask=create_bit_mask(21,0x828a22a2,0x80222a0a,0x88a2000a,0xa2aa8820,0x2aa02820,0x80aa8aaa,0x2aa8a082,0x8a22082,0x88802a00,0x8a282a0a,0x88088a8,0x880a0282,0xaaa02a8,0x228a220,0xa828a080,0x2000800,0x20a282a0,0x88a28800,0xa80aaa00,0x280a0222,0x202208);
remainders_pos[167].modulo_value=668; remainders_pos[167].bit_mask=create_bit_mask(21,0x22a08202,0x2a8a088a,0xaa2aa08a,0xa808082,0xa880aaa8,0x2a202208,0x2a02a000,0x800a2a,0x82008228,0x28088222,0x8a2a0a8a,0x28882202,0x2882008,0x8a8a0020,0x8000a80a,0xa208808a,0x2aaa022,0xa820202a,0x8a20aa8a,0x8a220a2a,0x80820a8);
remainders_neg[167].modulo_value=668; remainders_neg[167].bit_mask=create_bit_mask(21,0xaa280a8a,0xa2028002,0x22a22802,0x8208080a,0x20082220,0xa2a8aa00,0xa28a2888,0x880882a2,0xa880aa0,0xa0800aaa,0x2a28202,0xa000aa8a,0x8a00a880,0x20288a8,0x8882082,0x2a800802,0x8a2228aa,0x20a8a8a2,0x2a82202,0x2aa82a2,0x80a820);
remainders_pos[168].modulo_value=168; remainders_pos[168].bit_mask=create_bit_mask(6,0x220a2802,0x20208200,0x2008000,0x82080800,0x28a08800,0x80);
remainders_neg[168].modulo_value=168; remainders_neg[168].bit_mask=create_bit_mask(6,0xa2822002,0x28a00a00,0x82080088,0x2000080,0x20200008,0x8);
remainders_pos[169].modulo_value=13; remainders_pos[169].bit_mask=create_bit_mask(1,0x1ffe);
remainders_neg[169].modulo_value=52; remainders_neg[169].bit_mask=create_bit_mask(2,0x22220222,0x22222);
remainders_pos[170].modulo_value=680; remainders_pos[170].bit_mask=create_bit_mask(22,0x20802a82,0xa2220800,0x2a0a2208,0xa22a00a,0x2080a808,0x80220282,0x288802,0x2000a200,0x200a020,0x8228a820,0x820002,0x2a288280,0xa008008,0x8a000808,0x22280000,0x80880280,0x2a020882,0xa88a020,0x88a0a8a0,0x20888a20,0xa8020800,0x82);
remainders_neg[170].modulo_value=680; remainders_neg[170].bit_mask=create_bit_mask(22,0xa808220a,0x2a228080,0x2282a280,0x82222882,0xa0082000,0xaa0a0a,0x88200082,0x28882288,0x8a8028a8,0x2a02028,0x880a0882,0xa2200a00,0x2880000,0x2808080,0xa2a08008,0x8000a08,0xa20a0002,0x20020a8,0x202028,0xa0000228,0x200a0088,0x2);
remainders_pos[171].modulo_value=228; remainders_pos[171].bit_mask=create_bit_mask(8,0x82020022,0x28020000,0x820a288,0xa20008a0,0x28a08200,0x80282,0x80080820,0x8);
remainders_neg[171].modulo_value=228; remainders_neg[171].bit_mask=create_bit_mask(7,0x28208a2,0x20828808,0x282200,0x22880020,0x20200a08,0x8800820a,0x8000a0);
remainders_pos[172].modulo_value=344; remainders_pos[172].bit_mask=create_bit_mask(11,0xa2a228a,0x82aa0280,0x8020880,0xaa82022,0x8a8a0028,0xa228aa28,0xa02800a2,0x2088082a,0x82022080,0xa00280aa,0xa288a8);
remainders_neg[172].modulo_value=344; remainders_neg[172].bit_mask=create_bit_mask(11,0x82a2aa02,0xa228208,0x808a8008,0x8220a8aa,0x20288a0,0x2aa022a0,0x2820882a,0xa80080a2,0xa8aa808,0x288a0822,0x2a0020);
remainders_pos[173].modulo_value=173; remainders_pos[173].bit_mask=create_bit_mask(6,0xa3e1e652,0x13da8b7e,0x973a6209,0x56f22411,0xe1f15fb4,0x1299);
remainders_neg[173].modulo_value=692; remainders_neg[173].bit_mask=create_bit_mask(22,0x2a282a8a,0x8a0202a2,0xaa2aa80,0x8a2aa888,0x28288228,0xa0028a00,0x88a208a0,0x8088aaa2,0x880aaa2a,0xa22a0202,0x828aa0a2,0x22020a00,0xa882a2a0,0x8a802000,0x8820008,0xaa0a08a2,0x282aa002,0x20882288,0xa88800,0xa08a8002,0x202202a2,0x8028);
remainders_pos[174].modulo_value=696; remainders_pos[174].bit_mask=create_bit_mask(22,0x82820822,0x20a20220,0xa208088,0xa20a0000,0x28200a28,0xa002280,0x2020,0x28a28020,0x8088200,0x2022002,0x208200,0x80008208,0x20800880,0x28008220,0x8028a,0xa0080800,0x28028800,0x8a28a008,0xa00000a0,0x8220208,0x8208808a,0x882082);
remainders_neg[174].modulo_value=696; remainders_neg[174].bit_mask=create_bit_mask(22,0x20a00a2,0x28228a28,0x82280000,0x22820880,0x20a08220,0x8208a208,0x808820a0,0x20220828,0x80000288,0x28a2882,0x8a00a08,0x8080280,0xa0080000,0x20800a28,0x88008202,0x20800080,0x20820008,0x2202000,0x20880820,0xa28a00,0x2000002,0x2802);
remainders_pos[175].modulo_value=140; remainders_pos[175].bit_mask=create_bit_mask(5,0xa808020a,0xa208020,0x208a0000,0x202a080,0xa08);
remainders_neg[175].modulo_value=140; remainders_neg[175].bit_mask=create_bit_mask(5,0x20800a02,0x22808a0,0x20028088,0x8a022808,0x200);
remainders_pos[176].modulo_value=88; remainders_pos[176].bit_mask=create_bit_mask(3,0x20802a2,0x802a28a8,0x8a8020);
remainders_neg[176].modulo_value=88; remainders_neg[176].bit_mask=create_bit_mask(3,0x8a80822a,0x822a020,0x208a8);
remainders_pos[177].modulo_value=708; remainders_pos[177].bit_mask=create_bit_mask(23,0x2880882,0x28000,0x228a002,0x82020020,0x8220828,0x8228a20a,0xa0822082,0x20020208,0x28a200,0xa0002,0x8808a00,0xa20220a,0xa0000,0x8a28008,0x8080080,0x208820a2,0x8a28828,0x8208820a,0x80080822,0xa28800,0x280008,0x22022800,0x8);
remainders_neg[177].modulo_value=708; remainders_neg[177].bit_mask=create_bit_mask(22,0x82000002,0x820808,0x8a20208a,0x28a08a0,0xa28020,0xa202282,0x200a2802,0x28828a00,0x88202288,0x80820802,0x208,0x8228a282,0x80820880,0x220800,0x80008008,0xa0002822,0x220020,0xa000282,0x8000a2,0x8220008,0x88208080,0xa28a2080);
remainders_pos[178].modulo_value=712; remainders_pos[178].bit_mask=create_bit_mask(23,0x2a0a220a,0x2a8a88a8,0x8aaa80,0x2288222,0xa282002,0xa20aaa,0xaa2820a0,0x228822a2,0x800082a2,0x880802a,0xa220a028,0xa088a00,0x2022028,0x820002a8,0x8822888a,0x828aa8a,0xa08a000a,0x828a0aa,0x82288080,0xaaa20088,0x22a2a802,0x88a0a82a,0xa0);
remainders_neg[178].modulo_value=712; remainders_neg[178].bit_mask=create_bit_mask(23,0xa282aa82,0xa2020020,0x88022208,0x8aa00aaa,0x82a0a88a,0x882a8222,0x22a0a828,0xaa00aa2a,0x888022a,0x800808a2,0x2aa828a0,0x82800288,0x8a8aa8a0,0xa888a20,0xaa0002,0x80a02202,0x28028882,0x80a02822,0xaa00808,0x222a0800,0xaa2a208a,0x2820a2,0x28);
remainders_pos[179].modulo_value=716; remainders_pos[179].bit_mask=create_bit_mask(23,0x22822aa2,0xa2822008,0x2a22a082,0xaaa880a8,0x8222a00a,0x802a28a,0xa8088a02,0x2a200002,0x280880a0,0xa88a080a,0x8a08a2,0xa200208,0xa22a8a2,0x22028a02,0x8a80a0,0x2202a800,0xa802080a,0xa8882a28,0x22aaaa00,0xa88a82a0,0x8828a820,0x88288200,0x8aa);
remainders_neg[179].modulo_value=716; remainders_neg[179].bit_mask=create_bit_mask(23,0xaa0aa22a,0x2a0aa880,0xa2aa280a,0x22200820,0xaaa2882,0x80822a02,0x2080028a,0xa2a8888a,0xa0800828,0x20028082,0x8802802a,0x82a88a80,0x82aa202a,0xaa8a028a,0x88020828,0xaa8a2088,0x208a8082,0x2000a2a0,0xaa222288,0x20020a28,0xa020a8,0xa00a88,0x22);
remainders_pos[180].modulo_value=120; remainders_pos[180].bit_mask=create_bit_mask(4,0xa0080802,0x28020200,0xa008080,0x802020);
remainders_neg[180].modulo_value=120; remainders_neg[180].bit_mask=create_bit_mask(4,0x20800082,0x20028a00,0x2080008,0x28a0);
remainders_pos[181].modulo_value=181; remainders_pos[181].bit_mask=create_bit_mask(6,0x2a11fa3a,0x59937cf6,0x40878a4b,0x66b49478,0x151bcfb2,0x1717e2);
remainders_neg[181].modulo_value=724; remainders_neg[181].bit_mask=create_bit_mask(23,0xa28822a2,0x800aa02a,0x880a0282,0xaa2808a0,0x8882022a,0x8a8a8a2a,0x2088aa2,0x80a2aaa2,0x2808828,0x2a800a0a,0x208a2aa0,0x8a20222a,0x802a0020,0x82a0a0aa,0xa88288a,0xa2a20002,0x820088,0x88802080,0x200022a2,0xa88a08a8,0xaa8282a0,0x82002a00,0x20228);
remainders_pos[182].modulo_value=728; remainders_pos[182].bit_mask=create_bit_mask(23,0x2088202,0x28880a22,0x22aa02a0,0x20a28a2,0xa8a22080,0x8282800,0x802a0,0x2002aa0a,0x20002200,0xa0880202,0x822a000,0x20828282,0xa000a88,0x8808022,0x8008800,0xa0aa80,0x200a8020,0x2a002828,0x8002088a,0x888a28a0,0x280a80aa,0x8088a022,0x808220);
remainders_neg[182].modulo_value=728; remainders_neg[182].bit_mask=create_bit_mask(23,0x8a800a0a,0x20008222,0xa2228a28,0x8a02a02a,0x20222808,0x80a02088,0x88808220,0xa80a2202,0x2888aa80,0x2000828a,0x2a2888,0xa8020a0a,0x82800200,0x800800aa,0x80880008,0x282200,0xa08208a8,0xa20820a0,0x8828002,0x22020,0xa0020022,0x800282a,0x802a0);
remainders_pos[183].modulo_value=732; remainders_pos[183].bit_mask=create_bit_mask(23,0xa2022082,0xa28800,0x8a088208,0x2882822,0x800808,0x208a280,0xa20a2800,0x20028220,0x2000080,0x82082,0x8800a20,0x2080208,0x8a0022,0x8208200,0x80200008,0xa0882800,0x28a08,0x28a208,0x2020020,0x28828228,0x208220a,0xa00228a0,0x8208808);
remainders_neg[183].modulo_value=732; remainders_neg[183].bit_mask=create_bit_mask(23,0x228a2802,0x8220008,0x2000280,0x820020a2,0x8008000,0x8a002208,0x22822080,0x28820a28,0x8a088008,0x80802802,0x8228,0x8a008280,0x800208a2,0xa00208,0x8288080,0x20002080,0x8820200,0x88202280,0x828a08a0,0x20020a20,0xa00a282,0x208a2020,0xa00000);
remainders_pos[184].modulo_value=184; remainders_pos[184].bit_mask=create_bit_mask(6,0xa2082aa,0xa8222228,0x28a00,0x2a00a280,0xa0288888,0xaa8208);
remainders_neg[184].modulo_value=184; remainders_neg[184].bit_mask=create_bit_mask(6,0x82280a22,0x20aaaaa0,0x888a0288,0xa2802a08,0x28a00000,0x220a80);
remainders_pos[185].modulo_value=185; remainders_pos[185].bit_mask=create_bit_mask(6,0x4e72b16,0x2125e15,0x31d60191,0x22601ae,0x82a1e921,0x1a3539c);
remainders_neg[185].modulo_value=740; remainders_neg[185].bit_mask=create_bit_mask(23,0x882a228a,0x8a0a8280,0x280a8008,0x8aa20822,0x2282028,0xa02a0a00,0xa2000282,0x2a88a020,0x80a22082,0x8288800a,0xa0880a0a,0x20028a22,0x20088200,0x8a808080,0x82200a00,0x28080000,0x822a8202,0x20802200,0xa2020aa,0x20800008,0xa8802802,0x2008080,0x2002088);
remainders_pos[186].modulo_value=744; remainders_pos[186].bit_mask=create_bit_mask(24,0x28220a2,0x28020828,0xa200002,0x228808a2,0x200a08,0x220a282,0x20820002,0x208028,0x200088,0xa00228a2,0x800820,0x80000202,0x20020080,0x28800a08,0x8008a,0x2080022,0x820828,0x8a088080,0xa0080082,0x20228820,0x8a08a,0x20802880,0x8828028,0x8a);
remainders_neg[186].modulo_value=744; remainders_neg[186].bit_mask=create_bit_mask(24,0x20a2822,0x20828020,0x8228808a,0xa2000022,0xa08200,0x8a28220a,0xa00a0882,0x8a00820,0x88288000,0x208a2022,0x8008028,0x808828a,0xa0820800,0x20008200,0x88008002,0x828008a2,0x8020020,0x2000008,0x20800802,0x28a20028,0x88002002,0xa0082000,0x20820,0x2);
remainders_pos[187].modulo_value=748; remainders_pos[187].bit_mask=create_bit_mask(24,0xaa88020a,0x22220a08,0x2a8a0aa2,0x88002200,0xa0880202,0x2802a08,0x82080aa,0x20a8a22,0xa288aa,0x8880080,0x28280,0x28028282,0x28000828,0x2220028,0x28a00020,0x2a080aa2,0x2082088a,0x80280aa0,0x220a20a,0x80022808,0xa2a8008,0x88888aa,0x22aa20a,0xa08);
remainders_neg[187].modulo_value=748; remainders_neg[187].bit_mask=create_bit_mask(24,0x22008282,0xaa228280,0xa202822a,0x8aa80,0x28000a8a,0x8208a280,0x80a80822,0xa82022a,0x88220022,0x8808,0x888a0a00,0xa08a020a,0xa0088020,0x8aaa88a0,0xa02880a0,0xa280022a,0xa00a0002,0x8a08228,0x8aa02a02,0x882a080,0x2220880,0x80000022,0x82a22282,0x280);
remainders_pos[188].modulo_value=376; remainders_pos[188].bit_mask=create_bit_mask(12,0x82aa8a02,0x202208a8,0xa82000a,0x8a00082a,0x22a20880,0x280a2aa8,0x882aa8a0,0xa202208a,0xa0a82000,0x8a00082,0x822a2088,0x80a2aa);
remainders_neg[188].modulo_value=376; remainders_neg[188].bit_mask=create_bit_mask(12,0xa22028a,0xa8aa0020,0x820a8882,0x28880a2,0xaa2a8008,0xa082a220,0xa22028,0x2a8aa002,0x2820a888,0x8028880a,0xaa2a800,0x82a22);
remainders_pos[189].modulo_value=84; remainders_pos[189].bit_mask=create_bit_mask(3,0x2020022,0x8008a20,0x88008);
remainders_neg[189].modulo_value=84; remainders_neg[189].bit_mask=create_bit_mask(3,0x828a0822,0x800220,0x80);
remainders_pos[190].modulo_value=760; remainders_pos[190].bit_mask=create_bit_mask(24,0xa8a00a8a,0x80028002,0x2082a0a8,0x202280a,0x20882808,0x20200202,0x22208a22,0x28082088,0xa020a2,0x8008020a,0xaa0282,0x8800,0x2200,0x28280aa,0xa08020,0x288a080a,0x88220820,0x888a208,0x8808008,0x80202822,0x8a02880,0x22a0a82,0x2a800280,0xa2a00a);
remainders_neg[190].modulo_value=760; remainders_neg[190].bit_mask=create_bit_mask(24,0x20200202,0x80a0882,0x280a2020,0x8a82a082,0xa000a000,0xa8a8028a,0xaa2802a2,0x2000a000,0x8820a82a,0x808a02,0x88220a02,0x88080080,0x8882288,0x8a020822,0x80280828,0xa0020082,0x2a80a0,0x2200,0x80000880,0xa0a02a,0x80282008,0xa228202,0x22088208,0x222882);
remainders_pos[191].modulo_value=764; remainders_pos[191].bit_mask=create_bit_mask(24,0x820a2aa2,0x8082a008,0x88aa20a2,0xaaa0a00a,0x8aa0882a,0x88a8a80,0x2a020222,0xa22802,0x2a82800,0x8288022,0x22a80a2a,0x800828a,0x8a282002,0xa8a02a8,0x8802282,0x282a8,0x880288a0,0x888080a,0x202a2a22,0xaa8220aa,0x2a00a0aa,0x28a08aa2,0x2200a820,0x8aa8a08);
remainders_neg[191].modulo_value=764; remainders_neg[191].bit_mask=create_bit_mask(24,0xa82a22a,0x80a2880,0x22a82a,0x22282882,0x22800a2,0x20208,0xa28a8aaa,0x882aa08a,0x8a20a088,0x80a008aa,0xaa2082a2,0x80880a02,0x2a0a88a,0x82028a20,0x8008aa0a,0x888a0a20,0x8a0028,0x80008082,0xa8a2a2aa,0x220aa822,0xa2882822,0xa028022a,0xaa8820a8,0x220280);
remainders_pos[192].modulo_value=24; remainders_pos[192].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[192].modulo_value=24; remainders_neg[192].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[193].modulo_value=193; remainders_pos[193].bit_mask=create_bit_mask(7,0x9ba553de,0xc9c74c11,0xb07a092b,0x52417837,0x20cb8e4f,0xef2a9766,0x1);
remainders_neg[193].modulo_value=772; remainders_neg[193].bit_mask=create_bit_mask(25,0x2288a02,0xa8088,0x28a280a2,0x8a88a0aa,0xa8020282,0x22a20aaa,0xaa822a20,0xa0200aa,0xa8288a8a,0x280a28a2,0x880a8002,0x288a200,0xa20882a2,0xaa80a88a,0x820a822,0x80882802,0xaa2a0a0,0x20228000,0xa20222,0x82a2a800,0xa088082,0xa820820,0x8a80aaa2,0xa08822a8,0x2);
remainders_pos[194].modulo_value=776; remainders_pos[194].bit_mask=create_bit_mask(25,0xa2282222,0x80aa022,0x820aa22a,0x2aa20a80,0xa8a2080a,0xaaa82aa2,0x280002,0xa8a20802,0x80a20020,0x2200a282,0x2a00a8a0,0x22282202,0x88288882,0x2a00a880,0x8a00880a,0x8a0282,0x208a2a08,0x280080,0xa82aaa80,0x208a2a8a,0xa08aa8a0,0x8aa08202,0xaa020a8,0x88288a88,0x88);
remainders_neg[194].modulo_value=776; remainders_neg[194].bit_mask=create_bit_mask(24,0x2aa0aaaa,0x808228aa,0xa822aa2,0xa22a8208,0x202a8082,0x2220a22a,0x88a0888a,0x202a808a,0x82a88a8,0xaa882a02,0xa2882028,0xaaa0aa8a,0xa0000a,0xa2882008,0x2880082,0x88028a0a,0xa802a280,0x88a08808,0x20a22208,0xa802a202,0x28022028,0x2280a0a,0x8228a820,0xa00200);
remainders_pos[195].modulo_value=780; remainders_pos[195].bit_mask=create_bit_mask(25,0x800a0002,0x20228a20,0x2082200,0x82020082,0xa00008,0x80282082,0x802,0x28000208,0xa008002,0x80800002,0x808,0x88008200,0x20022000,0x8,0x202202,0x200a0800,0x28800,0x208,0x82802802,0xa00820,0x8082200,0x82080820,0x28808008,0xa00208a,0x800);
remainders_neg[195].modulo_value=780; remainders_neg[195].bit_mask=create_bit_mask(24,0x820882,0x28220220,0x200a288,0x2820802,0x200800,0x820200a,0x80080082,0x20008200,0x8200008a,0x80002,0x8800000,0x80280,0x20822080,0x8808800,0x8028a20a,0xa0820080,0x8020000,0x8080280,0x2002082,0x208020,0x80002288,0x20000a0,0x20000000,0x200a002);
remainders_pos[196].modulo_value=56; remainders_pos[196].bit_mask=create_bit_mask(2,0xaa8aaa2a,0xa8aaa2);
remainders_neg[196].modulo_value=56; remainders_neg[196].bit_mask=create_bit_mask(2,0x22022222,0x202222);
remainders_pos[197].modulo_value=197; remainders_pos[197].bit_mask=create_bit_mask(7,0x37c986d2,0xf8ea8fb6,0x352a1043,0x82152b33,0x7c55c7f0,0xd864fb1b,0x12);
remainders_neg[197].modulo_value=788; remainders_neg[197].bit_mask=create_bit_mask(25,0xaa000a0a,0x2022022a,0xa8a2888a,0xa88a2aa,0xa0880a28,0xa82282,0x20a8028a,0x80882a0a,0x288a80a2,0x8a28aaaa,0x2202288,0x8002aaa2,0xaa828282,0x22a20002,0x2088822a,0x88200008,0x8a820a80,0xa20a028,0xaa002a8,0x8a028228,0x80a820a8,0x8000208,0x2a208882,0xa00022a2,0xa0aa);
remainders_pos[198].modulo_value=264; remainders_pos[198].bit_mask=create_bit_mask(9,0x22002082,0x28020000,0x8a208008,0x800a2022,0x8a00200,0x208a288,0x802820,0x8008800,0x82);
remainders_neg[198].modulo_value=264; remainders_neg[198].bit_mask=create_bit_mask(9,0xa2882002,0x20028808,0x2280080,0x8228a2,0x200a08,0x82002200,0x800820a0,0x800008,0xa);
remainders_pos[199].modulo_value=796; remainders_pos[199].bit_mask=create_bit_mask(25,0x2a08aa2a,0x2aa22082,0x828a088a,0xa2a00808,0xa00a8080,0x808a0aaa,0x822a020a,0x2a80002,0xa8082a2a,0x820a22a2,0x88008222,0x22808a28,0x80200080,0x228a2028,0x28882002,0xa8a88a08,0xa8a8202,0x280002a8,0x2a080a88,0xaaaa0a20,0xa0202a00,0x220200a8,0x8a220a28,0x882088aa,0xa8aa20a);
remainders_neg[199].modulo_value=796; remainders_neg[199].bit_mask=create_bit_mask(25,0xa28022a2,0xa22aa80a,0xa028002,0x2a288080,0x28820808,0x8028222,0xaa28a02,0x8a20888a,0x2080a2a2,0xa82aa2a,0x880aaa,0xaa0802a0,0x8a88808,0xaa02a8a0,0xa000a88a,0x20200280,0x82020a8a,0xa0888a20,0xa2808200,0x222282a8,0x28a8a288,0xaa8a8820,0x2aa82a0,0xa80022,0x2022a82);
remainders_pos[200].modulo_value=40; remainders_pos[200].bit_mask=create_bit_mask(2,0x80820282,0x82);
remainders_neg[200].modulo_value=40; remainders_neg[200].bit_mask=create_bit_mask(2,0x80a0a0a,0x2);
remainders_pos[201].modulo_value=804; remainders_pos[201].bit_mask=create_bit_mask(26,0x2080822,0xa20220,0x88000200,0xa28200a0,0x20800200,0x8228020a,0xa20a28a2,0x8228200,0xa280200,0x882802,0x28000200,0x82002202,0x208a2080,0x8800820,0x8000288,0x2822000,0x8028a08,0x8288200,0xa28a08a0,0x8028828,0x800208a,0xa00828a0,0x8000220,0x8808a000,0x82020800,0x8);
remainders_neg[201].modulo_value=804; remainders_neg[201].bit_mask=create_bit_mask(25,0x828000a2,0x8228a28,0x88280,0x220a0820,0x28008a08,0xa208282,0x22822022,0xa20a08,0x82208288,0x80002082,0x20800a08,0xa08a28a,0xa0022800,0x8028,0x80088200,0x820a2880,0x820200,0x80200288,0x22020020,0x820020,0x8008a002,0x20802020,0x808a28,0x2088,0x28a0080);
remainders_pos[202].modulo_value=808; remainders_pos[202].bit_mask=create_bit_mask(26,0xaa820a0a,0x282a800a,0xa88a88aa,0x200220a,0x82282a20,0xa822208,0x80a0a20a,0x2288a222,0xa2aa2028,0x8800222a,0x28200880,0x8000aa80,0xa00a0a2,0xaa00028a,0x20082802,0x88002202,0x8aa8aa8,0x8a228828,0x8a0a0288,0x8882a0a0,0xa8288220,0x88008008,0x22a22aa0,0x2a828aa,0xa082aaa0,0xa0);
remainders_neg[202].modulo_value=808; remainders_neg[202].bit_mask=create_bit_mask(26,0x220a8282,0xa0a20882,0x20020022,0x8a88aa82,0xaa0a2a8,0x820aaa80,0x8282a82,0xaa002aaa,0x2a22a8a0,0x882aa2,0xa0a88008,0x8882208,0x8288282a,0x22888a02,0xa880a08a,0x88aa8a,0x80220220,0x2aa00a0,0x2828a00,0xa2828,0x20a00aa8,0x880880,0xaa2aa220,0x8a20a022,0x280a2228,0x28);
remainders_pos[203].modulo_value=812; remainders_pos[203].bit_mask=create_bit_mask(26,0x2028a02,0x2a200a80,0xa20a8222,0x8208a0aa,0x2a202880,0x8a0002a8,0x802802a0,0x80800a2a,0xa282080,0x822,0x80088802,0x88a20288,0xa0a0208,0x8a22208,0x22002228,0x802,0x828a0882,0x202020,0x2802a8a,0xa20a8,0x808a82a8,0xa2082022,0x2a08aaa0,0x8a8888,0x2808002a,0x80a);
remainders_neg[203].modulo_value=812; remainders_neg[203].bit_mask=create_bit_mask(26,0x8a8a020a,0x22a88200,0x22020aaa,0xa002822,0xa2a0a008,0x2880a20,0x8a08228,0x8088222,0x2a0a800,0x8880aa,0x800008a,0x228a00,0x82820a80,0x8022a280,0xaa88aa20,0x8888082,0x202800a,0x8828a828,0x8a00a202,0x88822820,0x20220,0x2a80a82a,0x22802220,0x80020000,0xa00008a2,0x82);
remainders_pos[204].modulo_value=408; remainders_pos[204].bit_mask=create_bit_mask(13,0xa20020a2,0x8028208,0x8088002,0x22020000,0x28020800,0x8008a208,0x20a28a0,0x28208a20,0x88002080,0x20000080,0x20800220,0x8a208280,0x8a0800);
remainders_neg[204].modulo_value=408; remainders_neg[204].bit_mask=create_bit_mask(13,0x22882822,0x820a00,0x8000008a,0xa20a0880,0x20828008,0x2280,0x82822020,0x20a00228,0x8a008,0xa0880800,0x28008a20,0x2280208,0x20000);
remainders_pos[205].modulo_value=205; remainders_pos[205].bit_mask=create_bit_mask(7,0x9863329a,0xac3bc0d0,0x2952041d,0x12a51002,0xf70d6e08,0x318642c0,0x1653);
remainders_neg[205].modulo_value=820; remainders_neg[205].bit_mask=create_bit_mask(26,0xaa2a02,0x20220800,0x208a8080,0x8220888a,0x2280a200,0xa80a0a08,0xaa0a8a82,0x20828280,0x802280a,0x22888822,0x20080a88,0x8220,0x2202a2a8,0x8aa00002,0xa2282202,0x80280088,0xa8020880,0xa00a020,0x80028082,0xa0008000,0x80282080,0x200a8202,0xa008088,0xa228880,0x2a8a022,0x22000);
remainders_pos[206].modulo_value=824; remainders_pos[206].bit_mask=create_bit_mask(26,0x822a0222,0x822a2a2,0x88aa20a0,0xa2002a22,0x2a808a0a,0xa2a82a2a,0x8002aa8,0x20202802,0xa08a80,0x82002aa2,0xaa0a2008,0x82020228,0x8222a022,0x82880a88,0xaa288080,0x822008a0,0x8aa800,0x802a20a,0x20802808,0x8a2aa800,0xa8a8a82a,0x8aa0a202,0x2288a800,0x200a08aa,0x828a8a88,0x8880a8);
remainders_neg[206].modulo_value=824; remainders_neg[206].bit_mask=create_bit_mask(26,0xaa28aaa,0x80aa2a2a,0x22a828,0x2a88a22a,0xa2080282,0x2a20a2a2,0x8088a220,0xa8a8a08a,0x88280208,0xa88a22a,0x2282a880,0xa8a8aa0,0xaaa28aa,0xa008200,0x22a00808,0xaa88028,0x88022080,0x808a2a82,0xa808a080,0x2a22088,0x202020a2,0x2282a8a,0xaa002088,0xa8828022,0xa020200,0x820);
remainders_pos[207].modulo_value=276; remainders_pos[207].bit_mask=create_bit_mask(9,0x22082882,0x20a00,0x828a208,0x20008a0,0x8008020,0x8200a200,0x208a2,0x8800a08,0x82282);
remainders_neg[207].modulo_value=276; remainders_neg[207].bit_mask=create_bit_mask(9,0xa2002002,0x8828208,0x80202280,0x82800020,0x800828,0xa082288,0x808a0022,0x8200,0xa20a);
remainders_pos[208].modulo_value=104; remainders_pos[208].bit_mask=create_bit_mask(4,0x2a82020a,0x20aa0808,0x8082a820,0xa0);
remainders_neg[208].modulo_value=104; remainders_neg[208].bit_mask=create_bit_mask(4,0xa20a8a82,0xa8228000,0xa20a8,0x28);
remainders_pos[209].modulo_value=836; remainders_pos[209].bit_mask=create_bit_mask(27,0x22a02222,0xaa200,0x20028002,0xa0a8aa00,0xa0820202,0xa0822288,0xa28a282,0x8a22082a,0x2020828,0x2a8a82a,0xa0020,0x22a8002,0x22202a20,0x8a80888a,0x2a8800,0x800a0008,0x82a2a800,0x8208080a,0x82088a22,0x28a28a0a,0x288820a8,0x80820a2,0xaa2a0a8,0x280080,0x8aa0008,0x8880a880,0x8);
remainders_neg[209].modulo_value=836; remainders_neg[209].bit_mask=create_bit_mask(26,0xaa20a2aa,0x88022a88,0x288a088a,0x28202280,0x280a0a8a,0x200aa200,0x82a02a0a,0x22a8022,0x8a8280a0,0xa2020a2,0x888288a0,0x8aa2008a,0xaa2822a8,0x2080002,0x80a20080,0x8820880,0x22a2088,0xa808002,0xa80022a,0xa0220282,0x2000a820,0x80a82a,0x822a2020,0x88208808,0x80228080,0x82000);
remainders_pos[210].modulo_value=840; remainders_pos[210].bit_mask=create_bit_mask(27,0x20880802,0x20000a20,0x2088208,0x2020082,0xa00a00,0x282208,0xa0020000,0x20000228,0x200080,0x82800022,0x20000000,0x208200,0x880020,0x220000,0x82080008,0x800,0x28200,0x80088,0x80000802,0x800a28,0x88280000,0xa00a0020,0x808000,0x82208082,0xa0000820,0x20220808,0x80);
remainders_neg[210].modulo_value=840; remainders_neg[210].bit_mask=create_bit_mask(26,0xa0000002,0x28008220,0x2000280,0x82020802,0x208208,0x88202280,0x200a0080,0x28808220,0x8208000,0x2080822,0x28000808,0x88200280,0xa0,0x8a28008,0x2008000,0x880080,0x20a00,0x8008000,0x800082,0x220,0x200088,0x20020820,0x8000000,0x220000a,0x20080020,0x28a20000);
remainders_pos[211].modulo_value=844; remainders_pos[211].bit_mask=create_bit_mask(27,0x8aa0a2aa,0x802220a8,0x28020a2a,0xa22aa220,0x88080288,0xaa00a282,0x8a02820a,0x82a8002,0xaa2a20a,0x88008822,0x828aa0a8,0x2888aa00,0x8280220,0x2820000,0xa2228088,0xaa28200a,0x200222a0,0xa8aa0882,0x2a820a08,0x280a2800,0xa00aaa08,0x2022828,0xaa88a228,0x8028088,0x88802a8a,0xa0aa22a0,0xaa8);
remainders_neg[211].modulo_value=844; remainders_neg[211].bit_mask=create_bit_mask(27,0x2282a22,0x8aaa820,0xa08a82a2,0x2aa22aa8,0x808a00,0x22882a0a,0x2820a82,0x80a2088a,0x822a2a82,0x8800aa,0xa022820,0xa0002288,0x80a08aa8,0x8a0a8888,0x2aaa0800,0x22a0a882,0xa88aaa28,0x2022800a,0xa20a8280,0xa082a088,0x28822280,0x8a8aa0a0,0x22002aa0,0x808a0800,0x8a202,0x2822aa28,0x220);
remainders_pos[212].modulo_value=424; remainders_pos[212].bit_mask=create_bit_mask(14,0x2202aa82,0x8a028820,0xaa022020,0xaaa0a0a,0xa208088,0x8808228,0xa8282aa8,0x8202202a,0x208a028,0xa0aaa022,0x880aaa0,0x2280a208,0xaa808808,0x82);
remainders_neg[212].modulo_value=424; remainders_neg[212].bit_mask=create_bit_mask(14,0xaa8a220a,0x28a00a8,0x228aa8a8,0x82228282,0x2a80800,0x80080aa0,0x20a0a220,0xa8aa8a2,0x8a8028a0,0x282228aa,0x80082228,0xaa002a80,0x22080080,0xa);
remainders_pos[213].modulo_value=852; remainders_pos[213].bit_mask=create_bit_mask(27,0x28a0802,0x8228a28,0x8008202,0x2022080,0x28a20200,0x8002200,0x820208a0,0x20008028,0x228a002,0x22020002,0x8008008,0x8a208202,0x828a00a2,0x28200a00,0x8a28a00a,0x2080820,0x8820020,0x88080008,0x8800a2,0x8228020,0x200a208,0xa2800880,0x8000808,0x2002088,0x82080820,0x28028a28,0x8020a);
remainders_neg[213].modulo_value=852; remainders_neg[213].bit_mask=create_bit_mask(27,0x82020082,0xa20220,0x8008020a,0x828a2800,0x20228a08,0x8008a288,0x28a0020,0x28800820,0x8a20208a,0xa28a0882,0x800800,0x2280282,0x2020822,0x20a08208,0x2202082,0x828000a0,0x28828,0x8080,0x80000822,0xa20828,0x8a082280,0x22080000,0x808000,0x8a08a000,0x28000a0,0x20820220,0x8282);
remainders_pos[214].modulo_value=856; remainders_pos[214].bit_mask=create_bit_mask(27,0x8a288aaa,0x8282220a,0xa20a2880,0xaa002288,0x280802a0,0xa020aaa8,0x22222002,0xa2a02aa,0xa8a82800,0x2aa00a02,0x20a20880,0x22828a88,0x882080a2,0x22000000,0x888a0208,0x822a282,0xa802208a,0x2a80a00a,0xa000282a,0x88aa80a8,0xa800888,0x282aaa08,0xaa0a8020,0x8a228800,0x820228a0,0xa2a08882,0xaaa228);
remainders_neg[214].modulo_value=856; remainders_neg[214].bit_mask=create_bit_mask(27,0x2a00222,0xa0aaa82,0x2a82a008,0x2288a200,0xa0808a28,0x28a82220,0xaaaaa88a,0x82a28a22,0x2020a088,0xa228828a,0xa82a8008,0xaa0a0200,0xa8082a,0xaa888888,0x28a80,0x80aa2a0a,0x200aa802,0xa2082882,0x2888a0a2,0x220820,0x82088000,0xa0a22280,0x228208a8,0x2aa0088,0xa8aa028,0x2a28000a,0x222aa0);
remainders_pos[215].modulo_value=860; remainders_pos[215].bit_mask=create_bit_mask(27,0xa80202,0xa82a2,0x288a2288,0x828228a0,0xa0808208,0xa8aa0a80,0x22088880,0x2220,0xa202802,0x8208020a,0x80202800,0x22a8200,0x200a8000,0x8a22088a,0x2a00,0x20082a88,0x20028080,0xa080208,0x802808a,0x80888000,0xa0222208,0xa02a0aa2,0x22082020,0x80a28828,0x2288a22,0x8a82a00,0x80802a0);
remainders_neg[215].modulo_value=860; remainders_neg[215].bit_mask=create_bit_mask(27,0x88200a8a,0x88020222,0x2002a200,0xa02a028,0x20080a00,0x20220208,0xaa000000,0x888a2a8,0x82a0a08a,0x2808a02,0x8a82088,0x8a220a80,0x28820080,0x2a28002,0x8088a208,0xa8802200,0xa80a0800,0x2800280,0x80820802,0x808,0x28aa2a80,0x28228222,0x2280a0a8,0x82200a0,0x82a0022a,0x80202288,0x8a20);
remainders_pos[216].modulo_value=24; remainders_pos[216].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[216].modulo_value=24; remainders_neg[216].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[217].modulo_value=868; remainders_pos[217].bit_mask=create_bit_mask(28,0xa02220a,0x20880080,0x820a0a88,0x22a2800,0x2220800,0x82a002a8,0xa8aa2282,0x2a0a0a22,0x2200aa00,0xa2000082,0x820800a2,0x88208a82,0x2800800,0x882a8222,0x2002808,0x2a208220,0xa0020828,0x200008a8,0xaa00888,0x8a0a0a80,0x288aa2a8,0xa800a828,0x2088802,0x28a8800,0x2a0a0820,0x20022082,0x8880a00,0xa);
remainders_neg[217].modulo_value=868; remainders_neg[217].bit_mask=create_bit_mask(28,0x28aaa02,0x28008800,0x2828200,0x8a22a088,0x82a28088,0xa280a20,0x2022a20a,0xa282822a,0x2a082280,0x2288880a,0xa00882a,0xa0020a,0x8a000088,0xa202aa,0x8a88a000,0x22a80aa0,0x208a00a0,0xa8088020,0x82208000,0x2820208,0xa0022220,0x20882020,0xa800082,0x82020088,0xa2028028,0xa882a80a,0x80000288,0x2);
remainders_pos[218].modulo_value=872; remainders_pos[218].bit_mask=create_bit_mask(28,0x8208aa82,0x882a0020,0xaa222a8,0xa2a0a2a,0x22aaa2,0x80a88a28,0x8282200a,0xa80aa22,0x2aa82088,0x20a20002,0xa02020,0x28880222,0xa28a2aa0,0x2828008,0xa8a28a20,0x8022280a,0x80a0088,0x8a0808,0x82aa880,0xaa02a022,0x8828288,0xa22a02a0,0xaa880028,0xa0a8a08a,0x888aa0a8,0xa8222a,0xaa208208,0x82);
remainders_neg[218].modulo_value=872; remainders_neg[218].bit_mask=create_bit_mask(28,0xa80220a,0xa288a8,0x822aaa20,0x82a282a2,0x88aa222a,0x82002a0,0xa0aa882,0x820822aa,0xa220a800,0xa82a888a,0x8828a828,0xa0008aaa,0x2a02a228,0x8a0a0880,0x202a02a8,0x8aaa082,0x80828800,0x88028080,0x80a22008,0x228a28aa,0x800a0a00,0x2aa28a28,0x220088a0,0x20202802,0x22820,0x8820aaa2,0x22a80a80,0xa);
remainders_pos[219].modulo_value=876; remainders_pos[219].bit_mask=create_bit_mask(28,0xa28200a2,0x20220828,0x200080,0x228a20a2,0x8828800,0x220820a,0x28208a0,0x8a00208,0x2008208,0x8a2002,0x28800800,0x88280288,0x2080,0xa08,0x8000020a,0x2822020,0x228228,0x8a200002,0x20080800,0xa20208,0x8280208,0x208800a2,0x28220a08,0x8a288002,0x8008a0,0x8808020,0x828a282,0x8a0);
remainders_neg[219].modulo_value=876; remainders_neg[219].bit_mask=create_bit_mask(28,0x220a0822,0x28a28020,0x88288008,0xa2022822,0x20008,0x8a280282,0x820a0020,0x208a00,0x8a080280,0x80022882,0x20008008,0x208200,0x80882800,0x8808200,0x8088282,0x20a28a0,0x8a20a20,0x228808a,0xa0800080,0x8228a00,0x80208280,0xa0000822,0x20a28200,0x220008a,0x80080020,0x820,0x8020220a,0x20);
remainders_pos[220].modulo_value=440; remainders_pos[220].bit_mask=create_bit_mask(14,0x88a220a,0x20a8080,0x2028228,0x200080,0x2882808,0x202a08,0xaa088a22,0x88a220,0x8020a808,0x202822,0x80020008,0x80288280,0x200202a0,0xa088a2);
remainders_neg[220].modulo_value=440; remainders_neg[220].bit_mask=create_bit_mask(14,0x80022282,0x8a020800,0xa8a02a0,0x88a08800,0x82002000,0x80a82280,0x220002a2,0x8002228,0x8a02080,0xa8a02a,0x88a0880,0x8200200,0x280a8228,0x20002a);
remainders_pos[221].modulo_value=221; remainders_pos[221].bit_mask=create_bit_mask(7,0xd3110af2,0x86e27e78,0x408322a5,0xb74120bb,0x69513040,0xc79f91d8,0x13d42232);
remainders_neg[221].modulo_value=884; remainders_neg[221].bit_mask=create_bit_mask(28,0xa88822a,0xa22a220,0x800aaa28,0x2a08a822,0xa0882888,0x2020800,0x8a08aaa2,0x22aa882,0x28008202,0xa088a088,0xa20a882,0x80a2aa80,0x80222a22,0x2a2088a,0x2a808022,0x2820020,0x208a0002,0x8a00080a,0xa8208808,0x282a282,0xa0028800,0xa0a00008,0x820aa0a2,0x288808,0x28822808,0x20a02000,0x80aa0200,0x2200);
remainders_pos[222].modulo_value=888; remainders_pos[222].bit_mask=create_bit_mask(28,0x82022002,0x28a28008,0x8200a288,0x20a20a0,0x220828,0x82202,0x208a28a0,0x28000008,0x82202082,0xa0082022,0x8020020,0x8000282,0x800a0000,0xa20220,0x208808a,0x200000a0,0x20828000,0xa080080,0x82880820,0x28820808,0x8200000,0xa28a2,0x808820,0x80282088,0x820a08a0,0x28228a00,0x8220028a,0x800880);
remainders_neg[222].modulo_value=888; remainders_neg[222].bit_mask=create_bit_mask(28,0x28a2882,0x20220800,0xa082200,0x82822820,0x8a28020,0x8800a28a,0xa0022020,0x20808800,0xa28a002,0x208028a2,0x828828,0x8008820a,0x20880,0x8228a28,0x8a000002,0xa0880820,0x28020808,0x82008008,0x20000a0,0x20028000,0x80288088,0x822022,0x8000028,0x820a000,0x2820020,0x20a20208,0xa208202,0x80000);
remainders_pos[223].modulo_value=892; remainders_pos[223].bit_mask=create_bit_mask(28,0x2a820a0a,0x82a022a,0x8a828aaa,0xa00aaa8,0x8a880020,0x88220288,0xaa08228,0x208a2802,0x220a88a2,0xa8aaa202,0x20002a82,0x80020a02,0x808a80a2,0x2828a00,0x200a2828,0x28a02a20,0x880a0800,0xa82a8000,0x8808aaa2,0x88a22a08,0x8028a20,0x228820aa,0x22280882,0x80022a,0x22aaa00a,0xaaa282a,0x8a880a82,0xa0a082a);
remainders_neg[223].modulo_value=892; remainders_neg[223].bit_mask=create_bit_mask(28,0xa20a8282,0x80a28aa2,0x20a0222,0x82882220,0x20088a8,0xaa8a00,0x2280aa0,0xa802a08a,0xaa82002a,0x20222a8a,0xa888a20a,0x88a828a,0x802082a,0x8a0a0288,0xa882a0a0,0xa028a2a8,0x828088,0x20a20888,0x80222a,0x2aa280,0x808a02a8,0xaa00a822,0xaaa0800a,0x88088aa2,0xaa222882,0x8222a0a2,0x200820a,0x28280a2);
remainders_pos[224].modulo_value=56; remainders_pos[224].bit_mask=create_bit_mask(2,0x82002a22,0x88a800);
remainders_neg[224].modulo_value=56; remainders_neg[224].bit_mask=create_bit_mask(2,0xa88a22a,0x2080);
remainders_pos[225].modulo_value=60; remainders_pos[225].bit_mask=create_bit_mask(2,0xa08a2882,0x8228a20);
remainders_neg[225].modulo_value=60; remainders_neg[225].bit_mask=create_bit_mask(2,0x20022002,0x220220);
remainders_pos[226].modulo_value=904; remainders_pos[226].bit_mask=create_bit_mask(29,0xaa2882aa,0x8a022a28,0xa0820808,0xaa008a22,0x80a80a2,0x2a2a8a00,0x22802a20,0x202a8222,0xa080202a,0xa80a8aa,0x2082a002,0xa8a28a02,0x2022a088,0x2882008,0x8228000,0xa880820,0xa28a2a22,0xa820880,0x2a02a080,0x8020aaa,0x82a808a8,0xa8028888,0xa2a8a808,0x2a02000,0xa200aa8a,0x20820a88,0xa880a220,0x8228aa28,0xaa);
remainders_neg[226].modulo_value=904; remainders_neg[226].bit_mask=create_bit_mask(29,0x22a00a22,0x28aa2a0,0x280a8080,0x228802aa,0x8082082a,0xa2a20288,0xaa08a2a8,0xa8a20aaa,0x2808a8a2,0x82082022,0xa802288a,0x202a028a,0xa8aa2800,0x8a00a880,0x80aa0888,0x820080a8,0x2a02a2aa,0x820a8008,0xa28a2808,0x808a8222,0xa208020,0x208a0000,0x2a202080,0x8a28a888,0x2a082202,0xa80a8200,0x20082aa8,0xaa022a0,0x22);
remainders_pos[227].modulo_value=908; remainders_pos[227].bit_mask=create_bit_mask(29,0xa2208202,0x2aa008a,0x8a2a222a,0x8a8aa822,0x80a0a2,0xaaaa220a,0xa0aa2080,0x28802802,0x8880002a,0x282a2a82,0xa0020208,0x88808208,0x2a800a80,0x8a288802,0x228a2a0a,0x2a8802,0x2022202a,0x800a208,0x8a828208,0x22282a,0x80228a80,0x8aa0a802,0x8aaaa020,0xa0200a08,0xaa2a28a0,0x8a8a2882,0xaa80a88,0x2088aa20,0x808);
remainders_neg[227].modulo_value=908; remainders_neg[227].bit_mask=create_bit_mask(29,0x2aa80a8a,0x8a228802,0x2a2aaa2,0x20220aa,0x8808282a,0x2222aa82,0x2822a808,0xa008a082,0x888a2,0xa0a2a20a,0x288a8a80,0x80a80,0xa2088208,0x2a0008a,0xaa02a282,0x88a2008a,0xa8aaa8a2,0x80882a80,0x20a0a80,0x88aaa0a2,0x8aa0208,0x228208a,0x22228a8,0x28a88280,0x22a2a028,0x202a00a,0x82208200,0xa80022a8,0x80);
remainders_pos[228].modulo_value=456; remainders_pos[228].bit_mask=create_bit_mask(15,0x22000082,0x28a20a00,0x2200282,0x20a0800,0x28008800,0xa08a288,0x20000880,0x20000828,0x8a20a002,0x22002822,0x20a08000,0x80088000,0xa08a2882,0x8800,0x82);
remainders_neg[228].modulo_value=456; remainders_neg[228].bit_mask=create_bit_mask(15,0xa2800802,0x20228208,0xa28820a,0x82820080,0x20800008,0x82002200,0xa0880000,0x28008020,0x228208a,0xa28820a2,0x28200800,0x8000088,0x20022002,0x8800008,0xa);
remainders_pos[229].modulo_value=229; remainders_pos[229].bit_mask=create_bit_mask(8,0xe1bda3a,0x77ab7c32,0xc82f58d1,0x6240919a,0xc6bd04d6,0xfb57ba2,0x16f61c13,0x17);
remainders_neg[229].modulo_value=916; remainders_neg[229].bit_mask=create_bit_mask(29,0x828202a2,0xaa22a0aa,0xa28008,0xaa880802,0xa20880a,0x8228a22a,0x8a2a808a,0xaa28a8a,0x28a20a88,0x882282a2,0x808aaa80,0xa280200,0x2a22a880,0x20a0aa8,0xa2a0222a,0x202800a0,0x20aa8802,0x8a8aa2a8,0x288a8000,0x8220282,0xa880a2,0x20800082,0x22808a80,0x20a02208,0x8000a88a,0xaa2a8a8,0x2008aa82,0x82800a02,0x202a2);
remainders_pos[230].modulo_value=920; remainders_pos[230].bit_mask=create_bit_mask(29,0x222202,0x2a028a02,0x208aa008,0x8222a802,0x82200a28,0xa0282280,0xaa2a08a0,0x28,0x8220808,0xa2208202,0x220800,0xa02a0,0x2802028,0x202a0a0,0x80888222,0x800a0a80,0x280802,0xa80a0,0x8a002088,0x20808208,0x202088,0xaa280000,0xa0a20a8,0x82028828,0x8228a008,0x8802a88,0xa8200aa2,0x80a280,0x808888);
remainders_neg[230].modulo_value=920; remainders_neg[230].bit_mask=create_bit_mask(28,0x882a2a8a,0xa20a0282,0x28022080,0xaa2208a,0x2a88220,0x28a02a08,0x22220020,0x88880a0,0x80a28080,0x22a8020a,0x88aa0088,0x88028a20,0xa08a020,0x8a822828,0xa2a,0x8820200,0x88208082,0x8820028,0x280a800,0xa0080a00,0x80a82800,0x22208880,0x282a020,0xa0200a0,0x2a02800,0x80082200,0x20208222,0x8082208);
remainders_pos[231].modulo_value=924; remainders_pos[231].bit_mask=create_bit_mask(29,0xa2802022,0x20000820,0x2088282,0xa0080080,0x8a20008,0x82080280,0x808800a0,0x20208,0x2002002,0x20802,0x20000028,0x200a20a,0x82080800,0x28000020,0x80000000,0x20800002,0x20208,0x8a08a008,0x2800000,0x8020800,0x8008008,0x22080800,0x20a00220,0x280208,0xa20008a2,0x200200,0x88282208,0xa0820000,0x8808028);
remainders_neg[231].modulo_value=924; remainders_neg[231].bit_mask=create_bit_mask(29,0x22082022,0x28008020,0x8200020a,0x20000800,0x220800,0x2000208,0x20,0x8828a00,0xa00a082,0x8a0082,0x28008820,0x8a002282,0x2000080,0x20808028,0x8088000,0x20080882,0x828a00,0x2002080,0x82000800,0x820008,0x80080,0xa2800000,0x28208220,0x80208280,0x22080022,0x8a08a00,0x202280,0x200a0080,0x820);
remainders_pos[232].modulo_value=232; remainders_pos[232].bit_mask=create_bit_mask(8,0xaa80a8a,0xa2020822,0x2228a2,0x20080a8,0x2888002a,0x20808a8a,0xa02aa088,0xa2);
remainders_neg[232].modulo_value=232; remainders_neg[232].bit_mask=create_bit_mask(8,0x82208202,0x2a8a80aa,0x882aa02a,0x8a880820,0xa00088a2,0xa8080202,0x28a22000,0x2a);
remainders_pos[233].modulo_value=233; remainders_pos[233].bit_mask=create_bit_mask(8,0xf68de396,0xd59e4073,0x1a221585,0x5b336b74,0xa11160bb,0x9e6ae86,0x1ec5bf38,0x1a7);
remainders_neg[233].modulo_value=932; remainders_neg[233].bit_mask=create_bit_mask(29,0x2a002a0a,0x80288aa,0x82aa8808,0x82aaa2a8,0x2888a822,0x802a220a,0x820822a0,0x820aa02a,0xa0082a20,0x88a28222,0xaa0a20a8,0xaa08aa2a,0x80808a,0x2a2a88a,0x28282a0,0x880202aa,0x8a8a8aa0,0x2008a000,0xa20a000,0x22820888,0x2028aa02,0x2200a28a,0x228a280,0xa2202a8a,0x28088820,0x2000282,0xa8800288,0x882a8a8,0xa02aa020);
remainders_pos[234].modulo_value=312; remainders_pos[234].bit_mask=create_bit_mask(10,0x28a0822,0x8820020,0x80288008,0xa20a2080,0x200000,0x8a000008,0x20208a0,0x20200228,0x80080082,0x8820a2);
remainders_neg[234].modulo_value=312; remainders_neg[234].bit_mask=create_bit_mask(10,0x820200a2,0x28828,0x200080,0x22822800,0x8a00808,0x2088080,0x828a0020,0x28208a20,0x800800a,0x2022);
remainders_pos[235].modulo_value=940; remainders_pos[235].bit_mask=create_bit_mask(30,0x88282a8a,0xa2020082,0x2a0a2200,0x22200a8,0xa88a20,0x20880a00,0xa0a0082,0x28a0280,0x8822882,0x22a88222,0x88a80000,0x80200a80,0x80a8080,0x8202280a,0x20,0x88082080,0x2a020a02,0x802020,0x2a2202a,0x22a88000,0x88220888,0xa280822,0xa0a0028,0x2208820,0x22a0000a,0x888008a,0x8a0a82a0,0x808a008,0x82822820,0xa2a);
remainders_neg[235].modulo_value=940; remainders_neg[235].bit_mask=create_bit_mask(30,0xa02202,0x2a0a0802,0x2282a288,0x8aa28820,0x80200228,0xa8000288,0x82028802,0xa028208,0x8002a00a,0xa2200a2a,0x200888,0x8288200,0x820008,0xa82a002,0x80888828,0x802808,0xa20a8282,0x808a0a8,0x8a22a8a2,0xa2200800,0xaa0000,0x822080a2,0x28280a0,0x8aa000a8,0x22288802,0x80000802,0x2020a20,0x802080,0xa02a0a8,0x222);
remainders_pos[236].modulo_value=472; remainders_pos[236].bit_mask=create_bit_mask(15,0xa2a20a22,0x2a2aa80,0x82a0008,0x22088288,0xa820228,0xa8a80200,0x8a8a02a,0x20a2288a,0x2aa80a2a,0xa000802a,0x88288082,0x20228220,0x802000a8,0x8a02aa8a,0x88a08a);
remainders_neg[236].modulo_value=472; remainders_neg[236].bit_mask=create_bit_mask(15,0x2a2a82aa,0x822a2208,0x80a28880,0xaa800a00,0x820a8aa0,0x20208a88,0x802028a2,0xa82aa002,0xa22082a2,0x28880822,0xa0080a,0xa8aa0aa8,0x8a88820,0x28a2202,0x2802);
remainders_pos[237].modulo_value=948; remainders_pos[237].bit_mask=create_bit_mask(30,0xa20a2002,0x8a28208,0x2288,0x20a0802,0xa00200,0x8220220a,0xa0800020,0x28820208,0x8200008,0x2822022,0x20028,0x2008282,0x808a2000,0x28820a28,0xa002282,0x22882880,0x20228a08,0x800008a,0x82820,0x28028008,0x82088088,0x22820000,0x20a20808,0x88208000,0xa00a0880,0x8000800,0x8020a,0xa2022880,0x8a20828,0x8008a);
remainders_neg[237].modulo_value=948; remainders_neg[237].bit_mask=create_bit_mask(30,0x22822882,0x220a00,0x88082200,0x82820082,0x8208a08,0xa28a282,0x200808a0,0x20028a00,0x80288080,0x820a28a2,0x8828820,0x8a08020a,0x22080,0x20028220,0x8208a20a,0xa2002000,0x28a20200,0x80088002,0x808020a0,0x20820800,0xa000000,0xa20a0880,0x28228000,0x280088,0x20820000,0x808008,0x88008282,0x228a2000,0x228020,0x8002);
remainders_pos[238].modulo_value=952; remainders_pos[238].bit_mask=create_bit_mask(30,0x2a80220a,0x808aa0,0x80828aa8,0xa20a0a2,0x28002a0a,0x80220,0x88228220,0xa202a,0x2a208220,0x222288a8,0xa02028a0,0x82020a00,0x88000208,0x2800808a,0x822200,0x28008882,0x22a20200,0x82208000,0xa00a080,0x880a2808,0xa82a2288,0x88208,0x22a808a0,0x88288,0x28088020,0xa0a0a800,0x28a0a08,0x2aa282,0xa80aa202,0xa08802);
remainders_neg[238].modulo_value=952; remainders_neg[238].bit_mask=create_bit_mask(30,0xa208aa02,0x8000220,0xa0220,0x8228282a,0xa080a282,0x80800aa8,0xaa02a8,0x882a822,0x22a80aa0,0xa2aa0020,0x2828a020,0xa828288,0x880a00,0xa0880002,0x880aa208,0x20880002,0xa22a0a88,0xa280888,0x82802808,0x822080,0x2022a200,0x88800a00,0x22208020,0x80800a00,0x200008a8,0x28202088,0x8a020280,0x88a22202,0x20822a0a,0x280002);
remainders_pos[239].modulo_value=956; remainders_pos[239].bit_mask=create_bit_mask(30,0x228a02a2,0xa802a88a,0xa0228000,0x2a8aa8a0,0xa2828828,0x80088aa2,0x88a08aaa,0x222088a,0x22822,0x80222aa,0xa0a20a08,0x28200200,0x2aaa08a8,0x82200aa0,0x80a8208,0x22082a02,0x80aa0088,0x82a20aaa,0xa0080082,0x20a08a0,0xaa88802,0x8828800,0x2a220888,0x2aaa20a2,0xa8aa2200,0x82822828,0xa0a2aa2a,0xa0002880,0x8a22a802,0x8a80a28);
remainders_neg[239].modulo_value=956; remainders_neg[239].bit_mask=create_bit_mask(30,0xaa028a2a,0x208a2002,0x28aa0888,0xa2022028,0x2a0a00a0,0x880022a,0x280222,0x8aaa0002,0x888aa0aa,0x808aaa22,0x282a8280,0xa0a88a88,0xa2228020,0xaa88228,0x80820a80,0xaa80a28a,0x8228800,0xa2a8222,0x2880880a,0x8a828028,0x8220008a,0x800a0088,0xa2aa8000,0xa222a82a,0x2022aa88,0xa0aa0a0,0x282a22a2,0x2888a008,0x2aa208a,0x2082a0);
remainders_pos[240].modulo_value=120; remainders_pos[240].bit_mask=create_bit_mask(4,0x20882,0x28220800,0x2088,0x822080);
remainders_neg[240].modulo_value=120; remainders_neg[240].bit_mask=create_bit_mask(4,0x808a0002,0x20228000,0x808a000,0x22800);
remainders_pos[241].modulo_value=241; remainders_pos[241].bit_mask=create_bit_mask(8,0x6b15977e,0x3c67a311,0x4c8fa909,0x2fd21c17,0xc4cba0e1,0x98f24257,0xa35a2317,0x1fba6);
remainders_neg[241].modulo_value=964; remainders_neg[241].bit_mask=create_bit_mask(31,0xa2880aa2,0xa0aa2a88,0x80022080,0xa20a808a,0x8022828,0x2a8a8a,0x2a82aa8a,0xa220222a,0x8aaa0aa2,0x8a8aa002,0xa0a20082,0x880a8228,0x822000a,0x8aa2a828,0x2a808a28,0x208a8022,0x28020088,0xaaa228a8,0x2280a880,0x8aa20a08,0xaa008082,0xa00082,0x22222202,0x80028020,0x80802aa0,0x2822a8a0,0x8a80a208,0x8a22aa80,0x80200a0a,0xa88208,0x2);
remainders_pos[242].modulo_value=88; remainders_pos[242].bit_mask=create_bit_mask(3,0x82828282,0x82028280,0x828282);
remainders_neg[242].modulo_value=88; remainders_neg[242].bit_mask=create_bit_mask(3,0xa0a020a,0xa0a0a08,0xa0a0a);
remainders_pos[243].modulo_value=12; remainders_pos[243].bit_mask=create_bit_mask(1,0x802);
remainders_neg[243].modulo_value=12; remainders_neg[243].bit_mask=create_bit_mask(1,0x82);
remainders_pos[244].modulo_value=488; remainders_pos[244].bit_mask=create_bit_mask(16,0xa08a22a,0x202a280,0x800a2a02,0xaaa22882,0x282288,0x8080a8a,0xa0028a8,0x22aa88a2,0x2800a08a,0xa020202a,0x882800a2,0x288aaa22,0xa8a00282,0x8a808080,0x8a20a002,0xa8);
remainders_neg[244].modulo_value=488; remainders_neg[244].bit_mask=create_bit_mask(16,0x82802aa2,0x8a8a2a08,0x882a28a,0x222aa00a,0x88a0aa00,0x80008202,0x8288a020,0xaa22002a,0xa0882802,0x28a8a8a2,0xa0882a,0xa00222aa,0x20288a0a,0x2080008,0x2a8288a,0x20);
remainders_pos[245].modulo_value=140; remainders_pos[245].bit_mask=create_bit_mask(5,0xa0080a02,0x28080280,0x20280a0,0x200a028,0x80a);
remainders_neg[245].modulo_value=140; remainders_neg[245].bit_mask=create_bit_mask(5,0x2880020a,0x20008a00,0x28a0028,0x8a0028a0,0x2);
remainders_pos[246].modulo_value=984; remainders_pos[246].bit_mask=create_bit_mask(31,0x28228a2,0x820808,0xa20a202,0x22882000,0x20808a00,0x828a20a,0x208000a0,0x20220228,0xa082002,0x82802,0x8820020,0x8002280,0x20028a2,0x20828220,0x200a080,0x80000000,0x8020a00,0x80088282,0x208a2800,0x20028800,0x80082,0xa0802820,0x8800820,0x8288088,0x200a0002,0x8a08a28,0x8800a202,0xa0000822,0x808a08,0x80202082,0x8a2882);
remainders_neg[246].modulo_value=984; remainders_neg[246].bit_mask=create_bit_mask(31,0x820a2022,0x8028000,0x8228228a,0xa2002880,0x28000208,0x80202282,0xa0080820,0x28a28a20,0x200a08a,0x80802082,0x28828,0x8008a208,0x82882022,0x28020a28,0x8a082000,0x880880,0x828208,0x800020a,0xa0022080,0x28820008,0x8800800a,0x200820a0,0x8028,0x80200000,0xa0820082,0x200220,0x8228a,0x208800a2,0x8000200,0x828200a,0x22002);
remainders_pos[247].modulo_value=988; remainders_pos[247].bit_mask=create_bit_mask(31,0xa220a8a,0xa00a8222,0x20aa800,0x8a88a8aa,0x2a0a2000,0x28888008,0x880002,0x20aa282,0x82288a08,0x222082aa,0x280a282,0x200228a8,0x2820020,0x2880aa0,0x8082880a,0x20080200,0xa022820,0xaa0228,0x80800828,0x2a28800,0x8828a028,0x2aa82088,0x20a2288,0x828aa08,0x88000220,0x82002222,0x20008a0a,0xaa2a22a,0xa002aa08,0x8882a00,0xa2a088a);
remainders_neg[247].modulo_value=988; remainders_neg[247].bit_mask=create_bit_mask(31,0x82a28202,0x28820a2a,0x2822088,0x2002022,0xa2822888,0xa0000080,0x88008882,0x8a022a0a,0xaa00280,0xaaa80222,0xa082a02,0xa88aa020,0x8a0208a8,0x8a008228,0xa0002,0xa0808a88,0x828aa0a8,0x88220a20,0x80880a0,0x8a220080,0x2028a0,0x2220a800,0x8a82a200,0xa02280,0x8882a8,0xa80aaaa,0xa8080282,0x822a2a22,0x208a2280,0x8000a280,0x2a20002);
remainders_pos[248].modulo_value=248; remainders_pos[248].bit_mask=create_bit_mask(8,0x22a8a202,0x28aa022a,0x22a008,0x82a20802,0x80208a,0x28200a88,0x88a880aa,0x808a2a);
remainders_neg[248].modulo_value=248; remainders_neg[248].bit_mask=create_bit_mask(8,0x2a202a8a,0xa0228aa2,0x88aa2880,0xa2a808a,0x8008a802,0xa0a88200,0x200822,0x802a2);
remainders_pos[249].modulo_value=996; remainders_pos[249].bit_mask=create_bit_mask(32,0x820000a2,0x20228028,0x2000080,0xa2002820,0x8a08200,0xa088200,0x22020082,0x28a28a20,0x220228a,0x82820802,0x28800208,0x2280208,0x20020a0,0x20200800,0x208a00a,0x28a2800,0xa20800,0x200808a,0xa0800800,0x8028800,0x8002282,0x2082822,0x28808808,0x8a28a28a,0x20080880,0x8220a08,0x820a200,0x828008a0,0x20000800,0x80288080,0xa0000822,0x8);
remainders_neg[249].modulo_value=996; remainders_neg[249].bit_mask=create_bit_mask(31,0x2880822,0x28a20820,0x8a008008,0x228820a0,0x200a08,0x82000288,0xa28a0802,0x20220228,0x8a28a202,0x20a0082,0x20008a00,0x8a208280,0x2882820,0x28a08008,0x8a002082,0x82022080,0x8228008,0x8a080002,0x20080080,0x820008,0x8008a20a,0x828020a2,0x20000000,0x2202202,0xa0800000,0xa28200,0x80282288,0x2080020,0x28808008,0x8200008,0x208800a2);
remainders_pos[250].modulo_value=40; remainders_pos[250].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[250].modulo_value=40; remainders_neg[250].bit_mask=create_bit_mask(2,0x882a82,0x20);
remainders_pos[251].modulo_value=1004; remainders_pos[251].bit_mask=create_bit_mask(32,0x22a2a22,0x882aa00,0xa2a202a2,0xa2228a28,0x82a08800,0xa8a02a8a,0xa2822aa0,0x2a2aa82,0x80a0008,0x2028008a,0xa0200a8,0x882aa228,0x8208208,0x22880a88,0x8aaa800a,0x22088080,0x2aaa2020,0x2288a00,0x2082022a,0xaa822208,0x80a0288,0x28082a0,0xa020a20,0xa8a80200,0x8828a82a,0x80a2a0aa,0x20a82a2a,0x2888a002,0x8a8a28a,0xa82208a8,0x8a88000a,0x88a);
remainders_neg[251].modulo_value=1004; remainders_neg[251].bit_mask=create_bit_mask(32,0x8aa2a2aa,0x800a2288,0x2a2a8a2a,0x2aaa02a0,0xa280088,0x2028a202,0x2a0aa228,0x822a220a,0x80828880,0xa8a08802,0x828a8820,0xa22aa0,0x80a80a80,0xaa008200,0x2220882,0xaa800808,0xa222a8a8,0x8aa00288,0xa80a8aa2,0x220aaa80,0x80828a00,0x8a080a28,0x828a82a8,0x20208a88,0xa020a2,0x82a2822,0xa820a2a2,0xa000288a,0x80202a02,0x20aa8020,0x2008882,0x2);
remainders_pos[252].modulo_value=168; remainders_pos[252].bit_mask=create_bit_mask(6,0xa2080002,0x8a08020,0x280002,0x20a2080,0x208a08,0x80);
remainders_neg[252].modulo_value=168; remainders_neg[252].bit_mask=create_bit_mask(6,0x22800802,0x200820,0x8020808a,0x82022800,0x8a00200,0x8);
remainders_pos[253].modulo_value=1012; remainders_pos[253].bit_mask=create_bit_mask(32,0x8a2a028a,0xaa0a8800,0x200a8882,0x802800,0x22a2022,0x228a8a08,0x82202008,0x20a200a,0xa802282,0x20802022,0x828a8a,0x22022a2,0x2000a008,0xaa088a80,0x88008a82,0x2a8a02a2,0x228a80a,0xaa82a20,0x802a22,0x880200a0,0x2008a880,0x208a2a28,0x2a088080,0x8082880,0x882a008a,0x28820080,0x88020a2a,0x2008808a,0x800280,0xaa8222a,0x8a20022a,0xa280a);
remainders_neg[253].modulo_value=1012; remainders_neg[253].bit_mask=create_bit_mask(32,0x2228202,0x22020088,0xa882000a,0x8800a080,0x8aa228aa,0xa2020280,0xaa82880,0x8a82a802,0x8200aa0a,0x2808a0aa,0x880a0202,0x8aa8a22a,0xa8082800,0x22800208,0x880202,0xa2020a22,0x82a02082,0x220a2a8,0x8808a22a,0x828828,0x20802008,0xa802a2a0,0xa2800008,0x8000a008,0xa28802,0xa00a8800,0x8202a2,0xa0800802,0x88088a08,0x82202a22,0x2a08aa2,0x2a082);
remainders_pos[254].modulo_value=1016; remainders_pos[254].bit_mask=create_bit_mask(32,0x228a0aa2,0x80a22288,0xa0222200,0x228aa828,0x282202a,0x8aaaa8a,0x802aaa08,0x2a20a02a,0x802a0a20,0xa8a0002a,0xa0800008,0x2a2282,0x8280082,0x8aa2222a,0x2882220a,0x8200a082,0x28820a00,0xa2a08882,0x20a8888a,0x820028,0xa8288a8,0x2a200002,0x2a8000a,0xa808a0a8,0x2a80a08,0x2020aaa8,0x80a2aaaa,0x88a80882,0xa282aa2,0x2008888,0x8822888a,0x8aa0a2);
remainders_neg[254].modulo_value=1016; remainders_neg[254].bit_mask=create_bit_mask(32,0xaa02822a,0x82aaa00,0x28aaaa88,0x2a0220a0,0x8a0aa8a2,0x80222202,0x8a22280,0xa2a828a2,0x8a282a8,0x202888a2,0x28088880,0x88a2aa0a,0x80a0880a,0x22aaaa2,0xa00aaa82,0xa88280a,0xa00a8288,0x2a28000a,0xa8200002,0x800a88a0,0x820a0020,0xa2a8888a,0x8a208882,0x20802820,0x8a208280,0xa8a82220,0x82a2222,0x20800a,0x82a0a22a,0x8a880000,0xaa0002,0x2282a);
remainders_pos[255].modulo_value=1020; remainders_pos[255].bit_mask=create_bit_mask(32,0xa0800002,0x8220a20,0x800a208,0x82000882,0x8a20,0x80080280,0x800a0082,0x8808028,0x202002,0x2080020,0x28020200,0x8000282,0x2022000,0x800820,0x8002,0x8a0800,0x20a20,0x8002000,0x820020,0x8808,0x88280002,0x80802,0x800208,0x8008080,0x22802022,0x28200a00,0x280200,0x208a2000,0x8220008,0x208a002,0xa08a0882,0x8000020);
remainders_neg[255].modulo_value=1020; remainders_neg[255].bit_mask=create_bit_mask(32,0x20080882,0x228220,0x82280,0x2000002,0x800228,0x208,0x20802,0x20,0x8820a08a,0x82800820,0x20820a00,0x80088202,0x28a2000,0x8008028,0x8008000a,0x80020080,0x8020220,0x8a088,0x800208a0,0x800000,0x20008a,0x80000082,0x8200,0x80000008,0x22082822,0x20a00208,0x88208280,0x20022080,0xa28800,0x8200200a,0x20020002,0x8020);
remainders_pos[256].modulo_value=8; remainders_pos[256].bit_mask=create_bit_mask(1,0xaa);
remainders_neg[256].modulo_value=8; remainders_neg[256].bit_mask=create_bit_mask(1,0x22);
remainders_pos[257].modulo_value=257; remainders_pos[257].bit_mask=create_bit_mask(9,0xe6e7ab16,0x7e16541d,0x93128359,0x1f76811c,0xe205bbe3,0x6b052324,0xe0a9a1fa,0xa3579d9e,0x1);
remainders_neg[257].modulo_value=1028; remainders_neg[257].bit_mask=create_bit_mask(32,0x2a2a228a,0xa28a8880,0xa8a0a80,0x82aa0880,0x2a88222a,0xa288aaa8,0x28202822,0x2a8a0002,0x28aa0,0x20a020a2,0xaaa88a2a,0xa2208aa0,0x882aa0a,0xa828a80,0x88a8a28,0x8a22a2a0,0x2022082,0x208088aa,0x808280a8,0xa00288aa,0x8a2200,0x2088000a,0xa2a0a20,0x82aaa2,0xaaa08022,0x282a2822,0x88202,0x22288028,0x88a00280,0x80a080aa,0x8880820a,0x8220202a);
remainders_pos[258].modulo_value=1032; remainders_pos[258].bit_mask=create_bit_mask(33,0x82080802,0x28220028,0xa288082,0x828a08a2,0x20020200,0x8000220a,0x200a2822,0x8000a00,0x8028a082,0x20000822,0x20228a00,0x8220a200,0x800082,0x8800020,0x8202002,0xa0882080,0x8220a00,0x8082002,0x22080,0x20008,0x8a088282,0xa2880800,0x20000800,0xa280288,0xa0002082,0x28a00800,0x88000288,0x808008a0,0x20a28200,0x228a08a,0x882882,0x20208228,0x80);
remainders_neg[258].modulo_value=1032; remainders_neg[258].bit_mask=create_bit_mask(33,0x2800082,0x20a28020,0x8220000a,0x2020022,0x28828a08,0x808a282,0xa00220a2,0x808208,0x820200a,0xa08800a2,0x28a20208,0xa282288,0x80080802,0x8828,0x8028a08a,0x20002800,0xa28208,0x8000208a,0x808a2800,0x8828800,0x200020a,0x22000080,0x20808008,0x82208200,0x20882802,0x20208008,0x88200,0x80020,0x28220a08,0x8a202002,0x80002002,0x28a00a20,0x8);
remainders_pos[259].modulo_value=1036; remainders_pos[259].bit_mask=create_bit_mask(33,0xa82a22a,0x2028a880,0x22a8822,0x22280002,0x82202a82,0x80880208,0xa0828028,0x820aaa02,0x8a02,0xa20a080a,0x8a220820,0xa08a02,0x8a200888,0xa000828a,0x2000a808,0x80802a0,0x2020000,0xa00080a8,0x2000a202,0x8a2a28,0x20a00222,0x88a280a,0xa08a082,0x20000a02,0xaa08280a,0x2820a80a,0x2202280,0x80882208,0x288882a,0x2a880800,0xa2808882,0xa82a0022,0xa88);
remainders_neg[259].modulo_value=1036; remainders_neg[259].bit_mask=create_bit_mask(33,0x820a2a22,0x28a02000,0x82a200aa,0xaa20088a,0xaa0a20a,0x8000a80,0x280a00a0,0xa82220a,0x8880282,0x22828082,0x22a80a8,0x8820028a,0x2280000,0x28880202,0xa8882000,0x808a20,0x828a8888,0x28080020,0xa8802a8a,0x880222a0,0xa82882aa,0x2a002,0x2802802,0xa088828a,0x2200a082,0xa0a02082,0x8aa02a08,0x800a280,0x8a000022,0x22008080,0x2208000a,0x2022882a,0x200);
remainders_pos[260].modulo_value=520; remainders_pos[260].bit_mask=create_bit_mask(17,0x20000282,0xa20a8022,0x200a8228,0xa000022,0x80000a0a,0x882a0088,0x802a08a2,0x28000088,0x2828,0x20a80222,0xa8228a,0xa0000222,0xa0a0,0x82a00888,0x2a08a28,0x80000888,0x82);
remainders_neg[260].modulo_value=520; remainders_neg[260].bit_mask=create_bit_mask(17,0xa8880a0a,0x2a020822,0x208202a0,0x828088aa,0x880202,0xa20800,0x8228022,0x20088000,0x8880a0a0,0xa020822a,0x8202a02,0x28088aa2,0x8802028,0xa208000,0x82280220,0x880000,0x2);
remainders_pos[261].modulo_value=348; remainders_pos[261].bit_mask=create_bit_mask(11,0x28020a2,0x8220008,0x808008a,0x22082880,0xa00800,0x828a282,0x800200a0,0x228208,0xa200202,0x2000882,0x8a08028);
remainders_neg[261].modulo_value=348; remainders_neg[261].bit_mask=create_bit_mask(11,0x82082822,0xa28800,0x80008002,0xa2802000,0x8208008,0x8020220a,0x8a0020,0x8a20a00,0x8228828a,0x2880002,0x200820);
remainders_pos[262].modulo_value=1048; remainders_pos[262].bit_mask=create_bit_mask(33,0xaa800a0a,0x8028a2a,0xaaa288a2,0x8a828a88,0x20282022,0x2a2200,0xa02880aa,0xa0882222,0x288a0a20,0x882a0222,0x202aa00a,0x282a2aa2,0x8280802a,0x82000880,0x82a8a208,0xaa800020,0xaa0a00a0,0x82080002,0x82208a2a,0x82022000,0x28a80202,0x88aa8a8,0x22a00aa8,0x288880a8,0xa08a0a2,0xa888822,0xaa0228,0x80088a8,0xa2880828,0xaa22a282,0x208a228a,0xaaa8a280,0xa0a002);
remainders_neg[262].modulo_value=1048; remainders_neg[262].bit_mask=create_bit_mask(33,0x22088282,0x808a02a2,0x222a002a,0x20a0200,0xa8a0a8a2,0x88a2aa88,0x28a00822,0x2800aaaa,0xa00282a8,0xa28aaa,0xa8a22882,0xa0a2a22a,0xa0808a2,0xa888008,0xa202a80,0x220888a8,0x22828828,0xa80888a,0xaa802a2,0xa8aa888,0xa0200a8a,0x80022020,0xaa288220,0xa0000820,0x8280282a,0x820000aa,0x88228aa0,0x80880020,0x2a0080a0,0x22aa2a0a,0xa802aa02,0x22202a08,0x28288a);
remainders_pos[263].modulo_value=1052; remainders_pos[263].bit_mask=create_bit_mask(33,0x20aa282,0xa8828022,0x2a8280a8,0x8aa82a00,0xaa220aaa,0x8020aa88,0x28882088,0xaa80828a,0x2a002208,0xa220820a,0x28aa2228,0x88802200,0xa0020002,0xa00aa80,0xa20202a,0x28a88aa,0x20a0a08,0xaa22a28,0xa80808a,0xa02aa00a,0x28000800,0x80088022,0xa2888aa2,0x8a082088,0xa208800a,0x8a28202a,0x22208222,0xa22aa080,0x2aaa088a,0x800a82aa,0xa2a0282a,0x8802822,0x828aa08);
remainders_neg[263].modulo_value=1052; remainders_neg[263].bit_mask=create_bit_mask(33,0x8a822a0a,0x200a08aa,0xa20a0820,0x220a288,0x22aa8222,0x8a82200,0xa000a800,0x22080a02,0xa288aa00,0x2aa80a82,0xa022aaa0,0x8aa88,0x288a888a,0x82882208,0x82a8a8a2,0x8a020022,0x8a828280,0x822aa2a0,0x82080802,0x28a22882,0xa0888088,0x88008aa,0x2a00022a,0x280a800,0x2a800882,0x2a0a8a2,0xaaa80aaa,0x2aa22808,0xa2228002,0x8820a22,0x2a28a0a2,0x8008a0aa,0xa02280);
remainders_pos[264].modulo_value=264; remainders_pos[264].bit_mask=create_bit_mask(9,0x820a2022,0x28220a00,0x80200002,0x20002082,0x8000800,0x80282,0xa0882880,0x8a08200,0x88);
remainders_neg[264].modulo_value=264; remainders_neg[264].bit_mask=create_bit_mask(8,0x28220a2,0x20228208,0x828808a,0xa0882802,0x800008,0x8000820a,0x20002000,0x200a08);
remainders_pos[265].modulo_value=265; remainders_pos[265].bit_mask=create_bit_mask(9,0x29c51b5e,0xc435013,0x1b8a032d,0x82d0108d,0x202d0679,0x14762c4,0x2b08c2d3,0x628e5320,0x1eb);
remainders_neg[265].modulo_value=1060; remainders_neg[265].bit_mask=create_bit_mask(34,0xa0080282,0x800a8882,0x20282a0,0xa008800,0x20a08a20,0x888a2a08,0xa2000a02,0x2a0282a8,0x2a22022,0x228aa0a,0x888a0280,0x282082a2,0x2802288,0x2000088,0x80082a0a,0x802a088a,0xa0220a00,0x8802288,0x8020a800,0x82a8a220,0x802a0288,0x2080820,0x2808080,0x800020aa,0x22200220,0x822002,0xa02a8200,0x20008080,0x820808,0xa088a02a,0x82228a,0x88000a82,0x288a200,0x2);
remainders_pos[266].modulo_value=1064; remainders_pos[266].bit_mask=create_bit_mask(34,0xa2800222,0x202822a2,0x20a0008,0x220828a2,0x2802aa0a,0x8a0a02a0,0xa828a008,0x8802a222,0x88002a2,0x20020202,0xa008a800,0xa8220080,0x82202000,0xa0820028,0x2208aa20,0x2000220,0x808202a8,0x8000802a,0xaa208808,0x820a08,0x8088228,0x882a00,0x2a200a02,0x80800800,0x80022080,0x8a80228a,0xa282a88,0x80a0a220,0xaa80280a,0x282088a0,0xa0808a,0x88280820,0x80028a8a,0x88);
remainders_neg[266].modulo_value=1064; remainders_neg[266].bit_mask=create_bit_mask(33,0x2a008a2a,0x28a0aa22,0x2828880,0xaa00a02a,0xa0822282,0x2820228,0x20a02080,0xa2a2a,0x88a22,0xa08a8a8a,0x28002080,0x20a28808,0xaa82888,0x280a80a0,0xa2802228,0xa888aa0,0xa8a20,0x8080822,0x22a00080,0x88020280,0x808002a0,0x800a208,0x22a88282,0x88088,0x80aa008,0x200aa02,0x82202200,0x82822a8,0x2208a002,0x20a80020,0x80280802,0x2000a8,0x8820202);
remainders_pos[267].modulo_value=1068; remainders_pos[267].bit_mask=create_bit_mask(34,0xa2080882,0x28a00,0x202282,0x820a28a2,0x20a00228,0x2088288,0x220a0820,0x8000008,0x82280280,0x808228a2,0x8a08200,0x8200202,0x820008a0,0x8800000,0x8a080088,0x800a0002,0x8020220,0xa002088,0x20a2800,0x220220,0x82000,0x8200a2,0x20a20808,0x88202008,0x28828a2,0x20028,0xa088200,0x22080082,0xa08228,0x8a082288,0x808008a2,0x28000828,0x208a00a,0x822);
remainders_neg[267].modulo_value=1068; remainders_neg[267].bit_mask=create_bit_mask(34,0x22800002,0x8820208,0x8828a20a,0x2822022,0x28208a20,0x8a000200,0xa28200a0,0x808800,0xa208208,0xa2022,0x200a08,0x8028828a,0x2880020,0x8808,0x2008000,0x820882,0x828a28,0x8208a000,0x82822080,0x8a20a28,0x8800a088,0x800a0822,0x28228000,0x28a080,0x82002022,0x8828820,0x82000288,0xa2800802,0x8200a20,0x200a200,0x22,0x20808020,0x8a002082,0xa2);
remainders_pos[268].modulo_value=536; remainders_pos[268].bit_mask=create_bit_mask(17,0xaa220a8a,0x800a0822,0xa282aa02,0x2888008,0x22280802,0xa828a200,0xa00280a,0x88aa882,0x20a8aa2a,0xa0822aa2,0x2aa02800,0x88008a28,0x80802028,0x8a200222,0x280aa82,0xaa8820a0,0xa2a088);
remainders_neg[268].modulo_value=536; remainders_neg[268].bit_mask=create_bit_mask(17,0x22aa8202,0x88280aa,0x2a0a2282,0x8a000880,0xaaa0808a,0x20a02a88,0x8288a082,0x8002200a,0xa82022a2,0x280aa22a,0xa2282088,0x8802a0,0x808a8a0,0x2a88aaa,0x8a08220a,0x2200a828,0x2a2800);
remainders_pos[269].modulo_value=269; remainders_pos[269].bit_mask=create_bit_mask(9,0x43b16a72,0x67faba74,0x3293c24f,0xe3e8029b,0x5f1c528,0xf2533650,0x57f9bc90,0xa3708b97,0x1395);
remainders_neg[269].modulo_value=1076; remainders_neg[269].bit_mask=create_bit_mask(34,0x8a28a2aa,0xaa2222a8,0xaa0a0a82,0x2a208a02,0x882808a0,0x2a8aaa88,0x8a202008,0x2aa8020a,0xa0828a08,0x8a8200aa,0xa0802022,0x888aaa8a,0xa02880a0,0xaa028822,0xaa0a8282,0x88aa2222,0x2aa28a2,0x8820820,0xa0022220,0xa0028280,0x2a022882,0x888a0a88,0x8a008000,0x80822a2a,0x8a000aa2,0x28a002,0x2080a2a8,0x8028aa2a,0x28888000,0x202a08a8,0xa002a08a,0x200280a0,0x20880222,0x208);
remainders_pos[270].modulo_value=120; remainders_pos[270].bit_mask=create_bit_mask(4,0x200a2082,0x20020,0x8080080,0x8208a0);
remainders_neg[270].modulo_value=120; remainders_neg[270].bit_mask=create_bit_mask(4,0xa0822802,0x8028820,0x8008,0x20020);
remainders_pos[271].modulo_value=1084; remainders_pos[271].bit_mask=create_bit_mask(34,0xa8a822a,0x2a2aaa20,0x8a2228a0,0xaa088800,0x8208a08a,0x82820a00,0x2a0a880,0x882a8aaa,0x2a020a80,0x800200a2,0x2a20280a,0x2a828a0a,0x228228a,0x2aa22280,0xa8208882,0xa8800080,0x8a0202,0xa8080a20,0xa0200022,0x88222082,0x2888aa,0x8a288288,0x8a0a282a,0x2a02808a,0x88a00800,0x202a080a,0xaaa2a82,0x2022a0a8,0x200a0828,0xaa20a208,0x2002220a,0x80a2888a,0x8aaa8a,0xa882a2a);
remainders_neg[271].modulo_value=1084; remainders_neg[271].bit_mask=create_bit_mask(34,0x82020aa2,0xa2a222a8,0x2aaa028,0x22800088,0xa802802,0xa0a8288,0x8a282008,0xa20222,0xa28a0208,0x88a882a,0xa2a8a082,0xa20a0282,0x8aa0aa02,0xa22aaa08,0x20a8000a,0x20088808,0x88028a8a,0x208082a8,0x28a888aa,0xaaa80a,0x88a00022,0x2a00a00,0x282a0a2,0xa28a0802,0x288088,0xa8a28082,0x8222a20a,0xa8aa2820,0xa88280a0,0x22a82a80,0xa88aaa82,0x82a0002,0x88022202,0x200a2a2);
remainders_pos[272].modulo_value=136; remainders_pos[272].bit_mask=create_bit_mask(5,0x228a202,0x8a2880a,0x228a2028,0x8a2880a0,0x80);
remainders_neg[272].modulo_value=136; remainders_neg[272].bit_mask=create_bit_mask(5,0x8aa02a8a,0x80220082,0xaa02a8a0,0x2200828,0x8);
remainders_pos[273].modulo_value=1092; remainders_pos[273].bit_mask=create_bit_mask(35,0x820a0802,0x800,0x8280,0x82080022,0x220208,0xa002000,0x800a2020,0x8028820,0x228a202,0x208a0082,0x20228008,0x200280,0x82022000,0x20008208,0xa080002,0x800000,0x828208,0x828200a,0x2002,0x20a00,0x8200088,0x880822,0x28008000,0x288080,0x200a2082,0x8a28808,0x82280208,0x808a0020,0x800a00,0x8088000,0x80020822,0x28200008,0x2000000,0x20a0820,0x8);
remainders_neg[273].modulo_value=1092; remainders_neg[273].bit_mask=create_bit_mask(34,0x2820002,0x8808000,0x80080208,0x20008a2,0x8a20a00,0x82082088,0x8220a0,0x20028,0xa202282,0xa0020002,0x28220800,0x88208208,0x2822880,0x28800200,0x8200800a,0x80880,0x20a00,0x80202082,0x80802882,0x8820200,0x80288000,0x80000022,0x20800000,0x200008,0xa0022802,0x220000,0xa200280,0x20020,0x8208,0x80,0x8a00a2,0x20208800,0x8a008080,0x820200a0);
remainders_pos[274].modulo_value=1096; remainders_pos[274].bit_mask=create_bit_mask(35,0x2a22a2aa,0x802a2888,0x8aa0a0a,0x22a00280,0xa2aa008a,0x20088288,0x28820a0a,0xaa28002a,0x22222a2,0x2aa82002,0xa0a28820,0x8288aa20,0x82a00208,0x82aa0220,0xa0a008aa,0x88202a80,0x2022028,0x8808000,0xa8082228,0x200a0a02,0x80aa82aa,0x800a8208,0xaa228220,0x228a0a08,0x82aa808,0x88888080,0x28aa8a,0xa08228a8,0x822008a0,0xaa8a22,0x800a88a2,0xa0aa2002,0x28a802a0,0x8a88a822,0xaa);
remainders_neg[274].modulo_value=1096; remainders_neg[274].bit_mask=create_bit_mask(35,0xa2aa2a22,0x8a2a000,0x80228282,0xaa288a08,0x2a228802,0xa8800a00,0xa00a8282,0x22a088a2,0x8aaaaa2a,0xa220a88a,0x282a00a8,0xa0022a8,0x2288a80,0xa228aa8,0x28288022,0xa8a208,0x8a8aa8a0,0x80080888,0x2080aaa0,0xa882828a,0x8220a22,0x8820a80,0x22aa0aa8,0xaa028280,0x80a22080,0x808,0x88a02202,0x280aa020,0xaa88028,0x82202aa,0x882002a,0x2822a88a,0xa0208a28,0x20020aa,0x22);
remainders_pos[275].modulo_value=220; remainders_pos[275].bit_mask=create_bit_mask(7,0x80282,0x802a08a0,0x228a8020,0x88020802,0x20802a28,0xa20a80,0x8280200);
remainders_neg[275].modulo_value=220; remainders_neg[275].bit_mask=create_bit_mask(7,0x8880020a,0x8228020,0x2a0200a8,0x828082,0xa0082220,0x802a0208,0x208a80);
remainders_pos[276].modulo_value=552; remainders_pos[276].bit_mask=create_bit_mask(18,0x82022822,0xa20000,0x2280202,0xa2020800,0x8a08a20,0x88000208,0xa0080802,0x8200008,0x82288288,0x82022,0x20200a20,0x80002280,0xa20a2020,0x20808a08,0x80288000,0x8a0080,0x28808200,0x88);
remainders_neg[276].modulo_value=552; remainders_neg[276].bit_mask=create_bit_mask(17,0x20a20a2,0x8228808,0x8a20828a,0x22820080,0x200228,0x88280,0x20800082,0xa08800,0xa200200,0x808020a2,0x28a08228,0x808a208,0x22822820,0x28000200,0x8200088,0x80020800,0x20000a08);
remainders_pos[277].modulo_value=277; remainders_pos[277].bit_mask=create_bit_mask(9,0x7ae9369a,0xca938394,0x1ffa9ced,0xf331450,0x3c0d212c,0xfe028a33,0x54edce57,0xd78a7072,0x165b25);
remainders_neg[277].modulo_value=1108; remainders_neg[277].bit_mask=create_bit_mask(35,0xa220aa02,0x20a0a08,0x82220020,0x82aa8888,0xa080a8a0,0x228a02aa,0x8820028a,0xa02a8aa,0xa8a82a8,0xaa0280aa,0x208aa8,0x28a228a,0xa8082aaa,0x8aaa0828,0x2220888,0x82820020,0xa8222882,0x2802a202,0x82828822,0x22aa2aa2,0x8888a2,0xa80828a0,0x82a00028,0x2aa08220,0xa008008a,0xa00a82,0xa0a80280,0x28800802,0x822082aa,0x8a0002a0,0x28a080a,0x22888880,0xa2aa2aa2,0x2208a0a0,0x2a00a);
remainders_pos[278].modulo_value=1112; remainders_pos[278].bit_mask=create_bit_mask(35,0x2a08a02,0x222a0288,0xaaaa000a,0x282a8a8,0x28a222aa,0x22020228,0xaaa088a2,0x288028a2,0x828a00a,0x2a8820aa,0x8a000208,0xa2a22208,0x22082822,0x282a000,0xa0000808,0xa02220aa,0x8a02a882,0xa20a82a0,0xa822a80,0xaaa0888,0x80202000,0x88000a82,0x8a882820,0xa220888a,0xa8208000,0x20aa0822,0x28a00a28,0xaa8a2802,0x888a220a,0x28288080,0x80aa888a,0xaa2a2a82,0x88a000aa,0x802280a8,0x80a20a);
remainders_neg[278].modulo_value=1112; remainders_neg[278].bit_mask=create_bit_mask(35,0x8a28028a,0xaaa28a00,0x22228882,0x8a0a2020,0xa02aa222,0xaa8a8aa0,0x2228002a,0xa008a02a,0x80a02882,0xa200a822,0x2888a80,0x2a2aaa80,0xaa80a0aa,0x8a0a2888,0x28888080,0x28aaa822,0x28a200a,0x2a820a28,0x820aa208,0x82228000,0x8a8a888,0x8820a,0x200a0a8,0x2aa80002,0x20a80888,0xa82280aa,0xa02882a0,0x2202a08a,0x2aa82,0xa0a00808,0x8220002,0x22a2a20a,0x288822,0x8aa0820,0x82a82);
remainders_pos[279].modulo_value=372; remainders_pos[279].bit_mask=create_bit_mask(12,0x2800822,0x820a00,0x8088000,0xa28a2022,0x20a00820,0x82280,0xa0802882,0x28a08200,0x208808a,0x20000022,0x28000a08,0x88200);
remainders_neg[279].modulo_value=372; remainders_neg[279].bit_mask=create_bit_mask(12,0x20800a2,0x8028208,0x80000088,0x220228a2,0x20208028,0x8800a208,0x20082002,0x20200a08,0x8a000002,0xa08808a2,0x20808200,0x288);
remainders_pos[280].modulo_value=280; remainders_pos[280].bit_mask=create_bit_mask(9,0x88820a0a,0x20280022,0x200a0220,0x8200802a,0x82000000,0x8a80200,0x80880a0,0x22880028,0xa0a082);
remainders_neg[280].modulo_value=280; remainders_neg[280].bit_mask=create_bit_mask(9,0xa0202,0x282088a2,0x208282a8,0xa0008a2,0x2808808,0x80200a88,0x80000020,0x2a008020,0x202802);
remainders_pos[281].modulo_value=281; remainders_pos[281].bit_mask=create_bit_mask(9,0xb21747b6,0xcf26299f,0x463c175,0x7ad5a434,0x7a23b711,0x80b096ad,0xceba0f18,0x37e65193,0x1b78ba1);
remainders_neg[281].modulo_value=1124; remainders_neg[281].bit_mask=create_bit_mask(35,0x2a8a8a2a,0x2aaa002,0x88aa08a8,0xa20828a8,0xa2aa2a88,0x8280a20,0x2228280,0xaa2a880a,0xa02a0228,0xa2a8a202,0x2202808a,0xa0800a0a,0xaa282280,0x82288aa2,0xa888a8a0,0xaa00a882,0x8aa2002a,0x8202a28a,0x2aa20080,0x880aa000,0x8088802,0x88228a,0x82282002,0xa0aa8a0a,0x8a82a220,0xa2080200,0x22a02202,0xa8802008,0x828222a0,0x20a828aa,0x8020020a,0x828a208,0x8a00888,0xaa0002a8,0x20808022);
remainders_pos[282].modulo_value=1128; remainders_pos[282].bit_mask=create_bit_mask(36,0x2882082,0x8a20a00,0x8288008,0x20220a2,0x8220a28,0xa20a28a,0x80020,0x8820220,0x8200a002,0x80822802,0x28a20020,0x282288,0x20800002,0x20a00020,0x8200202,0x882880,0xa28000,0x82008208,0x28a0020,0x28220000,0x80082002,0xa0880,0x20808,0x88280080,0x8a2822,0x28820208,0xa008280,0x80822080,0x200008,0x8a08a008,0xa08820a2,0x8808028,0x228208a,0xa08a2020,0x8228000,0x82);
remainders_neg[282].modulo_value=1128; remainders_neg[282].bit_mask=create_bit_mask(36,0x82002802,0x220208,0x80200080,0x828a2822,0xa28220,0x82282202,0x808008a0,0x28228,0xa08208a,0xa2082,0x20228828,0x8820a200,0xa0080882,0x28208828,0x8028828a,0x80002000,0x8220808,0xa080280,0x820208a0,0x20a28800,0x800a08a,0x80820000,0x8828000,0x208008,0x220a2,0x20028a00,0x82080208,0xa2800,0x8a08800,0x2002080,0x20002822,0x820,0x8a20a002,0x200228a0,0xa20808,0xa);
remainders_pos[283].modulo_value=1132; remainders_pos[283].bit_mask=create_bit_mask(36,0xaa08220a,0x22828a88,0x22a2aa08,0xaaa0a22,0x20082a0a,0x8aa00aa2,0x88000200,0xa088a282,0x2802a2a,0x280080aa,0xa20a0822,0x22a8aaa,0xa88802a8,0x28280a2,0x8888280,0x8a28008,0x8022020a,0x280288a2,0x88028a2,0x28a20a08,0x22220200,0x28280028,0x222a8a0,0x2a8802a8,0xa08aaaa,0x20028882,0x80280aa0,0xa220aa8a,0x22828,0xaa2008,0x820088aa,0xaaa0a0a,0xa8a8888a,0x2828820a,0x820aa22a,0xa08);
remainders_neg[283].modulo_value=1132; remainders_neg[283].bit_mask=create_bit_mask(36,0x2280aa82,0xaa0a0200,0xaa2a2280,0x822282aa,0xa880a282,0x228822a,0x888a88,0x28002a0a,0x8208a2a2,0xa0880822,0x2a8280aa,0x8aa20222,0x20008a20,0x8a0a082a,0x80000a08,0x802a0880,0x8aa8a82,0xa08a002a,0x8008a02a,0xa02a8280,0xaaaa8a88,0xa0a088a0,0x8aaa2028,0xa2008a20,0x82802222,0xa88a000a,0x8a08228,0x2aa82202,0x888aa0a0,0x8822a880,0xa880022,0x82228282,0x20200002,0xa0a00a82,0xa822aa2,0x280);
remainders_pos[284].modulo_value=568; remainders_pos[284].bit_mask=create_bit_mask(18,0xa2800aa2,0x8a8aa0a8,0x2022208,0xaa082028,0xa8020802,0x280820a,0x22088808,0xa2a0aa2a,0x28aa0028,0x8a2800aa,0x88a8aa0a,0x80202220,0x2aa08202,0xaa802080,0x80280820,0xa2208880,0x8a2a0aa2,0x8aa002);
remainders_neg[284].modulo_value=568; remainders_neg[284].bit_mask=create_bit_mask(18,0x2a08822a,0x2022820,0x8a8aaa00,0x2280a8a0,0x208a808a,0x8a080a82,0xaa800080,0x2a2822a2,0xa02288a0,0x2a08822,0x202282,0x8a8aaa0,0xa2280a8a,0x2208a808,0x8a080a8,0x2aa80008,0x2a2822a,0x2288a);
remainders_pos[285].modulo_value=1140; remainders_pos[285].bit_mask=create_bit_mask(36,0x20822002,0x28028220,0x2082088,0x82000082,0xa00,0x80200,0xa20000a2,0x220,0x200a080,0x82802,0x28220000,0x2280002,0x20080,0x8028000,0x2080002,0xa0882082,0x20220a00,0xa002208,0x80820880,0x20a00a08,0x8082082,0x2080002,0x28,0x88002008,0x82880002,0x8,0x8080282,0x20a0,0x8a08800,0x8a000,0x802,0x8200a00,0x8082000,0x2822082,0x20808828,0x80088);
remainders_neg[285].modulo_value=1140; remainders_neg[285].bit_mask=create_bit_mask(35,0xa0022882,0x20020a20,0xa00a000,0x2800802,0x808208,0x88000288,0x22080822,0x8008228,0x8a002008,0x80802002,0x20a20800,0x8a208082,0x8a0000,0x820808,0x8200800a,0x20002802,0x28228200,0x208a200,0x20000,0x20208200,0x8000200a,0x2000882,0x8808020,0xa080,0x2000002,0x8800800,0x80008202,0x882020,0x200008,0x80002000,0x80880082,0x200200,0xa088,0x82022802,0x20000020);
remainders_pos[286].modulo_value=1144; remainders_pos[286].bit_mask=create_bit_mask(36,0xaa08822a,0x2202a220,0xa08a8a80,0xa8828020,0x80a2080,0x2882802,0x88202002,0xa2a282a,0x22a000a2,0x220aa002,0x82000,0xa0aa2228,0xa2202880,0xa8a00a8,0x82a8200,0x22800088,0x2020282,0x2008a0,0x800a2008,0x88828080,0x20220002,0xa00082a8,0x8a2a00a2,0xa022808,0x2888aa,0x88000820,0x88800aa0,0xa08a000a,0x22a828a8,0x80800808,0x20802822,0x2a0208a0,0xa080282,0x8802a2a2,0xaa088a80,0xa88220);
remainders_neg[286].modulo_value=1144; remainders_neg[286].bit_mask=create_bit_mask(36,0x228002a2,0xaa0a2a28,0x20020208,0x200a08a0,0x80822808,0x8200a08a,0xa8a882,0x8222a022,0xaa20882a,0x2a82208a,0x880a888,0x2822a2a0,0x2a20a008,0x82028820,0x80a20a00,0xa2080800,0x828a8a0a,0x88a80028,0x882a800,0x20800,0x282a888a,0x28880a20,0x2a2802a,0x20aa080,0x88a00022,0x8080a0,0x80228,0x20028802,0xa220a020,0x8088000,0xa80020aa,0x228a8028,0x82808a02,0xa222a,0x22000208,0x2002a8);
remainders_pos[287].modulo_value=1148; remainders_pos[287].bit_mask=create_bit_mask(36,0x8202aa02,0xa000020,0x828a80a8,0x22a00aa,0xa282a88a,0x88a80220,0xa8880028,0x88008822,0x8a88282,0x20a0022,0x82228a2,0x20a0a200,0x88a2020,0x80200208,0xa8002,0x20280202,0x208a8a82,0xa00088a,0x8a22000a,0x882a2a20,0x8080280,0x28002a00,0x2080080,0x80808a22,0x8a0a0,0x8a28882,0x8800a08,0x282822a2,0xa8822002,0x22800222,0xa08802a2,0xa22a828,0x2aa00a88,0x2a02a28,0x2080000a,0x80aa808);
remainders_neg[287].modulo_value=1148; remainders_neg[287].bit_mask=create_bit_mask(36,0xa8a220a,0x28888a0,0x2020820,0x82228822,0x2a022002,0x200aa8,0x200080a0,0x88002a,0x200a02,0x828288aa,0x802aa02a,0xa8202a88,0x800228a8,0x8a88280,0x88820802,0x28a08a82,0xa002020a,0x82088002,0x2a28882,0xa22228,0x80808208,0xa088a208,0xa808800,0x802aa,0x88002028,0x8022000a,0x80080280,0xa0a0a22a,0x200aa80a,0x2a080aa2,0x20008a2a,0x822a20a0,0xa2208200,0x8a2822a0,0xa8008082,0x822000);
remainders_pos[288].modulo_value=24; remainders_pos[288].bit_mask=create_bit_mask(1,0x820082);
remainders_neg[288].modulo_value=24; remainders_neg[288].bit_mask=create_bit_mask(1,0xa0802);
remainders_pos[289].modulo_value=17; remainders_pos[289].bit_mask=create_bit_mask(1,0x1fffe);
remainders_neg[289].modulo_value=68; remainders_neg[289].bit_mask=create_bit_mask(3,0x22202222,0x22222222,0x2);
remainders_pos[290].modulo_value=1160; remainders_pos[290].bit_mask=create_bit_mask(37,0x2a2a02,0x20228000,0x200a02a8,0x8a22882a,0x280820a,0x28022a02,0x220a00a2,0x882a0a8,0x8a222888,0x202002,0x8020888,0x8208220,0x2002a020,0x828280a2,0x20802000,0xa00a00,0xaa020a20,0x20002228,0x88000882,0xa080aa28,0xa00a0008,0x8020800,0x2828200,0xa80088a,0x82082008,0x20802008,0x8080022,0x2888a280,0xa822022,0xa0882a,0xa880288a,0x82028080,0x2288a2a0,0x80a008a8,0x288082a,0xa8a80000,0x80);
remainders_neg[290].modulo_value=1160; remainders_neg[290].bit_mask=create_bit_mask(36,0x88a2228a,0xa82a0880,0x28028220,0x2a200a2,0x82080a02,0xa08a228a,0xaa028022,0xa2020,0x2a2a000,0xa8a80a,0x808a0000,0x80280aa0,0x288a20a8,0xa02082a,0xa008a808,0x88280288,0x220a82a0,0x2888a220,0x80800a,0x20082220,0x20820880,0x800a8080,0xa0a0288,0x82008002,0x2802800,0xa8082880,0x800088a2,0x20002208,0x8202a8a2,0x80280022,0x20082002,0xa0a0800,0x2a002228,0x8208020,0x82008022,0x20200088);
remainders_pos[291].modulo_value=1164; remainders_pos[291].bit_mask=create_bit_mask(37,0x220a08a2,0x20828208,0x80202208,0xa2802000,0x220a20,0x8800a280,0x800a2882,0x220a28,0x82088002,0xa2020802,0x808a08,0xa200,0xa2002,0x808a28,0x88280008,0x80882802,0x8020820,0x2008082,0x20080000,0x8020820,0x82202082,0x2822802,0x20200200,0x8a00028a,0xa0000800,0x20200008,0x808a20a,0x22082802,0x8800800,0x8a00228a,0xa0022822,0x8800028,0x8028a08a,0x80802000,0x28208208,0xa088208,0x8a2);
remainders_neg[291].modulo_value=1164; remainders_neg[291].bit_mask=create_bit_mask(37,0xa2820022,0x28020a00,0x828a280,0x22082880,0x8a28228,0x82208,0x822002,0x8a28220,0xa00008a,0x228a0082,0x8000200,0x88082288,0x80822882,0x8000220,0x208080,0x2082,0x828028,0x8a08000a,0xa0800880,0x828028,0xa28a00a,0x820a2002,0x28a08a08,0x2088202,0x20880080,0x28a08800,0x80002282,0xa2802082,0x8008,0x208a202,0x208a20a2,0x8820,0x8202002,0x82080,0x20a00a00,0x82000280,0x22);
remainders_pos[292].modulo_value=584; remainders_pos[292].bit_mask=create_bit_mask(19,0xa88020a,0x22820228,0xa22a8aa,0x8a80a202,0x82a8200,0x8a02a20,0xa2a88a08,0x8808288a,0x8022a02,0xa88020aa,0x28202280,0xa22a8aa2,0xa80a2020,0x82a82008,0x8a02a200,0x2a88a080,0x808288aa,0x8022a028,0xa0);
remainders_neg[292].modulo_value=584; remainders_neg[292].bit_mask=create_bit_mask(19,0x82008a82,0xaa0a8aa0,0x82aa2022,0x2082a8a,0x80a20a88,0x8028a2a8,0x22200280,0x80a002,0x808aa28a,0x2008a822,0xa0a8aa08,0x2aa2022a,0x2082a8a8,0xa20a880,0x28a2a88,0x22002808,0x80a0022,0x8aa28a0,0x28);
remainders_pos[293].modulo_value=293; remainders_pos[293].bit_mask=create_bit_mask(10,0x8763c652,0x3fe249fa,0xcd9e22bb,0x5a082953,0x5373b2b0,0xa5041683,0x511e6cf2,0xe491ff37,0x98f1b857,0x12);
remainders_neg[293].modulo_value=1172; remainders_neg[293].bit_mask=create_bit_mask(37,0xaaa0a8a,0xa22a8022,0x2aa22,0x8280a08a,0x8aaa2a28,0x28888a0a,0x8882a02a,0x280822aa,0x28208a,0x20a0028a,0x2080a288,0x2a088aaa,0x8888a2a0,0xa2aa8a82,0x280a08a2,0xaa000288,0xaa22a22,0x82aa8220,0x280828a,0xaa2280,0xa0022022,0xa82882aa,0x20208a0,0x88828080,0xa02a0208,0x8a200088,0xa28820a,0x2aa002aa,0xa8208a28,0x88800228,0x88202a02,0x80a088,0x82882020,0xaaa08a0a,0x22022002,0xa22a80,0x80a0);
remainders_pos[294].modulo_value=168; remainders_pos[294].bit_mask=create_bit_mask(6,0x22880022,0x208800,0x80000288,0x22080022,0x228800,0x88);
remainders_neg[294].modulo_value=168; remainders_neg[294].bit_mask=create_bit_mask(5,0xa2000822,0x8a00000,0x88200,0xa20008a2,0x8a20008);
remainders_pos[295].modulo_value=1180; remainders_pos[295].bit_mask=create_bit_mask(37,0xa8202a8a,0x800202a2,0x28822200,0x8a22880a,0x20888008,0xa0282288,0x8a080800,0xa0280,0x828200a8,0x8208020a,0x80aa200a,0xa2020a02,0x8822a0,0x220a008,0x80222a,0x88200282,0x8a220a80,0x28800008,0x80000000,0x22000022,0x202a088a,0x8280082,0xa888020,0x200a088,0xa0a88220,0x280a0808,0x2a008aa0,0x2a080208,0x2a00828,0x20280a00,0xa002020a,0x82288280,0x22002220,0x8a02288a,0x20088822,0xa8a80800,0xa2a8082);
remainders_neg[295].modulo_value=1180; remainders_neg[295].bit_mask=create_bit_mask(37,0x20a82202,0xa8a22,0x200aa288,0x2a20082,0xa0000800,0x28a02a00,0x2008080,0x8828208,0xa028820,0x2808a02,0x8222882,0x2a0a8282,0x800a228,0x8aa02880,0x8008aa22,0xa80a0a,0x2228200,0x20088080,0x8808888,0xa288882a,0xa8a20002,0x80208802,0x20000a8,0xa802800,0x20200a28,0xa0820080,0xa2080220,0x22808280,0x8a2080a0,0xa0a08208,0x288a0a82,0xa200200,0x2a88a2a8,0x282a002,0xa080002a,0x20200088,0x2220802);
remainders_pos[296].modulo_value=296; remainders_pos[296].bit_mask=create_bit_mask(10,0x220822a2,0xa80aaa0a,0xa80202a2,0xa2202000,0x80820228,0x8088a28,0x80802a00,0xaaa02a8a,0x882088a0,0x8a);
remainders_neg[296].modulo_value=296; remainders_neg[296].bit_mask=create_bit_mask(10,0xaa80aa2a,0x20822282,0x208a8a2a,0x2aa82888,0x80a8aa0,0x808002a0,0x808a288,0x2228a202,0xa80020,0x2);
remainders_pos[297].modulo_value=132; remainders_pos[297].bit_mask=create_bit_mask(5,0xa2020002,0x20228,0x8808000a,0x808a2,0x8);
remainders_neg[297].modulo_value=132; remainders_neg[297].bit_mask=create_bit_mask(4,0x228a0082,0x8028a20,0x8082,0x80800022);
remainders_pos[298].modulo_value=1192; remainders_pos[298].bit_mask=create_bit_mask(38,0x8a222a8a,0x880a8882,0xa80a2a00,0x82aa20a8,0x220aa80a,0xaa228208,0x2a082020,0x208008a,0x2aaa0280,0x22a202,0xa2aa82a0,0xa0208aa8,0x200a2a28,0x2228a282,0x280a800,0xa8080a20,0xa88aa020,0x20828880,0x2008022,0x22820888,0xaa22a02,0xa0202a08,0x2a028008,0x8a288800,0xa8a00882,0xa2080a28,0x82aa8a2a,0x8a88000a,0x80aaa880,0x208002,0x820a8a2,0x8288aa08,0x2aa08820,0x8aa82a0,0xa8a02a2a,0x22a02200,0xa888a282,0xa2);
remainders_neg[298].modulo_value=1192; remainders_neg[298].bit_mask=create_bit_mask(38,0x2aaa202,0x82000a,0x2082a288,0xa22a820,0xaa822082,0x22aa0a80,0xa280a8a8,0x8a808802,0xa2228a08,0x88aa2a8a,0x2a220a28,0x28a80220,0xa882a2a0,0x2aa02a0a,0x8a082088,0x208082a8,0x200228a8,0xa80a0008,0x8a8808aa,0xaa0a8000,0x822aa28a,0x28a8a280,0xa28a0880,0x2a00088,0x2028800a,0x2a8082a0,0xa2202a2,0x2008882,0x8222008,0x88a8088a,0x80a8202a,0xa002280,0xa22000a8,0x80220a28,0x2028a2a2,0xaa28aa88,0x20002a0a,0x2a);
remainders_pos[299].modulo_value=1196; remainders_pos[299].bit_mask=create_bit_mask(38,0xa2200222,0xa0aa822,0x202a080,0x2a802aa2,0x88a20220,0xaa8020a8,0x88820280,0x8a228a22,0xaaa0a0a2,0x802a2,0x82808,0x20088208,0x220a0a,0x220008aa,0xa88a08,0x828800a0,0x8288280a,0x8082002,0x220888a8,0x820202a2,0x82282800,0x2282a02,0x22a000a0,0x8820a,0x8800aa2,0x22008a0a,0x82000208,0x2000202,0xa0aaa8a8,0x288a28a0,0x822288a,0x802aa028,0x8a222a0,0x802a8088,0xa80808aa,0xaa0a0020,0x88a882,0x888);
remainders_neg[299].modulo_value=1196; remainders_neg[299].bit_mask=create_bit_mask(37,0x2a288aaa,0x8282202a,0x828a2808,0xa200a22a,0x2a0aa8,0x2208a820,0xa0a00,0x22a02aa,0x2228282a,0x8880822a,0x880a080,0xa8800a80,0x88a28202,0xaa888022,0x88200200,0x2008820,0xa00a082,0x80288a,0xaa800020,0xa8a8a22,0x220a088,0x8aa0a28a,0xaa288028,0x8000a82,0x8008822a,0xaa800282,0xa808a80,0x8a888a0a,0x20222020,0xa0022028,0x80aa2002,0x8a228a0,0x802aa220,0x8220800,0x20808022,0x22828028,0x800200a);
remainders_pos[300].modulo_value=120; remainders_pos[300].bit_mask=create_bit_mask(4,0x802802,0x28028020,0x80280,0x802802);
remainders_neg[300].modulo_value=120; remainders_neg[300].bit_mask=create_bit_mask(4,0x80082082,0x20020820,0x8008208,0x2082);
remainders_pos[301].modulo_value=1204; remainders_pos[301].bit_mask=create_bit_mask(38,0xa888a2a,0x22a02002,0x82028a28,0xa2082808,0xa0000088,0x800a20,0xa80282a2,0x200aa00a,0x22080a02,0x22020202,0x22028082,0xaa802a80,0x22a0a00,0x28228008,0x28880000,0x880a082,0x20a28a02,0x8200202a,0xaaa2888a,0x82a2228,0xa08a8080,0x22080a28,0x228820a0,0x82800002,0x88020028,0x2aa00a0a,0x8802a80,0x8882028,0x8880808,0x880a02,0x2aa00aa,0x2008a828,0xa08a00,0x8a22000,0x8220282,0xa8828a28,0x2a080080,0xa8a22);
remainders_neg[301].modulo_value=1204; remainders_neg[301].bit_mask=create_bit_mask(38,0x82000222,0x2a28a082,0x28a02a0,0x2a00a080,0x28808800,0x880802a8,0x200a022a,0xa8822802,0x2a808282,0xa28a8a8a,0xaa0a080a,0x2200a208,0x8aa20280,0xa0aa0080,0xa0008808,0x82802,0xa02a028a,0xa0828a2,0x22220002,0x80a22aa0,0x28020008,0xaa808220,0x2200a820,0x208888a,0xa88a0,0xa2208282,0x80082208,0x8000a0a0,0x80008000,0x8008282,0x82228822,0xa80020a0,0x88200288,0x80222888,0x80aa820a,0x200a0220,0x22808800,0x202aa);
remainders_pos[302].modulo_value=1208; remainders_pos[302].bit_mask=create_bit_mask(38,0x28aaa82,0xa8220800,0x2882a080,0xaa0a22a,0x2a22aaa0,0xa208a22,0x2802a2aa,0xaa82080a,0x20288880,0x22282822,0xa88882a,0x20a8a280,0xa00202a8,0x22208a20,0xaa000200,0x20220a00,0xa8a0288,0xaaa8a228,0x82800082,0xaa820002,0xa0288a2a,0x82280a2,0xaa00a088,0x88008000,0xa08a208,0x82a8080,0xa0028a2a,0x88a82222,0x8882828,0xaa022228,0x28a02082,0xa0aa8a80,0xa888a208,0xa00aaa88,0x28a88a0a,0x2a020a82,0x80002088,0x82aaa2);
remainders_neg[302].modulo_value=1208; remainders_neg[302].bit_mask=create_bit_mask(38,0x8a02220a,0x20aa8088,0xa00a2808,0x82282aa2,0xa22a2228,0x82a802aa,0xa08a2a22,0x220a8082,0xa8a00008,0xaaa0a0aa,0x820000a2,0xa8202a08,0x288a8a20,0xaaa802a8,0x22888a88,0xa8aa8288,0x82028a00,0x22202aa0,0xa08880a,0x220a888a,0x28a002a2,0x80aa082a,0x22882800,0x800888,0x82802a80,0x80a20808,0x288a02a2,0x20aaaa,0x8000a0a0,0x228aaaa0,0xa028a80a,0x28220208,0x20002a80,0x28822200,0xa0200282,0xa28a820a,0x888a800,0xa222a);
remainders_pos[303].modulo_value=1212; remainders_pos[303].bit_mask=create_bit_mask(38,0x22802082,0xa28220,0x8a200088,0x82020882,0x28a00808,0x828a20a,0x820882,0x8008008,0x8000a00a,0x208a2,0xa00a28,0x82282282,0x82020822,0x800020,0x220a000,0x20882,0x20228800,0x8808000a,0x2082800,0x20028208,0x8a000202,0x22880,0x8220800,0xa088,0x20800020,0x28820808,0x8288288,0x28a00a0,0x28a20800,0xa00a000,0x2002002,0x8220820,0x8a08a282,0x220200a2,0x28220808,0x220008a,0x808828a0,0x8208028);
remainders_neg[303].modulo_value=1212; remainders_neg[303].bit_mask=create_bit_mask(38,0xa2082802,0x8220a28,0x2288000,0x28a0002,0x20208000,0x80202282,0x800a0002,0x800800,0x8082082,0x808a0022,0x8208220,0xa20a20a,0x28a00a2,0x8008828,0x8a282088,0x808a0002,0x28a20008,0x8082,0x82802080,0x28820a00,0x208828a,0x808a2000,0xa28000,0x88082000,0xa00808a0,0x20028000,0x80200200,0x82020820,0x20228008,0x82082088,0x82882882,0xa28028,0x200220a,0xa28a0822,0x20228000,0x8a288002,0x2020,0xa00820);
remainders_pos[304].modulo_value=152; remainders_pos[304].bit_mask=create_bit_mask(5,0x8a02822a,0x280a2000,0x2822aa88,0xa20008a0,0xa88280);
remainders_neg[304].modulo_value=152; remainders_neg[304].bit_mask=create_bit_mask(5,0x2820aa2,0xa082a888,0x20aa2200,0x2a888028,0x200a08);
remainders_pos[305].modulo_value=305; remainders_pos[305].bit_mask=create_bit_mask(10,0x108f4396,0x81224af7,0x30c6551d,0x42812404,0x3331a179,0x50a7a16,0x8c308092,0x1206e2a9,0xc423bd49,0x1a70b);
remainders_neg[305].modulo_value=1220; remainders_neg[305].bit_mask=create_bit_mask(38,0x88020a0a,0x82a8222,0x280a8080,0x8a00a888,0xa2a82820,0x8802228a,0x280802,0xa8a2220,0x8a22080,0x2a02a02,0x8222882,0x22228a80,0x80a000,0x8a22008a,0x20a0aa2a,0x88a80288,0x80a80a0,0x220aa080,0x8282008a,0x82a08202,0x8002200,0xa0088a2,0x2880088,0x20828,0x80a02000,0xa80a0a80,0x80020222,0x80220a8,0x802200a2,0x82200822,0x2020008a,0xa8280aa2,0x2028080,0x8200000,0x8008a00a,0x88802808,0x22000822,0x2082a080);
remainders_pos[306].modulo_value=408; remainders_pos[306].bit_mask=create_bit_mask(13,0x22000822,0x20828020,0xa000000,0xa2002880,0x820a28,0x8220220a,0x82080020,0xa08808,0x8a28a082,0xa0022800,0x8000000,0x88080282,0x882000);
remainders_neg[306].modulo_value=408; remainders_neg[306].bit_mask=create_bit_mask(13,0xa28800a2,0x28020828,0x82088088,0x22082000,0x8028220,0x228a282,0x28008a0,0x8200000,0x220200a,0x208a2080,0x808800,0x820a,0x2800);
remainders_pos[307].modulo_value=1228; remainders_pos[307].bit_mask=create_bit_mask(39,0x8a82020a,0x8aa8a28,0xaa02282a,0xa02a222,0x82aaa82a,0x220228,0xa8a08,0x8882a2,0x28a2a882,0xaa02802a,0x20082080,0xa222a8a,0x8202aa88,0x8a88aaa2,0xa0000a20,0x880a8280,0xa08a8008,0x82002280,0x8a8a0220,0xa2a2802,0x80082088,0x2a20a028,0x2a022200,0xa028,0xa22a208a,0xa80828aa,0x888a022a,0x82008a2a,0x280aa020,0xa8a28a80,0x22200822,0x2a0008a8,0x880020a,0xaaa82288,0xa80a0a82,0x880aa888,0x2aa20a82,0x82a228a,0xa08);
remainders_neg[307].modulo_value=1228; remainders_neg[307].bit_mask=create_bit_mask(39,0x20a8a82,0x802202a0,0x228aa0a2,0x828a2aaa,0xa2220a2,0x88aa8aa0,0x88820280,0x88000a2a,0xa02a200a,0x228208a2,0xa880a808,0x82aaa202,0xa8a2200,0x200222a,0x288882a8,0x820a08,0x28020880,0xa88aa08,0x2028aa8,0x82a2a08a,0x880a800,0xa2a828a0,0xa28aaa88,0x888828a0,0x2aa2a802,0x2080a022,0x28aa2,0xa8802a2,0xa08228a8,0x202a0208,0xaaa880aa,0xa2888020,0x80088a82,0x2220aa00,0x2082820a,0x822000,0xa22a820a,0x80a2aa02,0x280);
remainders_pos[308].modulo_value=616; remainders_pos[308].bit_mask=create_bit_mask(20,0x28aa202,0x20200220,0x208a0288,0x220020,0xa022a8a,0x800a2a88,0x20808008,0x8082280a,0x28008800,0x202808aa,0x220028aa,0x28820200,0x20208a0,0xa8a00220,0xa880a022,0x8800a2,0x80a20808,0x80080822,0x8aa28008,0x80);
remainders_neg[308].modulo_value=616; remainders_neg[308].bit_mask=create_bit_mask(20,0x8a02220a,0x28288aa0,0xa0028a00,0x882a88a0,0x82822202,0x822200,0xa8080080,0x80aa002,0x20800080,0x20a08022,0xaa08a022,0xa0028288,0x8a0a0028,0x202882a8,0x20082822,0x8000822,0x2a8080,0x80080aa,0x2220800,0x8);
remainders_pos[309].modulo_value=1236; remainders_pos[309].bit_mask=create_bit_mask(39,0x20828a2,0x20a28008,0x8a00a082,0x22020022,0x8820,0x8000a208,0x80020020,0x8a00a28,0x2080282,0x2020022,0x820220,0x8028820a,0x20008a2,0x28000820,0x88000082,0x2022020,0x28028a20,0xa282208,0xa0820080,0x20a28a28,0x8a002008,0x2820882,0x8008a28,0x2208088,0x2882000,0x8008200,0x8028a200,0x200a0822,0x8008808,0x8088008,0xa2082802,0x228a00,0x208008,0x208a0,0x8808220,0xa288008,0xa08820a0,0x8020028,0x8a282);
remainders_neg[309].modulo_value=1236; remainders_neg[309].bit_mask=create_bit_mask(39,0x82802022,0x28220800,0x208200a,0xa28a0822,0x8800028,0x8082280,0x8a08a0,0x208220,0x8a00820a,0x828a08a2,0x8028a28,0x8200282,0x82880022,0x20808028,0x8800a,0x828a28a0,0x20820220,0x8220a280,0x200a0800,0x28220220,0x208a080,0x820a0002,0x800220,0x8a280000,0x82002880,0x800a08,0x8202288,0xa08200a2,0x800000,0x80000080,0x22802082,0x8a20208,0x88280080,0x808a0020,0xa28,0x82200080,0x20002820,0x828820,0x220a);
remainders_pos[310].modulo_value=1240; remainders_pos[310].bit_mask=create_bit_mask(39,0x282a028a,0xaa0a8a00,0x8a8200,0xa800080,0x2288a22a,0x20a22a00,0x8080880,0x2208a020,0x8200a020,0x22202822,0x8820208,0x8220a22,0x2880a228,0x2022888,0x802a20,0xa8802002,0x800a0a82,0x8002a0,0x802008a0,0x2800002,0xa2008,0x20a8002,0x2a82a0a0,0x800802,0x8008a802,0x28222880,0x20288a02,0x2088a088,0x88208082,0x82882808,0x88080a00,0x20080a20,0x8022020,0x8800a88a,0xa0a88a22,0x20002,0xaa0082a2,0x2800a2a0,0xa280a8);
remainders_neg[310].modulo_value=1240; remainders_neg[310].bit_mask=create_bit_mask(39,0x20a20a02,0x22020280,0x8020288,0x82008808,0xa2002a22,0xa82a2288,0x80008000,0x2a8020a8,0xa0028a8,0xa2a8a02a,0x800a0a80,0x802a82a2,0x200022a0,0x8a82a000,0x8008a228,0x2008288a,0x8020202,0x8088228,0x8a08028,0x8208880a,0x88822080,0x8a020882,0x220a2028,0x8800808a,0x80200a,0xa0aa2008,0xa8200282,0x28002000,0xa00802,0x200a000,0x800288,0x280082a0,0x8aa0a8,0x802002,0x2020022a,0x808a088a,0x22080a22,0x20882228,0x220820);
remainders_pos[311].modulo_value=1244; remainders_pos[311].bit_mask=create_bit_mask(39,0x82a82a22,0x8aa2800,0x82820282,0x2aaaa288,0x2880a208,0xa0802a80,0x8a8aa2,0xaaa0a0aa,0x88aa8288,0x2208888a,0x220222,0x2800220a,0x22a8028,0xaa280802,0x2a20aa00,0xa288,0x8a0a2208,0x80228a0a,0xa00a2aa0,0xa0220880,0x20aa8a00,0x2a0a2880,0x2088a0a,0x8228a000,0xa00aa08a,0x802028a,0x82802a88,0xa088002,0x88880880,0x2a222208,0xa2282aa2,0xaa0a0aa,0xa8aa2a20,0x802a8020,0x8208a022,0x2228aaaa,0x8280828,0x20028aa2,0x88a82a8);
remainders_neg[311].modulo_value=1244; remainders_neg[311].bit_mask=create_bit_mask(39,0xa20a2aa,0x8022a088,0xa0a8a0a,0xa2222a00,0xa0082a80,0x2808a208,0x8802022a,0x22282822,0x220a00,0xaa000002,0x88aa8aaa,0xa088aa82,0x8aa208a0,0x22a0808a,0xa2a82288,0x88882a00,0x282aa80,0x8aa0282,0x2882a228,0x28aa8008,0xa8220288,0xa282a008,0x8a800282,0xaa02888,0x28822802,0x808a8a02,0xa08a200,0x8280088a,0x8008,0xa2aaaa80,0x2aa0a22a,0x82282822,0x2022a2a8,0x8a208a8,0xa8028aa,0xaaa02222,0x80a080a0,0xa88a022a,0x20a20);
remainders_pos[312].modulo_value=312; remainders_pos[312].bit_mask=create_bit_mask(10,0xa2800882,0x8220a20,0x82282000,0x2002020,0x800a00,0x8000a002,0x82080800,0x20000828,0x8a08a088,0x822002);
remainders_neg[312].modulo_value=312; remainders_neg[312].bit_mask=create_bit_mask(10,0x22080002,0xa28228,0x220a088,0x828828a0,0x8000208,0x808208a,0x2800080,0x28008020,0x2002000,0xa2082);
remainders_pos[313].modulo_value=313; remainders_pos[313].bit_mask=create_bit_mask(10,0x2f4d3b5e,0x65710db,0x1aad9c5,0x2be1b9f,0x79c9da11,0x2216e4e,0x3e761f5,0x828e6d56,0xd36c23a9,0x1eb72cb);
remainders_neg[313].modulo_value=1252; remainders_neg[313].bit_mask=create_bit_mask(40,0xa280a282,0x8a8a8802,0x88220008,0x8a228202,0xa0000288,0x8aa8a282,0x8a2aa828,0xa02a08a,0xaa0aa20,0x2022aa02,0x2a8202aa,0x28022a8,0xa28a882a,0xaa88a0aa,0x2a0a28,0x22888a00,0x208a020a,0x8a888002,0xa2a008a,0x20a20a28,0x8aa020a8,0xaa888080,0xa2a08a22,0xa0888220,0x20a02aaa,0xa088008,0x28808200,0x22a82a0,0x2a28028,0xa0022220,0x200a00a2,0x8a02a0aa,0x28002080,0x82080088,0x82aaaa02,0xa2822088,0xaaa22882,0xa8808088,0x820a8202,0x2);
remainders_pos[314].modulo_value=1256; remainders_pos[314].bit_mask=create_bit_mask(40,0xa2220222,0x2a22a882,0xa2a20a0,0xaa028a00,0xa228828,0x202802aa,0x88a028a,0x28a0aa2a,0xa28022aa,0x2aa8aaa,0x82000080,0x820022a,0x88a0200a,0x2a2082a0,0x202002a8,0xa0082882,0x20aaa082,0x202a0a20,0x22028802,0x888222a2,0x2280888a,0xa0a80880,0xaaa0808,0x28200a82,0x80080882,0x8208a82a,0x80a220a,0x800820a0,0x82a8,0xa2aa8002,0x88028aaa,0xaa0a28aa,0x80a220a8,0x802808a2,0x2288a0aa,0xa280aa28,0x8a8a000,0x2a88a80a,0x80888a82,0x88);
remainders_neg[314].modulo_value=1256; remainders_neg[314].bit_mask=create_bit_mask(39,0x2aaa8aaa,0xa2aa200a,0x82a2a828,0x228a0288,0x82aa00a0,0xa8a08a22,0x80028a02,0xa02822a2,0x2a08aa22,0x8a220222,0xa888808,0x80a88aa2,0x28a882,0xa2a80a28,0xa8288a20,0x2880a00a,0xa822280a,0xa8a282a8,0xaa8a008a,0xaaa2a,0xaa080002,0x28208008,0x82228080,0xa0a8820a,0x880800a,0xa8020a2,0x8082aa82,0x880a828,0x88880a20,0x2a22088a,0x8a0222,0x2282a022,0x82aa820,0x8a0802a,0xaa002022,0x2a0822a0,0x80202888,0xa2002082,0x800020a);
remainders_pos[315].modulo_value=420; remainders_pos[315].bit_mask=create_bit_mask(14,0xa08a2002,0x8000800,0x208,0x82002802,0x20200808,0x2208,0x20000080,0x8800000,0x2008082,0x2800822,0x8000008,0x2000202,0x8a20a0,0x8);
remainders_neg[315].modulo_value=420; remainders_neg[315].bit_mask=create_bit_mask(13,0x20022802,0x8000,0x88280,0x2002082,0x20a08000,0x88082280,0xa0080000,0x8008,0xa000002,0x82080022,0x800,0x8a008282,0x22020);
remainders_pos[316].modulo_value=632; remainders_pos[316].bit_mask=create_bit_mask(20,0xa20a2aa,0x8802a888,0xa020a82,0xa2200aa2,0x82882,0xa8228200,0x8aa0088,0x282a080a,0x222a802,0xaaa8a08a,0x80a20a2a,0x28802a88,0x20a020a8,0x2a2200aa,0x8288,0x8a822820,0xa08aa008,0x2282a080,0xa0222a80,0xaa8a08);
remainders_neg[316].modulo_value=632; remainders_neg[316].bit_mask=create_bit_mask(20,0x82a82a22,0x8a2000,0x828a020a,0x2aa8822a,0x8880a00a,0x20aa0a88,0x80228800,0xa0a28082,0x8aaa208a,0x22202802,0x82a82a2,0xa008a200,0xa828a020,0xa2aa8822,0x88880a00,0x20aa0a8,0x28022880,0xaa0a2808,0x28aaa208,0x220280);
remainders_pos[317].modulo_value=317; remainders_pos[317].bit_mask=create_bit_mask(10,0x97818ed2,0xbe6a1df4,0xd6aea24f,0x1a0343b9,0x31736588,0x469b3a3,0xe770b016,0x7c915d5a,0x4bee159f,0x12dc607a);
remainders_neg[317].modulo_value=1268; remainders_neg[317].bit_mask=create_bit_mask(40,0xa08020a,0x22a28028,0xa222a82,0x828a8a20,0xa8aaa800,0x88a02a2a,0x2aa8288a,0xa0088082,0x82228802,0x8a00a8a2,0xa28a802,0x2a008a22,0xa2080880,0x8a88a0aa,0xaaa2a028,0x800aaa8,0x80228a8a,0x220aa222,0x80a00a2a,0x82820082,0xa828aa2,0xa22020aa,0x2a822200,0xa8a08082,0x8000a,0x80882a02,0xa2000a08,0xa02a88a8,0x20a2208a,0xa002a808,0x2282080a,0xaa02a882,0x228a88,0x8808828,0x800202a,0x828aa800,0x20aa2080,0x2228022,0xa0a82a82,0xa2a8);
remainders_pos[318].modulo_value=1272; remainders_pos[318].bit_mask=create_bit_mask(40,0xa2000002,0x20828a08,0x8828a002,0x82882082,0x20a08220,0xa28020a,0x200a0020,0x28820200,0x20200a,0x208808a2,0x20008808,0x20220a,0x22822080,0x28228028,0x20a000,0x82020802,0x820800,0x2280008,0x800002,0xa28a20,0x8a28a,0x80800002,0x200028,0x82002082,0x802080,0x28000a08,0x88280288,0x20882,0x8a08808,0x8202200,0x8a2022,0x28a00808,0x8008082,0xa00800a0,0x8a08028,0x8208820a,0x22820822,0x8800a28,0x8a20a282,0x800000);
remainders_neg[318].modulo_value=1272; remainders_neg[318].bit_mask=create_bit_mask(40,0x22880882,0x28020200,0x20208a,0x2002802,0x28200a28,0x82208282,0xa08208a0,0x20028a08,0x8828a082,0xa0000022,0x28800000,0x8820a282,0xa20a2800,0x20a20820,0x88282088,0x28a0082,0x8028008,0x8a208080,0x80080802,0x8220228,0x88002202,0x80882,0x8a08820,0xa08a00a,0x80082800,0x20808200,0x208200,0x808a0002,0x200000,0x8028a288,0x800228a2,0x20200000,0x8008000a,0x20800820,0x200820,0xa000282,0xa20a00a2,0x8220,0x2282202,0x80880);
remainders_pos[319].modulo_value=1276; remainders_pos[319].bit_mask=create_bit_mask(40,0x8a2282aa,0xa02aa208,0x202a0a00,0x28800a20,0x8a882880,0xa0a08288,0xa002002,0x8aa8880a,0x88800202,0xa828822,0x8820a02,0x88aa222,0x82222000,0xaa08a82,0x200a2820,0xaa200202,0xa082aa00,0x2a808028,0x2880028,0x20880,0x220800,0x82800228,0xa280202a,0xa00aa820,0x8808008a,0x828a00,0x282a20aa,0x8888,0x888aa22,0x80a0822,0x2882282a,0x28080022,0xa0222aa,0xa800800a,0x222820a0,0x8022822a,0x808a0022,0xa00a0a80,0x2208aa80,0xaa8288a);
remainders_neg[319].modulo_value=1276; remainders_neg[319].bit_mask=create_bit_mask(40,0x2aa0222,0x28222a80,0xa8228288,0xa00882a0,0x2002008,0x20280a00,0x8288a08a,0x2200002,0x8a8a,0x20a00aa,0x800a828a,0x800222aa,0xa2aa888,0x8220020a,0xa882a0a0,0x22a80a8a,0x200a2288,0xa2080820,0x8a008820,0x88828008,0xaa8088,0xa088aa0,0x2a08a0a2,0x280220a8,0x800802,0x880a0280,0xa0a22822,0x80880000,0x800022a2,0x8082802a,0xa002a0a2,0x200088aa,0x828aaa22,0x20880082,0xaa20a828,0x8aa02a2,0x80288a2,0x28820208,0x22802208,0x220a002);
remainders_pos[320].modulo_value=40; remainders_pos[320].bit_mask=create_bit_mask(2,0xa0280a02,0x80);
remainders_neg[320].modulo_value=40; remainders_neg[320].bit_mask=create_bit_mask(1,0x28a0028a);
remainders_pos[321].modulo_value=1284; remainders_pos[321].bit_mask=create_bit_mask(41,0x20a2022,0x28020020,0x8020a082,0x22020000,0x800008,0x8208228a,0xa2000882,0x20a20000,0x8a00a002,0xa2822082,0x8220a28,0x28028a,0x22822,0x8800228,0x8a20820a,0x80000802,0x20220,0x8282008,0x200200a2,0x20228200,0x828808a,0xa0080080,0x828208,0x88080002,0x2000020,0x8208a28,0x8800220a,0x82880002,0x28028008,0x8a08820a,0x208828a2,0xa00a28,0x8a088,0x220008a0,0x28820828,0x200a,0x80882,0x20a08020,0x80080288,0x808a0800,0x8);
remainders_neg[321].modulo_value=1284; remainders_neg[321].bit_mask=create_bit_mask(40,0x828228a2,0x20828828,0x828200a,0xa28a0080,0x8008800,0xa00a202,0x22880002,0x28228808,0x208208a,0x220a2802,0xa28220,0x88208202,0x808a20a2,0x8a20,0x2280282,0x880082,0x8028a28,0x8020a080,0xa08a0822,0x28a20a08,0x80200002,0x20800800,0x8020a00,0x808a,0x828808a0,0xa00220,0x8a282,0x2000882,0x20820800,0x2000282,0xa0002022,0x8208220,0x88002000,0xa2880020,0x20028020,0x8808a082,0x80800002,0x28200828,0x8008200,0x20080);
remainders_pos[322].modulo_value=1288; remainders_pos[322].bit_mask=create_bit_mask(41,0xa022a0a,0x82808a2,0x82020888,0x8220a8aa,0x8aa2880a,0x8022200,0x20202282,0xa8a02,0x20082800,0x22a820a,0xa00020a2,0xa820aa00,0x80802a28,0x800a2222,0xa202a822,0x28800082,0x80280,0x822a080,0x8a02888,0x802200a0,0x880200,0x280a200a,0xa882022,0x80200002,0x22802,0x2a808a82,0x88a00288,0xa8020288,0xaa082a28,0x8000a00,0x82a8808a,0x282008a0,0xa2a00000,0x88080880,0x88802082,0x228aa200,0x2a0882a0,0x208082aa,0x20282022,0xa880a08a,0xa0);
remainders_neg[322].modulo_value=1288; remainders_neg[322].bit_mask=create_bit_mask(41,0x820aa202,0xa08022,0x28a8000,0xa202022,0x2220082,0x808a2a88,0xa8a8220a,0x8882020a,0x2880a080,0x82a20282,0x2808a82a,0x20a02288,0x8082220,0x882a2aa,0x2a8a202a,0x20088802,0x80808a08,0x2a2808,0x8020a000,0x8aa0828,0x80008288,0xa082a802,0x200a8a2,0x28888a,0x880aa08a,0xa200020a,0x200a00,0x208a8200,0x2280a220,0x880280,0x2200802,0xa0288028,0x2a208088,0x800008,0x8a00a,0xaa022a08,0x22800a20,0xa0080a22,0xa820a8a2,0x20002802,0x28);
remainders_pos[323].modulo_value=1292; remainders_pos[323].bit_mask=create_bit_mask(41,0x22808a82,0x880202a0,0x2002200a,0x820222aa,0xa8a02a8a,0xa8a22a,0x88a08882,0x22022822,0x880a2a28,0xa08a0000,0xaa822802,0x8082a00a,0xa08a0820,0x22082008,0x8020228a,0xa80000,0x88a0a00,0xa888a800,0xa828202,0x20208a8,0x8080a0a,0x282a02a2,0xa222a808,0xa220002,0x2a0000a,0x80802000,0x82088a28,0xa20a200,0xa8202082,0x882aaa00,0xa20a802,0x8a022000,0x8808828a,0x20a22882,0xa2a00822,0x80a2aa88,0x88082a2a,0x88008aa8,0x8022a00,0x202880a8,0x82a);
remainders_neg[323].modulo_value=1292; remainders_neg[323].bit_mask=create_bit_mask(41,0xaa00020a,0x828a28,0x288aa882,0xa0aaa22,0x2028a202,0x802022a2,0x28000a,0x2a0aa0aa,0x82a2a0,0x28028888,0x220aa082,0x80a2882,0x28020028,0xaa80a880,0xa8a202,0x88208888,0x80020288,0x20002008,0x82020a8a,0x8a828020,0x80008282,0x20a28a2a,0x22aa2080,0x82aa888a,0xa288082,0x808a888,0xa0002a0,0x82a82a80,0x20a8a80a,0xa22200,0x82a8208a,0x28a2088,0x800a02,0xa02a200a,0x2a2880aa,0x8222200,0x80a222,0x80220,0x8082a288,0xa0a00820,0xa2);
remainders_pos[324].modulo_value=24; remainders_pos[324].bit_mask=create_bit_mask(1,0x8a28a2);
remainders_neg[324].modulo_value=24; remainders_neg[324].bit_mask=create_bit_mask(1,0x22022);
remainders_pos[325].modulo_value=65; remainders_pos[325].bit_mask=create_bit_mask(3,0x28c3521a,0x612b0c50,0x1);
remainders_neg[325].modulo_value=260; remainders_neg[325].bit_mask=create_bit_mask(9,0xa00a0a82,0xa8228000,0xa20a8,0xa828028,0x22800022,0x8202888,0x8a020800,0xa208,0x2);
remainders_pos[326].modulo_value=1304; remainders_pos[326].bit_mask=create_bit_mask(41,0xa28022a2,0x820a2a2a,0x80a8002,0xa22a20a2,0xa8222008,0x8aaa8222,0xaa2aa00,0x882080aa,0xa2a0a828,0x2a882a22,0x2880202a,0xa020220,0x8a288828,0x200a00a,0x80008aa0,0x22280022,0x202208aa,0x80a8a20a,0x20a282aa,0x2a820020,0xa8808202,0x8080082,0x2aa828a,0x8a08a2a,0x88aa2088,0x2880028,0x800aa200,0xa2a00a00,0xa0282228,0x28088080,0xa8a80802,0x8a88a822,0x22282a0a,0xa0aa0208,0xa200aa8a,0x2a8882aa,0x8a200888,0x208a08a8,0x828002a0,0x8aa8a8a0,0x8a8802);
remainders_neg[326].modulo_value=1304; remainders_neg[326].bit_mask=create_bit_mask(41,0x2a08aa2a,0xa82a2a2,0x8082088a,0x2aa2a82a,0x20aaa880,0x2220aa2,0x822a2288,0xa80822,0x2a2820a0,0xa200a2aa,0xa008a8a2,0x828a8aa8,0x2a000a0,0x8a882882,0x8880228,0xaaa088aa,0xa8aa8022,0x8202a82,0xa82a0a22,0xa20a88a8,0x20080a8a,0x8080880a,0x8a220a02,0x802802a2,0x22a800,0x8a0008a0,0x8822a88,0x2a288288,0x28a0aaa0,0xa0800808,0x2020808a,0x20020aa,0xaaa0a282,0x28228a80,0x2a882202,0xa2000a22,0x2a88000,0xa8028020,0xa088a28,0x2202028,0x2008a);
remainders_pos[327].modulo_value=1308; remainders_pos[327].bit_mask=create_bit_mask(41,0x20a0002,0x20a20208,0x808a28a,0x820800a2,0x28a28808,0x8200020a,0x2800882,0x20008808,0x800a08a,0xa0802882,0x28200208,0xa208200,0x22880020,0x28028220,0x8a080202,0x82,0x20820220,0x2208008,0x80080000,0x828208,0xa28028a,0x22082820,0x200,0x82002088,0x880820,0x28200000,0x8808020a,0x80882802,0x800228,0x8008208a,0xa2080082,0x8228020,0x8a20a002,0x2022000,0x28220028,0x8a080008,0x220228a2,0x8a00208,0x8a28a202,0x20808a0,0x8000a08);
remainders_neg[327].modulo_value=1308; remainders_neg[327].bit_mask=create_bit_mask(41,0x82820882,0x28228a00,0x80002202,0x2800822,0x20220000,0xa088282,0x82080002,0x28800000,0x80082002,0x20082002,0x20a08a00,0x82280288,0xa20008a0,0x20820a28,0x200828a,0x80880802,0x28028a28,0x8a280080,0x800880,0x8020a00,0x82208202,0xa28020a0,0x8808a08,0x208a000,0x800000a0,0x20a08808,0x8282,0x2082,0x8008a20,0x800a002,0x22800802,0xa20828,0x228208a,0x828a2880,0x20a28820,0x2008080,0xa28a2022,0x200a00,0x220228a,0x82800020,0x808200);
remainders_pos[328].modulo_value=328; remainders_pos[328].bit_mask=create_bit_mask(11,0xaa882a0a,0x222008a,0x20a20a28,0x82a222a8,0xa0288000,0x2280a00,0x888a8200,0xa08a082a,0x888028,0xa822aaa2,0xa0);
remainders_neg[328].modulo_value=328; remainders_neg[328].bit_mask=create_bit_mask(11,0x2200a282,0x8aaa8802,0xa82a82a0,0x22aaa20,0x28a00888,0x8aa08288,0x20a88,0x280280a2,0x80008a0,0x20aa222a,0x28);
remainders_pos[329].modulo_value=1316; remainders_pos[329].bit_mask=create_bit_mask(42,0x82082222,0x282222,0x808282a2,0x22200000,0x8200802,0x8aa20aa0,0x800022a0,0xa280802a,0xa00a82,0x82a0822a,0x2a0a28a2,0x2802a208,0xa2a0a80,0xaa008,0x2a882a20,0x20802a82,0x22220080,0x8080000,0xa0022a0a,0x820a2220,0x888a2220,0x888a0820,0xa8800a0,0x2020a,0x20088880,0x2a802080,0x8a822a88,0xaa0000,0x2a0a8a02,0x8a80280,0xa28a0a82,0x8820a828,0x2a00a00a,0x802028a8,0xa880002a,0xaa08aa20,0x2008200,0x8888,0xa8282020,0x88828008,0x88820828,0x8);
remainders_neg[329].modulo_value=1316; remainders_neg[329].bit_mask=create_bit_mask(41,0xa80aa2a,0x8a02aa2,0xa0a2a,0xaa288888,0x80a0808a,0x22a0228,0x888a228,0x2a080022,0x8288202,0x2280aa2,0xa202a02a,0xa0822a80,0x82a20208,0x88822000,0xa200a228,0x2808a202,0xa2aa8808,0x80008888,0x2882a282,0xa822aa0,0x2a2a8,0x28028,0x2008820,0x808a8a82,0x28000008,0xa200a808,0x20a2200,0x88228088,0xa282020a,0x208a00,0x2202820a,0x2820a0,0xa2802882,0x8a82020,0x200880a2,0x22802228,0xa080a80,0x80880000,0x2020a8a8,0x20880,0xa00a0);
remainders_pos[330].modulo_value=1320; remainders_pos[330].bit_mask=create_bit_mask(41,0x80880082,0x28028220,0x80208,0x80022800,0x20000a20,0x8220a,0x220800a2,0x8008008,0x8200200a,0x22,0x28820200,0x82000282,0x20002020,0x208808,0x80200200,0xa0002000,0x20020000,0xa08a000,0x80002020,0x800200,0x88,0x80020022,0x8000200,0xa20a008,0x800800,0x8000a00,0x80080200,0x22080000,0x8000820,0x80008208,0x80822882,0x8008288,0x20020a0,0x208820,0x8820008a,0xa00008a0,0x28800208,0x80200000,0x82802820,0x220208,0x82);
remainders_neg[330].modulo_value=1320; remainders_neg[330].bit_mask=create_bit_mask(42,0x2,0x20020a20,0x8008280,0x822080,0x20800228,0x80002282,0xa2000822,0x800000,0xa00a082,0x880822,0x20020a08,0xa088202,0x200820a0,0x8a00000,0x288200,0x20882880,0x20028800,0x2002088,0x802820,0x8a08,0x8080000,0xa08a2,0x808208,0x82202080,0x80080000,0x800200,0x8000280,0x22800080,0x808028,0x80200,0x22002,0x8008800,0x80200,0x82802020,0x200028,0x280002,0x20080020,0x20000200,0x208088,0x2082020,0x8a20a00,0x2);
remainders_pos[331].modulo_value=1324; remainders_pos[331].bit_mask=create_bit_mask(42,0xaa28aaa,0x88aaa00a,0x2222820,0x228a2a08,0xa2a82080,0x2802a222,0x880a2,0xa8aa0a8a,0x2082a80,0xa8802a2,0x28a0282,0x8a880aa2,0x202a00a2,0xaa820280,0xa0882a22,0x808880a,0x8aa28a0,0x8aa28082,0x2228a088,0x282a800,0x208,0xa8280208,0xa2888002,0x28aa2220,0x8aa20820,0x220200a2,0x8220aa02,0x82aa88a,0xa808028,0x22a28a0,0xa2808aa,0x22a0828,0x820808a8,0xaa2a02a,0x22000a2a,0xa80288a0,0x82a8a888,0x8a288020,0x8888020a,0xaaa22082,0x28aa0a00,0xaaa);
remainders_neg[331].modulo_value=1324; remainders_neg[331].bit_mask=create_bit_mask(42,0x822a0222,0x222882,0x8aaaa0a8,0xaa02a280,0x2a20a808,0xa08a2aaa,0x8880082a,0x20228202,0x8a80a208,0x82008a2a,0x8a02820a,0x200822a,0xa8a2882a,0x220a8a08,0x2800a2aa,0x80800082,0x8022a028,0x22a080a,0xaaa02800,0x8a0a2088,0x88888a80,0x20a08a80,0x2a00088a,0xa022aaa8,0x22a80a8,0xaa8a882a,0xaa8228a,0x80a22002,0x820808a0,0x8aa2a028,0x82a08022,0x8aa280a0,0xa808020,0x822a28a2,0xaa8882a2,0x208a0028,0xa202000,0x2a008a8,0x8a82,0x222aa80a,0xa0228288,0x222);
remainders_pos[332].modulo_value=664; remainders_pos[332].bit_mask=create_bit_mask(21,0x222a8202,0x20828aaa,0x2802a0aa,0x20a2880,0x8a008880,0x200a2a0a,0x8a080022,0xa8028022,0x28208aa0,0x2a888aaa,0xa8202808,0x28aaa222,0x2a0aa208,0xa2880280,0x8880020,0xa2a0a8a0,0x80022200,0x280228a0,0x8aa0a80,0x88aaa282,0x8082a8);
remainders_neg[332].modulo_value=664; remainders_neg[332].bit_mask=create_bit_mask(21,0xaaa20a8a,0xa80a0222,0xa0822822,0x8a82a008,0x2880008,0xa882a282,0x28088aa,0x208a08aa,0xa0a80228,0xa2000222,0x20a8a080,0xa0222aaa,0x22822a80,0x2a008a08,0x800088a8,0x2a282028,0x88aaa88,0xa08aa028,0x80228208,0x222a0a,0x80a20);
remainders_pos[333].modulo_value=444; remainders_pos[333].bit_mask=create_bit_mask(14,0x2000882,0x228200,0x8028228a,0x82080820,0x28a20a00,0x20a200,0xa0880020,0x800220,0x8008a080,0x200a08a2,0x20820208,0xa288280,0x82880,0x8220008);
remainders_neg[333].modulo_value=444; remainders_neg[333].bit_mask=create_bit_mask(14,0x82880002,0x8a20a08,0x820a202,0x28000a0,0x20228208,0x88282288,0x200008a0,0x8008a28,0x8002000,0xa0820022,0x28028a00,0x82200208,0x80002000,0xa28800);
remainders_pos[334].modulo_value=1336; remainders_pos[334].bit_mask=create_bit_mask(42,0xa88aa2a,0x2a220a2,0x820288a2,0x22a8a8aa,0x80a88280,0x2080a20,0x22a8828,0x28a82202,0xaa28aa00,0x20aa0a,0xa20222a2,0xa00a2a,0x2aa00820,0xa2a22808,0xa8288022,0x8a00a8a2,0x2a82880a,0x80080802,0xa20882a2,0xa20a2202,0x200880,0x8a022008,0x8a8088a0,0x28a8220,0xa8802020,0xa2a02282,0x2a8a2a00,0x8a880228,0xa820288a,0x8200a,0x8aa8a00a,0x8a8880,0xaaa0aa08,0x2800aa28,0x8080882a,0x802822a8,0x208a020,0x8802822a,0x82aa2a2a,0x808a2280,0xa08a088a,0xa8aa22);
remainders_neg[334].modulo_value=1336; remainders_neg[334].bit_mask=create_bit_mask(42,0x820022a2,0x8a2aa82a,0xa8a002a,0xaa202022,0x8200a08,0x8a808228,0x8aa200a0,0xa020aa8a,0x22a02288,0x88a82282,0x2a8aaa2a,0x882882a2,0xa22880a8,0x2a2aa080,0x20a008aa,0x288202a,0xa20a0082,0x880808a,0x2a800a2a,0x2a82aa8a,0x88a88008,0x28aa880,0x2080028,0x8a020aa8,0x2008a8a8,0x2a28aa0a,0xa202a280,0x2008aa0,0x20a8a002,0x8880a882,0x2202882,0x88020008,0x22282280,0xa08822a0,0x80800a2,0x8a0aa20,0x8a8028a8,0x8a0aa2,0xa22a2a2,0x802aa08,0x28028002,0x2022aa);
remainders_pos[335].modulo_value=1340; remainders_pos[335].bit_mask=create_bit_mask(42,0xa0a02a02,0x22a8000,0x20a8000,0x8aa2888a,0x208a22,0xa0aa0288,0x88208020,0x28028288,0x8882a0a8,0x8288a802,0x20220888,0xa200202,0x20080220,0x20208a0,0x20882822,0xa00a0008,0x80208a2,0x2020000,0x82a2008a,0x22a8000a,0x8802808,0x82028022,0x2a0002a8,0xa2008a8,0x808,0xa8a20802,0x82000a00,0x8828220,0x80a20808,0x880200,0x8808008a,0x22220880,0x2802a228,0x82a0a822,0x22282802,0xa0802082,0x2280aa0,0x288a2080,0xa2228aa,0x2a08,0xa0002a88,0x80a80a0);
remainders_neg[335].modulo_value=1340; remainders_neg[335].bit_mask=create_bit_mask(42,0x2828228a,0x8a220880,0xa820080,0x2220002,0x80a8022a,0x28220a00,0x2808a0,0x208a0200,0x22820,0x200200a,0xa8aa0000,0x82288a82,0x288082a8,0x8a828028,0xa000a02a,0x28820880,0x800a8022,0xa8a8088,0x2228802,0xa2208802,0x80082080,0xa0a08a2,0x22888220,0x82a08020,0x80888000,0x202a008a,0xa088280,0xa0228,0x8228080,0x80008a08,0x800802,0xaa2a8000,0x208a22a0,0xa2020aa,0xa2a0a00a,0x2800280a,0x8a208220,0x2002a008,0x82a2a022,0x8088a200,0x28882200,0x20820);
remainders_pos[336].modulo_value=168; remainders_pos[336].bit_mask=create_bit_mask(6,0x2020022,0x8008a20,0x2288008,0xa2002020,0x808008,0x88);
remainders_neg[336].modulo_value=168; remainders_neg[336].bit_mask=create_bit_mask(5,0x828a0822,0x800220,0x82200080,0x220828a0,0x8000800);
remainders_pos[337].modulo_value=337; remainders_pos[337].bit_mask=create_bit_mask(11,0x1f2573de,0x81d78eb1,0xc876cd01,0x4a2bd995,0x587b220b,0x7868b7b4,0x514b4113,0xb84ea66f,0xae0602cd,0x93e235c7,0x1ef3a);
remainders_neg[337].modulo_value=1348; remainders_neg[337].bit_mask=create_bit_mask(43,0x82a8aa02,0x80a0228,0xaa0088,0x82a20008,0x80a2aa82,0xa0a02a28,0x8882888a,0x20822aa2,0x228a8a00,0xa2aa80a,0x2a2022a2,0x80aaa282,0x28a8a22,0x2aa20820,0x888a088a,0xa2a0282a,0xaaa2808,0x80022a0a,0x8802a800,0xa2028080,0x2a8aa08,0xa00802a2,0x8a82a208,0xa802a88a,0xa022aa8a,0xa82000a0,0x282a0208,0x88a08880,0x28a20020,0x208082aa,0x82000a82,0x2222022,0xa80020a2,0xa0808220,0x228a2a,0x88828882,0x202a0a00,0x80020a88,0xaaa20282,0x8aa00aa8,0x22a0a8a8,0xa0080288,0x2);
remainders_pos[338].modulo_value=104; remainders_pos[338].bit_mask=create_bit_mask(4,0x82828282,0x82828202,0x82828280,0x82);
remainders_neg[338].modulo_value=104; remainders_neg[338].bit_mask=create_bit_mask(4,0xa0a0a0a,0xa0a0a0a,0x20a0a08,0xa);
remainders_pos[339].modulo_value=1356; remainders_pos[339].bit_mask=create_bit_mask(43,0x220a2822,0x20820800,0x8228800a,0x220020a2,0x20828208,0x8000202,0xa28200a0,0x8a28008,0x8228a008,0x82820022,0x8028228,0x8a080200,0xa0000800,0x28000220,0xa080002,0x828a20a0,0x8020808,0x8088082,0xa2000800,0x20820800,0x8000808a,0x20002000,0x8208a20,0x8a002,0x22020002,0x8020820,0x8a282202,0x20a00a0,0x28800,0xa088,0x20a2002,0x28020008,0x8282288,0xa2882880,0x28a20200,0x828a200,0x200a0,0x28208808,0x80088208,0x228828a0,0x8208a00,0x8a088002,0x882);
remainders_neg[339].modulo_value=1356; remainders_neg[339].bit_mask=create_bit_mask(43,0xa28220a2,0x28028008,0xa200082,0xa2882822,0x28020a00,0x8008828a,0x220a0820,0x220800,0xa202080,0x20a08a2,0x820a20,0x2008288,0x20880080,0x20808a28,0x8200808a,0x2022820,0x828000,0x8000000a,0x22880080,0x28028008,0x8080002,0xa0882880,0xa00228,0x8800208a,0xa20a0882,0x828028,0x220a28a,0x82820820,0x8820008,0x88082000,0x82822882,0x20828800,0x8020a200,0x22002000,0x20228a08,0x80202288,0x808a0820,0x20a00000,0x280,0xa2002020,0xa00208,0x200008a,0x2);
remainders_pos[340].modulo_value=680; remainders_pos[340].bit_mask=create_bit_mask(22,0x8a8028a,0x8a020020,0x2020220,0x2028822,0xa88020,0xa00a2a8a,0x8008022,0x8088228,0xa208808,0xa2008008,0x28aa2802,0x2008a80,0x2208a020,0x82202020,0x2002028,0xa8a00a88,0x22a00a2,0x22808008,0x80808088,0x80a208,0x802a2008,0xa2);
remainders_neg[340].modulo_value=680; remainders_neg[340].bit_mask=create_bit_mask(22,0x80200a02,0x20288a0,0xa8a82a8,0x8a0200aa,0x80200828,0x20822202,0x800808a2,0x8002a0,0x82a00080,0x22880800,0xa0222082,0x8a080200,0x2a802028,0xaa0a8a8,0x8288a020,0x20280200,0x8a220822,0x2a080080,0x8000800,0x80082a00,0x8222880,0x22);
remainders_pos[341].modulo_value=1364; remainders_pos[341].bit_mask=create_bit_mask(43,0x22222222,0x2a02a800,0x2a82aa,0xa8228282,0xa2820820,0xa000028a,0x20800020,0x228802,0xa02288aa,0x20828082,0xa0280a2,0xa082228,0x8a202a08,0x8a2002a8,0x82200,0x2820,0x820a2a8a,0xa20a820,0xaa0020a,0xaa02a2aa,0x22222000,0x88888a22,0xaa00088,0xaa0aa8a8,0x8a0a0800,0x82082a0,0xa2a8a,0x8280,0x8a200882,0x8a22a800,0xa020a80,0xa028882,0x2088a028,0x80a82028,0x800aa228,0x20880228,0xa08000,0x28aa2800,0x82a08208,0x80082828,0xa8aa82a,0x888002a8,0x88888);
remainders_neg[341].modulo_value=1364; remainders_neg[341].bit_mask=create_bit_mask(42,0x2aaaa2aa,0xa20a2088,0x88a20a22,0x20aa0a02,0x220a00a8,0x20888a02,0xa80888a8,0x88aa000a,0x28220022,0x280a080a,0x828a082a,0x8280a2a0,0x220a280,0x2a88a20,0x8880aa80,0x888820a8,0x2822202,0x82a820a8,0x82288a02,0x22822a22,0x2aaaa088,0x2aa,0x82288000,0x22022020,0x2828008,0x80a80a20,0x88822202,0x80880a08,0x2a88002,0x2aa2008,0x82828208,0x28a000a,0xa80028a0,0x820a0a0,0x8022aa0,0xa0008aa0,0x88280880,0xa0222088,0x2280a80,0x800a0a0,0x82022022,0x8a20);
remainders_pos[342].modulo_value=456; remainders_pos[342].bit_mask=create_bit_mask(15,0xa2022802,0x220828,0x88280,0x82082080,0x808a08,0x820228a,0xa0002000,0x8000a00,0x88082000,0xa20200a2,0x8208220,0x82200002,0x20880002,0x28808a28,0x80);
remainders_neg[342].modulo_value=456; remainders_neg[342].bit_mask=create_bit_mask(15,0x22822082,0x8a28020,0x8000208,0x2802800,0x8000200,0x8028a202,0x20882880,0x8208,0xa088,0x228a0822,0xa00a20,0xa28808a,0xa0000882,0x20000220,0x8);
remainders_pos[343].modulo_value=28; remainders_pos[343].bit_mask=create_bit_mask(1,0xa08020a);
remainders_neg[343].modulo_value=28; remainders_neg[343].bit_mask=create_bit_mask(1,0x2808a02);
remainders_pos[344].modulo_value=344; remainders_pos[344].bit_mask=create_bit_mask(11,0x22020aa2,0xaa8222a8,0x202a20a8,0x2280080a,0xa2a22800,0x8a008200,0x8800288a,0x8a02002,0xaa2a08a8,0x882a8882,0x8aa080);
remainders_neg[344].modulo_value=344; remainders_neg[344].bit_mask=create_bit_mask(11,0xaa8a822a,0x220aa220,0xa8a2a820,0xaa088082,0x2a2aa088,0x2880a88,0x8a002,0x8028a88a,0x22a28020,0xa2000a,0x22808);
remainders_pos[345].modulo_value=1380; remainders_pos[345].bit_mask=create_bit_mask(44,0x80000882,0x28820,0x2002008,0x2000082,0x20a00800,0x80002280,0x22080020,0x8800200,0xa00800a,0x22002002,0x28a20a08,0x2208282,0x2820020,0x8828008,0x82200200,0x82000,0x28200a20,0x2000080,0x2020800,0x8020,0x8208a,0x220a0880,0x20820000,0x8020000a,0x2080800,0x20000800,0x8a008280,0x820000,0x8008820,0x282200,0x80082802,0x28208800,0xa08a288,0x800882,0x200a08,0x800220a,0x80020880,0x28800020,0x200a080,0x20000800,0x800808,0x82280002,0x22000020,0x8);
remainders_neg[345].modulo_value=1380; remainders_neg[345].bit_mask=create_bit_mask(43,0x80002,0x8020020,0xa08a080,0x82800802,0x20208008,0x8082208,0xa20008a0,0x8208,0x82000082,0xa2882002,0x20220200,0x8a280202,0x20a0020,0x20800,0x2288208,0x80802880,0x20208220,0xa088008,0x82820080,0x800828,0x80002002,0xa2020000,0x28028008,0x8208082,0x82800000,0x28800008,0x2000200,0xa0080,0x800028,0x80202208,0x802082,0x20200000,0x2002200,0x80000002,0xa08200,0x80082202,0xa0000,0x20008028,0x8a002000,0xa0880000,0x8000000,0xa208082,0x228800a0);
remainders_pos[346].modulo_value=1384; remainders_pos[346].bit_mask=create_bit_mask(44,0x8a888a2a,0x2aa2a202,0xaa020a20,0x2a8a0828,0x88882288,0xa20aa0,0x2802a800,0x20280a02,0x28aa0a8a,0x28aa2a2,0x222a0002,0x82a2aaa0,0x8220200,0x2a2080a0,0xa822a0a8,0xaaaa802,0x888a0022,0x80288228,0xa00828a0,0x2a20a2,0x8082a202,0x8002088,0x8a820222,0x8a80080,0x28200a8a,0x8228020a,0xa22228,0x2aaaa088,0xa882a80,0x208a82a,0x8088200a,0xaa8a8200,0xa8880a,0x8aa28080,0xa0aa288a,0xa02808a2,0x2a802880,0xa08a0000,0x8822220a,0x20a2a822,0xa080aa28,0x8a8aa808,0xa222a280,0xa8);
remainders_neg[346].modulo_value=1384; remainders_neg[346].bit_mask=create_bit_mask(44,0x20002a2,0xa22a2a8a,0x228a82a8,0xa20280a0,0xaa00,0x882a8228,0xa08a2088,0xa8a0828a,0xa0228202,0x8a022a2a,0xaaa2888a,0xa2a2228,0x80aa8a88,0xa2a80828,0x20aa2820,0x8222208a,0x2882a,0x8a00aa0,0x2880a028,0x88a2a82a,0x80a2a8a,0x8088a800,0x20a8aaa,0x80208808,0xa0a88202,0xaa08a82,0x882aaaa0,0xa2222800,0x8200a208,0x8a8020a2,0x800a882,0x22020a88,0x88200082,0x22a0808,0x2822a002,0x28a0802a,0xa208a008,0x20028888,0xaaaa82,0xa82a20aa,0x280822a0,0x2022080,0x2aaa2a08,0x20);
remainders_pos[347].modulo_value=1388; remainders_pos[347].bit_mask=create_bit_mask(44,0x2288a282,0xa0aa8002,0x2a228200,0xa22a280,0x28a00aa2,0x8a2aaa0a,0x80a8aa28,0x28280aa,0x2a202020,0x888aaa,0x2a08a22,0x2a888228,0x80800222,0xaa8a8a8,0x2a002a0a,0x82208020,0x20828000,0x828802a8,0x80aa0808,0x282a8a08,0x88aaa00,0xa0a0822,0xaa220882,0x2a82800a,0xaa0220a,0x2282202,0x282082a8,0x20882000,0x800a8080,0xa2aa0a0a,0x2022a2,0x222a8888,0x20a80288,0x2220088a,0x808a8aaa,0x28280080,0xa2a02aa0,0xaa8a228a,0xa28a0a,0xa88a08aa,0x288a8028,0x2aa0a008,0xa2288800,0x828);
remainders_neg[347].modulo_value=1388; remainders_neg[347].bit_mask=create_bit_mask(44,0xaa002a0a,0x2822088a,0xa2aa0a88,0x82aa2a08,0xa028822a,0x2a22282,0x82022a0,0x8a0a0822,0xa2a8a8a8,0x88000222,0x822802aa,0xa2000aa0,0x8088aaa,0x82202020,0xa288a282,0xaa808a8,0xa80a0888,0xa008a20,0x8228080,0xa0a20280,0x80022288,0x828280aa,0x22aa800a,0xa20a0882,0x8228aa82,0x8aa0aa8a,0xa0a80a20,0xa800a888,0x8820808,0x2a228282,0x88a8aa2a,0xaaa20000,0xa8208a00,0xaaa88002,0x8020222,0xa0a08808,0x2a28a228,0x2202aa02,0x882a0282,0x20028022,0xa00208a0,0xa2282880,0x2aa00088,0xa0);
remainders_pos[348].modulo_value=696; remainders_pos[348].bit_mask=create_bit_mask(22,0x828a2002,0x8820a08,0x208a080,0x82022820,0x200,0x2280288,0x20080000,0x828808,0x20a208,0x220a0822,0x28008a28,0x8828a200,0x8820a0,0x208a08,0x8202282,0x80000020,0x228028,0x82008000,0x80082880,0x20020a20,0x8220a082,0x8008a2);
remainders_neg[348].modulo_value=696; remainders_neg[348].bit_mask=create_bit_mask(22,0x2022882,0x28200,0x8a002008,0x28a20a0,0x8808a08,0x8a208200,0xa0800080,0x8020000,0x88282280,0x228200a2,0x20800220,0x202288,0x80002820,0x8a00200,0x8028a20a,0x8808a0,0x8a20820,0xa080008,0x802000,0x28828228,0x228200a,0x80022);
remainders_pos[349].modulo_value=349; remainders_pos[349].bit_mask=create_bit_mask(11,0xaedbd23a,0x130b2630,0xf9eb7a7d,0x7e58f510,0x6a9bc884,0x82340b10,0x8844f655,0xc22bc69f,0x2f97b5e7,0x43193432,0x1712f6dd);
remainders_neg[349].modulo_value=1396; remainders_neg[349].bit_mask=create_bit_mask(44,0x22020aa2,0x8a82aaa8,0x2022a2a0,0xa2802888,0xa2020008,0xaa88288,0x882a88,0xaa20a02,0xa20a282a,0x8a80a8aa,0x8a8a2a00,0x8802a28a,0x2aa8a80a,0x82a0a282,0x202822a,0x808aa088,0x288a08aa,0x28800202,0x2088a00a,0x882a2a20,0x20aaaa0a,0x222a8202,0xa22a280,0x2880a000,0x8a2a2020,0x8820aa08,0x8822a2aa,0x8a8008a0,0xa2a88a00,0x2802282,0x22820a,0xa880a808,0x80008202,0x808aa020,0xa200080a,0xa02820,0x8aa2a0a2,0xa88028,0xa2088288,0x8208aaa2,0x2a28882a,0x808a0202,0xa2280002,0x200a2);
remainders_pos[350].modulo_value=280; remainders_pos[350].bit_mask=create_bit_mask(9,0x80002a02,0x22088800,0x820028,0xa2288a8,0xa0008200,0x2a2288,0x88280082,0x2002220,0x80a800);
remainders_neg[350].modulo_value=280; remainders_neg[350].bit_mask=create_bit_mask(9,0x888220a,0x2a000080,0xa80a0,0x82220020,0x20800a08,0x88a22a00,0x208002,0xa88a228,0x2080);
remainders_pos[351].modulo_value=156; remainders_pos[351].bit_mask=create_bit_mask(5,0x828800a2,0x20020208,0x82000008,0x22080800,0x8a00228);
remainders_neg[351].modulo_value=156; remainders_neg[351].bit_mask=create_bit_mask(5,0x2000822,0x28828a00,0x2088080,0xa2800080,0x200a20);
remainders_pos[352].modulo_value=88; remainders_pos[352].bit_mask=create_bit_mask(3,0x2a20228a,0xa8020080,0xa28808);
remainders_neg[352].modulo_value=88; remainders_neg[352].bit_mask=create_bit_mask(3,0xa2a8a202,0x200a8808,0x2a0080);
remainders_pos[353].modulo_value=353; remainders_pos[353].bit_mask=create_bit_mask(12,0x62ef8b16,0x3406dedd,0x795e5353,0x8712a01e,0xbb05118d,0xbf631bf4,0xc6228374,0xe0152386,0x2b29ea79,0xeded80b3,0xa347dd1a,0x1);
remainders_neg[353].modulo_value=1412; remainders_neg[353].bit_mask=create_bit_mask(44,0xaa22028a,0xa88a0200,0xa0828a8a,0xa8a2882,0x22888800,0x22aa8228,0xaaa0aa8,0x2888aa0a,0xa2a022a0,0x2020082a,0x2a8a0082,0x8028aa0,0xa0802022,0x2a202a2a,0x82a888a0,0xaa82aa82,0xa20aaa20,0x888a20,0x8a28a80,0x8a8a082a,0x20288aa,0x8a0222a8,0x222a082,0x882a2a8,0x28a08080,0x808208a0,0x20888aaa,0x2000a20a,0x8002800a,0x8880282,0x202a202a,0x2a2a8a00,0x82a8a2,0x8aa08022,0x28aa2a22,0x22a0200,0xa008882a,0xa000a0,0x22800228,0xa8888228,0x882080aa,0x80828a02,0xa2a08800,0x82a2200a);
remainders_pos[354].modulo_value=1416; remainders_pos[354].bit_mask=create_bit_mask(45,0x228020a2,0x20228828,0xa00800a,0xa20a2800,0x20020000,0x8a008202,0x808a08a2,0x8220a20,0x8008208,0x20022822,0x20a08228,0x2080202,0x22820,0x20828820,0x202088,0x800882,0x20828000,0x8a20a202,0xa0002002,0x28828028,0x8002000,0x20a0020,0xa08000,0x8002008,0x2822800,0x8000a28,0x8a08a280,0x2820880,0x20020000,0x8080082,0x22820822,0x28800008,0x80208008,0x820a0880,0x28800828,0x82002088,0xa0882020,0x20a20208,0x8200a28a,0x800880,0x28a08a00,0x200a000,0x228808a0,0x8028828,0x8a);
remainders_neg[354].modulo_value=1416; remainders_neg[354].bit_mask=create_bit_mask(45,0xa2082822,0x20a20020,0x82080082,0x22822080,0x28828808,0x208028a,0x20022,0xa28228,0x80080280,0xa08a2022,0x28200a20,0x8a00828a,0x808a20a0,0x28020028,0x8828a000,0x80080002,0x28020808,0x228228a,0x20882882,0x20020820,0x8008a088,0x828208a0,0x8200808,0x8a080,0x820a2080,0x808220,0x2002208,0x820a0000,0x28828808,0x8000800a,0xa20a00a2,0x20008000,0x8280080,0x2820000,0x20008020,0xa08a000,0x200028a0,0x28228a00,0xa082202,0x80080000,0x20200208,0x8a082088,0xa2000020,0x820020,0x2);
remainders_pos[355].modulo_value=1420; remainders_pos[355].bit_mask=create_bit_mask(45,0xa8022a0a,0x82a0882,0x228a0000,0x22228aa,0x82088a22,0x8022282,0xa0080022,0x2088020,0x808008aa,0xa020820a,0x802a0a82,0x2008a02,0xa08a008,0x2a22802,0xa2208a00,0xa8a02808,0x80a8280,0x2028208,0x808888,0x22800800,0x2080228a,0x800a8002,0x2a002000,0x80208800,0x288a28,0x20200002,0x28080222,0x2a020208,0x80a2a028,0x2088a202,0x88a8000a,0xa20a0802,0x20080200,0xa80280a,0x2080a82a,0x202a08,0x22080aa2,0x200a080,0x88020880,0x22082828,0x8888088a,0xa288aa2,0xa820000,0x8802a822,0xa0a);
remainders_neg[355].modulo_value=1420; remainders_neg[355].bit_mask=create_bit_mask(45,0x208a2282,0x80228002,0x2a028008,0x8aa2a022,0x280022a,0x808a2a0a,0x280088a2,0xa8000a8,0x8008022,0x20a80a02,0x8a2020a,0x8a080282,0x2802080,0x8a22a08a,0x22a80208,0x20282080,0x80020a00,0xa8a0280,0x88000000,0x22088008,0xa8082a02,0x8020882,0x2288a088,0x8a00088,0x80a00220,0xa8a8088a,0xa0008aa2,0x228a8280,0x2228a0,0xa0002a0a,0x200882,0x2a028082,0x28808288,0x8200a082,0xa0082022,0x88a82280,0xaa008222,0xa082008,0x828008,0xa280a020,0x2,0x82200222,0x20a8088,0x8220aa,0x202);
remainders_pos[356].modulo_value=712; remainders_pos[356].bit_mask=create_bit_mask(23,0x2220a22,0x2a2a080,0x28a282a8,0x2a00aa0a,0x2200082a,0x288a2282,0x82000888,0xaa00a8a,0xa828a28a,0x20a8a802,0x8a088800,0x2220a228,0x2a2a0800,0x8a282a80,0xa00aa0a2,0x200082a2,0x88a22822,0x20008882,0xaa00a8a8,0x828a28a0,0xa8a802a,0xa0888002,0x88);
remainders_neg[356].modulo_value=712; remainders_neg[356].bit_mask=create_bit_mask(22,0x8aaa82aa,0x8a2a2808,0xa02a0a20,0xa2882282,0xaa8880a2,0xa002aa0a,0xa888000,0x82288202,0x20a02202,0xa820208a,0x2800088,0xaaa82aa0,0xa2a28088,0x2a0a208,0x2882282a,0xa8880a2a,0x2aa0aa,0xa888000a,0x22882020,0xa022028,0x820208a2,0x2800088a);
remainders_pos[357].modulo_value=1428; remainders_pos[357].bit_mask=create_bit_mask(45,0xa2800802,0x28008800,0x8208028a,0x82020822,0x800800,0x2200200,0x20a0,0x28820200,0x8080202,0xa2000020,0x8200800,0x80208202,0x22800,0x808020,0x208a202,0x20000002,0x220028,0x80080002,0x82020800,0x820a20,0x8800a28a,0x800a2800,0x200a00,0x220028a,0x200a28a0,0x8208a08,0x200208,0x80080002,0x828008,0x8080000,0x200808a2,0x8020,0x80200288,0x82080820,0x8a00200,0x2008000,0x22880802,0x808,0x8800a080,0x20000800,0x8200200,0x8288208,0x28a2802,0x28a00220,0x80200);
remainders_neg[357].modulo_value=1428; remainders_neg[357].bit_mask=create_bit_mask(45,0x22080002,0x20800000,0x2008202,0x20a00a2,0x8008008,0x82280288,0x80882020,0x20028a08,0x8282,0x228808a0,0x208000,0x820028a,0x808a2000,0x8000028,0x8a00220a,0x20880882,0xa20820,0x800808a,0x2820080,0x8020228,0x82202,0x822000,0xa08200,0x82288202,0x20022020,0xa00200,0x88280280,0x800082,0x8020800,0x8000,0xa0800022,0x8000028,0x8208200,0x28000a0,0x208208,0x8a000008,0x22000082,0x8000,0x82008,0xa0800080,0xa00a08,0x200280,0x82022002,0x20208a20,0x8280);
remainders_pos[358].modulo_value=1432; remainders_pos[358].bit_mask=create_bit_mask(45,0xaaa028a,0x8aaa0820,0x20a88aa,0x8280a880,0xaa0a8822,0x20228aa2,0x8020a22a,0x208282a,0x20a888,0x80a22022,0x28a2208a,0x22082a20,0x220a808a,0xa2aa22a,0x28a2a888,0xa2a8028,0x802a2022,0x80a00200,0xa828228,0x80a2aa88,0xa0008008,0xa000aa28,0xa820082,0xa28aa00,0x2200200,0xa022aa8a,0x2288282,0x200800a,0xa08808a8,0x282802a8,0xa0222a8a,0x88a88aa8,0x88a202a0,0x2808a820,0x2a2088a,0x88088a,0x80222a08,0x2a82820,0x8a88a08,0xaa8aa288,0x828822a0,0x80022a02,0xa2aa22a0,0xa00820aa,0xa280aa);
remainders_neg[358].modulo_value=1432; remainders_neg[358].bit_mask=create_bit_mask(45,0x82228a02,0x22280a8,0x8a820022,0xa082008,0x228200aa,0xa8a2022a,0x8a82aa2,0x8a80a0a2,0x88a82000,0x82aa8aa,0xa02aa802,0xaa80a2a8,0xaa820802,0x82a22aa2,0xa02a2000,0x82a208a0,0x8a2a8aa,0x8288a88,0x820a0aa0,0x82a2200,0x28880880,0x288822a0,0x820a880a,0x82a02288,0x8aa88a88,0x28aa2202,0x8aa00a0a,0xa880882,0x28008020,0xa0a08a20,0x28aaa202,0x200220,0x2a8a28,0xa08020a8,0x8a2a8002,0x88008002,0x8aaa280,0x8a20a0a8,0x80200280,0x22022a00,0xa00aa28,0x88aa28a,0x2a22aa28,0x2880a822,0x2a0822);
remainders_pos[359].modulo_value=1436; remainders_pos[359].bit_mask=create_bit_mask(45,0x820a02a2,0x88022aaa,0x80aa02a8,0x228000a0,0x8a20a820,0x2a8aaa8,0x8288208a,0x2822a2aa,0x82828888,0xa8802a8a,0x8a88a08a,0x22022208,0xaa200228,0xa0a0200,0x8a0a222,0xa20a0008,0x200a8228,0x28822000,0xaa08a8a0,0x802a282a,0x8aa2200a,0x828a0000,0x2a02080a,0x20000a28,0x2a0088aa,0xaa828a80,0x80a2a20a,0x80008822,0xa2882a00,0x2000a08,0x888a0a2,0xa0080a0a,0x8288008a,0x22088808,0xaa20a22a,0x2a2a8022,0x82222828,0x2aa8a882,0xa208228,0x22aaa2a8,0x8082a08a,0x20a00028,0x22a80aa0,0x2aaa8802,0x8a80a08);
remainders_neg[359].modulo_value=1436; remainders_neg[359].bit_mask=create_bit_mask(45,0xa828a2a,0x8aa222,0x8228a20,0xaa088828,0x2a820a8,0x8a202220,0xa00a802,0xa0aa2a22,0xa0a0000,0x2008a202,0x2002802,0xaa8aaa00,0x22a88aa0,0x82828a88,0x80282aaa,0x2a828880,0xa8820aa0,0xa00aa888,0x22802028,0x8a2a0a2,0x22aa882,0xa028888,0xa28a8082,0xa88882a0,0xa2880022,0x220a0208,0x82a2a82,0x88800aa,0x2a00a288,0x8a888280,0x8000282a,0x28808282,0xa008802,0xaa800080,0x22a82aa2,0xa2a208aa,0xaaaa0a0,0xa220200a,0x82a80aa0,0xaa222a20,0x80a2802,0xa82888a0,0xaa208228,0xa222008a,0x208280);
remainders_pos[360].modulo_value=120; remainders_pos[360].bit_mask=create_bit_mask(4,0x80002002,0x220a20,0x208a088,0x800800);
remainders_neg[360].modulo_value=120; remainders_neg[360].bit_mask=create_bit_mask(4,0x882882,0x8228220,0xa002000,0x80);
remainders_pos[361].modulo_value=76; remainders_pos[361].bit_mask=create_bit_mask(3,0xaaa2aaaa,0xa8aaaaaa,0xaaa);
remainders_neg[361].modulo_value=76; remainders_neg[361].bit_mask=create_bit_mask(3,0x22222222,0x20222222,0x222);
remainders_pos[362].modulo_value=1448; remainders_pos[362].bit_mask=create_bit_mask(46,0x2288202,0x20aa008a,0x28aaa222,0xa88a800,0x2822a28a,0x2a0a2a8a,0xa2a82a02,0x20020a02,0xa2202888,0x8a20aaaa,0x802a8a00,0x2a80828a,0x208aa080,0x2200000a,0xaa28882a,0x202a0a2,0x2022a028,0x28208020,0x80a08202,0x82aa808,0xa222200,0x22a08aa0,0x8a82a288,0xa20a8822,0x8888a00a,0x2aa82000,0x820a0220,0x2082880,0xa880808,0xa808028,0x2228aa8a,0x88a8,0xaa208a0,0x8202a802,0xa2a802a2,0xaa08a200,0x28088aaa,0xa0800822,0xa82a8a80,0xa8a0a880,0x8a8828a2,0x2a22a0a2,0x8aaa2800,0xaa0888,0x822880a2,0x80);
remainders_neg[362].modulo_value=1448; remainders_neg[362].bit_mask=create_bit_mask(46,0x8aa00a8a,0xa8228802,0xa0222aaa,0x82002088,0xa0aa2a02,0xa282a202,0x2a20a28a,0xa88a828a,0x2aa8a000,0x2a82222,0x8a20288,0xa2080a02,0xa8022808,0xaa888882,0x22a000a2,0x8a8a282a,0x28aa28a0,0xa0a808a8,0x8280a8a,0x80a22080,0x82aaaa88,0xaa280228,0x20a2a00,0x2a8200aa,0x2882,0xa220a888,0xa828aa8,0x8a80a008,0x82008080,0x820808a0,0xaaa02202,0x88880020,0x822a8028,0xa8a208a,0x2a208a2a,0x22802a88,0xa0800222,0x280880aa,0x20a20208,0x20202008,0x200a02a,0xa2aa282a,0x222a088,0x88228000,0xaa0082a,0x8);
remainders_pos[363].modulo_value=132; remainders_pos[363].bit_mask=create_bit_mask(5,0x2802002,0x28028028,0x80280280,0x802802,0x8);
remainders_neg[363].modulo_value=132; remainders_neg[363].bit_mask=create_bit_mask(4,0x82082082,0x20020820,0x8208208,0x80082082);
remainders_pos[364].modulo_value=728; remainders_pos[364].bit_mask=create_bit_mask(23,0x2a008a2a,0xa02202,0x2820a88,0x2a02008a,0x80820888,0x28,0x8202280,0x80a8222,0x8280a20,0x80a0022a,0x200a8808,0x8a2aa8a,0x22022a0,0x20a8800a,0x2008a028,0x208882a0,0x28808,0x2280000,0xa8222082,0x80a20080,0x22a082,0xa880880a,0xa8a200);
remainders_neg[364].modulo_value=728; remainders_neg[364].bit_mask=create_bit_mask(23,0xa2880222,0x828aa02,0x820a8200,0xa20a8802,0x8020000,0x888808a0,0x80a8a200,0x80020a2a,0xa082a0,0x2882a2,0x28020080,0x80222202,0x8aa02a28,0xa8200082,0xa8802820,0x20000a20,0x808a0080,0x8a200888,0x20a2a80a,0x82a0800,0x882a200a,0x20080002,0x202280);
remainders_pos[365].modulo_value=365; remainders_pos[365].bit_mask=create_bit_mask(12,0x114b22d2,0xe472ce12,0x2b8ad4b1,0x2e7a8c1,0x26210070,0x11221b98,0x21190676,0x79d00380,0xd47520c5,0xd389e34a,0x34a2121c,0x12d1);
remainders_neg[365].modulo_value=1460; remainders_neg[365].bit_mask=create_bit_mask(46,0x88822a0a,0x282a0282,0x22020028,0x8a222008,0xa2a88828,0x88aa0200,0xa0008aa2,0x20088208,0x8a0a808,0x2002a02,0xa82a0a80,0xa8028a00,0xa82a0,0x8202a002,0x2080a828,0x28820880,0x8a2a8800,0x280202a8,0x88a088aa,0x20802222,0xa0a00202,0x8a0a02a0,0x282a208,0x8080a002,0x2080220,0x88222282,0x8020222,0x28000088,0x8000a2,0x8008a800,0x882a0882,0xa2080802,0x2a82a200,0x88020080,0x2008a002,0xa0aa0080,0x8228022,0x2a08a008,0x8800a088,0x80000a,0xa2280,0x20080880,0x2208a220,0x82020a2,0x80808220,0x2002);
remainders_pos[366].modulo_value=1464; remainders_pos[366].bit_mask=create_bit_mask(46,0x820a08a2,0x8828028,0x8220a200,0x22800002,0x28a00020,0xa208288,0x82020020,0x8228a08,0xa282088,0x200008a2,0x20a00208,0xa202200,0x20822802,0x20008228,0x88082000,0x80800020,0x28228220,0x8008200,0x220a2800,0xa28a00,0xa200202,0x800a0080,0x8020,0x2080200,0xa00200a0,0x808008,0x8800a28a,0x200028a0,0x28008200,0x2088288,0x22080002,0x8000820,0x8288200,0xa0802882,0x8008808,0x820800a,0xa08a2000,0x20220828,0x8220a288,0xa0080080,0x28228208,0x8808000a,0x82800002,0x20008a08,0x82280282,0x8a20a0);
remainders_neg[366].modulo_value=1464; remainders_neg[366].bit_mask=create_bit_mask(46,0x2820022,0x20820,0xa282288,0xa2080882,0x20208828,0x82280200,0x28a08a0,0xa20200,0x8220a000,0xa0880022,0x28208a00,0x8228a288,0xa00a2082,0x28800220,0xa088,0x808a0,0x20a20a28,0x80080288,0xa2822080,0x8220208,0x228828a,0x820800,0x8800828,0x8a008288,0x208a0820,0x8000800,0x82202,0xa0882020,0x20800a08,0x8a000200,0xa2800882,0x808028,0x80200288,0x20082002,0x800000,0x80280082,0x20022800,0x28a28020,0xa282200,0x20800800,0x20a20a00,0x8082,0x2080882,0x20800200,0xa20820a,0x22820);
remainders_pos[367].modulo_value=1468; remainders_pos[367].bit_mask=create_bit_mask(46,0xa082a0a,0x22220aaa,0x2228a80,0x8aaa82a8,0x882a8a0a,0xa82a8a,0x8a220800,0x8088a822,0x2280880,0xaa88282,0x8208aa20,0x28020aa0,0xa880228a,0x8a0a2002,0x2a22a08a,0x8a0222a,0xaaa28882,0x200a02a,0x28200a2,0xa00a0002,0xaa02088a,0x80028888,0x280a282,0x2828a028,0xa2222800,0xaa22080a,0x8000a00,0x8a00828,0xaa80a008,0x82228aa,0x8a8880a2,0x2a20a88a,0xa8008a0a,0x8a288022,0x20aa0802,0x8aa208,0x82822aa,0x20220288,0x2882a220,0x2088a,0x2a2a82a0,0x2a0a2a82,0x2a82aaa,0x802a2888,0xaaa0888,0xa0a820a);
remainders_neg[367].modulo_value=1468; remainders_neg[367].bit_mask=create_bit_mask(46,0x8280a282,0xaaaa8222,0x8aaa0208,0x2220a20,0xa20282,0x8820a202,0x2aa8088,0x80020aa,0x8aa08008,0x82200a0a,0xa8022a8,0xa08a0228,0x2008aa02,0x282a88a,0xa2aa2802,0x8028aaa2,0x222a000a,0x8a8828a2,0x8a0a882a,0x2882888a,0x228a8002,0x88a0000,0x8a082a0a,0xa0a028a0,0x2aaaa088,0x22aa8082,0x80888288,0x802880a0,0x22082880,0x80aaa022,0x200082a,0xa2a82002,0x20880282,0x2a008aa,0xa822808a,0x88022a80,0x80a0aa22,0xa8aa8a00,0xa00a2aa8,0x888a8002,0xa2a20a28,0xa282a20a,0x8a20a222,0x8a2a000,0x82228000,0x2820a82);
remainders_pos[368].modulo_value=184; remainders_pos[368].bit_mask=create_bit_mask(6,0x2208aa82,0x800a0a00,0x282aa208,0x2208aa8,0x8800a0a0,0x82aa20);
remainders_neg[368].modulo_value=184; remainders_neg[368].bit_mask=create_bit_mask(6,0xaa00220a,0x8828288,0xa0a22a80,0x8aa00220,0x882828,0xa22a8);
remainders_pos[369].modulo_value=492; remainders_pos[369].bit_mask=create_bit_mask(16,0x82800022,0x28020820,0x8082200,0xa28a0880,0x8008828,0xa002208,0xa0820820,0x8a20000,0x820a000,0x800a0082,0x20020208,0xa28a282,0x82020022,0x8028008,0x282082,0x880);
remainders_neg[369].modulo_value=492; remainders_neg[369].bit_mask=create_bit_mask(15,0x20808a2,0x20828028,0x8000a288,0x22020000,0x800020,0x8208a280,0x200a00a0,0x228808,0x282088,0x820802,0x28828a00,0x8220220a,0x28a08a2,0x820800,0x8820a002);
remainders_pos[370].modulo_value=1480; remainders_pos[370].bit_mask=create_bit_mask(47,0x288a020a,0x2a2a0200,0x88a20a8,0xa022882,0xa2888008,0x8a2a80,0x2208222,0xa080080,0x28022,0x2228202a,0x280a8a,0x80220a82,0x822a0,0xa202020,0x2280aa20,0x88a82080,0x220a02a2,0x82a0,0x8a80800a,0x8000a028,0x8200882,0x22200020,0x22808028,0x2028800,0x88828,0x20082008,0xa000282,0x202a228,0x820000a0,0x80a0880a,0x82a228a,0xaa028802,0x808a008,0x88200008,0xa088020a,0xa0280082,0x82888a2,0x28000a8,0x20a088,0x82088002,0xa8a20088,0x2228a02,0x2880a020,0x8a22082,0x80a8a82a,0x80a22800,0xa0);
remainders_neg[370].modulo_value=1480; remainders_neg[370].bit_mask=create_bit_mask(47,0xa0020a82,0xa2228a80,0x2a020,0x8282200a,0x22000800,0x88022208,0x8a280aa2,0x2808008,0x888208a2,0xa2a0a822,0x88a00202,0x82a8202,0x800a228,0x82a0a8a8,0xa2082228,0x202808,0xaa028a22,0x8880228,0x2000882,0x882820,0x80a8000a,0x2a2888a0,0x2a0800a0,0x8a820088,0x80800020,0xa8802880,0x82008a02,0xa8a22a0,0xa808828,0x280002,0x80a22a02,0x220a0002,0x802080,0xa08880,0x20008a02,0x20a0080a,0x80200022,0xa088020,0x88a02800,0x280080a,0x202a0800,0x8a2a0282,0x200820a8,0x8022a80a,0x202022,0x82a2080,0x20);
remainders_pos[371].modulo_value=1484; remainders_pos[371].bit_mask=create_bit_mask(47,0x22800222,0x2a08a2a2,0x20028288,0xa20a0020,0x20a2800a,0x8a282220,0xa002a0a8,0x8280880a,0x220088a2,0x22aa028a,0xa020a08a,0x802a28a,0x8282800,0x8800082,0x20a8802,0x22280a00,0x200a8802,0x88208020,0xaa002808,0x8a820800,0x80aa0a8,0x8008022,0xaa80880,0x2aa0208,0x20020022,0xaa020880,0x82a22a0,0x800aa002,0x20822202,0x2a008080,0x2888802,0x2a08000a,0x220802,0x82820820,0xa8020002,0xa080aa28,0xaa88a20,0x20088a28,0x202828a2,0xa800aa02,0x828a22a0,0x28a08088,0xa08aa00,0x28008080,0xa20a8228,0x2888a8,0x888);
remainders_neg[371].modulo_value=1484; remainders_neg[371].bit_mask=create_bit_mask(46,0xaa088a2a,0x22802a22,0xa08a0a00,0x2a0288a8,0x28220882,0x2a02aa8,0x288a2020,0xa080002,0x2a880022,0xa2228a02,0x28282802,0x80822a02,0x80a02088,0x8008800a,0x8a82000a,0x2aa08280,0xa082008a,0x2808a8,0x2280a000,0x20a0088,0x80822020,0x8088082a,0x2208000,0x82228a80,0xa00a88aa,0x22828008,0x80a22a28,0x882208a,0xa80aaa0a,0x22880800,0x8200008a,0xa2000882,0x88a2808a,0xa0a00a8,0x208a808a,0x28082220,0x22002a0,0xa08002a0,0xa820a022,0x2080228a,0xa022a28,0xa0280000,0x82802208,0x20880800,0x22020aa0,0x88200020);
remainders_pos[372].modulo_value=744; remainders_pos[372].bit_mask=create_bit_mask(24,0x228a0882,0x220000,0x208200a,0x2802082,0x20000220,0xa08828a,0x8a2822,0x28008800,0x8082080,0x800a0082,0x28800008,0x8828220a,0x228a0,0xa00220,0x8202082,0x22002802,0x28a20000,0x8220a088,0x800008a2,0x8028008,0x8208082,0x8800a0,0x20a28800,0x82);
remainders_neg[372].modulo_value=744; remainders_neg[372].bit_mask=create_bit_mask(24,0x22020002,0x8a28808,0x8a00a082,0x82082802,0x20808a28,0x82000202,0x800220a2,0x20800008,0x8000a008,0x820802,0x20008800,0x20a282,0x80822020,0x8208a28,0x8028a00a,0xa2882082,0x20220808,0xa282000,0x880022,0x820800,0x8028000a,0x80000820,0x28220008,0xa);
remainders_pos[373].modulo_value=373; remainders_pos[373].bit_mask=create_bit_mask(12,0xfa63369a,0x889b43f0,0xefda0ad5,0x1fbdbcb0,0x57190644,0x212a8639,0x3aa71855,0x7e089826,0xfdc34f6f,0x446ad416,0x97c3f0b6,0x165b31);
remainders_neg[373].modulo_value=1492; remainders_neg[373].bit_mask=create_bit_mask(47,0x22aaaa02,0x28a28,0x22028208,0x82202028,0x8a808a88,0xa8a20aa0,0xa22a8088,0xa28000aa,0x200a82a2,0x88a2088a,0xa0a282a,0xa8a82a8,0xa28280aa,0x82288aa0,0xa802288,0xa2a2a,0xaa22aa8,0x8228a888,0x880a882a,0x2022088a,0xa0220a0,0x8a000082,0xaaaa20a2,0x2a202,0xa0820a20,0xa0a28aaa,0x2a2a0a22,0xa88088a2,0x22802880,0xa88808,0xaaa80022,0x80a02020,0x2288822a,0x820a0088,0xa00a82,0xa0a80280,0x8802820,0xaa2088a2,0x82020280,0x22000aaa,0x8088a80,0x808a00a2,0x2288808a,0xa2282a2a,0xaaa8a282,0x2282082,0x2a000);
remainders_pos[374].modulo_value=1496; remainders_pos[374].bit_mask=create_bit_mask(47,0x82a02222,0xa022220,0x282028a,0xa0280a20,0x88a02a2a,0x22a80200,0x8a882,0xa22a20a,0x2882a082,0x20a028a8,0x280aaaa0,0x2aa28a,0x282000,0x2a0a0800,0x882000,0x220028a,0xaa2082,0xa8002288,0xa008a22,0xa8020020,0x2202a820,0x2080a082,0x22028222,0x88088220,0x8888280,0x88820a02,0x2a082a80,0xa0080080,0x2a88a200,0x228800,0x808208aa,0xa28008,0xa8000822,0x20a0,0x828,0x28a28aa8,0x80aaaa0,0x282a280a,0xa0820a82,0xa08a88,0x88822a20,0x2200802a,0xaa8a80a,0x8008a028,0xa0a28082,0x82088880,0x88880a);
remainders_neg[374].modulo_value=1496; remainders_neg[374].bit_mask=create_bit_mask(47,0xa28a2aa,0x8202aaa8,0x8a0a8a02,0x282082a0,0x2822a2,0xa2208a88,0x8880200a,0x2aa2a02,0xa002280a,0x2828a020,0xa0822220,0x88a22202,0x8820a808,0xa2828088,0x8800a080,0x8aa80a02,0x8022280a,0x2088aa00,0x8280022a,0x208288a8,0x2a0a20a8,0xa808280a,0xa28a02aa,0xaa8,0xa08,0xa8282,0xa2802208,0x20808800,0xa2002a88,0x88aa0008,0x8028022,0x82a0080,0x208880aa,0x88882028,0x880880a0,0xa0220220,0x80822220,0xa0222082,0x200a820a,0x80280200,0xaa228,0x2a8008a2,0x2202082,0x88028a0,0x282a0002,0xa000008,0x2);
remainders_pos[375].modulo_value=60; remainders_pos[375].bit_mask=create_bit_mask(2,0x20882,0x8220800);
remainders_neg[375].modulo_value=60; remainders_neg[375].bit_mask=create_bit_mask(2,0x808a0002,0x228000);
remainders_pos[376].modulo_value=376; remainders_pos[376].bit_mask=create_bit_mask(12,0xaa82a22a,0x80a2080,0x22aa2822,0xa2282002,0xa8a00a8,0x220280,0xa0028088,0x8a2a00a2,0x88800828,0x208828aa,0xaa0208a0,0xa88a82);
remainders_neg[376].modulo_value=376; remainders_neg[376].bit_mask=create_bit_mask(12,0x220a2aa2,0x80822808,0xaa22a0aa,0x2aa0a88a,0x82028820,0x88aa8a08,0x288a0800,0x2a2802a,0x880a0,0xa800a022,0x228a8028,0x20020a);
remainders_pos[377].modulo_value=377; remainders_pos[377].bit_mask=create_bit_mask(12,0xc2ed8b16,0x502ed679,0x53361741,0x13800cde,0x1151a83,0x31af2d3c,0xf2d3d6,0x230562a2,0x29ecc007,0x2a0ba1b3,0xe79add0,0x1a346dd);
remainders_neg[377].modulo_value=1508; remainders_neg[377].bit_mask=create_bit_mask(47,0xa20028a,0x88a20a20,0x822a8a88,0x8a088002,0x8888020a,0xa822a0a0,0x882a0202,0xaa08aa2a,0xa020088a,0x2282202a,0x2a02008,0xa02a8a00,0x2a00028a,0xa208020,0x202aa022,0x82aa8880,0xa08aa2a8,0x20aa0202,0x8888282a,0x828a8200,0xa20a0008,0x28888a8a,0x22802282,0x20828a00,0x802a022a,0x80808822,0xaaa08000,0xa2808208,0x20288880,0x82800a20,0x2000a02,0x88000088,0xa002a22,0x2a0820a2,0x82a8a02a,0xa0802200,0x2a2a028a,0x28228208,0x8a22a00,0x20088000,0xa0a02800,0x2022800,0xa280888a,0xaa882080,0x80800082,0xa20880,0x822a202a);
remainders_pos[378].modulo_value=168; remainders_pos[378].bit_mask=create_bit_mask(6,0x220a2802,0x20208200,0x2008000,0x82080800,0x28a08800,0x80);
remainders_neg[378].modulo_value=168; remainders_neg[378].bit_mask=create_bit_mask(6,0xa2822002,0x28a00a00,0x82080088,0x2000080,0x20200008,0x8);
remainders_pos[379].modulo_value=1516; remainders_pos[379].bit_mask=create_bit_mask(48,0x8a208aaa,0xa882aa2a,0x20022880,0x2a00822a,0x828228a,0x2282a228,0xa2a020a2,0xa80882,0x8aa8000a,0xa02aa8a,0xa22208,0x2820a88,0x2aa2028a,0x82a28888,0x202000a,0x2a82aaa0,0x80a22a0,0x88828a8,0xa2a0880a,0x808208a0,0xa8808a2a,0x202a208a,0x8200800a,0x228,0x20082288,0x8a808a00,0x2022aa20,0x8202a8a,0x20a8a0a2,0x82220a02,0x8a0202a2,0xa82a80a8,0x80800aa,0x28a82a00,0x8aa8222,0x8280a28,0x88a0022a,0xa80a0208,0x2aa2a2a,0x2a00a00,0x80a8a822,0xa82888a0,0x82820288,0x200a8a28,0x88008a88,0xa822a022,0x208a2a8a,0xaaa);
remainders_neg[379].modulo_value=1516; remainders_neg[379].bit_mask=create_bit_mask(48,0x2a80222,0x200a22a2,0xa88aa008,0xa2880aa2,0x80a0aa02,0xaa0a2aa0,0x2a28a82a,0x8820800a,0x2208882,0x828a2202,0x882aaa80,0x820a8200,0xa22a8a02,0xa2a0000,0x8a8a8882,0xa20a2228,0x8082aa28,0x8000a020,0x2a280082,0x80a8028,0x200802a2,0xa8a2a802,0xa880882,0x88888aa0,0xa880aa00,0x2080288,0xa8aa22a8,0x80a8a202,0xa820282a,0xaaa828a,0x28a8a2a,0x20a20820,0x80808822,0xa020a288,0x80220aaa,0x80a082a0,0x288aa2,0x20828a80,0x8a22a2a2,0x8a288288,0x82020aa,0x20a00028,0xa0a8a00,0xa88202a0,0x880200,0x20aa28aa,0xa802a202,0x222);
remainders_pos[380].modulo_value=760; remainders_pos[380].bit_mask=create_bit_mask(24,0x80802282,0xa82a8822,0x88a8080,0xa220022,0xa00000,0x808220a,0xa088202,0xa0,0x880088a,0xa0202a22,0x28822a82,0x28288020,0x28080228,0xa82a882,0x2088a808,0xa22002,0xa00a0000,0x20808220,0xa08820,0xa000000a,0x20880088,0x2a0202a2,0x28822a8,0x828802);
remainders_neg[380].modulo_value=760; remainders_neg[380].bit_mask=create_bit_mask(24,0x8002a0a,0x202200a2,0x20008,0x82a288aa,0x80288808,0x80802282,0x82000a82,0x8088028,0x80008002,0x20a8a22a,0xa00a2202,0xa02008a0,0x208002a0,0x8202200a,0xa0002000,0x882a288a,0x28028880,0x28080228,0x882000a8,0x20808802,0xa8000800,0x220a8a22,0xa00a220,0x2008a);
remainders_pos[381].modulo_value=1524; remainders_pos[381].bit_mask=create_bit_mask(48,0xa2882022,0x28220020,0x8208a202,0x228800a0,0x28020200,0xa002288,0x80080,0xa08200,0x82008202,0x208228a2,0x8000220,0x200a280,0x800a2802,0x20220000,0xa28a208,0x2020802,0x20828,0x80080080,0xa08a2002,0x20020200,0x828008a,0xa2022882,0x28200220,0x8a20208a,0x828a2080,0x8a08800,0x82082288,0x8a2002,0x20a00808,0x28008a,0x2002,0x8028208,0x8a080208,0x808208a2,0x200008,0x808028a,0x20028a0,0x20808800,0x828a288,0xa0080820,0x820,0xa002002,0x2822880,0x28800808,0x820a002,0x828808a2,0x28a08008,0x88082);
remainders_neg[381].modulo_value=1524; remainders_neg[381].bit_mask=create_bit_mask(48,0x220028a2,0x20a28828,0xa00228a,0x22000820,0x20828a08,0x8208a200,0x80800800,0x8200a08,0xa08028a,0xa00a2022,0x808a28,0x8a082208,0x822082,0x28a28808,0x82202280,0x828a0082,0x8828020,0x8008008,0x20022882,0x20828a08,0x80208002,0x228a2002,0x20a08a28,0x228a002,0x2022800,0x200008,0xa00a200,0x80022882,0x28208000,0x88208002,0x80882882,0x820a00,0x2008280,0xa0022,0x8a08800,0x80008202,0x82882020,0x28000008,0x80202200,0x208000a0,0x8808028,0x8208a08a,0x820a2000,0x20008000,0x8028208a,0x2000022,0x20200800,0xa);
remainders_pos[382].modulo_value=1528; remainders_pos[382].bit_mask=create_bit_mask(48,0xaa22028a,0xa8aa8820,0xa082088a,0x82888822,0xa288a002,0x20a2a2a8,0x22a2a0a,0x288a002a,0x2a800028,0x2000a80a,0xa802202,0x2028aaa2,0xa200082a,0x22a22a80,0x20a80aaa,0x282aaa80,0xa02aa088,0xa02022,0x802020a,0x82aa0882,0x2288882,0x88a28a,0xa288008,0x82a220,0xa0088a82,0x200228,0x80a28a22,0x82822228,0x208220aa,0xa08080,0xa88080a,0x28220aa8,0x802aaa8,0x88aaa02a,0x8a02a88a,0x8a82000,0xa08aaa28,0x8808802,0xa8a02a00,0x28280002,0x80a800a2,0x8a0a8a8,0x8a2a8a8a,0x82800a22,0xa882222,0x2aa22082,0xaa0822aa,0xa28088);
remainders_neg[382].modulo_value=1528; remainders_neg[382].bit_mask=create_bit_mask(48,0x22aa8a02,0x202200a8,0x280a8002,0xa0000aa,0x2a00288a,0x282a2a20,0x8aa2a282,0xa00288a2,0xa20888a0,0xa8882082,0x8208aa8a,0xa8a0222a,0x2a8880a2,0xaa2aa208,0xa8208222,0xa0a22208,0x28a22800,0x8828a8aa,0x808a8a82,0xa22800a,0x8aa0000a,0x88002a02,0x82a00880,0x880a2aa8,0x2880020a,0x88a88aa0,0x82a02aa,0xa0aaaa0,0xa80aa822,0x80280808,0x82008082,0xa0aa8220,0x808a2220,0x2228a2,0x28a2002,0x8020a888,0x280222a0,0x8008008a,0x2028a288,0xa0a0888a,0x820882a,0x80282020,0x2a20202,0xa0882aa,0x8200aaaa,0xa22aa80a,0x2280aa22,0x2a0800);
remainders_pos[383].modulo_value=1532; remainders_pos[383].bit_mask=create_bit_mask(48,0x22228a02,0xa028088,0xa80a8222,0x8a0a8828,0x8aa002aa,0xaa8822aa,0x2a82282,0x808080a2,0x88288a28,0xaaaa282a,0x88008080,0x8a882aa,0xa2022,0x2a2a2aa2,0xa08000,0x208220a2,0x882a2a2a,0xa08a00a,0x8a200,0xaa802,0x20a20282,0x888a02a0,0xa22a0a82,0xa820888,0xa222082a,0x282a0a88,0x80a80a22,0x82808a0,0x802aa00,0x8a200,0x2a00a20a,0x8a8a8a82,0x8a08820,0x800020a0,0x8aa8a8a,0x8808a00,0x2aa822a2,0xa0202002,0x2a828aaa,0x228a2282,0x8a02020,0xa82882a8,0x2aa8822a,0x2aa800aa,0xa2822a0a,0x8882a02,0x8220280a,0x80a2888);
remainders_neg[383].modulo_value=1532; remainders_neg[383].bit_mask=create_bit_mask(48,0xaaaa028a,0x828a0800,0x20820aaa,0x28200a0,0x2288a22,0x2200aa22,0x8a20aa0a,0x808082a,0xa002a0,0x2222a0a2,0x880808,0x200a22,0x8882a8aa,0xa2a2a22a,0x88280888,0xa80aa82a,0xa2a2a2,0x82802882,0x88802a88,0x8882208a,0xa82a8a0a,0x28a28,0x2aa2820a,0x820a8000,0x2aaa80a2,0xa0a28200,0x82082aa,0x80a08028,0x808a2288,0x88802a88,0xa2882a82,0x202020a,0x802800a8,0x888a828,0x80220202,0x80080288,0xa220aa2a,0x28a8a88a,0xa20a0222,0xaa02aa0a,0x8028a8a8,0x20a00a20,0xa2200aa2,0xa2208822,0x2a0aa282,0x8000a28a,0xaa8a082,0x82a000);
remainders_pos[384].modulo_value=24; remainders_pos[384].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[384].modulo_value=24; remainders_neg[384].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[385].modulo_value=1540; remainders_pos[385].bit_mask=create_bit_mask(49,0x808020a,0x22008a00,0x20080,0x880008a2,0x2000280a,0x80880a00,0xa0200022,0xa0200,0x208082,0x220880a,0x28020802,0xaa228202,0x20a22a8,0x8200a2,0x80882202,0x8082002,0x2028002,0x20002028,0x280880a,0x80a00020,0x2880008a,0x8020002,0x2880220,0x82008080,0x2020082a,0x88022800,0x80200,0x20002288,0x8000a02a,0x2202800,0x8080008a,0x280802,0x820208,0x8822028,0xa0082008,0xa88a0808,0x8288aa2,0x2080288,0x2208808,0x2020800a,0x80a0008,0x800080a0,0xa022028,0x2800080,0xa200022a,0x20080008,0xa200880,0x8020200,0xa);
remainders_neg[385].modulo_value=1540; remainders_neg[385].bit_mask=create_bit_mask(49,0x80800202,0x2a080280,0x8a8008,0x8022,0xa0802002,0x288,0x282880a2,0x8828208,0x8a00802,0x2a80002,0xa00a008a,0x22220282,0xa022220,0x8802802a,0xaa02,0x802882,0x820a0882,0x2808a0a0,0x8a000002,0x200828,0x20080002,0x800a8802,0xa0082a0,0x2000808,0xa0288022,0x822080,0x88000a80,0x2088a200,0x8802822,0x2a8a000,0x802,0x8208082,0x8028280,0x800220a0,0x2000a000,0x20020000,0x200222,0x2800200,0x82200080,0xa0a00802,0x80820800,0x8000020,0x28aa020,0xa008800,0x22888222,0xa8000880,0x82208000,0x8a0280,0x2);
remainders_pos[386].modulo_value=1544; remainders_pos[386].bit_mask=create_bit_mask(49,0xa2882aa2,0xa0aa2028,0x88022002,0x2a28000a,0x8a2a222,0x8202aa0a,0xa228a80,0xaaa2a00a,0x8882a2a,0x88aa8802,0x28aa20a2,0xa22802a0,0x2a82202,0xa20082a,0xa8800882,0x202888a2,0xaa020000,0x808220a0,0xa002a282,0x220208a0,0xaaa82022,0xaa22a880,0x2a200a02,0x288208,0x82280082,0xa008a820,0x2a88aa80,0x82aaa02,0x20808888,0x8a800a0a,0x8820282,0x80aa0a,0x22280800,0x20022a8a,0x2008a082,0x882a80a8,0x80288a80,0x8aa280a,0x22aa228a,0xa8222080,0xa8aaaa8,0xa288a0a0,0xaa808202,0x8a8a20a0,0x28a888,0x88022a0,0x8aa0a80,0xa8228a28,0x8a);
remainders_neg[386].modulo_value=1544; remainders_neg[386].bit_mask=create_bit_mask(49,0x2a00a22a,0x2822a8a0,0x8aa88a,0xa2a08882,0x802a2aaa,0xa8a2282,0x82aa0208,0x222a2882,0x8000a2a2,0x22008a,0xa022a82a,0x2aa08a28,0x8a20aa8a,0x82a880a2,0x2008800a,0xa8a0002a,0x228a8888,0x80aa828,0x288a2a02,0xaa8a8028,0x2220a8aa,0x22aa2008,0xa2a8828a,0x88a00a80,0xaa0880a,0x288020a8,0xa2002208,0x80a2228a,0xa8080000,0x2088282,0x800a8a0a,0x88082282,0xaaa08088,0xa88aa202,0xa880280a,0xa20820,0x8a00208,0x8022a082,0xaa22aa02,0x20aaa808,0x82022220,0x2a002828,0x22080a0a,0x202a828,0x88a02000,0x8008aa28,0x80228208,0x20aa02a0,0x2);
remainders_pos[387].modulo_value=516; remainders_pos[387].bit_mask=create_bit_mask(17,0x20a2082,0xa20200,0x8000080,0x2882022,0x8820028,0x8228a208,0xa00800a2,0x20800828,0x82002080,0xa00200a2,0x8a28828,0x80082202,0x80822802,0x20000208,0x808a000,0x208a0800,0x8);
remainders_neg[387].modulo_value=516; remainders_neg[387].bit_mask=create_bit_mask(16,0x82822802,0x8228208,0x80088008,0x820028a2,0x28820,0xa202280,0x20000822,0x28008020,0xa08a008,0x208a0822,0x220020,0x800a28a,0xa2082,0x28808a00,0x80002088,0xa0020080);
remainders_pos[388].modulo_value=776; remainders_pos[388].bit_mask=create_bit_mask(25,0x8a000a0a,0x2022880a,0xaa228a02,0x28a22a8,0x808a2022,0x8280028a,0x2800282a,0x808a202a,0xa88a2808,0xa288aa2,0x2288088,0xa000a2a,0xa000a0aa,0x22880a8,0xa228a022,0x28a22a8a,0x8a20220,0x280028a8,0x800282a8,0x8a202a2,0x88a28088,0xa288aa2a,0x22880880,0xa000a2a0,0xa0);
remainders_neg[388].modulo_value=776; remainders_neg[388].bit_mask=create_bit_mask(25,0x2888282,0xa8aa0082,0x22aa028a,0x8a02aa20,0x802a8aa,0xa088a02,0xa088a0a2,0x802a8a2,0x2002a080,0x82a00222,0x8aa00800,0x828882a2,0x28882822,0x8aa00820,0x2aa028aa,0xa02aa202,0x802a8aa8,0xa088a020,0x88a0a20,0x802a8a2a,0x2a0800,0x2a002222,0xaa008008,0x28882a28,0x28);
remainders_pos[389].modulo_value=389; remainders_pos[389].bit_mask=create_bit_mask(13,0x531b2af2,0xccd27618,0xe8e3f63f,0xa7a7845b,0xa241f222,0x9bdab50,0xb56f6412,0x13e09142,0x88797951,0x1bf1c5f6,0x1b92ccff,0xd5363286,0x13);
remainders_neg[389].modulo_value=1556; remainders_neg[389].bit_mask=create_bit_mask(49,0x8a82a22a,0xaaa80,0x202a2aa2,0x2a2a0882,0x2a882aaa,0x80202288,0x28a2a88a,0x8a28088a,0xa0a088,0x8228082a,0x820a0022,0x8aaaa0a,0xaaa8828a,0x2820a82,0x80a20a20,0x282802a0,0x80a28888,0xaa28a288,0x20200a88,0xa08aa08a,0x82a2a2aa,0xa2a02208,0xaa80022a,0x2a0a880a,0xa02202a2,0x8000a880,0x20022aa,0x288a22a,0x8a000202,0x2a208a00,0x200880aa,0xa888208,0x28288882,0xa8a02a8,0x82aa20a2,0x280a2,0x8a00288,0xa280a000,0x22822aa0,0xaa028a8,0x20888a0a,0x82088a8,0x2a808802,0x8028822a,0x20200028,0x2a2288a0,0xaaa20020,0x808a8000,0x2202);
remainders_pos[390].modulo_value=1560; remainders_pos[390].bit_mask=create_bit_mask(48,0xa0020802,0x8028200,0x2000208,0x820228a2,0x20800820,0x8800008a,0x80022,0x220,0x220a00a,0xa0882022,0x28200000,0x80288200,0x20020,0x8200820,0x8020a,0x22000,0x28028020,0x8082200,0x82800022,0x800000,0x200208,0xa2000000,0x808020,0x2200082,0x80080020,0x820008,0x8a080202,0x208008,0x82000002,0x20880002,0x28008820,0x80280,0x880,0x20a08020,0x82008,0x2080080,0x28008228,0xa000008,0x80880822,0xa00a08,0x88000,0x22880020,0x8a20000,0x82082002,0x808a2880,0x20208000,0xa008280,0x802080);
remainders_neg[390].modulo_value=1560; remainders_neg[390].bit_mask=create_bit_mask(49,0x208a0082,0x20a00,0x2088280,0x2822022,0x20000028,0x80002,0x800008a2,0x8008228,0x8a202082,0x20002022,0x20a00808,0x8200280,0x8200a0,0xa08028,0x80008202,0x808a2880,0x20020820,0x2288,0x20008a2,0x8800,0x88280280,0x22080880,0x8000028,0xa20800a,0x800820,0x8020800,0x2008282,0x880000,0xa00800,0x208800a,0xa0000882,0x20000020,0x8008200,0x80000000,0x20200828,0x88002080,0x2000800,0x20800220,0x82008080,0x22,0x8200200,0x80000080,0x220000a0,0x220808,0x200a00a,0x22000,0x28200800,0x2080208,0x2000);
remainders_pos[391].modulo_value=1564; remainders_pos[391].bit_mask=create_bit_mask(49,0x8a08a22a,0x220228a0,0xa00a288a,0x2222a0a2,0x808088a,0x82a00a88,0xa80a08,0xaaa0082,0xa0800082,0x82a28008,0xa80aa2a0,0x80880a8a,0xa22a020a,0x2800002,0x80a2008,0xaa2802a2,0x80000a2a,0x20288a2a,0xa8022282,0xa2828202,0x88222,0x80a80828,0x82888020,0x2022820,0x2200088,0x20828808,0x20802228,0x28202a0,0xa8882200,0xa8082828,0x88288802,0x2a8a2280,0xaa8a0000,0x8a8028a,0x2008a02,0xa8000028,0x2a080a88,0xaa2a0220,0x20a8aa02,0xa20028a8,0x8200020,0x8200aaa,0x220a02a0,0x22a00a8,0x8a220202,0xa8a0a888,0x8a228a00,0x20a28808,0xa88a20a);
remainders_neg[391].modulo_value=1564; remainders_neg[391].bit_mask=create_bit_mask(49,0x2002aa2,0xaa82a028,0x2882a002,0xaa22282a,0x80808002,0x2288200,0x88200280,0x222880a,0x2808880a,0xa2a0080,0x20822a20,0x8008202,0x2aa28a02,0x8a08888a,0x8082a080,0x22a08a22,0x88802a2,0x28a002a2,0x2082aa0a,0x2a0a0a8a,0x80000aaa,0x82080a0,0x20008a8,0x8a0aa0a8,0xaa88800,0xa80a0080,0xa800aaa0,0x8a0a8a20,0x2000aa88,0x20802020,0xa0008a,0xa202a208,0x22028088,0x80200a02,0x8a88028a,0x20808820,0xa2808200,0x22228aa8,0xa8202282,0x2288a020,0xa888a8,0xa88222,0xaa828a28,0x82a28820,0x2aa8a82,0x20282000,0x22a0208,0xa82a0080,0x2002282);
remainders_pos[392].modulo_value=56; remainders_pos[392].bit_mask=create_bit_mask(2,0x82828202,0x808282);
remainders_neg[392].modulo_value=56; remainders_neg[392].bit_mask=create_bit_mask(2,0xa0a0a0a,0x80a02);
remainders_pos[393].modulo_value=1572; remainders_pos[393].bit_mask=create_bit_mask(50,0x22822082,0x20828800,0x88080080,0x2802000,0x8a00200,0x200a202,0x200a2822,0x8000008,0xa208202,0x828a2,0x8a28220,0xa00a280,0x28228a2,0x28800a28,0x8000220a,0x228228a0,0x820208,0x80208000,0x20220a2,0x28820228,0xa000200,0x80880020,0x8200800,0xa20008a,0x820a0822,0x20008a08,0x200820a,0x80022020,0x8000a00,0x88082280,0xa0880802,0x208028,0x8082000,0xa2882882,0x8800020,0x8a00228a,0xa2882802,0x28a00a08,0x8828a200,0xa2820000,0x8208a08,0x208,0x828a0082,0x8a00808,0x800a208,0x802800,0x20020220,0x2282080,0x20882880,0x8);
remainders_neg[393].modulo_value=1572; remainders_neg[393].bit_mask=create_bit_mask(49,0xa20a2802,0x28020008,0x8008,0x82082880,0x208a00,0x8a08228a,0xa08220a2,0x808800,0x8228028a,0x80802022,0x220a28,0x82082208,0x820a2022,0x20008220,0x808a282,0xa20a2020,0x8028a00,0x8280088,0x828a2822,0x20028a20,0x82080288,0x8a0,0xa08008,0x82288002,0x28200a2,0x28800200,0x8a080282,0x8a28a0,0x808208,0xa208,0x20000082,0x8a00820,0x8000a088,0x22002002,0x8828,0x208a202,0x22002082,0x20208200,0x202288,0x220a0880,0xa00200,0x88088280,0x2020802,0x208000,0x80082280,0x80082080,0x28828a28,0x8a20a008,0xa0002000);
remainders_pos[394].modulo_value=1576; remainders_pos[394].bit_mask=create_bit_mask(50,0xaa0aaaa,0x8082a28a,0x802282a,0xaa28020a,0x28aa88,0xa0088222,0x8008a20a,0x20288aaa,0x882a2002,0x2a880a0a,0xa2808228,0x20a20a02,0xa222222,0x8202a0a2,0x8028228a,0x2880a0a8,0x2a22aa20,0xaa800088,0xaa002208,0x2aa22288,0x20088008,0xa8a0a2a8,0x8a802822,0xa08202,0xa,0x820a00a0,0x2802a280,0x8a0a2a88,0x220082a,0x888aa820,0x8800aa22,0x2aa20,0xaa88a822,0xa022808,0x8828022a,0xa8082a2,0x8888a08a,0xa08a0888,0x82028a80,0xa022a828,0x8a822a0,0xa2280880,0x8a2002aa,0x82200aa0,0xaa280088,0x8028aa22,0x288020a0,0x8a8202a8,0xaa0aa0a2,0xaa);
remainders_neg[394].modulo_value=1576; remainders_neg[394].bit_mask=create_bit_mask(50,0x82282222,0x80a2a02,0x808aa0a2,0x22a08a82,0x88a02200,0x28800aaa,0x8802a82,0xa8a00222,0xa2a88a,0xa2008282,0x2a080aa0,0xa82a828a,0x82aaaaaa,0xa8a282a,0x8a0aa02,0xa0082820,0xa2aa22a8,0x22088800,0x22882a80,0xa22aaa00,0xa8800880,0x20282a20,0x208a0aa,0x88280a8a,0x88888882,0xa828828,0xa08a2a08,0x282a200,0x8aa880a2,0x220a8,0x8822aa,0x888a22a8,0x220020aa,0x828aa080,0xa08aa2,0x82080a2a,0x2802,0x28028000,0xa8a0208,0x28aa20a0,0x8020aa28,0x2aa08008,0x2a88a22,0xaa88220,0x22a08800,0x8a022aa,0xa008a828,0x20a8a20,0x2282282a,0x22);
remainders_pos[395].modulo_value=1580; remainders_pos[395].bit_mask=create_bit_mask(50,0xa20202,0xa2200a2,0x2a8a2088,0xa020020,0x2200aa22,0xa0a02288,0x8a0a88a2,0x8820200,0x8a00a080,0x80a0222a,0x88202a82,0xaa2082a2,0x802a0,0x82000828,0x22080028,0x80000888,0x222a0822,0x20a8088,0x8002002,0x80a00828,0x8088000a,0x288a82,0x2822000,0xa802888,0x22088a08,0x802a020a,0x88280222,0x22800000,0x220282a,0xa02a00,0x80020282,0x2a080800,0xa888220,0x2882,0x2088222,0x82280,0x2000282,0x208aa0a8,0x808228a8,0x80a0282a,0xa00a2a88,0x8220020,0x2a0a2008,0x80a0a8a2,0xa0088228,0x80a088a,0x8a2a8080,0x88a0220,0x8a008a0,0x808);
remainders_neg[395].modulo_value=1580; remainders_neg[395].bit_mask=create_bit_mask(49,0x882a0a8a,0x822a8822,0x22022000,0x828288a8,0xa288222a,0x28282a00,0x2020022,0xa8288,0x2802808,0x28aa22,0xa8220a,0x22280a22,0x8808228,0xa8080a0,0xa2808820,0x8880000,0xaa2280a2,0xa820000,0x8080a88a,0x288020,0x8000882,0x88200202,0xa0aa008,0x8200a000,0xa2800200,0x8a20a82,0x208aa2,0x2a088088,0x8aa0a0a2,0x8028a208,0x88a0a0a,0xa2008080,0x20002a0,0x8880a00a,0x82800a2a,0x88802a08,0x8a088a02,0x28022020,0x802a020,0x28a022,0x28822200,0x2a88a0,0x2282a080,0x820202a,0x20800a20,0x80820002,0x2220800,0x282a8,0x80208028);
remainders_pos[396].modulo_value=264; remainders_pos[396].bit_mask=create_bit_mask(9,0x20800a2,0x220828,0x82088000,0xa0020802,0x20800a28,0x2208280,0x20880000,0x208028,0x8a);
remainders_neg[396].modulo_value=264; remainders_neg[396].bit_mask=create_bit_mask(9,0x82800022,0x8228020,0xa000088,0x208a0082,0x28000220,0x82280208,0xa0000880,0x8a00820,0x2);
remainders_pos[397].modulo_value=397; remainders_pos[397].bit_mask=create_bit_mask(13,0xee895e1a,0x3039d3e,0x3caeda2b,0xd7d0df5a,0x499731a,0xd00d3b89,0x2c02e619,0xa6482477,0xc2fad633,0xdd4f16be,0x70303516,0xa45ddf2e,0x161e);
remainders_neg[397].modulo_value=1588; remainders_neg[397].bit_mask=create_bit_mask(50,0x22008282,0x8a8a00a2,0xa02202a2,0xa080282,0x8800aa82,0x880a200,0xa08a2a80,0x2a80a8aa,0xa220aaa,0x828a22,0xa8a8a88a,0x288002a2,0xa20a8a82,0xaa2a0008,0x288a8a8,0x82228a08,0xa2aa8222,0x2aa8a80a,0x800aa288,0x88022808,0x820aa800,0x2a0a0082,0x8a2a0220,0x2228028a,0xa20a0802,0x2222a8a0,0x208082a8,0xa02a22a0,0xa2828aa0,0xa88aa800,0xaa88a822,0x288200,0x200a808,0x22822280,0x82a8a082,0x20080808,0xa208aaa0,0x88228000,0x80202aa,0x8aa08808,0x20a22082,0x802000a2,0x8a00080a,0x88aa8020,0xa88aa20a,0xa0a2800a,0x2a0282a8,0x808202a2,0xa2220aa0,0x2828a);
remainders_pos[398].modulo_value=1592; remainders_pos[398].bit_mask=create_bit_mask(50,0x2208202,0x28a08aa,0xaaa220a2,0x8a882020,0x8822a8a8,0xa8a22282,0xaa022a22,0x2a80282a,0x80200202,0xaa220a8a,0xa028aa0a,0xaa8a200,0xa80828a8,0xaa20800,0xa0082a,0x8080a220,0x22a2aa2a,0x282a80,0x20022a0,0x82822208,0x88080228,0xa2a2880,0xa20a2200,0xa008a082,0x82a28a22,0xa88a28a,0x8a820a20,0xa00088a0,0x220228a8,0x82288020,0x80208882,0xa8800,0x8802a828,0x2a8aa8a,0x88a02,0xa0a8200a,0x2a00208a,0xa02a2820,0xa008a2a,0xaaa0aa28,0x2a2a088,0xa8808008,0xaaa82802,0x2a88a880,0x2282888a,0xa22a2a88,0xaa080822,0x808a088a,0x80aa20a2,0x808208);
remainders_neg[398].modulo_value=1592; remainders_neg[398].bit_mask=create_bit_mask(50,0x8aa80a8a,0x8a028022,0x222aa82a,0x200a8a8,0xaa2020,0x202aaa0a,0x228aa22a,0xa208a0a2,0x8a88a8a,0x22aa8202,0x28a02282,0x82202a88,0x2080a020,0x822a8088,0x882880a2,0x8082aa8,0xaa2a22a2,0x88a0a208,0x8a88aa28,0xa0aaa80,0x808aa0,0x82a2a008,0x2a82aa88,0x2880280a,0xa2a02aa,0x82002a02,0x20a82a8,0x28880028,0xaa8aa020,0xaa008a8,0x8a8000a,0x88820080,0x8a20a0,0x8a202202,0x8880028a,0x2820a882,0xa288a802,0x28a2a0a8,0x828802a2,0x222822a0,0x8a2a2800,0x20080880,0x2220a08a,0xa2002008,0xaa0a0002,0x2aa2a200,0x228080aa,0x8028002,0x822a82a,0x80a80);
remainders_pos[399].modulo_value=1596; remainders_pos[399].bit_mask=create_bit_mask(49,0x22822822,0x8a00000,0x20800a,0xa20a0022,0x828800,0x88200280,0x20080080,0x20028008,0xa288280,0xa0000882,0x8008820,0x8a008082,0x2020880,0x8020200,0x2202,0x20028a0,0x20020000,0x82280008,0x22020080,0x8020820,0x820020a,0x8a0000,0xa00820,0x82288002,0x20000000,0x8002288,0x8200a0,0xa20,0xa080082,0x80820802,0x20200808,0x82000288,0x800,0xa28008,0x8088000,0x80802,0x20220808,0x820200a,0xa0822002,0x8220000,0x8028228a,0x82002800,0x20200200,0x280082,0xa0022820,0x8800a08,0xa002080,0x800000a2,0x8828828);
remainders_neg[399].modulo_value=1596; remainders_neg[399].bit_mask=create_bit_mask(50,0xa2022022,0x208800,0x280082,0x220208a2,0x8020008,0x280208,0xa0800000,0x28020800,0x2200200,0x20880002,0x20,0x200000a,0x828a0000,0x828208,0x8008a20a,0x2882020,0x20828808,0xa208000,0xa2820800,0x820028,0x80288282,0x20800,0x8208020,0x220008a,0x80080080,0x28808808,0x80082200,0x800a0020,0x8808228,0x2008002,0xa0082,0x28200000,0xa008200,0x80880080,0x8220000,0x80000008,0x800002,0x20a28000,0x8020a082,0x20022882,0xa20808,0x820a202,0x2882000,0x28a08200,0x8020800a,0x200a20a0,0x8200,0x82082008,0x880022,0x20020);
remainders_pos[400].modulo_value=40; remainders_pos[400].bit_mask=create_bit_mask(2,0xa8aa2a8a,0xa2);
remainders_neg[400].modulo_value=40; remainders_neg[400].bit_mask=create_bit_mask(2,0x20222202,0x22);
remainders_pos[401].modulo_value=401; remainders_pos[401].bit_mask=create_bit_mask(13,0x32554fb6,0x878ebb99,0x674f6361,0x6217e0dc,0x94a71401,0x14bfb255,0xf4a07b78,0x94a6a937,0xa11a00a3,0xcb98ec1f,0xc7861b1b,0xa9326775,0x1b7ca);
remainders_neg[401].modulo_value=1604; remainders_neg[401].bit_mask=create_bit_mask(50,0xaa88822a,0xa022200,0xaa82aaa8,0xaa8a2800,0x82a8888,0x88222a88,0x2828a2a0,0x82a20aa,0x2882882a,0x2002082,0xa0a8282,0x20aaaaa8,0x2a02a02,0xaaaaa822,0xa0a8280,0x8200202,0xa08a08a2,0xa822a082,0x2a28a0a2,0x8aa22088,0x888aa080,0xa28aa8,0xaaaa0aa8,0x2220280,0xa2088aa8,0x88a202,0x82a222a8,0xa0000a,0x820aa8,0x8a008888,0x8a22008a,0xa082028,0x8a022802,0x8a08a02,0xa2aa28a2,0x8280a0a2,0x2800000a,0xa02202a2,0xa22,0x8280a0a8,0x8a2aa2a2,0x28828822,0xa2028a0,0x2082820,0x8022288a,0x888028a8,0xa8208008,0x2800a,0xa222a0a8,0x2288800a);
remainders_pos[402].modulo_value=1608; remainders_pos[402].bit_mask=create_bit_mask(51,0x22002002,0x28820a08,0x80282200,0x828a2880,0x8a00a28,0x8a002202,0x82020082,0x20028a28,0x2002208,0x20800022,0x200a28,0x8a28020a,0x8208a0,0x20a00008,0x82280,0x228a0820,0x20228220,0xa208,0x82822080,0x20228000,0x280082,0x80000080,0x20200a08,0x80208008,0xa20a2020,0x8a08a00,0x2080208,0xa0080820,0x220,0x280002,0x2880882,0x8828200,0x8a000002,0x82880820,0x20a28808,0x88200008,0xa0802,0x20820020,0x8028a20a,0xa00800a0,0x20828,0x88008088,0xa2800820,0x808228,0x8800a282,0xa00a2080,0x28a28228,0x88280202,0xa0822800,0x8008820,0x80);
remainders_neg[402].modulo_value=1608; remainders_neg[402].bit_mask=create_bit_mask(51,0xa2882882,0x20028200,0x820a280,0x2022000,0x208220,0x208a28a,0x28a0802,0x28820220,0x8a08a280,0xa00808a2,0x8a00220,0x2208282,0x800a0020,0x28208800,0x8800a208,0xa20200a0,0x28a20a28,0x88082280,0x20a2800,0x28a20808,0x8820800a,0x880800,0x28a08200,0x8280080,0x228228a0,0x200208,0x8a008280,0x20800020,0x8808a28,0x8820808a,0x82000002,0x20a08,0x208808a,0x20000a0,0x28220000,0x208080,0x80820082,0x28028828,0x8202282,0x20800820,0x8828020,0x80000,0x220800a0,0x8000a20,0x8220a,0x20822800,0x20220a20,0x20828a,0x200a2080,0x800028,0x8);
remainders_pos[403].modulo_value=1612; remainders_pos[403].bit_mask=create_bit_mask(51,0xaa8028a,0x8a828820,0x2228aa8,0x82028020,0x200a2a22,0x280aa82a,0x820a2a0,0x880a222a,0xa8a,0x800aa002,0x8882200,0x28208a82,0xaa02800a,0x2288082,0x88a0220a,0xa002aa28,0x28a2a8a,0x88822202,0x8288080,0x8002a0a0,0x8002a080,0xaa088a08,0x20880a8,0x220a0200,0x8a00228,0xa20802,0xa088288,0x22080008,0x220aa2a0,0xa800220a,0xa8002020,0x228200a0,0x88222020,0x8a280808,0xa800aa2a,0x80a2228a,0x22880a08,0x280aa820,0x20828a00,0x8222082a,0xaa002008,0x800,0x8a022a2a,0xa0820a88,0xaa0280a8,0x8a008a82,0x2808288a,0x28880080,0x282a22aa,0x2aa0082,0xa28);
remainders_neg[403].modulo_value=1612; remainders_neg[403].bit_mask=create_bit_mask(51,0x2208a02,0x20a0028,0x82aa0220,0xa8a08a8,0xa08222aa,0xa08220a2,0x80a82a20,0x2aaa2,0x88088202,0x882208a,0xaa88,0xa0a8020a,0x22820882,0x8aa0080a,0x28aa02,0x208a22a0,0x8a022202,0xa2a8a,0x80a00808,0x88a2820,0x80a2008,0x22800280,0x8a800020,0x2a828a88,0x80288a20,0x8822808a,0x82800a00,0xaa808800,0xa2822a20,0x2088aa82,0x208828a8,0x2a0a8828,0xaaa8a0,0x2208080,0x208822a2,0x2aa202,0x2a008280,0xa08220a8,0xa8020288,0xa2a80a2,0x2288a800,0x80888088,0x28aa2a2,0x28020200,0x228a0820,0x2880202,0xa000a002,0xa0000808,0xa0a2a222,0xa22880a,0x2a0);
remainders_pos[404].modulo_value=808; remainders_pos[404].bit_mask=create_bit_mask(26,0x82aa2222,0x2a822,0x80a2a082,0x2a280a02,0xaa000208,0x22aa0a20,0xa8888a22,0xaa08a0a,0x8a820800,0xa0280a02,0x820a8,0xa82882a8,0x2228888a,0x82282aa2,0x820002a,0xa0280a2a,0x2082a280,0xa20aa000,0xa2222aa0,0xa0aa8888,0x8000aa08,0xa028a820,0xa8a0280,0x2a800082,0x88aa8288,0x88);
remainders_neg[404].modulo_value=808; remainders_neg[404].bit_mask=create_bit_mask(25,0xa22aaaa,0x888a20aa,0x82a280a,0xa2a0828a,0x22888a80,0xaa2282a8,0x200002aa,0x82280282,0x20a8088,0x28a0028a,0x8880a820,0x20a00a20,0xaaa00002,0xaa0a22a,0x80a888a2,0x28a082a2,0xa80a2a08,0x2a822888,0x2aaaa228,0x28220000,0x8882280,0x28a020a8,0x82028a00,0xa208880a,0x220a00);
remainders_pos[405].modulo_value=60; remainders_pos[405].bit_mask=create_bit_mask(2,0xa0080802,0x8020200);
remainders_neg[405].modulo_value=60; remainders_neg[405].bit_mask=create_bit_mask(2,0x20800082,0x28a00);
remainders_pos[406].modulo_value=1624; remainders_pos[406].bit_mask=create_bit_mask(51,0xa0aa22a,0x20822a0,0x82228a0a,0xaa208882,0x2000088,0x82282a80,0xa8002288,0xa8882202,0x22000880,0x2028000a,0xa820a02a,0xa0822a80,0x2222a20,0x20820220,0xa080a00,0x28280022,0xa2a200aa,0x28080808,0x2aa00282,0x8220880,0xa0a2a280,0x8a00080a,0x2008280,0x20a280a0,0x202802,0x800002,0x8802808,0x800a028a,0xa2028200,0xaa02000,0x20028a8a,0xa8022088,0x2882800a,0x8a202020,0x28aa008a,0xa0880028,0x800a020,0x80088082,0xa08a888,0x2a02a882,0x8a80a08,0x88a00028,0x2a022000,0x2a808822,0x82228800,0x8002a828,0xaa220000,0x82822208,0x80a0a288,0xa00a8820,0xa88aa0);
remainders_neg[406].modulo_value=1624; remainders_neg[406].bit_mask=create_bit_mask(51,0x82822a22,0xa80aa20,0x22a0282,0x2228000a,0x8a808800,0xaa02208,0x2088a200,0x2000aa0a,0x2a888000,0x20a08882,0x202828a2,0x2802a208,0x8aaa22a8,0xa80282a8,0x82808208,0x20a088a2,0x222a8822,0xa0008000,0xa2208a0a,0x80aa0008,0x202a2208,0x2888002,0xa880a00,0xa02a0828,0x8828208a,0x8800888a,0x80082080,0x8828202,0x2a8a0a00,0x228a880,0xa08a0202,0x200aa800,0xa0020882,0x2a828a8,0xa0228002,0x28008020,0x8828a0,0x80080a,0x2002000,0xa282200a,0x80200280,0x2880a0,0xa282a808,0x220800a2,0x2aa0088,0x80a20a0,0x22a28808,0xa0a2a80,0x8282200,0x20820028,0x200220);
remainders_pos[407].modulo_value=1628; remainders_pos[407].bit_mask=create_bit_mask(51,0xa282a282,0xaa220000,0x880a0000,0x88a02a80,0x22802a82,0x8028820a,0x2aa8aa8a,0x88022822,0x2a0a8a0,0x2088880a,0x2888820,0x82882288,0x2208aa80,0xa20a0a00,0x200aa200,0x222828a0,0x200a8282,0x8aa02020,0x8820,0x20820200,0x2a08808a,0x28020a00,0x82a22200,0x88002aa8,0xa0808aa,0xa0a00,0x2aa2020a,0x22aa8002,0x800888a8,0x800a0802,0x8a20220a,0x80820,0x20822000,0x808080aa,0x88282a00,0x80a28288,0xa008aa00,0x800a0a08,0x202aa208,0x2288228,0x80822228,0xa022220,0x20a2a0a8,0x88828802,0x2a2aa2aa,0x8a082280,0x282a8028,0x202a80a2,0xa0000a02,0xa000088a,0x828a828);
remainders_neg[407].modulo_value=1628; remainders_neg[407].bit_mask=create_bit_mask(51,0x2a0a220a,0x222a8888,0x828888,0x282200,0xaa08220a,0xa00a82,0xa2202202,0x8aa02a,0x8a202020,0x28000082,0x8a0000a8,0xa00a200,0xaa002208,0x2a828288,0xa8822a80,0xaaa02028,0xa0820a0a,0x228a0a8,0x88880028,0xa8028a88,0x22800802,0x208a8288,0xa2aa288,0x8a220,0x82808022,0x88828280,0xa2220a82,0xa222088a,0x8800020,0x882800a,0x2a0aa82,0x8808028,0xa80aa888,0x8080022,0x20a288,0x2a0a00,0x28802280,0x8820280,0xa0a22a80,0x8aa00aa0,0x80a2a20,0x8282aaa8,0x282a2820,0xa008a,0xa2a22222,0x200aa00,0xa0a208a0,0xa8a20822,0x2888028a,0x20088002,0xa020a0);
remainders_pos[408].modulo_value=408; remainders_pos[408].bit_mask=create_bit_mask(13,0x82080882,0x20228a20,0xa00a,0x20a2820,0x220028,0x80208200,0x2020080,0x8208,0x80280088,0x828a0,0x8a00a00,0x8208a288,0x822020);
remainders_neg[408].modulo_value=408; remainders_neg[408].bit_mask=create_bit_mask(13,0x2800002,0x28a20228,0x88082082,0x820220a0,0x8a28820,0x280288,0x828a0800,0x8800a00,0x8208000,0x80802020,0x208200,0xa002200,0xa2820);
remainders_pos[409].modulo_value=409; remainders_pos[409].bit_mask=create_bit_mask(13,0x4b97977e,0x106f6315,0x1e2f29b1,0x2b0834f5,0x8649cfa9,0x815c06dd,0x732fd33,0x86ed80ea,0x5257ce49,0xe2bcb043,0x223653d1,0x4aa31bd8,0x1fba7a7);
remainders_neg[409].modulo_value=1636; remainders_neg[409].bit_mask=create_bit_mask(52,0x820a0aa2,0x88a2aa88,0x82a2a028,0xa280a828,0xa800220,0x8808a00,0x8aaa20aa,0xa200822,0x8a8a0280,0x2a20288a,0xaaaa8a08,0x822a8200,0x20222a2a,0xa20aa2a2,0xaaa8020a,0x22a0828a,0x8a8a88a0,0x22800a02,0xaa8a2080,0x882a822,0xa800288,0xa282200,0x2a08a0a8,0x2888a02a,0x82088aaa,0x80222a82,0x88a282,0x2a088820,0xa08a020,0xa22820a8,0x82aa80aa,0x280288a8,0x8a208002,0xa0aa822a,0x8808082,0x828a022a,0xa2a80000,0x200a200,0x20222222,0xa2802280,0xa080000a,0x882a2028,0x82a08080,0x28aa20aa,0xa200082,0xa08a88a0,0x22aa80aa,0x280a820a,0x2a020288,0x80020888,0xa0a288,0x2);
remainders_pos[410].modulo_value=1640; remainders_pos[410].bit_mask=create_bit_mask(52,0xa00a0a82,0x800288a0,0xa2020,0x280282a,0x82200220,0x8aa2a88,0xa2a0a22,0x22220,0x8a288aa,0x82282802,0x80a82a08,0xa0200280,0x2820208,0xa00a0a2,0x2888222,0x20882008,0x8028820,0x2a800080,0xa22022,0xa02020,0x20880000,0x802a02a2,0x2a802028,0x8a822820,0xa2080002,0x820080,0x208a02,0x2882a280,0x802a808,0x80a80228,0x22088a,0x80a0000,0x88a0008,0x2a888,0x22802002,0x8220808,0x82228020,0xa00a088,0x8082808a,0x80080a20,0xa82a0202,0x28288220,0x228a2080,0x888000aa,0xa0a8a008,0xa8aa2088,0x80088222,0x28028008,0x8a000a8,0x22800208,0xa0a00a0a,0x82);
remainders_neg[410].modulo_value=1640; remainders_neg[410].bit_mask=create_bit_mask(52,0x2882020a,0x80a0020,0x882a0a8,0x8200a0a2,0x2a88a28,0x80222200,0x822282a2,0x88aa2a8,0x220022,0x2a0a00a,0x8202280,0x28288a00,0xa0a8280,0x8280282a,0x82000a22,0xa8002880,0x800a00a0,0x22088008,0x8822a8aa,0x8028a828,0xa8000888,0x8228a22,0x2208a0a0,0x202a0a8,0x2280800a,0x880a0808,0x88280282,0x200a2208,0x80822080,0x200a20,0x88aa0002,0x80028880,0x28080,0x88822000,0xa200a80a,0x80aa0080,0xa2a08a0,0x2882000,0x8020802,0x8228,0x20a20a8a,0xa0200aa0,0x2a02a008,0x8822,0x20202800,0x20222800,0x8000aa2,0x208a0080,0x80208820,0x22088a00,0x28280282,0x2);
remainders_pos[411].modulo_value=1644; remainders_pos[411].bit_mask=create_bit_mask(52,0xa2000822,0x28a20a20,0xa008208,0xa2822800,0x8220020,0x220028a,0xa0802082,0xa00200,0x8a280,0xa28a0882,0x8228a08,0x80202202,0x2822880,0x28220008,0x82088088,0x20000,0x28820200,0xa280002,0x200a08a0,0x8820828,0x82000288,0x80880,0xa28008,0x20820a,0x80080080,0x20008228,0x2002288,0x20800020,0x28a00a08,0x2000200,0x82022,0x8220228,0xa008282,0x28a00a2,0x8228800,0x8000008,0x22082000,0x8828220,0x88280200,0x80802022,0x28820808,0xa28a20a,0xa2000822,0xa00028,0x8020a008,0x880820,0x8820a28,0x8828a080,0x200a0002,0x8a28208,0x8a08a,0x882);
remainders_neg[411].modulo_value=1644; remainders_neg[411].bit_mask=create_bit_mask(52,0x228800a2,0x20228228,0x82080280,0x220a2080,0xa28828,0x8a288202,0x20082802,0x8208a08,0x88002208,0x22020002,0xa20200,0x828a28a,0x820a2000,0x20a28800,0xa000000,0x808a0880,0x20028a08,0x8220808a,0xa0820020,0x28020,0xa088200,0x80800000,0x8220800,0x88280282,0x800800,0x28800a20,0x8a08a200,0xa00808a0,0x20208200,0xa088288,0x808028a2,0xa28a20,0x8208020a,0x82020822,0xa20008,0x80088080,0xa2802880,0x20a28,0x208288,0x828a2,0x20028000,0x82202282,0x228800a2,0x8208820,0x8282080,0x800000a0,0x28220,0x202000,0xa0820882,0x220a00,0x88002002,0x2);
remainders_pos[412].modulo_value=824; remainders_pos[412].bit_mask=create_bit_mask(26,0xaa022a0a,0x200a8a8a,0xa0820888,0x8a28020a,0x2a8a222,0x8a800202,0x20280280,0x808002a,0x2888a2a8,0xaa08028a,0x82220820,0xaa2a2a00,0xaa0a880a,0xaaa022a0,0x8200a8a8,0xaa082088,0x28a28020,0x202a8a22,0x8a80020,0xa2028028,0x80808002,0xa2888a2a,0xaa08028,0x8222082,0xaaa2a2a0,0xa0a880);
remainders_neg[412].modulo_value=824; remainders_neg[412].bit_mask=create_bit_mask(26,0x228aa282,0xa8820202,0x280a8000,0x2a08a02,0x8a202aaa,0x2088a8a,0xa8a08a08,0x808088a2,0xa0002a20,0x22808a02,0xaaa80a8,0x22a2a288,0x22820082,0x2228aa28,0xa882020,0x2280a800,0xa02a08a0,0xa8a202aa,0x802088a8,0x2a8a08a0,0x808088a,0x2a0002a2,0x822808a0,0x80aaa80a,0x222a2a28,0x282008);
remainders_pos[413].modulo_value=1652; remainders_pos[413].bit_mask=create_bit_mask(52,0xa200a202,0x22a88002,0x822a82a0,0x8a2888a2,0x2000028a,0xa002208,0x82222a,0x2000802a,0xaa020a0,0x22000202,0x200888a8,0xaa08282,0x28820a0,0x8028208a,0x220a082a,0x20a88880,0x82020002,0x2028202a,0x2a00800,0x82a22208,0x20822002,0x8a8a0000,0xa2888a2,0x202a0aa2,0x2a0008aa,0xa022802,0xa880288,0x808aa200,0x8a08aa0a,0x2a28a222,0x2080000a,0xa8280088,0xa8020888,0x80800200,0x82a8082,0xa0880008,0x8802222,0x802a820a,0x280a2082,0xaa00a082,0x882820,0x882a222,0xaa080800,0x80a080,0x200a8020,0xa0a8888,0x820880,0x8a2a2800,0x8828a222,0xa880a82a,0x8a80022,0x808a0);
remainders_neg[413].modulo_value=1652; remainders_neg[413].bit_mask=create_bit_mask(52,0x2a882a0a,0x22200882,0x2a20a28,0x220002a,0xa8808a02,0x82882a80,0x880aa2a2,0xa8880822,0x228a820,0xa2888a0a,0xa8000020,0x82200a0a,0x8a002828,0x8a0a002,0xaa828022,0x28200000,0x282888a,0xa820a8a2,0x8a208088,0xa2a2a80,0xa80aa08a,0x2028808,0x2a00022,0x20a2822a,0xa2088022,0x8282a08a,0x82000a00,0x8022288,0x2802202,0x22a02aa2,0xa0088882,0x20208000,0x20828000,0x8080a88,0x80a2000a,0x28008800,0x8aaa2,0xa20a82,0xa002a80a,0x2280280a,0x880020a8,0x800a22aa,0x22808008,0x8082800,0xa08208a8,0x82020000,0x88028000,0x2a22088,0xa022aa,0x20082022,0x2088a2,0x8028);
remainders_pos[414].modulo_value=552; remainders_pos[414].bit_mask=create_bit_mask(18,0x20000a2,0x28220228,0x8200,0x22002080,0x20208808,0x8a288208,0x200a2082,0xa00220,0x8000028a,0x800a00a2,0x8a00808,0x8228a282,0x22080820,0x8008820,0x82000002,0x80882800,0x8028,0x8a);
remainders_neg[414].modulo_value=552; remainders_neg[414].bit_mask=create_bit_mask(18,0x82080822,0x20a28a20,0x88080288,0xa2802800,0x28a00000,0x2200280,0xa0822802,0x8208a28,0x8088202,0x820022,0x208000,0xa20220a,0xa2800020,0x800028,0xa08808a,0x2080,0x8800820,0x2);
remainders_pos[415].modulo_value=1660; remainders_pos[415].bit_mask=create_bit_mask(52,0x28a8228a,0xa2220280,0x88282a0,0x8a202002,0xa0080a20,0x28880a82,0x8288800,0x88aa228,0x80028022,0x80080a,0xa0a20880,0xaa0a0a00,0xa828000,0xaa20802,0x22808200,0xa8220808,0x2228a22,0x88a02a8,0x80800002,0xa2a02022,0x8022200,0x280282a0,0xa88a080,0x8020022,0xa2088808,0x82000,0xa0008200,0x2022208,0x8800802,0x8020a22a,0xa82802,0xa0088802,0x288080a8,0x8000020,0x2a80a22,0xa88a2888,0x82020882,0x82028,0x80208aa,0xa000282a,0xa00a0a0a,0x2208a0,0x2a020020,0x8802800,0x288aa22,0x80022282,0xa82a0222,0x208a0200,0x800808a,0xa0a82822,0x80280888,0xa2882a2);
remainders_neg[415].modulo_value=1660; remainders_neg[415].bit_mask=create_bit_mask(52,0xa0202a02,0x2a2a8a00,0x20228,0x2a0a88a,0x20808228,0xa000020a,0x80200080,0x222a0,0x88208aa,0x80088002,0x282a0008,0x22028280,0x20a0088,0x8222808a,0xa2080a08,0x20aa0080,0x8a2a02a2,0x28220,0x800888a,0x2228a82a,0x808a2a88,0xa00a0a20,0x2002008,0x808280aa,0x22800000,0x88802888,0x28080a80,0xa8aa280,0x8000808a,0xa82a22,0x8820208a,0x28000082,0x20080020,0x800088a8,0x8220822a,0x20022000,0xa0a8002,0x880a0a0,0x80828022,0x2088a022,0x28820282,0x882a8020,0x228a80a8,0x8000a088,0x82002222,0x88a2a0a,0x20228aa2,0x28028288,0x80800802,0x2020a02a,0x8a00000,0x2200a22);
remainders_pos[416].modulo_value=104; remainders_pos[416].bit_mask=create_bit_mask(4,0x2aa0a22,0x8822020,0xa0aa8008,0x88);
remainders_neg[416].modulo_value=104; remainders_neg[416].bit_mask=create_bit_mask(3,0x8a2282aa,0x800aa828,0x20220880);
remainders_pos[417].modulo_value=1668; remainders_pos[417].bit_mask=create_bit_mask(53,0x82822082,0x8a20020,0x88008008,0x82800020,0x220000,0x28820a,0x22822022,0xa08,0xa002008,0xa28a0822,0x20800020,0x8008a20a,0xa20a00a2,0x28028228,0x8228800a,0x20000822,0x20828a28,0x80200282,0x82800000,0x20220a20,0x8208a002,0x22002,0x20000a08,0x80080088,0x20022880,0x8220a08,0xa08820a,0x22880082,0x20020020,0xa000082,0x880002,0xa20828,0x8a088088,0x2820,0x28008020,0x8a282088,0x82000082,0x228828,0x8828028a,0xa00a08a2,0x8a20028,0x8000208a,0x820a28a0,0x800a08,0xa000002,0x80882882,0x8228008,0x8800a,0x80002820,0x200220,0x8008a202,0x20882820,0x8);
remainders_neg[417].modulo_value=1668; remainders_neg[417].bit_mask=create_bit_mask(52,0x20a2802,0x228828,0x80080,0x20808a0,0x8a28008,0x88200282,0xa20a28a2,0x8808200,0x8208a080,0x220200a2,0x28008828,0x8002282,0x22820822,0x20820a20,0xa200082,0xa08800a2,0x28020220,0x828820a,0x2080880,0x28a28228,0xa00208a,0x800a2882,0x28808200,0x8008000,0xa08a2000,0xa28200,0x82000282,0xa2000802,0x28828828,0x8208800a,0x80000882,0x8228020,0x2000000,0x808820a0,0x20800828,0x220a000,0x2880802,0x8a20020,0x208202,0x20820022,0x228820,0x808a002,0x2822020,0x8008200,0x8208808a,0x2002,0xa20800,0x88000082,0x8820a0,0x8a08a28,0x800228a,0xa00020a0);
remainders_pos[418].modulo_value=1672; remainders_pos[418].bit_mask=create_bit_mask(53,0xa80020a,0x28228a28,0x82a882a,0x88808220,0x88aa2a0a,0x80aa0280,0x22008aaa,0x820a2002,0xa222000,0x2a808002,0x28022800,0x2a02a02a,0xa080208,0xa28880a2,0x2002a020,0xa8222820,0xa28a8008,0xaa202022,0x8a20820a,0x82a222,0xa00880,0x2000088a,0x228a8080,0x28002888,0x20820020,0xa0a880a0,0x22a0a00,0x82080a,0x28002808,0x2a28822,0x20000802,0x200a00a2,0x8a820002,0x8208a288,0x808aaa0,0x2a28a88,0x28882a20,0xa800808,0x2228a08,0x8020a08a,0xa80a820,0x288028a8,0x202a800,0x888a080,0x8a08200,0xa2008880,0x80aa02aa,0xa8aa2202,0x820222a0,0x22a82008,0xa28828a8,0x8002a028,0xa0);
remainders_neg[418].modulo_value=1672; remainders_neg[418].bit_mask=create_bit_mask(53,0x82008282,0xa02a02a0,0xa200a2,0x80aa0,0x222282,0x228208,0xaa880222,0xa02a80a,0x82a2a888,0x2208088a,0xa08aa080,0xa28a20a2,0x82000a80,0x2a00082a,0xa08a28a0,0x20aa20a8,0x22020880,0x22a8a82a,0x2a80a02,0x88022aaa,0x8288008,0x28888002,0xaa020008,0xa008a000,0xa80a80a8,0x28200820,0x8a220288,0x800a8082,0xa088a080,0x8a2a0022,0xa880808a,0x2882082a,0x20a888a,0x2802200,0x80002228,0x8a2a0200,0xa000a220,0x82080080,0x82a20280,0x8a82802,0x2082028,0xa000a020,0xa8a2088,0x80002008,0x80280288,0x2a080008,0x8228a22,0x2022aa82,0xa8a2a20,0xa220a880,0x2a002020,0x88a2820,0x28);
remainders_pos[419].modulo_value=1676; remainders_pos[419].bit_mask=create_bit_mask(53,0xa2282a22,0x8a2220,0x800a02aa,0xa2a0028a,0xa822a202,0x228a2a8a,0x800a2020,0x80a0a08a,0xa8aa8a2,0x8088a022,0x222aaaa2,0x822aa288,0xa2aa82a0,0xa000802,0x8008a28,0x88a2,0xa8222a08,0x20020288,0x28282a08,0xa882aa02,0x8208a8,0x8808a002,0xaa8080a,0xa82aa02,0x88822a80,0x80a08828,0x20a02208,0x88222282,0xa82a002a,0x2aa080a,0xa2022a02,0x8200800,0xa822a2a2,0x8282880a,0x800820a,0x8882a228,0x2000020a,0x200208a2,0xa028a,0x2aa8a802,0xaa8820a8,0xaa888228,0xa22028aa,0xaa2a0880,0xa0a028a2,0x8a002a20,0x8a288080,0xa882aa2a,0xa8a808,0xa002a28,0x8a200aa8,0x8288a088,0x88a);
remainders_neg[419].modulo_value=1676; remainders_neg[419].bit_mask=create_bit_mask(53,0x2aa0a2aa,0x8802aaa8,0x8828a22,0x2a288a02,0x20aa2a8a,0xaa02a202,0x882a8a8,0x8282802,0x8202202a,0x80028aa,0xaaa2222a,0xaa22a00,0x2a220a28,0x82888082,0x808802a0,0x8888002a,0x20aaa280,0xa88a8a00,0xa0a0a280,0x200a228a,0x880a8020,0x80288a,0x82208082,0x820a228a,0xaa208,0x82800a0,0xa828aa80,0xaaaa0a,0x20a288a2,0x8a228082,0x2a8aa28a,0x80a88088,0x20aa2a2a,0xa0a0082,0x80880a82,0xa2aa0,0xa8888a82,0xa88a802a,0x88828a02,0xa220208a,0x2200a820,0x22000aa0,0x2aa8a022,0x22a28008,0x2828a02a,0x288a2a8,0x2a00808,0x200a22a2,0x88202080,0x8288a2a0,0x2a88220,0xa002800,0x2);
remainders_pos[420].modulo_value=840; remainders_pos[420].bit_mask=create_bit_mask(27,0x802002,0x8200200,0x2008200,0x20228a2,0x20800208,0x8000200,0x800a0020,0x8000200,0x8002080,0xa2880802,0x8200808,0x8008200,0x802000,0x8020028,0x82002000,0x20082000,0x20228a20,0x8002080,0x80002002,0xa00200,0x80002008,0x80020800,0x28808020,0x8200808a,0x80082000,0x8020000,0x80);
remainders_neg[420].modulo_value=840; remainders_neg[420].bit_mask=create_bit_mask(26,0x80082802,0x208a00,0x2080288,0x82022022,0x20008a00,0x80080288,0x200a0,0x808208,0xa000,0x22000002,0x200000,0x80000280,0x82080,0x828020,0x208a008,0x20802880,0x20220220,0x8a008,0x802882,0x200a08,0x8082080,0xa0000,0x20000020,0x2000002,0x2800,0x820808);
remainders_pos[421].modulo_value=421; remainders_pos[421].bit_mask=create_bit_mask(14,0x9e339aba,0x948b705a,0x6077ec19,0x36c73fba,0xb2d5c9c,0xe975c2e8,0x52001291,0xd0eba5e2,0x4ead3405,0x7f38db0e,0xdfb8197,0x83b44a66,0x56731e56,0x17);
remainders_neg[421].modulo_value=1684; remainders_neg[421].bit_mask=create_bit_mask(53,0x2aa0222,0x802a882,0xa8aa2080,0xaa0aa222,0x82a08000,0x20a80a20,0x8a888a08,0x822282a,0x8220a888,0xa2a00282,0x8022080a,0xa2882aa,0x8aaa828a,0xaaa8a8a,0x8a2828a,0x82200aaa,0x2a2a80,0xa8220a0a,0xa2208088,0x888a8aa0,0x80a82082,0x82a0822,0x2a82a800,0x22a8aa22,0xaa008008,0x2aa0208,0x2a22222,0xa008a2a0,0x228aa8a,0x80222008,0x28aaa802,0xa8228a0,0x88828a28,0x220a0080,0x28088a8a,0x2aa0a0a2,0x228a8020,0x8a000aa,0xa08282,0x800080,0x20a08280,0x2a800288,0x200a8a2,0x228282aa,0x28a8880a,0x80802822,0xa28a088,0x28a20a8,0xa00aaa8a,0x8022200,0xa8aa8a20,0x2a28802,0x222a0);
remainders_pos[422].modulo_value=1688; remainders_pos[422].bit_mask=create_bit_mask(53,0xa2888a82,0xa80a0880,0x2a2202,0x8a028a08,0xa0202aa0,0x82288aaa,0xa222aa22,0x2002a82a,0x228a8a22,0xa028a00a,0xaaa28880,0xa08228,0x20002a08,0x2aaa2828,0x8a0aa8a0,0x82000822,0x82a0a88,0x808220aa,0x2aa2220,0x220028,0x88228220,0x2a2a0000,0x82a08a00,0x202aa8a0,0xa0a802a2,0x88820a88,0x22028280,0xa22a082,0x88a802a,0x820a2aa8,0xa800a20a,0x220000a8,0x88288,0x80280088,0x20888aa,0x20aa0882,0x8222a0a8,0xa2882000,0xa80a2aa0,0x82828aa,0x20a800,0xaa28820a,0xa02228a,0x88a00a28,0x888a2a2,0x8aa82a80,0x8288aa88,0xaaaa228,0xa20aa808,0x20a280,0x2a8088a8,0x8a0220a0,0x82a222);
remainders_neg[422].modulo_value=1688; remainders_neg[422].bit_mask=create_bit_mask(53,0x2a00020a,0x20828008,0x88a2aa8a,0x28a0280,0x28a8a228,0xaa00222,0x2aa222aa,0xa88a20a2,0xaa0202aa,0x28a02882,0x222a0008,0x88280aa0,0xa888a280,0xa222a0a0,0x2822028,0xa8880aa,0x80a28200,0x80aa822,0x8a22aaa8,0x88aa88a0,0xaa0aa8,0xa2a28888,0xa280288,0xa8a22028,0x28208a2a,0xa8200,0xaa8a0a08,0x82aa280a,0x800208a2,0xa82a220,0x20882a82,0xaa888820,0x8800a00,0x8a08800,0x8a800022,0xa822800a,0xaaa2820,0x2a00a888,0x2082a228,0x80a0a022,0x88a82088,0x22a00a82,0x828aaa02,0x2882a0,0x80002a2a,0x220a208,0xa002200,0x82222aa0,0x2a822080,0x88a82a08,0xa2080020,0x28aa828,0xa2aaa);
remainders_pos[423].modulo_value=564; remainders_pos[423].bit_mask=create_bit_mask(18,0x828a0802,0x20220828,0xa00000a,0x82000822,0x20a20800,0x8082288,0x800a28a0,0x20020008,0x80282000,0x800082,0x220008,0x200a28a,0xa0822882,0x8200208,0xa088200,0x808a0000,0x28228208,0x8020a);
remainders_neg[423].modulo_value=564; remainders_neg[423].bit_mask=create_bit_mask(18,0x2020082,0x28a20020,0x82088082,0x28800a2,0x28228008,0x8000a200,0x822020,0x28828000,0x820a088,0x80080802,0x8a28800,0x8a082202,0x200a2002,0xa08a00,0x82000288,0x20880,0x20a20a00,0x8282);
remainders_pos[424].modulo_value=424; remainders_pos[424].bit_mask=create_bit_mask(14,0xa2a82aa,0xa20aa008,0x822a0808,0x22822222,0x2208a8a0,0x20a8aa00,0x80000280,0xaa2a0802,0x2a208800,0x8882880a,0x20a88288,0xaa08a20,0x82a8a020,0xaa);
remainders_neg[424].modulo_value=424; remainders_neg[424].bit_mask=create_bit_mask(14,0x82a20a22,0x2a822880,0xaa28080,0xaa0aaaaa,0x2a802028,0xa8202288,0x8888a08,0x22a2808a,0xa2a80088,0xa0082,0xa8200a00,0x822002a8,0xa2028a8,0x22);
remainders_pos[425].modulo_value=85; remainders_pos[425].bit_mask=create_bit_mask(3,0x42d2316,0x8328c53,0x1a312d);
remainders_neg[425].modulo_value=340; remainders_neg[425].bit_mask=create_bit_mask(11,0x88a02a8a,0x80220082,0x2a02a0a0,0x2200828,0xa0288a08,0x2000028a,0x2288080,0x200820a8,0xa820802,0x808a20,0x2);
remainders_pos[426].modulo_value=1704; remainders_pos[426].bit_mask=create_bit_mask(54,0x22822022,0x20028200,0x28a20a,0x220a08a0,0x820a28,0x280208,0xa20a2080,0x8208800,0xa00800a,0x20a2822,0x20208820,0x8208a202,0xa2822882,0x228,0x82008002,0x22002000,0x20a20808,0x80202000,0x20802882,0x20028808,0xa288200,0x828820a0,0x20200020,0xa280080,0xa2000000,0x228200,0x88002202,0x82880080,0x8a00,0x28a000,0x80802,0x8228208,0x8228a00a,0x22800800,0x28020820,0x8080282,0x208a0800,0x8008820,0x2008200,0x80000080,0x28828a28,0x8a208282,0x22080880,0x28a08008,0x200a088,0x220820a0,0x8a08a00,0x80280002,0xa0820020,0x20a08828,0x8a28000a,0x828008a0,0x8828800,0x88);
remainders_neg[426].modulo_value=1704; remainders_neg[426].bit_mask=create_bit_mask(53,0xa20a28a2,0x28820a08,0x88202202,0xa2820020,0x8028220,0x88208280,0x22822800,0xa00008,0x82080082,0x828220a2,0x28a00028,0xa002282,0x220a2002,0x8808a20,0xa08008a,0xa2882880,0x28228000,0x828a088,0xa0082002,0x28820000,0x82200288,0x2002820,0x28a08828,0x82208008,0x22880880,0x8a20a08,0x8a28a,0x2000800,0x800208,0x88202088,0x80800082,0xa20a00,0xa202082,0xa2080080,0x20828028,0x8000820a,0xa0020080,0x28,0x8a080288,0x880800,0x20020220,0x228020a,0xa2800000,0x20200800,0x8a082000,0xa2802820,0x200208,0x820808a,0x200a08a0,0x28200020,0x2208082,0x2080020,0x20008);
remainders_pos[427].modulo_value=1708; remainders_pos[427].bit_mask=create_bit_mask(54,0xa8a0a0a,0x2088802,0x20a88aa,0x20220a8,0x2aa2220a,0x8a202a80,0x2aa2a8,0x22082002,0x8088a00,0x22208282,0x20080888,0x200aa80,0xa082288,0x80aa0280,0x8a820822,0x28a80022,0x80a202a2,0xa20a0808,0x88a28202,0x8a8a28a0,0x200000,0x8a0028,0xa808880,0x82a0088a,0x88002002,0x282a202,0x8020280,0xa8280028,0x80022808,0xa82800,0x202a0a22,0xa200022,0x800280,0x8a2a2000,0x28a220a2,0xa08a808,0x8a02202,0x2a288a8,0x82a2880,0xaa02882,0x820a0028,0xa0080228,0x200802a,0x20888222,0x22020828,0x8208800a,0xaa800800,0x808a22a8,0x88aa802a,0x88080a08,0x2a0802a0,0x22080aa2,0xa2a0802,0xa0a);
remainders_neg[427].modulo_value=1708; remainders_neg[427].bit_mask=create_bit_mask(54,0x82028202,0xa800082,0x82820022,0x8a0aa820,0xa222aa82,0x2282208,0x88a22220,0xaa80a80a,0x800280,0xa2a80a0a,0xa8008000,0x8a802208,0x82802a00,0x8228208,0x20a802a,0x202088a2,0x2a8a2a,0x2a028080,0x220a8a,0x2022028,0x8a88088,0x88028820,0x2080000,0x2288002,0x8a88a,0x8a022a8a,0x808a0a08,0x20a080a0,0x882a000,0x820a080,0xa0a282aa,0x822888aa,0x88008a08,0x2a22888,0xa02aa02a,0x82802000,0x28aa02,0x822a0020,0x8022a008,0x8220a00a,0xa8208a0,0x288082a0,0x8a880822,0x20000aa2,0xa28a80a0,0xa000882,0x22008088,0x8022a20,0x2200a2,0x808200,0x22808a20,0xa280022a,0x8222808a,0x282);
remainders_pos[428].modulo_value=856; remainders_pos[428].bit_mask=create_bit_mask(27,0xa200a282,0xaaaa0a22,0x8a2200a8,0x822802a0,0x202a88,0x88088280,0xa0a082a,0x22022a82,0x80800028,0x288222a,0x88a20a8,0xaaaa2a0,0xa008a88a,0xa282828,0xa0a22a20,0x200a8aaa,0x802a08a2,0x2a88822,0x88280002,0xa082a880,0x22a820a0,0x28220,0x8222a808,0xa20a8028,0xaa2a0088,0x8a88a0aa,0x828a00);
remainders_neg[428].modulo_value=856; remainders_neg[428].bit_mask=create_bit_mask(27,0x2a882a0a,0x222282aa,0x2aa8820,0xaa08228,0x88a8a200,0x800a08,0x828280a2,0xaa8aa20a,0x80888a0,0x8a00aaa2,0x8002a820,0x82222a28,0x28802002,0x82a0a0a0,0x282aa2a8,0xa8820222,0x822802a,0x8a2000aa,0xa0888a,0x280a2008,0xaa20a828,0x888a0aa8,0xaaa2080,0x2a8208a0,0x22a28800,0x2002822,0xa0288);
remainders_pos[429].modulo_value=1716; remainders_pos[429].bit_mask=create_bit_mask(54,0x220a00a2,0x8028008,0x82200280,0x208028a0,0x20820228,0x20a000,0x882,0x220a00,0x8080,0x22080082,0x28800208,0x8000a20a,0x22020000,0x20020200,0x8000202,0xa2822800,0x20820008,0x2080082,0x80800,0x28220,0x2088000,0x20002820,0x20a20208,0x828a008,0x800a0022,0x20800a00,0x8a280282,0x20882802,0x200a00,0x8208800a,0xa08200a2,0x820808,0x8008280,0x22,0x8828,0x8000202,0x20882002,0x28a20008,0x2000288,0x880800,0x8800808,0x200008,0x228a08a0,0x8820800,0x82002,0x80002020,0xa08,0x80082200,0x208000a0,0x20828808,0x8820a280,0x2002800,0x8820028,0x8a00a);
remainders_neg[429].modulo_value=1716; remainders_neg[429].bit_mask=create_bit_mask(54,0xa2820022,0x20800,0x2288208,0xa0082020,0x28020a20,0x80282088,0x80880002,0x8228208,0x88000008,0x22800002,0x20008a00,0x8082282,0xa2020880,0x28828a08,0x80088202,0x220a2080,0x20028800,0x8a00000a,0x80800000,0x8820a20,0xa000088,0xa08820a0,0x28228200,0x202080,0x8208a2,0x28008200,0x220020a,0xa0002002,0xa08208,0xa000002,0x20020822,0x8028000,0x80080200,0x800800a2,0x8000020,0x8008828a,0x20002882,0x20220800,0x82008200,0x80000080,0x8000,0x80208080,0x22020020,0x20008,0x8800a08a,0x828a0,0x8008200,0x800a280,0xa0080020,0x20020000,0x282208,0x82802000,0x28820,0x2002);
remainders_pos[430].modulo_value=1720; remainders_pos[430].bit_mask=create_bit_mask(54,0x28802a0a,0x28228282,0x8202a0,0x8aa20088,0x80a8aa20,0x80822288,0xa2080a0,0x28080208,0x200002a,0xa2200a22,0xa8080008,0x2a028a20,0x802a020,0x820220a2,0x20280228,0x8200280,0x82a88a0,0x22002220,0x22a8a2,0xa0a0a828,0x880a0a00,0x88020282,0x2000008,0x8882a000,0x2200a20a,0x20800208,0x200a80,0x802a008,0x88208002,0x22a08a00,0x80000a82,0x22200000,0x22828080,0xa00a0a0,0x282a0a,0x888a2a88,0x20088800,0x200a22a8,0x8028008,0x82288028,0x208a0880,0xa8080a80,0x2a08a280,0x8a200020,0x8088a008,0x28a80000,0xa0208020,0x20a0208,0x2228882,0xa208aa2a,0x22008a,0x280a8082,0x28828288,0xa0a802);
remainders_neg[430].modulo_value=1720; remainders_neg[430].bit_mask=create_bit_mask(54,0xa0082282,0xa02a0202,0x80a8228,0x2228800,0x202228,0x80a2a00,0x82280820,0x20808280,0x8a8088a2,0x22a8822a,0x20800880,0xa20a02a0,0x8a20a0,0xa82a82a,0xa0a08a20,0x80a80a08,0x80220020,0x2a8822a8,0x88a2202a,0x20282020,0x820288,0xa8a02,0x2888080,0x22888,0xa2882a02,0xa8080a80,0x88288200,0x8a2080,0xa0080a,0xa2280208,0x888020a,0xaa288880,0x2a0a0008,0x82802828,0x80a0a202,0x22200,0xa8000080,0x2882a220,0x820880,0x2a00820,0xa8020008,0x20008200,0x22802208,0x2a088a8,0x2000,0xa0200888,0x282808a0,0xa828280,0x8aa2000a,0x22002222,0x88aa0802,0xa0020802,0x200a0200,0x20208a);
remainders_pos[431].modulo_value=1724; remainders_pos[431].bit_mask=create_bit_mask(54,0xa20002a2,0xa22aaa8a,0xa80a8,0xa200a882,0x28228008,0x8802a2a0,0x2a8a28a8,0xa8a088a2,0xa0a28a20,0x882aa02,0x828a8a08,0x800a82aa,0x88a0a8a8,0x222088a,0x20202822,0xa02aa,0xa282020a,0xa800a22,0x28820828,0x88222820,0x20208200,0xa8080aa2,0x22aa08a0,0x8a202a88,0x202a02aa,0x2000088,0x80aaa88,0x22aaa02,0x82200008,0x2aa80a80,0x822a808a,0xa0a20aa8,0x88aa0202,0x20082080,0x80828882,0x2820822,0xa88a002a,0xa080828,0x8aa80a00,0x8828080,0x2a220888,0x22a2a0a2,0x2aa82a00,0x20a2a28,0xa80aa822,0xa08a28a0,0x88a220a2,0x22a28a2a,0x80a8a802,0xa2002882,0x822a008,0xa2a02a00,0xaa2aaa88,0x8a80008);
remainders_neg[431].modulo_value=1724; remainders_neg[431].bit_mask=create_bit_mask(54,0x2a888a2a,0x2aa22202,0x88820820,0x2a88200a,0xa0aa0880,0x8a2a28,0xa202a020,0x2028002a,0x282a02a8,0x800a228a,0xa020280,0x8820a22,0x282020,0x8aaa0002,0xa8a8a0aa,0x88828a22,0x2a0a8a82,0x820882aa,0xa00a80a0,0xaaa0a8,0xa8a80a88,0x2080822a,0xaa228028,0x2a8a200,0xa8a28a22,0x8a888800,0x80822200,0x8aa2228a,0xaa88880,0xa2208208,0xaa20802,0x282a8220,0x228a8a,0xa880a808,0x80a000a,0x8a0a80aa,0x200288a2,0x828080a0,0x2208288,0x800a0808,0xa2aa8000,0xaa2a282a,0xa220a288,0x8a82a2a0,0x208220aa,0x2802a028,0x2aa82a,0xaa2a02a2,0x820208a,0x2a88a00a,0x80aa2880,0x2a28a288,0x22a22200,0x208880);
remainders_pos[432].modulo_value=24; remainders_pos[432].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[432].modulo_value=24; remainders_neg[432].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[433].modulo_value=433; remainders_pos[433].bit_mask=create_bit_mask(14,0xf473b5e,0x87f12bf,0x8106cd55,0x8a68b75d,0xda5b9b31,0xc80b0296,0xe79d19f1,0x404e3e62,0x696da503,0x59463367,0x8206ebb4,0xf842aacd,0x8bc3f523,0x1eb73);
remainders_neg[433].modulo_value=1732; remainders_neg[433].bit_mask=create_bit_mask(55,0x828aa282,0x80a28a22,0x88a0088,0x2a02a80,0x28202a8,0x828a0a,0x2a008028,0x8882a2aa,0xa0a0288a,0x808aaaaa,0xa8a2228,0x208a2200,0x20a28aa,0xaa2022aa,0xa8a28202,0x2228822,0xa2228a80,0xaaaa8808,0x88a0282a,0xaa2a088a,0xa00802a2,0x828a0800,0xaa020a02,0xaa02a00,0x88028880,0x228a2808,0xa2a8a0a,0xa08020a2,0xa8208220,0x8882a888,0xa02a00aa,0xa0a2a00a,0xa8a08282,0x2a8aa0a,0x88a02002,0x282a0880,0xa8800000,0x80822208,0x288222aa,0xa2820802,0x2222002,0x820a2a0,0xa2208a20,0x222080aa,0x8a88,0x882a0a00,0x2028880,0x2a8aa020,0xa0828aa8,0x2a282a0,0x802a02a8,0x8aa088aa,0x20820a88,0x82008282,0x2);
remainders_pos[434].modulo_value=1736; remainders_pos[434].bit_mask=create_bit_mask(55,0x220a0a22,0x8a028a0,0x822202a0,0x2a020028,0x22022008,0x8a882a80,0x808202aa,0x202220a,0xa288220,0x822808aa,0xaa00288a,0xa000a28a,0xaa02028,0xa002a20a,0x2a080020,0x208aa00,0x802a0000,0x8282080,0x22800080,0x822222a8,0xa28280,0x80088000,0x2a20a022,0x22a28028,0x2222008,0x822088a,0xa02208,0x880a0082,0x20882020,0x88880a2,0x28a8820,0xa08a828,0x2200288,0x828a0000,0x88888202,0x2882a,0x8282002,0xa80202,0xaa208000,0x20a800,0x8a800a08,0x80aa0a0,0x8a000a28,0x2800aaa2,0x202882a2,0x8228a0aa,0x88808008,0x808202a0,0xa822a2aa,0x8808802,0x80a820,0x80888228,0x280a200a,0xa0a0880a,0x88);
remainders_neg[434].modulo_value=1736; remainders_neg[434].bit_mask=create_bit_mask(54,0x2a82822a,0x28a020,0x2aa8a28,0xa20a88a0,0xa282a880,0x2002208,0x80a8222,0x8a8aaa02,0x2200aa0,0x2a08022,0x2208a002,0x28802a02,0x822028a0,0x288a2282,0xa2808828,0xa802280,0xa20888,0x8020a808,0xaa008808,0xaaa2a20,0x882a0208,0x8800808,0x22a828a2,0xa22a08a0,0x8a2aa800,0x80a28002,0x88282a80,0x82800a,0xa800a820,0x822,0x820200a8,0x20020a0,0x8aa08a00,0xa020888,0x28a,0x808a0022,0xa0a882,0x80208a8a,0x22280888,0x88202088,0x2080280,0x80822028,0x2888220,0x20802222,0xa0a00a2a,0xa202822,0x880,0x80a0a28,0x20aa2222,0x8008000a,0x80820a0,0x2a0,0xa002a882,0x28200082);
remainders_pos[435].modulo_value=1740; remainders_pos[435].bit_mask=create_bit_mask(55,0x80080082,0x8228220,0x2008288,0x20020a2,0x20008020,0x8282200,0xa2080022,0x20028200,0x8800a08a,0x802,0x20820a00,0x8080202,0x200000a0,0x8028008,0x228a002,0x80820080,0x20800,0x208a208,0x22002,0x8800,0x8020000a,0x80000080,0x28028028,0x8208088,0x2080800,0x8000200,0xa208000,0x208a0000,0x20000,0x2080008,0x20820002,0x28028220,0x2280,0x802020,0x20000a00,0x88000002,0xa2080800,0x8000208,0x8202002,0xa2880020,0x28020800,0x8200,0x20200a0,0x8208808,0xa,0xa0022802,0x28008a20,0x208a008,0x82820880,0x20008008,0x80080080,0x200808a0,0x28820228,0x2002088,0x820);
remainders_neg[435].modulo_value=1740; remainders_neg[435].bit_mask=create_bit_mask(55,0x800802,0x220a20,0xa080200,0x82802822,0x20800828,0x80202288,0x220000a2,0x28820208,0x2002,0x880002,0x28020208,0x80008282,0x20880020,0x820800,0x8220200a,0xa0800,0x8028000,0xa002200,0x80822882,0x800008,0x280082,0x80800,0x20820020,0x80200000,0x82800000,0x800a08,0x82280080,0x20020080,0x8828800,0x82008000,0xa00a0882,0x20020a20,0x808a208,0x800028a0,0x20808208,0x8008a,0x22000080,0x808200,0x20a08a,0x22000820,0x20820008,0x88080280,0x2820020,0xa00000,0x80088002,0x208a2082,0x20000220,0xa002080,0x2020000,0x20800800,0x8000008,0xa0000020,0x20028220,0x8a002000,0x20);
remainders_pos[436].modulo_value=872; remainders_pos[436].bit_mask=create_bit_mask(28,0xaa2082aa,0xa0022808,0x228a0a80,0x22020202,0x280a828a,0xa880a200,0xaaaa0822,0x22a8820a,0x28008a0,0x88a282a,0x28880808,0xa02a0a,0x8aa20288,0x2aaaa820,0x808aa208,0xa80a0022,0x202228a0,0x28a22020,0x200280a8,0x822a880a,0x20aaaaa0,0x8a022a88,0x82a02800,0x808088a2,0xa0a28880,0x28800a02,0x8208aa20,0xaa);
remainders_neg[436].modulo_value=872; remainders_neg[436].bit_mask=create_bit_mask(28,0x22a80a22,0x288aa080,0xaa028208,0xaa8a8a8a,0xa0820a02,0x20082a88,0x222280aa,0xaa200a82,0x8a088028,0x8002a0a2,0xa0008000,0x8828a282,0x22a8a00,0xa22220a8,0x8022a80,0x208288aa,0xa8aaa028,0xa02aa8a8,0xa88a0820,0xaa20082,0xa8222228,0x28aa200,0xa28a088,0x8002a,0x282a0008,0xa008828a,0xa8022a8,0x22);
remainders_pos[437].modulo_value=1748; remainders_pos[437].bit_mask=create_bit_mask(55,0x2208202,0xaa80aa,0x2aa2a20a,0xa8228a2,0x802888a,0x28a20208,0x22a02820,0x202820a,0x82208000,0x888222a2,0xa8222000,0xaaa0000a,0xa80a2000,0x8a0288a0,0x2280a828,0x2aa800,0x2220aa80,0x22088800,0x8220a2a,0xa020000,0xa02a2282,0x228a020,0x8a008082,0xa20a8288,0x2a2aa228,0xaa80282,0x82202a8,0x82080a02,0xaa02a808,0x8a882802,0x8a288aa,0xa22282a,0x88082020,0x80a080a2,0xa08288a,0x82000008,0x88a8a08,0x88800222,0x80002aa0,0x288002aa,0xa2282a0,0x2a0a228,0xaaa0008a,0x82aa0000,0x22200088,0x8828a888,0x8000020,0xa88a0828,0xa2808280,0x2020808,0x2a0a2228,0xaa88a288,0xa00a08a8,0x880aa02a,0x80820);
remainders_neg[437].modulo_value=1748; remainders_neg[437].bit_mask=create_bit_mask(55,0x8a200a8a,0x88220822,0x222a2a82,0x8202a02a,0x808a0002,0xa02a8280,0xaa2820a8,0x8a0a0a82,0xaa80888,0xaa22a,0x20aaa880,0x22288882,0x20822808,0x28a0028,0xa20820a0,0x88a22080,0xaaa82208,0x2a800008,0x80aa82a2,0x82828888,0x20a2aa0a,0xaa028a8,0x288080a,0x2a020a00,0xa2a222a0,0x82208a0a,0x80228a20,0xa80828a,0x228a2080,0x2002082,0x802a0022,0x82aa20a2,0x80a0a8,0x28082a,0x8280a002,0xa888800,0x80020200,0x88aaa,0x880a220,0xa0088a22,0x2aa0a28,0x8a282aa0,0x22288802,0x2228088,0xaaa88800,0x202000,0x800888a8,0x200280a0,0x2a080a00,0x8a828080,0xa2822aa0,0x22002a00,0x20820020,0x8228a2,0x80a8);
remainders_pos[438].modulo_value=1752; remainders_pos[438].bit_mask=create_bit_mask(55,0x828a2882,0x8020000,0x8082088,0x2820882,0x20a28028,0xa08a202,0x228a2080,0x20800a20,0xa28a200,0x20820822,0xa00028,0x80000280,0x200808a0,0x28200220,0x88282202,0x228a0800,0x28028a00,0x8208200a,0x2020,0x28820a20,0x2200,0x802882,0x20220,0x8200a00a,0x20882080,0x20a08800,0x828a,0x20080,0x8a28200,0x800220a,0x82020822,0xa00a00,0x88080,0x822802,0x28008800,0x8a082,0x82080800,0x28a00820,0x8800a280,0x220020a2,0x28808828,0x8088008,0x20a2020,0x28000,0x828000a,0xa0882082,0x8008a28,0x8808a002,0xa00208a2,0x8808a20,0x8028028a,0x20822082,0x20220820,0x82000080,0x8228a2);
remainders_neg[438].modulo_value=1752; remainders_neg[438].bit_mask=create_bit_mask(55,0x2022002,0x828808,0x8000a000,0x820a0002,0x28220820,0x8200228a,0xa2022800,0x28008228,0x82202288,0xa00a00a2,0x8208820,0x8088208,0xa0800020,0x20a08a28,0x20a28a,0x22020080,0x20820208,0xa00a082,0x808828a0,0x20028228,0x8808a288,0x80082002,0x8828a28,0xa082082,0xa0002800,0x28200000,0x88080202,0x808a0800,0x220a08,0x8008a282,0x28a00a2,0x8208208,0x88000008,0x800a2082,0x20800008,0x8800200a,0x2800080,0x20208028,0x82208,0xa2882822,0x20000020,0x80000080,0x828228a0,0x8820008,0x80208082,0x20002802,0x800220,0x208a,0x208a0022,0x228,0x8208202,0xa00a2802,0x28a20028,0xa088008,0xa2022);
remainders_pos[439].modulo_value=1756; remainders_pos[439].bit_mask=create_bit_mask(55,0xaa80a22a,0x2822a880,0x2082aa2a,0x22aa2800,0x888a22aa,0x88a0208,0x8828208,0x28a8aa,0x828022a8,0x820882,0x820880aa,0x82a0aaa2,0xaa8aa0a,0x20a8088,0x2002a22a,0x280a0280,0x228a0800,0x8a2a802a,0xaa0a0a22,0x20aaa008,0xa228020,0x8222a28a,0x8222228a,0x80088008,0xa28aaa0,0x8a08080,0x2a002a20,0x82200088,0x8a800a,0x2020a2,0x20aaa28a,0x22002200,0x2a288888,0xa28a888,0x8080288a,0xa200aaa0,0x288a0a0a,0x8a802a8a,0x80020a28,0x80280a02,0xa88a800,0x2202a08,0x2a0aa2aa,0x28aaa0a8,0xaa02208,0x28220820,0x2a88028,0xaa2a280,0x2082822,0x22080a22,0x8aa88a22,0x80028aa8,0x8a8aa820,0xa022a882,0xa88a02a);
remainders_neg[439].modulo_value=1756; remainders_neg[439].bit_mask=create_bit_mask(55,0x22082aa2,0xa0aa2008,0xa80a22a2,0xaa22a088,0x2aa22,0x80028a80,0x800a0a80,0x88a02022,0xa08aa20,0x880a800a,0xa800822,0xa28222a,0x82202282,0x8a020800,0xa88a2aa2,0xa0828a08,0xaa028088,0x2a208a2,0x228282aa,0xa8222880,0x82aa08a8,0xaaa2a02,0xaaaaa02,0x8800880,0x82a02228,0x80280808,0xa288a2a8,0xaa88800,0x88020882,0x88a8a82a,0xa8222a02,0xaa88aa88,0xa2a00000,0x82a02000,0x808a002,0x2a882228,0xa0028282,0x208a202,0x88a82a0,0x8a0828a,0x82002088,0x8aa8a280,0xa2822a22,0xa0222820,0x8228aa80,0xa0aa80a8,0x8a2008a0,0x822a2a08,0x8a80a0aa,0xaa8082aa,0x22002aa,0x88a0220,0x20220a8,0x28aa200a,0x20028a2);
remainders_pos[440].modulo_value=440; remainders_pos[440].bit_mask=create_bit_mask(14,0x20a20202,0x2a2288a0,0x2a0a8200,0x80028a0,0x22a00020,0x20080200,0x82208202,0x8808208,0x88008020,0x2008000a,0xa80a2800,0xa80082a0,0x80a2288,0x80808a);
remainders_neg[440].modulo_value=440; remainders_neg[440].bit_mask=create_bit_mask(14,0xa82a028a,0xa22a0020,0x22820288,0x8080a020,0xa2280828,0xa0800a88,0xa280a82,0x80200,0x8008a8,0x20808802,0x20822088,0x20080220,0x2a200,0x802);
remainders_pos[441].modulo_value=84; remainders_pos[441].bit_mask=create_bit_mask(3,0xa28a2822,0x28a08a20,0x8828a);
remainders_neg[441].modulo_value=84; remainders_neg[441].bit_mask=create_bit_mask(3,0x22022022,0x20200220,0x202);
remainders_pos[442].modulo_value=1768; remainders_pos[442].bit_mask=create_bit_mask(56,0xaa28028a,0xaa820200,0x208a0a88,0x8a080882,0x280828,0xa2a2a8a0,0xaa80a02,0x220a0822,0x88a022a2,0x280028,0x2a800802,0x20020a00,0x20828a02,0xa202882a,0x8a202002,0xa2228080,0x800a20a2,0x2aa028aa,0x88028a8,0x82020222,0x2228a0,0xa0a8,0x22aa0002,0x882888,0x82288a8,0x800080a0,0x200a82a0,0x82008220,0x82a00808,0x200020a,0x2288200a,0x2822002a,0xaa8822,0xa000080,0x2888002a,0x8080820a,0x28022088,0x280aa82a,0x8a002aa,0x2888a8a,0x808a202,0x22808a80,0xa28208a8,0xa0800880,0x2002a800,0x280080,0x880a2228,0x20a0888a,0xa02aa088,0x2a8a8a80,0x2028000a,0x2020a228,0xa0a20882,0x8082aa22,0x8028aa00,0xa2);
remainders_neg[442].modulo_value=1768; remainders_neg[442].bit_mask=create_bit_mask(56,0x22a08a02,0x22028a08,0xa0028200,0x200800a,0x88a000a0,0x222a2028,0x82208282,0x2a0280aa,0x28aa2a,0x88a080a0,0x22088082,0xa88a8288,0xa802020a,0x2a8a00a2,0x2a8a00a,0x22aa0808,0x882282a,0xa2282022,0x8000a020,0xa8a8aa2,0x882aa028,0x88882820,0xa222808a,0x800a000,0xaa0020,0x8800828,0xa8820a28,0xa880a20,0x2288080,0x8a888a02,0xaa002882,0xa0aa80a2,0x882200a2,0x82080808,0xa00088a2,0x8000282,0x208aa800,0xa00220a2,0x80208a22,0x82000202,0x80802a0a,0x22080208,0x2a0a8020,0x28080008,0xa88a2080,0x88a08800,0x2aa20,0xa8280002,0x28a22000,0x22020208,0xa8a00882,0xa8a02aa0,0x2822800a,0x80a222a,0x202288,0x2a);
remainders_pos[443].modulo_value=1772; remainders_pos[443].bit_mask=create_bit_mask(56,0x828aaa82,0xa0020a20,0xa8a2000,0x82a22208,0x8222a2a0,0x82a8a22,0xaa2a82a8,0xa880aa22,0xaa880208,0x20a8aa,0x22a0a02a,0x8a22208,0x2880aa8,0x2a08a02,0xa8a8228,0x22a002a2,0x8a28888,0xaa80a828,0x2028,0x20228aa2,0x8008802a,0xa2a00a00,0x8a088200,0xa280808,0x2a28888,0xa82aa882,0xa28828a,0xa0a0002,0x228a0800,0xaa82aa28,0x28a80822,0x28a0222,0x220a2202,0xa8a008,0x2200200a,0x28808a80,0x800008aa,0xa02aa280,0x28a20282,0xa88222,0x2a2a08a8,0x20a80288,0x228080a,0x88a202aa,0xa0a88208,0xa0800a80,0x22aaaa2,0xa022a208,0x2a8aa88a,0x2a8202a8,0xa888288a,0x88a820a8,0x8a2a0208,0x800a000,0xaa28208a,0x82a);
remainders_neg[443].modulo_value=1772; remainders_neg[443].bit_mask=create_bit_mask(56,0xa02220a,0x288a82a8,0x8202a888,0xa2aaa80,0xaaa2a28,0x80a202aa,0x22a20a20,0x200822aa,0x22008a80,0x88a82022,0xaa2828a2,0x802aaa80,0x8a008220,0x8228028a,0x82020aa0,0xaa288a2a,0x802a0000,0x220820a0,0x8888a8a0,0xa8aa022a,0x88008a2,0x2a288288,0x2800a88,0x82a08080,0x8a2a0000,0x20a2200a,0x82a00a02,0x8282888a,0xaa028088,0x220a22a0,0xa02080aa,0x8a028aaa,0xaa82aa8a,0x88202880,0xaa88a882,0xa0080208,0x8888022,0x28a22a08,0xa02a8a0a,0x88200aaa,0xa2a28020,0xa8208a00,0x8aa08082,0x2a8a22,0x28200a80,0x28088208,0x8aa2222a,0x28aa2a80,0xa2022002,0xa20a8a20,0x2000a002,0x20a820,0x2a28a80,0x80882888,0x22a0a802,0xa2);
remainders_pos[444].modulo_value=888; remainders_pos[444].bit_mask=create_bit_mask(28,0xa20a0822,0x828800,0x8a288280,0x22020880,0x28020000,0x820020a,0x820080,0x200820,0x8a080082,0x80000802,0x20220808,0x28228a,0xa0022820,0x28820a08,0xa20a082,0x82880,0x8a28828,0x2202088,0xa2802000,0x820020,0x82008,0x20020082,0x28a08008,0x88000080,0xa2022080,0x28228,0x8a002282,0x8820a0);
remainders_neg[444].modulo_value=888; remainders_neg[444].bit_mask=create_bit_mask(28,0x228200a2,0x8020008,0x2200208,0xa28a0000,0x20828808,0x80288282,0x800a0800,0x8a08028,0x2008002,0x880082,0x28a28000,0x8820a202,0x200a20a0,0x20028200,0x8228200a,0x80802000,0x220020,0x8a28a000,0x22082880,0x8028828,0x8800a080,0x208a0802,0x20200800,0x88008,0x228a2800,0x8820a20,0x200a20a,0x2820);
remainders_pos[445].modulo_value=445; remainders_pos[445].bit_mask=create_bit_mask(14,0x18a13a9a,0xc41318f6,0x581ea0a1,0x8667b4c8,0x111d8b5a,0x2409520a,0x6184b7,0x3b486180,0x1412a409,0x56b46e22,0x84cb7998,0xe1415e06,0x1bc63208,0x16572146);
remainders_neg[445].modulo_value=1780; remainders_neg[445].bit_mask=create_bit_mask(56,0x80282202,0x80a8022,0x822028,0xaa22800,0x80800202,0xa8800a82,0x88280822,0x2280a008,0x88822080,0x8228a22a,0x8022000,0x28088282,0x20aa280,0x8a82a88a,0x288aa0a,0xa00a2a82,0x820a0880,0x8002200,0x8aa228a2,0x20082208,0x8880280a,0xaa2080a0,0xa0a8008,0x82020008,0xa22a,0x80a02208,0xa200a80,0x220220a0,0xa20822,0x288802a,0xa8a0220a,0xa2000202,0x2088a2a0,0xa880,0x80880a0a,0xa820280a,0x88200,0x28820020,0x8282a02a,0xa2088820,0x80a20000,0x80008000,0x22808000,0x88228000,0xa0a08208,0x208a2a02,0x88000200,0x20a8020,0x880a80a,0x80002828,0x8882008a,0x2282a2,0x28aa020,0x88a02822,0x22802a00,0x22208);
remainders_pos[446].modulo_value=1784; remainders_pos[446].bit_mask=create_bit_mask(56,0x2aa2222,0x20022a02,0xa2aaa282,0x22288280,0xa2a02808,0xa00a2aa0,0x2288aa00,0x8a2002a,0xa22a08a,0x80828a2a,0x82802aa,0xa82a222a,0xa8a2a88a,0x2aaaa228,0x8220000,0x880208,0xa0222028,0x8002a828,0xa020828a,0xa08a0220,0x2aa208,0xaa00882,0xa0020aa,0x28a82a02,0xa828822,0x22820002,0xa2a022aa,0x82222002,0x8a800888,0x88aa880a,0xa0800082,0x28882282,0xa080a82a,0xa0aa0800,0x82200a,0xa208aa8,0xa0880a2,0x2a28208,0xa282a80,0x280888,0x20208022,0xa8000088,0x2a288aaa,0x2aa22a8a,0x20a888a8,0x2aa8028,0xa0a8a282,0x20a20a88,0x88a8008a,0xa00aa22,0x8a0aa8a0,0x8820280a,0x8a028228,0x8828aaa,0x8080a880,0x8888aa);
remainders_neg[446].modulo_value=1784; remainders_neg[446].bit_mask=create_bit_mask(56,0x8a22aaaa,0xa88aa28a,0x2a222a0a,0xaaa00a08,0x2a28a080,0x2882a228,0x2a002288,0x802a88a2,0x82aa2802,0x80a02a2,0x80a08a22,0x20a2aaa2,0x202a2002,0xa2222aa0,0x80aa8888,0x88008a80,0x28aaa8a0,0x88a20a0,0x28a80a02,0x28028aa8,0x88a22a80,0x8228800a,0x8288a822,0xa020a28a,0x820a00aa,0xaa0a888a,0x2a28aa22,0xaaaa88a,0x2088000,0x220082,0x2808880a,0xa000aa0a,0x280820a2,0x28228088,0x800aa882,0x82a80220,0x8280082a,0x8a2a0a80,0x82a0a208,0x88a08000,0xa8a808aa,0x20888800,0xa2a00222,0xa22aa202,0xa8200020,0x8a2208a0,0x28202a0a,0xa82a8200,0x208802,0x828822aa,0x2822028,0xa8a082,0x28a0aa0,0x800a0222,0x8082008,0x22);
remainders_pos[447].modulo_value=1788; remainders_pos[447].bit_mask=create_bit_mask(56,0x2000002,0x20828a28,0x8a20a202,0x2880820,0x8828a20,0x8808020a,0xa20a08a0,0x28800220,0x8008282,0x80000882,0x8228008,0x8208028a,0xa00800a0,0x8a08028,0x282002,0x200a20a0,0x28208,0x2280082,0x820228a2,0x8020a20,0x808a200,0x20020080,0x828220,0x88000002,0x20822002,0x8800800,0x80000208,0x28a2882,0x28228a28,0x2080000,0x80020022,0x28008820,0x8000002,0x80882820,0x200800,0x8a202,0x208a0802,0x8a28808,0x8200288,0x82082800,0xa08a00,0x8008280,0xa28020a2,0x20a00200,0xa280208,0x22002882,0x8220000,0x88282002,0xa0880022,0x20a20a08,0xa080202,0x8a2822,0x20820228,0x8808a08a,0x28a2820,0x8000008);
remainders_neg[447].modulo_value=1788; remainders_neg[447].bit_mask=create_bit_mask(56,0x82880882,0x28020220,0x228228a,0x820000a0,0x20228,0x8282,0x22820020,0x20008a28,0x8008020a,0x880002,0xa20800,0xa008202,0x20800820,0x200820,0x8820a08a,0xa0822820,0x8820a00,0x8a20800a,0x28a2022,0x828228,0x80002288,0xa08a0800,0x8020a28,0x8808a,0xa00a2882,0x8008,0x8088280,0x82022002,0x20a20220,0x8a008088,0x8a08a2,0x20800028,0x8000808a,0x20a0,0x8a08008,0x8800228a,0xa0020082,0x220000,0x80288200,0x2802080,0x8200208,0x80080208,0x22082822,0x28208a08,0x82208280,0xa2882002,0xa28808,0x20a08a,0x200008a2,0x28228200,0x8200828a,0x80022022,0x28028a20,0x2002,0x820220a0,0x808800);
remainders_pos[448].modulo_value=56; remainders_pos[448].bit_mask=create_bit_mask(2,0xaa08020a,0xa08020);
remainders_neg[448].modulo_value=56; remainders_neg[448].bit_mask=create_bit_mask(2,0x22808a02,0x2808a0);
remainders_pos[449].modulo_value=449; remainders_pos[449].bit_mask=create_bit_mask(15,0x12d54fb6,0xabae7399,0x37876149,0xe64dc47e,0x14013243,0x8535c297,0x7830b5fe,0xfeb4307b,0xa50eb285,0x93200a3,0xf88ec99f,0x4a1b87b1,0x6739d756,0xb7caad22,0x1);
remainders_neg[449].modulo_value=1796; remainders_neg[449].bit_mask=create_bit_mask(56,0x8a08822a,0x2222aa00,0xaa0aa880,0x2a8008a2,0x8888aa8a,0x8a80a0a,0xa0a82822,0x2228a8a2,0x28822a08,0x80aa882a,0x20020002,0x82820a28,0xaaa00a8a,0x2a0220aa,0xa82202a0,0x8a802aaa,0xa2820a0a,0x20020,0xa08aa80a,0x82a208a2,0x28a8a220,0x20a0a82a,0x8280a882,0x8aa8888a,0x28800aa2,0x8aa82aa,0x2aa2220,0xa2088288,0x8288a202,0x222002a8,0x28008aa,0xaa8820,0x88880082,0x880a8280,0x280a0a22,0x22080820,0x8a2028a,0xa8008a02,0x2aa2aaa0,0xa0a2820a,0x2a8080,0x2a22800,0xa22a022,0x80aa0000,0x20a28280,0xaaa2aa28,0x28800a82,0xa0228820,0x8082228,0x28280a02,0xa0a808a2,0x80088880,0x8aa8020,0x8800a002,0xa002222a,0x2288a08a);
remainders_pos[450].modulo_value=120; remainders_pos[450].bit_mask=create_bit_mask(4,0x80820082,0x28200,0x2008280,0x820082);
remainders_neg[450].modulo_value=120; remainders_neg[450].bit_mask=create_bit_mask(4,0xa0802,0x8020a00,0xa080208,0x20802);
remainders_pos[451].modulo_value=1804; remainders_pos[451].bit_mask=create_bit_mask(57,0x2a02a22a,0xaa8a0,0xaa088a,0xa022a820,0x22a28aa,0x82080202,0x22a0028a,0x82280a02,0xa00a020,0x888022,0x8008a8a2,0xa288028a,0xa0228228,0x822a8a82,0x8a02,0x8a08280a,0x2820800,0x2020aa02,0x8a222a08,0xa8228208,0x2a08828,0x888088a,0x8aa28202,0x20008a88,0xa0088000,0x20802aa0,0x22a82,0x800a8088,0x2a002000,0x88000220,0x8020882a,0x2200a0aa,0x20008000,0x28aa222a,0x2220808,0x20a80a22,0x2882a282,0x888a2208,0xa080820a,0x828080a,0x820a2002,0x20000a02,0x2a88280a,0x2880a82a,0x228a288,0xa2002a28,0x222008a2,0xa00a0880,0x2882080,0xa8880a,0x2082a28,0x8a880808,0xa880aaa2,0xaa00082,0xaa000a22,0xa80a80a2,0xa88);
remainders_neg[451].modulo_value=1804; remainders_neg[451].bit_mask=create_bit_mask(57,0xa28a22a2,0x88022028,0x88228002,0x20aa20a0,0x8aa22022,0x2808a8a,0xaa288a02,0xaa0820a,0x28028a8,0x80008aa,0x880202a,0x2a008202,0x282a0aa0,0xaa2020a,0x88880282,0x2802082,0x820a8088,0xa8a8228a,0x2aaa200,0x20a20a00,0xa2800a0,0x80008002,0x22a028a,0xa8080200,0x28800088,0xa808a220,0x888a220a,0x820800,0xa288a888,0x880a28,0x8a000a2,0x2a882822,0xa8880888,0xa022a2a2,0x8a228080,0xa82082aa,0xa00a2a02,0x22a80,0x20080a82,0x80208082,0xa82a80a,0xa880828a,0x2200a082,0xa00820a2,0x82a02200,0x2a08a2a0,0xaaa8802a,0x28828000,0x8a002808,0x200082,0x8a80a2a0,0x2008000,0x2000222a,0x228880a,0x228882aa,0x20820022,0x200);
remainders_pos[452].modulo_value=904; remainders_pos[452].bit_mask=create_bit_mask(29,0x8200aa82,0xa22a0200,0x88aa2020,0x8228a20a,0x2022a88a,0x202a228,0xaa80208,0x802aa0a,0x88a80802,0x22a88082,0x8a2882a,0x808aa22a,0x80a88a0,0x2aa00820,0x200aa828,0x22a02008,0x8aa2020a,0x228a20a8,0x22a88a8,0x202a2282,0xaa802080,0x802aa0a0,0x8a808020,0x2a880828,0x8a2882a2,0x8aa22a0,0x80a88a08,0xaa008200,0x82);
remainders_neg[452].modulo_value=904; remainders_neg[452].bit_mask=create_bit_mask(29,0xa88220a,0x2aa28a88,0x22a8a8,0xaa02a82,0xa8aa2002,0x8a8a2aa0,0x82208a80,0x808a2282,0x20808a,0xaa20080a,0x802200a2,0x8022aa2,0x80820028,0xa22880a8,0xa88220a0,0xaa28a880,0x22a8a82,0xaa02a820,0x8aa20020,0xa8a2aa0a,0x2208a808,0x8a22828,0x20808a8,0xa20080a0,0x2200a2a,0x8022aa28,0x8200280,0x22880a88,0xa);
remainders_pos[453].modulo_value=1812; remainders_pos[453].bit_mask=create_bit_mask(57,0x82880002,0xa20a28,0xa282082,0x828208a2,0x228808,0x8080200,0xa0000822,0x20a20,0x2000082,0xa20a00a2,0x20008a00,0x2002282,0x20002820,0x8a08808,0x88288202,0xa0002080,0x20020020,0x8800220a,0x2822802,0x20228,0x82000208,0x80002822,0x808220,0x8a200002,0x820a0880,0x20a28228,0x82280208,0x8a08a2,0x20000a00,0x8828a20a,0xa0820802,0x8228828,0x8a20220a,0x20080000,0x208820,0x8288280,0x20800,0x28028808,0x2280288,0x8a0880,0xa08008,0x82202080,0xa2080822,0x820220,0x8008280,0x882880,0x8a00a20,0x808a00a,0x82000,0xa08a08,0x2088200,0x80000802,0x28220228,0x8a08a208,0xa0082082,0x28228a08,0x80002);
remainders_neg[453].modulo_value=1812; remainders_neg[453].bit_mask=create_bit_mask(57,0x2000882,0x8228220,0x8220a00a,0x20a0022,0x8220000,0x80008288,0x208800a2,0x8828228,0x8a08800a,0x22820822,0x28800208,0x8a08a20a,0xa08820a0,0x200000,0x20028a,0x20882800,0x28828828,0x8a282,0x820a2082,0x8828a20,0xa088280,0x8820a2,0x8000a28,0x220808a,0x2820000,0x28220a20,0xa208280,0x80020022,0x28808208,0x202282,0x200a0082,0xa20020,0x228a282,0xa0800880,0x8a00028,0x80200208,0x808a0080,0x20820000,0x8a208200,0x80020000,0x8200800,0xa28a008,0x228000a2,0x8028a28,0x80080208,0x80002000,0x208228,0x80002082,0x80802880,0x8200200,0x8a000288,0x880082,0x20a28a20,0x2002280,0x20802802,0x20a20200,0x808a);
remainders_pos[454].modulo_value=1816; remainders_pos[454].bit_mask=create_bit_mask(57,0x8a08aa2a,0x2a8228a2,0xa2020a02,0xa2a2800a,0x28a8888a,0x82820a22,0x888208a8,0xa80022,0xa0a82802,0x202aa,0x882a2a20,0xa0a8aa20,0x2a822a8,0xa200a02a,0xaa20222,0x2802a02a,0x80a0802,0x20288a20,0xa2aaaa20,0x280a0002,0xa80aa2a8,0xa288802a,0xa2828808,0x88082220,0x82020088,0xa2a200aa,0x228022a0,0x8a08208,0x20080020,0x8820820a,0x8a0a8802,0x82aa008a,0x22220080,0x8a088820,0x8a202282,0x2aa80222,0x282a8aa0,0x8a8000a0,0x808aaaa,0x2008a228,0x288020a0,0xa0a80a80,0x8a88808a,0x80a80a00,0xa2a882a,0x2208aa2a,0x8a8a8,0xaaa8080,0x80282a,0x2288002a,0x822a2082,0x2888a082,0x8aa2222a,0x8aa0028a,0xa880a080,0xa28a2882,0xa8aa20);
remainders_neg[454].modulo_value=1816; remainders_neg[454].bit_mask=create_bit_mask(57,0x28022a2,0xa20aa02a,0x2a8a828a,0x2a2a0882,0xa0200002,0xa0a82aa,0xa8020,0x882088a2,0x2820a08a,0x888a8a22,0xa2a2a8,0x282022a8,0x8a20aa20,0x2a8828a2,0x822a8aaa,0xa08a28a2,0x8082808a,0xa8a002a8,0x2a2222a8,0xa082888a,0x20822a20,0x2a0008a2,0x2a0a0080,0x80aaa8,0xa8a8800,0x2a2a8822,0xaa08aa28,0x80280a80,0xa88088a8,0xa80a82,0x282008a,0xa228802,0xaaaa8808,0x28000a8,0x2a8aa0a,0xa2200aaa,0xa0a20228,0x2088828,0x80802222,0xa8802aa0,0xa008a828,0x28208208,0x2000802,0x8208288,0x82a200a2,0xaa8022a2,0x88802020,0x82220808,0x8808a0a2,0xaa0088a2,0xaa2a80a,0xa000280a,0x22aaaa2,0x2288a02,0x20082808,0x2a02a00a,0x2022a8);
remainders_pos[455].modulo_value=1820; remainders_pos[455].bit_mask=create_bit_mask(57,0x20820a02,0x2008a20,0x220a0080,0x82000808,0xa2800a08,0x8822080,0x880082a2,0x2802a028,0xa2,0xa2a8800a,0x28082880,0x88020280,0x28000a8,0x88828088,0x2002208,0x20082200,0x82020022,0x22002008,0x2800,0x82a80220,0x820000a,0x2a000020,0xa008000,0x2000808,0xa0082002,0x80802282,0xa0a0a20,0x808a000,0x20800,0xa202,0x208a0a0a,0xa8288020,0x8008200,0x2020008,0x8000200a,0x80000a,0x2a000082,0x8802a8,0x80028000,0x22008008,0x88800808,0x88200,0x22088008,0x2202822,0x22a00028,0x80280802,0xa0228202,0xa0022a8,0x88a00000,0x2280a802,0x8a82002,0xa0208822,0x220a0028,0x82020008,0x200a08,0x808a2008,0x80a0820);
remainders_neg[455].modulo_value=1820; remainders_neg[455].bit_mask=create_bit_mask(57,0xa80a020a,0xa080220,0x22828008,0xa008080,0x22000200,0x800a2808,0x80222,0x200a2020,0x8808822,0x22200002,0x20002008,0x28a00,0xa000020,0x20000,0x8288aa00,0x20802a80,0x20a88a2,0x2a082080,0x8880a088,0x2200a20,0x80280082,0xa2088820,0x2880080,0x2808080,0x2000a80a,0x8002a0a,0x820202a0,0x802008,0x80828008,0x882a02,0xa0020282,0x202008a0,0x800280,0x8a020880,0x88a002,0x88080002,0x22088802,0x8220,0x8020888,0xa2800800,0x80000,0x80000280,0x2a800000,0xaa020a2,0xa2288820,0x8200082,0x28220a82,0x2882220,0x208088,0x2208200a,0x202882,0x202000a2,0x2a0280a0,0xa828800,0x80a80200,0x8022080,0x20020);
remainders_pos[456].modulo_value=456; remainders_pos[456].bit_mask=create_bit_mask(15,0x20028a2,0x820228,0xa08228a,0x22022020,0x208008,0x2208280,0x820a0,0x8200000,0x8208800a,0x2080002,0x8808820,0x8820a008,0x808200a2,0x28008028,0x8a);
remainders_neg[456].modulo_value=456; remainders_neg[456].bit_mask=create_bit_mask(15,0x82802022,0x8028a20,0x200a202,0xa28a28a0,0x8a00800,0x8a280208,0x80802820,0x208808,0xa000082,0x82800882,0x20,0x282080,0xa0822,0x20800820,0x2);
remainders_pos[457].modulo_value=457; remainders_pos[457].bit_mask=create_bit_mask(15,0x3b2f53de,0x87cf8455,0x40929b1b,0xc29d7855,0x6bcda16d,0xb200cb2c,0x1dc03b39,0x700ee37b,0x4c013673,0x16cf58d3,0x7ae50eda,0x65240aa8,0x87cf8763,0x2bd372a8,0x1ef);
remainders_neg[457].modulo_value=1828; remainders_neg[457].bit_mask=create_bit_mask(58,0xa2a28a02,0xa020888,0x880a0282,0xa00a088,0xa20028a0,0x2a8802a0,0x8008a2a0,0xa8822aa2,0x8088aaaa,0x8a02800a,0xa2288202,0xa8a88220,0xa020aaa,0xa20aaa20,0x822a2022,0x28022aa,0xa8aaaa82,0xa2282208,0x28a0208,0x880a800a,0x8aaaaa8,0x800a2aa2,0x8aa02a28,0x2282a00,0x28028a0,0x80888828,0x2820a02,0x2a288882,0x82a2028a,0x88882020,0xa0a282a2,0x28888a82,0x82a82a8,0xa02822aa,0x202a008a,0x20aa88,0x8a2,0xaa80a888,0xa2a082a0,0xa2282208,0x80000808,0x22a82a2,0x22202280,0x2000a202,0xa2a0aa,0x22880800,0xa288220a,0xaa82a082,0x88a80,0x228800,0x208aa82,0x2a8802a,0x82aa20a,0x8a0aa0aa,0x82a0a888,0x88a2a0a2,0xa0820208,0x2);
remainders_pos[458].modulo_value=1832; remainders_pos[458].bit_mask=create_bit_mask(58,0x2222a202,0xa82000a,0xa00220a8,0xa28a8a2,0xaa8028aa,0x2288028a,0x2a8a202a,0xaa022a0a,0x8802aa28,0x28822202,0x202a0a20,0xaa88a2a0,0x8a820820,0xa2aaaa08,0x200828a,0x8088a000,0x800a28a2,0x2a2a0208,0x882a20a0,0xa882a222,0xa0082002,0x80202022,0x82202a20,0x800082a8,0x20a0082a,0xaa020808,0x80a80a22,0x2220aaa2,0x8a82a202,0xaa088880,0xa02a028a,0x2080aa88,0x200a0820,0x820002a8,0xa808822a,0x8080208,0x8200a88,0x8a822a80,0x8a82288,0x80a8a80a,0x28a00220,0xa22028a,0x82008000,0xaaaa8aa2,0x2082a220,0x8a22aa08,0xa0a8080a,0x88822808,0xaa802280,0xa880aa28,0x8a2a8a0,0x802288a8,0x2802aaa2,0x2a28a0aa,0x8800a8a,0x82a02a,0x8a8888a0,0x80);
remainders_neg[458].modulo_value=1832; remainders_neg[458].bit_mask=create_bit_mask(58,0xaaaa2a8a,0x820a8882,0x288aa820,0x82a0202a,0x2208a022,0xaa008a02,0xa202a8a2,0x228aa282,0x8a22a0,0xa00aaa8a,0xa8a282a8,0x22002a28,0x20a80a8,0x2a222280,0x8a880a02,0x8002888,0x882a02a,0xa2a28a80,0xa2a828,0x200a2aaa,0x2880a88a,0x8a828aa,0xaa8a2a8,0x8880a20,0xa82880a2,0x228a8080,0x82082aa,0xaaa8222a,0x20a2a8a,0x22800008,0x28a28a02,0xa8082200,0xa88280a8,0xa888a20,0x20800aa2,0x80808a80,0x80a88200,0x20aa208,0x8020aa00,0x8202082,0xa0288aa8,0x82aa8a02,0xa880888,0x2222022a,0xa80a2aa8,0x2aa2280,0x28208082,0xaa080,0x2208aa08,0x200822a0,0x802a2020,0x8aa0020,0xa08a222a,0xa2a02822,0x80088202,0x880a28a2,0x2000028,0x8);
remainders_pos[459].modulo_value=204; remainders_pos[459].bit_mask=create_bit_mask(7,0xa20020a2,0x8028208,0x8088002,0x22020000,0x28020800,0x8008a208,0x8a0);
remainders_neg[459].modulo_value=204; remainders_neg[459].bit_mask=create_bit_mask(7,0x22882822,0x820a00,0x8000008a,0xa20a0880,0x20828008,0x2280,0x20);
remainders_pos[460].modulo_value=920; remainders_pos[460].bit_mask=create_bit_mask(29,0x280a0a0a,0x22a8222,0x8828000,0x8a02802a,0xa2082200,0x88000a88,0x82020080,0x8082000,0x22020,0x8208a22a,0x280a2008,0x28220a80,0x2a880000,0xa028888,0xa0a0aa0a,0xa8222280,0x28000022,0x2802a088,0x822008a0,0xa88a20,0x20080880,0x82000820,0x22020080,0x8a22a000,0xa2008820,0x20a80280,0x80000282,0x288882a8,0xa0a0a0);
remainders_neg[460].modulo_value=920; remainders_neg[460].bit_mask=create_bit_mask(29,0xa0020282,0x8a220aa2,0xa0088,0x28208a2,0x2280aa08,0x880200,0xa0a0800,0x80a088,0x8882a8a8,0x2802222,0xa0822880,0xa02a8200,0x22008008,0x82820000,0x20282202,0x20aa2a00,0xa00888a2,0x208a2000,0xaa08028,0x80200228,0xa0800008,0xa0880a0,0x2a8a8008,0x2222888,0x22880028,0xa8200a08,0x8008a02,0x20000220,0x202828);
remainders_pos[461].modulo_value=461; remainders_pos[461].bit_mask=create_bit_mask(15,0x47fb4272,0x6d622a92,0xfbb33659,0x630eebd3,0x72c1ee14,0xa0301af8,0x83b0acda,0x4370752b,0x30156cd,0xe0d387d6,0xdc318a1d,0x3377f2f5,0x11ada69b,0xb7f89255,0x1390);
remainders_neg[461].modulo_value=1844; remainders_neg[461].bit_mask=create_bit_mask(58,0x8a228aaa,0xa0aaa20a,0x222aaa80,0xaa82220a,0xaa082288,0x28a88220,0xa282002,0x8aa8a8a2,0x8a888a00,0x280a0a0a,0xa80280,0xaaaa2a28,0x88202a02,0x2a200a88,0xa0828a08,0x888a8022,0xaa02a020,0xa2a2aa,0xa00a00a8,0x8a828280,0x8802888a,0x8228a8aa,0xa20020a2,0xa82208a8,0xa88a2082,0x2282220a,0x280aaaa2,0x8a822aa8,0x2aa8a22,0x80822080,0xa8280022,0x82220000,0x8800a222,0x28028a20,0xa20808a2,0x22820a2a,0xa8800808,0x80808882,0xaa0a8282,0xaa80aa0,0xa0000202,0x888a2a02,0x8a022a80,0x2028a002,0x288880aa,0x2a02a,0xaa82020,0xa82a82a8,0x8080a0a0,0x8aa088,0x20a20808,0x822aa28,0xa00a2288,0x80088228,0x2220a222,0xa0a8000,0x2080a200,0x82);
remainders_pos[462].modulo_value=1848; remainders_pos[462].bit_mask=create_bit_mask(58,0x82880002,0x8200000,0x220828a,0x800028a0,0x20820800,0x82202288,0xa0802080,0x8020a20,0xa200002,0x200a0022,0x8000800,0xa208202,0x82002020,0x200008,0x88082000,0x882822,0x20220a00,0x82208000,0x2802800,0x220028,0x28a000,0x2002020,0x8800200,0x200,0x82080082,0x28000a00,0x80000200,0x808a2000,0x808800,0x2002202,0x20008a2,0x28008000,0x8200a000,0x820020,0x20008000,0x80008002,0x80800,0xa28,0x80280088,0x82002802,0x8000208,0xa088,0x22882822,0x820,0x82200008,0xa0080800,0x20808208,0x8002000,0xa08800a0,0x20800008,0xa08a080,0x82020802,0x8228808,0x2002082,0x800a2800,0x20a28208,0x82000008,0x800022);
remainders_neg[462].modulo_value=1848; remainders_neg[462].bit_mask=create_bit_mask(58,0x2000002,0x208800,0x82280202,0x82020,0x28020008,0x2282200,0x20082000,0x828228,0x2208082,0x208208a2,0x8008,0x8220028a,0x20828a0,0x8a08000,0xa000,0x20a2,0x20a28208,0xa280088,0x82002000,0x8a20820,0x8202088,0x82882820,0x8200,0x80088288,0x2000802,0x20808200,0x8080288,0x22080,0x8000008,0xa08a202,0x82800022,0x20000808,0xa002088,0x800a00a0,0x28000008,0x808000a,0x800080,0x800220,0x208000,0x2802082,0x800a00,0x88002000,0x22002022,0x8808020,0x2288080,0x20000080,0x28000a00,0x80082080,0x20000020,0x20008800,0x2002000,0x28a0002,0x220000,0xa00a00a,0x822080,0x28220200,0xa088000,0x808a2);
remainders_pos[463].modulo_value=1852; remainders_pos[463].bit_mask=create_bit_mask(58,0x2aaa0a8a,0xa20a0002,0x28a2aa2,0x8282280a,0xaa80820,0xaaa0aa02,0x88a2a888,0x220808aa,0x280aa0,0x820802aa,0x88a888a,0x28200a20,0x20888a0,0x82208a02,0x8020aa2,0x8a82288a,0xa822228a,0xa882a8a0,0x2220222,0x800aa28a,0xa802a0aa,0x22a288,0x22200822,0x8a802800,0x28a2a002,0x82a08a0a,0x800820a,0x8aaa8288,0x2028000,0x20002808,0x2282aaa,0x2a082002,0x8a0a20a8,0x2800a8a2,0x8002802a,0x8820088,0xa228a880,0x2aa0a802,0xa28aa00,0xa8880888,0xa0a2a822,0x2a288882,0xa22882a,0x28aa0802,0x80a2088,0x80a22208,0x8a0082,0x2a222a22,0xaa80208,0x80aa0280,0x2aa20208,0xa222a8a2,0x80aa0aa,0x208202aa,0xa028828,0xa8aa8a28,0x88000a08,0xa2a0aaa);
remainders_neg[463].modulo_value=1852; remainders_neg[463].bit_mask=create_bit_mask(58,0xa2228202,0x2a82888a,0x8a02a22a,0xa0aa082,0x822080a8,0x2228228a,0x2a2000,0xaa808022,0x88a08228,0xa808a22,0x80020002,0xa0a882a8,0x8a800028,0xaa8028a,0x808a022a,0x20aa002,0x20aaaa02,0x200a2028,0x8aaa8aaa,0x8822a02,0x208a2822,0x88aa2a00,0xaaa880aa,0x208a088,0xa02a288a,0xa280282,0x80880a82,0x2220a00,0x8a8a0888,0xa888a080,0x8aa0a222,0xa280a88a,0x282a820,0xa088202a,0x88a08a2,0x800a8800,0x2aa02008,0xa228208a,0x82a02288,0x20008000,0x282a20aa,0xa2a0000a,0x82aa00a2,0xa022808a,0x8082a800,0x82aaa80,0x8802880a,0xa2aaa2aa,0x82208a80,0x8228a08,0xa22a8a80,0x2aaa202a,0x80822822,0xa80a8a22,0x828a00a0,0x202202a0,0x888280,0x2a28222);
remainders_pos[464].modulo_value=232; remainders_pos[464].bit_mask=create_bit_mask(8,0x28022a2,0x8a2a200a,0x280a008a,0x2a28a880,0xa02802,0x8a8a2a2,0x880280a0,0x8a);
remainders_neg[464].modulo_value=232; remainders_neg[464].bit_mask=create_bit_mask(8,0x8a08aa2a,0x2a2a882,0xa0028802,0xa2a02008,0x8828a08a,0x80202a2a,0x8a0028,0x2);
remainders_pos[465].modulo_value=1860; remainders_pos[465].bit_mask=create_bit_mask(59,0x20082802,0x28820,0x2000200,0x82822800,0x8000,0x82202,0x800a0000,0x8808208,0x80202002,0xa2000022,0x20020000,0xa088200,0x208200a0,0x8820820,0x228a202,0x20820882,0x28020820,0x2088280,0x20020,0x20000228,0x20200a,0x82080882,0x28008,0x2208000,0x80002,0xa20a08,0x2000200,0x208a0000,0xa08020,0x8020a00a,0xa2080,0x8000800,0xa08a000,0x20002,0x208808,0x280000,0x22020822,0x808008,0x8800008a,0x80080002,0x28220800,0x82080280,0x22082080,0x8a28808,0x82082208,0xa0082080,0x8220a00,0x80080,0x800008a0,0x808028,0x8202208,0xa0022,0x8820000,0x200008,0x2882820,0x8000800,0x82280000,0x2820080,0x8);
remainders_neg[465].modulo_value=1860; remainders_neg[465].bit_mask=create_bit_mask(58,0x20802082,0x8020020,0xa088288,0x2022080,0x800808,0x8800228a,0x20880,0x200,0x820a08a,0x22880822,0x28820808,0x82000280,0x20020020,0x28028,0x8220220a,0xa00a0002,0x20020020,0xa000208,0x808208a0,0x20808a20,0x88282082,0x2000002,0x8820000,0x8a200088,0x80800802,0x8220200,0x8a088280,0x20020080,0x8200820,0x282002,0x80822800,0x8000,0x2002088,0x80820882,0xa00000,0x80200088,0xa20a00a2,0x8000000,0x8002,0x802,0x20a20008,0xa008200,0x22802000,0x220000,0x200a200,0x20802800,0x228200,0x8000008,0x800020,0x820,0x80282280,0x800200a2,0x28008,0x88208080,0x82002020,0x800008,0xa208080,0x20a0000);
remainders_pos[466].modulo_value=1864; remainders_pos[466].bit_mask=create_bit_mask(59,0x8aa08aaa,0xa8a2280a,0x220a28a8,0x220a0208,0x88280882,0x208a82aa,0x22a88200,0x22aa008a,0xa88a80,0x28022282,0xaaa8008,0xaa80a8a,0xa020202a,0xa202082a,0xa2222200,0x28a2a20a,0x2a2a2a00,0x80a800a0,0xaa8000a0,0x8222a828,0x80880aa2,0x82a0022a,0xa2880220,0x2808a2a,0x88a82880,0xa2a0a222,0x820a228,0xa8220808,0x8a0080,0xa20000,0x20882a02,0x8a082020,0x8a0a8a28,0x282a2288,0xa2028002,0x80228aa8,0x800a8208,0xa02202a8,0x2a88828a,0x2aa28,0x2a020a,0xa8a8a80a,0x8a8a2800,0x88888aa0,0x20808a00,0x8080aa8,0xa02aa0a8,0x2aaa0a2,0x88802820,0xa22a0082,0xaa8802,0x822a88a2,0x82a20800,0x202822aa,0x80a08882,0x28a08820,0x288a2a2a,0xa20aa2a0,0xaa);
remainders_neg[466].modulo_value=1864; remainders_neg[466].bit_mask=create_bit_mask(59,0x2280222,0x202aa082,0xaa82a020,0xaa828a80,0xa0800a,0xa8020a22,0xaa200a88,0xaa228802,0x88200208,0xa08aaa0a,0x82220880,0x82208202,0x28a8a8a2,0x2a8a80a2,0x2aaaaa88,0xa02a2a82,0xa2a2a288,0x8208828,0x22088828,0xaaa20a0,0x800822a,0x2288aa2,0x2a008aa8,0x8a0802a2,0x20a008,0x2a282aaa,0x80a82aa0,0x20aa8080,0x88028808,0x882a8888,0xa800a28a,0x280a8a8,0x28202a0,0xa0a2aa00,0x2a8a088a,0x8aa0220,0x8820a80,0x28aa8a20,0xa2000a02,0x888a22a0,0x88a28a82,0x20202082,0x202a088,0x228,0xa8080288,0x80808220,0x28a22820,0x8a22282a,0x8a0a8,0x2aa2880a,0x822008a,0xaa2002a,0xa2a8088,0xa8a0aa22,0x828000a,0xa02800a8,0xa002a2a2,0x2a822a28,0x22);
remainders_pos[467].modulo_value=1868; remainders_pos[467].bit_mask=create_bit_mask(59,0x822aaa02,0x220208,0x2828028,0x22aa88a,0x2222a28a,0xaa822aa2,0x28aa080a,0x82820aa2,0x20aa8088,0x288aa082,0x28200228,0xa8a2a2a8,0xa0088022,0x2828a2a,0x2802a0aa,0x820a0a80,0xaa22a800,0x8082008,0x8aa8a0a0,0x280220a0,0x2a00288a,0x28a0a22,0xa2802088,0x8a0002,0x8088888,0x20008a82,0xaa0a0a82,0x8a88aaa0,0x80008a22,0x20002a0a,0xa22a288a,0xa0aa0aa,0x2000882a,0x2202082a,0xa200222,0x8028a800,0xa280220,0x800a888a,0x88028a22,0xa2aa20a0,0x820200a0,0xa88aa200,0xa082002,0xa802802a,0x28280aa0,0x2200aa8a,0xa8a2a880,0x8282a8,0xaa228288,0x2aa08820,0x8282220,0xaa288aa,0x882aaa02,0xa88888aa,0xaa880a28,0x28280a22,0x8800280,0xaa882208,0x80a);
remainders_neg[467].modulo_value=1868; remainders_neg[467].bit_mask=create_bit_mask(59,0xaa2228a,0x88aa8a80,0x8a0a08a0,0x8aa22002,0xaaaa2a02,0x220aa22a,0xa0228082,0xa0a822a,0xa8220800,0xa002280a,0xa0a88aa0,0x202a2a20,0x288008aa,0x8a0a02a2,0xa0822822,0xa828208,0x22aa2088,0x8080a880,0x2202828,0xa08aa828,0xa288a002,0x8a0282aa,0x2a08a800,0x8802888a,0x80800000,0xa888020a,0x2282820a,0x2002228,0x88802aa,0xa888a282,0x2aa2a002,0x82822822,0xa88800a2,0xaa8a80a2,0x82a88aaa,0x8a02088,0x82a08aa8,0x8820002,0x8a02aa,0x2a22a828,0xa8a8828,0x20022a88,0x8280a88a,0x208a08a2,0xa0a08228,0xaa882202,0x202a2008,0x880a0a20,0x22aa0a00,0xa22800a8,0x80a0aaa8,0x822a0022,0xa2228a,0x20000022,0x220082a0,0xa0a082aa,0x80088a08,0x2200aa80,0x82);
remainders_pos[468].modulo_value=312; remainders_pos[468].bit_mask=create_bit_mask(10,0x22820002,0x20a20808,0x8000a000,0x820208a0,0x28000828,0x82282000,0x20a2080,0x8000a00,0x8820208a,0x800082);
remainders_neg[468].modulo_value=312; remainders_neg[468].bit_mask=create_bit_mask(10,0xa20a0882,0x28228000,0x82088,0x28a0020,0x20800020,0xa20a088,0x82822800,0x8208,0x28a002,0x80002);
remainders_pos[469].modulo_value=1876; remainders_pos[469].bit_mask=create_bit_mask(59,0xaa80a22a,0x200022a0,0x208208a2,0xaa288822,0x8a20880,0xa800208,0x2aa00a,0x8080022,0xa080a0,0x2282880a,0x8200202a,0xa2828a8a,0x2280020,0x800a8022,0x2020222,0xa800a22,0xa20220,0x8a0a2820,0x20020a8a,0x8a0a22a0,0x8280280,0x808028,0x2aa00220,0xa8280,0x82288082,0x88a2a808,0x82082220,0x20002228,0x280aa82a,0x288aaa2,0x8a82aa,0x8228880,0xa2208882,0x882202a8,0x82022,0xaa80282a,0x20008800,0x82028020,0xa202802,0x80a88a,0xa2a2a08,0xa000828a,0x2a008808,0x8088a00,0x288808,0x8808802a,0x28a08002,0x82a2a28,0x288a8080,0xa00a0228,0x200a020,0x80088002,0x2a0a00aa,0xa2020800,0x8aa02208,0x20888222,0x88a208,0x2aa0a880,0xa88a0);
remainders_neg[469].modulo_value=1876; remainders_neg[469].bit_mask=create_bit_mask(59,0x22082a22,0x2888aa20,0xa00a8022,0x222000aa,0x80228008,0x82080a80,0x88a22082,0x8080882a,0x8280820,0xa20a0082,0xa0828a2,0x2a020202,0x8aa008a8,0x88200aa,0x8a8a8a2a,0x20882a2,0x802a8aa8,0x202a0a8,0xa0828202,0x2822a28,0x80a08208,0x88080820,0x22288aa0,0x80820a08,0xa20080a,0x222080,0xa802aa8,0xa888a220,0xa0822022,0xa002222,0x80020a22,0x802a0008,0x2aa0000a,0xaa0a20,0x8880a0aa,0x2200a022,0x28880080,0x28a08a8,0x8228a08a,0x88002002,0x82a22280,0x28880202,0xa2880000,0x800280,0x80a00080,0x8a2,0xa020088a,0x80a222a0,0xa0020008,0x28828a20,0xa8828a0,0x80088a,0xa2028022,0x2a828088,0x2282a80,0xa80002aa,0x88002a00,0x22282000,0x20028);
remainders_pos[470].modulo_value=1880; remainders_pos[470].bit_mask=create_bit_mask(59,0xa0000282,0x8a2a08a2,0x2020228,0xa022880,0x20808208,0x8a02208,0x222208a2,0x2a8222a8,0xa200aa,0x280aa0a,0xa0802808,0xa80802a0,0x2002a0a8,0x8a220022,0x20282808,0xa0200888,0x22a0202,0x28880008,0xa820802,0x280a820,0xa00a2080,0x22000002,0x22022000,0xa00a008,0x2882822,0x20a02882,0xa2228a80,0x8020,0x8aa20008,0xa2808202,0x20008a,0x8a080200,0x802a288,0x8082280a,0xa0882822,0x88200a00,0x88000880,0xa800000,0x800208a0,0xa0082a02,0x28802082,0x80200022,0xa8080a8,0x8222008,0xa2202828,0x8880088,0x2a2a0a80,0xa0a8020,0x80202802,0xa0aa02,0xa8aa008a,0x882a8882,0x208a2088,0x820880a,0xa0208202,0x80022880,0xa2288080,0xa8a20a8,0x828000);
remainders_neg[470].modulo_value=1880; remainders_neg[470].bit_mask=create_bit_mask(59,0x28880a0a,0x2220022,0xa8a82a0,0x8282a008,0xa0080a00,0x80282a80,0xaa2a8022,0x220aa220,0x88228822,0x82082202,0x28082080,0x20008a20,0x288a2020,0x2a2882a,0xa0a0a000,0x28a80000,0x8a228a82,0x20008080,0x8202808a,0x82082020,0x28822808,0xaa088882,0x2a8aa088,0x82802880,0x200a02a,0xa828200a,0x2a2a0200,0x88800a8,0x2228880,0x22080a0a,0x80a80802,0x2008a80,0x8a2200,0x802a082,0x2000a02a,0xa80288,0x88000,0x2088088,0x8828028,0x2080a20a,0xa008280a,0x82888a2,0x2000020,0x80a2a880,0x22a8a020,0x80000800,0xa2228200,0x28200a8,0x8a0208a,0x8028220a,0x20220802,0x220002,0x2802a000,0x80a00082,0x20a8020a,0x88a2008,0x2a200800,0x202a020,0x20888);
remainders_pos[471].modulo_value=1884; remainders_pos[471].bit_mask=create_bit_mask(59,0x220028a2,0xa28a28,0x800a082,0x22002080,0xa28a00,0x2008288,0x80882802,0x208800,0x8028a288,0x82882022,0x28800228,0xa088208,0x820882,0xa08008,0x208828a,0x200a0002,0x8228228,0x2008202,0xa2002082,0x20020008,0x280088,0x208a2000,0x20220a20,0xa088280,0x82002,0x28808a00,0xa20a208,0xa0820,0x28808000,0x80000000,0x2022,0x820a00,0x8208a08a,0xa2022,0x8008200,0x8028220a,0x808a0880,0x8a20,0x82200280,0xa2000800,0x8208008,0x8082008,0x82882882,0x8000a00,0xa282208,0x20020a0,0x8220820,0x8208220a,0x22880022,0x28808228,0x228a280,0x20022080,0x8028220,0x2282008,0x800a28a0,0x208008,0x820a002,0x828a28a0,0x8a28008);
remainders_neg[471].modulo_value=1884; remainders_neg[471].bit_mask=create_bit_mask(59,0xa2882022,0x8220220,0x8008200a,0xa2882800,0x8220208,0x8a080200,0x2082,0x8a00008,0x8202200,0x20028a2,0x20008a20,0x82000280,0x800a0002,0x8200800,0x8a000202,0xa0820882,0xa20a20,0x8a08028a,0x22882802,0x28828800,0x88208000,0xa0022880,0x28a28228,0x82000208,0x80802882,0x20000208,0x82282280,0x808200a0,0x20000808,0x8088088,0x808828a2,0x8028208,0xa002002,0x808228a2,0x800208,0x820a282,0x20000,0x8800228,0xa288208,0x22880080,0xa00800,0x8000a080,0x2002002,0x808208,0x8220a280,0x82882820,0xa28028,0xa00a282,0xa20008a2,0x20000a20,0x8a202208,0xa08a2800,0x820a28,0xa20a080,0x822020,0x8a00800,0x8028208a,0x2022020,0x220800);
remainders_pos[472].modulo_value=472; remainders_pos[472].bit_mask=create_bit_mask(15,0x8a8a220a,0x228a82a8,0x20022820,0xa20aaa0,0x22aa2a00,0x80802a28,0x20808802,0x88a00a2,0x2802202,0x8828a802,0xa000a8aa,0x80aaa08,0x88082880,0xa22a82a2,0xa088a2);
remainders_neg[472].modulo_value=472; remainders_neg[472].bit_mask=create_bit_mask(15,0x202aa82,0xa2020a20,0xa88aa0a8,0x82a82228,0xaa22a288,0x808a2a0,0xa808008a,0x8002882a,0x8a08aa8a,0xa0200a,0x28882022,0x80822280,0x80a008,0x2aa20a2a,0x28002a);
remainders_pos[473].modulo_value=1892; remainders_pos[473].bit_mask=create_bit_mask(60,0xa2888282,0xa82a8080,0x22020a,0x8008282,0xa000a8,0x2208a02,0x2228a800,0xa02222a,0x20a228a0,0x20808002,0xaa880282,0x28a8822a,0xaa022800,0x2a280882,0x28a20a80,0xa2a00a82,0x2aa0880,0xa8a000a2,0x8aaaa208,0x8220a00,0x28220008,0x28028a2,0xa222a222,0x220000a8,0x28020288,0x80a800,0x20020a08,0xa0aa8202,0x8a28080a,0xa0a0a08,0x2028a22,0x82aa0aa,0xa080088,0x2a02002,0x28080280,0xa0000882,0x88a888a2,0xa2802808,0x88288,0xa088202,0x8aaaa20,0xa000a2a2,0x220aa808,0x2a00a8a0,0x2a08a288,0x22028a80,0x2880aa8,0x8822a280,0x28022aaa,0x202088,0xa288a088,0x88880a00,0x2a2888a,0xa208800,0xa000a008,0x28200202,0x8088008,0x202a82aa,0x282228a0,0x8);
remainders_neg[473].modulo_value=1892; remainders_neg[473].bit_mask=create_bit_mask(59,0x2a00020a,0x20220008,0x88aa8a82,0x80880a02,0x88280820,0x82a8028a,0xaa202088,0x828aaa22,0xa822a028,0x2808088a,0x22008a0a,0xa02002a2,0x220aa080,0xa2a0800a,0xa02a8200,0x2a28020a,0x82228008,0x2028082a,0x2222a00,0x80a28288,0x20aa8880,0x8a08a02a,0x22aa22aa,0xaa088820,0xa08a8a00,0x88082080,0xa88a0280,0x20220a8a,0x2a08002,0x82828200,0x8a8202aa,0xa22822,0x82808800,0x8a20a08a,0xa0008a08,0x2888800a,0x200022,0x2a082080,0x800a00,0x82800a8a,0x80222228,0x28802a2a,0x2a822080,0xa2882028,0xa2802200,0xaa0a0208,0x8a008220,0xaa2a00,0xa08a2222,0x8028a800,0x2a002800,0x8208,0x8a220002,0x2a80088,0x28882880,0xa0a88282,0x80000880,0xa8a20a22,0xa0aaa020);
remainders_pos[474].modulo_value=1896; remainders_pos[474].bit_mask=create_bit_mask(60,0x82020822,0x20828a20,0x8280280,0x22022022,0x28800a28,0x8a080202,0x80882800,0xa20a20,0x82000,0x228a0802,0x28220800,0xa28a28a,0xa0820020,0xa20200,0x228028a,0x28000a0,0x8028220,0x282082,0x20000000,0x228820,0x8a20a080,0x28a2820,0x8820020,0x8008a008,0x800220a0,0x20200028,0x8202202,0x820a0080,0x20820000,0x8002082,0x820882,0xa08200,0x88082002,0x80880,0x8800228,0xa20020a,0x822020,0x28a28008,0xa08a208,0x22880002,0x808,0x8280000,0x82802082,0x28008,0x8028800a,0x808a00a2,0x820a00,0x8a28a008,0x208828a2,0x20a28800,0x8200080,0xa08a0000,0x28220208,0x8020a200,0xa0022880,0x8808828,0x80282088,0xa2820802,0x20808208,0x88);
remainders_neg[474].modulo_value=1896; remainders_neg[474].bit_mask=create_bit_mask(59,0x28a00a2,0x28020228,0x80200208,0xa28a28a2,0x20008220,0x200828a,0x2080,0x8228228,0x8800a088,0xa2020082,0x20a28008,0x82202202,0x200a00a0,0x8228a08,0x8a208202,0x82080820,0x820a28,0x8820a00a,0xa0880880,0x8a20028,0x2282008,0x820220a0,0x28828,0x8002080,0x8a2820,0x28a08820,0x8028a28a,0x2820800,0x28028808,0x8008a00a,0x800a0002,0x8200a08,0xa082,0x80800000,0x8a20,0x82288282,0x800a28a0,0x20220800,0x82002280,0xa2000882,0x8808000,0x208088,0x2082802,0x8820800,0x8200082,0x20822,0x8028208,0x2202080,0xa0002022,0x28220008,0x80288008,0x20020880,0x20a28a00,0x8282288,0x208a2000,0x20,0x820a000,0x220a0082,0x28000a00);
remainders_pos[475].modulo_value=380; remainders_pos[475].bit_mask=create_bit_mask(12,0x8802020a,0x280a0000,0x2802a288,0x820008a0,0x22a88200,0x8a0282,0x88280a20,0x200822a8,0x80a20008,0x8228a802,0x20000a02,0xa080802);
remainders_neg[475].modulo_value=380; remainders_neg[475].bit_mask=create_bit_mask(12,0x820a82,0xa0028880,0x208a2200,0xa808028,0xa2200a08,0x8802020a,0x2082a0,0x2800a220,0x8228880,0x2a0200a,0xa8880282,0x2008082);
remainders_pos[476].modulo_value=952; remainders_pos[476].bit_mask=create_bit_mask(30,0x2880a22,0x28a0a280,0xa08a8280,0x2208888a,0x200202,0x202a08,0x800aa208,0x28020802,0x208aa00,0x20a8080,0x88080080,0xaa222208,0x80282a20,0x28a0a2,0x288a0228,0x2880a2,0x28a0a28,0xaa08a828,0x22208888,0x80020020,0x800202a0,0x2800aa20,0x2802080,0x208aa0,0x20a808,0x88808008,0xaa22220,0x280282a2,0x80028a0a,0x88a022);
remainders_neg[476].modulo_value=952; remainders_neg[476].bit_mask=create_bit_mask(30,0x8a00822a,0x20202a00,0x20020a08,0xaa000002,0x88a08a8a,0x80a82280,0x8822280,0x208a800a,0xa802280,0x82820808,0x8800,0x22a2aa80,0x8a02228,0x88a0202a,0xa0028220,0x8a00822,0x820202a0,0x220020a0,0xaaa00000,0x88a08a8,0x80a8228,0xa0882228,0x208a800,0x80a80228,0x8282080,0x880,0x822a2aa8,0xa08a0222,0x88a0202,0x2822);
remainders_pos[477].modulo_value=636; remainders_pos[477].bit_mask=create_bit_mask(20,0x22022882,0x8028820,0x8a002000,0x28a0802,0x8208008,0x8008208,0xa00828a0,0x20028,0x208a008,0xa08a2022,0x8808a20,0x200a208,0xa2800800,0xa28200,0x2082002,0x2002082,0x28020a28,0x800a,0x80822802,0x8228808);
remainders_neg[477].modulo_value=636; remainders_neg[477].bit_mask=create_bit_mask(20,0xa28a2002,0x820028,0x208a088,0x82020082,0xa00800,0x80080280,0x20802020,0x8828820,0x8a002080,0x200228a2,0x228,0x8a002280,0x22080080,0x8220a08,0x8a00a08a,0x82882802,0x20828220,0x88080082,0xa2002,0xa20000);
remainders_pos[478].modulo_value=1912; remainders_pos[478].bit_mask=create_bit_mask(60,0xaa22a8a,0x802a80a2,0x880aa828,0x2a28088,0x8aaaa000,0xa820a28a,0xa088a282,0x2a0a20a2,0x282a000a,0x202a0a82,0x888a2220,0x82a28,0x2822080,0xaa082288,0x2022aa20,0xa20022a,0xa88228a0,0xaa8a2282,0x882028aa,0x2a222088,0x2280a02a,0x20aaa028,0x20a00a0,0x282088a,0x80820a28,0xaaaa0000,0x888a8202,0x882800a8,0xa20a802a,0x802200,0x8a008802,0x22a802a0,0x222a0028,0xaa8082a2,0x20000aa,0x8028a082,0x80a22082,0x80a00a0,0x88280aaa,0xa8a80a02,0x22220888,0xaaaa2808,0x2a8288a2,0xa00a2882,0x8a88008,0xaa08aa88,0x80228820,0x20882,0x2228a820,0x80888a2,0x2882a0a8,0xa8a000a8,0xa8a08a0,0x2a828a22,0xa2a28a08,0x80000aaa,0x2222028a,0x2282aa0,0xa08a02a8,0xa2a88a);
remainders_neg[478].modulo_value=1912; remainders_neg[478].bit_mask=create_bit_mask(60,0x822aa202,0x8a2082a,0x8220a0,0x8a2a0800,0x2222888,0x20a82a02,0x28002a0a,0xa282282a,0xa0a28882,0xa8a2820a,0x2aaa8,0x8880a2a0,0x8a0aa808,0x2280aa00,0xa8aa22a8,0x82a88aa2,0x200aa028,0x2202aa0a,0xa8a022,0xa2aaa800,0xaa0828a2,0xa82228a0,0x8a828828,0x8a0a8002,0x80a82a0,0x22228888,0x20a8a,0xa08820,0x2a8208a2,0x8808aa88,0x288008a,0xaa208a28,0xaaa288a0,0x22080a2a,0x8a888822,0x8a0280a,0x82aa80a,0x80828028,0xa08222,0x2020828a,0xaaaa8000,0x2222a080,0xa20a002a,0x2882a00a,0x80200880,0x22802200,0x8aa00a8,0x888a800a,0xaaa020a8,0x8080002a,0xa00a2820,0x20288820,0x82028028,0xa20a02aa,0x2a2a0280,0x8888222,0xaaaa8a02,0x8aa0a228,0x28028a20,0x2a2002);
remainders_pos[479].modulo_value=1916; remainders_pos[479].bit_mask=create_bit_mask(60,0x82280222,0x280aa882,0x8a8aa228,0xaa80a202,0x88008200,0xa8282a22,0xa0a008aa,0xa0228a0a,0x88a0008a,0x82a28aa2,0xaa088002,0x8a28a280,0x80a2aa8a,0xaa8880aa,0x88888a8,0x20222222,0x2a2200,0x200082a,0x2a082082,0xaaa2280,0x802080a,0x22aa822,0x28208a8,0x22a828,0x880a0a0,0x2a8a2aa2,0xa882a00,0xa0880202,0xa2002a0,0x88aa08a,0xa20aa22,0xa0a8008a,0x8080220,0x800a822a,0x8aa8a2a,0xa0a022,0x282a880,0x2a20828,0x882aa88,0xa020802,0x80288aaa,0x820820a,0xa820008,0x80088a80,0x8888880,0xa2a22222,0x2aa0222a,0x2a2aa8a0,0xa028a28a,0x2800220a,0x28aa28a8,0xaa2000a2,0xaa0a2880,0xaaa200a0,0x288a8282,0xa0082002,0x2808a02a,0x8288aa2a,0x2822aa02,0x8880288);
remainders_neg[479].modulo_value=1916; remainders_neg[479].bit_mask=create_bit_mask(60,0xaa08aaa,0xa082200a,0x2022aa0,0x22082a8a,0x880a88,0x20a0a2aa,0x28288022,0x28aa0282,0x288802,0xa2a022a,0x2280088a,0x2a02a08,0x82a2202,0x22000822,0x20,0xa8aaaaaa,0x88a2aa88,0x8a8880a2,0xa280a80a,0x8222aa08,0x808a8082,0x8aa220aa,0x8a0a8020,0x88aa20a0,0x80082828,0xa202a22a,0x8200a288,0x28008a8a,0x82a88a28,0x80022802,0x82a822aa,0x28208802,0x80808aa8,0x8820aa2,0x802202a2,0x882828aa,0x8a0a2008,0x8a2a80a0,0x800a2200,0x828a808a,0x8a00222,0x80a80a82,0x820a8880,0x8800208,0x80000008,0x2a2aaaaa,0xa228aaa2,0xa2a22028,0x28a02a02,0xa088aa82,0xa022a020,0x22a8882a,0x2282a008,0x222a8828,0xa0020a0a,0x2880a88a,0xa08028a2,0xa0022a2,0xa0aa228a,0x8a00);
remainders_pos[480].modulo_value=120; remainders_pos[480].bit_mask=create_bit_mask(4,0x200a2082,0x20020,0x8080080,0x8208a0);
remainders_neg[480].modulo_value=120; remainders_neg[480].bit_mask=create_bit_mask(4,0xa0822802,0x8028820,0x8008,0x20020);
remainders_pos[481].modulo_value=481; remainders_pos[481].bit_mask=create_bit_mask(16,0xcb1d977e,0x5a672151,0xa6033901,0xb3c43c37,0xad490881,0x4c36288f,0x5580d17,0x652fd29a,0xa2c06a81,0xc451b0cb,0x4424ad7,0xb0f08f36,0x2730197,0x2a13996a,0xfba6e34e,0x1);
remainders_neg[481].modulo_value=1924; remainders_neg[481].bit_mask=create_bit_mask(61,0x2800aa2,0x82aaa808,0x228aa088,0x2a0820aa,0x20800008,0x80aaa002,0x88808082,0xa8220a02,0x2a08a200,0x8882002,0x88a820a,0x282802aa,0x8a22880a,0xa28a00a2,0x222a2a02,0x2a2a220,0x28028a2a,0x808a228a,0xaa00a0a2,0x820a8882,0x208882,0x22882a2,0x28220a8,0x808088a,0x2aa80a,0x80000822,0xa82082a0,0x882a8a22,0x80aaaa08,0x2a800a00,0xa0aa0022,0xa000080a,0x802880,0x2082802,0x28a8aa8a,0xa00022a2,0x8a8a880,0x82282a0,0x2802288,0x888a0aa2,0x88802202,0x20aa000,0x80008a82,0x20822820,0x220200a0,0x80202222,0xa208202,0xa8800082,0x2a82020,0x22008880,0xa82888a0,0xa200a022,0xa0a20808,0x8a8a8802,0xa2000280,0xaa8a8a22,0xa082028,0x8a008020,0x8000280,0x2a82a8,0x2);
remainders_pos[482].modulo_value=1928; remainders_pos[482].bit_mask=create_bit_mask(61,0x228aa02,0xa8a28,0x20a28020,0x2aa202a,0xa8a28888,0xa08a2a2a,0x8a220a2a,0x280828a,0x2a0aaa02,0x2a2a00a2,0x2a022,0x28aa2288,0xa882a0aa,0x2a020888,0x8a202a88,0x802a2082,0x88a2a028,0xa028808,0x82200820,0x2a02aaa8,0xaa02022,0xa000a022,0x828282a2,0x20a22080,0x20208a00,0x88820800,0x2a2002a8,0x2a820a20,0x2080aaaa,0xa00822a8,0x88200a82,0xaa02082a,0xa082a8aa,0x8008a808,0x2082222a,0xa2080800,0x88a0800,0x82828202,0xa000a8a,0x80aa088,0xaa80a888,0x2008822a,0x2280a008,0xa8a2220,0x8a80228,0xa808a282,0x2080a822,0xa822a22,0x88aa28aa,0xa800022,0xa8a888,0xaaa0a88a,0x82028080,0xa088a2a2,0xa8a20aa8,0x228a2aa8,0x8aa8022,0x28a08a8,0xa2a00008,0xaa288028,0x80);
remainders_neg[482].modulo_value=1928; remainders_neg[482].bit_mask=create_bit_mask(61,0x8aa0228a,0x888202a0,0xa82a08a8,0x8a22a8a2,0x202a0000,0x2802a2a2,0x2aa82a2,0x8a080a02,0xa282228a,0xa2a2882a,0x888a28aa,0xa022aa00,0x200a2822,0xa28a8000,0x2a8a200,0x8a2a80a,0x2a28a0,0x828a0080,0xaa880a8,0xa28a2220,0x8228a8aa,0x288828aa,0xa020a2a,0xa82aa808,0xa8a80288,0xa8088,0xa2a88a20,0xa20a82a8,0xa8082222,0x2880aa20,0xa8820a,0x228a80a2,0x280a2022,0x8802080,0xa80aaaa2,0x2a808088,0x80028088,0xa0a0a8a,0x82888202,0x80822800,0x22082000,0xa8800aa2,0xaa082880,0x8202aaa8,0x80208aa0,0x20802a0a,0xa80820aa,0x820aa2aa,0x22a022,0x820888aa,0x88202000,0x22282002,0xa0a0808,0x28002a2a,0x202a8220,0xaa02a220,0x802208aa,0x8a028020,0x2a288880,0x22a008a0,0x8);
remainders_pos[483].modulo_value=1932; remainders_pos[483].bit_mask=create_bit_mask(61,0x82000002,0x20a00a00,0x8028808a,0x2020022,0x20228a00,0xa282200,0xa0020002,0x8828828,0x220a002,0x82080082,0x8000008,0x8a002202,0x820020,0x28820008,0x80082000,0x20802802,0x20800028,0xa082082,0x80820000,0x8a20208,0x208202,0xa0080802,0x20000000,0x82088000,0x80000082,0x8808,0x8a000288,0x20002000,0x800800,0xa280202,0x28a0802,0x200808,0x80008002,0xa2000,0x20000228,0x2202,0x22082820,0x8000,0x200a000,0x20800802,0x8a20808,0x8202208,0x820a0000,0x208820,0x80208280,0x82002802,0x8228000,0x8200200,0x800a2080,0x20808,0x2082200,0xa0880820,0x28220800,0x800a282,0x828a0800,0x28808008,0x808000a,0x22802880,0xa08a20,0x8200a,0x800);
remainders_neg[483].modulo_value=1932; remainders_neg[483].bit_mask=create_bit_mask(61,0x2080802,0x28208200,0x200002,0x820208a2,0x28a20208,0x82202288,0x208a0082,0x20020,0xa282082,0x2800002,0x8800,0x200a28a,0x800a0820,0x20028000,0x800a008,0x20082082,0x20008820,0x200a00a,0x20880,0x220a00,0x8028028a,0x20800002,0x28808800,0x2000088,0x80802,0x8800000,0x2000200,0xa0882080,0x8008008,0x2200282,0x82020082,0x8208000,0x800008a,0x80822880,0x28808220,0x8808a20a,0x228020a0,0x800808,0x8a082080,0xa0000082,0x220000,0x8028a280,0x2820800,0xa00020,0x280208,0x2082082,0x220808,0x80280288,0x822000,0x8828000,0xa00a280,0x200000a0,0x20220008,0x8000220a,0x2020080,0x20000000,0x80008002,0x22082000,0x200228,0x8800a082,0x80);
remainders_pos[484].modulo_value=88; remainders_pos[484].bit_mask=create_bit_mask(3,0xaaaaa2aa,0xaa2aaaa8,0xaa8aaa);
remainders_neg[484].modulo_value=88; remainders_neg[484].bit_mask=create_bit_mask(3,0x22222222,0x22222220,0x220222);
remainders_pos[485].modulo_value=485; remainders_pos[485].bit_mask=create_bit_mask(16,0x91832ad2,0xa6521470,0x5acee01d,0x9ae12c69,0xa395e602,0x8c032302,0xeb21c010,0x30120334,0xe135cb,0x31300c42,0x19ea7150,0x8d21d650,0x1dcd6a5,0x8a12996e,0xd5306243,0x12);
remainders_neg[485].modulo_value=1940; remainders_neg[485].bit_mask=create_bit_mask(61,0x80a220a,0x2a0a88a0,0x2022080,0x220a0a0,0x22082a0a,0x8a2a8a,0x22280880,0x288a82a8,0x8820a802,0xa0a88002,0x80222888,0x280a80,0x8000228,0x28200a2,0xa8aa0a,0xa8a8028a,0x20a0282,0x8228,0x80a000a2,0xa220080a,0x8a82888,0xa8208a00,0xa88a200,0x80a220aa,0xa2880008,0xa082228a,0x28220282,0x22020028,0x8a82a008,0x22808028,0x80220282,0x80080a82,0x2208a200,0x82020a2,0x82008222,0x80000202,0xa088aa0,0x800220,0x802a008,0xa8800a,0x20080808,0xa0088a2,0x2a8202a8,0x82a0208a,0x8220,0x8a00288,0x82208000,0x288202a0,0xaa020aa,0x22888028,0x8080802,0x28008a80,0x802a008,0x2200080,0x82a88828,0xa0200000,0x22208020,0x22820208,0x80228822,0xa0a80800,0x2200);
remainders_pos[486].modulo_value=24; remainders_pos[486].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[486].modulo_value=24; remainders_neg[486].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[487].modulo_value=1948; remainders_pos[487].bit_mask=create_bit_mask(61,0x2aa00a8a,0xa8228a02,0x82a2aa22,0xa00000a,0xa888220a,0x22aa820a,0xa80888aa,0xa08aa08a,0xa8282820,0x8200220a,0x8828a08,0x2a8a88,0xa80a2208,0x2020880,0xa28220a8,0x8a08020a,0x8a8aa028,0x82a02a22,0xaaa288,0xa22a202,0xaa8a2282,0xa0a00288,0x22828a0,0x22a00228,0x88022,0xa220028a,0xaa828288,0x80a02aa,0x8a00880,0xa8008a82,0xa8080202,0x82a2002,0x2200a2,0xaaa80a02,0xa228282a,0xa280088,0x88802200,0x28800a8,0xa0a28288,0xa22800a0,0x8288a2a,0x808a88a,0x2228aaa0,0x288a80a8,0x2280aa2a,0xaa08020a,0x2a08828,0xa0220808,0x2088a02,0x22a2a80,0x220a2822,0xaa088008,0xa0828282,0xaa20aa20,0x8aa22202,0xaa082aa8,0xa088222,0x2a00000a,0xa88aa8a8,0x880a2882,0xa2a00aa);
remainders_neg[487].modulo_value=1948; remainders_neg[487].bit_mask=create_bit_mask(61,0xa2288202,0x20aa028a,0xa2a22aa,0x82888882,0x2000aa82,0xaa220a82,0x20800022,0x28022802,0x20a0a0a8,0xa88aa82,0x800a0280,0x88a20200,0x2082aa80,0x8a8a8008,0x2a0aa820,0x2808a02,0x20228a0,0xa28a2aa,0x88222a00,0x82aa2a8a,0x2202aa0a,0x28288a00,0x8aa0a028,0xaa288aa0,0x888008aa,0x2aa88a02,0x220a0a00,0x80828a22,0x80288008,0x2088020a,0x20808a8a,0x80a2a88a,0x88aa882a,0x2220828a,0x2aa0a0a2,0x82a08800,0x8aa88,0x8a008820,0x282a0a00,0x2aa08828,0x80a002a2,0x80802002,0xaaa02228,0xa0020820,0xaa0822a2,0x22808a82,0x8a2800a0,0x28aa8080,0x8a80028a,0x8aa2a208,0xaa82a0aa,0x22800880,0x280a0a0a,0x22a822a8,0x22aaa8a,0x2280a220,0x82800aaa,0xa2888882,0x20022020,0x82a00a,0x2a28822);
remainders_pos[488].modulo_value=488; remainders_pos[488].bit_mask=create_bit_mask(16,0x22208a02,0xa2a8aa8,0xa822022a,0x828a00aa,0x28000aa0,0x202022a2,0x22280080,0xa82a08a,0x2888a2,0x88080802,0xa000288a,0xa2820a,0x80882aaa,0xa2a8a0a8,0xa208882a,0x80);
remainders_neg[488].modulo_value=488; remainders_neg[488].bit_mask=create_bit_mask(16,0xaaa8028a,0x82a20220,0x20aa8aa2,0xa028822,0xa0888228,0xa828aa2a,0xaaa08808,0x820a2802,0x88a0002a,0x80808a,0x2888a002,0x882a0a82,0x800a222,0x2a202020,0x2a8000a2,0x8);
remainders_pos[489].modulo_value=1956; remainders_pos[489].bit_mask=create_bit_mask(62,0x22820822,0x28820800,0xa200000,0x22080822,0xa20220,0x88000200,0x82800080,0x20a08200,0x8008200a,0xa28a00a2,0x200,0x8288202,0x20a20a0,0x28808220,0x82280282,0xa20a28a2,0x8a20a00,0x82002208,0xa0802822,0x20208,0x8a280200,0x800a2802,0x28228020,0xa080208,0x880800,0x28000200,0x82202202,0x22822080,0x8a00000,0xa200082,0x208a2082,0x20008a08,0xa208,0x20882880,0x8808820,0x8000288,0x2022000,0x8020a00,0x80288282,0x28a0020,0x8028a28,0x8080000,0x828020a2,0x8800828,0xa08a202,0xa28a08a0,0x28028828,0x88202288,0xa08a0800,0x8228200,0x8000008,0xa00a28a0,0x820028,0x820a08a,0x20002820,0x8000220,0x8808a000,0x82020880,0x8a08,0x2082280,0x82082880,0x8);
remainders_neg[489].modulo_value=1956; remainders_neg[489].bit_mask=create_bit_mask(61,0xa20a00a2,0x20028008,0x82288088,0xa28000a2,0x8228a28,0x88280,0x2080800,0x28200a08,0x800a082,0x22020822,0x8808a08,0x8020028a,0x82822820,0x20000a28,0xa20820a,0x22822022,0x228208,0xa08a280,0x200820a2,0x8828a00,0x2208288,0x822082,0x20a20828,0x82008280,0x80000080,0x20800a08,0xa28a28a,0xa20a2800,0x208808,0x8228800a,0xa0022802,0x28800200,0x88082280,0xa0002000,0x28,0x80088200,0x828a2880,0x828208,0x820020a,0x820208a0,0x820220,0x80008088,0x2082822,0x8020,0x8200228a,0x22020020,0x20820020,0x28a200,0x20020080,0xa20a08,0x80088080,0x20822020,0x8028820,0x80282002,0xa08820a0,0x808a28,0x2088,0x28a0000,0x8800200,0x8a00a208,0x2802000);
remainders_pos[490].modulo_value=280; remainders_pos[490].bit_mask=create_bit_mask(9,0x8800220a,0x200aa0,0x220a8088,0xa208800,0xa0800002,0x88002208,0x2202a0,0x220aa008,0xa08800);
remainders_neg[490].modulo_value=280; remainders_neg[490].bit_mask=create_bit_mask(9,0x882a02,0x8288220,0x22820000,0x82200088,0x2000880a,0x882a80,0x882a8220,0x2a822000,0x200080);
remainders_pos[491].modulo_value=1964; remainders_pos[491].bit_mask=create_bit_mask(62,0x28a22a2,0xa802a22a,0xa8a2000a,0x22a800a2,0x82800022,0x82208aa2,0x2828a020,0xaa22aa2,0x8820a80a,0x888a0aa2,0x28a8a2aa,0xa0028280,0x2a2202a0,0x8aa8aa82,0x8aa22280,0x20802a2,0x22a0888,0x808a0020,0xa8280a88,0x20aa0a0a,0x22000820,0xa2080282,0x282a028,0xa0880088,0x8a28a820,0xa080228,0x888aaaa,0x20082aa0,0xa002aa88,0xa80880a,0x2080882,0x202a0822,0xa800aa02,0x8200822a,0xa22200aa,0x20a0aaa,0xa28a2288,0x220a082,0xa8280220,0x208a280,0x88828,0xaa08082,0x282aa0a,0xa20222a,0xa880080,0x2080222,0x88aa28a8,0xa2aa2028,0x88a882a,0x2800a0a8,0xa2a28028,0xa222aa8,0xa08228aa,0x88aa0a02,0xa28288aa,0x20882080,0x2828aa,0x2a88880,0x8a2a8a0,0xa802aa00,0x8a280a88,0x8a8);
remainders_neg[491].modulo_value=1964; remainders_neg[491].bit_mask=create_bit_mask(62,0x8a02aa2a,0x208a2aa2,0x202a8882,0xaa20882a,0xa0888aa,0xaa8022a,0xa0a028a8,0x822aa22a,0xa82082,0x2822a,0xa0202a22,0x288a0a08,0xa2aa8a28,0x220220a,0x22aaa08,0x8a80822a,0x8aa28000,0x80288a8,0x20a08200,0xa8228282,0xaa8880a8,0x2a808a0a,0x8a0a28a0,0x28008800,0x2a020a8,0x82808aa0,0x80002222,0xa880a228,0x288a2200,0x82080082,0x8a80800a,0xa8a280aa,0x2088228a,0xa880aa2,0x2aaa8822,0x8a828222,0x2a02aa00,0x8aa8280a,0x20a08aa8,0x8a802a08,0x888000a0,0x8228080a,0x8a0a2282,0x82a8aaa2,0x82008808,0x8a808aaa,0x22a020,0x2a22a8a0,0x800200a2,0xa0882820,0x2a2a08a0,0x82aaa220,0x280aa022,0x22828a,0x2a0a0022,0xa800a808,0x88a0a022,0x8a200008,0x802a2028,0x208a2288,0x2a08200,0x20);
remainders_pos[492].modulo_value=984; remainders_pos[492].bit_mask=create_bit_mask(31,0x228a0082,0x28a20020,0x208820a,0x2800820,0x8a08228,0x8202,0x880880,0x8020a00,0x220000a,0x20000022,0x20a20808,0x280288,0x22080082,0x8a28a08,0xa288080,0xa0082820,0x20020228,0x8820a28a,0x820020,0x8228028,0x820208a,0x80880000,0x20a00008,0xa080,0x22022,0x20808200,0x8028820a,0x80082002,0x28a08220,0x8808008a,0x8200a2);
remainders_neg[492].modulo_value=984; remainders_neg[492].bit_mask=create_bit_mask(31,0xa2020802,0x20228828,0x8a000282,0x820800a0,0x200a20,0x8808028a,0x80000000,0x828208,0xa288082,0xa08808a2,0x28228000,0x88208200,0xa2800802,0x220200,0x82200000,0x208020a0,0x28828a20,0x282202,0x800a08a0,0xa20820,0x8028a002,0x880,0x28208800,0x88082008,0x808a20a2,0x28000a08,0x8200282,0x802882,0x20200a28,0x2,0xa0822);
remainders_pos[493].modulo_value=493; remainders_pos[493].bit_mask=create_bit_mask(16,0xca017e1a,0x293357fa,0xe44a8a09,0x5b2ed29a,0x95b9d43c,0x1cf031b,0x8cd90613,0x4618b803,0x26cc7007,0x3ce03218,0xe76a7630,0xdd368f0a,0x5489d652,0xb3252414,0xa014d7fa,0x161f);
remainders_neg[493].modulo_value=1972; remainders_neg[493].bit_mask=create_bit_mask(62,0x288a282,0xa0a28a22,0x28020280,0x82220a02,0x82008a0,0x80028a82,0x828a,0xa80208a,0xaa00a88a,0x2028aa80,0x2aa2aaa0,0xaa0282,0x88000a0a,0x2aa0a888,0x28880222,0xa20a8a82,0xa2220088,0x8888a82a,0x2828000,0xa20a02a8,0x202aaa2a,0xa80aa8a0,0x8288a802,0x288200a,0xa8a0800,0x820a8a00,0x8288020,0xa2028222,0x280a0200,0x2228a28,0xa20a288a,0x22208020,0xa8280082,0xa20822a0,0x28a02282,0xa2822288,0x808aa080,0x822aa880,0x80aa28,0xa002a808,0x82a0800,0x22002000,0x82a002a0,0x888aaa80,0x22000800,0xa2088a82,0x88228000,0x2220a8,0xa8888008,0x2a080aa,0x2202a0,0x2a080002,0xa0028008,0x8080080a,0xaa208a2a,0xa880808a,0x20a28082,0x28a08a2,0x822a0a2,0xa0a82a2,0x82222080,0x28200);
remainders_pos[494].modulo_value=1976; remainders_pos[494].bit_mask=create_bit_mask(62,0x220202a2,0x8822aa0a,0x22228028,0xa2808082,0x2220808,0xa0a020,0x8a02822,0x2a228aaa,0x8a00a220,0xa08a282,0x2aa88a82,0x82a0080,0x2aa22808,0x2a800288,0xa0aaa022,0x202a28,0x220a0008,0x28822a00,0xa8a80000,0xa82a020,0xa0008800,0x28008a0,0x2a2202a0,0x20008200,0xa0282208,0xaa200a0a,0x8288222,0x28a8a02,0x802a8220,0x20800220,0x820220a2,0x88a0880,0x2088002,0x800882a8,0x2080a2a2,0xaa888228,0xaa0a008,0x8208828,0xa8008200,0x800a8088,0xa0a2002,0xa0002200,0x2a080a82,0x2800002a,0x8800a882,0x2000a0,0xa28a808,0xa8880aaa,0xa8228002,0x2020288a,0xa80200a8,0xa082a22a,0xa2828a20,0xa8088a00,0x20aaa288,0x88280a,0x80080a0a,0x8a202088,0x88820202,0x22280288,0x88a0aa88,0x8a8080);
remainders_neg[494].modulo_value=1976; remainders_neg[494].bit_mask=create_bit_mask(62,0xaa828a2a,0xaa2202,0x22aa08a0,0x2a08080a,0x8aaa0080,0x882820a8,0x8028a0a2,0xa22a0222,0x2882aa8,0x8280220a,0x22200202,0x80a28808,0xa2222080,0xa2088a00,0x2022282a,0x80a8a2a0,0xaa828880,0xa00a2208,0x20208888,0x820228a0,0x28080088,0xa088028,0xa2aa8228,0x28880a88,0x28a0a280,0x22a08282,0x80200aaa,0x8a02020a,0xa20aa8,0xa8088aa0,0xa8a282a,0x80020008,0x8a800882,0xa20,0xa8082a2a,0x220002a0,0x2282800,0x80a800a0,0x20800a88,0x8820800,0x282a80a,0x2088aa88,0xa280820a,0xa08800a2,0x88200a,0x88288820,0x82202080,0x20008222,0x20aa0082,0x28a8a002,0x208a0820,0x28022aa2,0x220a02a8,0x20800208,0xa0222a00,0x8800a002,0x8800282,0x2a0a800,0xa8a82,0x2a208a00,0x282200,0x20008);
remainders_pos[495].modulo_value=660; remainders_pos[495].bit_mask=create_bit_mask(21,0x8a2002,0x28000,0x2008208,0x80,0x800808,0x202208,0xa2080822,0x808220,0x8020a008,0x2822,0x20008,0x80288280,0x200200a0,0x8a08820,0x80088202,0x20020880,0x20200,0x8002000,0x20820,0x20000028,0x8008a);
remainders_neg[495].modulo_value=660; remainders_neg[495].bit_mask=create_bit_mask(20,0x80022082,0x8020800,0xa080280,0x80800800,0x80282280,0x220000a2,0x8000228,0x8202080,0x882022,0x8820800,0x8200200,0x200a0020,0x200028,0x202,0xa08a0000,0x28a00,0x8a088,0x80820020,0x20808820,0x2);
remainders_pos[496].modulo_value=248; remainders_pos[496].bit_mask=create_bit_mask(8,0xa808a2a,0x822a02,0x80a8820,0xaa8a202a,0x20a808a2,0x822a0,0xa080a882,0xa8a202);
remainders_neg[496].modulo_value=248; remainders_neg[496].bit_mask=create_bit_mask(8,0x20802a2,0x880aa28a,0x808200a8,0x2202a8a2,0xa020802a,0x8880aa28,0x2808200a,0x202a8a);
remainders_pos[497].modulo_value=1988; remainders_pos[497].bit_mask=create_bit_mask(63,0xa202a202,0x2a808a22,0x80028020,0x2282802,0x82800880,0x80282200,0x28020208,0xa202020a,0x8a02aa2,0x288202,0x2008a020,0xaaa22282,0x88082208,0xa0aaa280,0x808aa000,0x228a202,0x2a880a,0xa2aa828,0x82208088,0xa222aa8,0x20288022,0x208a000,0x2aa02882,0x8202022a,0x820200a2,0x220a008,0x8800a08,0xa0a200,0x200a000a,0x22880aa0,0x22a022a,0x880a880a,0xaa02288a,0xa0080,0x8a0a00a,0xa002200,0xa08802,0xa0080822,0x88080828,0x2280aa8a,0xa20808,0x80228080,0xaa888a08,0x20208822,0x82aa8a02,0x22a8002,0x8a2880a,0xaa2028,0x28aaa0a0,0x8820220,0x2888aaa2,0x80a20088,0x8228000,0xaa80a208,0x80808a8,0x808028a,0x8828022,0x22002820,0x2828800,0x80280028,0x8a202a80,0x8a808a8,0x8);
remainders_neg[497].modulo_value=1988; remainders_neg[497].bit_mask=create_bit_mask(62,0x2a8a2a0a,0x220802a2,0x8a0828,0x8a20a08a,0xa008008,0x8a02a88,0xa08a8280,0x2a8a8a02,0x28a222,0x80a00a8a,0xa80028a8,0x2222aa02,0x802a80,0x28222208,0x8022808,0xaa02a82,0x80a20082,0x822220a0,0xaa00800,0x2aa2220,0xa8a000aa,0x8a802808,0x2228a002,0x28a8aa2,0xa0a882a,0x8aa02880,0x80080280,0x88282288,0xa0828802,0x2a008220,0x82a28aa2,0x20082,0x2282a002,0x88820808,0x80282082,0x8288aa08,0x8280082,0x200080aa,0x80a0,0xaa002202,0x882a0080,0x8aa0008,0x22000200,0x28a800a2,0x222028a,0x8a22088a,0x80220082,0x882228a0,0xa0222028,0x800a8a28,0x20002222,0x2a8800,0x802a0888,0x22002a80,0x80800020,0x80800202,0x800a082a,0x2a88a0a0,0x820a0088,0x82088a0,0x2a0a208,0x80200020);
remainders_pos[498].modulo_value=1992; remainders_pos[498].bit_mask=create_bit_mask(63,0xa2082882,0x8028800,0xa202088,0x82080000,0x20808a28,0x220a208,0x20a28a2,0x828208,0xa080282,0xa28a2022,0xa00a20,0xa002200,0x22080880,0x8000028,0xa208002,0x22820020,0x28820028,0xa28a082,0x80882000,0x20228028,0x28028a,0x22000002,0xa08020,0x82008282,0x20a0,0x20020220,0x88208,0xa2882080,0x8200028,0x8800a088,0x80082002,0x8200200,0xa002280,0x82022,0x8228a28,0x82200002,0x80800820,0x8000008,0x8200820a,0x20a0082,0x8808,0x80280080,0x28808a2,0x8220228,0xa28a000,0x822882,0x828828,0x208a008,0x2080,0x20208828,0x8800a002,0xa00a0000,0x8a28a08,0x8020a088,0x82820082,0x28a08020,0x8a08808a,0xa2020820,0x208228,0x808a000,0x22802022,0x28208a00,0x82);
remainders_neg[498].modulo_value=1992; remainders_neg[498].bit_mask=create_bit_mask(63,0x22802002,0x820008,0x8220a000,0x2800880,0x28000220,0x8a282280,0x82822022,0x8020a00,0x8200820a,0x220228a2,0x8208228,0x8208a288,0x22800000,0x808820,0x8228008a,0xa20a08a0,0x20028820,0x8220200a,0x2880,0x28a20820,0x88208202,0xa2880882,0x8200828,0xa08020a,0x80882820,0x28828a28,0x88000280,0x22002800,0xa08820,0x82000,0x802882,0xa08a08,0x8208a208,0x800028a2,0xa20220,0xa28808a,0x800a0,0x808800,0xa080282,0x82820802,0x8800000,0x8208008,0x82000022,0xa28a20,0x82202080,0x800a2002,0x8020020,0x8a002080,0x80882800,0x28a00020,0x8208a,0x20820880,0x220200,0x8282000,0x20a0802,0x20200828,0x2000002,0x228a00a0,0x8a00a20,0x80002088,0xa20828a2,0x20a00208,0xa);
remainders_pos[499].modulo_value=1996; remainders_pos[499].bit_mask=create_bit_mask(63,0x2aa88aaa,0x8a82200a,0x808aaaa0,0x22082228,0xa02a0,0xa02a8a2a,0xa8228820,0x22282282,0x2a288022,0xa88082a2,0xa820080,0x8880aa0a,0xaa2aa802,0xa2a0028,0x88008202,0x8a02820a,0x8a2aa202,0xaa80828a,0xa0008020,0x802a220a,0x2a8a0282,0xa2a202a,0xaa208088,0x8a08888,0xa208a00a,0x8200a828,0xa82aa80,0x88a288a8,0x22a20,0xa88a0228,0x2200082,0x880000,0xa22a820,0x88000288,0x28a2208a,0xa82a02a2,0xa008202a,0xa208a282,0x20a20a00,0x208aa222,0x8a8a0220,0xa2a8a80,0x8a802828,0x2000aa08,0x202aa080,0xaa8a2a28,0x280a2808,0x20022a08,0xa8a0808,0xaa8aa280,0xa0222802,0x82a0a0a,0x2022a020,0x228a88a8,0x82888880,0x2882a828,0x2a80a082,0xa000a8a,0x820880a8,0xaa202288,0x882a20aa,0x22aa8a00,0xaaa);
remainders_neg[499].modulo_value=1996; remainders_neg[499].bit_mask=create_bit_mask(63,0xa2200222,0x20aa882,0x8022228,0xaa80aaa0,0x88828a28,0x28a202a2,0x20aa00a8,0xaaa0aa0a,0xa2a008aa,0x20080a2a,0x820a8808,0x82282,0x22a2208a,0x82a288a0,0x880a8a,0x2820a82,0x2a22a8a,0x22080a02,0x288808a8,0x8a2aa82,0xa2028a0a,0x82a2a8a2,0x22a80800,0x80280000,0x2a802882,0xa8820a0,0x820a2208,0x2a0020,0x888aa2a8,0x20028aa0,0x8aa8880a,0x88008888,0x82aa20a8,0x888a00,0xa02aa802,0x20a28a2a,0x2880a8a2,0x2a802a0a,0xa82a8288,0xa8022aaa,0x2028aa8,0x82a20208,0x208a0a0,0xa8882280,0xa8a22808,0x2202a2a0,0xa082a080,0xa88aa280,0x82028080,0x22022a08,0x28aaa08a,0x80a28282,0xa8aa28a8,0xaa020020,0xa000008,0xa00a20a0,0xa208280a,0x82888202,0xa800820,0x22a8aa00,0xa2a822,0xaa220288,0x222);
remainders_pos[500].modulo_value=40; remainders_pos[500].bit_mask=create_bit_mask(2,0xa0280a02,0x80);
remainders_neg[500].modulo_value=40; remainders_neg[500].bit_mask=create_bit_mask(1,0x28a0028a);
remainders_pos[501].modulo_value=2004; remainders_pos[501].bit_mask=create_bit_mask(63,0x828a00a2,0x28220208,0x80280080,0xa28a0022,0x28208028,0x208202,0x820820a0,0x200028,0x88002202,0x80020082,0x8820008,0x2088208,0xa28220a0,0x8220a20,0x8a000000,0x2020022,0x28a20,0x82208000,0x220a2822,0x8820028,0x2088082,0x2000880,0x20008a08,0x200202,0x208208a0,0x8200028,0x820a202,0x28a28a0,0xa00a20,0x220a008,0x28a0882,0x28028a28,0x8808220a,0xa00200a0,0x28008a00,0x8200a28a,0x820808a0,0x20828000,0x8000a208,0x880800,0x8020a20,0x8002200,0x22082022,0x8828008,0x8828828a,0x20,0x8008a28,0xa288008,0x82000000,0x28a08a08,0x800a088,0x22020822,0x220008,0x2282008,0x80080880,0x8228000,0x8000a082,0x82880820,0x28a28020,0x8028800a,0x82802002,0x28220808,0x8a00a);
remainders_neg[501].modulo_value=2004; remainders_neg[501].bit_mask=create_bit_mask(63,0x2020822,0x20a28a00,0x8208008,0x220208a2,0x20a00820,0x8828020a,0x2802820,0x8a08820,0x8a28a,0x8a0802,0x28800,0x8a000280,0x220a2820,0xa28228,0x2088088,0x828a08a2,0x8820228,0xa280088,0xa28220a2,0x28820,0x8a00000a,0x82880000,0x28800200,0x8828828a,0xa00a0020,0xa08820,0x80282282,0x82022020,0x8208228,0x8a282080,0x82020002,0x20820220,0xa282,0x208a0820,0x20800208,0xa082202,0x2800020,0x28020808,0x8082280,0x80000080,0x828228,0x8008a288,0xa28028a2,0x20800,0x200202,0x808808a0,0x800220,0x82200080,0x2880880,0x20200200,0x80082000,0xa28a00a2,0x8a28800,0x8a20a080,0x800000,0xa20808,0x808200a,0x20000a0,0x20220828,0x8200082,0x2082882,0x20a28000,0x2082);
remainders_pos[502].modulo_value=2008; remainders_pos[502].bit_mask=create_bit_mask(63,0x2a02020a,0x20aa8228,0x8a8a2a8a,0x8a0aa200,0xaa88a028,0x808802a2,0x8aaa0288,0x228a82aa,0x20222820,0x80028a2,0x222a2880,0xa0028a00,0x2008aa20,0xaa022a0,0xa282a822,0xa20a8a8,0x2820808,0x2a00a228,0x8aa2a02,0x82aa0a20,0x20222aa0,0x2aa8aa88,0x222a2208,0x80802a28,0xa0008002,0xa88a8882,0x8800202,0xa0882a,0x20808aa2,0x800a2080,0xa2a02822,0x8a8a00a2,0x288280a,0x80208a0,0x8aa202,0x20a8220a,0x2a808002,0xa8222a2,0x2800200,0x8828a802,0xa82088a8,0x822aa2a,0x820aa888,0x2008a0aa,0xa880a8aa,0x80288a00,0xa0202082,0x8a2a2a08,0xa0882a82,0x80a880a,0xa08aa20,0x8800a280,0x200228a8,0x88a2800,0x88082888,0xa2aa82a2,0x22280aa,0xaa8a8022,0xa2280a22,0xa2008aa0,0x8a2a8a2,0xa82882aa,0xa08080);
remainders_neg[502].modulo_value=2008; remainders_neg[502].bit_mask=create_bit_mask(63,0xa28a8a82,0xa8220aa0,0x202a202,0x2822a88,0x220028a0,0x8008a2a,0x2228a00,0xa2020a22,0xa8aaa0a8,0x8088a02a,0xaaa2a008,0x288a0288,0xa88022a8,0x8228aa28,0x2a0a20aa,0x82a82020,0x8a0a8080,0xa2882aa0,0x8022a28a,0xa2282a8,0xa8aaa228,0xa2202200,0xaaa2aa80,0x808a2a0,0x2888088a,0x2002000a,0x80088a8a,0x882800a2,0xa808022a,0x882a808,0x2a28a0aa,0x202882a,0x8a00a082,0x808a8028,0x88022a8a,0xa820aa82,0xa208088a,0x820aaa2a,0x8a088a88,0xa0200a,0x20a80020,0x80aa22a2,0xa822000,0xa8802822,0x20082022,0x8a00288,0x28a8a80a,0x2a2a280,0x2800a20a,0x80820082,0x828022a8,0x882a08,0xa88aa020,0x8002a088,0x80a000,0x2a220a2a,0x8aaa0822,0x220208aa,0x2aa082aa,0x2a880228,0x802a202a,0x20a00a22,0x280808);
remainders_pos[503].modulo_value=2012; remainders_pos[503].bit_mask=create_bit_mask(63,0x8228a202,0x208a000a,0xa222a0,0x8aaa8882,0x80822882,0xaa0a2a8a,0x882a28,0x2a80aa2a,0xa2820a0a,0xa888a2a2,0x8008aa0,0xa80280,0xa88aaaa0,0xa282a028,0x8aaa0888,0x2208a880,0x80022a20,0xa082222,0x22020a8,0xa02aa800,0x2aa2aa0a,0x22202802,0x8a080808,0x2a008282,0xa22a8080,0x82802080,0x8a2a0200,0x20a20,0x882a8a22,0x8228a808,0x208a02aa,0x822a0a88,0x2aa80a20,0x2202a288,0x88a2a82,0x208a0800,0x20080a8a,0xa0208028,0x80202a88,0x2828200a,0x8202020a,0x88028088,0xaa0aa8aa,0x2aa80,0x2a08088,0x2888820a,0x808a8800,0x2022a208,0xa2220aaa,0xa280a828,0xaaaa22,0x2802a0,0xa0aa2002,0xa8a8a222,0x8a0a0828,0xa8aa02a,0xa28a8220,0x2a2a8a0a,0x28228820,0x8222aaa,0x80a888a0,0x2a000a20,0x808a288);
remainders_neg[503].modulo_value=2012; remainders_neg[503].bit_mask=create_bit_mask(63,0xaa02a8a,0xa8028882,0x882aaa28,0x222000a,0x80aa00a,0x2282a202,0x8800a2a0,0xa20822a2,0x2a0a8282,0x20002a2a,0x80880228,0x88208a08,0x20022228,0x2a0a28a0,0x2228000,0xaa002008,0x88aa2a8,0x8280aaaa,0x8aa8a820,0x28a22088,0xa22a2282,0xaaa8a08a,0x2808080,0xa2880a0a,0x2aa20808,0xa08a808,0x2a28a88,0x888a82a8,0xa202aa,0xaa02080,0xa8028a22,0xaa28200,0xa22082a8,0xaa8a2a00,0x8002a20a,0xa8028088,0xa8808202,0x28a808a0,0x8a8a200,0xa0a0a882,0xa8a8a82,0x8a0800,0x22822022,0x888a2208,0x8a280800,0xa0000a82,0x8020088,0xa8aa2a80,0x2aaa8222,0x2a0820a0,0x882222aa,0x88a08a28,0x2822a88a,0x20202aaa,0x28280a0,0x820228a2,0x2a020aa8,0xa2a20282,0xa0aa00a8,0x80aaa222,0x8200028,0xa28882a8,0x802a00);
remainders_pos[504].modulo_value=168; remainders_pos[504].bit_mask=create_bit_mask(6,0x82002822,0x20808800,0xa,0x220208a0,0x28008200,0x88);
remainders_neg[504].modulo_value=168; remainders_neg[504].bit_mask=create_bit_mask(5,0x2882022,0x28000000,0x80088082,0xa20a0020,0x20800a08);
remainders_pos[505].modulo_value=505; remainders_pos[505].bit_mask=create_bit_mask(16,0x992d53de,0xc3630451,0x201a9389,0xd2071485,0xc10dc069,0x2c282bcc,0x2b384211,0x7b0ca015,0x52a014c3,0xd2210873,0xccf5050,0x2e580ec2,0x1284a381,0xe472560,0x6628831b,0x1ef2ad2);
remainders_neg[505].modulo_value=2020; remainders_neg[505].bit_mask=create_bit_mask(64,0xa00a02,0xa2a8880,0x28820200,0xa828808,0x800820,0xa0a02200,0xa2208a80,0x22802088,0x8a20880a,0x2a8002a,0x80020888,0xa2008202,0xa082208,0x8282a8a8,0xa2a00a02,0x2022220a,0x2a2a8222,0xa0a0280,0x8280a8aa,0x2288220,0xa0208,0xaa008882,0x228aa000,0xa228088,0x22288820,0x28280a88,0x8000220,0xa802080,0x8a0808a,0xa2800202,0x2800088a,0x82220280,0x88aa028,0x82a28000,0x82080822,0x882a0080,0x222a88a2,0x882028,0x8802028,0x8280202a,0xa8000208,0x880a8000,0x2280a2a0,0x28020a8,0x8820a,0x80a02080,0x22000022,0x22022220,0x80a20000,0x8202,0xa0282088,0xa20a8200,0x882a280,0xa8000a8,0xa0882000,0x882a0200,0x800a0200,0x220a0208,0x8a288aa,0x2080802a,0xa0a20808,0x880000a2,0x208a02a8,0x2);
remainders_pos[506].modulo_value=2024; remainders_pos[506].bit_mask=create_bit_mask(64,0xa20222a2,0x8222a028,0x822808a,0x28a00020,0x2a02080a,0x2a2a200,0xaa080820,0xa220822,0x22a00aaa,0x8a8000a,0x288aa2a2,0x2a08028a,0x8288820,0x820082a8,0xa028a2a2,0x2a22a82,0x22008002,0x22800208,0x8a8020a,0xa0220888,0x2080a8,0x8820200,0x220a0a8,0x202000a8,0xa00208a2,0xaa2880,0xa0222202,0x20a882,0x28a82aa8,0x2800a02,0x82002a02,0x820022,0xa8008288,0xa0028080,0xa82a2880,0x2a08002a,0x88880a82,0x28aa0080,0x20800a02,0x8088a,0xa08802a,0x8082202a,0x2080000,0x20880a2a,0x802a2022,0x800288a0,0x2008820,0xa88a8080,0x8a280a82,0x8200828a,0x2228202a,0x8020a808,0x8aa228a2,0x2a208a,0xa00a88a0,0x2088a0aa,0x2020aa88,0x8a8a8008,0x2080a800,0xa28a0,0x2882008,0xa8882a2,0x88808a28,0x8a);
remainders_neg[506].modulo_value=2024; remainders_neg[506].bit_mask=create_bit_mask(64,0x2a0aa22a,0xa2a28a0,0x80aa0802,0xa02088a0,0xa28a0082,0x822a2a88,0x228000a8,0x82aa802a,0xaa208222,0x208082,0xa0022a2a,0xa2808202,0x80200028,0xa880a20,0x28a02a22,0x8a2a2202,0xa288088a,0x2a088a80,0x80208a02,0x28a28000,0xa80820,0x800a8a88,0x8aa82020,0xa8288820,0x288a802a,0x8822a000,0x28a22a8a,0x80a8200a,0xa020a220,0x8a08020a,0xa80a28a,0x80a88aa,0x20880200,0x288a0008,0x2022a008,0xa2808822,0x8202,0xa0220808,0xa0088282,0x88808002,0x2800822,0x802a8a2,0xa808888,0xa00082a2,0x8a2a0aa,0x80a0028,0x8a0800a8,0x20020800,0x2a0020a,0x2800a02,0xaaa0a8a2,0x8a82000,0x222202a,0x8a2a802,0x28820028,0xa8002022,0xa8282200,0x2020880,0xa8082000,0x88822028,0x8200a880,0x82000a22,0x80220,0x2);
remainders_pos[507].modulo_value=156; remainders_pos[507].bit_mask=create_bit_mask(5,0x2800802,0x28028028,0x80280280,0x2802802,0x8020028);
remainders_neg[507].modulo_value=156; remainders_neg[507].bit_mask=create_bit_mask(5,0x82080082,0x20820820,0x208208,0x82082082,0x820820);
remainders_pos[508].modulo_value=1016; remainders_pos[508].bit_mask=create_bit_mask(32,0xaa2228a,0xa88a0aa0,0x880a0a28,0xaa28000,0x2aaa0802,0x208282a2,0xa8028220,0x2088802,0xa8022208,0x80882802,0x88a82820,0x8020aaa,0x200028aa,0xa28a0a02,0xaa0a22,0xaa2888aa,0xaa2228,0x8a88a0aa,0x880a0a2,0x20aa2800,0x22aaa080,0x208282a,0x2a802822,0x80208880,0x2a802220,0x8088280,0xa88a8282,0xa08020aa,0x2200028a,0x2a28a0a0,0xa00aa0a2,0xa2888a);
remainders_neg[508].modulo_value=1016; remainders_neg[508].bit_mask=create_bit_mask(32,0x822aaa02,0x20028228,0x8282a0,0x22a0888,0xa222808a,0xa80a0a2a,0x208a0aa8,0x8a80008a,0x208aaa80,0x800a08a,0x20a0a8,0x808a8222,0xa888a022,0x2a02828a,0x882282aa,0x22a00022,0x8822aaa0,0x2002822,0x8008282a,0xa022a088,0xaa222808,0x8a80a0a2,0xa208a0aa,0x8a80008,0xa208aaa8,0x80800a08,0x20020a0a,0x2808a822,0xaa888a02,0xa2a02828,0x2882282a,0x2a0002);
remainders_pos[509].modulo_value=509; remainders_pos[509].bit_mask=create_bit_mask(16,0x67b34a72,0x43c238f0,0x7a3f02dd,0xef5a2ffb,0xd7532624,0x6abb7280,0xd3888098,0x29a2e4b3,0xf349d165,0x86404472,0xc053b755,0xc91932ba,0xb7fd16bd,0xaed03f17,0x83c710f0,0x1394b379);
remainders_neg[509].modulo_value=2036; remainders_neg[509].bit_mask=create_bit_mask(64,0xaa2a82aa,0x8a0aa028,0xa2a28a00,0x2282a222,0xa8aaaa8,0xa222aa08,0xa000800,0xa02a282a,0x2a8008a8,0xa8888aa,0x88a2a88,0x80aa22,0x2a208a20,0x2208a28a,0xa0a8828,0x8a082aa0,0x802aa082,0x20a08a82,0xa28a2882,0x228822,0x8222a808,0x808aa288,0xa2a8888a,0x28a8800a,0x82a0a2a0,0x28008002,0x8082aa22,0x20aaaa8a,0x2a222a0a,0x88028a2a,0xa8a02a82,0x2aa0aa2,0x80200a0,0xa882802a,0x20202082,0xa20a022,0x8a808000,0xa8222002,0x282aa8a,0x82a020a,0x200aa88,0x8a808888,0x22888200,0x2aa8a802,0x82022882,0xa228820,0x2a82808a,0xa0028a00,0xa0aa0028,0xa2282880,0x20208208,0x8aaa208a,0x88a2200a,0x80a88020,0x80200888,0x2a0808aa,0xa0a02820,0x220aa8aa,0x80a8a002,0x82280000,0x2022202,0xa08aa082,0x20082a00,0x280);
remainders_pos[510].modulo_value=2040; remainders_pos[510].bit_mask=create_bit_mask(64,0x80882802,0x20020200,0x88200,0x820020a2,0x20208208,0x80202288,0x800208a2,0x20808000,0x800000a,0x22002800,0x220a00,0x280282,0x220a0020,0x8a00008,0x28a00a,0x20820000,0x28220200,0x80008,0x822800,0x20008020,0x8000200a,0x20000022,0x20800220,0x20a088,0x2880802,0x208,0x8000200,0x820020,0x20800,0x220800a,0x80822082,0x200000,0x2000008,0x80820882,0xa00208,0x2080,0x20080082,0x8000,0x80208000,0x802022,0x8220a08,0x8088002,0x2880000,0x8a00800,0x80200,0x2882,0x28200020,0x8008088,0x82,0x20a00a28,0x8820000a,0x800a0,0x828028,0x8800a088,0x20002800,0x8a00000,0x2000202,0x28a2080,0x8228808,0x82208208,0x8a0800,0x8008220,0x2008080,0x802822);
remainders_neg[510].modulo_value=2040; remainders_neg[510].bit_mask=create_bit_mask(64,0x2082,0x28028a00,0x8000288,0x2002822,0x20a00a00,0x282200,0xa0022,0x28000008,0x80008082,0xa2882000,0x8a20200,0x88208202,0x22820020,0x208800,0x80202002,0xa00a0880,0x20220a00,0x8008080,0x80022080,0x20800828,0x8082082,0xa00808a2,0x20008228,0x88202000,0x2000002,0x8800a00,0x80088280,0xa00a0,0x8828008,0x82280002,0xa2802,0x8208000,0xa088080,0x20002,0x208a00,0x88002008,0xa0000802,0x8000008,0x8200088,0x80082822,0xa20200,0x82,0x2000080,0x208008,0x80008200,0x80882002,0x20208820,0x80000,0x80800002,0x20208220,0x280082,0x80000820,0x8020020,0x2000,0xa0082000,0x200808,0x82088282,0x2022000,0xa20000,0x2280200,0x80020080,0xa20,0xa080008,0x2022);
remainders_pos[511].modulo_value=2044; remainders_pos[511].bit_mask=create_bit_mask(64,0xa0aaa2a,0x28828a2,0x80220802,0x2220a828,0x22a08282,0x8a080a28,0xa0820008,0x888a802a,0x28080080,0x22808202,0x800028aa,0x88a08a00,0xaa20028,0x8882a208,0x8a800a20,0xa808a02,0x2820a02,0x2802a802,0xa2800a02,0xaa0800,0x28802022,0x8000aa2a,0xa882a00,0x222a82a0,0x2a0022,0x22a28a00,0x80822280,0x880a2008,0x20802808,0xa8088a2,0x8020820a,0x82,0x28200000,0xa082080,0x88a2202a,0x22028020,0x22008a02,0x80288820,0xa28a8,0x88800a80,0xa82a88,0x200a822a,0x8a8aa000,0x8808022,0xa0020aa0,0x880a0028,0x802a802,0x80a0828,0x280a202a,0x208a002a,0x208a822,0x228008aa,0x200a20a2,0x8aa28000,0x88082028,0x20200202,0xaa802a22,0x22000820,0x828a020a,0x882820a8,0x2282a088,0x8020880,0x8a28228,0xa8aaa0a);
remainders_neg[511].modulo_value=2044; remainders_neg[511].bit_mask=create_bit_mask(64,0x82822222,0xa00a022,0xaa808a,0xaa2820a0,0xaa200a0a,0x28002a0,0x200a8080,0x20822,0x20808800,0xa2080a8a,0x808a022,0x200288,0x822a08a0,0xa2280,0x2088228,0x2080282,0x820a828a,0xa00a208a,0x2a00828a,0x88220088,0xa008a0aa,0x8882222,0x200a280,0xa2a20a28,0x882288aa,0xaa220280,0x80a2a08,0x82a080,0xa808a000,0x2080022,0xa80a82,0x8808880a,0xa0a08888,0x82802808,0x2aa022,0xaa8a0828,0x2a880282,0xa000a8,0x8802a020,0x8208,0x88202200,0xa88202a2,0x2022808,0x800a2,0x208a8228,0x288a0,0x8082208a,0x808200a0,0xa082a0a2,0xa8028822,0xa8020a2,0xa2088022,0xa802282a,0x2220888,0x8028a0,0xa8a8828a,0x2208a22a,0x2a8880a0,0x2028a82,0x20a820,0xaa022800,0x80820008,0x802a02a0,0x2022202);
remainders_pos[512].modulo_value=8; remainders_pos[512].bit_mask=create_bit_mask(1,0x82);
remainders_neg[512].modulo_value=8; remainders_neg[512].bit_mask=create_bit_mask(1,0xa);
remainders_pos[513].modulo_value=228; remainders_pos[513].bit_mask=create_bit_mask(8,0x22000082,0x28a20a00,0x2200282,0x20a0800,0x28008800,0xa08a288,0x20000880,0x8);
remainders_neg[513].modulo_value=228; remainders_neg[513].bit_mask=create_bit_mask(7,0xa2800802,0x20228208,0xa28820a,0x82820080,0x20800008,0x82002200,0xa0880000);
remainders_pos[514].modulo_value=2056; remainders_pos[514].bit_mask=create_bit_mask(65,0x8a8a822a,0x22a2820,0xaa2aaa20,0x220aa820,0x8a28828a,0x2280a08,0x88808882,0x8a2aa0a2,0xa0a22a00,0x80008002,0xa082a8a,0x2802a00,0xa8220aaa,0xaa222a20,0xa82a2a88,0x2a820200,0xa2a28022,0x8020280a,0x20222008,0xa2280a,0xa02a82a0,0x8028a0aa,0xaa8aaa80,0xa0220a02,0xa002082,0x888a8882,0xa0a822a2,0x82882088,0x2800a220,0x2000200a,0x282022aa,0x2280808a,0x2028800,0x880828a2,0x80008aa,0x8a0028a0,0x8228208,0x882a0a22,0x22a2228a,0x800a082,0xa0880a82,0xaaa2aa80,0xa280202,0x82a80aaa,0x288a000a,0x88808a0,0x28080220,0x28a8aa0,0x8082a888,0xa8a82a00,0xa888aa22,0xa0882a08,0xa80280aa,0xa820a000,0x20002a2,0xa88a0a80,0xaa8a200,0x2202228a,0xa0288082,0x8228a220,0x2aa088a2,0xaaa8aa08,0x28a88008,0x82a2a208,0xa8);
remainders_neg[514].modulo_value=2056; remainders_neg[514].bit_mask=create_bit_mask(65,0x2020aa2,0x8aa2a0a8,0x22a222a8,0xaa8220a8,0x2a00a02,0x8aa08280,0x8000a,0x2a2282a,0x282aa288,0x888088a,0x8280a202,0x8a08a288,0x20aa8222,0x22aaa2a8,0x20a2a200,0xa20a8a88,0x2a2a08aa,0x8a8a082,0xa8aaa880,0x882aa082,0x28a20a28,0x8a02822,0x22022208,0x28aa828a,0x8288a802,0x2000a,0x2820aa2a,0xa00a800,0xa0882aa8,0xa888a882,0xa0a8aa22,0xaa080802,0x8a8a0088,0x80a02a,0x80888022,0x288a028,0x80aa0a80,0xa282aa,0xaa2aaa02,0x8088280a,0x2800820a,0x222a2208,0x82a08a8a,0xa208222,0xa0028882,0x80008028,0xa0808aa8,0x8a020228,0x80a2000,0x2020a288,0x200022aa,0x2800a280,0x208a0822,0x20a82888,0x8a888a2a,0x20028208,0x82202a08,0xaa8aaa02,0x28a0080a,0xaa02aa8,0xa228002a,0x22202280,0xa0200880,0xa2a2a80,0x20);
remainders_pos[515].modulo_value=2060; remainders_pos[515].bit_mask=create_bit_mask(65,0xa0800a82,0xa22a02a0,0xa2280,0x2020808,0x20a02002,0x8002228a,0xa2088aa2,0x28802220,0x80a2a82a,0x8000802a,0x88a02888,0x280a8222,0xa82a200,0x2822822,0xa0082a08,0xa08a0000,0xaa020802,0x82a028,0x808208a2,0x80080208,0x8802208a,0x20280200,0x2a08a220,0x80002800,0x80a82000,0x220808,0x80a8a00,0x28882080,0x282888a,0x80aa22,0x880080a,0x22a0820,0xa880208,0x220082,0xa0200a02,0x2828088a,0x82228a22,0x2a020020,0x880000a,0x82a02202,0x80002000,0xa20a8002,0x2808088,0x88022008,0x2002a20,0x8202208,0xa82008a2,0x80aa280,0xa20a802,0x8200a000,0x8828020a,0xa82a0882,0x2a028008,0x80a22888,0x20002222,0xa8a02a80,0x80228a82,0x2208a088,0x880028aa,0x80a08a28,0x8080800,0x8a000202,0xa88a028,0x20a0a8,0x82a);
remainders_neg[515].modulo_value=2060; remainders_neg[515].bit_mask=create_bit_mask(65,0x2808020a,0x2a228a20,0x882a208,0x8a828000,0xa028a80a,0x88a2a02,0x2a000222,0x2008a2a8,0x82220a2,0x880822,0x282000,0xa0020aa2,0x20a2288,0x8a02a0aa,0x2080a200,0x28020888,0x220a8082,0x80a20a0,0x802802a,0x808a00,0x8a2802,0xa8208a80,0x228022a8,0x880a088,0x20a808,0x88aa0080,0x80020280,0x2000a008,0xa020002,0x8008222a,0x80080082,0x8a2280a0,0x2008280,0x88a2880a,0x20a8820a,0xa0a00002,0xa2a02a2,0x228a80a8,0x80008882,0x228aa0a,0x8882888,0x22020882,0xa080000,0x82a880,0x8288a228,0x80a82a80,0x20288022,0x822208,0x82a0208a,0x2882808,0xa00a82,0x20228002,0x228a0080,0x822a000,0xa008aa2a,0x20282208,0x82a0202,0x2a802000,0x80a022,0x280220,0x80800088,0x2088a82,0x20020a0,0x88a02820,0x22);
remainders_pos[516].modulo_value=1032; remainders_pos[516].bit_mask=create_bit_mask(33,0xa2002022,0x20000,0x200a08a,0xa2822082,0x8220a28,0x88280202,0x20002,0x20200228,0x8800808a,0x80002,0x8028228,0x8a088208,0x208828a2,0x20a00808,0x8000a,0x808008a0,0x20020228,0x20000a,0x200a08a0,0x28220820,0x8220a28a,0x82802020,0x200028,0x2002280,0x800808a2,0x800028,0x80282280,0xa0882080,0x8828a28,0xa008082,0x8000a2,0x8008a00,0x88);
remainders_neg[516].modulo_value=1032; remainders_neg[516].bit_mask=create_bit_mask(32,0x228828a2,0x8828008,0x8a082002,0x220a2802,0xa28220,0x20828a,0x800a0882,0x28a08a20,0x80002,0x80800882,0x820a20,0x2000280,0xa0002022,0x28208000,0x88008082,0x80020,0x28828a20,0x88280082,0xa0820020,0x20a28028,0xa282202,0x20828a0,0xa08820,0x8a08a208,0x800022,0x8008820,0x820a208,0x20002800,0x20220,0x8208000a,0x80080822,0x800208);
remainders_pos[517].modulo_value=2068; remainders_pos[517].bit_mask=create_bit_mask(65,0x2a08220a,0xa220aa8,0x2a20a80,0x88a8aa82,0xa8880802,0x80020a82,0xa280002,0x8888020a,0x2002820,0x822002a,0xa200a8a2,0xaa2202a2,0x82022888,0x2000802,0x28a220a0,0xaa0008a,0x2a8202,0x2228a288,0x80020028,0xa2020a00,0x222a888,0xa8022a2a,0x2208228,0xa00202a0,0x888820,0xa28282,0x82000a00,0x8088aa0a,0xa8a88a00,0x822a020a,0x8222800a,0x2082a0aa,0x20882a82,0x882aa0a8,0x882a0028,0xa2aa080a,0x20200a22,0x82a0aa2,0xa0000a00,0x20082828,0xa08222,0x8800a808,0x2a28820,0x880a8a88,0x8a222a8,0x200a08,0x88828008,0x800228a2,0xaa08082a,0xa28a2000,0x800a088,0x8280200,0x8aa22288,0x8a8a808,0x8208a2a0,0x80a8008,0x22208280,0x8a0a0802,0x280002,0x22a82a08,0xa2280202,0xa8082aa2,0x8a002a08,0xa82aa08,0xa0882);
remainders_neg[517].modulo_value=2068; remainders_neg[517].bit_mask=create_bit_mask(65,0xa280a282,0x822a0220,0x8a2a8208,0x202202,0x2000008a,0x8a820a,0x82a0888a,0x8202,0x8a80a0a8,0xaa88a2,0x2a88202a,0x22aa822a,0xa0aa000,0x8a88800a,0xa02aa820,0x82280802,0x80a20a8a,0xaaa02a00,0x88a8820,0x2a828280,0xaaa2000,0x208aa2a2,0x8aa802a0,0x280a8a28,0x80000a8,0x882a0a02,0xa880288,0x2282,0x20200288,0xaa28a02,0x2a20882,0x280a2822,0xa800a20a,0xa22020,0x2288a0,0x2a228082,0xa82882a2,0x80a2022a,0x20888288,0xa880a0a0,0x88280a2a,0x802080,0xa2200a8,0x820200,0x802aa220,0x88288280,0xa0880,0x88aa022,0x228000a2,0x2202a888,0x80882800,0x80a08a08,0x222aa00,0x202080,0xa802228,0x80820080,0xaa280a08,0x282808a,0x88a08882,0xaa202280,0x22a08a0a,0x2080a22a,0x288a200,0x82022280,0x2800a);
remainders_pos[518].modulo_value=2072; remainders_pos[518].bit_mask=create_bit_mask(65,0x228a8a02,0x8008080,0x2202800a,0xa00282a,0xaa00028a,0x88a02a20,0x88aaa000,0xaa02822a,0x2828a222,0x82220022,0xa20a0008,0x2880a20a,0x820820a0,0x8828a2a2,0x8088020,0x20202a80,0x222a0828,0x8828a080,0x8208a0a,0x8820200,0x888220a,0x20820022,0x222088a2,0x28022a,0x82200022,0x8002,0xa000aa8,0xa8a00220,0xa80a002,0x2a02020,0x82a880aa,0x8002280a,0x2028280,0x82a02880,0x80aa022a,0xa008080a,0x2a800a02,0xa008800a,0x2aa000,0x82800200,0x880008,0x88a88028,0x88a2208,0x20880082,0x20a08822,0x20008082,0x22a0a208,0x88020a28,0x82820a8,0x2002a808,0x22080220,0x828a8a28,0x280a0820,0x8aa08a02,0x822000a0,0x28880088,0xaa888a28,0x22a88280,0x22000aaa,0xaa08a80a,0xa0a28000,0x88a82800,0x20a00280,0x88020200,0x80a2a2);
remainders_neg[518].modulo_value=2072; remainders_neg[518].bit_mask=create_bit_mask(65,0xaa02020a,0x880800,0xa28a0882,0x820820a2,0x22808a02,0x2822a8,0x222088,0x228a0a22,0x20a02aa2,0x2aa88aa,0x2a028880,0xa0002a82,0xa002828,0xa0222a,0x80800828,0x28a8a200,0xa2a280a0,0x202008,0x80a00282,0x800a0a88,0x8000a282,0x280a882a,0x2aa80022,0x80a08aa2,0xa2888aa,0x8880088a,0x82800220,0x202882a8,0x8208280a,0xa28a8a0,0x2200822,0x80aa002,0x8a820a08,0xa282008,0x82282a2,0x20808002,0x22088282,0x20800882,0x88222888,0xa008a88,0x88000880,0x2000a0,0x8002aa00,0x28008802,0xa02800aa,0xa8080802,0xaa202a80,0x8a02a0,0x80a0a020,0xa80a2000,0x2a808aa0,0x20202a0,0xa00280a8,0x220028a,0xaa80028,0xa0008000,0x22000220,0x2a200a00,0x22888222,0x22002082,0x28220888,0x202088,0xa8288208,0x828a08,0x82a22);
remainders_pos[519].modulo_value=2076; remainders_pos[519].bit_mask=create_bit_mask(65,0x28a20a2,0x228028,0x88288202,0xa28820a0,0x20000020,0x2088282,0xa0000080,0x8a00828,0xa008288,0x82880822,0x8a20228,0x80082282,0x80002880,0xa08208,0x8a08208a,0x82880082,0x20020208,0x8200020a,0x200a0020,0x28a20208,0x82282200,0x80020800,0x20020008,0xa008082,0xa0022002,0x28a00020,0x208a20a,0xa2880800,0x20000828,0x202008,0x8a0882,0x28028,0x2000008,0x2802800,0x8220a20,0x82008080,0xa2820000,0x20228,0x8a08a208,0xa08000a2,0x8008800,0x8820200a,0x22000800,0x20020800,0x80088288,0x820808a2,0x20800a00,0x8a080008,0x22080800,0x28200228,0xa20820a,0x220820a0,0x20228000,0x8288200,0x228808a2,0x8820228,0x228200a,0xa28200a2,0x200000,0x88282208,0xa0800000,0x20a08228,0x8082282,0x2802880,0x8a08a28);
remainders_neg[519].modulo_value=2076; remainders_neg[519].bit_mask=create_bit_mask(65,0x82022822,0x8a20820,0x20028a,0x22002820,0x28808828,0x8a00020a,0x20880800,0x208020,0x82080200,0x20000a2,0x228a20,0x800a20a,0x882000,0x8200a00,0x200a002,0x2000802,0x28828a00,0xa088282,0xa08208a0,0x20228a00,0xa20a288,0x8a0080,0x28828800,0x8208000a,0x208a2882,0x20208828,0x8a002282,0x22000080,0x28808020,0x8828a080,0x80020002,0x8820820,0x8a088080,0x82082080,0xa28228,0xa080008,0x220a0880,0x828a20,0x2002280,0x20080822,0x800008,0x28a082,0xa2880080,0x28828008,0x8000200,0x2800022,0x28008208,0x2008080,0xa2800080,0x20a08a20,0x82280282,0xa2802820,0x28a20808,0x80200288,0xa2000022,0x28a20,0x8a20a082,0x220a0822,0x8a08808,0x202280,0x20080880,0x28200a20,0x8000a20a,0x82082000,0x200220);
remainders_pos[520].modulo_value=520; remainders_pos[520].bit_mask=create_bit_mask(17,0x8280a8a,0x8a228802,0x2a200,0x200280a,0xa0282222,0xa0022880,0x88020082,0x820a0,0x8200000,0x80220a,0x28800a82,0x88280a02,0x28008088,0x8a8000a0,0x2288a200,0xa0282080,0xa2);
remainders_neg[520].modulo_value=520; remainders_neg[520].bit_mask=create_bit_mask(17,0x80a00202,0x22a0002,0x8a2288,0x8a80a082,0x20a02a2a,0x288a2008,0xa8802,0x800a028,0x80a08888,0x8008a202,0x2008020a,0x208282,0x20800000,0x2008828,0xa2002a08,0x20a02808,0x22);
remainders_pos[521].modulo_value=521; remainders_pos[521].bit_mask=create_bit_mask(17,0xa6752f36,0x4ebeb531,0x451b05a3,0xbaf6477e,0x893c427,0x32910347,0x181d17f4,0xc5647f96,0xf89a8eb5,0xa2e061a7,0x22530bf,0x8f24438b,0x89bd7790,0x836289fb,0xb5f5cb16,0xd2b99632,0x1b3);
remainders_neg[521].modulo_value=2084; remainders_neg[521].bit_mask=create_bit_mask(65,0x2aa8a2aa,0x822228a8,0x8882882a,0x222a8aa2,0x800a08aa,0xaa088a8a,0x80808a28,0x8a8a20a,0x20020228,0x2a28a82a,0x8aa8a822,0x2a88a02,0x20aa08,0xaaa0022,0x2828028a,0xa200aaa,0x2282a02a,0xa0a2aa80,0xaa828a00,0x20022002,0xaa0082a8,0xaa8a0288,0xa2a220a8,0x22a0a8,0xa880a202,0x80a8228,0x82a8a288,0x800a8a88,0xa222a882,0x88a2a8a,0x220aa08a,0xaaa0a8a2,0x2002aa28,0x80a0008,0x8a00a222,0x802088a0,0x88022200,0x8080aa82,0x82080288,0x2280a8a8,0xa20a8808,0xa022aa2,0xa220208,0x82a08008,0x28aa008,0xaa22aa28,0xa0828002,0x80020a0a,0x2202822a,0xaa20a0,0x82a82820,0x2aa000a0,0xa00a2aa2,0xa08802a8,0x28080082,0x28082022,0x22a2aa20,0xa20808a8,0x208a8a80,0x8088a008,0x8a0aa80,0x802220,0x28828882,0x8222280,0x2080028);
remainders_pos[522].modulo_value=696; remainders_pos[522].bit_mask=create_bit_mask(22,0x2880882,0x20020820,0x202082,0x20000a0,0x28800028,0x828a,0xa00a2080,0x28028a20,0x208220a,0x220820a2,0x808800,0x88002202,0x808a0820,0x28a08820,0xa08a280,0x208a0,0x28a28200,0x80280002,0xa0000,0x8820808,0x80082080,0x822022);
remainders_neg[522].modulo_value=696; remainders_neg[522].bit_mask=create_bit_mask(22,0x82000002,0x28828028,0x8828a00a,0x82880820,0x20008820,0x88080202,0x20822000,0x20820228,0x8a00a282,0x22802822,0x8000008,0x8a28a,0x200a0,0x20200028,0x82002208,0x808a0020,0x20220a08,0x820800a,0x80820880,0x28000,0xa008,0xa28a2);
remainders_pos[523].modulo_value=2092; remainders_pos[523].bit_mask=create_bit_mask(66,0x2a02220a,0x8aa8288,0x828aaa08,0x8a282280,0xaa282a20,0x22222aa,0x2a8a0282,0x2a0aa08a,0x2a00882a,0xaaa2002,0x8a2000a0,0x2a020a88,0x880a000,0x82208a82,0xa282a880,0x2a2aa2a8,0x80aa0280,0x8200800,0x8a2a200a,0x228a0228,0x80aaa82a,0x8222028a,0x2a82aa8,0x808aa880,0x8080a22a,0x822802,0x8a8a0a82,0x80000888,0x822880a0,0xa2a08a0,0x22a28002,0x80a20a00,0x2802888a,0x8a02a22,0x28a8800a,0xa8a0800,0x228820a2,0x20a0,0xa2a2222,0x8820082a,0xa0202802,0xaa202a88,0x82a80022,0x88822aa,0xaaa02a28,0xa288a82,0x8a8a2288,0x820a00,0xaa02002,0xaa8a8028,0xa828a2a8,0x20882022,0xa022082a,0x80a8000,0x8a222a,0x8aaa00a0,0x200a8800,0xaa0a8a82,0xa2a8a20,0x88880828,0x828aaaa8,0x828a208a,0xaa282028,0x2aa2020a,0x880a8228,0xa08);
remainders_neg[523].modulo_value=2092; remainders_neg[523].bit_mask=create_bit_mask(66,0xa28aaa82,0x80220a00,0xa022280,0x2a0aa08,0x22a0a2a8,0x8aaaaa22,0xa2028a0a,0xa2822802,0xa28800a2,0x8222a88a,0x2a88828,0xa28a8200,0x80082888,0xaa8020a,0x2a0a2008,0xa2a22a20,0x8228208,0x80a88088,0x2a2a882,0xaa028aa0,0x82220a2,0xaaa8a02,0x8a20a220,0x8022008,0x8082aa2,0x880aa08a,0x202820a,0x8888000,0xaa00828,0x82a28028,0xaa2a088a,0x82a8288,0xa08a0002,0x8028a2aa,0xa0200882,0x82028088,0xaa00a82a,0x8888a828,0x82a2aaaa,0xa880a2,0x28a8a08a,0x22a8a200,0xa2088aa,0x8000aa22,0x2228a2a0,0x82a0020a,0x202aa00,0x880a8288,0x8228a88a,0x220208a0,0x20a02a20,0xa800a8aa,0x28aa80a2,0x80820888,0x8802aaa2,0x2228828,0xa8820088,0x2282020a,0x82a202a8,0x80a0,0xa022220,0xa02a802,0x22a0a8a0,0xa22a8a82,0x820aa0,0x280);
remainders_pos[524].modulo_value=1048; remainders_pos[524].bit_mask=create_bit_mask(33,0x82a82222,0x202aa202,0x828aa08a,0xa2aaa2a0,0x8000802,0x28020a28,0x8800a882,0x88a00a0a,0xa22208,0xa0022a0a,0x8028822,0x2028a,0xaaa8a802,0xaa2820a8,0xaa808a20,0x82a82808,0x82222888,0xaa20282a,0xaa08a202,0xaa2a0828,0x802a2a,0x20a28080,0xa882280,0xa0a880,0x2220888a,0x22a0a00a,0x28822a00,0x2028a080,0x8a802000,0x820a8aaa,0x8a20aa2,0x82808aa8,0x88882a);
remainders_neg[524].modulo_value=1048; remainders_neg[524].bit_mask=create_bit_mask(33,0xa20aaaa,0xa8a22a8a,0xa022802,0x2a222a28,0x80888082,0xa08a82a0,0x88200a,0x288282,0x882aaa80,0x288aa282,0x808a00aa,0x888a8a02,0x2220208a,0x22a0a820,0x220802a8,0xa20a080,0xaaaa000,0x22a8a0a2,0x22802a8a,0x22a280a0,0x880822a2,0xa82a0808,0x8200aa08,0x88282008,0xaaa80002,0xaa282882,0xa00aa288,0xa8a02808,0x208a888,0xa820222,0x802a822a,0xa080220,0xa2);
remainders_pos[525].modulo_value=420; remainders_pos[525].bit_mask=create_bit_mask(14,0x20002,0x8008a20,0x2088008,0x82002020,0x808008,0x2288,0xa0002,0x28800008,0x202002,0x80800822,0x220800,0x8a200202,0x80000,0x8);
remainders_neg[525].modulo_value=420; remainders_neg[525].bit_mask=create_bit_mask(13,0x808a0802,0x220,0x2000080,0x20028a0,0x800,0x88082200,0x80020082,0x20008000,0x820a082,0x80022,0x8220008,0x2208282,0x800080);
remainders_pos[526].modulo_value=2104; remainders_pos[526].bit_mask=create_bit_mask(66,0x2a228aaa,0x80aaa80a,0x2aaa880,0xa2800228,0x820a2282,0xa80882a0,0xa008a0,0x82a8aaa2,0x2280a20,0x8a08aa22,0x820a00,0xa0a80a28,0x882a282a,0x222882a8,0x22080802,0x2aa2a082,0x2a222220,0x228a0200,0x22a8a8a2,0x88028822,0x282028,0xa820a80a,0x8aa0a28a,0xa22008a0,0x8a00a822,0xa2000802,0xa08aa0a,0x8a0288a8,0x28220a2,0xa822aa82,0x8a880002,0x20a8000a,0x8220,0x8088200,0xa2a0002a,0x2a800022,0x8082aa88,0xa28a0882,0xa02a2280,0x8aa0aa20,0xa2802000,0x8a882a00,0xa20a2008,0x2aa28a0a,0xa02a08,0x22280828,0x88882280,0x888a2a2a,0xa80080a2,0xa8088888,0x88820a8a,0x88802020,0x222a8228,0xaa828a8,0x28a02a,0xa200a082,0x8088aa20,0x8208a028,0x8aaa2a,0x2a0a200a,0x820a8220,0x8a8288a0,0x80288002,0x2022aaa,0xa8a02aaa,0xaaa288);
remainders_neg[526].modulo_value=2104; remainders_neg[526].bit_mask=create_bit_mask(66,0xa2aa0222,0x8222082,0x8a222008,0x2a088aa0,0xa82aa0a,0x20800a28,0x88288028,0xa20222a,0x8aa08228,0x28022aa,0x880a8288,0x282082a0,0xa2a0a2,0xaaa00a20,0xaa80808a,0xa22a280a,0xa2aaaaa8,0xaa028a88,0xaa20202a,0x8a00aa,0x88a0a8a0,0x20a82082,0x2282a02,0x2aa88028,0x28820aa,0x2a88808a,0x82802282,0x28a0020,0x8a0aa82a,0x20aa220a,0x200888a,0xa8208882,0x88880aa8,0x80800a88,0x2a2888a2,0xa20888aa,0x80a2200,0x2a02800a,0x28a2aa08,0x22822a8,0x2a08a888,0x200a280,0x2a82a880,0xa22a0282,0x8828a280,0xaaa080a0,0xaa08,0x2a2a2,0x2088082a,0x20800000,0xa8202,0x8a8a8,0xaaa20aa0,0x8220a020,0x88a028a2,0x2a88280a,0x80022a8,0xa8028a0,0x880222a2,0xa282a882,0xa820aa8,0x20a0028,0x8a0088a,0x8a8aa222,0x2028a222,0x222a00);
remainders_pos[527].modulo_value=2108; remainders_pos[527].bit_mask=create_bit_mask(66,0x22008282,0xa28208a2,0x800a02a2,0x8a080a20,0xa0a0a088,0x2a8208,0x800a2a20,0x2882a88a,0xa22aaaa,0x2a02aa8,0xa08aa880,0x88a00200,0x802880a,0x22a8882a,0x8000808a,0x80a28828,0x2208aa,0x20a0082a,0x82220aaa,0x80822008,0x22202a2,0x288aa822,0xa2200028,0x8000a80a,0x802,0x2200a20,0xa8808288,0x228a00a2,0x22282800,0x2882a202,0x80a80a8,0x8228208,0xa0a0a88,0x22a0a0a,0x2082882,0x82a02a02,0x8808a822,0x80028288,0xa8a00a28,0x2282022,0x8a0088,0x28020000,0xaa02a000,0x82800088,0x882aa22,0x28a80888,0x22008820,0x8aaa0888,0xa8200a0,0x2aa20880,0x228228a0,0x8a202000,0xa8222a8,0x2a022802,0xa00800a2,0x22aa20,0x2aa80a8,0x8aaaa88a,0x2a22a822,0x8a8a00,0xa2082a80,0x2220a0a0,0x208a020a,0xa8a80a00,0x88a20828,0x8282008);
remainders_neg[527].modulo_value=2108; remainders_neg[527].bit_mask=create_bit_mask(66,0x2a880a0a,0x2a02802a,0x8828a2a,0x20082a8,0x20282800,0x80a20a80,0x882a2a8,0x200a2002,0x822a2222,0x8a28a220,0x28022000,0x288a88,0x80820002,0xaa2000a2,0x8880002,0x82a00a0,0x88aa0022,0xa82880a2,0xaa28222,0x80aa880,0x8a2a822a,0xa00220aa,0x22a888a0,0x8882082,0x888800a,0x8aa882a8,0x20080a00,0xaa028822,0xaaa0a080,0xa00a2a0a,0x80820820,0xaa0280,0x82828200,0x8aa20282,0x8a80a00a,0x220a28a,0x8020aa,0x80a0a00,0x202882a0,0x8220a8aa,0x88028800,0x208a8888,0x228a2888,0xa008800,0x800a22a2,0xa0208000,0xaa880028,0x2220000,0x820a8028,0xa22a8008,0xaa0a2028,0x2a8a088,0x8202aa20,0xa28aa08a,0x2800882a,0x88aa2228,0x82220820,0x2222002,0x22aa20aa,0x88020280,0x2a80a208,0xaaa82820,0x28028a82,0x20208208,0x2a80a0,0xa0a080);
remainders_pos[528].modulo_value=264; remainders_pos[528].bit_mask=create_bit_mask(9,0xa2020002,0x20228,0x8808000a,0x808a2,0x20200028,0x20228a,0x808000a0,0x808a28,0x80);
remainders_neg[528].modulo_value=264; remainders_neg[528].bit_mask=create_bit_mask(9,0x228a0082,0x8028a20,0x8082,0x80800022,0x28a00820,0x8028a202,0x80820,0x8000220,0x8);
remainders_pos[529].modulo_value=92; remainders_pos[529].bit_mask=create_bit_mask(3,0xaa2aaaaa,0xaaaaaaaa,0xaaaaa8a);
remainders_neg[529].modulo_value=92; remainders_neg[529].bit_mask=create_bit_mask(3,0x22222222,0x22222222,0x2222202);
remainders_pos[530].modulo_value=2120; remainders_pos[530].bit_mask=create_bit_mask(67,0xa82202,0x200a0822,0x22822200,0x8aa028a0,0x2a00,0x282a0a88,0x2208aa2,0xa822208,0x82028082,0xa2880a2a,0x282a2200,0x88000202,0x22008228,0x82a0a028,0xa88a2a,0x208a280a,0x28aa0,0x28008228,0x800820,0x22080200,0x208a2208,0xa2288880,0x22002020,0xa0800a,0x8280a200,0xa0220082,0xa02a0,0x802020,0x8022a888,0x28000a,0xa022020a,0x28200a22,0x220802a0,0x80208882,0xa008280a,0x80880a88,0x2800a0,0x2a8802a0,0x8020022,0x80a00008,0x880a0a,0x8a028282,0x20a0000,0x80088a0,0x22288a08,0x88a20802,0x80208820,0x20020000,0x82002808,0xa2800028,0x28a2080a,0xa22a00a0,0xa0a82a8,0x82008828,0x80002228,0x88a82880,0xa0228a00,0x28082a8,0x8882a082,0xa2088020,0xa0a8288a,0xa8000022,0x280aa200,0x8882880a,0x20a00800,0x882a0088,0x80);
remainders_neg[530].modulo_value=2120; remainders_neg[530].bit_mask=create_bit_mask(66,0x88202a8a,0xa80280a2,0x2a0aa288,0x220a028,0x88a208,0xa0a20200,0x8a280222,0x20aa280,0xa82080a,0x22008222,0xa0a22a88,0x8a82,0x2a8802a0,0xa2028a0,0x80200222,0xa8022082,0x880a0220,0x208802a0,0x88008028,0xa2808a08,0xa8022a80,0x2a200000,0x2a88a0a8,0x88200882,0x2082a08,0x28aa080a,0x88028a20,0x808a0a8,0x8a22000,0x80a08802,0x28aa0a82,0xa02802a2,0x2a808228,0x8a0000a,0x2080a002,0x8000200,0x88208820,0x22008228,0x808288a2,0x288800,0x88000282,0x20a0a02,0xa828088,0x80800028,0xa2200200,0x2a008a,0x82800a0,0x288a8088,0xa80a080,0x22088820,0xa02a0082,0x2a228020,0x2820220,0xa8000a0,0x88aa20,0x202008,0x282a0280,0xa080220,0x2280a,0x22800828,0x28202002,0x200888a2,0x20822288,0x20082,0xa0208008,0xa20800);
remainders_pos[531].modulo_value=708; remainders_pos[531].bit_mask=create_bit_mask(23,0xa2820822,0xa28a00,0x8280008,0x22080080,0x8820228,0x88280200,0x882022,0x20a20808,0xa28020a,0xa0000022,0x8208000,0x208200,0x800000a0,0x8028a08,0x208a08a,0x80822002,0x8028228,0x88082200,0x20020882,0x28200,0xa28a002,0x820828a0,0x8);
remainders_neg[531].modulo_value=708; remainders_neg[531].bit_mask=create_bit_mask(22,0x220a00a2,0x220208,0x80208080,0xa2800800,0x28a20,0x208288,0x800028a2,0x28228000,0x82208282,0x20880822,0xa00808,0x88280288,0x880820,0x820200,0x8a002002,0xa2882,0x820a20,0xa288,0xa08a0002,0x8820a08,0x8220208a,0x2802020);
remainders_pos[532].modulo_value=1064; remainders_pos[532].bit_mask=create_bit_mask(34,0x8a802a0a,0x8000a82,0x22220820,0xa20008a,0x228202,0x82222288,0x80008020,0xa00a8a0a,0xa82a82,0x2a0a2a,0x88208020,0x80022888,0x8a080828,0x888a2000,0x2008208,0x2a282a00,0xa0aa0a80,0xa828a802,0x82008000,0x8a22220,0x2020a200,0x28800228,0x2082222,0xa0a80008,0xa82a00a8,0xa2a00a82,0x20002a0,0x88888208,0x82880022,0x8a080,0x208888a2,0xa0002008,0xa802a282,0xa0);
remainders_neg[532].modulo_value=1064; remainders_neg[532].bit_mask=create_bit_mask(34,0x200a202,0x888202,0x22aa80a8,0x82288802,0x88a20a8a,0xaaa2200,0x88800a8,0x28020202,0x820a202,0x80a282a2,0x2808a0,0x882a000,0x28000a0,0x2a088,0x82880a00,0x22a0a280,0x20228208,0x2020200a,0xa800888,0x80222aa8,0xa8a82288,0x20088a20,0xa80aaa2,0x20208880,0x20228020,0x2a20820a,0x8a080a28,0x280,0xa00882a,0x8802800,0xa000002a,0x28082880,0x20822a0a,0x28);
remainders_pos[533].modulo_value=533; remainders_pos[533].bit_mask=create_bit_mask(17,0x13898ed2,0xf44a9d14,0x36a234cd,0x8247ee9b,0xc8e70fa8,0x5203b9e3,0x9029412,0x3f11e16a,0x3f283305,0x2415a1e2,0x12920a50,0xc4f1e770,0x90457c39,0x5b365df8,0x8beccb11,0x720a2e54,0x12dc64);
remainders_neg[533].modulo_value=2132; remainders_neg[533].bit_mask=create_bit_mask(67,0x8a00020a,0x28820008,0xa22aa800,0x28a2202,0x2a0220,0x8a8a202a,0x808a0882,0xa20828a2,0x22a0aa88,0xa888202a,0xa0a8288,0x8282aa8,0x880a0a0,0x82aa8020,0x2200200,0xa282a288,0x28a00a8,0x2a0a28a8,0x22008a,0xaaa0802,0x28088020,0xa0a08028,0xa8280aa,0x2088a88a,0xa82a22a0,0xa082288a,0x82880a28,0x228a8a08,0x2a002a0,0x228a0022,0xaaa22a02,0x208a000,0x28880,0xaa228282,0xa2a08882,0xaa208,0x222a002,0xa22a80,0x822202aa,0x82088080,0x8a082028,0x20008820,0x8a0a0020,0x80a00808,0x2000882,0x88282a82,0x2a2a88,0x2aa088a0,0xa02082a2,0x2a80800,0x80080aa0,0x22a08202,0x28882aa,0x88aa2a00,0x20aa0a08,0xa0880020,0x88080280,0x2002828,0x82088002,0x8a020828,0x80808820,0x2aa02220,0xaa2280,0x2002a220,0x8822a800,0xa08882a2,0xa22a);
remainders_pos[534].modulo_value=2136; remainders_pos[534].bit_mask=create_bit_mask(67,0x820020a2,0x28228228,0x808028a,0xa2020082,0x8800a00,0xa20a280,0x2022000,0x20200820,0x8a002288,0xa08a0082,0x20808a28,0x8220a,0xa2082080,0xa00828,0x82202080,0xa0022822,0x20220a08,0x2280080,0x22020020,0x28020a08,0x8200008,0x208a2882,0x8820020,0x80080000,0x22800082,0x28220800,0x220a208,0x20028a2,0x28808a00,0x82200280,0xa0802082,0x200000,0xa208002,0xa0800002,0x800208,0xa000008,0x82820802,0x28028008,0x8000a202,0x808a2800,0x28208a08,0x88002088,0x2820002,0x20000020,0x8080082,0x208228a2,0x28200008,0x8820a080,0x80080080,0x8020028,0xa20a088,0x82882880,0x20808,0x8a28200a,0x20820,0x8a08820,0xa28a202,0xa28200a2,0x8228800,0x80082008,0xa0000880,0x20028a08,0x8a00a002,0x20820080,0x28a28020,0x82288288,0x8a0800);
remainders_neg[534].modulo_value=2136; remainders_neg[534].bit_mask=create_bit_mask(67,0x2882822,0x20a20a20,0x80008202,0x228a0802,0x8208,0x82282208,0x828a2880,0x28a08028,0x208a200,0x20020802,0x28000220,0x8800a282,0x22802800,0x8208020,0xa28a008,0x208a20a2,0x28a28200,0x8a208008,0xa28a08a0,0x20820200,0x80288080,0xa0022002,0x28828,0x8008088,0xa2080802,0x20a28008,0x8a282280,0x82882022,0x20000208,0xa288208,0x20002802,0x8a08808,0x8228008a,0x20080882,0x8008a00,0x82088080,0x20a0082,0x20820800,0x808228a,0x22080,0x20a00200,0x8a000,0x820a0882,0x28808828,0x8000800a,0xa00a2022,0x20a08800,0x282008,0x800800,0x828820,0x82282000,0x2002000,0x828000,0x220a082,0x808a00a0,0x200028,0x8220228a,0x220a0822,0xa20008,0x800a080,0x20880000,0x28820200,0x208208a,0x200a0800,0x20220828,0xa200200,0x20080);
remainders_pos[535].modulo_value=2140; remainders_pos[535].bit_mask=create_bit_mask(67,0xa882020a,0x280a8200,0x2a8a22a0,0xa020022,0x2228a828,0x808a2208,0x882a8000,0x28a0088,0x8a208aa,0x2080a00a,0x80000,0x880a0a82,0x808280,0x820220aa,0x82a8a800,0x28882a02,0x20a8080,0x2200a228,0x20880,0x82882a20,0x282a0008,0x82220a02,0x228a8200,0xa2088aa,0x80208228,0x800a0002,0x220222,0x20a22a8,0x80a088,0x8280a020,0x80000882,0x2208800,0xa002288,0xa20008a,0x228800a,0x20022088,0x28220000,0x80a028,0x220a020,0x2a88a08,0x28880880,0x28000a00,0x2882080,0x8aa2208a,0x20082a28,0x880a0888,0x22000a82,0x8a8228,0x80220800,0x288a008,0x80202a08,0x280a8222,0x2002a2a8,0xaa08808,0x20282020,0x82a0a02,0x80000200,0xa00a020,0xaa208a2,0x22200a28,0x20002a82,0x82088a20,0x282a288,0x8880080a,0x80a88a2a,0xa0082a02,0xa080822);
remainders_neg[535].modulo_value=2140; remainders_neg[535].bit_mask=create_bit_mask(67,0x200a0a82,0xa0020a80,0x2202a228,0x828280aa,0xa2a02020,0x8022a80,0x220880,0xa028000,0x80228022,0xa0082802,0x88800888,0x28202,0x8080208,0xa82a822,0x2202008,0xa000228a,0x8a020800,0x2a8822a0,0x88828008,0x200a228,0xa0a20880,0xa2a8282,0x2a020288,0x82a00022,0xa80a20,0x882088a,0x882a8aa2,0xa82a220,0x88002800,0x2082828,0x8880002,0x8a280080,0x288a200,0x82a08802,0x82a00802,0xa88a2800,0xa02a8880,0x80820a0,0x8aa028a8,0x82200200,0xa0000008,0xa0088280,0xa00a008,0x2222802,0xa080a220,0x820000,0xaa088202,0x80202a0,0x8a28088,0x82002800,0x8a82280,0xa0020aa2,0x288a2220,0x82200080,0xa0a0a828,0x80a2028a,0x88a80,0x28820a8,0x8222802a,0xa2a88220,0xa888220a,0xa0002a0,0xa0a2200,0x8082,0x200222,0x2880228a,0x20080a2);
remainders_pos[536].modulo_value=536; remainders_pos[536].bit_mask=create_bit_mask(17,0x820a22a2,0xa822200a,0x8aaa8222,0x2aa0a820,0xa00202a,0x80008a28,0x22280022,0x20a280aa,0x8808202,0x88aa028a,0x2880028,0xa028a200,0xa8a80800,0xa2082a0a,0x2a8882aa,0x82a00888,0x8a88a0);
remainders_neg[536].modulo_value=536; remainders_neg[536].bit_mask=create_bit_mask(17,0xa82aa2a,0x20aaa882,0x2220aa2,0xa22820a8,0x8288a8a2,0x88802a0,0xaaa088aa,0xa82a0822,0x80080a8a,0x228a02,0x8a0008a0,0x28a02a88,0x20208088,0x2a80a282,0xa2000a22,0xa288000,0x20028);
remainders_pos[537].modulo_value=2148; remainders_pos[537].bit_mask=create_bit_mask(68,0x82882802,0x20220a08,0x200088,0x2820000,0x828a08,0x2000280,0x208a2,0x28800a00,0x208208a,0x8008a2,0x220220,0x20a202,0xa2082802,0x20a28000,0xa08208a,0x820828a0,0x28a20208,0x82088202,0x828028a0,0x28228820,0x8228000a,0x200200a0,0x20020228,0x8002202,0x82022880,0x8a28820,0x280,0x2022882,0x8000a00,0xa00828a,0x20000002,0x20208800,0x8a088282,0xa00a00a0,0x28220a20,0x2208088,0x80,0x28200a08,0xa00020a,0x22880800,0x28000008,0x8228a200,0x22880820,0x8800200,0x88080088,0xa0080082,0x28820,0x8228828a,0xa2802820,0x8220820,0x808a288,0xa2820822,0x20820a00,0x28a08a,0x28208a0,0x8a08008,0x88088008,0xa2002000,0x20820808,0xa00228a,0xa2080000,0x28000808,0xa282000,0x82802,0x20008000,0xa088082,0x2822822,0x8);
remainders_neg[537].modulo_value=2148; remainders_neg[537].bit_mask=create_bit_mask(67,0x2002082,0x28a28200,0x88288000,0x820a0880,0x8020200,0x8a008208,0x808a0022,0x20008208,0x8a00a002,0x80080022,0x8a28a28,0x8828228a,0x22802082,0x28220808,0x8200a002,0x2802020,0x20228a00,0xa00028a,0x2082020,0x20a20028,0xa208082,0xa08a0820,0x28828a20,0x8008a28a,0x28a2000,0x220028,0x88088208,0x28a2002,0x808208,0x82080202,0xa0880882,0x28a00008,0x200020a,0x20820820,0x20a28228,0x8a280000,0x80880800,0x20a08200,0x82088282,0xa2000080,0x20808800,0xa202288,0xa20000a0,0x8a08,0x8000,0x20800802,0x8820028,0xa200202,0x220820a0,0xa28028,0x80002200,0x220a00a2,0x28028208,0x88202002,0x820a0020,0x200800,0x80,0x22882880,0x28028000,0x8208a202,0x22800880,0x20808000,0x8220a088,0x80802082,0x28800808,0x8200000a,0x820a20a2);
remainders_pos[538].modulo_value=2152; remainders_pos[538].bit_mask=create_bit_mask(68,0x2a88020a,0xa828208,0xaaaaa22,0x8a802aa2,0x2888a800,0x8a2a0a28,0x2a8080a8,0x8a08a2aa,0x220aa8,0x2a22a00a,0x208082,0x282a0a2a,0x882000,0xaa22882,0xaaa2222,0x280a8282,0xa20a8802,0xa822a880,0xa28280,0xa22220,0x8aa28822,0x282aaa28,0x2aa020a0,0x20228a8a,0x2aa0aa02,0xa0880022,0x80200208,0x20880a8a,0x882820a0,0x808aa808,0xa2002a,0x80a22000,0x8028a282,0x8a00a2a8,0x8a28022a,0x88a0282,0x8a0000,0x2aa202a8,0x8282220,0xa022080a,0x800802a2,0x220a20,0xaa0aa888,0xa2880880,0x80aa8a2,0xaaa8280a,0x228aa228,0x888a0088,0x828a0008,0x2a882a02,0x22a08a02,0x82a02880,0x88aaa082,0x288aa088,0x8220082,0xa0a82800,0x20800a8,0xa88a882,0xa08800a0,0x8a20a22a,0x202a8aa,0xa0a8a22a,0x2a222828,0xa802a200,0xaaaaa08a,0x8282a088,0x8022a820,0xa0);
remainders_neg[538].modulo_value=2152; remainders_neg[538].bit_mask=create_bit_mask(68,0xa2008a82,0x820a0a80,0x822222aa,0x208a22a,0xa0002088,0x2a282a0,0xa2080820,0x2802a22,0x88aa8220,0xa2aa2882,0x88a8080a,0xa0a282a2,0x8800a888,0x822aa00a,0x8222aaaa,0xa0820a0a,0x2a82008a,0x20aa2008,0x882a0a08,0x882aaaa8,0x22a00aa,0xa0a222a0,0xa228a828,0xa8aa0202,0xa228228a,0x2800882a,0x8a88a80,0xa8008202,0xa0a828,0x8022080,0x882a88a2,0x82aa888,0x8a02a0a,0x2882a20,0x2a08aa2,0x80028a0a,0x88028888,0xa22a8a20,0x80a0aaa8,0x28aa8082,0x8808a2a,0x88aa82a8,0x22822000,0x2a008008,0x8082202a,0x2220a082,0xaa022aa0,0x28800,0xa028880,0xa200a28a,0xaa28028a,0xa28a008,0x22280a,0xa0022800,0x80aa880a,0x2820a088,0x8a808820,0x8200200a,0x20008828,0x2a82aa2,0x8a8a2022,0x28202aa2,0xa2aaa0a0,0x208a2a88,0x22222802,0xa0a2800,0x8aa20a8,0x28);
remainders_pos[539].modulo_value=308; remainders_pos[539].bit_mask=create_bit_mask(10,0x2080222,0x2828a0,0xa28a8020,0xa8020802,0x20802a08,0x2a20a80,0x28a80208,0x8000802a,0x800a282,0x88802);
remainders_neg[539].modulo_value=308; remainders_neg[539].bit_mask=create_bit_mask(10,0x8a80822a,0x820a020,0x220208a8,0x200a8082,0xa8002280,0x822a0208,0xa0208280,0x8880822,0x802a02,0x8a);
remainders_pos[540].modulo_value=120; remainders_pos[540].bit_mask=create_bit_mask(4,0x20882,0x28220800,0x2088,0x822080);
remainders_neg[540].modulo_value=120; remainders_neg[540].bit_mask=create_bit_mask(4,0x808a0002,0x20228000,0x808a000,0x22800);
remainders_pos[541].modulo_value=541; remainders_pos[541].bit_mask=create_bit_mask(17,0x9ef992ba,0x9623aa1c,0xb113dc25,0x3e4973f0,0x8c8dfbe6,0x385cfe52,0x62d8a055,0x582b6d3e,0xa0b00341,0x9f2db506,0x2a8146d1,0x529fce87,0x19f7ec4c,0x43f3a49f,0x690ef223,0x4e15711a,0x175267de);
remainders_neg[541].modulo_value=2164; remainders_neg[541].bit_mask=create_bit_mask(68,0x2200a22,0xaaa2280,0x288a00a8,0xa280aa28,0x222a,0xa080228a,0xaa002888,0x80a2a0a2,0x28288a88,0x2a0288a,0xa2088a08,0x8a02020a,0x802a2080,0x8a2a2802,0xa0822aaa,0x8288a882,0x8a8aaa02,0xa02aa8a,0x2a08a88a,0x8aaaa208,0xa00a2a2,0x880822a0,0x2a820202,0x828882,0xa288a02a,0x88a88a0,0xaa282a28,0x2888a002,0x28a2008,0x2aa22000,0xa0a2a80a,0x80a80288,0xa22aa,0x22228022,0xaaa22a82,0xa800220,0xa0882a8,0x20a802,0x82aaaa22,0x8828aa20,0x2002aa08,0x88a82028,0x820a0880,0x8aa02a08,0x80228882,0xa882a2a2,0xa0aa0228,0x82020a,0xa028a200,0xa0a08808,0x80008002,0x8282a000,0x8a028808,0x20800022,0x2a82a820,0xa08a8a20,0xa200a2a2,0x2a8a088,0x2820882a,0xa888088,0xa0020a02,0x8a08882a,0xaaa0822a,0x8200222a,0x8828200a,0xa80aa0,0x22aa8220,0x220aa);
remainders_pos[542].modulo_value=2168; remainders_pos[542].bit_mask=create_bit_mask(68,0x22a2aa02,0x2028208,0xa20a0088,0x8220a028,0xaa2088a2,0xaaaa2228,0x2a8880a8,0xa002a282,0x22a22a8,0xa82a288a,0x2080022,0x2aaa222,0x2a000aa2,0x28a0aa8,0x8008a0aa,0x80a828a8,0x28a22a2a,0x80202208,0x8808280a,0xa00a08aa,0x2800a082,0xa200aaa0,0xa2220002,0x22aa8a2,0xa0882028,0x8020022,0x228202aa,0x80a8880,0x8222000,0x82088a20,0x82a0a22,0xa88aa0a2,0x28a282a2,0x82a00202,0x2880800a,0x2a8a828a,0x208a0aa2,0x8288a0a8,0x2008a220,0x20000888,0x880222a0,0x20aa8082,0xa880080,0x80280822,0x8a8a2aa8,0x8a800088,0x280aaa00,0xa820a00,0x22aa20a0,0x2a02820,0x28208808,0x2a8a88a,0x22a282a,0x80aa0a20,0xa82aa0a2,0x808aa000,0x80888aaa,0x2a880020,0x80a228a8,0xa2a88a8,0xa8828a80,0xaa2a0222,0xaa2888aa,0x828a2208,0x8a280a08,0x802200a0,0x88208280,0x80aa8a);
remainders_neg[542].modulo_value=2168; remainders_neg[542].bit_mask=create_bit_mask(68,0xaa2a228a,0x8a8a0a80,0x2a828800,0xaa828a0,0x22a8002a,0x2222aaa0,0xa2000820,0x288a2a0a,0x8aa22a20,0x20a2a002,0x8a8088aa,0x8a222aaa,0xa288822a,0x8a028220,0x8802822,0x820a020,0xa02aa2a2,0x8a8aa80,0x80a082,0x28828022,0xa088280a,0x2a882228,0x2aaa888a,0x8aa2202a,0x2800a8a0,0x808a88aa,0xaa0a8a22,0x80820008,0x80aaa888,0xa8002a8,0x80a282aa,0x2002282a,0xa02a0a2a,0xa288a8a,0xa0080882,0xa2020a02,0xa802822a,0xa002820,0xa8802aa8,0xa8888000,0x8aaa28,0xa822080a,0x82008008,0x8a080aa,0x202a220,0x2088800,0xa0822288,0x820a8288,0xaa22a828,0x8a28a0a8,0xa0a80080,0x8a202002,0x8aa2a0a2,0x82282a8,0x20a2282a,0x8022888,0x8000222,0xa20088a8,0x82aa020,0x82a20020,0x200a0208,0x22a28aaa,0x22a00022,0xa02aa80,0x2a08280,0x8aa8828,0xa80a08,0x82202);
remainders_pos[543].modulo_value=2172; remainders_pos[543].bit_mask=create_bit_mask(68,0x820a2882,0x8220220,0xa002200,0x828a0080,0xa28020,0x800a288,0x228a0082,0x8820828,0x8008a08a,0x2020022,0x28a28828,0x8280288,0xa0880800,0x8800220,0x88000008,0x82000822,0x8a28200,0xa080002,0x822882,0x20a28a20,0x808a202,0xa2822020,0x20028020,0x80200000,0x8a0882,0x200228,0x80208202,0x820a0000,0x20000820,0x828000a,0xa20a0002,0x28808a00,0x8088082,0x2000082,0x8200008,0x88202202,0xa00a2022,0x8000a08,0x8a000282,0x20820000,0x20000a08,0x8082080,0x2880080,0x28220a20,0x80000080,0xa0802800,0x808828,0x8808a202,0x8a28a0,0x8228820,0x800020a,0x200828a2,0x28820008,0x2000002,0xa0880022,0x20220,0x82280282,0x28228a2,0x28800808,0xa20a200,0x82820822,0x8200a28,0x228a002,0x208028a0,0x200a28,0x8800a,0x20880882,0x8228a08);
remainders_neg[543].modulo_value=2172; remainders_neg[543].bit_mask=create_bit_mask(68,0x2822002,0xa28a28,0x8208a288,0x2020800,0x8220828,0x80082200,0xa2020802,0x28020,0x8002002,0x828a08a2,0x20220020,0x80208200,0x20000080,0x8a28,0x88080,0x28800a2,0x220a08,0x8200808a,0x800a2002,0x28220228,0x8000228a,0x220a28a0,0x28820828,0x8288088,0x80020002,0x8a08a20,0x828028a,0x2820880,0x28808028,0x80208082,0x22820882,0x20000208,0x8000000a,0x82880802,0xa08800,0x28a28a,0x208228a2,0x808200,0x208820a,0xa0020880,0x28808200,0x8000a008,0x82000800,0x20a28228,0x8088008,0x20082080,0x8000020,0x228a,0x80022020,0xa20028,0x80088282,0xa0802022,0x20028800,0x8a08808a,0x200008a2,0x8828a28,0xa20820a,0x820a2022,0x20008000,0x82282288,0x20a00a2,0xa08220,0x8a20200a,0xa0082020,0x8a08220,0x88000082,0xa0000002,0xa20200);
remainders_pos[544].modulo_value=136; remainders_pos[544].bit_mask=create_bit_mask(5,0x2a008a2a,0x2082a022,0xa820800,0xa200a888,0xa8);
remainders_neg[544].modulo_value=136; remainders_neg[544].bit_mask=create_bit_mask(5,0xa28802a2,0xa80228aa,0x820a8088,0x2a082000,0x20);
remainders_pos[545].modulo_value=545; remainders_pos[545].bit_mask=create_bit_mask(18,0xa4a72316,0x6632c437,0x5212259d,0x9e340d84,0x611417b,0xb106038c,0x285bc010,0x4354a510,0x9aa31567,0x2294ab0b,0x200f6850,0xc7018234,0x7a0a2180,0x86c0b1e7,0xe6912128,0xb08d319a,0xa3139497,0x1);
remainders_neg[545].modulo_value=2180; remainders_neg[545].bit_mask=create_bit_mask(68,0x282a2a8a,0xaa2a08a2,0xa8aa000,0x2a08008,0x82888822,0x288a0a00,0xa0020880,0xa882088,0x22288aa,0xa2082202,0xa8822008,0xa080aa0,0x2282a008,0xa00282a,0x2208a820,0x28002802,0x2a0a0802,0x8282a0,0xa000a2,0x20a88222,0xa0a00280,0x802a0a22,0x2a808280,0x802208a8,0x2208228,0xa88a2202,0x88208a82,0x8820028,0x28288a0,0x20888a08,0x80082a02,0x2a8a80,0x2882a2a8,0x8aa2a0a2,0x2000002,0x20802,0x80000aa0,0x20082288,0x80808822,0x808228,0x28a2088a,0x800a0880,0x22008000,0x28220a2,0xa02808,0x808a0008,0x20200a82,0x28a0200,0x280082a,0x2a80a22,0x820a82,0xa8208022,0x280aa020,0x800a022,0x2028a02a,0xa8020200,0x2880a0,0xa028008,0x8220a008,0x802222,0x88280080,0x88228a00,0x20808028,0x880808a,0xa2080222,0xa8000088,0x8200002,0x2020);
remainders_pos[546].modulo_value=2184; remainders_pos[546].bit_mask=create_bit_mask(69,0xa2020022,0x28200020,0x288288,0xa2002802,0x28020a00,0x2280008,0x80020000,0x20028008,0xa008202,0x8200a2,0x8028800,0x8002288,0x82020820,0x8208220,0x200200a,0x20882820,0x20828a20,0x2,0x20000802,0x8220220,0x82080,0x20802002,0x208020,0x8088,0x208a2,0x20828000,0x8a000200,0xa0822000,0x20800228,0xa000,0xa00a0002,0x820,0xa202000,0x2022000,0x8808000,0x808a000,0x20000000,0xa00a08,0xa000080,0x80020800,0x8820a28,0x8000a200,0x2820800,0x20800000,0x200008a,0x2080022,0x8020808,0x8200080,0x80882002,0x20000808,0x80,0xa2820880,0x28220808,0x8008008,0x820820a0,0x20808208,0x88002008,0x22802022,0x820000,0x8200a08a,0x2800880,0x800220,0x288000,0xa0802820,0x28008a00,0x82280080,0x82822,0x808a08,0x88);
remainders_neg[546].modulo_value=2184; remainders_neg[546].bit_mask=create_bit_mask(68,0x228a0822,0x20a08820,0x80200200,0x22082082,0x20820208,0x8a200080,0x8a0080,0x28020800,0x2080282,0x800a0022,0x20008,0x8000a200,0x28200a0,0xa00228,0x8a08a002,0x200020a0,0x20020228,0x8808008a,0xa0800082,0xa20a20,0x8800a008,0xa0082802,0x8a00020,0x80000,0x800a0022,0x28020808,0x2080288,0x200a2000,0x20008a20,0x8082080,0x20820882,0x8008028,0x8220a080,0x820a2880,0x8,0x80002008,0x20880880,0x208200,0x82008008,0x820080,0x20220,0x82288,0x820a0000,0x28000800,0x82088002,0x820008a2,0x28000,0x80280008,0x2082,0x28808000,0x8088000,0x22020000,0x20228000,0x80000000,0x2802820,0x28000200,0x82000,0x220828a2,0x28800,0xa082002,0x82000000,0x8000228,0x8200088,0x20082020,0x20800200,0x2208008,0x80002022,0x200);
remainders_pos[547].modulo_value=2188; remainders_pos[547].bit_mask=create_bit_mask(69,0xaaa0228a,0x88a20880,0xa8aa8aa0,0x2220882,0xa0a80202,0x2a22822a,0x8a2aa802,0x8a082a02,0x28a2a020,0x222a282a,0xa088882a,0x2080a82,0x2a8a000a,0xa2008000,0x2080802a,0xaa828a88,0x800a2a20,0x8082aa02,0x20a0028,0xaa2a28a2,0xaa2a8800,0xaa8200aa,0x2a28a82,0xa222280a,0xa0808880,0xa8082080,0x80a28228,0xa000820a,0x8a08808a,0x8aa02800,0x22888a8a,0x200200a,0x22882228,0x8a8002a,0x2a20802,0x82288a80,0x80080288,0x22288a00,0x80aa2a2a,0x220a2002,0x2000aa20,0x28a02a08,0x8202a288,0x2020a020,0x8888a022,0x28a80a02,0x82aa82a,0x2a8aaaa0,0x8a8aa002,0xa0808a2,0xa8202280,0x8a00280a,0x282aa08a,0x2020822a,0x2008aa80,0xa2a8000,0x2080a00,0x2220a82a,0x8a888a82,0xa8a28a82,0x820a2080,0xaa8a280a,0x288a8802,0x2a0aa88,0x8880808,0x2aa2a822,0x8a220aa,0x80aaa022,0xa28);
remainders_neg[547].modulo_value=2188; remainders_neg[547].bit_mask=create_bit_mask(69,0x2228aa02,0x2a8008,0x20220228,0x8aaa800a,0x28208a8a,0xa2aa0aa2,0x2a2208a,0x280a28a,0xa02a28a8,0xaaa2a0a2,0x280000a2,0x8a80820a,0xa2028882,0x2a880888,0xa80808a2,0x220a0200,0x882a2a8,0x80a2282,0x8a8288a0,0x22a2a02a,0x22a20088,0x220a8822,0x8a2a020a,0x2aaaa082,0x28080008,0x2080a808,0x82a0aa0,0x28880a82,0x2800802,0x228a088,0xaa000202,0x8a88a882,0xaa00aaa0,0x802088a2,0x8a2a808a,0xaa00208,0x8808a00,0xaaa00288,0x822a2a2,0xaa82a88a,0xa88822a8,0xa028a280,0xa8a2a00,0xa8a828a8,0x28aa,0xa020828a,0x80a220a2,0xa2022228,0x202288a,0x8280802a,0x20a8aa08,0x288a082,0xa0a22802,0xa8a80aa2,0xa8802208,0x82a20888,0x8a808288,0xaaa820a2,0x200020a,0x202a020a,0xa82a808,0x2202a082,0xa002008a,0x8a282200,0x80008080,0xa22a20aa,0x802aa822,0x82228aa,0x2a0);
remainders_pos[548].modulo_value=1096; remainders_pos[548].bit_mask=create_bit_mask(35,0x20a8a82,0xa80200a0,0x20822222,0xa882aa8,0x8a8228a2,0x820aaa0,0xaa2222,0x82002802,0x2a0a0a8a,0x280082a,0x888aa008,0xaaa08208,0xa2882a20,0xaa822a08,0x88882082,0xa00802a8,0x2a2a0800,0x20a8a828,0x80200a00,0x822222a,0xa882aa82,0xa8228a20,0x820aaa08,0xaa22220,0x20028020,0xa0a0a8a8,0x280082a2,0x88aa0080,0xaa082088,0x2882a20a,0xa822a08a,0x8882082a,0x802a88,0xa2a0800a,0x82);
remainders_neg[548].modulo_value=1096; remainders_neg[548].bit_mask=create_bit_mask(35,0x8a82020a,0x208a8828,0xa80aaaaa,0x8200a220,0x20aa02a,0x80a82228,0x8822aaaa,0xa88a08a,0xa2828202,0x8a0880a2,0x22880,0x22280a80,0x2200a2a8,0x220aa280,0xa80a,0x28808a20,0xa2a28088,0xa82020a0,0x8a88288,0x80aaaaa2,0x200a220a,0x20aa02a8,0xa822280,0x822aaaa8,0xa88a08a8,0x28282020,0xa0880a2a,0x228808,0x2280a800,0x200a2a82,0x20aa2802,0xa80a2,0x8808a200,0x2a280882,0xa);
remainders_pos[549].modulo_value=732; remainders_pos[549].bit_mask=create_bit_mask(23,0x2082022,0x28200,0x80082202,0xa2822882,0x200208,0x808028a,0x20028a0,0x20a28820,0x800a08a,0xa0002022,0x8200020,0x808a202,0xa0800082,0x8808000,0x8a20a002,0x8228a0,0xa28008,0xa280202,0xa2080080,0x28228828,0x8088200,0x82800,0x8808208);
remainders_neg[549].modulo_value=732; remainders_neg[549].bit_mask=create_bit_mask(23,0x828028a2,0x8820a08,0x800a28a,0x220a2002,0x8a08a00,0x80008202,0x82882020,0x28220028,0x80082002,0x208828a2,0xa08828,0x8000228a,0x20080802,0x8,0x228208a,0x800a2020,0x8220800,0x8220828a,0x22800800,0x20a20020,0x288,0x80802080,0xa00);
remainders_pos[550].modulo_value=440; remainders_pos[550].bit_mask=create_bit_mask(14,0x2820228a,0xa8020080,0xa828008,0x80222022,0xa80200,0x208a2288,0xa20,0x808a000,0x2288a2,0x200802a,0xa0880888,0x2a200282,0x28020080,0xa28808);
remainders_neg[550].modulo_value=440; remainders_neg[550].bit_mask=create_bit_mask(14,0xa0a82202,0x200a8800,0x20a0080,0x8a2a8a2,0x80200a08,0xa0022a00,0x880882a0,0x802008,0x88a2002a,0x2880822,0x28000000,0xa2288202,0x200a8008,0x220080);
remainders_pos[551].modulo_value=2204; remainders_pos[551].bit_mask=create_bit_mask(69,0x2200a22,0x88aaaa0,0x2a0200a8,0x22820a8a,0x80a00008,0xa288028a,0xaa80202a,0x8a202a8a,0x28a080,0x228088a2,0xa202a080,0xa00a808a,0xaa82088,0x2a82a8a2,0x828a0a8a,0x808088a0,0x888aa8a8,0x2222208,0x2a2a0020,0x82022820,0x8202,0x20028820,0x2882228,0xa88020a,0x2a082228,0x20a88822,0xa8022828,0x8a000200,0xa20800a0,0x882a0202,0x8a08a0aa,0x8a000202,0x202a02a,0xa088a800,0xa8220882,0x2a220,0x2a80a808,0x2808000a,0x2aa0a20a,0xa8080a82,0x20a00208,0xa008000a,0x82828802,0x888222a0,0x288820a,0xa08022a,0x82888228,0x822800,0x28082000,0x80828808,0x800a8a,0x22088888,0x22a2aa22,0x20a22020,0x8a2a0a28,0x8a2a82a,0xa22082aa,0xaa202a00,0x8020a808,0x8a22028,0x2020a280,0xaa2a808a,0xaa80802a,0x2a280228,0x820000a0,0x8a2a0828,0x2a0080a,0xaaaa22,0x88a0088);
remainders_neg[551].modulo_value=2204; remainders_neg[551].bit_mask=create_bit_mask(69,0x8aa082aa,0x80022228,0x220a8820,0xaa0a8202,0x8288880,0x2a008202,0x2208a0a2,0x228a202,0x88a02808,0x2a08002a,0x2a8a2800,0x28820802,0x82202800,0xa202202a,0x2028202,0x8080028,0x22020,0x8aaaaa00,0xa2a288a8,0xa82a0a8,0x80880a8a,0x288a00a8,0x8a00aaa0,0x82008a82,0xa28022a0,0xa82000aa,0x200aa0a0,0x2888a88,0x2a808820,0xa28a82,0x2802822,0x2080a8a,0x8a8a28a2,0x20002088,0x20aa800a,0x888a22a8,0xa2082000,0xa0808882,0x22202a82,0x2080820a,0x28288a80,0x28808882,0xa02008a,0xaa228,0x8a000a82,0x82008aa2,0xa000a20,0x880aa088,0xa080a880,0xa0080,0x88080202,0xaa800000,0xa22a22aa,0xa82a28a8,0x2a282a0,0x802a2022,0x2aa80a22,0x22a0a280,0x8a82080,0x2aa8a0,0xa8282a08,0x22a20802,0x220800a2,0xa2a08aa0,0xa088028,0x2a280a0,0x8a288082,0x82222a2,0x28800);
remainders_pos[552].modulo_value=552; remainders_pos[552].bit_mask=create_bit_mask(18,0xa20a0002,0x28820828,0xa00220a,0x82022020,0x20808208,0x80282200,0x80002022,0x820,0x8a00a280,0x20000002,0x8000208,0x88280288,0x82020800,0x8808220,0x8800a008,0x208228a0,0xa08a28,0x80);
remainders_neg[552].modulo_value=552; remainders_neg[552].bit_mask=create_bit_mask(18,0x22020882,0x20028020,0x8208a282,0x28228a0,0x28000a00,0x820a288,0x8828a2,0x8808028,0x2082208,0xa0880082,0x808a00,0x208200,0x28a0000,0xa28,0x82080,0xa00a2020,0x8200220,0x8);
remainders_pos[553].modulo_value=2212; remainders_pos[553].bit_mask=create_bit_mask(70,0xa820a0a,0x28088202,0x8002082a,0xa000088,0x2a822800,0xa8a2208,0x808288,0x2888802a,0x22288280,0x20a80222,0x28020000,0x2022a0a,0x820800a0,0x200882aa,0x822a28,0x2a8802a2,0x228a02a0,0x8a22a8aa,0x8aa02a02,0x8022a00,0x8020a2a2,0x820aaa08,0x2002800,0xa282a2,0xa8200002,0xa2222200,0x88a00a08,0x802a008,0x8a808a08,0xaa08222,0x28000a2,0x88800,0x80a2a082,0xa820208,0x282a8282,0x8082a08,0x20a8a022,0x2220008,0xa0002800,0x8820aa08,0xa202a28,0xa80202,0xa00a222,0x88888a2,0x82a0,0xa828a008,0x2800808,0xaaa0820,0xa8a08022,0xa880208,0xa80aa20,0xa2a88a28,0xa80a288a,0xa8022a80,0x8a882008,0xa8220082,0xa002082a,0xa880800,0x8028a,0x8802a080,0x28228888,0x80222280,0x2820200a,0x88a2a02,0x2882a82,0x20000a00,0x82080022,0x822028a,0xa082a08,0xa);
remainders_neg[553].modulo_value=2212; remainders_neg[553].bit_mask=create_bit_mask(70,0x820a8202,0x20800a82,0x8a00a2,0x82088800,0xa202a088,0x82022a80,0x88080200,0xa0000822,0x2aa00a00,0xa0208aaa,0xa00a8888,0x8a82a282,0xa800028,0xa8800222,0x880aa220,0x22008a22,0xa2028a28,0x22a2022,0x220a28a,0x808a2288,0x8a8222a,0xa822200,0xa88a000,0x802a0a2a,0x2028888a,0x2aa2aa88,0x280280,0x808a2080,0x2080200,0x2280aa2,0x8208882a,0x88000088,0x8222802,0x820a0a80,0xa0a2020a,0x8080a200,0x282028a2,0x82aa8880,0x2808a088,0xa02280,0x82a822a0,0x820828a,0x82882a2a,0x22,0x80880a28,0x20202880,0x8a008080,0x822200a8,0x202800aa,0x82008a00,0x20822a0,0x222002a0,0x2002a002,0x2082a208,0x2002880,0x20aa800a,0x288a8022,0x2008080,0x80808a02,0xa2808,0xa0a20000,0x82a2a08,0xa0a8a082,0x8002a20a,0xa00a202,0xa0888288,0xa0088aa,0x80a28a02,0x82802280,0x2);
remainders_pos[554].modulo_value=2216; remainders_pos[554].bit_mask=create_bit_mask(70,0x2800aa2,0xa2aaaaa8,0x2282a080,0x220a2828,0x200800,0x822aa20a,0x2880a22a,0xaaa2080a,0xaa0a2208,0xaa2200a,0xa0802a08,0xa22a822a,0x8a88a0a,0x2a0aa888,0xa282a828,0x2222a080,0x8828822,0x88a202a2,0x22222882,0x820a8a02,0xa0282802,0x8a88800,0x2200a088,0x8a002280,0xa8a02a,0x2000aa22,0x8a220,0x8820a8a2,0x2280220a,0x2aa0a200,0xa22aa8aa,0x82282820,0x20a8a02,0x82a80000,0x8200aa,0x2a82aa,0xa2a08000,0x28288280,0x2aa88a08,0x8a0aa8aa,0x88028800,0x2a0822a0,0x8a20008a,0xaa000808,0xa2a0088,0x8800a2a8,0xa008802,0x222a2022,0x28280a00,0xa2a08280,0x28888880,0x808a2282,0x2282208a,0xa888888,0x2a828a02,0x2aa0a828,0xa22a2022,0x82a88aa0,0xa8020aa8,0x88aa020,0x88a0aaa0,0x208aaa20,0x8a0228a0,0x8aa882a8,0x200800a0,0x28a08800,0xa828828,0xaaaa8a02,0xa002802a,0x8a);
remainders_neg[554].modulo_value=2216; remainders_neg[554].bit_mask=create_bit_mask(70,0x8a08822a,0x2a222220,0xaa0a2808,0xaa82a0a0,0x88a88088,0xaa22a82,0xa0082aa2,0x222a8082,0x2282aa80,0x822aa882,0x2808a280,0x2aa20aa2,0x80200282,0xa2822000,0x2a0a20a0,0xaaaa2808,0x800a00aa,0x2a8a2a,0xaaaaa00a,0xa82028a,0x28a0a08a,0x80200088,0xaa882800,0x288aa08,0x882028a2,0x288822aa,0x88802aa8,0xa8202a,0xaa08aa82,0xa2282a88,0x2aa22022,0xaa0a0a8,0x8a82028a,0xa208888,0x880a8822,0x88a20a22,0x2a280888,0xa0a00a08,0xa2200280,0x2822022,0x8a0088,0xa280aa28,0x2a88802,0x22888080,0x82a28800,0x882a20,0x8288008a,0xaaa2a8aa,0xa0a08288,0x2a280a08,0xa0000008,0x802aa0a,0xaa0aa802,0x82000000,0xa20a028a,0xa22820a0,0x2aa2a8aa,0xa200228,0x208a8220,0x800228a8,0x202228,0xa80222a8,0x28aa028,0x2200a20,0xa8808828,0xa0280088,0x820a00a0,0x2222028a,0x288a08a2,0x2);
remainders_pos[555].modulo_value=2220; remainders_pos[555].bit_mask=create_bit_mask(70,0xa0882882,0x220000,0xa00a088,0x82000002,0x8220,0x202282,0x820208a2,0x20800228,0x2200000,0xa2080822,0x28a00800,0x82088280,0xa0020,0xa00008,0x282202,0x8a0800,0x8020000,0x2080280,0x2822882,0x20808200,0x8080080,0x220208a0,0x8008200,0x200002,0x80082020,0x8800200,0x8208280,0x20000a0,0x8800008,0x82080008,0x80080802,0x28a20,0xa002008,0x2880,0x20,0x80000080,0x800a0022,0x28000200,0x200208a,0x2082802,0x220200,0x80200,0x208200a0,0x220028,0x82002008,0x800080,0x20020800,0x8088008,0x20200a2,0x20208020,0x88280008,0x2080822,0x8020028,0xa200000,0x82800002,0xa00808,0xa000200,0x22082080,0xa28028,0x208a002,0x880882,0x208000,0x8082288,0x808008a2,0x20000828,0x82088,0xa00a0800,0x8800220,0x8220a000,0x822);
remainders_neg[555].modulo_value=2220; remainders_neg[555].bit_mask=create_bit_mask(70,0x20002002,0x8228800,0x2082000,0x2800882,0x800a28,0x8828220a,0x20a0022,0x28008220,0x8a208080,0x22800022,0x20200008,0xa000200,0x200a0,0x8208800,0x8020a20a,0x80020080,0x28800,0xa008208,0x82022002,0x20000a08,0x80000008,0x220a0020,0x800208,0x8820808a,0x802820,0xa08,0x80200200,0x2880020,0x8800,0x2008000,0x800082,0x8020220,0x208a080,0x80802000,0x808828,0x80008,0x208a2,0x20808208,0x8a00a002,0x82802002,0x8a20a08,0x88008280,0x200a0020,0x8a28820,0x208a000,0x80080800,0x28028000,0x80,0x82820822,0x20200828,0x200080,0x820000a2,0x828020,0x82208088,0x2080002,0x8200000,0x82088280,0x22802000,0x8220820,0x8200200a,0x80000002,0x8200800,0xa200,0x22,0x20808020,0x88002000,0x20020080,0x8228,0xa202008,0x22);
remainders_pos[556].modulo_value=1112; remainders_pos[556].bit_mask=create_bit_mask(35,0x2a88a22a,0xa022aa0,0x82822822,0x2aaa8080,0x8a0282,0xa2a2a00,0x8288a08a,0xa8008a,0x20008822,0x2a00882,0xa2282a20,0x8a8a0a20,0xa20000a,0x2aaa8828,0x88282020,0x880a0882,0xa22a80aa,0x8a22aa88,0x22aa02a8,0x228220a0,0xa8080828,0xa02822aa,0xa2a00008,0x8a08a0a2,0x8008a828,0x882200a,0x882200,0x82a2002a,0xa0a20a22,0xa8a8,0xa88280a2,0x820202aa,0xa0882882,0xa80aa880,0xa88a22);
remainders_neg[556].modulo_value=1112; remainders_neg[556].bit_mask=create_bit_mask(35,0xa2002aa2,0x828aa228,0xa0aa0aa,0xa2220808,0x8802820a,0x82a2a288,0xa002802,0x88208802,0xa88800aa,0x8a28800a,0x2aa0a2a8,0x20282a8,0x82a88882,0xa22200a0,0xa0a8a8,0x82800a,0x2aa20822,0x2aa2200,0xaa228a20,0xaa0aa828,0x208080a0,0x2820aa22,0x2a288880,0x280282a,0x88020a0,0x800aa882,0x8800aa88,0xa2a88a2,0x282a82aa,0x88882020,0x200a082a,0xa8a8a22,0x2800a00a,0x20822008,0x2002aa);
remainders_pos[557].modulo_value=557; remainders_pos[557].bit_mask=create_bit_mask(18,0x374b86d2,0xd8c24d92,0x650e96fb,0x8e9fa773,0xbe61aab8,0x43ecff81,0x11406870,0xb5031eaa,0x33331ab5,0x302b6b56,0x80a2155e,0xcdf08385,0x619f607f,0x7e5c4755,0x5c29b3b9,0x90c6f7da,0x74bb126c,0x12d8);
remainders_neg[557].modulo_value=2228; remainders_neg[557].bit_mask=create_bit_mask(70,0xaa820a0a,0xa800a,0xa8820a22,0x2022aaa,0x22a82220,0x8a202208,0x8888a0a8,0x288a8222,0xaaaa8228,0xa8a2a28a,0x82a8882,0x280a08,0xa802a8a2,0xa2808a88,0x80a02a20,0x80a2a02,0xa8228aa0,0xa8028a00,0x802a8a20,0xa02a280,0x2822a028,0xa88a880a,0x228aa00,0x808280a0,0xaa088aa0,0xaa8a2a28,0xa0a20aaa,0x8a220a88,0x88a82888,0x20822022,0x2220aa,0xaaaaa202,0x2228208,0xaa800a80,0x8282820a,0x8000a282,0x22aa80aa,0x8a282,0x2aa2a200,0x8a200a22,0x8822a22,0x20888828,0xa0880a2,0x800000a2,0xa0082020,0x8a8a0088,0x22aa0a82,0x880aa008,0x2820a880,0xa0a82a02,0x2a8a8202,0xa80a2080,0x280aa002,0xa8aa0082,0xa82a020,0x820a202a,0xa808808a,0x2aa20802,0x28a8a0a8,0x8028880,0x8202,0x88282280,0x88822280,0x20880a08,0x228a22a,0xa2aa2228,0x88000022,0xaaa220a2,0x8000aa80,0xa0a2);
remainders_pos[558].modulo_value=744; remainders_pos[558].bit_mask=create_bit_mask(24,0x22882002,0x28a20228,0x20a008,0x82820802,0x800008,0x8200288,0x808800a2,0x808a28,0xa20a082,0x82802,0x220,0x8a00a208,0x80000020,0x28200008,0xa08a080,0xa2020082,0x220228,0x8008208a,0x20022,0x20828220,0xa080080,0x808a2820,0x8228828,0x80);
remainders_neg[558].modulo_value=744; remainders_neg[558].bit_mask=create_bit_mask(24,0x22002882,0x20228a20,0x88282080,0x20a0082,0x8800,0x80288200,0x822,0x8000220,0x8228200a,0x80802082,0x8808a28,0x2082280,0x8008a0,0x20a08800,0x82002008,0x228a0802,0x8a20a20,0x800a002,0x808a08a2,0x28020a28,0x82008008,0x220a0,0xa20020,0x8);
remainders_pos[559].modulo_value=2236; remainders_pos[559].bit_mask=create_bit_mask(70,0x8a028a2a,0x8aaa022,0x222a0a08,0x22008028,0x8a2822a8,0x8a2a20a2,0x8080a80,0x822a8802,0x8088aa80,0x88282,0x2822a000,0x8802a80,0x8aa08200,0xa2208aa8,0xa00a0a02,0x2a28888,0x8a0a88a0,0x2a020a80,0xa8220082,0x2220802,0xa028288,0x20002888,0xa28a2002,0x2a00aa02,0xa8a808a0,0xa0022aa,0x2a00022a,0x8aa2000a,0x2202a220,0x8088082,0x80082082,0xa0aa2888,0x82828080,0x8200022,0x800a82,0x82a0020,0x28800082,0xa0202828,0x22228aa0,0x8208200,0x88202202,0x2088a808,0x8a0008aa,0xa88000a,0xaaa8800a,0x80a202a2,0xa80aa00a,0x88008a28,0x2228000,0x228280a,0xa8020888,0x88200882,0x202a080a,0xa22a0a,0xa22228a8,0xa80a0a00,0x22aa2088,0x820aa,0x802a8022,0xa882,0x28282200,0x202aa220,0x8022a88,0x202a0202,0x28a08a8a,0x82a8828a,0x82802008,0x20a0a88,0x2880aaa2,0xa8a280a);
remainders_neg[559].modulo_value=2236; remainders_neg[559].bit_mask=create_bit_mask(70,0x28a02a2,0x8022202a,0xa2a28280,0xaa8808a0,0x2a02a20,0x2a2a82a,0x80008200,0xa22008a,0x8002208,0x8880020a,0x20aa2888,0x8008a208,0x2200a80,0x2aa80220,0x2882820a,0x822a0000,0x2820028,0xa28a0208,0x20aa880a,0x8aaa8082,0x820a0a00,0xa888a000,0x2202a08a,0x2288228a,0x20208028,0x8280aa22,0xa2888aa2,0x22a8802,0xa28a2a28,0x8080080a,0x880280a,0x2822a000,0xa0a0800,0x802088aa,0x8808820a,0x80a280a8,0x2008880a,0x28a8a0a0,0x2aa20228,0x80a80a88,0xa8aa0a,0xa0002080,0x2888022,0x82000882,0x22200082,0x82a8a22,0x20022882,0x8802a0,0x8aaa0088,0xa20a082,0x208a8000,0xa0800a,0xa8a28082,0x882aa202,0x22aaa020,0x20828280,0xaa222800,0x8880a822,0x8a208a2,0x8808200a,0xa0a02a88,0xa8a222a8,0x8aa200,0xa8a28a8a,0xa0200202,0x2200a02,0xa08a800,0x82828200,0xa008222a,0x2022082);
remainders_pos[560].modulo_value=280; remainders_pos[560].bit_mask=create_bit_mask(9,0xa08a2202,0x8000802,0x20208,0x8a20a802,0xa2202808,0x802a08,0x20208080,0xa802000,0x8088a2);
remainders_neg[560].modulo_value=280; remainders_neg[560].bit_mask=create_bit_mask(9,0x28022a0a,0x88082,0x8a8280,0x220208a,0x22a0a000,0x88082280,0xa8280000,0x208a008,0x22);
remainders_pos[561].modulo_value=2244; remainders_pos[561].bit_mask=create_bit_mask(71,0x28000a2,0x20020008,0x88288,0x200a20a0,0x20200800,0x8202,0x80080002,0xa28020,0xa202080,0x20820020,0x828800,0x2002288,0x800a0800,0x820828,0x20a002,0x82020802,0x20220208,0x8a28a28a,0x82022022,0x820208,0x282002,0xa0820802,0x828008,0x8a200200,0x8a0800,0x20020820,0x8202280,0x200a2800,0x8008,0x2080002,0x802022,0x28228020,0x8a088000,0x80020020,0x28000a00,0xa00028a,0x80080020,0x220a20,0x80288280,0x80802000,0x20808,0x200008,0x28a0082,0x28808200,0x82080080,0x20a2000,0x8008a20,0x282000,0x20820a2,0x828008,0x8082008,0x80880822,0x28a28a28,0x808808a,0x2080822,0xa08008,0x82082008,0x20a0022,0x28800800,0x2282002,0x80082080,0x20808a00,0x8028a000,0x20020,0x8200008,0x2008088,0xa08a0080,0x28220000,0x80082,0xa0002802,0x8);
remainders_neg[561].modulo_value=2244; remainders_neg[561].bit_mask=create_bit_mask(70,0x82080022,0x28028800,0x88000200,0xa0022820,0x28a00008,0x8008028a,0x800882,0x8220828,0x8220a008,0x200a08a0,0x8020000,0x8a08a200,0x20000,0x8028020,0x88282082,0x28a0082,0x20a20a00,0x2202202,0x2822822,0x8028a00,0x820a08a,0x200a0082,0x20000,0x2288288,0x20080,0x28828020,0x80282208,0xa0822080,0x8800800,0x8a00800a,0x800028a2,0x20a20028,0x2000088,0x8a00a0,0x20008208,0x82008202,0x8008a0,0x8220228,0x200208,0x82880,0x8828000,0x8208080,0x2020802,0x20000a08,0xa008000,0x82022880,0x800228,0x8820a080,0x82802022,0x20800,0x80002080,0x22,0x20220220,0x2,0x820000a2,0x8200000,0x200a080,0x828208a2,0x20008000,0x8a20208a,0x802800,0x28000200,0x8202008,0x80820820,0xa08800,0x8a080000,0x20020000,0x20220808,0x8800800a,0x20802082);
remainders_pos[562].modulo_value=2248; remainders_pos[562].bit_mask=create_bit_mask(71,0x8a2a2a8a,0xa20a00a2,0x280aa808,0x2a88808,0x20a8a28,0xa888aa80,0xa2822220,0xa8a28aa,0x8aa288,0x20802a2,0x82a2202a,0x20aaaa,0xa888220,0x22882a02,0x8280800,0xaa00822,0x2a02a08a,0x22a2022a,0x8a02a020,0x28aa00a0,0xa8a828a2,0xa028822a,0x228880a2,0xa2aaa,0x2a220280,0x2a8a2a0,0x820082a2,0x82080a8,0x22228200,0x8008880a,0x2080a2aa,0xa88802a8,0xa800a828,0xaa0a208,0x80202082,0x8080200,0x8a0aa082,0x2a002a20,0x80222a28,0x8a02082a,0x222002aa,0x828888a0,0x2082000,0x8200822a,0x8a2a808a,0x8088a80a,0xa8a00002,0x22288aa,0x82280a8a,0x282a2aa8,0xaa288a,0xa80a20a,0x808a8808,0xa80a8a8,0x200aa0a2,0x20282088,0xa8228800,0x8222a080,0xaa080008,0x88a82aa,0x802080a8,0x8aa2008a,0x28a2a022,0x88828aaa,0xaa222a08,0xa2a08002,0x222a8028,0x2aa02820,0xa08a20,0xa8a8a28a,0xa2);
remainders_neg[562].modulo_value=2248; remainders_neg[562].bit_mask=create_bit_mask(71,0x2a2a202,0x2a82882a,0xa0822080,0x8a200080,0x8a8202a0,0x20002208,0x2a0aaaa8,0x8202a022,0x88022a00,0x8a808a2a,0xa2aa8a2,0x88a82222,0x82000aa8,0xaa00a28a,0x80a08088,0x822880aa,0xa28a2802,0xaa2a8aa2,0x28a28a8,0xa0228828,0x2020a02a,0x28a00aa2,0xaa00082a,0x8882a222,0xa2aa8a08,0x8a202a28,0xa88022a,0x80a80820,0xaaaa0a88,0x8800082,0xa8082a22,0x20008a20,0x208820a0,0x82282a80,0x8a8a80a,0x80808a88,0x282280a,0xa288a2a8,0x8aaa2a0,0x28a80a2,0xaaa88a22,0xa000028,0x8a80a888,0xa880aa2,0x2a20802,0x8002082,0x2028888a,0x8aaa0022,0xaa08202,0xa0a2a220,0x8822a002,0x82082a82,0x8020080,0x82082020,0xa882282a,0xa8a0a800,0x20aa0088,0xaaa2808,0x22808880,0x80020a22,0x8a80820,0x22a0802,0xa02a28aa,0xa0222,0x22aaa280,0x2a28088a,0xaaa208a0,0xa228a0a8,0x882802a8,0x20202a02,0x2a);
remainders_pos[563].modulo_value=2252; remainders_pos[563].bit_mask=create_bit_mask(71,0x8222a202,0x2282080a,0x808aa020,0xa2a8020,0x82a22aa8,0x2222028a,0x28222aaa,0xa002880a,0x8aa22a8a,0x88a8028a,0x20028aa2,0x2088a2a8,0x228a0008,0x2a20a228,0x80a2a08a,0xa00a80a0,0x2282aaaa,0x2802a0aa,0xa0880,0x2a02a800,0x28082a28,0x88288082,0xaa8208a0,0x82228a2,0x20aa8a0,0xaa02208,0x880202,0xa20aa802,0x88a082,0xa888880,0x880a02,0xaa0082a0,0xa8022a28,0xa28a08a8,0x88088a02,0x22022a0a,0xa28a80a,0x8802a2a2,0x200aa28a,0x22000a8,0x222a080a,0xa2200022,0xaa08a820,0x2200802,0x80aa0808,0xaa080208,0x888200a2,0x82aa8a2,0x228220a2,0x82028820,0xa80a828a,0xa000002,0xa8028022,0xa8288aa0,0x2a00aaaa,0xa8a020a0,0xa08a8a20,0xa288288,0xa2208200,0x280082a8,0x2a228aa,0x88aa2a28,0x2800aa2a,0x88828a02,0x8888aaa,0x88a82a28,0x2a8a02aa,0xaa202080,0x8288080,0xa8882a02,0x808);
remainders_neg[563].modulo_value=2252; remainders_neg[563].bit_mask=create_bit_mask(71,0xaaa2a8a,0xaa0a8082,0x80228a8,0x82a208a8,0xa2aa220,0xaaaa8a02,0xa0aaa222,0x288a0082,0x22aa202,0x208a02,0xa88a022a,0xa8002a20,0xaa028880,0xa2a82aa0,0x82a2802,0x28820828,0xaa0a2222,0xa0822822,0x88828008,0xa28a2088,0xa080a2a0,0xa0080a,0x220a8028,0x80aaa02a,0x8a822028,0x8228aa80,0x88008a8a,0x2a82208a,0x8800280a,0x82000008,0x8800828a,0x22880a28,0x208aa2a0,0x2a028020,0x80028a,0xaa8aa282,0x82a02082,0x8a2a2a,0xa8822a02,0x8aa88820,0xaaa28082,0x2aa888aa,0x228020a8,0x8aa8808a,0x8228080,0x22808a80,0xa882a,0x80a2202a,0xaa0aa82a,0xa8a00a8,0x20820a02,0x8288888a,0x208a08aa,0x20a00228,0xa2882222,0x2028a828,0x280202a8,0x82a00a00,0x2aa80a88,0xa0880a20,0x8a2aa022,0x22a2a0,0xa08822a2,0xa028a,0x80000222,0x20a2a0,0xa2028a22,0x22a8a808,0x80a00808,0x2000a28a,0x80);
remainders_pos[564].modulo_value=1128; remainders_pos[564].bit_mask=create_bit_mask(36,0x228008a2,0x20820228,0xa000,0x220a0882,0x20020200,0x2088282,0x20002800,0x20a20208,0x8a28800a,0xa08a0022,0x820808,0x8000280,0x882822,0x8800808,0x8220a,0x208000a0,0x28828808,0x8a28a200,0x22822800,0x20820,0x8820000a,0x200220a0,0x28220020,0x80002088,0x20820002,0xa20a20,0x228a288,0x808a08a0,0x28000820,0x82208000,0x80800882,0x20a08800,0xa000082,0x80820800,0x20028828,0x8a);
remainders_neg[564].modulo_value=1128; remainders_neg[564].bit_mask=create_bit_mask(36,0xa2080022,0x28020a20,0x88082088,0xa2820002,0x28828a08,0x8a00020a,0xa0882080,0x28228200,0x2200082,0x200208a2,0x8028000,0x80088208,0x800020a2,0x8000,0x8800a282,0xa0080820,0x20020000,0x2202288,0xa20a2080,0x8828020,0x288082,0xa08a2820,0x20a28828,0x808a000,0x200a0882,0x8228228,0x8a202200,0x20020,0x20808028,0xa280088,0x80002,0x28200008,0x8208800a,0xa0080,0x28820020,0x2);
remainders_pos[565].modulo_value=565; remainders_pos[565].bit_mask=create_bit_mask(18,0x8c835a1a,0x250b9e72,0x38024239,0x5a44bd88,0x47396a3a,0x188553d0,0x88089e37,0x689ac501,0x40931081,0x45a04232,0x46028d6,0x463b1e44,0x3882f2a8,0x969715a7,0x7046f48,0x29271090,0x4c539e74,0x1616b0);
remainders_neg[565].modulo_value=2260; remainders_neg[565].bit_mask=create_bit_mask(71,0xa0282,0xa80202a2,0x208a82a0,0x82802000,0x82a0a222,0x80080a08,0x2a2,0x20020088,0x88028808,0x80288a2a,0x88a0200a,0x8a228280,0x200a2220,0xa02882a,0x8288a200,0xa0aa0808,0x800a02a0,0xa8a8228,0x28008a2,0x82a8282a,0x288a0880,0x8a028002,0x228022a0,0xa228822,0x2028880a,0x88a00a80,0x8a008aa2,0x2002080,0x88,0x80800a2a,0x282a0882,0x200a0a22,0xa882000,0x200a82a,0x280022a,0x80a0220a,0xa22002a2,0x802008,0x8222a828,0x20202020,0x88820880,0xaa2000a2,0x228882a8,0x8082882a,0x80000a,0x82a008a,0x20208880,0x2022080,0x8082002a,0x8020aa02,0x8a0880,0x82200820,0x8200a8,0x8a802080,0x2080220,0x80882880,0xa02a8200,0x2a002080,0x822020,0x888202,0x28802a08,0xa8000080,0x2a08a080,0x8aa088a2,0xa280022a,0x882088,0x82020202,0xa0aa220,0x88020080,0xa2a00222,0x20280);
remainders_pos[566].modulo_value=2264; remainders_pos[566].bit_mask=create_bit_mask(71,0x82200a22,0xaaaa2a0,0xa8a8220,0x2282220a,0x8200222,0xa288228a,0xa0282a28,0x88a08aaa,0x22a80202,0x28a882,0x8a22200a,0x2a02a282,0x80a02a80,0x2aaaa88a,0x20a0aaa8,0x208aa820,0xa80a2a22,0x2aa08a,0x20a8008a,0x8a2220,0xa0a2a28,0x2800,0x2a0a8088,0x2a02a80,0x22208282,0x82aa0aa,0xa8002288,0x8a0882a2,0x282a0000,0x28820820,0xaa28a082,0x22822222,0x8080a0a2,0xaa22,0xaa228a02,0xaa088220,0x80a288,0x288aa00,0x888a0a02,0xaa888882,0x28820a28,0x28082082,0xa20000a8,0x2a8a8220,0x20228800,0x88aa0aa8,0x80828208,0xa802a80a,0x2202a0,0xa0002800,0x28a8a0,0x80888a2,0xa2002a,0x2aa20aa8,0x888a8a0,0x8082aa2,0xa82aaa0a,0x2a22aaa,0xa802a80a,0xa2828a80,0xa00888,0x88822a28,0x2280802a,0xaaaa20a,0x8a28a828,0x20a28822,0x88888008,0xa0a08882,0xa00882a2,0x820a8aaa,0x88a008);
remainders_neg[566].modulo_value=2264; remainders_neg[566].bit_mask=create_bit_mask(71,0xaa882aa,0x82222a28,0x82020aa8,0xaa0aaa82,0x80a88aaa,0x2a00aa02,0x28a0a2a0,0x280222,0xa2208a8a,0x88a0200a,0x2aaa882,0xa28a2a0a,0x828a208,0xa2222002,0xa8282220,0xa80220a8,0x2082a2aa,0x88a22802,0xa8208802,0x8802aaa8,0x8282a2a0,0x8888a088,0xa2820800,0x8a28a208,0xaaa80a0a,0x80a22822,0x2088aa00,0x2800a2a,0xa0a28888,0xa00a80a8,0x22a0280a,0xaa0aaaaa,0x808282a,0x888822aa,0x22aa028a,0x22800aa8,0x88082a00,0x8a002288,0x2828a,0x2200000a,0xa00a82a0,0xa080a80a,0x2a888820,0xa2020aa8,0xa8aa0008,0x228220,0x80a0a80,0x208a2082,0x88aa8a28,0x2888a088,0x88a02028,0x8080002a,0x882a88a2,0xa22a8220,0x80002028,0x8080a22a,0x20a22282,0x8a2aa222,0x208a2082,0x2a0a0208,0x88288000,0xaa2a0,0xaa0808a2,0x82222a82,0x2a020a0,0xa82a00aa,0x880,0x2828000a,0x28800a2a,0xa820222,0x2880);
remainders_pos[567].modulo_value=84; remainders_pos[567].bit_mask=create_bit_mask(3,0xa2080002,0x8a08020,0x80002);
remainders_neg[567].modulo_value=84; remainders_neg[567].bit_mask=create_bit_mask(3,0x22800802,0x200820,0x808a);
remainders_pos[568].modulo_value=568; remainders_pos[568].bit_mask=create_bit_mask(18,0x8aa8228a,0xa2a28880,0x2a2a0a20,0x82200800,0x802a202a,0x2aa8aa22,0xa00a020,0x8a888202,0x822800,0xa2002882,0xa0808222,0xa8080a00,0x288aa2a,0x82a808a8,0xa8002008,0x8a08a0a8,0xa202228a,0xa2882a);
remainders_neg[568].modulo_value=568; remainders_neg[568].bit_mask=create_bit_mask(18,0x220aa02,0x2a2a0008,0xa2a28228,0xaa88088,0x8a2a8a2,0xa22022aa,0x828828a8,0x2000a8a,0x880aa088,0x2a88a00a,0x28080aaa,0x20808280,0x8a0022a2,0xa208020,0x2088a880,0x2802820,0x2a8aaa02,0x2a00a2);
remainders_pos[569].modulo_value=569; remainders_pos[569].bit_mask=create_bit_mask(18,0x161767b6,0xa3162b1f,0x2ce781ff,0xe6a78934,0xc2817d5d,0x452d635,0x68c32530,0x7e36c4b7,0xb75119d1,0xfa2e622b,0x5bb48db1,0x8032930c,0xeb1ad28,0x9eeafa05,0xd0b24795,0x17fe079c,0xa3e351a3,0x1b79ba1);
remainders_neg[569].modulo_value=2276; remainders_neg[569].bit_mask=create_bit_mask(71,0x8a8aaa2a,0x2a8aa282,0xa02a0822,0x2a2a00a8,0xa08a080,0x888a0aa8,0xa00aa8a8,0xa2aa082a,0x2a888008,0x22a2aaa2,0x82280028,0x8aa0a80,0x822820a0,0x2222288,0x82a8a08,0x8a228a00,0x2a2a882a,0xa02a0228,0xa2a0a202,0x228aa08a,0xa080028a,0x2200828a,0xa2088a22,0xa8802820,0xa2080a82,0x2a20a000,0x8aa22aaa,0xaa288008,0x802aa082,0x8888a8aa,0x8280aa82,0xa2a00828,0xa028a802,0x8aa22082,0x8a8a0a2a,0x202a2aa,0x20a08080,0x8a220080,0xa8082a02,0x28aa0202,0x800a8288,0x8088882,0x8a002a80,0xaa882002,0x220088,0xaa0a2020,0x80a8a20a,0x282a8802,0x2088a20a,0x828aa222,0x82aa8a00,0x8a008220,0xa20a0200,0x22a02202,0x28802028,0xa0822080,0xa08028aa,0x822222a8,0xa282288,0x80a008aa,0x2aa8228a,0x20228,0xaa888022,0x28a00208,0x800aa00,0xa08888,0x8a08a0a8,0xaa0202a,0x28a02a08,0x82008022,0x20008082);
remainders_pos[570].modulo_value=2280; remainders_pos[570].bit_mask=create_bit_mask(72,0x820802,0x228a00,0xa000080,0x820028a2,0x20200208,0x8202208,0x82080882,0x28000208,0xa208088,0x20000022,0x20800,0xa000002,0x200a20a0,0x28828,0x220200a,0x80800882,0x8020220,0x2080200,0x808200a0,0x800220,0x20008a,0x22000822,0x28000000,0x80200000,0xa2802022,0x28200800,0x200282,0x80080,0x808028,0x208000,0x20082002,0x20000220,0x8,0x28208a2,0xa08000,0x8002080,0x20a0002,0x20828000,0x8a,0x80000820,0x8200808,0x2080080,0x2020000,0x200028,0x80080002,0x20082882,0x8028a00,0x80288,0x2800,0x20008800,0x80088,0x800200a2,0x820208,0x8020800a,0x80802000,0x20020208,0x8088082,0x228000a0,0x8a00828,0xa00a,0x20800080,0x800,0x208a088,0x80002822,0x20208220,0x88082082,0x80080820,0x28008220,0xa08a,0xa2880002,0x20820000,0x80);
remainders_neg[570].modulo_value=2280; remainders_neg[570].bit_mask=create_bit_mask(71,0x80020082,0x8220200,0x2088008,0x2802022,0x20a08a00,0x80282280,0x2000002,0x20008200,0x82200000,0xa0880822,0x8820000,0x82088082,0x20822020,0x8820020,0x8228a002,0x80002,0x28a20,0xa008208,0x20820,0x8a28,0x88280002,0x220800a2,0x20808008,0x8208088,0x22082022,0x20a00008,0x88288202,0x800000,0x8000820,0x80280000,0xa0802882,0x28000a20,0x8088080,0x82020022,0x200808,0x80082008,0x82020802,0x28020008,0x88008002,0x880020,0xa00000,0x8a008000,0x28a0080,0x8a08020,0x800a,0xa0002002,0x20200,0x8008200,0x80802080,0x20800008,0x88000000,0xa0822,0x28200,0x8200082,0x82800,0x28820a00,0x80000002,0x22000020,0x208020,0x82002,0xa0080800,0x8008000,0xa002000,0x8020a2,0x20200a28,0x200a,0xa0,0x20800220,0x88002002,0x22000802,0x28020808);
remainders_pos[571].modulo_value=2284; remainders_pos[571].bit_mask=create_bit_mask(72,0x2a28a2aa,0xa202a0a8,0x8882888a,0x2220aa82,0x20aa20a0,0x2a80aa00,0x200a0a,0x28a8aa02,0x2202002a,0x28aa002a,0x88a2a2a2,0x2a08a0a,0x2208880,0x82080888,0x8aa022aa,0x280208a2,0x82a28a,0x2820aaa,0x2a80028a,0xa082080a,0x208228a,0x8a0008a8,0x8aa222a2,0xaaa2228,0x22028228,0x20880808,0x8880aa80,0x2080aa8a,0xaa8a8020,0x8282a8,0x28aa802a,0x88a82880,0x220a0028,0x8802220a,0x80a0280a,0x820,0x80a02082,0x88022a02,0xa088a08,0x82a22280,0x2aa28022,0x28200a80,0x2a2aa2a8,0xa0208080,0xa0222a2a,0x220802a,0x28088202,0x8aaa0288,0x88aa2288,0xa28a8,0x820802a2,0x820aa28,0x2a8a02,0x8280a28,0xa8200aaa,0x8028a28,0x80aa28a2,0x2082aa8,0x20880222,0x20a80022,0xa8a22a0a,0xaa288a8,0x8088a80,0xa2a28a80,0x80080a,0xa02a8a0a,0x8aa0800a,0xa08880a0,0x2822282a,0xa808aa22,0xa28a82a0,0xaa8);
remainders_neg[571].modulo_value=2284; remainders_neg[571].bit_mask=create_bit_mask(72,0xa2a02a22,0x2a8a2820,0xa0002,0xaaa8220a,0xa822a828,0xa2082288,0x88a88282,0xa020228a,0xaa8a88a2,0xa02288a2,0x2a2a2a,0x8a280282,0x8aa80008,0xa808000,0x228aa22,0xa08a802a,0x880a2a02,0x820a8222,0xa2088a02,0x280a8082,0x8a80aa02,0x2888020,0x22aaa2a,0x8222aaa0,0xaa8a0aa0,0xa8008080,0x82208,0xa8082202,0x220208a8,0x880a0a20,0xa02208a2,0x20a008,0xaa8288a0,0x8aaa82,0x828a082,0x888880a8,0x828a80a,0x8aa28a,0x82800280,0xa2aaa08,0xa22a08aa,0xa0a88208,0xa2a22a20,0x28a80808,0x28aaa2a2,0x8aa808a2,0xa0800a8a,0x2228a00,0x22aa00,0x8882a020,0xa808a2a,0x80a822a0,0x88a2028a,0x80a082a0,0x20a88222,0x808a02a0,0x822a02a,0x8a80a220,0xa8008aaa,0xa82088aa,0x202aa282,0x822a0020,0x80800208,0x2a2a0208,0x88088082,0x28a20282,0x2280882,0x28000828,0xa0aaa0a2,0x208022aa,0x2a020a28,0x220);
remainders_pos[572].modulo_value=1144; remainders_pos[572].bit_mask=create_bit_mask(36,0x82208202,0xa2a8a08,0x80a282a8,0x808aa800,0x202208a8,0x22a0000a,0x80080822,0x2020002,0xa80288a,0xa22802a,0x28000808,0x88820200,0x8a0000a8,0x22a20880,0x2002aa20,0x2a828a0,0x220a2a8a,0x28082088,0x88220820,0x80a2a8a0,0x80a282a,0x8808aa80,0xa202208a,0x222a0000,0x28008082,0xa0202000,0xa0a80288,0x80a22802,0x2800080,0x88882020,0x8a0000a,0x22a2088,0x2002aa2,0xa02a828a,0x8220a2a8,0x808208);
remainders_neg[572].modulo_value=1144; remainders_neg[572].bit_mask=create_bit_mask(36,0xaa8028a,0x82220200,0x2a0a20,0x8022080,0xa8aa0020,0xa2288882,0x88080a2,0x8a0a880a,0x8200a002,0x2aa00a2,0x20888080,0xa8288,0x2008820,0xaa2a8008,0xa88a2220,0x82202028,0xa282a202,0xa0802800,0xaa8028,0x8222020,0x2a0a2,0x802208,0x2a8aa002,0x2a228888,0xa088080a,0x28a0a880,0x28200a00,0x2aa00a,0x82088808,0xa828,0x80200882,0xaa2a800,0x8a88a222,0x28220202,0xa282a20,0x80280);
remainders_pos[573].modulo_value=2292; remainders_pos[573].bit_mask=create_bit_mask(72,0x22002802,0x228a08,0x82288088,0x28a20a2,0x8008228,0x208228a,0x82080082,0x20200,0x828800a,0xa0020082,0x20000028,0x2002280,0x220220a2,0x8220828,0x2008288,0xa0080000,0x8a28220,0x208a000,0x80002882,0x28220a28,0x80,0x808a0802,0x20a08220,0x2282202,0x200a2002,0x28a00000,0x82080200,0x28008a0,0x8020200,0x2080082,0x828a2082,0xa20800,0xa28228a,0x8808a2,0x20828a20,0xa00a002,0x208800a0,0x20008a28,0x8a08a202,0xa00a2882,0x28200208,0x808208a,0x2082002,0x28000808,0x820a200,0xa2800802,0x800000,0x8808008a,0xa0880882,0x20208820,0x8020a,0x82802000,0x228a08,0x8082280,0xa20000a2,0xa08828,0x8000002,0x82022820,0x8828208,0x808a088,0x802880,0xa28000,0x82082008,0xa0022,0x8200808,0x8082002,0x20a2882,0x28028820,0x8828a08a,0x80022022,0x8820a28,0x80280);
remainders_neg[573].modulo_value=2292; remainders_neg[573].bit_mask=create_bit_mask(72,0xa2882082,0x8a20200,0xa200000,0x82022822,0x800a20,0xa00a202,0x2800802,0x8828a08,0x80200082,0x208a0802,0x28808820,0x8a08a208,0xa28a2822,0xa28020,0x8a080200,0x20800880,0x220a28,0x8a002088,0x882002,0x20a28220,0x88088008,0x20082,0x28200a28,0x8a20a28a,0xa0822882,0x20208808,0xa008288,0x82080020,0x828a08,0x8200800a,0x2022802,0x8228008,0x8220a202,0x80000022,0x28020228,0x8208208a,0xa0000820,0x28800220,0x200228a,0x20822002,0x20a08a00,0x8000a002,0x82802882,0x20808000,0x80282288,0x22080082,0x8008808,0x8002,0x20000002,0x28a00028,0x88008282,0x2082880,0x8a20200,0x8000a208,0x22880822,0x8200020,0x8008808a,0x28a20a0,0x20a00,0x80002000,0x80082000,0x8220808,0xa00a080,0x808208a2,0xa08000,0x8000a08a,0x82822002,0x20820028,0x202002,0x8a28a2,0x28220,0x8208);
remainders_pos[574].modulo_value=2296; remainders_pos[574].bit_mask=create_bit_mask(72,0xaa0a822a,0x22282800,0xa2a28880,0x22022882,0x8aa28082,0x80802a08,0x80a00000,0xa008a00a,0x2080aaa2,0x2222080a,0x200a008a,0x8808a08,0xa20808,0xa8082220,0x2802a822,0x8002a22,0x8282aa,0x222820a2,0xa2022802,0x80020208,0x202022a8,0x80228,0x2a2028a0,0xa0a8820a,0x28208088,0x2082a08a,0xa82220,0x28a,0x808a082a,0xaa82a02,0x80a00a8a,0x220a8000,0x2802280,0xa880200,0x8a02022,0x28020,0x20080280,0xa088080a,0x80008022,0x88028802,0x20002a0,0xa0a2a00a,0x280a82a,0xa820a2,0xa28000,0x808882a,0x28a20a82,0xa220208,0xa8a0822a,0xa2808,0x8288020,0x22a8808,0x8a208080,0x88802880,0x8a0828,0x20aa8282,0x2888a800,0x2a882a80,0x88820,0x2020208a,0x820a202,0x88a200a0,0x8a02088,0xa8aaa02,0x2a00a20,0x200000a,0xa220a802,0x8882028a,0x8a822880,0x8802228a,0xaa002828,0xa882a0);
remainders_neg[574].modulo_value=2296; remainders_neg[574].bit_mask=create_bit_mask(72,0x22820a22,0x2aa0a080,0x222a0008,0xa20aa00a,0x222080a,0x8082280,0x8288088,0x28802802,0x28082222,0xa2aa8082,0xa8028802,0x80000280,0x882a0080,0x2080a2a8,0xa08a2022,0x88a2a2,0x800a0a22,0xaa20a82a,0x2a82a08a,0x88a0a00,0xa8a8a220,0x88808a20,0x22a8a020,0x20200a82,0xa0280000,0xa8022802,0x88202aa8,0x88888202,0x8028022,0x2202282,0x288202,0xaa020888,0x8a00aa08,0x82000a88,0x8020a0aa,0x888a0828,0x28808a00,0x20008082,0x80808aa,0x2008a,0x8a880a28,0x282a2082,0x8a082022,0x820a822,0x802a0888,0x800000a2,0xa022820a,0x82aa0a80,0x202802a2,0x882a000,0xa008a0,0x82a20080,0x2280808,0xa008,0x880200a0,0xa8220202,0xa0002008,0x2200a200,0x808000a8,0xa828a802,0x80a02a0a,0x2a0828,0x8028a000,0x8202220a,0xa2882a0,0x82888082,0x2a28208a,0x28a02,0x20a2008,0x8aa202,0x22882020,0x200a20);
remainders_pos[575].modulo_value=460; remainders_pos[575].bit_mask=create_bit_mask(15,0x20082a82,0x800a0a00,0x280aa208,0x22088a8,0x8000a020,0x80822a00,0x2208a2,0x8800208,0x882028a2,0xa000200a,0x20880080,0xaa0282a2,0xa002208,0x8200800a,0x82a);
remainders_neg[575].modulo_value=460; remainders_neg[575].bit_mask=create_bit_mask(15,0xa800220a,0x8028280,0x20822280,0x8aa00020,0x882828,0x80a2288,0x882a0022,0x88280,0xa0a02a,0x2088a002,0xa8000808,0x220a0a22,0x288a200,0xa800882,0x22);
remainders_pos[576].modulo_value=24; remainders_pos[576].bit_mask=create_bit_mask(1,0x8a28a2);
remainders_neg[576].modulo_value=24; remainders_neg[576].bit_mask=create_bit_mask(1,0x22022);
remainders_pos[577].modulo_value=577; remainders_pos[577].bit_mask=create_bit_mask(19,0x8bcf1b5e,0x4a6f505f,0x390219f7,0x92441e5d,0xa3eb551d,0x8f630a8c,0x81f6b95b,0x690c9812,0x99bb81d9,0x6e077667,0x2064c25a,0x6a75be05,0xc5431bc7,0xe2ab5f14,0xe9e08926,0xbe610272,0xe82bd94b,0xeb63cf47,0x1);
remainders_neg[577].modulo_value=2308; remainders_neg[577].bit_mask=create_bit_mask(73,0x2028282,0x82a28882,0xa08a802a,0xa888280,0x2a228880,0x2aa8200,0x82a2082,0xa080008a,0x220800,0xa28aaaaa,0xa8a80a82,0xa2a82288,0x88a820a,0x2a228288,0x202800aa,0x22a88aaa,0x20a20082,0x22aa028a,0x8a02aa20,0x8022822,0xaa88aa22,0xa800a022,0x8a0a22a2,0x820a8880,0x8a20aa2a,0xa80a8a8,0xaaaa8a2a,0x822002,0x88000828,0x822a082,0x20aaa02,0x88a22a0,0xa088a80,0xa00a8828,0x88a2a0a,0xa0a0202,0xa2a0a0a2,0xa02088a2,0x2880aa00,0x8088a0a8,0x22088aa,0xa000a2aa,0x8a0228a2,0x28aaa882,0xaa228aa8,0x20800002,0x80a80a0,0x200a2088,0x8880a280,0x220a08a,0x2a0aa802,0x20088002,0x2822a8a2,0x2002a082,0x82a00222,0x8aa20a20,0x880222,0xaa82a20,0x82822020,0xa28088a8,0x82280200,0x80a80808,0x8202,0xa8a22aa0,0x8aaa8a0a,0x8a2028a0,0xa28002a2,0x8882202a,0x828880aa,0x2a808a0a,0x88820280,0x8282a2a2,0x2);
remainders_pos[578].modulo_value=136; remainders_pos[578].bit_mask=create_bit_mask(5,0x82808282,0x82828282,0x82828282,0x82028282,0x82);
remainders_neg[578].modulo_value=136; remainders_neg[578].bit_mask=create_bit_mask(5,0xa080a0a,0xa020a0a,0xa0a0a0a,0xa0a0a0a,0xa);
remainders_pos[579].modulo_value=2316; remainders_pos[579].bit_mask=create_bit_mask(73,0x228a0022,0x8220200,0x8a28a000,0xa20a2882,0x20228008,0x80282208,0x82002000,0x228220,0xa20a208,0x882082,0x220208,0x80008282,0x828a0882,0x20a00a00,0x8a288080,0xa00a2022,0x820228,0x8228a082,0x20000802,0x8820a08,0x8800a000,0x22000000,0xa00828,0x200020a,0x20a2802,0x8808a08,0x8202282,0x80080082,0x8008a20,0x88288208,0x80802802,0x28008820,0x8002,0xa0000002,0x8808228,0x8a00008a,0xa2000,0x20220a20,0xa288,0x20000800,0x20028800,0x80202082,0x22822802,0x20020208,0x200208a,0x80820820,0x20220828,0x8a08020a,0x80802,0xa00a08,0x88282,0xa0022000,0x8220000,0x820a,0xa2882802,0x8200820,0x8a00a288,0x228a2880,0xa08020,0xa28200a,0x20002822,0x8800828,0x82200208,0xa08a0820,0x28800208,0x80082088,0x82802000,0x28808208,0x8a08a200,0xa28a2822,0x8820000,0xa288008,0x880);
remainders_neg[579].modulo_value=2316; remainders_neg[579].bit_mask=create_bit_mask(72,0xa20208a2,0xa28a08,0x2202088,0x22822002,0x28a20800,0x820a280,0x2882880,0x8a20a28,0x82282280,0x80002802,0x8a28a00,0x808020a,0x2020002,0x28208208,0x2200008,0x208228a2,0x8028a20,0xa20200a,0xa0880082,0x28200,0x82088,0xa2880880,0x8208020,0x8a088282,0x82822082,0x200,0x8028a20a,0x800802,0x800228,0x200280,0x82082,0x20800028,0x8808008a,0x20880882,0xa20,0x2088002,0x80822880,0x28a28228,0x88082200,0xa0880080,0x28820008,0x828a00a,0xa20a2002,0x28828a00,0x8a08a002,0xa00a0,0x28a28020,0x2008282,0x80800082,0x8208200,0x8800020a,0x208a2880,0xa28808,0x88080282,0x22002082,0xa08028,0x2082200,0xa2022000,0x8200828,0x8220a082,0xa08820a2,0x8020,0xa288280,0x200200a0,0x20008a00,0x800a000,0x2082080,0x20000a00,0x2002288,0x220220a2,0x28808,0x82200080);
remainders_pos[580].modulo_value=1160; remainders_pos[580].bit_mask=create_bit_mask(37,0x802020a,0x80a8820,0x8022280,0x8202a002,0x22a8aa22,0x2a020a,0xa220082,0x208a8080,0x82020080,0x2008082a,0x202a2080,0x20088a00,0x80a8008,0x8aa2a88a,0xa80828,0x28880208,0x822a0200,0x8080200,0x802020aa,0x80a88200,0x80222800,0x202a0020,0x2a8aa228,0x2a020a2,0xa2200820,0x8a80800,0x20200802,0x8082a8,0x2a20802,0x88a002,0x80a80082,0xaa2a88a0,0xa808288,0x88802080,0x22a02002,0x80802008,0xa0);
remainders_neg[580].modulo_value=1160; remainders_neg[580].bit_mask=create_bit_mask(37,0x808a0a82,0x800200a0,0xaa208,0xa82288a,0xa220222a,0x88a20a82,0x822a8002,0x28020008,0xa828808,0x20808022,0xa8a22808,0xa8000280,0x820080,0x2222002,0x80208020,0xa0000a80,0xa228a80,0x808208,0x8a0a822,0x200a08,0xaa2088,0xa82288a0,0x220222a0,0x8a20a82a,0x22a80028,0x80200088,0xa8288082,0x8080220,0x8a228082,0x8000280a,0x820080a,0x22220020,0x2080200,0xa808,0xa228a80a,0x8082080,0x20);
remainders_pos[581].modulo_value=2324; remainders_pos[581].bit_mask=create_bit_mask(73,0x22882a22,0x88a020,0xa2020202,0xaa2a20aa,0xa0828802,0x82a20a00,0x80228222,0xa8802802,0x828a022,0xa0880802,0x2022,0x220a280,0x8a20200,0xa8a28080,0x80aa800,0x2a88222,0x20280a8,0x8022208a,0x22200a02,0xa020288,0x8aa00a8,0xaa808222,0xa28a800,0x2288008,0x28220002,0x2000000a,0x80882a20,0x28a08200,0xa008aa20,0xa200a00,0x822a0a22,0x8a082802,0x22a2aa00,0x2022aa8,0x28880202,0xa08a2020,0x20a88a22,0x2280808a,0x8080802,0xa882aa88,0xa200aa8,0x88280282,0x8a088a0a,0xa00a00,0xa2808aa2,0x20200820,0x808a82,0x828a0000,0x88080008,0x8a020022,0x2aa002a2,0xa2088820,0xa02a00a,0x88822808,0x80280a00,0x80a2088,0xa802a028,0x2088822,0xa2a002aa,0xa2002028,0x88000808,0x28a0,0x20a88080,0x82080202,0x22a880a2,0x80280280,0xa8288828,0x20a00a08,0x8aa80228,0x8aaa08a,0x20080808,0x288080a2,0x88a82);
remainders_neg[581].modulo_value=2324; remainders_neg[581].bit_mask=create_bit_mask(73,0xaa00a22a,0x80028a0,0x22828a8a,0x2222a822,0x2802008a,0xa2a0288,0x8aa02aa,0x2008a00a,0xa028a2,0x2000808a,0x8808a8aa,0x8aa02a08,0x2a0a88,0x202a0008,0x80822008,0xa200aa2,0x828a0820,0x82aa802,0xaaa0828a,0x828a0a00,0x80228020,0x22080a2a,0x2a02080,0x82a00080,0xa02a888a,0xa8808882,0x80022a8,0xa0280288,0x28802228,0x2a88280,0x2a282aa,0x200a08a,0xaa222288,0x8a0a2220,0xa000828a,0x2802a828,0x282002a2,0xa2080802,0x8000808a,0x20022200,0x82a80220,0xa0820a,0x2800202,0x8288280,0x22080222,0xa82880a8,0x8800020a,0xa020888,0x808080,0x28a882a,0x22288a22,0x228000a8,0x820a2882,0x2a080,0x8a00288,0x8082a000,0x208a2820,0xa8000a2,0x22288a22,0x2a08a8a0,0x808080,0x88882028,0xa8200008,0xa808a0a,0x22200822,0xa08a08,0x202000a0,0xa8208280,0x2200aa0,0x80222002,0xa8808000,0x20080822,0x20a);
remainders_pos[582].modulo_value=2328; remainders_pos[582].bit_mask=create_bit_mask(73,0x2022082,0x8a28a20,0x88080200,0x82880820,0x28020208,0x80288288,0xa00200a2,0x28020200,0x8a20a00a,0x820a2022,0x28a08220,0x8288208,0x20020822,0x28a08200,0x80002000,0xa0800002,0x20220008,0xa28a08a,0x2820,0x20220008,0x8a08008a,0x228a0022,0x8000a28,0x82282282,0x80082020,0x8000820,0x208202,0x2000022,0x20a00028,0x82280282,0x800a0002,0x20a00800,0x88008082,0xa0880020,0x8a20,0x220a200,0x80820082,0x8a08,0xa08a200,0x22080022,0x8820200,0x200200a,0x828000a0,0x8828028,0x8028000a,0x880000,0x20808208,0x2082000,0x82080820,0x20828828,0x8828a000,0xa28800a2,0x8a20020,0x200088,0xa0082800,0x8a20a28,0xa200088,0x2800002,0x28000800,0x800820a,0x20882080,0x28208228,0x8208820a,0xa28808a0,0x28a00a08,0xa008080,0x28a0080,0x28228228,0x82208080,0x22082022,0x20008020,0x8008a28a,0x820880);
remainders_neg[582].modulo_value=2328; remainders_neg[582].bit_mask=create_bit_mask(73,0x828a2802,0x220228,0x8288,0x20000a0,0x20828a00,0x8200200,0x208a0822,0x20828a08,0x2282082,0x28228a2,0x20200a28,0x80200280,0xa08a00a2,0x20200a08,0x808a088,0x20080882,0x28a28800,0x82202002,0x808820a0,0x28a28800,0x2008002,0xa2020822,0x808220,0xa20a20a,0x8028a0,0x808028,0x8828028a,0x828808a2,0x28208820,0xa20820a,0x820882,0x28208008,0x8000a,0x200000a0,0x8800228,0x8a282288,0xa0802,0x8800200,0x82002288,0xa28008a2,0x28a08,0x8a08a082,0x2080820,0x20820,0x8208082,0x80000880,0x28000a00,0x8a00a088,0x28000a0,0x28020020,0x202088,0x22000822,0x228828,0x88288000,0x20802080,0x228220,0x82288000,0x82000882,0x20808008,0x80080282,0xa0002800,0x20a00a20,0xa000282,0x22000020,0x20208200,0x82080008,0x82020800,0x20a20a20,0xa280008,0xa20028a2,0x28800828,0x8002202,0xa0000);
remainders_pos[583].modulo_value=2332; remainders_pos[583].bit_mask=create_bit_mask(73,0x8aa0028a,0xa0020a20,0xa2220aaa,0x802a8a2,0xa000a,0xa088a202,0x2a28820a,0x88880a2a,0xa0a0a808,0x2028a08a,0x288a208a,0x8a220200,0x80000a08,0xa8a8808,0x208282,0x880a2008,0x2022a80,0xa28020a8,0x8a08020a,0x8a82a028,0x2a22800,0x2888a2a8,0x202820a,0xaa2a22a0,0x80820282,0x8020aa82,0x20a80228,0x20280208,0xa2202280,0x808002,0x2220028a,0x2aa0888,0xa8282aa,0x88880020,0x88280000,0xa8020882,0xa8080202,0x28220802,0x20000282,0x800222,0xaa8282a,0x82220aa8,0xa280088,0xa8002020,0x80288088,0x82080280,0x228802a0,0x282aa080,0xa8280820,0xa88a8a,0x8a082808,0x2a8a222,0x200288a8,0x2280a82a,0xaa08020a,0x2a08028,0x202a8808,0x2008a02,0x8282080,0x22022a2a,0x220a0000,0x8008088a,0x8a208a22,0xaa20a280,0x2202a0a0,0x8a8a0222,0xaa08228a,0x808a220,0xa000a00,0xa8a2a802,0xaaaa0888,0x208a0800,0xa2800aa);
remainders_neg[583].modulo_value=2332; remainders_neg[583].bit_mask=create_bit_mask(73,0x2288202,0x280a82a8,0x2aaa8222,0x808a2022,0x8820882,0x20002a8a,0xa2a00a82,0x8222,0x28202080,0x28a02802,0xa002a802,0x2a28288,0x8088280,0x82020080,0x88a80a02,0x822880,0x828aa208,0x2a08a820,0x2808a02,0x20228a0,0xa2aa088,0xa0002a20,0x8a8a0282,0x2222aa28,0xa8a0a,0x8a82202,0xa8200aa0,0xa0a08a80,0x2aa8aa08,0x8808080a,0xaaa08a02,0xa220000,0x820a0a22,0x80a8,0x208888,0x208a800a,0x20808a82,0xa0aa008a,0xa0888a02,0x88088aaa,0x8220a022,0xaa28220,0x2a08800,0x2088a8a8,0x8200000,0xa008a08,0xaa008a28,0xa0a22800,0x20a000a8,0x80200202,0x280a080,0x8a20222a,0xa8820020,0x2a0820a2,0x22808a82,0x8a2800a0,0xa8220080,0xa88028a,0x80a0a800,0xaa8a22a2,0xa2828888,0x8808002,0x2a8022a,0x22a02a08,0x2a822828,0x2028aaa,0x2280a202,0x80002aa8,0x82888288,0x202a2082,0x22220000,0xa0028008,0x2a08822);
remainders_pos[584].modulo_value=584; remainders_pos[584].bit_mask=create_bit_mask(19,0x22a02a22,0xaaa2a00,0x220a8082,0xa2a88a2a,0x2002aa28,0x20880208,0x8280a220,0xa02000a2,0x202a022a,0x80a80882,0x80aa8,0x8a02828a,0x80220808,0xaa800820,0xa22a8a28,0x2a088a8,0xa8aaa082,0xa80a8800,0x88);
remainders_neg[584].modulo_value=584; remainders_neg[584].bit_mask=create_bit_mask(18,0xaa28a2aa,0x8222a288,0xaa82080a,0x2a2002a2,0xa88a22a0,0xa8008a80,0x2082aa8,0x28a8882a,0xa8a28aa2,0x820800a,0x88808220,0x28a0a02,0x8aa8080,0x220880a8,0x2aa202a0,0xa280020,0x2022280a,0x20820088);
remainders_pos[585].modulo_value=780; remainders_pos[585].bit_mask=create_bit_mask(25,0x20000082,0x20028020,0x88208,0x2000022,0xa08,0x88280088,0x800a08a2,0x28000008,0x2008,0x20880022,0xa00208,0x80000202,0x20a0,0xa00808,0x2208208,0x80000880,0x28200,0xa002200,0x28228a2,0x220,0x8020a,0x22000880,0x28008208,0x8080,0x820);
remainders_neg[585].modulo_value=780; remainders_neg[585].bit_mask=create_bit_mask(25,0xa0880802,0x28020820,0x280,0x828008a2,0x800200,0x200000,0x20022,0x20008000,0x8800a080,0xa0000022,0x8200a00,0x8088282,0x802020,0x8208000,0x82280200,0x880000,0x8020a00,0x2082288,0x82022022,0x808a20,0x88000282,0xa2080000,0x20800200,0x8000008,0x20);
remainders_pos[586].modulo_value=2344; remainders_pos[586].bit_mask=create_bit_mask(74,0xaa0aaa2a,0x28a2082,0xa0a20a82,0x2220002a,0x2a0a8a88,0x88282aaa,0x2822008a,0x88a8820a,0xa088802a,0x8000a20a,0x80200228,0x8aa82a0a,0x28280200,0x20a2a22,0x88aaa802,0xaa0a228,0xaa028a82,0x220a2280,0xa220222a,0xa00a8220,0xa28082,0x888220a,0xa2a2a800,0x28222020,0x8aa2a8,0x2a80a028,0xaa8822aa,0x8a00220a,0x8802a88,0x2820a288,0x28808aa2,0xa0200028,0x22288080,0xa002aaa,0x82a280a2,0xa0028a20,0x8002000,0xa2800a00,0x28a8208,0xa800a08a,0x22888aa,0x80a02,0xa2022828,0x8a08288a,0xa8022022,0x8800a222,0x8822aaa0,0xa02a8aa,0x8aa20028,0x888282a,0x2a8a8a08,0x88222000,0x28a00a0,0x82a00a82,0x88088a08,0x88a088a8,0xa280aa02,0x8a0aa082,0x2aaa2228,0xa8a08080,0x80282888,0xa82aa200,0x800802a0,0x8a000228,0x2220aa0,0x822a22a8,0x8828a0,0xa82822a2,0xa2a0a8aa,0x88822,0xa08a0aa8,0x8a28082,0xaaa0aa82,0xa8);
remainders_neg[586].modulo_value=2344; remainders_neg[586].bit_mask=create_bit_mask(74,0x228222a2,0x8a02a80a,0x282a820a,0xaaa888a2,0xa2820200,0xa0a222,0xa0aa8802,0x200a82,0x280008a2,0x8882a82,0x8a88aa0,0x220a282,0xa0a08a88,0x8a82a2aa,0x22208a,0x82282aa0,0x228a020a,0xaa82aa08,0x2aa8aaa2,0x28820aa8,0x882a080a,0x8000aa82,0x2a2a2088,0xa0aaa8a8,0x88022a20,0xa20828a0,0x2200aa22,0x2882a82,0x8008a200,0xa0a82a00,0xa008022a,0x28a888a0,0xaaa00808,0x8288a222,0xa2a082a,0x288a02a8,0x8088a888,0x2a088288,0x8a020a80,0x20882802,0x8aa00022,0x8880828a,0x2a8aa0a0,0x280a002,0x208aa8aa,0x882aaa,0xaa2228,0x828a2022,0x22a88a0,0x8000a0a2,0xa2020280,0xaaa888,0x8a028828,0xa28820a,0x800280,0x280020,0x2a08228a,0x282280a,0xa222aaa0,0x20280808,0x8a0a000,0x20a22a88,0x8808a28,0x2888aa0,0x8aaa8220,0xaa2aa20,0x8800a028,0x20a0aa2a,0x2a282022,0x888000aa,0x28028220,0x802a080a,0x2228220a,0x20);
remainders_pos[587].modulo_value=2348; remainders_pos[587].bit_mask=create_bit_mask(74,0x22aa8a02,0x220088,0x2a020282,0x8a0288a8,0xa202a8a,0x2aa20a20,0x282a222,0xaa00aaaa,0xa208a0aa,0x8a0a20aa,0xa2008a88,0x80808220,0xa20aa882,0xa282a20,0xa8a8a220,0xa822aa,0xa222880,0xaa00828a,0x802802a0,0x82820a2a,0x2aaa2088,0x200008a0,0x82a88820,0x882880a0,0x2a0a2002,0x88288a2a,0x8222022a,0x88002882,0x80002822,0xa80002a,0x88880a,0x82a88288,0x820200a8,0x80a0220a,0xaa8a0a8a,0x8a22a88,0x2a0a8200,0x88a20008,0xa2aa22a,0x80a02a2a,0x8082a08,0x22a822a0,0x22200228,0x2a0a02,0x80002a80,0x80022882,0x8882822,0x22822a88,0x8a0a8a8a,0x22822800,0x22a820a0,0x8082,0x8aaa80a2,0x8282220,0x2802a8a,0x200aa0a8,0x888a0a28,0x82a00022,0xa2a2aaa8,0x828a0088,0xaa08a08a,0x20202822,0x2008a088,0x8a0a222a,0xa208aaa0,0xa00aaaa0,0xa8280aaa,0x8aa8888,0x808a008a,0x280a2a2a,0x80a82a2,0x8800828,0x2aa88220,0x80a);
remainders_neg[587].modulo_value=2348; remainders_neg[587].bit_mask=create_bit_mask(74,0xaa22028a,0x88aa8800,0xa28a8a0a,0x28a0020,0x82a8a202,0xa22a82a8,0x8a0a2aaa,0x22882222,0x2a802822,0x282a822,0x2a880200,0x8080aa8,0x2a82200a,0x82a0a2a8,0x20202aa8,0x8820aa22,0x82aaa008,0x22880a02,0x8a08228,0xa0a82a2,0xa222a800,0xa8888028,0xa2000a8,0xa00828,0xa282a88a,0xa002a2,0xaaa8aa2,0x88a00a,0x888a0aa,0x820888a2,0x88000082,0xa200a00,0xa8a8820,0x828aa82,0x22028202,0x802aa200,0xa2820a88,0x2a8880,0x82a22aa2,0x828a2a2,0x8080a280,0xaa20aa28,0xaaa88aa0,0x88a2828a,0x888a208,0x88aa00a,0x8000a0aa,0xaa0aa200,0x2820202,0xaa0aa088,0xaa20a828,0x8888080a,0x222082a,0x80a0aaa8,0x8a08a202,0xa8822820,0x282a0,0xa2888aa,0x2a2a2220,0xa028800,0x22802802,0xa8a8a0aa,0xa8802800,0x282aaa2,0x2a802228,0x28822228,0x20a08222,0x80220000,0x8028802,0xa082a2a2,0x80820a2a,0x800880a0,0xa2200aa8,0x82);
remainders_pos[588].modulo_value=168; remainders_pos[588].bit_mask=create_bit_mask(6,0x2802802,0x28008020,0x80280280,0x2002802,0x28028008,0x80);
remainders_neg[588].modulo_value=168; remainders_neg[588].bit_mask=create_bit_mask(6,0x82082002,0x20800820,0x208208,0x82082082,0x20820800,0x8);
remainders_pos[589].modulo_value=2356; remainders_pos[589].bit_mask=create_bit_mask(74,0x2a20a2aa,0x8022a0a8,0xa828802,0xa2208220,0xa0aa2088,0xaa288288,0x80000808,0x202a2a02,0x8022a28,0x8808002,0xa0822a20,0xa2880880,0x2222a80,0xaa0a0088,0xa28800aa,0x2aa00282,0xa2022800,0x22820a00,0x22020202,0x2802820a,0xa000a202,0x2a0a002a,0xaaa8008a,0x880282,0x280aa222,0x2808808a,0x8022a208,0x82000808,0x20a2a200,0xa02a2a0,0xa8808000,0x288a08a2,0x288822a8,0x80220822,0xa008a0a,0xa0a82a20,0xaaaa2822,0xa0a8828a,0xa008a82,0x802a0a20,0x22888208,0x2282a882,0x22a8a20a,0xa000020,0xa080a8a8,0x82008a8,0x80220200,0x28208a8,0x28a2022,0x200888aa,0xaaa82802,0xa8a2002,0xaa800a,0x28808a0,0x88a0828,0x28880808,0x8a00a08,0xaa800288,0x28a82800,0xaaaa002,0x8802200a,0x28a02a88,0x20a02202,0x22008a88,0x2080020,0x80828a88,0x280a8a,0x8aa20200,0xa0a22822,0x88a2208a,0x2a008820,0x80280228,0x8a82a0a8,0xaa8a0);
remainders_neg[589].modulo_value=2356; remainders_neg[589].bit_mask=create_bit_mask(74,0x22a02a22,0x8aa2820,0x20a008a,0x2aa80aa8,0x2022a800,0x22a00200,0x8888080,0xa822a28a,0x800aa2a0,0x8008088a,0x280aa2a0,0x2a008008,0x8aa22208,0x22828800,0x22008822,0xa2288a0a,0x2a8a2088,0xaa0a8208,0xaa8a8a8a,0xa0820a82,0x2888228a,0x228288a2,0x22208802,0x88008a0a,0xa082222a,0xa0800802,0x82a2a80,0xa888080,0xa82a2a80,0x828a2a20,0x20080888,0x2002002a,0xa000aa20,0xaa80aa,0x82880282,0x2020a2a8,0x2222a02a,0x28200a02,0x8280020a,0x82282a8,0x2a000a80,0xaa0a200a,0xaa202a82,0x828080a8,0x28082020,0x80288020,0x8aa8a88,0x8a0a0020,0x8a02a8a2,0xa8800022,0x2220208a,0x8202a08a,0x80220882,0x8a008028,0x800280a0,0xa0008000,0x80288280,0x22008a00,0xa020a088,0x2222882,0x8aa882,0xa028a200,0x2828a28a,0xaa880200,0x8a0088a8,0x80a0200,0x80a08202,0x22a8a80,0x282aa0aa,0x2a2802,0xa20800a8,0xa08aa0,0x20a2820,0x22028);
remainders_pos[590].modulo_value=2360; remainders_pos[590].bit_mask=create_bit_mask(74,0x80080282,0xa02a0a82,0x8a0228,0x8202a022,0xa0a820,0x88000a80,0xa2200020,0x280222a8,0x8aa22880,0xa2202a22,0xa8820802,0x8a2a0222,0x8800288,0xa008820,0x20a80a02,0xa0082a8a,0xa20202a0,0x882020,0x88202828,0x228280a,0x8022082,0x200008a2,0x2280a008,0xa2088a0,0x8080aa08,0x222000,0x2288280,0x2002220,0xa802000,0x2228,0x882a2a02,0xaa0082a0,0xa080208,0x822200a2,0x20a00a,0x80802008,0x820288a2,0x28a2280,0x200802,0x82a00a08,0xa08a0088,0xaa208020,0x220a8200,0x80a8a8,0xa0288800,0x80000802,0x80088800,0x28228,0x2000888,0xa020aa02,0x880a2208,0x8200a02,0x208a2000,0x80820880,0x22a02828,0x282808,0x8a080822,0xa0a8080,0x8a2a820,0xa080a02a,0x20082200,0xa2228002,0x2a8880a8,0x8a802082,0xa288a808,0x2802288a,0x8a2a8880,0x22080008,0x2a000,0x82082a0a,0x880a80,0xa2880a2,0x282a0a8,0x828020);
remainders_neg[590].modulo_value=2360; remainders_neg[590].bit_mask=create_bit_mask(74,0x8800a0a,0x20228202,0x80282a0,0xa8228aa,0x80282028,0x880208,0x2a2888a0,0x208aa220,0x222a008,0x22a8a22a,0x200a008a,0x2228aa2,0x88200,0x828000a8,0xa020820a,0x28802202,0x2a028a20,0x800a0a8,0xa0a0a0,0x82a0a002,0x808a280a,0xa8088022,0x2a082080,0x2a00028,0x82200,0x88aa2888,0x8a200a00,0xa88a2a8,0x8200a888,0x8088aa20,0xa2228a,0x22080220,0x2808280,0xaa2882a,0x80a82802,0x8082880,0xa0a0022,0xa02a208,0x8820808a,0x2288200,0x28020800,0x222808a0,0x2a820288,0x88002020,0x20a00008,0x888008a,0x8000080,0x88a02a0,0x8a808000,0x20a8220a,0x822a80,0x80288282,0x2802a088,0x8020008,0xa228a020,0x88a02080,0x20080a2,0x2820008,0x802220a8,0x20082822,0xa0802a88,0x2a2a0882,0x22000020,0x200a80a,0x22002000,0xa08a2002,0x2220000,0x2a808080,0x88822808,0x280a202,0x88000208,0x82200822,0xa0a2020,0x208a8);
remainders_pos[591].modulo_value=2364; remainders_pos[591].bit_mask=create_bit_mask(74,0x82820022,0x28028020,0xa28a008,0x220a2882,0x28a08820,0x82200200,0xa0882,0x8a08800,0x8a00a000,0xa28a2082,0x8008000,0x2088200,0x820008a2,0x28828208,0x8200a288,0xa0820820,0xa28808,0x8828808a,0x22020880,0x220020,0x220000a,0x20820820,0x20000a08,0x2080200,0x80022882,0x28820208,0xa282282,0x2080000,0x28a00a00,0x8a202002,0x200a2,0x28828808,0x88202000,0x82000080,0x20a00000,0x8280280,0x8a0802,0x8020a20,0x80280282,0x200000a0,0x20200008,0x80008082,0x2022822,0x28a00800,0x8800808a,0xa00a2,0x208,0x8828828a,0x22080822,0x8228800,0x80080208,0x820a0000,0x820820,0xa000088,0x80800880,0x20220808,0xa202282,0xa20228a0,0x20820820,0x8228a008,0x22082822,0x8a20008,0x82208,0xa0002002,0x28208a28,0xa00a,0x220a2,0x28220a00,0x80080088,0x808220a2,0x8228a08,0x8200a28a,0x20802802,0x8800828);
remainders_neg[591].modulo_value=2364; remainders_neg[591].bit_mask=create_bit_mask(74,0x20a08a2,0x20820828,0x82202080,0xa2822002,0x20200028,0xa288288,0x80820002,0x200008,0x2082088,0x22022802,0x800808,0x8a000288,0x2880022,0x20020a00,0xa082200,0x200a00a0,0x8220000,0x200002,0xa28a0000,0x8a28828,0x8a288082,0xa00a00a0,0x28808200,0x8a008288,0x8a2002,0x20028a00,0x8220a20a,0x82800880,0x20208208,0x228a08a,0x808a0822,0x20020000,0x28a088,0x2880800,0x28208808,0x80208208,0x80020082,0x828228,0x820820a,0xa0880820,0x28a08800,0x808000a,0x828a20a2,0x20208000,0x80002,0x80820822,0x8808a00,0x200202,0xa28000a2,0xa20008,0x8008280,0x2820880,0x8028028,0x82088000,0x80000,0x28a28000,0x8228a20a,0x228a2020,0x28028028,0xa202080,0xa28020a2,0x228800,0x8800a280,0x20882882,0x20a00220,0x88082082,0x808a2822,0x20228208,0x8008000,0xa2822,0xa20200,0xa082202,0xa0082082,0x8020);
remainders_pos[592].modulo_value=296; remainders_pos[592].bit_mask=create_bit_mask(10,0xa200a8a,0x80228202,0x802a2a8a,0x8a080828,0xa8aa2a00,0x2020a200,0xa8a80228,0x828802a2,0xa008a080,0xa2);
remainders_neg[592].modulo_value=296; remainders_neg[592].bit_mask=create_bit_mask(10,0x82a88202,0x8aa0a8a,0x8a2a202,0x28000a0,0x2022a288,0xa8a82a88,0x20208aa0,0xa008a2a,0x28802800,0x2a);
remainders_pos[593].modulo_value=593; remainders_pos[593].bit_mask=create_bit_mask(19,0xe2af8316,0x5ca644ff,0x181af7d7,0x1b526436,0x7715f997,0xa1c80b4c,0x1498a716,0x8945192e,0xf7e16937,0x1fbf2313,0x8a47b25a,0x64a1d262,0x4e15a394,0xa3b8cb40,0x2b63a67e,0x6061b099,0x94ebafbd,0xd51ffc89,0x1a307);
remainders_neg[593].modulo_value=2372; remainders_neg[593].bit_mask=create_bit_mask(74,0x2a220a8a,0x802a8822,0x80822a0a,0x828aa8aa,0xaa88200a,0x28008280,0x88002a8a,0x8880a2,0x2a28a0aa,0x8222aa8a,0x28a2a82,0xa8280aaa,0x82882a08,0x2a200288,0xa2aa2aa2,0xa8a82800,0x8222220,0x8822000,0x8aa02a8a,0x220882,0x22222080,0xa0a8a8,0x2aa2aa28,0x8a0022a2,0x82a08a08,0xaa80a0a8,0xaa28a02,0x8aaa220a,0xa828a2a2,0x28088802,0x8aa0008a,0xa0800a2,0x80208aa8,0xa8aa8a0a,0x82a2080a,0x208aa00a,0x8a8222a2,0x2228082,0xaa008a22,0xa8a20280,0xa0800800,0x8a2a80,0xaa8a0a8,0x8aaa0082,0xa888a820,0x2082802,0xa2200082,0xa08200a0,0xa0a8002,0xa08a0288,0x22aa088,0x20020022,0x80a0aa8,0x8a222228,0x88a22aaa,0x80220082,0xaa2288a0,0x222228aa,0xa828080a,0x20020a,0x82aa2022,0xa0288288,0xa82808,0x802082a0,0x80022282,0xa082020,0xa888aa0,0x802aa882,0x828aa820,0xaa28800a,0x8008280,0xa0228a80,0x28802a80,0x80a22022);
remainders_pos[594].modulo_value=264; remainders_pos[594].bit_mask=create_bit_mask(9,0x820a2022,0x28220a00,0x80200002,0x20002082,0x8000800,0x80282,0xa0882880,0x8a08200,0x88);
remainders_neg[594].modulo_value=264; remainders_neg[594].bit_mask=create_bit_mask(8,0x28220a2,0x20228208,0x828808a,0xa0882802,0x800008,0x8000820a,0x20002000,0x200a08);
remainders_pos[595].modulo_value=2380; remainders_pos[595].bit_mask=create_bit_mask(75,0x8082a0a,0x2a000aa2,0x28088,0x8a228008,0x20208002,0xa20a80,0x2a0222,0x88a2208,0xa20a082,0x20000220,0x800a2000,0x28028a02,0x800028,0x8802a020,0x2808008,0x8a80002,0x802a0202,0xa008020,0x8a02800a,0x82080200,0x88002208,0xa8000202,0x22080280,0x80028022,0x22282828,0x80022080,0x88020a00,0x88aa0a0,0x8008008,0x802202,0x80800882,0x2080080,0x2a008088,0x8a000008,0x2008a222,0x820002,0x8280,0x20000000,0x8200028,0xa2008800,0xa2888,0x200a8200,0x2080220,0x202020,0x80200822,0x20020808,0xaa220200,0x80220a0,0x8800200a,0x82888020,0x28002282,0x2088880,0x2a028,0x80022808,0x2082208,0x280a2008,0x200a0a00,0xa802080,0x2a20808,0x20280800,0xa8022200,0x200080,0x28028280,0x8a00280a,0x8000,0xa08a0088,0x8a220820,0xa800208,0x8a00888,0x2080802a,0x288a2800,0x28000200,0xa8220,0x820208aa,0xa0a);
remainders_neg[595].modulo_value=2380; remainders_neg[595].bit_mask=create_bit_mask(75,0x80802202,0x22008222,0x8a0000,0x2220880,0xa0a0080a,0x802a0208,0x882282a2,0x2a200,0x2a02802,0xa0888a28,0x8022880,0xa0020282,0x8080020,0x8220a8,0x82080000,0x200882,0x220a82,0x20800a8,0x2820882,0x2800a08,0x82280,0x20088a02,0x22808200,0x8208aa,0x2220a020,0x8822808,0xa0280,0x22020,0x80800800,0x8aa02,0x8000a,0x8a008000,0x22800000,0x2800880,0xa080222a,0x8802080a,0x8080a00,0x20088088,0x802088a0,0x22800008,0x88822000,0x28020280,0xa808228,0x8a0a8a0,0xa88022,0xa80a0080,0x22228a00,0x8a2028,0x80a082,0x2000828,0x20882a02,0x82000000,0x80220a0,0x882a080,0x82002a00,0xa0822080,0xa0028200,0x208a000,0x2228080,0xa0208008,0x20822a88,0x88280800,0x208a0208,0x280a002,0x880800,0x20020000,0x22a80a0,0x2000280,0x80200000,0xa0000022,0xa0022008,0x20088a80,0x8202a8,0x2028022,0x202);
remainders_pos[596].modulo_value=1192; remainders_pos[596].bit_mask=create_bit_mask(38,0xa20a02a2,0xa022a0aa,0x80220228,0xaa820880,0xa028022,0x820aaa20,0x2200808,0x2a2028a2,0x2822aa8,0x280a8a2a,0x8a82aa88,0x8808a280,0x8220200,0xa008aaa,0x2aa88028,0x80202208,0x80a28808,0x8aaa0a8,0x2a28a80a,0xaaa20a0,0x228a022a,0x88080220,0x22aa820,0xa200a028,0x808820aa,0x8a202200,0xaa82a202,0xa2a02822,0xa88280a8,0x2808a82a,0x2008808a,0xaaa08220,0x280a008,0x2082aa88,0x80880202,0xa880a28,0x80a08aaa,0x8a);
remainders_neg[596].modulo_value=1192; remainders_neg[596].bit_mask=create_bit_mask(38,0x2a828a2a,0x28aa2822,0x8aa8aa0,0x220a8008,0x828a08aa,0xa8222a8,0x8aa88080,0xa2a8a02a,0x8a0aa220,0xa08202a2,0x20a2200,0x802a08,0x80aa8a88,0x2880222,0xa22008a0,0x8a8aa80,0x82a0080,0x80222820,0xa2a02082,0x8222a828,0xaa028aa2,0x808aa8,0x8aa220a8,0x2a8828a0,0x800a822,0x2a8aa88,0x220a2a8a,0x2a28a0aa,0x200a0820,0xa08020a2,0xa8800802,0x22280aa8,0x8a002880,0xa80a2200,0x8008a8a,0x820082a0,0x8280222,0x2);
remainders_pos[597].modulo_value=2388; remainders_pos[597].bit_mask=create_bit_mask(75,0x82022882,0x28020a00,0x8808a080,0x28a08a0,0x20a28a00,0x8a08a280,0x22000022,0xa00,0x82088200,0x22002002,0x8a08820,0x8002202,0x200a0020,0x20220a28,0x2088008,0x8208a0,0x8228800,0x200a282,0x82020820,0x28020020,0x8a20820a,0x82080000,0x8820028,0x82202080,0x28020a2,0x20008020,0x88288202,0x20022020,0x8820a00,0x80000002,0x22002,0x28828a20,0x8a28200a,0x828a0002,0x28808828,0x8200a008,0xa2020802,0x8a20a08,0x8280208,0x228200a0,0x28228220,0x8a28000a,0x228a0082,0x8a28,0x280088,0x22080000,0x800a08,0x82208088,0x880822,0x28008020,0x8828a080,0x22002080,0x8228008,0x8a200002,0x28a0820,0x8208008,0x8008208,0x820828a0,0x20000228,0x800a208,0x80820022,0x828a08,0x200800a,0xa2080880,0x8808220,0x8280080,0x822,0x8800a00,0xa288000,0xa08028a2,0x28000a28,0x220a20a,0x28020a2,0x8200a08,0x82288);
remainders_neg[597].modulo_value=2388; remainders_neg[597].bit_mask=create_bit_mask(75,0x28a2002,0x20828208,0x2008,0x82020020,0x28220208,0x2002208,0xa2880822,0x8808208,0xa000288,0xa2882882,0x200028,0x8008a28a,0xa08208a0,0x28a28220,0x8a000080,0x800a0020,0xa20008,0x8a08220a,0x28a00a0,0x20828828,0x2280282,0x2800880,0x28820,0xa28a008,0x82082822,0x28800828,0x20028a,0xa08a28a0,0x28208,0x808808a,0x808a2882,0x20020220,0x220a082,0x2020882,0x20000020,0xa082080,0x228a0082,0x228200,0x80208280,0xa20a0820,0x20a20a28,0x2208082,0xa2020802,0x8800220,0x88208000,0xa2800880,0x8008200,0xa280000,0x800000a2,0x20800828,0x202008,0xa2882800,0xa20800,0x228808a,0x820200a0,0xa00800,0x80080280,0x2802020,0x28808a20,0x80082280,0xa08a2,0x8020200,0x8a080082,0x22800000,0xa28,0x80208008,0x808800a2,0x8208,0x82200088,0x20082022,0x20808220,0x8a282282,0x82082822,0xa08200,0xa200);
remainders_pos[598].modulo_value=2392; remainders_pos[598].bit_mask=create_bit_mask(75,0x8a080a0a,0x2222800a,0x222a8888,0x280028a,0xa08a2a08,0x82a80880,0x80aa2aa0,0x820aa20a,0x8288888a,0x2820228a,0x28200000,0x820aa20,0x28022222,0xa080082,0x2880a220,0xa2a02880,0xaa800022,0x2020082a,0xa20a080,0x8a2a0a82,0xa2000028,0x2a00022a,0xa882088,0x2820aa02,0x20a8228a,0xa20a202,0xaa200a20,0x2a282a2a,0x80828080,0x820088,0xa00a2,0xa8028800,0x8a0280,0xa802a8a0,0x80202082,0x82222008,0x2880808a,0x280a00a0,0x82a20202,0x2200888,0x2a820808,0xa2a80,0x2a0280a2,0x2280,0x8a00a0,0x2220082,0xa8020282,0xaaa8a828,0xa008a008,0x8808a08,0x28a2882a,0xa080aa08,0xa8220822,0x8aa88000,0xa2280000,0xa082a0a8,0x8020a08,0xaaa82008,0x8a880002,0x2802280a,0xa0088a02,0x28820020,0x20888880,0x2808aa08,0x28000008,0x82a28808,0x82a22222,0x2a08aa0,0x820aa8aa,0xa02202a,0x8020a8a2,0x88a28002,0x882222a8,0xa2a00288,0xa0a020);
remainders_neg[598].modulo_value=2392; remainders_neg[598].bit_mask=create_bit_mask(75,0x2008282,0xaaaa0802,0xa2a20000,0x8a008a02,0x28022280,0xa208008,0x8222220,0xa022a82,0xa000002,0xa0a8a202,0x20a88888,0x80a822a8,0xa082aa2a,0x8280880a,0xa0082a28,0x2228a000,0x220888aa,0x28a800a2,0x82a82808,0x2a28202,0x220888a0,0xa2888aa2,0x8200a000,0x2028228a,0xa820aa02,0x82a02a8a,0x22a082a8,0xa2a0a222,0xa0808,0x880a0800,0x8882082a,0x208a0088,0x88028200,0x200a2028,0x8a8a80a,0xaaaa000,0x20080802,0xa0828828,0xa228a82,0x8aa88000,0x220a8000,0x8082a208,0xa28a082a,0x80882a08,0x88028828,0x8aaa8802,0x200a8a0a,0x222020a0,0x28802080,0x280,0xa02a00a2,0x28002280,0x20aa00aa,0x2200808,0x22a08888,0x280a2020,0x808a0280,0x2220a880,0x2008802,0xa00aa082,0x2880028a,0xa00a80a0,0x28000008,0x20802280,0xa0808880,0xa2a0080,0x22aaa2a,0x82280228,0xa822022,0x820a28a2,0x8a8202a,0x2a0882,0x22aa20,0x2a288a00,0x2820a8);
remainders_pos[599].modulo_value=2396; remainders_pos[599].bit_mask=create_bit_mask(75,0x82822aa2,0x8aa22a28,0x8a2a002,0x22a80808,0xa00a008,0x8aa08282,0xaaa0a280,0x22208202,0xa088aa08,0xaa0a88a,0xaa28a0,0xa02a0280,0x28aaa8aa,0x8a88022a,0xa280a0a8,0x888a02,0x2aaa8882,0x8088a288,0x22088a8a,0x222a0202,0xa208,0x222a8a22,0x2a080a82,0xa2022028,0x20808a,0x80a80020,0x802aaa0a,0x2802a820,0x22280220,0x2888a280,0x28000a8a,0x28022a08,0x2a828a0a,0xa008a2a8,0x822a2a2,0x88020aa8,0x8a0a2080,0x20020800,0x20208a0a,0x2aa0802,0xa8a8a882,0x82a8a200,0x8a0a282a,0x820a8802,0x8a2a0002,0x8028a222,0x80880288,0x2082a802,0x2a0aaa80,0x8002a0,0xaa202080,0x82808808,0x882a020a,0x88a2a88,0x8208a000,0x88080a88,0x2a2a2208,0x8228a220,0x8222aaa,0xa80a2220,0x22a0a028,0x8a88022a,0xaaa2aaa2,0x280a80,0xa28aa0,0xaa22a0aa,0x820aa220,0xa8082088,0x2028a0aa,0x82820aa,0x8200a00a,0x20202a8,0x2800a8a2,0x228a88aa,0x8aa8828);
remainders_neg[599].modulo_value=2396; remainders_neg[599].bit_mask=create_bit_mask(75,0xa0aa22a,0x22aa2a0,0x802a288a,0xaa208080,0x82882880,0x2280a0a,0x22282a08,0xaaa80a8a,0x28002280,0x82282002,0x8822a028,0x28a28a08,0xa0222022,0x2008aa2,0x2a082820,0x8800028a,0xa222000a,0x8002a00,0xaa000202,0xaaa28a8a,0x88882a80,0xaaa202aa,0xa280820a,0x2a8aa8a0,0x88a80802,0x82088a8,0x8a22282,0xa08a20a8,0xaaa08aa8,0xa0002a08,0xa0888202,0xa08aa280,0xa20a0282,0x28802a20,0x80aa2a2a,0x8a8220,0x282a808,0xa88a8088,0xa8a80282,0x8a22808a,0x2020200a,0xa202a88,0x282a0a2,0xa82008a,0x2a2888a,0x8a02aaa,0x8008a00,0xa80a208a,0xa2822208,0x88088a28,0x22a8a808,0xa080080,0xa28a82,0x8002a200,0xa802888,0x808200,0xa2a2aa80,0xaa02aa8,0x80aaa222,0x2082aaa8,0xaa2828a0,0x2008aa2,0x222a222a,0x88a08208,0x882a0228,0x22aa2822,0xa822aa8,0x2080a800,0xa8a02822,0x80a0a822,0xa882882,0x8a8a8a20,0xa088202a,0xaa020022,0x2200a0);
remainders_pos[600].modulo_value=120; remainders_pos[600].bit_mask=create_bit_mask(4,0x20880002,0x228800,0x8002288,0x800022);
remainders_neg[600].modulo_value=120; remainders_neg[600].bit_mask=create_bit_mask(4,0xa0000882,0x8220000,0x8a200,0x8a2);
remainders_pos[601].modulo_value=601; remainders_pos[601].bit_mask=create_bit_mask(19,0x4f95b77e,0x3057e1b1,0x56036d2b,0x27aa9135,0x34f92ced,0x1b9c0285,0x430b8fb9,0x43f4ece,0x86705cd1,0x871afd63,0x822ce839,0x9cdcbf0,0x6277c743,0xb28500e7,0x92dcd27c,0xaab22557,0x3352db01,0xca361fa8,0x1fbb6a7);
remainders_neg[601].modulo_value=2404; remainders_neg[601].bit_mask=create_bit_mask(76,0x82082aa2,0xa88a2828,0x8a8aa0a2,0xaa2208a8,0xa820a020,0x82008a08,0x8a820220,0x88a28202,0xaa88008,0xa8220aa,0xaa020a0,0x80000228,0xaaaa0a8a,0x2a08882a,0xa000aa0,0x222aa88a,0xaa8a0288,0x2aa8220,0x20222a2a,0xaa02a2a2,0x8aa8220a,0xa2208a02,0x28288aa,0x82a02a80,0xaaaaa088,0xa8a82,0x2a80a200,0xa802820,0xaa82a822,0x28888008,0xa8a020a,0x2082202,0x20a88288,0x22a82028,0x8a88a882,0x88aa282a,0x8208a0a2,0x222aa0,0xa08a28a,0x28200882,0x88088080,0x2a280222,0x82880a28,0xa228a2a8,0xa2a080a2,0xaa888820,0x28028008,0x282a80a2,0xa20a802a,0x8080aaaa,0x8a000002,0x802a0288,0x88282aa,0xa08a2200,0xa2280082,0x202a000,0x20222222,0x228002a0,0x82a0800a,0x88002228,0xaaa0a0,0x2888a02a,0x80a00000,0x22aaaa80,0xa2a00a8,0xa2280aa,0xaa8800a0,0xa2820888,0x22a28082,0xa08aa28a,0x2a0a2808,0x8a2200a,0xa008088,0x28208802,0x28a288,0x2);
remainders_pos[602].modulo_value=2408; remainders_pos[602].bit_mask=create_bit_mask(76,0x2280a202,0xa880022,0xa22a8200,0x8a200020,0x88202880,0x8a82208,0x802aa28a,0x8028822,0xa202222,0x22a0a2a,0xa2aa8aa,0x82a00288,0xa022220,0xaa020,0x802828,0x20a888a2,0x8a822a,0xaa280802,0x8282a082,0x20a00,0x88a2a088,0xa002200,0x2a00880,0xa2a8082a,0xa02a2800,0x2802202,0xa802a8,0x20a00000,0x20802020,0x28a02222,0x2820882,0x8208000,0x2880a208,0x820828,0x200a22aa,0x808aa200,0x22028a0,0x8282820a,0x280880a0,0x8aa2020a,0x88a00800,0x208200aa,0x8a022828,0x2082020,0x20828000,0x880a2882,0x8020888,0xa0808,0x802a0000,0x8802802a,0x28a80a80,0x202a8a00,0x200a80a8,0x8800a002,0xa8a2200,0xa0800022,0xa828200,0x2028aa82,0x82a20080,0x222a08a8,0x2802008a,0xaa00028,0x8880a008,0x800a8208,0x2aa8a022,0xa0a880aa,0x8808a0a8,0x22802088,0x8aa80288,0x882a20a2,0x28082220,0x8a202,0x82a88a08,0x22a000,0x8a028888,0x80);
remainders_neg[602].modulo_value=2408; remainders_neg[602].bit_mask=create_bit_mask(76,0xaa082a0a,0x20080a2,0x22a20a88,0x22888a8,0xa0a008,0x80202a80,0x8222202,0x808a002a,0x2a8aaa2,0x82a282a2,0x82222022,0xa208a00,0x828a2aa0,0x888220a8,0x8808a020,0x28200022,0x80020aa2,0x2220008a,0xa02280a,0x888a0288,0x2a2000,0x8288aa08,0x2288000,0x222080a2,0x2822a088,0x8a00aa8a,0x88200a20,0xa8288088,0xa808a828,0x2028aaa2,0x820a800a,0x80280888,0xa0002a80,0x880200a0,0xa882a222,0x8022a08,0xaa8a020,0x20a0a82,0x20000828,0x2228a82,0x280088,0xa80a8022,0x28aa020,0xa80a8a0,0xa00a0088,0x2a00a,0x80828000,0x88820080,0x8a28088,0xa0822,0x20208200,0xa0a20288,0xa8020820,0x80288a,0x82022a88,0x280880a2,0x820a0a08,0x28a02202,0x22a8808,0xaa228020,0xa0820802,0x822808a0,0x82080,0x8820a00,0x222028a2,0x20200822,0x2820,0xaa00a800,0x2200a00,0xa2a02a,0xa080aa28,0x8802a02,0x2200280,0x882a2888,0x2820000,0x8);
remainders_pos[603].modulo_value=804; remainders_pos[603].bit_mask=create_bit_mask(26,0xa2020882,0x20820,0x8200a202,0x2880000,0x20200800,0x8828a200,0x2002802,0x8828800,0x28a20a,0xa08228a2,0x28a00800,0x88008208,0x80802020,0x8200220,0x200a282,0xa28820a0,0x8a28008,0x228220a,0x2800800,0x8a28228,0x2008080,0x22800,0x8a00820,0x82080008,0x220808a0,0x8);
remainders_neg[603].modulo_value=804; remainders_neg[603].bit_mask=create_bit_mask(25,0x228a0002,0x8828028,0xa082282,0x82000880,0x28a08008,0x202288,0x82882082,0x20008,0x88202282,0x200a2022,0x20200008,0x80280,0x828a0,0xa08a28,0x8a08220a,0x22002820,0x220800,0x8a20a282,0x82080080,0x220a20,0x8a080008,0x808a2080,0x208028,0xa008080,0xa2800020);
remainders_pos[604].modulo_value=1208; remainders_pos[604].bit_mask=create_bit_mask(38,0x2aa282aa,0x800a2028,0xaa88a8,0x22888a02,0x20a8288,0x2208a20a,0x2a8a82,0x82aa2022,0x800a0a8,0xa00000a,0x22a0a002,0x8808aa8,0x882a2a80,0xa08a208,0x82282a08,0x80a2228,0x22a22aa0,0x82808a00,0xaaa828aa,0x82aa282a,0x8800a202,0x200aa88a,0x822888a0,0xa020a828,0x22208a20,0x2002a8a8,0x882aa202,0xa0800a0a,0x20a00000,0x822a0a00,0x8808aa,0x8882a2a8,0x80a08a20,0x882282a0,0x80a222,0x22a22aa,0xa82808a0,0xaa828a);
remainders_neg[604].modulo_value=1208; remainders_neg[604].bit_mask=create_bit_mask(38,0xa22a0a22,0x882a8a0,0x88220020,0xaa00028a,0x8a020a00,0xaa802a82,0x88a2020a,0xa22a8aa,0x80882820,0x82888882,0xaa28288a,0x80080220,0xa2a208,0x82802a80,0xaa0a280,0x8082aaa0,0xaa2aa228,0xa080288,0x2220a022,0xa22a0a2,0x882a8a,0xa8822002,0xaa00028,0x28a020a0,0xaaa802a8,0xa88a2020,0xa22a8a,0x28088282,0xa8288888,0xaa28288,0x88008022,0xa2a20,0x82802a8,0xaa0a28,0x88082aaa,0x8aa2aa22,0x20a08028,0x220a02);
remainders_pos[605].modulo_value=220; remainders_pos[605].bit_mask=create_bit_mask(7,0xa0280202,0x280a0280,0xa0280a0,0x80a020,0x80a0280a,0xa0280a02,0x8080280);
remainders_neg[605].modulo_value=220; remainders_neg[605].bit_mask=create_bit_mask(7,0x28a0028a,0xa0028a00,0x28a0028,0x880028a0,0x282002,0x20a0028a,0x8a00);
remainders_pos[606].modulo_value=2424; remainders_pos[606].bit_mask=create_bit_mask(76,0x28808a2,0x28828a08,0x82082080,0xa20a2082,0x800020,0x8202,0x208a20a2,0x20208820,0x88288002,0x200a2082,0x28800200,0x8a00028a,0xa20a2002,0x28a00808,0xa088008,0x200a20a2,0x8028028,0x80202002,0x22000020,0x8228a20,0x8228220a,0x200a00a0,0x20020020,0x8288080,0x882800,0xa20020,0xa280,0x22822880,0x820028,0x2288008,0x22080822,0x20020008,0x8220828a,0x20a2882,0x20020,0xa082082,0xa0800080,0x8800,0xa002200,0xa0020002,0x820820,0x80080080,0x828228a0,0x8a28208,0x88200080,0x80882020,0x200228,0x88280082,0x22882,0x28a00,0x8008a,0x20002822,0x8020228,0x8080080,0x820a00a0,0x20a08828,0x8808a288,0x2080000,0x20800808,0x8280280,0xa08a08a0,0x28200220,0x8a20200a,0xa28008a0,0x28a28000,0x8008002,0x228208a0,0x8800228,0x8082208,0x8a08a2,0x808200,0x8a080002,0x828208a0,0x28020820,0x8020a282,0x8a2022);
remainders_neg[606].modulo_value=2424; remainders_neg[606].bit_mask=create_bit_mask(76,0x82000022,0x20020200,0xa00a008,0x22822802,0x8008828,0x8808028a,0xa0022822,0x28a00028,0x20008a,0xa0822802,0x20008a08,0x2088202,0x22822882,0x20208000,0x82000080,0xa0822822,0x820820,0x828a08a,0xa28808a0,0xa20228,0xa20a282,0xa0820820,0x28828820,0x80200008,0x80002080,0x8228828,0x88082208,0xa20a2000,0x8028820,0x8a200080,0xa28000a2,0x28828800,0xa280202,0x82822002,0x8028828,0x8200a00a,0x20080800,0x8800008,0x8208a288,0x208a0882,0x8028028,0x8008008,0x20a2020,0x220a00,0x288008,0x28a0,0x8a08a20,0x20800a,0x808a2002,0x8820208,0x88008002,0xa08820a2,0x828a20,0x80008008,0x2820820,0x28200020,0x2200,0x82800880,0x28008000,0x208208,0x20020020,0x20a08a28,0x228a082,0x22080020,0x20220808,0x8008008a,0xa20a0020,0x8a20,0x8000a280,0x80020022,0x8000a08,0x200808a,0x2020020,0x20828028,0x828220a,0x228a2);
remainders_pos[607].modulo_value=2428; remainders_pos[607].bit_mask=create_bit_mask(76,0x8a80220a,0x28a20288,0xa22a2a28,0x8aa002a2,0x800a0a20,0x8aaa2228,0x88a8aa2,0x8a0a0aa2,0x2a800888,0x8a20a82a,0xa0000a28,0xa0028a2a,0x88a8820,0xa08800a,0xa2008280,0x882a280,0xa2aa00a0,0xa82a0aaa,0xaa08008,0x22aa2802,0x280a0,0xa82a8008,0x2a080a22,0xaa0a2a88,0x82aa2282,0x28a20222,0x820aa8,0x2082aaa8,0x80a02882,0x2222aa00,0x88028282,0x8020222,0x20888002,0x288282aa,0x80aa802,0xa0808088,0x220a8820,0x288aa02,0x880aa228,0xa0822a08,0x2202020,0x8802aa02,0x8aa82822,0x8002220,0x28880802,0x88282802,0x200aa888,0x882280a0,0x2aaa820,0x82aa0820,0x288808a2,0xa8288aa8,0x822a8a0a,0xa88a020a,0x2002a82,0x80a02800,0x8028aa8,0xa20020aa,0xaaaa0a82,0xa00aa8,0xa028a822,0x282008,0xa00220a,0xa0822a22,0xaa8a2800,0x228a0000,0x8a82a08a,0x2222002a,0x8aa0a0a,0x28aa2a22,0x22888aaa,0x208a0a00,0xa8a800aa,0x828a8a88,0x222808a2,0xa08802a);
remainders_neg[607].modulo_value=2428; remainders_neg[607].bit_mask=create_bit_mask(76,0x208aa82,0xa02a8a00,0x2aa2a2a0,0x2288a2a,0x88282a8,0x222aaa0,0x8002022a,0x282822a,0xa2088000,0x2a820a2,0x288882a0,0x288a02a2,0x800200a8,0x82800882,0x2a880a08,0x800a2a08,0x2a228828,0x20a28222,0x2280880,0xaa22a08a,0x888a0828,0x20a20880,0xa28082aa,0x2282a200,0xa22aa0a,0xa02a8aaa,0x880a8220,0xa80a2220,0x828a00a,0xaaaa2288,0x8a0a0a,0x808a8aaa,0xa800088a,0xa00a0a22,0x8082208a,0x28080800,0xaa8200a8,0x8a00228a,0x822aa0,0x280aa280,0x8aa8a8a8,0x8a228a,0x220a0aa,0x8088aaa8,0xa000808a,0xa0a08a,0xa8822000,0xaa0828,0x8a2220a8,0xa2280a8,0xa000802a,0x20a00220,0xaa20282,0x20028a82,0x8a88a20a,0x828a088,0x808a0220,0x2a88a822,0x2222820a,0x88288220,0x28a020aa,0x88a0a880,0x8288aa82,0x280aa2aa,0x2202a088,0xaa028888,0x20a2802,0xaaaa88a2,0x80228282,0xa022a2aa,0xaa000222,0xa8028288,0x20208822,0xa020200,0xaaa0802a,0x28008a2);
remainders_pos[608].modulo_value=152; remainders_pos[608].bit_mask=create_bit_mask(5,0xa222aa02,0x220828,0xa82a0,0x8a282088,0x80aa88);
remainders_neg[608].modulo_value=152; remainders_neg[608].bit_mask=create_bit_mask(5,0x2aa2228a,0x88aa80a0,0x8820a28,0x2a0a800,0x82200);
remainders_pos[609].modulo_value=2436; remainders_pos[609].bit_mask=create_bit_mask(77,0x82080822,0x28800000,0x80080208,0x22022002,0x28800200,0x80002280,0x20020000,0x28,0x288082,0x200a0882,0x208000,0x8020a282,0x820000a0,0x8020008,0x8008202,0x200808a2,0x220200,0xa20800a,0xa2822802,0xa20a28,0x80082282,0x2022002,0x28808020,0x2080082,0x80020082,0x28000208,0xa282008,0x82002,0x8828020,0x808a002,0xa0000002,0x20020,0xa200008,0x20008a0,0x220220,0x8200800a,0x8a0,0x20808208,0x820208a,0xa2000002,0x200820,0x8808800a,0xa2000800,0x8a00,0x80080002,0xa0,0xa20208,0x80282208,0x820000,0x828a08,0x8000282,0x20080022,0x20020808,0x80202288,0x880800,0x28820028,0x8a08a008,0x28828a2,0x208a08,0x808800a,0xa2020080,0x8200208,0x80208,0xa0000822,0x28a08020,0x208008,0x220a0080,0x20228008,0x80000008,0x80082,0x28800020,0x8002280,0x880880,0x8020028,0x2282,0x82020820,0x8);
remainders_neg[609].modulo_value=2436; remainders_neg[609].bit_mask=create_bit_mask(76,0x2800022,0x20008800,0x8280,0xa20a2882,0x20008a08,0x8082208,0xa08a0080,0x8808820,0x8200002,0x20820002,0x8200808,0x820220a,0x2880820,0x828000,0x8008020a,0x20800022,0xa28a08,0x82280002,0x22022082,0x8220220,0xa20a,0x828a2802,0x20000820,0x8200800a,0xa0802,0x20808a00,0x82202080,0x80802082,0x20820,0x2082,0x20880882,0x8028828,0x82208080,0x82880020,0x8a28228,0xa080002,0x880020,0x20000a00,0x28a002,0x22800882,0x8a00028,0x82,0x22800000,0x8800200,0x808a,0x80080820,0x8228a00,0x8202280,0x800a0080,0x20200,0x88202,0xa08008a2,0x28028000,0x8202200,0x80000080,0x20028020,0x2002000,0x2002022,0xa00200,0x80000082,0x22020800,0xa00a00,0x88008280,0x20880022,0x20200020,0x80280080,0xa2020800,0x28a20800,0x8080080,0x80800002,0x20008828,0xa200,0x80000000,0x28820,0x8800a20a,0x28a0020);
remainders_pos[610].modulo_value=2440; remainders_pos[610].bit_mask=create_bit_mask(77,0x28a22a8a,0x880a0282,0x88a2020,0xaa00828,0x2088800,0x2822020a,0xa00888a2,0x2a0a8280,0x88028020,0xa2008a22,0xa8820802,0x82020a20,0x200000a0,0xa82a02a,0x80000a0a,0x28082208,0xa82a0000,0x28a0000,0x222a02a,0x22002222,0x28a00280,0xaa200802,0x2208a028,0x80a2a888,0x20008020,0x8aa2a00,0x20228282,0x28828008,0x82a002,0x2280a802,0x8080200a,0x8088a02,0x22820020,0x8880a0a0,0x20a8002a,0x28200888,0x82208802,0x20020,0x800000,0x22088208,0x20082880,0x2a0822,0xa0222a8,0x82880a,0xa2202008,0x8020280,0x2a0288a0,0xa820080,0x2822880,0x82880820,0xa8aa2082,0x2000800,0x2a8a0208,0xa208822,0x2008aa28,0x800a2880,0x88008802,0xa888088,0xa280a8,0xa82a00,0x88202800,0xa0000220,0xa82a0a0,0x8a8,0xa080820a,0x20822a08,0xa2008a80,0x2802288,0x82a0a808,0x22200a02,0x8088288a,0x222080a0,0x200aa000,0x8a22028,0x80a02208,0xa88a2882,0xa2);
remainders_neg[610].modulo_value=2440; remainders_neg[610].bit_mask=create_bit_mask(77,0xa02a2202,0x28a02,0x2a0a8,0x822080a0,0x82800008,0xa02a0a82,0x28000022,0x22820208,0x8208a8,0x2288022a,0x200a008a,0xa0a82a0,0x28888028,0x820220a2,0x888202,0xa0802a80,0x20228880,0xa028088,0x8aa228a2,0xa288aa2a,0x20280a08,0x22288082,0x2a8020a0,0x8222000,0xa0880828,0x80222288,0xa82a0a02,0x200a0080,0x8802288a,0xa208200a,0x8082882,0x80000282,0x2a0a80a8,0x2828,0xa0208822,0xa0a80000,0xa280002,0x88a80a8,0x88008888,0xa2800a00,0xa8802008,0x882280a2,0x28aa220,0x80020082,0x22a8a800,0x808a0a08,0xa20a0020,0x20a8008,0x8a02a008,0x2008028,0x2022280a,0x8a080080,0x22028280,0x82a000aa,0xa0802220,0x8822008,0x80082,0x2000000,0x88220820,0x8020a200,0xa82088,0x28088aa0,0x20a2028,0x88808020,0x20080a02,0xa80a2280,0x2a080200,0xa08a200,0xa202080,0xa2a8820a,0x8002002,0xaa280820,0x28822088,0x8022a8a0,0x28a200,0x2002200a,0x22);
remainders_pos[611].modulo_value=2444; remainders_pos[611].bit_mask=create_bit_mask(77,0x28202a2,0xaa222a0a,0x20aa0280,0xa288a820,0x22000a00,0x8a022,0x8aa282,0x20a0a2,0xa220aa00,0xa8aa22aa,0x20aa0800,0x880aa208,0xa2a08288,0x8a088a28,0xa00a2022,0xa2a20280,0x808a0202,0xa08a28aa,0xa28aa022,0x88020802,0x20022808,0x8a802a00,0xa880008,0x2882a80a,0x228208,0xa200820,0x88028222,0x82802822,0x802028,0x80088800,0x20282aaa,0x8022888,0x8822aa82,0x802a8888,0xa020082a,0xa8020a8,0x8800a,0x88a0a82,0xa220208,0x2200082a,0x802a0a00,0x80a2a0,0x2a802a82,0xa8822222,0x8802082a,0x82808222,0x22002aaa,0x80200002,0x80282280,0x28022882,0x8a0888,0x28800082,0xa8228208,0x22a0a02,0x802a2200,0x8800800a,0x8022202,0xaa28a802,0x8a20a880,0xa202aa2,0x8a8a808,0x8a00a028,0x220a2880,0x20a8a28a,0xaa022228,0xaa08208,0x8aa2a002,0xa088aaa8,0xa080000a,0xaa2008a0,0xa2000828,0x88880,0xa228a00a,0xaa08082,0x888aa028,0x8280a0a,0x8a8);
remainders_neg[611].modulo_value=2444; remainders_neg[611].bit_mask=create_bit_mask(77,0x8a0a8a2a,0x22aa2202,0xa0228a08,0x2a0020a8,0xaa880288,0x888028aa,0x88022a02,0x8828202a,0x2aa82288,0x2022a222,0x28228088,0x822a80,0x2a200a00,0x2800220,0x2882a82a,0x222a8a08,0x8028a8a,0x28022022,0x2a0228aa,0x8a8082,0xa80aa080,0x208a288,0x82008080,0x200a2082,0x8aa0a80,0x82a080a8,0x8a0aaa,0xa08a02a,0x8008a8a0,0x8800088,0xa0a02222,0x808aa000,0xaa2202,0x8220000,0x28a880a2,0x8208a020,0x8000882,0x8002820a,0x82a28a80,0xaa8880a2,0x8a28208,0x80082a28,0xa200a20a,0x200a2aaa,0x8a80a2,0xa080aa2,0xaa08a222,0x8a8888a,0x8a02208,0x208aa00a,0x88028000,0xa000880a,0x20aa0a80,0x8aa2820a,0xa2a288,0x880882,0x808a2a8a,0x22a0208a,0x2a82000,0x8228a22a,0x80202000,0x28820a0,0x2a82a008,0xa8202a02,0x2282aaa0,0x82280a80,0x22a2802,0x20002220,0x28088882,0x22a80028,0x2a8880a0,0x8800000,0x2a202882,0x8228080a,0x220a0,0xa08282,0x20);
remainders_pos[612].modulo_value=408; remainders_pos[612].bit_mask=create_bit_mask(13,0x2082002,0x8a28808,0x2082008,0x820800a0,0x28a20200,0x82080202,0x82002800,0x28808020,0x8200808a,0x800a0020,0x20200820,0x8020228a,0x800820);
remainders_neg[612].modulo_value=408; remainders_neg[612].bit_mask=create_bit_mask(13,0x82802882,0x220000,0x8a00a080,0x2000820,0x20228a08,0x200828a,0x2882080,0x20000828,0xa080002,0x8208a0,0x28a08020,0x828a202,0x80020);
remainders_pos[613].modulo_value=613; remainders_pos[613].bit_mask=create_bit_mask(20,0x7e6b169a,0xe28bcb30,0x79258df,0xbd99592,0x4ebdae2a,0xf5e89ef1,0x14641ef9,0x7518841e,0x91215b3b,0x4721388e,0x36a1225c,0x8462bb7,0xde098a1e,0xde45ebe7,0x1d6f5ca3,0x6a66f415,0xc6927812,0x34f451fe,0x5a359f83,0x16);
remainders_neg[613].modulo_value=2452; remainders_neg[613].bit_mask=create_bit_mask(77,0xa2a28a02,0x2a0202a8,0x8a0a8002,0x8200080a,0x822022a2,0x28200228,0x88a88220,0xa8800882,0x8a882a2,0x8aa8a002,0xaa28aa80,0x808aa22a,0x2800282,0x288222a,0x80a2802a,0xa2aa2888,0xa0aa08a,0xa8288822,0x82a8020a,0xaa0a8a,0x88a0aa82,0x2a828220,0xa2aa2a88,0xa280888,0x208a02a0,0xa02a2,0x2a880a0a,0xa8a2aaa2,0x28aa880a,0x8a88200,0x8008aa2a,0x8a88a08,0x20a022,0x202208a2,0x80020a2a,0xa828a80,0x202a200,0x8a2a28aa,0x2082a202,0xa2a00820,0x80aaa202,0xaa8a8082,0x2a2020a2,0x2aa208a2,0x8a22a0a,0xaa88a088,0x8a02008,0x82aa288,0x800a880,0x80220002,0xaaa0a0a8,0x8a2202a0,0x20aa02a0,0x20888a8,0x80288020,0x88a2282,0xaa2800a,0x28000a0,0x2800a2a8,0xa0a22888,0x2008a00,0xa888820,0x82a02a82,0x82a02228,0x8a8282aa,0x20002200,0x8a8008,0x8a2aa08,0x88020288,0x88288aa,0x282a2288,0x228822aa,0xa282022a,0xa080a8aa,0xa022aa80,0x20802a2,0x2a082);
remainders_pos[614].modulo_value=2456; remainders_pos[614].bit_mask=create_bit_mask(77,0xa2aa2a22,0x2082a200,0x822a0002,0x222a8a0a,0xaa828002,0x280a2a00,0x2822a220,0x28a0aa8a,0x8a80aa,0x8222a802,0x82008a8,0x220a02a2,0xaa2a82a0,0xa2a0828a,0x88282208,0xa022aaa8,0x88a2a820,0xaa280aa8,0xa2a22a08,0x2202002a,0xa82008a0,0x2088800,0x22a0a28,0x28288800,0x8a0208a2,0x80200082,0xa0a22a02,0xaa28a202,0x228808,0x808aa2a8,0xa08200a,0x2282080,0x20a82a22,0x82800aa0,0x802222aa,0xa02280a0,0x28a22aa,0x20020aa2,0x8088220,0x808aa080,0xaaa88a2,0x20a0288,0x82aa8888,0x80aa002,0x8088a82a,0xa0020828,0x2a00820,0x2a8aa2,0xaa202288,0xa808a28,0x280a88a,0xa2820008,0x288a2080,0x80002228,0x8028a0a8,0x2a002220,0x880a2008,0x8aa80080,0xaa20a88a,0x222aa028,0xa082a8a,0x222aaa88,0x8a208828,0xaaa2820a,0x880a82a8,0x208a80a0,0x822a2008,0x802a88,0x28aa02a2,0x28a2aa0a,0x28088a88,0xaa00a8a0,0x88800282,0x82a0a2a8,0x88000a8,0x8a008a82,0x88a8aa);
remainders_neg[614].modulo_value=2456; remainders_neg[614].bit_mask=create_bit_mask(77,0x2a22a2aa,0xa80a2a88,0xaa2888a,0xaaa20282,0x220a088a,0xa082a288,0xa0aa2aa8,0xa0282202,0x88020822,0xaa2208a,0x80a88020,0xaa828a2a,0x22a20a28,0x2a280a02,0xa0aa80,0x28aa2220,0x2a20a8,0x22a08220,0x2a2aa280,0xaa8a88a2,0x20a88028,0x8a800088,0x8aa282a0,0xa0a00088,0x28a802a,0x8a8880a,0x282aa28a,0x22a02a8a,0x88aa0080,0x8022a20,0x8280a882,0x8aa0a808,0xa820a2aa,0xa088228,0x8aaaa22,0x28aa0828,0x8a02aa22,0xa88a822a,0x80800aa8,0x8022808,0x8222002a,0x8a828a00,0xa220000,0x8082288a,0x80020a2,0x288a80a0,0x8a2880a8,0x8a2022a,0x22a8aa00,0x820802a0,0x8a082002,0x2a0a8880,0xa002a808,0x888aaa0,0x8a02820,0xa288aaa8,0x82a880,0x2208808,0x22a82002,0xaaa228a0,0x8280a202,0xaaa22200,0x2a800a0,0x222a0a82,0x820a20,0xa8020828,0xaa2a880,0x8808a200,0xa0228a2a,0xa02a2282,0xa0800200,0x22882028,0x88a0a,0xa282a20,0x80088820,0x288020a,0x2022);
remainders_pos[615].modulo_value=2460; remainders_pos[615].bit_mask=create_bit_mask(76,0x20082002,0x28028800,0x200a000,0x828200a2,0x20a00008,0x800228a,0x820800a2,0x28820008,0xa000088,0x2080002,0x28200800,0x82088282,0x2802080,0x808008,0x200200,0xa08a0880,0x20028a00,0x8088080,0x808008a2,0x20200208,0x200002,0x80822,0x200,0x8820a002,0x22082002,0x28020208,0x2080200,0x20800000,0x828820,0x8200020a,0x80002002,0x20020220,0xa008008,0x80000000,0x228,0x808000a,0x20008a0,0x20808200,0x82200088,0x82020,0xa20008,0xa000202,0x22880000,0x8200200a,0x20880800,0x28008000,0xa080008,0x80822820,0x20,0x80080208,0x82080802,0x28008208,0x800a082,0x80000,0x8820200,0x88000080,0x22080080,0x8a20020,0x80202202,0xa00a2800,0x220a20,0x80080,0x2002020,0x20208028,0x88282208,0x20082,0x8000208,0x8220000a,0x22000822,0x8a00208,0x8a288002,0x220000a0,0x8a00828,0x8000a008,0x80022802,0x8008200);
remainders_neg[615].modulo_value=2460; remainders_neg[615].bit_mask=create_bit_mask(77,0xa0802882,0x20020000,0xa082088,0x2020822,0x20208800,0x80082202,0x2000822,0x20028000,0x2008000,0x82800802,0x20a00008,0xa000202,0x2082000,0x8000800,0x80288200,0x20020000,0x28020200,0x8,0x22,0x20a08a00,0x8828008a,0x800000a2,0x8808208,0x20208a,0xa2802002,0x20820a00,0x8a008280,0x20080080,0x8020028,0x2080202,0x882882,0x28028a20,0x2080080,0x800880,0x808a20,0x80000082,0x82080020,0x28000208,0xa208000,0x80002820,0x8220800,0x82088282,0x22000080,0x8808808,0x208a002,0xa0000080,0x20000800,0x2008080,0x220a0,0x808828,0x8000280,0x2000082,0x20800200,0x8000200a,0x80800800,0x20a00,0x88000,0x22800000,0x228828,0x28a20a,0x20822000,0x8228220,0x8008008,0x828028a0,0x20a00820,0x202280,0x800a0802,0x808200,0xa208082,0xa2880022,0x200a00,0x2200082,0x22880020,0x208020,0x82000,0x822082,0xa00);
remainders_pos[616].modulo_value=616; remainders_pos[616].bit_mask=create_bit_mask(20,0x2a82822a,0x8082a00,0xa20aa0,0x280a2800,0x22220282,0x80220280,0x8a8a020,0x888a0022,0x20a020,0x82,0xa080082,0xa22208,0xa2a2088,0x80880208,0x80888802,0x28a02882,0xa08a0000,0xa820200a,0x8282a800,0xa8);
remainders_neg[616].modulo_value=616; remainders_neg[616].bit_mask=create_bit_mask(20,0xa20a0222,0xa280,0x802a8228,0xa002a080,0xaaa20a0a,0xaa0a08,0x802020a8,0x2882a,0x8a028a0,0x88880a,0x8200880a,0x8822a280,0x82222800,0x8008280,0x8000002,0x20282002,0x20028888,0x2028a882,0xa022008,0x20);
remainders_pos[617].modulo_value=617; remainders_pos[617].bit_mask=create_bit_mask(20,0xd24dcb96,0xd90e9af1,0x41eef5ab,0x534b025e,0xf51346e5,0x1089def4,0xaa0611f4,0x19071067,0xb8be7e19,0x5733a927,0xf9f47792,0x23826261,0x21815798,0xee4420be,0x8b22bcbd,0x34b2a9d,0xbdde09e9,0x65c26f56,0x4ec92e3d,0x1a7);
remainders_neg[617].modulo_value=2468; remainders_neg[617].bit_mask=create_bit_mask(77,0xa80020a,0x820228,0x88222822,0x8a828a82,0x288a8a28,0x88000228,0x228a8828,0x808a88aa,0x2022a280,0x8aaa20aa,0x2028aa0a,0xaa0aaaa8,0xa8088a00,0x2288a822,0x2aa2020,0x8a82a200,0x20028020,0xa80aa28a,0x8200a2a0,0x20aa02a,0x80a82a28,0x228a2a,0xa88200a,0xaa00022a,0x8a202022,0x80aa20a8,0x82a80288,0xa020aaaa,0xaa8a82a8,0x2022a822,0x88080a2a,0x8a22a88a,0x88a08a,0x88a0a200,0xa88a28a,0x208a0a8a,0x80220a2,0xa80a202,0xa2828200,0xa20a80aa,0xa22a8a2,0x80a08a22,0x820880a0,0xa20a0880,0x8a088aaa,0x88022080,0x20a8a880,0x28022a20,0x2808002,0xa2a08,0x82a80280,0xa200a88,0x2a2a2088,0x22aaa002,0xaa2880a0,0x20822aa0,0x20280a80,0x2200a2a8,0x20aa280,0x8200a80a,0x2a82aa20,0xa202808a,0x2a2002aa,0x2808822a,0xa088a802,0xa00a,0xa0082a28,0xa200080,0x82022a20,0x8808a8a,0x28808220,0x22aaa888,0x20808828,0x80828088,0x28222882,0x22a28aa2,0xa2aa80a8);
remainders_pos[618].modulo_value=2472; remainders_pos[618].bit_mask=create_bit_mask(78,0x22000082,0x8828820,0x8228808a,0x20a2802,0x28208008,0x88288200,0xa00a2800,0x20800200,0xa20228a,0x220a2802,0x28a20a08,0x8800a202,0x22082082,0x200008,0x8028208a,0x220a0800,0x228200,0x2000200,0x808a28a0,0x8828200,0x82280000,0x228a20a2,0x20208200,0xa08a080,0x22800820,0x20208a28,0x88008208,0x22002,0x20208020,0x20a000,0x82000022,0x28028228,0x808a000,0x200a2080,0x20a08a08,0x2008000,0x80800880,0x20220800,0x8200828a,0x208808a2,0x20020200,0x2008002,0xa20a0800,0x8a00820,0xa202002,0x82802800,0x8228,0xa080088,0x2080800,0x8800008,0x82002280,0xa2080820,0x20028828,0xa20a008,0x82080802,0x8a28800,0x28828a,0x82822000,0x28a20200,0x8000800a,0x82880000,0x20a08800,0x8280200,0x800a2,0x8208820,0x8a002282,0xa08a2880,0x28a08820,0x8808a080,0x800208a2,0x28a00a00,0x82282200,0x2082800,0x28a08020,0x2288280,0x228220a2,0x8808,0x82);
remainders_neg[618].modulo_value=2472; remainders_neg[618].bit_mask=create_bit_mask(78,0xa2880802,0x20028,0xa200002,0x82822002,0x20a00800,0x200288,0x20822080,0x28008a08,0x8228a202,0xa2822082,0x20228200,0x8228a,0xa2802802,0x8a08800,0x820a002,0xa2820080,0x8a20a00,0x8a088288,0x22020,0x20a08,0xa208088,0xa2022822,0x28a00a08,0x82002008,0xa20800a0,0x28a00220,0x80280,0x808a2882,0x28a00828,0x88282088,0x28808a2,0x20820a20,0x80002088,0xa0822800,0x28200200,0x8a080088,0x80000,0x28a28008,0xa080202,0xa0000022,0x28828a08,0x8208008a,0x22820080,0x208028,0x8228a08a,0x2082080,0x8800a20,0x82008000,0x82800080,0x8800,0xa08a208,0x228000a0,0x28820020,0x82282080,0x2000082,0x220008,0x88200202,0x20a2880,0x20228a08,0x8080082,0x2000880,0x28200008,0x80208288,0x80800822,0xa00028,0x208a20a,0x20022000,0x20200028,0x2008,0x8a0022,0x20208208,0xa20a288,0x82802080,0x20200828,0x8a200208,0xa20a2822,0x8800000,0xa);
remainders_pos[619].modulo_value=2476; remainders_pos[619].bit_mask=create_bit_mask(78,0xa088a2a,0x2aa2aa22,0x8a0a0a08,0x2a80a0a8,0xa0082882,0x828a2a22,0x2a822a20,0xa280822,0x220aaa80,0x20228aa2,0x2a080,0x282aa20,0x80aa00a8,0x22800082,0xaa88a288,0xa0aa088,0x2a0a8aaa,0x22002a22,0x2a888080,0x8a0282,0x88aa2a22,0x808880aa,0x82800202,0x222802,0x8a20822,0x2a0a2aaa,0xaa02a80,0xaaa8800a,0xa8aa280a,0x88880208,0x82aa0002,0x88a2a0,0xa28808a,0xa8088082,0x8a20a2,0x8220282a,0xa28282,0x82888008,0x20008222,0x22282888,0x28a00200,0x80882828,0x8a200a82,0x2202a8a0,0x228a0820,0xa2200a20,0xaa820a8,0x2222800,0x8aa2a208,0x22aaaa02,0x80aa0a00,0x8a0a802a,0x8a20aaa,0x88800882,0x282802,0x22202808,0x8aa22aa0,0xa20088a,0x222a8828,0x80088020,0x2a0a888a,0xaa0a0aaa,0xa22aa220,0x288228,0xaa02820,0xa82802a0,0xa800008a,0x28808020,0xaa0888aa,0x28a002a,0x882a8882,0x8a28208a,0x8200a88a,0xa02a8822,0xa0a22a0,0xa8aa820a,0x220a088a,0xa8a);
remainders_neg[619].modulo_value=2476; remainders_neg[619].bit_mask=create_bit_mask(78,0x828002a2,0xa22a22aa,0x2828280,0xa2082820,0x2880a00a,0xa02a2aa,0xa20aa2a8,0x82a080aa,0xaa822208,0xa8aa022a,0x888a2808,0x8a0a22a8,0x8228820,0xaa08880a,0x22002a00,0x82822800,0xa2820222,0xaa88a2aa,0xa2000808,0x8802820a,0x22a2aa,0x8000822,0xa088a8a,0x88aaa08a,0x802a80aa,0xa282a222,0x8228a208,0x22200882,0x2022a082,0x8a80,0xa22888a,0x88002a28,0x82a00802,0x2080080a,0x8802a82a,0xaa8a0a2,0x882a0a0a,0xa000880,0xa8880aaa,0xaaa0a000,0xa0288a88,0x800a0a0,0x2a8820a,0xaa8a2028,0xaa0280a8,0x2aa882a8,0x8220a820,0x8aaaa088,0x22a2a80,0xaa22228a,0x8228288,0x28208a2,0x802a8222,0x8800a,0x88a0a08a,0xaaa8a080,0x22aa228,0x82a88002,0xaaa200a0,0x88008a8,0xa2820002,0x22828222,0x2aa22aa8,0x88a00aa0,0x8228a0a8,0x20a08a28,0x20888802,0xa00808a8,0x22800022,0x8a0288a2,0xa2000a,0x2a0a802,0xa882002,0x28a200aa,0x8282aa28,0x20220a82,0xaa828002,0x202);
remainders_pos[620].modulo_value=1240; remainders_pos[620].bit_mask=create_bit_mask(39,0x22a82,0x82228220,0x882a228,0x2a028a8,0x2a08a02,0x88a0208,0x202000a0,0xa008008,0x8a208808,0x208000a,0x208a2a00,0x200a0202,0x808200,0xa2200a0,0x20a80208,0x80a8080a,0xa82202a2,0x28882288,0x88000088,0x22a8282a,0x28220000,0x2a228822,0x28a8088,0x8a0202a,0xa020802a,0xa0088,0x8008202,0x88080a0,0x8000a8a2,0xa2a00020,0xa0202208,0x8200200,0x200a0008,0x802080a2,0x8080a20a,0x202a280a,0x82288a82,0x88288,0x82a880);
remainders_neg[620].modulo_value=1240; remainders_neg[620].bit_mask=create_bit_mask(39,0x88a220a,0xa2a0aa0,0xa22a0,0x8a20a020,0x8228020a,0x80020a80,0xa8288820,0x2880080,0x2200080,0x82808802,0xa8022288,0xa8028a82,0x8000288,0x82a28828,0xa0208a00,0x8200082,0x202a0a22,0x2000a200,0x808800,0xa220a022,0xa0aa0088,0xa22a00a2,0xa020000,0x8020a8a2,0x20a80822,0x88820800,0x80080a82,0x80028,0x8802022,0x22288828,0x28a82a80,0x288a80,0x28828080,0x8a0082a,0x82a02,0xa0a22082,0xa200202,0x8800200,0x22008);
remainders_pos[621].modulo_value=276; remainders_pos[621].bit_mask=create_bit_mask(9,0x82022822,0xa20000,0x2280202,0xa2020800,0x8a08a20,0x88000208,0xa0080802,0x8200008,0x88288);
remainders_neg[621].modulo_value=276; remainders_neg[621].bit_mask=create_bit_mask(9,0x20a20a2,0x8228808,0x8a20828a,0x22820080,0x200228,0x88280,0x20800082,0xa08800,0x200);
remainders_pos[622].modulo_value=2488; remainders_pos[622].bit_mask=create_bit_mask(78,0xaa80020a,0x20820028,0xaaaa2aaa,0x2828aa0,0xa88a20,0x88a802a8,0x28a2a28a,0x82888882,0xa082aaa0,0xa20a0a2,0x280a2a0a,0x280a22,0x2a02a800,0x8200202a,0x2088228,0x28288aa0,0xa2220a20,0xa80aa222,0x88220288,0x880a20a8,0x882a228,0x22200a8,0x2a20a222,0xaa008828,0x882288a2,0x202a2aa2,0xaaa802a0,0x2220a82a,0xa0a020a8,0x20a0a00,0x8a00028a,0x22888882,0x80820208,0xa802a808,0xaa20880a,0xa008282,0x20002000,0x82aa28a,0xa2aa80,0x2002aa8a,0x8a28aa8,0xa0000800,0xaa828200,0x2aa02208,0x2202a80,0x88208082,0xa2822222,0x80a28000,0xa00a0a0,0x882a080a,0xaaa82a08,0x80a802a,0x228aa8a8,0xaa8a2288,0xa8282200,0x80888a08,0x202a0088,0x22288a82,0x222a08a0,0x2a228088,0x8a888aa0,0x2808a088,0x800aa228,0x82288220,0xa8a80800,0x2a80,0x2888a028,0xa0a0a8a0,0xa8a0a08,0x820aaa82,0x28822222,0x22a28a8a,0x2a802a,0x8008a22a,0xaa0aa282,0x8aaa8aa,0xaa280082,0xa08002);
remainders_neg[622].modulo_value=2488; remainders_neg[622].bit_mask=create_bit_mask(78,0x22088a82,0xa80a88a0,0x2222a222,0x8a0a0228,0x882002a8,0x208a20,0xa02a2a02,0xa00000a,0x280a2228,0x8228282a,0xa082a282,0x88a082aa,0xa28a2088,0xa88a8a2,0x8a800aa0,0xa0a00228,0x2aaa82a8,0x20822aaa,0xaa8a00,0x82a820,0x800a2aa0,0x8aaa8820,0xa2a82aaa,0x228800a0,0xaa002a,0xa8a2a22a,0x22208a28,0xaaa820a2,0x2828a820,0x8a828288,0x2888a02,0xaa00000a,0x80a8a80,0x208a2080,0x22a80082,0x82880a0a,0xa888a888,0x80a22a02,0x882a2208,0xa88a2202,0x802a0220,0x28888088,0x220a0a88,0xa228aa80,0x8aa8a208,0xa8080a,0x2a0aaaaa,0x82a0888,0x82802828,0xa28082,0x2220a280,0x808208a2,0xaa022020,0x2202aa00,0x20a0aa88,0x8000280,0xa8a28800,0xaaa0020a,0xaaa28028,0xa2aa0800,0x2000228,0xa0802800,0x8822aa0,0xaa00aa8,0x20208088,0x8888a208,0xa00028a0,0x28282028,0x82028280,0xa82220a,0xa00aaaaa,0xaa2a0202,0x88a208a2,0x8802aa2,0x22822a0a,0x80222022,0x22a0880a,0x28088a);
remainders_pos[623].modulo_value=2492; remainders_pos[623].bit_mask=create_bit_mask(78,0x280a202,0x22a88a02,0xa0a20000,0x8220a020,0x8820888,0x8a220288,0x882a8288,0xa02a22a,0x8208000,0xa2a80aaa,0x2228802a,0x8a2a28a,0x88822808,0x8002a080,0xa0880802,0x8a820,0x2a28a8a,0x882080a8,0x8082a80a,0x2208a8,0xa0022a,0xa00a2822,0x2000802,0x20a002aa,0xa22a028,0x2a200a88,0x82a00a20,0x80080022,0x8a2008,0xa000,0xa2a208a2,0x80882,0x22020082,0x820a0880,0x20228222,0x2888280a,0x2aa88800,0x82820008,0xa08280a,0x2a02820a,0x82000828,0x800222aa,0x8a028222,0x28882880,0x80220a08,0x8200808,0xa8220200,0x8a208a8,0xa000,0x22008a20,0x28800200,0x808a00a8,0x22a008a,0x8280a88a,0xaa800a0,0xa8020008,0x8828a00,0xa8800a0,0x22a20880,0x2a02a820,0x2a02082,0xa2a28a8,0xa082a200,0x28020220,0x2020a800,0x2028822,0x8a28a8a2,0xaa802288,0xaaa02a8,0x2082,0x2a88a80a,0x22282a82,0x228088a,0x22220822,0xa080a088,0x800008a0,0x80a22a8,0x808a028);
remainders_neg[623].modulo_value=2492; remainders_neg[623].bit_mask=create_bit_mask(78,0x8a082a0a,0x2a200282,0x202a8888,0xa2828a8,0x80028000,0x2aa0a00,0xa20200,0x828a2a22,0xa80080,0x22208222,0xaa2008a2,0x80222a02,0xa2080,0x88a2008,0x2800800a,0x88020a0,0x822a0202,0x280820,0x8022082,0x88aa0020,0x882882a2,0x2882a02a,0xa888082,0xa0288a22,0x822a28a0,0xa2a08200,0xa2802a8,0x88080aa,0x8802a800,0x8882880,0x2222802a,0x8800800a,0xaa82880a,0xa820008,0xa8aa02aa,0xa000a002,0x22200080,0x20a8880,0x8200a082,0xa2820a82,0xa8800a0,0x80aa222,0x28a0a2a,0x2000a000,0xaa8280,0x80288080,0x20a28a88,0x802a0020,0x88882088,0xaa880228,0x20088a80,0x28820,0x82228802,0xa002002,0x82200828,0x208a8080,0x800a0208,0x2008820,0xa22a8008,0xa20a20a8,0x8a20a80a,0x82a22020,0x280a2288,0x208a8a28,0x28a82080,0x828a00aa,0x220202a,0x2200aa00,0x82220a20,0x8888a00a,0xa2002002,0x2aa0a202,0x82a08002,0xaa2a80aa,0x28002800,0x8880020,0x8082a220,0x802820);
remainders_pos[624].modulo_value=312; remainders_pos[624].bit_mask=create_bit_mask(10,0x828800a2,0x20020208,0x82000008,0x22080800,0x28a00228,0x8828800a,0x82002020,0x8200000,0x82208080,0x8a0022);
remainders_neg[624].modulo_value=312; remainders_neg[624].bit_mask=create_bit_mask(10,0x2000822,0x28828a00,0x2088080,0xa2800080,0x20200a20,0x200082,0x28828a0,0x208808,0xa280008,0x200a2);
remainders_pos[625].modulo_value=5; remainders_pos[625].bit_mask=create_bit_mask(1,0x1e);
remainders_neg[625].modulo_value=20; remainders_neg[625].bit_mask=create_bit_mask(1,0x22202);
remainders_pos[626].modulo_value=2504; remainders_pos[626].bit_mask=create_bit_mask(79,0x2200222,0x2a2a28a2,0x2882a0a8,0x2a8222a2,0xa0a228,0x2a080222,0x2a8a0888,0xaaa2002a,0xaa000a80,0x80820aa2,0x8a22a20a,0xa2208208,0x22a288a,0xa28000a,0xa08aaa88,0x82282aa0,0x802aa2aa,0x2a2820a2,0xaa8aa02a,0x8002aa88,0x2a008008,0xa282020,0x2002a82,0x282280,0x80008a0a,0xaaa820a8,0x882022a0,0xa28a2200,0xa2022088,0xa28280,0x80aaa002,0x2aa2000a,0x88a08020,0x22a8a028,0x220a0aa2,0x2228028,0xa028822,0x8202028,0x22aa22a2,0x88aa8882,0x808208a,0x2280a028,0x2888088,0xa0a08828,0xa2a888a,0x20a2228,0x8aa808,0xaaa02a0,0x828a0080,0x8808a02,0x88a28a22,0x88082200,0x82aaa0a,0xa200022a,0x882800a0,0xa8008002,0x828a082,0x200a808,0xaa800220,0xaa2aa22,0x828a8a8,0x8aa8028a,0xa82882aa,0xaaa20a0a,0x28a022,0x28a880a0,0x82088aa2,0x8a88a220,0xa08202a0,0xa000aa8a,0x8aaa02,0x20a2a8a8,0x8020a822,0x8a0a0088,0x8882a828,0xa82288a,0x28a8a82a,0x8008808a,0x88);
remainders_neg[626].modulo_value=2504; remainders_neg[626].bit_mask=create_bit_mask(78,0x8aa88aaa,0xa2a2a02a,0xa00a2820,0xa20aaa2a,0x88282aa0,0xa2808aaa,0xa2028000,0x222a88a2,0x22888208,0x80a822a,0x2aa2a82,0x2aa80a80,0x8aa2a002,0x82a08882,0x28022200,0xaa0a228,0x8a22a22,0xa2a0a82a,0x220228a2,0x88a2200,0xa2880880,0x82a0a8a8,0x8a88a20a,0x88a0aa08,0x8880282,0x2220a820,0xa8aa28,0x2a02aa88,0x2a8aa800,0x882a0208,0x822288a,0xa22a8882,0x2808a8,0xaa2028a0,0xaa82822a,0x8aaa08a0,0x828a00aa,0x80a8a8a0,0xaa22aa2a,0x22000a,0x8080a802,0xaa0828a0,0x8a000800,0x282800a0,0x82a20002,0x8a82aaa0,0x88022080,0x82228a28,0xa028808,0x8008028a,0x2a02aa,0x80aa88,0x80a22282,0x2a888aa2,0xa08828,0x2088088a,0x80a0280a,0x8a882080,0x22088aa8,0x822a22aa,0x80a02020,0x2208a02,0x20a00a22,0x222a8282,0x88a028aa,0xa0200828,0xa80022a,0x2002aa8,0x280a0a28,0x28882202,0x8802228a,0xa82a2020,0x8a820aa,0x2828800,0xa20a0,0x820aa002,0xa02020a2,0x8800802);
remainders_pos[627].modulo_value=2508; remainders_pos[627].bit_mask=create_bit_mask(79,0x82822082,0x228800,0xa000008,0x8228a0,0x20220800,0x82008282,0xa2002022,0x8020228,0x800a002,0x22822882,0x820a20,0x8282008,0x820a2880,0x8008208,0x282002,0x200000a0,0x28200,0x8808a000,0x2020802,0x28028008,0x88082,0x20022002,0x8028228,0xa28a088,0xa08000a0,0x8208200,0x202,0x80802082,0x800a20,0x80000,0xa0022082,0x8820008,0x88088000,0x20a0000,0x20008808,0x80288,0xa08a0020,0x20000a20,0x20a,0x802,0x8a08,0xa20a08a,0x2000080,0x20008228,0xa080202,0x22022000,0x8220000,0x8800a200,0x2000820,0x200000,0x8020208a,0x820,0x20820808,0x20a008,0xa28a00a0,0x28020220,0x88008288,0x22000800,0x28028820,0x8080200,0xa2022802,0x28000000,0x8008,0x828000a0,0x20020800,0x8a082208,0x82820022,0x8200200,0x8828808a,0xa0020822,0x8020800,0x8008a288,0x20082880,0x8808828,0x88200002,0xa00a2,0x28800200,0x88282002,0x820);
remainders_neg[627].modulo_value=2508; remainders_neg[627].bit_mask=create_bit_mask(79,0x2022002,0x8220008,0x2088080,0x800a2020,0x28a20008,0x208020a,0x228828a2,0x28a20,0x8000208a,0x220a2002,0x8028220,0x8020a080,0x2022000,0x800a00,0x8020a082,0xa0880820,0x820a08,0x2008,0x828a0002,0x20820800,0x800000a,0x208a2882,0x820220,0x82202000,0x20080020,0xa00a00,0x8808028a,0x82802,0x8008228,0x88008000,0x20822802,0x20800,0x88,0x82820080,0x28000000,0x88008200,0x20020820,0x28800228,0x80008282,0x80880082,0x8800200,0x82202002,0x2880800,0x28800220,0x8200828a,0xa20a2880,0xa28808,0x82280,0x828800a0,0xa08808,0x8282002,0x80880020,0x20028000,0x8282080,0x22020820,0x20828228,0x80200,0xa2800080,0x20820020,0x288,0x228a2082,0x20808808,0x88080000,0x2000820,0x28028008,0x200a280,0x20a00a2,0x208a00,0x200002,0x208a00a2,0x820008,0x2200,0xa0802000,0x20,0x20800a,0x820822,0x20008a08,0x20a08a,0xa0);
remainders_pos[628].modulo_value=1256; remainders_pos[628].bit_mask=create_bit_mask(40,0x8a0a2a0a,0x20a80aa,0x22020888,0x822aa228,0x20aa000,0x8002a82,0x20a22aa2,0x888202,0x8aa80a82,0x2a82a282,0xaa2828a8,0x20082a02,0xa0880822,0x208aa88,0x8082a80,0x882000aa,0x88288aa,0x8022208,0xa2aa02a,0xa0aa0a8a,0xaa8a0a2,0x888020a8,0x22822020,0x822aa,0xa82020aa,0xaa208002,0x20220a22,0xa8200888,0x2828aa80,0x8a82a82a,0xa02aa282,0x82220082,0xa88a0880,0xa800208a,0xaa08082,0x8aa88200,0x20808828,0x2a08022,0xa8a0a2aa,0xa0);
remainders_neg[628].modulo_value=1256; remainders_neg[628].bit_mask=create_bit_mask(40,0x282a282,0x8a820822,0xaa8a8000,0xaa22aa0,0x8a822888,0x8088a20a,0xa82aa22a,0x88000a8a,0x220820a,0xa20a2a0a,0x22a0a020,0xa880a28a,0x280080aa,0x8a802200,0x8000a208,0xa88822,0x800a0022,0x808aaa80,0x82a228a2,0x28228202,0x8220282a,0x8a820,0xaa0aa8a8,0x8880aa22,0x20a8a822,0x22a8088a,0xa8aa82aa,0x20a88000,0xa0a02208,0x20a20a2,0x28a22a0a,0xaaa880a,0x20028008,0x2088a802,0x8228000a,0x2200a88,0xa80800a0,0x8a2808aa,0x20282a22,0x28);
remainders_pos[629].modulo_value=629; remainders_pos[629].bit_mask=create_bit_mask(20,0xe7f14272,0x2362a0d2,0x195fa049,0xfb5a6333,0x39e7c63e,0x741642a8,0x8972c0da,0xb20c3b,0xade0daa1,0x52ba2260,0xed419117,0x402156c1,0xa4770c13,0xb96c0d3,0xe705509a,0xb7df18f9,0xa6333196,0xb124817e,0xf9d2c151,0x1390a3);
remainders_neg[629].modulo_value=2516; remainders_neg[629].bit_mask=create_bit_mask(79,0x2a288aaa,0xaaa2280a,0x80822880,0x22022aaa,0xa02a0aa2,0xa08a8a20,0xaa0802,0x82a80a2,0x20280220,0x8a22aaa8,0x20880882,0x88a88a08,0x282a800a,0x820a080a,0x2a888002,0x2a028020,0x2aaa280a,0x28a800a2,0x20020888,0x828a082a,0x820022a0,0xa8a088,0xa2aaa228,0xa02a280,0x88aa020,0x82820a00,0xaa0a280,0x88a88a80,0x80882082,0xaa228a08,0xa020aa,0xaa08022,0x82a80228,0x8a882a00,0x82a02822,0xa202222a,0xa2080aaa,0xa22aa808,0x88a2a280,0x8002aa,0x20a0202,0x2208a800,0xa00000a8,0x2802220,0x80822822,0x28aa000,0xa82228,0xaa2088a,0x2000022a,0x8a088082,0x8820a28,0xa88088,0x82828208,0x888aa0a2,0xa02a2200,0xa8202,0x2882200,0x82888a08,0x28a022a,0xaa2028a0,0x82888a0,0x2208a0,0xa020a800,0x80222a02,0xa282a888,0x820a0a0,0x8808a80,0x8a282088,0x20808828,0x2a200002,0x288822a8,0xa220a80,0x8002a8a0,0x220a2080,0x822200a0,0xa800002,0xa8822,0x20202820,0x80);
remainders_pos[630].modulo_value=840; remainders_pos[630].bit_mask=create_bit_mask(25,0x80820802,0x20200020,0x80200,0x82000022,0x8280200,0x800a0,0x20800028,0xa20a082,0x22880002,0x20200800,0xa200202,0x820080,0x820000,0x8008a002,0x20080880,0x228800,0xa08a080,0x20882,0x200028,0x8028200a,0x8200,0x80200088,0x80800,0x20820208,0x80);
remainders_neg[630].modulo_value=840; remainders_neg[630].bit_mask=create_bit_mask(26,0xa0002,0x28208820,0x8288,0x20008a2,0x808808,0x80200288,0x80000020,0x28008020,0x2202002,0xa2000802,0x28200008,0x82208282,0xa0000,0x8028008,0x200a,0x20800000,0x220000,0x2002008,0x80820002,0xa00820,0x8202082,0x80080800,0x8800200,0x208000,0x80000000,0x28020a00);
remainders_pos[631].modulo_value=2524; remainders_pos[631].bit_mask=create_bit_mask(79,0x2a2a8aaa,0x8a8a220a,0x82a22800,0x2a8002aa,0x222a00a8,0x28028aa0,0xa2820,0x802880aa,0xa208a8a2,0x88a88a8a,0x28820008,0x80a28aa0,0x2028a202,0xaa0a8200,0x8888a08,0x2aa8a0a2,0x80a220a,0x8aa088,0x228a8002,0x20a828a,0x208820a,0x22a8aaa,0x82a2a228,0x20008288,0x22228828,0x2802282,0x20a8aa8a,0x82a0a88,0x8a20a802,0x202222aa,0x22880202,0x20aa0820,0x82aa802a,0xaa0a8a0,0x8088a802,0xaa00a02a,0x880a000a,0x82022aa0,0x808288,0x8000002,0x22282020,0x20aa8808,0xaa000a02,0x2a80a00a,0x802a220,0x20a2a0aa,0x8a802aa8,0x80820aa0,0x88080228,0x2aa88880,0x802a08a,0x822a0a82,0xa2aa2a0,0x88288028,0x82822888,0x22282000,0x288a8a8,0xaaa2a88,0xa082208,0x8a282a08,0x8002a28,0x220aa20,0x8a088a02,0x8a0a2aa,0xa20a2222,0x80082a0a,0x2808a280,0x80aa28a0,0x22000822,0xaa2a22a2,0x28a2a208,0xaa02280,0x80828a00,0x80aa2802,0x82a00a88,0x2aa8002a,0x200288a8,0x8a088a2a,0xaaa2a8a);
remainders_neg[631].modulo_value=2524; remainders_neg[631].bit_mask=create_bit_mask(79,0xa2a20222,0x202aa82,0xa2aa088,0xa2088a22,0xaaa28820,0xa08a0228,0x8882a0a8,0x8a00822,0x2a80202a,0x200202,0xa00a8880,0x82a0228,0xa8a02a8a,0x22820a88,0x80000280,0xa220282a,0x8082aa82,0x88022800,0xaa02088a,0x8a020a02,0x8a800a82,0x8aa20222,0xa2a2aa0,0xa8880a00,0xaaaa00a0,0x8a08aa0a,0xa8202202,0x80a28200,0x2a8208a,0xa8aaaa22,0xaa008a8a,0xa82280a8,0xa2208a2,0x82282028,0x800208a,0x228828a2,0x828882,0xa8aa228,0x88080a00,0x8088888a,0xaaa0a8a8,0xa8220080,0x2288828a,0xa2082882,0x808a2aa8,0xa82a2822,0x208a220,0x80a8228,0x808aa0,0xa2200008,0x808a2802,0xaa2820a,0x82a22a28,0xa008a0,0xa0aa000,0xaaa0a888,0x8a002020,0x8222a200,0x8280aa80,0x2a0a280,0x8088a2a0,0x8aa822a8,0x280028a,0x80282a22,0x2a82aaaa,0x880a282,0xa0802a08,0x822a028,0xaa8880aa,0x22a2aa2a,0xa02a2a80,0x8228aa08,0x80a0288,0x822a08a,0xa288200,0xa22088a2,0xa88a0020,0x28002a2,0x222a202);
remainders_pos[632].modulo_value=632; remainders_pos[632].bit_mask=create_bit_mask(20,0x22088a82,0xa02a80a0,0x222a22aa,0x8a08228a,0x282000aa,0x800aaa28,0x208228a0,0x20022,0x2a0a802a,0x828088a2,0xa88a2202,0xa802a0,0x8880080,0x20a2882,0x2828aaa0,0xa2aa0008,0x88a28820,0xaaa88a8,0x880a02a8,0x82a220);
remainders_neg[632].modulo_value=632; remainders_neg[632].bit_mask=create_bit_mask(20,0xaa80020a,0x28a20828,0xaaa22a22,0x280aa02,0xa0a88822,0x88222a0,0xa80aa028,0x888a88aa,0xa28208a2,0xa08002a,0x2002aa8a,0x88208a28,0x80008008,0x8a82a00a,0xa0a02228,0x2a228880,0x2a00a8,0x82220020,0x828a20,0xa2aa8);
remainders_pos[633].modulo_value=2532; remainders_pos[633].bit_mask=create_bit_mask(80,0x228a2002,0x820a28,0x200a200,0x2002080,0x8a00808,0x80000288,0x220000a2,0x8828a00,0x200200,0x200a0882,0x228a28,0x208020a,0xa0020080,0x220a00,0x88202082,0x20220a2,0x20a20820,0x8228a088,0x828808a0,0x28a20200,0x8a080202,0xa2082880,0x28208828,0x2002082,0x208a2822,0x20020820,0x280282,0x82000802,0x20028200,0x8020a28a,0x20020882,0x8800200,0x220028a,0xa2800,0x28808a00,0x2008080,0x2000820,0x8000228,0x8a20a282,0x800a0022,0x28a08a28,0x88000208,0x82000802,0x20200800,0xa202280,0x28a0000,0x28008800,0x800220a,0x22080080,0x28a08028,0x828008a,0x2000820,0x28028008,0x82080088,0x828a2080,0x20800808,0x82208288,0x228208a2,0x8020a20,0x808a288,0xa2022820,0x20a28820,0x8208a082,0xa0880800,0x20808228,0xa088008,0x200800a0,0x8020800,0x8a28800a,0x220a0082,0x8008008,0xa282200,0xa0000880,0x28000028,0x200a202,0x20800802,0x8a00800,0x8a082000,0x8a2882,0x8);
remainders_neg[633].modulo_value=2532; remainders_neg[633].bit_mask=create_bit_mask(79,0xa2022882,0x8028220,0x8a082288,0x82882800,0x208000,0x8088200,0xa2800822,0x20208,0x88288288,0xa0820002,0x8a20220,0x8a008282,0x208a0800,0x8a28208,0x28a00a,0x828a2822,0x28228028,0xa202000,0x2000020,0x20228a08,0x200828a,0x22802000,0x20a00020,0x8a08a00a,0xa00220a2,0x28828028,0x8820820a,0x2880082,0x28820a08,0x8282202,0xa08a0002,0x8a08,0xa288202,0x80822080,0x20000208,0x8a080008,0x828800a0,0x808a20,0x228220a,0x8208a2,0x20200220,0x88280,0x2880082,0x28a08008,0x8228a208,0x82020880,0x20800008,0x8008a282,0xa2800800,0x20200820,0x80208002,0x828800a0,0x20820800,0xa008000,0x2022800,0x28008000,0xa280200,0xa20a0022,0x828228,0x80002200,0x228a20a0,0x28220028,0xa00200a,0x20000080,0x28000a20,0x82000080,0xa0800820,0x828008,0x2200082,0xa2820802,0x800800,0x8220a288,0x20880000,0x20808820,0x8a08228a,0xa0080082,0x208008,0x200a088,0x80022002);
remainders_pos[634].modulo_value=2536; remainders_pos[634].bit_mask=create_bit_mask(80,0xaaa8a2aa,0x82022088,0xaa828a22,0x222a2a80,0x80a08a0,0x28008a8a,0x8a08882a,0xa82022,0x228228a2,0xaa00802,0xaa8808a2,0x8aa02a82,0x2a8a820,0x2a28000a,0xa020088,0xa8a00a08,0x20822a2a,0x82aa0282,0x2000aa8a,0x2222a022,0xaa222a02,0x280800a,0x8a2282a0,0x8002022,0xa0a8a0aa,0x20288aa2,0x2a0aaa8,0x8a2808,0x8002802a,0x2208a8,0x8222a8aa,0xaa20822,0xa0822a28,0xa8202888,0xa8a0808a,0x222a08a0,0x800a8020,0xa2822082,0x88a22,0x80000208,0xa2200020,0x8828a88,0x2a00282,0x20a88808,0x20a2a0a,0x28082aa2,0xa8820a22,0x208aa028,0x2a888288,0x208800aa,0x280022a,0x28a200a8,0xaa0a8020,0xa228082a,0xa2a0a8a,0x80020aa,0x8288a288,0x202800a,0xa888aaa0,0xa888880,0xaa000888,0x80aa82a2,0xa8820882,0xa00a2aa8,0x80a020,0x28a822,0x2a2a80a0,0xa80aa208,0x2022aa82,0x200aa08a,0x28828880,0x82a008a,0x2220a288,0xa20028a8,0x20a020a2,0xa8a8880a,0xa282aa02,0x8808288,0x8a2aaa22,0xaa);
remainders_neg[634].modulo_value=2536; remainders_neg[634].bit_mask=create_bit_mask(80,0x22202a22,0xa8aa800,0x220a02aa,0xaaa2a208,0x80828028,0xa0880202,0x28000a2,0x8820a8aa,0xaa0aa02a,0x8228808a,0x2200802a,0x228a20a,0x8a2020a8,0xa2a08882,0x828a8800,0x20288280,0xa80aa2a2,0xa228a0a,0xa8882202,0xaaaa28aa,0x22aaa28a,0x8a080882,0x2aa0a28,0x8088a8aa,0x28202822,0xa8a0022a,0x8a282220,0x8802a080,0x88a08a2,0x882a8020,0xaaa2022,0x822a80aa,0x280aa2a0,0x20a8a000,0x20280802,0xaaa28028,0x88208a8,0x2a0aa80a,0x888002aa,0x8888a80,0x2aa888a8,0x800a0200,0x8a288a0a,0xa8200080,0x8a82a282,0xa080a22a,0x200a82aa,0xa80228a0,0xa2000a00,0xa8008822,0x8a088aa2,0xa02a8820,0x228208a8,0x2aa080a2,0x82a28202,0x8088a822,0xa002a00,0x8a8a0882,0x20002228,0x82000008,0x22888000,0x8220a2a,0x200a800a,0x2882a220,0x880828a8,0x88a020aa,0xa2a20828,0x20822a80,0xa8aa220a,0xa8822002,0xa00a0008,0x80a28802,0xaaa82a00,0x2a88a020,0xa828a82a,0x20200082,0x2a0a228a,0x80080a00,0x2a222aa,0x22);
remainders_pos[635].modulo_value=2540; remainders_pos[635].bit_mask=create_bit_mask(80,0x200202,0x2a2a8282,0x28822020,0x2808882,0xa08a22,0x2800220a,0x2a220a02,0x2280a208,0x20288a,0xa280aa22,0x802a0888,0x8a228280,0x8802a0,0xaa00080,0x22a08802,0xa0aa2802,0x820a8a02,0x2a0082a0,0x80a2a820,0x2a0aa20,0x28280008,0x80288000,0xa080228,0x8028002,0x88a000,0x8a2208,0x2a2a0aa0,0x802a0,0x8a220808,0x202200,0xa888000a,0x80200a0,0x828028,0x80828088,0x2280882a,0x8080208,0x200a0222,0x88a088,0x8a820800,0x22088a28,0x82a228a,0xa2200002,0xa008220,0x2020888,0x20288208,0x28202a82,0x28200220,0x8020280,0x222a0a0,0x80800a00,0x88a2008,0x2000202,0xa8a80a8,0x8a2000aa,0xa2200208,0x28020000,0x20a0800,0x22802288,0x2828000,0xa0a80200,0xa8a0208a,0x200a8082,0x2a0820a8,0x8aa0a80a,0x20a88802,0xaa0802,0x2200020,0x288a20a8,0xa802028,0xa028a222,0x8080088a,0xa0288a22,0x88a8208,0x8002880a,0x20a00a08,0x2028088a,0x88228822,0x2a8a8080,0x800828,0x808);
remainders_neg[635].modulo_value=2540; remainders_neg[635].bit_mask=create_bit_mask(79,0x88a80a8a,0xa2220a02,0x200aa0a8,0xa00000a,0x8028022a,0xa0882a82,0xa22a8282,0x2a082280,0x88a0a002,0x2208222a,0x8a20000,0x22a0a00,0x8008228,0x82208808,0xa228000a,0x2822208a,0xa020282,0x22880228,0x82220a8,0x82282228,0xa0a00880,0x8200880,0x28082a0,0x8082088a,0x80002808,0x88022a80,0xa2228220,0x8808228,0x2a28080,0x80a8aa08,0x20000882,0x800a8820,0x80a00a0,0x8020800,0xa2080022,0x80000a80,0xa8028aa2,0x8002000,0x2028088,0xa2800220,0x80a22a02,0x2a288882,0x28802a8,0x8a828000,0xa0a00a00,0xa0a8220a,0xa0288aa0,0x8a8208,0x8aa22828,0x88208,0x80022880,0x8a008a82,0x2020020,0x2a08822,0x22a88a00,0xa08a0888,0x8a028080,0x2a08a200,0x8a020888,0x20208a08,0x20282802,0xa8020802,0x2280a020,0x2202082,0xa020000a,0x8822008a,0x8a2888a0,0x20022020,0x8200a8a0,0x20a02a2a,0x8080002,0x282002a2,0x20280,0x8820082,0xa0288200,0xa8a00002,0x2a00a2,0x22020008,0x880080a0);
remainders_pos[636].modulo_value=1272; remainders_pos[636].bit_mask=create_bit_mask(40,0x82082822,0x8828220,0x8000800a,0xa28008a2,0x8808a08,0x2002202,0x22800,0xa20a28,0x8080002,0x802082,0x8208020,0x8000202,0x28a08a0,0x28800,0x8088008,0xa20a2022,0x28a20028,0xa002000,0x20882822,0x28828208,0x8208282,0xa0882822,0x28000800,0x8a28008a,0x208808a0,0x200220,0x80002280,0x200a20a2,0x20808000,0x80208,0x20820802,0x800020,0x28a08a,0x80002880,0x20808800,0x8a20a202,0x28a2002,0x20a00200,0x82088282,0x882820);
remainders_neg[636].modulo_value=1272; remainders_neg[636].bit_mask=create_bit_mask(40,0x28020a2,0x20a28,0x8080082,0x22080022,0x200,0x8a08a28a,0x808a2080,0x8228220,0x8000808a,0x80082802,0xa00828,0x8000828a,0x82020020,0x8820008,0x80000080,0x228228a2,0x20228820,0x8208a088,0xa0002022,0x20020a00,0x8028020a,0x200020a2,0x20808008,0x2208002,0xa0000020,0x8a08a28,0x808a208,0xa0822822,0x28000808,0x88008280,0xa00a0082,0x8000828,0x88202002,0x882000,0x28000008,0x228228a,0x82022882,0x28208a08,0xa000202,0x20a0);
remainders_pos[637].modulo_value=364; remainders_pos[637].bit_mask=create_bit_mask(12,0x2a82020a,0x20a80800,0x80828820,0x820208a0,0xaa00080a,0x82a82020,0xaa080,0x8082a02,0x282020a2,0x2a08082,0x82a8002,0xa08);
remainders_neg[637].modulo_value=364; remainders_neg[637].bit_mask=create_bit_mask(12,0xa20a8a02,0x28208000,0xa00a8,0xa0a8028,0x22800082,0xa2028a8,0x88822000,0x8000a20a,0x20a8a822,0x8228000a,0x22088a,0x280);
remainders_pos[638].modulo_value=2552; remainders_pos[638].bit_mask=create_bit_mask(80,0x820aa282,0x88028a20,0x8020228,0xa82200,0xa2a000a8,0x8088aa80,0x2228002a,0x8280a022,0xa0a02a0a,0x22aaa00a,0x208a222a,0x20a2820a,0xaa0a0828,0x228082aa,0x8220000,0x82080a2a,0x80aa8208,0x2a8a800,0xaa02800,0x282220a8,0x200a2028,0xaa882a00,0x8a880002,0x88228008,0xa02008a2,0x28aaa220,0x20882,0x2028a080,0x20a08202,0x22200a,0xa20002,0x20280a,0x220a0a82,0x8028a022,0xa000888,0xa80a8202,0xa8a22802,0x88222288,0x2208288,0x828000a2,0x808a0002,0x22228208,0x2a228888,0x2a80288a,0xa08082a0,0x2222000,0x88880a28,0x82a0a0,0xa02808,0x80008a,0x8a00888,0x880820a,0x20a28,0x28822080,0xa088aaa,0x228a2008,0xa2200288,0xaa800022,0x800a822,0x282808a0,0xa02a0888,0x8000280a,0x2002a2a,0x822082aa,0x20a8a020,0x88000088,0xaaaa8202,0x82820a0,0x8a0828a,0x88a888a2,0xaa00aaa,0x82a0a80a,0x88880a02,0x2a80028,0x8a02aa22,0x2a000a,0x2000882a,0x22288080,0x8208a280,0x828aa0);
remainders_neg[638].modulo_value=2552; remainders_neg[638].bit_mask=create_bit_mask(80,0xa82220a,0xa02a8,0x800a8aa0,0x8820aa80,0x2a280820,0x2208,0xaaa080a2,0xa08282a,0x2820a282,0x2a222882,0xa802aaa2,0xa82a0282,0x220280a0,0xaa000a22,0x80aa8880,0xa8002a2,0x220a80,0x8a202008,0x8228a008,0xa0a2a820,0x2082a8a0,0x2200a288,0x200808a,0x2a0880,0x28a8002a,0xa0222aa0,0x888a000a,0xa0a02808,0xa8280a82,0x88aaa802,0x8822888a,0x828a082,0xaa82820a,0x8a020aa,0x82088000,0x2082028a,0x202aa082,0xaa2a00,0x2a80a00,0xa08882a,0x802880a,0xaaa20a80,0x22a20000,0xa208a002,0x28080228,0x8a2aa888,0x8220,0x880a2820,0x88282080,0x80088802,0x80288000,0x80080a02,0x888282a0,0x200a2808,0x82800222,0xaa02a080,0x2a288a00,0x220888a2,0x808820a2,0xa0a00028,0x20228000,0x888a082,0x8a88a222,0xaa00a22,0x282020a8,0x888800,0x2222028a,0x20a828,0x80280a02,0x200022,0x82280222,0x2202082,0x828a,0x8a208820,0x28222aa,0x8a28802,0xa88800a2,0xaaa00008,0x2002a08,0xa0228);
remainders_pos[639].modulo_value=852; remainders_pos[639].bit_mask=create_bit_mask(27,0xa28008a2,0x8828028,0x2002208,0xa2082020,0x28020800,0x200820a,0x22080800,0x20a08a28,0x8280008,0x820800a2,0x8a08a08,0x80202200,0x22800002,0x28800000,0x80280000,0xa2000880,0x8220a20,0x8208a002,0xa0820002,0x8828a20,0x8000202,0x28a0820,0x8a00208,0x8008082,0x22020880,0x28a28028,0x8a200);
remainders_neg[639].modulo_value=852; remainders_neg[639].bit_mask=create_bit_mask(27,0x22080022,0x20820,0x8a08a200,0x228028a0,0x20828008,0x8a080282,0xa2800080,0x28200220,0x80208080,0x2800822,0x200200,0x828a280,0xa2080882,0x20008808,0x8208088,0x22880000,0xa28228,0xa00208a,0x200a0882,0x20228,0x8008828a,0x820200a0,0x208a00,0x8008000a,0xa28a0000,0x20220820,0x2288);
remainders_pos[640].modulo_value=40; remainders_pos[640].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[640].modulo_value=40; remainders_neg[640].bit_mask=create_bit_mask(2,0x882a82,0x20);
remainders_pos[641].modulo_value=641; remainders_pos[641].bit_mask=create_bit_mask(21,0x16556fb6,0x8b9e3139,0x2f83a7e3,0x6a6541dc,0xe4b5f60f,0x86d5e20d,0x10015352,0x96669633,0x773ca05d,0x7c10c5bb,0x768c20fb,0xe814f3bb,0x31a599a6,0x2b2a0023,0xc11ead85,0xc1beb49e,0xee0a995b,0x1f9707d0,0x7231e747,0xb7daa9a2,0x1);
remainders_neg[641].modulo_value=2564; remainders_neg[641].bit_mask=create_bit_mask(80,0x8a88a22a,0x202a8a0,0xa20a2a2a,0xa2a88800,0x28282a82,0xa082a80,0x88888a8a,0xaaa0aaa,0xaaa02880,0xa0880822,0xaa00a822,0x20882a22,0xa828002a,0xa2a288aa,0x8822008,0x8222802,0x22820082,0x820a8a08,0xaaa82a8a,0x2a02202a,0xa02202a0,0x8aa0aaaa,0x828a820a,0x8020a20,0xa22082,0x80220882,0xa88a2a28,0xa000a0aa,0x22a08822,0x20a802aa,0x2080882a,0x8a02aaa,0xaa82aa80,0x8a88888a,0xaa08282,0xaa0a0a0,0x88aa2a,0xa2a28228,0x28aa0202,0xa2288a88,0x80882202,0xa2a00828,0x22820202,0x20088aa8,0xa0a00a0,0x828a00aa,0x88888082,0x80028000,0x2a08aa,0x288a8a20,0x2a80a20,0x288a0220,0xa0aaa02,0x20208800,0x88a22a88,0x8a220aa2,0x20a2a8a2,0xa280828a,0xa0080,0x2a22a00,0x2a22a022,0x80280000,0xa080a280,0x8aa28228,0xa82228a2,0xaa2288a2,0x8820208,0x2aa82800,0x20288a20,0x280aa002,0x28a88a02,0x882a0002,0xa000aa,0x80888880,0x8028a0a0,0x8028282a,0xa8880202,0x2020a20a,0x802a2a0,0x2208808a);
remainders_pos[642].modulo_value=2568; remainders_pos[642].bit_mask=create_bit_mask(80,0x22020802,0x220808,0x8808808a,0x20a2020,0x28a00820,0x8a200282,0x820820a2,0x8820828,0x8228800a,0x828a08a2,0x20020200,0x8002282,0x200a0002,0x20a00a00,0x8208a202,0xa0082022,0x28220a08,0xa2882,0x8028a28,0xa082,0x800028a0,0x28a28a20,0x8020000a,0x22082800,0x20008200,0x80280202,0xa2802822,0x228820,0x8220a202,0x800082,0x28800200,0x8208080,0x2082080,0xa20000,0x8280002,0x200020a2,0x8808808,0x88202280,0xa0822020,0x8820a00,0x88082208,0x22022002,0x8000820,0x28208a,0x8a0080,0x8208000,0x2082002,0x80022802,0x20000,0x8a088282,0x22880080,0x28028a08,0x80280288,0x82000880,0x28208800,0x80200,0xa28a28a0,0x28000208,0xa00000a,0xa2802082,0x28a00028,0x82,0xa0882800,0x8200a20,0x8a208288,0xa00a0880,0xa00800,0x88002080,0x80800882,0x20a28200,0x228828a,0x208220a0,0x8208228,0x8008a28a,0x200a2882,0x8a08008,0x2202208,0x208800a2,0x20808820,0x80);
remainders_neg[642].modulo_value=2568; remainders_neg[642].bit_mask=create_bit_mask(80,0xa28a0082,0x8a28000,0x2,0x828220a0,0x20208028,0x228820a,0x2802822,0x28020,0xa200082,0x2020022,0x28828a08,0x8008a20a,0xa0820882,0x28208208,0xa00228a,0x208028a2,0x20228200,0x88088088,0x80822002,0x820220,0x8808200a,0x882020,0x20220228,0x8288082,0xa2802080,0x28800a08,0x820828a,0x220820a2,0x8a20028,0xa28228a,0x80080802,0x20008a08,0x80280008,0x82802800,0x8228808,0x8020808a,0xa0882822,0x28a208,0x200a28a0,0x28208,0xa280,0xa28a2882,0x800028,0x8820a002,0x80020800,0xa00808,0x8a00a08a,0x8a2082,0x8828808,0x200020a,0xa2000800,0x20820200,0x8208200,0x2880000,0x20a00008,0x80008288,0x22022020,0x20808a00,0x82088082,0x22082802,0x20208820,0x8808800a,0x20002080,0xa08228,0x2280200,0x20820000,0x8208008,0x8a008,0x80002,0x28220a08,0x8a200202,0xa00a2820,0xa00a20,0x8002202,0xa0822002,0x200800,0x8a28a280,0xa0000822,0x28000028,0x8);
remainders_pos[643].modulo_value=2572; remainders_pos[643].bit_mask=create_bit_mask(81,0x2a080a0a,0xa22880a,0xaa2888a,0x82880aaa,0x88a82002,0x22a2a02,0xa222228a,0x800aa22a,0x8a2a28aa,0xa8a28a2a,0xa000002,0xa0202a20,0x800a208a,0x828288a0,0x2a020022,0x8808088a,0xaa2a80a2,0x20088a,0x882202aa,0x802a08a0,0x2280aa02,0xa88a228,0x22a8aa80,0x2a008002,0x22a2a208,0xaa8a8082,0x220a0a08,0x2882aa28,0x80a8a802,0xaaaa82a8,0x8208200a,0x20808200,0x8802aa00,0x8888800,0x80808a82,0xa8a0220a,0x2a20a0a,0x22080280,0xa2088282,0x82a28082,0x28a82802,0x2208a820,0x2088828,0x8a80028,0x80a2aa0a,0x20202a08,0x2222082a,0xa8022002,0x2020800a,0x82082008,0x2aaaaa00,0xa2a022a8,0xa8228802,0xa08828a,0x2a2aa20a,0xa8a88820,0x200a8208,0xa2a88800,0xa22a002a,0xa0288288,0xa80280a,0x88220a2,0x800aa8,0x2a8aaa22,0x20228a0,0x80a8a22,0x28282880,0x8a0020a2,0x8080aa20,0xa008a,0x28a2a800,0x8a8a2a8a,0xaa002aa2,0x8888aa88,0x8a880a28,0x82a2280a,0x2282800,0x28aa0aaa,0x288a0a22,0x20a8a02,0xa0a);
remainders_neg[643].modulo_value=2572; remainders_neg[643].bit_mask=create_bit_mask(81,0xa2808282,0x82aa0082,0x822a0002,0xa008222,0x20a88a,0x8aa2a28a,0x2aaaaa02,0x8822aa2,0x2a2a022,0x202a02a2,0x8288888a,0x28a8a2a8,0x882a802,0xa0a0028,0xa28a88aa,0x808002,0x22a2082a,0x88a88002,0xaa8a22,0x8a28028,0xaa082282,0x82002aa0,0xaa202208,0xa288088a,0xaa2a2a80,0x2202080a,0xaa828280,0xa00a22a0,0x820208a,0x22220a20,0xa80a882,0xa8080a88,0x8a2288,0x80000088,0x808020a,0x2028aa82,0x8a2a8282,0xaa808a08,0x2a800a0a,0xa2a080a,0xa020a08a,0xaa8020a8,0x8a8000a0,0x802088a0,0x82a2282,0xa8a8a280,0xaaaa80a2,0x208aa88a,0xa8a80882,0xa80a880,0xa2222288,0x2a28aa20,0x20aa008a,0x82800a02,0xa2a22a82,0x202000a8,0xa8820a80,0x2a200088,0x2aa288a2,0x28a00a00,0x8208a082,0x800aa82a,0x88088220,0xa20222aa,0x8a8aa028,0x808202aa,0xa0a0a008,0x288a82a,0x80822a8,0x88828802,0xa02a2088,0x202a202,0x2288a22a,0x2200,0x20082a0,0xa2aa082,0x8aa0a088,0xa0228222,0xa00282aa,0x8a82028a,0x282);
remainders_pos[644].modulo_value=1288; remainders_pos[644].bit_mask=create_bit_mask(41,0x220a0222,0x20002082,0xa22a0080,0xaa008082,0xa282a002,0x2a0a28,0x80802aa,0x802a22a,0x8200020,0x22028222,0x8828088a,0x80008208,0x88a80200,0xa802020a,0x8a0a800a,0xa828a2,0x20200aa8,0x200a88a8,0x20800080,0x880a0888,0x20a02228,0x20822,0x22a00802,0xa008082a,0x282a002a,0x2a0a28a,0x80802aa0,0x802a22a0,0x82000200,0x20282220,0x828088a2,0x82088,0x8a802008,0x802020a8,0xa0a800aa,0xa828a28,0x200aa80,0xa88a82,0x8000802,0x80a08882,0x88);
remainders_neg[644].modulo_value=1288; remainders_neg[644].bit_mask=create_bit_mask(40,0xaa028a2a,0x2888a802,0x22a28808,0x2200080a,0x2a02288a,0x88a202a0,0x80800222,0x808a2a22,0xa888a0,0xa28a02aa,0x208002,0x8800a80,0x200a08,0x208a8282,0x2820802,0x820a022,0xa0a88220,0x28020020,0xa8008808,0x820000,0xa028a2a0,0x888a802a,0x2a288082,0x200080a2,0xa02288a2,0x8a202a02,0x8002228,0x8a2a228,0xa888a08,0x28a02aa0,0x208002a,0x8800a800,0x200a080,0x8a82820,0x28208022,0x820a0220,0xa882200,0x8020020a,0x80088082,0x820000a);
remainders_pos[645].modulo_value=2580; remainders_pos[645].bit_mask=create_bit_mask(81,0xa0820082,0x228020,0x2088280,0x2802800,0x20208808,0x8028228a,0x82020820,0x800020,0x208008,0x22000022,0x800200,0x8280202,0xa0,0x8820008,0x820a,0x80022800,0x20228a00,0x82200,0x80000022,0x208a00,0x88200202,0x802,0x20800220,0x8a202002,0xa2000802,0x8000000,0x2088282,0x80,0xa08028,0x80202,0x22002,0x8200220,0x80080,0x82020820,0x20a00820,0x8820228a,0xa00a0080,0x8820000,0xa20000a,0x2082820,0x8020a08,0x8a008282,0x220a0000,0xa00008,0x8220200a,0xa08a2880,0x8208200,0x8208,0x82002002,0x8800,0x80088,0xa0080802,0x28020,0x8002000,0x2082822,0x8a00000,0x8a280200,0x20880080,0x8800,0x82280200,0x80080800,0x200a20,0x88000,0x80800882,0x200a28,0x288,0x220a0820,0x20008,0x8200a000,0x20080802,0x8800800,0x80088000,0x20020020,0x8208000,0x80208208,0x808a2882,0x28020220,0x8000280,0x80002822,0x20a08028,0x82008);
remainders_neg[645].modulo_value=2580; remainders_neg[645].bit_mask=create_bit_mask(81,0x200a0802,0x8220020,0xa000208,0x82002080,0x20a00000,0x8202202,0x20a00a0,0x8008028,0x88200080,0xa2880822,0x8000a08,0x80208282,0x880020,0x28800,0x80080202,0x8a2080,0x28220200,0x8002288,0x8008a2,0xa00208,0x28028a,0x80080082,0x20008228,0x220a08a,0x22880002,0x800808,0x8a000202,0x880000,0x8200820,0x8000820a,0x808a2882,0x208a20,0x8008008,0x28200a0,0x20208028,0x282202,0x20020800,0x28008,0x2208082,0x82802020,0x820200,0x2080202,0x22820080,0x8208800,0x228a002,0x20022000,0x200a00,0x8080280,0x2802882,0x8,0x88000000,0x20000082,0x8820028,0x8000a088,0x82802022,0x200800,0x2208280,0x20000000,0x8800008,0x2208208,0x800080,0x8208220,0x8000088,0x2,0xa08220,0x80080200,0xa20200a0,0x8828000,0xa002088,0xa0800002,0x8,0x8000000,0x208a00a0,0xa00808,0x280200,0x22002,0x20028a20,0x88208,0x8020a2,0x20200820,0x2080);
remainders_pos[646].modulo_value=2584; remainders_pos[646].bit_mask=create_bit_mask(81,0xaa0a2aa,0xa0222a88,0x80a0822,0xaa2a0a82,0x80880282,0x20808202,0x8088a0aa,0xa2a000a,0x80220200,0x88a22828,0x82aa0022,0xa8aa8802,0x88a22008,0xa000820,0xa00802a2,0x28800828,0x20a22228,0x80a08028,0x22a28a2a,0x2a022080,0x20202222,0x22a8a,0x820a8020,0x20a280a,0x2a882022,0xa8a80828,0xaa20a200,0x2088a20,0x8008088a,0xa0028220,0x2208802a,0xa22a0028,0xa0008aa2,0x8a00aa,0x8a88200a,0xa88282a0,0xa0000202,0xa0288280,0x20220228,0xa880,0x800002,0x8022a00,0xa288088,0xa028228,0x2a808000,0xa20a8282,0xa00822,0xaaa00a2,0x8a8aa200,0x882800a8,0xaa80220,0x2088280,0x80a22020,0xaa08a220,0x2a008a08,0xa828202a,0x80880822,0x82a028a0,0x802a0,0x8a2a880,0xa8888808,0x88020880,0x2a8a28a,0x828020a,0x2828888a,0xa282002,0xa08a8020,0x22082000,0x2a20088a,0x828022aa,0x228800aa,0x228288a,0xa0008088,0x2a000a8,0x8aa0a22,0x2808202,0xaa828022,0x2082a0a8,0xa8820a0,0xa022a888,0xaa8a0a);
remainders_neg[646].modulo_value=2584; remainders_neg[646].bit_mask=create_bit_mask(81,0x82202a22,0x28a2a200,0x8280aa,0x2222820a,0x8008a0a,0xa008028a,0x8002822,0x2228882,0x8aa8a88,0x2aa0a0,0xa2288a2,0x2022008a,0x2a2800,0x828880a8,0x2080822a,0xa00880a0,0xa82a2aa0,0x8280820,0xaa2202a2,0xa282a808,0xa828aaaa,0x88aa202,0x28208a8,0x8a82a082,0x2200a0aa,0x202080a0,0x22282a88,0x8a8002a0,0x8808002,0x288a0a20,0xaa8008a2,0x2aa200a0,0x2888022a,0x80020822,0x200a882,0x20020a28,0x28888a0a,0x28200a08,0xa8a28aa0,0x80882008,0x808888a,0x8aa288,0x82a00800,0x828a02a0,0xa2080880,0x2a020a0a,0x8828802a,0x8222882a,0x2022280,0xa08820,0x82200aa8,0x8a800a08,0x22a8a8,0x22802aa8,0xa2080280,0x20a0a822,0x80aa,0xa20a028,0x8808a28,0x2a2008,0x20000080,0x8a8000,0x8a202202,0x80a08a02,0xa0200002,0x82a0a08a,0x280208a8,0xaa802880,0xa2a88002,0xa002a22,0xaa008822,0x8220a002,0x28880800,0x82288820,0x8022822a,0xa080a8a,0x220208aa,0xa80a2820,0x200a820,0x28aa2000,0x220202);
remainders_pos[647].modulo_value=2588; remainders_pos[647].bit_mask=create_bit_mask(81,0x22aaaa02,0x8228208,0x20020080,0x222a08a,0x2a808a8a,0x2a880a28,0xaa8a8288,0xa280a282,0x2a0a08aa,0x8882a822,0x2a02882a,0x2a8222a0,0x8000a882,0xaaa88a80,0xa820008a,0xa2a02808,0x88aa28a0,0xa0082288,0x28880aa,0xa8828a8a,0xa202208,0x220a8202,0xa2a8002a,0x80222208,0xa8082820,0xa8a022a0,0x2000022a,0x2aaa2a02,0x8a008a20,0xa80a808,0xa2200a2,0x828088a0,0x2a080022,0x2000a08,0xa200a20a,0x2a00a082,0x88a8202,0x8aa88228,0x8a2aa2a,0x8a28a,0x2a0a80,0xa28a200,0x2a8aa8a2,0x28822aa,0x88082a22,0xa820a00a,0xa08a008,0x820a0008,0x2880020a,0xa22028,0x8a0088a,0x2202a02a,0x808a200a,0x880a8aaa,0xaa880000,0xa0a880a2,0x20828202,0xa2088880,0x8a8002a8,0x8082a08,0xa208808a,0xa2a2822,0xaaa02228,0x22288200,0xa0a28aa2,0xa20280a8,0xaa200082,0x202a22aa,0x8822a000,0x80a8882a,0x2a82280a,0x8882a822,0xaaa20a0a,0xa828a028,0x82282a2a,0x828a022a,0xa2a202a,0x8a20a888,0x200800,0x82082882,0x80aaaa8);
remainders_neg[647].modulo_value=2588; remainders_neg[647].bit_mask=create_bit_mask(81,0xaa22228a,0x80aa0a80,0xa88a8808,0x8aaa2802,0xa2080202,0xa20082a0,0x22020a00,0x2a082a0a,0xa2828022,0xa20aa,0xa28a00a2,0xa20aaa28,0x888200a,0x22200208,0x20a88802,0x2a28a080,0x22a028,0x2880aa00,0x8a000822,0x200a0202,0x82a8aa00,0xaa820a8a,0x2a2088a2,0x8aaaa80,0x2080a0a8,0x2028aa28,0xa8888aa2,0xa222a28a,0x28802a8,0x82082080,0x82aa882a,0xa080028,0xa28088aa,0x8a888280,0x2a882a82,0xa288280a,0x80020a8a,0x2200aa0,0x802a22a2,0x88802a02,0x88a28208,0x82a02a88,0xa202202a,0x8a00aa22,0x80a2aa,0x20a82882,0x82802880,0xa828880,0xa0088a82,0x882aa8a0,0x80288002,0xaa8a28a2,0x802a882,0x820222,0x22008888,0x2820082a,0xa80a0a8a,0x2a800008,0x2088a20,0x8080a280,0x2a800802,0x82a2a0aa,0x2228aaa0,0xaaa00a88,0x282a022a,0x2a8a0820,0x22a8880a,0xa8a2aa22,0xaa2888,0x82000a2,0xa20aa082,0xa20aa,0x222a8282,0x20a028a0,0xaa0a2a2,0xa028aa2,0x82a2a8a2,0x2a82000,0x88a88088,0xa80a00a,0x822220);
remainders_pos[648].modulo_value=24; remainders_pos[648].bit_mask=create_bit_mask(1,0x820082);
remainders_neg[648].modulo_value=24; remainders_neg[648].bit_mask=create_bit_mask(1,0xa0802);
remainders_pos[649].modulo_value=2596; remainders_pos[649].bit_mask=create_bit_mask(82,0xa00a22a,0x20222880,0xa0a82,0x20a02020,0x80a802aa,0xa0220a,0x8aa80888,0xa28020a,0x8800020a,0x20828002,0xa8020aa8,0x88802228,0x8a2082a8,0x82228000,0x20200222,0xa882288,0xa0028280,0xa0882a2,0x2a82202,0xa2a28a22,0x2a2aa8aa,0xaa02228a,0x82820220,0x2a2a08,0x8a800a0a,0x20208a20,0x220a2200,0x2288000a,0x888aa08,0xa8a220,0x820aa82,0x8a000a,0xa2828200,0xaa8a8200,0x28008880,0xa80a8220,0x2822aa00,0x80002020,0x20220a82,0x28008a2,0x88aaa228,0xa2002802,0x2a088088,0x80800028,0xaa88280,0x882a02a0,0x22200280,0x82a2aa0,0x82828a0,0xa2000,0x2aa0820a,0x88a2a008,0xaa22200,0x22882,0x88a088a,0x8a208080,0xa002a20,0xa8a800a,0x88082822,0x28880aa0,0xa2aa8a8a,0x8a28a8aa,0x882a808,0xa8220a08,0x282800a8,0x28822a00,0x88008082,0x288828,0xa8208a20,0x88802222,0xaa0802a2,0x282082,0x8000228,0x8028a0a,0x2202aa2a,0x880a002,0xa802a02a,0x8080a08a,0x2a0a0000,0x22888088,0x88a00a00,0xa);
remainders_neg[649].modulo_value=2596; remainders_neg[649].bit_mask=create_bit_mask(82,0x828822a2,0xa02aa008,0x8882820a,0xa828a8a0,0x8200a22,0x8028aa82,0x2208000,0x82a08a02,0x808a82,0x280a080a,0x208a8220,0x8a2a0,0x2280a20,0xaaa0888,0xa8a88aa2,0x82002a00,0x20820a08,0x82800a2a,0x8a20aa0a,0x2a2202aa,0x22a22022,0x228aaa02,0xa0a82a8,0x822a280,0x2088282,0xa8a802a0,0xaa822a88,0xa2008882,0x80002280,0x88202a28,0x80a0220a,0x8028082,0x2a0a0a88,0x22020288,0xa0080008,0x20820aa8,0xa0aa2280,0x88828a8,0xa02a820a,0x8a08802a,0x222a20,0x2a80a08a,0x22800800,0x80888a0,0x82200208,0x228a28,0xaaa88a00,0x80a2a220,0x80a02028,0x8082a888,0xa2280a82,0x2a2800,0x8222aa88,0x88a200a,0x80028002,0x2a80008,0x8208a2a8,0x82020882,0x80a0a2,0xa0000228,0x22220202,0x2a02022,0x800a2000,0x20a28280,0x20a08820,0xa00aa288,0x88000a,0x882000a0,0x20a80228,0x8aaa2,0x22800a2a,0x80a0a80a,0x80888aa0,0x808a0202,0xaa8222a2,0x288a,0x208a28a2,0x8082002,0xa2028888,0xaa000800,0x288280,0x2);
remainders_pos[650].modulo_value=520; remainders_pos[650].bit_mask=create_bit_mask(17,0xaa0a02,0x8020020,0x208a8008,0x8a022088,0x82202002,0xa8800808,0xa2288a0,0x200002a8,0x80000882,0x2288a02a,0x20022a0a,0x8088220,0x880a280,0x2a20822,0x802020,0xa0aa0008,0x80);
remainders_neg[650].modulo_value=520; remainders_neg[650].bit_mask=create_bit_mask(16,0x8822028a,0x800a8820,0x20020080,0x282a800,0x2a8280a,0x20080080,0x822a0020,0x28088220,0x880800a,0xa2002022,0x288a2282,0x80000aa0,0x2208,0x8a2280aa,0x8008a828,0x20220880);
remainders_pos[651].modulo_value=2604; remainders_pos[651].bit_mask=create_bit_mask(82,0x22082022,0x20200a00,0x80088282,0xa20028a0,0x820200,0x88202280,0x802022,0x28820020,0x8000202,0x20a0022,0x820,0x82202288,0x828208a0,0x8828020,0x8008002,0x2080080,0x20a20228,0xa00a082,0xa2820800,0x8a20000,0x80282,0xa2800,0xa08000,0x8000a,0x82000800,0x20a20a00,0x2082208,0x82002,0x8000200,0xa200080,0x828800a0,0x20008800,0x828a,0x2882880,0x20008028,0x2080008,0x80882,0x20200028,0x88080002,0x80020080,0x820,0x8002082,0x2022020,0x808800,0x2000280,0x2080822,0x20008200,0x82280280,0x20000022,0x20008a28,0x2282002,0x8a00a0,0x20020,0x82000008,0x82080800,0x8a08208,0x8200a,0x2000002,0x20a00a00,0x8a000000,0x2000002,0x8a20828,0x828a000,0xa00a0002,0x8a08820,0x2080288,0x20020020,0x28220800,0x8282080,0x808820a2,0x228,0xa080082,0x20880,0x8228808,0x80200080,0x80822880,0x8200028,0x8008a008,0x220020a2,0x808828,0x8208800a,0x880);
remainders_neg[651].modulo_value=2604; remainders_neg[651].bit_mask=create_bit_mask(81,0x22802822,0x28a08200,0x20a,0x22082020,0x28a08,0x282208,0x800820a2,0x20028828,0x88282,0x828208a2,0x8008028,0xa20a200,0x2020020,0x20028,0x8008000a,0x2800800,0x20220a20,0x8208200a,0x22020080,0x220808,0x8800820a,0x80822000,0x8200800,0x80008082,0x2080000,0x28228208,0x8a002280,0x80802082,0x808a00,0x2288000,0x2000820,0x28000008,0x88000202,0x82002000,0x28800020,0x82008000,0x800002,0x20a08820,0x808a,0x20800,0x8800028,0x8008a00a,0x828a2820,0x8000000,0x82088208,0x820000a2,0x28800a08,0xa200208,0xa08800a2,0x28800220,0xa20a082,0x80020020,0x8028828,0xa008080,0x2800080,0x200200,0x8800a002,0x2880882,0x20208208,0x2088080,0x82800882,0x220020,0x202088,0x20820802,0x200020,0x82008200,0xa00a08a0,0x20a28008,0x80202008,0x2022,0x8008a20,0x2008002,0x808a0000,0x220000,0x8208008,0xa2000,0xa08020,0x8002000,0x22882822,0x20,0xa000082);
remainders_pos[652].modulo_value=1304; remainders_pos[652].bit_mask=create_bit_mask(41,0x8aa80a8a,0xaa220202,0x2022a82a,0x8a02088a,0x800a0820,0xa282aa02,0x228a8228,0xa008a882,0x8a888000,0x2a0020a,0xa80802,0x222a2a08,0xa200a000,0x2a288822,0xa828a288,0xa00280a,0x80a2082,0xa8808a22,0x88aaa82,0x2aa2808,0x80a8aa2a,0x202028aa,0x2a82aaa2,0x2088a202,0xa08208a0,0x2aa02800,0xa8228a28,0x8a882228,0x88000a00,0x20a8a8,0x8080202a,0xa2a0800a,0xa000222,0x88822a20,0x8a2882a2,0x280aa82,0xa20820a0,0x8a22080,0xaaa82a88,0xa2808088,0xa2a02a);
remainders_neg[652].modulo_value=1304; remainders_neg[652].bit_mask=create_bit_mask(41,0x2208202,0x22aa8a8a,0xa8aa20a2,0x28a8002,0x88280a8,0x2a0a2282,0xaa020aa0,0x2880200a,0x2000888,0x8a288a82,0x8820808a,0xaaa2a280,0x2a882888,0xa2a000aa,0x20a02a00,0x8288a082,0x8082a80a,0x200802aa,0x8002220a,0x8a22a080,0x82022a2,0xa8a8a022,0xa20a222a,0xa8002a8a,0x280a8028,0xa2282088,0x20aa02a0,0x200aaa0,0x888288,0x88a82020,0x808a8a2,0x2a280882,0x82888aaa,0xaa2a8,0x2a00a2a,0x8a08220a,0x2a80a828,0x802aa808,0x2220a200,0x2a080800,0x2a28a2);
remainders_pos[653].modulo_value=653; remainders_pos[653].bit_mask=create_bit_mask(21,0x9389aed2,0x9c7a9514,0x1ce6f645,0x9e27ea39,0x20c3c7ee,0xd68d8beb,0x3bbf2bc,0x36815342,0x25c82903,0x537f0d7a,0x3fb28330,0x4e917ac,0xa05b3025,0xf77010b2,0x6c5acf53,0xf0c135f4,0xf91e5df8,0xd9ce2715,0x578e689b,0x64724a2a,0x12dd);
remainders_neg[653].modulo_value=2612; remainders_neg[653].bit_mask=create_bit_mask(82,0xa00220a,0xa20888,0x802a2a88,0x2aa22a0,0xa80a0a22,0xa000222,0x8a222a20,0xaa088a8a,0xa800a08a,0x8aa280a2,0xa22a0aa8,0x88208a20,0x2882a8a8,0x2aa8882a,0xa082028a,0x2808a828,0x20828020,0x2aa88,0x8a02a002,0xa8aa82a2,0xa8028a00,0x8a2a0aa8,0x2002a02,0x208aaa00,0xa0200a08,0x28a0a880,0xa28a0208,0xa2a088aa,0x88a8aa08,0x28228820,0x88aa82a2,0xaa280a2a,0xaa882800,0x8a8a8882,0x8022a222,0xaa220002,0x2228280,0x82a22aa,0x8aa2a0,0x80888228,0x82822002,0x8a82aa22,0x8aa82288,0x28aa0200,0x22a00220,0x200a8282,0x2a82aaa2,0x80822202,0x80028880,0x200aa828,0x88020a8,0x28220280,0x88a2882,0x208a008,0x8200088a,0x828a2a0,0x2a0a880a,0x8a28a0aa,0xa2aaa000,0x2082a02a,0xa80800a0,0x80aa002,0xa0820280,0xaaa2aa02,0x8a288002,0xa82a2a82,0x8a082808,0x2082a2,0x88202888,0x28880802,0x220a208a,0x8800a0,0xa8020a82,0xa0008a0a,0x20808088,0xa0aa2022,0xa80222aa,0x2a220a0,0x2a8a0220,0xaa88020,0xa0a888a2,0xa22a);
remainders_pos[654].modulo_value=2616; remainders_pos[654].bit_mask=create_bit_mask(82,0x22022822,0x8820a20,0x208282,0xa2000882,0x828020,0x8a282202,0x228820a2,0x8208020,0x288082,0x808800a2,0xa20,0x208a208,0x2802800,0x228a08,0x8220220a,0x200828a2,0x8a20a08,0xa08a000,0xa0002820,0x28a28a20,0x2002282,0x2000000,0x28200a28,0x82280080,0x20802000,0x828,0x80202202,0xa0800022,0x28a00a00,0x88200082,0x820028a2,0x20028808,0x8208800a,0x220a0820,0x20800,0x82202000,0x20a0082,0x20800a20,0x8200820a,0x22002080,0x200220,0x88088008,0x82020800,0x8a08200,0x8008a002,0x828200a0,0x808,0x88002080,0x820820a0,0x8a00220,0x82202280,0x228a2800,0x28820008,0xa00a00a,0x2880002,0x808808,0x8282000,0x82000802,0x28020028,0x8028a008,0x80000000,0x28828800,0xa08a28a,0xa0082800,0x20000a20,0x820a08a,0x828a2820,0xa08808,0x8020a288,0x80002802,0x208a20,0x208a000,0x8a0022,0x20820228,0x88080208,0xa28a0822,0x808828,0x8a080282,0x822000,0x20828208,0x8808a082,0x882880);
remainders_neg[654].modulo_value=2616; remainders_neg[654].bit_mask=create_bit_mask(82,0xa28a20a2,0x28228,0x8828020a,0x22880002,0x8020828,0x220a28a,0xa2002822,0xa00828,0x8820000a,0x822,0x8808228,0x8a002280,0x82082080,0x8a20200,0xa28a282,0xa0802022,0x228200,0x82002088,0x208820a0,0x20220228,0x8a08a20a,0x82880880,0x20a08220,0x2208008,0xa0082880,0x8808020,0x828a28a,0x200808a2,0x20208208,0x28800a,0x2882022,0x28820000,0xa000082,0xa28200a0,0x8828008,0xa28a088,0x82820802,0x28000228,0xa080282,0xa2882800,0x8a08a28,0x80,0x28a0080,0x200a08,0x800208a,0x20a0820,0x8808000,0x8a008,0x2802820,0x208a28,0xa28a208,0xa2022080,0x20028800,0x82082082,0x82000882,0x8000000,0x8020a088,0x2880082,0x20828820,0x8202080,0x880880,0x20020008,0x82002202,0x20802080,0x28008228,0x80282002,0x20220a0,0x8200000,0x8282200,0x882082,0x8a00228,0x8a002088,0x800208a2,0x28028a20,0x8280,0x220200a2,0x8000020,0x200820a,0x800a2080,0x28020a00,0x200a,0x2000);
remainders_pos[655].modulo_value=2620; remainders_pos[655].bit_mask=create_bit_mask(82,0x882a028a,0xa20a0a20,0x22028280,0xa80a822,0x22088a22,0x20802a80,0xa2000a0,0x28082000,0x8880288a,0x8208a82a,0x80288a,0x82228aa0,0xa008008,0x202282a,0x80880800,0x882a0880,0x280a2,0xa880220,0x222a880,0x80208a08,0x8020a82,0xa20208a2,0x2a000088,0x8882a002,0x280a2a,0x28220082,0xaa228222,0x808288,0x2822882,0xa000080a,0x202a0a,0x200282,0x200a220,0x8aa0a0a0,0xa028a802,0xa8002a08,0x80020280,0x8802200,0xa020888,0x20808208,0xa0082,0x88200a00,0x2082020,0x222080a,0x20088022,0xa0280800,0xa20a8002,0x2802a280,0xa0a0aa,0x88a008,0x8280080,0xaa0a8080,0xa020000,0x8228828,0xa2282020,0x8888288a,0x8200882,0x2a8a0280,0x8800a822,0xa220000a,0x8a20808,0x282a0802,0x20a2080,0x22a888,0x88022a,0x28a02800,0x20220a82,0x20220,0xa828808,0x2200200a,0xaa2888,0x2a228020,0x2a82a208,0x8a228022,0x8202,0x80a0008a,0x802a8020,0x88a2208,0x8882a02a,0xa0282808,0x208a0a08,0xa280a82);
remainders_neg[655].modulo_value=2620; remainders_neg[655].bit_mask=create_bit_mask(82,0xa20a02,0x2a0282a0,0x2a8a0208,0x820020aa,0xa2800222,0xa8082208,0x82288820,0x2080a088,0xa002,0x2802022,0x88082002,0xa2a0220,0x2880080,0x8a82a0a2,0x8008,0xa20008,0x880a0822,0x20082a8,0x8aa22008,0xa80200,0x808a020a,0x2a0a8022,0x22888000,0x2288a,0x80a08222,0xa0aa080a,0x222a0aa2,0x8080200,0x8a02a00a,0x20888002,0x88a82282,0x88288a02,0xa8822a8,0x2202828,0x20a0200a,0x20882280,0xa8a00,0x8a288,0x82828000,0xa0080a00,0x8882080a,0x288280,0xa80a0a8,0x8aa28082,0xa080082a,0x28a00088,0x2a020882,0x208a2208,0x88202822,0x80002800,0x80a00808,0x22020800,0x28a8088,0x80a20020,0x22a0a828,0x2002,0x80288002,0x22028208,0x8020aa,0x22a88802,0x802a0080,0xa0228082,0xa82a008,0x88a22000,0x80008a22,0xa0282088,0xa82a8202,0x88a82a8,0x82020080,0xa280a802,0x88222000,0xa22a08a0,0x220a2280,0x2a208aa,0x80880a0a,0x8280802,0x82208a0,0x2a280,0x228a2,0x20a0a000,0xa8020280,0x2208202);
remainders_pos[656].modulo_value=328; remainders_pos[656].bit_mask=create_bit_mask(11,0x82a00222,0x2a0a28a2,0x88a2200,0xa28a0a80,0x8800a828,0x2a002228,0xa0a28a28,0x88a22002,0x28a0a800,0x800a828a,0x88);
remainders_neg[656].modulo_value=328; remainders_neg[656].bit_mask=create_bit_mask(10,0xa288aaa,0xa282a02a,0x8002aa88,0x22028208,0x8820a0,0xa288aaa0,0x282a02a0,0x2aa88a,0x20282088,0x8820a02);
remainders_pos[657].modulo_value=876; remainders_pos[657].bit_mask=create_bit_mask(28,0x2880002,0x20820228,0xa20a08a,0x82802002,0x8228200,0x8202200,0xa2880800,0x8000808,0x8002202,0xa08020a2,0x28200200,0x82288282,0xa00a2020,0xa00008,0x8a00a200,0x22882080,0x828828,0x8020a008,0x800208a0,0x20808,0x228a202,0x80820002,0x28820008,0x80282008,0xa08a0800,0x8208a20,0x2280288,0x800);
remainders_neg[657].modulo_value=876; remainders_neg[657].bit_mask=create_bit_mask(28,0x82000882,0x28028a20,0x82282002,0x2082882,0xa20a08,0x8028a288,0x22000080,0x808000,0x8008a28a,0x20082822,0x20a08a08,0xa20020a,0x208228a0,0x8208800,0x2082288,0x22002800,0x8020020,0x8282080,0x8a0020,0x8828000,0x8a20228a,0xa0882,0x20028800,0x820a080,0x20020080,0xa00220,0x8a208200,0x80);
remainders_pos[658].modulo_value=2632; remainders_pos[658].bit_mask=create_bit_mask(83,0xaa000a0a,0x28000a02,0xa0aa8a8a,0xa082828,0x2000000a,0x828a2288,0xa8280288,0x8a88a002,0x288822a2,0xa2888a02,0x222008a,0x228a00,0x20222a8,0x28228020,0x2800208,0x8a802a2,0x20a0888,0x20202828,0x88220202,0x8a220a00,0xa0a20208,0xa0822008,0x22a02880,0x202a0a22,0x820a0a8,0x2a00888,0x82aa02a0,0x8822028,0x202a22a,0x20802aa0,0x82a202aa,0xa0088000,0x2208802,0x88080080,0x80a82002,0x82008208,0x2a28aa20,0x202880a0,0x80000808,0xa082a800,0x80aa2000,0x8aa0200,0x2a820a00,0x20000200,0x2280820,0xaa28a80a,0x82008208,0x82a0220,0x202280,0x22088002,0x2200a80,0x808a8200,0xa80208aa,0x8a80800a,0x88220a8,0x80aa8228,0x200a800a,0xa082022,0xa0a8082a,0x280a8888,0x8820a02,0x808a0a20,0xa088a220,0x80882200,0x28080880,0x20a08028,0x802a2022,0x8002808a,0x2882820,0x88808008,0xa288002a,0x888000,0xa2228aa2,0x88222880,0xa22a28a,0x80282a80,0x88a28222,0x822,0x2820a0a0,0xa2aa0a28,0xa00028a2,0xa000aa80,0xa0);
remainders_neg[658].modulo_value=2632; remainders_neg[658].bit_mask=create_bit_mask(83,0x22888202,0x20880282,0x20220202,0x8200a0a0,0xa8808882,0xa022a00,0x20a08200,0x200200a,0x2000aa22,0x2200028a,0x8a2a8802,0x88a20288,0x8a8a2a20,0xa0aa0028,0x8a088a00,0x208a22,0x82828000,0xa828a0a0,0xa28a8a,0x2aa0280,0x282a8280,0x280aa800,0x2a28a000,0xa0a282aa,0x282820,0x8a208000,0xa220a28,0x800aa0a0,0x8a8a2a22,0x2808a220,0x22a8a22,0x28000888,0x8aa0008a,0x800808,0x820a08a,0xa880a00,0x222022a0,0xa0a00828,0x8088080,0x28022088,0x8222888,0x80228288,0xa2028208,0x28888a80,0x82a080a8,0x22202082,0xa800a80,0x80a20aa8,0x88a82208,0xaa80080a,0xaa88200,0x20a88,0x200a8022,0x2000882,0x800a2020,0x82202a0,0xa8820802,0x280a8a2,0x202080a2,0xa0020000,0x8002820a,0x80202a8,0x280022a8,0x800aa08,0x20808000,0xa02808a0,0x822a8a2,0x8820802,0x8a0020a8,0x80080,0x2a008822,0x8000880,0x22aa022a,0x2aa008,0x82a22a02,0x8a02208,0x2a02aa,0x8088802a,0x20a82820,0x222282a0,0x2808a02a,0x28802208,0x28);
remainders_pos[659].modulo_value=2636; remainders_pos[659].bit_mask=create_bit_mask(83,0x820222a2,0xa202a82a,0xaaa02aa,0x2aa02080,0xaaa00008,0x8222a220,0xa2820080,0x2022288a,0x28802088,0xaa02a22a,0x288aa22,0xa820a28a,0x8a288028,0x22a8aa8a,0x8028a088,0xa0a20a0a,0x2202820a,0x282a0a2,0xa88a82a8,0xa222a0aa,0x802a28a,0x28088802,0xa022000,0x280a0aa0,0x8a0a8888,0x82882802,0x2820aa02,0x2008a2,0xaa208202,0x828a020,0x80220a82,0x880a8008,0x28aa2080,0x2088a8a2,0x2a8a0802,0x88088a2a,0xaaa80028,0x28a800a2,0xa8002aa,0xa00a8800,0x88a8a00,0x2a220208,0x2a00a00a,0x2a0002,0x2a28aa8,0x2aaa8a0,0x22022280,0xa2a8a8a,0xa2208802,0x8aa288a2,0x2a022020,0x8802200,0xa282082a,0x208aa080,0x800808,0xa08288a2,0x8228280a,0x2a0a2802,0xa028222,0x880a00aa,0x22028000,0xa8020802,0xa888aa28,0x2a22aaa0,0xa82802a8,0x280888a0,0x8a0aa08,0xa2802a0a,0xa2a88220,0x228a2a2a,0xa082a280,0xa2280a28,0xa80aa88a,0x80228a88,0x88808220,0x828aa22,0xa8882020,0xaaa088,0x80aa8200,0xaaa0020,0xa808aaa8,0x88082a82,0x8a8);
remainders_neg[659].modulo_value=2636; remainders_neg[659].bit_mask=create_bit_mask(83,0xa8aaa2a,0x2a8a20a2,0x82228a22,0xa228a808,0x22288880,0xaaa2aa8,0x2a0a8808,0xa8aaa002,0xa008a800,0x228a2aa2,0x8a0022aa,0x20a82a02,0x2a008a0,0xaa202202,0x8a02800,0x282a8282,0xaa8a0a82,0x8a0a282a,0x20020a20,0x2aaa2822,0x80822a02,0xa080008a,0x828aa888,0xa0828228,0x2820000,0xa00a08a,0xa0a8228a,0x88a8802a,0x22a80a8a,0x80a028a8,0x8aa820a,0x820880,0xa022a808,0xa800202a,0xa202808a,0x8002a2,0x222088a0,0xa020882a,0x82088a22,0x28820088,0x80020288,0xa2aa8a80,0xa2882882,0x88a2888a,0x8a2a0220,0x8a222028,0xaa8aaa08,0x82a20202,0x2aa8008a,0x22a002a,0xa28aa8a8,0x8008aa88,0x2a0a80a2,0xa8022808,0x88088080,0x280a002a,0xaa0a082,0xa282a08a,0x828a0aaa,0x828822,0xaa8a0888,0x208a808a,0x200022a0,0xa2aa2228,0x20a08a20,0xa0800028,0x80282280,0x2a08a282,0x2a200aa8,0xaa02a2a2,0x280a2a08,0x2aa082a0,0x20822002,0x8aa0200,0x80aa8,0x80a022aa,0x2000a8a8,0x88222800,0x8220a88,0x822288a8,0x20802220,0x80a20a,0x20);
remainders_pos[660].modulo_value=1320; remainders_pos[660].bit_mask=create_bit_mask(42,0xa0802082,0x228a00,0x8000200,0x80020020,0x200208,0x200202,0x2000882,0x8020,0x8a200002,0x20082802,0x820a08,0x8a280282,0x80000,0x8008020,0x80082200,0x80080800,0x220800,0x2008008,0x80000800,0x20a00a28,0x8282080,0xa00a0802,0x20000228,0x2008000,0x20880020,0x20200200,0x88200200,0x2002000,0x200008,0x8028a200,0xa0820082,0x28200820,0x8a280,0x2000080,0x20008008,0x80080082,0x80080080,0x800220,0x80002008,0xa2880000,0x8020a00,0x82);
remainders_neg[660].modulo_value=1320; remainders_neg[660].bit_mask=create_bit_mask(42,0x20082002,0x8220200,0x88288,0x8208a0,0xa00a00,0x8028028a,0x82080002,0x8800028,0x220808a,0x20802002,0x8020200,0x2208202,0x80,0x800828,0xa200,0x800080,0x228000,0xa080080,0x800000,0x20208220,0x202008,0x20020082,0x28808220,0x8a000088,0xa0000820,0x28a00a00,0x280280,0x2882080,0x8a08800,0x202208,0x20020802,0x20208020,0x8002208,0x82800000,0x20000800,0x800000a,0x800,0x8000228,0xa080,0x22000800,0x820208,0x2);
remainders_pos[661].modulo_value=661; remainders_pos[661].bit_mask=create_bit_mask(21,0xae13da3a,0x19abfe52,0x28a34e51,0xfe14391a,0xef0b64f6,0x591f337a,0x589474b1,0x80a9de6f,0x49cf8269,0x83805f9a,0x70724093,0xe4967e80,0x4065907c,0x46bd9ee5,0x26a34b8a,0x3dd7b33e,0x1fdbc9b4,0x4516270a,0x66229cb1,0x1d529ff5,0x1716f2);
remainders_neg[661].modulo_value=2644; remainders_neg[661].bit_mask=create_bit_mask(83,0x228a02a2,0x8022228a,0xa02a8288,0x2288a082,0x2282a82a,0x8082aaa2,0x8008a828,0x82002a2,0x80020aa0,0xa088202,0xa8aa880a,0x280aa208,0x88a808a0,0x8a200228,0xaa2a8202,0xa00a2aa2,0x82020028,0x888aaa82,0xaaaa0028,0x808a0228,0x8a8a8a2a,0x2880aa2,0x2aaa8a2,0xaa8888a0,0x2020a0a,0xa28028a0,0xaa2aa2a,0x228a02,0x80a888a2,0x2a80a028,0x8aa8a882,0x8828280,0x82000a02,0x20802a,0xa8800a2a,0xaa0808a0,0xaa0a222a,0x288a22a0,0xaa02a08,0x2220088a,0x28a228a,0x82a0222a,0x22208220,0xa088aa,0x8828a02a,0xa00a0220,0xa0002220,0x880a08a8,0x2aa020aa,0xa2802a8a,0x88a2a0aa,0x8a8282,0x80228808,0xa882a0a,0x2aa20888,0xa2a082,0x82002002,0xa2aa082a,0x8000a0a2,0x2aa0888,0x82a20800,0x8200a8,0x8a802080,0x822a0,0x88882aa0,0xa2828000,0xaa082aa2,0x20020020,0x2082a280,0x88822aa,0xa82a08a8,0x808a200,0xa0a0a880,0xaa82a288,0x28aa00a2,0xaa8202aa,0x8a882808,0x82220002,0x82202802,0x2a028a08,0x2a888280,0x82208222,0x202a0);
remainders_pos[662].modulo_value=2648; remainders_pos[662].bit_mask=create_bit_mask(83,0x228aa282,0xa0828822,0x2a0a0008,0xaa20220,0x8a8008a8,0x2a8a0a,0x2820a88a,0x808222a2,0x2a2002a8,0x22a02a8a,0x2aa222aa,0xa2a0228a,0x802288a,0x82aa2aa8,0x88a0020a,0x2020a022,0x20820088,0xa28aa8aa,0xa0088a0,0x2aaa8028,0x28282a20,0x80002a20,0x8aa0a82a,0x820a08,0xa28a2008,0xa2a288a,0xaa08822a,0x200280a2,0x22a8a800,0x2a020088,0x22002082,0x2a022000,0xaa202080,0x228a8802,0xa282202,0x802aa088,0xaa8080a0,0xa200a808,0xa0a02a22,0x828a08aa,0x822228,0x828282,0x82288882,0xaaa20a2,0x8a88a80a,0xaa202a00,0x20a0202,0xa0220aa8,0x88808828,0xaa8022a2,0xa8020808,0x88000880,0xa8820800,0x88220080,0x8002a2a,0xaa8a0280,0xa0a88220,0x8aa228a8,0x2008a2,0xa220a082,0x2a82a0a,0x2808a800,0xa808a828,0xa02802aa,0x8a0a2200,0x8aa2aa2,0x8220082,0x22880a08,0x82a0800a,0x202aa8aa,0x8aa22880,0xa8a2880a,0x88aa888a,0xa8a2a80a,0x22a8008,0x288a8882,0xa22a08,0xa2a0a2a8,0xa02a2002,0xa808808a,0xa2000a0,0x88882282,0x828aa2);
remainders_neg[662].modulo_value=2648; remainders_neg[662].bit_mask=create_bit_mask(83,0xaa022a0a,0x280a00aa,0xa2828880,0x822a8aa8,0x2088020,0x88a20282,0xa0a82002,0x80aaa2a,0xa2a88a20,0xaa28a202,0xa22aa222,0x2a28aa02,0x808aa002,0xa22a220,0x288a82,0xa8a828aa,0xa80a8800,0x2a022022,0x82880028,0xa22208a0,0xa0a0a2a8,0x888a2a8,0x22820a2,0x880a8280,0x2a02a880,0x82a2a002,0x22800aa2,0xa88a082a,0xaa202088,0xa28a8800,0xaa88a80a,0xa28aa888,0x22a8a808,0xaa02008a,0x82a0aa8a,0x8a22800,0x22080828,0x2a882080,0x2828a2aa,0xa028022,0x880aaaa0,0x880a0a0a,0xaa0000a,0x8222a82a,0x2002082,0x22a8a288,0x8a828a8a,0x28aa8220,0x800a0,0x2208aa2a,0x208a8080,0x88008,0x200a8088,0xaa8808,0x8088a2a2,0x22028a08,0x28200aa8,0x22aa020,0x88a8802a,0x2aa8280a,0x8a20a282,0xa0802088,0x208020a0,0x28a08a22,0x282aa88,0x8022a22a,0x80aa880a,0xaa008280,0xa280882,0xa8a22022,0x22aa008,0x202a0082,0x220002,0x202a2082,0x8aa20880,0xa002000a,0x882aa280,0x2a282a20,0x28a2a88a,0x20800802,0x82a88828,0xaa0a,0xa022a);
remainders_pos[663].modulo_value=2652; remainders_pos[663].bit_mask=create_bit_mask(83,0x22080802,0x20028,0x2008288,0x20a2002,0x28a00a00,0x80082082,0x828a2082,0x8020a08,0x8a08a280,0x800a20a0,0xa00,0x288202,0xa0802002,0x8828800,0x8808a000,0x22000800,0x28820208,0x8082088,0x80820020,0x8820000,0x208a082,0x80000820,0x220228,0x220a08a,0x2020,0x28808208,0x2200280,0x2022820,0x20200a00,0x88208,0xa28a0882,0xa20000,0x2000000,0x82022800,0x200,0x82200208,0x20000800,0x28a00,0x8008288,0x828a0002,0x808028,0x280280,0x22802020,0x8000a28,0x2282002,0x800a2800,0x20020000,0x2080088,0x20080000,0x28808,0x8,0xa00008a0,0x8220a28,0x82082200,0xa0080,0x828808,0x80280088,0x2082022,0x808000,0xa20a088,0x22880880,0x820000,0x820a208,0x20000822,0x800820,0x82208202,0x82080822,0x20020008,0xa202,0xa0022822,0x8008020,0x8082280,0x200a0000,0x20a08a00,0x28a20a,0x220a0802,0x28208a28,0x88208200,0xa00a2,0x8008a08,0x2282008,0x82800800,0x8020208);
remainders_neg[663].modulo_value=2652; remainders_neg[663].bit_mask=create_bit_mask(83,0xa2800082,0x8828820,0x82080200,0x82022882,0x20200208,0xa00a,0x2022802,0x28200,0x2002208,0x822020,0x8808200,0x8820028a,0x20002802,0x20008,0x2008,0xa2880080,0x20020a00,0x80002000,0x208a0,0x28800,0x8a00200a,0x8800a0,0xa28220,0xa282002,0x8828a0,0x20000a00,0x8a288208,0x828a2020,0x20a08208,0x88000200,0x22020002,0x8228008,0x8a088080,0x20a2080,0x8808a08,0xa208280,0x20880080,0x8020208,0x80000200,0x2020882,0x8000820,0x208208,0xa20828a0,0x800220,0x8a20a082,0x822080,0x28028808,0x8a008000,0xa0800080,0x8820000,0x88080080,0x20800020,0xa28220,0xa00a208,0x80020800,0x8020000,0x200000,0x828028a2,0x8000800,0x82282000,0xa2000000,0x8028000,0x282280,0xa0880022,0x8008028,0xa28028a,0x2800022,0x20820800,0x8808228a,0x208220a2,0x800828,0x8000a200,0xa0020880,0x20200208,0x88202282,0x22820082,0x20a00220,0x200288,0x80820822,0x800200,0x8220a000,0x2080080,0x820200);
remainders_pos[664].modulo_value=664; remainders_pos[664].bit_mask=create_bit_mask(21,0xa02aa2a,0x8aaa282,0x228882,0x2a2200a8,0xa228a0a8,0x8220222,0xa220280a,0x802aa80a,0x8a288,0x2a0a282,0x80080020,0x828a0a,0x2228a20,0x8aa02aa8,0x20a02808,0x8a888088,0xa82a0a28,0x2a0088,0x20822288,0xa0828aaa,0xa8aa80);
remainders_neg[664].modulo_value=664; remainders_neg[664].bit_mask=create_bit_mask(21,0x828a22a2,0x80222a0a,0x88a2000a,0xa2aa8820,0x2aa02820,0x80aa8aaa,0x2aa8a082,0x8a22082,0x88802a00,0x8a282a0a,0x88088a8,0x880a0282,0xaaa02a8,0x228a220,0xa828a080,0x2000800,0x20a282a0,0x88a28800,0xa80aaa00,0x280a0222,0x202208);
remainders_pos[665].modulo_value=2660; remainders_pos[665].bit_mask=create_bit_mask(84,0x80020a02,0x82082a0,0x28a0228,0x82020808,0x22200002,0x88a00200,0x200002,0x2a000,0x8802000,0x2220880a,0x80222080,0x2228082,0xa802220,0x802082,0x20080028,0x20a80a80,0x220a82a0,0x8008008,0xa228882,0x2a02000,0x28a20288,0x2a008a02,0x22800020,0x82088a,0x82208008,0xa8202a0a,0x80200a80,0x800a000,0xa820820,0x22002220,0x200a080a,0x22200822,0x8828088,0x20,0x2000002a,0x288a0000,0x22200200,0x20a0000,0x8a008082,0x2080a202,0x82a08,0x280a0282,0xa820008,0x8a02082,0x20200a28,0xa0808,0x8008880,0xa2280,0x80000080,0x8000000a,0x20282200,0x82008882,0x20a0088,0x8880088a,0x82082a00,0xa00200,0x2a008020,0xa8082a0,0x20882a,0x22082002,0x8000288a,0xa200a80,0x2808a288,0x80a802,0x22288a00,0x200208,0xa82a0882,0x2a02a080,0x80020080,0x20802002,0x88802a08,0x20288800,0x20888028,0x2208880,0x80220a,0xa80000,0x8000,0x800a228,0x8880,0x2080828,0x880a2802,0xa8208202,0xa080020,0x8);
remainders_neg[665].modulo_value=2660; remainders_neg[665].bit_mask=create_bit_mask(83,0x882020a,0x280a20,0x20282a0,0xa028080,0xa2a0880a,0x280288,0x88288082,0x80a2008,0xa880,0xa2a80002,0x82a2800,0x8a220802,0x20822a8,0x8800a00a,0xa0808820,0x28200200,0xa2020a20,0x80000,0x82a2000a,0x82202808,0xa02a0200,0x22080202,0x2a0880a0,0x80028002,0x2280000,0x20a02282,0x8280200,0x882088,0x82028028,0x2288aa20,0xa0820082,0xaa2800a2,0x20000,0x808008a8,0xa0888022,0xa0020808,0x2a288a00,0x2828088,0x200080a,0xa0002a0a,0x8802280,0xa0028202,0x20a8000,0x20a002,0xa0a88220,0x88020080,0x80000000,0x8822208,0x8808000,0x888802,0x28a02a80,0x2080002,0x2028000,0x8002,0x2802208,0x88280208,0xa2080820,0x2000220,0x80a000a2,0x2200a80a,0x8802002,0x82280200,0x20802200,0x8800200a,0x22200200,0x80a80a80,0x20228002,0x22822000,0x8820808,0xa008200a,0x82200,0x20200080,0x200000a0,0x8a200008,0x8000aa02,0x88200888,0x88000080,0x882220,0x8800000,0x82808020,0x2208a,0x20200282,0x28000a8);
remainders_pos[666].modulo_value=888; remainders_pos[666].bit_mask=create_bit_mask(28,0x220820a2,0x28028a08,0x88000282,0xa2002000,0x820228,0x8088208,0x80802800,0x28a00a08,0x88208080,0x22082,0x8a20a20,0x200a288,0x200000a0,0x20820,0x8082080,0x800a0000,0x20228a00,0x8a08a,0x22820880,0x28020208,0x220a00a,0x20002802,0x208220,0x8a288082,0x22000800,0x28828000,0x8820a280,0x8a0820);
remainders_neg[666].modulo_value=888; remainders_neg[666].bit_mask=create_bit_mask(28,0xa2802822,0x20820200,0x8820a,0x22882880,0x8028a20,0x80000280,0x82080,0x20208200,0x280000,0x808a2802,0x228228,0x8a082200,0xa0080820,0x8828028,0x8000a008,0x820880,0x28a20208,0x88002002,0xa20a0000,0x20828a00,0x8a282082,0x20882082,0x8a00a28,0x220000a,0xa2880080,0x20020808,0x202208,0x200a0);
remainders_pos[667].modulo_value=2668; remainders_pos[667].bit_mask=create_bit_mask(84,0x8a22228a,0xa08a80a0,0xaa0a0808,0x82a28882,0x22082208,0x2080a228,0x228a22a0,0x208022a,0x28a28a0a,0xa800aa,0xa08a2a22,0xa0002a2a,0x8082a22,0xa028aa8,0x2a00808,0x2a008a02,0xa880,0xa800022,0x88a280a0,0x22aa8288,0x2800a2a,0x88082aa,0x22020800,0x2282a08,0xaa882aaa,0x80aa200a,0xa82802a,0x82a2a2a8,0xa2a00200,0x880800a2,0x80080aa,0x2a088082,0x208022a0,0x88808208,0x8808aa2,0x28a80002,0x8880882a,0xa228882,0xa2200,0x2820a282,0x282a02,0x8020888,0x82800222,0xa082880a,0x8a000828,0x288a0008,0x20222822,0x2a28a82,0x20220800,0x202228aa,0x80208208,0x220a80a8,0x20020820,0x2022aa0,0xa8a8a0,0xa8a82008,0x282a02a8,0x8aa02a80,0x822aaa00,0x82880aaa,0x808820a,0x20220002,0x280aa8,0x2aa88a8a,0x28a22228,0x2a00a0,0xa0000880,0x200a8022,0xa8080a,0x280a0202,0x820202aa,0x8000a88a,0x8a20aa8a,0x2a0088a,0x28a28aa0,0x2080a0a,0x8a288a88,0xa02080a8,0x82088288,0x28a82208,0xa0aa822,0x2a20a202,0x888a20a0,0xa28);
remainders_neg[667].modulo_value=2668; remainders_neg[667].bit_mask=create_bit_mask(84,0x22aaa02,0x28020828,0x22028080,0xa22000a,0xaa80aa80,0xa8082aa0,0xaa022228,0x8a808aa2,0xa02a0282,0x8208022,0x2802a2aa,0x2888a2a2,0x8080a22a,0x82820220,0x8a288080,0xa2880282,0x88882008,0x208882a,0x2a0828,0xaa220a00,0x820882a2,0x80080a22,0xaa8a8088,0x8a20a280,0x22002222,0x822a882,0x820208a2,0xa2a2a20,0x2a288a80,0x80082a,0x80880822,0xa200080a,0xa808a228,0x80a80,0x8008022a,0xa020800a,0x800a2,0x82aa000a,0x882aa80,0xa0a82a0a,0x8a0a28a,0x808a8000,0xa008aaa,0x200a0082,0x28880a0,0xa0028880,0xa82aa02a,0x8a2a020a,0xa8aa8088,0xa0a2a022,0x8a80a80,0xaa820820,0xa88a00a8,0x8a8a2228,0x88202028,0x2020a080,0xa0a28a20,0x228a200,0xaa22208,0xa008222,0x80000a82,0xa8aa8882,0x88a08220,0x22200202,0xa02aa2a0,0x88a28828,0x20888008,0x288208aa,0x88208082,0xa0028a8a,0xa8a8a22,0x8802002,0x2a02202,0x8a288002,0xa02a0228,0x8a800202,0x2a00200,0x28a80820,0x2800200,0xa020aa80,0x828220aa,0xa2a82a0a,0x22828,0x2a0);
remainders_pos[668].modulo_value=1336; remainders_pos[668].bit_mask=create_bit_mask(42,0x22a08202,0x2a8a088a,0xaa2aa08a,0xa808082,0xa880aaa8,0x2a202208,0x2a02a000,0x800a2a,0x82008228,0x28088222,0x8a2a0a8a,0x28882202,0x2882008,0x8a8a0020,0x8000a80a,0xa208808a,0x2aaa022,0xa820202a,0x8a20aa8a,0x8a220a2a,0x280820a8,0xa22a0820,0xa2a8a088,0x2aa2aa08,0x80a80808,0x8a880aaa,0x2a20220,0xa2a02a00,0x800800a2,0x28200822,0xa2808822,0x28a2a0a8,0x82888220,0x288200,0xa8a8a002,0xa8000a80,0x2a208808,0xa02aaa02,0xaa820202,0xa8a20aa8,0x88a220a2,0x80820a);
remainders_neg[668].modulo_value=1336; remainders_neg[668].bit_mask=create_bit_mask(42,0xaa280a8a,0xa2028002,0x22a22802,0x8208080a,0x20082220,0xa2a8aa00,0xa28a2888,0x880882a2,0xa880aa0,0xa0800aaa,0x2a28202,0xa000aa8a,0x8a00a880,0x20288a8,0x8882082,0x2a800802,0x8a2228aa,0x20a8a8a2,0x2a82202,0x2aa82a2,0xa080a820,0x2aa280a8,0x2a202800,0xa22a2280,0x8208080,0x2008222,0x8a2a8aa0,0x2a28a288,0x880882a,0xa0a880aa,0x2a0800aa,0xa02a2820,0xa000aa8,0x88a00a88,0x2020288a,0x20888208,0xa2a80080,0x28a2228a,0x220a8a8a,0x202a8220,0x2aa82a,0x80a82);
remainders_pos[669].modulo_value=2676; remainders_pos[669].bit_mask=create_bit_mask(84,0x828808a2,0x8820828,0x80002280,0xa20a2800,0x8200a20,0x8220a282,0x228a0080,0x20220200,0x8082088,0x8020a2,0x20a00000,0x8a00a208,0x20800002,0x228000,0xa088002,0x808a2880,0x8a20200,0x8228200a,0x20022802,0x8020008,0x200220a,0x82822002,0x20800200,0xa00a200,0x828020a2,0x8020228,0x8008a288,0x2000882,0x20000a20,0x8a000008,0x20822802,0x28,0x82282008,0x20882080,0x20020220,0x808028a,0x820a2822,0x28228020,0x800a282,0x800a2000,0x8a08208,0x8a28808a,0xa20a2022,0x220820,0x200008a,0x828828a0,0x8208028,0x208828a,0x8a2802,0x20808808,0x88202082,0x20082,0x20828000,0xa280288,0x820000,0x8008a00,0x282200,0x20228a2,0x28228808,0xa08a080,0x208008a0,0x28200800,0x8080088,0x20a0880,0x820008,0x88280288,0xa20a0082,0x20200808,0xa00228a,0x80080022,0x20800028,0xa280000,0xa08208a0,0x20000000,0x208a080,0x80822082,0x28800808,0x8820200a,0x820828a0,0x8a08a00,0x20028a,0x22002880,0x28228208,0x8a202);
remainders_neg[669].modulo_value=2676; remainders_neg[669].bit_mask=create_bit_mask(84,0x2000022,0x28020,0x808a208,0x22822080,0xa08228,0xa28220a,0x22020800,0x28a28a08,0x8000a000,0x80082822,0x28208808,0x2082280,0xa0080882,0x8a20808,0x8200008a,0x22000,0x228a08,0xa20a082,0xa08a2082,0x828800,0x8a08a282,0x20a2882,0x28008a08,0x82082288,0x2082822,0x828a20,0x8002200,0x82880002,0x28808228,0x2088080,0xa00a2082,0x8808820,0xa20a080,0xa0002800,0x20828a28,0x80008202,0x28220a2,0x20a20828,0x8008220a,0x822880,0x200a00,0x2200002,0x228228a2,0x8a28028,0x8a088002,0x2002020,0xa00820,0x8a000202,0x80022082,0x28000000,0x28a00a,0x808a0802,0x28020808,0x82208200,0x800a0880,0x800208,0x8820a288,0x828a2022,0x20a20000,0x82002008,0xa0080020,0x20a08008,0x80008000,0x82820000,0x8028800,0x208200,0x22820802,0x28a08000,0x8208a202,0x8008a2,0x28008820,0x82208088,0x200a0020,0x28808808,0x8a002008,0xa2802,0x20008000,0x28a082,0x2802020,0x200208,0x88288202,0xa2882000,0x20a20a00,0x228a);
remainders_pos[670].modulo_value=2680; remainders_pos[670].bit_mask=create_bit_mask(84,0x8888020a,0x2a028820,0x2a02a020,0x8282a0a2,0x2008a20a,0x88822a80,0xa0088800,0xaa2a0,0x80a28880,0xa2a0802a,0x80a2080,0x22080a22,0x8002208,0xa222088,0x80000a,0x88222800,0x202a0082,0x2a0a2028,0x828228a2,0x2802822,0x20a80000,0xaa2a8802,0x2082280,0x2002080,0x20282020,0x808a200a,0xaa280220,0x208aa208,0x88822020,0x20a02a28,0xa0202882,0xa0a00a0,0xa8200,0x8a80800a,0x200002a,0x88a0088a,0x2a000280,0x8200a8,0x2020082,0x20280220,0x88280280,0x228880,0x22022288,0x8028028,0x80088028,0x820080,0xa82a0082,0x22028000,0x80a2200a,0xa2a80000,0xa00202,0xa00082a0,0xa0a00a0,0x8822808,0x2228a80a,0x8080882,0xaa208aa2,0x2088028,0x8a008a2,0x80080828,0x80020800,0xaa028820,0x88022a8,0x8000002a,0x82882802,0xa88a2882,0x82808a0,0x228200a8,0x2888,0xa0a00002,0x20220888,0x88208800,0x2088a020,0x8a0208a0,0x2a8020a,0x2228a,0xa0a8aa0,0x22002220,0x802a882,0x82a08a20,0xa88a0a82,0xa8080a80,0x22082280,0xa08022);
remainders_neg[670].modulo_value=2680; remainders_neg[670].bit_mask=create_bit_mask(84,0xa82,0xa20a00a0,0x228a20a0,0xa02282a,0xa0802a02,0xa2208,0x28000080,0x8822228,0x8220008,0x22280822,0x80822808,0xaa0082a2,0x88a280,0x82a2a800,0x80088802,0xaa2088,0xa8228802,0x2282a0a0,0x202a02a,0x8208a02a,0xa8200888,0x22220082,0xa80a208,0x8a80a808,0xa0a0a828,0x8022882,0x22208aa0,0x28022200,0x2a8a8,0xa028a220,0x28a8200a,0x82028820,0x8820288,0x2000882,0x82888822,0x200002,0xa2088a00,0x80a8020,0x8a82880a,0xa0a08a28,0xa00a08,0x882a0000,0x2a8aa200,0x8208a0,0x800820,0x880a0808,0x20228802,0x2a8a0088,0x822a882,0x22208808,0x88280a8a,0x28080a20,0x2828028,0x8002a080,0xa2a02002,0x8080000a,0x22280222,0xa8000a0,0x8020802a,0x808020,0x80a0088,0x220a00a0,0x8a220,0x88088a2,0x200a00a,0x2002200a,0x80208020,0x2a0a8020,0x8880a000,0x20288802,0xa8aa0000,0x280080,0x280020a8,0x2828028,0x82208a02,0x888a2a02,0x82020220,0x2a8822a8,0x8082200a,0x2280228,0x2002020a,0x20008200,0x2a80a208,0x2008aa);
remainders_pos[671].modulo_value=2684; remainders_pos[671].bit_mask=create_bit_mask(84,0xa2aa0222,0x8022080,0x882a0088,0xa80a8a22,0x8a02220a,0xa0002a80,0x8800800a,0x820a222,0x8aa0a8aa,0x20a02082,0xa8028a88,0x8028022a,0x82208080,0x82a0820a,0x888802a2,0x82882a2a,0xa2aaa000,0x8828aa80,0x88a2a800,0x8a82808a,0x88888a8,0x20022222,0x22a0202,0x2a200820,0x280000a2,0x28a80000,0x80220a,0xaa202,0x280280a,0xa2a280a,0x82a0aa,0x220208a0,0xa088828,0x80202002,0x80828a0,0x2aa0aa2,0x2800aa20,0x2880a82,0x820280,0xa22880a2,0xa288a20,0x8828008,0x2002822,0xa08a228a,0x8a02288,0x280820,0x882a0228,0x8aa002,0x8aa0aa8,0x20a28202,0x8008080,0x8282220a,0xa20808,0xaa0a820,0xa028a8a,0xa028028,0x808aa00,0x8a088020,0x800002a2,0x88a00002,0x82008a,0x88080a88,0x8888800,0x22a22222,0x2a20282a,0x2002a8a2,0xa02aa282,0x2000aaa8,0x2a8a8228,0x28a80222,0x2a0820a8,0x20202088,0xaa880280,0x822a2802,0x282080a0,0xaa2a0aa,0x2888a082,0xaa002002,0x202a8000,0xaa08880a,0x288a2a02,0x2200a82,0xa0208802,0x8880aa8);
remainders_neg[671].modulo_value=2684; remainders_neg[671].bit_mask=create_bit_mask(84,0x2a2282aa,0x800aa808,0xa28800,0x208202a2,0x28a2a82,0x2008a208,0x880882,0x80a82a2a,0x2202022,0x2828a80a,0x208a0200,0x8a082a2,0xa280808,0xa280282,0x8a22,0xa0022a2,0x22222888,0xa02208,0x2a2008,0x2020802,0x20,0xa88aaaaa,0x8aa2828a,0xa22880a8,0xa088882a,0xa0208880,0x88082a82,0x80822a8a,0x8a00a082,0x82a2a002,0x88022822,0x2a8a8028,0x828000a0,0x8a8a08a,0x8000a028,0x8a22822a,0xa0882220,0x8a00020a,0x800a8a08,0x2aa0082a,0x82a00228,0x80020880,0xa88a0aa,0x2002aa02,0x8028a200,0x882080a8,0xa28aa0,0x88022882,0x80220220,0xa02a0a8a,0x80880808,0xa0a2a02,0x88228080,0x22820a8,0x828a0202,0x828a00a0,0x80002288,0x28008a8,0x8888a22,0x280882,0x800a8802,0x808200,0x80000008,0xaa22aaaa,0x22a8a0a2,0xa88a202a,0x2822220a,0xa8082220,0xa2020aa0,0xa0208aa2,0xa2802820,0xa0a8a800,0x22008a08,0xaa2a00a,0xa0a00028,0x22a2822,0xa000280a,0x2288a08a,0xa8220888,0x22800082,0xa002a282,0xaa8020a,0x20a8008a,0x8220);
remainders_pos[672].modulo_value=168; remainders_pos[672].bit_mask=create_bit_mask(6,0x220a2802,0x20208200,0x2008000,0x82080800,0x28a08800,0x80);
remainders_neg[672].modulo_value=168; remainders_neg[672].bit_mask=create_bit_mask(6,0xa2822002,0x28a00a00,0x82080088,0x2000080,0x20200008,0x8);
remainders_pos[673].modulo_value=673; remainders_pos[673].bit_mask=create_bit_mask(22,0x3fa573de,0x85f744b1,0x9aba4f21,0xc231d597,0x98dd0c01,0xe944c76e,0xd993e915,0xd20e0f2b,0x225208ab,0x616259fb,0xf8b7b47c,0x7e691a18,0x54412913,0x53c1c12f,0xa25f266f,0xdb8c8a5e,0xc2ec65,0xa6ae310e,0x13c97567,0x348bbe86,0xef3a97f2,0x1);
remainders_neg[673].modulo_value=2692; remainders_neg[673].bit_mask=create_bit_mask(85,0xa228aa02,0x82a8828,0x22282a8,0xaa8080a,0x8088,0x20880aa2,0xa2088,0xa8282a2,0xaa8a8022,0xa8aa8022,0x202a28a0,0xa2a08280,0x8888a08a,0x8a0808a2,0x2a82aaa2,0x2000282,0x880a20a8,0x2a22a882,0x8a00a8aa,0xa802220a,0x22a20a2a,0xa2822a20,0x822200aa,0xa8a8028a,0x8aa22a2,0xa822808a,0xa000200,0x2aaa0aa2,0x2880828a,0x8828888a,0xa082a2a,0x28a2a020,0x200aa8a8,0x200a8aaa,0x2a0a0a82,0x88228002,0x2a808820,0x88080002,0x8080aa80,0xaa0a2202,0xa08aa080,0x2a8a228,0x220802a2,0x8a008a08,0xa220a00a,0x800a8a82,0xaaa8a88a,0x288a8022,0xaa82288a,0x80a0a022,0x80aa22,0x800aa20,0x2a020828,0x2028a0aa,0x88882880,0x828a8820,0xa00020,0xa2aaa0a2,0x8a82280a,0x22008a0,0x82a80802,0xaa22280,0x20228200,0x20a20222,0xa222a800,0x80aa080,0x88022020,0xa20a882,0x82aaa2a8,0x28022,0x8a8a082,0x8a088882,0x828a0200,0x8202a2a,0x2a80080a,0x2a808002,0x28280a2,0x8a20aaa2,0xa88a28,0x8a8aaaa2,0xa8a800a8,0x28222a0,0x288028a8,0xa0082208,0x2);
remainders_pos[674].modulo_value=2696; remainders_pos[674].bit_mask=create_bit_mask(85,0x22080aa2,0xa8aaa288,0xa00aa028,0x2202a0a8,0x20020a22,0x8a88,0x2822282a,0x80228a02,0x822a2aa0,0xaa8a08aa,0x8a808202,0x200a0222,0xa22a2a82,0x8a02a880,0x282aa82a,0x200888a,0xaa0a88a8,0x20a28aaa,0x28a208a0,0x2a22020,0xa2080aa8,0xa8a202,0x2a2202a8,0x8a2082a,0x820a2a,0x880a000,0x888aa2a8,0x28002820,0x8802a080,0x8020220a,0x22a0aa22,0xa2828082,0x8a08002,0x202280,0xa0822a8a,0x28222822,0x808aaaa0,0x20a2aa28,0xa02a222,0x2a00aa08,0x82000808,0xa8a228,0x8a2a0082,0x20008228,0xaa00a820,0x8a80a020,0xaa8a0888,0xaaa20228,0x2888280a,0xa8820a88,0x880800a2,0x20a2002,0x2828a80,0xaa0a8882,0x88080288,0xa8022a0,0x28002802,0x8aa22208,0xa02202a,0xa0820000,0x208a20a8,0x8088a8a8,0x8a2a002a,0xa0208a80,0x88a802a,0x208a2808,0xa28a080a,0x22a0aaaa,0x2200802a,0x2aa828a2,0x2a80a2a8,0xa8a88a02,0x80a00882,0x8202a288,0x20a2aa80,0xa8a882aa,0xa288020a,0x28882880,0xa20000a8,0xa0800822,0xa808888,0xaa00a2a,0x8aaa2a28,0xa0208822,0x8a);
remainders_neg[674].modulo_value=2696; remainders_neg[674].bit_mask=create_bit_mask(85,0xaa80822a,0x20222a00,0x288228a0,0xaa8a2820,0xa88a82aa,0x88880200,0xa0aaa0a2,0x8aa028a,0xaa2a228,0x22028022,0x2080a8a,0xa8828aaa,0x2aa2a20a,0x28a2008,0xa0a220a2,0x8a880002,0x22820020,0xa82a0222,0xa02a8028,0x8a2aa8a8,0x2a808220,0x88202a8a,0xa2aa8a20,0x802a80a2,0x880a82a2,0x80082888,0x22a20,0xa088a0a8,0x8a2808,0x8a8aa82,0xaa2822aa,0x2a02080a,0x8028088a,0x88a8aa08,0x280aa202,0xa0aaa0aa,0x8022228,0xa82a22a0,0x828a2aaa,0xa2882280,0xa888080,0x88202aa0,0x2a2880a,0xa8880aa0,0x228820a8,0x20828a8,0x22028000,0x222a8aa0,0xa000a082,0x200a8200,0x80882a,0x8a82a88a,0x8a0a0208,0x2282000a,0x808a00,0x8208aa28,0xa088a08a,0x22aaa80,0x828aa8a2,0x280a8888,0xa802a820,0x8002020,0x2a288a2,0x28a80208,0x800208a2,0xa802a080,0x2a028082,0xaa282222,0xaa8808a2,0xa220a02a,0xa2082a20,0x2020028a,0x828800a,0xa0a2a00,0xa82a2208,0x20200a22,0x2a008a82,0xa000a008,0x2a888820,0x280880aa,0x82080000,0x822882a2,0x222a2a0,0x28a800aa,0x2);
remainders_pos[675].modulo_value=60; remainders_pos[675].bit_mask=create_bit_mask(2,0x802802,0x8028020);
remainders_neg[675].modulo_value=60; remainders_neg[675].bit_mask=create_bit_mask(2,0x80082082,0x20820);
remainders_pos[676].modulo_value=104; remainders_pos[676].bit_mask=create_bit_mask(4,0xaaaa8aaa,0xaaaaaa2a,0xa2aaaaa8,0xaa);
remainders_neg[676].modulo_value=104; remainders_neg[676].bit_mask=create_bit_mask(4,0x22220222,0x22222222,0x22222220,0x22);
remainders_pos[677].modulo_value=677; remainders_pos[677].bit_mask=create_bit_mask(22,0x361e652,0x5fda437e,0xe5f2e0a1,0xc6aa0cb3,0xa9773bf6,0xa9c422db,0xdf8f1ada,0x2f411913,0x58320ac1,0xb91e6262,0x5973a6b8,0x919e2747,0xd4130691,0x2620bd20,0xd63c7ef2,0xd108e556,0xf73ba576,0x4c1558db,0x41d3e9f3,0xb096fea1,0x99e1b01f,0x12);
remainders_neg[677].modulo_value=2708; remainders_neg[677].bit_mask=create_bit_mask(85,0x8aa82a8a,0x82028aa2,0x282a2828,0xa22802a,0x20aaa22a,0x2008aa02,0x2028202,0xa288808a,0x80aa8208,0x2082aaaa,0x80aa2a20,0x802aa8a,0x88a8a08,0xaaa820a8,0xaa0a22a,0x880288a,0x2aa228aa,0x80888002,0x880a202a,0x280a2228,0x282882,0xa0a0028a,0x2280a208,0x228088a2,0x8880aa0,0xa22aa200,0xa00882a8,0x282a8288,0x20aaaaa2,0x8a8880a8,0xaa008082,0xa2a80a8a,0xaa082022,0xaa80aaa,0x88a2882,0xa020288,0xa8802202,0x2aa82202,0xa2282a2,0xa2a0a2a0,0x8a0208a0,0xa0aa8a2a,0xa00828a,0xa0802080,0x20a08a2,0x20aa020a,0x220282,0x8802a228,0xa0a2a22a,0x88a882a2,0xa28820,0xa00000a8,0x2022a28,0xa00080a8,0x80828a8a,0xa280a88,0x28220000,0xaa088280,0x22080288,0x88aaa200,0x822a00a8,0x8222088a,0xa08a20a,0x2aa002aa,0xa8228828,0xa8882220,0x8a802a20,0x22aa88,0x88a00822,0xa0882a,0x220a,0x88a80a28,0xa8a8a080,0xa808002,0x8a2a2020,0xa800002,0x8208a280,0xa2a08a88,0xaa22a282,0xa22a008,0x20a02200,0x28202a82,0xa2882820,0x820082,0x8028);
remainders_pos[678].modulo_value=2712; remainders_pos[678].bit_mask=create_bit_mask(85,0x2020002,0x8a20028,0x8a00a002,0x2080882,0x8a28a20,0x28220a,0x828a2880,0x20828820,0x8a008000,0xa28a2802,0x20228a00,0x82202208,0x80082020,0x200a08,0x220200a,0xa2820880,0x20220020,0xa08a,0x82082020,0x8a20028,0x8828a082,0x80820,0x20008208,0x820800a,0x20a2822,0x20220008,0x8200020a,0x22022880,0x28228028,0x8288080,0x22020822,0x220820,0x280,0x828000a0,0x820a28,0x8208,0x200a2880,0x8020,0x8020a200,0x2800080,0x20008228,0x8220a00a,0x828a00a2,0x8a00a08,0x80288200,0x2020002,0x8a08,0x8080200,0x228a0,0x208200,0x8228a082,0xa0002,0x28000,0x88082088,0x20882080,0x28020228,0x88280288,0x82022880,0x8a08000,0x80200088,0x208828a0,0x8a00208,0x208200,0x22082020,0x20820a28,0x8228008a,0x80820,0x8a20a00,0x8a080088,0x80022082,0xa00808,0x220a008,0x82080820,0x8208808,0x8a00a288,0xa28028a2,0x8000200,0x82082282,0x228a2,0x20a08828,0x8008a28a,0xa2822020,0x20800a00,0x8028008a,0x800080);
remainders_neg[678].modulo_value=2712; remainders_neg[678].bit_mask=create_bit_mask(85,0x828a0882,0x228820,0x208208a,0x82800002,0x220228,0x8820a282,0x2022000,0x28020028,0x2080088,0x22022082,0x28a20208,0xa28a280,0x8028a0,0x8a08200,0x8a28a082,0x220a0000,0x28a28828,0x88082002,0x28028a0,0x228820,0x20200a,0x808000a0,0x28800a00,0x80280082,0x820220a2,0x28a28800,0xa088282,0xa28a2000,0x20a20820,0x80200008,0xa28a00a2,0x8a28028,0x88088208,0x2080820,0x8028220,0x88080280,0xa0822000,0x8800828,0x282288,0x82080800,0x28800a20,0xa282082,0x2020822,0x208200,0x8200288,0x828a0882,0x8800200,0x80008288,0x808a2020,0x8a00a08,0xa20200a,0x80820882,0x8820808,0xa000,0xa0002800,0x20828a20,0x208200,0x28a2000,0x200808,0x8288000,0xa0002020,0x208a00,0x88280288,0xa28028a0,0x28028220,0xa208002,0x808000a0,0x228200,0x2008000,0x8a2802,0x8208000,0x8a282080,0x28000a0,0xa00000,0x2082200,0x22082022,0x808a08,0xa00a20a,0x808a2022,0x28200020,0x8002202,0x220a2820,0x28008208,0x8208002,0x80800);
remainders_pos[679].modulo_value=2716; remainders_pos[679].bit_mask=create_bit_mask(85,0x8a82a22a,0x28a280,0x202208aa,0xaa2a2880,0xaa002080,0x282280,0x202aa22a,0x8008880a,0x8800282,0x80280802,0xaa0888a2,0x20820a88,0x2a800a0,0x80202a8,0xa28a0802,0x8a80000,0x82a28088,0x80202082,0xaa80820a,0x8a0a22a8,0x80880288,0xa20a020a,0x2802202,0x22000008,0xa82a008a,0x8020a08,0x80222228,0x800820aa,0x8a828a22,0x2008a002,0x8228882a,0x22002002,0xa0a20802,0x8a2202a0,0x20aa020a,0x888222,0x20a88080,0xa2800a08,0x80002828,0x88a08a00,0xaa0020,0x8022208,0x2000008,0x2088802,0x20800aa0,0x200a20a2,0xa2828000,0x820a0028,0x2022a0,0x88882220,0x2a080aa0,0xa0a8088a,0x880208a0,0x28008008,0x8a822288,0x2800a200,0x288a282a,0x2aa08200,0x2888880,0xa20a0802,0x8a200a82,0x2000008,0xa8088028,0x2a080a08,0x22280220,0xa2a88a0a,0x2a08202a,0x28208080,0x22028a8,0xa00002a2,0x8020a28,0x2a80802,0x80a002a8,0xa22a0820,0x28a2220a,0x8020280,0x28280022,0x8a022200,0xa88aa80,0xa0288280,0xa020800a,0x80228a8a,0xaa20880,0x2028a280,0xa88a82a);
remainders_neg[679].modulo_value=2716; remainders_neg[679].bit_mask=create_bit_mask(85,0x20a2a22,0x8a02a00,0xa0aa8022,0x2222a008,0x2280a808,0x88a02a08,0xa8a222a2,0x8800002,0x88a02,0xa08082,0x2200002a,0xa8028200,0x8a200828,0x808a8220,0x2a02800a,0x208880,0x22a0800,0x828a80a,0x22000a82,0x2822a20,0x8008200,0x2a828a02,0xa08aa82,0xa2888880,0x20228802,0x80828280,0x8aa2aa0,0x880a022,0x20a022a,0x28802882,0x2a000a2,0xaa08a88a,0x2822808a,0x2aa0228,0xa8228282,0x88000a2a,0x28200800,0x22088280,0x808a0a0,0x200288,0x882208a8,0x808aa280,0x8a888800,0xa800082,0xa0088228,0xa802282a,0x2a020888,0xa8208a0,0x88a8a228,0xaa28,0x22808220,0x20208002,0xa8028,0xa0800880,0x20a2a00,0xa0882288,0xa002a022,0x22200a80,0x82000008,0x2a02808a,0x2a0820a,0x8a880880,0x208000a0,0xa2808200,0x2aa08aa0,0x22200282,0xa200a8a2,0xa0a00808,0x8aa82020,0x2808822a,0x808a8220,0xa208082,0x288a20,0x2a2280a8,0xa022aa82,0x808a0a08,0xa0a080aa,0x28aaa08,0x2002200,0x20a00a08,0x28280882,0xa20202,0x822a0008,0xa8a02208,0x2002022);
remainders_pos[680].modulo_value=680; remainders_pos[680].bit_mask=create_bit_mask(22,0x20802a82,0xa2220800,0x2a0a2208,0xa22a00a,0x2080a808,0x80220282,0x288802,0x2000a200,0x200a020,0x8228a820,0x820002,0x2a288280,0xa008008,0x8a000808,0x22280000,0x80880280,0x2a020882,0xa88a020,0x88a0a8a0,0x20888a20,0xa8020800,0x82);
remainders_neg[680].modulo_value=680; remainders_neg[680].bit_mask=create_bit_mask(22,0xa808220a,0x2a228080,0x2282a280,0x82222882,0xa0082000,0xaa0a0a,0x88200082,0x28882288,0x8a8028a8,0x2a02028,0x880a0882,0xa2200a00,0x2880000,0x2808080,0xa2a08008,0x8000a08,0xa20a0002,0x20020a8,0x202028,0xa0000228,0x200a0088,0x2);
remainders_pos[681].modulo_value=2724; remainders_pos[681].bit_mask=create_bit_mask(86,0x20a00a2,0x20a08,0x80288200,0x22802882,0x208a20,0x80288200,0x802020,0x28200200,0x8200a000,0x80002822,0x20a20808,0x82002202,0x828a0820,0x8808200,0x8088200,0xa00008a2,0x20820a28,0x2000202,0x228800a0,0x820228,0x8a20228a,0x228a28a2,0x8028a20,0x8a20a202,0x2002800,0x8220200,0x28a282,0x80822880,0x8800208,0x8a080208,0x2082082,0x28220220,0x88082,0x20a20a0,0x20a00028,0x8000200,0xa0082020,0x20000208,0xa202288,0xa00a08a0,0x208,0x82008282,0x28a2802,0x28200828,0x8000008,0xa20a00a2,0x28808a00,0x8000082,0x808200a2,0x8000200,0x8000a080,0xa08a0802,0x20220000,0x88088288,0x20820800,0x8020a28,0x8002202,0x22882022,0x28a28000,0x8088208,0x2800800,0x8a08a20,0x8a280208,0xa28a2880,0x28808a28,0x8808200a,0xa0022882,0x8000800,0x8a082088,0xa20000a2,0x8220208,0x8202200,0x820a2820,0x8800820,0x208a088,0x82800022,0xa00828,0x8008280,0x80802000,0x8228020,0x8a208000,0x22802880,0x8228028,0xa080000,0xa00a0802,0x8);
remainders_neg[681].modulo_value=2724; remainders_neg[681].bit_mask=create_bit_mask(85,0x82820822,0x8828200,0x8200288,0xa2082002,0x8a00228,0x8200288,0x800828a0,0x20a08a00,0xa082088,0x8820a2,0x28228000,0xa08a28a,0x20200a0,0xa08,0x80000288,0x20880022,0x28028220,0x8a08828a,0xa2000820,0x8028a20,0x228a202,0xa2022022,0x820228,0x228228a,0x82882080,0xa28a08,0x8820220a,0xa2000,0x8a00,0x2008280,0x82802802,0x20a28a28,0x8800000a,0x82822820,0x28208820,0x80080288,0x208028a0,0x28808a00,0x8228a200,0x20820020,0x8808a00,0xa08020a,0x82022082,0x20a08020,0x80088080,0x22820822,0x20000208,0x8008800a,0xa0822,0x808a08,0x8082008,0x20020082,0x28a28808,0x200,0xa00a0080,0x828220,0x8008a28a,0xa20028a2,0x20220808,0x80000280,0x82080080,0x200228,0x2208280,0x22022000,0x20000220,0xa082,0x208a2002,0x808008,0x200a000,0x22880822,0xa28a00,0x8028a288,0x28220a0,0x8028,0x8a002000,0x20808a2,0x8208020,0x80080208,0x82880,0xa20828,0x2280088,0xa2082000,0xa20820,0x82008088,0x20820082);
remainders_pos[682].modulo_value=2728; remainders_pos[682].bit_mask=create_bit_mask(86,0xa0a020a,0x22a8028,0x8028a82,0x800aaaa2,0x82aa2008,0x80282a82,0x8a82808,0x80aa02a,0x8802a082,0x8aaa8aa,0x220aa88a,0x22200200,0xa2000220,0xa2080a80,0x28200a20,0x28280008,0xa2220282,0x22088008,0x2880a22,0x82228a82,0xa0a0028,0xa080a20a,0x228820a0,0x82228080,0xa2220028,0x2008aa80,0x28220282,0x2028aa88,0xa20820a2,0x820a8028,0x222222a8,0x222aa0aa,0x8808800,0x88800000,0xa8228a00,0xa00a00,0x2888a820,0x820028,0xa288aa00,0xa8200000,0x2a28002,0xa0802a80,0xa00a0a0,0xa8020a0a,0x28a8002,0x82a80,0xaa228a00,0x820000,0x2a222828,0xa00a0008,0xa2882a00,0x22200,0x22022000,0xaa88800,0x888888aa,0x2a0822a,0x8208a28,0xaa28088a,0x80882822,0xaa200882,0x888a02,0x2888228,0x8228802,0x8a020a0a,0xa0a0a0,0xa2888228,0xa0228082,0x2208888,0x80888a20,0x282882,0xa0082820,0xa0208a08,0x80008a02,0x80088808,0x2aa08800,0x2aaa20a2,0xa8022aa,0xaa02082,0x282a20a8,0xa8280220,0x8aa8282,0xaaa00220,0xa280208a,0x2a88082,0x80a0a028,0xa0);
remainders_neg[682].modulo_value=2728; remainders_neg[682].bit_mask=create_bit_mask(86,0x2828282,0x8a2208a0,0x808a020a,0x8822222,0x2222880,0xa0a20a,0x8020a080,0x80822822,0x2280a,0x222022,0xaa822002,0xaaa88288,0x2a008aa8,0x2a808208,0xa0a882a0,0xa0a00880,0x22aa0a0a,0xaa800880,0x8a00822a,0xaa2020a,0x28280a0,0x28082a82,0xaa00a028,0xa2a0808,0x2aaa8820,0xa8802200,0xa0aa0a0a,0xa0a02200,0x2a80a822,0xa820820,0xaaa2aa20,0x2aa22822,0x80080088,0x88088,0x202a0288,0x80288288,0xa00020a0,0x880a08a0,0x22002288,0x20288888,0x8a2a080a,0x2800a208,0x2882828,0x20828282,0x8a02008a,0x8800a208,0x22aa0288,0x880a0880,0xa2aa20a0,0x20828880,0x2a00a288,0x888aa208,0xaa82a888,0x2200088,0x22,0x8a280222,0x802802a0,0x22a08002,0x800a0a2,0x22a80002,0x8000028a,0x8a000aa0,0xaa000a,0x2828282,0x8282828,0x2a000aa0,0x20aa000a,0x8a280000,0x80002a8,0x88a0a002,0x280020a8,0x20a80280,0x888028a,0x8800000,0xa2200088,0x2222a82a,0x8208aa22,0x8228a00a,0xa0222820,0x20a08aa8,0x80220a02,0x22280aa8,0x2208a002,0x8a20080a,0x8282820,0x28);
remainders_pos[683].modulo_value=2732; remainders_pos[683].bit_mask=create_bit_mask(86,0xa2808a82,0xa22a8880,0x8022a002,0xa00aa2a,0xa88,0x20aa222,0xaa20aaa0,0x2a820222,0xa8a88a2,0xa22820a2,0xa2aa8a08,0xa8a822aa,0x8200a028,0xa2882888,0x228288a0,0x20280a82,0xa8220a2a,0xa2aa0228,0x8008028a,0xa2a0820,0x8a0000a8,0x822802a0,0x82a02aaa,0xaa28a280,0x8020aa2,0xa0208a80,0xa8808288,0x88a02a0,0x828220a,0x8a2220a2,0x2020a82a,0x80008a0,0x88228200,0x28828a0,0x882202,0x80088a8a,0x8aa80028,0xaa888802,0x82a202aa,0xaa00802a,0x880aa808,0x82a2200,0xa0a022a,0x8a820a88,0xaa022008,0x200aa202,0x8a82a80,0x222aaaa8,0x2aa2802,0x22002280,0x82200a2a,0x82280808,0x288220a2,0x20008,0xa08080a2,0x888a2a82,0x828208a0,0xa220a08,0x2022a0a8,0x2080a228,0x802002a,0xa28aa8aa,0x80a82028,0x2882aaa,0xa20a8,0xa8a02a0,0x2002082,0xaa8aa28,0x882a288,0x2808a8a,0x2828882a,0x8228a0a2,0xa0082222,0x82a2a280,0x2aa8aaa8,0x8288a20a,0x2a2a08a0,0x82a88a2,0xa08aa888,0xaa0800aa,0x888,0xa00a022a,0xa8802a8a,0x2a88a800,0x2028a022,0x82a);
remainders_neg[683].modulo_value=2732; remainders_neg[683].bit_mask=create_bit_mask(86,0x2a08020a,0x2aa20008,0x8aa288a,0x828822a2,0x88888200,0x8a822aaa,0x22a82228,0xa20a8aaa,0x8202002a,0x2aa0a82a,0x2a220280,0x2020aa22,0xa8828a0,0x2a00a000,0xaa0a0028,0xa8a0820a,0x20aa82a2,0x2a228aa0,0x8808a02,0x82a280a8,0x2888820,0xaa08228,0xa28a222,0x22a02a08,0x808a822a,0x28a80208,0x20080a00,0x80028a28,0x80a0aa82,0x2aaa82a,0xa8a820a2,0x80888028,0xaa0a88,0x8a00a028,0x8800aa8a,0x8800202,0x22088a0,0x2200008a,0xa2a8a22,0x228808a2,0x822080,0x80a2aa88,0x82828aa2,0x20a8200,0x228aa880,0xa8822a8a,0x8020a208,0xaaa22220,0x8a22a08a,0xaa88aa08,0xaa882a2,0xaa08080,0xa00aa82a,0x888a8880,0x2808082a,0x2a20a,0xa0a8028,0x82aa8280,0xa8aa2820,0xa8082aa0,0x808a88a2,0x2a022022,0x820a8a0,0x8a00a222,0x8882a820,0x82028a28,0x8a88a80a,0x822022a0,0x800a2a00,0x8a080202,0xa0a000a2,0xaa0282a,0x2880aaaa,0xa2a2a08,0xa2202220,0xa002a82,0xa2a28028,0x80a2002a,0x28022000,0x22808822,0x88888000,0x28828aa2,0x2008a202,0xa2002088,0xa8a028aa,0xa2);
remainders_pos[684].modulo_value=456; remainders_pos[684].bit_mask=create_bit_mask(15,0x82020022,0x28020000,0x820a288,0xa20008a0,0x28a08200,0x80282,0x80080820,0x20200228,0x80200008,0x820a2882,0x20008a00,0x8a08200a,0x802822,0x808200,0x88);
remainders_neg[684].modulo_value=456; remainders_neg[684].bit_mask=create_bit_mask(14,0x28208a2,0x20828808,0x282200,0x22880020,0x20200a08,0x8800820a,0x8000a0,0x28208a20,0x8288080,0x2822002,0x28800200,0x200a082,0x800820a2,0x8000a08);
remainders_pos[685].modulo_value=685; remainders_pos[685].bit_mask=create_bit_mask(22,0x8897a1a,0x6f339c96,0x100e1429,0xc2a2306a,0xf48de912,0x14e75b08,0xc25bca11,0x23d81319,0x82301449,0x208d8f0a,0x84085194,0x6c410a62,0x310543c,0x6f1248a,0xf690e632,0xb9ca2214,0xec4bc436,0x1150d225,0x1c021583,0x733da50a,0xa4441a4e,0x1617);
remainders_neg[685].modulo_value=2740; remainders_neg[685].bit_mask=create_bit_mask(86,0x80002282,0xa22a0002,0x88280a0,0xa22a8a2,0x2000200a,0x882a0280,0xa020280,0x2a008280,0xaa08880,0xa0000202,0x8800808,0xa00882a2,0x28880a0,0x8a20a802,0x22082a2a,0x20022a88,0x200208a2,0x88802a8,0x8082880a,0xa2a02802,0x28880282,0xa20a8a82,0x2a0a0088,0xa00a02a,0x82808a08,0x20aa0088,0x22288200,0x208aa200,0x8aa2a082,0x200a822,0x2828088a,0x822a0880,0x28808008,0x82020000,0xa008882a,0x800a0802,0x880a0202,0x200a02a0,0x82802000,0x8228aa22,0x28280a08,0xa0002a2,0x220a2000,0x80a2a820,0x2020022a,0x20882088,0x82020000,0x2a0aa220,0x888200a0,0xa280a220,0xa8022880,0xa2000880,0x80aa2a0,0x280a88a,0x20288820,0xa0800888,0x20802,0xa028200,0x222a000,0x228a208,0x80880a80,0xa0288082,0x20002208,0x820088a0,0x80208000,0x20288,0x8a028822,0x880a080,0x8a2000a8,0xa2200822,0xa0280002,0x20000020,0x82a008,0x8a020888,0x80a00208,0xa808288a,0x2282a2,0x200a8088,0x80a2808a,0x20888222,0xa82a0280,0x20208222,0x8820000,0x2020882,0xa280aa20,0x2020a);
remainders_pos[686].modulo_value=56; remainders_pos[686].bit_mask=create_bit_mask(2,0x82002a22,0x88a800);
remainders_neg[686].modulo_value=56; remainders_neg[686].bit_mask=create_bit_mask(2,0xa88a22a,0x2080);
remainders_pos[687].modulo_value=2748; remainders_pos[687].bit_mask=create_bit_mask(86,0xa2000882,0x20020220,0x8228a08a,0x820a0022,0xa00,0x208288,0xa28808a2,0x820800,0x8a28220a,0xa0800882,0x8a20808,0x88202282,0x28020a0,0x8228820,0x280288,0x8a0880,0x28820a08,0x800820a,0x80820,0x28008,0x8220a000,0x208a2,0x28a00808,0x8228028a,0xa08220a2,0x820a20,0x82008200,0xa2020022,0x20028028,0x88200082,0x20082802,0x8008820,0x2208002,0x2022820,0x808000,0x820820a,0x80022000,0x20020828,0xa00a28a,0x8a0082,0x200008,0x8088288,0x2022880,0x228808,0x2282202,0x2000080,0x8200a20,0x8a28a00a,0x22820800,0x8800,0xa082082,0x2020,0x828808,0x8002088,0x80822002,0x28028200,0x88200082,0xa2802800,0x28800808,0x82008,0xa08a0820,0x28a08820,0x8a280288,0x20200a2,0x28a20800,0xa088,0x2002800,0x820200,0x8a082002,0x20a0822,0x220a20,0x2280280,0x822882,0x20a08028,0x8288082,0xa20208a2,0x28220020,0xa08828a,0xa0020820,0x8a20228,0x2282080,0x200a0000,0x28800a08,0x8a20a288,0xa0880800,0x8220008);
remainders_neg[687].modulo_value=2748; remainders_neg[687].bit_mask=create_bit_mask(86,0x22880002,0x28828a28,0xa202002,0x28208a2,0x8808208,0x88280200,0x22000022,0x8028008,0x220a282,0x20080002,0x228000,0x28a20a,0x82082820,0xa20028,0x88208200,0x80020000,0x20028200,0x80080282,0x808000a0,0x8820800,0xa282088,0x808a0022,0x20208000,0xa208202,0x200a2822,0x8028228,0xa080288,0x228a08a2,0x28820820,0x28800a,0xa0802082,0x800028,0x8a28008a,0x828a20a0,0x8000808,0x80280282,0x8a2880,0x28828020,0x82082202,0x80020802,0x8a08800,0x80000200,0x828a2000,0x8a20000,0x8a20a28a,0x82880800,0xa08228,0x2202082,0xa20a0080,0x8800008,0x8200a002,0x808828a0,0x8020000,0x8008a000,0xa2882,0x20820a08,0x28800a,0x22082080,0x20008000,0x8800a080,0x200200a0,0x20200028,0x2208200,0x828a0822,0x20228008,0x88082000,0x82882080,0x8028a08,0x200a08a,0x828200a2,0x8a28228,0x8a208208,0x800a2002,0x28200820,0x8020000a,0x228a0022,0x20a28828,0x82000202,0x200a00a0,0x228a20,0x8a20a008,0xa0820880,0x20008200,0x2282200,0x20000080,0xa28800);
remainders_pos[688].modulo_value=344; remainders_pos[688].bit_mask=create_bit_mask(11,0xa2a228a,0x82aa0280,0x8020880,0xaa82022,0x8a8a0028,0xa228aa28,0xa02800a2,0x2088082a,0x82022080,0xa00280aa,0xa288a8);
remainders_neg[688].modulo_value=344; remainders_neg[688].bit_mask=create_bit_mask(11,0x82a2aa02,0xa228208,0x808a8008,0x8220a8aa,0x20288a0,0x2aa022a0,0x2820882a,0xa80080a2,0xa8aa808,0x288a0822,0x2a0020);
remainders_pos[689].modulo_value=689; remainders_pos[689].bit_mask=create_bit_mask(22,0xa23f0736,0x44062f57,0x847f33bd,0x3212aa14,0xf876759,0x60921df7,0x44cc2192,0x2524a106,0x14953969,0xf5c21f,0x2b51d71d,0xbc02e3ae,0xa4a3e10e,0x92925a72,0xcc898214,0x24192610,0x87c3bee1,0x21326b9b,0xf884a155,0x808af733,0xf117abd1,0x1b383);
remainders_neg[689].modulo_value=2756; remainders_neg[689].bit_mask=create_bit_mask(86,0x2aa28aaa,0x888aa20a,0xa2aa20,0xaa8a2288,0x20a2a80,0xa80a802a,0x8800a802,0xa828288a,0x8808a0a0,0x88280282,0x22880a80,0x20822a22,0x28222882,0xa0a82aa,0xa08,0xa080aa88,0xa0a2228,0xa8aa2202,0x20082808,0x8022aa2,0x280a2208,0x82a02a0a,0x822280a2,0x2aa20080,0x80a08022,0x222a8a8,0xa2228282,0x8aa80828,0x82800000,0xaa0a8280,0x8a220a2,0x22a20822,0xa808a22,0xa00a088,0x2828808a,0x88a0a0a8,0xa8008a,0xa00a80aa,0xaa28202,0x8a228aa8,0x22aa2800,0x822a8888,0xaa8a2aa2,0x200002,0x88002082,0x88200220,0x2208a,0xa28000a8,0xa80a200,0xaa80a80,0x80a0880,0x8a802028,0x8a0a80a0,0x208a8028,0x20820220,0xa0208a2,0x82802002,0x2aaa808a,0x8a80082,0x82022208,0x80022a2,0x2a8a0088,0xaa20002,0x882228a,0xa0220282,0xa2208820,0x22a828,0x8028aa20,0xa2000808,0x222020a2,0x800a8808,0x80aaaa20,0x200a0a8,0x88202820,0x20208222,0xa88202,0x80a8288a,0x200a882,0x8828080a,0xa80aa800,0x2280a800,0x8000a280,0x8220000a,0x200208a8,0x82000882,0x20020);
remainders_pos[690].modulo_value=2760; remainders_pos[690].bit_mask=create_bit_mask(87,0xa0082082,0x28228000,0xa080000,0x20028a2,0x800028,0x88280288,0x2000800,0x800228,0x220a002,0x2080022,0x820200,0xa088282,0x228a2000,0x828820,0x82082208,0x20000800,0x200,0xa082088,0x2022020,0x20208808,0x200082,0x20200a0,0x8820028,0x88002002,0x22002020,0x8200008,0x82208280,0x208a2020,0x208008,0x208,0x80000002,0x8020,0x2008280,0x8020a2,0x20000220,0x280202,0xa00a00a0,0x800008,0xa208080,0x82020,0x28a00000,0x8a000002,0x2080000,0x208000,0xa200,0xa2880,0x8200000,0x208a008,0x20002,0xa00a20,0x8028000a,0x80000880,0x8020008,0x8200808a,0x2000002,0x208,0x80000080,0x2080020,0x8a20820,0x82088208,0x82002,0x8008820,0x8002208,0x822080,0x808028,0x8000a,0x22080882,0x8808020,0x820a008,0x80000022,0x20000800,0x88208200,0x22820020,0x8a28808,0x8220a000,0x80820082,0x208000,0xa088088,0x80020080,0x20008028,0x80282200,0x20022,0x28008028,0x20a08a,0x2882800,0x8200a00,0x82);
remainders_neg[690].modulo_value=2760; remainders_neg[690].bit_mask=create_bit_mask(87,0x20002802,0x20220800,0x2008088,0x82802022,0x8820,0x200200,0x82080080,0x8008220,0x8a20208a,0x82800022,0x8020a08,0x82000202,0x22022000,0x8020028,0x200a200,0xa0880080,0x8008a00,0x200a000,0x828228a0,0x20a00000,0x8028000a,0x820a0820,0x28020,0xa08a,0xa2882820,0xa00800,0xa200200,0x200220a0,0x8a00800,0x80080200,0x880882,0x8000820,0xa080208,0x80002822,0x20808a28,0x8820020a,0x20020820,0x8008000,0x82200000,0x80802820,0x20200808,0x2088082,0x2800080,0xa00808,0x80082208,0x80822000,0x208800,0xa002080,0x80820882,0x208228,0x8200082,0x80000,0x820000,0xa000002,0x82880802,0x8800200,0x8088000,0x28000a0,0x228028,0x2000200,0x80802882,0x20,0x8a280,0x22800,0x820,0x88000082,0xa2000002,0x28,0x80202080,0x80822,0x28800008,0x280280,0x220200a0,0x220000,0x2282008,0xa0802,0x8200800,0x2000000,0x820000,0x20800820,0x8202288,0x800a0822,0x20800020,0x88202002,0x82002000,0xa00208,0x2);
remainders_pos[691].modulo_value=2764; remainders_pos[691].bit_mask=create_bit_mask(87,0xaa888a2a,0x22822022,0xaa8a0828,0x228a882a,0x82a0088,0x82a02a28,0x200a82a2,0x2a2a822a,0xa08a0820,0xa00aa20a,0x228a202,0xa88a22,0xa0a20220,0xaaa80000,0x8028082,0x8a8282a0,0x88a20882,0xaa208aa8,0x2a2aa80a,0x20a2228,0x88808002,0x8a02822a,0xaa2a2200,0x88828a8a,0xa00080a0,0x2288002,0x22882220,0xa0a020a,0x2a0aa2a8,0xaaa0000a,0x8a88282a,0x82202aa8,0x88208a88,0x8802a80a,0xa2822802,0x8a0080a8,0xa02a880,0x80a80a08,0x2a80a2a0,0xa20208a2,0xa2820000,0xa88a08a0,0x82220008,0x8882000,0xa22a200,0x828a0a2,0x808a000,0xa02a88a2,0x2a020a8,0xa80a020a,0x200a2022,0x8828a2a0,0xa8022802,0x20822a02,0x8088222a,0x822a22aa,0xaaaa82,0xaa0a8a00,0xa0a02a8,0x82288a08,0x22880088,0x2000a800,0x282220a0,0x8a8aaa2a,0x280a2008,0x20222a88,0x8a080800,0xaa8a8288,0x208aaa02,0x8a222aa,0x282a2822,0x280200a8,0x2aaa820,0x8a0a000,0x22a00088,0xa288088a,0xaa00a808,0xa20aa08,0x2a8a8082,0x2a008a88,0x80a828a8,0xa82028a,0x2a288220,0xa2aaa82,0x88288282,0x222aa880,0xa8a);
remainders_neg[691].modulo_value=2764; remainders_neg[691].bit_mask=create_bit_mask(87,0x220002a2,0xaa0aa8aa,0x220280a0,0xaa0200a2,0x80a28800,0xa28a2a0,0xa8820a2a,0xa2a20aa2,0x280280a8,0x28822a82,0x8aa02a8a,0x882002aa,0x282a8aa8,0x22208888,0x808a080a,0x20a0a28,0x2a800a,0x22a80220,0xa2a22082,0x8a82aaa0,0x8088a,0x2820aa2,0x22a2aa88,0xa0202,0x28880828,0x8aa0088a,0xaa00aaa8,0x82828a82,0xa2822a20,0x22288882,0x200a0a2,0xaa8a220,0xa80200,0x8a2082,0x2a0aa08a,0x2880820,0x828a2008,0x8208280,0xa2082a28,0x2a8a802a,0x2a0a8888,0x20028028,0xaaa8880,0x8000a888,0x82aa2a88,0x80a0282a,0x80802888,0x28a2002a,0x8a28a820,0x20828a82,0xa882a8aa,0xa02a28,0x208aa08a,0xa80aa28a,0x800aaa2,0xaa2aa22,0x8822220a,0x22820288,0x82828a20,0xaa00280,0xaa008800,0xa8882088,0xa0aaa828,0x20222a2,0xa082a880,0xa8aaa200,0x2808088,0x22020a00,0xa802228a,0x802aaa22,0xa0a2a0aa,0xa08a8820,0x8a2220a8,0x80282888,0xaa288800,0x2a008002,0x22882080,0x82a82280,0xa202080a,0xa2880200,0x820a020,0x820a8a02,0xa2a00aa8,0x82a2220a,0xa00a0a,0xaaa22008,0x202);
remainders_pos[692].modulo_value=1384; remainders_pos[692].bit_mask=create_bit_mask(44,0xa2a0a202,0x28a8a2a,0x822a2208,0x2a22000,0xa0a00aa0,0x288a0288,0x2a8028,0x800222a,0x8222a2,0x2aa28a8a,0xa02282a,0xaa8a8288,0x200a2a28,0x208a888,0x800a8880,0x2282802a,0xa0a2280a,0xa800aa00,0x88200088,0x2802088a,0xa8aa8a2a,0x202808a0,0xa2aa2a0a,0x208028a8,0x822a2,0xaa002a22,0x288a0a00,0x28288a0,0x22a002a8,0x2a208002,0xa8a00822,0x82a2aa28,0x2880a022,0xa28aa8a8,0x888200a2,0x8800208a,0x2a800a8,0x80a22828,0xa00a0a22,0x88a800a,0x88a88200,0xa2a28020,0x8a0a8aa8,0x80);
remainders_neg[692].modulo_value=1384; remainders_neg[692].bit_mask=create_bit_mask(44,0x2a282a8a,0x8a0202a2,0xaa2aa80,0x8a2aa888,0x28288228,0xa0028a00,0x88a208a0,0x8088aaa2,0x880aaa2a,0xa22a0202,0x828aa0a2,0x22020a00,0xa882a2a0,0x8a802000,0x8820008,0xaa0a08a2,0x282aa002,0x20882288,0xa88800,0xa08a8002,0x202202a2,0xa8a08028,0x2a22a282,0xa808a020,0x8880aa2a,0x2288a2aa,0xa0028288,0x8a0a0028,0xaa288a20,0xa2a8088a,0x202880aa,0xa2a22a0,0xa00828aa,0x2a022020,0xa882a,0x88a802,0x8a208820,0x2aa0a0,0x288282aa,0x80020882,0x200a88,0x2a2a08a8,0x2820220,0x8);
remainders_pos[693].modulo_value=924; remainders_pos[693].bit_mask=create_bit_mask(29,0x28a2002,0x20200220,0x80288,0x20020,0x8020a08,0x80082288,0x20800000,0x820808,0x8008000,0x200808a2,0x20000828,0x8000200,0x20208a0,0x28a00220,0x8800a002,0x8800a2,0xa20808,0x80080002,0x82820000,0x8a20200,0x2002,0x82020820,0x20000020,0x2288200,0x20a0802,0x800800,0x82280200,0x880080,0x8008a28);
remainders_neg[693].modulo_value=924; remainders_neg[693].bit_mask=create_bit_mask(29,0x82022002,0x28208a20,0x80008200,0x800a08a0,0x820200,0x2200,0xa0080080,0x8028000,0x80,0x20800022,0x28008020,0x80008288,0x820a0020,0x20208228,0x82002,0x822,0x228000,0x800808a,0x2020800,0x220a08,0x808a08a,0x28a0020,0x28808020,0x82200288,0x82020082,0x8008000,0xa200288,0x80000000,0x800220);
remainders_pos[694].modulo_value=2776; remainders_pos[694].bit_mask=create_bit_mask(87,0xaa08aaa,0x8882a82a,0x20aaa28,0x220a8aa8,0x88228a,0xa2028222,0xa8808200,0x2aaaa882,0x2080808,0x28a0a282,0x2288a20a,0x2a0aa00,0xa8a82a0a,0x22808080,0x2280222,0xaa08a808,0x8aaa828,0xaaa02a80,0xa8822020,0x2a220,0x20a28228,0x2222200a,0x820a20aa,0x2aaa822,0x22880a22,0x2a000a2a,0x8aa80,0x8a00828,0xa822a8a8,0x8a822222,0x28080a8a,0xa02a0a0,0x8802aa0,0xa0820a2,0xa8a2a282,0x28a8,0x8a880288,0x82a20aa2,0x288aa222,0x80a22082,0xa2a800,0x2888820,0x8a00a028,0xa2000000,0x80280a00,0x82222,0x2002a8a,0x2882088a,0x82888aa2,0xa28aa08a,0x228022,0x2a2a2800,0xa0828a8a,0x208a0820,0xa00aa802,0x280a0a80,0xa2a2a020,0x2a888882,0x202a2a88,0x28200a,0xa802aa20,0x88a8a000,0x8088a022,0x82882aaa,0x88aa08a0,0x8a00888,0x28828a,0x2a088a80,0xaa080882,0x2002a80a,0xaa282aaa,0x80202a20,0x88888028,0x2a020202,0x80a0a82a,0x8800aa0a,0x28a08a22,0x80828a0a,0xa8202020,0x2a822aaa,0x8a008202,0x888280,0x88a28822,0x802aa2a0,0x2228aaa0,0xa0a82a82,0xaaa20a);
remainders_neg[694].modulo_value=2776; remainders_neg[694].bit_mask=create_bit_mask(87,0x82280222,0xa20a2,0x8a8222a0,0xaa820220,0x8800aa02,0x2a8a0aaa,0x20080a88,0xa222200a,0x8a808080,0xa0282a0a,0xa2002a82,0x8a282288,0x2020a282,0xaa080808,0x8aa08aaa,0x22802080,0x802220a0,0x2228a208,0x200aa8a8,0x888a2aa8,0xa82a0aa0,0xaaaaa882,0xa82a822,0x8a2220aa,0xaa0082aa,0xa28882a2,0x88802208,0x802880a0,0x20aa2020,0x20aaaaa,0xa0808202,0x828a2828,0x8008a228,0x8280a82a,0x202a2a0a,0x8888a020,0x2008a00,0xa2a822a,0xa0022aaa,0x82aa80a,0x882a2088,0x8a0000a8,0x28828a0,0x2a888888,0x8a08288,0x8880aaaa,0x8a88a202,0xa00a8002,0xa00022a,0x2a022802,0x88aa08aa,0xa2a2a088,0x280a0202,0xa80280a8,0x2882200a,0xa0828208,0x2a2a28a8,0xa200000a,0xa8a2a200,0x88a0a882,0x208a22a8,0x202888,0x80028aa,0xa00a222,0x228028,0x80288000,0x88a00a02,0xa2800208,0x2280800a,0xa88a2082,0x22a0a222,0x8a8a2a8,0x8a0,0xa28a8a8a,0x82820a2,0x882282,0xa02802aa,0x80a0282,0x20a8a8a8,0xa20aa222,0x2880a8a,0x88000a08,0x2a00aa,0x8a22a28,0xaaa02228,0x2820a20a,0x222a82);
remainders_pos[695].modulo_value=2780; remainders_pos[695].bit_mask=create_bit_mask(87,0x200a0282,0x88228282,0x220a0228,0x82808802,0x22808022,0xa80a88,0x2808a0,0x20002280,0x882280a0,0x20a88a22,0xa8aa0a88,0x82a8202,0x2a882200,0x828080a8,0xa220a200,0x8088280a,0x200a0880,0x2a8a8080,0x8a80082a,0x88a200,0xa08a2880,0x22088a80,0x2082200,0x222a820,0xa2002a08,0x82,0x82288a22,0x220a2220,0x88000a0,0x22082a08,0xa000000a,0x228a80,0x288,0x20a002,0xa00802,0xa8a8280a,0xa02a8a00,0x200aa020,0xa00a002,0x22080822,0x80880a82,0x8220082,0x2888208,0xa02080a,0x2082228,0x28200882,0x882a0220,0x8820208,0x8800a00a,0xa080aa00,0xa00a2a80,0xa0282a2,0x80200a0,0x800a080,0x2280000,0xa02a2880,0x8a000000,0x20a8208,0x80a00022,0x20888a08,0x88a2288,0xa8200000,0x20a8008,0x82a888,0x80088208,0xa02a2208,0x228a20,0x2008a220,0x8a82002a,0x80202a2a,0x20220a00,0xaa028220,0x2008a088,0x82a02028,0x8822a,0xa8082a82,0x822a0aa2,0x288a22a0,0x80a02882,0x288000,0xa20280,0x822a02a0,0x28802028,0x88022028,0x22880a08,0x88282882,0x8280a00);
remainders_neg[695].modulo_value=2780; remainders_neg[695].bit_mask=create_bit_mask(87,0xa8820a0a,0x2a0a02,0x2a8282a0,0xa00008a,0xa208002a,0x88200200,0x88208020,0x2888a208,0xa20828,0xa020022a,0x20220200,0x80220a82,0x2200a288,0xa000820,0x22a82a08,0x8002082,0xa8028000,0x22020008,0x20080a2,0x80002a08,0x28022008,0xaa000200,0xa80a288,0x8aa220a8,0x2288a200,0x8888080a,0xa2002a2,0x2a82a2a8,0x80008828,0xa280a200,0x28880882,0x882a0200,0x8888200,0x88a0288a,0x8028800a,0x20202082,0x28220280,0x288220a8,0x8280288a,0xa2808022,0x800020a,0x802a8802,0xa000280,0x82828082,0x8280aa20,0xa0a8000a,0x228aa0,0xa8280,0x802882,0x20082208,0x28822208,0x820a0a22,0x8a8028,0x80802808,0x82a08808,0x28a22008,0x2080880,0xa820280,0x82088aa,0xa0000200,0x80022a00,0x20288880,0xa820080,0x88022000,0x800a00,0x28a22a80,0x882a02a0,0x288022a8,0x20288a2,0xa8a222,0xa8aa0288,0x220a0aa0,0x28802000,0x220a8a0,0x80800a22,0x2080220a,0xa228222,0x2002a228,0x820a00a,0x80a00808,0x882a0a08,0xa228a20,0x2008a0a0,0x82a8a0,0xa2008200,0xa0200a,0x208280);
remainders_pos[696].modulo_value=696; remainders_pos[696].bit_mask=create_bit_mask(22,0x82820822,0x20a20220,0xa208088,0xa20a0000,0x28200a28,0xa002280,0x2020,0x28a28020,0x8088200,0x2022002,0x208200,0x80008208,0x20800880,0x28008220,0x8028a,0xa0080800,0x28028800,0x8a28a008,0xa00000a0,0x8220208,0x8208808a,0x882082);
remainders_neg[696].modulo_value=696; remainders_neg[696].bit_mask=create_bit_mask(22,0x20a00a2,0x28228a28,0x82280000,0x22820880,0x20a08220,0x8208a208,0x808820a0,0x20220828,0x80000288,0x28a2882,0x8a00a08,0x8080280,0xa0080000,0x20800a28,0x88008202,0x20800080,0x20820008,0x2202000,0x20880820,0xa28a00,0x2000002,0x2802);
remainders_pos[697].modulo_value=697; remainders_pos[697].bit_mask=create_bit_mask(22,0x3b655bde,0x8d471c13,0x81daa987,0xc259309f,0x4c494017,0x6223d3c4,0x852849df,0x5a8c3bcb,0x30088335,0xb2fa3a49,0x7b85d090,0x34242e87,0x3249717d,0x6ab30440,0x874f70c5,0x1bee4852,0xc88f2f11,0xfa00a48,0x7e43269,0xc786556e,0x7320e38a,0x1ef6a9b);
remainders_neg[697].modulo_value=2788; remainders_neg[697].bit_mask=create_bit_mask(88,0xa2a88202,0x82808a,0x802200a,0x200a802,0x8080888a,0xa2aa0a08,0x8a08002,0x200a202,0x28800aa8,0x2a22a280,0xa2080800,0xa8a8a20a,0xaa80a820,0xa22a8888,0xa82a000,0x8222808a,0x22288,0x82288280,0x8a20aaa0,0xa0a8822,0xaa282a02,0x2022a202,0xa2aa022a,0x828202a0,0x228a208a,0xa2082aa8,0xa08,0x22088a22,0xa82880a,0xa228002a,0xaa8888a,0xa8a820a8,0x822a8228,0x22a00080,0x8a00a2a,0x200aa80,0x28820228,0xaa2a0008,0x8088282,0x2028888,0x8200a8,0x8028022,0xaa2a880a,0xa2a20208,0xa0802000,0x2a82a880,0xaa28a22,0x888280a8,0x8288a0a8,0x2aa00002,0x22a20828,0x8008a2a8,0x20a008aa,0x8a8a0220,0x2200228a,0x8280800,0x808800a8,0xa882200,0x288280a0,0x2088a220,0xa0aaaa20,0x28a200,0x88208208,0x2a28200,0x22a0000a,0xa2022220,0xa0280002,0x88020a2,0x82082,0x8280228a,0x82022aaa,0x8a022288,0xa80280a0,0x80802208,0x80a8008,0x2200080a,0xa8a8a200,0x82022028,0xaa8802,0xa20aa288,0xaa0a0822,0xa0a00002,0x888a8288,0xa80a80a0,0xaa2228a2,0x82808aa0,0xa2800200,0x2);
remainders_pos[698].modulo_value=2792; remainders_pos[698].bit_mask=create_bit_mask(88,0x82a2aa02,0x2a220a08,0x80820200,0x2208828,0x2a2a0a8,0xaa082228,0xa0288a28,0xaa02aaa2,0x2aa888a,0x2a20080a,0xa2a8aa0,0x28a2022a,0x8a0808aa,0x22000222,0xa2a2228a,0x202a0028,0x882aa80a,0x8820a2a2,0x802800aa,0x288a8a80,0x800a0aaa,0x828a22a2,0xaa820220,0x882000a0,0x2a8a8080,0x28800aa8,0x2882020a,0x2a20a800,0x2082aa0,0xa2208222,0xa08222aa,0x82008a8,0x202022a2,0x202a0080,0x2a0a8aa,0xa0008880,0x2a020002,0xa0082088,0x2a82228,0x22a80a02,0x8a88288a,0x202aa2a2,0x288a0a2,0xa82a002,0xa228080,0x8aa8088a,0x2822a28a,0xa02a88a2,0x882a8080,0x8200a28,0x80a822,0x22000a80,0x2a0a8002,0xa808aa,0x88080802,0x2008208a,0x88820a2a,0x82088aaa,0xa8208088,0x2a08a80a,0x80822800,0xa00228a0,0x2a2a82a,0x82202,0x8082aa0a,0x88a28208,0xa0a0028a,0xa2a228aa,0x280202,0x8a0822aa,0x2aa8228a,0xa808a0,0x888a8a28,0x800088a2,0x2020a288,0x808a28aa,0xa2a8a0a8,0x2008a80a,0x22aa80a0,0xaa80aaa2,0xa2280a8a,0x8820aa28,0xa8a8028,0x2208802a,0x80820228,0xa088a800,0xaa8a8220,0x80);
remainders_neg[698].modulo_value=2792; remainders_neg[698].bit_mask=create_bit_mask(88,0xa2a228a,0xa2aa8280,0x80a8a88,0x8aa800a0,0x8a2a2820,0x2280aaa0,0x28a002a0,0x228a222a,0x8a220002,0xa2a88082,0x82a20228,0xa02a8aa2,0x2808022,0xaa888aaa,0x2a2aaa02,0xa8a288a0,0xa22082,0xa82a2a,0x8a08822,0xa0020208,0x8828222,0xa02aa2a,0x220a8aa8,0xa88828,0xa2020808,0xa0088220,0xa00a8a82,0xa2a82088,0x8a80a228,0x2aa80aaa,0x280aaa22,0x80a88020,0xa828aa2a,0xa8a28808,0x8a282022,0x28880008,0xa28a888a,0x2880a800,0x8a20aaa0,0xaa20828a,0x200a002,0xa8a22a2a,0x8a00282a,0x820a288a,0x82aa0808,0x2208002,0xa0aa2a02,0x28a2002a,0xa20808,0x80a882a0,0x880820aa,0xaa888208,0xa282088a,0x88208022,0x80808a,0xa880a802,0xa82a2,0xa800222,0x20a80800,0xa2802082,0x80aa088,0x288aa028,0x8a2a20a2,0x8880aa8a,0x80a2282,0x2a0a80,0x28288a02,0x2a2aa022,0x88a08a8a,0x280aa22,0xa220aa02,0x88208028,0x202a0,0x888002a,0xa8a82a00,0x802a022,0x2a202020,0xa8802082,0xaa220828,0x2208222a,0x2aa08202,0xa822a0,0x820208a0,0xaa8008a2,0x80a8aa0,0x28002088,0x22020aa8,0x8);
remainders_pos[699].modulo_value=2796; remainders_pos[699].bit_mask=create_bit_mask(88,0x2822022,0x220a20,0x20a08a,0xa2082880,0x20a00a28,0x2200288,0xa28a2880,0x8220020,0x2000282,0xa0000802,0x20228220,0x8008288,0x200208a2,0x8820a00,0x8008a202,0xa0800882,0xa20828,0x82008082,0x22822820,0x28a28a00,0x82208280,0x28228a2,0x8000008,0x280208,0x8a0000,0x8208008,0xa08220a,0x20002080,0x28020228,0x2088002,0xa08a0082,0x20800208,0x8820020a,0xa0080800,0x8828228,0x8208028a,0x80880,0x8a20000,0x8200288,0x800000a0,0x28a20020,0x8a002008,0x2880080,0x20008808,0x2280202,0x800a2020,0x8a28200,0x2080,0x8200a0,0x8a20228,0x2000000,0x2082022,0x28220a28,0x200a288,0x822002,0x208a08,0xa20a208,0x22080820,0x8028800,0x80008288,0x820a0020,0x20820a08,0xa200200,0x2800000,0x20208,0x88280200,0x208828a2,0x28a28028,0x8828800a,0x20082082,0x8a00820,0x20a282,0xa2002822,0x8220808,0x800800a,0x200208a2,0x28808228,0xa088,0x80802,0x8820828,0x8a28a080,0x880022,0xa08228,0x8208a28a,0xa0800022,0x8800a20,0x8828008a,0x880);
remainders_neg[699].modulo_value=2796; remainders_neg[699].bit_mask=create_bit_mask(87,0x820a28a2,0x8a28228,0x88282002,0x22802000,0x28208220,0x8a288200,0x22022000,0xa28828,0x8a08820a,0x20880082,0x28a20a28,0x80080200,0xa08a0022,0x28208,0x800228a,0x20080002,0x8228020,0xa08000a,0xa20a20a0,0x20220208,0xa280208,0x820a2022,0x808800,0x88208280,0x80020880,0xa00800,0x8200a282,0xa0882800,0x20828a20,0x8a00008a,0x20020802,0x28008a00,0x288282,0x20800080,0x20a20,0xa008202,0x80800000,0x228808,0x80288200,0x880820,0x20228828,0x208a080,0x82000800,0x28800000,0x8a20828a,0x8228a0,0x220a08,0x8808a008,0x800a0820,0x228a20,0xa088088,0x828028a2,0x20a28220,0x8a082200,0x800a2882,0x8a00200,0x82282280,0xa28000a0,0x820008,0x8080200,0x28208a0,0x28028200,0x82288288,0x82080880,0x8828a00,0x208288,0xa0002022,0x20220820,0x200082,0xa0802802,0x208028,0x8828220a,0x228820a2,0xa28000,0x80080082,0xa08a0022,0x20000a20,0x88082000,0x80800082,0x28020,0x2202000,0x800008a2,0x8200a20,0xa002202,0x200808a2,0x8228,0x208002);
remainders_pos[700].modulo_value=280; remainders_pos[700].bit_mask=create_bit_mask(9,0xa808020a,0xa208020,0x208a0000,0x202a080,0x8020aa08,0x8020a80,0xa00000a2,0x2a080208,0xa08020);
remainders_neg[700].modulo_value=280; remainders_neg[700].bit_mask=create_bit_mask(9,0x20800a02,0x22808a0,0x20028088,0x8a022808,0xa02200,0x808a0208,0x28088022,0x22808200,0x2008a0);
remainders_pos[701].modulo_value=701; remainders_pos[701].bit_mask=create_bit_mask(22,0xf35b22f2,0x80526e1a,0xcb2b5433,0x37f5c4f3,0xcec77a78,0x8592fbb8,0xafe52ad0,0x70e022e6,0xc8b14037,0x2f892762,0x229a18bf,0x3f461651,0xd1b9247d,0xbb00a344,0x59d101c3,0x42d529fd,0xc777d268,0x797b8dc,0xf3c8ebfb,0x730ab534,0xd61d9280,0x13d136b3);
remainders_neg[701].modulo_value=2804; remainders_neg[701].bit_mask=create_bit_mask(88,0x2a82aa2a,0x88aa282,0x2a288aa,0xaa28082a,0x20aa2a0,0x80a2220,0x2228a208,0xa828aa2a,0x2888aa,0xa200aaaa,0xaa028022,0xa28a8a88,0x820a8a0,0x22882a88,0x8008880a,0x8a08a020,0xaaa0aa0,0x8a0a2000,0x2a002222,0xaa8228a8,0xa800a08,0x8a08aa2a,0x82a2a882,0xa882800a,0xa0a8a20a,0x8a222002,0x80002282,0x882a82aa,0x8202882,0x22808880,0x808aa08a,0x2828a820,0xa88a8a8a,0x2a200a02,0x2aaa802,0xaaa888a0,0x22a2a8a0,0x808228a2,0x222280,0xa82a2a82,0x2a080a2,0x82a88a2a,0xa20a2a88,0x2a2aa0a,0xa200a002,0x2888020,0x2a02088,0x28020a8a,0x2aa28020,0x8a8a8222,0x2220822,0xa0802,0x2aa0888,0x2022a800,0x8802a0aa,0x28208080,0x8a8a2808,0x82208a00,0x28aa888a,0x2882882a,0xaa800280,0x2082822a,0xa02aa22,0x8800a208,0x280aa82,0xa0028802,0x80a02008,0x8008a0aa,0xa0280822,0xa082222a,0xaaaa20,0xa08a00a0,0xaa8a2a08,0x8220a888,0x28a88028,0x820a080a,0xa0088080,0xa2022a82,0x2aa0000a,0x28000888,0x22202008,0xa8a8a208,0xa2aa2220,0x200a0200,0x2a028a8,0x88a00882,0x80228200,0x2002);
remainders_pos[702].modulo_value=312; remainders_pos[702].bit_mask=create_bit_mask(10,0xa2800882,0x8220a20,0x82282000,0x2002020,0x800a00,0x8000a002,0x82080800,0x20000828,0x8a08a088,0x822002);
remainders_neg[702].modulo_value=312; remainders_neg[702].bit_mask=create_bit_mask(10,0x22080002,0xa28228,0x220a088,0x828828a0,0x8000208,0x808208a,0x2800080,0x28008020,0x2002000,0xa2082);
remainders_pos[703].modulo_value=2812; remainders_pos[703].bit_mask=create_bit_mask(88,0x2a80220a,0x820888,0x2a22aa8,0x82a22a22,0x28a8020a,0x88000228,0x8a28a2a2,0x880a8aa2,0x8a000a22,0x88a888a,0xa82802a2,0x2a080820,0xaa022a08,0x82000820,0x808a80aa,0x8888200,0x28028,0xa2028200,0x8a288a8,0xa28222a0,0xa82a0a88,0xaa08008,0x22aa2802,0x22028080,0xa8088008,0x20200822,0xaa020a88,0x80aa0aaa,0xa8a2222,0x2a0222,0x28a2a88a,0xa2800a00,0x2882a280,0x8008a0a0,0x2220222,0x82aa80,0x8020222,0x8080082,0xa082aa82,0x8aa020,0x880080a,0xa000880a,0xa0228a0a,0x288a200,0xa008a228,0xaa0a2880,0xa022000,0xa020022,0xa080aa20,0x82aa820,0x8200202,0x8880802,0x2aa820,0x28880888,0x80a0a200,0xa028a822,0x800a0028,0xa22a8a2,0x8880a80,0x28888a2a,0xaaaa0aa0,0x822a080a,0xa8820080,0x82002202,0x80202808,0x8028aa8,0xa20020aa,0xa22a0a82,0xa88828,0xa2a228a2,0x82808,0x2802800,0x20082222,0x2aa02a20,0xa0820008,0x820a880a,0xa082020a,0x8a80282,0x2a222a22,0x288a000a,0x28aa2a02,0x28a8a28a,0x82880002,0x2a0802a2,0x88a88a8,0x2aa88a8,0x82220820,0xa08802a);
remainders_neg[703].modulo_value=2812; remainders_neg[703].bit_mask=create_bit_mask(88,0xa200aa82,0x880a8000,0xa2aa220,0xa2a22aa,0xa0208a82,0x8882a0,0x2a02a2a,0x2022a,0x28882a2,0x80020002,0x20a08a22,0xa28080a8,0x220a2280,0xa8880a8,0x20822,0x80000a88,0x888a08a0,0x2a8a0208,0x802a0020,0x2a02aa28,0x20a28200,0x2280880,0xaa22a08a,0xaa8a0808,0x20800080,0xa8a880aa,0x22028200,0x8228222,0x8202aaaa,0x88a28aa2,0xa02a2002,0x2a080208,0xa00a2a08,0x802828,0x8aaa8aaa,0x800a2208,0x808a8a2a,0x8080880a,0x2802220a,0x880228a8,0x80082,0x28880082,0x28aa0282,0x8a002288,0x28802aa0,0x2202a000,0x828aa888,0x828a88aa,0x280822a0,0x802220a8,0x80a80a8a,0x8000808a,0x80a220a8,0xa0008000,0x8282288,0x28a0202a,0x88288a0,0x82a2202a,0x8208,0x200002a2,0x22228228,0xaa28082,0x200a8008,0xa80aa8a,0x828a080,0x808a0220,0x2a88a822,0x2aa28202,0x88200020,0x2a2a202a,0x8880a080,0x8208a088,0xa080aaaa,0xa228a2a8,0x280a8800,0xa820082,0x28028a82,0x80200a0a,0x22aaa2aa,0xa0028882,0xa022a28a,0xa0202202,0xa008882,0xa2008a2a,0x80020020,0x8a220020,0xa2a80a0,0x28008a2);
remainders_pos[704].modulo_value=88; remainders_pos[704].bit_mask=create_bit_mask(3,0x20802a2,0x802a28a8,0x8a8020);
remainders_neg[704].modulo_value=88; remainders_neg[704].bit_mask=create_bit_mask(3,0x8a80822a,0x822a020,0x208a8);
remainders_pos[705].modulo_value=2820; remainders_pos[705].bit_mask=create_bit_mask(89,0x20022802,0x20220a00,0x88208,0x82000000,0x8020,0x808220a,0xa2000022,0x20000,0x2008088,0x82800002,0x8000800,0x8a208282,0x20000020,0xa20008,0x200a00a,0x20022000,0x28220000,0xa008008,0x82802082,0x20008a00,0x80202082,0xa2020882,0x8820228,0x820200a,0x2880022,0x8200a08,0x80080282,0x220a0,0x220020,0x80280200,0x80022802,0x20000020,0xa082288,0x800082,0xa08,0x88080202,0x20000,0x20000228,0x82208082,0x20080002,0x208,0x82088000,0x2822020,0xa20020,0x8008a00a,0x80882800,0x220820,0x8000000,0x20082,0x20208828,0x88000088,0x80002,0x8020220,0xa000008,0x20002002,0x28820a08,0x80000082,0x2880020,0x8028028,0x80088000,0xa0880000,0x28020020,0xa008208,0x80022802,0x808208,0x8808220a,0x220808a2,0x20808028,0xa200088,0x20802820,0x200a08,0x88282,0x880080,0xa00800,0x8a00a,0x80000082,0x28208a20,0x2000208,0x2820,0x20200808,0x80002,0x800008a0,0x8820208,0x8020000a,0x820,0x8220000,0xa088082,0x2880080,0x8);
remainders_neg[705].modulo_value=2820; remainders_neg[705].bit_mask=create_bit_mask(87,0xa08a2082,0x28220200,0x8000280,0x2800880,0x800828,0x80002282,0x220808a2,0x8828008,0x8a000000,0x2080802,0x800008,0x2280202,0x208800a0,0x8228800,0x82082002,0xa08a2880,0x20228800,0x2080080,0x2002802,0x20800200,0x828200a,0x220a0002,0x28220,0x8020a082,0x2000822,0xa00200,0x8008202,0x8a2020,0x8a28828,0x208208,0x8a2082,0x28008820,0x200a200,0x80000802,0x808200,0x28a,0x800a0880,0x28808220,0xa20000a,0xa0800802,0x8800a00,0xa000080,0x20220a0,0x8228828,0x2002,0x2080,0x8228020,0x88088,0x80820802,0x20a00020,0x80000,0x80000882,0x828228,0x82008080,0xa0882002,0x20020200,0x8088002,0x20000a0,0x820820,0x8,0x20000800,0x20028820,0x2080280,0x822082,0xa00,0x2282,0xa2000022,0x28000020,0x82208000,0xa0082020,0x8a00200,0x8000202,0x8208008,0x80002002,0x880802,0x20200220,0x2088280,0x808020a0,0x20a08000,0x8800008a,0x80020,0x28200,0x8208082,0x80880020,0xa20808,0x82000002,0x2000000);
remainders_pos[706].modulo_value=2824; remainders_pos[706].bit_mask=create_bit_mask(89,0xa82a22a,0x82aa2a0,0x222a2a,0xaa2a8822,0x822828a0,0x820a2288,0xaa0aaa08,0x88280aaa,0x2008200,0x8080a88a,0x8a2aa022,0xa8a22a00,0x208082,0x8a808a8a,0x22082800,0xa220a22,0x2aa0a80,0xa0282a80,0xa8022a20,0x2a2aa88a,0xa2a2280a,0x2aa28208,0xa2820022,0xa8220208,0x88002020,0x2022a800,0x80282a0a,0x80a002aa,0x20a220aa,0xa828a222,0x808a808a,0x8a8a2aa0,0xa0220802,0x2a002082,0x880a8a82,0xa28aa2a0,0xa8288a,0xa000a000,0x8220a288,0x8282288,0x2880200a,0x20222aa2,0x20028a0,0x220280aa,0x2808800,0x280080aa,0xa888080a,0x802288a,0x882820a0,0x8a088222,0xa000a22,0x282a0000,0x8aa28aa2,0xa2a0220a,0x800a882,0x20880a82,0xa8a2a280,0x2a2020a,0x8a282aa2,0x88a0888,0x800a02aa,0xa82802aa,0x2a8808a0,0x8002200,0x80882a08,0x828a20,0x828aa888,0x288a8a20,0x2aa8a8a0,0xa8802aa2,0xa8280a08,0xa0aa8002,0xa088a002,0x28208888,0xa202a200,0x20800a2,0xa88a2a82,0xaa8a200,0x2a020288,0x820080a2,0xa0282200,0xaaa0aaaa,0x88a08220,0x28288222,0x22a8aa0a,0xa8880088,0x8aa820a8,0x8a82a00a,0xa8);
remainders_neg[706].modulo_value=2824; remainders_neg[706].bit_mask=create_bit_mask(89,0x820a2aa2,0x80a22a28,0x88aaa2a2,0x22a200aa,0xaa0a028,0xa82aa00,0x22822280,0xa08222,0x8a880a88,0x8082002,0x2a228aa,0x202aa288,0x88a8080a,0x2080202,0xaa80a088,0x82aa82aa,0x8a228208,0x28a0a208,0x208aa2a8,0xa2a22002,0x2a2aa082,0xa22a0a80,0x2a0a88aa,0x20aa8a80,0x88a8a8,0xa8aa2088,0x8a0a282,0x8288a22,0xa82aa822,0x20a02aaa,0x8020802,0x202a228,0x28aa808a,0xa288a802,0x82020a,0x2a022a28,0x8820a002,0x28882888,0xaa82a00,0x80a0aa00,0xa008a882,0xa8aaa22a,0x8a88a028,0xaa8a0822,0x8a080088,0xa0880822,0x20008082,0x808aa002,0xa0a828,0x2800aaa,0x828882aa,0xa0a28888,0x22a022a,0x2a28aa82,0x8088200a,0xa800820a,0x202a2a08,0x8a2a8a82,0x2a0a22a,0x80028000,0x8828a22,0x20a08a22,0xa2008028,0x8088aa88,0x800a280,0x880a02a8,0xa022000,0xa00202a8,0xa2202028,0x2008a22a,0x20a08280,0x2822088a,0x2800288a,0xa0a80000,0x2a8a2a88,0x8a80882a,0x2002a20a,0x82202a08,0xa28a8a00,0xa88082a,0x28a0aa88,0x22282222,0x280aa8,0xa0a00aaa,0xaa202282,0x20008800,0x220a820,0x20a2882,0x20);
remainders_pos[707].modulo_value=2828; remainders_pos[707].bit_mask=create_bit_mask(89,0x82088a02,0x200882a0,0x2a2022a,0x82080008,0x8082028a,0x80022a28,0xa0a20022,0xa02222a,0x2a882282,0x222808aa,0xa8282882,0x82a2820a,0x80802880,0x8802a0a2,0x882a80a,0x2020a0,0xa2820028,0xa2aa82a,0x88a02a02,0x2022880,0x882a2008,0xa08a0000,0x2a802802,0xa080822a,0x200a88a2,0x80a0220a,0x80000288,0xa88a0000,0x28888a20,0x200822a0,0x200a8202,0x202a2808,0x8a000a0a,0xa02000,0x2020808a,0x8820000,0x8888282,0xa2880800,0x800aa808,0xa822a00,0x8aa20a00,0x888802a0,0xa02a800,0x2208a20,0x20880a0a,0xa80a008a,0x2222002,0x8aa20a8,0x882a000a,0xaa00200a,0x228a202,0x22220002,0x8220828,0x20808000,0x80a00a20,0xa2000,0x8a808a0a,0x2a008202,0x82008808,0x222280a8,0xa22a08a,0x2000,0x80a02228,0x2a008a08,0x2020a8a2,0x802a8a88,0xa20a802,0x8a822000,0x88080200,0x80a22022,0xaa8a080a,0x828aa82,0x80800280,0xa82200a0,0xa8022a02,0x802028a0,0x28a82022,0x8282aa08,0x2888822,0x822a8aa2,0x880a0828,0x8a0aa88,0x88002880,0x820228a,0x2082a28,0x8a80200,0x22008a88,0x220820a8,0x80a);
remainders_neg[707].modulo_value=2828; remainders_neg[707].bit_mask=create_bit_mask(89,0xa80020a,0x28800a20,0x822a8aa2,0xa008880,0x8028a02,0x88a22a0,0x282a80aa,0x828aaa22,0x2200aa02,0xa2a00022,0x2020a00a,0xa220a82,0x8082008,0x8a202a,0x800a2002,0x8a8a820,0x220a88a0,0x822220a2,0x20a28a,0x8a8a2008,0xa2a080,0x28028808,0x2208a082,0x20080aa2,0xa802002a,0x820aa82,0x8880a00,0x20028088,0xa0000228,0x2880aa20,0xa0820a8a,0xa822a080,0x2808282,0x88282888,0xa8280002,0x800a8808,0xa02,0x22008088,0x8022080,0x8202a288,0x22a0288,0x8228,0x828a2008,0xaa802a0,0xa0008282,0x20028802,0x8aa2a88a,0x80222020,0xa28082,0x2288a802,0xaa02a82,0xa2aa888a,0x802a80a0,0xa8000888,0x82802a8,0x8882a088,0x2080202,0x22880a82,0x2880080,0x2a2a0820,0x82a22802,0x88882888,0x828a2a0,0xa2880200,0x28a82022,0xa20200,0x8228208a,0x202a888,0x800a88,0x82aa0aa,0x22028002,0xa02202,0x8a08,0x202a8828,0x2082a28a,0x8a82028,0xa020a0aa,0xa0a2200,0xa0000a2,0x2a2022a,0x280a0,0x80202200,0x882008,0x80a8a202,0x8a80a220,0x208a00,0xa2880200,0xaa00a820,0x82);
remainders_pos[708].modulo_value=1416; remainders_pos[708].bit_mask=create_bit_mask(45,0x2880882,0x28000,0x228a002,0x82020020,0x8220828,0x8228a20a,0xa0822082,0x20020208,0x28a200,0xa0002,0x8808a00,0xa20220a,0xa0000,0x8a28008,0x8080080,0x208820a2,0x8a28828,0x8208820a,0x80080822,0xa28800,0x280008,0x22022800,0x28808828,0x280000,0x228a0020,0x20200200,0x82208288,0x228a20a0,0x8220828,0x20208a,0x28a2002,0xa00020,0x8808a000,0xa20220a0,0xa00000,0x8a280080,0x80800800,0x8820a20,0x8a288282,0x208820a0,0x808228,0xa288008,0x2800080,0x20228000,0x82);
remainders_neg[708].modulo_value=1416; remainders_neg[708].bit_mask=create_bit_mask(45,0x82000002,0x820808,0x8a20208a,0x28a08a0,0xa28020,0xa202282,0x200a2802,0x28828a00,0x88202288,0x80820802,0x208,0x8228a282,0x80820880,0x220800,0x80008008,0xa0002822,0x220020,0xa000282,0x8000a2,0x8220008,0x88208080,0xa28a2080,0x20000020,0x8208088,0xa20208a0,0x28a08a08,0xa280200,0xa2022820,0xa28020,0x8828a002,0x82022882,0x8208028,0x2088,0x228a2820,0x8208808,0x2208008,0x80080,0x28228,0x220020a,0xa0002820,0x8000a20,0x82200080,0x82080800,0x28a20808,0xa);
remainders_pos[709].modulo_value=709; remainders_pos[709].bit_mask=create_bit_mask(23,0x3e399aba,0xda83f81e,0x80977c09,0xa61317d8,0x60af75b6,0xb57f9e68,0x74ca0e97,0x19695cda,0x686c90e3,0xcb753ba2,0x1489c315,0x30e44a00,0x772bb4ea,0xc24d8591,0xcea5a631,0x5c14cb96,0x9e7fab7a,0x6bbd4185,0xfa32195b,0xfba4046,0x7f056e4,0x56671f1e,0x17);
remainders_neg[709].modulo_value=2836; remainders_neg[709].bit_mask=create_bit_mask(89,0xa2a00222,0x20a2082,0x80aa080,0x2a8a8a00,0xa822a82a,0x28a202a0,0xa802820a,0x80a08002,0xa0a0082a,0x2a8a22a,0x88000a88,0xa8288288,0xaaa22822,0xa800808,0x2282aa8,0x8088020a,0x2a222a2,0xa2208808,0x22aa8082,0x8222888a,0x8a288a28,0x8aaa8282,0xaaa8a8a,0x88a28a0a,0x8a2208a2,0xaaa2288,0x88222a08,0x222a0080,0x880a2a,0xa0a20282,0x800a80aa,0xa22aa880,0x8a0aa20,0x8000888a,0x28aa008a,0x80282aa2,0x8280aa0,0xa00aa00,0x228a282,0xaa20a82a,0x8a8aa2a0,0x2a808002,0x22820008,0x2a2a08,0x2aa22222,0x8228a020,0x8028aaa2,0x8082aa8a,0x200a0200,0x22a0280a,0xa0a28208,0x28aaa00a,0x2a8a00a8,0x8220028,0xaa808aa0,0x8a0888a,0x220a200a,0xaa8a8800,0xa000a80,0x8aa282a2,0x222020a8,0x288a8aa0,0xa8a022,0x20888220,0x88208a2,0xa0a082,0x800080,0x20828280,0x22882088,0x2208882,0x22028a80,0x2a8a88a,0x8a820222,0x22a0a2a8,0x80a80028,0x8a8aa,0x28022822,0xa8888288,0x2a880aa,0xa002208,0xa8028aa,0xa802aa8a,0x820a282,0x280a02a2,0x80202802,0xa8aaa080,0xa2aa8a00,0x2028a20,0x222aa);
remainders_pos[710].modulo_value=2840; remainders_pos[710].bit_mask=create_bit_mask(89,0x802a0202,0x200200a2,0xa822028,0xa020082,0xa220a20a,0x202a0a8a,0x88000802,0x2a00a008,0x88a02082,0x8008aa22,0xa802228a,0x2a288222,0x22008020,0xa82002a,0x8208a228,0x80880000,0x20228aa0,0x2a0aa220,0x8a0a0a0,0x2a82028,0x8a80a82,0xa8228822,0x2080028,0x8800a028,0x20008200,0x808280a,0x200a02,0x20a2220,0x80828800,0xa08a2a,0xa0802802,0x8a220022,0x8002228,0x2a00022,0xa88002,0x28080200,0xa200282,0x2a0880a8,0x802220a8,0x2200000,0xa0a02082,0x22008282,0x228a2008,0x8022800a,0x2888222,0x88a00288,0x882008a2,0xa828200,0x8082080a,0x2000008,0xa82a0888,0xa02a0220,0x28828008,0x8020,0x8080022a,0x2088000a,0xa2288800,0xa880088,0x802802,0x2a8a20a,0x80002282,0x888a0,0x2080a008,0x8a02820,0x22008200,0x80280a00,0x2a280020,0x20882288,0x8082a02a,0x2028082a,0xa80a0a0a,0x8088aa0,0x20aa288,0x82000022,0xa0288a20,0x88a80082,0xa8080200,0x2a888228,0x2a28880,0x2288aa20,0xa882080a,0x22200a00,0x8802000,0x8aa2a0a8,0xa0a08a08,0xa0820080,0x8280882,0x28a0080,0x8080a8);
remainders_neg[710].modulo_value=2840; remainders_neg[710].bit_mask=create_bit_mask(89,0x8a20a8a,0xa80a8822,0x20aa020,0x8282880a,0x22a82a02,0xa8a20202,0x88082,0x22882080,0x20a80a,0x80222a,0x208a2a02,0xa2200aa2,0x2a8800a8,0x820288a2,0x2802a20,0x8000888,0xa82a0220,0x228222a8,0x80202828,0x220a820,0x8020020a,0x202a00a2,0xa8080a0,0x8028a0,0xa0880a08,0x80802082,0x88288282,0xa82a2a8,0x20088,0x80280222,0x2808208a,0x22a88a2,0x88a2a0,0x8a2088aa,0x8020080a,0xa0800a88,0x82288a02,0x22000020,0x8a2a820,0x82a88808,0x2828280a,0xaa080a02,0x2a02a080,0x8a20882,0x82000a2a,0x280a00,0x208022,0x20a0288,0x8028082,0x82888800,0x20a20000,0x28228aa0,0x200a0080,0x888008a8,0x88a22,0xa8000882,0x2a200080,0x2008000,0x8800a08a,0x82202a02,0x8882a0a,0x88000020,0x28082080,0x8020a0a8,0xa2880208,0x8a00288,0xa22088a0,0x2800a200,0x80228a2,0xa0a08022,0x20820282,0x80000220,0xa822200,0xa80882a,0x20a00228,0x20080a,0x20008a80,0x220002a0,0x8a220008,0xa2002228,0x200a0082,0xaa288280,0x8a080,0x2222820,0x20280200,0x280a0808,0x80208002,0xa028008,0x820);
remainders_pos[711].modulo_value=948; remainders_pos[711].bit_mask=create_bit_mask(30,0x20020a2,0x8028808,0xa000282,0xa20008a2,0x800,0x88208200,0x8a0080,0x28220808,0x220a002,0xa2882082,0xa20a28,0x8002288,0x208020a0,0x28220028,0x8288,0x82822820,0x20828008,0x200a080,0xa0022880,0x28a28a08,0x88082082,0x828800a0,0x20020208,0x8220200a,0x820,0x8a00200,0xa08a200,0x2082800,0x8020228,0x8a080);
remainders_neg[711].modulo_value=948; remainders_neg[711].bit_mask=create_bit_mask(30,0x82882822,0x820000,0x8208020a,0x22880022,0x8808008,0x280288,0x80020800,0x20a28000,0x8a28208a,0x22002802,0x8228220,0x8008a200,0xa0082020,0x20a28820,0x88080200,0x20a20a0,0x28020800,0x8a082008,0x208a2000,0x20220200,0xa00a,0x2000820,0x28828a00,0xa28a082,0x808800a0,0x208a08,0x82002288,0x82802080,0x828a20,0x2008);
remainders_pos[712].modulo_value=712; remainders_pos[712].bit_mask=create_bit_mask(23,0x2a0a220a,0x2a8a88a8,0x8aaa80,0x2288222,0xa282002,0xa20aaa,0xaa2820a0,0x228822a2,0x800082a2,0x880802a,0xa220a028,0xa088a00,0x2022028,0x820002a8,0x8822888a,0x828aa8a,0xa08a000a,0x828a0aa,0x82288080,0xaaa20088,0x22a2a802,0x88a0a82a,0xa0);
remainders_neg[712].modulo_value=712; remainders_neg[712].bit_mask=create_bit_mask(23,0xa282aa82,0xa2020020,0x88022208,0x8aa00aaa,0x82a0a88a,0x882a8222,0x22a0a828,0xaa00aa2a,0x888022a,0x800808a2,0x2aa828a0,0x82800288,0x8a8aa8a0,0xa888a20,0xaa0002,0x80a02202,0x28028882,0x80a02822,0xaa00808,0x222a0800,0xaa2a208a,0x2820a2,0x28);
remainders_pos[713].modulo_value=2852; remainders_pos[713].bit_mask=create_bit_mask(90,0x2228a02,0x2a228aa8,0x8a8a8082,0x2000028,0x20202a8,0x28202228,0x8080a82,0x888202a2,0x80082,0x2aa8000a,0xa00aa082,0xa282a220,0x82202a,0x8a822282,0x22a00228,0x80200020,0x2a88282a,0x2a2028aa,0x808a8282,0x88aa08a0,0xa080222a,0x2880aa0,0x8a0228a2,0x8282a80,0xa88aa220,0x88082882,0x22a20a0a,0x8aa2a8a0,0x200aa0a0,0x2a202000,0xa88a200,0x8020a22,0xa2aa022,0x8028222a,0xaaa2082a,0x80028000,0x88a0800,0x80822a02,0x20a20a80,0x200a220,0x200aa02,0x8a88a000,0x22a2080a,0x2200aa8a,0x280a028a,0x2aa0088a,0x208a88a,0xa22a2a,0xaa00800,0x88a00808,0x2a08a080,0xa882020,0x20a2208,0x280020,0x8208aaa0,0x8882802a,0x80aa8a0a,0x8a080208,0x8a22a08,0x808a80,0xa0aa0080,0xa2a8aa20,0xa08a880,0x2282022a,0x88aa22a8,0x2a828200,0xa2880a20,0xaa022808,0x888020a0,0xa20aa22a,0x282a2020,0xa2808a88,0x82822a8a,0x8000802a,0x8800a880,0x28882a22,0x80882008,0x88a828aa,0x20aa00a0,0x2aa88,0x2002000a,0xa8082228,0x2a020208,0x88808288,0xa8080802,0x80000802,0x202a2a22,0xaa288a88,0xa288802,0x8);
remainders_neg[713].modulo_value=2852; remainders_neg[713].bit_mask=create_bit_mask(89,0xa2a028a,0xa2aa0220,0x202080a,0x8a8088a0,0x828a8a20,0xa0a8aaa0,0x8080020a,0xa8a2a,0x8800880a,0xa2208082,0x2882280a,0x2a0a2aa8,0x8802a822,0x20aaa0a,0xaa288aa0,0x8a888a0,0xa20020a2,0x22a8a022,0x8020a0a,0x228028,0x2008a2a2,0x8a008228,0x28aa02a,0x8020a208,0x20022a28,0x80a00a,0xaa228282,0x22a2028,0xa8822820,0xa2a82888,0x82002a88,0x8a82aa,0x82a220aa,0x8a0aaa2,0x222a80a2,0x8a0808,0x80028088,0x80aa28a,0xa82a8200,0x8a082aa8,0xa88228a,0x2002888,0xaa2a8082,0xa2802202,0xa0828a02,0xa2288002,0x8a002002,0x882a22a2,0x82288088,0x208080,0xa2802808,0x8200a0a8,0x8a822a80,0x88a088a8,0xa802228,0xa0022,0x8220282,0x2808a80,0x802aa200,0x88080200,0x28228808,0x2a2022a0,0x2802008,0x2a0a8aa2,0x22aa20,0xa20a0a88,0x220082a8,0x228aa080,0x8a828,0x2a022aa2,0xa022a8a8,0x2a080200,0xa02a202,0x88808a2,0x802008,0xa00022aa,0x800a880,0x20a022,0xa8220028,0x888a2200,0xa88a8882,0x2080aa20,0xa28a8280,0x80a00,0x20808082,0x888808a,0x28a2a22a,0x22a00200,0x82a0008a);
remainders_pos[714].modulo_value=2856; remainders_pos[714].bit_mask=create_bit_mask(90,0x82882022,0x208020,0x82000282,0xa20a2002,0x28a00008,0x2082208,0x80080,0x820a28,0x202202,0x82080800,0x20000020,0x8800a20a,0xa0020,0x28a08008,0xa00820a,0x80822,0x20020800,0x8820200a,0xa2022000,0x8820208,0x80288282,0xa0020020,0x20000220,0x2080282,0x20080,0x20008200,0x8082200,0xa0002022,0x28828800,0x202000,0x82,0x28208008,0x88000280,0x82002000,0x20800228,0xa00a008,0x2802022,0x20200020,0x80288088,0x2000,0xa28,0xa200,0x22820022,0x800a00,0x80000200,0xa0020000,0x828800,0x8a000088,0xa0000000,0x8000028,0x2280200,0x80822,0x8028008,0xa00a088,0x80020820,0x8008228,0x80002200,0x2082802,0x20,0x8080082,0x22822800,0x8000a00,0x88202088,0x82000800,0x800000,0x80208002,0x80000882,0x800a08,0x82280208,0x80822082,0x8808a20,0x8082200,0x208008a0,0x20200000,0x8200a088,0x20a28a0,0xa00020,0x8a002208,0x8a0,0x20208208,0x88080000,0xa0820080,0x200028,0x88208002,0xa2820,0x8a08a20,0x80008280,0x2080082,0x8228208,0x88);
remainders_neg[714].modulo_value=2856; remainders_neg[714].bit_mask=create_bit_mask(89,0x2002822,0x8a00820,0x208820a,0x22022882,0x20208800,0x82002280,0x80800000,0x8028220,0x828a282,0x2800080,0x28008820,0x2282,0x80820820,0x20200000,0x82080202,0x8000a2,0x20820008,0x28a082,0x22822880,0x20a00,0x820020a,0x208a0820,0x20808a20,0x8200820a,0xa0800,0x28800a08,0x80002288,0x208820a2,0x20020008,0x828a000,0x80880802,0x20200000,0x8208,0x2882880,0x28008220,0x82002000,0x20828a2,0x20208828,0x8200000,0x80802880,0x8800220,0x8082288,0xa20a0822,0x8008200,0x88280,0x200a0880,0x8020008,0x2080000,0x20880080,0x808820,0xa200280,0x808000a2,0x20800,0x82002000,0xa00a0,0x800220,0x8a208,0x2802082,0x808828,0x8000800a,0xa2022080,0x800200,0x28a000,0x2880000,0x8008800,0x28008a,0x80002,0x8000200,0xa200280,0x22002,0x228,0xa280,0xa0080020,0x20208808,0xa002000,0x2822020,0x8208028,0x208a200,0x880020,0x20a00a00,0x8008,0x20020800,0x8a00020,0x28008a,0x80822020,0x200220,0x208,0x82000802,0x220a00);
remainders_pos[715].modulo_value=2860; remainders_pos[715].bit_mask=create_bit_mask(90,0x88a2028a,0x880a0220,0x200a80a0,0x880a082,0x2280a08,0x20222082,0x2288000,0x20a2208,0x82a02008,0x2028000a,0x20822880,0x8228222,0x2a0022a0,0x8a800002,0x82800,0x2a0808,0xa00a8280,0x8800280,0x22,0x82a02a00,0x80882,0xa0802a2,0x28a0280,0xa000880,0x80222,0xa8a20080,0x22080aa2,0x200a0208,0x82a00800,0x8280a200,0x222082,0xa88a2,0x28800a0,0x8008008,0xa0282008,0x8020280,0x8228880,0x8820080,0x80a2a800,0x82880808,0xa20200,0xa0020820,0x22082000,0x2000882,0x8,0x80200,0x82088822,0x800a000,0x8a00082,0x2282008,0xa8a02202,0x8220002,0x28820020,0x8020022,0x8280a028,0x20020200,0x2280200,0x2a0000a0,0x888008a2,0xa0282820,0xa82008,0xa008002,0x2088208,0x8a2a8aa,0x2000020,0xa0888,0xa280022,0x20a0028,0x20008a8,0x80a82822,0xa,0x220880,0x2a00a028,0xa800028,0x82000202,0x2a2002,0x800a8800,0x288200a8,0x88208888,0x2808022,0x80a82a00,0x8a080200,0x22880208,0x88808000,0x2880820,0xa022020a,0x2a008820,0xa0220a0,0x8a22088,0xa28);
remainders_neg[715].modulo_value=2860; remainders_neg[715].bit_mask=create_bit_mask(90,0x2a0202,0x28a20,0x20820028,0x80002802,0x82a00200,0xa0aa280a,0x8a200880,0xa02a200,0xa20a880,0x20a08002,0x280a2008,0x802a02a2,0x2200a228,0x200888a,0x8080a000,0x80a20080,0x20020a00,0x80208,0x8880882a,0x220a200,0x800000a,0x82008a22,0xa028208,0x2008008,0x80808a2a,0x20220800,0xaa000222,0x20828200,0x2208088,0x2082a08,0x88a2280a,0x8020022,0xa008020,0x80000080,0x2020a800,0x808a0208,0x2a0000,0xa0008,0x222088,0x2008000,0x882a0a08,0x200280a0,0x2a80a088,0x8a80000a,0x80888000,0x88000a80,0xa0000a2,0x882080,0x8020000a,0x2a0a800,0x20282a8a,0x80228802,0x200280a8,0x82882a,0x2082820,0xa88a0288,0x8a200a80,0x22888028,0x8022,0x20202028,0x80202880,0x82080082,0xa800200,0x80222022,0x2808828,0x88820000,0x82208022,0x20280a0,0x8a808020,0x202022,0x88880882,0x802a8000,0x220820a0,0x82008820,0x280820a,0x8a2288a,0x8020080,0x20028020,0x200000,0x8208082a,0x202280,0x2000a80,0x22000280,0x888,0x82008020,0x28220a82,0x220800a0,0x28aa028,0x22a000,0x220);
remainders_pos[716].modulo_value=1432; remainders_pos[716].bit_mask=create_bit_mask(45,0x22822aa2,0xa2822008,0x2a22a082,0xaaa880a8,0x8222a00a,0x802a28a,0xa8088a02,0x2a200002,0x280880a0,0xa88a080a,0x8a08a2,0xa200208,0xa22a8a2,0x22028a02,0x8a80a0,0x2202a800,0xa802080a,0xa8882a28,0x22aaaa00,0xa88a82a0,0x8828a820,0x88288200,0x22aa28aa,0x22008228,0x2a082a28,0x880a82a2,0x2a00aaaa,0x2a28a822,0x88a02080,0x2a80,0x880a02a2,0xa080a280,0xa08a2a88,0x208008,0x2a8a20a2,0x28a020a2,0xa80a0220,0x2a800008,0x2080a220,0x82a28a80,0xaaa00a88,0xa82a022a,0x8a820a88,0x88200882,0x8aa882);
remainders_neg[716].modulo_value=1432; remainders_neg[716].bit_mask=create_bit_mask(45,0xaa0aa22a,0x2a0aa880,0xa2aa280a,0x22200820,0xaaa2882,0x80822a02,0x2080028a,0xa2a8888a,0xa0800828,0x20028082,0x8802802a,0x82a88a80,0x82aa202a,0xaa8a028a,0x88020828,0xaa8a2088,0x208a8082,0x2000a2a0,0xaa222288,0x20020a28,0xa020a8,0xa00a88,0xaa22a022,0xaa880aa0,0xa280a2a0,0x820a2a,0xa2882222,0x22a020aa,0x28a808,0x8888a208,0x828a2a,0x28082a08,0x2802a200,0x88a80880,0xa202a82a,0xa028a82a,0x20828aa8,0xa2088880,0xa8082aa8,0xa2a0208,0x22288200,0x20a28aa2,0x20a8200,0xa8800a,0x2200a);
remainders_pos[717].modulo_value=2868; remainders_pos[717].bit_mask=create_bit_mask(90,0x82800002,0x28a28208,0x8a20200a,0x82802800,0x20228228,0x8a082288,0x208a0802,0x820208,0xa008008,0xa0082002,0x20020008,0x220a20a,0x82800800,0x800020,0x2082202,0x820228a2,0x20a08,0x8820a280,0x20022,0xa20220,0x282008,0xa08808a0,0x28820208,0x288088,0x8020a0,0x220228,0x8a200200,0xa2820,0x8828200,0x228a202,0x2022082,0x8200008,0x800a,0x22822822,0x28800200,0x82002080,0x800802,0x20820208,0x8a008288,0x208a2082,0x8a20a20,0x8800a00a,0xa2802022,0x8820a28,0xa00000a,0x220a0000,0x28a28a08,0x2288080,0xa20a00a0,0x20808a08,0xa28208a,0x20822820,0x20020808,0x8280200,0x22802080,0x28800800,0x88288,0x820a0020,0x8020000,0x88082088,0x220808a2,0x828,0x8a20828a,0x80000800,0x20028808,0x8000a080,0x22822022,0x20a20808,0x8000a202,0xa0020082,0x8808,0x82288008,0x28a0,0x8220a08,0x808a288,0x20080882,0x28208000,0x88000200,0x8a08a0,0xa20008,0xa080082,0x20020020,0x20820808,0xa28020a,0x80822882,0x28228828,0x8a200280,0xa28a0080,0x28220828,0x80000);
remainders_neg[717].modulo_value=2868; remainders_neg[717].bit_mask=create_bit_mask(90,0x2080882,0x20220a00,0x228a082,0x2082080,0x28a20a20,0x200a200,0xa0020082,0x8020a00,0x82080080,0x20802882,0x28828800,0x8a282282,0x2080080,0x8008828,0x8a00a28a,0x28a2022,0x8828200,0x282208,0x808a08a2,0x8228a28,0x8820a080,0x20000020,0x20028a00,0x88200000,0x80082820,0x8a28a20,0x2288288,0x808220a0,0x20a08,0x8a20228a,0x828a2802,0xa08800,0x88080082,0xa20a20a2,0x20008a08,0xa08a008,0x80080082,0x28028200,0x2080200,0xa0022802,0x228228,0x82082,0x220828a2,0x28220,0x82088082,0xa2820880,0x20220200,0x8a200008,0x22820820,0x28000200,0x8220a002,0xa00a20a0,0x28828000,0x80208288,0xa2082800,0x20008008,0x88000200,0x28208a0,0x828808,0xa000,0xa2800022,0x8808020,0x2280202,0x880080,0x28820000,0x8082008,0xa20a28a2,0x28228000,0x808228a,0x208a0802,0x8800000,0xa200080,0x80882020,0xa28200,0x80002200,0xa0800002,0x20a00808,0x88288,0x80020020,0x8228800,0x8200800a,0xa08a08a0,0x28028000,0x82208282,0xa2002,0x20a20020,0x2288208,0x22020800,0x20a28020,0x8088);
remainders_pos[718].modulo_value=2872; remainders_pos[718].bit_mask=create_bit_mask(90,0xaa222a8a,0xa02a0282,0xa8822a80,0xaa82888,0xa2088008,0x2a808280,0xaaa008a2,0xa8a82,0xaaaaa0a0,0x80a802a2,0xa2a088a2,0xa2a0a20,0x82082a00,0x22222a28,0x20888a0a,0x8a222820,0x822aa00,0xaa0828,0x82208088,0xa8020002,0xa28a0822,0xaaa22828,0x22a2022,0x8282200,0x228a082,0x82aaa2a8,0xa88a8a22,0xa828a00a,0x8aa00228,0x2a282220,0x20a0888a,0x88202222,0xaaa028a2,0xa00a020,0x82088a02,0x202a80a,0xaa0a0000,0x28080aa,0x2208aa00,0xa828a80,0xa8aa88a2,0x8882800,0xa802288,0x282a02a,0x802220,0x80088802,0xa0a80a82,0x20228802,0x2a002822,0xa08a22aa,0x8802a282,0x8000aa20,0xaaaa0202,0x800000a0,0x82a02a80,0xa080a220,0xaa080a00,0x228a280a,0x8888808,0xa8a2220a,0xa2088828,0x2a28800a,0x2aa00a28,0x8288a2a2,0x802a8aaa,0x20820a28,0x80008828,0xaa8808a8,0x8a28288a,0x2a8820a2,0x82800080,0x220208,0x202820aa,0xa200aa88,0x8082888,0x88a0a222,0x8228a888,0xa000a820,0x8a08a0a8,0x28a220a,0xaa8a802a,0xa0aaa,0xaa82a2a0,0xa88a200a,0x8a028202,0xa0200220,0x2a22282a,0xa02a882,0xaa8280a8,0xa2a888);
remainders_neg[718].modulo_value=2872; remainders_neg[718].bit_mask=create_bit_mask(90,0x22aaa202,0x28a28a0a,0x200aa208,0x8220a000,0x2a800880,0xa2080a08,0x2228802a,0x8882020a,0x22222828,0x8208a2a,0x2a28002a,0x82a28228,0xa80a288,0xaaaaa2a0,0xa8000282,0x2aaa0a8,0x80aa2288,0x882280a0,0xaa80800,0x208a888a,0x2a0280aa,0x222aa0a0,0x8aa2a8aa,0x80a0aa88,0x8aa0280a,0xa222a20,0x200202aa,0x20a02882,0x2288aa0,0xa2a0aaa8,0xa8280002,0xa8aaaa,0x2228a02a,0x828828a8,0xa80028a,0x8a8a2082,0x22828888,0x8a080822,0xaa802288,0x820a0208,0x2022002a,0x8000a088,0x8208aa00,0x8a0a28a2,0x8808aaa8,0x880008a,0x2820820a,0xa8aa008a,0xa288a0aa,0x2802aa22,0x8a2a0a,0x88822a8,0x22228a8a,0x8888828,0xa28a208,0x28082aa8,0x22808280,0xaa02a082,0x80000080,0x202aaa82,0x2a8000a0,0xa2a00882,0xa22882a0,0xa002a2a,0x8a20222,0xa80a82a0,0x88800a0,0x22008020,0x2a0a002,0xa200a82a,0xa088808,0x88aa8a80,0xa8a0a822,0x2a882200,0x8080a000,0x282aaa,0xaa02000,0x288820a8,0x2802820,0x8a02aa82,0x220208a2,0x88828222,0x220a2a28,0x2002a882,0x28a0a8a,0x28a88aa8,0xa2aaa0a2,0x828a200a,0x220a0820,0x2a2000);
remainders_pos[719].modulo_value=2876; remainders_pos[719].bit_mask=create_bit_mask(90,0x22a82a22,0x208aa820,0xa002808a,0xaaaa0208,0x80228a08,0x222a0a00,0x8a828080,0xaaa28882,0x822a8822,0xa02a882a,0x800a0a2,0x82280288,0x20a2a828,0x2a28088,0x8a8aaa20,0x80a0828a,0x2802aaa0,0xa82a80a0,0xa888822a,0xaa028aaa,0x8888a080,0x8a88aa82,0x22220888,0xa002202,0x2a282222,0x20a80,0x8a2220,0xa82a00a0,0x28000aa0,0x20a282a,0xa8800202,0xa22a080a,0x20a00828,0xa20aa08a,0x8282aa2,0xa200a8,0x8a2008a,0xa220800,0xaa2a0a02,0x2a828088,0x228208aa,0x228a8000,0x822a0aa8,0xa8a00228,0x880a008,0xa200a022,0x228800a2,0x82aa0a88,0x80002a28,0x8aa20828,0xa220282a,0x80a0a8a,0x2088a,0xa2008a2,0x2a008a0,0xa8aa8282,0x8a20aa08,0xa28200a0,0xaa020a88,0x8080022,0x8a828a08,0xa0aa0002,0xa00a82,0x888a20,0x802a0800,0x888828a,0x8808800a,0x22220888,0x282aa22a,0xa020a222,0xaaaa280a,0xaa882222,0x80a02a82,0x20aaa802,0x2a2820a0,0x8aaa2a,0x822028a8,0x2282a8a0,0x2280288,0xa8a0a002,0x2a822a80,0xa8822a88,0x282228aa,0x8020282a,0x200a0a88,0xa20a2880,0xa2080aaa,0x8a202800,0x8082aa20,0x88a82a8);
remainders_neg[719].modulo_value=2876; remainders_neg[719].bit_mask=create_bit_mask(90,0xaa20a2aa,0xa80220a8,0x288a0802,0x22228a80,0x8aa0280,0xaaa28288,0x20a0808,0x222a000a,0xaa200aa,0x28a200a2,0x8088282a,0xaa08a00,0xa82a20a0,0x8a2a0800,0x20222a8,0x8280a02,0xa08a2228,0x20a20828,0x20000aa2,0x228a0222,0x2808,0x200220a,0xaaaa0000,0x8288aa8a,0xa2a0aaaa,0x888a8208,0x8802aaa8,0x20a28828,0xa0888228,0x8a82a0a2,0x20088a8a,0x2aa28082,0xa82880a0,0x2a822802,0x80a0a22a,0x882a8820,0x802a8802,0x82aa8088,0x22a2828a,0xa20a0800,0xaa0a8022,0xaa020888,0xaa28220,0x20288aa0,0x80082880,0x2a8828aa,0xaa00882a,0xa228200,0x888a2a0,0x22a80a0,0x2aa8a0a2,0x80828202,0x888a8002,0x82a8802a,0x8a288028,0x20220a0a,0x2a82280,0x2a0a8828,0x228a8200,0x808088aa,0x20a0280,0x2822888a,0x8828820a,0x880002a8,0x8a28088,0x80000a02,0x800882,0xaaaa8000,0xa0a22aa2,0x28a82aaa,0x2222a082,0x2200aaaa,0x828a20a,0xa822208a,0xa2a0a828,0x880222a2,0xaa8a020,0xaa0a2028,0x8aa08a00,0x2028288a,0xa20aa208,0x200aa200,0xa0aaa022,0x8a8a0a2,0xa8828200,0x2a82a008,0x2a808222,0x2a8a088,0x80a22a8,0x20a20);
remainders_pos[720].modulo_value=120; remainders_pos[720].bit_mask=create_bit_mask(4,0xa0080802,0x28020200,0xa008080,0x802020);
remainders_neg[720].modulo_value=120; remainders_neg[720].bit_mask=create_bit_mask(4,0x20800082,0x20028a00,0x2080008,0x28a0);
remainders_pos[721].modulo_value=2884; remainders_pos[721].bit_mask=create_bit_mask(91,0xaa80822a,0xa000,0x22828a20,0x22080820,0x2822a280,0x8282208,0x28028080,0x88aa80a,0x8808822,0x20088022,0x2a02000a,0x80a02a82,0x8a22800,0xa08aa0a0,0x82820008,0xa000a02,0xa2a20280,0x880a8008,0x22220082,0xaa2020,0x88aa0202,0xa20a220a,0x2aa8aaa2,0x8222822a,0x202a88a,0x2022a802,0x8022220,0x800a808a,0xa022a28,0x2800280,0x80020a0a,0x282a8828,0xa22880,0xaa02808,0x800202a2,0x20088022,0x20880882,0x80aa8882,0x80a00a2,0x8220a080,0xa2a20a0,0x20808220,0x228a0a20,0x280000,0xa2080aa8,0xaa0208aa,0x28002,0x8a0a2880,0x88202080,0xa08a8a00,0x20a08820,0xa00a0200,0x222aa028,0x22022088,0x80220088,0xa8080028,0x280aa08,0x2288a002,0x822a8280,0xa080022,0x2800280a,0x8a880a00,0x202a0022,0x8888020a,0x2a88080,0x22a80808,0x8828882a,0xaaa2aa8a,0x88a08a8,0x80aa22a,0x808aa008,0x20088880,0x2a0228,0x2808a8a2,0xa000a00,0x82828,0xa0aa20a2,0x288a200,0x2a80a020,0x80a88,0x8022008a,0x82202208,0x2aa2208,0x2028028a,0x8828200,0x28a88282,0x82020880,0x8a282880,0xa00000,0x88202aa0,0xa);
remainders_neg[721].modulo_value=2884; remainders_neg[721].bit_mask=create_bit_mask(91,0x22080a22,0x8882880,0xa20a02a8,0xaa008028,0xa0a22a08,0x80a02a80,0xa08a0008,0x80022002,0x800a2,0xa08008aa,0xa20a8882,0x820a20a,0x802a2088,0x28022028,0xa0a8800,0x2888282,0x222a8a00,0x20880,0xaaa2880a,0x882228a8,0x22828a,0x2a82aa02,0x22202222,0x2aa0aa2,0x8a0a2002,0xa8a2208a,0x808a2aa8,0x8820002,0x28aa220,0xa088a00,0x8a8282,0xa02200a0,0x8822a008,0x82282080,0x88a822a,0xa880082a,0x28008002,0x22000a,0x8002882a,0xaa02808,0x82a22828,0xa00802a8,0xaa028228,0x8a08880,0x22808220,0x220a8022,0x8882088a,0x2822008,0xa8a008,0x28020208,0x282800a0,0x20828a88,0xaa2228a0,0xaa82a800,0x82a0800,0x208080a0,0x8a082200,0x2a002882,0x2a20a08,0x820088aa,0xa080a082,0x2000288,0xa8a280aa,0x8a02,0xa200800,0xa2208080,0x2000a2,0x22222202,0x80020020,0x808222a2,0x8022800,0x28800000,0x80a28aa0,0xa000202a,0x82808288,0x888020a0,0x2822a02a,0x8a002a08,0x220828a0,0x80808200,0x82a0802,0xaa0aa80,0x8a222a80,0xa8a08202,0x800a0a08,0x20200a02,0x28a8008,0x220a008,0x88208888,0xa82228,0x2);
remainders_pos[722].modulo_value=152; remainders_pos[722].bit_mask=create_bit_mask(5,0x82828282,0x80828282,0x2828282,0x82828282,0x828282);
remainders_neg[722].modulo_value=152; remainders_neg[722].bit_mask=create_bit_mask(5,0xa020a0a,0x80a0a0a,0xa0a0a0a,0xa0a0a0a,0xa0a0a);
remainders_pos[723].modulo_value=2892; remainders_pos[723].bit_mask=create_bit_mask(91,0x820a0082,0x28828800,0x2080002,0x828808a2,0x228a20,0x8220a208,0x20020a2,0x28008020,0x8202200,0xa2082822,0x28828208,0xa00a28a,0x20800822,0x820a20,0x8808a28a,0x80802,0x20228200,0x828000a,0x20220a0,0x828800,0x808a000,0x200208a2,0x28028008,0x208a082,0xa0022080,0x20020a28,0x808828a,0xa28020a0,0x8008800,0x8220a28a,0x22002,0x820a00,0x82282088,0xa0080,0x8020000,0x80208002,0x80882080,0x28028028,0x8288288,0x80080800,0x8a20,0x2200208,0xa2820880,0x20020008,0xa00820a,0x200a0802,0x8008a08,0x828a200,0x880022,0x20000208,0x200208a,0x22820002,0x28028228,0x82202280,0x20802020,0x8020800,0xa000000,0x82882020,0x8200220,0x8800000a,0xa0882800,0x20020a28,0x8028a002,0x220200a0,0x8008a28,0x8800a28a,0xa2080020,0x28028820,0x8008200,0xa20208a2,0x28200000,0x88080002,0x28200a0,0x28808a00,0x2000008,0xa2022802,0x8200a28,0x20808a,0xa00a0882,0x28228a28,0x8208a208,0x80820882,0x20028008,0x80080080,0xa08828a0,0x28800208,0x228208a,0x20808a2,0x28228800,0xa082002,0x820);
remainders_neg[723].modulo_value=2892; remainders_neg[723].bit_mask=create_bit_mask(91,0x2820802,0x20020008,0x8a00808a,0x2000022,0x8a20228,0xa282280,0x82882822,0x20800828,0x8028a288,0x228020a2,0x20020a00,0x82082202,0xa00800a2,0x8028228,0x2202,0x80800082,0x28a20a08,0x80208082,0x828a2820,0x8020008,0x80002088,0xa08a0022,0x20820800,0x8a00200a,0x208a2800,0x28828220,0x80000202,0x22082820,0x800008,0xa282202,0x808a2882,0x8028208,0xa20a000,0x80820800,0x828808,0x828008a,0x2800,0x20820820,0x80200200,0x800080,0x8800228,0x8a288280,0x220a0000,0x28828800,0x82080282,0xa0820082,0x800200,0x80202288,0x800008a2,0x28808a00,0x8a08a002,0xa20a0882,0x20020a20,0xa28a208,0xa00828a0,0x828008,0x82088088,0x20028a0,0xa08a28,0x88082,0x20002080,0x28828220,0x820208a,0xa28a0820,0x800220,0x82202,0x228008a0,0x20820028,0x80080288,0x228a0022,0x20a08808,0x808a,0x820a0820,0x20000208,0x8a088080,0x228a2082,0xa08220,0x88280002,0x20820002,0x20a20220,0xa002280,0xa0002,0x20820800,0x8008008,0x20002020,0x20008a00,0x8a20a002,0x82800022,0x20a20008,0x8200a08a,0xa0);
remainders_pos[724].modulo_value=1448; remainders_pos[724].bit_mask=create_bit_mask(46,0x2a00aa2a,0x88228a2,0x828a0a,0x22a08028,0xa8aa2,0x20202a2,0x8a80022a,0x82a222a,0x8a0800a0,0xa2088282,0xa802a228,0x2a8aaa2,0x8a288a8,0xa282822,0x8200a002,0x2a2a888a,0x80a8800,0x8a808,0xa888aa2a,0x20028020,0x220a0a28,0xa88a288,0xa2aa8aa0,0x8a22a00a,0xa0a08822,0x2800828,0xaa222a08,0x2a2000a8,0x22a02020,0x22a8a800,0xa0082a2,0x2828a080,0x228a2088,0xaa2a802a,0x8a802a8a,0x82208a28,0x20a282,0x88a8200a,0x8002a2a8,0x808080a8,0xa2a0008a,0x20a888a,0xa2820028,0x288220a0,0xaa00a88a,0xa8);
remainders_neg[724].modulo_value=1448; remainders_neg[724].bit_mask=create_bit_mask(46,0xa28822a2,0x800aa02a,0x880a0282,0xaa2808a0,0x8882022a,0x8a8a8a2a,0x2088aa2,0x80a2aaa2,0x2808828,0x2a800a0a,0x208a2aa0,0x8a20222a,0x802a0020,0x82a0a0aa,0xa88288a,0xa2a20002,0x820088,0x88802080,0x200022a2,0xa88a08a8,0xaa8282a0,0x82002a00,0x2a220228,0x2aa2882,0x282800aa,0x8a0880a0,0x22aaa280,0xa2a88820,0xaa28a8a8,0xaa202088,0x82880a2a,0xa0a02808,0xaa02a800,0x22a208a2,0x208a202,0xaa802a0,0x88a82a0a,0x20a882,0x88a2a20,0x8000820,0x2a288802,0x8a820002,0x2a0a88a0,0xa00aa828,0x22882002,0x20);
remainders_pos[725].modulo_value=145; remainders_pos[725].bit_mask=create_bit_mask(5,0x11c122d2,0xca7a0456,0x794e4489,0xe21a881,0x12d12);
remainders_neg[725].modulo_value=580; remainders_neg[725].bit_mask=create_bit_mask(18,0x88082a0a,0x2228882,0x20028000,0x82a02008,0x8028a00a,0x80202a0a,0xa0020,0x88a2200,0x82a0808a,0x2880202,0xa0200880,0x28208aa2,0x200a2288,0x8a002880,0xa2280200,0xa8820a08,0x880202a2,0x2008a000);
remainders_pos[726].modulo_value=264; remainders_pos[726].bit_mask=create_bit_mask(9,0x22880022,0x228800,0x88000288,0x20880022,0x220800,0x80002288,0x22880022,0x228800,0x88);
remainders_neg[726].modulo_value=264; remainders_neg[726].bit_mask=create_bit_mask(8,0xa20000a2,0x8220008,0x88200,0xa00008a2,0x8a20008,0x8a200,0xa20008a2,0x8a20008);
remainders_pos[727].modulo_value=2908; remainders_pos[727].bit_mask=create_bit_mask(91,0x8a0a2a0a,0x2a2a008a,0x200a8a00,0x22a2a2a,0xa28a8000,0x8aa22a02,0xa88228a0,0xa8aaa02,0x8a88a00a,0x888a22a,0xaa2882aa,0x88000a8a,0x2082002,0x2a8a2200,0x28a0a8a0,0x88a0088,0x880a808a,0x22022820,0x80a8a082,0x280a8082,0xaa8aa220,0x88a200a,0x20aa082,0x82228a28,0x20002a8,0x2a0a888,0x202a0a2a,0x8a888a28,0x2a228a0,0x8222822a,0x2820a22a,0x8200a820,0xa28aaa88,0xaaa20aa8,0x20800202,0x2222800a,0x22020088,0x20282a8,0x8a200a80,0x88022820,0x2080a80,0x808aaaaa,0x2a88080,0x8080aa82,0x8282022a,0x20280280,0x2a880828,0x82aa020,0x202022a8,0xaaaaa20,0x202a0208,0x20828802,0x2a008a,0x82a82808,0x82200808,0x8a002888,0xa8080020,0xa2aa08aa,0x222aaa28,0x8082a008,0x2a88a082,0xa882888,0x20a288a8,0x828a222a,0xa8a0a80,0x222a0a8,0x22a80008,0x28a2888,0x820aa08,0xaa008a22,0x8088aa2a,0x28202a02,0x8820a2a0,0x20828808,0xa202a02,0x82200a22,0x80a2a0a2,0x88a2a,0x28008208,0xaa2a0002,0xaa8228a,0x2a88a222,0xa00a22a,0xa80aaa2a,0x20a28822,0xa80a88aa,0x2a28,0x8a8a8a88,0x800a2a00,0x2a200a8a,0xa0a8a0a);
remainders_neg[727].modulo_value=2908; remainders_neg[727].bit_mask=create_bit_mask(91,0x282a282,0xa2a28802,0xa8820288,0x8aa2a2a2,0x2a020888,0x22aa28a,0x200aa028,0x8202228a,0x2002882,0x80002aa2,0x22a00a22,0x888202,0x8a80a88a,0xa202aa88,0xa0282028,0x80028800,0x820802,0xaa8aa0a8,0x820280a,0xa082080a,0x22022aa8,0x8002a882,0x8a02280a,0xaaa02a0,0x8a888a20,0x8a282000,0xa8a282a2,0x20002a0,0x8a2aa028,0xaaa0aa2,0xa0a82aa2,0xa8820a8,0x2a022200,0x222a8220,0xa8088a8a,0xaaaa0882,0xaa8a8800,0x8a8a0a20,0x2a88208,0x8aa0a8,0x8a808208,0x8022222,0x8a200808,0x808220a,0xa0a8aa2,0xa8a08a08,0xa20080a0,0x80a228a8,0xa8a8aa20,0x822222a8,0xa8a28a80,0xa80a008a,0x88a28802,0xa20a080,0xaa88080,0x288a000,0x208088a8,0x2a228022,0xaaa222a0,0x80a2880,0xa200280a,0x8200a000,0xa82a0020,0xa02aaa2,0x82028208,0x8aaa2820,0xaa208880,0x8a02a000,0x80a82280,0x228802aa,0x80022a2,0xa0a8a28a,0xa82a28,0xa80a0080,0x82a8a28a,0xaa882aa,0x82a282a,0x888002a2,0xa0880a80,0x22a2888a,0x8220aa02,0xa2002aaa,0x82882aa2,0x208222a2,0xa82a00aa,0x20820022,0x8888a2a0,0x2020200,0x882a288,0xa2a88202,0x2820282);
remainders_pos[728].modulo_value=728; remainders_pos[728].bit_mask=create_bit_mask(23,0x2088202,0x28880a22,0x22aa02a0,0x20a28a2,0xa8a22080,0x8282800,0x802a0,0x2002aa0a,0x20002200,0xa0880202,0x822a000,0x20828282,0xa000a88,0x8808022,0x8008800,0xa0aa80,0x200a8020,0x2a002828,0x8002088a,0x888a28a0,0x280a80aa,0x8088a022,0x808220);
remainders_neg[728].modulo_value=728; remainders_neg[728].bit_mask=create_bit_mask(23,0x8a800a0a,0x20008222,0xa2228a28,0x8a02a02a,0x20222808,0x80a02088,0x88808220,0xa80a2202,0x2888aa80,0x2000828a,0x2a2888,0xa8020a0a,0x82800200,0x800800aa,0x80880008,0x282200,0xa08208a8,0xa20820a0,0x8828002,0x22020,0xa0020022,0x800282a,0x802a0);
remainders_pos[729].modulo_value=12; remainders_pos[729].bit_mask=create_bit_mask(1,0x8a2);
remainders_neg[729].modulo_value=12; remainders_neg[729].bit_mask=create_bit_mask(1,0x22);
remainders_pos[730].modulo_value=2920; remainders_pos[730].bit_mask=create_bit_mask(92,0x28220a8a,0x880a8222,0x282a088,0xa8280a8,0x2082808,0x280a2280,0x200a02,0x8822a8,0x880008a8,0xa2a08a22,0x88a2a00,0x8220aa0,0x208a2200,0x2a200a2,0x80200808,0x8222800,0x2a0a08a0,0x882a208,0x800280a,0x80208202,0x2282,0x2a2a8200,0x220202a8,0x2000a2,0xa2a8a200,0x28820202,0xa8228282,0x880a028,0x8020a002,0x20880820,0x288a2802,0x22888a2,0xa0202a0,0x8a2a020,0x80a80022,0xa2000,0xa8020082,0xa8800a8,0x8a00028,0xa020a02a,0xa0aa0200,0x80288820,0x2880280,0x88808002,0x20202200,0x82,0x88080882,0x2022200,0x80228080,0x22280202,0x80aa0a08,0xa080a00,0xa20a8,0x22a028,0x802a2a,0x8a00082,0x2a0200,0xa8a2088,0x8080a008,0x2228800a,0x28a2288a,0x20220880,0xa080208,0xa022080,0x82882a28,0x80822882,0x8a2a8a80,0x80000,0x8080888a,0x82a8a82a,0x88000000,0x82080282,0x28002080,0x8a8220a0,0x20a0a820,0x2888200a,0x20080200,0x8a8020,0x88a2088a,0xa0882000,0xa8a2200a,0xa20a8a00,0x20002288,0x8822002a,0xa008002a,0x88a02880,0x28208002,0x282a020,0xa82802a,0x82a02222,0xa0882888,0xa2);
remainders_neg[730].modulo_value=2920; remainders_neg[730].bit_mask=create_bit_mask(92,0xa0aa0202,0x20aa2,0xa0a2000,0x82020820,0x8280a000,0xa0822a08,0x80288282,0x800a220,0x808020,0x2228022a,0x80022288,0x802a8220,0x2802a288,0x8a22882a,0xa88000,0xaa2088,0xa2028020,0xa2280,0x8080a082,0xa80a0a,0x88882a0a,0xa2220a80,0x2a8a8220,0x88a0882a,0x22202a08,0xa00a0a82,0x202a0a02,0x820a0,0x8a0288a,0xa0008028,0xa002208a,0x8a200022,0x28a8228,0x802228a8,0x20882a,0x88822888,0x200a8802,0x2008020,0x802088a0,0x20a82822,0x28220a88,0x82000a0,0xa008208,0x8a,0xa0a8aa08,0x8888080a,0x8002,0xa8aa288,0x8a20808,0xa2a08a0a,0x8220280,0x82008280,0x8822020,0x88a228a0,0x8008a222,0x8028080a,0x88228a80,0x202a000,0x8002880,0xa2a00802,0xa02a2002,0xa8228000,0x2808280,0x8282a808,0x200a220,0x80a200a,0x2220200,0x8808088,0x8000002,0x2202022,0x80888,0xa008a02,0x2088a008,0x202a828,0xa0282028,0xa0002882,0xa8008a80,0x80200a8,0x228002,0x2000a808,0x202a2882,0x2a020280,0x2888a200,0xa288a2,0x20808822,0x282008,0xa0280882,0xa0a20a8,0x20208a2,0x228aa2a,0x28002000,0x22);
remainders_pos[731].modulo_value=2924; remainders_pos[731].bit_mask=create_bit_mask(92,0xa2a82a22,0x20a22020,0x80028202,0x2a2a0a28,0x882a800,0xa20a0a80,0x80228200,0x22a202a2,0x8aaa0aa0,0x82a20828,0x202aa0,0xaa00a202,0x28a88a00,0x82a08000,0xa8a8a002,0x22028a8a,0x88822a80,0x822020a8,0x2220aaa8,0x228288a2,0x8822288a,0xa0820aa2,0x2a2aa82,0x80a0008,0x2208028a,0x8802088,0x88882208,0x8a200020,0x22202808,0x88aa000a,0x2000a,0xa0aa020,0x20aa8a28,0x80220220,0xaa88882a,0xa2000a8,0x200a088,0x8280280,0x28080a88,0x802a8a08,0x2aa0282,0x8aa00a,0xaa208280,0x288a8a08,0x828a080,0x22088080,0xa2820020,0x2a228020,0x208aa20a,0xaa200028,0xaa80a00,0x2a802828,0x202820a,0x282022a,0xa0080028,0x8a0220,0x222aa2a0,0x8802a82,0x2aa08088,0xaa0a028a,0x8000080,0xaa22a00,0x80888a00,0x8a2202,0x82222080,0x80220208,0x2088220,0xa020a28,0xa8a80200,0x820a82a,0x888228aa,0x28288a22,0xa08888a2,0x808822aa,0x882222a0,0x2808802a,0xa2a2aa2a,0x20a82800,0x22a28000,0xa00aa00a,0x80800808,0x8a820aa,0xaaa2282,0x8a880aa,0x288028a8,0xa08a008,0xa822002a,0xa8a8002,0x2800228a,0x88a08808,0x82a8a080,0x88a);
remainders_neg[731].modulo_value=2924; remainders_neg[731].bit_mask=create_bit_mask(92,0x2a20a2aa,0xa822a0a8,0x88a0a8a,0xa22282a0,0x800a2088,0x22828208,0x82a0a88,0x2a2a8a2a,0x2228228,0xa2a80a0,0x88a8a220,0x22882a8a,0xa0200200,0xa280888,0x2020208a,0xaa8a0202,0xa2208,0xaa82820,0xaaa02220,0xaa0a002a,0x2aa002,0x280a822a,0x822a220a,0x80828880,0x2a808a02,0x8008a800,0xaa80,0x2a888a0,0xaaa8a000,0x228802,0x888a8882,0x828220a8,0xa82202a0,0x8a20aa8,0x220000a2,0x82a08820,0x8a882800,0x80208a08,0x20808200,0xa20280,0x8a228a0a,0x8022882,0x22a80a08,0xa0020280,0x80a02000,0xaa800808,0x2a0a8828,0xa2aa08a8,0xa8022282,0x222888a0,0x82200288,0xa208a0a0,0x8a820a82,0x8a0a8aa2,0x280088a0,0x88028aa0,0xa2a22a28,0x8008a20a,0x22280800,0x22828a02,0x80880808,0x822aa280,0x8000288,0x8802aa0a,0xaaaa808,0xaa8280,0x8a800aa8,0x828a02a0,0x20208a88,0x80a020a2,0xaa022,0xa020022a,0x2800002a,0xaa22,0xaaaa28,0x208008a2,0x2a2222a2,0xa820a088,0xaa2a0880,0x28822882,0x8088000,0x20a822,0x8222a20a,0x80200822,0xa0082020,0x82802880,0x20a288a2,0x8202008a,0xa008aa02,0x280080,0x2202808,0x2);
remainders_pos[732].modulo_value=1464; remainders_pos[732].bit_mask=create_bit_mask(46,0xa2022082,0xa28800,0x8a088208,0x2882822,0x800808,0x208a280,0xa20a2800,0x20028220,0x2000080,0x82082,0x8800a20,0x2080208,0x8a0022,0x8208200,0x80200008,0xa0882800,0x28a08,0x28a208,0x2020020,0x28828228,0x208220a,0xa00228a0,0x28208808,0xa202208,0x800a2880,0x28a08820,0x80288282,0x80080,0x208a28,0xa20a280,0x2002822,0x20200008,0x8208,0x808800a2,0x20208020,0x8a002,0x80820820,0x8020000,0x8a088280,0x800028a0,0x28a20,0x80202002,0xa2882822,0x208220,0x8a00228a,0x820880);
remainders_neg[732].modulo_value=1464; remainders_neg[732].bit_mask=create_bit_mask(46,0x228a2802,0x8220008,0x2000280,0x820020a2,0x8008000,0x8a002208,0x22822080,0x28820a28,0x8a088008,0x80802802,0x8228,0x8a008280,0x800208a2,0xa00208,0x8288080,0x20002080,0x8820200,0x88202280,0x828a08a0,0x20020a20,0xa00a282,0x208a2020,0x20a00000,0x8228a280,0x822000,0x20200028,0x820020a,0x80800800,0x8a00220,0x82282208,0x828820a2,0x28a08800,0x88080280,0x822,0x28a00828,0x8800208a,0xa0020,0x828808,0x2000208,0x882020,0x8820228,0x828a08a,0x220020a2,0xa00a28,0x208a202,0xa0000);
remainders_pos[733].modulo_value=733; remainders_pos[733].bit_mask=create_bit_mask(23,0xce4b561a,0x2fabc738,0x2f4afa17,0xdb089312,0xa21bfc40,0xd1aa6fbb,0x9cd85c79,0x261af5b7,0x874dd5,0xbb597c1a,0x4b00b818,0xa6761b99,0x46074034,0x160fa6b7,0x2aecb840,0x7b6bd619,0xe78e86ce,0x777d9562,0xc08ff611,0x12324436,0x3a17d4bd,0xc738f57d,0x161ab49c);
remainders_neg[733].modulo_value=2932; remainders_neg[733].bit_mask=create_bit_mask(92,0x2828a82,0xa2220aa0,0xa282228a,0x2800a8a,0x2a822088,0x822a222,0x80a0,0xaa82282a,0x880a8008,0x2280a082,0x82882080,0x2aaa8200,0x8a8a88a8,0x8a822a2a,0xa2202088,0xa2a20a80,0x2a020a02,0xaaa008aa,0x8022a88,0x8aaa88aa,0xa28a0820,0xaa028a0,0x8a822800,0x8000a20a,0x2828a02a,0x8820828a,0x82a88aaa,0xa88aa200,0xa2a8802a,0x2a028202,0x280a822a,0x88882022,0x8aa2a20a,0xa0a88a8a,0x8020aaa,0x2008208a,0x8a08280a,0xa8800a80,0x2a0a20a,0x80280800,0xa2222a20,0x88220a,0x2a8a800a,0x2a8a220a,0x2a8222,0xa20a8a0a,0x2aa0a080,0x80222280,0x8020a220,0x8aa0aa80,0x2200a228,0x2a8a2022,0x2aaa8a8,0x8800a20a,0xa08a80aa,0xaa20a828,0xa8a08a28,0xa0000a2,0x808088,0x8880a202,0xa8222a28,0xa0202280,0x202a282,0x88002a88,0x28aa200,0x28800088,0x2820828a,0xaa802a08,0x8000a20a,0xaaa822,0x820a082a,0x8a28a0,0xa8a00880,0x88022,0xa02008aa,0x202a0a2,0x220a80a2,0x80888a2a,0x80802022,0x280880,0x828aa280,0x822a8a28,0xa8828a0a,0x8008aa80,0xaaa02822,0x28aa0a8a,0x80222202,0x82a88a22,0x820080aa,0x22008222,0x82aa00a2,0x28082);
remainders_pos[734].modulo_value=2936; remainders_pos[734].bit_mask=create_bit_mask(92,0x22200222,0xa0a2282,0x2a0aa2a8,0xa282aa80,0xa002a222,0x288002a2,0xa20a2028,0xa8a0800a,0x2a0020a8,0x2280aaaa,0xaa208208,0x2a2288,0x80a80aa2,0xa222082a,0x20a88a2,0x20880a02,0x828aa0aa,0x2a288802,0x2aaa288a,0x8822282a,0x822a20a2,0xa82aa0a0,0x2aa88aaa,0x8800,0x8a0a0028,0x820a2022,0x20282228,0x20882000,0x82a88820,0x200a0082,0xa2a0a88a,0x20880a2,0x8028a222,0xa200a80a,0x882002a,0x28a28a20,0x20000a82,0x80a2aa0,0xaa8a08,0x282a20a2,0x202aa88,0x22202aa,0x2a800282,0xa80200a0,0x228220a0,0x8222aa22,0x8888aa88,0x2a0a0882,0xa80a0080,0x80828002,0x80aa8088,0x2822aa80,0x8a08a8,0x2020a2aa,0x80aa8a0,0x2882a000,0x2008a28a,0x8aa80082,0x2a02a00,0x80888a28,0x8a8a0220,0x8a22a0a,0x828200a0,0x808222a,0x8000822,0x82288828,0xa28808a0,0x2800a0,0xa8002200,0x2aaaa22a,0x820a0aa8,0x228a08a8,0xa8a82888,0xa8a228aa,0x82802228,0x8aa0aa2,0x8080a022,0x8a8a22a0,0x2a82088,0x8aa02a,0xaa2288a8,0x88208208,0xa8aaaa02,0x2a2a0800,0x8aa0020a,0x282808a0,0xa8a8002,0x8a888a80,0xa802aa82,0xa02a8aa0,0x888288a0,0x888008);
remainders_neg[734].modulo_value=2936; remainders_neg[734].bit_mask=create_bit_mask(92,0xaaa88aaa,0x8282aa0a,0xa2822a20,0x2a0a2208,0x288a2aaa,0xa0088a2a,0x2a82a8a0,0x20280882,0xa288a820,0xaa082222,0x22a80a80,0x88a22a00,0x820822a,0x2aaa80a2,0x8a82002a,0xa800828a,0xa022822,0xa2a0008a,0xa222a002,0xaaa0a2,0xaa2a82a,0x20a22828,0xa2200222,0x88880088,0x28288a0,0xa82a8aa,0xa8a0aaa0,0xa800a888,0xa2000a8,0xa882880a,0x2a282002,0x8a80082a,0x8a02aaa,0x2a882082,0x800a88a2,0xa02a02a8,0xa888820a,0x8082a228,0x88220280,0xa0a2a82a,0x8a8a2200,0x8aaa8a22,0xa2088a0a,0x208a8828,0xaa0aa828,0xaaa22aa,0x2200,0xa282800a,0x20828808,0x80a088a,0x8220800,0xa0aa2208,0x88028020,0xa8a82a22,0x80822028,0xa00a2888,0xa8802a02,0x220800a,0x8a28a288,0x80002a0,0x2028aa8,0x802aa282,0xa0a8828,0x8080aaa2,0x808880aa,0xaa000a0,0x2a008028,0x88a08828,0x2088aa88,0xa2222aa2,0xa828220,0xaa028020,0x2020a000,0x202aa022,0xa08aaa0,0x8022822a,0x80828aa,0x202aa28,0x8a20a800,0x880228a2,0x22aa0020,0xa80a80,0x2022228a,0xa2a28088,0x2288a82,0xa0a08028,0x8202088a,0x2000208,0x208a220a,0x28a20228,0xa0028,0x880);
remainders_pos[735].modulo_value=420; remainders_pos[735].bit_mask=create_bit_mask(14,0x20802,0x28200800,0x88,0x82022080,0x208,0x88282208,0x80000020,0x8828220,0x8000002,0x20880822,0x20000000,0x2008282,0x2080000,0x8);
remainders_neg[735].modulo_value=420; remainders_neg[735].bit_mask=create_bit_mask(13,0x808a0002,0x20208000,0x88000,0x2022800,0x808a00,0x202280,0x800a0,0x20228,0x8082,0xa0000022,0x28000808,0x8a000202,0x2800080);
remainders_pos[736].modulo_value=184; remainders_pos[736].bit_mask=create_bit_mask(6,0xa2082aa,0xa8222228,0x28a00,0x2a00a280,0xa0288888,0xaa8208);
remainders_neg[736].modulo_value=184; remainders_neg[736].bit_mask=create_bit_mask(6,0x82280a22,0x20aaaaa0,0x888a0288,0xa2802a08,0x28a00000,0x220a80);
remainders_pos[737].modulo_value=2948; remainders_pos[737].bit_mask=create_bit_mask(93,0x280a282,0xaa0a8a20,0xaaa28080,0x2022a0,0xa8020880,0x20820a,0x880080a8,0x2008202,0x8200a20a,0x2000280a,0xa8a222a,0x20a22220,0xaa00a228,0x820008a8,0x80822,0x20a22022,0x22aa8082,0xa0a88aa8,0xa2008,0xaa028a8a,0x2808828,0x2008a222,0xa20a80aa,0xa80a8228,0x22288022,0xa0880a02,0x8a8a02a8,0x80228002,0xaa88a828,0x80aaa20,0x20222822,0x20808002,0xa880020a,0xa22802a8,0x22222820,0xa2228a80,0x80a00022,0x8228020a,0x2080202,0xa808008a,0x82082000,0x88200aa,0x2a202000,0x80a2aa8,0x228a82a8,0xa280a00,0xa028a0a,0xa82a2880,0xaa8a0202,0x808a82,0xa0082200,0x82082a,0x200202a0,0x802080a,0x8028828,0x8000a02a,0x2a2888a8,0x82888880,0xa80288a0,0x80022a2,0x20208a,0x82888088,0x8aaa0208,0x82a22aa0,0x288022,0xa80a2a28,0xa0220a2,0x80228888,0x882a02a8,0xa02a08a2,0x88a2008a,0x82202808,0x2a280aa2,0x8a000a,0xaa22a0a2,0x202aa882,0x8088a088,0x82020008,0xa2000828,0x88a00aa2,0x8888a082,0x888a2a00,0x280008a,0x8a0082a,0x820080a,0xa0200228,0x8208002,0x220802aa,0xa8808000,0x2028aaa0,0x8a2a0aa0,0x28a02800,0x8);
remainders_neg[737].modulo_value=2948; remainders_neg[737].bit_mask=create_bit_mask(92,0x8a08220a,0x220202a8,0x222a0800,0x88a8aa20,0x208a0008,0x80a80a82,0x880820,0x8a880a0a,0xa802a82,0x2888a082,0x82022aa2,0xa82aa2a8,0x22082aa0,0xa888020,0x888080a2,0xa82a28aa,0xa222080a,0x28200220,0x8082a800,0x22820202,0xa0800a0,0xa8802aaa,0x2a820022,0x20020aa0,0xaaa008aa,0x28008282,0x2020a20,0xaa080a,0x220020a0,0x80822228,0xa8a2a0aa,0x2808088a,0x20088a82,0x2aa08220,0xaa2aa0a8,0x2aa20208,0x82888a2,0xaa00a82,0x82808a8a,0x20808802,0xa80a808,0x80028822,0x22a8a888,0x82a220,0xaa020220,0x82208288,0x828a0282,0x20a2a000,0x22020a8a,0x8008020a,0x2880aa88,0x880a8022,0xa8828228,0x8a8082,0x808a00a0,0x88820a2,0xa2200020,0xa000008,0x208a0020,0x80882a2a,0x8028a802,0xa000800,0x2228a00,0xa22a228,0x8a008aa,0x2082a2a0,0x828aa02a,0x82a0000,0xa28a20,0x28a28022,0x2a0802,0x2a8a080,0xa2a0822a,0x88028802,0x22a2282a,0x28a2200a,0x8002800,0xa8a0080,0x2a0880a0,0x28822a,0x2802,0x22288,0x82088802,0x802880a2,0x80a88002,0x20a08aa0,0xa8088a,0xaa808a22,0x20080088,0xa8202228,0x2a28228,0xa028a080);
remainders_pos[738].modulo_value=984; remainders_pos[738].bit_mask=create_bit_mask(31,0xa2882802,0x220008,0x200208,0x828220a0,0x20208000,0x2280200,0x808a0000,0x20820828,0x88008,0xa00228a2,0x8220a20,0x200828a,0xa20a2802,0x20228820,0x8000082,0x200a00a0,0x8820000,0x8a082288,0x808028a0,0x20a28200,0xa08a088,0x8a2880,0x8200220,0x2282082,0x800000a2,0x8008028,0x82000208,0xa0882,0x208008,0x8a200088,0x802822);
remainders_neg[738].modulo_value=984; remainders_neg[738].bit_mask=create_bit_mask(31,0x22002082,0x8a28800,0x88288280,0x20a2820,0x28a00808,0x8a208288,0x20880,0x28028020,0x8000080,0x208a2022,0xa28228,0x8a080202,0x22822082,0x28a20028,0x80088002,0xa0820820,0x28808,0x200a200,0x82020,0x28220a08,0x82002000,0x80022000,0xa08a28,0x8a20a00a,0x880022,0x800820,0xa088280,0x80820002,0x8a00800,0x2280000,0x820a2);
remainders_pos[739].modulo_value=2956; remainders_pos[739].bit_mask=create_bit_mask(93,0xaaa82aa,0x820aa228,0x802a0aa0,0xaa8a8280,0x2228a8a8,0x2220aa0a,0x820a020,0x8a2a8802,0xa020000a,0xaaa00022,0x22a82a08,0x22a22a28,0x802a8280,0x822222aa,0xa888a00,0x88820208,0xa002288a,0x82a8a800,0x220a88a2,0x88222880,0x822aa20a,0x82820828,0x820aa802,0xa0028a02,0xa28a0a0a,0x88008a20,0x2088a202,0x220288aa,0xa0200a08,0x208aa82a,0x8a8a2202,0x822202a2,0x8888a2a,0xa00aa00,0x808808aa,0x80a008a0,0xaaa80022,0xaaa8a808,0xa2008202,0xa828a28a,0x802888a8,0x20208882,0xa020020,0x280aa2a,0x88028a28,0xa0002a0,0xa0000,0x280220a8,0xa028028a,0x80a0a8a,0x20808080,0x22802822,0xa282a2a2,0x2008aa28,0xa2aaa808,0x2aaa202,0xa02880,0x22020a2,0xa00a0aa2,0x2222000a,0x8882a8a,0x8a2a28a8,0xaa208808,0x80aa82,0x2808820a,0xa2208aa2,0x20022808,0xa28a08a,0x2800aa0a,0xaa08280a,0x8282802,0xaa882282,0x88822a08,0x2a088022,0xa2a828a2,0x8800a002,0x8222a22,0x222a0208,0x8888200a,0x2a802aa8,0x88a88028,0x82a8828a,0xaaa20a,0x80a880,0x2a8a2a00,0xa0820802,0xa0888080,0xa2888a0a,0x2a2aa2a2,0xa802028,0xaa0820aa,0x2aaa0288,0xaa8);
remainders_neg[739].modulo_value=2956; remainders_neg[739].bit_mask=create_bit_mask(93,0x82220a22,0xa822aa0,0x8a28228,0x22020a08,0xaaa02020,0xaaa82282,0x80a828a8,0x2a2008a,0x28a88882,0x222888aa,0xaa20a280,0xaa2aa2a0,0x8a20a08,0xaaaaa22,0x82000288,0xa8a80,0x288aa002,0xa202088,0xaa82002a,0xaaa008,0xaa22a82,0xa0a80a0,0xa82208a,0x288a0282,0x2a028282,0x8802a8,0xa8002a8a,0xaa8a0022,0x28a88280,0xa80220a2,0x202aa8a,0xaaa8a2a,0x800002a2,0x82882288,0x8008022,0x8288028,0x222088aa,0x22202080,0x2a880a8a,0x20a02a02,0x8a00020,0xa8a8000a,0x828a88a8,0x8a0822a2,0x8a02a0,0x82888a28,0x88828888,0xa08aa820,0x28a08a02,0x80828202,0xa8080808,0xaa08a0aa,0x2a0a2a2a,0xa88022a0,0x2a222080,0x8a222a8a,0x8828a008,0x8aa8a82a,0x2882822a,0xaaaa8882,0x8000a202,0x2a2a020,0x22a80080,0x8808220a,0xa0800a82,0x2aa8022a,0xa88aa080,0x82a02802,0xa0882282,0x2280a082,0x80a0a08a,0x2200aa0a,0xaa280,0xa28008aa,0x2a20a02a,0x88288a,0x80aaa2aa,0xaaa28a80,0xa882,0xa208a220,0x2008a0,0xa200a02,0x88222a82,0x88082008,0xa202a288,0x280a808a,0x28000808,0x2a000282,0xa2a22a2a,0x8208a8a0,0x2280a822,0xa2228a00,0x220);
remainders_pos[740].modulo_value=1480; remainders_pos[740].bit_mask=create_bit_mask(47,0xa22a02,0x2020a00,0x20820080,0x22200aa,0x82a0a820,0x28a20288,0x2a088a02,0x220020a8,0x822a802,0x2000802,0x28000282,0xa80a02a2,0x28000288,0x2000808,0x2a88208,0xa0800888,0xa220a82,0x2808a288,0x82a0a822,0xa0088800,0x2008208a,0xa080800,0xa88a000,0xa22a028,0x2020a000,0x8200800,0x22200aa2,0x2a0a8200,0x8a202888,0xa088a022,0x20020a82,0x822a8022,0x20008020,0x80002820,0x80a02a22,0x8000288a,0x20008082,0x2a882080,0x8008880,0xa220a82a,0x808a2880,0x2a0a8222,0x888008,0x8208aa,0xa0808002,0xa88a0000,0x80);
remainders_neg[740].modulo_value=1480; remainders_neg[740].bit_mask=create_bit_mask(46,0x882a228a,0x8a0a8280,0x280a8008,0x8aa20822,0x2282028,0xa02a0a00,0xa2000282,0x2a88a020,0x80a22082,0x8288800a,0xa0880a0a,0x20028a22,0x20088200,0x8a808080,0x82200a00,0x28080000,0x822a8202,0x20802200,0xa2020aa,0x20800008,0xa8802802,0x2008080,0x2002088,0x82a228a0,0xa0a82808,0x80a80088,0xaa208222,0x22820288,0x2a0a000,0x2000282a,0xa88a020a,0xa220822,0x288800a8,0x880a0a8,0x28a22a,0x882002,0xa8080802,0x2200a008,0x80800008,0x22a82022,0x8022008,0xa2020aa2,0x8000080,0x88028022,0x2008080a,0x20020880);
remainders_pos[741].modulo_value=2964; remainders_pos[741].bit_mask=create_bit_mask(92,0xa2000822,0x20a28820,0x80008,0x22822802,0x28a20a00,0x2082002,0x808000a2,0x228800,0x88282202,0x820a0002,0x208800,0x8008082,0xa2800080,0x20,0x82002000,0x800200a0,0x8820220,0xa282202,0x208a0880,0x28200,0x80280002,0x2802020,0xa20008,0x2280202,0x200a0080,0xa00820,0xa000200,0x82882002,0x20a28008,0x2088202,0x2000822,0x20020828,0xa2820,0x8008008,0x88200208,0x2820800,0x20a08a00,0x8a200202,0x80800,0x20820228,0x8228a200,0xa20a2002,0x800200,0x8a282080,0x80022820,0x208a20,0x8a208288,0x20002082,0x8a00800,0x8828008a,0x208008a2,0x28808,0x8a200202,0xa0880800,0x28000a20,0x82200208,0x2020800,0x20020,0x828a,0x800000,0x8028208,0x8088200,0xa0880822,0x28220028,0xa080082,0xa0000800,0x808200,0x8800200a,0xa0080802,0x28020008,0x80208080,0x80002,0x20800828,0x8a00220a,0x22080882,0x208808,0x820a008,0x80,0x28a08000,0x2002000,0x80082020,0x8200220,0x8228000a,0x80080882,0x20200228,0x808a000,0xa2880082,0x28800a08,0x80288,0xa0820002,0x8a08228,0x88200);
remainders_neg[741].modulo_value=2964; remainders_neg[741].bit_mask=create_bit_mask(93,0x228000a2,0x28220028,0x8080,0xa20a2082,0x20220208,0x8a00a08a,0x80822,0x8220008,0x20a28a,0x2820082,0x8a00000,0x8008000a,0x22000800,0x8808828,0x208a008,0x8a0820,0x28a28,0x8220220a,0xa0020000,0x8820a00,0x820808a,0x20828a0,0x8228000,0xa20828a,0xa0820000,0x8208028,0x82088288,0x2002802,0x20220800,0x8a000282,0x828800a2,0x28820020,0x88088080,0x800220a0,0x800800,0x288280,0x20a0000,0x28200208,0x220828a,0x80800080,0x28028a20,0x2202288,0x22822882,0x8000208,0x220a008,0xa20a0,0x8200228,0x2280200,0xa0882002,0x208008,0x200002,0xa0000022,0x820000,0x228820a,0x20000080,0x20808228,0xa280280,0x82820080,0x8828820,0x8080202,0x80080880,0x820200,0x288,0x200000a2,0x20228820,0x8200800a,0x20880000,0xa00,0x8a082,0x20800082,0x20828800,0x280000,0x80000882,0x28008020,0x208a202,0x22800002,0x8a00000,0x80202080,0x880800,0x20200808,0x8208a088,0x8020a0,0xa00a28,0xa208082,0x800002,0x28208a20,0x80002080,0x22000002,0x20000200,0x88008200,0x20020882,0x200a20,0x208);
remainders_pos[742].modulo_value=2968; remainders_pos[742].bit_mask=create_bit_mask(93,0xa882a0a,0x2008a82,0x2a8aa0,0x8a222808,0x882a802,0x82000a08,0x882a8080,0xaa88a022,0xa28a082,0x2820aa2,0x880888a2,0x20228a82,0x28,0x20a820aa,0xa02a02a,0xa002220,0x22802a,0xa008a808,0x82200000,0x82aa2028,0x20228080,0x2008a80a,0x228020a0,0x22820a20,0x2a280a,0x82222088,0x20a88,0xa822802a,0x88a0a0a,0x228a8a0,0x22a0802a,0x2202822,0x2802200a,0x8aaa2008,0x802a202a,0x88888200,0x2280a200,0x208200,0x8000082,0x8020820a,0x8aa20a88,0x8880a0,0x22008228,0x28a8a0,0x82228a00,0x2800a080,0x28020080,0x82020a00,0xa288,0x880a2a28,0x288200,0xa20a0222,0x222a08a,0x20a08208,0x820000,0x88008208,0x22008a02,0x2008222,0xa2a808a8,0x282008aa,0x80a00880,0x88882808,0x80a8020a,0x200a2a28,0x2aa0a0a2,0xa80288,0x8222a080,0x220888,0x88a028a8,0x8808a082,0x80a0802,0x8a02a20,0x82020288,0x822808aa,0xa000008,0x202a20,0xa0a80288,0xa0088800,0x8a80a80,0xaa082a,0x8280000,0x2282a288,0x808a2220,0xa08aa082,0xaa820a28,0x22880a22,0x820202a8,0x2020a000,0xa2802a82,0x202888,0x800aa2a8,0xa082a200,0xa0a822);
remainders_neg[742].modulo_value=2968; remainders_neg[742].bit_mask=create_bit_mask(93,0x8200a202,0xa880202,0x80a20228,0x22aa080,0x2208a,0xa880280,0xa20008,0x2200282a,0x2a02802,0x820a822a,0x2a,0xa8a2020a,0x888808a0,0xa820a022,0x828a2822,0x288aaa0,0x80aa08a2,0x28002080,0xaa08808,0xa2228a0,0xa8aa0008,0xa8802002,0x2a08a820,0xa20a82a8,0x8022a082,0xaa2a800,0x888a0200,0x20aa00a2,0x80028202,0xaa02020,0xa22808a2,0x8a2820aa,0xa082a882,0x2222880,0x8a2a0a2,0xa08,0x2a082a80,0x80a80a88,0x80088802,0x8a00a82,0x22a0200,0x88000028,0xaa880a20,0x8a02020,0x22a0288,0xa0082808,0xa0828808,0xa8a0288,0x88882200,0x82a220,0x8a00a80,0x228282aa,0x8a2a2802,0xa8200a80,0x880a0888,0x880280,0xaa88020a,0xa880aa2,0x22208020,0xa0288022,0x8208008,0x2080,0x8208282,0xa882a220,0x22202822,0x80208a00,0xa2a2808,0x88a28000,0x282020,0x80200a,0x8082800a,0x28a220,0x28a8a00,0xa208022,0x82808880,0x88a822a8,0x28208200,0x20800008,0x208200,0x802280a2,0x80208888,0xaa022a00,0x8022aa8,0x2802200a,0x220a0220,0x2a0082a2,0x28a8a20,0xa8282888,0x2a00a20a,0x88a82000,0x8822220,0x280a2a00,0x2820a2);
remainders_pos[743].modulo_value=2972; remainders_pos[743].bit_mask=create_bit_mask(93,0x2008282,0x8aaa02a2,0xa0a2802a,0x2082a22,0x2882a082,0x8888a288,0xa8a20000,0x2a8208aa,0x82222808,0x82aa0a02,0xa8a82808,0x2aaaa2a8,0xaa0882a2,0xaaa2002a,0xa2288a80,0x2000a088,0x800a802a,0xaa08a88,0x8a8a208,0xa022a8a,0x8a00a88a,0x28802080,0xaa80a2a8,0x2a002208,0xaa20a008,0x2a822a28,0x8a828220,0x20228200,0x2802a288,0x2aa2202,0x2aa880aa,0x20882228,0x88002a02,0x228000aa,0x880a,0xa0202008,0x800a22a0,0x888a0a82,0x8a00a2a0,0x88200022,0x22222aaa,0xa20a208,0xa28a8a28,0x8280880,0x800200aa,0x2a8a880a,0x8a0a0a0a,0x2a022a2a,0xaa00800,0xa0220282,0x28a2a28,0x8208a08a,0x2aaa8888,0x28800082,0x20a8a00a,0x282a0a22,0xa0a88a00,0x2008080,0x8a022000,0x2aa00028,0x880a8002,0x82888220,0xaa022aa,0x88088aa8,0x8228a802,0x20082880,0x8088282a,0xa28a882a,0x8200a08a,0xa208800a,0x82a8a02a,0x20208022,0xa22a00a,0xa2a880a,0x208a2a2,0x222a20aa,0x8a802a00,0xa220a000,0xa02a2288,0xaa8008aa,0x88a8220a,0xa2a8aaaa,0x220282a2,0x280a0aa8,0x82028888,0xaaa2082a,0x200008a2,0x8228a222,0x820a822,0xa88a8208,0x2a8028a0,0x8a80aaa,0x8282008);
remainders_neg[743].modulo_value=2972; remainders_neg[743].bit_mask=create_bit_mask(93,0x8a880a0a,0x2228a2a,0x282a08a2,0x8a80a2aa,0xa00a280a,0x2a00,0x202a8888,0xa20a8022,0xaaaa080,0xa22828a,0x2020a080,0xa2222a20,0x22800a2a,0x222a88a2,0x2aa00208,0xa8882800,0x88208a2,0x82280200,0x80202a80,0x828aa202,0x2882002,0xa008a808,0x22082a20,0xa288aa00,0x22a82880,0xa20aa2a0,0x20a0aa8,0xa8aa0a88,0xa08a2a00,0x8a22aa8a,0xa2200822,0xa800aaa0,0x88a28a,0xaa088822,0x88880082,0x28a8a880,0x882aa28,0x2820a,0x2882a28,0xa888aa,0xaaaaa222,0x82a82a80,0x2a0202a0,0x80a08008,0x88a8822,0xa2020082,0x2828282,0xa28aa2a2,0x82288088,0x28aa8a0a,0x8a02a2a0,0xa802802,0xa2220000,0xa008880a,0xa8202882,0xa0a282aa,0x28200288,0x8a880808,0x28aa888,0xa22888a0,0x82088a,0xa000aa8,0x8228aa22,0x800220,0xaa0208a,0xa880a008,0x800a0a2,0x2a0200a2,0xa882802,0x2a800882,0xa2028a2,0xa8a808aa,0x82aa2882,0x82a20082,0x8a802a2a,0xaaa2a822,0x208a288,0x2aa82888,0x28a2aa00,0x22088022,0x20aa82,0x2a202222,0xaa8a0a2a,0xa0828220,0xa8a0000,0x222a80a2,0xa888802a,0xaa02aaa,0x80a820aa,0x20020a80,0xa208a028,0x80208222,0xa0a880);
remainders_pos[744].modulo_value=744; remainders_pos[744].bit_mask=create_bit_mask(24,0x28220a2,0x28020828,0xa200002,0x228808a2,0x200a08,0x220a282,0x20820002,0x208028,0x200088,0xa00228a2,0x800820,0x80000202,0x20020080,0x28800a08,0x8008a,0x2080022,0x820828,0x8a088080,0xa0080082,0x20228820,0x8a08a,0x20802880,0x8828028,0x8a);
remainders_neg[744].modulo_value=744; remainders_neg[744].bit_mask=create_bit_mask(24,0x20a2822,0x20828020,0x8228808a,0xa2000022,0xa08200,0x8a28220a,0xa00a0882,0x8a00820,0x88288000,0x208a2022,0x8008028,0x808828a,0xa0820800,0x20008200,0x88008002,0x828008a2,0x8020020,0x2000008,0x20800802,0x28a20028,0x88002002,0xa0082000,0x20820,0x2);
remainders_pos[745].modulo_value=745; remainders_pos[745].bit_mask=create_bit_mask(24,0xad8d335e,0x665348d1,0x30ca7121,0x16b41107,0x35090603,0x8d885acc,0x81018037,0xb9c47b8,0x8ab4202f,0x711084b,0x91445a50,0x4a794ae2,0x688a251d,0x42238029,0x10b54748,0x88e743d0,0x6020477,0x6846c7b0,0x8242b0cd,0x20b5a301,0x394c3382,0x4b299a12,0x32c6d62c,0x1eb);
remainders_neg[745].modulo_value=2980; remainders_neg[745].bit_mask=create_bit_mask(94,0x20002a82,0xaa0a8000,0x2802a0a8,0x8a20888a,0xa0808a0a,0x200,0x80808a2,0x2008220,0x220a8a2,0x82888002,0x8880288,0x82208222,0x2002a080,0xaa208a0,0x80288a00,0x2a0a08,0x8a0a88a2,0x208a0228,0xa822800,0xa0282a08,0xa0802a0a,0x82200282,0x2a0a02a0,0x82a22022,0x22082a02,0x82a0a00,0xa02a82a0,0xa8082a0,0x882000a2,0x8288a202,0xa002288a,0xa0088282,0x2a800288,0x202882,0x2208082a,0x88822208,0x8a088a00,0x22020008,0x20228a8,0x80802208,0x22880,0x8280200,0x228a8288,0xa28888,0x82a8a82a,0x20000a,0x220aa0,0x2a8a22a8,0x8000080,0x80800a20,0x80800208,0xa2280820,0x802a2a8,0x8222888a,0x20a228,0xa8a22208,0xa0088080,0x22028088,0x8a0a028,0x80282a20,0x80a80002,0x82088a00,0x802a200,0x82008080,0x2a82802,0x8a0082,0x2000a02,0x208020a8,0x8020a02a,0x22200202,0xa02002,0x822a0282,0x200a8200,0x80202800,0xa0208028,0x20a00a0a,0x8008080,0x208022a0,0x80802888,0xa0000a,0x88022a0a,0x220a0282,0x28880a0,0x8a808088,0x82222,0x208a2282,0x82888a2,0x228aa2a0,0x8082080a,0x80882000,0x8022800,0xaa008000,0xaa228,0x2);
remainders_pos[746].modulo_value=2984; remainders_pos[746].bit_mask=create_bit_mask(94,0x820a0aa2,0xa0a22a88,0x82a222a8,0x22808088,0x2a202a28,0x802aa00,0x28a2028,0x220a00a,0x80aa2202,0x2802a82a,0xaaaa888a,0xaa0a2208,0x222200a,0x22882a00,0xaa208228,0xa0aa8a8a,0xaa028a08,0x22880828,0x28aa288a,0x8082a82a,0xaaa28000,0x2aa0a022,0xa0a8002,0xa0a202a2,0x22aa80,0x22a0a,0x8a8aaa82,0x8202802,0x82208820,0xa00828a8,0xa08a082,0x20008080,0x8228228a,0x22aaa028,0x2000aa22,0x8a220,0xa8208880,0xa802802,0x22a2a220,0x82a0aa0a,0xa8a82a20,0x202aa002,0x8228202a,0x2888a8a,0xa080222,0xa2888022,0x8200a0,0x2228a0a,0x8020a088,0xa2228088,0x82882a2,0xaa808a8,0xa82a2a80,0xaa0a8208,0x8a8a88a0,0x2802a008,0x22082a80,0x8a200002,0xaa000808,0xaaa8888,0x88288228,0x20008a2,0xa20a002,0x28200a82,0x2208822a,0x28082008,0xaaa2a280,0xa8800082,0xaa8800a0,0x808a0a02,0x2a0a08a,0xa0aa880,0x28aaa88,0x2a820200,0x28aa28a8,0x202288a2,0xa280aa28,0xa2aa0a20,0x8208aaa2,0xa8228828,0x8888000,0x88a0aaa0,0x22aaaa20,0x2a8028a2,0x88aa02a8,0xa088080,0x8a280a0,0xaa802028,0xa808a800,0x2028828,0x888a8222,0xa88a0a2a,0xa0a08222,0x8a);
remainders_neg[746].modulo_value=2984; remainders_neg[746].bit_mask=create_bit_mask(94,0xa82822a,0x282aa200,0xa2aaa20,0xaa080800,0xa2a8a2a0,0x808a2288,0x8a02a8a0,0x8aa82882,0x822aa8a,0xa08a20a2,0x22220002,0x2282aa80,0x8aaaa882,0xaa00a288,0x22a80aa0,0x28220202,0x228a0280,0xaa0080a0,0xa022a002,0x80a20a2,0x222a0888,0xa22828aa,0x8282088a,0x282a8a2a,0x88aa2208,0x888aa282,0x202220a,0x80a8a08a,0xaa800a8,0x2880a020,0x8280280a,0xa8880808,0xaa0aa02,0xaa2228a0,0x288822aa,0x88802aa8,0x20a80008,0x8208a08a,0xaa2a2aa8,0xa282282,0x2020a2a8,0xa8a2288a,0xaa0a8a2,0x8a000202,0x82808aaa,0x2a0008aa,0x880a8828,0x8aaa0282,0x8a82800,0x2aaa0800,0x80a00a2a,0x82208020,0x20a2a208,0x22820a80,0x2020028,0xa08a2880,0xaa80a208,0x2a8888a,0x22888080,0x82220000,0xa00aa0,0x8a88802a,0x82a8288a,0xa0a8820a,0xaa800aa2,0xa080a880,0x222a2a08,0x2008880a,0x22008828,0x802828a,0x8a282802,0x82822008,0x8a022200,0xa20a8a88,0xa022a020,0xa8aa002a,0x2a0822a0,0x2a2282a8,0xa80222a,0x20aa00a0,0x80000888,0x202228,0xaa2222a8,0xa208a02a,0x228a20,0x82800808,0x802a0828,0x2208a8a0,0x20802088,0x8a8a00a0,0x20aaa,0x200282a2,0x28280aaa,0x2);
remainders_pos[747].modulo_value=996; remainders_pos[747].bit_mask=create_bit_mask(32,0x220a0002,0x20828a28,0x800a08a,0x20a2880,0x8008800,0x8220a,0x82080022,0x28028020,0x8208280,0x228808a2,0x28200808,0x828a202,0x220a2000,0x20800200,0x8080000,0xa28028a0,0x20200,0x8002080,0x8a0880,0x8a28200,0x2008288,0xa2022882,0x28208208,0x80280280,0x80020820,0x8820008,0x220020a,0x228a0800,0x20a00200,0x8a28208a,0xa0882,0x8);
remainders_neg[747].modulo_value=996; remainders_neg[747].bit_mask=create_bit_mask(31,0xa2820882,0x28020220,0x80002002,0x82822000,0x800008,0x8800a282,0x28008a2,0x20820828,0x80280208,0xa2000022,0x20a08000,0x8020228a,0x22822880,0x28008a08,0x80008088,0x22082020,0x8828a08,0x8008a008,0x80020000,0x220a08,0x8a080200,0x228a2002,0x20a00a00,0x8208208,0x8a00a0,0x28800,0x8a288282,0xa2020080,0x28208a08,0x220a002,0x80820002);
remainders_pos[748].modulo_value=1496; remainders_pos[748].bit_mask=create_bit_mask(47,0xaa88020a,0x22220a08,0x2a8a0aa2,0x88002200,0xa0880202,0x2802a08,0x82080aa,0x20a8a22,0xa288aa,0x8880080,0x28280,0x28028282,0x28000828,0x2220028,0x28a00020,0x2a080aa2,0x2082088a,0x80280aa0,0x220a20a,0x80022808,0xa2a8008,0x88888aa,0x22aa20a,0x8020aa08,0x20a08aa8,0xa0aa2222,0x22002a8,0x80202880,0x2a08a08,0x80aa028,0xa8a22082,0x288aa020,0x8008000a,0x28280088,0x28282000,0x828280,0x20028280,0x20022,0x80aa228a,0x2088a2a0,0x80aa0208,0xa20a802,0x22808022,0xa8008800,0x888aa0a2,0xaa20a088,0xa08022);
remainders_neg[748].modulo_value=1496; remainders_neg[748].bit_mask=create_bit_mask(47,0x22008282,0xaa228280,0xa202822a,0x8aa80,0x28000a8a,0x8208a280,0x80a80822,0xa82022a,0x88220022,0x8808,0x888a0a00,0xa08a020a,0xa0088020,0x8aaa88a0,0xa02880a0,0xa280022a,0xa00a0002,0x8a08228,0x8aa02a02,0x882a080,0x2220880,0x80000022,0x82a22282,0x8282280,0x28280220,0x2822aaa2,0x8aa80a20,0xa8a000,0x8a280280,0x80822820,0x2022a80a,0x200220a8,0x8808882,0xa0a00000,0xa020a888,0x88020a08,0xa88a0a00,0x880a08aa,0x22aa02,0xa0002a28,0x8228a00,0x2a0208a,0x2a0808aa,0x20880088,0x22022,0x22282800,0x28082a);
remainders_pos[749].modulo_value=2996; remainders_pos[749].bit_mask=create_bit_mask(94,0xa2820a22,0xaa020a0,0x228200,0x2a08008a,0x2aa22a0a,0xaa02288,0x20822a,0x22808222,0x20882aa2,0x80888082,0xa02a2880,0x2020a202,0x2a00880,0x2a228,0xa0008200,0xa020,0x22a8a0a,0xaa088008,0xa2a20080,0x22a0888,0x2882a2a0,0x802a00,0x22288282,0x200a0280,0x20a08a2,0xa002a008,0x2a2a08,0x288882a2,0xa808022a,0x800882,0x2828aa2,0x28202800,0x20800,0xa22a00,0x2008802a,0x2a022820,0x808a2a0,0x22080888,0x222aa088,0x222080a,0x82a20820,0xa08a202a,0xa282a22a,0x2880082,0x80020a20,0x2828202a,0x8a22820a,0x828a0828,0x2a8082,0x28008a08,0x28a82002,0x20aa88a8,0xa82a808a,0x88008208,0x888a0208,0x88220aa,0x2022202,0xa80a8a2,0x808288,0xa00a8022,0xa88,0x82800208,0x28000280,0x2008aa28,0x2a82200,0x228a8802,0x8008a822,0xa20a8a,0x80200a8,0x88a20a,0x8880280a,0x20082822,0x22800a80,0x8800a8a8,0xa8a2220a,0xaa02008,0x88020022,0xa0a2a,0xa080a0,0x820,0xa80288a8,0x80802200,0x80a808a0,0x2020228a,0x20882022,0x2888aa82,0x80088820,0xaa0a8820,0xaa822880,0xa8a0a88,0x800a2002,0xaa000828,0x28a0a080,0x88a08);
remainders_neg[749].modulo_value=2996; remainders_neg[749].bit_mask=create_bit_mask(94,0x2a0a822a,0x228a820,0x80aa0a88,0xa2008002,0xa222a282,0x82282a00,0x88a802a2,0xaa080a2a,0x2800a222,0x80a,0x2822a008,0xa8a02a8a,0x8a280008,0x888a22a0,0x28880a08,0x88828a0,0x82220282,0x22000880,0x2a228808,0x8aa20000,0xa00a2228,0x8808a208,0x2aa00a02,0xa0828a08,0x8a02802a,0x28822880,0x88a22280,0xa000022a,0x20808a22,0x8088002,0x820a0222,0xa028a088,0x88828088,0x882a2288,0xa88000a2,0xa28aa028,0x802a20,0xa2808000,0xaa222800,0x8aa28082,0xa2a00a8,0x2802a0a2,0x2a0a2a22,0xa000802,0x8a82a8,0xa020a8a2,0x2a20a82,0xa0200a0,0x88a2000a,0xa0880200,0x2020a882,0xa0220020,0x20220802,0x800a80,0x20a80,0x800aa022,0x828aaa0a,0x2082022,0x80080a00,0x280208aa,0x88808200,0xa080a80,0xa0888208,0xa8802220,0xa20aa80,0xa202008a,0x80020aa,0x88228202,0x808a0820,0x88002282,0x8a002,0x2880a0a2,0xa2088208,0x82020,0x2022aa82,0x82282880,0x8a80aa,0x88828222,0x8280820,0x808880a8,0x200a0020,0x800aa88,0x8200028,0xa8a0a202,0xa800a82a,0x20002202,0x8000a8,0x220200a8,0x2202a008,0x82020200,0x882a08a,0x22888020,0x20282800,0x280);
remainders_pos[750].modulo_value=120; remainders_pos[750].bit_mask=create_bit_mask(4,0x200a2082,0x20020,0x8080080,0x8208a0);
remainders_neg[750].modulo_value=120; remainders_neg[750].bit_mask=create_bit_mask(4,0xa0822802,0x8028820,0x8008,0x20020);
remainders_pos[751].modulo_value=3004; remainders_pos[751].bit_mask=create_bit_mask(94,0x8a20aaaa,0xa0a220aa,0x2a0aa80a,0x2220a28a,0x88288080,0x22a08220,0x8200022a,0xa28a2aa,0x88a2a88,0x822020a,0x80a0a8a8,0x88880a80,0x80aa2882,0xaa22a800,0x202aa280,0x800a8a2,0x88022008,0xa880aaa,0xaa8282a8,0x20028a20,0xa8080882,0xa22828a,0x8a02a880,0xa020a82,0xaa200a82,0x20a0882,0x8a22a2a0,0x28820aa8,0xa00a0a00,0x2a202,0xa2a88aa2,0x88202aa2,0x2a0800a8,0x2aa00880,0xa20802a,0x2a02a222,0xa020282a,0x828a88a2,0xa2008222,0x82082,0x808aaa8a,0x288a2808,0xaa2a20a2,0x82082888,0x2a00280,0x288828,0x2882,0x8228000,0x2822280,0x202800a8,0xa2228208,0x88a08a8a,0x22028a22,0xa2aaa20,0xa8208200,0x28882008,0xa8a22a28,0x8a828080,0x888a80a,0x8a80208a,0x802200aa,0x22a0020a,0xa8aa8082,0x8aa22a8,0xa808a800,0x800a0a00,0x22aa0822,0xa8a88a,0xa8220a08,0x82a008a,0x282a080a,0x22a80a,0xaa28288a,0x88220202,0xa08a2800,0x2a8282a,0x2aaa022a,0x2008802,0x88a2a002,0xa028aa80,0x2002a88a,0x28228aa0,0x202a0222,0x2a2a0a0,0xa080882,0x22a8a22,0x2aa8a28a,0x8a880008,0x208820a8,0x80202282,0x8a28a088,0xaa02aa0a,0x2aa088a0,0xaaaa08a);
remainders_neg[751].modulo_value=3004; remainders_neg[751].bit_mask=create_bit_mask(94,0x2a82222,0x282aa822,0xa2822082,0xaaa82a02,0xa00808,0xaa280aa8,0xa888aa2,0x82a02a22,0x8002a200,0x80aa8a82,0x8282020,0x8208,0x822a00a,0x22aa2088,0xa8a22a08,0x8088202a,0x8aa880,0x82008222,0x220a0a20,0xa88a02a8,0x2080800a,0x82aa0a02,0x28a2008,0x828a020a,0x22a8820a,0x8a82800a,0x2aa2a28,0xa00a8220,0x28828288,0x888a2a8a,0x2a20022a,0xa8a22a,0xa2808820,0xa2288008,0x82a808a2,0xa28a2aaa,0x28a8a0a2,0xa02002a,0x2a880aaa,0x8880a80a,0x8022202,0xa002a080,0x22a2a82a,0xa80a000,0x8a288a08,0x88a000a0,0x8888a00a,0x80aa0888,0x8a0aaa08,0xa8a08820,0x2aaa0a80,0x280202,0xaa8a02aa,0x82a222a8,0x20a80a88,0xa000a880,0x202aa2a0,0x20a0808,0x80002082,0x208a802,0x8aa8822,0xaa288a82,0x2022080a,0x8022aa20,0x20802088,0x8828288,0xaa2280aa,0x88202002,0x20aa8280,0x80a28802,0xa0a28082,0x88aa2082,0x22a0a002,0xaa8a8a,0x2802a088,0x8a20a0a2,0xa2228aa2,0x8a88008a,0x2a288a,0x28a02208,0xa88a2002,0xa0aa0228,0xa8a28aaa,0x8a2a2828,0x8280800a,0x8aa202aa,0xa2202a02,0x2008880,0xa800a820,0x8a8aa0a,0x2a02800,0x228a2282,0xa2280028,0x2222802);
remainders_pos[752].modulo_value=376; remainders_pos[752].bit_mask=create_bit_mask(12,0x82aa8a02,0x202208a8,0xa82000a,0x8a00082a,0x22a20880,0x280a2aa8,0x882aa8a0,0xa202208a,0xa0a82000,0x8a00082,0x822a2088,0x80a2aa);
remainders_neg[752].modulo_value=376; remainders_neg[752].bit_mask=create_bit_mask(12,0xa22028a,0xa8aa0020,0x820a8882,0x28880a2,0xaa2a8008,0xa082a220,0xa22028,0x2a8aa002,0x2820a888,0x8028880a,0xaa2a800,0x82a22);
remainders_pos[753].modulo_value=3012; remainders_pos[753].bit_mask=create_bit_mask(95,0xa2002882,0x8228000,0x8820a288,0x2080880,0x8200,0x82208280,0x2882800,0x28000,0x208a002,0x80020022,0x8a20a28,0x82280202,0xa00a00a0,0x20200008,0x80282000,0x82020020,0x28020a20,0x828220a,0x80880082,0x28220808,0x208a282,0xa28a0000,0x8a20020,0x8228a20a,0x20022882,0x28a28,0xa288200,0x808220a2,0x8008808,0x8220a082,0x228200a2,0x20020208,0x20a008,0x80002020,0x28028028,0x200a208,0xa2822882,0x20020008,0x288202,0xa0002,0xa08a00,0xa082208,0x2080000,0x8808200,0x8a282088,0xa2080,0x8a00228,0x8800a20a,0x208a0002,0x20828a20,0x8202202,0x20800,0x8820a00,0xa20a002,0xa0000,0x8228008,0x80088,0x228828a2,0x8a00808,0x80280282,0x80800022,0xa08000,0x8080082,0xa0082882,0x20a08828,0x2200208,0xa0882022,0x8228a08,0x8a280000,0x22880082,0x8a28828,0x8008a20a,0xa28a0,0x28a20800,0x2088288,0x20022022,0x8828208,0x8a08028a,0x80080820,0x828020,0x8080,0xa00a00a2,0x8028820,0x8a08a208,0x80080022,0xa20808,0x280008,0x2822800,0x28208820,0x8200000,0x22020800,0x28a08220,0x288202,0x228008a0,0x8);
remainders_neg[753].modulo_value=3012; remainders_neg[753].bit_mask=create_bit_mask(94,0x22882002,0xa20808,0x282200,0x82800000,0x8800a08,0xa280208,0x82002080,0x820808,0x8a00208a,0x8a08a2,0x228220,0xa20828a,0x20820820,0x28a08800,0x820a088,0x28a08a0,0x20828228,0x8020a282,0x802,0x20a28000,0x8a00220a,0x22020880,0x228828,0xa202282,0xa08a2002,0x8820220,0x82200288,0xa2822,0x800000,0xa28200a,0xa20a0822,0x28828a00,0x88282080,0x8828a0,0x20820820,0x8a082280,0x220a2002,0x28828800,0x8820028a,0x80820802,0x8200208,0x8200a280,0x82800880,0xa08,0x220a000,0x80822800,0x208a20,0x82282,0xa0020882,0x28020228,0x8028a28a,0x808a0080,0x28208,0x8228208a,0x80820880,0xa20800,0x88008000,0xa2002022,0x208000,0x820820a,0x808a2,0x8200808,0x8000800a,0x20802002,0x28200020,0x8a288280,0x200028a2,0xa20200,0x2208088,0xa2000802,0x220020,0x8002282,0x80822020,0x20228008,0x8a000200,0xa08a28a2,0x20a00,0x2008202,0x8000a0,0x8020828,0x88080008,0x20820822,0x820028,0x2002280,0x8008a2,0x8228000,0x88208080,0x820a2080,0x20a00028,0x80288088,0xa28a0080,0x20200a28,0x8820028a,0xa2080020);
remainders_pos[754].modulo_value=3016; remainders_pos[754].bit_mask=create_bit_mask(95,0x8a80822a,0x2802aa00,0x220a2a28,0x2a8820a2,0x282822aa,0x8820000,0x88aa2a2,0xa280a8a,0x80a80a,0x222808a,0x22008088,0x8a2aa0,0x8aa0a22a,0xaa800080,0x808a0002,0x220a0820,0xa0208,0x800a2222,0x2828888a,0x22a22a0,0x22aa0a8,0x88282a2a,0x82008222,0x222a80,0x208a228a,0x20202882,0xa0000a0,0x2202228,0x80882820,0x200aa80,0xa2a02aa2,0x2820a028,0x8aa08a82,0x8a288002,0x2208008a,0x2082a0,0xa8aa20a,0x88822288,0x28028aa0,0x80a800a0,0x8820,0xa2a288a0,0x200282a,0xa280020,0x20208022,0x8002a820,0x220a800a,0x2a08800,0x2a8002a0,0x2080808,0x28a088,0x28008008,0x228a8aa8,0x2200000a,0x2a0208,0xa280280a,0x8882220a,0x8aa2a022,0x820800a0,0x20880a,0x228a2a2,0xa20aa280,0xa082882,0xa80a8a28,0xaa00808a,0x28220202,0x88088008,0xa028,0x2808080a,0x88a20882,0xa88800a2,0x82008202,0xa8282288,0xaa880a8,0x88a8802a,0x2228280a,0x88a002a2,0x80a00088,0x20a08820,0xa20208,0x2aa80,0x8a0aa202,0xa8a200a8,0x200880a,0x2888022,0x2a0200a2,0xa028a0a0,0x8aa220a2,0x82208a,0x88282800,0x822a8aa,0xa8a0888a,0xaa802828,0x8202a200,0xa8);
remainders_neg[754].modulo_value=3016; remainders_neg[754].bit_mask=create_bit_mask(95,0x2080aa2,0xa08a2208,0xa202a2a0,0xa200a82a,0xa0a02a22,0x800a8888,0x80022222,0x82208202,0x88082082,0xaaa0002,0x2a880800,0x8802a228,0x2202aa2,0x22008808,0x802880a,0xa28280a8,0x88828a80,0x8822a2a,0xa0a00002,0x8aa2aa20,0x82222820,0xa0a2a2,0xa8802aa,0x8aaa208,0xa8022a02,0xa8a0a00a,0x82888828,0x8aa8aa20,0xa0a0,0x8a882208,0x2a28222a,0xa02828a0,0x2280202,0x220088a,0xaa808802,0x88a80228,0x2022a82,0xaaa00,0x20820228,0x8208828,0x88880028,0x222a0028,0x8a80a0a2,0x82a008a8,0xa8a808aa,0x88a20a0,0xaa020802,0x8a280088,0xa2088228,0x2808080,0x88a02800,0xa0800880,0xaa020220,0xaa880802,0x80a28a80,0x2a08a082,0xa2a82,0x22a28a2,0xa808820,0x88280082,0x8a202a2a,0x2a822208,0x280a00a,0x208202a0,0x22800002,0xa0aa8a8a,0x800800,0x8828a0,0xa0808082,0x2a000a,0x2000882a,0xa800a82,0x2020aa00,0x82200820,0x2000a2,0x2aa0a002,0x288a2a,0x8208800,0xa02800a8,0x882a8a00,0x808a2208,0x2822a8a,0x202a0820,0x8a880082,0x8a0008a2,0xa20a882a,0x28a02820,0x22aa02a,0x80aa802,0x20a088,0x80a22022,0x20280002,0x2208a020,0x28a2288,0x20);
remainders_pos[755].modulo_value=3020; remainders_pos[755].bit_mask=create_bit_mask(95,0x20202202,0x22a8802,0x2002a2a0,0x8aa28080,0x20000028,0x288a0282,0x820a02a0,0x22020228,0x8022a82a,0x2008822a,0x2822008a,0x8a200282,0x28820088,0x8222a88a,0xa020a808,0x880280,0xa0028282,0x2208a008,0x2822028,0xa0a0aa0a,0x8082028a,0xa22a00a0,0x2280a2a8,0x802a0aa,0x280800,0x28800800,0xa0a0a20,0x82000,0x88820882,0xa0208820,0x80a2802,0xa220a82,0x2008a028,0x8820,0x22882002,0x8a80202,0x2a088082,0xa02a080,0x8000a800,0x82888800,0x8820a00,0xa0200a0,0x2a8a2088,0xa000020,0x8a82a,0xa0000a88,0x88288882,0x22822208,0xa822,0xa200022a,0xa0a82,0x8a2a8080,0x80a0220,0x82200a0,0x2228200a,0xa0002002,0xa80a0002,0x220a8020,0x2a20820,0x82288808,0x20000800,0xa2008082,0x88a0280,0x8a02082a,0x2080a802,0x8222082,0x82000822,0xa0a0000,0x22808a,0x2800002,0xa8020002,0xa0288aa0,0xa88a2a8,0x82020a0,0xa0a0aa28,0x88280a0a,0xa2088280,0x2800a200,0x2200828,0xa080a028,0xa8882202,0x8228a22,0x8a2220,0x8828828,0x22008a20,0xa8802a88,0x8088a82,0xa082288,0xa2280a8,0x8828,0x28aa2280,0xa8008020,0x2a8800a8,0x80808802,0x808);
remainders_neg[755].modulo_value=3020; remainders_neg[755].bit_mask=create_bit_mask(94,0xa8a82a8a,0x8a220082,0x288a2228,0x2220808,0xa0088820,0xa0020a0a,0xa028a20,0x2a8a82a0,0x8a220a2,0xa0800a22,0xa0aa0802,0x2288a02,0x200a8000,0xaa22002,0x20a82000,0x88000a08,0x280a0a02,0x2a802080,0x8a02a8a0,0x20282202,0x80a0a02,0x2a228820,0x2a082220,0x80822822,0x80a08008,0xa0080088,0x820282a0,0x880a088,0x2800a,0x20a80028,0x8082208a,0x822a8202,0x288020a0,0x888000a8,0xa200a80a,0x80200a8a,0xa2000802,0x28a2008,0x8802088,0x2000008,0x800a0288,0x820a8820,0x22022000,0x828088a8,0x80802022,0x28880200,0x200002,0x2a0aa280,0x888020aa,0x22888a22,0x8882020a,0x2220800,0x8282a8,0x80a28828,0xa2a0a802,0x2888288a,0x20028882,0x2a8200a8,0x8a2280a8,0x2a00000,0xa8880088,0x2a080002,0x28208,0x28280a2,0xa008200a,0x80aa280a,0xa0880a2,0x2828088,0x88a20802,0x8208880a,0x208a088a,0x28200220,0x2002220,0x80a0a828,0x20282220,0xa00282,0x2a000a00,0x20882288,0x8aa080a0,0x20082820,0x20002a0a,0x802a02a2,0x802a2a8,0x800200a0,0xa2880228,0x20082200,0x80000202,0x280a200,0x82a20820,0x80880020,0xa0222a08,0x200808a0,0x22008020,0x800008a);
remainders_pos[756].modulo_value=168; remainders_pos[756].bit_mask=create_bit_mask(6,0x2020022,0x8008a20,0x2288008,0xa2002020,0x808008,0x88);
remainders_neg[756].modulo_value=168; remainders_neg[756].bit_mask=create_bit_mask(5,0x828a0822,0x800220,0x82200080,0x220828a0,0x8000800);
remainders_pos[757].modulo_value=757; remainders_pos[757].bit_mask=create_bit_mask(24,0x5a2b3e9a,0xee3b5192,0x4e5ebe51,0x83a95098,0x9a8bc67c,0xefa3ce81,0x6cd73495,0x1fd8bbbe,0x76b4805,0x34b11712,0x87819d7c,0x21089b78,0x7847b644,0x4b0fae60,0xb8123a23,0xfe2804b5,0xcd9f7746,0x7dea4b3a,0x56605cf1,0x704f98f4,0x9c8642a5,0x1de29f5e,0x169262b7,0x165f35);
remainders_neg[757].modulo_value=3028; remainders_neg[757].bit_mask=create_bit_mask(95,0x82a2a202,0x22a2880a,0x82822288,0xa208800,0x2020aa0,0x222a0208,0xa00aa808,0x82002222,0x8aa28088,0xa8288a8a,0xa0800a0,0xa88002a0,0xa08a2a88,0x828222a8,0x208aa2aa,0x22a08828,0x2aa8a,0xa02282a2,0x8aa88028,0xa8820028,0xa8a28,0x802a0282,0x8a0aaa2a,0xa8a82a8,0xaa8288aa,0x2a00aa2,0x8a80020a,0x208a8a2,0x8aa88a0,0xa2028a0,0xaa00022a,0x882a228a,0x2a8820a0,0x220a0aaa,0xa28828aa,0x8a88a,0x82802a,0x88a0a828,0xa2a8a8a,0x20020888,0xaa802a22,0x2a22080,0x82020082,0x8822802a,0x220a0800,0x8a2a208a,0x2a2a0a80,0x2022a202,0x208a80a0,0xa2208a20,0x288aa8a0,0xa2802a82,0x2a28aa2,0x800a8a22,0xaa22202a,0x20a888a2,0x8808080,0x8aa8280a,0xaaa02a82,0x82008808,0xa2200828,0x802000a0,0x288a0a28,0xa0008220,0x20a022aa,0x8aa082a,0xa2aa0880,0x80820808,0x2a0a2aa,0x800200aa,0xa00882,0xa0880280,0x2a802000,0xaaa282a0,0x88082080,0x882aa2,0x2a082a88,0xaaa20282,0x2208002,0x8a28288a,0x82800200,0x8a080222,0x88088020,0xa0aa02aa,0x280a0aa8,0x808088,0xa2888a82,0xaa02222a,0x2228a800,0xa2a8a2a0,0x20aa00a2,0x828aa88a,0x2288222,0x280a882,0x2a202);
remainders_pos[758].modulo_value=3032; remainders_pos[758].bit_mask=create_bit_mask(95,0xa208a282,0x80aa8202,0x82a00a8,0x228aa02,0x20000aa2,0xaaa8a00,0x8a88088a,0x288020aa,0xa2802822,0x222a82a2,0x288a0a20,0x22aa22a0,0x28a2aa2,0xaa8aa0a0,0x2a2a2822,0x2aa8288,0x20220a88,0x20a00080,0x8a88a022,0xa8aa2088,0x80a8a202,0x80208a2,0xaa28a822,0x28282a00,0x8200aa0,0xa2a8a228,0x80a8208,0x200802a2,0x880888a,0xaa0a222a,0xa22a2a8a,0x8002a880,0x20202882,0x800228,0x2082aa0a,0x20002200,0xa0882a02,0x80222a20,0x2a820202,0x2a882228,0xa880800a,0x8000a088,0xaaaa2aa0,0x822a200,0xa028a2a0,0x800a880a,0x8a202a2,0x20828282,0x28a808a,0xaa022a0,0x200a8a28,0xaa008a88,0x20aa8aa,0x2a220a00,0xa8a00202,0xa8288822,0x2808082,0xa08a888,0x880a822,0x8008800,0xa0aa82,0x8288002,0x2822808,0x8a022a80,0xaaa2a8a8,0x20a888a0,0x8a22202,0x208a8020,0x8a2082a0,0x20288a2a,0x280aa008,0xaa00a828,0x20882a28,0x28a2080,0x2a808a2a,0xa22208aa,0x8880a22,0x802000a,0x8022a088,0xa82282aa,0xaa8828a8,0x800a0aa2,0x888aa8a2,0x280a88aa,0x8808a0a2,0x8a8a82a8,0x28882802,0xa2aa0802,0xa0a22022,0x800a2aa,0x808aa000,0x2080aa28,0x22a00a8,0x8a8082aa,0x828a20);
remainders_neg[758].modulo_value=3032; remainders_neg[758].bit_mask=create_bit_mask(95,0x2a802a0a,0x8220a8a,0x80a28820,0x8aa0228a,0xa888822a,0x82220288,0x2008002,0xa008a822,0x2a08a0aa,0xaaa20a2a,0xa00282a8,0xa222aa28,0x8a02a22a,0x22022828,0xa2a2a0aa,0x8a220a00,0xa8aa8200,0xa8288808,0x20028aa,0x2022a800,0x8202a8a,0x808a802a,0x22a020aa,0xa0a0a288,0x80a88228,0x2a202aa0,0x80820a80,0xa8808a2a,0x80080002,0x2282aaa2,0x2aa2a202,0x88a2008,0xa8a8a00a,0x88088aa0,0xa80a2282,0xa888aa88,0x2800a28a,0x8aaa2a8,0xa20a8a8a,0xa200aaa0,0x20080882,0x8882800,0x2222a228,0x80aa2a88,0x28a02a28,0x8820082,0x802a8a2a,0xa80a0a0a,0x8a020802,0x8228aa28,0xa88202a0,0x22880200,0x8a822022,0xa2aa8288,0x20288a8a,0x20a000aa,0x8a08080a,0x82802000,0x800820aa,0x80880008,0x8828220a,0x80a0088a,0x8a0aa080,0x28aa208,0x222a2020,0xa8200028,0x802aaa8a,0xa80208a8,0x2a80a28,0xa8a002a2,0xa0822880,0x228820a0,0xa800a2a0,0x8a02a808,0xa20802a2,0x2aaa8022,0x800082aa,0x808a8882,0x8aa2800,0x20aa0a22,0x2200a020,0x882822a,0x2202a,0xa0820022,0x80282a,0x2020a20,0xa000a08a,0x2a22808a,0x282aa8aa,0x80882a22,0x8022888,0xa80822a0,0x8aa28820,0x2080a22,0xa02a8);
remainders_pos[759].modulo_value=3036; remainders_pos[759].bit_mask=create_bit_mask(95,0x22000022,0x28228200,0xa080088,0xa08228a0,0x820a20,0x82200,0x220820a0,0xa20a08,0x8288,0x8a2082,0x28008,0x8208288,0x800a2020,0x28808000,0x80002280,0x82800002,0x808208,0x28820a,0x808a2802,0x8a20a20,0x208008a,0x80802880,0x28a08000,0x2088088,0x20000022,0x28220a00,0x82002200,0x80020002,0x200800,0x280200,0x2020082,0x28020008,0x8a08028a,0x20002000,0x220a28,0x8208008,0x8a2002,0x220228,0x2288200,0x80082002,0x808200,0x8220a008,0x20a0880,0x800,0x8200a000,0x2020,0x28008a08,0x88200082,0x20a2002,0x20808000,0xa008,0x20000,0x20220a08,0x200a088,0x20082020,0x8008200,0x82288,0x2880880,0x8008a20,0x2002082,0x828a0880,0x20008000,0x8a28020a,0x2000802,0x8200808,0x80280,0x20020080,0x28000800,0x80088008,0x800a0882,0x8800000,0x82202208,0x200020a2,0x228020,0xa200208,0x208a08a2,0x8028a20,0xa082280,0x22082020,0x28000028,0x80288000,0x20002022,0x808a00,0x2282082,0x2002800,0x8208a20,0x2282000,0x820a08a0,0xa08208,0x88200,0xa08a0820,0xa28820,0x8220020a,0x80082882,0x8800008);
remainders_neg[759].modulo_value=3036; remainders_neg[759].bit_mask=create_bit_mask(95,0xa20800a2,0x20220a08,0x82008000,0x20022020,0x8020228,0x8000a288,0xa2802820,0x8228200,0x88000200,0x22002,0x8820800,0x80280200,0x22820,0x20000808,0x808a200,0x2080882,0xa00,0x8200282,0x22002,0x228228,0x2008002,0x82000,0x20200008,0x8a000000,0xa08808a2,0x20a28200,0xa002288,0x8a0882,0x8a08008,0x88200208,0x82820802,0x20828800,0x2008202,0xa0882080,0x8228220,0x80280000,0x80022882,0x8a20a20,0x82200280,0x802882,0x8000a08,0xa202080,0x2820000,0x808008,0xa082088,0x800828a0,0x20000200,0x288002,0x82822882,0x20000808,0x88082080,0x808a0800,0x28a20200,0xa082000,0xa08028a0,0x800208,0x8800a200,0x82000000,0x800220,0x8a08200a,0x2020000,0x28800800,0x2208202,0x2800082,0xa08000,0x88008208,0xa08a0000,0x20008008,0x8000080,0x20002,0x808,0x228a280,0xa0802822,0x8a20828,0x82208280,0x20020022,0x820228,0x8200a208,0xa20020a0,0x20808820,0x8200080,0xa0882822,0x208,0x8a20a00a,0x82882000,0xa00228,0xa20a088,0x2820020,0x8200200,0x80000288,0x200200a0,0x8220020,0xa280282,0x802002,0x8800);
remainders_pos[760].modulo_value=760; remainders_pos[760].bit_mask=create_bit_mask(24,0xa8a00a8a,0x80028002,0x2082a0a8,0x202280a,0x20882808,0x20200202,0x22208a22,0x28082088,0xa020a2,0x8008020a,0xaa0282,0x8800,0x2200,0x28280aa,0xa08020,0x288a080a,0x88220820,0x888a208,0x8808008,0x80202822,0x8a02880,0x22a0a82,0x2a800280,0xa2a00a);
remainders_neg[760].modulo_value=760; remainders_neg[760].bit_mask=create_bit_mask(24,0x20200202,0x80a0882,0x280a2020,0x8a82a082,0xa000a000,0xa8a8028a,0xaa2802a2,0x2000a000,0x8820a82a,0x808a02,0x88220a02,0x88080080,0x8882288,0x8a020822,0x80280828,0xa0020082,0x2a80a0,0x2200,0x80000880,0xa0a02a,0x80282008,0xa228202,0x22088208,0x222882);
remainders_pos[761].modulo_value=761; remainders_pos[761].bit_mask=create_bit_mask(24,0x22bf0736,0xc06e7f7,0xfe37f51d,0x225a0a34,0xd7338b53,0x77fc0fb5,0xa64a518,0x6249b46,0x486591a5,0x51b5278f,0x8875e599,0xb51f5770,0x443babe2,0x2a669eb8,0x4bc792b6,0x82962698,0x418b6491,0xb8629499,0xaeb7c0ff,0x132b4733,0xfcb14169,0xc2e2bfb1,0x13bf9d80,0x1b383f5);
remainders_neg[761].modulo_value=3044; remainders_neg[761].bit_mask=create_bit_mask(95,0xaa228aaa,0x808a2a2a,0x22aa2880,0xaa8282a8,0xaaa028a,0xaa208228,0x82a82880,0x8aa8028a,0x80a80828,0x8828aa02,0xa82800,0x28828aa8,0x88a2222a,0xa2aa0220,0x820a0a2a,0xa0aaa00,0x8802a808,0x20aa0800,0x222a0822,0x8aa28aaa,0x202888a0,0xa2a2228,0x8a220008,0xa02a0a28,0x2288a282,0xa2808002,0xa020a222,0x2a882888,0xa222aa8a,0xa8222082,0x880082,0x828080aa,0x820802aa,0xaa2aa282,0x28882202,0x8a2a222,0xa800aa8a,0xa08800a0,0xa80a02a8,0xaa88a080,0xaa0a8a00,0x22a808a0,0xaa80a208,0xaaa8a02,0xaa20aa0a,0x880808a2,0x22aaa2a2,0x8002aa8a,0x2000220,0x8a8a882,0xa00a2002,0xa08000a0,0xa20a8002,0x8a80228,0xa080a00a,0x8a08800a,0x2a0a80a,0xaa88a08,0x800aa80a,0x220208a0,0xa2288822,0x82002002,0x2a8a282,0xa8a8280,0x8aa88aa0,0x8a222802,0x80022202,0x88288020,0x220a2a08,0xaa8a8202,0x82088222,0x20a02202,0xaaa22088,0x222020a8,0x8882a28,0x82008a,0x28a02220,0xa8a00a22,0xa802a88a,0xa000a0a8,0x20a0a28a,0x22a00200,0x2222088a,0x828820,0xa8280aa8,0xa008288a,0x28a80a82,0x82a80088,0x88280280,0x228a200a,0x82a000a8,0x2828000,0x88200228,0x20208a8a,0x822000);
remainders_pos[762].modulo_value=3048; remainders_pos[762].bit_mask=create_bit_mask(96,0x82800802,0x20808,0x8a20820a,0x2802880,0x220a28,0x2280280,0x200028a0,0x28808a28,0x8a28a20a,0x8a0082,0x20200a08,0xa288288,0xa0020022,0x8020828,0x2008200,0x220a2022,0x28220000,0x88202088,0x80820822,0x220a28,0x2082,0x820a00a2,0xa08,0x82080082,0xa28208a0,0x20808028,0x8a200280,0x20820822,0x8800020,0x8002082,0x20080822,0x20228a20,0x82202200,0xa08a2082,0x28000820,0x202282,0x22080080,0x8a08028,0x8280,0x80002000,0x28200008,0x228000a,0x228a00a0,0x800020,0x8800a,0xa2802082,0x808820,0xa00a08a,0x220200a2,0x8028a08,0x2200082,0x200a0,0xa28808,0x28800a,0x828a0,0x8000220,0x82000000,0x20a2002,0x208828,0x88080002,0x20002882,0x8a20a08,0x88088282,0xa2880800,0x20200808,0x8002088,0x22082,0x20820808,0x8008a288,0x2020802,0x20828a28,0x20820a,0xa0000082,0xa08220,0x800008a,0xa0880082,0x20820228,0x8082288,0x882822,0x8a08800,0x82008088,0x20802000,0x800a28,0x8228a088,0xa0080822,0xa20020,0x8a28a282,0xa20228a0,0x28000828,0x8028800a,0xa0880002,0x28028028,0x8208a202,0x208000a0,0x20028220,0x80);
remainders_neg[762].modulo_value=3048; remainders_neg[762].bit_mask=create_bit_mask(96,0x2080082,0x8828000,0x2280282,0x2082000,0x8a28220,0x8a208208,0xa0882020,0x20000220,0x2202282,0x80020802,0x28a08200,0x82200200,0x208a08a2,0x828020,0x8a080288,0xa28228a2,0x20a28808,0x28a000,0xa00a2,0xa28220,0x8808a00a,0x2820822,0x8808200,0xa00800a,0x220a0020,0x28000820,0x2288208,0xa00a00a2,0x8828,0x8008a00a,0xa08000a2,0x28a20228,0xa28a288,0x20022802,0x20808028,0x8828a20a,0xa2800800,0x200820,0x88080208,0x882880,0x20a08800,0x8a208082,0xa2020820,0x8008828,0x88000082,0x22082802,0x8000028,0x82082002,0xa28a0822,0x820200,0x8a28800a,0x80820820,0x8220000,0x88200082,0x80802020,0x808a28,0xa088088,0x82822882,0x8a00020,0x808a,0xa0882002,0x228200,0x20a,0x22000080,0x28a08000,0x8008a000,0x808a2802,0x28020000,0x8002200,0x828a0082,0x28020220,0x88280282,0x20880802,0x8200a28,0x80088002,0x20000802,0x28028a20,0x8000a200,0x800020a2,0x200008,0xa080000,0xa0082880,0x8008220,0xa202000,0x208000a2,0x8228828,0x220220a,0x228a2020,0x20808020,0x8200082,0x20000882,0x20820820,0xa00228a,0xa0080820,0x28820a28,0x8);
remainders_pos[763].modulo_value=3052; remainders_pos[763].bit_mask=create_bit_mask(96,0xaa822a0a,0x80282,0xa08a8828,0x8a220828,0x2808208,0xa280208,0xa0808222,0x220a2a2a,0x22882222,0x828a8a82,0x80280022,0x2a222a88,0x80a2280,0x20802220,0x80080a28,0x8002aa0,0xa2228a08,0x80022808,0xa80800a,0x822800,0x288020a0,0x8a808228,0x22a80282,0x2800a0a,0x80280aa,0x2020a80,0xa00a28,0x28a820aa,0x28aa00a,0x2a00aa22,0x8002808a,0xa2228008,0xaa20008a,0x8a002a0,0xa0a880aa,0x808a0802,0x2a200080,0x2080aa0,0x80a0002,0x88808882,0x8000080,0x8088822a,0xa828a20,0x2208a22,0x20208088,0xaa20a002,0x20a82,0x8020088,0x8000220,0xa08aa82a,0x20808800,0x20880220,0x282a088a,0x2220208a,0x20a88,0x20222020,0xa020822,0x2080800,0x8a80aa,0xa202020,0x22a0a802,0xa20aa0,0x8aa0a8,0x2888aa20,0x28002200,0xa00a8a20,0xaa28088a,0x82a28a00,0xa00aa0,0x808028a,0x2802002a,0x280aa0,0x2a88a0a,0x202a2828,0x80228288,0x882000a0,0x202a0002,0x88002a00,0x2888a202,0x8002020a,0x20020aa,0x8020828a,0x8a020088,0x888a8028,0x280222a,0x2a282880,0x8228882a,0x8a088888,0x2020aa8a,0x28a0888,0x20280208,0x88a2208,0x2a20a282,0x2000282,0x882aa828,0xa0a);
remainders_neg[763].modulo_value=3052; remainders_neg[763].bit_mask=create_bit_mask(96,0x220aa202,0x8808a02,0x200200a0,0x22a80a0,0x8a000a80,0x82a00a80,0x280802aa,0xaa82a222,0x2a00aaa2,0x202020a,0x820882a,0xa2a2a200,0x80822a08,0xa808a2a8,0x8808220,0x88a220,0x22aa0280,0x80aa080,0x82000882,0x880a2088,0xa008a028,0x2080a20,0x2a208a02,0x82088282,0x800a0822,0x8a828208,0x882802a0,0xa020a022,0x8a022802,0x228822a2,0x8a0802,0x2a2a0880,0x22a08802,0x80280a28,0x28200022,0x802800a,0x22a88800,0x82800228,0x8002888a,0xa,0x80880808,0x80002a2,0x820a0228,0xaa802a2,0xa0a80800,0x2228288a,0x8882820a,0x808a0800,0x808882a8,0x28022022,0x28080080,0xa0008aa0,0xa0228002,0xaaa0a802,0x888a0200,0xa8aaa0a8,0x828a802a,0xa808080,0x80020822,0x8228a8a8,0xaa20208a,0x882a0228,0x88022020,0xa0002228,0x2008aa80,0x208202a8,0x22208002,0xa220288,0x88280228,0x80808202,0xa08a8822,0x8a08220,0x82200282,0xa822a0a0,0x8a20a00,0xa80828,0xa8a2808a,0x88a208,0x20002282,0x8a8a82,0x8a08a822,0x8a00a02,0x28a0800,0x200a0,0x8a08aa22,0x22a0a000,0x2a000a2,0x2000000,0xa8a02202,0x8a020000,0xa8a08280,0x2aa00,0x22a82a02,0x82888a0a,0x2220a0,0x282);
remainders_pos[764].modulo_value=1528; remainders_pos[764].bit_mask=create_bit_mask(48,0x820a2aa2,0x8082a008,0x88aa20a2,0xaaa0a00a,0x8aa0882a,0x88a8a80,0x2a020222,0xa22802,0x2a82800,0x8288022,0x22a80a2a,0x800828a,0x8a282002,0xa8a02a8,0x8802282,0x282a8,0x880288a0,0x888080a,0x202a2a22,0xaa8220aa,0x2a00a0aa,0x28a08aa2,0x2200a820,0x28aa8a08,0x8820a2aa,0x28082a00,0xa88aa20a,0xaaaa0a00,0x8aa0882,0x2088a8a8,0x22a02022,0xa2280,0x202a8280,0xa0828802,0xa22a80a2,0x20800828,0x88a28200,0x20a8a02a,0x80880228,0x282a,0xa880288a,0x20888080,0xa202a2a2,0xaaa8220a,0x22a00a0a,0x28a08aa,0x82200a82,0x8aa8a0);
remainders_neg[764].modulo_value=1528; remainders_neg[764].bit_mask=create_bit_mask(48,0xa82a22a,0x80a2880,0x22a82a,0x22282882,0x22800a2,0x20208,0xa28a8aaa,0x882aa08a,0x8a20a088,0x80a008aa,0xaa2082a2,0x80880a02,0x2a0a88a,0x82028a20,0x8008aa0a,0x888a0a20,0x8a0028,0x80008082,0xa8a2a2aa,0x220aa822,0xa2882822,0xa028022a,0xaa8820a8,0xa0220280,0xa82a22,0xa080a288,0x20022a82,0x22228288,0x8022800a,0xa0002020,0xaa28a8aa,0x8882aa08,0xa8a20a08,0x280a008a,0x2aa2082a,0xa80880a0,0x2a0a88,0xa82028a2,0x8008aa0,0x8888a0a2,0x2008a002,0xa8000808,0x2a8a2a2a,0x2220aa82,0xaa288282,0x8a028022,0xaa8820a,0x22028);
remainders_pos[765].modulo_value=1020; remainders_pos[765].bit_mask=create_bit_mask(32,0x880082,0x8020020,0x2000200,0x2020822,0xa08020,0x8008228a,0x22,0x8008228,0xa208008,0xa2000000,0x28a20800,0x2008280,0x22082020,0x200020,0x2002008,0xa0800880,0x220020,0x2008008,0x80800080,0x808208,0x80282008,0xa00208a2,0x8,0x200208a,0x2882002,0x28800000,0xa288200,0x8020a0,0x8820808,0x80008,0x800802,0x8200220);
remainders_neg[765].modulo_value=1020; remainders_neg[765].bit_mask=create_bit_mask(32,0x80000802,0x28820,0xa088288,0x820200a2,0x200828,0x2202,0x800808a2,0x800220,0x82200080,0x22880800,0x20220000,0x8a080200,0x22802020,0x8a08828,0x8208a000,0x20080000,0x8220820,0xa080080,0x800,0xa00,0x8202080,0x200a0022,0x808000,0x8a00a002,0x2002802,0x20000808,0x82200280,0x82020,0x28000,0x80008000,0x80080082,0x208220);
remainders_pos[766].modulo_value=3064; remainders_pos[766].bit_mask=create_bit_mask(96,0xa0aa22a,0x222aa8a0,0x8022aa0a,0xa222a000,0xa2882a82,0x82a00a82,0x2a800aaa,0xa8a8a88a,0xa000a200,0x82820002,0xa028a8a8,0x2080aa82,0x2822080a,0x202028a,0x2888a828,0x8aa088a,0xa0020202,0x22208822,0x28208a28,0x2822802a,0x88a2aaa,0xa0a22a88,0x8a0222aa,0x22aa20a0,0x8a0a2002,0x222a0,0xa880220a,0x20002088,0x202a8228,0x28208a28,0x2288a22,0xa2a2a2aa,0x2088a008,0xa8280888,0x2082a2a2,0xa8a228,0x2800a8a,0x8808082a,0x2aaa282,0xaa20aaa,0x20880808,0x8000aa80,0x280aa02,0x2802882,0x8aaa0222,0x20a0022a,0xaa080022,0x2200a0,0xaa0a0088,0x8880020,0xa2a8800a,0x808880aa,0x80822802,0x280aa02,0x802aa00,0xa0202022,0x80aaa08a,0x22828aaa,0x80a82020,0xa2a002,0x8288a2a,0x2a8a8a82,0x8222028,0x8a200a22,0x80aa8a8a,0x2888a228,0x828a208,0x82882a8,0x2a220800,0xa08802,0xa20a8880,0x888008a0,0xa20a08aa,0xaaa8880,0x2022a88a,0x28aaa8a2,0x28a80288,0x8828a208,0xa882208,0x20808080,0x28a220aa,0x80282a22,0x28a28080,0x8a02088,0xa82aa02,0x822a2a28,0xa800082,0x2a008a00,0xa8a22a2a,0x82aaa002,0x8a82a00a,0x8a82a822,0x2000a88,0x88a0aa88,0xa00a2aa8,0xa88aa0);
remainders_neg[766].modulo_value=3064; remainders_neg[766].bit_mask=create_bit_mask(96,0x82822aa2,0xaaa22028,0x8aa2282,0x2aaa2888,0x2a00a20a,0xa28820a,0xa2088222,0x20202002,0x28882a88,0xa0a888a,0x28a02020,0x2808220a,0xa0aa8082,0x8a8a8a02,0xa00020a0,0x80228002,0x288a8a8a,0xaaa800aa,0xa0a802a0,0xa0aa08a2,0x8002a222,0x282aa200,0x28aaa22,0xaa22a828,0x282a88a,0x888aaa28,0x2008aa82,0xa888a800,0xa8a20aa0,0xa0a802a0,0x8aa002aa,0x2a2a2a22,0xa8002880,0x20a08000,0xa80a2a2a,0x88202aa0,0x8a088202,0x8080a2,0x8a222a0a,0x822a8222,0xa8008080,0x8882208,0x8a08228a,0x8a08a00a,0x2228aaa,0xa8288aa2,0x228088aa,0x88aa8828,0x22828800,0x800088a8,0x2a200882,0x8000822,0x80aa08a,0x8a08228a,0x808a2288,0x28a8a8aa,0x8222802,0xaa0a0222,0x820a8a8,0x802a288a,0x80a002a2,0xa202020a,0x80aaa8a0,0x2a882aa,0x8220202,0xa0002aa0,0x80a02a80,0x80a00a20,0xa2aa8088,0x8828008a,0x2a820008,0x88028,0x2a828022,0x82220008,0xa8aa2002,0xa022202a,0xa0208a00,0xa02a80,0x8200aa80,0xa8080808,0xa02aa822,0x8a0a2aa,0xa02a0808,0x8028a800,0x820a228a,0xaa2a2a0,0x8208880a,0xa2880288,0x202aa2a2,0xa22288a,0x20a2882,0x20a20aa,0x8a888200,0x282200,0x2882a220,0x200228);
remainders_pos[767].modulo_value=3068; remainders_pos[767].bit_mask=create_bit_mask(96,0x228a8282,0x80a20822,0x22020280,0xa802282,0xa2020a8,0x80a8888a,0x8008aa00,0x8200a8a2,0x2820802a,0xaa202,0x2aa8a80a,0xa8a88282,0x8a0228aa,0xa2a022a,0xa8880228,0xa28aa080,0x208202a0,0x88aa0880,0x208008,0x88aa8820,0x880a282a,0x888022a0,0x8a2a8000,0x8aa8082,0x82a0022,0x2aa20082,0x28088a22,0x808222,0xa0a20020,0x228a28aa,0xaa220022,0x280a8028,0xa280200,0xa00aa220,0x888082,0x208280,0xa0a2000,0x80a02000,0x8a0002aa,0xaa08a0,0x8202a0a,0x228022a8,0x8202002a,0xa0282882,0x8a08a802,0xa0a8a80,0x88228822,0xa080208,0x2208020a,0x8822882,0x202a2a0a,0xa802a20a,0x28228280,0x8a800808,0x2a88028,0xa0a8082,0x20a20aa0,0x2aa8000a,0x80a0,0x8a0a,0x282080,0xa8202220,0x88aa00,0x8008028a,0xa2802a02,0x8880088a,0xaaa28a28,0x8008a0,0x88882020,0x888a2202,0x82008aa,0x8800a82,0x28202aa2,0x20002a8a,0x20a88022,0x2a828a02,0x822aa2,0x22002080,0x80220aa2,0xa0a80820,0xa020aa28,0x2880222,0x2a880a8a,0xaaa2880a,0x882822a2,0xa02a2aa,0x8808aa00,0x2a802082,0x28a2a008,0x200aa200,0xa2222a0,0x2a0808a,0x8828802a,0x20280808,0x888208a0,0x8282a28);
remainders_neg[767].modulo_value=3068; remainders_neg[767].bit_mask=create_bit_mask(96,0xaa020a0a,0x2a802a,0xa28a8a08,0x8208aa0a,0x82a82820,0x8200002,0x8802280,0xa08202a,0xa0a808a2,0x8882220a,0x22202082,0x20200a0a,0x282a022,0x82a28aa2,0x20008a20,0x22022808,0xa8028a28,0x220008,0x88a80880,0x2200a0,0x2a0a2,0x8aa28,0x2a20088,0x22080a,0x80a288aa,0xa222880a,0xa08002aa,0x88080a2a,0x202a88a8,0xaa02a022,0x22aa08aa,0xa08200a0,0x82a08a80,0x28022aa8,0x8800080a,0x88a80208,0x282a888,0x828a888,0x2008a22,0x88228028,0x80a8a202,0xa208aa20,0xa8a88a2,0x28a0200a,0x280208a,0x82820200,0x2a00a2,0x82808a80,0xaa808282,0xaa00a,0xa8a2a282,0x20822a82,0xa0aa0a08,0x2080000,0x822008a0,0x8282080a,0xa82a0228,0xa2208882,0x88880820,0x88080282,0x80a0a808,0x20a8a2a8,0x8002288,0x8808a02,0x2a00a28a,0x88002,0x222a0220,0x80088028,0xa828,0x22a8a,0x80a88022,0x80088202,0xa028a22a,0xa888a202,0xa82000aa,0x2202028a,0x880aa22a,0xaa80a808,0x8aa822a,0x28208028,0x20a822a0,0x8a008aaa,0x22000202,0x222a0082,0xa0aa22,0x820a2a22,0x802288,0xa208a00a,0x202a2880,0xa8822a88,0x82a2a228,0x8a280802,0xa00822,0xa0a08080,0xa8028,0xa022a0);
remainders_pos[768].modulo_value=24; remainders_pos[768].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[768].modulo_value=24; remainders_neg[768].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[769].modulo_value=769; remainders_pos[769].bit_mask=create_bit_mask(25,0xcb1f977e,0x524fab55,0xee673911,0x3b061055,0x2f69a18b,0xd4743e17,0x818839b3,0xad6d08b6,0x6ca368cf,0xa6dd1c43,0xd5e011d,0xd39a1f90,0x27e1672f,0xe201eac0,0x8e2ed96,0xcc5b14db,0xb442dad7,0x36704605,0xa1f0b8af,0x46165bd3,0xa8218373,0x227399de,0xab57c92a,0xfba7e34e,0x1);
remainders_neg[769].modulo_value=3076; remainders_neg[769].bit_mask=create_bit_mask(97,0x2820aa2,0x8a822288,0x22aaa088,0xa28a8888,0xa2a02802,0x8a8a28a,0x800a02a,0x20a0802a,0xa02aa002,0x2a00808a,0x80828880,0xa028208,0xaa28aaa2,0x2a882208,0x802a200a,0x828802,0x288a020a,0xaaa88a88,0x28282022,0x8a8a820a,0x20a80aaa,0xaaaa0002,0x228a00a2,0x222a2a82,0xaa2a220,0x28028a22,0x2aaaa,0xaa80a822,0x820a8a8a,0x2020a0a2,0x8a88aaaa,0x820288a0,0x8a0802,0x8022a00a,0x82208aa2,0x2aa8a2a8,0x820a0282,0x88a0808,0x880802a0,0x2aa02a,0xa0082822,0xa0280082,0x8a28a882,0xa02a2a,0x888a8a2a,0x882aaa20,0x8a220a88,0x2a820a00,0xa0a28022,0x80a2208a,0x20002888,0x2080888a,0x202a0aa0,0x88082080,0x8aa82a02,0x2828aa02,0x2a002aa2,0x2a8a880,0xa8a088aa,0x82a0a288,0x2080022,0x8a2288,0xaa022a82,0xa8a08aa0,0x882a282,0x8808a,0xa0a2a20,0x8080a282,0x280a8000,0x2aaa2,0x2082a820,0x220200a2,0x80202222,0xaa08222,0xaaa00002,0xa80a22,0xa2808080,0x2a282820,0x80880002,0xa2a08828,0xa8828aa0,0xaa202a82,0xa2288020,0x82008,0xa282a0a2,0x88828a88,0x8a8aa02a,0xaa002a00,0x2a8a0a22,0x2a0aa8a0,0x820808a0,0xa82a0200,0x88808202,0x8a000228,0x82228088,0xa282a8,0x2);
remainders_pos[770].modulo_value=3080; remainders_pos[770].bit_mask=create_bit_mask(97,0x20002202,0xa288220,0x200a00a8,0x80202082,0x200002,0x80a02208,0x88080002,0x8022228,0x800a8a2,0x22088022,0xa200a,0x82028202,0xa020280,0x8a2008a,0xa0800a22,0x20200802,0x222a8802,0x8080000,0xaa0a002,0x80802808,0xa82082,0x200a0822,0x2a802200,0x822088a8,0x80002,0xa0220000,0x8200a20,0x80280,0x88008802,0x2080020,0xa0a00882,0x28000022,0x28822200,0xa20000,0x80200020,0x80820000,0x20008282,0x220002a0,0x200a000,0xa802,0x222800,0xa8208080,0x220a0000,0xa8028a0,0x82280202,0x8202800,0x22000080,0xa2220,0x88a00082,0x8808,0x28082002,0x80288200,0x2802a080,0xa0880a,0x2082a00,0x28880002,0x2a000000,0xa008080,0x80008800,0x8200080a,0x820282,0x80200,0x8a0008,0x88822800,0x2800,0x200a0a88,0x208082,0x22002208,0x80200080,0xa0082002,0x880a08,0x200000,0x22088280,0x8802a82a,0x20a00800,0x82a0088,0x28020282,0xa0aa020,0x202080,0x22a88800,0x20080880,0xa0020a80,0x8a2088,0x8080a0a2,0x82808202,0x8a00080,0x22088a0,0x2a002088,0x8880208a,0x202228,0x880a0280,0x80020,0x8080280,0xa00882,0x8228a02a,0x88000808,0x80);
remainders_neg[770].modulo_value=3080; remainders_neg[770].bit_mask=create_bit_mask(96,0xa888220a,0x2200aa0,0x20828020,0x820a802,0x80a0080a,0x282a80,0x8082,0x8aa220,0x802022,0x2280082a,0x88022882,0xa020282,0x20a0208,0x80228002,0x20088222,0x28a80082,0xa2220082,0x8088,0x8220280a,0x2000,0x820200a,0xa8028022,0x2208a280,0x2200020,0x8000880a,0x28a20880,0x802802a0,0x808208,0x80000a,0x2808820,0x2020000a,0x200888a2,0x2002a288,0x88220088,0x288028,0x80a0808,0x28080a02,0x22880228,0x82002888,0x8080200a,0x88aa2008,0x20200000,0x2a828008,0x200a020,0x2a0820a,0x80282088,0xaa008800,0x88222a0,0x20000a,0x80880000,0x20802882,0x200a00,0x20022008,0x8200082,0x82802208,0xa000008a,0xa2088800,0x2880000,0x800080,0x2080002,0x80020a0a,0x88000a80,0x8028000,0x2a008,0x88a000,0xa0820200,0x88280002,0x2a00a280,0x8a00808,0x2080a002,0x88000200,0x288880,0x22800208,0x22022,0xa0208008,0xa20800,0xa00a8202,0x2822028,0x820a800,0xa2200008,0xa8000000,0x28020200,0x22000,0x800202a,0x2080a0a,0x200800,0x2280020,0x2208a000,0xa002,0x80282a20,0x820208,0x88008820,0x800200,0x8020800a,0x2202822,0x800000);
remainders_pos[771].modulo_value=3084; remainders_pos[771].bit_mask=create_bit_mask(97,0x28828a2,0x28a20a08,0x88002202,0xa20800a0,0x20a08020,0x820a,0x822002,0x20a28208,0x8208a200,0x22882,0x20800820,0x28a202,0x20002022,0xa20808,0x8a00a28a,0xa2802880,0x8228208,0x8208a008,0xa0000880,0x28220a20,0x82000282,0xa0822,0x28828,0x82088200,0x82020802,0x20028a28,0x8200a280,0x28a0800,0x808008,0x228a008,0xa0020822,0x8008220,0x280002,0xa0022,0x20800a00,0x8828a082,0x80002880,0x20a20808,0x8a288,0x80020802,0x8000808,0x88082282,0x820a2882,0x28200800,0xa208008,0x808a2020,0x800008,0x200282,0x800000,0x200828,0x8a202200,0x208a0080,0x828200,0x8a082002,0x82022822,0x20828,0x8002202,0xa2000802,0x8a08228,0x80002202,0xa2822022,0x208820,0xa00000a,0x2800880,0x20020800,0x800a088,0xa2880882,0x20200200,0xa280200,0xa0082002,0x28008028,0x808228a,0x22082802,0x28000008,0xa000282,0x82882,0x8828828,0xa08a,0xa2082022,0x28820200,0x8028a208,0xa00a2022,0x8a00a28,0x80008202,0xa2800880,0x208808,0x88000082,0xa2082822,0x28a08008,0x88200208,0x20000800,0x20a08a08,0x208a080,0x800220a0,0x8a28808,0x8228020a,0x8a2);
remainders_neg[771].modulo_value=3084; remainders_neg[771].bit_mask=create_bit_mask(97,0x82002022,0x20228200,0x8a28a,0x22800820,0x28200828,0x88080282,0x800a2882,0x28220a00,0xa002288,0x808a2002,0x28008028,0x8820228a,0xa08828a2,0x8228000,0x2082202,0x22082000,0xa20a00,0xa002080,0x20880000,0x20a28228,0xa08820a,0x808200a2,0x8820020,0xa000288,0x28a0082,0x28820220,0xa082208,0x82020080,0x8000800,0x8a202080,0x208a00a2,0x800a28,0x8820808a,0x808208a2,0x28008208,0x20200a,0x882000,0x28228000,0x88002200,0x8a0082,0x808000,0xa20a,0x2822002,0x20a08008,0x82280080,0x228a0,0x8008800,0x8828820a,0x80080880,0x8a08020,0x228a288,0xa0020800,0x8020a08,0x200a08a,0x28a20a2,0x8828020,0x8008a20a,0x22880082,0x200a20,0x808a28a,0x220a28a2,0x8a00028,0x82088082,0x82080000,0x28828008,0x80082000,0x22000002,0x28a08a08,0x82208288,0x20802882,0x20800820,0x8000a202,0xa2802082,0x20808800,0x8208820a,0x80802002,0x20020,0x88082002,0x228028a2,0x20028a08,0x8202280,0x208228a2,0x208220,0x808028a,0x22080000,0x8a00000,0x8800a,0x228020a2,0x20200000,0x288280,0xa0880080,0x28200200,0x8a002008,0x8a2820,0x220000,0xa208282,0x22);
remainders_pos[772].modulo_value=1544; remainders_pos[772].bit_mask=create_bit_mask(49,0x8aa0028a,0x88820800,0xa02a082a,0x2002822,0x208a8a0a,0xaa2a8222,0x220aa2a8,0x828a8822,0x20a00202,0xa082a02a,0x82088a,0x8a002a88,0x2a800a2a,0x22082002,0x80a820aa,0x800a08a,0x822a2828,0xa8aa0888,0x882a8aa2,0xa2a2088,0x8280080a,0x820a80a8,0x208222a,0x2800aa20,0xaa0028aa,0x88208008,0x2a082a8,0x2002822a,0x8a8a0a0,0xa2a82222,0x20aa2a8a,0x28a88222,0xa002028,0x82a02a2,0x82088aa,0xa002a880,0xa800a2a8,0x20820022,0xa820aa2,0x800a08a8,0x22a28280,0x8aa08888,0x82a8aa2a,0xa2a20888,0x280080a0,0x20a80a88,0x208222a8,0x800aa200,0xa2);
remainders_neg[772].modulo_value=1544; remainders_neg[772].bit_mask=create_bit_mask(49,0x2288a02,0xa8088,0x28a280a2,0x8a88a0aa,0xa8020282,0x22a20aaa,0xaa822a20,0xa0200aa,0xa8288a8a,0x280a28a2,0x880a8002,0x288a200,0xa20882a2,0xaa80a88a,0x820a822,0x80882802,0xaa2a0a0,0x20228000,0xa20222,0x82a2a800,0xa088082,0xa820820,0x8a80aaa2,0xa08822a8,0x2288a022,0xa80880,0x8a280a20,0xa88a0aa2,0x80202828,0x2a20aaaa,0xa822a202,0xa0200aaa,0x8288a8a0,0x80a28a2a,0x80a80022,0x288a2008,0x20882a20,0xa80a88aa,0x820a822a,0x8828020,0xaa2a0a08,0x2280000,0xa202222,0x2a2a8000,0xa0880828,0xa8208200,0xa80aaa20,0x8822a88,0x2a);
remainders_pos[773].modulo_value=773; remainders_pos[773].bit_mask=create_bit_mask(25,0x723ce52,0x7742d1f8,0xacde803f,0xca4a45db,0xbff75aa2,0xb92f262b,0x4fe42c32,0x8d93479b,0xce193b53,0x1a00a8c2,0xf1417992,0x92bc73a8,0x738f5273,0x67a0a3c5,0xc5401612,0xb7261cd0,0x78b26c72,0xd09fcb6,0x193d2753,0x56bbff75,0xe89494d1,0x5ecd76,0xe2d0bbbf,0x9cf13807,0x12);
remainders_neg[773].modulo_value=3092; remainders_neg[773].bit_mask=create_bit_mask(97,0x8aaa028a,0xaa8a0820,0x200208a2,0x2828802,0x222a822a,0x20a2aaa2,0x8228a0aa,0xa8a02,0x280a28a,0x8288200a,0x2888a00a,0xa20aa20,0xaa028aaa,0xaa282a08,0x8888a8a,0x2aaa8008,0xa02aa0aa,0x8080202a,0x80a0aa8a,0x8a2222a8,0x20080808,0x888200aa,0x2a082222,0x28a08a,0x28a0028a,0x2082a288,0x2088a22,0x808022a8,0x22228880,0xa82808aa,0x20080a8a,0x2aa02aa0,0xaaaa2a8,0x88888080,0xa0a2aa8a,0x8a02a882,0xa82282aa,0x2888a022,0x208a0a80,0x280a0280,0x8a80028a,0x28a20a02,0xaa2822a8,0xaa2222a,0x8a0a02a2,0x82002200,0x828aaa28,0x2aa8820,0x2a0828a,0x828a2880,0xa2882000,0xa08aa22a,0xa202a0,0x20002222,0x8280228,0x8082a0a2,0xa82082aa,0x8a2a82a0,0x882a80a0,0x28022a08,0xa0800282,0xa028802,0x88808002,0xaa8a88,0x280200,0xaa2a002a,0x280080a8,0x222008a8,0x8a8a8882,0xa2a8022a,0x8a222088,0x8288202,0x2aa002aa,0xa0208a08,0x88822228,0xaa200aa2,0x2088a8a8,0xa880222,0x8a80800a,0x2a002a2a,0x200a00,0x88a8aa80,0x20008088,0xa008a028,0x20a00082,0x882a200a,0x8280aa08,0x82a0aa28,0xaaa0820a,0x2282a080,0xa200a08,0x22220002,0x82a02280,0xaa22a882,0x800208a2,0x8a28a0,0x82a0);
remainders_pos[774].modulo_value=1032; remainders_pos[774].bit_mask=create_bit_mask(33,0x220208a2,0x28820228,0x282088,0x22800802,0x20a20800,0x8a008200,0x80002882,0x8a00000,0x8a280088,0x800a0882,0x20828000,0x8820020a,0xa08a0022,0x8200a20,0x208008,0x822020,0x8820000,0x2080008,0xa0082020,0xa20a08,0x80082288,0x28208a0,0x20a00200,0x28a282,0xa2022,0x28000200,0x8200a282,0x208a0800,0x20028800,0x8280080,0x80822822,0x20808828,0x8a);
remainders_neg[774].modulo_value=1032; remainders_neg[774].bit_mask=create_bit_mask(33,0xa28a0022,0x20028220,0x8820a000,0xa2080082,0x28228008,0x2080288,0x82002,0x208808,0x2208000,0x820002,0x28020808,0x288282,0x200208a2,0xa08228,0x88280080,0x800a28a0,0x28808,0x8a000080,0x208028a0,0x8228200,0x800a200,0x820a0020,0x20208a08,0x8820220a,0x808228a2,0x20808a08,0xa08220a,0xa0020080,0x28820008,0x80208008,0xa20a2,0x28000020,0x2);
remainders_pos[775].modulo_value=620; remainders_pos[775].bit_mask=create_bit_mask(20,0x8800a0a,0x20a02,0x80a8020,0x8a82202a,0x20a80822,0x82280,0xa0008882,0x2a88a200,0x202808a,0x2000822a,0x28080a88,0xa22a8a20,0x2000a008,0x82000822,0x2a08028,0x882a2882,0x2a020a80,0x8000080,0x22080a,0xa0a);
remainders_neg[775].modulo_value=620; remainders_neg[775].bit_mask=create_bit_mask(20,0x80282,0x880a8282,0x8200a8,0x202a8a2,0xa020802a,0x88802a08,0x28080002,0x22002288,0x8a020802,0xa0880a22,0xa0800200,0x2a2202a0,0x28802080,0xa8080aa,0x82280820,0xa2200a,0xa20a0200,0x888008,0x88a28082,0x202);
remainders_pos[776].modulo_value=776; remainders_pos[776].bit_mask=create_bit_mask(25,0xa2282222,0x80aa022,0x820aa22a,0x2aa20a80,0xa8a2080a,0xaaa82aa2,0x280002,0xa8a20802,0x80a20020,0x2200a282,0x2a00a8a0,0x22282202,0x88288882,0x2a00a880,0x8a00880a,0x8a0282,0x208a2a08,0x280080,0xa82aaa80,0x208a2a8a,0xa08aa8a0,0x8aa08202,0xaa020a8,0x88288a88,0x88);
remainders_neg[776].modulo_value=776; remainders_neg[776].bit_mask=create_bit_mask(24,0x2aa0aaaa,0x808228aa,0xa822aa2,0xa22a8208,0x202a8082,0x2220a22a,0x88a0888a,0x202a808a,0x82a88a8,0xaa882a02,0xa2882028,0xaaa0aa8a,0xa0000a,0xa2882008,0x2880082,0x88028a0a,0xa802a280,0x88a08808,0x20a22208,0xa802a202,0x28022028,0x2280a0a,0x8228a820,0xa00200);
remainders_pos[777].modulo_value=3108; remainders_pos[777].bit_mask=create_bit_mask(98,0xa2882002,0x20808200,0x280008,0x820200a2,0x800000,0x88082200,0x880080,0x828000,0xa002200,0x20008a2,0x8020020,0xa202208,0x20a0820,0x20a08008,0xa000002,0x20000000,0xa20a00,0x8a002082,0x80022082,0x20028,0x820a208,0x20802822,0x8a08a00,0x2008208,0x20228a2,0x28808208,0x8200288,0x208220a0,0x8228,0x8a002,0x28a0822,0x28020820,0x200282,0x2800000,0x28220000,0x82008008,0x22802082,0x808220,0x2002002,0x2822880,0x8000220,0x8800000a,0x80880800,0x8000220,0x2080000,0xa0022802,0x8208000,0x8a280208,0x800a0020,0x8028a20,0x208202,0x28800a0,0x20808,0x88000200,0x2022020,0x220,0x8800020a,0x22882800,0x800800,0x88202008,0x20802880,0x200828,0x88282,0x2800,0x28008000,0x82080288,0x820a2800,0xa20008,0x88200008,0xa0882082,0x28008200,0x8202282,0xa2880802,0x8200808,0xa20a202,0x82802080,0x8a08208,0x80080002,0x20880022,0x20800a28,0xa08a008,0x80,0xa00,0x20a088,0x820a0802,0x8808a00,0x80080202,0xa2000800,0x8800a08,0x282000,0x20022000,0x8820220,0x2000,0xa0080820,0x28008,0x8202082,0x8228a0,0x8);
remainders_neg[777].modulo_value=3108; remainders_neg[777].bit_mask=create_bit_mask(96,0x22002802,0x28000a00,0x80208080,0x20a0822,0x8008808,0x2288,0x80000000,0x8020808,0x208a280,0x82880022,0x28828,0x8220a280,0x820200a0,0x28200000,0x8208800a,0x20880880,0x228208,0x208a00a,0x822802,0x8820820,0x80282280,0x20082022,0x200200,0x82080280,0x820a2022,0x20000a00,0x80200200,0xa00a2020,0x8800a20,0x8002082,0x820200a2,0x20028028,0x8820820a,0x82080880,0x20a28008,0x2080000,0x22082802,0xa28,0x8a08a08a,0x82022000,0x800a28,0x88082,0x808a20,0x82008088,0x200a2082,0xa00808,0x2200280,0x8200a0,0x20228,0x8280282,0x82000820,0x8028000,0x8288,0x828a20a0,0x8808228,0x88202,0x22002080,0x8008,0x28a080,0xa0002000,0x8a00020,0x8800020a,0x80802000,0x20800800,0x2008200,0x2022080,0x8228800,0x280000,0x20002002,0x20800a08,0x28a202,0x22000082,0x208000,0x8220228a,0x2082800,0x200200,0x800000a,0x200008a2,0x20008220,0x82002080,0x80800800,0x8800200,0x88282000,0x2820002,0x200,0x828a,0x22080080,0x8200,0x88202088,0xa08a2080,0x28228,0x808a080,0x208000a0,0x8020800,0x20a00a,0x800a2020);
remainders_pos[778].modulo_value=3112; remainders_pos[778].bit_mask=create_bit_mask(98,0x2a22028a,0xa0aa0a20,0x808a8a02,0x8a8aa822,0x8a288a0a,0x20808228,0x8802082a,0x2a88a82a,0xa0000028,0x2288a88a,0x22aaa082,0xa80a0aaa,0xa08220a,0xa222aa22,0x2002aa80,0x8888a200,0x20022828,0xa880228,0x8080aa28,0x2a002a,0x2202020a,0x20082a8,0xa20a28a,0x8aaa28aa,0x82a202,0x20a00820,0xa2a0820a,0xa228028a,0x2aa0a2a2,0x8a802aa0,0x80a8200a,0xaa2822a8,0x88882822,0xaa2aa208,0x220a8002,0xa0a22002,0xa8002228,0x22000a0,0x82228a00,0xaa008808,0x80282aaa,0xa8802808,0x8a2028a2,0x2088228a,0x8080a088,0x8a822800,0xa02a080,0x202a20a0,0x820082a2,0x8a8088a,0xa80a00a,0x2882a202,0xa020200,0x88220822,0x2808a2a2,0x28022a8a,0xa8280220,0x2200aaaa,0xa2888220,0x88000,0x88002a0a,0x88a0a28,0x2a08880,0x8aa8aa80,0x28222220,0x8828aa88,0x82a022a,0xa802a2a0,0x8a0aa80a,0x80288a8a,0x820a8aa2,0x200a08a0,0x8a820008,0x28aaa280,0x8a08a0aa,0x820080a2,0x8080882a,0xa800a0,0xaa0202a8,0x8022a028,0x28800828,0x8a222228,0xaa800800,0xaa888a02,0x8820a088,0xa0a02aa0,0xaaa88aa,0x2a228882,0xaa2,0x2a22a828,0x208022a8,0x820208a8,0xa228a228,0x2aa2a2a0,0xa2a20288,0xa0aa0a80,0x8088a808,0xa2);
remainders_neg[778].modulo_value=3112; remainders_neg[778].bit_mask=create_bit_mask(98,0xa2aa8a02,0x282282a8,0x802028a,0x20220aa,0x2a00282,0xa8080aa0,0x8a80a2,0xa20020a2,0x288888a0,0xaa002002,0xaa22280a,0x20828222,0x8280aa82,0x2aaa22aa,0xa88a2208,0x2a88,0xa88aa0a0,0x82008aa0,0x80822a0,0x88a288a2,0xaa8a8a82,0x8a880a20,0x82a82a02,0x222a022,0x880a2a8a,0xa82880a8,0x2a280a82,0x2aa08a02,0xa2282a2a,0x208a228,0x820a882,0x22a0aa20,0xa0aa,0x22a22a80,0xaa82088a,0x282aa88a,0x20882aa0,0x8aa88828,0xaaa0288,0x22880080,0x8a0a222,0x2008a080,0x2a8a02a,0xa800aa02,0x8082800,0x20aa088,0x828a2808,0xa8a2a828,0xa880a2a,0x80208002,0x82082882,0xa00a2a8a,0x828a8a88,0xaa80aa,0xa0802a2a,0xa08aa202,0x20a08aa8,0xaa882222,0x2a000aa8,0x88800888,0x88a282,0x800282a0,0x8a280008,0x2202208,0xa0aaaaa8,0xa02200,0x80a28aa2,0x208a2a28,0x2822082,0x8a00202,0xa82022a,0xa8828028,0x20a8880,0xa0222a08,0x2802822,0xa88082a,0x80800a2,0x88208828,0x228a8a20,0x8aa28a0,0xa00880a0,0x2aaaaa0,0x22088088,0x2200028a,0xa82800,0x2828a220,0x82220022,0xa2aa000a,0x8888822a,0xa2aa20a0,0xa808aa20,0xa8a8020,0x2aa02aa0,0xa22a2a28,0x2a2a8a00,0x28228208,0x8002080,0x2a);
remainders_pos[779].modulo_value=3116; remainders_pos[779].bit_mask=create_bit_mask(98,0xa2002aa2,0xa8aaa008,0xa022222,0xa220a88a,0x8028002,0x80208200,0x82200aa2,0x8220a802,0x2a802a2,0xa88a80a2,0x2aa80220,0xa082022,0x282802a2,0xa20088a2,0xa080a8a0,0xa0a882a8,0x2a8a0aa8,0x8820822,0xa2800a80,0x2a22a228,0x8082022a,0x282a8282,0x2aaa8800,0x8a822080,0x20802a2,0xa8288a,0x802aa200,0xa0a0a00,0x80808a8a,0x8200800,0x22aa00aa,0x88228a,0x28000280,0x20200a20,0x8820202a,0xa0882228,0x82880a20,0x888882,0x20208aa0,0x8082a02,0xa080aa0,0x8a0aa028,0xaa080088,0xa2aa8228,0x8008aa0a,0x82022028,0x8888a,0x88228800,0x20800a2,0x288228a0,0x22000002,0x88082a22,0xa2002280,0x2aa8aa0a,0x20aa288,0xaa0a2220,0x20a0280,0x820200aa,0x2080880a,0x222000aa,0x2282822,0x8220a08a,0x80822288,0x808a80,0x2808a,0x82200028,0xaa88a28,0x820aa0,0x20202002,0xa0a0a2a,0xaa80200a,0x82a00008,0x2080a22,0x882a28a8,0x2aaa8020,0x2a828002,0x8202828,0xa88a8a88,0x28a288,0x822002a,0xa2a8882,0x22a0a2aa,0xa020a2a8,0x2008a0a2,0x28288a2,0x820a08a8,0x2aa8880,0x2a22a088,0x2a808a0,0xa08828a8,0x882802,0x208028aa,0x28020008,0xa088a800,0x880a0a22,0xaaa2a888,0x8008a200,0x8aa);
remainders_neg[779].modulo_value=3116; remainders_neg[779].bit_mask=create_bit_mask(98,0x2a80a22a,0x20222880,0x28aaaaa,0x22a82002,0x808a088a,0x8a80288,0xaa8822a,0xa28208a,0xa208a2a,0x2002082a,0xa2208aa0,0x8280a8aa,0xa0a00a2a,0x2a88002a,0x20082020,0x28200a20,0xa2028220,0x800a802a,0x2a088208,0xa2a22a20,0x80a8aa2,0x20a20a0a,0xa2220088,0x20aa808,0x8a80822a,0x8820a002,0x8222a88,0x82828288,0x8080202,0x80a80080,0xaa228822,0x88002a02,0xa0888a08,0xa0a882a8,0xa0a8a2,0x2800aaa0,0xa008228,0x8800000a,0xa8a00228,0x8000a28a,0x2808228,0x28228a0,0x22808800,0x2a2202a0,0x802282,0xa0aa8a0,0x88800002,0xaa0088,0x8a808822,0x200aa028,0xaa88088a,0x80a2aa,0x2288aa08,0xa2202282,0x8a822a00,0x2282aa20,0x8a828a08,0xa828822,0xa8080082,0x2aa88822,0x8aa0a02a,0xaa82802,0x80aa200,0x88080208,0x880a0802,0xaa880a0,0x822002a0,0x880a8220,0xa8a8a88a,0x828202a2,0x22082882,0x2288880,0x8a8082aa,0xa2a020,0xa2220828,0xa202088a,0x80a0a0a0,0x20020200,0x8a02a00,0x80aa88a2,0x8222000a,0xaa282222,0x28a82a20,0xa800282a,0x8a0a002a,0x2828020,0x8a220000,0xa2aa2800,0x8a200028,0x2800a020,0xa08a,0xa808a022,0xa08a8880,0x28002008,0x8282aa,0x22222000,0x8802a80,0x22);
remainders_pos[780].modulo_value=1560; remainders_pos[780].bit_mask=create_bit_mask(49,0x800a0002,0x20228a20,0x2082200,0x82020082,0xa00008,0x80282082,0x802,0x28000208,0xa008002,0x80800002,0x808,0x88008200,0x20022000,0x8,0x202202,0x200a0800,0x28800,0x208,0x82802802,0xa00820,0x8082200,0x82080820,0x28808008,0xa00208a,0xa0002800,0x28a20800,0x82200202,0x20082020,0x8820,0x8208200a,0x802802,0x208000,0x8002280,0x20a0,0x808808,0x8200000,0x22000880,0x8200,0x2202000,0xa0800002,0x28800200,0x208000,0x2802000,0x820828,0x8220000a,0x80820080,0x8008820,0x208a288,0x8000a0);
remainders_neg[780].modulo_value=1560; remainders_neg[780].bit_mask=create_bit_mask(48,0x820882,0x28220220,0x200a288,0x2820802,0x200800,0x820200a,0x80080082,0x20008200,0x8200008a,0x80002,0x8800000,0x80280,0x20822080,0x8808800,0x8028a20a,0xa0820080,0x8020000,0x8080280,0x2002082,0x208020,0x80002288,0x20000a0,0x20000000,0x200a002,0x20882000,0x20220008,0xa288282,0x20802020,0x800028,0x200a002,0x80082082,0x8200800,0x8a200,0x80002820,0x80280088,0x22080000,0x8800208,0x8a20a088,0x20080802,0x20000a08,0x80280080,0x2082080,0x8020020,0x2288002,0xa0800,0x20,0xa002200,0x20);
remainders_pos[781].modulo_value=3124; remainders_pos[781].bit_mask=create_bit_mask(98,0xa22a2aa,0xa00a22a8,0xa220802,0xa8a08280,0x22280880,0xa288a200,0xa00820aa,0x882a802a,0x8a020808,0xaaa0002,0xa8822020,0x28828220,0x200000,0xa2a00a0a,0x8080aa80,0x28080802,0x22222a88,0x208808a0,0x8228aa0a,0x2a228280,0x2880008a,0x2280820a,0xa20aa200,0x22020280,0x22020202,0x280a0202,0x20022a82,0xa282080a,0xa2880008,0x80a0a22,0x2282a8a2,0x20288088,0xa08aa222,0xa008080,0x280aa808,0x282802a,0xa0000020,0xa8220a08,0x80202208,0x8a0002aa,0x88808202,0x2aa00aa0,0x2aa82080,0x2002288a,0xa80880a2,0x800a0828,0x2a008222,0x80aa2280,0xaaaa2a22,0xa0288a8a,0xa80288a,0x288820,0x2a2820a,0x88a022,0xaa8a2288,0xaa802082,0x2220aa00,0xa280820,0x802aa800,0x82a20880,0xa20a08,0x28008000,0x28a8028,0xa0202aa,0x20a02020,0x808888aa,0x28822022,0x208a2a8,0x28a88a0a,0x28a20002,0x8a0208,0x2882a88,0x888080a,0x8880808,0x8a02808,0x288008aa,0x228a0820,0x8a8a2000,0x88202828,0x208a0aa2,0x8880a202,0x2822a88,0x20280202,0xa8a02aa0,0x800a0a00,0x22800000,0x22a08828,0xaa008088,0xa28000a,0x82220208,0xaa802a,0x28aaa082,0x888008a2,0xa2a02202,0x8a002820,0xa80208,0x8a02a88a,0xaa8a8);
remainders_neg[781].modulo_value=3124; remainders_neg[781].bit_mask=create_bit_mask(98,0x82aa2222,0x2802aa20,0x82aa800a,0x20280a00,0xaaa00008,0x22002a88,0x2880a822,0xa20822,0x2828080,0x222888a,0x200aa8a8,0xa00a02a0,0x88288888,0x2a288282,0x8082200,0xa080008a,0xa2aaa200,0xa8008028,0xaa02202,0x22a20a08,0x20088802,0xaa080a82,0x2a822288,0xaa0a8a08,0xaa8a8a8a,0xa0828a82,0xa88a220a,0x220a8082,0x22008880,0x8082822a,0xaa02202a,0x28a00800,0x28022aaa,0x82880008,0xa0022080,0x8a0a08a2,0x288888a0,0x202a0280,0xa8aa80,0x2888a22,0x80a0a,0xa2208228,0x2220a808,0xa88aa002,0x2080002a,0x80280a0,0xa2800aaa,0x822aa00,0x222222aa,0x20a00202,0x8208a002,0x88a00028,0x8a220a82,0x80028aa,0x2202aa00,0x2208200a,0xaa282288,0x82a080a8,0x8a22080,0xa2a0008,0x802a8280,0xa0880888,0x8a020820,0x82828a22,0x2828a0a8,0x8000022,0xa00aa0aa,0x8a002a20,0xa0200282,0xa02a8882,0x88020a80,0x8200a200,0x80008082,0x80008000,0x8020a080,0x20088022,0xaa0280a8,0x202a088,0x28a0a0,0xa802822a,0x82a82,0x8a0a2200,0xa0a08a82,0x2028a228,0x8828208,0xaa008888,0x2a2800a0,0x22880800,0x82a08082,0xa2a8a80,0x82208a2,0xa0222802,0x8002a,0x2228aa8a,0x288a0a8,0x88208a00,0x2822002,0x22020);
remainders_pos[782].modulo_value=3128; remainders_pos[782].bit_mask=create_bit_mask(98,0xa2208a02,0xa220088,0x88020082,0xa02888a,0x202020a2,0xa28822a0,0x8802220,0x28228aa,0x88a828aa,0xaa8aa020,0x80228a80,0xa8a02282,0x8a022a22,0x2a88282a,0x20220020,0x82000a82,0xa8282202,0x800a202,0x80220aaa,0x8a8a8a2a,0x2020aa0a,0xa8802000,0xa2a0a808,0xa2a0008,0x2a0828a0,0x8aaa000,0x8a00a00,0x2aaa2a80,0x80a00a08,0x80200000,0x8000a02a,0x2a202a8,0x82a20028,0x20802aa2,0x2a28a22a,0x80202800,0x20022a0,0x82022a08,0x8808222,0x82280080,0x82808,0x20082282,0xa220a88,0x22022880,0xa20a2a22,0x80888080,0xa20aa228,0x88aa020,0x82a08202,0x2080820a,0x8a080aa2,0x2288aa0,0x8a020222,0x8888a8a0,0xa0022880,0x822a088,0x828820,0x82202820,0x20020028,0x82888202,0x8020a880,0x20a8800,0xa8002808,0x8a88a28,0x828aa802,0x8028008a,0x22a808a,0x2a80a00,0x2000008,0xa820a00a,0x2002a8aa,0x2000a00a,0xa8000aaa,0xa00a2820,0x8a2000a8,0x2a202a0a,0x8000802,0xa2a0aa08,0x2a8a2a2,0x20aaa088,0x2a808a00,0x82808828,0x882a000,0xa8080088,0xa2a82822,0x2a88a880,0x282880a,0xaa02a288,0x22080aa2,0x80aa282a,0x20aa2882,0x8a088802,0x80a8822,0xa08a0808,0x22a22280,0xa0820080,0x8a220088,0x80a208);
remainders_neg[782].modulo_value=3128; remainders_neg[782].bit_mask=create_bit_mask(98,0x2a28028a,0x82a28800,0x8a880a,0x82020002,0xa8a8a82a,0x2200aa28,0x80082aa8,0xa0aa022,0x20a022,0x220220a8,0x8aa0200,0x2028aa0a,0x28aa22a,0xa200a0a2,0xa8aa80a8,0xa888202,0x20a02a8a,0x882a8a,0x8a28222,0x20202a2,0xa0282282,0x2008a888,0x22282080,0x82228880,0x2280a028,0x80222888,0x80208288,0xa222a200,0x8288280,0x8a80808,0x88828a2,0x8a2a8220,0xa2a80a0,0xa808222a,0xa2a02aa2,0x8a0a008,0x8a88aa28,0xa0aa280,0x80080aa2,0x2a08808,0x880a080,0x2880aa0a,0x82aa8200,0xa28aa008,0x2a82a2a2,0x8000808,0x2a022a20,0x800228a8,0xa28028a,0xa8000a82,0x280022a,0x8aa00228,0x2820aaa,0x2028,0x280aa008,0x80aa2000,0x800a00a8,0xaa8a0a8,0x288a8820,0xa000a8a,0x8a82008,0x8a820080,0x2088a080,0x200220,0xa02208a,0x8a08002,0x82a20802,0x8a200288,0x8a888880,0x20202882,0xa88a2022,0xa8082882,0x20808222,0x2082a0a8,0x2a88820,0x22a82282,0x8088808a,0x2a282280,0x8a202222,0xa8222800,0xa2080208,0xa080020,0x800a2088,0x20808800,0x2a2020a2,0xa2002008,0xa020082,0x228a2a00,0xaa00822a,0x22a0a2,0xa022a00a,0x280008a,0x200aa,0x28028080,0xaa2aaa08,0x28028800,0x2aa8800,0x82a00);
remainders_pos[783].modulo_value=348; remainders_pos[783].bit_mask=create_bit_mask(11,0x828a2002,0x8820a08,0x208a080,0x82022820,0x200,0x2280288,0x20080000,0x828808,0x20a208,0x220a0822,0x8008a28);
remainders_neg[783].modulo_value=348; remainders_neg[783].bit_mask=create_bit_mask(11,0x2022882,0x28200,0x8a002008,0x28a20a0,0x8808a08,0x8a208200,0xa0800080,0x8020000,0x88282280,0x228200a2,0x800220);
remainders_pos[784].modulo_value=56; remainders_pos[784].bit_mask=create_bit_mask(2,0xaa8aaa2a,0xa8aaa2);
remainders_neg[784].modulo_value=56; remainders_neg[784].bit_mask=create_bit_mask(2,0x22022222,0x202222);
remainders_pos[785].modulo_value=785; remainders_pos[785].bit_mask=create_bit_mask(25,0x90cd4b96,0xc12a58d1,0x13ca7381,0x5a01acee,0x15056221,0x1875a4e,0xe652d6d4,0x3144a009,0x16860401,0x1992033,0xb1ee1854,0xa229c235,0x51147338,0xde36b10e,0x6600a861,0x85a33012,0x8a320081,0x299e4014,0x8600adad,0x82a1c96b,0x16a111a,0x4f21dcd6,0x520e0739,0xcc262c69,0x1a74a);
remainders_neg[785].modulo_value=3140; remainders_neg[785].bit_mask=create_bit_mask(98,0x800020a,0x8228000,0xa02a208,0x82802022,0x8088aa28,0x880a0282,0x2a0a0a00,0x28882080,0x8a028888,0x8000a82a,0x28220088,0x2a200aa0,0x888a2a0,0x2a22882,0xa2082028,0x82a288a,0x22a8800,0x20028088,0xa802020,0x2280022,0x88a20882,0x82280002,0xa8282a0,0xa2a0a0,0x2a02a02,0x282a2802,0x2a0a0a80,0xa208,0x882288a,0x2000a202,0x20200a82,0x880a0020,0x8aa200,0x88a2a080,0xa020822a,0x8a22a00,0x2a288882,0x2a8022a0,0x880220a0,0xa0a80008,0x888a028a,0x82088a0,0x28282a0,0xa028088,0xa2a8880a,0x20200a08,0x822a0282,0xa2080,0x82000080,0x82a8a202,0x88202a8a,0x82200282,0x208a2220,0x8800008,0x8280a020,0x820288,0x80a08a2,0x2808088,0x8aa00800,0x220a808,0x2a000a,0x88080022,0x200200a0,0x82280a,0x82000800,0xa0000a8a,0x2a208882,0x8a2228,0x8202a822,0x80208822,0xa00a2a80,0x80208020,0x28002028,0x802200a2,0x2000a22,0x82008a,0xaa280282,0x8820080,0xaa020a2,0x22288022,0x88822a0a,0xa80002a0,0x80020a8,0xa208000,0x80200228,0x80882,0x2a0022,0xa822028,0x802a888,0x8080a000,0x88280808,0xa0208022,0x280a088,0x8802,0x22288208,0xa0022082,0xaa0208a0,0x228aa0a8);
remainders_pos[786].modulo_value=3144; remainders_pos[786].bit_mask=create_bit_mask(99,0x28a08a2,0x8a28028,0x80202088,0x22880820,0x20800a20,0xa28820a,0x20002,0x20200820,0x208a20a,0x20000082,0x20828a08,0x2288288,0x228a0082,0xa00200,0x88280202,0x28a0080,0x28a20a20,0x8808a008,0x220a0882,0xa20a00,0x2282208,0xa0802800,0x20000028,0x2082082,0xa2022002,0x8208220,0xa28a202,0xa00a0800,0x20000228,0x80200288,0x80802022,0x28008800,0x200008,0x828000a2,0x20a00808,0x82280282,0x82800022,0x800220,0x80008208,0x828a2820,0x20008220,0x8282200,0xa28228a2,0x20800020,0x288200,0x20880020,0x8220a08,0xa000088,0x8000a0,0x20000,0xa00a,0xa0882022,0x220820,0x82280008,0x20800,0x28828a08,0x8828208a,0x82000800,0x28a28208,0x82000208,0x80020020,0x28208,0x80288288,0x200a0882,0x28220,0x8008a,0x22002820,0x8020200,0x80080288,0x80000822,0x20a00a28,0x8a28a000,0x82082080,0x8808a08,0x8208080,0x882,0x28020a28,0x88288000,0xa08a0020,0x20a08800,0xa202282,0xa08a2820,0xa28008,0x80282202,0x800a0080,0xa28800,0x82288082,0xa2820822,0x820,0x8a208082,0x200808a0,0x800008,0x8228a080,0xa00208a0,0x20228808,0x8080208,0x28a2022,0x20a28028,0x8a);
remainders_neg[786].modulo_value=3144; remainders_neg[786].bit_mask=create_bit_mask(99,0x82020022,0x220820,0x828a000,0xa20000a0,0x28008220,0x82200282,0x808a0882,0x28a08028,0x8a002282,0xa0880802,0x28020200,0x8a200200,0xa2020802,0x8208a08,0x20828a,0x82020800,0x20228228,0x2080,0xa2820002,0x8228208,0x8a202280,0x20082080,0x28808820,0x8a00a00a,0x228a2882,0xa00a28,0x8220228a,0x20820080,0x28808a20,0x8288200,0x828a2,0x20800008,0x88288080,0x2080822,0x28208000,0xa20820a,0x20808a2,0x8008a28,0x8080280,0x20220a0,0x28800a28,0x8020a288,0x220a2022,0x28008828,0x88200288,0xa00008a0,0xa28200,0x82088000,0x80080820,0x8828808,0x88082082,0x200028a2,0x8a28028,0xa208000,0x808a0080,0x20020200,0x20a002,0x2880080,0x20220a00,0xa088280,0x8a08a0,0x8820a00,0x8200200,0xa0820002,0x8820a28,0x88008002,0xa28820a0,0x828a08,0x8008200,0x8000a2,0x28208220,0x2202088,0x2802800,0x200,0x80280008,0x80880002,0x20828220,0x200088,0x200208a0,0x28200008,0x8228a20a,0x200220a0,0x8220800,0x820a28a,0x820800,0x8220008,0xa20000a,0x220a00a2,0x8808028,0x228000a,0xa0800020,0x8008800,0xa202008,0x208a0020,0x28a20000,0x80008280,0x820228a2,0x28220820,0x2);
remainders_pos[787].modulo_value=3148; remainders_pos[787].bit_mask=create_bit_mask(99,0x2a00020a,0x8820a28,0x88aaaa82,0x8a8a82a2,0x8a8802a,0x8802222a,0x2a2a0028,0xa28a28a2,0x2202a222,0xa22880a,0xa2aaa08,0xa02a8aa2,0x882a8,0xaa222828,0x80228222,0x28008208,0x8a8aa0,0x2802000a,0x82a282,0x828a8a08,0x80a2280a,0xaaa808,0x2a802a2,0x28a0a8a2,0xaa02802a,0x20282080,0xaa00a88,0xa0002a88,0xa0882a22,0x82020a02,0x80a2aa2,0xaa8aa08a,0x2022a82,0x8a2aa0a8,0x8a08aa22,0xa0888008,0xa22aaa0,0x200a820,0x80008288,0xa20882a2,0x880a8882,0x20820808,0xaa808088,0x888aa220,0xaa202280,0xa020200,0x2208,0x828a228a,0x8aaa80a0,0x2aaa2802,0x8a2820a0,0x80000a28,0x80a0208,0x808aa008,0xaa222028,0x202aa088,0x8208220,0x2a022202,0x2208a822,0x200028a8,0xa0080228,0xa88a0082,0x2220a0aa,0xa20a2200,0xaa8a288a,0x880802a0,0xaa2aa82a,0x8a020a20,0x80828aa,0x8220a80a,0x8000a88a,0xaa022a,0x8280822a,0x280aa020,0xa0a28a80,0x2a808a2,0xaaa008a8,0x88a02202,0x2a282a02,0xa820020a,0x8028828,0x2a200a00,0x200280aa,0x28802208,0x888aa888,0x22000282,0x2a80a2a8,0xaa8a08aa,0x288a020a,0xa8088a02,0x8a28a888,0xa8a88a2,0x88022280,0x22a20a88,0x2a2a2a80,0xaaa228a8,0x822082a,0xa828a,0xa08);
remainders_neg[787].modulo_value=3148; remainders_neg[787].bit_mask=create_bit_mask(99,0xa2888a82,0x800a82a0,0x22220a,0x2020a2a,0x802008a2,0x8aaaa2,0xa2a288a0,0x2a02a02a,0xaa8a2aaa,0x82aa0082,0x82a22280,0x28a2022a,0x88800a20,0x22aaa0a0,0x8aa0aaa,0xa0880a80,0x88020228,0xa08a8882,0x880a2a0a,0xa020280,0x82aa082,0x88222080,0x8a208a2a,0xa028202a,0x228208a2,0xa8a0a808,0x82288200,0x2888a200,0x2800a2aa,0xa8a828a,0x8082a22a,0x22022802,0x8a8aa20a,0x2a22820,0x28022aa,0x28000880,0x82aa2228,0x8a8820a8,0x8880a00,0x2a800a2a,0x82000a,0xa80a8080,0x22080800,0x22aa8,0x22a8aa08,0x828a8a88,0x8888aa80,0xa02aa02,0x2220828,0xa222a08a,0x2a0a828,0x88882a0,0x80828a80,0x8022880,0x22aaa8a0,0xa8a22800,0x80a80aa8,0xa28aaa8a,0xaa8020aa,0xa888a020,0x28808aa0,0x2002880a,0xaaa82822,0x2a82aa88,0x2202a002,0x808a28,0x22a220a2,0x28a82a8,0x8080a022,0xaa82082,0x8882002,0x88228aa2,0xa080aa2,0xa08228a8,0x282a0208,0x8a20802a,0x22288020,0x28aa8a,0xa2a0a28a,0x20a88a82,0x808a00a0,0xa2a88288,0xa88a0822,0xa008aa80,0x22000,0xaa888a0a,0xa2082a20,0x22028022,0xa0028a82,0x2080028a,0x2a02000,0x8202002a,0x8aaa08,0xaa2a8200,0xa2a2a208,0x222aa020,0x80aa80a2,0x88820a02,0x280);
remainders_pos[788].modulo_value=1576; remainders_pos[788].bit_mask=create_bit_mask(50,0x22888282,0xa8aa8aa2,0x202a0002,0x82002a22,0x280082a0,0x8820aa0a,0xa8208a02,0x800a282,0xa002082a,0x2a02222,0x8aa8aa00,0x88a222a,0x220a0a0a,0xaa2a888a,0xa8000aa2,0xa88880,0x20a8208,0x82a828a0,0x82280a20,0x28a0aa0,0x820a820,0x80888a80,0xa2a8000a,0x2888aa2a,0x28282822,0xaa222888,0x2a8aa8,0xa22202a0,0x2a082002,0xa0a28008,0xa028820a,0x282a8208,0x82a0800a,0x222a0020,0xa0002a02,0x22a8aa8a,0xa0a088a2,0x88a220a0,0xaa2aa2a8,0x880a8000,0x20800a88,0x8a0020a8,0xa2082a82,0xaa082280,0x820028a0,0xa800820a,0xa80888,0xa2aa2a80,0x8222888a,0x82);
remainders_neg[788].modulo_value=1576; remainders_neg[788].bit_mask=create_bit_mask(50,0xaa000a0a,0x2022022a,0xa8a2888a,0xa88a2aa,0xa0880a28,0xa82282,0x20a8028a,0x80882a0a,0x288a80a2,0x8a28aaaa,0x2202288,0x8002aaa2,0xaa828282,0x22a20002,0x2088822a,0x88200008,0x8a820a80,0xa20a028,0xaa002a8,0x8a028228,0x80a820a8,0x8000208,0x2a208882,0xa00022a2,0xa0a0a0aa,0x22aaa000,0x88a20220,0x2aaa8a28,0xa280a88a,0x282a0880,0x28a00a82,0xa0a20a80,0xa280882,0xaaa288a8,0x2888a28a,0xaa202202,0x2828002a,0x2aa828,0x22a22a20,0x820888,0xa8088200,0x288a820,0x2a80a20a,0x2280aa00,0xa88a028,0x20880a82,0x88208000,0x2a22a208,0xaaa0002,0xa);
remainders_pos[789].modulo_value=3156; remainders_pos[789].bit_mask=create_bit_mask(99,0xa2002022,0x28228a20,0x2082,0x228228a0,0x28820028,0x8a200282,0x80822020,0x28208808,0x8202,0x20a00a2,0x28020828,0x8200820a,0x800a2,0x8a08000,0x208000,0xa2800802,0xa20008,0x208202,0xa28a0022,0x20828a08,0x200a00a,0x20020082,0x20208020,0x80088082,0x20200a2,0x8800a28,0x8202208,0x2002020,0x28020208,0x8a082280,0x28a2882,0x8200220,0x2280202,0x800a2880,0x8200200,0x8288000,0x800000,0x8020a28,0x8200a282,0x28200a0,0x8000228,0x8820a202,0x22080880,0x228820,0x208a20a,0xa20a20a0,0x20000028,0x8a20a208,0x20022822,0x808a20,0x8a288288,0x8208a0,0x8a28000,0x800a08a,0x800a08a2,0x8808228,0x82202202,0x20808a0,0x28028800,0x820a008,0x20828a0,0x20028a08,0x82000000,0x82000022,0x200800,0x8800228a,0x82080802,0x28008800,0xa28228a,0x2802882,0x8020808,0x82008080,0x22020880,0x8028a00,0x28a008,0x80882022,0x808020,0x8082008,0x208a00a0,0x28a20a28,0x8008800a,0xa0080820,0x28a20008,0x80080200,0xa2000020,0x20,0x828a002,0x28a0820,0x8028208,0x8a00a,0x82880820,0x20220220,0x8a208088,0x22882800,0x28828008,0xa288,0x82882080,0x8a08a28,0x88080);
remainders_neg[789].modulo_value=3156; remainders_neg[789].bit_mask=create_bit_mask(99,0x228828a2,0x20a20228,0x8808a00a,0xa20a2020,0x20028820,0x228820a,0xa28a0,0x20a00000,0x8808020a,0x82820822,0x20828020,0xa080282,0x80800822,0x200808,0x88280088,0x22080082,0x8228800,0x8828028a,0x220208a2,0x28020200,0x8a082082,0xa08a0802,0x28a00828,0x800000a,0x828a0822,0x8220,0x8028a280,0x828828a0,0x20828a00,0x200a208,0x82022002,0xa08a28,0x8a20828a,0x822000,0xa08a08,0x80200088,0x80080880,0x828220,0xa08220a,0x820a0820,0x808a20,0x282282,0xa2800000,0x8a20028,0x8a002282,0x22822820,0x28808820,0x2282280,0xa08a20a2,0x8000228,0x2200200,0x800a0020,0x220808,0x80082002,0x820022,0xa20,0xa28a28a,0x82800020,0x20820008,0x80282080,0x82802020,0x28820200,0xa088088,0x28808a2,0x8a08008,0x8a202,0x2800082,0x20800008,0x8220a202,0x82082002,0x828000,0xa080008,0xa28a0000,0x820208,0x88202080,0x28a2,0x8000828,0x8000a080,0xa0020820,0x20228220,0x8000082,0x208000a0,0x20228800,0x8008288,0x228808a0,0x8808828,0x8020208a,0x820200a0,0x820a00,0x88002082,0x20000a0,0x28a28a28,0x2280000,0xa2002080,0x20020800,0x88082200,0x2002800,0x200220,0x8);
remainders_pos[790].modulo_value=3160; remainders_pos[790].bit_mask=create_bit_mask(99,0x288a2a0a,0x220a0882,0x28200a0,0x2222808,0x228820a,0x88880a80,0xa2228082,0x208a0228,0x822088a8,0xa0880a02,0xa008028a,0x82088a82,0x28002288,0x8a202000,0x2202800,0xa8282080,0xa020002,0x2a02a0a0,0x20082a,0xa0882000,0xa8a02802,0x280082a2,0x2a8a0028,0x2a000a0,0x220a220,0xa8022a02,0xa0000a02,0xa882008,0xa000002,0x20880228,0xa82a2a8a,0x2200020,0x2280a200,0x82000aa,0x2220aa0a,0x28200a88,0x2a280aa2,0x8828080,0x88a20080,0xa0880002,0x88220280,0x200a0800,0x2020020,0x8880808a,0x8020aa00,0x20222082,0x820288a0,0x20822208,0x802088,0x80020,0x8220802,0x82208882,0x80a2280,0x2820888,0x2200aa08,0x80a20202,0x8080080,0x220020a0,0xa028088,0x22800022,0x2002008a,0xa8020282,0x288aa028,0x8822a008,0x20a0aa08,0x88aa0008,0x80008a02,0x2a080008,0x8a2a8a8,0xa0288022,0xa0800000,0xa200822,0x2a80a000,0x8080a880,0x80088a08,0xa80a000a,0x282800a2,0x2a8a8200,0xa80280a,0x822,0xa8a82008,0xa00a0a80,0x2a800080,0x80020828,0xa2002808,0x28000808,0x82228800,0xa82a220,0xaa28020,0x8280a022,0x82a2208,0x8a2880a2,0x22820288,0x8002a022,0x80a08228,0x80202888,0x880a0082,0x288220a0,0xa0a8a2);
remainders_neg[790].modulo_value=3160; remainders_neg[790].bit_mask=create_bit_mask(99,0xa0022282,0xaa028002,0xa0a0028,0x8aa2a080,0x82a00a02,0x208,0x2a2a0802,0x280282a0,0xaa00020,0x2000820a,0x28800a02,0xa000202,0x2088a200,0x2a0a888,0x82a8a008,0x20a02808,0x820a8882,0x228a2028,0x88a080a2,0x2000a808,0x2028208a,0xa0080a22,0x22028020,0x8a208828,0x82a82a28,0x208a228a,0x28088282,0x200a080,0x8280888a,0xa0008a20,0x20a22202,0x8a2888a0,0x2a082280,0x80a08822,0xa2a82202,0xa0a80200,0xa2208222,0xa0008,0x228808,0x2000880a,0xaa0a08,0x28028080,0xa8a80a8,0x802,0xa82208,0xa8aa280a,0xa0a0020,0x280aa280,0x8800a800,0x80808828,0x80aa008a,0x2280002,0x82a208,0x8a028000,0xa2882200,0x82a0a8a,0x80008800,0x2a88a028,0x82820800,0xa208882a,0xa88a0802,0x200a8a02,0x200220a0,0xa22880,0xa0282200,0x220880,0x8080282,0x22808080,0x80222020,0x20a0082a,0x28080888,0x822080a2,0x22082088,0x8002008,0x800200,0x20820882,0xa0208822,0x22020288,0x8200a082,0x8088802a,0x20202880,0x28020200,0x22088008,0x88280a0,0x2288a000,0xa0880080,0xa2a0080,0x20a22a8,0x822208a8,0x208282a,0x80a22a80,0x2200022,0x2a0a8200,0x88228aa,0x280a20,0x8a82000,0x28802,0x200aa028,0x20202a);
remainders_pos[791].modulo_value=3164; remainders_pos[791].bit_mask=create_bit_mask(99,0x82820222,0x2202882,0x2aa8288,0x2a08a820,0xaa0a808,0x80aa0220,0x828008,0x880022a,0x28a02280,0xa0a8022a,0xa0028000,0xaa00a288,0x8aa228a8,0x208a8220,0x20880008,0x2a80a82,0xa280a2,0x8a0a2882,0x28a0880a,0x82820288,0xa0aaa280,0x80880800,0x2a88aaa2,0xa0888a80,0x22088888,0x22202202,0x2002a00,0xa02a0000,0x82a22,0xa082a00,0x2020820a,0x800200a2,0x8228a00,0x800a0828,0xa0288202,0x208a22a,0x8002882,0x22800020,0x2282008,0xa0202a8a,0x88200088,0xa822a208,0xa0a800a,0x2822a2,0x2002888a,0x200a0a0,0x2a880,0x888a2208,0x880a0220,0x2a02080a,0x20880a02,0x2088a22,0x22a800,0x80a0a008,0xa222800,0x8a88280,0xaa002a0a,0x2208a882,0xa2200082,0xa2a8080,0x82008288,0x800028,0x8228002,0xaa88a208,0x28082280,0x2820a00,0x200a2882,0x88a00800,0xa082080,0xa820a,0xa88a8200,0xa80,0x800a8008,0x88088088,0xa2222208,0x802a2220,0x28aaa22a,0xa0020220,0x2028aaa0,0x82280828,0x2a0220a2,0x8228a0a,0x8a028a0,0x882a02a8,0x82000220,0x20882a20,0xa2a288aa,0xa228a00a,0xa0002800,0x8a8802a0,0x2880a2,0xa880022,0x22002820,0x880aa0,0x8202a0aa,0x82a20a,0x2282aa8,0x28228088,0x8880828);
remainders_neg[791].modulo_value=3164; remainders_neg[791].bit_mask=create_bit_mask(99,0xa0a8a2a,0xaa8a002,0x82220a00,0xa20020a8,0x82202080,0x8220aa8,0x880a0080,0x80088a22,0x2028aa00,0x20208aa2,0x28020888,0x22802a00,0x22a2020,0xa80202a8,0xa8008800,0xa208202,0x802a082a,0x202a00a,0xa0200082,0xa0a0a00,0x28222208,0x8008008,0x22002222,0x20000208,0xaa000000,0xaaa0aa8a,0x8a882288,0x28a28088,0x8880a22a,0x280a280,0xa0a80a82,0x80a882a,0x80a20288,0x88200a0,0x28a0028a,0x8a802a22,0x88a002,0xa20888a8,0x8220a880,0x28a0a202,0xa80800,0x20aa2280,0x82820802,0x8a0aa22,0xa08a0002,0x8a082828,0x88822008,0x22a80,0x8282a8,0xa28a8002,0x28008282,0x828002aa,0x82a2088,0x8202880,0x82aa2088,0x80200208,0x2288a202,0x2a802002,0x22a8880a,0x82220808,0xa800a00,0x880808a0,0x80aa008a,0x22002a00,0x2080aa00,0x820a8288,0xa802a00a,0x208080,0x82802808,0x88820282,0x20020a08,0x8888200,0x820880,0x800,0x2aa2aa80,0x8a22aa8,0xa02222a2,0x288a8a28,0x28a02220,0x2a080a0,0xa20aa82a,0x80a20202,0x80282028,0xa28220,0xa888a28,0x2800a2a0,0x222a0022,0x2a202882,0x2880a088,0x2000a28,0x88a0002a,0x8200882a,0x2a88a0a0,0x80008228,0xa0a2822,0x88022282,0x8aa02220,0xa0aa0000,0x8020);
remainders_pos[792].modulo_value=264; remainders_pos[792].bit_mask=create_bit_mask(9,0x22002082,0x28020000,0x8a208008,0x800a2022,0x8a00200,0x208a288,0x802820,0x8008800,0x82);
remainders_neg[792].modulo_value=264; remainders_neg[792].bit_mask=create_bit_mask(9,0xa2882002,0x20028808,0x2280080,0x8228a2,0x200a08,0x82002200,0x800820a0,0x800008,0xa);
remainders_pos[793].modulo_value=793; remainders_pos[793].bit_mask=create_bit_mask(25,0x9b655bde,0xc9471433,0xa3322d8d,0xd24b989d,0x4db4059,0x640f8504,0xc3e063d3,0xd10c18cf,0x3a883bcb,0x385347,0x6ab9281a,0xb410aad5,0xb6877b85,0x5aad5420,0x1605275,0x738b2870,0x2f4f7045,0xfcc60c2,0x9b2f181f,0x808287c0,0x2e680b6c,0x16e46749,0x4ec6d133,0x6730a38a,0x1ef6a9b);
remainders_neg[793].modulo_value=3172; remainders_neg[793].bit_mask=create_bit_mask(100,0x2a88202,0x8a882a,0x22aaa000,0xa820000,0x88020880,0xa8020888,0xa28aa02,0x8008002,0xa200a202,0x88a0828a,0x2220a082,0x28882208,0x2a02a208,0x822080a8,0x88a88a28,0xa202a0a8,0xa282a888,0x8200280a,0x2a28082,0x80a0a00,0x222082a0,0x8a28aa80,0x820a00aa,0x2aa82a02,0x2022a202,0xaaa2022a,0x820a02a0,0xa28aa802,0x22200aa8,0x80802a08,0x2a000282,0x20a080a,0xa2a80a0,0x22888aa,0xa888a82a,0x2208a288,0x2a0a808,0x88a0822a,0x22208220,0x288a0828,0x22a8a08,0x820228,0xa2820008,0xaa02a8,0x888882,0xa800882,0xaa200002,0x8800002a,0xaa02a08a,0x22a20208,0x8802a008,0x2aa22880,0xaaaa0000,0x88aa8022,0x80888aa0,0x28008a2,0xaaa20208,0x22a288a8,0x808022a0,0x28208820,0x22882228,0x228a008a,0x80a0280,0x8008a228,0x28008808,0x88802a0,0xa002020,0xa0a0a2a8,0x828aa020,0xa02a0a82,0xa82028,0x88008200,0x220a282,0x22a0000a,0xa2022220,0x20280002,0x8a0a282,0x80002080,0x28a020a,0xa0a0a82a,0x828202a8,0x280aa282,0x88020200,0xa02a008,0x880880,0x80a2288,0xa200a028,0xa2288028,0xa0a2208,0x80820882,0xa202a200,0xaa8a88a2,0xa0082022,0x80a28800,0x8882a888,0xaaa200aa,0xaa00002a,0x8808a22,0xa20802a0,0x2);
remainders_pos[794].modulo_value=3176; remainders_pos[794].bit_mask=create_bit_mask(100,0x82a02222,0x2a2aa002,0x82a202,0xaaa8a222,0x28a00a22,0xa82002a0,0x2a8a20,0x8a20080a,0xaa82aa0a,0xa0222a82,0x808082a,0x8820a202,0x2aa0a22,0xa8aa0a8,0xa2280808,0x22822aa8,0x20a2282,0x8a0808aa,0x20aa0228,0x28a288a8,0x22aa08a0,0x8aaaa022,0x2a8aa280,0x8288a22a,0x2aaa8a2,0x82820800,0x80202208,0x8a8200,0x2222a00,0x82a08a0,0xa280882,0xa08822a0,0xa2a008a8,0x82228220,0x22080022,0x80a8a8a8,0x2a80a00,0x28822020,0xa8a2a20a,0x2a0028a8,0x80028022,0x2080a002,0x2aa0a8aa,0x280a2080,0x82a02aa,0x220aa,0x8aa22208,0x2022a202,0x282aa00,0x8882222a,0xaa8280a8,0x8a880800,0x888aa280,0x8800020,0x80a820aa,0x8a028aa,0x2a0aa802,0xa0208aa,0x2800280,0x2800a888,0x8a8a2a2a,0x88228a0,0xa02a8008,0x2a2a0200,0x20882a,0x82888288,0x200a8a08,0x88220a2a,0x2028a00a,0x20a82082,0xa888800a,0x82a20000,0x88080200,0x20828220,0x2aaa8000,0x8a22828a,0x8aa2a8a8,0xaaaa202,0x20aa8888,0x228a280a,0x80aa082a,0x2020a228,0x88a080aa,0xa8828882,0x20288a2a,0xaa2a020,0xa0aa802a,0x8a082288,0x20202080,0xa8880aa8,0xaa82aa82,0x2008a2a0,0xa2a800a0,0x80082a08,0xa00a280a,0x8a2aaa88,0x8a820088,0xaa8a880,0x880a8280,0x88);
remainders_neg[794].modulo_value=3176; remainders_neg[794].bit_mask=create_bit_mask(99,0xa28aaaa,0xa2a2288a,0x880a2a8a,0x22202aaa,0xa02882aa,0x20a88a28,0x88a202a8,0x2a88082,0x220a2282,0x28aaa20a,0x808080a2,0xa82a8a,0x8a2282aa,0x82022820,0x2aa08080,0xaa0aa220,0x8a82aa0a,0x2808022,0xa8228aa0,0xa02a0020,0xaa228028,0x22228aa,0xa2022a08,0xa002aa2,0x8a22202a,0xa0a8088,0x8a8aa80,0x88020a88,0x8aaaa288,0x80a28028,0x82a0800a,0x2800aa28,0x2a288020,0xaaa0aa8,0xaa8088aa,0x8202020,0x8a208288,0xa00aa828,0x202a2a82,0xa288a020,0x88a08aa,0xa808288a,0xa2282022,0xa082a808,0x80a28a22,0x888aa822,0x22aaa80,0xa8aa2a8a,0x8a0a2288,0xaaaa2,0x220a0820,0x2008088,0x22a08,0x800888a8,0x820a822,0x8028a022,0xa282208a,0x828a8022,0x8a088a08,0xa0882000,0x202a2a2,0x800aa028,0x28a20880,0xa2a28a88,0x88a800a2,0xa000a00,0xa8820280,0xaa82a2,0xa8a02882,0xa820a80a,0x20000882,0xa2a8888,0x808a88,0xa80a0aa8,0xa2220888,0x2aa0a02,0x22a2020,0x82222a8a,0xa8220000,0xaa02a082,0x82280a2,0xa8a82aa0,0x280822,0x200a000a,0xa8a002a2,0x822a28a8,0x202208a2,0x280aa00,0xa8a8a808,0x20008220,0x220a220a,0xa8802a28,0x2a208828,0x880a280,0x2882a082,0x2a22200,0x20a8800,0x82202008,0x820a08);
remainders_pos[795].modulo_value=3180; remainders_pos[795].bit_mask=create_bit_mask(99,0x808a0882,0x8020a00,0x8a200,0x2820020,0x20800828,0x800028a,0x82020022,0x20020020,0x80200080,0x22882022,0xa20008,0x8a208200,0x220220a0,0x8208000,0x2000208,0xa0880082,0x28028800,0xa080208,0x80822020,0x800028,0x20220a,0x22080000,0x8800208,0x2000008,0x2800800,0x8a20208,0x2208282,0x8200a0,0x8a28800,0x2008008,0x20002882,0x200800,0x8280,0x22002,0x800a20,0x8020028a,0x800a0820,0x8,0xa208000,0x802800,0x28000800,0x88280280,0x2082080,0x808808,0x20a,0x802082,0x28208a00,0x2088000,0x2020080,0x200,0x8080008,0x22080020,0x20828000,0x8020000a,0x820,0x20200a08,0x82080202,0x2822020,0x28028,0x80200002,0x208a0002,0xa002200,0x802082,0x200a28,0x8800008a,0x20000800,0x800028,0x80008002,0x20080822,0x28a20200,0x8200002,0x208800a0,0x8a20828,0x280208,0x80002,0x220200,0x2088208,0x80800000,0x200a08,0x88202280,0x20a0080,0x28028208,0x220a002,0x80820,0x20820208,0x88088000,0x208a20a0,0x8a00008,0x82288200,0x802880,0x8008020,0x8082080,0x20880,0x208a28,0x8280282,0xa2000080,0x8020008,0xa20200a,0x822);
remainders_neg[795].modulo_value=3180; remainders_neg[795].bit_mask=create_bit_mask(100,0x20002,0x28200,0x8002288,0x820208a0,0x20008020,0x80080202,0x20a08a2,0x28828028,0x8208008,0xa2002822,0x8220800,0x2200280,0x228a2020,0xa00808,0x82088200,0x20000802,0x20020000,0x2008280,0x22820,0x8820,0x88282282,0xa2000880,0x8200,0x8a008080,0x82080000,0x220a00,0x8a280202,0xa0020,0x220008,0x82080000,0xa0882002,0x8200000,0x8080208,0x80822882,0x8228,0x8280202,0x200a0,0x8808000,0x82200080,0x80082000,0x20800008,0x208200,0x2802000,0x8000000,0x80088202,0x80082802,0x20200200,0xa000088,0x82820800,0x808a08,0x80000080,0xa20000a0,0x28020008,0x8208082,0x80880020,0x28a00200,0xa008282,0x20a20a0,0x8820820,0x28800a,0xa0020882,0x8008800,0x208a288,0x80002802,0xa08220,0x80002,0xa0080080,0x8008020,0x800008a,0xa0800022,0x20220a08,0x80288002,0x20000020,0x228020,0x80208200,0x80800882,0x8228a00,0x2000280,0x880,0xa08200,0x282208,0x82020800,0x20820200,0x8a20208a,0x80800020,0x28020a00,0x80,0x20022020,0x208800,0x2200208,0x80082000,0x820,0xa008,0x80820000,0xa00220,0x8020020a,0x22080800,0x28000,0x8220a082,0x22);
remainders_pos[796].modulo_value=1592; remainders_pos[796].bit_mask=create_bit_mask(50,0x2a08aa2a,0x2aa22082,0x828a088a,0xa2a00808,0xa00a8080,0x808a0aaa,0x822a020a,0x2a80002,0xa8082a2a,0x820a22a2,0x88008222,0x22808a28,0x80200080,0x228a2028,0x28882002,0xa8a88a08,0xa8a8202,0x280002a8,0x2a080a88,0xaaaa0a20,0xa0202a00,0x220200a8,0x8a220a28,0x882088aa,0xaa8aa20a,0x22a08aa2,0xa2aa2208,0x8828a088,0xa2a0080,0xaa00a808,0xa808a0aa,0x2822a020,0xa02a8000,0x2a8082a2,0x2820a22a,0x88800822,0x22808a2,0x88020008,0x2228a202,0x82888200,0x2a8a88a0,0x80a8a820,0x8280002a,0x2a080a8,0xaaaa0a2,0x8a0202a0,0x8220200a,0xa8a220a2,0xa882088a,0xa8aa20);
remainders_neg[796].modulo_value=1592; remainders_neg[796].bit_mask=create_bit_mask(50,0xa28022a2,0xa22aa80a,0xa028002,0x2a288080,0x28820808,0x8028222,0xaa28a02,0x8a20888a,0x2080a2a2,0xa82aa2a,0x880aaa,0xaa0802a0,0x8a88808,0xaa02a8a0,0xa000a88a,0x20200280,0x82020a8a,0xa0888a20,0xa2808200,0x222282a8,0x28a8a288,0xaa8a8820,0x2aa82a0,0xa80022,0x22022a82,0xaa28022a,0x2a22aa80,0xa02800,0x82a28808,0x22882080,0x20802822,0xa0aa28a0,0x28a20888,0xa2080a2a,0xa0a82aa2,0x880aa,0x8aa0802a,0x8a8880,0xaaa02a8a,0xa000a88,0xa2020028,0x82020a8,0xa0888a2,0x8a280820,0x8222282a,0x28a8a28,0xaa8a882,0x202aa82a,0x200a8002,0x2022a8);
remainders_pos[797].modulo_value=797; remainders_pos[797].bit_mask=create_bit_mask(25,0x323ee52,0x77525b58,0xe69246bd,0xc268ecd9,0xefc3b2ac,0x2b6b167b,0xa3050cbe,0x8751eb9f,0x107b138f,0xb8885fa,0x544b4230,0xd9bac3b2,0xf507382b,0x9370d766,0x310b48a,0x17e84474,0x7c723782,0x7e75e2b8,0x1f4c2831,0xf79a35b5,0xcd5370fd,0xe6cdc590,0xaf589259,0x6b692bb,0x129df130);
remainders_neg[797].modulo_value=3188; remainders_neg[797].bit_mask=create_bit_mask(100,0x8aaa228a,0xaa8a8280,0x2a0a8a20,0xaa02000,0x220a2a20,0xa2a28aa2,0x2a888022,0xa882202,0x88a28a02,0x82000822,0x88828aa8,0x220a2a,0x288aa0a2,0xaa80aaa,0x8a882802,0x8a2088a8,0xa0aaaa0a,0xa2a82a20,0x8280a0a8,0x2a02a828,0x8aa820,0x2a000808,0x22800a80,0x8a2a0a22,0x8a0028a8,0x88a8a002,0x222282a2,0xa00a800a,0x808002,0xa020aa88,0x8a0aa02,0x28a8280a,0x2822a0aa,0x8a82aaa8,0x88a88822,0x8200a08a,0xa2aa80aa,0x2282a88,0x8aa28220,0x8aa8a08,0x8a208002,0x82028a28,0xa202208a,0x2a20088a,0x222a8a2a,0x8022a282,0xa000202a,0xa8228a82,0x880a0a8a,0x828a22aa,0xa8800220,0x280080a0,0xaa028082,0x2a802a2a,0x22228202,0x20202080,0xa20088aa,0xa2808a22,0x20882082,0x8a2aa8a,0x88a080,0x22aa2282,0x2088028,0xa2800a80,0x8808a0a,0x80822888,0x28280002,0x8200a02,0x8a0a828,0x2a02a00a,0x8aa8800a,0xaa02aa8a,0x2220aa80,0x8820282,0xa002aa08,0x2088082a,0x822220a0,0xa02a80aa,0x8aa8a8aa,0xa02a2800,0x82882802,0x2080a0a,0xa0a2028,0x2080a000,0x8088088a,0xa2a828,0x882000a8,0x2aa20a00,0x888020a2,0xa2880082,0x88228aa,0x80a2a082,0x8022a228,0x2022a88,0xa2220082,0xaa2020,0xa02aaa2a,0x800a2080,0x8a8280,0x8220);
remainders_pos[798].modulo_value=3192; remainders_pos[798].bit_mask=create_bit_mask(100,0x2820002,0x20800820,0x88002,0x82022802,0x28a28008,0x80082288,0x20a0,0x8028820,0x200a280,0x800800a2,0x20208000,0x8220a08a,0x20a20a0,0x20020228,0x8020a,0x22080080,0x220828,0x8a002000,0x2020880,0x220008,0x82202,0x20822820,0x28800000,0x8200800a,0x20082020,0x8200808,0x280280,0x8a2080,0x28000208,0x2202082,0xa08a0022,0x8000020,0x8a002280,0x82020,0x28828020,0xa008,0x20002022,0x20020,0x2,0x80820802,0x20828,0x88000282,0xa2080000,0x8000220,0x8a,0x800a0000,0xa00200,0x2280088,0xa0082082,0x828000,0xa000282,0x80820820,0x220028,0x200800a,0xa0,0x20a20000,0x8a088000,0x22000020,0x828000,0x2282080,0x802082,0x800000,0x8080080,0x880800,0x28200a00,0x80282,0xa2080820,0x20028800,0xa080000,0x808800a2,0x28820808,0x208000,0x208a2,0x20028028,0x8202008,0x82080820,0x28a00200,0x8000002,0x82882,0x808820,0x80200088,0xa2022080,0x800,0x88282088,0x20020,0x8a08020,0x80288080,0x820a08a0,0x8a20a08,0x2000208,0x808a0020,0x20028a00,0x82280,0x20a0800,0x28228820,0x8220028a,0x802880,0x8800220,0x80082002,0x800082);
remainders_neg[798].modulo_value=3192; remainders_neg[798].bit_mask=create_bit_mask(100,0x82020802,0x28008020,0x8a,0x20a2082,0x20220800,0x8002200,0x80882020,0x20028,0xa082200,0x800822,0x28200800,0xa202002,0x82822820,0x28828220,0x80008202,0x22800800,0xa28020,0x208a008,0x82820000,0x8a20800,0x8800a28a,0x200a2020,0x20008800,0x2080082,0xa00020a0,0xa08000,0x88200208,0x80022000,0x20808a00,0xa28a002,0x200208a2,0x828,0x200a208,0x808028a0,0x20020028,0x88082000,0x20882822,0x828828,0x8800808a,0x20082,0x8820020,0x8820a,0x22800800,0x808220,0x80088002,0x20880,0x8208a08,0x8a200000,0x20802002,0x8020808,0x2080202,0xa00a0,0x228820,0x8a000082,0x80880820,0x28228008,0x2000008,0x228008a0,0x20808,0xa20a008,0x80002802,0x8000808,0x80008008,0x80000000,0x20208200,0x8000820a,0x220000a0,0x28820000,0x82000088,0x22,0x20028000,0x280080,0x808a0022,0x28020820,0x8020a000,0x28000a0,0x20208208,0x8008800a,0x802002,0x28,0x8288000,0x22822000,0x8800008,0x20a000,0x808a0820,0x200820,0x200000,0x2020020,0x220200,0x8a080280,0x200a0,0x28820208,0x800a200,0x82820000,0x20220028,0xa208202,0x80082000,0x8228,0x800a00a,0x80802);
remainders_pos[799].modulo_value=3196; remainders_pos[799].bit_mask=create_bit_mask(100,0x2a2882aa,0x82222a08,0x82828a88,0xaa022220,0xa0aa80aa,0x20288a00,0x88a02a02,0x20282202,0xa8000a20,0x2a00a800,0x8a88080,0x288a2a8a,0xa88a08,0xaa88a000,0xa8a88222,0x8a020082,0x82022002,0x8a80282,0x28800808,0x888288a2,0xa222822a,0x800082a,0xa2a08a88,0x8a880200,0xa828822,0x8a20a02,0x2a8aa200,0x82022800,0xa2aa82,0x8a0808,0x80220822,0xa2a2a220,0xa0a8022,0x8a80a8a,0xa288a82,0x8822020,0xaa082200,0xa202202a,0x80008220,0xa20a2022,0xaa02a80,0x2882aaa0,0xa80a0a8,0x8a80a820,0x2a8220a0,0x280008,0x28888a00,0x2202080a,0xa280008a,0xa0080,0xa0200a00,0x8a200028,0x8a020808,0xa2222,0x82000280,0x20a0882a,0x82a02a,0x82a0a02a,0xaaa822,0xa02a80aa,0x28808a08,0xa0882000,0xaa808808,0x8820a,0x808822,0x82a228a,0xa2a02a2,0xa8802a0a,0x2088a8a8,0x8820880,0x2020a20,0x282aa8a0,0x80028808,0x8aa2a,0x80a08a2,0x2882282a,0xa008022a,0x22a20a8,0xaa820002,0x2a882888,0x88a22822,0x2020022,0x282802a2,0x28008808,0xa820080a,0xa88822a2,0xa22a,0x820a22a0,0xa2a8a22,0x802022a2,0xa002a00a,0x808a0002,0x28088280,0x880a80a2,0xa00a2280,0xaaa02aa0,0x2088880a,0x222a2828,0x820a8888,0xaa8228a);
remainders_neg[799].modulo_value=3196; remainders_neg[799].bit_mask=create_bit_mask(100,0xa2a00a22,0xaa22280,0xa0a0200,0x220aaaa8,0x28220822,0xa0a00288,0x28a28a,0x28a0a28a,0x208882a8,0xa2882088,0x80200800,0xa002a202,0x88200200,0x22002808,0x202002aa,0x28a880a,0xa8a288a,0x80208a0a,0xa0008080,0xa0022,0x2a2a0aa2,0x808880a2,0x22280200,0x2008a88,0x20a00aa,0x802a828a,0xa2022a88,0xa8aa080,0x882a220a,0x88028000,0xaa80aa,0x2a2a22a8,0x828208aa,0x80200202,0x82a0020a,0x8002a8a8,0x2200aa88,0x2a0aa8a2,0x8880aa8,0x2282a8aa,0x8228a208,0x200a2228,0x82002820,0x20820a8,0xa20aa820,0x88a08880,0xa0000208,0xaa8a8082,0x2a080002,0x88828808,0x28a80288,0x2a888a0,0x2828080,0x8882aaaa,0xa088208,0xa82800a2,0x800a28a2,0xa2828a2,0x82220aa,0x28a20822,0xa0080200,0x2800a880,0x22080080,0x88800a02,0x880800aa,0x80a2a202,0x82a28a22,0x20082282,0xa8002020,0x80028008,0x8a8a82a8,0x20222028,0x88a0080,0x808022a2,0x8082802a,0x200aa0a2,0x28808aa2,0x82a2a820,0x220a8882,0xa200a000,0x2aa02a,0x8a8a88aa,0xa0a0822a,0xa0080080,0x20a80082,0x2000aa2a,0x88802aa2,0xa82aa28,0x822202aa,0x8a0aa2a,0x208a2882,0x802888a,0x20800a08,0x82082a,0x2882aa08,0x22282220,0xa8000082,0xaaa2a020,0xa820000,0x220a202);
remainders_pos[800].modulo_value=40; remainders_pos[800].bit_mask=create_bit_mask(2,0x80820282,0x82);
remainders_neg[800].modulo_value=40; remainders_neg[800].bit_mask=create_bit_mask(2,0x80a0a0a,0x2);
remainders_pos[801].modulo_value=1068; remainders_pos[801].bit_mask=create_bit_mask(34,0x2020822,0xa28000,0x8208288,0x22002802,0x20000828,0x8082282,0x82000880,0x8a00a08,0x8828a28a,0x20882802,0x8008800,0x220a208,0x220a0800,0x8200a00,0x8008a082,0x200000a2,0x8a20820,0x8082,0xa20028a0,0x820820,0xa08800a,0xa0880002,0x20020208,0x82208002,0xa2822802,0xa20a28,0x8220a,0x82020022,0x8828,0x80088282,0x20820802,0x28a00228,0x8080000,0x882);
remainders_neg[801].modulo_value=1068; remainders_neg[801].bit_mask=create_bit_mask(34,0x828a00a2,0x8220808,0x80280200,0xa2882082,0x28808020,0x8000a20a,0x2880000,0x208200,0x202202,0xa0002082,0x800008,0x8a282280,0xa2820080,0xa08208,0x800200a,0xa0880822,0x228028,0x8808000a,0x22882020,0x8020028,0x82000082,0x20000882,0x28828a00,0xa28008a,0x220a2082,0x8228220,0x8800a282,0x28a08a2,0x8800020,0x800020a,0xa0020082,0x20208a20,0x80008088,0x2);
remainders_pos[802].modulo_value=3208; remainders_pos[802].bit_mask=create_bit_mask(101,0xa28228a,0xaaa282a0,0xa220a08,0xa2a88a0,0xa88a2828,0x28828a28,0x88880200,0xa88a800a,0x8822288a,0xa2a08022,0xaaaa2222,0x800a0a08,0xa2008aa2,0xa0a0882,0xaaaa2220,0xa880a2a2,0x2aa802,0x8820022,0x8a880002,0x2a028028,0x282a0020,0xa0022a08,0xa0aaa08,0xa282a220,0x2a82a08,0xa02802a2,0x22028208,0xa000a0aa,0xa022aa08,0x2aa02828,0x2a82a02a,0xaaa88088,0x2aa288a2,0xa8002aa2,0x20a8802,0x22200002,0x88a0a0aa,0x82a202,0xa0a0aa82,0x22200008,0x2a8a0202,0x88222882,0xaa808800,0xa2a88880,0x2082882a,0x28208808,0x88020a8,0xa0a220aa,0x2820008,0x822820aa,0x8288200,0x8280aa,0x88a0a20,0x80220aa,0x2208282a,0x22820820,0x222a8aa8,0x2202aa02,0x28882200,0x80a2a882,0x88880,0xaa0a0a20,0x8a820082,0xa0a2280,0x888aa,0x22a08080,0xa8002a80,0x228aa88a,0x22aaa8a,0xa82a822,0x280aa8a8,0xaa880a28,0xa000a20,0x828088aa,0x80280a20,0xa82a808a,0x8a828a20,0xaaa0a008,0xa8800a20,0xa82820,0x280a808,0x22a228,0x822080,0x2aa80088,0x8a022a80,0x88aaaa8a,0x20a0a008,0xa2008a82,0xa0a0028a,0x88aaaa20,0x20a8a88,0x28882288,0x2a22aa2,0x802222a0,0xa2822800,0x28a22a28,0x22a8a028,0xa088a00a,0x828aaa20,0x8828a00a,0xa2);
remainders_neg[802].modulo_value=3208; remainders_neg[802].bit_mask=create_bit_mask(101,0x82a0aa02,0x222a0a28,0x82aa8280,0x82a20028,0x2002a0a0,0xa00a02a0,0x8a88,0x20020882,0xaaa002,0x2a2808aa,0x2222aaaa,0x8828280,0x2a88022a,0x8282800a,0x2222aaa8,0x20082a2a,0x88a2208a,0x800a88aa,0x200888a,0xa28a08a0,0xa0a288a8,0x288aa280,0x82822280,0x2a0a2aa8,0x8a20a280,0x28a08a2a,0xaa8a0a80,0x28882822,0x28aa2280,0xa228a0a0,0xa20a28a2,0x22200800,0xa22a002a,0x2088a22a,0x8a82008a,0xaaa8888a,0x282822,0x88022a8a,0x2828220a,0xaaa88880,0xa2028a8a,0xaaa00a,0x22080088,0x2a200008,0xa80a00a2,0xa0a80080,0x8008a820,0x282aa822,0x8a0a8880,0xaa0a822,0x80a00a88,0x880a0822,0x800282a8,0x808aa822,0xaa80a0a2,0xaa0a80a8,0xaaa20220,0xaa8a228a,0xa000aa88,0x82a200a,0x88800008,0x228282a8,0x20a880a,0x8282aa08,0x88800022,0xaa280808,0x2088a208,0xaa022002,0x8aa22202,0x820a20aa,0xa0822020,0x220082a0,0x828882a8,0xa080022,0x8a082a8,0x20a20802,0x20a02a8,0x22282880,0x200882a8,0x8820a0a8,0x8a082080,0x88aa2aa0,0x880aa808,0xa2208800,0x28aa208,0x222202,0xa8282880,0x2a08020a,0x28288a02,0x2222a8,0x8a820200,0xa000aa00,0x8a2aa22a,0x8aaaa28,0x2a0aa088,0xa02aa2a0,0xaa2028a0,0x28002882,0xa0222a8,0xa02882,0x2a);
remainders_pos[803].modulo_value=3212; remainders_pos[803].bit_mask=create_bit_mask(101,0xa22aa202,0x8028028,0x2028020,0x882800a2,0x82000282,0xa0a80a0a,0x20a822aa,0x8282020a,0xaaa22222,0x28022002,0x80aa8aa,0xa8802aa,0x820aa220,0x8a880a82,0x88aa00a0,0x20a22002,0x208aa880,0x220088a,0x82882020,0x20a2202a,0x2228a88a,0xa0888820,0x8220a082,0x8a2aa208,0xa0aa0282,0x8802802,0x8800020a,0x20288a02,0xa220280a,0x8828,0xa8802882,0x28a20800,0x22888aa8,0x20022082,0xa8082828,0x2a80020a,0xa20220,0x88020a0a,0xa20288,0x2000a280,0x888aa000,0x8880000a,0xa8a0808,0x8802082,0x82a02000,0xa2a8a00,0x2aa0820a,0x880a8a80,0xaa0a22a8,0x80008800,0x20002a0a,0x8a0aa002,0x2a0222a8,0x20aa802a,0x2a8a0a08,0x80a8200a,0x80220000,0xa2a0820,0x222202,0xaa222a00,0xa0008000,0x8a00028,0x8022228,0x8a00a0a,0x2a8088,0x8202aa08,0x88008282,0x22288820,0x8a282aa,0x8022a002,0x20000822,0x8088a282,0x22808a02,0x2280a,0x80220a08,0xaa0a802,0xaa8a2828,0xa0882208,0x2220a820,0xa2888082,0x88a08a22,0x82282a80,0x880080,0xaa208a22,0x88a08022,0xaa22800,0x22a20a0,0xaa08282a,0x22a0088,0xaa020aa8,0x88028aa2,0x88aaa800,0x8282888,0x82a08a08,0x2a0aaa8,0x82a0a,0x2822828,0x280808a0,0x28020080,0xaa88a280,0x808);
remainders_neg[803].modulo_value=3212; remainders_neg[803].bit_mask=create_bit_mask(101,0x2aa2228a,0x800a08a0,0x8a8a08a8,0xa08822,0xa880a0a,0x20208282,0xa020aa22,0xa0a8a02,0x2222aaaa,0x208aa88a,0x80822022,0x82008222,0xa022aa8,0x200820a,0x228820,0x282a288a,0xa0022008,0x8aa88002,0xa00a828,0xa822a8a2,0x2aa02002,0x280000a8,0xaa8200a,0x2222a80,0x28228a0a,0x8008a082,0x880a82,0xa0a0028a,0x2aa8a082,0x88880020,0x2000a00a,0x202a8088,0xaa000220,0xa88aa00a,0x2000a020,0xa2088a82,0x882a8aa0,0x8a0282,0x802a8a00,0xa8882a08,0x22808,0x8882,0x2028080,0x8008a00a,0xa28a088,0x82220288,0xa2280a82,0x820200,0x22822a20,0x880088,0xa888a282,0x282280a,0xa2822a20,0x282208a2,0xa2028280,0x820a082,0x82a8888,0x82a280a8,0x88aaaa82,0x22aa2288,0x20880888,0x802088a0,0x808aaa20,0x80208282,0x8a20800,0xa8a2280,0x88020a,0xaa2000a8,0x802a0a22,0x8aa2882,0xa80800aa,0x2a0a,0xaa08028a,0x888aa002,0x8a28280,0x228208a,0x2202a0a0,0x2800a280,0xaa2820a8,0x2200080a,0x2802a2,0xaa02208,0x80008808,0x22a802aa,0x28082a,0x8222a088,0xaa2a828,0x2280a0a2,0xaa28000,0x220a8220,0x8a022a,0x222080,0x80a02000,0x2280280,0x8a282220,0x8880a202,0x8a02a0a0,0x20808028,0xa08a8800,0x22002208,0x80);
remainders_pos[804].modulo_value=1608; remainders_pos[804].bit_mask=create_bit_mask(51,0x2080822,0xa20220,0x88000200,0xa28200a0,0x20800200,0x8228020a,0xa20a28a2,0x8228200,0xa280200,0x882802,0x28000200,0x82002202,0x208a2080,0x8800820,0x8000288,0x2822000,0x8028a08,0x8288200,0xa28a08a0,0x8028828,0x800208a,0xa00828a0,0x8000220,0x8808a000,0x82020800,0x20808228,0xa202200,0x80002000,0x28200a08,0x800200a,0x228020a2,0x20a28a28,0x8228200a,0xa2802000,0x8828020,0x80002000,0x20022022,0x8a20808,0x88008202,0x80002880,0x28220000,0x8028a080,0x82882000,0x28a08a00,0x8028828a,0x800208a0,0x828a00,0x8000220a,0x808a0000,0x20208008,0x88);
remainders_neg[804].modulo_value=1608; remainders_neg[804].bit_mask=create_bit_mask(50,0x828000a2,0x8228a28,0x88280,0x220a0820,0x28008a08,0xa208282,0x22822022,0xa20a08,0x82208288,0x80002082,0x20800a08,0xa08a28a,0xa0022800,0x8028,0x80088200,0x820a2880,0x820200,0x80200288,0x22020020,0x820020,0x8008a002,0x20802020,0x808a28,0x2088,0x28a0080,0x28000a20,0x8228a288,0x882800,0x20a08200,0x8008a082,0xa2082822,0x28220220,0xa20a082,0x22082880,0x20828,0x800a088,0xa08a28a2,0x228000,0x8028a,0x882000,0x20a28808,0x8202008,0x2002880,0x20200208,0x8200202,0x8a0020,0x8020208,0x808a282,0x20880,0x28a00800);
remainders_pos[805].modulo_value=3220; remainders_pos[805].bit_mask=create_bit_mask(101,0x2808220a,0x222088a0,0x2820288,0x2228800,0x80802202,0x8a82a80,0x88280280,0x88a8020,0x8a2,0x80002,0x802a2802,0x2200a02,0x8002008,0x88,0xa000022a,0x8280802,0xa2008282,0x2a0a0,0x88a20802,0x82000a28,0x20220280,0x82028000,0x2082000,0x2200a8,0x220800a,0x8220202,0x2200280,0x208200a8,0xa020800,0x2202000,0x8002080a,0x822888a2,0x2a000200,0x8022828,0x80a0820a,0x2a00,0x2a2,0x20008088,0x80220080,0xa2a00a02,0x800200,0x80000200,0xa888228,0xa08820,0xa0a8800a,0x2008080a,0x8a220202,0x20a0000,0x8822208a,0x2080a028,0x20882a82,0x822280a0,0x80a2088,0x8a20000a,0x880808,0xa0aa0202,0xa00a0022,0x2a008220,0x228822,0x20000800,0xa8a00800,0x80280a00,0x802008,0x22020,0x8a800,0xa0200a80,0x20a0820,0xa828288,0x88200800,0x28a222,0x880a0208,0xa000080,0x20800208,0x8802a008,0x82002800,0x88080808,0x800a0020,0x802a008,0x8200082,0x80800028,0x8202808,0xa2228a00,0x80208,0x8a0a0a8,0x82082820,0xa80202,0xa8800,0x2022000,0x88020080,0x80280a00,0x8028a,0x80002,0x2208a200,0x8220802a,0xa2002802,0x20202a82,0x88080880,0x28000228,0x88822808,0x280a220,0xa0882);
remainders_neg[805].modulo_value=3220; remainders_neg[805].bit_mask=create_bit_mask(100,0xa0002a02,0x2a280020,0x20a8200,0x8a220088,0xaa0a,0x80202208,0x200200,0x20028,0x8808022,0x8080800a,0x822208a,0x8a208282,0x882000,0x88808000,0x20888a22,0xa00082,0x22080a02,0x80a2028,0x22808a,0x2880220,0xa0aa0208,0xa0a0800,0xa80a080,0x80228820,0x82280802,0x80a20a8a,0x8a200a00,0x280a8020,0x82828008,0x2a82800,0x8a0082,0xa200022,0x22808288,0x808220a0,0x280202,0x88882208,0x8088a22,0x20880000,0x8228800,0x2220820a,0x8080a88,0x8088280,0x2000220,0x2000a0,0x20200802,0xa8000082,0x2228a82,0xa820088,0xa2a002,0xa0002820,0x28002202,0x22a0820,0x22000,0x2a08882,0x80000000,0x2822028a,0x28028822,0x228802a0,0x80a2002a,0xa0088008,0x20200088,0x8200280,0x808a080,0x882a828,0x802000,0x20a80208,0x820280a0,0x2020200,0xa00088,0x8020222a,0x820a00,0x2088800,0x20008280,0x22880,0x280a008,0x800080,0x80280a0,0x8a2000,0xa08002,0x88820,0x80282080,0x2a220200,0x8800280,0x80202020,0x280a020,0x8200a82,0x20080,0xa0aa088,0x828808,0xa00208,0x88800202,0x88008802,0x2a002280,0x2a008a2,0x2208a00a,0xa8a0220a,0x208882a0,0x22000,0x2082a20,0x2000a);
remainders_pos[806].modulo_value=3224; remainders_pos[806].bit_mask=create_bit_mask(101,0x22800aa2,0xa2aaa008,0x20aa280,0xaa0aa808,0x22020a,0x228002,0x88a80,0xa0220a02,0x282822a2,0xa822802a,0x20800a08,0x8a2aa,0x8222a822,0x2a0088aa,0xa0880a22,0x802a8200,0x2a8202a2,0xa0aa0a2a,0x200088a8,0x882a8880,0xa82a80a8,0x8220a220,0x2a20a080,0xa222a08,0x20882a00,0x2882202a,0x22208a80,0xa202820,0x2228a80,0x80080a22,0x80280808,0xaaa2888,0x800a0800,0x82002020,0x80288202,0xa08a02a2,0xaa02200,0x228008,0x8a2a228,0xaa0a0002,0x82280820,0x20082028,0xa20a0202,0x88a222a0,0x822aa880,0x8228a2a2,0x2000a2,0xa028a8,0x20282,0x2a82088a,0xa8008200,0xa22082,0x828080,0x2a280a,0x828a0008,0x828a8a28,0x22022aa8,0x8a0a888a,0x88080a0,0x82280820,0xaa082028,0x208000a0,0x288a8a,0xa0200288,0xa00880a,0x28a80a2,0x82808228,0x2080800,0xa00020a0,0x22228aa,0x2202028,0x8088a020,0xa002a288,0x88082808,0x2802a208,0x8a80882,0xa000a822,0xa820a888,0x82020a08,0x2a088a08,0x222a02a8,0x80222a8,0xa2a2200,0xa8a8a0aa,0x28a8082,0xa0082a8,0xa888a022,0x82aa2200,0x882a88,0x8aa8a20,0x2a20a002,0x28a80288,0xa8a8828,0x80a088,0x2a220,0x800288,0xaaa08088,0x80202aa0,0x8a028aa0,0x88200aaa,0x8aa002);
remainders_neg[806].modulo_value=3224; remainders_neg[806].bit_mask=create_bit_mask(101,0x2a08822a,0x2a222800,0x82822a08,0x22822080,0x80aa0a82,0x88aa088a,0x88800200,0x282a828a,0xa020aa2a,0x20aa00a2,0x28088280,0x88802a22,0xaa220aa,0xa2880022,0x2800822a,0xa20a88,0xa20a0a2a,0x282202a2,0xa8880020,0xa20000,0x20220020,0xaa82aa8,0xa2a82008,0x2aaa280,0xa800a208,0xa002a8a2,0xaaa80208,0x82a8a028,0x82aa0200,0x88082aa,0x8a00080,0x222a000,0x8828080,0xa08a8a8,0x8a00a8a,0x2002822a,0x228aa88,0x88aa0880,0x80222aa0,0x2202888a,0xaa08028,0xa080a8a0,0x2a828a8a,0x222a28,0xaa22008,0xaa02a22,0x8828882a,0x88282020,0x888a820a,0x220a8002,0x20880a88,0x8822a00a,0x880a0808,0x88a2a002,0x2028880,0xa020220,0xaa8a2220,0x2820002,0x80080820,0xa2080a0,0x2280a8a0,0xa8088028,0x8a00202,0x28a88a00,0x82800082,0x8a02082a,0x2080a20,0x82808088,0x2888a828,0x8aaa2022,0x8a28a8a0,0x80028a0,0x280a2a00,0x80a080,0xa0822280,0x20800a,0x288820aa,0x20a02000,0xa8a0280,0xa2800200,0xa2a28a20,0x808aaa20,0x82a22288,0x20202822,0x8a020802,0x82080a20,0x2000282a,0xa22a288,0x800a200,0x802202a8,0xa2a02882,0xa0208a00,0x82020020,0x82800,0x888a2aa8,0x88080a00,0x22280800,0xa8a220,0x20a0228,0xa88222,0x2208a);
remainders_pos[807].modulo_value=3228; remainders_pos[807].bit_mask=create_bit_mask(101,0xa28a2882,0x20028020,0x8002200,0x2820022,0x8a28,0x8800820a,0xa2822820,0x20808000,0x208828a,0xa2000882,0x28a08228,0xa008208,0x808a0880,0x20220a28,0x800a200,0xa0082802,0x8828a28,0x8a082082,0x80802800,0x28220008,0x88080000,0xa00800a0,0x200208,0x2080288,0xa2820082,0x8000208,0x8208820a,0xa08a2002,0x20a00208,0x8220200a,0x808028a2,0x28220228,0x82002280,0x2020820,0x20a00000,0xa208280,0x80882880,0x220000,0xa00a202,0x20002082,0x28800008,0x2088202,0x22080000,0x8000a28,0xa202080,0x228a0082,0x8028000,0x8a20808a,0x2882880,0x828,0x80200,0x2820000,0x20228228,0xa20208a,0x80002802,0x8200a28,0x20808a,0x828a0002,0x208,0x88082208,0x82000022,0x8208000,0x808a00a,0x20000880,0x228220,0x8028208a,0xa0,0x20820808,0x80288008,0x22880882,0x28a28020,0x8a008088,0xa20800a0,0x28008a20,0xa082208,0xa2080002,0x8200828,0x2280208,0xa2080080,0x20a00200,0x80000202,0x22000882,0x20028020,0x820820a,0xa28a2822,0x8028200,0x8008a002,0x228a0880,0x220a20,0x8208200a,0xa28820a2,0x8220008,0x8a282208,0xa0002020,0x20828828,0xa082002,0x28a2000,0x8800828,0x80088002,0xa0802800,0x8228a28);
remainders_neg[807].modulo_value=3228; remainders_neg[807].bit_mask=create_bit_mask(101,0x22022002,0x28820828,0x8008a288,0x820a08a2,0x8800220,0x80282,0x220a20a0,0x28000808,0x8a000202,0x22880002,0x20200a20,0x82080280,0x20000,0x28a28220,0x80082288,0x20802082,0x20220,0x200a00a,0x82080,0x20a28800,0x8088,0x20800820,0x8a08a00,0x8a008200,0x220a0802,0x808a00,0xa000282,0x20022882,0x28208a00,0xa28a082,0x82022,0x20a28a20,0xa08a208,0x828a00a0,0x28208808,0x82280208,0x2000,0x8a28808,0x8208228a,0xa0882802,0x20008800,0x8a00028a,0xa2800880,0x808220,0x8228a008,0xa2020802,0x820808,0x2280002,0x82002000,0x8808020,0x88008288,0x820a0880,0x28a20a20,0x8228a002,0x882082,0xa08220,0x88280002,0x2020882,0x808a00,0xa280,0x28808a2,0xa00808,0x80002082,0xa0880000,0x8a20a28,0x820a002,0x80880820,0x28028000,0x8200080,0xa2000002,0x20220828,0x2080000,0x22800820,0x20800228,0x8200a280,0x22800882,0xa08020,0x8a208280,0x22800800,0x28208a08,0x808828a,0xa2880002,0x28820828,0x80280282,0x220220a2,0x820a08,0x800208a,0xa2020000,0x8a28228,0xa00a082,0x22002822,0xa28800,0x2202280,0x208828a0,0x28020020,0x8200a08a,0x82022880,0x8020,0x800008a,0x20082080,0xa20220);
remainders_pos[808].modulo_value=808; remainders_pos[808].bit_mask=create_bit_mask(26,0xaa820a0a,0x282a800a,0xa88a88aa,0x200220a,0x82282a20,0xa822208,0x80a0a20a,0x2288a222,0xa2aa2028,0x8800222a,0x28200880,0x8000aa80,0xa00a0a2,0xaa00028a,0x20082802,0x88002202,0x8aa8aa8,0x8a228828,0x8a0a0288,0x8882a0a0,0xa8288220,0x88008008,0x22a22aa0,0x2a828aa,0xa082aaa0,0xa0);
remainders_neg[808].modulo_value=808; remainders_neg[808].bit_mask=create_bit_mask(26,0x220a8282,0xa0a20882,0x20020022,0x8a88aa82,0xaa0a2a8,0x820aaa80,0x8282a82,0xaa002aaa,0x2a22a8a0,0x882aa2,0xa0a88008,0x8882208,0x8288282a,0x22888a02,0xa880a08a,0x88aa8a,0x80220220,0x2aa00a0,0x2828a00,0xa2828,0x20a00aa8,0x880880,0xaa2aa220,0x8a20a022,0x280a2228,0x28);
remainders_pos[809].modulo_value=809; remainders_pos[809].bit_mask=create_bit_mask(26,0x169d67b6,0xa11e695b,0xbec39347,0xee2ba176,0xc1253035,0xcd5efaa7,0x44a99794,0xd6f406c7,0x7200511,0xf8a32e13,0x7de44437,0xd87133f4,0x4bb74b18,0x32386c63,0x889ef8bf,0xd3147fb0,0x80138321,0x80bdae22,0xa6548b8d,0x7deacca7,0x32920f95,0x1751deb0,0x270df5ba,0x59e2178b,0x9ae5a36a,0x1b7);
remainders_neg[809].modulo_value=3236; remainders_neg[809].bit_mask=create_bit_mask(101,0x8a00aa2a,0x2882a082,0x220a0a8a,0x22a228aa,0x8a8a8a8,0x82222a,0x88200a08,0xa288a0a,0x8aa88888,0x202aa28a,0xa02888aa,0xa8aa28,0x822a8280,0xaaa0a0aa,0x22022,0xa88a228,0x88a0aa8,0x82022aa,0x2a880200,0xa022002a,0xaa0a8208,0x8a880228,0xaa802822,0x802a8a02,0x2282aa2,0xa202a02a,0xa00a2aa0,0xaaa028a,0x8a8a20a0,0x82a8a200,0x2028820a,0x8aa2a002,0x20800200,0x8882aa20,0x8a80aa82,0xa228,0x2aaa2022,0xa20aa828,0xa8000a0a,0xa028a2a8,0xa022a888,0xaa8a8a2a,0xa2808888,0x208a8288,0x8008280,0xa882a222,0x2a20a8a8,0x8222a8a2,0x8a28a82,0x28a082a,0x202a2a8,0x28a082a8,0x808208a0,0x8022282,0x80a2022,0x22028808,0x828aa8a2,0x82808a2a,0x888a8208,0x20808008,0x88022a00,0x2082a08,0xa0aaa808,0x2800a200,0x2a200028,0x220aaaa2,0x800a8088,0x20028882,0xa2aa8a2a,0xaa02008a,0xa2882a22,0xa2080280,0xa20808a,0x82a000aa,0x20aa00,0x2202a20a,0x2822a0,0xa0802a82,0x282a8002,0x22a88082,0xa280a008,0x2aa22a08,0xa2a88020,0x22a28aa,0xa088a0,0x220880a8,0x2a22aaa8,0xa0a0002,0x82802280,0x20080aaa,0x8882a08,0x82002a20,0x88880080,0xa08820a8,0xa0aa2880,0x22228aa8,0x80808a0,0x8220228,0x80a0a220,0x8a028820,0x200aa082);
remainders_pos[810].modulo_value=120; remainders_pos[810].bit_mask=create_bit_mask(4,0x80002002,0x220a20,0x208a088,0x800800);
remainders_neg[810].modulo_value=120; remainders_neg[810].bit_mask=create_bit_mask(4,0x882882,0x8228220,0xa002000,0x80);
remainders_pos[811].modulo_value=3244; remainders_pos[811].bit_mask=create_bit_mask(102,0xaa80822a,0x22a220,0x2a8aaaa0,0xaa88a802,0x880aaaa8,0xa282200,0x282aa020,0xaa82a82,0x2822aa0,0x280202a,0xa02288a,0x20822a28,0xa8228a02,0x2288a8a2,0x22222082,0x288008a0,0x82820280,0x820a880a,0xa08a8000,0x8a8a22a0,0x8a828288,0x28a8202a,0xaa220aaa,0x82020822,0x2aa880a0,0x8a0282,0x8a282a20,0x8008a28a,0x20800288,0x200a8a0,0x2220a0a,0x28280882,0x8a2aaa02,0xa02a202,0x20aa0a8a,0x88a822aa,0x88a2888,0xa0082022,0x288a8208,0x82a0808a,0xa82208a,0xa80a8aa,0x2a8008a,0xa08a00a0,0x82a8a800,0xa22a88a0,0x200002,0x8aa022,0xaaa80002,0x288808,0x20008a2a,0x22800a8a,0x2aaa202,0xaa200800,0x800880,0x2a88a800,0xa2a820a2,0xa20a002,0x2a800a0,0xa02a0a20,0x882a0aa2,0x20a82a20,0x2a228a20,0x8200a208,0x8a220880,0x82a22222,0xaa08aa8,0xa80a0a2a,0xaa8a2808,0x282880a,0x8880822,0xa0080a0a,0x2080a2,0xa2002228,0x828a2a28,0xa20008a,0x22aa8828,0x80820a0,0x88aa882,0x82a28aaa,0x282a2a80,0x8a2a2228,0x2a20a0a8,0x2a082000,0x8282a02,0x228028,0x888880a2,0xa2288820,0x2882a8a2,0x8820880a,0x880a028a,0x80280a22,0x88280a80,0x82aa00aa,0xaa82882a,0x828a0080,0xaa022008,0xa22aa2aa,0xaa2a8802,0xa88000aa,0x202aa088,0xa88);
remainders_neg[811].modulo_value=3244; remainders_neg[811].bit_mask=create_bit_mask(102,0x22080aa2,0x88aa2aa8,0xa2022228,0x2200208a,0x822220,0x82a0aa88,0xa0a228a8,0x8220a20a,0x8a0aa228,0x8a08a8a2,0x828aa002,0xa80aa2a0,0x20aa028a,0xaa00202a,0xaaaaa80a,0xa0088028,0xa0a8a08,0xa820082,0x28020888,0x202aa28,0x20a0a00,0xa020a8a2,0x22aa8222,0xa8a80aa,0xa2200828,0x8802820a,0x2a0a2a8,0x8802a02,0xa8088a00,0x8a882028,0x8aaa8282,0xa0a0800a,0x2a2228a,0x828a2a8a,0xa8228202,0x20aa22,0x8002a000,0x2880a8aa,0xa0020a80,0xa280802,0x820aa802,0x82082022,0x8a208802,0x28028828,0xa202088,0x2aa20028,0x88a8888a,0x880228aa,0x2220888a,0x88a00080,0xa88802a2,0xaa088202,0x8a222a8a,0x22a88088,0x88088008,0xa2002088,0x2a20a82a,0x82a8288a,0x8a208828,0x28a282a8,0xa2822a,0xa820a2a8,0xa2aa02a8,0xa882a80,0x2aa8008,0xa2aaaaa,0x82280220,0x208282a2,0x2202a080,0x8a0a0082,0x800080aa,0x28808282,0x88a8082a,0x2a88aaa0,0xa02a2a0,0x82a88802,0xaa2200a0,0x8080a828,0x8002200a,0xa2a0222,0xa0a2a208,0x2a2aaa0,0xa2a82820,0xa280a888,0x80a0a28a,0x88aa08a0,0x82a,0x2aa000a8,0xa00a202a,0xa80082,0x828a02,0x8a082aa,0xa08208,0xa228822,0x220a00a2,0xa028808,0x228aa880,0x2aa22a22,0x22a2008a,0x20088822,0xa8a22800,0x200);
remainders_pos[812].modulo_value=1624; remainders_pos[812].bit_mask=create_bit_mask(51,0x2028a02,0x2a200a80,0xa20a8222,0x8208a0aa,0x2a202880,0x8a0002a8,0x802802a0,0x80800a2a,0xa282080,0x822,0x80088802,0x88a20288,0xa0a0208,0x8a22208,0x22002228,0x802,0x828a0882,0x202020,0x2802a8a,0xa20a8,0x808a82a8,0xa2082022,0x2a08aaa0,0x8a8888,0x2808002a,0x28a0280a,0xa80020,0xa82222a2,0x8a0aaa20,0x2880820,0x2a82a2,0x802a08a0,0xa2a802,0x82080808,0x8220a2,0x88802000,0x20288800,0xa020888a,0x222080a0,0x222808a,0x802220,0xa0882000,0x2020828,0x2a8a002,0xa20a8028,0xa82a8000,0x82022808,0x8aaa0a20,0xa88882a0,0x8002a008,0x80a280);
remainders_neg[812].modulo_value=1624; remainders_neg[812].bit_mask=create_bit_mask(51,0x8a8a020a,0x22a88200,0x22020aaa,0xa002822,0xa2a0a008,0x2880a20,0x8a08228,0x8088222,0x2a0a800,0x8880aa,0x800008a,0x228a00,0x82820a80,0x8022a280,0xaa88aa20,0x8888082,0x202800a,0x8828a828,0x8a00a202,0x88822820,0x20220,0x2a80a82a,0x22802220,0x80020000,0xa00008a2,0xa020a082,0x882008a8,0x20aaa22a,0x2822220,0xa0080a0,0x80a20a2a,0x8228028,0x8822208a,0xa800080,0x880aa02a,0x8a008,0x28a00080,0x20a80002,0x2a280828,0x8aa20802,0x88082aa8,0x2800a088,0x8a828020,0xa202882,0x228208a0,0x20220888,0xa82a000,0x22202a8,0x20000228,0x8a2800,0x82a00);
remainders_pos[813].modulo_value=3252; remainders_pos[813].bit_mask=create_bit_mask(102,0xa2800082,0x28828020,0x8020808a,0x20208a0,0xa08a08,0x8800a20a,0xa28a2820,0x8828028,0x228a,0x20080002,0x28800208,0x2200,0xa2022022,0x28020800,0x82202088,0x8a0020,0x220800,0x8208a20a,0xa0082,0x8820a00,0xa282080,0x22020020,0x8a20228,0x2080,0x208a08a0,0x20820208,0x288288,0x80082000,0x20a20228,0x80200202,0x800820a2,0x28208,0xa080280,0x2822882,0x8220,0x8200288,0xa0882022,0x28008200,0x2202202,0xa0022000,0x20820008,0x2000080,0x828a2000,0x28a00a08,0x8820a28a,0x2822800,0x828828,0x8288202,0xa2880820,0x28200a08,0xa08000a,0x828a0002,0x28a20a00,0x82008202,0x20080822,0x28028228,0x82200288,0xa28a28a0,0x28220a00,0x800008a,0x20802000,0xa20008,0x88000088,0x2880880,0x20a00820,0x82088082,0x22800,0x28008820,0xa082288,0x2802,0x220828,0x8028a082,0xa0880800,0x228808,0x80000082,0x20822822,0x20820808,0xa20a,0xa2002080,0x8828808,0x8a008008,0x22002082,0xa08,0x828200a,0x800a08a2,0x20000208,0x8820800a,0x2822080,0x8a00208,0x88088,0x22800880,0x820800,0x80002,0x220a2880,0x28a28028,0x220828a,0xa00a08a0,0x8020a20,0x8020a208,0x228a2020,0x28a08028,0x82000);
remainders_neg[813].modulo_value=3252; remainders_neg[813].bit_mask=create_bit_mask(102,0x22080802,0x20020828,0x8280002,0x828a0020,0x8200200,0x82282,0x220220a0,0x20820,0x88082202,0xa0800882,0x20008a00,0x8808a288,0x228a28a2,0x20828008,0xa28a000,0x800208a0,0x8a28008,0xa002282,0x80820802,0x28208,0x8220a008,0xa28a08a0,0x228a20,0x8808a008,0xa0020020,0x28028a00,0x88200200,0x802880,0x28228a20,0x828828a,0x802822,0x8820a00,0x82008208,0x820a2002,0x8800a28,0x80288200,0x200028a2,0x20800a08,0x8a28a28a,0x208a2880,0x28028800,0x8a088008,0x2022080,0x20208200,0x282202,0x820a2080,0x8020020,0x8020028a,0x220000a0,0x20a08200,0x82008082,0x2020882,0x20228208,0xa08028a,0xa08000a2,0x20820a20,0xa288200,0x22022020,0x20a28208,0x80088002,0xa0082880,0x8228800,0x88000,0x82000000,0x28208028,0xa00000a,0x808a2080,0x20800028,0x8200a200,0x80882082,0x8a28020,0x820200a,0x20000080,0x8a20000,0x808800a,0xa00a20a2,0x28028000,0x88082282,0x22882800,0x20000,0x2080080,0xa2882802,0x8808200,0x8020a082,0x820022,0x28808a00,0x280082,0x820a2800,0x208a00,0x88000000,0xa2080000,0x8028008,0x8800808a,0xa2822000,0x20220820,0x8a280202,0x20820020,0x828228,0x8282280,0xa20228a0,0x20200820,0xa088);
remainders_pos[814].modulo_value=3256; remainders_pos[814].bit_mask=create_bit_mask(102,0x8aaa82aa,0x820a2808,0xa0220828,0xa08802a0,0xaa802aa,0xa000aa02,0x28082a2,0x802a000a,0x2a808080,0x8a0a022,0x2a808008,0xaaa002a0,0xa2082a8,0x8a220228,0x8228a20,0xa000088,0x22aa8a,0xa2880008,0x828a008,0x8822a28,0x220a8a2,0xa2228,0xaa8a0228,0xa0280280,0x22200082,0x28222220,0x2822a22,0x282a80a,0x8820a080,0x8822202a,0xa2000a22,0x28202008,0x88a0828,0xa8a8a082,0xa0000228,0xa08a8aa0,0x88208220,0xa8222220,0x28a00,0x2a00aa00,0x88aa0a00,0x22220a08,0x88a8880,0xa08aa02a,0x2028282,0xa2200aa0,0x28800,0x8028882,0x8828222a,0x80282082,0x8000,0x2000200,0x22820828,0x20a88828,0x822280,0x8a002280,0x800aa008,0xa828280,0x20a80aa2,0x880222a2,0x2220a088,0xa800a0aa,0xaa00,0x2a00a280,0x22088888,0xa088208,0xa0aa2a2,0x2a288000,0x20820a2a,0x282820a2,0x8a200808,0x2288a000,0x22a80888,0x80020a08,0x80a02a82,0x2888a882,0x88088888,0xa820008,0xaa028028,0x2880a2,0x802888a0,0x208a2a08,0x2028a882,0x8a200a28,0x200022,0xa0a2aa88,0x20220000,0xa208a288,0xa0288088,0xaa2a8208,0xa80a800a,0x20200202,0xa8880a0a,0x2020202,0x80a000a8,0xa8a8202,0xa080aa00,0xaaa802a,0xa0a8022,0x82282088,0xa22028a0,0xaa82aa);
remainders_neg[814].modulo_value=3256; remainders_neg[814].bit_mask=create_bit_mask(102,0x2220222,0xa02a080,0x28aa80a0,0x28000a20,0x82200a22,0x2088228a,0x8a080a2a,0x8a28802,0xa2000800,0x2828aa,0xa2080880,0x22288228,0x82280a20,0x2aa8aa0,0x80aa02a0,0x82880800,0x80aa2202,0x2a008080,0x80a02800,0x8002a2a0,0xaa8202a,0x882aaa0,0x220282a0,0x28208a08,0xaaa8880a,0xa0aaaaa0,0x8a0222aa,0x820a2082,0xa82808,0xaaa822,0x2a8082aa,0x20a8a800,0x800280a0,0x2020200a,0x28088aa0,0x20020228,0xa80aa0,0x20aa2aa8,0x808a0288,0xa2882288,0x220208,0xaaa28280,0x20008,0x280228a2,0x8a8a020a,0x2a288220,0x888a0088,0x808a0002,0xa02aa2,0x20a80a,0x88880888,0x8a888a08,0xaa0280a0,0x282000a0,0x880aa208,0x288a208,0x8022880,0x820a0a08,0x28208222,0x8a2a2a,0xa2a82800,0x20882822,0x88882208,0xa2802a08,0x2a800000,0x82800a80,0x8282222a,0xa2200888,0xa80a8222,0xa0a0a822,0x2a80080,0xa2002888,0xa2208000,0x88a8200,0x820a20a,0x2000200a,0x800000,0x820a0080,0x220a08a0,0x88a0082a,0x8a00020,0xa8022280,0xa0a02002,0x2a882a0,0x88a8882a,0x28222200,0x282a8888,0x2a802a00,0x28a00000,0x22220a80,0x20820882,0xa8a88282,0x20000282,0x828a8a8a,0x8288820,0x2020a0a,0x28002288,0x22208a2,0x828208aa,0xaa0a000,0x2a20a028,0x220a22);
remainders_pos[815].modulo_value=3260; remainders_pos[815].bit_mask=create_bit_mask(102,0x802a2a02,0x28028a20,0x8a8220,0x2200008,0xa2008a00,0xa8000a82,0xa02a0a02,0x808288,0x2a28882,0x20a8802a,0x82a288a,0xa00a8222,0x2888208,0x820280a0,0x82202028,0x820882,0x8a2a88a0,0x808a028,0x80a0a000,0x20a0aa28,0x882a0a82,0xa2008080,0xa0a80a8,0x88a2a880,0x82888a00,0x20220888,0x2a020202,0x2a000020,0x220082,0x22282a08,0x88020082,0x20000820,0x2a882028,0xa020a2,0xa0200002,0x8020a0a,0x20288882,0x280a0288,0x282200a,0x80880228,0x80200082,0x802a0020,0xa020008,0x202282a,0x2808028,0x22a80,0xaa020a80,0x228aa088,0x8882002a,0x20800a20,0x880008a,0x8a200022,0x208a0020,0x8a800822,0xa220aa28,0x2a080a,0x2a8800,0x2802028,0xa828808,0x2200080a,0x20800a80,0x28200080,0x2880220,0x8a008828,0x82280a02,0x8222280,0xaa0a0802,0x8000080,0x88a080a0,0x8280822a,0x20820000,0x88200802,0x20a8288,0x88200880,0x8080000a,0x8808080a,0x22220880,0x200a2228,0x22a8a2,0xa2a02a0a,0x20202008,0x882a0a82,0x228aa0a0,0xa0a0,0x2280a202,0xa22a8a,0x28220820,0x22808088,0xa02808,0xa2082228,0x8882a00,0x8a228a82,0xa8022a0,0x82228a8,0xa2282020,0xa80a0a80,0xa82a0002,0xa2008,0x2000088,0x80882a20,0x208a2802,0x80a8a80);
remainders_neg[815].modulo_value=3260; remainders_neg[815].bit_mask=create_bit_mask(102,0x8a2228a,0xa00a02a0,0x80202a8,0x8aa08880,0x22880208,0x20880202,0x28228282,0x8080200,0x8a22000a,0xa0200822,0x80a22002,0x28020aa2,0xa000280,0xa820828,0x2a8a820,0x880a000a,0x2220020,0x8020a0,0x8202888,0xa0282220,0xa2020a,0x2a080800,0x2820020,0x222008,0x2000208,0xa8aa0000,0xa20a8a82,0x228880a8,0x88a2880a,0xa2a0a200,0x8a080a,0xa80880a0,0x2200a0a0,0x8820a82a,0x20a8880a,0x808a0282,0xa8200002,0x20828200,0x8a02a882,0x8a20,0x8a8080a,0x82288a0,0x28a8080,0x8a82a0a2,0x82080820,0x808a2208,0x220a8200,0x2a022000,0x288a2,0xa0088228,0x80080802,0x22888a2,0x280280a8,0x20080aa,0x22a82220,0x88a20082,0x88220080,0xa08a0a0,0x82020080,0xa2888002,0xa8080208,0xa0288800,0xa0082a8,0x28000a0,0x2a0820a,0x80aa2a08,0x22028002,0x888008,0x200828,0x2080a22,0xa80a0888,0x288082,0xa820200,0xa08008,0x88802,0x800082,0xaa2a8000,0x2882a2a0,0x88a2202a,0x2228a202,0xa8a82880,0x228202,0x2a022028,0x88802828,0xa2082a0a,0x882a2202,0xa02280a0,0x2a080000,0x8820a080,0x2280aa20,0x80002288,0x22a0202,0x208a228,0x80a2a020,0x22a0a828,0x20820208,0x20228882,0x882a080,0x8a808800,0xa228,0xa802208a,0x20200);
remainders_pos[816].modulo_value=408; remainders_pos[816].bit_mask=create_bit_mask(13,0xa20020a2,0x8028208,0x8088002,0x22020000,0x28020800,0x8008a208,0x20a28a0,0x28208a20,0x88002080,0x20000080,0x20800220,0x8a208280,0x8a0800);
remainders_neg[816].modulo_value=408; remainders_neg[816].bit_mask=create_bit_mask(13,0x22882822,0x820a00,0x8000008a,0xa20a0880,0x20828008,0x2280,0x82822020,0x20a00228,0x8a008,0xa0880800,0x28008a20,0x2280208,0x20000);
remainders_pos[817].modulo_value=3268; remainders_pos[817].bit_mask=create_bit_mask(103,0x2a820a0a,0xa802a,0xa8a08a2,0x2028228,0xa882800,0x8080200,0x8228a028,0x2a0a802a,0x880a8a22,0x88828282,0x8a800,0x8028a882,0x22882882,0xa020aa2,0x80222822,0x88002020,0x2aa20a28,0x80022202,0x82088200,0x8a222aa0,0x2800880a,0x2002a2aa,0x22aa820,0x8882aa80,0x8a02a02a,0x22a8aa22,0xa02a028a,0xaaa088a,0x20aaa200,0xaa2a0020,0x808800a2,0x2aa2228a,0x2088208,0x2220008,0xa2822aa2,0x20200088,0x20a22008,0x2a820282,0x8a08a22,0x8a8a00a,0xa88002,0xa0a0888,0x228a808a,0xa00a82a2,0xa028a20a,0x2008080,0xa08a80,0xa20a0200,0x28828a80,0xa00a8002,0x82820aa2,0xaa08282a,0x2a00a8,0x2a282288,0x80a08a0,0x2a20a000,0x20200800,0x8a280a0,0xa82a00aa,0x202a2888,0x220a0a0a,0x22a002,0xa2a208,0x8a20a20a,0x28082a88,0x88a088,0x20008082,0xaa8828a2,0x88808,0x8220802,0x2888aa82,0xa002202a,0x800a8aa8,0x8aaa080,0x220aaa00,0x280a80aa,0x8aa2a88a,0x80a80a28,0x2aa8222a,0x82aa8800,0xa8a80080,0x220028a,0xaa888a2a,0x8220820,0x8880020,0x8a08aa88,0x80800222,0x82888020,0xaa080a08,0x22822888,0x22a28028,0x2a20008,0x28282220,0x8a2a0228,0x802a0a88,0x80a2882a,0x8020202,0x2822a00,0x88280800,0xa20a2a02,0x802a0008,0xa082a8a,0xa);
remainders_neg[817].modulo_value=3268; remainders_neg[817].bit_mask=create_bit_mask(103,0xa2028282,0x888200a2,0x202802a,0x8a8a0aa0,0x8200a088,0x80808288,0xa2028a0,0xa20208a2,0x8202aa,0xa0a0a,0x88802080,0x8a0200a,0xaa002002,0x828a822a,0xaaa0aa,0x88a8a8,0xa22a82a0,0x88a2a0a,0xa800a88,0x2a2a228,0xa0880082,0x288a2a22,0x82a220a8,0xa2208,0x28a20a2,0xaa2022aa,0x28228a02,0x82228002,0xa8222a08,0x22a288a0,0x800882a,0xa22a2a02,0x8a800a80,0x82a28880,0x2a0aa22a,0xa8a88800,0xa82aa800,0xa20a8a0a,0x2002aa,0x80202882,0x820088a,0x82828000,0xaa020802,0x2882022a,0x28a02282,0x8a080808,0x88280208,0x2a828a88,0xa00a0200,0x2802088a,0xa0a022a,0x2280a0a2,0x80a28820,0xa2a0aa00,0x80828028,0xa2a82800,0xa8a88088,0x80220828,0x20a28822,0x28a2a000,0xaa820282,0x88aa288a,0x882a2280,0x2a82a82,0xa000a200,0x80002800,0xa888080a,0x2200a022,0x88800080,0x80aa008a,0xa000220a,0x208aa822,0x8820220,0x80222808,0xaa822208,0xa0820822,0x2222002,0x82082a0,0x2220aaa2,0xa220088,0x20208808,0xaa88202,0x220002a2,0x802a80a8,0x800088a8,0x2802200,0x8088aa2,0xa0000a8,0x22800280,0xaa0aa000,0xa22a08a0,0x8a2a8880,0xa020aaa8,0x2a28a20,0x8a28200,0x82200a2,0x808a8a8a,0xa0aa288,0xa08080,0x2a82a28a,0x8a28080,0x8280a202,0x2);
remainders_pos[818].modulo_value=3272; remainders_pos[818].bit_mask=create_bit_mask(103,0x22aaaa02,0x28020a28,0x22020088,0x2200888,0xaa20a280,0xa8202aa0,0x2a0a800a,0xaa80a882,0x2a2aa220,0x8a80882a,0xa0a2aa8,0x228a22a0,0x80828a8a,0x2aa0202,0xa08a2aa,0x8200222a,0x2a2a2800,0x8220aaa2,0xa2a8020,0xa8220882,0xaa20a228,0xaa8882a0,0x8aa80008,0x8828008a,0x22a82a0a,0x20828a22,0xa0280222,0x8aa82880,0xaaa80080,0x2888008,0x220a200a,0x88a22808,0x2a8020a2,0xa228a,0xa8202022,0x222aa28a,0x208a0a0,0xa2a002a0,0x80828282,0x2208220,0x20a0aa,0x288a8088,0x22002020,0x880a800a,0xaa80a022,0x2a2800,0x820a200a,0x88aa22aa,0x8aa2a228,0x20a2a828,0xa0000228,0x80000a82,0x2a8a0828,0x8a8aa228,0x88aa2228,0x8a082aa,0x28a800a0,0xa02aa00,0x2a02288,0x80088a0,0x2a22808,0xa080022,0x820880aa,0x82820208,0x800a8a82,0xa20800a,0x8aa8880a,0x8082aa2,0x88a00088,0x802a8a2,0x288a228a,0x8a08820,0x22280a0,0x2aaa20,0x282aa202,0x80280a02,0xa2820888,0xa82a8888,0x2822a0,0x2aa2a2,0x8222aa20,0x8a08aa0a,0x20882a28,0x2a8a082,0xaa088208,0x28a8a88a,0x88008200,0x8a20a0a8,0x80aa80aa,0xa2820280,0x88a288a2,0xa8a0a00a,0x2202a22a,0x8aa8a8a8,0x2a02aa08,0x2a0a882,0xa8082aa0,0x8a08aa0a,0x20088002,0x808822,0xa0802822,0xaaaa8828,0x80);
remainders_neg[818].modulo_value=3272; remainders_neg[818].bit_mask=create_bit_mask(103,0xaa22228a,0xa08a82a0,0xaa8a8800,0x8aa88000,0x22a82a08,0x20a8a228,0xa2820882,0x2208200a,0xa2a22aa8,0x20800a2,0x8282a220,0xaa02aa28,0x80a0202,0x8a228a8a,0x82802a22,0xa88aaa2,0xa2a2a088,0xaa8222a,0x82a208a8,0x20aa800a,0x22a82aa0,0x22000a28,0x2208880,0xa08802,0xaa20a282,0xa80a02aa,0x28a08aaa,0x220a008,0x22208808,0x8a000880,0xaa82a882,0x2aa080,0xa208a82a,0x8882aa02,0x20a8a8aa,0xaaa22a02,0x8a802828,0x2a288a28,0x80a020a,0x8aa80aa8,0x88a82822,0xa0020800,0xaa88a8a8,0x820882,0x220828aa,0x88a2a088,0xa82a882,0x22aa22,0x22a2aa0,0xa82a20a0,0x28888aa0,0x888820a,0xa20280a0,0x2022aa0,0x22aaa0,0x80280a22,0xa0208828,0x828a2288,0x8a28aa00,0x80880028,0x8a2aa080,0x828088aa,0xa800822,0xa0a8a80,0x882020a,0x82a80882,0x2200082,0x8080a22a,0x288800,0x808a202a,0xa002aa02,0x802800a8,0x8aaa0828,0x88a222a8,0xa0a22a8a,0x8a0828a,0x2a0a8000,0x20a20000,0x88a0aa28,0x88a22a2a,0xaaa22a8,0x2802282,0xa800a2a0,0x8a20280a,0x22800a80,0xa0202002,0x880a88,0x2a82820,0x8220822,0x2a0a0a08,0x2a002a,0x20282882,0xaa8a2aa2,0x2202020,0xa28a2280,0x8a28200a,0x2080a228,0x2802282,0xa880088a,0x880800aa,0x2808a0aa,0x222200a0,0x8);
remainders_pos[819].modulo_value=1092; remainders_pos[819].bit_mask=create_bit_mask(34,0x22000822,0xa00200,0x2000288,0x22020082,0x820808,0x8,0x2080,0x8028220,0x8280200,0x80800022,0x20028808,0x820a28a,0x20020a0,0x20a08008,0x8a008,0x208800a0,0x28808,0x2280000,0xa0022082,0xa20000,0x20a082,0xa0800802,0x28a08200,0x228008a,0x80002022,0x8028208,0x88280200,0x20800000,0x2082002,0x20080882,0x28000808,0x800a002,0x82000880,0x8);
remainders_neg[819].modulo_value=1092; remainders_neg[819].bit_mask=create_bit_mask(34,0xa2880022,0x8208a00,0x82088200,0xa20a0802,0x8020000,0x88080080,0x80882000,0x20a28,0x208280,0x800a2,0x28020000,0x80202202,0x82802820,0x28200000,0x88002000,0x20000820,0x820000,0x8a200088,0x20822802,0x8220800,0x8828200a,0x20080002,0x20200200,0x2208002,0x828a2,0x820a00,0x200288,0xa0080000,0x808808,0xa00a082,0xa0800002,0x20008000,0x80002082,0x2080000);
remainders_pos[820].modulo_value=1640; remainders_pos[820].bit_mask=create_bit_mask(52,0x8822228a,0xa82a8080,0x28020008,0x2a00002,0xa2082a08,0x20820280,0x22020202,0x280a0208,0x82a082,0xa200002a,0xa8800200,0x88080aa0,0x2a8a2220,0x220888a,0x22a0aa02,0x8a00800,0x200a8000,0x28820a8,0x882080a,0x20880808,0x8a02808,0xa8020a82,0x2880000,0x82a20008,0x8220202a,0x28aa2888,0x8088222,0x8a82a8,0x228020,0xa0802a00,0x280a2082,0x20220820,0x20822020,0x82280a0,0x2a0082a,0x200a2000,0xaa0a8800,0x22088080,0x88a2a8a2,0xa0202208,0x80022a0a,0x8a00,0xa8200a8,0x80a02882,0x80808820,0x80820880,0xa8208a02,0xa8020,0x802880,0x2a82a20,0x88882202,0xa2);
remainders_neg[820].modulo_value=1640; remainders_neg[820].bit_mask=create_bit_mask(52,0xaa2a02,0x20220800,0x208a8080,0x8220888a,0x2280a200,0xa80a0a08,0xaa0a8a82,0x20828280,0x802280a,0x22888822,0x20080a88,0x8220,0x2202a2a8,0x8aa00002,0xa2282202,0x80280088,0xa8020880,0xa00a020,0x80028082,0xa0008000,0x80282080,0x200a8202,0xa008088,0xa228880,0x2a8a022,0xa0222000,0x80000aa2,0x8020220,0x88a208a8,0x20082208,0xa082280a,0xa82a80a0,0x280aa0a8,0x80a20828,0x82208022,0xa8822888,0x22020080,0x2a800008,0x22202a,0x2028aa00,0x88a2282,0x88080280,0x20a8020,0x820a00a,0x80028,0x80a0008,0x20280282,0x88200a8,0x8800a008,0x220a228,0x2a8a,0x22);
remainders_pos[821].modulo_value=821; remainders_pos[821].bit_mask=create_bit_mask(26,0x77d902f2,0xc04a2ebe,0xf3039289,0x275f45b1,0xb473b73c,0xad2df7a,0xe5ccb8fe,0xdb6244a2,0x9062f0af,0x1845627a,0xaca2e093,0xf533a679,0x123d2809,0x2be4052f,0x4d679973,0x863241d1,0x825791a8,0xb6fd43d1,0xe9d14891,0xd41fc74c,0x8b57bed2,0xb90f3b73,0x33e368be,0x80e45270,0xfb9f5d14,0x13d026);
remainders_neg[821].modulo_value=3284; remainders_neg[821].bit_mask=create_bit_mask(103,0xaa008a2a,0x882a222,0x2a8a0a00,0xaa828828,0x28aa2aa0,0x820a02a2,0x28002022,0x2aa882a,0x8aa2822,0x8088aaa2,0x202a280a,0x28aa2aa0,0x8aa0a080,0xa22888a2,0x80aa00aa,0xaaa8808,0xa8a0820,0x2a208a08,0x20088008,0x8820a80,0x28a220a,0x2082a2aa,0xaa2aa022,0x8288aa8,0x22028088,0x828a08aa,0xa0222a8,0x88a08088,0x2aa2a8aa,0x2a082220,0x228a02aa,0x82088282,0x880200a,0x8822a080,0x828a8082,0x8aaa8020,0x2a80880,0x88a22aa8,0x282a8a28,0xa2a8a008,0xa2a0202a,0xa8880a80,0xa2a8822a,0x8aaa0220,0x2000a2a0,0x2820a20,0xa2a8a22a,0x8a0aa82a,0x828aa0a0,0x2a088002,0x8802aa22,0xa88202a2,0xa0222002,0x8282aa88,0xa08a0a00,0x2002800,0x82a02208,0xaa2a20a2,0x8a020a,0x20a22a0,0x88002288,0x20a80a8,0x2002a2a,0x2828aa08,0x8882080,0x2a80022,0x8a88a8,0x828a2a80,0x28828a80,0x88aa8a02,0xa280aa2a,0x82228288,0xa02002a0,0x2a2228,0x8800802,0xa0a88a8a,0x2880222,0xa22008a0,0x28a88a82,0x20080088,0x8a222a00,0x82a00202,0x88a0a220,0xaa2a80a2,0x2028aa88,0x80a8a08a,0xaa28a0,0xa88a880,0x22000aa0,0x820888,0x82a8a0a,0x2a2a0020,0x8a80a820,0x8a20008,0x2a22820,0xa8202880,0xa2822a2a,0x82202a0,0x800a0020,0x80282882,0x88aaa0a0,0xa0022202,0x208a);
remainders_pos[822].modulo_value=3288; remainders_pos[822].bit_mask=create_bit_mask(103,0x82082002,0x820208,0x228a200,0x828a0020,0x20020808,0xa082282,0x808808a2,0x28800a28,0x8208288,0x828220a2,0x20028220,0x8808020a,0x228a00a0,0x20820,0x8a20a080,0x200a2820,0xa20a28,0x200200a,0x22080,0x20a20000,0x8a282280,0x200000a0,0x28828820,0x808a202,0xa00028a0,0x8208a00,0xa280280,0x882800,0x800220,0xa282208,0x20000802,0x20020a00,0x228a28a,0x22822882,0x20028028,0x282000,0x2000820,0x20a28a08,0x80002208,0xa0880802,0xa20020,0x2008202,0x82082002,0x28800800,0x88088000,0x20802000,0x20a20200,0x80008080,0x22822,0x20828a20,0x8208082,0x208a00a2,0x8820208,0x8a282,0x2882880,0x8020200,0x800808a,0x22000802,0x28000220,0x82002002,0x80800820,0x808200,0xa08008a,0x2802022,0x8208800,0x8020a28a,0x82000,0x8000828,0x88280280,0x80822882,0x8a28a28,0x800a080,0xa0802000,0x208828,0x88002,0xa0002822,0x20028028,0xa00a208,0x200a2800,0x28808a20,0x8082282,0xa20a0000,0x8028828,0x8a,0x80020880,0xa00800,0x828a08a,0xa20828a0,0x20a08,0x88082080,0x220a00a2,0x8a08020,0x82088280,0x208a0882,0x28228208,0x228a002,0xa08a2022,0x8828820,0x82202080,0x800800a2,0x8a28,0x82208082,0x800820);
remainders_neg[822].modulo_value=3288; remainders_neg[822].bit_mask=create_bit_mask(103,0x2802882,0x8028a00,0x8a202288,0x20208a0,0x28828000,0x8200a20a,0x22,0x20008220,0x80280200,0x20a2822,0x28820a28,0x8282,0xa2020820,0x8828028,0x2282008,0xa08220a0,0x8228220,0x8a08a082,0x808a2800,0x28228808,0x220a208,0xa0880820,0x20020028,0x8000228a,0x20882020,0xa00208,0x82208208,0x80002080,0x8008a28,0x220a280,0xa0880082,0x28828208,0x8a202202,0xa20a2002,0x28820820,0x8820a088,0x828800a0,0x28220200,0x808a280,0x20000082,0x8228828,0x8a08028a,0x2802882,0x20008008,0x88,0xa0082880,0x28228a08,0x8080000,0x808a20a2,0x28020228,0x8028000a,0xa0020822,0x28a00,0x8800220a,0x82002000,0x828a08,0x80080002,0xa2880082,0x20808a28,0xa08a08a,0x800a0,0x8000a08,0x82008002,0x820828a2,0xa00008,0x8282202,0x80802880,0x808020,0x208208,0xa2002,0x220220,0x80082008,0x20082880,0x8a00020,0x8800008a,0x208820a2,0x28820820,0x82082280,0xa0822080,0x20000228,0x8000a20a,0x22820080,0x820020,0x88088002,0x8a0000,0x8208008,0x80202002,0x22802020,0x8828200,0xa008,0xa2820822,0x200828,0xa000208,0xa0020002,0x20a20a00,0x8a20208a,0x200228a2,0x20028,0xa282008,0x800822,0x8800220,0xa28000a,0x800a0);
remainders_pos[823].modulo_value=3292; remainders_pos[823].bit_mask=create_bit_mask(103,0xaa82a8a,0x822288a2,0xa222828,0x8222a00a,0x888aa222,0xa0288a02,0x2a200,0xa08a0282,0xa0aa88a,0x28a2aaa,0xa88a2aa0,0xa008aa08,0x8a80a828,0xa28a080,0x2a2a0020,0x2a08288a,0x8a82a22a,0xa2808008,0xa080880,0x200aa8a8,0x8002,0xaaaa08aa,0x22088000,0x2880822a,0xaa020a,0x20aa282,0x802a8a08,0x2a20828a,0x2288808a,0x80002aaa,0x2aaa8022,0x2a88aaa,0xa282a020,0x2a082a22,0xa0222aa,0xa2808088,0x80808220,0x2082a8aa,0x2a022a28,0x22a820a0,0x82202280,0x820aa800,0x2828000,0x82280220,0xaaa8a0a,0x20a82a88,0x2220a82,0x88a0888,0x888282a8,0x88a20a0,0xa0028822,0xa0080200,0x8822800,0x20a08a22,0x2a82822,0x2220a22,0x882a0888,0x22a82a0,0x2a0a2aaa,0x880288,0x20002828,0x2002aa08,0x80288088,0x80a082a8,0x828a880a,0x2aa2a820,0xa0882020,0x2202028,0x8aa8880a,0xa88a820a,0x80a828,0x8aaa22a8,0x28802aaa,0x8aaa8000,0x8a202228,0x2a28208a,0x20a2a80,0x828aa08,0x8a080aa0,0x8a882022,0xa0002208,0xaa20aaa,0x88002000,0x2a2aa00,0xa022020a,0x22002028,0x8a88a82a,0x8a22820a,0x800a8a,0x2020a28a,0xa282a02a,0xa20aa200,0xaa8a22,0xaaa8a28,0xaa22aa0a,0x8280a20,0xa008a800,0x280a2280,0x2888aa22,0xa00a888,0x2282888a,0x8a22888,0xa2a82aa);
remainders_neg[823].modulo_value=3292; remainders_neg[823].bit_mask=create_bit_mask(103,0x8220a202,0xaaa002a,0x82aaa0a0,0xaaa2882,0x22aaa,0x28a0028a,0x888a2a88,0x28028a0a,0x82822002,0x8a02a222,0x2002a228,0x28802280,0x20820a0,0x82a02808,0xa2a288a8,0xa280a002,0x20a2aa2,0x2a080880,0x82808008,0xa8822020,0x8888088a,0x22228022,0xaa800888,0xa0080aa2,0x88228a82,0x8a022a0a,0x8a20280,0xa2a80a02,0xaa000802,0x888a222,0xa22208aa,0x8a200222,0x2a0a28a8,0xa280a2aa,0x828aaa22,0x2a080800,0x8080aa8,0xa80a2022,0xa28aa2a0,0xaa20a828,0xaa8aa08,0xa822088,0x8a0a0888,0xaa08aa8,0x82220282,0xa820a200,0x8aaa820a,0x80028000,0xa0a20,0x8002a828,0x288a00aa,0x28808a88,0x800aa088,0xa82802aa,0x8a20a0aa,0x8aaa82aa,0xa28000,0x8aa20a28,0xa282a222,0x88008a00,0xa888a0a0,0xa88a2280,0x8a00800,0x8280a20,0xa020082,0xa22a20a8,0x2800a8a8,0x8aa8a8a0,0x2200082,0x20020a82,0x880820a0,0x222aa20,0xa008a222,0x2220888,0x2a8aaa0,0xa2a0a802,0x8a82a208,0x80a02280,0x2808228,0x200a8aa,0x2888aa80,0x822a8222,0x88a888,0x8a2a2288,0x28aa8a82,0xaa88a8a0,0x20020a2,0x2aa0a82,0x88088202,0xa8a82a02,0x2a0a28a2,0x2a822a88,0x882202aa,0x822202a0,0x22aa2282,0x80a082a8,0x28802088,0xa082aa08,0xa00022aa,0x82882000,0xaa0a0002,0x802aa000,0x2a20a22);
remainders_pos[824].modulo_value=824; remainders_pos[824].bit_mask=create_bit_mask(26,0x822a0222,0x822a2a2,0x88aa20a0,0xa2002a22,0x2a808a0a,0xa2a82a2a,0x8002aa8,0x20202802,0xa08a80,0x82002aa2,0xaa0a2008,0x82020228,0x8222a022,0x82880a88,0xaa288080,0x822008a0,0x8aa800,0x802a20a,0x20802808,0x8a2aa800,0xa8a8a82a,0x8aa0a202,0x2288a800,0x200a08aa,0x828a8a88,0x8880a8);
remainders_neg[824].modulo_value=824; remainders_neg[824].bit_mask=create_bit_mask(26,0xaa28aaa,0x80aa2a2a,0x22a828,0x2a88a22a,0xa2080282,0x2a20a2a2,0x8088a220,0xa8a8a08a,0x88280208,0xa88a22a,0x2282a880,0xa8a8aa0,0xaaa28aa,0xa008200,0x22a00808,0xaa88028,0x88022080,0x808a2a82,0xa808a080,0x2a22088,0x202020a2,0x2282a8a,0xaa002088,0xa8828022,0xa020200,0x820);
remainders_pos[825].modulo_value=660; remainders_pos[825].bit_mask=create_bit_mask(21,0xa0020002,0x20220,0x8080008,0x8a2,0x20200028,0x20228a,0x800000a0,0x808228,0x2000080,0x2002822,0x8000a00,0x8088280,0x20002000,0x228820,0x8000a000,0x808a2880,0x28000,0x208a200,0x20002,0xa08808,0x80008);
remainders_neg[825].modulo_value=660; remainders_neg[825].bit_mask=create_bit_mask(21,0x208a0082,0x8028a20,0x8080,0x80800022,0x20a00820,0x80282202,0x80820,0x8000220,0x8a008008,0x2882022,0x800208,0x80000200,0x20082080,0x8a20028,0x82000,0x22000,0x20800,0xa002288,0x80820802,0x200000,0x80);
remainders_pos[826].modulo_value=3304; remainders_pos[826].bit_mask=create_bit_mask(104,0x8a088a2a,0x280a822,0xa2028a88,0xa200a08a,0x8202a82,0x2280a20,0x28aa0202,0x808a802,0x22880880,0x2280a2a,0x820a080,0x2280aa8a,0xaa00888,0xa80000a2,0xa022002,0x880a0a0,0xa222082a,0x8000802,0x2a802008,0x8a8a0a20,0x8aa002a,0xa2822828,0x2200a082,0x2028a,0x2282082,0x2222000a,0x2a02aa0,0xa8828228,0xa2008222,0x2008a02,0xa80822,0x800020a0,0x80222080,0x88a82a28,0x2002a0aa,0x88802820,0x20a80a02,0xa0028a22,0x2208aa,0x8220888a,0x8a00008,0x20aa820a,0x82000028,0x28a888a0,0x228808,0x2222a0a0,0x28a22080,0x82020028,0xa000820a,0x80888002,0x20802802,0x88,0x28020822,0x2220280,0x82000a80,0x8082a0,0x88a2828,0xa888802,0x2288000a,0x222a2820,0x820a,0x82aa0828,0xa20a0,0x22088220,0x208800a2,0xa2800aaa,0xa02a0888,0x28022280,0xa800808,0xa82a22aa,0x8880228,0x8000202,0x202a000a,0xa2008088,0x82008a80,0x82822a88,0xa80a8028,0x88880a,0x82880a0,0x80800082,0xa0088a2,0x28828a82,0xaa2028,0xa0a2a2a8,0x802a808,0x20002020,0x20888a80,0xa0220a8,0x880a00a,0x2a80,0x200aa08a,0xaa028822,0xa0820a2,0xa0288002,0x202288a8,0x2a202002,0x80aa2880,0xa0288080,0xa8082008,0xa008a82,0xa2808aa2,0x2a028022,0xa220a288,0xa8);
remainders_neg[826].modulo_value=3304; remainders_neg[826].bit_mask=create_bit_mask(104,0x2800222,0x20820a2,0x228a0200,0x2a082802,0x80a0a20a,0x8aa002a8,0xa022828a,0x8080200a,0x2a008000,0x82a08222,0x80282808,0xaa002202,0x82280000,0x2088802a,0x828aa80a,0x82820,0x22a280a2,0x8008808a,0xa200a880,0x20202a8,0x802280a2,0x2a0aa020,0x2a882802,0x8a8a02,0x8a20a80a,0xaaa28882,0x8a282228,0x200a02a0,0x2a880a2a,0xa880282,0x802080aa,0x808a028,0x8a2a808,0x2022a0,0xa88a2022,0x8a028,0x28208282,0x208a02aa,0x882a8022,0xaa00002,0x80280880,0xa8220282,0xa888820,0x20200020,0x80aa0080,0xaa2a2828,0xa022a800,0xa8a08a0,0x28880282,0x800080a,0x2808a082,0x80888800,0xa00a80aa,0x8aa20a08,0xa880208,0x88080228,0x8002a020,0x2000082,0xa2008882,0xaa22a0a8,0x80800a82,0xa2200a0,0x8882a028,0xaa800a28,0x28008822,0x22088222,0x28228000,0xa082aa08,0x82080000,0x20a2a222,0x80008a20,0x888a82,0xa0a28882,0x2a080800,0xa800208,0xa022200,0x208200a0,0x88000002,0xa00820,0x8880a,0x8208002a,0xa002020a,0x82228a0,0x282a2220,0x808a2000,0x2888a8a0,0xa0000208,0x820aa820,0x80002882,0x88882208,0xa8822002,0x228a002a,0x280a822,0x20a0088a,0xa82a0020,0xa2a0a88a,0x8222008,0x28200008,0x2080a800,0x2880202,0x2208022a,0xa20a08aa,0x2aa02a00,0x20);
remainders_pos[827].modulo_value=3308; remainders_pos[827].bit_mask=create_bit_mask(104,0x2008282,0xaa8a8882,0xa0a2022a,0xa0aa220,0xa8a28088,0xa00a200,0x202a20a2,0x2a80a282,0xa22aa00,0xa0202a0a,0x88a00a0a,0x22a0022a,0xaa8aa8aa,0x222288aa,0x28aa88,0x80a8a2aa,0x82aa880a,0x288000a0,0xa882280,0x8a22222,0x2a2a0822,0x828aaa20,0x8288820a,0x2a208000,0xa8aa28,0x2a08a20,0x2aa88228,0x80a08020,0xa2aaa28,0xa028a22,0x80a88002,0x2288a280,0x2888888,0x20aa08a2,0xa282aaa,0x2a02a200,0xaa800820,0x88a20020,0x2208,0x8802002,0x22008aa8,0xa80282a8,0x828280a8,0xaa8008,0xa88a082a,0x82882880,0x202a882a,0x82a22a08,0x28288802,0x8a88,0x8a8a2202,0xa0a0a2a,0x8a2a2a8a,0x20000808,0x2282822a,0x88a82802,0x2a80820a,0x82282a82,0xa22a022,0x2aa00a82,0x28282200,0x2802a2a0,0x200882a8,0x802202aa,0x80000800,0x8a22208,0x2aa080,0xa80a8082,0x828a0008,0xaa08aaa,0x222808a2,0xa2288222,0x22a02028,0x280a0800,0xaa8a088a,0x20a0228a,0x22aa8080,0x20a80288,0xa2a0008a,0x208a828a,0x22282000,0xaa282a08,0xa8a808a,0x88a20882,0x822a0888,0x228028,0x2aa820a0,0xa2a02a02,0xa2800aa8,0x2888822a,0xaa2aaaa2,0xa88aa2,0xa22a88,0x8080aa0a,0xa88a0a0a,0xa02a800a,0x8a808828,0xa00a08a0,0x28a2a008,0xaa0a0220,0x8a0a088,0x2a2aaa88,0x20080822,0x828);
remainders_neg[827].modulo_value=3308; remainders_neg[827].bit_mask=create_bit_mask(104,0x8a880a0a,0x2202000a,0x282a8aa2,0x82822aa8,0x202a0800,0x82882a88,0xa8a2a82a,0xa2082a0a,0x82aa2288,0x28a8a282,0x288282,0xaa288aa2,0x22022022,0xaaaa0022,0x88a02200,0x8202a22,0xa220082,0xa0088828,0x8200aa08,0x802aaaaa,0xa2a280aa,0xa0222a8,0xa000a82,0xa2a80888,0x882022a0,0x822802a8,0xa2200aa0,0x82808a8,0x82a222a0,0x828a02aa,0x820088a,0xaa002a08,0x8a000000,0xa822802a,0x82a0a222,0xa28a2a88,0x220880a8,0x2a88a8,0x8888aa80,0x8008a88a,0xaa880220,0x208a0a20,0xa0a0820,0x88220880,0x200280a2,0xa00a008,0xa8a200a2,0xa2aa280,0xa0a0008a,0x88880200,0x202aa8a,0x828282a2,0x2a2a202,0xa8888080,0xaa0a0aa2,0x20a08a,0xa2080a82,0xaa0a20a,0x82aa28aa,0xa228820a,0xa0a0aa88,0xa08a2a28,0xa8800a20,0x8aa8a22,0x8888088,0x802aaa80,0x88a22808,0x2082080a,0xa028880,0x82280222,0xaaa0802a,0x2aa00aaa,0xaa28a8a0,0xa0828088,0x22028002,0xa828aa02,0xaa220808,0xa8208a00,0x2a288802,0xa8020a02,0xaaa0a888,0x22a0a280,0x82020802,0x2a800a,0xaa28000,0x88aa08a0,0xa220a828,0x2a28a28a,0x2a088220,0xa0000aa2,0x22a2222a,0x8820022a,0x882aa200,0x8082282,0x20028282,0x28a20882,0x20800a0,0x28828028,0xa02a2880,0x22828aa8,0x80282800,0xa2a22200,0xa88080aa,0xa0);
remainders_pos[828].modulo_value=552; remainders_pos[828].bit_mask=create_bit_mask(18,0x22082882,0x20a00,0x828a208,0x20008a0,0x8008020,0x8200a200,0x208a2,0x8800a08,0x88282282,0xa0022082,0x20800020,0x8a00828a,0x2002000,0x20008008,0x8a28200a,0xa0800020,0x28208800,0x82);
remainders_neg[828].modulo_value=552; remainders_neg[828].bit_mask=create_bit_mask(18,0xa2002002,0x8828208,0x80202280,0x82800020,0x800828,0xa082288,0x808a0022,0x8200,0x20a20a,0x208a2002,0x28008828,0x2080202,0x82882800,0x28800800,0x220a082,0x200808a0,0x20a00008,0xa);
remainders_pos[829].modulo_value=829; remainders_pos[829].bit_mask=create_bit_mask(26,0xa93fa3a,0x7d9b36d2,0x50630cfb,0x6a2c343a,0x778d82b8,0xccd76378,0x726dfab9,0xe13b5107,0x1f59901f,0x404abf8e,0xc352cab5,0x85af8ec8,0x831b294,0x4a536304,0xc4dc7d68,0xab54d2b0,0x1c7f5480,0xfe0266be,0xb822b721,0xe757ed93,0x87b1bacc,0x87506c7b,0x970b0d15,0xb7cc3182,0x12db366f,0x1717f254);
remainders_neg[829].modulo_value=3316; remainders_neg[829].bit_mask=create_bit_mask(104,0x820a22a2,0xa002aa0a,0x88aa8022,0xa2a0a8a2,0xaa000a20,0xaaaa0,0xaaa02220,0x28a2888a,0x82800882,0x88822202,0xa8a0228,0x8220200,0x80a82a08,0x828aaa88,0x800a0a0,0x22880a28,0x80a28808,0x228aaa22,0x20aa2aa8,0x2a8a200a,0xa282200,0xa88a0a2,0xa828088,0x2a00a80a,0x8a02aaa2,0x8a8a2a88,0x888aa28a,0xa22aaa02,0x8280a802,0x80880a0a,0x8228288a,0xa00220a2,0x2280228a,0x20aaa2a8,0xa22aa8a,0x20808a28,0x80a2808a,0x8282800,0x88aaa8a,0x8082a0a8,0x80020220,0x88a2028a,0xa022208,0xa208800a,0xaa888a28,0x22202a,0xa82aaa80,0x28228002,0x8a28a82a,0x2a200aa8,0xa82aa00,0x222a2282,0x80a28220,0x202aa002,0x208800aa,0x28202808,0x2802aa82,0x2aaa8000,0x80002a22,0xa2082088,0xa0a0aa88,0x888a222,0xaa8082a2,0x8a8a22a2,0x88a80a02,0x28a08000,0xa8aa828,0x8a208a82,0x20a8208a,0xa208002,0x82280200,0xaa00822a,0x22820a22,0x8a808828,0x8280a0a8,0x2202a80a,0x8882a000,0x80008200,0xa0888020,0xa0220002,0x80a0a80a,0x80a88a82,0x20a20a08,0x802aa228,0xa20aa20,0x82280020,0xa822000,0x8228a882,0xa8a820a8,0x828a0a0a,0xa888000,0x28a8a028,0x80aaa2a2,0x888822a0,0x8282a222,0x82288aa,0x8882,0xaaaa222a,0xa00a0000,0x20a20aa,0x882080a,0xaa022a80,0xa280a002,0x20220);
remainders_pos[830].modulo_value=3320; remainders_pos[830].bit_mask=create_bit_mask(104,0x800a82,0x8a0a0aa0,0x2082a288,0x8200082a,0x80202208,0xa0228a,0x20008020,0x20828200,0x8822a80a,0x20a82022,0x888a2088,0x82220220,0x228aa028,0x282202a,0x2a8aa28,0x800a2000,0x2a0a8202,0x20822280,0x88a0280a,0x8288080a,0x202a0a08,0x2a8a80,0x228080a8,0x22200a,0x8220a020,0x28200808,0x88288a20,0x2a0a0220,0xa0202a,0xa0088a02,0x2880000a,0x88208022,0x88a080,0x202808,0x2280220a,0x80a20080,0xaa2a00a2,0x28000000,0x222082,0x80280002,0x88222202,0x280a0080,0x20a2008,0xa00028,0x22a08202,0xa82a0a8a,0x80020a02,0x8822228,0x20a8a2,0x8080000a,0xa8002080,0x82008a82,0x2a82a200,0x2020800,0xa00002,0x208a2a08,0x2288882,0x2a80a080,0x88a2a0a8,0x80820a,0x8028000a,0x282008a0,0x220200a0,0x2808888,0x800028,0x28820888,0xaa000000,0x28a00a8,0x8802008a,0xa08802,0x202808,0x22020a22,0x28880208,0xaa00002,0x80a220,0xa8a8080a,0x220880a0,0x2808a228,0x82202008,0x80a08,0x88a00888,0x2a0202,0x802a2a8,0x8220a0a8,0x22a02022,0x8a0280a,0xa8028882,0x28082a0,0x800008a0,0x8028aa2a,0x88a80882,0x82280aa2,0x22088088,0x82208a2,0x2288082a,0x8a02a88,0x8008282,0x80200,0x2a2880a,0x82208808,0x8a82000,0xa2228a82,0xaa0a0,0x82a002);
remainders_neg[830].modulo_value=3320; remainders_neg[830].bit_mask=create_bit_mask(104,0x8808020a,0x2028220,0x28022200,0xa8080a2,0xa8aa00,0x88282a02,0xa80808a0,0x280a0288,0xa22082,0xa020a82a,0x22800,0xa2a8aa0,0x2a0220a0,0x8a02a8a2,0x82202220,0x8822888,0xa2020a82,0x280aa208,0x20a082,0x2008002,0xa8a20280,0x88220200,0x2a080020,0x88a2a082,0x2a82828,0xa0a80080,0x2002a0,0x228282a8,0x8820a8a2,0x2080020a,0xa0080882,0x2808a2,0x8002008,0x8820a080,0xa208aa02,0x82a0808,0x22228822,0x20888088,0x88a2a80a,0xa0880a,0xaa2a8a,0xa0028800,0xa82a080,0x882088a0,0xa2280a02,0x20a20202,0x80a8282,0xaa2a0,0x88a0202a,0x88802,0x20882808,0xa080202,0x220a2288,0x8a828088,0x8028080a,0xa8022280,0x8a200002,0x22082008,0x222820,0x80080a02,0x8a00882,0xa0288020,0x2a8a8028,0x8a000000,0x80088820,0xa00a0000,0x22088880,0xa028020,0x828802,0x8028000a,0x88a82080,0xaa0a82a2,0x20008280,0x8220888a,0x80082a28,0x20200002,0xaa000820,0x208022a0,0xaa0a880,0x80808200,0x280000,0x88228a82,0x8a2220,0xaa02820,0xa228a82a,0x80202082,0x200a0002,0xa080228,0x8808028,0xa02222,0x20000a,0xa208222,0x2a800000,0x80a2802a,0xa2008022,0x282200,0x80080a02,0x8808288,0x8a220082,0x2a80000,0x80202888,0x2a2a0202,0x8822028,0x2288a);
remainders_pos[831].modulo_value=3324; remainders_pos[831].bit_mask=create_bit_mask(104,0x82822022,0x8228800,0x282082,0xa20800a0,0x28a00a28,0x80002208,0xa08208a2,0x220a20,0x8800a20a,0x82800882,0x8000820,0x80000208,0x808a2082,0x828a20,0x8028200a,0xa2000800,0x20028a08,0x8a088280,0xa2000002,0x28828828,0x8200a000,0x808a2080,0x8808220,0x8828a282,0x808a08a2,0x8808808,0x2202202,0x20022,0x8000020,0x8082202,0x22080022,0x28a00a08,0x200200,0x28a2880,0x28020200,0x2000288,0x22822880,0xa08028,0x800020a,0x2080022,0x20828a28,0x820a282,0x2022802,0x800820,0x800808a,0x20820,0x20808a28,0xa000,0xa00a2080,0x8208028,0xa200082,0x880802,0x8020220,0x820008a,0xa2802082,0x208a00,0x8000a000,0x28a2020,0x820800,0xa202002,0x820020,0x8028808,0x8828a082,0x28a2820,0x8800208,0xa080002,0x828020a0,0x228828,0x82280008,0x80802,0x228a28,0x80080080,0x820a00a2,0x8800208,0x8088202,0x800002,0x8800800,0x8088088,0x22022022,0x28a20a20,0x828a282,0x20882022,0x20208a20,0x8000a008,0xa2822822,0x28000008,0x8028220a,0xa20a2800,0x20020008,0xa008280,0x208a2820,0x28208a20,0x2080000,0x20820002,0x28220028,0xa08a002,0xa08a0880,0x28a20820,0x82088000,0xa28a00a2,0xa00208,0x8208280,0x20022882,0x8808828);
remainders_neg[831].modulo_value=3324; remainders_neg[831].bit_mask=create_bit_mask(103,0x20a28a2,0xa20008,0x8820a00a,0x22800820,0x20208220,0x808a280,0x200a0022,0x8a28228,0x82282,0x2080002,0x808028,0x8088280,0x22802,0x8020228,0x820a082,0x22880080,0x28820200,0x2000208,0x22880882,0x20020020,0xa082088,0x22800,0xa28,0x20220a,0x20022,0x8a28a28a,0x808a08a2,0x808828,0x8000a28a,0xa28008a2,0x20208200,0x88288288,0x82022000,0x20828a08,0x8a088200,0xa20a2000,0x8200820,0x80088282,0x828008a2,0x28020220,0x8028220a,0x828a2082,0x8008028,0x80080002,0x808a00a0,0x28000220,0x88082088,0x20822800,0xa00820,0x8228800a,0x80000082,0x828a28,0x80288002,0x22082802,0x8a00208,0x8082088,0x820228a0,0x8028008,0x8228a08a,0x800208a0,0x820000,0x20200a,0x820220a0,0x8a00,0x8200808a,0x2082820,0x8a20020,0xa208080,0x80800082,0x8a20220,0x8008008,0x2820822,0x8a00,0x8000028a,0x80080882,0x8008,0x80000000,0xa28a28a2,0x20228228,0x8020220a,0xa00028a2,0x28a00228,0x8082080,0x220a20a2,0x20808800,0x820a282,0x22822080,0x28828800,0x82080208,0xa00220a0,0x20a00228,0x8a008088,0xa00a0882,0x20a28820,0x8200200a,0x20020000,0x20228028,0xa000088,0x22020822,0x8208a00,0x80280208,0xa08a2002,0x20);
remainders_pos[832].modulo_value=104; remainders_pos[832].bit_mask=create_bit_mask(4,0x2a82020a,0x20aa0808,0x8082a820,0xa0);
remainders_neg[832].modulo_value=104; remainders_neg[832].bit_mask=create_bit_mask(4,0xa20a8a82,0xa8228000,0xa20a8,0x28);
remainders_pos[833].modulo_value=476; remainders_pos[833].bit_mask=create_bit_mask(15,0x208a202,0x8a08802,0x228a0028,0x8a2880a0,0x28a20280,0x82880a00,0x88202008,0x2880a022,0x22008082,0x800a0228,0x202808a2,0x80a0228a,0x2800a28,0x80220a2,0x808a208);
remainders_neg[833].modulo_value=476; remainders_neg[833].bit_mask=create_bit_mask(15,0x8a802a0a,0x200082,0xa20288a0,0x2200828,0xa0228a08,0x2000288,0xa8a080,0x2008282a,0x2a880802,0x828aa0,0xa8208022,0x820aa02,0x8a080220,0x808aa02a,0x802200);
remainders_pos[834].modulo_value=3336; remainders_pos[834].bit_mask=create_bit_mask(105,0xa28a08a2,0x20820808,0x8028a000,0xa2882800,0x28020028,0x800a202,0x28a0802,0x28200220,0x2280000,0x82822002,0x8a00808,0x88208202,0x82022882,0x228a00,0x8a00a002,0x82002,0x8a28200,0x8808228a,0xa2882820,0x8020208,0x8a20800a,0x200a0822,0x8200220,0x88202080,0xa00a0,0x20020220,0x220a202,0x28028a2,0x8220808,0x228208a,0x20800822,0x28820000,0x8220a080,0x20080000,0x208808,0x82000080,0xa20828a2,0x28028000,0x80002282,0x80022082,0x20820800,0x88280082,0xa2020080,0x28a00220,0x228200a,0xa08000a2,0x20028820,0x820a002,0xa0080000,0x28000a08,0x8820820a,0x800000,0x20000,0x82082200,0xa00028a0,0x200a20,0xa082000,0x22800880,0x20a08,0x828808a,0x800a28a0,0x808a08,0x282202,0x20820882,0x8800200,0x88000282,0x2802882,0x28208a00,0x828a,0x22080002,0x200820,0xa088200,0x822802,0x20020800,0x8288088,0x208820a2,0x28028020,0x8a08808a,0x80800880,0xa00008,0x808220a,0x80082002,0x20a00808,0x208a288,0x808020a0,0x28228a20,0x88202208,0x828a20a2,0x8200000,0xa00a280,0xa2880080,0x28808200,0x82082282,0x200a2080,0x8828220,0x288080,0x80082800,0x20a28008,0x8a002080,0x802880,0x28228a28,0xa280200,0x20820800,0x20a28a20,0x8a);
remainders_neg[834].modulo_value=3336; remainders_neg[834].bit_mask=create_bit_mask(105,0x22020022,0x28028000,0x8202088,0x22002080,0x20828020,0x8008228a,0x82020082,0x20a08a28,0x8a208088,0x20a2882,0x208000,0x28028a,0x28a2002,0x8a20208,0x208208a,0x80802882,0x220a08,0xa202,0x220020a0,0x828a00,0x2280082,0xa00200a2,0xa08a28,0x28a008,0x80820820,0x28828a28,0x8a28228a,0x82082022,0xa28000,0x8a20a002,0xa00800a2,0x20028808,0xa282008,0xa0800880,0x8a00000,0xa088008,0x22802022,0x20820808,0x808a20a,0x8a2802,0x28028008,0x20800a,0x228a0800,0x20208a28,0x8a20a082,0x20080822,0x28820028,0x8028208a,0x20800880,0x20808200,0x280282,0x80080880,0x8828808,0xa00a288,0x20882020,0x8a08228,0x82002088,0xa2080000,0x8828200,0x80200002,0x822020,0x8000200,0x8820a28a,0xa00a0002,0x8a08,0x8820a,0x82082002,0x20a00208,0x88080202,0xa2800882,0x8a08028,0x82000288,0x800a2082,0x20828008,0x80200000,0xa0002822,0x20820828,0x2000002,0x80000,0x8208800,0x8000a282,0x802882,0x28208000,0x8a002200,0x82820,0x20a20228,0x28a280,0x2022822,0xa08808,0x82082208,0x22000800,0x20000a08,0xa00a20a,0xa0822800,0x20a28,0x88200008,0x802080,0x28220800,0x208a008,0x80082000,0x20a20220,0x82208288,0xa00a0080,0x28220228,0x2);
remainders_pos[835].modulo_value=3340; remainders_pos[835].bit_mask=create_bit_mask(105,0x2822228a,0xa82a80a0,0xa828280,0x82a28800,0x82882808,0x20a20200,0xa8220822,0x20082020,0xa2288aa,0x2000002,0x80a82a02,0xaa288a20,0x22000200,0x2a008a2,0xa2082a2a,0xa88a2002,0x800a0800,0x8880220,0x202a008,0xa028880a,0x208a0000,0x200a8002,0x2008280,0x8280a08a,0xa2a08a28,0x800a2a02,0x80028a02,0x2080008,0x8220820,0x2288a02,0xa802288a,0xaa020882,0x2200a028,0x88028882,0x82a02222,0xa0822a08,0xa8000022,0x828008,0x2a00880,0x82a88808,0x8020000a,0x80200a20,0x2a8a8288,0x200808a,0xa0a0a800,0xa0882008,0xa080880,0x20820228,0x82000822,0x208022a,0x2800200a,0x80880a0,0x22020000,0x800280a0,0x2080a00,0x82a88,0x8208882,0x20a0288,0x8220a022,0xa0a0a200,0x20080002,0x2a2a8a20,0x802228,0x80208a,0x22a82a00,0xa80202,0x28200022,0x2a200,0x8820a880,0x80a8220a,0x28022888,0xa0088822,0x80aa280,0x8802a822,0x22880a22,0x882080a,0x2080082,0x28002200,0x8a00280a,0x20a8a80a,0xa028228a,0x20080a20,0x2a008028,0xa208800,0x2280a000,0xa8080a02,0x2220200,0xa002088,0x8a22a002,0x8208a800,0xa80a8a,0x888a2,0x228aa008,0x82a0208a,0x8080a,0x288a0800,0x82008aa2,0x882a080,0x8a08882,0x82282008,0x28a82202,0x282a0002,0x2a82a028,0x888280a0,0xa28);
remainders_neg[835].modulo_value=3340; remainders_neg[835].bit_mask=create_bit_mask(105,0xa0aa2a02,0x20220820,0x20a0208,0xa220088,0x200a000,0xa82a0a08,0x202a80a2,0x2880a0a8,0x82a20022,0x8288880a,0x820228a,0x222002a0,0x2a888288,0x8a20802a,0x2280a222,0x2002288a,0x8028080,0x82a8,0x8a822880,0x20a00002,0xa8020888,0xa8020882,0xa880208,0xa002802,0x22280220,0x882228a,0x80a0282,0xa808080,0x80a280a8,0x82a0020a,0x208a2002,0x220a8002,0x2a8820a0,0x82000a,0x228aa2a,0x280a2280,0x200888a2,0x80a0080,0x8a208008,0x2200000,0x8a80882,0x82882a0,0x22020200,0x8a800802,0x20282008,0x28002880,0x2008000,0x280a82a0,0xa8080aa,0x82808a22,0xa0882882,0x80000820,0x2a8a8088,0x8820828,0x82808208,0x88802200,0x80280002,0xa828200,0xaa028aa,0x20282a08,0xa880088a,0xa22202a0,0x808a2a0,0x8800a802,0xa220a208,0x88200a8a,0xa02888a2,0x8a2288,0xa02008,0x20aa02,0xa08a2000,0x280000a2,0x822208,0x8220aa,0xa200822a,0x800a0082,0x8a008802,0x2088a288,0x280a082,0xa0202002,0x28a02a02,0xa80082a0,0x228800a0,0x82a00088,0xa2082808,0x2080028a,0x8a2a8a80,0x288a000,0x222288a,0x2802008,0x88200202,0x88000022,0x2a022080,0xa20a802,0x80808002,0xa0020088,0xa080222,0xa2008,0x8020000a,0x2a0a800,0xa0202a8a,0xa0228882,0x220a20a0,0x20828,0x220);
remainders_pos[836].modulo_value=1672; remainders_pos[836].bit_mask=create_bit_mask(53,0x22a02222,0xaa200,0x20028002,0xa0a8aa00,0xa0820202,0xa0822288,0xa28a282,0x8a22082a,0x2020828,0x2a8a82a,0xa0020,0x22a8002,0x22202a20,0x8a80888a,0x2a8800,0x800a0008,0x82a2a800,0x8208080a,0x82088a22,0x28a28a0a,0x288820a8,0x80820a2,0xaa2a0a8,0x280080,0x8aa0008,0x8880a880,0x2a022228,0xaa2002,0x280020,0xa8aa002,0x820202a,0x822288a,0xa28a282a,0xa22082a0,0x20208288,0x2a8a82a0,0xa00200,0x22a80020,0x2202a200,0xa80888a2,0x2a88008,0xa00080,0x2a2a8008,0x208080a8,0x2088a228,0x8a28a0a8,0x88820a82,0x80820a22,0xaa2a0a80,0x2800800,0x8aa00080,0x880a8800,0x88);
remainders_neg[836].modulo_value=1672; remainders_neg[836].bit_mask=create_bit_mask(52,0xaa20a2aa,0x88022a88,0x288a088a,0x28202280,0x280a0a8a,0x200aa200,0x82a02a0a,0x22a8022,0x8a8280a0,0xa2020a2,0x888288a0,0x8aa2008a,0xaa2822a8,0x2080002,0x80a20080,0x8820880,0x22a2088,0xa808002,0xa80022a,0xa0220282,0x2000a820,0x80a82a,0x822a2020,0x88208808,0x80228080,0x82000,0xa20a2aa0,0x8022a88a,0x88a088a8,0x82022802,0x80a0a8a2,0xaa2002,0x2a02a0a2,0x22a80228,0xa8280a00,0xa2020a28,0x88288a00,0xaa2008a8,0xa2822a88,0x2080002a,0xa200800,0x88208808,0x22a20880,0xa8080020,0xa80022a0,0x2202820,0xa820a,0x80a82a2,0x22a20200,0x82088088,0x2280808,0x820008);
remainders_pos[837].modulo_value=372; remainders_pos[837].bit_mask=create_bit_mask(12,0x228a0882,0x220000,0x208200a,0x2802082,0x20000220,0xa08828a,0x8a2822,0x28008800,0x8082080,0x800a0082,0x28800008,0x8220a);
remainders_neg[837].modulo_value=372; remainders_neg[837].bit_mask=create_bit_mask(12,0x22020002,0x8a28808,0x8a00a082,0x82082802,0x20808a28,0x82000202,0x800220a2,0x20800008,0x8000a008,0x820802,0x20008800,0xa282);
remainders_pos[838].modulo_value=3352; remainders_pos[838].bit_mask=create_bit_mask(105,0x8a00020a,0x28a20a08,0xa8222a82,0x8a882aa2,0x800a8a2a,0xaa202a2,0xa8220808,0xa88888a2,0x22a2808a,0xa8a0880a,0xa02828a,0xaa028aa0,0x8a82aa88,0x22282022,0x2028a200,0x2828a08a,0x800a0220,0x82a2aa0,0x220,0x80aa822a,0x28aa2080,0xa020882a,0x22802022,0x22aa822a,0xa0aa02a8,0xa888a000,0x8880a20,0xa00a0aaa,0x80022802,0x2a822022,0x8a2a022a,0x20082028,0x800a8a8a,0xaaaaa022,0x2028aa22,0xa0aa8a00,0x8282a22,0x82a208a,0x28222aa2,0x280802a,0x82a00880,0x82a0aa00,0x8a080082,0x820220a8,0x8888008a,0xa2280208,0xa200a8a8,0x80aa8202,0x28808020,0x22280200,0xa2082280,0xaaa088a0,0xaa8a00a2,0x8a0a220a,0x88028820,0x28008028,0x2080202,0x8a8082aa,0x8a2a2a00,0x22208028,0x82a20022,0xa22a0880,0x82820020,0x8200aa0a,0x8002200a,0x28a80202,0x208aa888,0x20a208a8,0xa88a828,0x800a2aa,0xaa88aa28,0x2880aaa,0x8a2a2a0,0xa2280820,0xa8a880a8,0x2880882,0xa802880,0x20aaa0a0,0x2a08a022,0xa000a22,0x882a80aa,0x88a882aa,0xa880802,0x28a82208,0x20208aa,0xa882aa,0x20088000,0x20aa8a8,0x280880a0,0x8a20a28,0x88008a28,0xa2880828,0xaa22aa82,0xa00aa280,0x2aa28280,0x88a0220a,0x2aa2028a,0x2a8a2222,0xa0202088,0x28a808a,0xa2a8a2a0,0x2a8aa822,0x2882a888,0xa220a08a,0xa08000);
remainders_neg[838].modulo_value=3352; remainders_neg[838].bit_mask=create_bit_mask(105,0x2888a82,0xa02a8280,0x20aaa20a,0x200a22a,0x88202a2,0x822a8a2a,0x20aa8080,0x2000002a,0xaa2a0802,0x20280082,0x828a0a02,0x228a0228,0x20a2200,0xaaa0a8a2,0xa8a02a88,0xa0a02802,0x8828aa8,0x80a2a228,0x88888aa8,0x8220aa2,0xa022a808,0x28a800a2,0xaa08a8aa,0xaa220aa2,0x28228a20,0x20002888,0x800082a8,0x28828222,0x88aa08a,0xa20aa8aa,0x2a28aa2,0xa880a8a0,0x8820202,0x222228aa,0xa8a022aa,0x28220288,0x80a0a2aa,0x80a2a802,0xa0aaa22a,0x8a0808a2,0xa288008,0xa282288,0x280880a,0xa8aa820,0x8802,0x2aa08a80,0x2a882020,0x8220a8a,0xa00808a8,0xaaa08a88,0x2a80aa08,0x22280028,0x2202882a,0x282aa82,0x8a00a8,0xa08808a0,0x8a808a8a,0x2080a22,0x2a2a288,0xaaa808a0,0xa2a88aa,0x2aa28008,0xa0a88a8,0xa882282,0x88aa882,0xa0200a8a,0xa8022000,0xa82a8020,0x820020a0,0x80882a22,0x220022a0,0x8a008222,0x802a2a28,0x2aa080a8,0x20200820,0x8a00800a,0x8208a008,0xa8222828,0xa28028aa,0x828882aa,0xa20822,0x200a22,0x8200808a,0xa020aa80,0x8a8a8022,0x88200a22,0xa8800888,0x8a822020,0xa0800828,0x802a82a0,0x8802a0,0x2a0080a0,0x22aa220a,0x28822a08,0xa22a0a08,0x28aa82,0xa22a8a02,0xa202aaaa,0x28a8a800,0x8a020802,0x2a202a28,0xa20220aa,0xa00a2000,0x2aa82802,0x280888);
remainders_pos[839].modulo_value=3356; remainders_pos[839].bit_mask=create_bit_mask(105,0x82200a22,0x22aa20a0,0x88a82a8,0x2a800aa8,0x8802a2a,0xa202028a,0x200008aa,0x88a28082,0xa2880a82,0x220a282a,0x822aaa88,0xaa20a288,0xa0a00880,0x82aa2028,0xa0808a82,0x2280a000,0xaaa2208a,0xa80a8,0x2a8a0882,0x820aa0a8,0xaa2aa8aa,0xa88a282,0xa8080a2,0x8aa8088a,0x8080aaa0,0x888aa8a8,0x2222208,0x2a2a2020,0xa0022800,0x2a028222,0x2202882a,0x800a08,0x28a0020,0x8200a028,0x2aa8a22,0x8800202a,0x2a088228,0x2a282aa8,0x800a0a02,0xa82820a8,0x28802a22,0x8a2208,0x8288a200,0xa20800a0,0x8220a02,0xaaa88a2a,0x8a880022,0xaa22820a,0xaa008080,0x222a002,0x8008a00a,0xa88a2828,0xa8220882,0x22828a22,0xa00a200,0xa800a888,0xa020200a,0x2a08288a,0xa880022a,0xa8a22aa,0xa80a0882,0x20a00208,0x8a228,0x82088a20,0xa88a8022,0x22a08282,0x880a0a00,0x82aa828a,0x2288220a,0xa808002,0x288a2aa8,0x280a008,0x800a28,0x820a0020,0x8a822808,0xa888280a,0x80028800,0x808a8a,0x22088888,0x22a2aa22,0x20aaa020,0xa2202aa,0x8a0202a,0xa828a22a,0x2aa2aa8a,0x82a0aa08,0x8220a2a,0xa2a02a00,0x8a2088aa,0xa000a028,0x282a2020,0xa2808aa8,0xa02200a0,0x2228a08a,0x822aaa88,0xaa828a08,0x282a0228,0x882028a2,0xaaa20000,0xa280808,0x8a8a8022,0x2aa002a,0x82a82a22,0x20a08aa8,0x88a0088);
remainders_neg[839].modulo_value=3356; remainders_neg[839].bit_mask=create_bit_mask(105,0xaa882aa,0xaa22a828,0x80020a20,0xa2088220,0x8008a2a2,0x2a8a8a02,0xa8888022,0x2a080a,0x2a00820a,0xaa82a0a2,0xaa22200,0x22a82a00,0x28288008,0xa22a8a0,0x2808020a,0xaa082888,0x222aa802,0x88820820,0xa202800a,0xa822820,0x22a22022,0x82002a0a,0x8208082a,0x2208002,0x8082228,0x22020,0x8aaaaa00,0xa2a2a8a8,0x288aa088,0xa28a0aaa,0xaa8a00a2,0x88088280,0x8a0288a8,0xa8828a0,0x8a2202aa,0x88a8a2,0xa2800aa0,0xa2a0a220,0x882828a,0x20a0a820,0xa008a2aa,0x8802aa80,0xa002a88,0x2a808828,0x80aa828a,0x222002a2,0x20088aa,0x22aa0a82,0x22880808,0x8aaa288a,0x8802882,0x2002a0a0,0x20aa800a,0xaa0a02aa,0x82882a88,0x20882000,0x28a8a882,0xa280a002,0x20088aa2,0x8202aa22,0x2082800a,0xa8288a80,0x88802aa0,0xa8002a8,0x200208aa,0xaa280a0a,0x828288,0xa220a02,0xaa00aa82,0x8208088a,0xa002a220,0x8a082880,0x880882a0,0xa8288a8,0x20aa080,0x2000a082,0x88a0088,0x88080202,0xaa800000,0xaa2a22aa,0xa82228a8,0x82aa8a22,0x8028a8a2,0x20a02aa2,0xa22a2202,0xa282280,0x80aa82a2,0x2a28a288,0x2a80022,0x288828a0,0xa0a2a8a8,0x2a080220,0x28aa8828,0xaaa02802,0xaa22200,0x220a0280,0xa0a28aa0,0xa8a02a,0x222a8888,0x82a08080,0x20208aa,0x8a2288a2,0xa20a2aa,0xa8280220,0x28800);
remainders_pos[840].modulo_value=840; remainders_pos[840].bit_mask=create_bit_mask(27,0x20880802,0x20000a20,0x2088208,0x2020082,0xa00a00,0x282208,0xa0020000,0x20000228,0x200080,0x82800022,0x20000000,0x208200,0x880020,0x220000,0x82080008,0x800,0x28200,0x80088,0x80000802,0x800a28,0x88280000,0xa00a0020,0x808000,0x82208082,0xa0000820,0x20220808,0x80);
remainders_neg[840].modulo_value=840; remainders_neg[840].bit_mask=create_bit_mask(26,0xa0000002,0x28008220,0x2000280,0x82020802,0x208208,0x88202280,0x200a0080,0x28808220,0x8208000,0x2080822,0x28000808,0x88200280,0xa0,0x8a28008,0x2008000,0x880080,0x20a00,0x8008000,0x800082,0x220,0x200088,0x20020820,0x8000000,0x220000a,0x20080020,0x28a20000);
remainders_pos[841].modulo_value=29; remainders_pos[841].bit_mask=create_bit_mask(1,0x1ffffffe);
remainders_neg[841].modulo_value=116; remainders_neg[841].bit_mask=create_bit_mask(4,0x2222222,0x22222222,0x22222222,0x22222);
remainders_pos[842].modulo_value=3368; remainders_pos[842].bit_mask=create_bit_mask(106,0xa20aa282,0xa8a20822,0x80a8020,0xaaa0282,0x220020a0,0x8008aa80,0x2a282aa8,0xa882888a,0x22800828,0x200a222,0x2082a8aa,0xaa88220a,0x2a0a222a,0xaa0a2a0a,0xa802222a,0x2280aa0a,0xa08a8a20,0x882aaaa,0x2802028,0x282a2a00,0x20088022,0xa88aa882,0x8a2208a0,0x82080a82,0xaa020a8,0xa20aa2a8,0xa2028282,0xa80200,0xa2880a2a,0x208280a8,0x880a08a2,0xaa228800,0x28222a88,0x82aaa020,0x88a82a2a,0x8a000002,0x822a2080,0xa800a00a,0xa0002222,0xa0202020,0x80002220,0x8a20a228,0xa2a00802,0x8222220a,0x880828aa,0x20208882,0xaa880028,0xa22a8008,0xaa0a2a,0xa8a282a0,0x80a2a80,0xa20228a2,0x82828200,0x28808a00,0xa8a0208a,0x828a2a02,0xa0aa000a,0x2a88aa8,0x22aa20,0x22080828,0x28202282,0x888882aa,0x200a8aa0,0x8a08a280,0x88000228,0x8080a08,0x88000a08,0xa002a88,0x8a882a0,0xa202,0xa82a2280,0xaaa82a8,0xa8882808,0x2288aa22,0x20a02200,0x282088a,0xa0228a2a,0x802a00a8,0x82808a00,0x8aa08a82,0x80aa02a,0xa020822a,0x2088a282,0x2aa22a0a,0x2200882,0xa8a82888,0x8028000,0xaa822028,0xa2a20aaa,0xaa028808,0x88802aa0,0xa8a0aaa8,0x88a0a8a0,0x8822aaa8,0x2a8208a0,0x8a0080aa,0x20028888,0x22822a28,0xa828a8a2,0xaa20022a,0x8008802,0x80aaa00a,0x2a02082,0x208a2a08,0x8aa08a88,0x82);
remainders_neg[842].modulo_value=3368; remainders_neg[842].bit_mask=create_bit_mask(106,0x2a822a0a,0x202a80aa,0x808208a8,0x82228a0a,0xaa88a828,0x8802208,0xa2a0a220,0x200a0002,0xaa0880a0,0x8a882aaa,0xa80a2022,0x2200aa82,0xa282aaa2,0x2282a282,0x208aaaa2,0xaa082282,0x280202a8,0x800a2222,0x8a08a8a0,0xa0a2a288,0xa88008aa,0x2002200a,0x2aa8028,0xa80820a,0x8228a820,0x2a822a20,0x2a8a0a0a,0x88208a88,0x2a0082a2,0xa80a0820,0x82802a,0x22aa0088,0xa0aaa200,0xa2228a8,0x20a2a2,0x288888a,0xaa2a808,0x20882882,0x2888aaaa,0x28a828a8,0x888aaa8,0x2a82aa0,0x2a28808a,0xaaaaa82,0x80a022,0xa8a8000a,0x220088a0,0x2aa20880,0x882282a2,0x202a0a28,0x8082a208,0x2a8aa02a,0xa0a0a88,0xa0080288,0x2028a802,0xa02a28a,0x28228882,0x8a200220,0x88aa22a8,0xaa8080a0,0xa0a8aa0a,0xa22,0xa8820228,0x2802a08,0x888aa0,0x80808280,0x888280,0x8288a200,0x80200a28,0x88882a8a,0x20a2aa08,0x82220a20,0x2000a080,0xaa0022aa,0xa828aa88,0x8a0a8002,0x28aa02a2,0x8a28820,0xa080288,0x228020a,0x808228a2,0x28a80aa2,0xa8002a0a,0xa22aa282,0x8aa8800a,0x2020a000,0x808a0888,0x220aa8a0,0x2a2a8222,0x228a0080,0x8a228,0x20282220,0x282020,0xaa2220,0xa20a8028,0x2880822,0xa88a0000,0xaa0aa2a0,0x20a0202a,0x22a88aa2,0x8088008a,0x8222882,0x8a28a80a,0xa802a280,0x2280200,0xa);
remainders_pos[843].modulo_value=3372; remainders_pos[843].bit_mask=create_bit_mask(105,0x2080002,0x8820208,0xa20200a,0x828a2080,0x28828800,0x8a202282,0x228008a0,0x20020a00,0x20228a,0x820a2822,0x28220200,0x2082288,0x828a28a0,0x8000828,0xa008002,0x828220a2,0x828220,0x88208,0x20008a0,0x220208,0x8a00a080,0x200a28a2,0x8008208,0x220a288,0xa2002800,0x28208008,0x80280280,0x80022820,0x8a28028,0x82200008,0xa0820822,0x8a28200a,0x82820880,0x28a00228,0xa208202,0x2080802,0x800808,0x8208a20a,0x20020a2,0x8a00000,0x80200082,0x820a0880,0x8808000,0x88000088,0x8a0082,0x200228,0x80088,0x20a2002,0xa20800,0x200a088,0x82820882,0x28028a20,0x828208a,0xa0080822,0x8a00220,0x8a080002,0x2000800,0x800220,0xa200288,0x22820,0x20220220,0xa082000,0x802022,0xa20820,0x80080000,0xa20828a0,0x200a08,0x2080202,0x208a0802,0xa28808,0x8282288,0x828a2022,0xa00,0x820a000,0x882882,0x28a20200,0x88002280,0x2802082,0x20828028,0x8008a200,0xa0880002,0x20020228,0x8a008208,0xa00a28a2,0x8800020,0x80208,0x220000a2,0x28200208,0x88282088,0x200a08a0,0x20800,0x8a282282,0x820800a2,0x8828228,0x8a082008,0x80800882,0x8008a28,0x28800a,0x808a20a2,0x28228828,0x8a282002,0x808a0020,0x8220a00,0x2080208,0x800);
remainders_neg[843].modulo_value=3372; remainders_neg[843].bit_mask=create_bit_mask(106,0x82800882,0x28a00,0x8228a082,0x2022800,0x20020008,0x228a20a,0xa2080020,0x28828208,0x8828a202,0x28220a2,0x20a28a08,0x8a00a200,0x2022020,0x808020,0x8208008a,0x20a2822,0x8020a28,0x88000280,0x82880020,0x8a28a00,0x2082008,0xa0822022,0x800a00,0x8a282200,0x22882080,0x20a00800,0x8208208,0x8a20a0,0x220820,0xa288080,0x200a00a2,0x8808808,0x220a082,0x20a0000,0x20208a20,0x8228028a,0x82800082,0x8008000,0xa002282,0x82882822,0x208808,0x828800a,0x2820000,0x808,0x88000,0x80020802,0x8a08a20,0x88008000,0x82822882,0x8228008,0x8a082000,0x20a0002,0x20820228,0x8020a002,0x208000a2,0x208a28,0x200808a,0x82880080,0x8008a28,0x82288200,0x808a20a0,0x28a20a28,0x8200a088,0x800828a2,0x8228028,0x8008088,0x22802020,0x8a08200,0x8a00828a,0xa0020082,0x8220000,0x8020a200,0x20228a2,0x8808208,0x80282088,0x80002002,0x20228a08,0x8a208,0x82082802,0x28020820,0x8002288,0x20000882,0x28828a20,0x2080280,0x20822022,0x8828,0x88008280,0xa2880822,0x20a08a00,0x20a000,0xa0820020,0x8828008,0x220a20a,0x2800822,0x20a20,0x200a080,0x2,0x800220,0x88200082,0x22822,0x20a20020,0x220a08a,0x208a0,0xa28208,0x8a008280,0x80);
remainders_pos[844].modulo_value=1688; remainders_pos[844].bit_mask=create_bit_mask(53,0x8aa0a2aa,0x802220a8,0x28020a2a,0xa22aa220,0x88080288,0xaa00a282,0x8a02820a,0x82a8002,0xaa2a20a,0x88008822,0x828aa0a8,0x2888aa00,0x8280220,0x2820000,0xa2228088,0xaa28200a,0x200222a0,0xa8aa0882,0x2a820a08,0x280a2800,0xa00aaa08,0x2022828,0xaa88a228,0x8028088,0x88802a8a,0xa0aa22a0,0xa2aaaa8,0x220a88aa,0x20a2a802,0xaa220280,0x80288a22,0xa282880,0x2820aaa0,0xa80028a0,0x2a20a082,0x88220aa,0xaa0a8880,0x8aa00828,0x80220288,0x20000082,0x28088028,0x8200aa22,0x222a0aa2,0xa0882200,0x20a08a8a,0xa28002a8,0xaaa08280,0x22828a00,0x8a228020,0x28088aa8,0x2a8a080,0xa22a0888,0xaa8a0a);
remainders_neg[844].modulo_value=1688; remainders_neg[844].bit_mask=create_bit_mask(53,0x2282a22,0x8aaa820,0xa08a82a2,0x2aa22aa8,0x808a00,0x22882a0a,0x2820a82,0x80a2088a,0x822a2a82,0x8800aa,0xa022820,0xa0002288,0x80a08aa8,0x8a0a8888,0x2aaa0800,0x22a0a882,0xa88aaa28,0x2022800a,0xa20a8280,0xa082a088,0x28822280,0x8a8aa0a0,0x22002aa0,0x808a0800,0x8a202,0x2822aa28,0x82a22220,0xaa820022,0xa82a208a,0x22aa8a08,0x8a002aa,0x82a0a008,0x20a82228,0x2088a028,0xa2a8280a,0x800aa822,0x22820008,0x22880a0,0x8aa8a00,0xa888880a,0xa08008a0,0xa8822aa,0xaaa2822a,0x2800aa88,0xa8280202,0x2a088a20,0x22280a08,0xaa0a0288,0x2aa08a8,0xa0800220,0x8a202808,0x2aa28000,0x220282);
remainders_pos[845].modulo_value=65; remainders_pos[845].bit_mask=create_bit_mask(3,0xa1294a52,0x294a5214,0x1);
remainders_neg[845].modulo_value=260; remainders_neg[845].bit_mask=create_bit_mask(8,0x28a0028a,0xa0028a00,0x28a0028,0x8a0028a0,0x282002,0x28a0008a,0x80028a00,0x20a0028);
remainders_pos[846].modulo_value=1128; remainders_pos[846].bit_mask=create_bit_mask(36,0xa2822022,0x8020000,0x2282002,0xa2082002,0x8820028,0x200280,0xa0020080,0x8220020,0x88000008,0x208828a2,0x28020820,0xa288282,0x808a00a2,0x20000a20,0x220a208,0xa0822820,0x28a00,0x88002202,0xa2800080,0x28820a00,0x8a088008,0xa0000820,0xa20208,0x8228a08a,0x20802882,0x28220808,0x228a,0x882020,0x800a08,0x80080002,0x822002,0x8208a28,0x8288080,0x802080,0x8828a00,0x88);
remainders_neg[846].modulo_value=1128; remainders_neg[846].bit_mask=create_bit_mask(35,0x220a28a2,0x820808,0x8a20a08a,0x22802882,0x28820,0x88288208,0x208a0800,0xa28028,0x88080,0xa0002022,0x20828028,0x8220020a,0x20822,0x28808228,0x8a282280,0x200a20a0,0x8820208,0x8a28a,0x22080800,0x20028200,0x2000080,0x208800a0,0x8228a00,0xa202002,0x20082002,0x20a28000,0x8808a202,0x800028a0,0x8008200,0x800808a,0x800a2882,0xa00220,0x80200008,0x80082800,0x20208);
remainders_pos[847].modulo_value=308; remainders_pos[847].bit_mask=create_bit_mask(10,0xaa08020a,0xa208020,0x20aa0802,0xaa080,0x80202a08,0x20a80,0xa08020aa,0x8a08020a,0xaa08020,0xa0802);
remainders_neg[847].modulo_value=308; remainders_neg[847].bit_mask=create_bit_mask(10,0x22808202,0x22808a0,0xa022808a,0x88022800,0x8a02280,0x808a0208,0x2808a022,0x2808a02,0x22008a0,0x2808a);
remainders_pos[848].modulo_value=424; remainders_pos[848].bit_mask=create_bit_mask(14,0x2202aa82,0x8a028820,0xaa022020,0xaaa0a0a,0xa208088,0x8808228,0xa8282aa8,0x8202202a,0x208a028,0xa0aaa022,0x880aaa0,0x2280a208,0xaa808808,0x82);
remainders_neg[848].modulo_value=424; remainders_neg[848].bit_mask=create_bit_mask(14,0xaa8a220a,0x28a00a8,0x228aa8a8,0x82228282,0x2a80800,0x80080aa0,0x20a0a220,0xa8aa8a2,0x8a8028a0,0x282228aa,0x80082228,0xaa002a80,0x22080080,0xa);
remainders_pos[849].modulo_value=3396; remainders_pos[849].bit_mask=create_bit_mask(107,0x20220a2,0x20228008,0x8200202,0xa2800882,0x20a00008,0x8020a288,0x200a00a0,0x20208800,0x8200,0x800a0002,0x20a20220,0x8282280,0x820000,0x228a20,0x208228a,0xa08800a0,0x820808,0x2002088,0xa08a2802,0x28020008,0x820a20a,0x80020080,0x828220,0x8a282,0xa2022802,0x20a28200,0x8a28a28a,0x20a2822,0xa20228,0xa288002,0x220a0802,0x8020008,0x82282080,0x800200a2,0xa28828,0x88200202,0x80820800,0x28008828,0x8a208200,0x228a2002,0x20800,0xa20a080,0x22022820,0x28008,0x2080002,0x882020,0x28028020,0x8a282008,0x80002820,0x8800a28,0x2002082,0x800a2022,0x28220200,0x808828a,0x808a0020,0x20800808,0x8a002208,0x82800022,0x828a20,0x80280282,0x80822000,0x20800,0x280008,0x82880882,0x20a08a00,0x2080000,0x8a2880,0x8208a28,0x82200280,0x2082022,0x8008220,0x8228a008,0xa0080022,0x20828828,0x80200,0x20a0882,0x228a08,0x8808a008,0x828a0802,0x28a28a28,0x828a08a,0x28808a0,0x28a20008,0x88282008,0x20080020,0x8a08200,0x8028a,0x28a20a2,0x20800808,0x2082002,0xa00220a2,0x28820800,0x8a28800a,0x82000,0x28828200,0x8808a080,0xa0020,0x8200008,0x2208080,0xa00a0080,0x28a08020,0xa082,0x220028a2,0x8008208,0x288088,0xa0880802,0x8);
remainders_neg[849].modulo_value=3396; remainders_neg[849].bit_mask=create_bit_mask(106,0x828a2822,0x28a20800,0x8028828a,0x22080002,0x28208800,0x8282200,0xa0820820,0x28a00008,0x80080288,0x820882,0x28228a28,0x8020a208,0x800a0880,0x8a20228,0x8a00a202,0x20000820,0x8028000,0x8a08a000,0x20022082,0x20828800,0x80282282,0x8a0800,0x8020a28,0x8800220a,0x228a2082,0x28220a08,0x2202202,0x828220a2,0x8228a20,0x8220008a,0xa2820082,0x828800,0xa20a008,0x8a0822,0x8220020,0x28828a,0xa0080,0x20800020,0x2280288,0xa2022882,0x8828008,0x82282008,0xa28a20a0,0x8820800,0x8a00800a,0x800028a0,0x20820828,0x220a080,0x8820a0,0x8220,0x8a08a00a,0x8228a2,0x20a28a08,0x80000202,0x208a0,0x28008000,0x208a280,0x20808a2,0x8020228,0x820820a,0xa2880,0x8828008,0x88208080,0x2000002,0x28200208,0x8a008088,0x80022000,0xa00220,0xa288208,0x828028a2,0x800a28,0xa202080,0x208008a2,0x28020020,0x88008288,0x82820002,0x8a20200,0x2080,0x2020082,0x20220220,0x80202002,0x82000020,0x20228800,0x20a080,0xa08008a0,0x200a08,0x88008202,0x82022822,0x28008000,0x8a00a08a,0x208a2822,0x20028008,0x2200082,0x80802880,0x20020a08,0x2008,0x808208a0,0xa08800,0x8a280008,0x20820800,0x20200828,0x8808200a,0xa2882022,0x800a00,0x88200000,0x20000082);
remainders_pos[850].modulo_value=680; remainders_pos[850].bit_mask=create_bit_mask(22,0x28000a0a,0x20028022,0xa820000,0x8200a888,0x882a28,0x80a0220a,0x82080020,0x888008,0x8a22a8a2,0xa0202a00,0x8002082,0xa8080a82,0x2a88a200,0x222008a,0x208220,0x880a0208,0xa82200a0,0x2a008228,0x82a022,0x2800800,0xa0002888,0xa0);
remainders_neg[850].modulo_value=680; remainders_neg[850].bit_mask=create_bit_mask(22,0xa0880282,0xa80208a2,0x20a8088,0xa002000,0x8000a220,0x282a82,0xa0088a0,0x8000080,0x2a2202a,0x20a8a208,0x80882802,0x20008202,0x22002208,0x8aa28802,0x80a80228,0x820a80,0x202a0820,0x228802a0,0x880228aa,0x82088008,0x28082000,0x20);
remainders_pos[851].modulo_value=3404; remainders_pos[851].bit_mask=create_bit_mask(107,0x82020aa2,0xaa822288,0x2aa2208,0x2282282a,0x8a00200a,0x808a288,0x2202028,0xa2a202,0xa28a2080,0x2aa20082,0xa8a08800,0x2a02a280,0xa822a022,0xa2882a2a,0xa08228,0xa20880a0,0x88888002,0x2aa0a0aa,0xaa2a0828,0x20a880a,0xa0aa82a0,0x22228a82,0x200002a,0x2a22a882,0x2888808a,0x8022a00a,0x802a28a,0xa808aa02,0xa222080,0x20080082,0x8aaa8a8a,0x2088800,0xa00200a,0xa0008aa8,0xa2808002,0x20280020,0xaa022200,0x2a22880a,0x8a282a80,0x80a00820,0x28082000,0x80a8088,0xa028002a,0x228800aa,0x2882080a,0x880828a2,0xa8288a8a,0x82282a0,0xa8aa8202,0xa0020882,0x88800a80,0x88a0020,0x28a0a20,0xa280208,0xa22008a,0x222080,0x800a02a,0x2aa2a822,0x28820808,0x2282a0a8,0x82022a2a,0x82288a2,0x2288a02,0x280aaa0,0x2a020a80,0x82028220,0xa02000,0x828a2082,0x288a802a,0x880aaa02,0x2808008,0x2028a080,0x2000a800,0x800a02aa,0x22080a00,0x2aaa2002,0x2008a2a,0x888a0820,0xa202a020,0xa802080a,0xa8802a28,0x22228a00,0xa88a8a20,0x80822,0x28888a80,0x2aa0a82a,0x2a0800a8,0xa8aaa02,0xa0aa8282,0x22222aa0,0x2208a800,0x20a000a0,0x8228a288,0xa882aa8a,0xa80a8880,0x20a2a028,0x8aa8002,0x8a28a820,0xa8a00020,0x80880808,0xa2020280,0x800a2228,0x88288a00,0x8aa80a82,0x882aa208,0x8082228,0x8aa);
remainders_neg[851].modulo_value=3404; remainders_neg[851].bit_mask=create_bit_mask(107,0xa0a822a,0x220aaa00,0x8a22aa80,0xaa0220a2,0x288a882,0x80802a00,0x8aa828a0,0x882a2a8a,0x2a02a800,0xa22a800a,0x20280088,0xa28a2a08,0x202a282a,0x2a00a2a2,0x88280aa0,0x2a800820,0x88a,0x22282022,0x22a280a0,0x8a820082,0x20220a28,0x2aaa020a,0x8a8888a2,0xa22a200a,0xa0000802,0x8aa2882,0x80822a02,0x2080228a,0x82aaa808,0xa880080a,0x2220202,0x8a800008,0x8288a082,0x28880220,0x2a08088a,0xa0a08828,0x228aaa88,0xa2aa0082,0x2a0a200,0x82880a8,0x20802888,0x80820800,0x28a088a2,0xa2008822,0xa00a8082,0x80a022,0x20200202,0x80aa0a28,0x20220a8a,0x2802800a,0x88208,0x800288a8,0x8a0202a8,0x82a08a80,0x82aa8002,0x88aaa008,0x808828a2,0xa22a20aa,0x200a8000,0xaa0a2820,0xa8aa2a2,0x80aa0022,0x8aa0028a,0xa002228,0xa28a8208,0xa8a0aa8,0x8028a888,0xa02a80a,0xa0020822,0x2228a,0x8a080880,0xa8a02808,0xa0802088,0x8828a22,0xaa808288,0xa222288a,0x8a8802a2,0x200a8,0x2a8a20a8,0x208a8082,0x2008a2a0,0xaaaa0208,0x200202a0,0x888080aa,0xa0000200,0xa22820a2,0x22008820,0x8202228a,0x28220a0a,0xa2aaa228,0xaa802088,0xa8288028,0xa202a00,0x200a2202,0x20820008,0x282228a0,0x8022088a,0x2a020a8,0x202808a8,0x8008080,0x2a828a08,0x882a2a0,0xa00288,0x220820a,0xa22a00,0x8080aa20,0x22);
remainders_pos[852].modulo_value=1704; remainders_pos[852].bit_mask=create_bit_mask(54,0x28a0802,0x8228a28,0x8008202,0x2022080,0x28a20200,0x8002200,0x820208a0,0x20008028,0x228a002,0x22020002,0x8008008,0x8a208202,0x828a00a2,0x28200a00,0x8a28a00a,0x2080820,0x8820020,0x88080008,0x8800a2,0x8228020,0x200a208,0xa2800880,0x8000808,0x2002088,0x82080820,0x28028a28,0x8028020a,0xa28028a0,0x20208228,0x8008008,0x20002022,0x20028a20,0x8a008002,0x2882020,0x220008,0x20228a,0x822020,0x20208008,0xa28a208,0xa00828a0,0xa28200,0x8208a28a,0x2002080,0x808820,0xa288080,0x2000880,0x20808228,0x8800200a,0x808a2800,0x8808000,0x82002002,0xa2882080,0x20a28028,0x80);
remainders_neg[852].modulo_value=1704; remainders_neg[852].bit_mask=create_bit_mask(54,0x82020082,0xa20220,0x8008020a,0x828a2800,0x20228a08,0x8008a288,0x28a0020,0x28800820,0x8a20208a,0xa28a0882,0x800800,0x2280282,0x2020822,0x20a08208,0x2202082,0x828000a0,0x28828,0x8080,0x80000822,0xa20828,0x8a082280,0x22080000,0x808000,0x8a08a000,0x28000a0,0x20820220,0x8208282,0x22082020,0x20a00a20,0x80080080,0xa08828a2,0x28820228,0x208008a,0x820028a0,0x8a28800,0x8828a202,0x800a28a0,0x28200800,0x82202280,0x20802020,0x8220a08,0xa002202,0x82882800,0x8000028,0x82200008,0x82880000,0x28000a20,0x8a082,0x22080,0x808,0xa08a08a,0x22002800,0x28220820,0x8);
remainders_pos[853].modulo_value=853; remainders_pos[853].bit_mask=create_bit_mask(27,0xcec9561a,0x2f838f1c,0x3d2a7eaf,0x5b8217f0,0x608bb102,0x5aa65bb9,0x7c688aff,0xce1895db,0x1af6d7d9,0xc9759532,0x6c1a1814,0xce3b1be1,0x15930808,0x6a2f61bd,0x1cc40432,0xda1f637,0xa4ca0616,0xd6132a6b,0x1ce6fadb,0x8fb6ea46,0x56bfd445,0x41a77699,0x76902374,0x2f03fa10,0x7d3d5f95,0xdcce3c70,0x161aa4);
remainders_neg[853].modulo_value=3412; remainders_neg[853].bit_mask=create_bit_mask(107,0x2008a82,0xa20a0280,0xa0a2a222,0x820a8a28,0xa802288a,0x822a8220,0xa0a00222,0x2800802,0x822a0a00,0xa808aa,0xa0828088,0x2a28228,0x880a8080,0xa2a2a820,0x800888aa,0x80282aaa,0x28028a8a,0xa8aa2a2,0x802a2202,0x22a228a,0x8a220888,0x882aaa00,0xaa08aaa8,0xa28a2288,0xa0a08208,0x2a0a8,0xa8a8228,0x2a0000a2,0x82828a8,0x228a2882,0xa882a88a,0xaaa088aa,0x82228802,0x22a20088,0x22a00a8a,0x2a8a8202,0x8a00a22a,0xa0a00a8a,0x88800aaa,0xaa2a2aa8,0xa808820,0xa2a0008,0xa0828a2,0x80a80088,0x82a20aa8,0x800a0002,0x282a00,0xaa20a22,0xa200a822,0x8a820a88,0x2a2828a2,0x2822a22,0x8802000a,0xa880a20a,0x82a0aaa2,0x20222022,0x80820828,0xa20880a2,0xa2280a,0x2aa220a2,0xaa8aa028,0x282aaa0,0xa8800a2,0xa0a88aa8,0x20a20828,0x80a8aaa0,0x200a288a,0x88880020,0xa0000aa,0xa08080aa,0x8020220a,0x222a280,0x22080aa,0x22888aa2,0x2a882,0x8800088a,0x82208802,0x28a28820,0xa0280828,0xa20aaa,0xaaa82280,0xa080a02,0x8208a28a,0xa0088220,0x28880008,0x208aa000,0x22a888a2,0x2a808220,0x80a2a220,0xa8220200,0x2a808082,0xaa800028,0x2000888,0xa88a2802,0x2aa8a80,0x8a082282,0xaa88a82,0x228008a8,0x82aaa0a0,0xa02a8aa,0x228222aa,0xa80a2280,0xa2808822,0xa082080,0xa2022202,0xa2aa82a0,0x2808a);
remainders_pos[854].modulo_value=3416; remainders_pos[854].bit_mask=create_bit_mask(107,0x22822222,0xa20a022,0x22228082,0x2a2a0880,0x2820a02,0x820802a8,0x28028280,0xa00082a,0x2020a220,0x2088aaa,0x82020a0,0x2a208288,0x2200aa0,0xa88222a8,0xa28a200a,0x802802,0xa08a0a8a,0x8a222000,0xa082aa0a,0x82a20088,0x28082028,0x28822800,0x22a8a0a0,0xa28800a2,0xa028082a,0x2aa28a0a,0x2a2aa8,0x80002000,0xa8020020,0x28800020,0x2020a,0x2208280a,0x28a00a88,0x82020828,0x8a008a,0x22008020,0x20880a22,0x228a8080,0x200200a8,0x2280008a,0x8a222800,0x88202200,0x2a08a802,0xa0aa02,0x22a0000,0xaa20a822,0x82a02008,0x88820a80,0xa082a002,0xa0002220,0x2202a80,0x220028a,0x2202202a,0x88808202,0x80a80880,0x80a28008,0xa02a808,0xa088800,0x22800a82,0x8202a082,0xaa20080a,0x80882a08,0xa8,0xa880aa0a,0x22802a20,0xa2008808,0x88002888,0x8a20002,0x882a0080,0x80202a2,0x8888a022,0x80200,0x82a200a2,0x28282080,0x8822a00a,0xa02820,0x28a08080,0x2a080002,0x2080080,0x800,0xa82aa8a8,0xaa0a28a,0x8aa82028,0x888a0022,0x280a0a2a,0x28002882,0x82280820,0xa22008a,0xa2a0aa82,0xa000888,0xa2a0a2,0x8a802802,0x2aa008a2,0x802a8882,0xa80aa008,0x20228208,0x800a0808,0x8aaa220,0xa0088a08,0x28a82000,0x82028280,0x802a8020,0xa880a082,0x880220a8,0xa0820288,0x88880a08,0x888882);
remainders_neg[854].modulo_value=3416; remainders_neg[854].bit_mask=create_bit_mask(107,0xaa0aaa2a,0x2a828a2,0xa2aa080a,0xa2228008,0x8a02828a,0xa000a20,0xa08a0208,0x82888022,0x28a82aa0,0x82800222,0x8028a828,0xa2a00a00,0x8aa80228,0x200aa220,0x2a02a802,0x808a082,0x20028202,0x22aa888,0x28022282,0xa2a0800,0x2080a0a0,0xa00aa008,0x2a202820,0x2200882a,0x282080a2,0xa2220282,0x88a22220,0x888a088,0x20828828,0x200888a0,0x808a8a82,0xaa00a082,0xa0208200,0xa8a00a0,0x88028002,0xaa880828,0x28008222,0xa2020808,0xa80a8820,0xaa008802,0x2aa2088,0xa8a288,0xa280200a,0x282282,0x82a28888,0x222820aa,0xa20a880,0xa0208,0x280a208a,0x2888aa28,0xaa8a200,0x82a80a02,0xaa0aa8a2,0xa8a,0x8200008,0x82a0080,0x828a2000,0x2800080,0xa208820a,0xa0a2802,0x22a08082,0x8002280,0x88888020,0x20082202,0x2a08a2a0,0x22880080,0x8a000,0x8022888a,0xa20808,0x808a822a,0x282a,0x8808a80,0x22a882a,0xa020a808,0xa22082,0x882820a8,0xa0280008,0xa280880a,0xa808800,0x80888088,0x20222020,0x2202a02,0x22028a0,0x280aa,0xa0828222,0x2088a002,0x2a080a8,0x822a8802,0x2a20220a,0x82800000,0x882a202a,0x208a00a,0x22288022,0xa2000a,0x20022880,0xa8a20a80,0x8820080,0x80222228,0x28800200,0x2020a880,0x28a0a08,0x82208a8,0x2000280a,0x8a2820,0x200a8200,0x8200,0x2);
remainders_pos[855].modulo_value=1140; remainders_pos[855].bit_mask=create_bit_mask(36,0x80802082,0x28228820,0x8088080,0x2020022,0xa00000,0x808220a,0x2080002,0x20,0x800008a,0xa0002822,0x28820a00,0x8288000,0x20080020,0x8828800,0x8a008,0x822002,0x20020000,0x8200,0x800820,0x20000008,0x80088,0x220200a2,0x800228,0x82008002,0x22800022,0xa00a08,0x2088280,0xa2000,0x8008000,0x2080002,0xa00a0882,0x8000000,0x2088008,0x82802022,0x20208228,0x82080);
remainders_neg[855].modulo_value=1140; remainders_neg[855].bit_mask=create_bit_mask(36,0x2802,0x20220020,0x8,0x828208a2,0x208808,0x80002282,0x82000882,0x8008028,0x80008002,0x20882022,0x20020200,0x80200080,0x208000a0,0x20008,0x80002000,0x800a2882,0x28028800,0x8080208,0x800000a0,0x20808800,0x88000000,0x220a0822,0x8008220,0xa00008a,0xa2080022,0x8200200,0x8a000200,0x822080,0x800808,0x82008002,0x20820002,0x800,0xa000080,0x20028a2,0x20a00a20,0x2008);
remainders_pos[856].modulo_value=856; remainders_pos[856].bit_mask=create_bit_mask(27,0x8a288aaa,0x8282220a,0xa20a2880,0xaa002288,0x280802a0,0xa020aaa8,0x22222002,0xa2a02aa,0xa8a82800,0x2aa00a02,0x20a20880,0x22828a88,0x882080a2,0x22000000,0x888a0208,0x822a282,0xa802208a,0x2a80a00a,0xa000282a,0x88aa80a8,0xa800888,0x282aaa08,0xaa0a8020,0x8a228800,0x820228a0,0xa2a08882,0xaaa228);
remainders_neg[856].modulo_value=856; remainders_neg[856].bit_mask=create_bit_mask(27,0x2a00222,0xa0aaa82,0x2a82a008,0x2288a200,0xa0808a28,0x28a82220,0xaaaaa88a,0x82a28a22,0x2020a088,0xa228828a,0xa82a8008,0xaa0a0200,0xa8082a,0xaa888888,0x28a80,0x80aa2a0a,0x200aa802,0xa2082882,0x2888a0a2,0x220820,0x82088000,0xa0a22280,0x228208a8,0x2aa0088,0xa8aa028,0x2a28000a,0x222aa0);
remainders_pos[857].modulo_value=857; remainders_pos[857].bit_mask=create_bit_mask(27,0xe6a5a316,0x76be4e1b,0x907ea3c5,0x17b4edf6,0x442570b1,0x6b2a17d4,0xd45bd79e,0x1174d1a,0x1782cda9,0x38900eb3,0x31ddb38,0x36e70644,0x799f437c,0x7b9a3167,0xb0fb0be6,0x89839d,0x70736ee3,0xa335c024,0x256cd07,0xad62cba2,0x59e7af68,0x88afa153,0xa2343a90,0x25bedcb7,0xba8f15f8,0x9f61c9f5,0x1a31695);
remainders_neg[857].modulo_value=3428; remainders_neg[857].bit_mask=create_bit_mask(107,0x2a282a8a,0xaa228282,0x8a22a08,0x8a28202a,0x88a8a828,0xa2a28a08,0x8820a02,0x888a8082,0x8a0a0020,0xa008822a,0x8a8002a0,0xaa2a8a88,0xa0028aa0,0xa202a8aa,0x2822822a,0x88000a00,0xa8aaa22a,0x2aa808a8,0x8a8a008a,0x20aa022a,0x802a22a0,0x22288a,0x2aa8a208,0xa822002a,0x22028820,0x2a80028,0xa02a8a08,0xaa00828a,0x220a000,0x20a8208a,0xaaa2a002,0x20008228,0xa00a88a2,0xa8202a22,0x8a8aa202,0xaaa28802,0xa8a8a880,0x8aa22a,0x20a00280,0x22aa20a,0x2aa8aa,0xa2a82a8a,0xa888a8a,0x80282a00,0x828aa208,0x88882002,0x882080a,0x2a2a0282,0xa888828a,0xa288a0a8,0x2882a020,0x22a882a2,0xa2a20a0a,0x828aa0,0xa0a2020a,0x2880220,0x2a028822,0xa08820a,0x82888808,0x82a02020,0xa8a288a2,0xaa288880,0xa2008282,0xa0282a88,0x808880aa,0x80280200,0x8002aa0,0xa20022a0,0x82aa0a20,0x22008aaa,0x88080800,0xa882000a,0xa2008082,0x202a2802,0x880aa02,0x228aaa22,0xaa020008,0x8a280a22,0xaa0a22a0,0x828aa00a,0xa0802200,0x2aa802a8,0x2882a228,0x2aa2280a,0xa2080020,0x88222aa8,0x2202a80,0x22a00a2a,0x8aa08080,0x8a80020,0x22000808,0xa0aaa880,0x2282282a,0x802a200,0x82aa00,0x8080200a,0x2aa8088,0x2288aa0a,0x2aa0a080,0x8a80888a,0xa0a288a2,0xa0820202,0x28080888,0x2a282088,0xa02208a0,0x82822008,0x80282022);
remainders_pos[858].modulo_value=3432; remainders_pos[858].bit_mask=create_bit_mask(108,0x2020082,0x20228820,0x82080288,0x880080,0x8a20a00,0x88008,0x820a2,0x8020228,0x820a088,0x20020a2,0x800a00,0x88288202,0x2022820,0x8220a28,0x282202,0x828a0020,0x820800,0xa20208a,0x2020,0x8228a00,0xa20a008,0x80000,0x8820220,0x8000,0xa0020802,0x8a00220,0x8200028a,0x800002,0x20000228,0x8a00a002,0x80822882,0x28a20020,0x8a280,0x20082002,0x28208000,0x28020a,0x800802,0x820820,0x2202280,0x20800020,0xa00020,0x2000,0x2822080,0x20820028,0x820000a,0x80080800,0x28200220,0x88200200,0x882080,0xa28000,0x80088288,0x2000020,0x20a20800,0x2008002,0x208a0880,0x8000,0x82200208,0x28a0022,0x8220000,0x80082202,0x80082800,0x20200208,0x82000,0x8208a0,0x8828028,0x8000002,0xa0000,0x20808,0x88088008,0x20820002,0x20020008,0x80280080,0x20828a0,0x8200800,0x8a200080,0x8a2802,0x28820208,0xa00a282,0x80000880,0x20028,0x80008280,0x800a20a2,0x20800a08,0x2000080,0x80822000,0x200008,0xa08a000,0xa2882020,0x8000000,0x808a008,0x208200a2,0xa28200,0x88280008,0xa0882080,0x28808028,0x82282208,0xa0020080,0x8008000,0xa08208a,0x80802022,0x8200028,0x220008a,0xa08a2020,0x220000,0x80208202,0x22880822,0x808008,0x82);
remainders_neg[858].modulo_value=3432; remainders_neg[858].bit_mask=create_bit_mask(108,0x828a0002,0x28220028,0x2008200,0x80000800,0x220208,0x80000080,0x80802822,0x28a20,0x80202000,0x2882022,0x8008208,0x20028a,0x820220a0,0xa28220,0x8820a202,0x20208a0,0x28008,0x82282002,0x80882820,0xa20200,0x2282080,0x80800880,0x28228,0x8080088,0x208a0082,0x208a20,0xa080202,0x80080802,0x20808a20,0x208200a,0x22002,0x20228828,0x88002200,0xa0002082,0x20200808,0x88208282,0x80082,0x8020028,0x8220a208,0xa00808a0,0x8208828,0x8000a088,0x20a2800,0x28020820,0x80288082,0x80,0x20208a28,0x288280,0x80002000,0x220808,0x8000200,0x82800820,0x28228008,0xa08000a,0xa0020000,0x8800008,0xa280280,0x820208a2,0xa28800,0x800228a,0x802080,0x28a08200,0x800a008,0x80020020,0x20820,0x8008808a,0x80820000,0x28000,0x80,0xa00a0882,0x28820800,0x208000,0x82002020,0xa08008,0x2208008,0x22082,0x20028a00,0x8200220a,0x80000,0x8828820,0x80200,0x822822,0x20000200,0x8a088008,0xa2800,0x8208800,0x2002088,0x220020a0,0x808008,0x80002080,0xa0020822,0x8220a00,0x200000,0x20002800,0x20000820,0xa202200,0x20820800,0x800800,0x8200a002,0x820a2,0x208020,0xa288002,0x200228a0,0x8a20808,0x28028a,0xa2000022,0x800,0xa);
remainders_pos[859].modulo_value=3436; remainders_pos[859].bit_mask=create_bit_mask(108,0xa8aaa2a,0x22aa2282,0x802a082a,0x2a280808,0x82288a80,0x8a02aa0,0x2202aa88,0xaaa88022,0x2a28228a,0x20822a02,0x82008a2a,0x88820a82,0xaa2220a0,0x8220222a,0x28000a0a,0xa8228000,0x88a20282,0x800a2228,0xa8a00a80,0x208aa282,0x80802888,0x228aa0a8,0xa282a808,0x2a820a2a,0x2aa00a2a,0x82800280,0x2822a8a,0xa02008a,0xaa0aaa,0x8082a8,0x800828a,0x8a2a00a,0x2a202802,0x28a220aa,0x200a0802,0xaa2a02a8,0x22208882,0xa00a0a88,0xa0aaaa08,0x8a0282aa,0x800088a8,0x22282888,0x8a0a028a,0x2880822a,0x808a808a,0x208a0,0x8888aa20,0x22a2000a,0x8a2800a8,0x80aa0220,0xaa22a2a0,0x880a280,0x28a0880,0x20008002,0xa280800,0xa0220022,0xa88aa028,0xaa020a8,0x28a2088,0x8a882a0,0xa2222a00,0x800008a,0x2a2020a2,0x20228a20,0xa0a2a88,0x82888a28,0x20002222,0x280a22a2,0xaaa0aaa8,0xa00a20a,0x2088822a,0xa8aa822,0xa0082a8,0x88a28802,0x808a8aa0,0xa8a20802,0x20020a00,0x20200a28,0xaa002a8,0x80a0aaa,0x88280a20,0x282a2a,0xaa8028,0x82a8a8a,0xa828aa8a,0xaa288202,0x802022a0,0xaa208222,0xa2a828,0x8a00202a,0x8a22288,0x2882a828,0x28000,0x80882a0a,0x888aaa88,0x82220a0,0x2008282a,0x88208a8a,0x828a880a,0x22aaaa28,0xa8088880,0x80a2022a,0x228820aa,0x28a802a,0xa802202,0x8aa88a82,0xaa2a0828,0xa8a);
remainders_neg[859].modulo_value=3436; remainders_neg[859].bit_mask=create_bit_mask(108,0x820222a2,0xaa22aa0a,0x8a280a2,0xa2a08080,0xaa00208,0x8028a228,0xaa8a2200,0x222008aa,0xa2a0aa02,0xa80aa28a,0xa8802a2,0xa820a,0x22aaa828,0xaa8aaa2,0xa0888282,0x20aa0888,0x2a8a0a,0x882aaa0,0x20288208,0xa8022a0a,0x808a000,0xaa022820,0x2a0a2080,0xa20a82a2,0xa22882a2,0xa088a08,0x820aa202,0x828a8802,0x88228222,0x88080a20,0x80880a02,0x802a2882,0xa2a8a08a,0xa02aa822,0xa882808a,0x22a28a20,0xaaa8000a,0x28828200,0x28222280,0x28a0a22,0x8880020,0xaaa0a000,0x2828a02,0xa0080aa2,0x8020802,0x888a8028,0x22a8,0xaa2a8882,0x2a08820,0x8228aa8,0x22aa2a28,0x80082a08,0x8a028008,0xa888088a,0x82a08088,0x28aa88aa,0x200228a0,0x8228a820,0x8a02a800,0x80200a28,0x2aaaa288,0x80888802,0xa2a8a82a,0xa8aa02a8,0x8282a200,0xa0002a0,0xa888aaaa,0xa082aa2a,0x22282220,0x82882a82,0xa8000aa2,0x820220aa,0x82880a20,0x2a008a,0x8020228,0x202a808a,0xa88a8288,0xa8a882a0,0x82288a20,0x80828222,0xa082a8,0x88a0a2a2,0x882208a0,0x80a20202,0x20a02202,0x22a00a8a,0x8a8aa28,0x22a80aaa,0x882a20a0,0x288a8a2,0x802aaa00,0xa00a20a0,0x888a0888,0x800a282,0x22200,0x80aaa828,0xa880a0a2,0xa80202,0xa020082,0xaa2222a0,0x20800008,0x82a8aa2,0xaa00a822,0x8a0208a2,0x8208aa8a,0x220020a,0x22a280a0,0x202);
remainders_pos[860].modulo_value=1720; remainders_pos[860].bit_mask=create_bit_mask(54,0xa80202,0xa82a2,0x288a2288,0x828228a0,0xa0808208,0xa8aa0a80,0x22088880,0x2220,0xa202802,0x8208020a,0x80202800,0x22a8200,0x200a8000,0x8a22088a,0x2a00,0x20082a88,0x20028080,0xa080208,0x802808a,0x80888000,0xa0222208,0xa02a0aa2,0x22082020,0x80a28828,0x2288a22,0x8a82a00,0x280802a0,0x200a8020,0x8000a82a,0x288a228,0x8828228a,0xa080820,0xa8aa0a8,0x2208888,0x20000222,0xa0a20280,0x8208020,0x8020280,0x22a820,0xa200a800,0x8a22088,0x800002a0,0x20082a8,0x82002808,0xa0a08020,0x802808,0x88088800,0x2a022220,0xa02a0aa,0x82208202,0x280a2882,0x2288a2,0x8a82a0,0x80802a);
remainders_neg[860].modulo_value=1720; remainders_neg[860].bit_mask=create_bit_mask(54,0x88200a8a,0x88020222,0x2002a200,0xa02a028,0x20080a00,0x20220208,0xaa000000,0x888a2a8,0x82a0a08a,0x2808a02,0x8a82088,0x8a220a80,0x28820080,0x2a28002,0x8088a208,0xa8802200,0xa80a0800,0x2800280,0x80820802,0x808,0x28aa2a80,0x28228222,0x2280a0a8,0x82200a0,0x82a0022a,0x80202288,0xa0008a20,0x288200a8,0x8802022,0x82002a20,0xa02a02,0x820080a0,0x2022020,0x8aa00000,0xa0888a2a,0x282a0a08,0x802808a0,0x8a8208,0x8a220a8,0x22882008,0x802a2800,0x8088a20,0xa880220,0xa80a080,0x20280028,0x88082080,0x80,0x228aa2a8,0x82822822,0x2280a0a,0xa082200a,0x882a0022,0x8020228,0x8a2);
remainders_pos[861].modulo_value=3444; remainders_pos[861].bit_mask=create_bit_mask(108,0x22082822,0x8a00820,0x80080082,0xa2000002,0x20228208,0x82282208,0x820080,0x8808220,0x2282280,0xa2000082,0x8020020,0xa20020a,0x80802080,0x800008,0xa082000,0xa2,0x20028000,0xa080,0x22000082,0x8820020,0x8a28a,0x800a2820,0xa00a00,0x82000208,0xa0022000,0x8028200,0x2002200,0x80022002,0x28020a00,0x8002200,0x820002,0x8028028,0x20a280,0x828a2880,0x20200208,0x2080002,0x20082800,0xa00,0x82280000,0x2082,0x8800008,0x80082008,0x82820022,0x28200200,0x2080002,0x800a20a2,0x220808,0x8080208,0x208220a2,0x28820a20,0xa000002,0x82080080,0x20208028,0x8a20a082,0x808820a0,0x8228020,0xa002002,0x22880000,0x20808a08,0x208a088,0x80020802,0x220208,0x808a08a,0x82880002,0x28200800,0x288008,0x22020082,0x20000,0x88282080,0x2,0x800a00,0x8000282,0x80880002,0x28220800,0x8000228a,0x20028a0,0x20028028,0x2080008,0x2800880,0x200a08,0x8080088,0x2000880,0xa00828,0x8200088,0xa0020800,0x200a00,0x828a,0x220a28a2,0x8808008,0x82000,0x8020a0,0x28,0xa08a000,0x20000082,0x20220000,0x8a002080,0x20a0800,0x8a08008,0x88082000,0x22002882,0x20008820,0x88202008,0x80820882,0x8a20828,0x280000,0xa2082002,0x8808200,0x88282);
remainders_neg[861].modulo_value=3444; remainders_neg[861].bit_mask=create_bit_mask(108,0xa2802022,0x208020,0x800a,0x22080882,0x28a20a00,0xa202280,0x800a0000,0xa28,0xa20a200,0x22880802,0x28828,0x82208282,0x82800,0x8008000,0x8200a000,0x880822,0x20820808,0x88082008,0xa2800802,0x20828,0x88002202,0x822020,0x8208200,0x2088280,0x200a2080,0x820a08,0x8a082288,0x8a2082,0x20828208,0x82280,0x800a0882,0x20820,0x88202208,0x2022000,0x28a08200,0x8a00800a,0x20802080,0x808208,0xa208088,0x80002802,0x800,0x800a080,0x20a0822,0x20a08a00,0x8200808a,0x22822,0x8a28000,0x80000280,0xa00a2022,0x20028228,0x2088082,0x2800800,0x28200820,0x220200a,0x2820,0xa20020,0x8208a00a,0x22000880,0x20000200,0x8a002000,0x820002,0x8a20a00,0x80002002,0x2000802,0x20a08000,0x80200080,0xa20a0802,0x8828808,0x202008,0x80880082,0x8000208,0x88202,0x882,0x20228008,0x800a202,0x82802020,0x28820020,0x8a008000,0x2080000,0xa08200,0x80008000,0x82800000,0x8200020,0x80288000,0x208a0000,0xa08200,0x80080202,0xa2022022,0x800,0x88002088,0x82020,0x8808820,0x2002080,0xa0880802,0x28228808,0x200a008,0x82820080,0x200000,0xa008,0x22882002,0x20800028,0x28a000,0x20002,0x220020,0x88208088,0x22802802,0x200,0x20a);
remainders_pos[862].modulo_value=3448; remainders_pos[862].bit_mask=create_bit_mask(108,0x8a282a8a,0x8a0282a2,0x2822a880,0x8a2880aa,0xaa820,0xa02a8a88,0x2a20080,0x8088a08a,0x888aa208,0x20aa822a,0xaaa2a220,0xa822aa82,0xa0888080,0x2a0a20a2,0x808000a,0x28222a82,0x8aaa2a22,0x22a8220a,0xaa2000,0xa00a0008,0x808aa28,0x8020228a,0xa822088,0xa20802a0,0x8022a82,0x2a2828a0,0x202282a0,0x2a02822a,0xaa082820,0x28022a8,0xaa02a8a2,0x888a2280,0xa0822a2a,0x82008a8,0xa8aaa0aa,0x2aaa200a,0x80a22802,0x22202000,0xa2802228,0x20aaa8a8,0x20a00a0,0xa8a888a,0x2800228,0x2a220200,0x8022800a,0x88a20088,0xa08a088a,0xa8aa202,0xa880802a,0x8a2800aa,0x200a8820,0x8a880228,0x820282a0,0x802820,0x82082802,0xa20a8280,0x8288022,0xa20822a0,0x2aaa0028,0xa0a80202,0xa808aa2,0x22a220a2,0x222008a,0xa8a00288,0x80008088,0xa0288002,0x80a222a2,0x80a00a0,0x8a2a2aaa,0x88288802,0x2000808,0xa880288a,0x2aa008aa,0x20aa0aaa,0xa2a2008,0x22a8a882,0xaa0288a2,0x808a2a80,0xaa2a8802,0xa8082820,0x8a88280,0xa80a8288,0x200a2828,0x8a82a880,0xa00a8020,0x2220882,0x20a28808,0xa28aa20,0x2000a0,0x880008aa,0xa2a0882a,0x2888a8aa,0x2082a888,0xa8a00020,0xa8a08a0,0x2a020222,0xaa82aa88,0x8088a8a,0x22a882aa,0x2208aa2,0x80a20a22,0xa02008a,0x22a2a8,0xa2082aa0,0x28aa0228,0xa2022a88,0xa28a8280,0xa2a828);
remainders_neg[862].modulo_value=3448; remainders_neg[862].bit_mask=create_bit_mask(108,0x2a0a202,0x28a0a2a,0xa0aa2008,0x2a00822,0x888220a8,0x28a20200,0x8a2a8808,0x8002802,0x22a80,0xa8220aa2,0x222a2aa8,0x20aa220a,0x28000808,0xa282282a,0x80808882,0xa0aaa20a,0x222a2aa,0xaa20aa82,0x8822a888,0x28828880,0x808022a0,0x8a8aa02,0x820aa800,0x2a808a28,0x808aa20a,0xa2a0a028,0xa8aa0a28,0xa28a0aa2,0x2280a0a8,0x8a08aa20,0x228a202a,0x2aa08,0x280aa2a2,0x80a88020,0x20222822,0xa222a882,0x82aa08a,0xaaa8a888,0x2a08aaa0,0xa8222020,0x8a828828,0x82020002,0x8a088aa0,0xa2aa8a88,0x8aa0882,0x2a8800,0x28028002,0x82022a8a,0x200808a2,0x2a08822,0xa88200a8,0x2008aa0,0xa8a0a28,0x8808a0a8,0xa80a08a,0x2a820a08,0x80a008aa,0x2a80aa28,0xa22288a0,0x28208a8a,0x8208022a,0xaa2aa82a,0x8aaa8802,0x20288a00,0x8880800,0x28a0088a,0x82aaa2a,0x80828028,0x2a2a222,0xa0008a,0x8a888080,0x2008a002,0xa2288022,0xa8228222,0x82a2a880,0xaa20200a,0x228a002a,0x802a208,0x22a2008a,0x2080a0a8,0x80200a08,0x20820a00,0xa882a0a0,0x20a2008,0x288208a8,0x8aaa800a,0xa82a0080,0x82a022a8,0x88a88828,0x888022,0x2a2800a2,0xa0002022,0xa80a2000,0x202888a8,0x82028028,0xa28a8aaa,0x220a2200,0x80800202,0xaa200a22,0x8aa8022a,0x82a82aa,0x828a8802,0x88aa2a20,0x2a80a228,0xa0228aa0,0x2a8aa200,0x2a020a08,0x2a20a0);
remainders_pos[863].modulo_value=3452; remainders_pos[863].bit_mask=create_bit_mask(108,0x22828a82,0xa22282a0,0xa00aa00a,0x2822208,0xa20088a,0x82a82a2,0x8a88800,0xa880002a,0x820aa8aa,0x822080aa,0x8a822202,0x28aaa20a,0xa88a0a0,0x2a8aaaa8,0xa0a0a2aa,0x2082a20,0x88200aa,0xa0888a20,0x8a22a2a8,0x8aa08a0,0xaaa80888,0xa282002,0x82a8a882,0x8808080a,0xa88a0022,0x88228000,0x8aa82a2,0x80a0022,0x2aaa08a2,0x8aa8220,0x2a2a2a2,0x8a20200a,0xaa8a082,0xa222a000,0x28228022,0x20080882,0xa8822228,0x802a8a22,0x2880a28a,0x80082828,0xa0000200,0xa8282202,0x2880020,0xa888a02,0x2a288a,0x20028a,0x80aaaa20,0xaaa22022,0x880a00a2,0x8222a882,0x22888a0a,0x2a802a8,0xa80a0888,0xa020a08,0xa20a080a,0x2220a02,0x82a802a8,0x2a0a2228,0x2822a888,0xa8a00a02,0x288088aa,0x8aaaa0,0xa280080,0xa228a80,0x80a222a,0xa0800228,0xa8088282,0x20080000,0x82828200,0x2a28a022,0xa88a2a80,0x82888822,0x88220200,0xa8802882,0xa888,0x2820a2aa,0xa00808a,0x8a8a8a8,0x80882aa2,0x8a20aaa,0x8800a02,0x28a82aa2,0xa0002882,0x28800a22,0x2a020202,0x822a2a8,0xa800828a,0x22202aa,0x20a20aa2,0xa2a8a88a,0x8a2220,0xaa00822,0xa08a8208,0x8aa8a0a0,0x2aaaa2a,0x80a0a22a,0x2a08aaa2,0x2808882a,0x2aa02088,0xaaa2aa08,0xa800022,0x222a2,0x8a82a82,0xa22008a,0xa2088828,0xaa00aa00,0x80a82888,0x82a2828);
remainders_neg[863].modulo_value=3452; remainders_neg[863].bit_mask=create_bit_mask(108,0xaa0a020a,0x2aaa0a28,0x28822882,0x8a0aaa80,0x82a88002,0x80a20a2a,0x80200088,0x200888a2,0xa822022,0xaa80822,0x20aaa8a,0xa0222a82,0x82002828,0xa2022220,0x28282a22,0x8a80a2a8,0x800a8822,0x280002a8,0x2aa2a20,0x80228028,0x22208000,0x82a0a88a,0xa20200a,0x808082,0x200288aa,0xaa0888,0x220a2a,0x808288aa,0xa222802a,0x80220aa8,0x8a2a2a2a,0x2a8a882,0x8220280a,0x2aaa2888,0xa0aa08aa,0xa880800a,0x200aaaa0,0x8a202aa,0xa0082a02,0x880a0a0,0x28888a88,0x20a0aa8a,0x8a0088a8,0x8200028a,0x88a2a002,0x88a88a02,0x82222a8,0x222aa8aa,0x82882a,0xaaa200a,0xaa000282,0x8a208a20,0x20828000,0x828a8280,0x2a828082,0x8aaa828a,0xa208a20,0xa282aaa0,0xa0aa2000,0x2028828a,0xa0080022,0x88022228,0x82a08808,0x82aa0208,0x8082aaa2,0x28088aa0,0x20800a0a,0xa8808888,0xa0a0a88,0xa2a028aa,0x2002a208,0xa0000aa,0xaa8a88,0x2008a00a,0x88882000,0xa0a82a22,0x82880802,0x80202020,0x800a22a,0x802a8222,0x8008828a,0xa020a22a,0x2888a00a,0xa00882aa,0xa28a8a8a,0x80aa2a20,0x20880a02,0x8aaa8a22,0xa82a822a,0x2a202002,0x8802aaa8,0x822880aa,0x28020a80,0x2202828,0x8a2222a2,0x8282aa2,0xa280222a,0xa08000a2,0xa228a800,0x222a2280,0x820888aa,0x888aaa2a,0x8020a20a,0x82aa8802,0x2a8000a0,0x22882288,0x820a000,0xa2a0a0);
remainders_pos[864].modulo_value=24; remainders_pos[864].bit_mask=create_bit_mask(1,0x880022);
remainders_neg[864].modulo_value=24; remainders_neg[864].bit_mask=create_bit_mask(1,0x8a2);
remainders_pos[865].modulo_value=865; remainders_pos[865].bit_mask=create_bit_mask(28,0xb92753de,0xc56b0615,0x22960111,0xda913cc7,0x82a98901,0x240c4314,0x875a5817,0x79cc4121,0xc8e12821,0x40ac4c6b,0x10850992,0x2c3a0af2,0x11c20135,0xca37b14e,0xb2010e21,0x3d4170d2,0x26428421,0x58c8d409,0x10521c4f,0x1208ce7a,0xa0696b86,0xa308c093,0x2465504,0x8cf2256e,0x2201a513,0xa1835a8e,0xef2b9276,0x1);
remainders_neg[865].modulo_value=3460; remainders_neg[865].bit_mask=create_bit_mask(109,0x20aa0a02,0x8228a80,0x2a0a8088,0x200a00a,0x2200008,0xa8800a88,0xa028082,0x200080a8,0x20a0a8,0x80208022,0x8808000a,0xa0228222,0x80a80a8,0x2a22882,0x22888228,0xa088280a,0xaa0a0820,0x80,0x88828082,0x82800222,0x28a0220a,0x2a000802,0xa8a8088,0x22a8a2,0xa288280a,0x280a0202,0x22220aa2,0x2a822220,0x20280a2,0x80a08a2a,0x28aa2002,0x880a8a82,0x8002a0,0x822028a2,0x22000a0a,0x80a088a,0x8000002,0x208282a8,0x80a08828,0xa2088a22,0x8a22a00,0xa80a8082,0x220a2028,0x8000808a,0x2008200a,0xa8282002,0xa8080020,0x80a0280,0x8a8008aa,0x80002200,0x80280200,0x880a82a2,0xa8a2080,0x282a820,0x20008222,0x8820008a,0x2008882,0x22882280,0x8222a88a,0xa8000a,0x80a02880,0x2a288802,0x2a082008,0x8a20a822,0x8288aa00,0x28202200,0x8a008800,0x200200a0,0x80a00a,0x20880a02,0x820808,0xaa2a88a0,0x880a0a0,0x80a2a020,0x282200,0xaa0a82,0x80008882,0x2a000020,0x820882,0x280a220,0x20220002,0x220222,0x2280a020,0x88208002,0x2a20,0x88800082,0xa82a8020,0x220a0020,0x2a28080,0x82808802,0x88aa2a82,0x8208002,0x28088208,0x2808020,0x80200228,0x8802882,0x22020a00,0x2a88a080,0xa822880,0x2082a22,0x88a2a08,0x8a0280a0,0xa8000,0x8aa220a8,0xa208a228,0x88802000,0x800208a0,0x20800228,0x2);
remainders_pos[866].modulo_value=3464; remainders_pos[866].bit_mask=create_bit_mask(109,0x222a0222,0x20022a82,0xa82aa028,0xa2008a20,0xa222a208,0xa0222aaa,0x8aa02088,0x2822020a,0x882a,0x202a0a0a,0xaa2a8288,0x802a82a0,0xa2aa880a,0xa80820a,0x80222a2,0xa2822882,0x2822a20,0xa0a28a8,0x2800888a,0xa8aa82a,0xa8a202,0x222aa8a0,0xaa2aaa2,0xaa008aa0,0x28a22820,0x822a88a8,0xaa8a2aaa,0x208002,0x8802280,0x28220828,0x8aa00a,0x200aa,0x8002222,0xa2080aaa,0x280080a2,0x888aa820,0x820a0a0,0x202282a8,0x8822820a,0x222a8a2,0xa28280a2,0xa8800200,0x2802a80,0x8280200a,0xa0a02a28,0x288aaaa0,0xa2a22820,0x8a2a0080,0x222a08,0xa2022200,0x208aa208,0x2a00280a,0x8022aa28,0x22a02222,0x880a8882,0xaa880288,0x2800a828,0x8aa208a0,0x88808a20,0xa8880000,0xa8a220,0x288a8a02,0xaaa22808,0xa80a0a0a,0x8028228,0xa80280a0,0x80022a02,0x2828a00,0x2a88808a,0x8288228a,0x828808a0,0xa08202a,0x2aa2220a,0x2002808,0xa0208a8a,0x880020aa,0x800088,0xaa200aa,0x208828a0,0x88022028,0x2080002,0xa8a2aa80,0x22a882aa,0x288a282a,0xa200aa08,0xaa8aa00a,0x2aa8888a,0x8a2a000a,0x2aa2a080,0x220028a8,0x28a0a0a2,0xa882802a,0x28828a08,0x88802082,0x8202a08a,0x22aa8aa0,0x82a802a0,0x82a8aa0a,0xa0a80822,0x220000a0,0x808828a8,0x80aa2a0,0xa8880a22,0x8a888aaa,0xa2008a20,0xaa82a08,0xa8800828,0x80a88882,0x88);
remainders_neg[866].modulo_value=3464; remainders_neg[866].bit_mask=create_bit_mask(108,0xaaa28aaa,0xa88aa20a,0x20a228a0,0x2a8802a8,0x2aaa2a80,0x28aaa222,0x228a800,0xa0aa8a82,0x888800a2,0xa8a28282,0x22a20a00,0x8a20a28,0x2a220082,0x82080a82,0x808aaa2a,0x2a0aa00a,0x8a0aa2a8,0x8282a020,0xa0880002,0x820220a2,0x88202a8a,0xaaa22028,0x822a222a,0x22880228,0xa02aa0a8,0xaa20020,0x2202a222,0x88a8088a,0x8008aa08,0xa0aa80a0,0x88022882,0x888a8822,0x8088aaaa,0x2a808222,0xa088082a,0x220a8,0x80a82828,0xa8aa0a20,0xaa0a82,0x8aaa202a,0x2a02082a,0x20088a88,0x8a08a208,0xa08a882,0x2828a2a0,0xa0022228,0x2a2aa0a8,0x2a28808,0x88aaa280,0x2a8aaa88,0xa8022a80,0xa288a082,0x8aa22a0,0xaa28aaaa,0x82000a,0x22008a00,0xa08820a0,0x22a8028,0x802a8,0x20008888,0x88202aa8,0xa002028a,0x222aa080,0x20828282,0x808a0aa0,0x208a0828,0x88aa28a,0x8a0a0288,0xa2000802,0xa00aa02,0xa008028,0x8280a8a2,0xa22aaa82,0x8a88a080,0x28a80202,0x88a822,0x88088800,0x822a8822,0xa800a028,0x8aa8a0,0x8a80888a,0x202a2208,0xaa200a22,0xa002a0a2,0x2a882280,0x22022882,0xa2200002,0x2a28882,0xa22a2808,0xaa88a020,0xa028282a,0x200a08a2,0xa00a0280,0x8a80a,0xa0a2802,0xaa220228,0xa208a28,0xa202282,0x282080aa,0xaa888828,0x800a020,0x80822a28,0x200082aa,0x2000222,0x2a8802a8,0x8220a280,0x200880a0,0x820000a);
remainders_pos[867].modulo_value=204; remainders_pos[867].bit_mask=create_bit_mask(7,0x2802802,0x28028028,0x80080280,0x2002802,0x28028028,0x80280280,0x802);
remainders_neg[867].modulo_value=204; remainders_neg[867].bit_mask=create_bit_mask(7,0x82082082,0x20820820,0x8008208,0x82082082,0x20820820,0x208208,0x82);
remainders_pos[868].modulo_value=1736; remainders_pos[868].bit_mask=create_bit_mask(55,0xa02220a,0x20880080,0x820a0a88,0x22a2800,0x2220800,0x82a002a8,0xa8aa2282,0x2a0a0a22,0x2200aa00,0xa2000082,0x820800a2,0x88208a82,0x2800800,0x882a8222,0x2002808,0x2a208220,0xa0020828,0x200008a8,0xaa00888,0x8a0a0a80,0x288aa2a8,0xa800a828,0x2088802,0x28a8800,0x2a0a0820,0x20022082,0x8880a00,0xa02220aa,0x8800800,0x20a0a882,0x22a28008,0x22208000,0x2a002a80,0x8aa22828,0xa0a0a22a,0x200aa002,0x20000822,0x20800a2a,0x8208a828,0x28008008,0x82a82220,0x20028088,0xa2082200,0x208282,0x8a8a,0xaa008882,0xa0a0a800,0x88aa2a88,0x800a8282,0x2088802a,0x28a88000,0xa0a08200,0x220822,0x8880a002,0xa0);
remainders_neg[868].modulo_value=1736; remainders_neg[868].bit_mask=create_bit_mask(55,0x28aaa02,0x28008800,0x2828200,0x8a22a088,0x82a28088,0xa280a20,0x2022a20a,0xa282822a,0x2a082280,0x2288880a,0xa00882a,0xa0020a,0x8a000088,0xa202aa,0x8a88a000,0x22a80aa0,0x208a00a0,0xa8088020,0x82208000,0x2820208,0xa0022220,0x20882020,0xa800082,0x82020088,0xa2028028,0xa882a80a,0x80000288,0x28aaa022,0x80088000,0x28282002,0xa22a0880,0x2a280888,0xa280a208,0x22a20a0,0x282822a2,0xa082280a,0x288880a2,0xa00882a2,0xa0020a0,0xa0000880,0xa202aa8,0xa88a0000,0x2a80aa08,0x8a00a02,0x80880202,0x2208000a,0x28202088,0x222200,0x882020a,0xa8000822,0x20200880,0x20280288,0x882a80aa,0x288a,0x28);
remainders_pos[869].modulo_value=3476; remainders_pos[869].bit_mask=create_bit_mask(109,0xa28202a2,0xa2022a88,0x2220008,0xa088a802,0x8a222800,0x2aa20a,0x8aa8a82a,0x2a0020a,0xa082a022,0x2a2aa022,0xa0888000,0x80228280,0xa20a2a8,0x22020a20,0xa82020,0xa828a2,0x828002,0xa0800822,0x8aa80000,0x8028228,0x888200a,0xa2828882,0xaa8a202,0xa822a208,0x82002a2,0x88008022,0x80222a82,0xa2200a02,0x8008a0a,0x208220,0x2a20aa2a,0x28aa8082,0x8a0a2a02,0x20888208,0xa008280,0xaa88a2,0x80082800,0xa080220,0xa0a80200,0x20a80228,0x82022020,0x28228022,0xa2008aa,0x888280a,0x2aa2a000,0x2a082a20,0x2a0220,0xa8aa8a82,0x2aa002a0,0xa2228a82,0xa8882800,0x2220200,0xa2022880,0x20a288a,0x80a8a2a,0x8aa228a,0x88002288,0x22a00808,0x88a00282,0xaa882a28,0xa2a0a800,0x80082a2a,0xa80880a,0xaa808a82,0x220000a8,0x8a0a0282,0x828aa200,0x8288028,0xa0808088,0xa0a28802,0xa000802,0x208802,0xa0000282,0xa08a22a,0x20802820,0xa220822,0xa2880a8a,0x8a88202a,0x800a8aa0,0x2008820,0x88aa0a20,0x80280a00,0x2282a88,0x82088020,0x82a8a800,0xaa0208a8,0x28a808a2,0x22082228,0x20a0082,0xaa228828,0x20a00002,0x20088200,0xa0080028,0xa008a282,0x8808082,0x8a008a08,0x8022a8a0,0x20a02828,0x8a800022,0x20a880aa,0xa80880a8,0xaa2a0800,0x800a82a2,0x8a2a08aa,0x20a00288,0x880802a2,0x8a20008,0x28a22a88,0x8a808);
remainders_neg[869].modulo_value=3476; remainders_neg[869].bit_mask=create_bit_mask(109,0x2a0a822a,0x2a0aa200,0x8aaa0880,0x28002082,0x2aa2088,0x80a22a82,0x22020a2,0x8a288a02,0x280228aa,0x22a228aa,0x28000888,0x8aa0208,0x82282220,0xaa8a82a8,0x8820a8a0,0x8820202a,0x800a088a,0x280880aa,0x2208808,0x80820aa0,0xa882,0x2a0a000a,0x8220220a,0x202a2a80,0x80a88a2a,0x8808a2,0x8aa220a,0x22a8828a,0x80880282,0x88a80a28,0xa2a022a2,0x2022080a,0x282a282,0xa8000280,0x82080a08,0x8822002a,0x880a080,0x82800aa8,0x20208a88,0xa8208aa0,0xa8aa828,0x20a208aa,0x2a88022,0x8000a082,0xa22a2088,0xa200a2a8,0x88a28aa8,0x20220202,0xa2280a28,0x22aa020a,0x2000a088,0x82aa8a08,0x2a82a008,0xa82a002,0x808202a2,0x8022a202,0x8aa00,0xaa288080,0x288a02,0x220022a0,0x22282088,0x800a2a2,0x82080002,0x2200020a,0x2a888820,0x2828a0a,0xa022288,0x802008a0,0x28080800,0x282a0082,0x8288008a,0x80a0008a,0x28888a0a,0x82802a22,0xa800a0a8,0x2aa80aa,0x2a008202,0x200a0a2,0x8020228,0x8a8800a8,0x2282a0,0x8a00288,0x82a0a200,0xa8008a8,0xa202008,0x22828020,0x2020802a,0xaa80aaa0,0x8a82800a,0x222a00a0,0xa828888a,0xa8800280,0x288008a0,0x20802a0a,0x8008080a,0x2880200,0x8a22028,0x2828a0a0,0x20888aa,0xa8200022,0x20000820,0x22a28008,0x8820a22,0x2a20022,0xa0288a00,0x808a2a,0x802a8800,0xa022a200,0x2080);
remainders_pos[870].modulo_value=3480; remainders_pos[870].bit_mask=create_bit_mask(109,0x80002882,0x20028a00,0xa08a280,0x2000882,0x208808,0x208,0x82000002,0x8028228,0x80208082,0x20082022,0x8a20208,0x200202,0x82080,0x228820,0x200800a,0xa08a0880,0x28220020,0xa008200,0x20822,0x20208028,0x80082002,0xa00808a0,0x28000,0xa080,0x22002020,0x20200a08,0x2088000,0x822020,0x8220820,0x2202000,0x8a2802,0x228a00,0x8080288,0x800800,0x200228,0x8028200a,0x82000020,0x20000200,0xa,0x82802800,0x220008,0x8288200,0x22022080,0x8020,0x280002,0x800a0002,0x208200,0xa008000,0x828220a0,0x208820,0x88202088,0x80,0x820200,0xa200080,0xa0000000,0x20008,0x8082,0x22020000,0x220808,0x82082208,0xa00a0882,0x200,0x2008208,0x8000a0,0x800028,0x88080200,0x20020880,0x8228,0x82200088,0x2802,0x8a00000,0x82008000,0x2080000,0xa00828,0x288008,0x20002002,0x228020,0xa288,0x808028a2,0x20000808,0x82088,0x80080882,0x8000220,0x8820a008,0x80800,0x20a00,0xa000280,0x20a2020,0x8800820,0x280208,0xa0882080,0x28008200,0xa080088,0x800220a2,0xa00200,0x82288,0x20820,0x20808008,0x820808a,0x2880820,0x20820208,0x82288280,0x800000,0x208000,0x80202208,0xa0822000,0x8028a20,0x200a280,0x822800);
remainders_neg[870].modulo_value=3480; remainders_neg[870].bit_mask=create_bit_mask(109,0x882002,0x28020200,0x2002208,0x82800002,0xa00000,0x88080280,0x2080082,0x820220,0x820000a,0x20802822,0x220a00,0x88288282,0x802000,0x8a20028,0x82080002,0x20020000,0x20228820,0x2080208,0x808200a2,0x20a00820,0x208a,0x20000020,0x8820008,0x88002008,0xa2882820,0x28a00200,0x8a000080,0xa20a0,0xa28020,0x8228a008,0x80022082,0x8220200,0x8200,0x80000080,0xa08a20,0x8202082,0x20808a0,0x28808208,0x8008082,0x2082000,0x8a20800,0x80200280,0x22822000,0x8800828,0x8020800a,0x820882,0x8200a00,0x2080088,0x2022820,0xa00028,0x282000,0x80080800,0x8028208,0x82200008,0x20880800,0x8820800,0x88080002,0x228a0080,0x8a28000,0x200a200,0x20020002,0x8008a00,0xa080280,0x80000820,0x8020,0x288,0xa00a0000,0x8800220,0xa208000,0x80882002,0x200808,0xa000080,0x2800080,0x8208020,0x80200000,0xa0882802,0x8220820,0x8082200,0x2022,0x20808000,0x88002000,0x2,0x800228,0x202080,0x80800000,0x8820208,0x82088200,0x28220a0,0x8028,0x80208200,0x20002800,0x20000a00,0x2008000,0x822022,0x208a08,0x88002200,0xa00a0,0x28000000,0x80200002,0x82000020,0x28020a00,0xa200200,0x80080,0x8a00808,0x28a200,0x200a2880,0x20220,0x2082208,0x22080);
remainders_pos[871].modulo_value=3484; remainders_pos[871].bit_mask=create_bit_mask(109,0x2a0a822a,0xa0aaa00,0x80aaa880,0x2a002002,0x228a2a82,0x802a288a,0x82a02020,0xaa2828aa,0x2220202a,0x88008a,0x8200088,0x8a8a8a,0x8aa2082a,0xa088a82,0x88800820,0x80a02228,0xa20a0082,0x2208008a,0x22828082,0x80028880,0xa02008a8,0x2a8a208,0xaaa2028a,0x208a2880,0xa0828a0a,0xa80280a8,0x80880a8a,0xa22a020a,0x2a00200,0xa08202a,0x82280282,0x88000820,0xa8202,0xa8282288,0xaa82022,0xa00a222,0x22888a0a,0x8088022a,0xa8220a2a,0xa8220088,0x2282a822,0x80020a00,0x20a80a2,0x28a20202,0xa22aa02a,0x88888222,0xa00000a8,0xa8080020,0x80aa2088,0x2088200,0xa8808a00,0x220aa8,0x2828220,0x280aa00,0x2200088,0xaa028,0x882828,0xa2aa0880,0xa2022,0x20082208,0xa2208aa0,0xa0800202,0x22a00000,0xa8882222,0x8a80aa88,0x80808a2,0x28a02a08,0x800a0800,0xa882a828,0xa2200882,0x2a8a0882,0x8a880220,0xa0a2228,0x888a00a,0xa88082aa,0x2288282,0x28082a00,0x20820002,0x8280288,0xa80820a,0xa00800a8,0x2a080a88,0xaa2a0220,0xa2a02802,0x8a0a2820,0xa0228a20,0xa2808aa,0xa208a2a8,0x22a20080,0x80222800,0x88202828,0xaa200208,0x28200a08,0x228880a0,0x820022,0x282a220a,0xa8208aa,0xa2a2a20,0x2200082,0x8a200220,0xaa808088,0x2aa2828a,0x208080a8,0x8a228a80,0x882a8a28,0x2800800a,0x22aaa0,0x800aaa0a,0xa882a0a);
remainders_neg[871].modulo_value=3484; remainders_neg[871].bit_mask=create_bit_mask(109,0xa2820aa2,0x82822208,0x222000,0xa288a88a,0xaa02220a,0x8a2a002,0xa28a8a0,0x2220a022,0xaaa8a8a2,0x88008002,0xa80800,0x88020202,0x22280a2,0x8280020a,0x88028,0x28aaa0,0x2a82880a,0xaa800802,0xa20a080a,0x88a0000,0x28288020,0x8a202a80,0x222a8202,0x2802a008,0x280a0282,0x20820820,0x8008202,0x2aa28a02,0x82288a88,0x8280a8a2,0xaa00a0a,0x8880a8,0x88820a82,0x2020aa00,0x8220a8aa,0x828022aa,0x2a000282,0x8008aa2,0x20a282a2,0x20aa8800,0xaa0a202a,0x8a8288,0x8a82082a,0x202a0a8a,0x2aa228a2,0xaa2,0x28088820,0x208088a8,0x822a000,0xa800a88,0x20080288,0x88a28220,0x8a0a02a8,0x8a082208,0x82a88800,0x888228a0,0x880020a0,0x2a228008,0x8882a8a2,0xa800aa80,0x2a280228,0x2808828a,0x2a288888,0x2000aaaa,0x2002200,0x8080802a,0xa028a200,0x828088,0x200a20a0,0x2aa80002,0xa202800a,0x2008aa0,0x8202aaa0,0x80002882,0x20080222,0xaa00a0a,0xa080a288,0xa802088a,0x80a08a00,0x82080a02,0x20808820,0xa2808200,0x22a20aa8,0x2a28a08a,0x282a0a0,0x202a02a8,0x82a08022,0x2a802220,0x2a2a8808,0x8aaa088,0xa0a0a0,0x22a88a80,0xa0a88200,0xa2000828,0x880a882a,0xa0a22a82,0x820a8022,0x82a2a2a0,0x8a28880a,0x2a88aa8,0x22080000,0x222a0a02,0xa8000820,0x2a20208,0xa202a0,0xa0880802,0x80aa2228,0x8822282,0x2002282);
remainders_pos[872].modulo_value=872; remainders_pos[872].bit_mask=create_bit_mask(28,0x8208aa82,0x882a0020,0xaa222a8,0xa2a0a2a,0x22aaa2,0x80a88a28,0x8282200a,0xa80aa22,0x2aa82088,0x20a20002,0xa02020,0x28880222,0xa28a2aa0,0x2828008,0xa8a28a20,0x8022280a,0x80a0088,0x8a0808,0x82aa880,0xaa02a022,0x8828288,0xa22a02a0,0xaa880028,0xa0a8a08a,0x888aa0a8,0xa8222a,0xaa208208,0x82);
remainders_neg[872].modulo_value=872; remainders_neg[872].bit_mask=create_bit_mask(28,0xa80220a,0xa288a8,0x822aaa20,0x82a282a2,0x88aa222a,0x82002a0,0xa0aa882,0x820822aa,0xa220a800,0xa82a888a,0x8828a828,0xa0008aaa,0x2a02a228,0x8a0a0880,0x202a02a8,0x8aaa082,0x80828800,0x88028080,0x80a22008,0x228a28aa,0x800a0a00,0x2aa28a28,0x220088a0,0x20202802,0x22820,0x8820aaa2,0x22a80a80,0xa);
remainders_pos[873].modulo_value=1164; remainders_pos[873].bit_mask=create_bit_mask(37,0x82000802,0x20228808,0x8a208202,0x28a20a0,0x820020,0x8200028a,0x20002822,0x820028,0x88082008,0x20808a2,0x208008,0xa00020a,0xa00020a2,0x208028,0x8228a002,0x20822882,0x8a20220,0x8002088,0x800200a0,0x8a20220,0x88208088,0xa2882822,0x20800800,0x8000a280,0xa08a0,0x20800a08,0x2080200,0x820228a2,0x8200200,0x80008280,0x82882,0x8200a28,0x8a280080,0x208a20a0,0x28808808,0x82202,0x802);
remainders_neg[873].modulo_value=1164; remainders_neg[873].bit_mask=create_bit_mask(37,0x2880082,0x28a20000,0x228028a,0x82022820,0x8028828,0xa088202,0xa08820a2,0x8028820,0xa080,0x82800022,0x8a00800,0x82088282,0x20882822,0x8a00820,0xa20208a,0xa00a2002,0x228a28,0x8008a000,0x8a0820,0x228a28,0x280000,0x22002022,0x28008008,0x8082208,0x80820020,0x28008200,0x8a008288,0x28a2022,0xa08a08,0x8080208,0x80802002,0xa08220,0x2208008,0xa0022020,0x20000000,0x8800a28a,0x82);
remainders_pos[874].modulo_value=3496; remainders_pos[874].bit_mask=create_bit_mask(110,0x2a00aa2a,0x2882a882,0x28a8a02,0x22a2008a,0x202aa082,0x8a2220,0xa880008,0xa2aaa22,0x8a08a828,0xa0aa028a,0x800a0820,0x82882822,0x80220828,0xa20aa088,0x2a88000,0x28028020,0xa0882a8,0xa20a028,0x200a0202,0x22220828,0x80020aaa,0x2a008808,0xa228a8aa,0x8a22aa80,0x2028200,0x22802a8a,0x20022a80,0x8a20222a,0x822a8020,0xa2a00022,0x8aa082,0x220a0002,0xa0000008,0xa088a88a,0x222000a2,0xaa282820,0x2082a220,0xa0a80a0a,0xa8200280,0xa82a82,0x220aaa88,0x2a888a00,0x82880882,0xa2822028,0xa0828a0,0xa00080a0,0x282808,0x80a22000,0x8aa8aa80,0x2a222020,0x2220a00,0x82808aa0,0x80222080,0x80228802,0x8002008,0x22880220,0x8880280,0xa2028202,0xa088800a,0x888a800,0xaa2aa208,0x88a0202,0x28280000,0x2000a20,0x2820a00a,0x8828a0a,0x20228228,0xa222a882,0xaaa08800,0xa82a0022,0x80082a82,0xa02a0a02,0x8a8208a0,0x2828aa08,0x88808,0x2a220a8a,0xaa2,0xa08820,0xaa20080,0xa8a82,0x2a88288,0x8808a208,0xa88008a8,0xa8028802,0x828080a2,0xaa88a200,0x2a288a02,0x2200a8aa,0xa0800220,0x208888aa,0x80a00828,0xa08a080,0x8220a028,0x280282a,0x22a8008,0xaa08a00,0x20880222,0x28228228,0x20a00288,0x80aa0a08,0x2a20a2a2,0xaaa8a028,0x22a088,0x88a20020,0xaa80808,0x8a8882,0xa2a280a2,0x2a822880,0xaa00a882,0xa8);
remainders_neg[874].modulo_value=3496; remainders_neg[874].bit_mask=create_bit_mask(110,0xa20022a2,0xa00a200a,0xa02028a,0xaa228802,0xa8a2280a,0x8802a2a8,0x82000880,0x822222aa,0x28020a0,0x28228202,0x88280a0,0xa00a0aa,0x8aa0020,0x2a822800,0x82200888,0xa08a08a0,0x82800a20,0x2a82820,0xa8828a8a,0xaaa280a0,0x8a8222,0x22880080,0x2aa02022,0x22a2208,0x8a8a0288,0xaa08a202,0xa802a208,0x2a8aaa2,0xaa208a8,0x2a2808a2,0x8802280a,0xaa82088a,0x28888080,0x20002002,0xaaa8882a,0x22a0a028,0xa80a2a28,0x28208282,0x20a08a00,0x8820a20a,0x2a822200,0xa2000288,0xa00800a,0x220aa0a0,0x8280a028,0x28080828,0x8820a080,0x82aa888,0x2202200,0xa2a2a8a8,0x8aaa0288,0xa080228,0xaa2808,0x8aa008a,0x8088a880,0xaa008228,0x80008a08,0x2a820a8a,0x28000802,0x2088,0x22a22a80,0x80028a82,0xa0a08088,0xa8882a8,0xa0282882,0x800a0282,0xa0aa0aa0,0x2aaa2002,0x22280088,0x202208aa,0x880a20a,0x20a2828a,0x2028028,0xa0a02280,0x88800000,0xa2aa0202,0x8880822a,0x882800a8,0x22a8008,0x8882020a,0x8a200a00,0x802200,0x20088020,0x200a008a,0xa080822,0x22002a88,0x22a00282,0xaa882022,0x28080aa8,0xa0000022,0x2880a0,0x82802808,0xa2828a0,0x8a08a022,0x8aa20880,0x82200288,0xa8008aaa,0x20aa0aa0,0xa8280a00,0x8228280,0xa2a8222a,0x222020a0,0x882a2800,0x2a88a8,0x82208000,0x88020002,0x2a2a082a,0xa20a2000,0x2288200a,0x20);
remainders_pos[875].modulo_value=140; remainders_pos[875].bit_mask=create_bit_mask(5,0xa08a2202,0x8000802,0x20208,0x8a20a802,0x808);
remainders_neg[875].modulo_value=140; remainders_neg[875].bit_mask=create_bit_mask(4,0x28022a0a,0x88082,0x8a8280,0x220208a);
remainders_pos[876].modulo_value=1752; remainders_pos[876].bit_mask=create_bit_mask(55,0xa28200a2,0x20220828,0x200080,0x228a20a2,0x8828800,0x220820a,0x28208a0,0x8a00208,0x2008208,0x8a2002,0x28800800,0x88280288,0x2080,0xa08,0x8000020a,0x2822020,0x228228,0x8a200002,0x20080800,0xa20208,0x8280208,0x208800a2,0x28220a08,0x8a288002,0x8008a0,0x8808020,0x828a282,0x200a28a0,0x20828a28,0x80202,0xa20a2002,0x28800228,0x820a088,0x208a0022,0x208028,0x820808a,0xa2002020,0x800008,0x80288288,0x2080882,0xa08000,0x20a000,0x22020800,0x28228028,0x2002,0x808008a2,0x20208200,0x8020800a,0x800a2082,0x20a08208,0x88002282,0x8a08a2,0x8020008,0x8a282088,0x8a0082);
remainders_neg[876].modulo_value=1752; remainders_neg[876].bit_mask=create_bit_mask(55,0x220a0822,0x28a28020,0x88288008,0xa2022822,0x20008,0x8a280282,0x820a0020,0x208a00,0x8a080280,0x80022882,0x20008008,0x208200,0x80882800,0x8808200,0x8088282,0x20a28a0,0x8a20a20,0x228808a,0xa0800080,0x8228a00,0x80208280,0xa0000822,0x20a28200,0x220008a,0x80080020,0x820,0x8020220a,0xa0822020,0x28020220,0x8800828a,0x22822882,0x20008a20,0x80282000,0xa00208a2,0x8a00820,0x80280002,0x228828a0,0x8008800,0x8200200,0x82800002,0x8200808,0x88282088,0xa28a0080,0x20a20020,0x8808a08a,0x80022,0x28a00a08,0x8280082,0x822802,0x28200a00,0x8a20a,0x80020022,0x820800,0x220a000,0x20802);
remainders_pos[877].modulo_value=877; remainders_pos[877].bit_mask=create_bit_mask(28,0xfee1169a,0x88a30b74,0xf5b2caef,0x9f15b152,0x8539810a,0x284e92fb,0x3a76ced9,0x9c8afe4e,0xbfd4f5b9,0x1546701e,0x1cd12898,0x791f4f03,0xa501263e,0xf213e069,0x20296581,0xbe279f19,0x22ce303c,0x98aa0645,0xcaff5e03,0xd44e676b,0xdb971c9f,0x5c8526dc,0x672877d2,0x6a3e5420,0xd36bd2a3,0x31447dd4,0x21dfcbb4,0x165a);
remainders_neg[877].modulo_value=3508; remainders_neg[877].bit_mask=create_bit_mask(110,0x22288a02,0x2a82a8,0x282a0222,0x288288a,0x8a00882,0xa0820a22,0xa2aa0200,0x22282,0x22082820,0x888aa882,0x8008a000,0xa082828a,0x2888aaa2,0x2a8a28a0,0xa8a0a808,0x22aa0280,0xaa02a8a0,0x228a88,0x222828a,0x882aaa2,0x20a8002,0x8008aa00,0xaaa0aa0a,0xa2a288a8,0xaa20a2a,0xa888a008,0xa8020228,0xaa0a8a82,0xa8a20200,0x80802888,0x2aa2822a,0xa8a88a2a,0xa82a82a,0x2a880,0x82000a82,0x22aaa08,0x28a0a22,0xa88a8a20,0x2028aa02,0xa80a02aa,0xa080a828,0xa028a28a,0x2a2aa888,0xa8a0a08,0x88002880,0x2208aa88,0x20a082,0x2a0a2200,0x280202aa,0x82228208,0x2088028,0xa288a08a,0x22202a0,0x20aa0aa0,0xa20288a2,0xa220882,0x22aa00a0,0x820a02a2,0xa2a08a08,0x22882a88,0xa828a282,0xa02002a2,0x2aaaa220,0xa2228a0a,0xa8888008,0x80aa882a,0x2028a0a0,0x2a088800,0x8a008208,0xa808280a,0x2a2002a0,0x8802a008,0x82aa2080,0x22a220a0,0xa288a000,0xaaa280aa,0x80aa8802,0x8002802,0x202088,0x8a802282,0x808882a,0xa00aa2a2,0xa8028000,0x880822a2,0xa8aa08,0x20020a2,0x80882,0xaa80a00a,0xa2aaa008,0x88a2aa80,0x22a20002,0x2aa08282,0xa0088082,0x22a0802,0x80a82a0,0x8028a80a,0x882a0820,0x8a020008,0xaa808282,0x888aaa08,0xa2228800,0xaaaa2828,0x2028222,0x8a0aa2a0,0x8a220a2,0x82a288aa,0x28208828,0x2aa222a0,0x22280280,0x2a088);
remainders_pos[878].modulo_value=3512; remainders_pos[878].bit_mask=create_bit_mask(110,0x82a88a02,0xa80a8,0x8aa8202,0xa820028,0xa0a20a82,0x20a22a20,0x20aaaa20,0x28008082,0xaaa80a80,0x28aa20aa,0xaa20a882,0xaa88828a,0x22808222,0x2a22a8a0,0x82a8a02,0x222aa8,0xaa22028,0xa202a802,0x8222220a,0x8828820,0x220aa808,0xaa0a8aa2,0xaa0a0aa2,0xa820a820,0x22008288,0x2088a8a8,0x2280208,0xaa0828a0,0x28a2a822,0x2808088a,0x8828aa2,0xa280a28,0x200a0a0,0x220080a0,0xa8a800a2,0x8a288288,0xa22222,0xa2a802a2,0xa82a2200,0xa800222a,0x22a08028,0x2a080200,0x2228a82,0x828880,0x22880a20,0xa2008,0x2a80a800,0x228a8aa8,0x2a20000a,0xa20220a,0xa280a20a,0xa82aa280,0xa2a28008,0x880a80aa,0x82a08802,0x2280220a,0x8aaa02a0,0x2a20028a,0x8a028aa8,0xa0a08a02,0xa8a08808,0x88a00008,0xa82aa2a2,0x2a02,0x882008a0,0x8a022,0x80022282,0xa882a288,0x88008020,0x2a28020a,0x2aa88800,0x8a0088a8,0x8a802a,0xa288888a,0x2a228228,0x888a002a,0x800a0200,0xa00a0a00,0x2028a028,0x288aa282,0x28a22020,0xaa882a8a,0x800a2820,0x8208028,0x882a2a22,0x2a228200,0xaa082a08,0xaa8aa0a0,0x888aa2a0,0x20202aa0,0x82082282,0x8aa08888,0xa0802a80,0x28088a,0x202aa888,0xa880a2a8,0x880a2a88,0xaa888202,0xaaa28222,0x28822a08,0xaaaa08aa,0x2802a02a,0x8820200,0x808aaaa,0xa08a88a,0xa082a08a,0x20280082,0x8082aa,0x822a02a0,0x80a22a);
remainders_neg[878].modulo_value=3512; remainders_neg[878].bit_mask=create_bit_mask(110,0xa20028a,0x88820820,0x80220a8a,0x820a88a0,0x282a820a,0xa82aa2a8,0xa82222a8,0xa088080a,0x22208208,0xa022a822,0x22a8200a,0x22000a02,0xaa080aaa,0xa22a2028,0x80a2028a,0x88aaa220,0x822aa8a0,0x2a8a208a,0xaaaaa82,0x800a00a8,0xaa822080,0x2282022a,0x2282822a,0x20a820a8,0xaa880a00,0xa8002020,0x8aa08a80,0x2280a028,0xa02a20aa,0xa0808002,0x800a022a,0x82a082a0,0x8a882828,0xaa880828,0x2020882a,0x2a00a00,0x882aaaaa,0x2a208a2a,0x20a2aa88,0x2088aaa2,0xaa2808a0,0xa2808a88,0x8aaa020a,0x880a0008,0xaa0082a8,0x8882a880,0xa2082088,0xaa020220,0xa2a88882,0x82a8aa82,0x2a082a82,0x20a22a08,0x2a2a0880,0x820822,0xa28008a,0xaa08aa82,0x2228a28,0xa2a88a02,0x28a0220,0x2828028a,0x20280080,0x288880,0x20a22a2a,0x8888a28a,0xa88028,0x888028aa,0x88aaa0a,0x200a2a00,0x8008a8,0xa2a08a82,0xa2200088,0x2880020,0x880208a2,0x2a000002,0xa2aa0aa0,0x288a2,0x8828a88,0x28828288,0xa8a028a0,0xa0022a0a,0xa02aa8a8,0x2200a202,0x882a0a8,0x80a808a0,0xa2a2aa,0xa2aa0a88,0x2280a280,0x22022828,0x22a28,0xa8a8a228,0xa80aa0a,0x2280000,0x2808a208,0x88a08002,0xa8a22000,0x20082a20,0x82a200,0x22000a8a,0x222a0aaa,0xa00aa280,0x22228022,0xa08a28a2,0x800a8a88,0x80802222,0x82802002,0x280a2802,0xa8a0880a,0x88080a22,0xaa28a28,0x82aa2);
remainders_pos[879].modulo_value=3516; remainders_pos[879].bit_mask=create_bit_mask(110,0x220800a2,0x28020228,0x82088280,0xa20228a2,0x828820,0x8a00a288,0xa0002802,0x20208020,0x82200008,0x20882,0x28a00000,0x8800a208,0xa00a2880,0x28820808,0x8a002000,0x828208a0,0x828828,0x20a282,0x220208a2,0x8828008,0x2080080,0x808a0882,0x8228a28,0xa08a002,0x80880820,0x8200,0x8820a288,0x2020882,0x20000820,0xa08228a,0xa0822022,0x8a00200,0x82,0x82020022,0x28228208,0x82280202,0x80020880,0x8000828,0x20020a,0x20020802,0x28a08a00,0x2208200,0x220000a0,0x20800a20,0x8a28a000,0x208a2,0x20a28808,0x8282088,0x28028a0,0x20000a00,0x820020a,0x880,0x28020208,0x80088280,0xa2080,0x20208a00,0x80282200,0x2080802,0x220000,0x8a080082,0xa0000,0xa28028,0x82208282,0x20228a0,0x28a20800,0x8000a28a,0x808a0020,0xa00008,0x80082088,0x800a20a2,0x8020800,0xa080080,0x22820002,0x20220800,0x88080288,0x22082882,0x8800808,0x8200000,0xa00800a2,0x8808820,0x8a28820a,0x820000,0x28220808,0x228a082,0x20082000,0x820220,0x800a20a,0x228a2882,0x8220a20,0x200208,0x82002822,0x8a20808,0x828a080,0x22822820,0x20a20828,0x8000800a,0xa2020822,0x20228a00,0x8208a002,0xa2,0x28220800,0x82000088,0xa0802080,0x28028000,0x228a00a,0xa0822820,0x28a28808,0x80282208,0x82880802,0x8a00208);
remainders_neg[879].modulo_value=3516; remainders_neg[879].bit_mask=create_bit_mask(110,0xa2800822,0x20828a20,0xa000208,0x228a2022,0x8020028,0x2082200,0x20882082,0x28a00828,0xa288080,0x808a0002,0x20208808,0x82280,0x20822000,0x20028000,0x208a088,0x20a0020,0x8020020,0x8828220a,0xa28a0022,0x20800,0x8a008008,0x20002,0xa20220,0x8200208a,0xa0,0x8800a08,0x282200,0x828a0002,0x28808028,0x8200a202,0x200a28a2,0x208a08,0x8808800a,0x28a08a2,0x20a20a00,0xa20828a,0x8a0000,0x808020,0x88288282,0xa08a0082,0x20200208,0x8a280288,0xa2880820,0x28008228,0x2202088,0x808a0022,0x28220000,0x8020a000,0x82082020,0x28808208,0x80288282,0x80880000,0x20828a00,0x8000208,0x80822800,0x28a00208,0x820a288,0x82800082,0x8a28808,0x200800a,0x80820880,0x8220820,0xa28020a,0x828a2020,0x20228000,0x8082202,0x208a0,0x8208800,0x800a000,0x822822,0x828008,0x82008008,0xa20a0882,0x28a28008,0x8200,0xa2802002,0x8000,0x80288088,0x20800822,0x28,0x2200282,0x800a0880,0x20a28000,0x8a20200a,0xa0802880,0x8028a28,0x80082282,0xa2022002,0xa28228,0x88288280,0x28820a2,0x228000,0x80202008,0xa20a20a0,0x28228020,0x8080082,0x228a00a2,0x28a20208,0xa00208a,0x80880822,0x20228008,0xa088000,0x20082800,0x20820808,0x8a202082,0x200a20a0,0x20220000,0x820a280,0x2000082,0x208a00);
remainders_pos[880].modulo_value=440; remainders_pos[880].bit_mask=create_bit_mask(14,0x88a220a,0x20a8080,0x2028228,0x200080,0x2882808,0x202a08,0xaa088a22,0x88a220,0x8020a808,0x202822,0x80020008,0x80288280,0x200202a0,0xa088a2);
remainders_neg[880].modulo_value=440; remainders_neg[880].bit_mask=create_bit_mask(14,0x80022282,0x8a020800,0xa8a02a0,0x88a08800,0x82002000,0x80a82280,0x220002a2,0x8002228,0x8a02080,0xa8a02a,0x88a0880,0x8200200,0x280a8228,0x20002a);
remainders_pos[881].modulo_value=881; remainders_pos[881].bit_mask=create_bit_mask(28,0x67d2f36,0x209e3f71,0xa7df15ab,0x26a06b5e,0x8111ee45,0xe6177b57,0x9ecf3976,0x84b66682,0x545e9091,0x62238103,0x174dbb35,0x741e031d,0xa5e1b79c,0xeb5cd05d,0x1e96e82c,0xe0b8e7b6,0xcba2e301,0x111ab376,0xe8ab0207,0xb4862425,0xcde50599,0xa19dba73,0x2207ab7b,0x159289de,0xef95eb58,0xe41356a3,0xf9823bf1,0x1b3d2);
remainders_neg[881].modulo_value=3524; remainders_neg[881].bit_mask=create_bit_mask(110,0x8aa0a2aa,0xa802a2a8,0x2a028822,0xaa28a282,0x8882288,0x2a8aa28a,0x202a0aa,0x82aaa0a,0x88820808,0xaaaa088a,0x8a8022a8,0xa8828a80,0x28282a00,0x22800880,0x820a20a0,0x28202a2a,0x22a2a88,0x88822aaa,0x20228a8a,0x280aa8a8,0x288800,0x280022aa,0xaa8a22a2,0x880a0002,0x22082280,0x288a8a2a,0x200aa228,0x82a02a0a,0xa22a8022,0xa28a88a0,0xa208222,0x28088,0x2a228aaa,0xaa2000a2,0x88a002,0xa8aa80a0,0x8a8a2020,0xaaa2088a,0x8aa2a202,0xa2a020a0,0x2822820a,0x8800a20,0x2a0a0a0,0xa8a08a8,0xaa200a88,0x8882aaa8,0x8082088a,0x82aaa080,0xa82a0202,0x8a2a8aa2,0x8a208882,0xa28a2a8,0x208a02a2,0xaa2a00aa,0xaa282a88,0x80282002,0xaa02008,0x2a08a20,0x20820a2,0x888a2088,0x802082,0xa2a02802,0x8a000282,0x88a28a8a,0x28880,0x80aa2008,0x8a080a8,0xa0a02a8,0x20aa88aa,0xa282282a,0xa2a0200,0xa202008a,0x88a20002,0x2a208080,0xa80080a,0xaa088aaa,0xaaa2002,0x822022,0x8a82aaa0,0x228a20a8,0x8808202,0x2a80220a,0xa0220282,0x2200aa20,0x20808828,0x8228a220,0xaaa0a88a,0x2208002,0x22aa822,0xa8882aa0,0x800a82a,0x80822a28,0x228880,0x802022a0,0x202a2828,0xa20a280,0x88aa822a,0xa028282a,0x8082880a,0x22a808a,0x88a00008,0xa8a28880,0xa00028a8,0xa02a2a0,0x82008020,0x822888a0,0x82082008,0x2882a022,0x202a802,0x20a0088);
remainders_pos[882].modulo_value=168; remainders_pos[882].bit_mask=create_bit_mask(6,0x82820002,0x808200,0x82008282,0x82020082,0x828200,0x80);
remainders_neg[882].modulo_value=168; remainders_neg[882].bit_mask=create_bit_mask(6,0x20a0802,0x8000a00,0x208020a,0x20a0802,0x8020a08,0x8);
remainders_pos[883].modulo_value=3532; remainders_pos[883].bit_mask=create_bit_mask(111,0x2aaa2a8a,0x8a0a8882,0x8822820,0x82a00888,0xa88aa22,0x2a028a82,0x2002a800,0x208a82a2,0xa2a8220,0x802a2a22,0xa082822a,0x2a20aaaa,0xa8028888,0x22a2228a,0xa022820,0xaa82028,0x80a2822,0x88a2a88a,0x82800802,0x200a22aa,0x8a8a20a,0x8208008,0xa2a2000,0x8080a2a8,0x880a8022,0x8aaaa000,0xa228a02,0xaa02800a,0x82088282,0xa800020a,0xaa02a22a,0x82a2a08,0xa88082a0,0xaa28a2aa,0x882022a2,0x882a882,0xa2aa0a00,0x2008220a,0xa082a282,0xa8a00288,0xa08a82a0,0x88808a8,0x220aa002,0x2880a2,0x8a0a2800,0x8080aa00,0x8a008288,0xa0228a8,0xa00aa888,0x20a80aa,0x8022028a,0x22a80a08,0xa08a22a2,0x22028228,0x80008a,0x200802,0x28088a20,0x8a20a288,0x2a888a8,0x880220a,0x2a080a28,0xaa00aaa0,0x880a0222,0x200a22a2,0xa0202228,0x8a0a200a,0x22800002,0xaa0888a0,0x2220800,0x2a20a2a2,0xa2a0a8,0xa820a228,0x82008828,0xaa8aa08,0xa822000a,0x80822822,0xa28aa8a8,0x2022aaa8,0x8a8200a8,0xa80aa20a,0x2aa88,0x22082a08,0x280aa828,0x288a0a00,0xaaaa280a,0x2a022000,0xa0202880,0x8a8a02a8,0x20820000,0xa2a20200,0x8a008a08,0x282aa8,0xa8a22802,0x8a020a22,0x82aa0882,0x880a0280,0x88a88082,0x2802aa28,0xa08a8222,0x2820aaaa,0x8a802a88,0x2a8a088a,0x2a208088,0xa80088a8,0x280a8002,0xa22a082a,0xa8288a,0x88220222,0x2a0a2082,0x8aaa8822,0xa2a);
remainders_neg[883].modulo_value=3532; remainders_neg[883].bit_mask=create_bit_mask(111,0xa222a202,0x282000a,0x800aa0a8,0xa288000,0x820022aa,0xa28a020a,0xa88a2088,0xa8020a2a,0x82a20aa8,0x8a2a2aa,0x280a0aa2,0xa2a82222,0x208a0000,0xaa2aaa02,0x828aa0a8,0x8220a8a0,0x8082a0aa,0x2a2002,0xa08808a,0xa882aa22,0x80202a82,0x80a80880,0x82a2a888,0x8082a20,0x8208aa,0x2222888,0x82aa028a,0x22820882,0xa800a0a,0x20888a82,0x228a2aa2,0x80a2a280,0x20080a28,0x22a02a22,0xa8aa2a,0x800a200a,0x2a228288,0xa880aa82,0x280a2a0a,0x20288a00,0x28020a28,0x80008020,0xaa82288a,0x88a0082a,0x282a088,0x8082288,0x2880a00,0x828aa020,0x28822000,0x8a820822,0x8aa8a02,0xaa208280,0x2802aa2a,0xaa8a0aa0,0x88088802,0x88a8808a,0xa08002a8,0x2a82a00,0x8a200020,0x8008aa82,0xa28082a0,0x22882228,0x828aaa,0xa882aa2a,0x28a8aaa0,0x282a882,0xaa08888a,0x22800028,0x8aaa8088,0xa2a82a2a,0x882a2820,0x20a82aa0,0xa8800a0,0x82202280,0x20aa8882,0x80aa0aa,0x2a022020,0xa8aa2220,0x20a8820,0x20822a82,0x888a2200,0xaa80a280,0xa08220a0,0xa0028288,0x2222a082,0xa28aa888,0x28a8a008,0x2028a20,0xa80a8888,0x2a2a8a88,0x2880280,0x88a0a220,0x202aa08a,0x28a82aa,0xa22800a,0x828a08,0x20080a,0xa08a22a0,0x28020aaa,0xa0a82222,0x208a200,0xa2028002,0xa2a80800,0x20880020,0xa082088a,0x2aa280a2,0x8820a002,0xaa8aaa,0xa282a80a,0x22200aa,0x2a2);
remainders_pos[884].modulo_value=1768; remainders_pos[884].bit_mask=create_bit_mask(56,0x82000aa2,0x82a22a28,0x8222a0,0xa20020aa,0x28002000,0x828a8088,0x2802222,0xa22200a,0xa0880a8a,0x28002000,0x2a82002,0x82a2208,0x8a2a22a,0x8a2a8002,0xa208002a,0x820a88a8,0xa802088a,0x2880082,0x20a00080,0x8a0a2a02,0x280a0088,0x28288880,0x282202a,0x8a00080,0x200aa080,0xa820a888,0x8228a88,0xaa28aa00,0xa2882000,0x2a082a22,0xaa00822,0xa2002,0x8828002,0x222828a8,0xa02802,0xa8a0a222,0xa0880,0x228002,0x20802a82,0x22a082a2,0x208a2a,0x2a8a2a8,0x8a8a2080,0x88a820a8,0x82a8020,0x8002880,0xa0220a00,0x888a0a2,0x880280a0,0x2a28288,0x8002822,0x8008a00,0x888200aa,0xa88a820a,0xa0008228,0x8a);
remainders_neg[884].modulo_value=1768; remainders_neg[884].bit_mask=create_bit_mask(56,0xa88822a,0xa22a220,0x800aaa28,0x2a08a822,0xa0882888,0x2020800,0x8a08aaa2,0x22aa882,0x28008202,0xa088a088,0xa20a882,0x80a2aa80,0x80222a22,0x2a2088a,0x2a808022,0x2820020,0x208a0002,0x8a00080a,0xa8208808,0x282a282,0xa0028800,0xa0a00008,0x820aa0a2,0x288808,0x28822808,0x20a02000,0x80aa0200,0x22a02200,0x2200a888,0xa280a22a,0x822800aa,0x8882a08a,0x800a0882,0xaa202020,0x8828a08a,0x202022aa,0x8828008,0x882a088a,0xa800a20a,0xa2280a2a,0x88a80222,0x2202a20,0x202a808,0x202820,0x80a208a0,0x8088a000,0x282a8208,0x8000282a,0x8a0028,0xa2a0a00,0x808820aa,0x80800288,0x28822,0x20020a02,0x20080aa0,0x2);
remainders_pos[885].modulo_value=3540; remainders_pos[885].bit_mask=create_bit_mask(111,0xa2802,0x220820,0x2008208,0x20008a2,0x20208a08,0x88280282,0x220208a0,0x820000,0x8800a082,0x22000022,0x20a08,0x88008200,0x822000,0x808808,0x2008200,0x200a0002,0x8020000,0x200a000,0x200a0,0x20a00a20,0x8282080,0xa0020022,0x8008220,0x8000082,0x800000,0x28a20208,0x282,0x20200a0,0x28,0x228a20a,0x80082,0x8000,0x8008208,0x2002822,0x20a08220,0x80282008,0x2000022,0x28020028,0x2008000,0x2802000,0x8020200,0x88080002,0x22080080,0x8008800,0x82000002,0x22082,0x28008a20,0x2080208,0x82022000,0xa08828,0x8820220a,0x80020082,0x8020228,0x82200082,0xa2800020,0x28a00a00,0x88208000,0x2082000,0x228808,0x82282008,0xa00a0880,0x8228220,0x8000088,0x2820802,0x8a20,0x8282088,0x2080000,0x8800220,0x2202002,0x2080002,0x28000808,0x8000080,0x2800020,0x8028008,0x80288000,0xa0820082,0x8008820,0x2088280,0x20820,0x20,0x88082002,0xa08a2,0x8028000,0xa008,0xa2882800,0x20020808,0x2000000,0x2082000,0xa08820,0x82088008,0xa0802082,0x8a00,0x800a008,0x20000a0,0x800008,0x808000a,0x20000820,0x20020220,0x2200088,0x820,0x8820a08,0x2288000,0x200820a0,0x8800008,0x8220a208,0x80882802,0x8020a20,0x808a200,0x80020820,0x8208,0x8028a);
remainders_neg[885].modulo_value=3540; remainders_neg[885].bit_mask=create_bit_mask(111,0x80822082,0x228020,0xa080280,0x82800022,0x20a00200,0x20020a,0xa20a0020,0x8028008,0x200a,0xa2880822,0x8820200,0x80280,0xa2080,0x8000000,0x82080208,0xa0820882,0x28800,0xa082088,0x80820820,0x20208228,0x80202008,0x200a08a2,0x800228,0x800a,0x80080800,0x20220a00,0x88088202,0x28a0020,0x8808820,0x82202202,0x80800802,0x8000000,0x80280,0x828020a2,0x20200a28,0x8202080,0x820808a2,0x20828020,0x8a000088,0x82082800,0x820a08,0x8082,0x22800000,0x800008,0x208800a,0x808a2802,0x20000220,0xa008280,0x2822880,0x200020,0x282282,0xa0802,0x828220,0xa20000a,0x22080820,0x20200208,0x280080,0x2802080,0x8a20000,0x220a000,0x20820000,0x220a20,0x88000,0x82020082,0x800228,0x80202000,0x82000880,0x8228,0x8a20a00a,0x82800802,0x20800000,0x80088000,0x20800a0,0x820800,0x200008,0x20020802,0x20,0xa000208,0x808200a0,0x808828,0x208a,0x80020022,0x820008,0x88002080,0x22002000,0x28820000,0x8a088080,0x2802080,0x8200028,0x2000000,0x20082002,0x8000200,0x82080,0x82800820,0x8800,0x80000082,0xa00800a0,0x28028228,0x8a208000,0x80880020,0x20200,0x8a200080,0x20802020,0x8800,0x2282200,0x2082,0x28220,0x2288,0x8200a0,0x800a00,0x202);
remainders_pos[886].modulo_value=3544; remainders_pos[886].bit_mask=create_bit_mask(111,0xaaa282aa,0x882a2208,0x22a20828,0xaa8a0a20,0xaa0a8a88,0x2002a20a,0x8202aa80,0x80a8820a,0x82a02a20,0x28088082,0xa888802,0x208a0a20,0x2aa02280,0x2288a22a,0x22a2aa00,0xa882a8a,0x208aa0a0,0x82a88000,0x28280800,0x80aa28a,0xa820a802,0x8a882228,0xa220aa28,0x22002020,0x2a8aa0a0,0x800280aa,0x2200aaa2,0x2222282a,0xaa22028,0x82aa8200,0x80200a,0x2aa22a0a,0xa220a2a,0x28808820,0xa280822,0xa8a2a8,0xa8282082,0x88028aa8,0x8a2aaa,0x2880aa0a,0x2022080,0x8802aa0,0x2a002022,0xa08a2a82,0x8880aa20,0x88a822a8,0x2a02828a,0x880a8a20,0x2a280a2,0x2aa2a80,0x80a000a2,0xa0800880,0xa2022a20,0x20288828,0x820008a2,0x82800002,0x88a2000,0x8a282228,0xa08a880,0x2022002,0x808a000a,0x8002a8aa,0x228a028a,0xa808a2a0,0x22a28280,0x222a882a,0xa08aa02,0xa882a8a2,0x20880800,0x800aa802,0x28020880,0xa0aa02,0x22aaa8a2,0x2a2aa280,0x820828,0xa02a8a2a,0x28882028,0xa0082202,0xa8a8a088,0xa0a88a,0x82a00802,0xa00082aa,0x8828088a,0x88a82888,0x28aaa00,0xa8aa0280,0x880a0aa2,0x8a080800,0xa228aa08,0x2a288822,0x20802a08,0x28a28aa0,0x82002028,0x800022a,0xa00a0aa2,0x88a2a822,0x8800aa8a,0xa8a88a22,0x802880a,0xa008a0a2,0x28802222,0x82820220,0x208a80a,0x82a0822a,0x802aa80,0xaaa08a80,0xaa22a2a0,0x8808a0a2,0x2228208a,0xaa2088a8,0xaa828a);
remainders_neg[886].modulo_value=3544; remainders_neg[886].bit_mask=create_bit_mask(111,0x222a0a22,0xa2aa80,0xaa2a80a0,0x220282a8,0x22820200,0xa88a2a82,0xa8a2208,0x8200a82,0xa28a2a8,0xa080080a,0x8200008a,0xa80282a8,0xa228aa08,0xa2002aa2,0xaa2a2288,0x8200a202,0xa8022828,0xa200888,0xa0a08088,0x80822a02,0x20a8208a,0x200aaa0,0x2aa822a0,0xaa88a8a8,0xa2022828,0x88a0822,0xaa88222a,0xaaaaa0a2,0x822aa8a0,0xa220a88,0x8808a882,0xa22aa282,0x82aa82a2,0xa00800a8,0x82a080aa,0x88202a20,0x20a0a80a,0x8a0220,0x8802a222,0xa0082282,0x8a8aa808,0x8008a228,0xa288a8aa,0x2802a20a,0x822a8,0x20aa20,0xa28a0a02,0x8202a8,0x8a2a082a,0x8a22a208,0x828882a,0x28088008,0x2a8aa2a8,0xa8a000a0,0xa88802a,0xa08888a,0x8002a888,0x2a0aaa0,0x82802008,0x8a8aa88a,0x8028882,0x88a2022,0xaa028a02,0x20802a28,0xaa2a0a08,0xaaa200a2,0x8280228a,0x200a202a,0xa8008088,0x882200a,0xa08a8008,0x8828228a,0xaa22202a,0xa2a22a08,0x880a80a0,0x28a202a2,0xa000a8a0,0x2880aa8a,0x20202800,0x88282002,0xa28808a,0x28880a22,0xa08002,0x20a000,0x8a022288,0x20228a08,0x82822a,0x2808088,0x2aa02280,0xa2a000aa,0xa808a280,0xa02a0228,0xa88a8a0,0x80888aa2,0x2882822a,0x2a20aa,0x882202,0x202002aa,0x808a0082,0x2880282a,0xa008aaaa,0xa0a8aa8,0x8a802082,0xa280aa2,0x808a2208,0x22280208,0x22aa2a28,0x80282a,0xaaa0a802,0x22a80020,0x220a02);
remainders_pos[887].modulo_value=3548; remainders_pos[887].bit_mask=create_bit_mask(111,0x82a0a202,0x28a002a,0x80a2a008,0xaa020a0,0x802222aa,0x288a2200,0x88aaa888,0x2028802,0x8882aa82,0xaa8a2a8a,0x2082a2a,0xa888228a,0xa2082000,0x2a80a82a,0x2820aaa,0x80a2aaaa,0x80822a22,0xa2282888,0x88088a8,0x28828200,0x80a80a02,0x2aa882a0,0x8a008aa0,0x20aa20a8,0xa08aaa80,0xa2828828,0x8aa80a88,0x2208a880,0xa0022a20,0xa082202,0x22820a2,0x8028aa00,0x2a822028,0x2000a802,0xa02a280a,0xa208a82,0x2a22aa8a,0xa0882022,0x8a2a2220,0x82a0880,0x8a0a8000,0x2a008002,0xa28800a0,0xa2202aa8,0xa28a8208,0x82000080,0xa220200,0xa8a8a00,0x220aa2,0x8220a220,0x88aa2a88,0xa8028008,0x82a2828,0x820aa2a8,0x280a00aa,0x822a0a88,0x2aa00a02,0x2a8aa08,0xa2828a82,0x22002802,0x222a8aa2,0x88a088,0x8aa0880,0xa2a2a,0x2008088a,0xa0200008,0xa2082a28,0xa2aa8088,0x80a00228,0x2800200a,0x2a0a,0x20220a82,0xa0888a8a,0x88808220,0xa2aa88a,0xa82a208a,0x8a028a80,0x8802a000,0x2280882a,0xaa280,0x8a08288,0xa808820a,0x808a8800,0x2022a208,0xa22a02aa,0xa2822828,0x802aaa20,0x22a08aa0,0x80aa200a,0x20a822aa,0x880a02a0,0x82822,0xa2a22022,0x22228288,0x288a8820,0xaaaa8a0,0x8aaa0828,0xaa82a02a,0xa0008208,0xa288222,0xaa8a8208,0x2a2a8a2a,0x82aa822,0x28022808,0x8222aaa2,0x20088a22,0xaa88880,0x20a080aa,0x200a8a0,0x2a800a28,0x808a0a8);
remainders_neg[887].modulo_value=3548; remainders_neg[887].bit_mask=create_bit_mask(111,0xa282a8a,0x8a0288a2,0x82a2880,0x8228a828,0x8aaaa22,0xa002aa88,0x222000,0x8a8a008a,0xa220a,0x2202a202,0x8a80a2a2,0x2000aa02,0x2a80a888,0xa20820a2,0x8a0a8222,0x82a2222,0x80aa2aa,0x2aa0a000,0x80080020,0xa00a0a88,0x820828a,0xa2200a28,0x2880228,0xa822a820,0x28022208,0x2a0a00a0,0x2208200,0xaa002008,0x288aa2a8,0x8280aa8a,0x8aa0a82a,0x8a02288,0xa20aa8a0,0xa888208a,0x28a2a082,0x82a8020a,0xa2aa2202,0x2800a8aa,0x2a2aaa8,0x80a28008,0x2820888,0xa288088a,0x2a008828,0x2aa8a220,0x2a020a80,0xa888808,0x82aa8a88,0x82020288,0x88aa822a,0xaa82aa8,0x22a200,0x208a0880,0x80a2a0a0,0xa822a20,0xa0828822,0xaa28200,0xa228828a,0x8a202280,0x2a0a020a,0xaa88a08a,0xaaa2022a,0x88002800,0x80228008,0x8882a2a2,0xa8808002,0x28a88880,0x2a80a2a0,0x2a220800,0x8288aa0,0xa088a882,0x8888a282,0xa8aa820a,0x28000202,0x80aa8,0x82a22002,0x20a2a802,0x28a0208,0x8a2888,0xaa0800a2,0x88822a08,0x80280a00,0x20800a82,0x8020088,0xa8aa2a80,0x2aa28a22,0x2a0aa0a0,0x8a222a8,0xaa280228,0x822a882,0xa820aa22,0x828a28,0x8880a0aa,0x2a2aa8aa,0xaaaa0a00,0xa00200a8,0x82222028,0x22280a0,0x220a28a2,0x28880a80,0x82a00aaa,0x22020a80,0xa2a202a2,0x80a220aa,0xa08aa080,0xaaa222a,0xa88002aa,0x82200008,0xa8280822,0x8a882028,0xa20882a0,0x802820);
remainders_pos[888].modulo_value=888; remainders_pos[888].bit_mask=create_bit_mask(28,0x82022002,0x28a28008,0x8200a288,0x20a20a0,0x220828,0x82202,0x208a28a0,0x28000008,0x82202082,0xa0082022,0x8020020,0x8000282,0x800a0000,0xa20220,0x208808a,0x200000a0,0x20828000,0xa080080,0x82880820,0x28820808,0x8200000,0xa28a2,0x808820,0x80282088,0x820a08a0,0x28228a00,0x8220028a,0x800880);
remainders_neg[888].modulo_value=888; remainders_neg[888].bit_mask=create_bit_mask(28,0x28a2882,0x20220800,0xa082200,0x82822820,0x8a28020,0x8800a28a,0xa0022020,0x20808800,0xa28a002,0x208028a2,0x828828,0x8008820a,0x20880,0x8228a28,0x8a000002,0xa0880820,0x28020808,0x82008008,0x20000a0,0x20028000,0x80288088,0x822022,0x8000028,0x820a000,0x2820020,0x20a20208,0xa208202,0x80000);
remainders_pos[889].modulo_value=3556; remainders_pos[889].bit_mask=create_bit_mask(112,0xa008a2a,0x8802022,0x20a8880,0x2202882a,0x200880,0x822a22a8,0xa0280020,0x28a2022,0x8080882,0x2888aaa,0x2008200a,0x2800a08,0x82a808a0,0xa8a08002,0x8a202,0xa20a220,0x82aa8080,0x2228a880,0x2202a000,0x82020808,0x28802280,0x28a800a,0xa8802a0,0x2200a28,0x8220280a,0x22828202,0xa222288,0xa8aaa2aa,0x2282aa2a,0xa208a22,0x220a020a,0x220280a0,0x8a80a280,0x8a002a00,0x8a2800a,0x2080a20,0x2208082,0xa220002a,0xaa0808a8,0x2280080a,0x80002228,0x28aa0228,0xaa0a0008,0xa002880,0x80208280,0x8a028020,0x8082aa88,0x8a020880,0xa02a2022,0xa2082000,0x2000aa22,0x2200880,0x8202a08a,0x880088a,0x2822020,0x28aaa288,0x80882802,0x22002200,0x20a8082a,0x2200880a,0x8aa00080,0x8208a8,0x808a80a0,0x22080a28,0x2aa82020,0x80280a22,0x28208020,0x22800a00,0xa0aa0,0x880aa282,0x88800022,0x2002882,0xa2020aaa,0x800088a2,0x2020880a,0x8a020808,0x28a200,0xa800a2a,0x28a02a20,0xa0280880,0x80a0880,0x8a208a0a,0x8aa82888,0xa8aaa2aa,0x28888a0a,0x8282882,0x2808828,0x8a00880a,0xa8022a02,0x2a2800,0x2880228a,0x2080820,0xa80882,0x22a28880,0x202aa820,0x88a08a00,0x8a20000,0x20a2a8,0xa202a828,0xa002800,0x820082,0xaa22280a,0x2202020a,0x808a2808,0x800280a8,0xa88a8820,0x22008002,0x82280880,0x222a080a,0x80802200,0x8a200a08,0xa);
remainders_neg[889].modulo_value=3556; remainders_neg[889].bit_mask=create_bit_mask(112,0x82880222,0x8a8a2,0x82820008,0x2a0a00a2,0x88a08008,0xaa22a20,0x28a080a8,0x8a02a82a,0x808002,0x82000222,0xa800a882,0x8a008280,0xa200028,0x2028008a,0x88802a0a,0x2a82aa0,0x2220808,0xaa202008,0xaa822888,0x28a0080,0xa008a208,0x8a020802,0x2008a20,0x82a882a0,0xa28a082,0xaa020a8a,0x82aa2a00,0x20222222,0xaa0a2222,0x2a802a2,0xa2828a82,0xaa0a0828,0x2002a08,0x2882288,0x802a0082,0x8a008228,0xaa80802,0x22a888a2,0x22008020,0xaa008082,0x8882aa0,0xa02282a0,0x22828800,0x288a000,0xa80a08,0x20a08a8,0x8022200,0x28a0008,0x28a2a0aa,0x2a80a808,0x288822a2,0x82a08008,0xa0a2802,0x80008002,0x8a0a28a8,0xa0222200,0x800a00a,0x2a88aa80,0xa02080a2,0xaa080082,0x2208808,0x880a0020,0x8020028,0xaa808220,0x2220a8a0,0xa082aa,0xa02808a8,0xaa000288,0x88820228,0x82220a,0x8882a,0xa88a002,0x228a8222,0x808002a,0xa8a00082,0x28a0080,0x88a02288,0x82088222,0x2028a2a0,0x20a08008,0x80028008,0x2a00282,0x2202000,0x20222222,0xa0000202,0xa0a002,0x820800a0,0x2080082,0x2082a28a,0x88a22088,0xa008a202,0x8a808028,0x8208002,0xa22a0008,0xa82220a8,0x200288,0x802a0888,0x88a82220,0x2a8a2020,0x288a000,0x800a880a,0x222aa082,0xaa828a82,0x8022080,0x88a0020,0x20020028,0x2a880882,0x2a08008,0xaa228082,0x800aa88,0x2a80280,0x2);
remainders_pos[890].modulo_value=3560; remainders_pos[890].bit_mask=create_bit_mask(112,0x20880282,0xa82a0082,0x20028088,0x8a0288a0,0x2020a222,0x8202a02,0x28088882,0x20000a8,0x8228020,0x2288020a,0xa8a20080,0x88280222,0x228a0220,0xa22082a,0xa2280a2a,0xaa0a02,0x222a8820,0x288082a0,0xa028802,0x80a80228,0x2820088a,0xa000000,0x2a8a20a8,0x2a2a0a8,0xa0a0020a,0x20000288,0xaa008a20,0x2828000,0x8002a882,0xa228200a,0x800028a,0x22082a2,0x80200,0x80a00820,0x2028aa2a,0x880088a,0xa0280280,0x802a080,0x222008a,0x2a82800,0x20022080,0x202000a0,0x20020a0,0x88220a0,0x2228,0x802a0a82,0x282082a0,0x228a2080,0x882008aa,0x20a08808,0x2822200a,0xa0020202,0x220a0080,0x8008882,0x82208a20,0x80820288,0xa2088222,0x22002008,0xa08882,0x80800a02,0x8882880,0x220a08a0,0x20082220,0x8a288aa,0x82082802,0xa0a8020a,0x88000082,0x8822028,0x800800a,0x8080a,0x880080a,0x282a8002,0x888000,0xa8020a2,0x80280a02,0x20022002,0xaa2808a2,0x200a02a8,0x80200008,0x82088000,0x8000208a,0x8288aa2,0x2a8002a0,0x2828082,0xa200aa00,0x80000808,0x800a0a22,0xa8a80a0,0x8a2a82a,0xa02a,0x20082800,0x802a02a2,0x2280a028,0x82022880,0x22a8880a,0xa0aa0008,0xa0288a80,0x2088a0a8,0x80a288a8,0x80282208,0x8a2a88,0x80228802,0x28820a0,0x8008,0x2220282a,0xa8082082,0x8a080880,0x2280a288,0x280080a,0xa82a22,0x80220882,0x82);
remainders_neg[890].modulo_value=3560; remainders_neg[890].bit_mask=create_bit_mask(112,0xa8000a0a,0x20228802,0x288a0000,0x2820028,0xa0a82a2a,0x80a8228a,0xa0000002,0xa888020,0x80a200a8,0xa2008a02,0x202a0808,0x208aa2,0x2a0282a8,0x82a280a2,0x22a08222,0x8822028a,0xaa2200a0,0x20080228,0x8282008a,0x200a20,0xa0a80002,0x82088880,0x2202a020,0x8a222820,0x20288a02,0xa8880a00,0x220802a0,0xa0a0088,0x882200a,0x22a0a802,0x80800a02,0x8a280a22,0x8808288,0x82080a8,0xa0a02222,0x80080002,0x28208a00,0x8a2008,0x8aa28802,0x8220a008,0xa88a2808,0xa8288820,0xa88a028,0x8002a828,0x8088aa20,0x8a2020a,0xa0280a20,0x2a02a008,0xa08022,0xa0280000,0xa0aa2882,0x280a8a82,0x22828008,0x8080000a,0x2a80228,0x80a0a00,0x2a000aa2,0x2a88a080,0x8820000a,0x8820a,0x80002008,0xaa028020,0x2880a2a8,0x220022,0x280a00a,0x28200a82,0x88802,0xaa0a0,0x80800882,0x80808002,0x80080082,0xa0220882,0x8000088,0x8200a82a,0xa0820a,0xa88a2882,0x22208022,0x28828220,0x8a08880,0x2800808,0x8882802,0x80200222,0x22088228,0x8a02080a,0x22882208,0x8880080,0x8028222,0x2020028,0x802220a2,0x80882822,0xa8802088,0x8228a22,0x2a0820a0,0xa82a008,0xa2200002,0x28220880,0x28200200,0x28002020,0x8220020,0xa0aa00,0x88022200,0x82a0082,0xa00a028,0x88800880,0xa2a8a022,0x2080280a,0x2008000,0x2a082200,0x8a000082,0x8020a22a,0x8aa000a,0x2);
remainders_pos[891].modulo_value=132; remainders_pos[891].bit_mask=create_bit_mask(5,0x20800a2,0x220828,0x82088000,0xa0020802,0x8);
remainders_neg[891].modulo_value=132; remainders_neg[891].bit_mask=create_bit_mask(4,0x82800022,0x8228020,0xa000088,0x208a0082);
remainders_pos[892].modulo_value=1784; remainders_pos[892].bit_mask=create_bit_mask(56,0x2a820a0a,0x82a022a,0x8a828aaa,0xa00aaa8,0x8a880020,0x88220288,0xaa08228,0x208a2802,0x220a88a2,0xa8aaa202,0x20002a82,0x80020a02,0x808a80a2,0x2828a00,0x200a2828,0x28a02a20,0x880a0800,0xa82a8000,0x8808aaa2,0x88a22a08,0x8028a20,0x228820aa,0x22280882,0x80022a,0x22aaa00a,0xaaa282a,0x8a880a82,0xaa0a082a,0xa2a820a0,0xa082a022,0x88a828aa,0xa00aaa,0x88a88002,0x88822028,0x20aa0822,0x2208a280,0x2220a88a,0x2a8aaa20,0x220002a8,0x280020a0,0x808a80a,0x802828a0,0x200a282,0x28a02a2,0x880a080,0x2a82a800,0x88808aaa,0x88a22a0,0xa08028a2,0x2228820a,0xa2228088,0xa0080022,0xa22aaa00,0x20aaa282,0xa8a880a8,0xa0a082);
remainders_neg[892].modulo_value=1784; remainders_neg[892].bit_mask=create_bit_mask(56,0xa20a8282,0x80a28aa2,0x20a0222,0x82882220,0x20088a8,0xaa8a00,0x2280aa0,0xa802a08a,0xaa82002a,0x20222a8a,0xa888a20a,0x88a828a,0x802082a,0x8a0a0288,0xa882a0a0,0xa028a2a8,0x828088,0x20a20888,0x80222a,0x2aa280,0x808a02a8,0xaa00a822,0xaaa0800a,0x88088aa2,0xaa222882,0x8222a0a2,0x200820a,0x228280a2,0x2a20a828,0x280a28aa,0x20a022,0x88288222,0x20088a,0xaa8a0,0xa02280aa,0xaa802a08,0xaaa82002,0xa20222a8,0xaa888a20,0xa088a828,0x80802082,0x8a0a028,0x8a882a0a,0x8a028a2a,0x80082808,0xa20a2088,0x80222,0x8002aa28,0x2808a02a,0xaaa00a82,0x2aaa0800,0x288088aa,0x2aa22288,0xa8222a0a,0x20200820,0x28280a);
remainders_pos[893].modulo_value=3572; remainders_pos[893].bit_mask=create_bit_mask(112,0xa202a282,0xa0820202,0x80a0028,0x82aaaa20,0xa2a0008a,0x822a8280,0xaaa8082a,0xa800a08a,0x8aa08aa2,0x202002aa,0x8800020,0x2a2200a,0xa882802,0x220a2a22,0x82220202,0x2000aa28,0x20228088,0xa8a0028,0x88a000a0,0x20022aa0,0x228a,0xa80a022,0xa2288828,0xa00a888,0x22882282,0x228880a8,0x2288282,0x28820a2,0x228208a,0x220a08a2,0x20a8088a,0x820a208a,0x2888a802,0x80a088a2,0x220280a,0x228a2000,0x882aa200,0x80280028,0x20a0028a,0x20880a20,0x8a2a800,0x22020222,0x8222a282,0x200a08a,0x8280222a,0x20200008,0x8aaa0020,0xaa2a882a,0xaa882800,0xaa080aa,0x280a0aa2,0xa88002a,0x8022aaaa,0x28a00280,0x2a020208,0xa0a2a02,0xa880a8a,0xa2820808,0x80202800,0x2a0aaaa8,0x28a8002,0xaa08aa0a,0x2aaaa020,0x8aa00282,0xaa2a822a,0x8080800a,0x28220000,0x80a8880,0x882a20a0,0x88828a8,0xa2088808,0x208002a8,0xa0808a02,0x802a2800,0x82228002,0x288008aa,0x8800008a,0xa02a0280,0x2288a220,0x82802a2,0xa08a208a,0x88a2202,0x8808a20a,0x280a2082,0x8808a082,0x28882822,0x2882a022,0xa082882,0x88a222a0,0x2a028222,0x880a0,0x8a2880,0xa220aa88,0x2a00a000,0x8082800a,0x822028,0x88228aa0,0x8880808,0x2a088a8a,0xa8080282,0x220a0088,0x80808000,0xaa2aa800,0x2a8aa20,0xaaaa20a0,0x882a8202,0xa8a0282a,0xa82a2000,0x2008aaa,0x20a2800a,0x8a80808,0x828a8);
remainders_neg[893].modulo_value=3572; remainders_neg[893].bit_mask=create_bit_mask(112,0x2a822a0a,0x280a0a8a,0x8288a0,0xa2222a8,0x2a288802,0xaa20208,0x222080a2,0x20082002,0x228022a,0xa8a88a22,0x800888a0,0x8a2aa882,0x8200208a,0xaa82a22a,0x2aa8a8a,0xa88822a0,0xa8aa0800,0x82028820,0x288828,0xa882a220,0x8888aa02,0x20828aa,0x2aa000a0,0x82882000,0x2a00a20a,0xaa000820,0x8a200a0a,0x8a00a82a,0x8aa0a802,0xaa828022,0xa0208002,0xa822802,0xa000208a,0x28002a,0x8aa8a082,0xaa02a888,0x222a08,0x8a088a0,0xa8208a02,0xa80082a8,0x2a2088,0xaa8a8aaa,0xaa22a0a,0x8a882002,0xa08aaa2,0xa8288880,0x22288a8,0x22a200a2,0x22002080,0x82280822,0xa082022a,0x820088a2,0xaa2222,0xa0288a08,0xa28a8280,0x8282a20a,0x82008202,0x2a028080,0x8a8a088,0x22822220,0x8a02080a,0x22802282,0xa22220a8,0x2288a0a,0x22220aa2,0x8080882,0xa0aa8880,0x80820000,0xa2a828,0x80002020,0x2a800080,0x20088a20,0x2808028a,0x8a2a088,0xaaa080a,0xa0088022,0x808802,0x20a28a08,0x2a002aa8,0x80a08a2a,0x2802a802,0x8002a28a,0x802a82,0xa002a80a,0x80280a,0xa000a0aa,0xa00a28a2,0x8280a00a,0x2a2a28,0xa2820aaa,0x80800828,0x8802a008,0x2aa82200,0xa2882808,0x80a0882,0x880228a0,0xaa0228,0x8080,0xa2800202,0x20808a0a,0xaa828000,0x8080088,0x22222088,0x8a2022a8,0x2222a828,0xa20a82,0x2028a0a2,0x20a22808,0x8a880222,0xa02a0882,0x80208080,0xa020);
remainders_pos[894].modulo_value=3576; remainders_pos[894].bit_mask=create_bit_mask(112,0x22082822,0x8a28200,0x8208820a,0x22802000,0x20828208,0x80202202,0x82022080,0xa00a08,0x28a28a,0xa00820a2,0x20028820,0x8a202282,0x80002880,0x20808800,0x800000a,0x20880,0x28228a20,0xa00208a,0xa20a0082,0x20220208,0x208208,0xa28a0,0x28a28a08,0x8028200a,0x8a0822,0x20a00028,0x88282200,0x228200a2,0x28200,0xa202008,0xa00a2802,0x208008,0x20200a,0xa0800000,0x28000028,0x820820a,0x822022,0x20828020,0x82280,0xa2000020,0x28808228,0x28a288,0x82880882,0x8800a28,0x2002200,0x20800a2,0x20020828,0x8000000a,0x80802802,0x8820220,0x220220a,0x20820002,0x8a20a00,0x80208082,0x22820000,0x200820,0x88082008,0x2000082,0x20820208,0x800a08a,0x80800082,0x20a08808,0x2088082,0x2802802,0x8a00000,0x80282080,0x808a0020,0x20008800,0x8228a002,0x822022,0x28228a28,0x8a288202,0x80000,0x8028820,0x80282,0x20880882,0x28a08208,0xa280000,0x2,0xa00808,0xa200208,0xa08028a0,0x200808,0x88008280,0x228a0082,0x8008828,0x28000a,0x28820a2,0x28a00828,0x20a28a,0xa28a0,0x8208208,0x8a208088,0xa08200a0,0x28a20800,0x8a288,0x20022080,0x8a00000,0x2002202,0xa2022800,0x8828808,0xa082280,0x8a0820,0xa28a28,0x8220a00a,0x2020880,0x8808808,0x88208282,0x82000802,0x20a08220,0x8800828a,0x882820);
remainders_neg[894].modulo_value=3576; remainders_neg[894].bit_mask=create_bit_mask(111,0xa28020a2,0x220a08,0xa000282,0xa2082880,0x28020a00,0x828a28a,0x28a2800,0x8208200,0x88202202,0x20802822,0x28820028,0x228a20a,0x882000,0x28000008,0x80088082,0x808a0000,0x20a20228,0x8208a002,0x22820802,0x28a28a00,0x88280280,0x80822020,0x20220200,0x820a082,0x800200a2,0x28208820,0x20a288,0xa20a0822,0x8820a08,0x8228a080,0x20822082,0x8a00800,0x8820a082,0x20080880,0x20808820,0x80280282,0x800a28a2,0x28020828,0x8800a208,0x228808a0,0x20000a20,0x88202200,0x2000002,0x8220,0x8a08a288,0x82800822,0x28828020,0x8088082,0x82082,0x28a28,0x8a28a282,0xa00a0802,0x228208,0x828000a,0xa20a0880,0x8a08028,0xa080,0x82880802,0x28028a00,0x80082002,0x80802,0x28200000,0x8a00000a,0x82082082,0x208808,0x820a008,0x208a0,0x28800008,0xa20208a,0x800a28a2,0x20a20220,0x220028a,0x80800880,0x820028,0x8800820a,0xa0000002,0x20200a00,0x82208088,0x80880882,0x8208000,0x82288280,0x20082020,0x8a08000,0x80208,0xa2020802,0x800020,0x88208082,0x82002822,0x20200020,0x88282202,0x80822020,0xa00a00,0x2280000,0x200a0820,0x20228008,0x88002200,0xa08a2800,0x208808,0x8a08a28a,0x228a2080,0x20000,0x8200a208,0x800200a0,0x8220220,0xa282082,0x828a0000,0x280202,0x2880082,0x28200a28,0x80202,0x20a0);
remainders_pos[895].modulo_value=3580; remainders_pos[895].bit_mask=create_bit_mask(112,0x28000a0a,0x20228822,0xa8a8288,0x282882a,0xa028222a,0x800202,0x2a280220,0xa882208,0x80228822,0x82808a2a,0x808280a,0x88008a22,0x2a8a82a8,0x8a208080,0x22800200,0x28800888,0x28228020,0x8000020,0x8a80a082,0x82800000,0x80aa0888,0xa080a22,0x20202a0,0x8a2288aa,0xa808,0x8088220a,0xa8200280,0xa808228,0x822802,0x2208a820,0x8088220a,0x22200aa2,0x20080,0x8802888a,0x80a202,0x2022000a,0x2a2a8a02,0xa082200,0x88a2a8a2,0xa0a80800,0xa0222a00,0x2a0a8a00,0x2a0a2080,0x20000,0x80802a22,0xa02282,0x20a02,0x88200a8,0x8280a002,0xa088a000,0x88a220a,0x800880a0,0x820208,0x82020002,0x80a00828,0x822a80,0x202a8820,0x228200a0,0x8000808,0x22080820,0xa02200,0xaa088a22,0x2000a220,0x800a028,0x2a00822,0x80a0800,0x282880a0,0x88a8020,0x80000800,0x80208a0a,0xa00a2a0a,0xa00a8880,0x200202a0,0x8a2a8a2,0x8008820a,0x880a2a8a,0xa000880,0x2808a020,0xa222802,0x80200800,0x28aa0088,0x8a088220,0x82a208,0x8028820,0xa288202a,0x20280082,0xa088220,0x2202a000,0xaa2288a,0xa80808,0x288a020a,0x22220aa0,0x20000028,0x820a02a,0x80800002,0x80802882,0x82220022,0x20080028,0x8020208a,0x22a82a2a,0x88a2002,0x2a028202,0x2a8a2028,0x8822880,0x8208822a,0x88028a,0xa8080020,0xa888280,0xa822828,0x82282a2a,0x88822880,0xa0a0002);
remainders_neg[895].modulo_value=3580; remainders_neg[895].bit_mask=create_bit_mask(112,0xa0880282,0xa82a00a2,0x2020200,0x8a0200a2,0x20a0aa22,0x88000a8a,0xa2208aa0,0x200a280,0x8a200aa,0x2080222,0x80802082,0x802a2,0x22020220,0x2a00808,0xa2088a08,0xa0080000,0xa02a08a0,0x8880a8,0x200280a,0x2088808,0x8220000,0x820082a2,0xa8a8228,0x2a20022,0x80882000,0x8002a82,0x20288a00,0x20802a0,0x8802a08a,0xa2802028,0x8002a82,0xaa288222,0x88a8008,0x820002,0x80082a0a,0xa8aa0882,0xa2220282,0x280a288,0x22202a,0x20208008,0x28aa2288,0xa2020280,0x2282a008,0x88828888,0x8a22a,0x88282a0a,0x880a8282,0xa8020,0xa00288a,0x20002808,0x80022282,0x8000820,0x80a8280,0xa82888a,0x288020,0x880a2208,0xa82200a0,0x2a0a8028,0x80808080,0xa2808028,0x88282a88,0x220002a2,0x288822a8,0x808028a0,0x8228802a,0x80820088,0xa0200820,0x200a8,0x8808088,0xa80202,0x28822282,0x28020000,0x280a8228,0x8022202a,0x800a02,0x822202,0x82088000,0x208020a8,0x82a2a08a,0xa88008,0xa0220800,0x2000aa0,0x80a2280,0x808200a8,0x2200a822,0xa8a0080a,0x82000aa0,0x2a8a2088,0x8222a002,0x80208000,0xa0020a82,0xaa2a8220,0x288880a0,0x80a028a2,0x8880a,0x8082002,0xa2a88a2,0x288080a0,0x8a0a802,0xa220a222,0x8002288a,0xa20a0a82,0x2202a0a0,0x8002a008,0x2800a22,0x88000a02,0x200088a0,0x2000208,0x8202a0a0,0x2a0a222,0xa2008,0x2028882);
remainders_pos[896].modulo_value=56; remainders_pos[896].bit_mask=create_bit_mask(2,0x82002a22,0x88a800);
remainders_neg[896].modulo_value=56; remainders_neg[896].bit_mask=create_bit_mask(2,0xa88a22a,0x2080);
remainders_pos[897].modulo_value=3588; remainders_pos[897].bit_mask=create_bit_mask(113,0x20a0082,0x8a28220,0x88,0x82822802,0x8020820,0x80008080,0x880020,0x8028020,0x80200088,0xa0020002,0xa00208,0x8082202,0xa0000822,0x20800228,0xa282202,0x22820000,0x208,0x2080008,0x82020800,0xa20820,0x8028800a,0xa20028a2,0x20000020,0xa082200,0xa08a0802,0x20a08008,0x88008202,0xa20a0022,0x20028228,0x208008a,0x80082822,0x828828,0x2208282,0x20000020,0x28a00228,0x8008a00a,0x20822802,0x820208,0x80008200,0x82000000,0x20a20,0x220a280,0xa2000822,0x20800028,0x2208082,0x82002802,0x20028,0x8800200a,0x200a0080,0x20008800,0x8080008,0x20820080,0xa20a08,0x88000002,0x220a2880,0x8028200,0x828020a,0x228a0880,0x220,0xa08a008,0x20082082,0x20200,0x2200082,0x200a0080,0x800220,0x8008000a,0x2800822,0x20208808,0x80002088,0x820008a2,0x28a08808,0x8a080000,0x820,0x8200020,0x8082000,0x2882082,0xa20028,0x8800a28a,0x80000082,0x28208800,0x82282008,0x82820022,0x20020808,0x8828008a,0x800a08a2,0x8200228,0x20a088,0x20a20a2,0x8820a08,0x80000080,0xa28008a0,0x228028,0x8208a00a,0x2080820,0x20800,0x8000002,0x82882,0x8828a00,0x88002088,0x820000a2,0x8820208,0x800a008,0x800a2,0x20008028,0x80280202,0x80022000,0x20200020,0x82080200,0x2882820,0x20000008,0x8828a202,0x200a0800,0x8);
remainders_neg[897].modulo_value=3588; remainders_neg[897].bit_mask=create_bit_mask(112,0x82020802,0x220a28,0x80088000,0x2022082,0x820028,0x8080008,0x800008a0,0x20828,0x8288000,0x208a0082,0x8208a00,0x8000a28a,0x20800022,0x28008a20,0x8220a20a,0xa20a0880,0x8808a00,0xa000080,0x28a0080,0x8228020,0x200082,0x22882022,0x28808028,0x200a288,0x20020082,0x28200800,0x28a,0x22820822,0x20820a20,0x8a000002,0x8020a2,0x8020020,0x8a280202,0xa00808a0,0x20208a20,0x8002002,0x200a2082,0x8028a00,0x8000280,0x2880880,0x8828228,0x82282208,0x228800a2,0x20000820,0x8a28000a,0x2882082,0x8028820,0x8a082,0xa0820000,0x28800008,0x80008080,0xa0020800,0x8220200,0x8800a,0xa2822000,0x820208,0x80200282,0xa2020000,0x8808a20,0x82002080,0xa0802802,0x8828200,0xa28800a,0x20820800,0x8008a28,0x8008082,0x82080022,0x20a00000,0x808a000,0x2080022,0x20200000,0x2008080,0x800000a0,0xa08828,0x8000a088,0x2002802,0x8228820,0x2202,0x880002,0x20a00008,0x220a080,0x20a0822,0x28820000,0x208002,0x820022,0x208a20,0x8282000,0x82822022,0x28200,0x88008,0x22000020,0x8a20820,0xa002002,0x828000a0,0x8828008,0x8000008a,0x80802002,0x20200,0x82000,0x2880822,0x28200,0x82080,0x80800822,0x28800820,0x820820a,0x8a2800,0x20a08828,0xa008280,0x820020a0,0x28808800,0x202282,0xa0020080);
remainders_pos[898].modulo_value=3592; remainders_pos[898].bit_mask=create_bit_mask(113,0x2aa8228a,0x82820aa0,0xaaa0820,0x8a20a802,0x28280a2a,0xa808aaaa,0x88882,0x82880802,0x88228aa8,0x200a288a,0x80a2a0a2,0x2222aa88,0xa00aa2a,0x8aa2800a,0x882a200,0x2a208a0a,0x222aaaa,0xa0a2a080,0x2a08aa,0x2202a802,0x88080280,0x8000088a,0x22200822,0x2a08282a,0x8820aa02,0xa80a220a,0xa20a8280,0x2a82228,0x222802a2,0x8280a208,0xa220a80a,0xa00a2880,0x2828a022,0x28aa2220,0x88aaaa82,0x82a8a880,0xa802a22a,0x8a02aa2,0x8a020a00,0x222aa,0xa08a2020,0xa20288a0,0xaa820082,0x200aa0a0,0x80022220,0xa020a88,0x8820aa22,0x822880,0xa8a88288,0x8888aaa2,0x8a802,0x20a82820,0xa80a2080,0x28a000a2,0xa2828a,0x8228002a,0x288200,0x828a00a8,0xa28a2,0x8a02a8a,0x282a0802,0x2a200008,0xaa222280,0x822a2a8a,0x28820022,0xaa082202,0xa080a088,0x88800222,0xaa00808,0x82aa0a,0x22808a82,0x8a20a0a,0x88800008,0xa080a2aa,0xa80a2000,0x8a802a8a,0x2a2a82a8,0xaaaa2202,0x88aa2882,0xa282808,0x28a00a88,0x2a088a02,0x8a0282a0,0x80288820,0x882a808a,0x82a08a28,0x88a02a02,0xaa0822a0,0x2820a880,0x200888a8,0x20000288,0x802022a2,0x2a808802,0x20a80080,0xa8a0aaa,0xaa888002,0xa208a8aa,0x8a8220a0,0x28aa200,0xaa00a0a0,0xaa8888a8,0xa8a0222,0x28a0088a,0xa28822a2,0x2022822a,0x22200080,0xaa202a82,0xa02828aa,0x2a08a2a8,0x20aaa080,0xa0828208,0x882aa80a,0xa2);
remainders_neg[898].modulo_value=3592; remainders_neg[898].bit_mask=create_bit_mask(113,0xa220aa02,0xa0a8228,0x822280a8,0x2a8208a,0xa0a082a2,0x20802222,0x8880000a,0xa00808a,0xaa0220,0xa882a002,0x82a282a,0xaaaa2200,0x828822a2,0x22a0882,0x800a2a88,0xa2a80282,0x8aaa2222,0x282a2808,0x88a28022,0xaa8a208a,0x808a08,0x8888002,0xaaa880aa,0xa280a0a2,0xa8228a,0x2082aa82,0x2a820a08,0x8a20aaa0,0xaaa08a2a,0xa082a80,0x2aa82082,0x2882a008,0xa0a028aa,0xa022aaa8,0x22220a,0xa202008,0x208a2aa2,0x8028a22a,0x28a8288,0x888aaa22,0x2802a8a8,0x2a8a0028,0x220a8802,0xa8822828,0x88aaaa8,0x828a8200,0xa822aa,0x880aa008,0x20200a00,0x222a,0x8880208a,0xa820a0a8,0x2082a808,0xa028882a,0x882a0a02,0xaa088a2,0x88a00a88,0xa028820,0x8882a02a,0x8028a202,0xa0a2808a,0xa2a88880,0x22aaaa08,0xaa2a202,0xa00a88aa,0x2280aa8a,0x28082800,0x88aaa,0x82288080,0x880a2282,0xaa08020a,0x802a8282,0x88880,0x28082a22,0x2082a888,0x208a202,0xa2a20a20,0x2222aa8a,0x22a00a,0x82a0a080,0xa0288200,0xa280028a,0x28a0a28,0x8a000a8,0xa20802,0xa2802a0,0x28a28a,0x2280aa28,0xa0a82008,0xa8800020,0xa8888a00,0x8a8aa2a,0xa208008a,0xa8208808,0x82028222,0x2200088a,0x2a802022,0x20aa828,0x8a022a08,0x22882828,0x22000020,0x82028aaa,0xa0288002,0x2a00aa2a,0xa8aa0aa2,0xaaa88808,0x22a8a20a,0x28a0a022,0xa2802a20,0xa8222808,0x280a0a80,0xa22082,0x2a);
remainders_pos[899].modulo_value=3596; remainders_pos[899].bit_mask=create_bit_mask(113,0x2aa0222,0x2002a0a2,0x8a2a2200,0x2a082200,0x82a08a0a,0xa20a2aa8,0x80288288,0x20aa22,0x82002002,0xaaa20a,0xa80a200a,0xa82822a0,0xa220a0a2,0x2a220280,0x8aa8a08a,0x8a0088,0x80222a00,0x8a0a8222,0xa2828888,0xaa0080,0xaa82aa,0xa0a222a0,0x28a8808,0xa0aaaa22,0xa8a82220,0xa208a202,0x828a208a,0xa2aaa0a2,0x28000802,0x8820a08,0x88228800,0x820200a,0x80002828,0x82000a88,0x8882822,0x20022a8,0x2a802a80,0x200a080a,0xa028222,0x8008aa08,0x28822aaa,0x28000222,0xa808082,0x2820882a,0x8a08008,0xa880a028,0x88002a02,0xa8aa0a02,0x8028aa88,0x2208a08,0x828822,0x82080220,0x88a28082,0x8880002a,0x280aa8aa,0x8a800808,0x2a2208,0xaa028202,0x222aa2,0x28a22a80,0x2082820,0x28200088,0x20880882,0xa280220a,0xaa2a22a,0x8002280a,0xa022a80a,0x20a20280,0x20828200,0x202a0a82,0x28820,0x88228888,0xa2002aaa,0x280a020a,0xa008888,0x802a8a02,0x8008002a,0x822202a8,0x82882,0x8000222a,0x80820282,0x28822a00,0x8220002,0x2820a,0xaaa8a802,0x8a2828a0,0xa208aa20,0x82a2a808,0xaaa0a088,0x2a28088a,0x88a0a202,0x2aa000a8,0xaa00aa8,0x2828a020,0x2a0a2222,0x88802888,0xa20000a,0xa2aa2220,0x88a8a20,0xa088a028,0x8282a8a0,0x8a02a0a8,0xaaa00a00,0x80082a08,0xa0800800,0x2280288a,0x8a08a228,0x20a822aa,0x820a8a0a,0x8a8a2008,0xa8008008,0xaa808a0,0x888);
remainders_neg[899].modulo_value=3596; remainders_neg[899].bit_mask=create_bit_mask(112,0xa228aaa,0xa88a282a,0x222aa88,0xa280aa88,0x2280282,0x2a82a220,0x8a00200,0x88a822aa,0xa08a88a,0x8222a82,0x2082a882,0x20a0aa28,0x2aa0282a,0xa2a28a08,0x2202802,0x88028800,0x8aa2288,0x2820a2a,0x2a0a0000,0x88228808,0x80220222,0x282aaa28,0x8a020080,0x282222aa,0x20202a28,0x2a802a8a,0xa02a802,0x2a22282a,0xa0888082,0x800a8280,0xaa0088,0x28a882,0x888a0a0,0xa888200,0x8000a0aa,0x8288a220,0xa208a208,0xa8828082,0x28a0aaa,0x8002280,0xa00aa222,0xa0888aaa,0x8200080a,0xa0a000a2,0x80280880,0x200828a0,0x88a20a,0x2022028a,0x8a02200,0x82a80280,0x880a00aa,0xa8082a8,0x2a080a,0x808a2,0xa0822022,0x2088000,0x88a2aa80,0x228a0a82,0x88aaa22a,0xa02aa200,0x8a00a0a8,0xa0a88800,0x2800800a,0x2a08aa82,0x822a22a2,0x88aa082,0x20aa2082,0x282a8a08,0xa80a0a88,0xa8a2820a,0x880a00a8,0xa20000,0x2a88a222,0xa0828a82,0x82800000,0x8a2020a,0x88088a2,0xaaa8a20,0x8080200a,0x888aaa2,0x80a8a0a,0xa00aa288,0x80aa008a,0x888a0a82,0x2220208a,0x2a0a028,0x2a802220,0xa2a2080,0x22282800,0xa2208002,0x282a82,0xa2288820,0x82288220,0x20a020a8,0xa282aaaa,0x8a000,0x2a88882,0x2222aaa8,0x800202a8,0x280028a0,0xa022028,0x20a2820,0x22288288,0x880a280,0x28088008,0xaa00a002,0x2802aa0,0xa020aa22,0xa820282,0x2022880,0x20880880,0x82200028);
remainders_pos[900].modulo_value=120; remainders_pos[900].bit_mask=create_bit_mask(4,0xa08a2882,0x28228a20,0xa08a288,0x8228a2);
remainders_neg[900].modulo_value=120; remainders_neg[900].bit_mask=create_bit_mask(4,0x20022002,0x20220220,0x2002200,0x22022);
remainders_pos[901].modulo_value=901; remainders_pos[901].bit_mask=create_bit_mask(29,0x8ad1f23a,0x3d13aed0,0x3303aee7,0x7a2c7010,0x192928e2,0x50f273d0,0xc8e7ec5f,0x40393ba7,0x5bb48959,0x961a59b4,0xc2420cf5,0x4c972a59,0x2afec0a,0xa31700ca,0xc03a3140,0xdfd5014,0x953a4c94,0xcc1090e6,0x66961a6b,0xa44b768b,0x772700a6,0x8df9c4f9,0xf393c2be,0xc5252602,0x38d1791,0xdd703302,0xdd722f39,0x13e2d442,0x17);
remainders_neg[901].modulo_value=3604; remainders_neg[901].bit_mask=create_bit_mask(113,0x2082aa2,0xa0822208,0xaa8a222a,0xa220a888,0xa0a02a,0x802aaa08,0x2a2082,0x8a0a22a,0x82280080,0xa828028,0xa8a8020,0x8002a280,0x8a222002,0x2a8a8802,0x8a2a088,0x80208888,0x8a20008,0x88082a,0xaa028222,0x2882aa02,0xaa2a882a,0x200820,0x220a0a22,0x8a8aa8a,0x2008a08,0xa8aa8a,0xaaa2a0,0x8a2a0882,0x82a28a8a,0x2aa80208,0xa8a8002a,0x8802028a,0xa8a88082,0x8282228a,0x80200222,0x8aa2a820,0xaa08a2a0,0xa02aa02,0x80880222,0x22882a0,0x88200880,0x2a288088,0x8a8aa088,0x22228a00,0x2a000a00,0xa8a800a,0xa0a8020,0xa208a0,0x28288008,0x22a002a2,0xaaa00a08,0x28280082,0xa8222aa0,0x228aa888,0x22082aa2,0xa0820082,0x8200222a,0x2228aa2,0x2220028,0x28088800,0x282a0020,0x288a0,0x228a08a,0x2822022a,0xaa0a888,0xa0aa00a2,0x80aa0a80,0x80202a80,0x2228a0aa,0x8082a082,0x20088a00,0x28088a08,0x20aa88aa,0x828a0288,0x80a022a8,0x2002a002,0x8a0208,0x228a2800,0x828222a2,0x8008202,0xa8828a08,0x8008082,0x2022a8,0x28882a8,0x20800082,0xaa088a0,0xaa20200,0xa0808008,0x828a08a,0xa0208008,0x22a220a0,0xa28a2,0x8202880,0xa002a002,0x8aa20280,0x8a028a0,0x2a88aa82,0x828880a,0x80288802,0x2082a080,0xaa828a22,0x80aa0200,0x80a82a80,0x22802a82,0x88a82a8,0x2a20220a,0x28828a20,0x288a000,0x2002a0a,0x8880a,0xa002220,0xa2a8a220,0x20020);
remainders_pos[902].modulo_value=3608; remainders_pos[902].bit_mask=create_bit_mask(113,0x22a8202,0x28228088,0x288200a2,0x800a8000,0x2a020082,0xa2202a0a,0xa880aa2,0x8a00222a,0x22208808,0x28a0a80a,0xa800808a,0x8aa022a2,0x880aaa00,0xaa0282aa,0x2828a222,0xa2200022,0x228a2008,0x808822a,0x820a0220,0x8002aa20,0x2a88a000,0x208020a2,0xa28aa22a,0x828a2a0,0x88208028,0x8a80280,0x82a02aa,0xa022a880,0x2280828,0x80282a08,0xa800a002,0xa288882,0x808a828,0x820202,0x2a022020,0x880020a,0xaa8aa2,0xa0a20a00,0x80a8aa02,0x20002022,0x8a22082a,0x820222a,0x2a00022,0x888002,0x220082a0,0x8a280200,0xa08008a,0x882200a0,0x2aa008a8,0x2080a002,0x2a200200,0x82a02020,0x80a0828a,0x228828aa,0x8208220a,0x8022a082,0x20a82a0,0x82820a88,0x88a08820,0x2aa2822,0x82a2820a,0xa808080a,0x8008008,0xa8800a02,0x222a200a,0xa00a0088,0xa2a20020,0x88008028,0xa8200,0x80800222,0x2088000a,0xa2a88808,0x8a82088,0x2880800,0xa80aa2a,0xa08a,0x208aa2aa,0xa8a08002,0x80880,0x20808082,0xa0282a20,0x2a822228,0x2800a00,0x8020a828,0xa282028,0x20022a88,0x20aa80a8,0x2202802a,0x20280208,0x8a0a8a28,0x8a88aa2,0xa88a0802,0x2000a22,0x8208aa80,0x200880a0,0x88a88220,0x8a2008a2,0x28880008,0xaa888a28,0x22aa8280,0xa200aaa0,0x2a8a880a,0x28a20200,0x88a02a0a,0xa2202208,0xa0a88800,0x8a8aa022,0xa8a0a808,0x2820080,0x280002a0,0x288a0082,0x80220288,0x8082a8);
remainders_neg[902].modulo_value=3608; remainders_neg[902].bit_mask=create_bit_mask(113,0x8aa2028a,0xa02a0800,0xa00a882a,0x820880,0xa28a080a,0x22a8a282,0x8200822a,0x288aa22,0x2aa00080,0x20282082,0x20880802,0x228a22a,0x22288,0x228a0a22,0xa0a02aa2,0x2aa808aa,0xa202a880,0x80800aa2,0xa828a28,0x8822228,0x22002888,0xa808a82a,0x2a0222a2,0x80202a28,0xa800a0,0x80208a00,0x80a20a22,0x20aa2008,0x8aa080a0,0x8a02200,0x2080288a,0x2a0000a,0x808020a0,0x880a828a,0xa202a8a8,0x80088a82,0x88220222,0x282a0288,0x20228a,0xa808a8aa,0x2aa8022,0x80a0aaa2,0xa2888aa,0x8800088a,0xa2880228,0x2208a88,0x82808802,0xaa8820,0xa2280020,0x2008288a,0xa2a88a88,0xa28a828,0x8200a02,0x2a00a022,0xa80aa82,0x8aa2002,0x8a020a28,0xa0a8200,0x2800a0,0x8a2220aa,0x22a0a02,0x20808082,0x80880800,0x2000828a,0x2aa2a882,0x28828000,0x2a2a80a8,0x808a0,0x88820a88,0x8088aa2,0xa8000882,0x22200080,0x8020a800,0x8a008008,0x820022a2,0x8802802,0xa8022a22,0x2028008a,0x88008008,0xa808080a,0x2820a2a0,0xa20a2aa0,0x82088288,0x8a820a0,0x82a0a820,0xa082a200,0x28220820,0xaa8a08a2,0xa8a08280,0x20202a0,0x20022a,0x20028082,0x8a8802aa,0x2802208,0xa8800828,0x200a28,0x2a08022,0x20008880,0x220002a0,0xaa220208,0x2a082228,0xa2020002,0xa02a8a80,0x282282,0x22a8aa80,0x28200088,0x202202a,0x20202080,0xa0a8808,0xa0888a28,0xa002800a,0x82a0a00,0x80a20);
remainders_pos[903].modulo_value=3612; remainders_pos[903].bit_mask=create_bit_mask(113,0xa2820802,0x20000200,0x80000202,0x20228a0,0x20a00a08,0xa002208,0x80002,0x20828a08,0x808a200,0x820800a2,0x20228800,0x8000828a,0x820008a0,0x28828208,0x8a00a,0x20882022,0x20028000,0x88000000,0x2800802,0x8820828,0x8808208a,0x82020800,0x20200a20,0x80008008,0x20080080,0x28000008,0x2000288,0xa0822080,0x8000208,0xa08a200,0x82800002,0x20208028,0xa20220a,0x808020a0,0x8820000,0x82002202,0x2000020,0x20028020,0xa200082,0x2802882,0x200a28,0x88088,0x2000020,0x28800200,0x2280002,0x80002,0x20820a00,0x2200008,0x82080,0x20028200,0xa008282,0xa0080080,0x8028000,0x8a202000,0x82880880,0x8000200,0x8080202,0x20080002,0x20220228,0x808a,0xa0002802,0x200200,0x8828200a,0x82800,0x208200,0x82000088,0xa0820,0x8000200,0x88000288,0x80022,0x800008,0x2202200,0x28a0080,0x8228028,0x8820008a,0x802800,0x20800008,0x8088008,0x20000822,0xa08020,0x8a08808a,0x22802080,0x8000028,0x8a20a,0xa2080002,0x208820,0x82280008,0x82000002,0x20200200,0x82002000,0x208a0080,0x20020808,0xa208202,0x2820822,0x28020028,0x80000002,0x80002800,0x8808220,0x8a00a200,0x22082822,0x20a20008,0x8a282000,0x20022880,0x8a00208,0x8008a202,0x20a2820,0x8000200,0x8208800a,0x20a00a0,0x20a28808,0x88080000,0xa0080000,0x8020828);
remainders_neg[903].modulo_value=3612; remainders_neg[903].bit_mask=create_bit_mask(113,0x220a0002,0x28808200,0x8828a,0x820a2020,0x28208200,0x82082280,0x80000082,0x28020200,0x2280,0x2800822,0x28220008,0x8000202,0x2880020,0x20020200,0x88002002,0x200028a2,0x20820808,0x80088,0x82000082,0x20020,0xa002,0x28a0000,0x20a08220,0x80080,0xa0000800,0x20808800,0x8a080200,0x200a2000,0x808a00,0x2002280,0x2080882,0x28200820,0x8220a282,0x2820,0x28008,0xa08a20a,0x28808a0,0x20820828,0x228800a,0x82002002,0x8a00220,0x88000000,0x82880820,0x20008a00,0x8220808a,0x80000882,0x28028208,0x8a280080,0x80802000,0x28020a08,0x2080202,0x20800800,0x20808,0x220a088,0x2000000,0x808200,0x8000820a,0x20800882,0x20a28a20,0x88080002,0x20802082,0x8a00a08,0x20a082,0x80802000,0x8a00a00,0x2088000,0x800200a0,0x808a08,0x80200,0x808000a2,0x8008800,0xa28a200,0x82020800,0x220820,0x208002,0x80082080,0x28008000,0x80000000,0x208800a2,0x200828,0x2000002,0x22002800,0x800820,0x88002282,0x22800802,0x8a00020,0x2208080,0x2080082,0x28a08a08,0xa082088,0xa0020000,0x28828000,0x2280282,0x820a00a2,0x20028820,0x800808a,0x882080,0x228,0x2082200,0x228020a2,0x20228800,0x220a088,0xa0822000,0x200a00,0x800228a,0x82822020,0x808a00,0x2000082,0x82020820,0x28220000,0x88,0x20800080,0x828020);
remainders_pos[904].modulo_value=904; remainders_pos[904].bit_mask=create_bit_mask(29,0xaa2882aa,0x8a022a28,0xa0820808,0xaa008a22,0x80a80a2,0x2a2a8a00,0x22802a20,0x202a8222,0xa080202a,0xa80a8aa,0x2082a002,0xa8a28a02,0x2022a088,0x2882008,0x8228000,0xa880820,0xa28a2a22,0xa820880,0x2a02a080,0x8020aaa,0x82a808a8,0xa8028888,0xa2a8a808,0x2a02000,0xa200aa8a,0x20820a88,0xa880a220,0x8228aa28,0xaa);
remainders_neg[904].modulo_value=904; remainders_neg[904].bit_mask=create_bit_mask(29,0x22a00a22,0x28aa2a0,0x280a8080,0x228802aa,0x8082082a,0xa2a20288,0xaa08a2a8,0xa8a20aaa,0x2808a8a2,0x82082022,0xa802288a,0x202a028a,0xa8aa2800,0x8a00a880,0x80aa0888,0x820080a8,0x2a02a2aa,0x820a8008,0xa28a2808,0x808a8222,0xa208020,0x208a0000,0x2a202080,0x8a28a888,0x2a082202,0xa80a8200,0x20082aa8,0xaa022a0,0x22);
remainders_pos[905].modulo_value=905; remainders_pos[905].bit_mask=create_bit_mask(29,0x30c74b96,0x8f22d095,0x710ae111,0xcad78cae,0x76010903,0x1d81224e,0x4870ce54,0x78c65135,0x2630827,0x95140023,0xa48412d4,0x81594281,0x785eb3b0,0x924235a0,0xb1092733,0x35e87816,0xa6a0437,0x20849605,0xa2a4ad,0x43190310,0x298c7b90,0xcc384ab2,0x1206e0a9,0x4201b9c9,0xc7ad4f02,0x1d4239d4,0x2d13c622,0x4b8c32a4,0x1a7);
remainders_neg[905].modulo_value=3620; remainders_neg[905].bit_mask=create_bit_mask(113,0xa80a020a,0x22a0800,0x28822088,0x2020022,0xa288800a,0x80082a82,0x80280280,0x200a80a8,0x8aa280aa,0x88882a,0x28080a08,0x8008a00,0x20822228,0xa82a828,0x2080aa2a,0xa820208a,0x22288a2,0x28080288,0x88222820,0x82808a08,0xa0002208,0x208222,0xa8a2020,0x8a802000,0x220820a,0x808a2008,0xa00a0aa2,0x200a228,0x202a02a,0x8028a228,0x880a2a82,0x22008022,0xa8a8208,0x8a800020,0x20002022,0x2a2208,0xa088220,0x20888288,0x80a020a2,0x22008a00,0x20aa288a,0x8228820,0xa82a2a8,0x820a0aa,0x80a222,0x80a00288,0x88008280,0x2a8aa088,0x8022a80a,0xa008a80a,0x80080a00,0x8a2a0aa0,0x2028008,0x8a22002,0xa2008822,0x80a80082,0xa2028202,0xa880280,0x882a002,0xa2200822,0xa8822282,0x200080,0x2088a208,0x8800a80a,0xa0002a00,0x88000000,0x820a8880,0x28a0288,0x20288a8,0x2082822,0x20080,0x280208a0,0x8000228,0x8080a020,0x2280a0a,0x80880a02,0x220880a0,0x220022a8,0xa22a828,0x22a88002,0xa080008a,0x2a0a8220,0xa080,0x2002882,0x2200a228,0x200a2280,0x208a00,0x2a008280,0x828220a8,0xa2288000,0x2820008a,0x220a0aa2,0x220aa200,0x80828088,0x20202808,0x80a8280a,0x8020280,0xa0a2000,0x28820,0x20008020,0x20220a08,0xa00a88a0,0x2088a028,0x88088a8,0x2800000,0x8802a00,0x82280a80,0x82288,0x8a808002,0x22a0a220,0x88220802,0xa82002a0,0x2280a008);
remainders_pos[906].modulo_value=3624; remainders_pos[906].bit_mask=create_bit_mask(114,0xa2802822,0x28820200,0x200008a,0xa28a2082,0x28028020,0x202208,0x80082002,0x28220208,0xa28208a,0x82022882,0x8208228,0xa28028a,0x80000,0x20808020,0x8000a20a,0x800808a0,0x8220808,0x80280202,0x228a0022,0x28220a00,0x8a282200,0xa0080002,0x28a08a08,0x8200200a,0xa20220a0,0x8828a00,0x8a002200,0x20822082,0x8200228,0x80008202,0x808a2022,0x20028000,0x82080202,0x2820,0x28008008,0xa288,0x200a2020,0x228020,0xa002280,0x208220a0,0x28808820,0x8a080088,0x82002002,0x28a20a08,0x28a288,0x208000a0,0x20800208,0x208002,0x20000820,0x28808220,0xa20a208,0x80082022,0x20a00,0x82208200,0x800008a2,0x28220,0x800200a,0x828000a0,0x20000208,0x8208828a,0xa0800000,0x8200200,0x8a08a088,0x82022820,0x20000808,0x2080008,0x80020880,0x20820,0x8a28000a,0xa08a2822,0x8008220,0x20a280,0x22022822,0x8820808,0x8800a00a,0x2880002,0x8a00808,0x8a000008,0x2002822,0x28000020,0x80208208,0x2800880,0x8a20200,0x82000288,0x80082080,0x8820828,0x8800a282,0xa2822000,0x8808a00,0x800820a,0xa20a28a0,0x200a20,0x8828a280,0xa0882800,0xa28800,0x80280288,0x20882080,0x20200220,0x8a00020a,0x20208a0,0x200008,0x8028a000,0x820820a2,0x28808228,0x828a082,0x808828a2,0x8200220,0x88080080,0x2802820,0x8a28a08,0x8082,0x808228a2,0x28028a00,0x88);
remainders_neg[906].modulo_value=3624; remainders_neg[906].bit_mask=create_bit_mask(113,0x220820a2,0x20028a08,0x8a088002,0x22022802,0x20020828,0x8828a280,0x802882,0x20a28a00,0x8220a002,0x28a2002,0xa00a20,0x82208202,0x80800880,0x28000828,0x8082282,0x800020,0xa28000,0x820828a,0xa20208a2,0x20a28208,0x220a288,0x20800882,0x20200200,0xa00a082,0x228a2820,0x20208,0x208a288,0xa00a2802,0xa08a20,0x808028a,0x228a2,0x28820808,0xa00828a,0x808820a0,0x20800800,0x88082200,0xa08228a0,0x8a20828,0x8208a208,0xa00a2820,0x20000028,0x2008000,0x2882882,0x20228200,0x88202200,0xa0080820,0x28008a00,0x8828008a,0xa08800a0,0x20000a28,0x82282280,0x8028a2,0x8828208,0xa280288,0x880022,0x8820a28,0x8008a082,0x2080820,0x28808a00,0xa000202,0x20080880,0xa08208,0x2002000,0x28a20a0,0x28808000,0x8a008080,0x8a0000,0x8828028,0x2208082,0x200220a2,0x800a28,0x88282208,0xa28a20a2,0x28000,0x82082,0x82000882,0x208000,0x2088080,0x828820a2,0x20808828,0x8280200,0x82080000,0x228a08,0xa088200,0x802800,0x28020,0x8220a,0x220a2880,0x208,0x80080282,0x22822020,0x8a08228,0x202208,0x20002080,0x8220008,0x8208200,0xa0002800,0x28a08a28,0x2088282,0x828a0020,0x8a08800,0x8202088,0x2802822,0x20000a20,0x8020200a,0x2022,0xa08a28,0x8008,0x820820a0,0x220200,0x8808000a,0xa2022,0x20820208);
remainders_pos[907].modulo_value=3628; remainders_pos[907].bit_mask=create_bit_mask(114,0x8a202a8a,0x88228a82,0xaa2aa808,0x8a282088,0xa08aa00a,0x220a8280,0x82000280,0x8a0822a2,0xa2aa2802,0x222a2202,0x8aa20aa8,0x280aaa0a,0x220802a8,0x2aaa088,0xa2a0822a,0xa202020,0x2aaa2820,0xaa88a88,0x8aa00a0,0x20228208,0x2a080a00,0x88aaa80a,0x22020080,0xa8020a82,0x20000088,0x82888aa2,0x22888a20,0xa0a20aa,0x82aa0280,0x8802882,0xa288222,0xa8880222,0xa0222aaa,0x888a028a,0x222a2a8a,0x8082a000,0xa8aa82a2,0x22a28a08,0x2a82a80,0xa20220,0x82a8a080,0x828a8a8,0xa808a28,0x88a22800,0xa0a00aa2,0x2028002,0x88a00288,0x80a8880,0x820aa080,0x8a0808a2,0x88aa0aaa,0x282a0a02,0x8202a802,0x802228a0,0xa222a000,0x820a0208,0x80200a8,0xa0822a0,0xa888a208,0x88802000,0xa80820a2,0xa828802,0xaa2280a,0x20a2aaa,0xaa0828a2,0x2a020020,0xa22022,0x28080228,0xa0a800,0x88a228aa,0x202a0002,0xa282028a,0xa2a822a2,0x8a00020,0x82a80088,0x28a8802a,0x2aa2a20a,0xa82028a8,0x8a888000,0xa222a2a,0x8880aa28,0x222aaaa,0x228a0888,0x80220888,0xaa82822,0x8a0a0028,0x22288aa0,0x2228208a,0x88aa,0x802a220,0x808882a,0xaaa22020,0x20a8a02,0x2880800a,0xaa20208,0x22aa00a0,0x8aaa822a,0x808a0082,0x20a8a080,0xaaa80a88,0x2088220,0xaa0282a8,0x8aa2a0a,0x8a8882aa,0x8aa8a808,0x820a2802,0x828a8,0x2a088028,0xaa20a028,0x828a2a00,0xaa8aa220,0x28822202,0x808a282a,0xa2a);
remainders_neg[907].modulo_value=3628; remainders_neg[907].bit_mask=create_bit_mask(114,0x2a8a202,0xaa020a,0x22a22080,0x2a0a800,0x28022882,0xaa820a08,0xa888a08,0x280aa2a,0x2a22a08a,0xaaa2aa8a,0x22a8220,0xa0822282,0xaa808a20,0x8a222800,0x2a280aa2,0x82a8a8a8,0xa222a0a8,0x82200200,0x80228828,0xa8aa0a80,0xa2808288,0x222082,0xaa8a8808,0x208a820a,0xa8888800,0xa00022a,0xaa0002a8,0x8282a822,0xa228208,0x8008a00a,0x82a00aaa,0x20008aaa,0x28aaa222,0x28a02,0xaaa2a202,0x80a2888,0x20220a2a,0xaa2a0280,0x8a20a208,0x882a8aa8,0xa202808,0x80a02020,0x820802a0,0x2aa088,0x2828822a,0x8a8a088a,0x288a00,0x80820008,0xa822808,0x280802a,0x228222,0xa0a2828a,0xa8a208a,0x8aaa028,0x2aaa2888,0xa828a80,0x808a8820,0x8280aa28,0x20002a80,0x8a888,0x2080a82a,0x820a008a,0x822aa082,0x8a82a222,0x2280a02a,0xa28a88a8,0x882aa8aa,0xa0808aa0,0x88282088,0x2aa022,0xa8a2888a,0x2a0a8a02,0x2a20aa2a,0x802888a8,0xa208800,0xa02008a2,0xa22a2a82,0x20a8a020,0x2000888,0x82aaa2a2,0x822a0,0x8aaa2222,0xaa028000,0x8aa8000,0x8220a0aa,0x28288a0,0xaaa00228,0xaaa0a802,0x88880022,0x808a2aa8,0x808000a2,0x222aa8a8,0x8a82028a,0xa0080882,0x822a8a80,0xaa228828,0x2220aa2,0x802880a,0xa8202808,0x22208200,0x8a800aa8,0x228a0a20,0x8022a282,0x2000a22,0x2202080,0xa82a08a,0x8880a020,0xa28008a0,0x22a828a0,0xa02a288,0x22022aa8,0xa00aaa8a,0x802a0a2,0x2a2);
remainders_pos[908].modulo_value=1816; remainders_pos[908].bit_mask=create_bit_mask(57,0xa2208202,0x2aa008a,0x8a2a222a,0x8a8aa822,0x80a0a2,0xaaaa220a,0xa0aa2080,0x28802802,0x8880002a,0x282a2a82,0xa0020208,0x88808208,0x2a800a80,0x8a288802,0x228a2a0a,0x2a8802,0x2022202a,0x800a208,0x8a828208,0x22282a,0x80228a80,0x8aa0a802,0x8aaaa020,0xa0200a08,0xaa2a28a0,0x8a8a2882,0xaa80a88,0x2088aa20,0x8202808,0xa008aa22,0xa222a02a,0xaa8228a2,0xa0a28a8,0xa220a008,0xa2080aaa,0x2802a0a,0x2a288,0xa2a82888,0x20208282,0x8208a00,0xa80888,0x888022a8,0xa2a0a8a2,0xa8802228,0x2202a002,0xa208202,0x28208080,0x2282a8a8,0x28a80002,0xa802802,0xaa0208aa,0xa088aa,0xa28a0a02,0xa2882aa2,0x80a888a8,0x8aa200aa,0x808208);
remainders_neg[908].modulo_value=1816; remainders_neg[908].bit_mask=create_bit_mask(57,0x2aa80a8a,0x8a228802,0x2a2aaa2,0x20220aa,0x8808282a,0x2222aa82,0x2822a808,0xa008a082,0x888a2,0xa0a2a20a,0x288a8a80,0x80a80,0xa2088208,0x2a0008a,0xaa02a282,0x88a2008a,0xa8aaa8a2,0x80882a80,0x20a0a80,0x88aaa0a2,0x8aa0208,0x228208a,0x22228a8,0x28a88280,0x22a2a028,0x202a00a,0x82208200,0xa80022a8,0x80a8a080,0x288022aa,0x2aaa28a2,0x220aa02a,0x8282a020,0x2aa82880,0x2a808222,0x8a082282,0x888a2a00,0x2a20a000,0xa8a80a0a,0x80a80288,0x88208000,0x8aa20,0x2a28202a,0x2008aaa0,0xaa8a288a,0x82a80a8a,0xa0a80808,0xaa0a2020,0xa020888a,0x8208a08a,0x228a8022,0x88280022,0x2a02828a,0x2a00a22a,0x8200020,0x22a8822,0x80a80);
remainders_pos[909].modulo_value=1212; remainders_pos[909].bit_mask=create_bit_mask(38,0x828a2022,0x28820,0x8020a082,0x22080802,0x28000208,0x2280200,0xa0880822,0x8a08a08,0x8a000000,0xa0080802,0x28,0x88288288,0x22080882,0x200a20,0x820000a,0xa0080822,0x20828200,0x8208a000,0xa20228a0,0x20a28808,0x8000a208,0xa0082820,0x8820200,0xa000082,0x808a0080,0x28220208,0x2282282,0xa2800000,0x28020200,0xa,0xa20a20a2,0x8820220,0x80080288,0x82080002,0x28020208,0x820a080,0x20822800,0x8808a28);
remainders_neg[909].modulo_value=1212; remainders_neg[909].bit_mask=create_bit_mask(38,0x20228a2,0x8820028,0x828200a,0xa2800082,0x20808a00,0x8a208288,0x200000a2,0x200200,0x2088088,0x20800082,0x8808820,0x200200,0xa2800002,0x8a08228,0x80288082,0x208000a2,0x28020a08,0xa002088,0x228a2020,0x28220000,0x8082280,0x208020a0,0x28a00,0x8208800a,0x20800,0x20a28a00,0x8a20a20a,0x22080880,0x20828a08,0x88088082,0x22822822,0x28a28,0x8008200,0x2800882,0x20028a00,0x80282008,0xa00a2080,0x220);
remainders_pos[910].modulo_value=3640; remainders_pos[910].bit_mask=create_bit_mask(114,0x88a020a,0x2a288200,0x20200a8,0x8a002020,0x82a02200,0xaa0888,0x80288282,0xa8000,0x8202882,0x82808022,0x200088,0xa0220a80,0xa802080,0x80a280a0,0x22080a20,0x200a00,0xa20a0802,0xa080020,0x8200008,0x82802a00,0x20082002,0x2080800,0x2208a020,0x2200000,0x8020082a,0xa8a00a8a,0x2220200,0x20008028,0x22028,0x20088a22,0xa20202,0x80008800,0xa200,0xa222820,0xa0280022,0x28882002,0x2280882,0x20800280,0x8822a828,0x2208800,0x80a82000,0x20208220,0xa00a020,0xa000002,0x2880800,0x88002002,0x88028a02,0x2080280,0x80802028,0x288800a,0x20800802,0x80008002,0xa020000,0x8a222800,0x82220,0xa8820000,0x20000,0x2a000080,0x82,0xa2088820,0xa0002888,0x2000080,0x8800200,0x80802002,0x2a00222,0x80280802,0x22028020,0x2280a280,0x80800800,0xa0002022,0xa0800000,0x8080a00,0x2088208,0x8000082a,0x22002208,0x8282a88,0x80028002,0x28822028,0xa800822,0xa0880028,0x82888,0x2008a00,0x2200,0x880808a,0x88a220,0x8280880,0x80280200,0x2a008088,0x2a2a00a,0x80a82008,0x88000008,0x80080a20,0x8002020,0x82800820,0x2000a802,0xa0200008,0x8a080020,0x8020a0,0x8800a008,0x208a020,0xa00a028a,0xa020802,0x2a088,0x82220008,0x20880202,0x822808,0x20002a0,0x828228,0x822220aa,0xa200880a,0x80080800,0xa82a0080,0x20008228,0xa080a2);
remainders_neg[910].modulo_value=3640; remainders_neg[910].bit_mask=create_bit_mask(114,0x80020a02,0x22200a00,0x28a8020,0x200a8a8,0x2202a08,0x88220000,0x8200202,0x8020008,0xa0a002,0x208002a,0x8280800,0x28228200,0x2002008,0x8220028,0xa2808228,0xa80280,0x22028082,0x20000a8,0x80a08880,0x2082200,0xa800208a,0x8a008000,0x2a8020a0,0x2a08888,0x288022,0x20200202,0x8a2a0a80,0x28880020,0x8082a820,0x20800222,0x802a0a8a,0x8080080,0x8802280,0x822220a8,0x20a0802a,0xa000200a,0xa208002,0x20088208,0x2220a0,0x82a00008,0x8202808,0xa02802a0,0x2882028,0x2800882,0x82008008,0x82882,0x20282,0xa800208,0x800a0a0,0x2000802,0x28080082,0x882,0x20a8088,0x2a2a008,0x80802a28,0x200a0088,0x880a0800,0x22888000,0x80808802,0x22000028,0x28802000,0x8a080000,0x88288,0x800a80a,0x2208a22,0xa0008a,0xaa0a0820,0x22002208,0x8000088,0x2088202a,0x28080808,0x8280,0x2000280,0x80080a2,0xa280a200,0xa02200,0x80a0082,0x2002a020,0x20080a2,0x20008820,0x80002000,0x8a000280,0x8882288,0x80000002,0x80002a20,0x200000,0x208a80,0x22080000,0xa222882,0x202800,0x800080,0x8008220,0x88a020,0x20080a8,0xa008200a,0x28a00880,0x20008a0,0x808a020,0x2800,0x2802820,0x20820202,0x20a8082,0x8822000,0xaa20880,0xa000820a,0x880a2000,0x2088a20,0xa02a0,0xa222822,0x22800002,0x8800080,0x20228000,0x28880220,0x200022);
remainders_pos[911].modulo_value=3644; remainders_pos[911].bit_mask=create_bit_mask(114,0x2002aa2,0xaaaaa208,0x2a8aa22a,0xa2228008,0x20a0880a,0x80088288,0x80202082,0x8022082a,0xaa088282,0x28a2808a,0x28282aa,0x2a22a0,0xa82288a2,0x2a288888,0xa0220a88,0xa008a088,0xa82a80a2,0xa22aa0,0xa880aaa2,0x2a2aa80a,0x88828a20,0xa8a20800,0xa80aaa0,0x808a8a2a,0x28a8200,0x800a8a8a,0x8828a2a8,0x8a80a8a0,0x222088a,0x28202a02,0x200820a2,0x820202aa,0x2a8a020a,0x8082022a,0x28002a02,0x2aa28820,0xa8820a22,0x2a00080,0x88002a20,0x2800882a,0x82a00a0,0x222822a8,0xa20288a8,0x22222080,0x882208a0,0xa80880aa,0x800a0a0a,0x22a0820,0xa0a2280,0xa288aa,0xa28a8aa,0xa20a22aa,0x2a22828,0xa2aa0888,0x880200,0xa2880000,0x800aa8a,0xaa2aa002,0x228,0xa0080220,0x2220aa8,0xa28288a8,0xaa88a08,0xaa2a28a,0xaa228a0,0x288a0a,0x20820a88,0xaa0a0a00,0x2aa02202,0x80a20882,0xa0208888,0x82a22808,0x2a88288,0x80a00a82,0x2a822002,0x8a8002,0xa02000a8,0x888a0822,0x808228aa,0x280a8002,0x8a880820,0x2a080a2a,0x8aa80808,0x88a08200,0x80a8082,0x2a220888,0x20a2a02a,0x22a8a282,0xa2a2a00,0x20082a28,0xa8a2a20,0xa0aaa02a,0x200208a2,0x808a2822,0xaa02aa8a,0x8aaa022,0xa0aa88a0,0xa8a02a82,0xa220a200,0x822a0880,0xa222228a,0x8a22882,0xa88a80,0x8aa82828,0xaa2028a2,0x2828220a,0x2a820880,0x28208080,0x82282200,0xaa0220a0,0x82002888,0xaa88aa2a,0x208aaaa,0x8aa8008);
remainders_neg[911].modulo_value=3644; remainders_neg[911].bit_mask=create_bit_mask(114,0x8a88a22a,0x22222a80,0xa2022aa2,0x2aaa0880,0xa8280082,0x8800a00,0x8a8a80a,0x8aa80a2,0x22800a0a,0xa02a0802,0x8a0a0a22,0x88a2aa28,0x20aa002a,0xa2a00000,0x28aa8200,0x28802800,0x20a2082a,0x882aa228,0x2008222a,0xa2a22082,0xa02a8,0x202a8088,0x82082228,0x80202a2,0x8a020a88,0x8820202,0xa02a20,0x2082028,0x8aaa0002,0xa0a8a28a,0xa880a82a,0xa8a8a22,0xa2028a82,0x80a8aa2,0xa088a28a,0xa22a00a8,0x200a82aa,0x8a288808,0x88a2a8,0xa08800a2,0x80a28828,0xaaa0aa20,0x2a8a0020,0xaaaaa808,0xaa8028,0x20800822,0x8828282,0x8aa280a8,0x8282aa08,0x882a0022,0x82a02022,0x2a82aa22,0x8a2aa0a0,0x2a228000,0x88008a88,0x2a008888,0x80882202,0x22a2288a,0x88888aa0,0x28808aa8,0x8aaa8220,0x2a0a0020,0x82200280,0x822a2a02,0x822aa028,0x88a00282,0xa80a8200,0x22828288,0xa228aa8a,0x82a800a,0x28a80000,0xa2aa080,0x8a200a00,0x828820a,0xa20aa88a,0x8802088a,0x28a88820,0x280aa,0x80aa022,0xa082088a,0x20080a8,0xa28082a2,0x2208080,0x280a88,0x8082080a,0xa2aa8000,0xa82a28a2,0xaa202a0a,0x82a2a288,0xa880a2a0,0x8202a2a8,0x282228a2,0xa88a802a,0x802a0aa,0x228a2202,0x802228aa,0x28220028,0x2028a20a,0x2aa82a88,0xaa28008,0x2aaaaa02,0x802aa00a,0x88200208,0x220a0a0,0x22a8a02a,0xa0a0aa82,0xa20a8008,0xa0a80808,0xaa0aa88,0x228aa828,0xa88a000,0x220022a2,0x8a802222,0x220880);
remainders_pos[912].modulo_value=456; remainders_pos[912].bit_mask=create_bit_mask(15,0x22000082,0x28a20a00,0x2200282,0x20a0800,0x28008800,0xa08a288,0x20000880,0x20000828,0x8a20a002,0x22002822,0x20a08000,0x80088000,0xa08a2882,0x8800,0x82);
remainders_neg[912].modulo_value=456; remainders_neg[912].bit_mask=create_bit_mask(15,0xa2800802,0x20228208,0xa28820a,0x82820080,0x20800008,0x82002200,0xa0880000,0x28008020,0x228208a,0xa28820a2,0x28200800,0x8000088,0x20022002,0x8800008,0xa);
remainders_pos[913].modulo_value=3652; remainders_pos[913].bit_mask=create_bit_mask(115,0x8a88220a,0xa0208a8,0x20228a20,0xa28a20,0x2a0802,0x80020a00,0x808a880,0x8088aa22,0x8a808280,0x2a0a8802,0x880aa222,0x8008022a,0xa2008,0xaaa8080a,0x820a280,0x288a2822,0x200a020a,0xaa008002,0x2202000,0x2a02aaa2,0x2028000,0x8a08a0a8,0x2822282,0xaa2880a0,0x882200a2,0x822a220,0xa80a0aa8,0x8282280a,0x22a8aa22,0x80a20a0a,0xaa8280aa,0x222a2080,0x28022088,0x2808a2aa,0xa220a00,0xa828828a,0xa0200,0x2aaa02a0,0x202202,0x802a8,0x82028022,0x20a288a2,0xa282082,0x8080aaa8,0x80228002,0xa2008008,0x222a808a,0x8a82a2,0xa080a8a,0x22a88808,0x8a88082,0x2820008,0x82a000,0x228a2802,0x228a2020,0xa8820280,0x82208a88,0x2a20882a,0x280822a2,0x808a2880,0x28a2880,0xa82008,0x82800,0x2022a202,0x222a888,0x2a020a02,0xa82a200a,0x202a8888,0x2008aa,0x288022,0xaaa02028,0x20828a02,0xa228a088,0x80280828,0xa8020008,0x8808002,0xa80aaa88,0x80a0000,0x282282a0,0xa088a0a,0xa8a20280,0x2088028a,0x208a8882,0xa0282aa0,0xa08a02a,0x8aa2a88a,0x2882828,0xaa0a02aa,0x88a88202,0xa0088220,0xa0228aa8,0x28882800,0xa0a20a28,0x280802,0xaaa80a80,0x808808,0x200aa0,0x80a0088,0x828a228a,0x28a08208,0x202aaa0,0x8a000a,0x88020022,0x88aa022a,0x22a0a88,0x28202a28,0x8aa22020,0x22a20208,0xa080020,0x20a8000,0x8a28a008,0x8a288080,0xa2080a00,0x8822a22,0xa);
remainders_neg[913].modulo_value=3652; remainders_neg[913].bit_mask=create_bit_mask(115,0x200a282,0x820a8020,0xa8a202a8,0x882a02a0,0x88a2008a,0x8a8288,0x80802008,0x800222a,0x2000a08,0x2282008a,0x822aaa,0x88082a2,0x802a880,0x22208082,0x80a82a00,0xa00220aa,0xa0828a82,0x2288088a,0x8aa8a808,0xa282222a,0xa8a0888,0x2802820,0x8a0aa20a,0x22200028,0xaa882a,0x80aa2aa0,0x20820220,0x20aa082,0xaa2022aa,0x82a8202,0x22020822,0x2aa2a808,0xa08aa800,0xa0002222,0x822a8288,0x20a00a02,0x88828a80,0xa2220a28,0x80a8aa8a,0x88808a20,0xa8a082a,0xa822002a,0x2a0a80a,0x8082220,0x8aa0082,0x2a080880,0xaaa20802,0x88020a22,0x82800202,0xa2200080,0x8020080a,0x8a0a8800,0x88022888,0x2a02a08a,0xaa0228a8,0x200a8208,0xa280200,0xa2a800a2,0xa080aa22,0x8022008,0x8202a008,0x8820a880,0x8880a008,0xa8a22a8a,0x2aa2000,0xa28a828a,0x20a2a082,0xa8220000,0x88a88022,0x88a008a2,0x222828a0,0xa00a028a,0x2aa02800,0x8a08020,0x20828880,0x8080a,0x20822200,0x80828088,0xa02a0a28,0x82800282,0x202a8a00,0xa8000a02,0xa002000a,0x28a0a228,0x82802822,0x2222002,0xa00a0a0,0x22828a22,0x20028a,0x28000aa8,0x28aa0220,0xa000a080,0x282a02a0,0x80a0808a,0x22208208,0x8080000,0x88a08228,0x828800,0xa02aa02,0xa0280280,0x8a0a2228,0x88028882,0x8a88a2,0x220aa2,0x82a28200,0xa0a8a2a0,0x22aa028,0xaa228a80,0x28088a8,0x8a820888,0x2a02080,0x2200808,0x2a808288,0x800aa2a2,0x2);
remainders_pos[914].modulo_value=3656; remainders_pos[914].bit_mask=create_bit_mask(115,0x2022aa2,0xaaa2a828,0x28aaa222,0xaaa00028,0x2a08800,0x8a28a200,0x20a80200,0x8228a02,0x20280a0a,0x2aa220aa,0x28822a2,0x8082280,0xaaa2aa0a,0x2aa0a80,0x228a8082,0xa220820a,0x80a0a22,0x28882a8,0xa22aa2a8,0x28aa20aa,0xa80a0a08,0x20aa8a02,0x2a008a88,0xa2888aa0,0xa2208800,0x20282888,0xa222aaa2,0x8a882822,0x2202a22a,0x28288080,0x22202,0x88282a22,0xa88a2208,0x88820a,0x808aa02a,0xa0800a28,0xa0a0a802,0xa200828,0x2002200,0x28882a8,0x20a0a8a8,0xa28a2202,0x82808220,0x80a002a2,0xa002000a,0x8228a8a0,0x22882aa,0xa220022,0xa0a82a20,0xa08228a0,0xa2a80a22,0xa208208a,0xa88a02aa,0x2aaa000a,0x22000828,0x28020002,0x22a2a8,0x8a880082,0x80282a,0x20008880,0xaaa828,0x80a22aa0,0x8208aaa,0xa02a8aa2,0x28820a88,0xa82a0a0a,0x88a008,0x82288088,0x2a2882aa,0x800a0a,0x800a02a0,0x8202828a,0x88a28a08,0x2a0a0880,0x8222802a,0x8800802a,0x2008a000,0x2a0a0a28,0xa0020a80,0xaa20228,0x822200a8,0x88a22aa0,0xa8282220,0x88800088,0x2282880,0x8a808802,0x2822a2a8,0xaa888a88,0x2828088a,0x22088a22,0xa2228a00,0xa200a80a,0xa2aa0822,0xa0a02a80,0x8aa2820,0x8aa88aaa,0x8222802a,0xa0a0202a,0x82088a88,0x2a288a0,0xa0aa8082,0xaa8aaa02,0x882020a0,0x88228002,0x88aa88a,0xa02808aa,0xa28820a0,0x802a0880,0x8a28a200,0x220a8000,0xaaa00,0x8aaa2828,0x2a8aaa88,0xa8808028,0x8a);
remainders_neg[914].modulo_value=3656; remainders_neg[914].bit_mask=create_bit_mask(115,0x8a8aa22a,0x222a20a0,0xa0222aaa,0x222888a0,0x8a280088,0x2a02a88,0xa8208a88,0x80aa028a,0xa8a08282,0xa22aa822,0x8a00aa2a,0x8080aa08,0x222a2282,0x8a228208,0xaa02080a,0x2aa80a82,0x808282aa,0x8a000a20,0x2aa22a20,0xa022a822,0x20828280,0xa822028a,0xa2880200,0x2a000228,0x2aa80088,0xa8a0a000,0x2aaa222a,0x200a0aa,0xaa8a2aa2,0xa0a00808,0x888aaa8a,0xa0a2aa,0x2002aa80,0x88000a82,0x80228a2,0x280882a0,0x2828208a,0x82a880a0,0x8a88aa88,0x8a000a20,0xa8282020,0x2a02aa8a,0x2080aa8,0x8288a2a,0x288a8882,0xaa02028,0x8aa00a22,0x82aa88aa,0x2820a2a8,0x280aa028,0x2a2082aa,0x2a80a802,0x20028a22,0xa2228882,0xaa8880a0,0xa08a888a,0x88aa2a20,0x200880a,0x8808a0a2,0xa8880008,0x882220a0,0x82aa228,0x80a80222,0x28a2022a,0xa00a8200,0x20a28282,0x88002880,0xaa00800,0xa2a00a22,0x88088282,0x8828a28,0xa8a0a02,0x2a0280,0xa2828008,0xaaa08a2,0x8808a2,0xa8802888,0xa28282a0,0x288a8208,0x822a8aa0,0xaaa8820,0x2aa228,0x20a0aaa8,0x88800,0x8aa0a008,0x208008a,0xa0aa2a20,0x22000200,0xa0a08002,0xaa8002aa,0x2aaa0288,0x2a882082,0x2a2280aa,0x2828a208,0x8022a0a8,0x2200222,0xaaa08a2,0x2828a8a2,0xa800200,0xa2a0028,0x2822080a,0x2202228a,0xa8a828,0xaa088a,0x80022002,0x28a08022,0x2a00a828,0x8a28008,0x2a02a88,0xaa820888,0x88822288,0x222a0a0,0xa2022200,0x200808a0,0x2);
remainders_pos[915].modulo_value=3660; remainders_pos[915].bit_mask=create_bit_mask(115,0xa0800002,0x20020,0xa00a000,0x828220a0,0x20808a28,0x8080208,0x200a0022,0x828028,0x8a200002,0x22002022,0x20a08,0x80288202,0x20022020,0x28000,0x82288208,0xa00a0880,0x220220,0x8000,0x820008a2,0x800008,0x8080282,0x22020082,0x8808200,0x8200208a,0xa0002820,0x20220808,0x2080280,0x20802080,0x8028028,0x282000,0x820882,0x20008820,0x88000,0x820820,0x200200,0x808008a,0x2020002,0x28820208,0x2208002,0xa2082800,0x8800a08,0x2008000,0x2000020,0x28820,0x8008a00a,0x80002800,0x28a00,0x8088080,0x82800000,0x20000a08,0x88002080,0x82020080,0x20020,0x8000000,0xa0080020,0x28820808,0x8a280000,0x28a2000,0x8228000,0x200a202,0x20080,0x8000000,0x8082080,0x80022020,0x8020,0x28220a,0x22020000,0x28000020,0x8000200a,0xa2002002,0x28000a00,0x80082,0x20080080,0x220000,0x8208a20a,0x20880002,0x8228800,0x8080208,0x2020800,0x800a20,0x8200008,0x22000082,0x20008000,0x8200082,0x82800822,0x28020000,0x80208280,0x2080020,0x8808028,0x8000a202,0x80082082,0x20220a20,0x8088008,0x2020820,0x200828,0x82200,0x200008a2,0x8800000,0xa00a088,0x22882022,0x28000208,0x88008000,0x22802080,0x8000808,0x8008820a,0x8a2880,0x28200800,0xa008280,0x2020880,0x20208208,0x8828228a,0xa00a00a0,0x8000000,0x20a080,0x800);
remainders_neg[915].modulo_value=3660; remainders_neg[915].bit_mask=create_bit_mask(114,0x20080882,0x8028820,0x2082088,0x2022820,0x20000220,0x80000280,0xa00208a2,0x8020020,0x220808a,0xa2882822,0x8820200,0x8200282,0x208a20a0,0x8820008,0x2200200,0x20820000,0x8228a20,0x8080088,0x2800022,0x8800,0x20a,0xa20a0802,0x208,0xa00a002,0x20882020,0x28a20000,0x8a008200,0x20082000,0x820820,0x8020a008,0x800a0002,0x28000020,0x8000088,0x800200a0,0xa08a08,0x80000002,0x820a0802,0x20028200,0x8a20008a,0x22802000,0x200,0x8a080080,0x28800a0,0x820028,0x2002,0x882080,0x8020200,0x8,0x2000880,0x20808200,0x82008,0x20a0800,0x8828028,0x80008088,0x20800820,0x20020000,0x2208080,0x2022080,0xa20808,0x82082202,0x808a0800,0x8800,0xa008,0x8228a0,0x800828,0x88202282,0xa20a0880,0x20808028,0x800a082,0x22882802,0x20800208,0x88008002,0x20800000,0x8a28808,0x2002202,0xa0000882,0x220000,0x8280,0x82820080,0x8228,0x80280080,0x22080802,0x28800008,0x8020800a,0x2080022,0x20820808,0x8280200,0x28000a0,0x820,0x220a,0x802802,0x28228220,0x80,0x828200a0,0xa08020,0x88002288,0xa0080022,0x8008,0x82002000,0xa2002822,0x20800a00,0x80080,0x22082000,0x808000,0x202,0x80022000,0x20208000,0x2080208,0x82820000,0x20a00a00,0x202202,0x20020820,0x800008,0x88202008);
remainders_pos[916].modulo_value=1832; remainders_pos[916].bit_mask=create_bit_mask(58,0xa0a8a2a,0x22aa2822,0x882a0880,0x2200808a,0x82a80082,0xaa02aa2,0x2a20802,0x822a0202,0xa02a8200,0xaa0a2a,0x8022208,0x82a08a88,0xa2aa2008,0x8a828220,0x2a28aaa2,0xa8a08828,0xa822008a,0x2022a20,0xa0020888,0x80aa8a0a,0x8820082a,0xa808080a,0xaa080208,0xa828aa80,0x8882002,0x822a2020,0xa880220a,0xa08828a,0xa2aa8a2a,0x8220a0a8,0x88022aa2,0x8a882a0,0x8222008,0xaa282a80,0x8020aa02,0x20202a20,0x200822a0,0xa2aa02a8,0x20800aa0,0xa8808022,0x882a08,0x220a2aa2,0xaa28a828,0x8282a28a,0x8aa8a08,0xa20a8220,0x88802022,0xa0aa0020,0x82a80aa8,0x80a88200,0x208a8080,0xa80aa080,0x2a828a,0x2008882,0x20a822a2,0x28aa8802,0xa2a0a088,0xa8);
remainders_neg[916].modulo_value=1832; remainders_neg[916].bit_mask=create_bit_mask(58,0x828202a2,0xaa22a0aa,0xa28008,0xaa880802,0xa20880a,0x8228a22a,0x8a2a808a,0xaa28a8a,0x28a20a88,0x882282a2,0x808aaa80,0xa280200,0x2a22a880,0x20a0aa8,0xa2a0222a,0x202800a0,0x20aa8802,0x8a8aa2a8,0x288a8000,0x8220282,0xa880a2,0x20800082,0x22808a80,0x20a02208,0x8000a88a,0xaa2a8a8,0x2008aa82,0x82800a02,0x2a2202a2,0xaa82820,0x8aa22a,0x80200a28,0x80aaa880,0x22a0a208,0x8a8228a,0xa8a8a2a8,0xa880aa28,0x2a228a20,0xa8088228,0x200808aa,0x8800a280,0xaa82a22a,0x22a020a0,0xa0a2a02,0x80220280,0x2a820aa8,0x8a8aa,0x282288a8,0xa208220,0x8200a88,0xa8020800,0x20822808,0x88a20a02,0x8a88000a,0xa820aa2a,0xa022008a,0x2a282800,0x20);
remainders_pos[917].modulo_value=3668; remainders_pos[917].bit_mask=create_bit_mask(115,0x820a8a02,0x208880,0x8a0222,0xa0aa88a,0x22820880,0x8aaa2a20,0x802a2282,0x8802a22a,0x20a80882,0x220288a2,0xa0228008,0x2a800208,0x8000808,0xa002a2a8,0xaa022200,0x22a00282,0x228a20,0x2002000,0x8082a280,0x8822820,0x8aa202a,0x20002820,0x2a800802,0x82a00a28,0x8a0080a0,0x822a088,0x8a2a0a08,0x28022280,0x8a028820,0x2200a020,0x82a2880a,0x2a208082,0x8028888,0x802a0828,0x800aa0a2,0xa0002200,0x22a88020,0xa20882a0,0x2a080820,0x2002000a,0x8a200000,0x2a2022,0x2202aa0a,0x2a002802,0x800080aa,0xaa080,0xa202882,0x8a022080,0x80a82228,0x2a008a08,0x22a00aa2,0xa2aa8a0a,0x820a2022,0xaa828008,0x2880288,0x88200222,0x8a820808,0x2a280a02,0x82220208,0x28088800,0x2aa22802,0x8220028,0xa8a8808a,0xa88a0a2a,0xa88aa00,0xa0220a20,0xa228882,0x8a002088,0x82280,0x2020aa,0xa8aa020,0x8880280,0x800a0aa8,0x8a28808a,0x800000,0xa8a0008,0x8a08202,0xa880a822,0xa08022,0x200880,0x8028a0aa,0x202820a,0x8a822228,0xa8282020,0x88a0228,0xa2080a0,0x2808228,0x8a202888,0x82020a0a,0xa2220a8,0xa820a020,0x2a828a00,0x880200,0xa2008280,0x220a808a,0x20208288,0x80028a8,0x80000080,0xa8808a28,0xaa82800,0xa00888,0x202a8a8,0x2a820200,0x80a20800,0x8820028,0xa088a228,0x2282202,0x802a88a8,0xaa28288a,0x28808a8a,0xa002208,0x200a22aa,0x8008880a,0x8202220,0x80a2a);
remainders_neg[917].modulo_value=3668; remainders_neg[917].bit_mask=create_bit_mask(115,0xa82020a,0x8a80000,0x80028aaa,0x82022002,0xaa028000,0x22222a8,0x8a2a20a,0x8a2a22,0x28208002,0xa28a002a,0x282a0880,0xa2008a80,0x80880080,0x288a2220,0x228aaa08,0x2a288a02,0x80aa02a8,0x8a08a888,0x8022a08,0x800a20a8,0x802220a2,0xa888a028,0x22088082,0x22882a0,0x2080828,0x80a22800,0x2a20280,0xa08aa208,0x28a0028,0x2a8828a0,0x22a0082,0xa228080a,0x80820000,0x8a200a0,0x882202a,0x2888aa08,0x222008a0,0x22800a28,0xa20080a8,0xa8828882,0x2a80888,0x88a2a0aa,0xaa8a2202,0x2288a082,0x880822,0x88022808,0x82a0a00a,0x28a2808,0x820a2a0,0xa2880200,0x2a288222,0x22220282,0xa02a8aa,0x22020800,0x8a000a00,0xa882aa,0x20a8000,0x22a08282,0x2aa8a80,0xa0000088,0xa222a08a,0x80aa08a0,0x20200002,0x20028222,0x2002280,0x20aa82a8,0x822a000a,0x280a800,0x88802a08,0x88a0a022,0x82022828,0x8a00,0x828220,0x2200802,0x88008888,0x82020880,0x8028028a,0x2008202a,0x82808a2,0x80a88008,0x8202822,0x8a820a82,0x20a2aa0,0x20a0a0a8,0x80028a20,0x2a80820,0x82080aa0,0x228a000,0xa828282,0x82aa2820,0x20a820a8,0xa20a0208,0x8008a80,0x22880a08,0xaa020802,0xa8a00a00,0x80882020,0x8888008,0x20080220,0x220a080,0x80288000,0x8a0a2020,0xa2028288,0x82a0088,0x800a80a0,0x28002a20,0xaa0aa82,0xa20020,0x2220a002,0xa0000202,0x82882a80,0xa882a222,0x8800002,0xa8aaa0,0x82a2);
remainders_pos[918].modulo_value=408; remainders_pos[918].bit_mask=create_bit_mask(13,0x82080882,0x20228a20,0xa00a,0x20a2820,0x220028,0x80208200,0x2020080,0x8208,0x80280088,0x828a0,0x8a00a00,0x8208a288,0x822020);
remainders_neg[918].modulo_value=408; remainders_neg[918].bit_mask=create_bit_mask(13,0x2800002,0x28a20228,0x88082082,0x820220a0,0x8a28820,0x280288,0x828a0800,0x8800a00,0x8208000,0x80802020,0x208200,0xa002200,0xa2820);
remainders_pos[919].modulo_value=3676; remainders_pos[919].bit_mask=create_bit_mask(115,0xaa2aa2aa,0xa20a2888,0xa8a28802,0xa2a20200,0xa2a00a2,0x20a0aa88,0x2280088a,0xa0aaa22a,0x28a28020,0x80a2022,0x88880028,0xa2800a22,0x2a88828,0x820a8a28,0x880a8888,0x28a82a00,0x8002aaa2,0x2a082a8,0x20aa2a8a,0x28820002,0x2880aa0,0x28088a08,0x8a28a080,0x228082aa,0x280a0a2,0x880aa008,0x2aaa2aa,0x28a880a,0x20a8288,0x20aa20a,0xa82a2,0x2a20008,0x2a0aa2a8,0x2a088828,0x2082000a,0x22a0aa28,0xa20aa282,0x8a20a802,0x80288aaa,0x280a8200,0xaaa200a,0xa0208800,0x2a22a80,0x28a2222,0x200aa082,0xaa0820a2,0xa2220882,0x82a2a0aa,0x82088a8,0x2828aa,0xaa088088,0x28288222,0x8082a200,0x8808082a,0x8202a8a,0x82880aa2,0x802220,0x2000008,0x20888020,0x8aa0228,0x2a2a8082,0x2a820202,0x8008a820,0xa8882282,0x220220a,0xaa28280,0x22a22082,0xaaa0a8a8,0xa8220888,0x88a0820a,0x820aa00,0x8888a28,0xa02a88a8,0x22080,0x8a008aaa,0x20082a02,0x2aaa2280,0xa802a08a,0x8828aa08,0x828aa0a8,0x8a000820,0x8282220a,0x2a8aa0a,0x20008a8,0x8a82a00,0xa08aa08,0x2282a08,0xa022a28,0x2aa8aaa8,0x200aa02,0x88a0a028,0x2aa82028,0x8020a28a,0x20a2202,0x80aa0228,0x88000822,0xa2a8aa0,0x22a820a8,0x88aaa800,0x200a82a2,0x22222a02,0x28a2a08,0xa28222a8,0x288a0028,0x2800222,0x88808a02,0xa08028a2,0x8a88aaa0,0x8a220028,0x22aa0a0,0xa8a00a8a,0xa00808a8,0xa80228a2,0xa2228a08,0xaa8aa8a);
remainders_neg[919].modulo_value=3676; remainders_neg[919].bit_mask=create_bit_mask(115,0x22a22a22,0x2a82a000,0x202a008a,0x2a2a8a88,0x82a2882a,0xa8282200,0xaa088002,0x28222aa2,0xa02a08a8,0x8082a8aa,0x88a0,0x2a0882aa,0x8a2000a0,0xa8202a0,0x820000,0xa020a288,0x88a222a,0x8a280a20,0xa822a202,0xa00a888a,0x8a008228,0xa0800280,0x2a02808,0xaa080a22,0x8a08282a,0x822880,0x8a222a22,0x8a020082,0x8a020a00,0x8a822a82,0x88820a2a,0x8a2a8880,0xa2822a20,0xa28000a0,0xa80a8882,0xaa2822a0,0x2a822a0a,0x2a8208a,0x8a00222,0xa0820a88,0x8222a882,0x28a80088,0x8a2aa208,0x8a02aaaa,0xa882280a,0x2280a82a,0x2aaa800a,0xa2a2822,0x80a80020,0x88a0a022,0x22800800,0xa0a00aaa,0x80a2a88,0x8080a2,0x80a8a202,0xa00822a,0x8808aaa8,0x8a888880,0xa80008a8,0x80228aa0,0xa2a2080a,0xa20a8a8a,0x88020a8,0x2000aa0a,0x8aa8aa82,0x822a0a08,0xaa2aa80a,0x22282020,0x20aa8000,0x280a82,0x80a82288,0x800002a0,0x28a20020,0x888aa808,0x2880222,0xa880a28a,0xa222aa08,0x208a2802,0xa02280,0xa022820,0x28880a8,0xa0aaa82,0x8a202282,0x8a888020,0x8020a288,0x82802280,0x8aa0a280,0x828aa2a0,0xa2202220,0x8a88228a,0x2828a0,0xa220a8a0,0x8a82a02,0x8a82aa8a,0x8228aa0,0x8880aa,0x82a20228,0xaa20a820,0x222088,0xa8820a2a,0xaaaaa28a,0x8a02a280,0x2a0aaa20,0xa00288a0,0x8a088aaa,0x8028a,0x2808a02a,0x2002228,0x2aa88a0,0x8aa22828,0x20288202,0x28808020,0x208aa02a,0x2aaa0280,0x2202202);
remainders_pos[920].modulo_value=920; remainders_pos[920].bit_mask=create_bit_mask(29,0x222202,0x2a028a02,0x208aa008,0x8222a802,0x82200a28,0xa0282280,0xaa2a08a0,0x28,0x8220808,0xa2208202,0x220800,0xa02a0,0x2802028,0x202a0a0,0x80888222,0x800a0a80,0x280802,0xa80a0,0x8a002088,0x20808208,0x202088,0xaa280000,0xa0a20a8,0x82028828,0x8228a008,0x8802a88,0xa8200aa2,0x80a280,0x808888);
remainders_neg[920].modulo_value=920; remainders_neg[920].bit_mask=create_bit_mask(28,0x882a2a8a,0xa20a0282,0x28022080,0xaa2208a,0x2a88220,0x28a02a08,0x22220020,0x88880a0,0x80a28080,0x22a8020a,0x88aa0088,0x88028a20,0xa08a020,0x8a822828,0xa2a,0x8820200,0x88208082,0x8820028,0x280a800,0xa0080a00,0x80a82800,0x22208880,0x282a020,0xa0200a0,0x2a02800,0x80082200,0x20208222,0x8082208);
remainders_pos[921].modulo_value=3684; remainders_pos[921].bit_mask=create_bit_mask(116,0x228800a2,0x8028028,0x80008000,0xa2082082,0x28228,0xa20a202,0xa00008a0,0x208820,0x2200088,0x2000082,0x20a00a00,0x208280,0x22082820,0x8208020,0x8a00a20a,0x20000020,0x20228a08,0x8800828a,0x22800080,0x8820200,0x8a088082,0x820a2080,0x28a20800,0xa000002,0x2002022,0x28a00228,0x8208a200,0x220a0882,0x28a28a20,0x8220228a,0x820a0882,0x28a00228,0x200a200,0x2822020,0x28a20000,0x82080082,0x82880822,0x8820808,0xa200202,0x880800,0x20228a08,0x20828a,0x2882000,0x20828228,0x82202008,0x200020a0,0x28200a08,0x2000280,0x22020800,0x20008800,0x282088,0x22802880,0x20228,0x8228a20a,0xa0820,0x800008,0x8a20a00a,0x28a2800,0xa08a20,0x200a,0x820a0002,0x8a28820,0x8808000a,0x22802882,0x20828000,0x2200082,0x2080880,0x28000800,0xa008280,0xa0800082,0x808820,0x88282082,0x822802,0x28208000,0xa28808a,0x2022002,0x8008a00,0x2082208,0x82022822,0x20020828,0x8a288,0x80882800,0x8a00800,0x8800a280,0x220a0822,0x28808828,0x8a28a28a,0x220a0880,0x8a20828,0x8800a280,0x80800802,0xa08,0x208a288,0x208a0820,0x20220a20,0x8082208,0x20002880,0x28200220,0xa28808a,0x80000082,0x8a00a20,0x8020820a,0x82820880,0x28208000,0xa00a080,0x20000800,0x20008808,0x82208002,0xa20000a0,0x8a08a00,0x88280008,0x208208a2,0x200028,0x80280200,0xa0022882,0x8);
remainders_neg[921].modulo_value=3684; remainders_neg[921].bit_mask=create_bit_mask(115,0xa2000822,0x820820,0x8080088,0x22802802,0x8820a20,0x8228228a,0x20880020,0x8a00028,0x8a288000,0x82800802,0x28208208,0x88280208,0xa28020a0,0xa00828,0x2082282,0xa08808a0,0x28a20200,0x80202,0xa2080800,0x28a08,0x200000a,0x2822800,0x20228008,0x8208808a,0x828828a2,0x20208a20,0xa002288,0xa2820002,0x20220228,0xa28a202,0x2820002,0x20208a20,0x8a082288,0x820a28a0,0x20228808,0xa00800a,0x20000a2,0x28000,0x8228828a,0x80000080,0x28a20200,0x88280202,0x82002880,0x28020a20,0xa28a080,0xa0882820,0x20a08200,0xa088208,0xa28a0080,0x28800008,0x8820a000,0xa2082000,0x8828a20,0xa202282,0x808200a0,0x8008800,0x2282082,0x82022080,0x8200228,0x8808a082,0x2820882,0x220028,0x8082,0xa2082002,0x28020808,0x8a28800a,0x82800000,0x20808008,0x82080208,0x20080802,0x8000028,0x20a00a,0x800a2082,0x20a00808,0x82200002,0x828a2882,0x800208,0x8a00a280,0x28a20a2,0x28828020,0x88002200,0x2080,0x208008,0x82208,0xa28200a2,0x20000020,0x2202202,0xa2820000,0x228020,0x82208,0x80082,0x8808200,0x8a002200,0xa00200a0,0x28a28228,0x8000a280,0xa0882000,0x20a08a28,0x82200002,0x880802,0x208228,0x8280282,0x20a0000,0x20a00808,0x82082008,0xa0880080,0x28800000,0xa28008a,0x22880820,0x200208,0x208080,0xa00a0022,0x8a08820,0x8208288,0x208a2002);
remainders_pos[922].modulo_value=3688; remainders_pos[922].bit_mask=create_bit_mask(116,0x2a822a0a,0xa02aa,0x828a0a20,0xa2282aa,0xaa88228,0x88082280,0xaa8880a2,0x2a080802,0x2a282aa0,0x88aaaaaa,0xa008a220,0xa0a8a88,0x28808aa2,0x8a80aa28,0x220aa8,0x282a2082,0xaa20080,0xa002020a,0xaaa008,0x2a222220,0x28a2282a,0x2288080a,0x2a08002,0x882a808,0x82a8022,0x822282aa,0x88aa0a02,0x2a228a08,0xa20a2a82,0x20228020,0x888a082,0x2282a0a0,0x28a00282,0x88a22a80,0x2a8a802,0x8222aa8a,0x820a8a8,0x20202822,0xaaa2222,0xaa08aa00,0xa0a2a2,0x282a8aa2,0x2aa28a20,0x80880022,0x8828200a,0xa0a2008a,0xaa088080,0x88a2208,0x20200000,0xa02a0028,0x8002a8a8,0xa8820a88,0xaa8228,0x20a82288,0x82800282,0xaaaa20a0,0x802002a0,0xa00a022,0x80080288,0x8aaaa0a,0x8002820a,0x882a0882,0x82aa0022,0xa0822a28,0x2a800222,0xa80a2a,0x80828,0x88a22000,0x220aa20,0x8a0a02,0x82822a2,0x2202a0,0xa28aa888,0xa2a82808,0x8a0a008a,0xaa20aa8a,0x88aaa000,0x28080888,0x2a082088,0xaa88822a,0x2a2a80a2,0xa88a2280,0x800a2802,0xa828882,0xa22200a,0x2880882,0xa8a08a08,0xa288a882,0xa0aa2220,0x82888280,0x2a820aa,0x2a822088,0x20a8020,0x20228880,0x288a28a0,0x8888a8a8,0xaaa0008,0x80800a20,0x8aa0a0,0x8a82802,0xa0880082,0xaa02a22a,0xa2022828,0xa2a0a08a,0x8a200a22,0xaaaa2208,0xa828a8aa,0x2020a80a,0x222aa80,0x8820228a,0x822aa002,0x8288a028,0xa0a282aa,0x80a00008,0xa882a8aa,0xa0);
remainders_neg[922].modulo_value=3688; remainders_neg[922].bit_mask=create_bit_mask(116,0xa20aa282,0x88828a22,0xa0282a8,0x82aa0a22,0x82200aa0,0x80aa08,0x2200082a,0xa280808a,0xa2a0a228,0x222222,0x28802aa8,0x82820200,0xa008022a,0x20822a0,0x88aa8220,0xa0a2a80a,0x822a8808,0x288a8a82,0x88222880,0xa2aaaaa8,0xa02aa0a2,0xaa008082,0x8a28088a,0x800a2080,0x80a208aa,0xaaa0a22,0x22828a,0xa2aa0280,0x2a82a20a,0xa8aa08a8,0x8000280a,0xaa0a2828,0xa0288a0a,0x2aa208,0x8a20208a,0xaaa2202,0x80a82020,0xa8a8a0aa,0x8222aaaa,0x22802288,0x88282a2a,0xa0a2022a,0xa22a02a8,0x800882a,0xa0a882,0x282a8802,0x22800808,0x8002aa80,0xa8a88888,0x28a288a0,0x88a2020,0x200a8200,0x88220aa0,0xa820aa00,0xa088a0a,0x2222a828,0x8a88a28,0x828828aa,0x8808a00,0x80222282,0x88a0a82,0xa2800a,0xa2288aa,0x280aa2a0,0xa2088aaa,0x882082a2,0x888080a0,0x2aa888,0x8aa822a8,0x8802828a,0x80a0aa2a,0x88aa8a28,0x2a022000,0x2a20a080,0x2828802,0x22a82202,0x222888,0xa0808000,0xa280a800,0x22000aa2,0xa2a2082a,0x2002aa08,0x882a08a,0x820a000a,0x82aaa882,0x8a00800a,0x20280280,0x2a00200a,0x2822aaa8,0xa000a08,0x8a20a822,0xa20aa800,0x8a8208a8,0xa8aa0008,0xa002a028,0x2020,0x82228880,0x80882a8,0x88022828,0x8020a08a,0x2000880a,0x228a2aa2,0x2a8aa0a0,0x2a282802,0x2a882aa,0x2222aa80,0x20a02022,0xa8a82082,0x8aaa2208,0xa8aa02,0xaa2288a,0xa0028a0,0x282a0a22,0x8288880,0x200a2022,0x28);
remainders_pos[923].modulo_value=3692; remainders_pos[923].bit_mask=create_bit_mask(116,0x22a88202,0x8a020a,0x202a2000,0x82808022,0xa8a02a82,0x2a820880,0x8a888222,0x22a02,0x2a208a28,0x2a222282,0x2a288282,0x8aa2282,0x2082aa00,0x2aa88820,0xa28aa02,0x80088a08,0xa282aaa8,0xa2280282,0x80a00880,0x2a8082,0x8a20820a,0x208220,0xaaa22020,0x2a82a80a,0xa0a8082a,0x228088a8,0x2a0002a0,0x808a8a22,0x2a0aa20,0x2a088028,0x80880202,0x8a808aa,0x8202222,0xa0a0,0xa00a,0xaaa8a088,0x2288a08,0xa8028a28,0x2a022a00,0x882a02a8,0x80a0a20,0xa2200008,0x82828202,0x8a0020,0x28a8a020,0x2aa8002,0x80020880,0x80880a20,0x808a0088,0x2aa08208,0xa808a,0x808a20,0x202a028a,0xa0a0028,0xa088aa28,0x2a8a2880,0x8820a82,0x2a0a8820,0x8220082,0x8a2a882a,0xa220a022,0xa0a028a,0xa808280,0x20200a28,0x2a00008a,0x20aa8a20,0xa202208,0x2202220,0x800208a,0x2aa80022,0xa2a28800,0xa200080,0x28282080,0x88a808,0xa020200,0xa82208a,0x880a82a8,0x2802a00a,0x2288028a,0xa2aaa20a,0xa0000220,0xa0000a00,0x808200a0,0x2a20888,0x2202aa2,0x220a8808,0xa0a80280,0x2a20208a,0xa888a,0x202880a8,0x2a0a2a2,0xa82a8a82,0x88aaaa02,0x20800080,0x208a2088,0x2a800a08,0xa02820,0x288a022,0xa828a828,0x220022aa,0xa28a020a,0x22aa880a,0xa8208082,0x8aa2000a,0x228a8828,0x888a8828,0x208a8828,0x8800028a,0x222a280a,0x82a8888,0x80a2a022,0x2028282a,0x8a808880,0xa200000,0x22a88a08,0x808);
remainders_neg[923].modulo_value=3692; remainders_neg[923].bit_mask=create_bit_mask(116,0xaa200a8a,0x88028a02,0xa0a2a808,0xa0808aa,0x2028220a,0xa20a8008,0x2000aa2,0x880aa28a,0xa2a802a0,0xa2aaa20a,0x22a00a0a,0x8022aa02,0xa8022288,0xa22000a8,0x82a0220a,0x800280,0x2a0a2220,0x2aa00a0a,0x8288008,0x8a20802,0x2280a82,0x88a80aa8,0x222aa0a8,0x220a2082,0x282080a2,0xaa000020,0xa2888a28,0x802022a,0x822822a8,0xa28008a0,0x8000a8a,0x80208022,0x80a8aaa2,0x88082828,0x88882882,0x22202000,0xaa00280,0x200a02a0,0xa282a288,0xa28a20,0x80828228,0x22a88880,0xa0a0a8a,0x880208a8,0xa02028a8,0x8a220882,0x8028008,0x80082a8,0x8028000,0x22280a80,0x88820802,0x880002a8,0xa8a28a02,0x82828820,0x200022a0,0xa2022008,0x800a020a,0xa28200a8,0x80aa8802,0x22200a2,0x2aa828aa,0x82828202,0x2080a08,0xa8a882a0,0xa2808002,0xa82202a8,0x82a8aa00,0x82a8aaa8,0x8088a802,0xa22008aa,0x2a2a0088,0x82a88800,0xa020a808,0x88002000,0x828a8288,0x20aa802,0x820a20,0xa0822882,0xaa008a02,0x2a222a02,0x20888aa8,0x28888288,0x80a0820,0x8a2a8000,0x8aa8a222,0xaa020080,0x28208a08,0xa2a8a002,0x8820002,0xa8a00820,0xa202a2a,0x20a2020a,0x22220a,0xa0088808,0xa802a800,0xa2080280,0x8828a0a8,0x8a0028a2,0x202020a0,0xa288aa22,0x2a028282,0x2a220082,0x20a8080a,0x2228882,0xaa0200a0,0x20020,0xa00200a0,0x888a02,0xaa222082,0x80a20000,0x82a28a2,0xa820a0a2,0x2080008,0x82a88088,0x2a200280,0x80);
remainders_pos[924].modulo_value=1848; remainders_pos[924].bit_mask=create_bit_mask(58,0xa2802022,0x20000820,0x2088282,0xa0080080,0x8a20008,0x82080280,0x808800a0,0x20208,0x2002002,0x20802,0x20000028,0x200a20a,0x82080800,0x28000020,0x80000000,0x20800002,0x20208,0x8a08a008,0x2800000,0x8020800,0x8008008,0x22080800,0x20a00220,0x280208,0xa20008a2,0x200200,0x88282208,0xa0820000,0x28808028,0xa280202,0x22000082,0x208828,0x8a008008,0x8a2000,0x8208028,0x8808800a,0x20002020,0x20200200,0x80002080,0xa2000002,0x200a20,0x8208080,0x2800002,0x28000000,0x82080000,0x80002020,0x8a08a00,0x280000,0x80802080,0x800800,0x2208080,0x820a0022,0x20028020,0xa20008a,0x80020020,0x8828220,0x8a082000,0x880802);
remainders_neg[924].modulo_value=1848; remainders_neg[924].bit_mask=create_bit_mask(58,0x22082022,0x28008020,0x8200020a,0x20000800,0x220800,0x2000208,0x20,0x8828a00,0xa00a082,0x8a0082,0x28008820,0x8a002282,0x2000080,0x20808028,0x8088000,0x20080882,0x828a00,0x2002080,0x82000800,0x820008,0x80080,0xa2800000,0x28208220,0x80208280,0x22080022,0x8a08a00,0x202280,0x200a0080,0x20000820,0x2208202,0xa2800802,0x8200020,0x2000080,0x80022080,0x200020,0x2,0x208828a0,0x20a00a08,0x8a008,0x22800882,0x8a00228,0x80200008,0x2080802,0x20808800,0x2008088,0x828a0,0x200208,0x88200080,0x82000,0x8008,0xa280000,0x2820822,0x28020828,0x2208002,0x8a08a0,0x20228,0x200a008,0x82);
remainders_pos[925].modulo_value=185; remainders_pos[925].bit_mask=create_bit_mask(6,0x1c211a9a,0xc423d256,0x5a6289,0x8e451968,0xe1a92f10,0x1656210);
remainders_neg[925].modulo_value=740; remainders_neg[925].bit_mask=create_bit_mask(24,0x80a80202,0x82a0a82,0x882a200,0x28000a0,0x2020a208,0xa8a82a88,0x20208aa0,0xa008228,0x8802800,0xa080022a,0xa80a0a82,0xa2220200,0xa008,0x2a28802,0xa0288820,0x208a2088,0xa0a20,0x80008,0x82022828,0x2888228,0xa002088a,0x2008a2,0x22880280,0x2);
remainders_pos[926].modulo_value=3704; remainders_pos[926].bit_mask=create_bit_mask(116,0x28222a2,0x8a22282a,0x2aa2028a,0xaaaa0022,0x22802008,0x8288822a,0xa08a80a0,0xa202082,0x28002288,0xaa202a82,0x20a2a0a2,0x82208,0x2a20a088,0xaa08a22a,0x202a228a,0xa2aa00a2,0x800a0aa2,0x80aa8088,0x2a0a2a0a,0xa8228aa2,0x802a8882,0x280a8aa0,0xa08200a,0xa2a80028,0x8a882a,0xaa88a222,0x2028aa22,0xa282aaa0,0x2a2aa828,0x8280020,0x2a000282,0x220082a,0xa2220880,0x28808a,0xa82aa802,0x20aa28a0,0x8a0080a8,0x288802a,0x22008228,0x80a020a0,0x888a800a,0x200a0aa,0x220aa002,0x82002a,0x202208a0,0xa88a0a20,0x28a228aa,0x20a020a,0x22802a20,0xa8822aa8,0x28a2a20,0x8a0a808a,0x20228882,0x8aa2a82,0x222aa000,0x8082a200,0xa0282220,0x82022282,0xa828880,0x2088828,0x88008a82,0x20000aa8,0x882a8aa,0xa2822288,0x80a202a0,0x2a08a8a2,0x882aa882,0x8008a802,0x28a080a0,0x2aaa288a,0x808a0a2,0xa2088,0x88a80082,0x80800aa0,0x22aa0a00,0x2a002a2,0x880a080a,0x80288200,0xa2a80222,0x82a0200,0x2a0a28aa,0x802aa8,0x8aa20228,0x80022088,0xa8a82008,0x20828000,0xa8080028,0x8a282aa8,0x80aaa82,0xaa88aa28,0x888a22,0x8aa822a2,0xa028002a,0x28a00820,0x20aa2a0,0x2a8222a8,0xa88aa288,0x2a0a8a0,0x22202aa,0x8a8aa0a0,0x88a00aa,0xaaa288a8,0xa8a88a20,0x220a08,0x8208820,0xaa8a0a8a,0x2882a808,0xa0228800,0xa820808,0x820a02a2,0x88a88222,0xaa200802,0xa88800aa,0xa2a2808a,0x80a82888,0x8a8882);
remainders_neg[926].modulo_value=3704; remainders_neg[926].bit_mask=create_bit_mask(116,0x8a0aaa2a,0x2aaa0a2,0xa22a8a02,0x222288aa,0xaa08a880,0xa000aa2,0x28020828,0x82a8a80a,0xa088aa00,0x22a8a20a,0xa82a282a,0x8880aa80,0xa2a82800,0x22802aa2,0xa8a22a02,0x2a22882a,0x882822a,0x8220800,0xa282a282,0x20aa022a,0x8a2000a,0xa0820228,0x8280a882,0x2a2088a0,0x880200a2,0x22002aaa,0xa8a022aa,0x2a0a2228,0xa2a220a0,0x80a088a8,0xa2888a0a,0x8aa880a2,0x2aaa8008,0x88a00802,0x20a2208a,0xa822a028,0x2880820,0x8a0008a2,0xaa880aa0,0x828a828,0x20882,0x8a882822,0xaa82288a,0x880a88a2,0xa8aa8028,0x200282a8,0xa02aa022,0x8a828a82,0xaa08a2a8,0x200aa220,0x8a02a2a8,0x2820802,0xa8aa000a,0x8022a20a,0xaaa22888,0x80a2a88,0x28a0aaa8,0xa8aaa0a,0x820a0008,0x8a8000a0,0x88020a,0xa8888220,0x800a2022,0x2a0aaa00,0x82a8a28,0xa280202a,0xa2200a,0x880208a,0xa0280828,0xa222a002,0x8080282a,0x8882a800,0x20800a,0x8088228,0xaa228288,0x8a288a2a,0x828082,0x8a00a88,0x2a208aaa,0x80a28a88,0xa282a022,0x8808a220,0x22a8aa0,0x88aa800,0x2020a880,0xa80a0888,0x208088a0,0x2a0a220,0x8082220a,0x220022a0,0x880002aa,0x220aa2a,0x28a088a2,0xa02880a8,0x8a822a28,0xa20aaa20,0x20022a00,0x8a282028,0x8aaa8a22,0x2022828,0x80028822,0x222a0020,0x202002a8,0x88aa8280,0x80a800a8,0x22028202,0xa00a2080,0x28aa0088,0x820a8080,0xa828a2a,0x200aaa,0x22a8808a,0x20008822,0x2a2a0802,0x820a000,0x2000a);
remainders_pos[927].modulo_value=1236; remainders_pos[927].bit_mask=create_bit_mask(39,0xa2022802,0x20028a08,0x80000088,0x82080002,0xa08220,0x8a000202,0x20080080,0x8000028,0x808a288,0xa2080082,0x220820,0x8a282200,0xa20a0802,0x28a00220,0x8200a088,0xa2082080,0x28a28020,0x288202,0x880020,0x20028028,0x80008002,0xa2880822,0x8a08028,0x8202082,0xa28220a0,0x8a08800,0x8a28020a,0x80000882,0x8a08208,0x2082002,0x20228a2,0x828000,0xa202002,0xa0080800,0x8208820,0x280002,0x822000,0x8a20a28,0x80288);
remainders_neg[927].modulo_value=1236; remainders_neg[927].bit_mask=create_bit_mask(39,0x228a2082,0x28820200,0x8088000,0x2800802,0x8200a28,0x208828a,0xa0800800,0x808820,0x80002200,0x22800802,0x8a28028,0x220a288,0x22820082,0x20208a28,0xa082000,0x22802800,0x20220820,0x8820028a,0x800008a0,0x28820820,0x808008a,0x220000a2,0x200820,0x8028a00a,0x220a2820,0x200008,0x2208282,0x880002,0x200a00,0x8a00a08a,0x828a2022,0x8020808,0x8228a08a,0x20800080,0xa00028,0x8820808a,0x800a2880,0x228220,0x8200);
remainders_pos[928].modulo_value=232; remainders_pos[928].bit_mask=create_bit_mask(8,0xaa80a8a,0xa2020822,0x2228a2,0x20080a8,0x2888002a,0x20808a8a,0xa02aa088,0xa2);
remainders_neg[928].modulo_value=232; remainders_neg[928].bit_mask=create_bit_mask(8,0x82208202,0x2a8a80aa,0x882aa02a,0x8a880820,0xa00088a2,0xa8080202,0x28a22000,0x2a);
remainders_pos[929].modulo_value=929; remainders_pos[929].bit_mask=create_bit_mask(30,0x22fd0f36,0x24aef5f1,0xff1b5581,0x2e9a46fc,0x3133420b,0x77572bc5,0xba16d75a,0xa47469ce,0x920c02e7,0x8503250f,0xac477093,0x3d15931f,0xf1bd9144,0x67e4219e,0xbc6b58f4,0xe6109f98,0x8a26f63d,0xe326a2f0,0x243b88d7,0xc2930287,0x9d00c127,0xce58b897,0x6bada175,0x8f53abb9,0x410b3232,0xfd8965d3,0x6ab63fc,0x3ebdd492,0xb3c2fd12,0x1);
remainders_neg[929].modulo_value=3716; remainders_neg[929].bit_mask=create_bit_mask(116,0xaa2082aa,0xa8222828,0x22828808,0xa2028a20,0xa8aa8a82,0xaa8aa208,0x228a0a82,0x28a8a002,0xa808a2a,0x88aa882,0x208aa80a,0xa0880a82,0x28200888,0xaa28a802,0x20a28028,0x2a880200,0x2aa2aa0,0x2aaa2a28,0xa8a2000a,0xa0a8a0a,0x8808aa,0x280200a,0xa22028a8,0x28a2220,0x8882aaaa,0x2000a80a,0x8a22aa20,0xa220080a,0x2a0a208a,0x882282a0,0x8080222a,0x22aa228a,0x80a80020,0xaaaa088a,0x22228a02,0xa8a02228,0x80200a00,0xa8808802,0x828a8282,0x800228aa,0xa2a2aaa2,0x2aa2aa00,0x2008aa0,0xa00a2820,0xa8a2a8,0x888020a2,0xa808828,0x80aa8822,0x8aa8882,0xa2880a82,0x28a8a2,0xa828a22,0x822a8aaa,0xa8aa8a8,0x228a022a,0x808a0a20,0xa0a220a8,0xaa0822a8,0x228a002,0xa220a08,0x20a08a88,0x22a0822a,0x80080a0,0x802288,0x208280a0,0x8082aa2,0x80a88202,0x888008a2,0x28800a82,0x288a80a2,0xa2a8888,0x2080aa2,0x2820aa08,0x8aa2aa,0xa002002a,0x2020a,0x822aa82,0x82808280,0xa88a8802,0xa0aa2a82,0x222a080a,0xa0822228,0x88a00002,0x2aa80a80,0x8220022a,0x222a8a80,0x2822880,0x8a20a022,0xa8aa2200,0x20022080,0xa80aaa2a,0x28880,0x222082a0,0x82a220a,0xaa2a82a8,0x8a88aa0,0xa080a0a0,0xaaa20800,0x20200020,0x2002a8,0xa2a8802a,0x2a820a2a,0xa8082008,0x88aa2822,0x80a88a08,0xa8008a22,0x880088a0,0x208a80a2,0xaa080820,0x80a08222,0xa2008002,0x80800808,0x2082a202,0xa882822a,0x28222808,0x28a2008);
remainders_pos[930].modulo_value=3720; remainders_pos[930].bit_mask=create_bit_mask(116,0x2,0x28228000,0xa082208,0x82820020,0x20208828,0x820020a,0xa0020820,0x20808220,0x8800000a,0x82082802,0x8020808,0x2208200,0x822080,0xa20008,0x200820a,0x8a2082,0x220000,0xa00a288,0x20800,0x200a00,0x8080002,0xa20000a2,0x28028020,0xa00a008,0x20002822,0x28820200,0xa280200,0x822020,0x8808800,0x80088002,0x20020880,0x20200020,0x2008008,0x22822,0x20008020,0x2008,0x20a0002,0x28808020,0x80202082,0xa0002822,0x20008,0x8a000280,0x20000a0,0x828020,0x82200200,0x80000880,0x20020220,0x8002088,0x80002080,0x20a08800,0x80200,0x20020002,0x20820028,0x8002000,0x22800000,0x20200008,0x8a000000,0x228a2080,0x8a28800,0xa202,0x80800,0x28820,0x8002000,0x820800,0x800828,0x80200080,0x220a0800,0x8000200,0x8002002,0x80800822,0x20000208,0x80088202,0x2820000,0x8008,0x8000a20a,0x800002,0x28000a20,0x8080288,0x2022882,0xa08008,0x8000080,0x2000820,0x28800008,0x2008088,0x80820,0x20800808,0x2200202,0x22022080,0x8820000,0x8028a008,0x80822800,0x28000800,0xa00a088,0x2802820,0x8a08,0x20208a,0xa0080080,0x20800000,0x8a00a000,0x880022,0x8a20000,0x82008082,0x8a00a0,0x8820020,0x82088002,0x20802000,0x28208220,0x2280,0x820208a0,0x20800a08,0x80082008,0x220808a0,0x828228,0x8820a008,0x2882820,0x80);
remainders_neg[930].modulo_value=3720; remainders_neg[930].bit_mask=create_bit_mask(115,0x880882,0x20220800,0x200a280,0x20208a0,0x20a00020,0x80280282,0x200a00a0,0x28000228,0x8082,0x2802002,0x820000,0x8a280280,0x22000,0x8228800,0x82080202,0x80022802,0x8220800,0x2082200,0x80820080,0xa08208,0x8000008a,0x22080822,0x20820028,0x82002080,0xa0882022,0x20020a08,0x82208280,0xa20a0,0xa,0xa08a0000,0x28208820,0xa080080,0x808220a2,0x20800828,0x80082080,0x82020882,0x20000028,0x820a00a,0x20082022,0x8820800,0x2088200,0x2880020,0x8020828,0x2288208,0x880000,0x28028a20,0x82000,0x802800,0x20200008,0x88000288,0xa00a0082,0x28028020,0x8000a088,0xa2080800,0x28a00800,0x2088080,0x22022000,0x220008,0x80082202,0x80800080,0x8020020,0x8a088,0x80020080,0x8020,0x8280008,0xa2020080,0x808208,0x8000a08a,0x80022,0x28000a00,0x8000282,0x20a0080,0x8800800,0x2202,0x80080882,0x20008220,0x8200,0x82822002,0x200800,0x80080008,0x820800a0,0x20008000,0x8a000000,0x80800020,0x28000000,0x8a288202,0x228a2000,0x28808,0x202000,0xa2080,0x20008000,0x2082000,0x20020a0,0x800200,0x88282002,0x20000800,0x20008008,0x2002088,0x80000822,0x220808,0xa080002,0x20020,0x28828,0x200000a,0xa0002880,0x20200a20,0x808a208,0x2820020,0x20000200,0x8002080,0xa2000020,0x8020220,0x202080,0x82002020,0x8800808);
remainders_pos[931].modulo_value=532; remainders_pos[931].bit_mask=create_bit_mask(17,0x8a02822a,0x28082000,0x20228a88,0xa20008a0,0x2aa08280,0x8a0280,0x88280220,0xa008222a,0xa00000,0x822a8882,0x20008802,0xaa80280a,0x8a02820,0x8080a200,0x2822a28,0xa200082,0xa8828);
remainders_neg[931].modulo_value=532; remainders_neg[931].bit_mask=create_bit_mask(17,0x2820a22,0x2080a880,0x20aa0200,0x2a088028,0xa2200a08,0x88020208,0xa082a8,0x2800aa22,0x8288880,0x2a2000a,0xa8080082,0x2200a082,0x802820a8,0x8082288,0x820aa220,0x2a88802,0x200a0);
remainders_pos[932].modulo_value=1864; remainders_pos[932].bit_mask=create_bit_mask(59,0xa288a282,0x808a0022,0xa220080,0xa222a20,0xa00020aa,0x8a2aa82,0xa80aa28,0xa8228a2,0x2880a2a8,0x2a0aaa,0x2282a820,0x228022a2,0x88080802,0x8a2a2002,0x8a0a0a28,0x8a8a22,0x2020228,0xa8802888,0x82a82888,0xaa0a8000,0xa8a0228a,0xa2882a02,0x8aa02a08,0x2aa8a202,0xa08000a8,0x8a888a0a,0x20088a00,0x800a2020,0x28a228a8,0x288a2828,0x8a0022a,0xa2200808,0xa222a200,0x20aa0,0x8a2aa82a,0xa80aa280,0xa8228a20,0x880a2a80,0x2a0aaa2,0x282a8200,0x28022a22,0x80808022,0xa2a20028,0xa0a0a288,0x8a8a228,0x20202280,0x88028880,0x2a82888a,0xa0a80008,0x8a0228aa,0x2882a02a,0xaa02a08a,0xaa8a2028,0x8000a82,0xa888a0aa,0x88a008,0xa20202,0x8a228a88,0x82);
remainders_neg[932].modulo_value=1864; remainders_neg[932].bit_mask=create_bit_mask(59,0x2a002a0a,0x80288aa,0x82aa8808,0x82aaa2a8,0x2888a822,0x802a220a,0x820822a0,0x820aa02a,0xa0082a20,0x88a28222,0xaa0a20a8,0xaa08aa2a,0x80808a,0x2a2a88a,0x28282a0,0x880202aa,0x8a8a8aa0,0x2008a000,0xa20a000,0x22820888,0x2028aa02,0x2200a28a,0x228a280,0xa2202a8a,0x28088820,0x2000282,0xa8800288,0x882a8a8,0xa02aa020,0xa002a0a0,0x80288aa2,0x2aa88080,0x2aaa2a88,0x888a8228,0x2a220a2,0x20822a08,0x20aa02a8,0x82a208,0x8a28222a,0xa0a20a88,0xa08aa2aa,0x80808aa,0x2a2a88a0,0x28282a00,0x80202aa0,0xa8a8aa08,0x8a0008,0xa20a0002,0x28208880,0x28aa022,0x200a28a2,0x228a2802,0x2202a8a0,0x8088820a,0x20002822,0x88002880,0x882a8a8a,0x2aa0200,0xa);
remainders_pos[933].modulo_value=3732; remainders_pos[933].bit_mask=create_bit_mask(117,0x22822882,0x8020200,0x8800a288,0x82802020,0x28208808,0x8a00828a,0xa0800802,0x28008a28,0x82282282,0x82020822,0x820820,0x2008200,0xa2000080,0x28800200,0x20020a,0xa00a2020,0x8a20808,0x8a202200,0x2800,0x20820200,0xa28220a,0x82002820,0xa08008,0x8828000a,0x2022,0x8a20808,0x88008282,0xa20200a2,0x8200200,0x208202,0x2022802,0x8008a28,0x8228820a,0x20000080,0xa08a20,0x8280280,0xa0020880,0x20a28020,0x2082282,0xa0000002,0x20228800,0x82288002,0x22802880,0x200020,0x8a282,0x22802,0x8020008,0x82080002,0x82020820,0x20a20a20,0x8800a20a,0x800828a2,0x8028a00,0x8820a28a,0x200a0880,0x28008820,0x200808a,0xa20a2002,0x8a20a08,0x808008a,0x28a2020,0x808220,0x8220220a,0x20a28a0,0x200a28,0x228a282,0xa08a08a0,0x8208a08,0x8208208,0x2080002,0x20008,0x80288,0x800828a2,0x28808000,0x88202280,0x80880022,0xa00228,0x8080000,0xa0882882,0xa08028,0x82002208,0xa0002802,0x808a20,0x88202000,0x20a0822,0x8028a20,0x80080288,0x82080820,0x8a00800,0x228a008,0xa2082820,0x20208,0x82288080,0xa00a0002,0x8220020,0x8a008280,0x208a0882,0x808,0x8a200280,0xa2000880,0xa20208,0x8000808a,0x228a0800,0x8a00800,0x8002000,0x20000820,0x8208208,0x88288208,0x2882882,0x20a28a20,0xa280200,0x828a2820,0x28220220,0x2208080,0x20228a0,0x28800808,0x82288);
remainders_neg[933].modulo_value=3732; remainders_neg[933].bit_mask=create_bit_mask(117,0xa20a2002,0x828a08,0x82200,0x20828a0,0x20a00000,0x2080202,0x20080082,0x20800220,0xa20a20a,0x20a00a2,0x8028028,0x8a080288,0x22880800,0x20008a08,0x88288282,0x208228a0,0x228000,0x228a288,0x80882080,0x28028a08,0x8220a282,0x28820a0,0x8200800,0x208082,0x808828a2,0x228000,0x8020a,0x228a0822,0xa08a08,0x8828028a,0x828a2082,0x800220,0xa200282,0xa0880800,0x8200228,0x80208208,0x208a0000,0x28220828,0x8a00a20a,0x20880882,0x28a20008,0xa20008a,0xa2082000,0x8a08828,0x8800220a,0x808a2082,0x828800,0xa00808a,0x28200a0,0x28228228,0x82282,0x802022,0x820208,0x282202,0xa0820000,0x20800028,0x8a080002,0x22822882,0x228200,0x80008002,0x820228a0,0x8000a28,0xa28a282,0x82822020,0x8a08220,0x8a20220a,0x20020020,0xa00200,0x80280280,0x82800882,0x8828800,0x88008200,0x802022,0x20000808,0x28a208,0x8a2,0x8208a20,0x80008088,0x20002002,0x8200820,0xa08a280,0x20882082,0x8000228,0x28a088,0x828200a2,0x820228,0x8008200,0x28000a0,0x208008,0x8a202080,0x228020a0,0x8828a00,0xa200008,0x20820882,0xa28828,0x2080208,0xa0020002,0x8808000,0x2288208,0x22880000,0x8228a00,0x8080002,0xa2020080,0x208008,0x8008a088,0xa08800a0,0xa00a00,0x200280,0x82002002,0x28220228,0x82208288,0x20220a0,0x20a28a28,0x8a280008,0x828a2020,0x20008000,0xa200);
remainders_pos[934].modulo_value=3736; remainders_pos[934].bit_mask=create_bit_mask(117,0xaa02822a,0x280a2a20,0x2aaaa800,0x2a0280a2,0xa0a8aa2,0x82aa028a,0x822022,0xaaaa228a,0x882a8a0,0xa288aa,0x82a00,0x808a8a80,0x8820a80a,0x2aaaa202,0x228882,0xaa2222a8,0x820a8028,0x20200820,0xa2808888,0x2a0888,0x22800a2,0x2aa2220a,0x8aa808a0,0x28a2282a,0x2020a0a0,0x828a2aa,0x822222aa,0xa2a08288,0xa828a20a,0x8280222,0x8a0200a2,0x22228882,0x828a002,0xa2a2002,0x22082a0a,0xa8008028,0x22002a08,0xa822a0a2,0xa02aa20a,0x8a820888,0xaa2a2888,0x80a28a28,0x2220082a,0x802aa802,0x2288,0xa2882a2,0x808a80a8,0x28aaaa80,0x820aaaa0,0x288a008,0x20000a08,0x228aa082,0xa002822a,0x80a0a082,0x82a02200,0x220a,0x20a82aa8,0x82a00a20,0x82880022,0x808a00a,0x2aa82a,0x82a08800,0x200880a,0xa820a0a,0x88a88280,0x8820aa2,0x8020a000,0x82200a22,0x280aaaa0,0x202aaaa,0xa02a02a2,0x8a8228,0x2228800,0x88802aa8,0x2a82008,0xaa28a28a,0xa22228a8,0xa222082,0x2aa08aa8,0x888a0a88,0x2a20a800,0x88280200,0xa0a0a820,0x208008a8,0x88800a28,0xa2822288,0x208a0080,0x2a888028,0x8aa08a28,0x8222820a,0x20aa8888,0x8aa8a28,0x280a0a08,0xa2a8288a,0xa80a202a,0x80a0888a,0x8a0028,0x8a2220a8,0x8222202,0x82082008,0xaa2802a0,0x2a8888,0xa8822288,0x22808aaa,0x2a02a08,0x2a2a2,0xa820,0x20aa228a,0xaa0a2a82,0xa288aa,0x82880882,0xa0a280aa,0xa88aa2a0,0xa88a0280,0x28002aaa,0xaa08a8a0,0xa88280);
remainders_neg[934].modulo_value=3736; remainders_neg[934].bit_mask=create_bit_mask(117,0x228a0aa2,0xa082a2a8,0xa2222088,0xa28a082a,0x8282022a,0xa228a02,0x880aa8aa,0x2222aa02,0x800a2028,0x882a0022,0x8880a288,0x8020208,0xa82082,0xa2222a8a,0x88a2000a,0x22aaaa20,0xa8208a0,0xa8a880a8,0x2a080000,0x88a28000,0x8aa0882a,0xa22aaa82,0x2208028,0xa02aa0a2,0xa8a82828,0x80a02a22,0xaaaaa22,0x2a280a00,0x20a02a82,0x80a08aaa,0x28a882a,0xaaaa000a,0x80a0288a,0x82a2a88a,0xaa80a282,0x208808a0,0xaa88a280,0x20aa282a,0x28a22a82,0x20a8000,0x22a2a000,0x82a02a0,0xaaa880a2,0x8a2208a,0x8888aa00,0x82a00a2a,0x8020820,0xa0222208,0xa822228,0x8a002880,0xa8888280,0xaa02280a,0x288a0aa2,0x828280a,0xa28aa88,0x8888aa82,0xa820a220,0xa2882a8,0xa0088aa,0x80802882,0x88a220a2,0xa280088,0x8a880082,0x820a8282,0x200a08,0x800a822a,0x8a82888,0xaa882aa,0xa0822228,0x8a8a2222,0x28a28a2a,0x88020aa0,0xaaa0088,0x8a220,0x8a20a880,0x22a02a02,0x2aaaa020,0x82aaa80a,0xa2280220,0x28200,0xa2a82088,0xa08a88,0x282820a8,0xa8088020,0x882a0,0x2a0aaa00,0xa8028808,0xa20008a0,0x22802a0,0xaaa0a82,0xa8220000,0x802202a0,0xa0828280,0x2a20a002,0x2082a8a2,0x8280002,0x880288a0,0x2aaa820,0x80aaaa8a,0xa80a880,0x22a08a28,0x88a20000,0x200aaa00,0xaa080222,0x8a28a280,0x888a2a2a,0x888820a8,0xa822aa02,0x2282a20a,0x882a0022,0xa00800a,0x282a0822,0x20022a28,0x20028a08,0xa088a222,0x22802028,0x200a08);
remainders_pos[935].modulo_value=3740; remainders_pos[935].bit_mask=create_bit_mask(117,0xa0082282,0xa0028220,0xa0200a8,0x222882,0x80800022,0x20a80,0x8a000880,0x282a028,0x88808028,0x22808020,0x8802200,0xaa220280,0x8082220,0x8a0008aa,0xa88200,0x208a0a0a,0xa2220080,0x208020a8,0x8a02a808,0xa200aa28,0x282080,0x88080080,0x22828000,0x802a088,0x2a80808,0xa8280282,0x80000a80,0x20880280,0x8a80808a,0x2002208,0xa020000a,0x2a2a0002,0x20800200,0x8002000a,0x220a220,0x8000208,0xa0020aa2,0x200a2228,0x802808,0x2802000,0x88282200,0x8000020,0x208a028,0x28080,0xa2802202,0xa0220000,0x88000800,0x2802a000,0xa002088,0x2220080a,0x808208a,0x8a000020,0xa82a220,0xa022802,0x8008882a,0x88820200,0x2280020,0x2880088,0x20800,0x2200228,0x20800288,0x20080822,0xa822200,0x802880a,0x2088a82a,0x80000a,0x8a208202,0xa020088,0x8220800a,0x2000a802,0xa0020002,0xa0000880,0x8088028,0x202800,0x280a208,0x20800002,0x88282,0x8028,0x82028020,0xa2888a00,0x8aa0800,0x2080002,0x2088a088,0x8a000800,0x80080020,0xa8000a8a,0xa000080,0x22088008,0x8a20202a,0x20280220,0xa02a0000,0x8280282,0x20202a8,0x8220a802,0x20002828,0x200202,0xa0208280,0x228aa008,0x8202a80a,0xa2a08020,0x80200888,0xa0a0a20,0x200822a0,0xaa2000a,0xa0888202,0x28088a,0x80088022,0x20802028,0x2802022,0x2280a828,0x22000a,0x202a0800,0x8800020,0x8228880,0xa2a0080a,0xa0882800,0x8288200);
remainders_neg[935].modulo_value=3740; remainders_neg[935].bit_mask=create_bit_mask(117,0x2880220a,0x28020aa0,0x28a8020,0x8822a002,0x8082a,0x808a0208,0x2088000,0xa0a2020,0x8a0,0x22080828,0x80082a80,0x222a8200,0xa228,0x2808022,0x80200200,0xa8020282,0x222a0800,0x2808a020,0x2822000,0x22802220,0x8202808,0x8800,0x220a0088,0x80022800,0x2208000,0x20a00a02,0x8080200,0x20008200,0x2000802,0x8288aa00,0x28a00882,0x22228082,0x28088288,0x8820082,0x82282a28,0x80800a80,0x280a8222,0x280222a0,0x8000a080,0x8208a808,0xa02a08,0x88a0,0xa8020a0,0x88820808,0x2208a202,0x282a0888,0x88000,0x208a2080,0x82802000,0xa2a88002,0x80802802,0x2088820,0x20222a8,0x282a08a,0x22,0xa0288,0x822088a0,0xa008000,0x8828080,0x82a80a20,0xa0080a00,0xa80080a2,0x20aa208,0x80820002,0x20002022,0x88080082,0x2280282,0x20a0000,0xaa00882,0xa0802002,0x288a088a,0x20088000,0x8000a0,0x80a0a008,0x82002a00,0x2808088a,0x88000a02,0x88800a0,0xa0208a0,0x22000208,0x80220000,0x8a000882,0x20002000,0x2808088,0x800828,0x20800202,0x2008800,0x2a800080,0x220a0a2,0xa0208a28,0x20a20888,0x80208a02,0xa8a0220,0x2a0208a,0xa088a020,0x88a80a02,0x28200a00,0x2a022000,0xa822082,0x22280028,0x8280000,0x820202a0,0x2880a220,0x82220882,0x20000a0a,0x88200002,0x8000822,0x2000a0a0,0xa00a8aa,0x22082020,0x88aa0082,0xa8228080,0x880a0,0x80a20000,0x22280002,0x20002088,0x200280);
remainders_pos[936].modulo_value=312; remainders_pos[936].bit_mask=create_bit_mask(10,0x28a0822,0x8820020,0x80288008,0xa20a2080,0x200000,0x8a000008,0x20208a0,0x20200228,0x80080082,0x8820a2);
remainders_neg[936].modulo_value=312; remainders_neg[936].bit_mask=create_bit_mask(10,0x820200a2,0x28828,0x200080,0x22822800,0x8a00808,0x2088080,0x828a0020,0x28208a20,0x800800a,0x2022);
remainders_pos[937].modulo_value=937; remainders_pos[937].bit_mask=create_bit_mask(30,0xaf0f335e,0x66ff00fd,0x22827f51,0x163cdf57,0x97dd1309,0x48c4a26,0x79dd3117,0x219e8dda,0x81a588e1,0x4b9d71fb,0xba46a63e,0x8a104ada,0xad090b35,0x1f41d351,0x6a7959ee,0x2e0be1de,0x4242d62b,0x482146b3,0x9589756d,0x3ae749f1,0x4696077e,0xc5e6121c,0x32ee796e,0x48c483a2,0x22efa591,0xecf1a243,0xf90513ab,0x3fd9a2b,0x33c3d6fc,0x1eb);
remainders_neg[937].modulo_value=3748; remainders_neg[937].bit_mask=create_bit_mask(118,0x2282aa82,0xaa228820,0xaa0aa288,0x8aa0028a,0xa008a80,0x880082aa,0xa000a88a,0xa8020002,0x8280028,0x8200a822,0x228a2aa2,0x2888202,0x208082a8,0x82880a88,0xa2a08022,0xa2822802,0x8a8a0aa2,0x80a88a2a,0x800a8a0,0xa22a8028,0x8a0a8aa2,0x82a8a80,0xaa22a0a2,0x80080a2a,0xaa222808,0x20282a8a,0x20888a22,0x8a2002a2,0xaa0a0a20,0x82aa2022,0x22882282,0x88222a00,0xa0222288,0x22aa8aa0,0x800880a2,0x22aaa280,0xa082282a,0x82880a8a,0xa22a2a8a,0x80a00a,0xa80828a8,0x8a8aa288,0xa2a2a82,0x2a2a00a2,0x8a0a2008,0x8208a80,0x8a00aa08,0x8a220208,0x20a2aa2,0xa08220a8,0xa8a000,0x2a0002,0x8a88a8,0x282aa08,0x2a880a88,0x82aa8a00,0x22a88a2a,0xa20208a,0xa20aaa,0x8a2a20a0,0x20880220,0xa0800280,0x80a8802a,0xa00282a8,0x8808aa8,0xaaa02aa8,0xaa080aa2,0xa228a0a,0x20a2a8,0xa2a22082,0xa00aa088,0x808a28a8,0xaa20a08a,0xaa02028,0x802020a2,0x82008082,0x828a808,0xaa0a8aa8,0x80202200,0x80a88280,0x28228a00,0x82000220,0xa88aa8a,0x800222,0x82222a0a,0xa0222888,0x8228822a,0x22200282,0x20a0a002,0x2aa208a,0x20888a22,0x80282a22,0xa8222000,0x20a8aa88,0xa022000,0x808028a2,0x80a08a,0x2a802202,0x80aa8a8,0x20880a8a,0xa08080,0xa8228202,0x2a8a0202,0x80a88282,0x28a8a28,0xa28882a8,0x208222,0x280aa282,0x2aa828a2,0xaaa2a808,0x880aaa02,0x28aa880,0x808aa0a0,0x82aa008a,0x8200a000,0x28822008,0x8002822a,0x2);
remainders_pos[938].modulo_value=3752; remainders_pos[938].bit_mask=create_bit_mask(118,0x82888a02,0x8280a80,0xaa0082,0x8200a00a,0x20822088,0x2a82a20,0x28028022,0x2000280a,0x2888a880,0x2aa8022,0xaa280802,0x8aa2a282,0xa002808,0xa822a00a,0x2a0a2a0a,0x22a82202,0x208a0a08,0xa2220008,0x222282,0x82220a88,0x200022a8,0x2888a800,0x2882a00,0x20228aa8,0xaa00a8aa,0xa0828000,0x8a200a08,0x8280200,0x28002,0x2aa08282,0x20a28a82,0x200a80a8,0x8a00a08a,0x800a2a80,0x2820000a,0x82800002,0x828a020,0xa22a8808,0x2208002a,0x28a08082,0x2020220,0x8828a2a2,0x208a020,0x2020a220,0x20008020,0xa020a802,0x80a80a,0x20200,0xa2a088,0x88022a00,0x2a288800,0xa020882a,0x2222082,0x8a222008,0x82880a20,0x8a0a20a,0x28808a20,0x28880a0,0x2828088,0x2228022,0xa202280a,0x8a0a2008,0xa02282a0,0x888a208,0x8888020,0x22080a82,0x2228a8a8,0xa8802200,0xa8a0000,0x80800022,0x2a020000,0x2a080aa0,0x2000880,0x8a080808,0xa208008,0x8a282208,0x8080808a,0x20a2808,0x208882,0x22a88aa8,0xa282020,0x28208,0x82880,0xa8a002a0,0xa00a202,0x2a008a2,0xa28a082a,0x820aa882,0x2800082,0x80282080,0xa008a200,0x2820a20,0x2a00aa00,0xa28808aa,0xa822802a,0x2a222800,0x88000800,0xa088822a,0x88880022,0x888a82,0xa0a20820,0x882a8820,0xa8a0a880,0xa882aa0,0x2800a0a0,0x8a8aa220,0x2028aa82,0x2aa8080,0x2a222888,0x28000802,0x28028a0,0xa82a8088,0x8820808,0xa008222,0xaa00a0,0xa0282082,0xa2228202,0x80);
remainders_neg[938].modulo_value=3752; remainders_neg[938].bit_mask=create_bit_mask(118,0xa00020a,0xa08200,0x80228802,0xa082882,0xa802a800,0x8a2022a8,0xa08a00aa,0xa888a002,0x20002000,0x822208aa,0x2220008a,0x2222a0a,0x82882080,0x20aa2082,0xa282a202,0x2a20aa82,0xa0028280,0x2a2a8880,0x80a2aa0a,0xaaa0200,0xa888a220,0xa0002008,0xa00a280,0xa0aa0220,0x22082022,0x28020888,0x2a80280,0x80a08208,0x888a080a,0x22280a02,0xa02a020a,0xa8020820,0x2802802,0x8822208,0xa0a88082,0xa00880a,0xa028a0,0x22a20080,0xaa0088a2,0xa020080a,0x8a8a0aa8,0xa0222a,0x8a802828,0x28a82aa0,0xa08808a8,0x2828208a,0x88002082,0x888a0a88,0x882a2000,0x8aa208,0x22a00080,0x20a800a2,0x8a2aa00a,0x2a2a880,0xa0002a8,0x80282282,0xa0080228,0xa000820,0x820a0800,0x8a2a08aa,0x2a02a082,0x2822880,0x28aa0228,0x80002a00,0x8a0,0xa280820a,0xaa202020,0x2000aa88,0x82020888,0x80880a2,0xa28a8808,0x22808220,0x82888008,0x2008080,0x82a00880,0x2a02a80,0x8080002,0x8a822000,0x8a80002,0xa2200220,0x8220a8a8,0x88820a80,0x88802008,0x20288228,0x82882a0a,0x2288022,0x220280a2,0xa02200a,0x8a00880a,0x8a02808,0x28802288,0x8a0a8228,0x22882280,0x22008022,0x202a0822,0xa2a2a088,0x880088,0x280002a2,0x882a,0x8000202,0x202a80a8,0x2200a8,0x20202008,0x82002228,0xa0882028,0x2022a28,0x28a02202,0x82220808,0xa22aa000,0xa080808a,0xa082028,0x20a20000,0x800a8000,0x2880aa2,0x80228828,0x2820a80a,0x2aa20a8a,0x8);
remainders_pos[939].modulo_value=3756; remainders_pos[939].bit_mask=create_bit_mask(118,0x820228a2,0xa20a08,0xa28208a,0xa2800822,0x28208000,0x8208200,0xa2882000,0x220200,0x88288282,0x802022,0x20a28020,0x8008020a,0x82080002,0x20a00220,0x8220228a,0x20a0020,0x28a28000,0x800a080,0x228808a2,0x28828820,0x828000a,0x820a08a0,0x28800828,0x200a282,0x22082,0x200200,0x8a08a282,0x22880880,0x8820220,0x2080282,0x880882,0x220220,0x8a080002,0xa28a08a0,0x8820808,0x8000a,0x820020a2,0x20a00200,0xa280,0x882002,0x20800220,0x28200a,0x828a2800,0x8208a00,0x8000088,0x820808a0,0x28028a20,0x8008282,0x2882800,0x20008828,0x8a080200,0xa0880,0x20a28820,0x80288208,0xa2820,0x808228,0xa002080,0x82020080,0x8,0x8082200,0x800a0020,0x20200020,0x8a000288,0x22802082,0x28a08208,0xa000082,0x20a2022,0x20008008,0x82280282,0x20020002,0x28028828,0x208208a,0x200a2,0x20820220,0x8a28200a,0x82800002,0x208a00,0x82200088,0xa0000800,0xa08028,0x80082008,0x20008a0,0x8220a00,0xa28a202,0x20a20a2,0x8800800,0x2200088,0x2080822,0x8220828,0x2288088,0xa20a2022,0x800828,0x88000008,0xa0080820,0x228828,0xa082282,0x28200a2,0x28228a00,0x2288082,0xa00208a2,0x28a28020,0xa080000,0x80882080,0xa08a28,0x2082088,0x2002800,0x28a08a08,0x80200080,0x22822880,0x8800828,0x8228a008,0x20820000,0x20828008,0x28a000,0x828a0882,0x8a00a20,0x8808220a,0x8a2);
remainders_neg[939].modulo_value=3756; remainders_neg[939].bit_mask=create_bit_mask(118,0x28a2022,0x8228200,0x8220a002,0x220800a2,0x20a00808,0x80280288,0x22002880,0x8a28a08,0x20020a,0x800828a2,0x28220828,0x8008282,0x2800882,0x28208a28,0xa28a202,0x828208a0,0x20220808,0x80082008,0xa2000022,0x20020028,0x80208082,0x2820020,0x20008020,0x8a08220a,0x808a2802,0x8a08a08,0x200220a,0xa2000000,0x28a28,0x8a00820a,0x80000002,0x8a28a28,0x200808a,0x22020020,0x28000,0x88008082,0x2882822,0x28208a08,0x88082208,0x80002882,0x28008a28,0x8820a082,0x2022080,0xa00208,0x80088000,0x2800020,0x20820228,0x8008020a,0x82002080,0x28800020,0x2008288,0x80820000,0x28220028,0x8200280,0x808220a0,0x8000a20,0x8208a008,0x28a0800,0x8808800,0x8000a288,0x8208a0,0x28a08828,0x2088200,0xa2082802,0x20200a00,0x8208800a,0x828228a2,0x28800800,0xa20020a,0xa08a0882,0x20820020,0x8a00a002,0x808a0822,0x28028a28,0x220a082,0x2080882,0x8a00208,0xa288000,0x20880080,0x8200820,0x800a080,0x82880020,0xa28208,0x8220228a,0x82822822,0x8008,0x8a288000,0x828000a2,0xa28020,0x8a200000,0x228228a2,0x8008020,0x88080,0x208000a0,0x8a20020,0x8200a20a,0x820a0822,0x20a20208,0x8a20000a,0x208a0022,0x20220828,0x82008088,0x2800,0x8200220,0x8a00a000,0x82882080,0x20200200,0x8208008,0xa20a2000,0x8020,0xa202080,0xa00a0880,0x28020800,0x88202088,0x2020002,0x208228,0xa282,0x22);
remainders_pos[940].modulo_value=1880; remainders_pos[940].bit_mask=create_bit_mask(59,0x88282a8a,0xa2020082,0x2a0a2200,0x22200a8,0xa88a20,0x20880a00,0xa0a0082,0x28a0280,0x8822882,0x22a88222,0x88a80000,0x80200a80,0x80a8080,0x8202280a,0x20,0x88082080,0x2a020a02,0x802020,0x2a2202a,0x22a88000,0x88220888,0xa280822,0xa0a0028,0x2208820,0x22a0000a,0x888008a,0x8a0a82a0,0x808a008,0x82822820,0x82a8aa2a,0x20082882,0xa2200a20,0x200a82a0,0x88a20022,0x80a0000a,0xa0082208,0xa02800a0,0x22882028,0x88222088,0x8000022a,0xa8088a,0xa8080802,0x2280a080,0x20820,0x82080000,0x20a02880,0x20202a0,0x2202a008,0x8800002a,0x2088822a,0x80822882,0xa00280a2,0x88200a0,0xa022,0x8008a22a,0xa82a0088,0x8a0088a0,0x22820080,0xa2a828);
remainders_neg[940].modulo_value=1880; remainders_neg[940].bit_mask=create_bit_mask(59,0xa02202,0x2a0a0802,0x2282a288,0x8aa28820,0x80200228,0xa8000288,0x82028802,0xa028208,0x8002a00a,0xa2200a2a,0x200888,0x8288200,0x820008,0xa82a002,0x80888828,0x802808,0xa20a8282,0x808a0a8,0x8a22a8a2,0xa2200800,0xaa0000,0x822080a2,0x28280a0,0x8aa000a8,0x22288802,0x80000802,0x2020a20,0x802080,0xa02a0a8,0x2202222,0xa080200a,0x2a2882a0,0x28820228,0x2288aa,0x288802,0x28802a80,0x28208820,0x2a00a0a0,0xa2a800,0x888a22,0x88200002,0x20008082,0x2a002008,0x888280a8,0x2808808,0xa8282008,0x8a0a8a20,0x2a8a2080,0x8008a2,0xa0000a22,0x80a200a,0x280a0822,0xa8028,0x888028aa,0x802222,0x20a20800,0x2080020,0x2a0a8008,0x2220a0);
remainders_pos[941].modulo_value=941; remainders_pos[941].bit_mask=create_bit_mask(30,0x67714272,0x436aeaf6,0x291ba2e1,0xe7d04b91,0x31738b74,0xfbbe563a,0xe79236de,0xa1e00e7f,0x3f10ca65,0x72d4ca5a,0xe6697899,0x8012b209,0xf2692e8b,0xbe21cf1,0xf52beaa0,0x11f40155,0x2593e3ce,0x5200745d,0xa599e413,0xcad3a647,0xc23f1694,0x1e16994,0x1279ff9c,0x9f77dedb,0x73a3171a,0x82f9cbb4,0x76252274,0xd5b0a1d1,0xa3b99bd5,0x1390);
remainders_neg[941].modulo_value=3764; remainders_neg[941].bit_mask=create_bit_mask(118,0xaaa88aaa,0x8aa2222a,0xa0822a28,0x2a088208,0xa8aa02a8,0x22228aa2,0x2a0aaa02,0x282882a2,0xa28802a8,0xaa080282,0x2aa0a000,0x88a2a80,0x2aa0a202,0x822a8028,0x28a22228,0x88288888,0xa80a2a02,0x8a88a880,0x2800288a,0x20a2a8a,0xaa28a08,0x8828a008,0x8aa02200,0x2aa0202,0xaa2a8a82,0xa200228,0xaaa8aaa8,0x8282808,0x2a200208,0xa0828a08,0x80820022,0xa880a0a0,0x80aaa8aa,0xa8a20022,0x20a8aa2,0x8a0202aa,0x8802202a,0x808028a0,0x828a2a,0xa28aa282,0x8a88a000,0xa808a88a,0x8a02a280,0xa08888a0,0x8a22228,0xa0a00aa2,0x8202282a,0xa00aa288,0xa800282a,0x2a0a0082,0xa0aa008a,0xa22a08a0,0x2202aa82,0xaa2a8a22,0xa0aa02a8,0x28820882,0x88a2a208,0xaaa2222a,0x2aa88aa,0x880,0x8802222,0x8828a202,0xa0288a2,0x200802a0,0xa2222080,0x22028002,0xa0a08a0,0xa0208aa0,0xa8028aa0,0xaa00282a,0xa2888200,0xa002822,0x8a200aa,0x8a082222,0xa08a0888,0xa80a8202,0x80808808,0x820aaa08,0x8aa28200,0x8a802082,0xa88a082a,0xa0802a22,0xa2a002a2,0x8020080,0xa822aa2,0x88000800,0x8a8a0a0a,0x8a022aa2,0x2028a002,0x28a8a2aa,0x8a828,0x20a80008,0x200822aa,0x2a28080,0x82a2a0,0x288aa22a,0xa8aa08,0xa2a8a082,0xa8208020,0x8080882a,0xa80a8808,0x2882a020,0x8288888,0x22882222,0x282a80,0x88a2a20a,0x2a8020,0xa00aaa0a,0x820282a8,0x282802a8,0xa0220288,0x2222a000,0x8020082,0xa02802a0,0x8a08a288,0x882022,0x2222,0x88);
remainders_pos[942].modulo_value=3768; remainders_pos[942].bit_mask=create_bit_mask(118,0x2080082,0x28828200,0x28808a,0x20808a0,0x8828228,0xa28a280,0xa0800022,0x28008028,0x88008280,0xa2800802,0xa00a00,0x220a200,0x208a20a2,0x28808820,0xa20a282,0x22822,0x20028a00,0xa28a20a,0x820808a2,0x8220820,0x8002080,0x820820,0x8020208,0x220a288,0x20000822,0xa08228,0x2088200,0x20022000,0xa08828,0x88282008,0x20080802,0x28a20228,0x8a208082,0x20020802,0x20208228,0x88000202,0xa08220a0,0x28208208,0x8a082288,0x82082020,0x20008820,0x200000,0xa28000a2,0x20200228,0x2000200,0x22080880,0x20020008,0x8a200202,0x2802802,0xa08a00,0xa008288,0xa08a0,0x20228a08,0xa000000,0xa0020080,0x28008820,0x8800a,0xa2820080,0x828820,0x8a082282,0x20082,0x28a00220,0xa082200,0xa0020080,0x8000000,0x20a288,0xa00a20a0,0x228200,0x8000a20a,0xa2802802,0x8808008,0x88200080,0x80022020,0x8008000,0x8a288008,0x8a0002,0x8000008,0x82082200,0xa2080820,0x28228820,0xa208208,0x220a0882,0x8808000,0x8288208,0xa2802080,0x28820208,0x828808a,0x22802020,0x200828,0x828220a,0x80000880,0x8220,0x828820a,0x80882000,0x20228a08,0x208080,0x20082082,0x20020800,0x82082088,0xa08a2020,0x8a08a28,0xa280,0xa0882880,0x28828a08,0x8082202,0x808a08a2,0x8a08,0x8a00a00a,0x22802002,0x28028200,0xa280200,0xa0880002,0x20028a28,0x80288282,0xa2020,0x28a20228,0x80008282,0x820020);
remainders_neg[942].modulo_value=3768; remainders_neg[942].bit_mask=create_bit_mask(118,0x82800802,0x20020a08,0x88200002,0x82800020,0x20a20,0x82202208,0x200808a2,0x20800820,0x80208,0x22080082,0x8208208,0x8a282288,0xa0022822,0x20000028,0x8228220a,0x808a20a2,0x28820208,0x82202282,0x2800022,0xa28028,0x8008a008,0x800a00a0,0x828a00,0x8a282200,0xa08800a2,0x8200a20,0x8a000288,0xa08a2880,0x8200020,0x20a080,0xa0800082,0x20228a20,0x228000a,0xa08a0082,0x28a00220,0x8828a,0x200a2820,0x20a00a00,0x200a200,0x28028a0,0x28800028,0x88288088,0x22080822,0x28a08a20,0x8a088288,0xa2800000,0x28828800,0x228828a,0x82082082,0x8200208,0x82080200,0x80820020,0x28a20200,0x2088088,0x208a0800,0x20800028,0x88000082,0x220a0800,0x8020028,0x200a20a,0x808a0802,0x20208a28,0x8200a288,0x208a0800,0x808808,0x88282200,0x20822820,0x8a20a08,0x8082282,0x22082082,0x800,0x288008,0x8a28a0,0x800808,0x2200080,0x80020882,0x808800,0xa00a288,0x228000a0,0x20a20028,0x82280280,0xa2820002,0x808,0x80200280,0x22082800,0x20028a00,0x80200002,0xa20828a0,0x8a08020,0x8020a282,0x880000,0x8800a28,0x80200282,0x2800,0x28a20200,0x88280008,0xa0802802,0x28828008,0xa00a000,0x200228a0,0x200220,0x88082208,0x20002000,0x20020200,0x8000a28a,0x20022,0x8800200,0x2082082,0xa2082882,0x20820a08,0x82208288,0x20000882,0x20820220,0x820020a,0x808228a0,0x20228a20,0x808020a,0xa08a0);
remainders_pos[943].modulo_value=3772; remainders_pos[943].bit_mask=create_bit_mask(118,0xa02020a,0xaa8808,0x8a0a2a82,0x2002a82,0xaaaaa222,0x22822a0,0xa2a0828,0xa8880a2,0x2a222808,0x8aa200aa,0x28208882,0xa008a82,0x2a888808,0x8288aaa8,0x20aaaa22,0xa288002,0xa2000a80,0x202a88,0x8aa00,0x8a8a0802,0x80028a02,0x80a88808,0xa00a028,0x822208a2,0x82880220,0xa0a88082,0x2a20a02,0x2a8aaa8,0x80288828,0x28020aa0,0x80a220aa,0xa000200a,0xa828008,0x82a2028,0x2882a20a,0x8822880a,0x20882a82,0x22a2200a,0xa820aa2,0x8a208202,0x2a8022aa,0x2200a8aa,0xaa028888,0x82aa2082,0x800028,0xa000200a,0xa2222200,0x8a020a82,0xa222880,0x228802,0xa2a08880,0x882a2200,0xa0a00082,0x808a08a,0x88880000,0xa00a282,0x88008282,0xa282002a,0x28a8a02,0xa80a2a28,0x2a800828,0x8282002,0x2828a00a,0x222,0xaa20a202,0x282000a0,0xa0088a82,0x2220a8,0x8022880,0x2022888a,0xa82a080a,0xa0088888,0xa008000,0x22800020,0xa8208aa8,0x8222280a,0x8aa2a008,0x2aa8802a,0x808208a,0x88aa082a,0x8a0088a8,0x282a8220,0x8a022882,0xa08a822,0x2808a82,0xa200282a,0x2a008000,0x8aa088a0,0x20aa0802,0x2822280,0x2aaa2a8,0xa80a08a8,0x282022a0,0x20880228,0x8a20888,0x2280a00a,0x220222a0,0x280a2800,0x8020a2a,0xaa200,0xa22a8080,0x2a0008,0x8800228a,0x288aaaa0,0x82aaa228,0x202222a,0x882a200a,0x28222082,0x8aa008aa,0x202888a,0x8a0222a,0x2820a8a,0xa0a88288,0x888aaaa,0x282a8008,0x82a8a0a,0x2022aa0,0xa08080a);
remainders_neg[943].modulo_value=3772; remainders_neg[943].bit_mask=create_bit_mask(118,0x820a8a82,0x88220080,0x282a20a,0x8280a20a,0x22222aaa,0x8aa0aa28,0x82a200a0,0x8200082a,0x22aaa080,0x22a8022,0xa0a8000a,0x8288020a,0xa2000000,0xa002220,0xa82222a2,0x82a00882,0x2a888208,0x8a8a200,0x88802288,0x202800a,0x8a028a,0x8200080,0x828828a0,0xa2a802a,0xa0082a8,0x2820080a,0x8a22828a,0x8a202220,0x8a000a0,0xa08a0228,0x82aa822,0x2888a882,0x820a0080,0x80a2a8a0,0xa0022a82,0xaa0002,0xa800a20a,0xaa2aa882,0x820a8222,0x2280a8a,0x2208aa22,0xaa882022,0x228a0000,0x222a80a,0x800888a0,0x2888a882,0x2a2aaa88,0x28a820a,0x82aaa008,0x8a2008a,0x2a280008,0xa2aa88,0x2828080a,0x80802802,0x8888,0x82882202,0x880a0a,0x2a0a88a2,0x8a02020a,0x2082a2a0,0xa2088020,0x80a0a882,0xa0a02882,0x8888aaa,0x22a82a8a,0xa0a88028,0x2080020a,0x88aaa820,0x808aa008,0xa82a0002,0x20a20082,0x28800000,0x82800888,0xaa0888a8,0x20a80220,0xaa22082,0x22a2880,0xa22008a2,0x8080a002,0x2280a2,0x2080020,0xa0a20a28,0x28aa00a,0x828020aa,0x8a080202,0x2288a0a2,0x22880888,0x2280028,0xa822808a,0x820aaa08,0xa222a20,0x20828020,0xa028aa28,0xa8008aa0,0x802a8000,0xaa002882,0xaa8aaa20,0xa082a088,0x808a02a2,0x88822a88,0x2aa20808,0x88a28000,0x88aa02,0xa0022228,0xa222a20,0x8a8aaaa2,0xa2a082,0xa0aaa802,0x2288022,0xa8a0002,0x8028aaa2,0x8a0a0202,0x20200a00,0x80002222,0xa0a20880,0x80220282,0x8a82a228,0x2808082);
remainders_pos[944].modulo_value=472; remainders_pos[944].bit_mask=create_bit_mask(15,0xa2a20a22,0x2a2aa80,0x82a0008,0x22088288,0xa820228,0xa8a80200,0x8a8a02a,0x20a2288a,0x2aa80a2a,0xa000802a,0x88288082,0x20228220,0x802000a8,0x8a02aa8a,0x88a08a);
remainders_neg[944].modulo_value=472; remainders_neg[944].bit_mask=create_bit_mask(15,0x2a2a82aa,0x822a2208,0x80a28880,0xaa800a00,0x820a8aa0,0x20208a88,0x802028a2,0xa82aa002,0xa22082a2,0x28880822,0xa0080a,0xa8aa0aa8,0x8a88820,0x28a2202,0x2802);
remainders_pos[945].modulo_value=420; remainders_pos[945].bit_mask=create_bit_mask(14,0x802002,0x8200200,0x2008200,0x20228a2,0x20800208,0x8000200,0x800a0020,0x8000200,0x8002080,0xa2880802,0x8200808,0x8008200,0x802000,0x8);
remainders_neg[945].modulo_value=420; remainders_neg[945].bit_mask=create_bit_mask(13,0x80082802,0x208a00,0x2080288,0x82022022,0x20008a00,0x80080288,0x200a0,0x808208,0xa000,0x22000002,0x200000,0x80000280,0x82080);
remainders_pos[946].modulo_value=3784; remainders_pos[946].bit_mask=create_bit_mask(119,0x8aa0a2aa,0x8002a0a8,0x280a0a22,0x2028aaa2,0x28882880,0x2208a20a,0xa008028,0x22a0a02,0x8820088,0x8a8882a,0x82802aaa,0x80a202,0x822a0020,0x20000aa,0x8a22a0,0x8a8822aa,0x22822088,0x8088288a,0x82828a20,0x20022228,0xa2800,0x2a88008a,0x820a820a,0xa282880,0x2a0aa0,0x28a88020,0x82a2220,0x8082aa0a,0xa2002022,0x2222220,0xa22a20a,0x20028882,0x220028a0,0x2a80002a,0x202aa8,0x882800aa,0xa080a082,0x8aa80020,0x2020aa80,0x2220aa2a,0x828208,0x88008a8a,0xa228020,0x2088088,0x22082a0,0xa2aa2a8,0x2aa00280,0xa00a8aa0,0x2a0282,0x20080880,0x8aa088a0,0x80a02228,0x2a8280a2,0x2208a028,0x88088820,0x82222,0x2020a820,0x8028a82,0xa0080,0xa00000,0xa2802002,0x2a080882,0x88200008,0x22202288,0xa208808,0x282a828,0x880a028a,0x220aa228,0x2020080a,0x80a80002,0xa2a00a82,0x800aa80a,0x8aa8a002,0x8208802a,0x220800a,0x288a022,0xa2002208,0x828200a2,0xaa088820,0xaa0808a8,0x2aa202,0xa020a08,0x282282,0xa80800aa,0x2a82a,0x280088a8,0x2280080a,0x8a88a082,0x888880a0,0x8008a08,0xaa820288,0x88a820a0,0x22a2808,0xa0a80008,0x2828a00a,0x82a08202,0x22a8a0,0x28a000a2,0x88800800,0xa2828228,0x28220208,0x88288a2,0x8822a222,0x88a200aa,0x800a,0xa882aa,0x8a020008,0xa8028280,0x222a20aa,0x8220a8,0xa0a88022,0x200a080,0x8a208828,0x282228a0,0xaa280802,0xa0a0288a,0xa800288,0x8a0aa22a,0xaa);
remainders_neg[946].modulo_value=3784; remainders_neg[946].bit_mask=create_bit_mask(119,0x2282222,0x80a2020,0xa08282aa,0xa8a02222,0xa0002008,0xa2802a82,0x820808a0,0x8aa2820a,0x80028800,0x2000a2,0xa08a222,0x8808228a,0xa2288a0,0x8a888822,0x8802aa20,0x2002a22,0xa20aa800,0x8002002,0xa0a0228,0xa882aaa0,0x882a088,0xa2008802,0x2820282,0x8220a008,0x88a28228,0xa02008a0,0x80a22aa8,0xa2282,0x2a88a82a,0x8aaaaa28,0x82a22a82,0x288a000a,0xaa88a028,0xa20080a2,0x8828a220,0xa08822,0x28082802,0x22008a8,0x20a82208,0xaaa822a2,0x880a0a00,0x800202,0x2aa08a8,0x8a800800,0x8aa80228,0x82222a20,0xa2288a08,0x28820220,0x88a20a0a,0xa0008008,0x2280028,0x828aa20,0xa202082a,0x2a8028a0,0x8000a8,0x8880a2a2,0xa82820a8,0x808a020a,0x88828800,0x88280888,0x2208288a,0xa280800a,0xa88800,0xaaa0aa00,0x2a80080,0x820a20a0,0x828202,0xaa022aa0,0xa8a88082,0x8208882,0x2a28020a,0x822002,0x220288a,0xa800822,0x8aa00882,0xa0028aa,0x2a80aa80,0xa0a802a,0x220000a8,0x22808020,0x88a22a82,0x28a0280,0x80a0aa0a,0x20808822,0x888a2022,0xa0800020,0x2a088082,0x200200a,0x28,0x80080280,0x220a8a00,0x20a820,0x8a222080,0x20208880,0xa0a02882,0xa280a0a,0x88a22028,0x2028882a,0x88080,0x2a0a02a0,0xa02a8a80,0x800a002a,0xaa2aa2,0x2a0822,0x80880882,0x88200a22,0x28a8800,0x20820a08,0x22a2a822,0x880aa820,0x282000aa,0x8a082808,0x2a800a0,0xa0aaa020,0x22a0000a,0x2028a002,0x82088a00,0x2822a22,0x22);
remainders_pos[947].modulo_value=3788; remainders_pos[947].bit_mask=create_bit_mask(119,0x2888a82,0x882a8a80,0x22a2228a,0x200a208,0x82208a0,0x2a88aaa,0xa00a8082,0x2a002202,0xa02a8002,0x22aaa02a,0xa28a0202,0x2a8a82a0,0xa80a0a28,0xaa220a0,0x2aa8a8aa,0xa20228a2,0x288a80a8,0xa8a22008,0x88088a82,0x88a28282,0x22aaa0,0x2028aa00,0xa2882822,0x2a288022,0x2a2aaa02,0x882088a,0x8028aa0,0x28282002,0x880a,0x2802a08a,0xa22aaa82,0xa8a0a082,0x20aa28a,0x228a22a8,0x80008082,0xaa20808a,0xa0a20808,0x8020a888,0x8aaaa,0xa082228,0x222a20a,0xa888200a,0x2a0220a2,0x208a8820,0x20200088,0x28880280,0x8282a028,0xa0200a0,0x8a820828,0xa800088a,0xaa00a800,0x88aa808a,0x2a02a80a,0x2200a8a,0x2288a280,0x882a8aa8,0x88808a2,0x200a202,0x2220aaa,0x8880a0a,0xa0080aaa,0x2220808,0x2a8228a2,0xa22882aa,0x880028,0xa80a8a2a,0x2aa22a02,0xa00aaa20,0x2a002,0x82a2a22,0x80a0282,0xa82820a0,0x2228280,0x808028,0x2a208220,0x880a8082,0x8222a8a0,0xa8880a00,0x820a0a08,0xa2000288,0xa0802aaa,0x8a0a222,0x208aa202,0x20002a20,0x8a288820,0xaa0802a8,0xa0a2aa28,0xaa88a820,0xa802882a,0x20000a20,0x82828a02,0x28020800,0x82200aa,0xaa8a8a22,0x228a880a,0x8228a880,0xa2808882,0xa880000a,0x28a220aa,0x22022828,0x88a2a82a,0x2a228200,0x8808a2a0,0xa2aa88a2,0x88aa0aa2,0xa02a0a0,0x2a2a828a,0xa28a0a8,0xaaa88808,0x2a80aa80,0x800a8800,0x2a00a808,0x22a80820,0x8820aaa,0xa00800a2,0x88a88208,0x2a822a28,0x2228002a,0x82a);
remainders_neg[947].modulo_value=3788; remainders_neg[947].bit_mask=create_bit_mask(119,0x8a00020a,0xa20208,0xaa2aaa02,0x8a882a80,0x80aa8028,0x8a200222,0x2882080a,0xa288aa8a,0x28a2088a,0xaa2228a2,0x2a028a8a,0xa2020a28,0x208282a0,0x822aa828,0xa2202022,0x2a8aa02a,0xa0020820,0x202aa880,0x80020a,0x2a0a0a,0x88aa2228,0xa8a02288,0x2a00a0aa,0xa2a008aa,0xa2a2228a,0x800a8002,0x808a0228,0xa0a0a88a,0x88880082,0xa0822802,0x2aa2220a,0x2028280a,0x8a822a02,0xaa02aa20,0x888080a,0x22a80802,0x282a8080,0x8a82000,0x88802222,0x8280aaa0,0x8aaa2a82,0x2000a882,0xa28aa82a,0xa80200a8,0xa8a88800,0xa0008a08,0xa0a28a0,0x828a8828,0x20a80a0,0x20888002,0x22882088,0x220802,0xa28a2082,0x8aa88202,0xaa002a08,0xa20220,0x8000802a,0x8a882a8a,0x8aaa8222,0x80008282,0x28808222,0x8aaa8080,0xa20aa02a,0x2aa00a22,0x880088a0,0x208202a2,0xa22aa28a,0x288222a8,0x888a288a,0x80a2a2aa,0x80828a0a,0x20a0a828,0x8aaa0a08,0x880808a0,0xa2a80aa8,0x82080a,0xaaa2028,0x20008288,0xa828280,0x2a888a00,0x2808a222,0x80282aaa,0xa8022a8a,0xa888a2a8,0x2a000a8,0x22808a20,0x282a22a0,0x220020a8,0x208a00a2,0xa88882a8,0xa0a028a,0xa08a8088,0x80aa8822,0x220202aa,0xaa020082,0xaa02008,0x2a08000a,0x20088882,0xa02aa822,0xaa8aa0a0,0x2a20a2,0xa2aa0a88,0x802a28,0x2a22002a,0x22822a,0x828a2828,0xa2a20a02,0x82a02820,0x22200080,0xa2082208,0x8820088,0xa2882080,0xaa2080a8,0x800a8222,0x2880882a,0x200a80,0xa20aa2a0,0xaaa088a2,0xa2);
remainders_pos[948].modulo_value=1896; remainders_pos[948].bit_mask=create_bit_mask(60,0xa20a2002,0x8a28208,0x2288,0x20a0802,0xa00200,0x8220220a,0xa0800020,0x28820208,0x8200008,0x2822022,0x20028,0x2008282,0x808a2000,0x28820a28,0xa002282,0x22882880,0x20228a08,0x800008a,0x82820,0x28028008,0x82088088,0x22820000,0x20a20808,0x88208000,0xa00a0880,0x8000800,0x8020a,0xa2022880,0x8a20828,0x28008a,0x208a20a2,0x28808a28,0x80200002,0x200020a0,0x20a00a00,0x2082202,0x208a0800,0x828820,0x2208200,0x2802822,0x28200020,0x2008,0xa28808a2,0x28228820,0x8800a002,0xa0822882,0x8a20228,0x82008000,0x800082,0x8828028,0x82088,0x80822820,0x20a20,0x88088208,0x800a00a0,0x20a08000,0x88000080,0x828a2022,0x8a08a20,0x80);
remainders_neg[948].modulo_value=1896; remainders_neg[948].bit_mask=create_bit_mask(60,0x22822882,0x220a00,0x88082200,0x82820082,0x8208a08,0xa28a282,0x200808a0,0x20028a00,0x80288080,0x820a28a2,0x8828820,0x8a08020a,0x22080,0x20028220,0x8208a20a,0xa2002000,0x28a20200,0x80088002,0x808020a0,0x20820800,0xa000000,0xa20a0880,0x28228000,0x280088,0x20820000,0x808008,0x88008282,0x228a2000,0x228020,0x88208002,0xa0022822,0x20000220,0x8288082,0xa0882820,0x28208208,0x8a00a28a,0xa0020080,0x8020028,0x8a280288,0x820820a2,0x20a08828,0x808a080,0x22000022,0x20a20028,0x8208a,0x200a2002,0x228a20,0xa080088,0x80080802,0x20820,0x8800a000,0xa20a0,0x8828228,0x280,0x820820,0x28200808,0x88008,0x20228a2,0x200228,0x8);
remainders_pos[949].modulo_value=949; remainders_pos[949].bit_mask=create_bit_mask(30,0x9ab19aba,0xbc83b05e,0x305fac21,0xb20d1658,0x98971d0,0x70b7e4aa,0x5648c437,0x3a7bf9a6,0xadca447,0x5c9e1786,0xbe11b19,0x68f58a40,0xa7655822,0x4281cdbc,0x61163,0x50b1a218,0xb94f6ce0,0xc59106a9,0xf400946b,0x4ea63621,0xd4187a1e,0x9738894e,0x9a9967f7,0x43bb08c4,0x641549fb,0x1342e3a4,0x83069a2c,0x4f610d7e,0x565e8370,0x175663);
remainders_neg[949].modulo_value=3796; remainders_neg[949].bit_mask=create_bit_mask(119,0x2280222,0x200a2802,0xa08220a8,0x2a808a80,0x80002808,0xa82a2822,0x828008a2,0xa222202a,0x8200288a,0x8002820a,0x2288280,0xa0280288,0x2aa080aa,0x8a080020,0x888a882a,0x282a80,0x2082a028,0x8080a20,0x288808a2,0x822aaaa0,0x800a282,0xaa00082,0x200a22a,0xa228008,0xa8888022,0x8a822a20,0xa8a02a0,0x82a88022,0x82820aa8,0x8a8a8aaa,0x820a0aaa,0x8aa08aa,0x28a8220,0xa20a882a,0x888a822,0xa228220,0x28020080,0x2a82a2,0x28008208,0xaaa20a0a,0x8088a02a,0x80808228,0x2a082022,0xa0a000a0,0x8a88880a,0x828aa0,0x82aa020,0xa02aa8,0x8a2008a,0xa00080a,0xa0020a82,0x22222a88,0x800a0aa0,0xa2a0aa28,0xa0000a20,0x880aa080,0x2208280a,0xa28020a8,0xa20200,0xa222222,0x8208a2a2,0xa220022a,0x2880a808,0xa80a8a00,0x20220aa,0xaa88000a,0x222a0000,0xa2088020,0xa08280a2,0x8a028aa,0xa2088a2,0x8a80022,0x8a2a2a00,0x80880082,0xa00a088,0x202a088a,0x88822a28,0x8a80228a,0xa08,0xa020a022,0x22a8a288,0xa8020280,0x20aa0aa2,0x88aa2082,0x82020800,0x82202880,0x8a82280,0xa28000a0,0x800080,0x80800080,0x28000a2,0x80a20a88,0x808a0220,0x80082020,0x20822a88,0xa2a82a82,0xa0200a,0x88a28aa2,0x22028202,0x88280000,0xa8a200a8,0xa2a2088,0x28882a02,0x88828028,0xa0800880,0x2a2a28,0x22000a88,0x22888228,0xaa8a0288,0xa280a082,0x2008822,0x80002a22,0x280008aa,0xaa822020,0x28a80a,0x880a808a,0xaa200222,0x22a28820,0x22228);
remainders_pos[950].modulo_value=760; remainders_pos[950].bit_mask=create_bit_mask(24,0xa0222a02,0x220820,0xa82a0,0x8a202088,0x280aa08,0x28a2228a,0xa0000200,0x8000280,0x88822820,0xa200802a,0x8282202,0x82200022,0x20880008,0x8a808828,0x22a80200,0x20082882,0xa028000,0x28008000,0x80a2888a,0xa220aa02,0x220808,0xa82a0,0xa082088,0x80a888);
remainders_neg[950].modulo_value=760; remainders_neg[950].bit_mask=create_bit_mask(24,0x28a2228a,0x882a80a0,0x8820228,0x2a0a800,0x82082200,0xa02a2202,0x28088a80,0x88208,0x2a0a8,0x22880822,0x80a02a82,0xa2888a2,0x28000080,0x20000a0,0xa2208a08,0xa880200a,0x820a0880,0x20880008,0x8220002,0x22a0220a,0x88aa0080,0x8020a20,0x280a000,0x2000);
remainders_pos[951].modulo_value=3804; remainders_pos[951].bit_mask=create_bit_mask(119,0x228a0822,0x28820220,0x88280200,0xa2080000,0x20820a08,0xa280288,0x20a20a2,0x28200208,0x28a080,0x82822082,0xa08,0x8808a280,0x828a00a0,0xa00220,0x828808a,0x20822080,0x8020800,0x80008280,0xa0020002,0x8a28208,0x8808a200,0x82822882,0x20a28008,0xa28a000,0x208a0822,0x8a08808,0x8228a,0x80880080,0x28828200,0x208808a,0x2000022,0x20220a08,0x8228a20a,0x20020000,0x208220,0x8082,0x828a0,0x8020228,0x2200200,0x22880,0x8a00208,0xa28028a,0x82822802,0x8208a20,0x20a208,0xa00a0022,0x20808,0x220200a,0xa28a2800,0x8000200,0x288208,0xa0802820,0x828208,0x80008008,0x82082002,0x20800200,0x8020228a,0x82002882,0x8808,0x8200082,0x22022000,0x28228008,0x8a288080,0x20080020,0x28008208,0x2002000,0xa2082820,0x828020,0x2082280,0xa0080002,0x28a28,0xa008088,0xa2020800,0x8800a00,0x208a080,0x208a2082,0x8028828,0xa28028a,0x20800a2,0x228800,0x80088,0x2880802,0xa28200,0x8202000,0x80882080,0x20000800,0x8a08a288,0x20a0880,0x8800008,0x8a202208,0x20082822,0x200220,0xa288200,0x820220a2,0x8820a20,0x8000a28a,0x220028a0,0x28228828,0x8a202,0xa20828a2,0x28000800,0x282000,0x80020802,0x208820,0xa202282,0x208800a0,0x20a00a28,0x28a202,0x220a0000,0x8208828,0x8020a280,0x2080082,0x8a08a08,0x8228028a,0xa20a0820,0x20000208,0x80080282,0x80880822,0x8820a28);
remainders_neg[951].modulo_value=3804; remainders_neg[951].bit_mask=create_bit_mask(119,0xa20200a2,0x20028a28,0x208288,0x22800880,0x28028200,0x82208200,0x82822822,0x20a08a00,0x88202008,0x20a2802,0x8808200,0x2208,0x2020820,0x8208a28,0x80200002,0xa00a2800,0x828008,0x8080208,0x208a0882,0x220a00,0x2288,0x20a2002,0x28220800,0x82202088,0xa00200a2,0x200000,0x8800a202,0x800,0x20020a08,0x8a000002,0x828808a2,0x28a28200,0xa202282,0xa08a0880,0x8a00a28,0x8808000a,0x80802020,0x828a20,0x8a288288,0x808a2000,0x208a00,0x82208202,0x20a2082,0xa00228,0x88282280,0x208208a2,0x8828000,0x8a28a082,0x22022080,0x808a08,0x88200280,0x200820a0,0x8020a00,0x8080080,0x2802882,0x28008a08,0x828a202,0x2882002,0x8800000,0x8028800a,0xa28a2880,0x20a20800,0x2200008,0xa08008a0,0x20800a00,0x8a08a088,0x228020a0,0x8020828,0x8a00a208,0x20800082,0x8820220,0x82080000,0x228a0080,0x8208,0x8a002008,0xa0022802,0x820020,0x82208202,0x82800822,0x8a20008,0x88008000,0x82000082,0x8220a08,0x8028a088,0x2800,0x28808008,0x2002200,0x82820000,0x8800,0x228a280,0xa08020a2,0x8a08a28,0x82200288,0x28a2822,0x28228,0x8082202,0xa2882020,0x20a20020,0x8800228a,0x22802022,0x20808008,0x8820a088,0x8a0082,0x8a00028,0x8228a20a,0xa0000820,0x28208220,0x8820228a,0x22820880,0xa00020,0x8282208,0x82800802,0x200200,0xa208202,0x228200a0,0x28808a00,0x800820a,0xa2,0x28220);
remainders_pos[952].modulo_value=952; remainders_pos[952].bit_mask=create_bit_mask(30,0x2a80220a,0x808aa0,0x80828aa8,0xa20a0a2,0x28002a0a,0x80220,0x88228220,0xa202a,0x2a208220,0x222288a8,0xa02028a0,0x82020a00,0x88000208,0x2800808a,0x822200,0x28008882,0x22a20200,0x82208000,0xa00a080,0x880a2808,0xa82a2288,0x88208,0x22a808a0,0x88288,0x28088020,0xa0a0a800,0x28a0a08,0x2aa282,0xa80aa202,0xa08802);
remainders_neg[952].modulo_value=952; remainders_neg[952].bit_mask=create_bit_mask(30,0xa208aa02,0x8000220,0xa0220,0x8228282a,0xa080a282,0x80800aa8,0xaa02a8,0x882a822,0x22a80aa0,0xa2aa0020,0x2828a020,0xa828288,0x880a00,0xa0880002,0x880aa208,0x20880002,0xa22a0a88,0xa280888,0x82802808,0x822080,0x2022a200,0x88800a00,0x22208020,0x80800a00,0x200008a8,0x28202088,0x8a020280,0x88a22202,0x20822a0a,0x280002);
remainders_pos[953].modulo_value=953; remainders_pos[953].bit_mask=create_bit_mask(30,0x7607e396,0x9fb60237,0xa80e87bd,0xc3ff6794,0x6a9565db,0x44e0a254,0x349f499a,0xf0055df7,0x9431d96d,0x14c6c31b,0x6ca43932,0x1a9a5680,0x3a43c166,0xaf38e35f,0x33715517,0xd7a2aa3b,0x73eb1c73,0x619a0f09,0xd805a965,0xa1327094,0xa7630d8c,0x3eda6e30,0xb3beea80,0x89664be4,0x58a9141c,0xf6e9aa5,0x54a79bff,0xe6f785c0,0xbbb101b7,0x1a71f81);
remainders_neg[953].modulo_value=3812; remainders_neg[953].bit_mask=create_bit_mask(119,0xaa8a2a0a,0x22a8aaa,0x20820a20,0xa22aa08,0xa208a802,0x88282a88,0xa8028002,0x2888802a,0x8a800a0,0x880a0a82,0xa028a0aa,0x82028a08,0xa28a08aa,0x22a02a82,0xaaa8888a,0xa2a22a2,0xaa2280aa,0xa8028280,0x8820a8,0x28aaa808,0x2aaa8000,0xa202a2a8,0x2a222208,0xaa8228,0x80208880,0x82a20228,0x882a0222,0x2a2a0808,0x2228882a,0xa02a8208,0xa220820a,0xa2a2a088,0xa0888082,0x2a0a2202,0x2008a202,0xa8000888,0x22a0a20a,0x2288222,0xaaa0aa2a,0xa8a0000a,0x880080aa,0xa8a820,0x22a80a0a,0xa282a80a,0xaaaa2a22,0x2a228888,0x8a2a0aa0,0x20aa882,0xa028828a,0x808aa828,0xa8820a82,0x88a02800,0xaaa008,0xa08a000a,0x82288aa0,0x228200a8,0x82082aa,0xa2002282,0x8aaaaa8a,0x28282a2,0x80000082,0x822aa200,0x28a28a2,0xaa28220,0x882288,0xaaa08a0a,0xaa000aa0,0xa82a0888,0x80a2880a,0x28008a82,0x82882a08,0x8800a2a0,0xa02082,0x8882202a,0x20200008,0xa8028202,0xa0a80220,0x28080aa0,0xa8aa88a,0xaaaa0800,0x200a0000,0x228822a0,0xa20a0222,0x88aaa800,0xa208aa28,0xa220a022,0x8a888a02,0x8a020202,0xa28a2208,0xa2802200,0x28882228,0xa8a02020,0x22a02888,0x22a20282,0x888a2a88,0x22800aaa,0xa2222028,0x202a208,0xaa800028,0xa800082a,0xa288aa8,0x8282a808,0xa82200a,0x22020a0,0x88880002,0x802a0220,0x8a08202,0xa082a280,0xa082a08,0x80a0a880,0xaa808a2,0x2a88882a,0xaa82a800,0x80282882,0xa808a208,0xa00220a2,0x20a28a28,0x8022aa,0xa0208000);
remainders_pos[954].modulo_value=1272; remainders_pos[954].bit_mask=create_bit_mask(40,0x20a00a2,0x20028008,0x82280008,0x22822022,0x20008820,0x28a200,0x80000080,0x28220800,0xa208000,0x80820802,0x20a08208,0xa208200,0x82882020,0x28828a28,0xa20000a,0x220808a2,0x220200,0x828a002,0xa08a0022,0x28020,0xa080280,0x208800a2,0x800a28,0x88008088,0xa08a2020,0x28a00008,0x8228a282,0xa0080822,0x8008208,0x220820a,0xa0802082,0x28000208,0x2002088,0x20000,0x8008a28,0x88082200,0x82880882,0x8200028,0x80200280,0x8a00a0);
remainders_neg[954].modulo_value=1272; remainders_neg[954].bit_mask=create_bit_mask(40,0x82820822,0x28820800,0xa208080,0xa20a28a2,0x28800028,0x88202288,0x880800,0x20a28008,0x82280088,0xa0082,0x28200a00,0x82200288,0x20028a0,0x20020220,0x82288082,0xa2800022,0x8a28a08,0x8020208a,0x20020822,0x8820828,0x82008208,0xa0000822,0x8008220,0x80000,0x200228a0,0x20208800,0xa20220a,0x208000a2,0x800a00,0x8a280282,0x20082802,0x20800a00,0x8a08a000,0x808a0880,0x800220,0xa288,0x2000002,0xa08820,0x8288200,0x20820);
remainders_pos[955].modulo_value=3820; remainders_pos[955].bit_mask=create_bit_mask(120,0x88880a0a,0x2220822,0x280202a8,0x282a888,0xa0a80a0a,0x82a08,0xa8228a00,0x200a0208,0x8a822082,0x22200202,0x282a2a82,0x8a200aa0,0x2a800208,0x82a0082a,0x2288a022,0x8802200,0xa220082,0x8002200,0x880020a0,0x8088a20a,0x20820002,0xaa000280,0x2888228,0x8080808a,0xa228200a,0x208a0a88,0x2a2a0a20,0xa022008,0x80800000,0x2802a08,0x2822008a,0x822a8aa2,0x82a088,0x8a08080,0x80200202,0x28022880,0xa020a22,0x8220,0x8a208aa,0x2208aa0a,0xa0020802,0xa02808a2,0x28a80a8,0x2822888,0xa8882a,0x8082802,0x82a0,0x20282a8,0x80a000a8,0x80880000,0xa0a88,0xa80a0882,0x20028280,0x8a228008,0x20208022,0x8820280,0xa0220002,0xa800288,0x200a088,0x20008222,0xa0080888,0x2a0220,0x880a228,0x8220800,0x20808028,0x288a2880,0x28008200,0xa02a028,0xa000822,0x220222a,0xa02000,0x280802a0,0x200002a8,0x820200a8,0x22a00802,0x88280a82,0x2a280222,0x280a2a0,0x800a8a2,0xa2088802,0x8a20a0a,0x20000aa2,0x80a0088,0x8802888a,0x802022,0x20a20808,0xa8200020,0x2a882220,0x88288aa,0x80280a20,0x20220a,0x880a0000,0xa8a8200,0xa20808a,0x8288a22a,0x20202a00,0x22280a20,0xaa288,0x8208028,0xa2202800,0x80022a08,0x800200a0,0x88a0008,0x80220820,0xa2288008,0xa82880,0x2a8a82,0x8a2208,0x8a8280aa,0x88882a,0x882a2808,0xa008820,0x2882a208,0x8200000a,0x2a0a20a,0xa8280a0a,0x8028222,0x88802a8,0x2222882,0xa0a);
remainders_neg[955].modulo_value=3820; remainders_neg[955].bit_mask=create_bit_mask(120,0x282,0x8a2a80a2,0x208a8220,0x8a022000,0x20208202,0x8802280,0x202a0280,0x28828280,0x202a80a,0xa2a88a0a,0xa0a2220a,0x2288220,0x22088280,0xa2080a2,0xa200282a,0x80082a88,0x822a8802,0x88a288,0x80a828,0x2a02,0xa80a088a,0x22088a00,0xa0002a0,0x8000802,0x22a0a802,0xa8020200,0xa22282a0,0x28aa080,0x8008888,0x8208a200,0xa0aa0802,0xa220222,0x80a2000,0x80200808,0xa88a0a,0xa08a2008,0x820a82a2,0x88802a8,0x80228022,0xa2802202,0x288a008a,0x28208022,0xa020020,0x8a02a000,0x80200022,0x8080208a,0x88080a20,0xa8a0220,0x8208820,0x8808,0x88820200,0x20028002,0x288a0208,0x2220880,0xa0a8082a,0x800a0a08,0x282a8882,0x2088200,0x8a802800,0xa0880a2a,0x28800000,0x88228aa0,0x822a0,0x80a28088,0xa0080820,0xa0022008,0xa0080a80,0x28a20a0,0x828080aa,0x82a8aa22,0x88282888,0xa0008a20,0x28888220,0xa828820,0xa228800a,0xa0020a,0xa2208aa2,0xa002228,0x8080202a,0x2280000a,0x802a0282,0xa8088222,0x828000,0x820002,0x8008a82a,0xa82a0080,0x202888a0,0x2200a2a8,0x80020022,0xa08228,0x88a82a82,0x28880,0x2020288,0x82a00802,0x2002a22,0xa8a82288,0xaa2082a0,0x8822200,0x80a008a0,0x22a8a008,0x88a2280,0x80a0820,0x28080,0x8a280a8,0x22a00800,0x88202008,0x88220202,0x802a280,0x2020822,0x80000022,0xa22080,0x820800a0,0x200a2280,0xa808882,0x82282a02,0x20a00282,0x800a0aa2,0x8220,0x8aa2a00a,0x202);
remainders_pos[956].modulo_value=1912; remainders_pos[956].bit_mask=create_bit_mask(60,0x228a02a2,0xa802a88a,0xa0228000,0x2a8aa8a0,0xa2828828,0x80088aa2,0x88a08aaa,0x222088a,0x22822,0x80222aa,0xa0a20a08,0x28200200,0x2aaa08a8,0x82200aa0,0x80a8208,0x22082a02,0x80aa0088,0x82a20aaa,0xa0080082,0x20a08a0,0xaa88802,0x8828800,0x2a220888,0x2aaa20a2,0xa8aa2200,0x82822828,0xa0a2aa2a,0xa0002880,0x8a22a802,0x28a80a28,0xa228a02a,0xa802a88,0xa022800,0x82a8aa8a,0x2a282882,0xa80088aa,0xa88a08aa,0x20222088,0xa0002282,0x8080222a,0xa0a20a0,0x82820020,0x2aaa08a,0x882200aa,0x2080a820,0x822082a0,0xa80aa008,0x282a20aa,0xa008008,0x2020a08a,0xaa8880,0x80882880,0x22a22088,0x2aaa20a,0x8a8aa220,0xa8282282,0xa0a2aa2,0x2a000288,0x88a22a80,0x8a80a2);
remainders_neg[956].modulo_value=1912; remainders_neg[956].bit_mask=create_bit_mask(60,0xaa028a2a,0x208a2002,0x28aa0888,0xa2022028,0x2a0a00a0,0x880022a,0x280222,0x8aaa0002,0x888aa0aa,0x808aaa22,0x282a8280,0xa0a88a88,0xa2228020,0xaa88228,0x80820a80,0xaa80a28a,0x8228800,0xa2a8222,0x2880880a,0x8a828028,0x8220008a,0x800a0088,0xa2aa8000,0xa222a82a,0x2022aa88,0xa0aa0a0,0x282a22a2,0x2888a008,0x2aa208a,0xa02082a0,0x2aa028a2,0x8208a200,0x828aa088,0xa202202,0xa2a0a00a,0x20880022,0x20028022,0xa8aaa000,0x2888aa0a,0x808aaa2,0x8282a828,0xa0a88a8,0x8a222802,0xaa8822,0xa80820a8,0xaa80a28,0x20822880,0xa0a2a822,0x82880880,0xa8a82802,0x88220008,0x800a008,0xaa2aa800,0x8a222a82,0x2022aa8,0x20a0aa0a,0x8282a22a,0xa2888a00,0x2aa208,0x2082a);
remainders_pos[957].modulo_value=3828; remainders_pos[957].bit_mask=create_bit_mask(120,0x2080002,0x20028808,0xa28820a,0x808a0880,0x8200200,0x82202282,0xa20820a2,0x8008208,0x8200a208,0x22880882,0x8020000,0x80208,0x220820a0,0x8008000,0x8088002,0x20a00a2,0x20228000,0x2002,0x82820000,0xa20200,0x220a00a,0x228a0080,0x20000228,0x8a080008,0x20020022,0x228000,0x2280080,0xa00a0800,0x8208020,0x208a008,0x80802882,0x28200820,0x8280,0x800a2,0x20000a20,0x8a200200,0x20800082,0x20a20000,0x8028a,0x2802822,0x8020228,0x20a,0x22002000,0xa20,0x82280282,0x82000880,0x20000800,0x282082,0x80820080,0x20000200,0xa08808a,0x80822802,0x28820800,0xa282082,0x820a2022,0x8020020,0x80088288,0x228208a2,0x808a00,0xa000082,0x20082000,0x28800a20,0x28a208,0x2022822,0x8208008,0x8a08808a,0x22882082,0x20200208,0xa080288,0x8a2022,0x20200800,0x80002008,0x882082,0x8200200,0x88202200,0x82802,0x8808a00,0x80,0x20a0800,0x28028808,0x88280,0xa08a2802,0x20800008,0x8a282000,0x800800,0x8a00,0x8a002,0x82802820,0x20208200,0x8082280,0x820200a2,0xa08020,0x8800020a,0x80002002,0x800028,0xa288008,0x820002,0x28828800,0x8800200a,0xa00a00a0,0x28200808,0x8,0x80880080,0x8000028,0x208a00a,0x2080022,0x8800020,0xa082,0x2020802,0x28800008,0x8282202,0x20208a0,0x8a20820,0x8828a082,0x82082880,0x20200800,0x8a00220a,0x8a0822,0x8020228,0x80002);
remainders_neg[957].modulo_value=3828; remainders_neg[957].bit_mask=create_bit_mask(120,0x82800082,0x28020000,0x82200282,0x20000,0xa00a08,0x228a20a,0x22802022,0x800a00,0xa002280,0x22000002,0x828808,0x88008280,0xa2002820,0x800808,0x80000082,0x82820822,0x20a20808,0x8808a00a,0x20a0800,0x8228a08,0x2282082,0xa2020800,0x28808220,0x2008080,0xa08a08a2,0x8a20800,0x8a200008,0x20820080,0xa00820,0x8a002000,0x2002,0x20208028,0x88080208,0x80800022,0x28008228,0x2288288,0xa0080802,0x28220808,0x8202,0x820820a2,0x828a20,0x88008282,0x22802880,0x8808228,0xa20820a,0x2080000,0x28808008,0x8820a002,0xa0800,0x20808a08,0x82000002,0xa2002,0x20028008,0x220200a,0x28228a2,0x828028,0x8000200,0xa20a0022,0x8000200,0x8208000a,0xa0002880,0x20008228,0x88202200,0x828220a2,0xa00000,0x2000002,0xa2002002,0x28208a00,0x82008200,0x800228a2,0x28a00008,0xa080,0x80002802,0xa08a08,0x20a288,0x802002,0x208,0x88088008,0x82020080,0x20820000,0x88000200,0x20022082,0x20000800,0x220a088,0x80080000,0x8800208,0x8002082,0x20820a0,0x28a00208,0x8000a208,0x28a0822,0x8200820,0x80282,0x882082,0x8008820,0x82200000,0x20882,0x20020008,0x8a082,0x20820020,0x20208000,0x88088080,0x800,0x800820,0x82002002,0x828008a2,0x8828,0x8000200a,0x28a0082,0x20008800,0x8020a28a,0x820a0020,0x228028,0x202002,0x2802000,0x20a08000,0x208a282,0x80020022,0x28a20,0x808a);
remainders_pos[958].modulo_value=3832; remainders_pos[958].bit_mask=create_bit_mask(120,0xaa002a0a,0x2280aa,0xa2a28a00,0x82a88a2a,0xa028aa28,0x8000020a,0x88882082,0x880aa222,0xa08828a2,0xaa8aa28a,0x8220a82a,0xa2008aa8,0xa88a82a2,0x82a0a882,0x20a0a080,0x80a0a0a,0x28020a28,0x2a282002,0x22008aa,0x22820aa8,0x202a2022,0x2a02800a,0x2aaa2080,0x280a8000,0x20a88888,0x2a2028a,0x22a00228,0x88a02a2a,0x22082a88,0x20a288a2,0x2208820a,0x888028a2,0x20202a08,0xa822aa02,0x2082a202,0x2888880a,0x2aaa80a8,0x2a8a2000,0x20aa82a0,0x222a202a,0xa800a282,0x2008aa22,0x22aa2820,0xa820a2a8,0xa0a0a8,0x8a8a0a0a,0x2880a02,0x2028088,0x88008aa2,0x280a22,0x820080,0x8208288a,0x822200a8,0x828a2888,0xa2aaaa,0x8820082a,0x208802,0xaaa08202,0xa822a,0xa02aa0,0xaa80a,0xaaa882a0,0x80820a,0x22802208,0xa82008,0x82aaaa8a,0x822228a2,0x822a0088,0xa22820,0x20082,0x2288a028,0x808aa200,0x80220280,0xa280a022,0xa0a0a2,0x2a2a0a0a,0x882a8a08,0x80828aa,0x2a88aa20,0x88828a00,0x8a808a8,0xa80a82aa,0xa80008a2,0x282a02aa,0x8a02222,0x2a808a82,0x880aa88,0x2220a808,0x888a2802,0x8a08220,0x888a228a,0x2222a820,0x88a8a80a,0x8028800a,0x8a2808a,0x2822222a,0xa80002a0,0xa80208aa,0x8a00280,0x888808a8,0x802aa082,0xa8aa2008,0x28800828,0x2028a080,0x8a0a0a0,0x82020a0a,0x2a822a0a,0x8a8a82a2,0x822aa200,0xaaa82a08,0xaa28aa2,0x228a2822,0x22888aa0,0x2820822,0xaa08000,0x8228aa28,0x8aa8a22a,0xa28a,0xaaaa0288,0xa0a800);
remainders_neg[958].modulo_value=3832; remainders_neg[958].bit_mask=create_bit_mask(120,0x2288a282,0x88aa0822,0x2a2a0288,0xa2002a2,0x28a022a0,0x8888a82,0xa80a,0x822aaa,0x2800a02a,0x22022a02,0xaa820a2,0x2a880220,0x20020a2a,0xa28200a,0x28282808,0x80828282,0xa08a82a0,0xa2a0a88a,0x8aa88022,0xaa0a8220,0xa8a2a8aa,0xa28a0882,0xa222a808,0xa0820888,0xa8200000,0x8a2a8a02,0xaa288aa0,0x28a2a2,0xaa80a200,0xa82a002a,0xaa800a82,0x8a02a,0xa8a8a280,0x20aa228a,0xa80a2a8a,0xa0000082,0xa2220820,0xa202a888,0xa8220a28,0xaaa2a8a2,0x20882a0a,0xa88022aa,0xaa22a0a8,0x20a82a20,0x88282820,0x2028282,0x8a00828a,0x8a8a0800,0x88022a,0x88a082aa,0x880a8808,0xa80a002,0xaaa8820,0xa02a000,0x882a2222,0xa880a2,0x88a8008a,0x22280a8a,0x88820aa2,0x8828a228,0x88822082,0x22200a28,0x88080a82,0xaa08aa80,0x8820a880,0xa222202,0xaaaa02a,0xaa28800,0x882aa0a8,0x888a880a,0xaa0028a0,0x8022a88,0x8aa8a08,0x2a0828aa,0x8028282a,0xa2a28282,0xa20280,0x8080a022,0xa20022a8,0xa0288,0x80208020,0x20820a22,0x2088802a,0xa0a28a22,0x8028aaaa,0xa208020a,0x80082200,0xaaa82080,0x2a08a,0x80280aa8,0x2aa02,0xaaaa20a8,0x202082,0x8a00882,0x802a0802,0xa0aaaaa2,0x20888a28,0x208a8022,0x80288a08,0x8020,0x8a2280a,0x2022a880,0xa00880a0,0xa8a02808,0x80282828,0xa8a8282,0xa20aa282,0x2020a2a,0xaa22a88,0x2220a280,0x822a022a,0xaa02a0aa,0xaa000228,0x8a0a80aa,0x82280888,0xaa022a0,0x2202aa2,0x88882a02,0x22228a00,0x282088);
remainders_pos[959].modulo_value=3836; remainders_pos[959].bit_mask=create_bit_mask(120,0x2882222,0x8082a22,0xa282808a,0xa2282082,0xa0002800,0x8a880aa8,0x880a022,0x82828022,0xa002000,0x80808a82,0x202aa822,0x8022828a,0x200a28,0xa0a8a008,0x880a8822,0x2822,0xa02a8aa8,0x8808a002,0xaaa20a82,0x8a8a0220,0xa0a82082,0xa8800200,0x22008082,0x80a28088,0x8822880a,0x8a22a80a,0x8aa822a8,0x80808280,0xaa8a8828,0x8888880,0x22220222,0x20220000,0x2a022a0a,0x82000,0xa00082,0x2820822,0x22280822,0x22a0a88,0x2a08a820,0xa20a008,0x88a0200,0x8028000,0xa882222,0x200000a0,0xa08aaa,0x82200a2,0xa2a02008,0x8228a0,0x800a00aa,0x8800028,0xa002a22,0x22a88a82,0x882a080a,0x8a02a02,0x22200,0x2880a2a8,0xa08a008,0x2280a00,0x808082a2,0x808a00a,0x2a00a202,0x8a82020,0xa0288,0x8200a20a,0x2a8a022,0x88800,0x280a80a2,0x8a020a82,0x82a22a8,0x88a800a,0x22800022,0xaa00a00,0xa0a28820,0x20080a8,0x8a00882,0x8aaa20a0,0xa00000,0x888822a,0x2802,0x80a22,0x8200a08a,0x82a20a,0x822a0a88,0x8820288,0x8820828,0x82000a0,0x80008200,0x8a0a880a,0x80000880,0x8880888,0xa0222222,0x22822a2a,0x20282020,0x22a882aa,0x2a02a88a,0x2a022882,0x822028a0,0xa8202008,0xa0080022,0x282082a0,0xa0880a2a,0x282a08aa,0xa800a202,0x2aa2a80,0x28828000,0xa8822a02,0x200a2a0,0x228a0080,0x8a282880,0x2882aa80,0x82a2020,0x2000800a,0x8802828,0x2880a022,0xa2aa022a,0xa0028000,0xa8208288,0xa202828,0x88a8202,0x8888228);
remainders_neg[959].modulo_value=3836; remainders_neg[959].bit_mask=create_bit_mask(119,0x8a00aa2a,0x80a2a2,0x220a0802,0x2a20a80a,0x2880a088,0x2000220,0x800820aa,0xa0a082a,0x288a880,0x8020a,0xa82220aa,0x8a20a02,0x80a802a0,0x28202080,0x82002a,0x888a0a2,0x20a20220,0x288a,0x2222820a,0x2020aa8,0x2820a00a,0x20088a08,0x2a880802,0x2a0800,0x2a0082,0x2a22082,0x2202a20,0x8080208,0x22020020,0xa2aa8aaa,0xa82a8888,0xa282a282,0x88802888,0x8828800a,0x8a0a802a,0x2aa080a2,0x82a28200,0xa20020a8,0x82a02880,0x80020a88,0x808a0088,0x8200aa2a,0x28888820,0x80280222,0x802a882a,0x2a20a880,0x880a2020,0x8828022,0x80088820,0x288a2a2,0xa220020a,0x228082,0x8020a28a,0x888a2a88,0xa0082220,0x82802800,0xaa08280,0x80a2a,0x80002882,0xa2802a8a,0x802028a8,0x88828200,0xa882a02,0xa202822,0x80800088,0xa002082a,0x282820a,0x80a22a20,0x80020082,0xaa08882a,0x2288280,0x202a00a8,0x8a080820,0x8020800a,0x2222828,0x88288088,0x80000a22,0x888a082,0x808082aa,0xa082802,0x88022282,0xaa20200,0x800a8200,0x800a8020,0xa88820,0x880a88,0x2020082,0x8808008,0x80000000,0x28aaa2aa,0xaa0aa222,0x28a0a8a0,0xa2200a22,0xa20a2002,0xa282a00a,0xaa82028,0x20a8a080,0x2880082a,0x20a80a20,0x200082a2,0xa0228022,0x20802a8a,0x8a222208,0xa00a0088,0x200aa20a,0xa882a20,0xa2028808,0x220a008,0xa0022208,0x80a228a8,0xa8880082,0x8008a020,0x200828a2,0x22228aa2,0x28020888,0x20a00a00,0x82a820a0,0x8002028a,0xa20);
remainders_pos[960].modulo_value=120; remainders_pos[960].bit_mask=create_bit_mask(4,0x20882,0x28220800,0x2088,0x822080);
remainders_neg[960].modulo_value=120; remainders_neg[960].bit_mask=create_bit_mask(4,0x808a0002,0x20228000,0x808a000,0x22800);
remainders_pos[961].modulo_value=124; remainders_pos[961].bit_mask=create_bit_mask(4,0x2aaaaaaa,0xaaaaaaaa,0x8aaaaaaa,0xaaaaaaa);
remainders_neg[961].modulo_value=124; remainders_neg[961].bit_mask=create_bit_mask(4,0x22222222,0x22222222,0x2222222,0x2222222);
remainders_pos[962].modulo_value=3848; remainders_pos[962].bit_mask=create_bit_mask(121,0xa2208a02,0x220a0808,0x822a0028,0x8a88000a,0x802020a8,0x200a00a2,0x8202022,0x802aaa2,0x8aa802a0,0xa82880a2,0x282a028a,0x8888a20a,0x2a0228aa,0x22a8002,0x828a8a22,0xa2020280,0x88822a8a,0x202a022a,0xaa00002,0x28a2822,0xa0002822,0x22882202,0xa2228008,0x28a8a80a,0xa08a08aa,0x20a0a882,0x8800200,0x288a2a02,0xa0aa8,0x8a00aaa0,0xa2082,0xa0a82a,0x80208820,0xa22888a2,0x88080a2a,0xa08202,0x28080800,0xa8822200,0xa2208208,0x282a8a02,0x28200222,0xa2aa00a0,0x20802a22,0x80220880,0x82a2a000,0x808282,0xaa0022a2,0x8208022,0xa2088080,0x2202800,0x8882800,0x2a00082,0x288a8,0x2a2a2822,0x2a0a220,0xa0a2a82,0xaa880088,0x2aa02080,0x8a0a220,0xa00a2208,0x88a00a82,0x8a0a2020,0x80aa808,0x22aa02,0xa8a0a022,0x8a0a8082,0x28a8a808,0x22800088,0xa802a,0x28222082,0x28088000,0x2208a00,0x2082002,0x8800aa88,0x8202008a,0xa8a8282,0x20880200,0xa8020802,0xaa8a88,0x8008280a,0xa2a82888,0x82088a80,0x88822a20,0x20202800,0x820a0000,0xa0202280,0x22288aa8,0x2208028a,0x2a0a0008,0x8a000a8,0xaa00a282,0xa0a0000a,0xa8a2282a,0x80022080,0x2a0a0800,0x20a20a82,0x2a2a28aa,0x2888aa0,0x88228820,0x28000a80,0x28a28088,0xaa088,0x80a80880,0xa88222a8,0x80808aa2,0xa2a28202,0x2a88088,0x2880a880,0x8a2222aa,0x80a828a0,0x2282aa2,0x802aa28a,0xaa80200a,0x808208a,0xa00888,0x808028a,0x22a22a,0xa882a0,0x20a08828,0xa2088a20,0x80);
remainders_neg[962].modulo_value=3848; remainders_neg[962].bit_mask=create_bit_mask(121,0x2aa8028a,0xaa828000,0x2a288a0,0x2000882,0x8a82820,0xa882882a,0x80a8a8a2,0x800a222a,0x2208a20,0x20a0002a,0x20a28a02,0x2a82,0xa202a022,0x8aa2088a,0xa02022a,0x228a8a08,0xaa202,0xa8a202a2,0x8228888a,0x8a02a0a2,0x2808a0aa,0x2a00aa8a,0x2aaa0080,0x20202082,0x28028022,0xa820200a,0x80008a88,0xa002a20a,0x80828220,0x2882228,0x8882280a,0x88282022,0x8a800a0,0x2a20002a,0x8082a2,0x8028028a,0x20808088,0x200aaa88,0x2aa00a80,0xa0a2028a,0xa0a80a2a,0x22228828,0xa808a2aa,0x8aa0008,0xa2a2888,0x88080a02,0x2208aa2a,0x80a808aa,0x2a800008,0xa28a088,0x8000a088,0x8a20880a,0x888a0020,0xa2a2a02a,0x822822a8,0x8282a20a,0x22000800,0xa228a808,0x282aa0,0x2802aa80,0x28820a,0x282a0a8,0x822080,0x88a2228a,0x202028aa,0x282080a,0xa0202000,0xa2088800,0x88820822,0xa0aa280a,0xa0800888,0x8aa80280,0x8200a88a,0x882200,0xa8a8002,0x2020a0a,0xa8008a88,0x2082008a,0x88220200,0x880a002,0x2220a000,0xa800208,0xa22a0,0xa8a8a088,0xa828880,0x2828aa08,0xaa200220,0xaa808202,0x22828880,0x80288820,0x22802a0a,0x28288082,0x202aa022,0x8aa808,0xa2828088,0x282a020a,0xa2a2a022,0x8a000220,0x2a00a8,0xa0888208,0xa0220000,0x8822800,0x8208008,0x2002aa20,0x808022a,0x2a2a0a0a,0x82200800,0xa0082008,0x2aa2a22,0x20a028,0x8aa0a222,0x8222a02,0x2208a882,0x8080a002,0x8280000,0x80808a02,0x88a22aa2,0x88200a28,0xa8280020,0x228002a0,0x8);
remainders_pos[963].modulo_value=1284; remainders_pos[963].bit_mask=create_bit_mask(41,0xa2002082,0x28a20a20,0x8a200088,0x820800a0,0x200a08,0x88088280,0x20a0822,0x20020a00,0x80000008,0x2882022,0x8820028,0xa28a280,0xa0082882,0x8200828,0x80202200,0x200a08a2,0x220820,0x2288002,0x80080002,0x20828800,0x2282080,0x20020,0x228808,0x82088008,0xa20a0080,0x8808028,0x82008200,0x228200a2,0x28a28a08,0x80082200,0x80822802,0x28,0xa080082,0x820a0800,0x28220228,0xa008000,0xa0020822,0x20008a20,0x8a08a282,0x208008a0,0x8);
remainders_neg[963].modulo_value=1284; remainders_neg[963].bit_mask=create_bit_mask(40,0x22882802,0x20228228,0x2288000,0x2800020,0x8a08200,0x208,0x828200a2,0x28828208,0x8088080,0x820028a2,0x28820,0x82202208,0x20802002,0xa08020,0x828a288,0xa0820022,0x8228028,0x8a20008a,0x800882,0x28020008,0x8a20a008,0x808a08a0,0x8a20000,0xa000080,0x22820800,0x820,0xa080288,0xa20a0822,0x20220200,0x800a288,0xa2082,0x8808820,0x8200800a,0x2820080,0x20a28a20,0x82080088,0x208a00a2,0x28800228,0x200220a,0xa0080020);
remainders_pos[964].modulo_value=1928; remainders_pos[964].bit_mask=create_bit_mask(61,0x2a00822a,0x2822a200,0x88aa808,0x2a820802,0x808aa0a0,0x88a20202,0xa20a2202,0x2aa8aaa2,0x222822a,0x202288a,0x282a880a,0x820aa0,0x80aa8882,0x22a20a0,0xa20802a0,0xa80208aa,0xa08a8800,0x222aa020,0xaa082008,0x22a8280,0x2288080a,0x8828880a,0xaaa2aa8a,0x88a08a8,0x808a228,0xa0aa2028,0x2082a80,0x2aa2208,0x8a88282,0x88200a80,0xa00822aa,0x822a2002,0x88aa8082,0xa8208020,0x8aa0a02,0x8a202028,0x20a22028,0xaa8aaa2a,0x222822a2,0x202288a0,0x82a880a0,0x820aa02,0xaa88820,0x22a20a08,0x20802a00,0x80208aaa,0x8a8800a,0x22aa020a,0xa0820082,0x22a8280a,0x288080a0,0x828880a2,0xaa2aa8a8,0x88a08a8a,0x808a2280,0xaa20280,0x2082a80a,0x2aa22080,0x8a882820,0x8200a800,0xa8);
remainders_neg[964].modulo_value=1928; remainders_neg[964].bit_mask=create_bit_mask(61,0xa2880aa2,0xa0aa2a88,0x80022080,0xa20a808a,0x8022828,0x2a8a8a,0x2a82aa8a,0xa220222a,0x8aaa0aa2,0x8a8aa002,0xa0a20082,0x880a8228,0x822000a,0x8aa2a828,0x2a808a28,0x208a8022,0x28020088,0xaaa228a8,0x2280a880,0x8aa20a08,0xaa008082,0xa00082,0x22222202,0x80028020,0x80802aa0,0x2822a8a0,0x8a80a208,0x8a22aa80,0x80200a0a,0xa88208,0x2880aa22,0xaa2a88a,0x22080a,0x20a808a8,0x8022828a,0x2a8a8a0,0xa82aa8a0,0x220222a2,0xaaa0aa2a,0xa8aa0028,0xa200828,0x80a8228a,0x822000a8,0xaa2a8280,0xa808a288,0x8a80222,0x80200882,0xaa228a82,0x280a880a,0xaa20a082,0xa0080828,0xa00082a,0x22222020,0x280202,0x802aa08,0x822a8a08,0xa80a2082,0xa22aa808,0x200a0a8,0xa882088,0x20);
remainders_pos[965].modulo_value=965; remainders_pos[965].bit_mask=create_bit_mask(31,0x81636252,0xf72c072,0x1d66231,0xde2220e1,0x4391487a,0xb5ec22ca,0xe17b0ad4,0x2202f339,0xcc970881,0xc798265a,0x19412912,0x86082a7,0x5b182421,0x8a762880,0xa2b0e21a,0x11c35152,0x451b9456,0x9063680,0x50418421,0x2520a639,0x990678d2,0x443a4cd6,0x33d01120,0xd437a1e7,0xd10deb4a,0x84a27094,0xc1111ed7,0x119ae021,0x80d3bc23,0x91b1a053,0x12);
remainders_neg[965].modulo_value=3860; remainders_neg[965].bit_mask=create_bit_mask(120,0x8aa2a8a,0x822a08a2,0x80aa2a8,0x2a2a828,0x82088022,0x28202a02,0x28228200,0x2a8a22a0,0x28008,0x200aa02,0x8088208a,0x80280a22,0x280a0a8,0x2a2a008,0x22282a02,0x880a088a,0x88020882,0x8aa208,0x888008a0,0xa0a82a20,0x8a200a,0x88228002,0x2a0880a8,0x8a2282a,0x802202,0x82a2808,0x8088202,0x80220a8,0x80220aa,0x28280a,0xa0a0028a,0x22008280,0x220082a8,0x88080a8,0xa2a08202,0x20080080,0xa2288202,0x882a2a0,0xa2088a8,0x22880200,0xa0a82a80,0x80088822,0x2a880028,0x82008882,0x82808a08,0xa0a22288,0x2a2a0202,0x280a0080,0x80a008a8,0x20880a22,0xa8020288,0xa000202,0x228aa080,0xa20a02a,0xa020a002,0x8820a02,0xaa2a0220,0x2a8080a0,0x82a208aa,0xa2a88228,0x28a,0x2080280,0xa0,0x2088a,0x80a82220,0x28022082,0x20228202,0x2002208,0x80a28800,0xa000a22a,0x88280282,0xa0200a0,0x280800a8,0x2288a0,0x200a220,0x80880202,0xa2088082,0x28088,0x8a8008a8,0x2002808,0x802a0008,0x202a82a0,0x8880088,0x2020002,0xa220a208,0xa0a88,0x88228282,0x22080288,0x8220008a,0x208820a,0x28a0028a,0xa8208820,0x28800220,0x88a00822,0x20a0a208,0x88a82800,0x8228222,0x20002020,0x8800888,0x82a0aa02,0x8002a00,0x880a0020,0xa8800a8,0x8880a000,0xa0808822,0x20200880,0x2228020,0x288a200,0xa80080a,0x82802028,0xa0a00a08,0xaa228002,0x8a280,0x8220020,0x2020a202,0xa082200a,0x2220a80,0x28882000,0x2800000,0xa00820);
remainders_pos[966].modulo_value=3864; remainders_pos[966].bit_mask=create_bit_mask(121,0xa2082822,0x8800220,0x80008082,0x22022802,0x8028208,0x2000208,0x800a2022,0x828000,0xa088002,0xa20000a2,0x20200820,0x8220020a,0x8a2800,0x820020,0x88000008,0x880022,0xa00800,0x220008a,0xa0822800,0x820a20,0x8a20a,0x80002022,0x8200820,0x82208008,0xa00828a2,0x28208000,0x82202280,0x80020,0x28800028,0x2002202,0x2820022,0x28000020,0x8820a00a,0x20820,0x8200200,0x808020a,0x2000000,0x20208828,0xa288000,0x802002,0x820020,0x80200,0xa2020820,0x20008000,0x80088288,0xa2080022,0x20028808,0x82208,0xa0020080,0x28020020,0xa200200,0x80800000,0x20028,0x20828a,0x82822020,0xa08020,0x2002,0x28800a0,0x20808208,0x8000002,0x20000000,0x8800000,0x80208202,0xa0022,0x800800,0x8208020a,0x80882,0xa28208,0x2280080,0xa0000002,0x28008008,0xa080080,0x20080,0x8208820,0x8a202280,0x2880020,0x8228220,0x8a000200,0x82080,0x8020,0x80082,0xa0800802,0x8000228,0x80282208,0x20000000,0x20a08020,0x8008,0x22082080,0x28a00a08,0x80080000,0x80880082,0x28808800,0x280002,0x82080020,0x28028808,0xa000208,0x828a2820,0x20200208,0x2082008,0x880800,0xa08a20,0xa08a082,0x80002882,0xa20008,0x200a,0x22880022,0x200000,0x80082,0x820028a2,0x8a08008,0x8a082008,0xa08a0000,0x800220,0x2000282,0x808808a0,0x20208000,0x88208280,0x2802880,0x20820200,0x8a088002,0x882820);
remainders_neg[966].modulo_value=3864; remainders_neg[966].bit_mask=create_bit_mask(121,0x22002022,0x8a20,0x8000a,0xa2022082,0x820a00,0x8a080280,0x8220a2,0x8020808,0x2000082,0x22880022,0x28208028,0xa208282,0x80022000,0x8028028,0x88000,0x8a2,0x208008,0xa288002,0x20022080,0x8020228,0x80002282,0x882822,0xa08020,0x2280080,0x20002022,0x20a00808,0xa202208,0x808000a0,0x20008820,0xa082282,0x820a08a2,0x20008828,0x202082,0x808a00a0,0xa08208,0x80008202,0x2880880,0x20a00020,0x82200080,0x80002882,0x8020828,0x88008288,0x228a0020,0x20800800,0x200,0x220008a2,0x28020000,0x88002280,0x208a0000,0x20828828,0x2288280,0x80880,0x8020820,0x88200202,0x20a28a0,0x8200028,0x8808a00a,0x2000820,0x20000a00,0x8008808a,0xa0800880,0x800,0x828028a,0x820822,0x8008000,0x2008282,0x80000002,0x8220a00,0x8a200008,0x20080082,0x20800800,0x2008000,0x80820800,0x200028,0x220a208,0x820008a0,0xa20228,0x2088208,0x802000,0x800828,0x8800800a,0x20000002,0x800a20,0x820a280,0xa0880800,0x28200820,0x80080,0xa2002800,0x20208200,0x88,0x2,0x20000008,0x8208082,0x28008a0,0x20020000,0x82008280,0x20220a0,0x28a08200,0x8a002000,0x80,0x200228,0x8200200a,0x802002,0x8220800,0x8808a002,0xa2000822,0x8a08800,0x80008002,0x2082022,0x200800,0x2002080,0x20020080,0x8a28,0xa088202,0x20,0x28200808,0x200208,0x82082000,0x28028208,0x200000a,0x20a0);
remainders_pos[967].modulo_value=3868; remainders_pos[967].bit_mask=create_bit_mask(121,0xaaa028a,0xa2aa8800,0x28a2a,0x2820822,0xaa2aa222,0xa2a08aaa,0x808a20a,0x80aa02a,0x2208082,0xa008a0aa,0x20a8020a,0x82280a2a,0x8888a08a,0xaa282aa8,0xa8a22888,0x8820828a,0xa022280a,0xa8880a22,0xa000a0a,0xa822aa0,0x2820a,0x888a8a,0x2a88a088,0x802a2008,0xa0a02008,0x88822a00,0x2008a82,0xa02aa20a,0xa0228a8,0xa280208a,0xaa08020a,0x8a828228,0xa82020,0x2a8a8288,0x8a220a02,0xa8282a80,0x80aa22a8,0x2200a2a8,0x222aa228,0xaaa8202,0x8a02828a,0xaa82a800,0xa2200282,0x88a80882,0x28a202a0,0x8820020a,0xa0802220,0x22222000,0xa08a0228,0xa0288082,0x22a8028a,0x288a8028,0x2a28a2a,0x22a280a8,0x28080288,0x800002,0x8a0a8888,0xaaa88a2,0x80082,0x8002aaa2,0x2a08020a,0x8aaa800,0x8200200,0x28a22aaa,0x2222a0a,0x88000020,0x82280202,0x2a028a8,0x8a8a28a8,0x82802a22,0xaa2802a8,0xa8202280,0x82880a20,0xa0008888,0x20888020,0x8a080082,0x20a808a2,0xa82202a2,0xa8280088,0x2002a82a,0xa28280a,0x88082aaa,0x8288aa88,0x22a8a008,0xa2a88aa0,0x202a8282,0x880a088a,0x2282a2a,0x208082a0,0xa288282a,0xaa08020a,0xa208028,0xa2a2880a,0xa08aa80,0x282a2008,0xa00a8822,0x220080a0,0x82008a80,0x220a22a,0xa2a2220,0xa082800,0xaa882a,0xaa0a000a,0xa88a0222,0x2a028880,0xaa282082,0xa22288a2,0x22aa828a,0x2a20a222,0x8a8a0288,0xaa0802a0,0xaa0a200,0x8202088,0xa80aa02,0xaa08a202,0xaaaa20a8,0x888aa8a,0x8820828,0xaa8a2800,0x22aa8,0xa280aaa);
remainders_neg[967].modulo_value=3868; remainders_neg[967].bit_mask=create_bit_mask(121,0x82228a02,0x2a220088,0x888a02a2,0x8a0a80aa,0x22a22aaa,0x2a280222,0x80802a82,0x808228a2,0x8aa8080a,0x28802822,0xa8208a82,0xaa082a2,0x2802,0x22a0a220,0x202aa000,0xa80a02,0x28aaa082,0x200082aa,0x82888282,0x820aa228,0x888a0a82,0x88000202,0xa2002800,0x8a2a880,0x2828a880,0xaa288,0x8a88020a,0x28a22a82,0x828aa020,0x2a08a802,0x22808a02,0x20a0aa0,0x8820a8a8,0xa2020a00,0x2aa828a,0x20a0a208,0x822aa20,0xaa882a20,0xaaa22aa0,0x82220a8a,0x28a0a02,0x220a2088,0x2aa88a0a,0x20800a,0xa02a8a28,0xa88a82,0x2808aaa8,0xaaaaa888,0x28028aa0,0x28a0080a,0xaa208a02,0xa00208a0,0x8a2a02a2,0xaa2a0820,0xa0808a00,0x8808888a,0x2820000,0x8222002a,0x8880880a,0x88a222a,0xa2808a82,0x80222088,0x80a88a88,0xa02aa222,0x8aaaa282,0x8888a8,0xaa08a8a,0x8a28a020,0x202a020,0xa08a2aa,0x22a08a20,0x20a8aa08,0xa0082a8,0x28880000,0xa80008a8,0x280880a,0xa820802a,0x20aa8a2a,0x20a08800,0xa88a20a2,0x82a0a082,0x80a222,0xa002200,0xaa202880,0x2a200228,0xa8a20a0a,0x828002,0x8aa0a2a2,0xa8080a28,0x2a00a0a2,0x22808a82,0x82a808a0,0x2a2a0082,0x82802208,0xa0a2a880,0x288200aa,0xaa880828,0xa880208,0x8aa82aa2,0x82a2aaa8,0x8280a088,0x882200a2,0x22828882,0x20028aaa,0xa28a0008,0x22a0a80a,0x2aaa002a,0xaa220a02,0xa2a82aaa,0x2028a00,0x22808a28,0x82282a88,0x80a8a800,0x8208228a,0x22802a8a,0x2222a820,0x80002202,0x800a80a0,0x2202a088,0x888aa220,0x2a08222);
remainders_pos[968].modulo_value=88; remainders_pos[968].bit_mask=create_bit_mask(3,0x82828282,0x82028280,0x828282);
remainders_neg[968].modulo_value=88; remainders_neg[968].bit_mask=create_bit_mask(3,0xa0a020a,0xa0a0a08,0xa0a0a);
remainders_pos[969].modulo_value=3876; remainders_pos[969].bit_mask=create_bit_mask(122,0x82800822,0x8a20820,0xa008000,0x22082002,0x28000008,0x2280200,0x880822,0x20220220,0x82088202,0x8000a0,0x28220200,0x8a000000,0x800800,0x20800808,0x82208280,0xa08200a0,0x8220000,0x8208000a,0xa28000a2,0x820228,0x8a200,0x2,0x20828208,0x20a008,0x228a00a2,0x200a00,0x88082200,0x820a20a0,0x28800a00,0x82280202,0xa20a28a2,0x8a20200,0x82080280,0x80882822,0x20000220,0x8a200280,0x20022802,0x8208028,0x820a,0x80020000,0x8000228,0xa28a202,0x820a2800,0x20828000,0x8280002,0x22082802,0x8200a08,0x8008080,0x2880080,0x20a00000,0x8000202,0x82082800,0x20220208,0x88002202,0xa2022080,0xa00200,0x82208000,0x2820020,0x28800008,0xa082082,0x208a2080,0x20820a00,0x2288,0x80082802,0x208820,0x800a000,0x208808a0,0x8800220,0x8088088,0x2820822,0x8000200,0xa088,0x20022800,0x20200200,0xa008200,0x2820882,0x28208,0x282088,0x28a0820,0x8a28a00,0x88000208,0x80022,0x8200000,0x8020820a,0x2880082,0x28008a28,0x88000080,0x82822020,0x28020828,0x808a200,0xa28a08a0,0x8028828,0xa002288,0xa08a0820,0x8820220,0xa008000,0xa00a2880,0xa08008,0x8282082,0x2,0x8a20008,0x88082000,0xa00028a2,0x20828,0x8820a,0xa00820a0,0x28208820,0x2002080,0x2002002,0xa20,0x8088280,0xa0002000,0x8220820,0x88088088,0x82022000,0x8028808,0x280,0x820882,0x200a08,0x8208a200,0x82002820,0x8);
remainders_neg[969].modulo_value=3876; remainders_neg[969].bit_mask=create_bit_mask(120,0x20000a2,0x228028,0x2080088,0xa2002882,0x20808800,0x82208288,0x800000a2,0x28228a28,0xa00028a,0x80080820,0x20a28a00,0x2088088,0x80080000,0x28008000,0x2280208,0x200a0820,0xa20808,0xa008002,0x22000822,0x8028a20,0x88002288,0x880882,0x20020a00,0x88282080,0x22020022,0x8a08208,0xa288,0x2822820,0x20008208,0xa208202,0x22822022,0x220208,0xa008208,0x20a2,0x28808a28,0x2208208,0xa08a2002,0x200820,0x88000282,0x8a0880,0x808a20,0x220228a,0x2822080,0x28020008,0x80208082,0xa2002082,0xa08200,0x80080008,0x82000000,0x28208808,0x8008028a,0x2802080,0x20a28a00,0x8a28a,0x220a2800,0x8208a08,0x2280088,0x820208a0,0x20008800,0x200a00a,0xa0022800,0x28028200,0x8808a200,0x802002,0x8200028,0x80082088,0xa0000020,0xa20,0x80000000,0x820200a2,0x808a08,0x80082000,0xa08a2080,0x20a08a08,0x82080208,0x20a0002,0x8820a00,0x8820a000,0x20200a0,0x220208,0x88200,0x808000a2,0xa08008,0x8280282,0x82000802,0x20800220,0x8000,0x20a28a0,0x20020020,0x80002288,0x22020020,0x820020,0x208a200,0x20020020,0x28a28,0x82000080,0x20822000,0x8200800,0x8020a00a,0x80880082,0x228000,0x2088,0x20082022,0x8828020,0x88000282,0x20002820,0x20a00028,0x82082008,0x82882882,0x808228,0x80000208,0x20882880,0xa28020,0x2822800,0x820000,0x8088208,0x800a0002,0x8a00200,0xa002288,0x28020a0);
remainders_pos[970].modulo_value=3880; remainders_pos[970].bit_mask=create_bit_mask(122,0xa8aa028a,0x8a2a0800,0x22828020,0x82800000,0x82a88a2a,0xa02a0a0a,0x82088820,0x80a2208,0x88008a2,0x82022,0x20820808,0xa0088a20,0x2880a288,0x8222a002,0xa0080a2a,0x808220a,0xa22a8222,0x20802288,0xa002,0x280a82a,0xa8080808,0x8000a20,0x2a0802a0,0x2800a,0x228a028,0x22020a,0x88028222,0x282a088,0xa2200a8,0x82202008,0x20822202,0x20288a22,0x28802a0,0x88808002,0x22a02202,0x20a02282,0xaa280a00,0x2000a280,0x88a200a8,0xa008002a,0x80a82888,0xaa200802,0xa02a208,0x200802a,0xa0a02200,0xa8002208,0x220000a0,0x802a200,0x8a00800a,0x82282008,0xa8a82882,0x88200a20,0x288200a8,0x8280a020,0x22082808,0x802080,0x82000080,0x202a2a8,0x822882,0x8a820,0x80000280,0x2a200002,0x28820008,0x8a808082,0x822a,0x8020002,0x28208802,0xa028220,0x822808,0xa008222a,0x282a2a08,0x8288282,0x200a220,0x8a8020a0,0x8800,0x88002a0a,0x880a0a20,0x2008000,0x8a80a0a8,0x2008aa20,0x282a0280,0x200a22,0x8a22a8,0x8a00082a,0xa028aa02,0x880a0800,0x880a8882,0x2022280,0x80228080,0xa228080a,0x88820888,0x8088280,0x88a020,0xa82802a,0x82802222,0x80880088,0xa2880a0,0x2800028,0x8020a8a0,0xa000200a,0x20202a08,0x2a028020,0xa0000a8,0x88020880,0x82a88a22,0x88202088,0xa0200aa0,0xa8882a8,0x8a022880,0xa2200a22,0x20820808,0x8200020,0x20022088,0x88a0208a,0x22208220,0xa0a80a08,0xa22a82a0,0x282a8,0x2828800,0x20a8a208,0x80aa2a00,0xa2);
remainders_neg[970].modulo_value=3880; remainders_neg[970].bit_mask=create_bit_mask(122,0x20220a02,0x2228080,0x2a0a00a8,0xa008888,0x2200222,0x28a20282,0xa0000a0,0x82a280,0x8000802a,0x8080a822,0xa80a0080,0x280002a0,0x20082200,0xaa2288a,0x20808222,0x80802a82,0x2a220aa2,0x2808a200,0x8880288a,0x82082022,0x20800080,0x80088220,0x22808228,0x88820882,0x82a02820,0x88aa0a82,0xa0aa2,0xa0a2000,0x82a28820,0x2a8a800,0xa80a2a8a,0xa82002a2,0xa008228,0x8a,0xa228aa0a,0xa8282a0a,0x22208280,0x28882208,0x228820,0x20808822,0x8202000,0x22288082,0x28a2280,0x8a8008a2,0x2028aa08,0x20882a80,0xaa088820,0x8a2288,0x2800882,0x2a0a800,0x2020200a,0x2882a0,0x200a8020,0xa0028a8,0xa280a000,0x88082808,0xa088800,0xa822220,0x8802a00a,0x80802028,0x8880a08,0xa2288882,0x200a8080,0x200080a,0x80880a22,0x808a088a,0xa0280082,0x28a02a8,0x8802a080,0x2000aa22,0xa0a22280,0x80200a02,0xa8822a8,0x200a828,0x80880008,0x882282,0x282a0,0xa880088,0x2002820,0xa0802228,0xa0a20a08,0x802882a2,0x802a220,0x28080a2,0x20a0220a,0x820088,0x20002,0xa8aa208,0x8a20808,0x22a08002,0xa0000,0x80000a00,0x80020a8,0x20208a2,0x208aa2a,0x8000800,0x82200820,0xa0880a0,0x8a02028,0x2088a802,0xa8a82280,0xa20a08a0,0x2888020,0x828008,0x220022a,0xa82800,0x28288220,0x2000220,0x282a008,0x22a8822a,0xa80a0080,0x802888a0,0x288aa000,0x20a802,0xa2a80a28,0x28200280,0x2a220a20,0x88a0220,0x8a020008,0xa0202a00,0x8222288,0x22);
remainders_pos[971].modulo_value=3884; remainders_pos[971].bit_mask=create_bit_mask(122,0x82222a22,0x2aa22820,0x2a280aa,0xaa20020a,0xa00a820,0xaa08220a,0x28800882,0xa2080aa,0xa28aa202,0x2228082,0x2a28a2a2,0x882082aa,0xaaa0a28,0xa2aa8200,0x22802220,0x80a880a2,0x880aaa02,0x800a0a0,0xaa0282a0,0xaaaa2880,0xaa002a88,0x28820020,0x2a8a88aa,0xaaa020aa,0x2a0aa80,0x88a08222,0x8a022202,0xa88aa2a2,0x2aa088a2,0xaa82a2a2,0x802a2,0x88080a,0x20082228,0x82080802,0x220a888a,0xa888a28,0x2a0028,0x8028a8,0x20802202,0x8028aa8a,0x2200aa,0x802a2080,0xa2a80a0a,0xa228282a,0x2a0a8002,0x8882a20,0x82888aa,0x2aa8a00,0xa2208280,0x80800a28,0x8a880820,0x80a2a08,0x828a8802,0x20802a28,0x800a22aa,0x828288a2,0x8028a02a,0xa828aa8,0x802a08,0x8a28a088,0x22088088,0xa28a2220,0x80200220,0x282a020a,0xa28022aa,0x28282a80,0x8a0028a2,0x80208aa8,0x2a28228a,0x8a020802,0x22a220a,0x202082,0x2088a28a,0x2aa80028,0x2282000a,0x82220aa2,0x2a0a808a,0x8288a800,0x2a8aa82,0x8a802a0a,0x8800020,0xa2802aa0,0x80208a2a,0x80200808,0xa8002a2,0x222a0280,0x2a08828a,0x2082a22,0x82008802,0x2200288,0x2000a02,0xa82aa8a8,0x20aa88a8,0xaa22a8a2,0x880a28a8,0x20a22808,0xa0a80888,0x80aaa02a,0x2a2a8aa0,0x8228aa2,0x800aa080,0x8aaaa22a,0x280aa022,0xa00200a8,0xaa0220a0,0x22a0280a,0x802888a0,0x2aa8a088,0xaaa0008,0x2082228a,0xa28a8aa8,0x288808a8,0xaa28a820,0x208a0808,0x228aa0,0x820aa822,0xa00a0a08,0x8aa082,0x28a80a08,0x88aa8aa0,0x88882082,0x88a);
remainders_neg[971].modulo_value=3884; remainders_neg[971].bit_mask=create_bit_mask(122,0xaaaa2aa,0xa22aa0a8,0x8a2a0822,0x22a88a82,0x828820a8,0x2280aa82,0xa008800a,0x82a80822,0x2a022a8a,0x8aaa080a,0xa2a02a2a,0xa80a22,0x822282a0,0x2a220a88,0xaa08aaa8,0x820082a,0x82228a,0x80882828,0x228a0a28,0x2222a008,0x2288a200,0xa00a88a8,0xa2020022,0x2228a822,0x8a282208,0x280aaa,0x28aaa8a,0x20022a2a,0xa228002a,0x220a2a2a,0x8880822a,0x88008082,0xa880aaa0,0xa80808a,0xaa820002,0x820002a0,0x88a288a0,0x8808a020,0xa808aa8a,0x8a02202,0x88aa8822,0x8a2a808,0x2a208282,0x2aa0a0a2,0xa282088a,0x8000a2a8,0x80a00022,0x8a220288,0x2aa80a08,0x80882a0,0x20080a8,0x8082a280,0xa02008a,0xa808a2a0,0x882aa22,0xa0a002a,0x8a028a2,0x820a0220,0x8808a280,0x2a02800,0xaa800800,0x2a02aaa8,0x8a88aa8,0xa0a28a82,0x2a08aa22,0xa0a0a208,0x288a02a,0x8a80220,0xa2a0aa02,0x28a808a,0x8aa2aa82,0x88a8a80a,0xa8002a02,0xa22088a0,0xaa0a8882,0xaaa822a,0xa2820802,0xa002088,0x8a20220a,0x208a282,0x800888a8,0x2a08a228,0x8a802a2,0x8a88080,0x82088a2a,0xaaa28a08,0xa2800a02,0x8a80a2aa,0xa88008a,0x8aa88a00,0x8a88828a,0x20a22020,0xa8220020,0x22aa202a,0x82a020,0xa82aa080,0x28208000,0x82228a2,0xa2a20228,0x80aa022a,0x8822808,0x2222aa2,0xa08228aa,0x288a8820,0x228aa828,0xaa28a082,0x8a00028,0xa2202800,0x82228880,0xa80aaa02,0x2a020220,0xa0008020,0x22a020a8,0xa8028080,0x88aa0228,0xa8220aa,0x28828280,0x8802280a,0xa0208280,0x220228,0xa80a,0x2);
remainders_pos[972].modulo_value=24; remainders_pos[972].bit_mask=create_bit_mask(1,0x802802);
remainders_neg[972].modulo_value=24; remainders_neg[972].bit_mask=create_bit_mask(1,0x82082);
remainders_pos[973].modulo_value=3892; remainders_pos[973].bit_mask=create_bit_mask(122,0x2a0a0a0a,0x2a080022,0x828a8a,0x820a88aa,0x2a000200,0x88820a08,0x88a2228a,0xaa82a,0x8a2a0,0x80a08a2a,0xa08200a,0xa0000a82,0x88882000,0x20800228,0x888a8800,0x28002200,0x202a0202,0x2202a0a2,0x820020a,0x802a00a8,0xa0228228,0xa00a8a0a,0x2a82a82,0x8020a,0x2828a002,0x22002088,0x82200208,0x8888822a,0xa000a202,0x20088222,0x2282020a,0x2a220880,0x8a022800,0xaa20888,0x20820022,0xa0882002,0x20028a0,0x2820080,0x2a0aa0aa,0x2a828a80,0x8a20a20,0x80a802a0,0x22820020,0x22282a02,0xa20202a0,0x88022000,0x20008a88,0x88a20008,0x28002088,0x820a8000,0xa802082,0x2a28828,0x2a2880,0x8aa0aa80,0x8a8a2228,0xa0828208,0x8020002,0x2a88a82,0xa28a8a08,0xa2200082,0x2a828282,0x88a82828,0x28a82000,0xa8020a2a,0x2082a22,0x20a80008,0x2a220828,0xaa22888a,0x80002aa0,0xa800228a,0x2a028228,0x8282080,0x280002a,0xa2222080,0x820008,0x2222a20,0x8a00088,0x8880a808,0x28880a82,0xa0208008,0xa200a802,0x2a808a08,0xa802a28,0x280aa0aa,0x8002008,0x20a0a280,0x20880082,0xaa088008,0xa222208,0x8a800288,0x28802208,0x8a0808,0xa88822,0x222808a0,0x882a8822,0x8820800,0x82822080,0x800a2,0xa80a0802,0xa82a82,0x80aa0a2a,0x80228828,0x8202a00a,0x88a0800,0x8088a0a8,0x288080a,0x22200880,0x2080022a,0x22228800,0xa00082,0xa082a00,0xa02a0082,0xa8a20,0xa8a2,0xa22a82aa,0x222a2888,0xa820a08,0x8200800,0x200aa22a,0xa8a2a28,0xa888002,0xa0a0a);
remainders_neg[973].modulo_value=3892; remainders_neg[973].bit_mask=create_bit_mask(122,0xa2828202,0x228088a2,0x800a0202,0xa020022,0xa2808288,0xa0280,0x2aa202,0x88822022,0x8802a20,0x2802a2,0x8200a882,0x2880820a,0x2888,0xa80882a0,0x20008,0x2088aa80,0xa0a28a8a,0xaa0a282a,0x80a08a82,0x8a20820,0x28aa02a0,0x28020202,0xa20a202,0x80808a82,0xa020288a,0xaa80a800,0xaa80a80,0x2a2,0x28882a0a,0x28800aa2,0xa20a8a82,0xa22a8008,0x282a088,0x822a0000,0xa80a80aa,0x2800a80a,0xa88a020,0x820a8808,0xa2022822,0xa2020200,0x802a02a8,0x8208228,0xaa0a8828,0x2aa0a282,0x228a8a28,0xaa888,0xa8800200,0x2a0880,0xa088a000,0xa820808,0x208a802,0x822a00a0,0x8822a008,0x2202208,0x2022aa0,0x280a0280,0x808a080a,0xa200202,0x22020280,0x2a28880a,0xa2020a0a,0x2020a0,0xa020a088,0x208a8222,0xa80a2a2,0xa0208880,0xa22a80a0,0x22a20002,0x8882228,0x2088a202,0xa28a0a20,0xa0a800,0x820888a2,0x222aa808,0x88028880,0x8aaa22a8,0x80288000,0x82000,0x20008202,0x20a80880,0x2a08208a,0xa2000280,0x820822a0,0xa0822022,0x8088a800,0x28282a00,0xa000880a,0x22000880,0x82a2aa80,0x2080a00,0xa008a280,0x88028000,0x82000a2,0xa2a08028,0x2200aa,0x80028088,0xa0a2808,0x8880802a,0x2082800a,0x820a202,0x2282a2,0x82a00a0,0xa822882,0x80020088,0x8002020,0x8a008002,0x2aa88000,0xa0088aa2,0xaa220088,0x8820880a,0x82802288,0x28a2800a,0x88820228,0x8882022,0x22a20a22,0xaa22a000,0x82028280,0x80a80088,0xa88222a2,0x8202a220,0x2000882,0x28282);
remainders_pos[974].modulo_value=3896; remainders_pos[974].bit_mask=create_bit_mask(122,0x28822a2,0x800aa22a,0xaa8a820a,0x22282822,0x80a00a22,0xa82aa22,0x8020a082,0x88a288a2,0x80000008,0xaa280a22,0x20aaa220,0x2802a2a0,0x80220a20,0x2a2a20a8,0x8aaa0880,0xa2202a22,0xa2a28800,0xaa88020a,0x28828aa0,0x220a8a2a,0x82a20aaa,0x88882aa0,0x2a000088,0xa882a00,0x2820a80a,0x8a082aa2,0x82aaaaa0,0x20222a82,0x208820a8,0x8028a2aa,0x80202a2a,0x2002082a,0x280a288a,0x8280222a,0x8a000002,0x220028a0,0xa0a80a28,0x2aa02880,0x888aaaa0,0x8a002888,0x2000a202,0x202080a2,0xa008288,0xa2a880,0xaa88202,0x82002a22,0x2a88a000,0x880a2020,0x2a20a22a,0x2a0202a8,0xa22000a,0x8220a820,0x88aaaaaa,0x82088208,0xa28a0a2a,0x82200280,0x2220aa0a,0x2282822,0x80a28080,0xa02200aa,0x82022882,0xa822880,0x2aa0088,0x8002028a,0x88882828,0x82a0aa08,0x8a028008,0x82a8a0a2,0x22208220,0x82aaaaaa,0xa0082a08,0xa8a00088,0xa82a8080,0x22a88a08,0xa80808a0,0x82000a22,0xa088a800,0x80822a,0xa0022a8a,0x8228200,0x88a0208,0xa2808a00,0x22222800,0xa80aaaa2,0xa02280a,0x8828a02a,0xa20a2800,0x82800000,0x28a88802,0x8a228a0,0x2a82080,0x2a8a808,0x8aa8a28,0x82a0822,0x8282a888,0xa20aaaaa,0x288aa820,0xa0a02a08,0xa800a822,0x22220000,0x820aa822,0x88aaa08a,0x28a8a2a0,0xaa0aa282,0x8aa08022,0x8a00228a,0xa288a808,0xa80220aa,0x22a08a8,0x2808a088,0x80a8a80,0xaa088aaa,0x288a028,0x22200000,0x28a228a,0xa0820a08,0x288aa82,0x8888a00a,0xaa882828,0x2a082a2,0x80a88aa0,0x8a8822);
remainders_neg[974].modulo_value=3896; remainders_neg[974].bit_mask=create_bit_mask(122,0x8a00aa2a,0x8822aa2,0x22020a82,0xaaa0a0aa,0x82882aa,0x820a22aa,0x8a8280a,0x2a002a,0x8888880,0x22a082aa,0xa8222aa8,0xa08a2a28,0x8aa82a8,0xa2a2a820,0x2228008,0x2aa8a22a,0x2a2a0088,0x22008a82,0xa00a0228,0xaa8202a2,0xa2a8222,0xa228,0xa2888800,0x8200a288,0xa0a82082,0x280a22a,0xa222228,0xa8aaa20a,0xa800a820,0x8a02a22,0x8a8a2a2,0xa88a80a2,0xa082a002,0xa08aaa2,0x288888a,0xaa88a028,0x282082a0,0xa228a008,0x22228,0x288a000,0xa8882a8a,0xa8a8082a,0x82880a00,0x882a2008,0x82200a8a,0xa88a2aa,0xa2002888,0x82a8a8,0xa2a82aa2,0xa28a8a20,0x82aa8882,0xaa820a8,0x222222,0xa800a80,0x2a0282a2,0xaa88a08,0xaaa82282,0x8aa0a0aa,0x82a0808,0x28aa8822,0xa8aa00a,0x820aa008,0x8a228800,0x88a8a02,0xa0a0,0xa282280,0x28a0880,0xa20282a,0xaaa80aa8,0xa222222,0x2880a280,0x20288800,0x20a20808,0xaa200280,0x20808028,0xa8882aa,0x28002080,0x88080aa2,0x288aa202,0x80aa0a88,0x80028a80,0x2a080288,0xaaaaa088,0x2082222a,0x828aa082,0xa028a2,0x2a82a088,0xa088888,0xa020008a,0x802aa028,0x8a20a808,0x8a202080,0x802202a0,0x80a280aa,0xa0a2000,0x2a822222,0xa00220a8,0x2828a280,0x208820aa,0xaaaa8888,0xa8220aa,0x222802,0xa0202a28,0x22822a0a,0x22808aa,0x288aa02,0x2a002080,0x208aa822,0x8aa28020,0xa0802800,0x80820208,0x22800222,0x8a0028a0,0xaaa88888,0x8a02aa02,0x280a8280,0x8a00220a,0x2882,0x2200a0a0,0x8a280a2a,0x8200228,0x200aa);
remainders_pos[975].modulo_value=780; remainders_pos[975].bit_mask=create_bit_mask(25,0x80880082,0x20020200,0x2000008,0x2000800,0x20a00228,0x8828000a,0x82000020,0x8000000,0x82208080,0xa2880022,0x8820800,0x8200202,0x820000,0x8220808,0x228a002,0x20882880,0x20020,0x82000,0x2822080,0xa08a00,0x80288,0x80002,0x8008200,0x2202008,0x820);
remainders_neg[975].modulo_value=780; remainders_neg[975].bit_mask=create_bit_mask(25,0x802,0x28028a00,0x2088080,0x82800080,0x20200a20,0x200082,0x20808a0,0x8008,0xa200008,0x22000022,0x20008,0x80288282,0x20080,0xa28000,0x8220200a,0xa0002000,0x8028820,0x8002088,0x82022800,0x200200,0x88000200,0x80000882,0x800208,0xa20a080,0x20);
remainders_pos[976].modulo_value=488; remainders_pos[976].bit_mask=create_bit_mask(16,0xa08a22a,0x202a280,0x800a2a02,0xaaa22882,0x282288,0x8080a8a,0xa0028a8,0x22aa88a2,0x2800a08a,0xa020202a,0x882800a2,0x288aaa22,0xa8a00282,0x8a808080,0x8a20a002,0xa8);
remainders_neg[976].modulo_value=488; remainders_neg[976].bit_mask=create_bit_mask(16,0x82802aa2,0x8a8a2a08,0x882a28a,0x222aa00a,0x88a0aa00,0x80008202,0x8288a020,0xaa22002a,0xa0882802,0x28a8a8a2,0xa0882a,0xa00222aa,0x20288a0a,0x2080008,0x2a8288a,0x20);
remainders_pos[977].modulo_value=977; remainders_pos[977].bit_mask=create_bit_mask(31,0xf247cb96,0xf7069895,0xc34a61bb,0xdf9daabc,0xb6014fcd,0x5a0f9264,0x4e446d52,0xd1d5617b,0x10fff271,0xe714288b,0xe1ac1610,0x19c42257,0x1653990c,0x92febab8,0x162b17e5,0x51a3733b,0xfd269fa3,0x29a07575,0x8e60c267,0xd61fa910,0xa39c21a0,0xfc234450,0xae2e393f,0x89cb7a1a,0xc1692ad8,0x1b49927,0xe7eecfca,0x4b0cf556,0x83bf7619,0x893ea465,0x1a74f);
remainders_neg[977].modulo_value=3908; remainders_neg[977].bit_mask=create_bit_mask(122,0x2a8a020a,0x2a8a0008,0xa82a822,0x2002220,0x2a888200,0x82820aa8,0x8288a08a,0x808a8a2,0x88222aa8,0x2a88a002,0x28208a88,0x8008aa8a,0x8a8a0080,0xa222220,0x8aa28a28,0x882aaaa2,0x20aa022a,0xa028a8a8,0x2a80aaa,0xa822088,0x2a00a828,0x20aa8888,0x22a2a0a2,0x2a282,0x8a0a200,0x882800aa,0x2a220202,0x8280288a,0xa22aa80,0xa228a8,0xaa02a82,0xa8a22802,0xaaa2280,0x88a00a08,0x20222a2,0xa800a08a,0x2282882,0xa2a0000,0x282a2a22,0x888aa822,0xa0a802a0,0x88220a80,0xaa80aa00,0xa8a8a02a,0xa202a820,0x2aaaa08a,0xa28a2a8a,0x22222280,0x8028a88,0x8aa88008,0x8a8820a2,0x288aa0,0xaaa2208a,0x28a88080,0x88288a0a,0xaa820a0a,0x2088aa0,0x22200200,0x20aa0a80,0x80028aa2,0x82028aa0,0x82a282,0x82aa8a,0x80a00a22,0xa2aa222a,0x88a2aa,0xa0a2800a,0xa0882880,0x8a880820,0x8a22000a,0x882aa0,0xa28808a,0xaa880082,0x8082a8a8,0x82222228,0x8020820a,0x8a000020,0x2802a200,0x2a08080a,0xa00a8000,0x80a2288a,0x2a80a0a,0x2800888a,0x20202822,0xaaa020a2,0x882822aa,0x8a0aa802,0x222a2a0,0xa0aa0882,0x822000a8,0xa822080a,0x802200a2,0x8220aa2,0x800220a8,0x882a828a,0xa2a22020,0xaa82882,0xa20a08a0,0x8202aaaa,0xa020222,0x88800a22,0x280aa028,0x8a2280a8,0xa802a8,0x8082a00,0x22a00a28,0x2880,0x20820082,0x222220a8,0x8aa0808a,0x8008aa8a,0x808a2820,0xaa088028,0x222882,0x808a8a8,0x8a088282,0xa28280,0xa2888028,0x222aa2aa,0x280280aa,0xaaa08022,0xa2a08028);
remainders_pos[978].modulo_value=3912; remainders_pos[978].bit_mask=create_bit_mask(123,0x28a2002,0xa20028,0x2082008,0x2002002,0x28820a08,0x80282200,0xa28828a0,0x8808a28,0x88200002,0x82822882,0x28200a28,0xa20a,0x22020880,0xa08a08,0x8a00228a,0x82020082,0x20820228,0x8a280200,0x80880002,0x28220a20,0x82002208,0xa0020022,0x28808,0x2202200,0x20802020,0x200a28,0x8a08020a,0x28a08a0,0x20800828,0x208208a,0x8208a2,0x8208220,0x8288200,0x8000a0,0x20a08008,0x82280,0x220a0820,0x20220228,0x8800a28a,0x22822800,0x20228200,0x202008,0xa2880882,0x20a00000,0x220820a,0x82822080,0x228000,0x80080280,0x80822020,0x20028a28,0x282000,0x80020080,0x28a20800,0x88082,0x80000,0x20200a08,0x80208008,0xa20a20a0,0x28208220,0xa200288,0xa20000a0,0x8a00,0x8008a00a,0x82082822,0x8a08a08,0x208020a,0xa0080820,0x200020,0x2200000,0x208a2882,0x800200,0x8280002,0xa2800800,0x8820228,0x80200208,0x2880002,0x8828200,0x82088002,0xa08a0,0x20228a00,0x8080082,0x82880820,0x28828800,0x8a002200,0x808808a2,0x20a08828,0x88200008,0x20a0802,0x20020,0x8228200a,0x82082000,0x20820008,0x820808a,0x200208a2,0x20a28028,0x8020a20a,0xa00800a0,0x8020828,0x88088008,0x22800000,0x800a20,0x88008288,0xa0882820,0x220208,0x8028a280,0x80820800,0x808228,0x8800a282,0xa20a00a2,0x20808820,0x8a000002,0xa00828a0,0x28828228,0x82282,0xa2022080,0x28228a28,0x8828020a,0xa0822800,0x8008020,0x8208080,0x8a0020,0x8a28028,0x80);
remainders_neg[978].modulo_value=3912; remainders_neg[978].bit_mask=create_bit_mask(123,0x82022882,0x8228820,0x8a00a080,0x82882882,0x20028200,0x820a280,0x22002020,0x220,0x28808a,0x20a2002,0x20a08220,0x88082282,0xa28a0000,0x8200200,0x208a202,0x28a0802,0x28028a20,0x2208288,0x882,0x20a28228,0xa08a280,0x208a08a2,0x8820000,0x8a28a288,0xa00828a0,0x8a00220,0x2008282,0x82020020,0x28008020,0x8a00a002,0x800a0022,0xa00a28,0x80200288,0x80080820,0x28200800,0x8800a208,0xa28200a0,0x28a28a20,0x82202,0xa20a2080,0x28a20a08,0x8828a080,0x22000002,0x28208808,0x8a280282,0x20a2800,0x8a20808,0x8008208,0xa28a0,0x28820220,0x8820a088,0x8a0800,0x20228008,0x8800000a,0x80800880,0x28a08200,0x8280080,0x22822820,0x20a00a28,0x82288200,0x22880820,0x8800208,0x8002082,0x28020a2,0x200200,0x8a008282,0x20800020,0x8a08828,0x8a288088,0xa0022002,0x8008a08,0x8020808a,0x22080080,0x28a20,0x8288280,0x82000882,0x20a08,0xa00008a,0x80820020,0x28a20208,0x8000800a,0x20000a0,0x20020008,0x208a288,0x22,0x28200020,0x208080,0x82820082,0x8828828,0xa20a082,0x2802880,0x28028800,0x80280002,0xa08a0022,0x28220820,0x8282282,0x20800820,0x828020,0x80,0xa2080880,0x8008228,0x80200,0x200020a0,0x8a28a00,0x8202208,0xa0080,0x8000a20,0x8220a,0x22820822,0x28000028,0x208808a,0x20802020,0x20020a20,0x8800a20a,0x228a2800,0x20a20220,0x208282,0x200a2080,0x800828,0x80280008,0x800208a0,0x220820,0x8);
remainders_pos[979].modulo_value=3916; remainders_pos[979].bit_mask=create_bit_mask(123,0xaa80a22a,0x28222280,0x20828822,0x28a808a2,0xa82a08a8,0x80820288,0xa02a,0x2a200a,0xa80a2aa,0x202aa8aa,0xaa02022a,0x8a820220,0x2a8a00,0x8208000a,0xa0a20202,0xaa2a0220,0x20aa0808,0xa2002020,0x808a8228,0x20222080,0xa088082,0x80820a28,0x8aa88280,0x88200200,0xaaa28288,0x2a880,0x8a222aa8,0x222aa208,0x8208208a,0x8a80a002,0x8a02822a,0x282a0000,0x28828a0a,0x88008082,0x88220,0x200a8aa0,0x8080002,0x8aa82222,0x220020a,0x82882a2a,0xa228220,0x2800a88a,0xa02282a,0xa88a8820,0xa28020a8,0x8a008020,0x8a880828,0xa8a28802,0x82002a82,0x828a020a,0x8a888028,0xa800,0x220280,0x2880a880,0xa80aaa82,0xa8a2200,0x2200a020,0x22a000a0,0xa00a0880,0x8882088,0x802a000a,0xa802000,0x82220a00,0xa00a220,0xa88022,0xa00880a0,0x8a2a0080,0xaa02a008,0xa022882a,0x8800022,0xa0000028,0x222a2002,0xa282280,0x80082a08,0x28a2a82a,0x22a2802,0x200a2282,0x8028a080,0x2a22a2a0,0x880a0082,0xa0028a82,0x288a0a22,0x82282088,0x880a8a,0x82aa2a08,0x2020888,0x2a008800,0x220000aa,0x20022088,0x28228820,0xa828a0a,0x280a2000,0xa02a2a88,0x82082800,0xaa888a20,0x888a2208,0xa80002aa,0x28aaa022,0x822228,0x22aa2008,0x8202028,0x220a028a,0x88808820,0x2a202020,0x8008a288,0xaa08080,0xa8aa202,0x8a0a088,0x2082808,0x2a800a00,0x82a200a,0x80aa088,0xaa808a88,0xa02a0aa2,0x8a800aa8,0xa0000a00,0x8202a80,0xa82a228,0x2a282a2,0x282088a2,0x88828882,0xa02aa028,0xa88);
remainders_neg[979].modulo_value=3916; remainders_neg[979].bit_mask=create_bit_mask(123,0x220822a2,0xa02aaa08,0xa80a00aa,0xa0208022,0x20a20020,0xa8a00,0x888828a2,0x88a2a802,0x82002222,0x28a22022,0x228a8aa2,0x20a82a8,0x88220288,0xa808882,0x282a8a82,0x22a20aa8,0xa0228080,0x2a88a8a8,0x8020a20,0xa8a22808,0x280080a,0x80a82a0,0x2200208,0x288a88,0x222a0a00,0x888a2000,0x2aa2220,0xa2a22a80,0xa80a802,0x208280a,0x2820aa2,0x20a28888,0xa00a0282,0x88000a,0x88000aa8,0xa8820228,0x80808882,0x2202aaa,0x82a88a82,0xa00a2a2,0x82aa0a28,0xa0002002,0x28aa0a2,0x200200a8,0x2a08a020,0x20808a8,0x20080a0,0x202a0082,0xa88220a,0x2028a82,0x20008a0,0x88882008,0x80a28a08,0x20082008,0x2082220a,0x8202a288,0xaa0828a8,0xaa288828,0x28828000,0x80002800,0xa28882,0x8208a888,0xaaa8208,0x2802aa8,0x82008aa,0x28800828,0x2a28008,0x220a2880,0x28aa00a2,0x800888a2,0x288808a0,0xa2a2a88a,0x82a0aa08,0x880a200,0xa02220a2,0xaa2a082,0xa882aa0a,0x8a02008,0xa22a2a28,0x82880a,0x288a0202,0xa00202aa,0x2a0a800,0x88008202,0xa22a200,0x8a828000,0x22880008,0xaa888822,0xa88aa000,0xa02a00a8,0x820a0282,0xa082a880,0x28a22200,0x280a088,0x220002a8,0x2aa00,0x20808a22,0x202220aa,0x880aaaa0,0xaa22a080,0x8028a8a0,0xaa828a02,0x800a0,0xa2a828a8,0x802a00,0x82280808,0x82022a0a,0x80202800,0xa802080,0xa2088288,0x80a2a082,0x80022800,0x22080200,0x28a28222,0x2080220,0x20888288,0x80a8a208,0x820a2a20,0x8a220a2a,0x20a0002a,0xa000a,0x28a220a0,0x200);
remainders_pos[980].modulo_value=280; remainders_pos[980].bit_mask=create_bit_mask(9,0xa0080a02,0x28080280,0x20280a0,0x200a028,0x80a0280a,0x80280a00,0x280a0280,0xa028020,0x80a020);
remainders_neg[980].modulo_value=280; remainders_neg[980].bit_mask=create_bit_mask(9,0x2880020a,0x20008a00,0x28a0028,0x8a0028a0,0x20a002,0x8a00288,0xa0028200,0x28a0028,0x28a0);
remainders_pos[981].modulo_value=1308; remainders_pos[981].bit_mask=create_bit_mask(41,0xa20000a2,0x20020808,0x2080280,0x22020002,0x28028208,0x8800a200,0xa28a0822,0x20a08208,0x2000080,0x8a2822,0x28800808,0x20220a,0x82820080,0x28a28820,0x8008a208,0xa00a0022,0x20220820,0x8202000,0x200200a0,0x228808,0x28a280,0x82022880,0xa00800,0x80008082,0xa0820880,0x28800a00,0x8208a200,0x208228a2,0x200828,0x8a088080,0x2020022,0x8828a20,0x80200008,0xa20820a0,0x28820a28,0x8008a002,0x82082802,0x8000808,0x80280208,0xa2020800,0x8a00008);
remainders_neg[981].modulo_value=1308; remainders_neg[981].bit_mask=create_bit_mask(41,0x22880822,0x28828000,0x8a008208,0xa28a0882,0x20820a00,0x82288,0x220200a2,0x28200a00,0x8a088008,0x800220a2,0x20008000,0x8828a282,0x20a0800,0x20220028,0x8002280,0x208208a2,0x28a28028,0x8028a088,0xa08a0820,0x8a20000,0x88202208,0x28a2000,0x8208008,0x8000a,0x200a0000,0x20008208,0xa002288,0xa00a2022,0x8a08020,0x2000008,0x828a08a2,0x20228,0x8288080,0x22802820,0x20028220,0x800208a,0x2802082,0x800000,0x8208280,0x228a0080,0x208800);
remainders_pos[982].modulo_value=3928; remainders_pos[982].bit_mask=create_bit_mask(123,0x2aa20a8a,0x802a8a02,0x808a2822,0xa80288a,0xaaa8280a,0xaa08a28a,0x8808,0x228a028a,0xa0088022,0xa0a2228a,0x808a82,0x882aaaa8,0x20a2a88,0xa28082aa,0xa28a0aa8,0x2a20228a,0x2a0220a0,0xa8a22808,0x800022a0,0x8822222,0xa282008,0x8a202aaa,0x2aaa8800,0x88a028a0,0xa2008008,0x22202a00,0x20a08282,0x8200288,0x882a82a0,0x22a8a022,0x2a2020aa,0x802200a,0x8028822a,0xaa28aa02,0x8a0a2882,0x2a222282,0x8aa20aa0,0x2a0888aa,0x80002a08,0x2a208aa8,0x2820a000,0x2288a8aa,0x2aaa8222,0x22080a02,0x22a028a8,0x2a202a0a,0xa0820080,0x8a820800,0x20a2a002,0x288880,0x8a8aa800,0x220a0280,0x88aa0082,0xa082222a,0x8aaaa082,0x8a008a8,0x28000082,0x2a80a0a8,0x208a8088,0x802a8228,0xa20022a0,0x8a020080,0x20a8800,0x82882a8,0xa82202a2,0x282a0a02,0x20820000,0xa22a200a,0xa820aaa,0x22a88882,0x888200aa,0xa20280a0,0x2aa2,0x8022228,0xa2800a8a,0xa002082,0xa8020082,0x88a0a808,0x882a280a,0xa880a020,0x888882aa,0x28aa2a22,0xa8000a08,0x22aa208,0xa820a800,0xa2aa2220,0xa80aa08a,0xa2828888,0xaa8228a0,0x280aa28,0x20a88228,0xa8a00880,0x88aa0808,0x22880a2a,0x200a82a8,0x8228008,0x8882820a,0x8a00a808,0x22200200,0xa80a280a,0xa20022aa,0xa0aaa808,0x20200828,0x2888882,0x2a0a8800,0xa820288a,0xa80a0880,0x8aa28808,0x8a2aa0a2,0x80aa8202,0x2222a8a0,0x2aaaa8,0xa82a202,0xaa2888a,0x88880220,0xa280a2,0xaa202200,0xaaa28a20,0xa0a0282a,0x2a22802,0x28828a2,0xa880a2a8,0xa2a08a);
remainders_neg[982].modulo_value=3928; remainders_neg[982].bit_mask=create_bit_mask(123,0xa22a8202,0x8a2028a,0x802a0aa,0x8208a002,0x2220a082,0x22802a02,0x88880080,0xaa028a02,0x288008aa,0x282aaa02,0x8808020a,0xa22220,0x8a82a200,0x2a080a22,0x2a028220,0xa2a8a202,0xa28aa828,0x202aa080,0x888aa28,0x800aaaaa,0x82a0a880,0x2a8a222,0xa2220088,0x28a028,0x2a880880,0xaaa8a288,0xa8280a0a,0x80a88a00,0xa20a28,0xaa2028aa,0xa2a8a822,0x808aa882,0x8a00aa2,0x22a0228a,0x282a00a,0xa2aaaa0a,0x22a8228,0xa2800022,0x888a280,0xa2a80220,0xa0a82888,0xaa002022,0xa2220aaa,0xaa80828a,0xaa28a020,0xa2a8a282,0x280a8808,0x20a8088,0xa82a288a,0x88a00008,0x2022088,0xaa828a08,0x22880a,0x280aaaa2,0x222280a,0x80288020,0xa088880a,0xa2082820,0xa8020800,0x8a20aa0,0x2a88aa28,0x28a8808,0x8a820088,0x80a00a20,0x20aa8a2a,0xa0a2828a,0xa80a8888,0x2aa2a882,0x820a8222,0xaa20000a,0xa8822,0x2a8a0828,0x8888a22a,0x808aaaa0,0x2a088202,0x8288a80a,0x200a880a,0x282080,0xa2a082,0x200828a8,0xa22,0xa022a2aa,0x20888280,0x8aa22a80,0x20a82088,0x2a22aaa8,0x20822802,0x2a0a0000,0x220aa028,0x8a0822a0,0xa8200aa0,0x20288008,0x228080,0xaa0082a2,0xa8820a20,0x80aa0880,0xa0a82,0x2882080,0xaaa88a88,0x2082a082,0x2a88aa22,0x28222080,0xa8a880a0,0x8a00000a,0xa2820088,0x20a8a002,0x20828008,0x22a0080,0x2a2282a,0x8220a8a,0xaaaa2028,0x88a22220,0x820a2a8a,0x822a0002,0x8aa8,0x882a082a,0x22a8aa88,0x222a02a8,0x2828a0a2,0x8a2aa08a,0x8a00a02a,0x20082a20,0x2a2802);
remainders_pos[983].modulo_value=3932; remainders_pos[983].bit_mask=create_bit_mask(123,0x2208a02,0x22a02a8,0xaaaa00a2,0x82880088,0x80220a2,0x208a2222,0x2a880a82,0x80aa8a,0xa08a000,0xa8aa8aa,0xa80aaa8a,0x8a822202,0x288a0aa2,0x8222a80a,0xa0a8a828,0x82282aa0,0x8882a202,0x2022aa82,0x8080a282,0x800a22a8,0x28088088,0x80a8a2,0xaa880a28,0xa8aa2a00,0x8aaaa228,0xa00a8a02,0x288002a8,0x288a0880,0xa0a28a08,0xaaa80080,0x2208a882,0xa0000a20,0x8282a2a,0x8220a282,0xaa20aa22,0x2a8200a,0x20a82,0x80202088,0xa2002a80,0x2a2aa082,0x22a08820,0x82aa222a,0x2a2a2008,0x8a88a08,0xa220a00,0xa08a0a88,0x20282800,0x88a20a0,0x822082a0,0x8a020802,0x2a00a88,0x2028200,0x22828020,0x2a2aaaa8,0xa2008200,0x82288a02,0x8aa28888,0xa20a0828,0x8000222a,0x808a882a,0x288aa00a,0x882a0a82,0x2a00aa22,0x2a822a20,0xaa888000,0x22820a08,0x222228aa,0xa80a2288,0x80082008,0x82aaaa8a,0x802828,0x82808,0x222a00a8,0x2802080a,0xa82088,0x80a08a22,0xa0028280,0x22a0a20,0xa088a,0x820a22a2,0x22008a8a,0x8a888aa8,0x808220a8,0xa820aa8a,0x202a8008,0x2208080,0x82a0800,0xaa0082a8,0x288aa08a,0x828a088,0xaa8a8282,0x808a0000,0xa822a208,0xa02002aa,0x820a28a0,0x80220a22,0xa2a80022,0x280a2a00,0xa288aaaa,0xa00a8aa2,0x28a022a,0x88a2a020,0x22202202,0x22a88a00,0x8828a020,0x282aa880,0x2808a822,0xa0aa8288,0x2282a2a0,0x8a02a888,0x28aa0a22,0xa808882a,0xa2aaa02,0xaa2aa2a,0xa20a,0x8a2aa020,0x882a022a,0x8888a20,0x28a08802,0xa2200228,0x8a00aaa,0x2a80a88,0x80a2088);
remainders_neg[983].modulo_value=3932; remainders_neg[983].bit_mask=create_bit_mask(123,0x8aa8028a,0x8aa28a20,0x2222882a,0xa008800,0x808aa82a,0xa802aaaa,0xa200820a,0x88082202,0x82802888,0x82022022,0x20822202,0x20aaa8a,0xa002822a,0xaaa2082,0x282020a0,0xaa0a228,0xa2a8a,0xa8aa220a,0x8082a0a,0x882aa20,0xa0800800,0x8808202a,0x220082a0,0x2022a288,0x2222aa0,0x2882028a,0xa0088a20,0xa0028008,0x282a0280,0x22208808,0xaa00200a,0x288882a8,0x80a0a2a2,0xaa82a0a,0x22a822aa,0x8a20a882,0x888a820a,0x8a8a800,0x2a88a208,0xa2a2280a,0xaa2800a8,0xa22aaa2,0xa2a2a880,0x80200280,0x82aa8288,0x28028200,0xa8a0a088,0x8002a828,0xaa80a28,0x28a808a,0x8a288200,0x8a8a0a88,0xaa0a08a8,0xa2a22220,0x2a880a88,0xaa0028a,0x22a0000,0x2a8280a0,0x888aaa2,0x80200a2,0xa0022882,0xa2820a,0xa28822aa,0xa20aa2a8,0x22000888,0xaa0a8280,0xaaaaa022,0x2082aa00,0x880a880,0xa222202,0x8808a0a0,0x8880a080,0xaaa28820,0xa08a8082,0x8820a800,0x82802aa,0x288a0a08,0x8aa282a8,0x88828002,0xa82aa2a,0xaa880202,0x2000220,0x80aa820,0x20a82202,0xa8a20880,0x8aa80808,0x80a28088,0x22880a20,0xa0022802,0x80a02800,0x22020a0a,0x8028888,0x20aa2a80,0x28a88a22,0xa82a028,0x8aa82aa,0x2a2088aa,0xa082a288,0x2a002222,0x2882022a,0x8a028aa2,0x2a28a8,0xaaa8aa8a,0xaa200288,0xa028a8,0xa0a22008,0xa08020aa,0x28220a00,0xaa0a2a28,0x28a2000,0xa02282aa,0x208000a2,0x82a2228a,0x822a22a2,0x88882a82,0x2a228a8,0xa28aa2,0x800002a8,0xa028008a,0x2aa88aa0,0x80288222,0x8a208200,0x82a800);
remainders_pos[984].modulo_value=984; remainders_pos[984].bit_mask=create_bit_mask(31,0x28228a2,0x820808,0xa20a202,0x22882000,0x20808a00,0x828a20a,0x208000a0,0x20220228,0xa082002,0x82802,0x8820020,0x8002280,0x20028a2,0x20828220,0x200a080,0x80000000,0x8020a00,0x80088282,0x208a2800,0x20028800,0x80082,0xa0802820,0x8800820,0x8288088,0x200a0002,0x8a08a28,0x8800a202,0xa0000822,0x808a08,0x80202082,0x8a2882);
remainders_neg[984].modulo_value=984; remainders_neg[984].bit_mask=create_bit_mask(31,0x820a2022,0x8028000,0x8228228a,0xa2002880,0x28000208,0x80202282,0xa0080820,0x28a28a20,0x200a08a,0x80802082,0x28828,0x8008a208,0x82882022,0x28020a28,0x8a082000,0x880880,0x828208,0x800020a,0xa0022080,0x28820008,0x8800800a,0x200820a0,0x8028,0x80200000,0xa0820082,0x200220,0x8228a,0x208800a2,0x8000200,0x828200a,0x22002);
remainders_pos[985].modulo_value=985; remainders_pos[985].bit_mask=create_bit_mask(31,0x2d0f335e,0x2e5b02d5,0x867319,0xe7631e5,0x170d0141,0x80a24a96,0x2f015611,0x28dc959c,0x82110a9,0x21bda053,0x906e4b9c,0xca2ace04,0x13962101,0x6558009,0xc319c952,0xc6a7958,0x812a4e63,0x224006a9,0x4e0211a7,0x2481cd51,0x10e749d8,0x432816f6,0x52542210,0xd0e6a4ec,0x621aa03,0xa1a54914,0xc20a02c3,0x29e31b9,0xd2633984,0xd2ad0369,0x1eb33c2);
remainders_neg[985].modulo_value=3940; remainders_neg[985].bit_mask=create_bit_mask(124,0xa0822a82,0xa2028a00,0x80aa280,0x82a2a828,0x82808808,0x82a020a,0xa2088a80,0x20000000,0x80a08820,0xa020280a,0x8a20200,0x2220280,0xa0aa088,0x8a800880,0x280000a,0x80a22080,0x82282a2,0x2a888220,0x8282082a,0xa0080008,0x882a0000,0x8a208a22,0xa88a288,0x8222820,0x82a8220a,0x28280088,0xa028a02,0xa02a020,0xaa2a008,0x2208a20,0x20222a0a,0x220282a2,0x2a802288,0x280802a,0x280202a,0xa0a2028a,0xaa088800,0x20828220,0x8a8020a2,0x22888a28,0xa08a2288,0x80280002,0xa088000,0x8aa2a082,0x20802208,0x280a2a0a,0xa000822,0xa8a8000,0x82800880,0x2200882a,0x28800a02,0x20280202,0x280a80a0,0x202088,0x82280000,0xa0800a88,0xa0a8202,0x2a088088,0x8080a0aa,0x2280a2a,0x828028a,0x220aa2,0x28820a0,0xa020a0,0x2088a00,0x88882000,0xa20080a0,0xa8200,0x8aa02088,0x82282a2a,0x8802808,0xa2280a22,0x208a8020,0x882a022,0x80a88200,0xa880008a,0x282a80a2,0x2000a020,0x82228820,0x82000008,0xa8882080,0xaa280a82,0x2008200,0x808028,0x2288008,0x20820a02,0xa8088002,0x2820208,0xa2280a0,0x22888220,0x802a0000,0x2028220,0x20022220,0x8220a0,0xa200002a,0x28022088,0xa00280a2,0xa882020,0x82200088,0x20202820,0x8008008a,0x200a0080,0x2a802080,0x8aa0a80a,0x20808208,0x20880000,0x20208a2,0xa020280,0x8aa280aa,0x80a88000,0x8800208a,0x822202a0,0x228200a8,0x8822280a,0x20880a00,0xa8aa2a0a,0x88202,0x228020a8,0x88828080,0x20000208,0x80002888,0xa0028202,0x28208,0x2);
remainders_pos[986].modulo_value=3944; remainders_pos[986].bit_mask=create_bit_mask(124,0x82280222,0x22a82,0x8802a220,0x2202aaa2,0xa880a800,0x20a22a22,0x80a0222a,0x2a20802a,0xaa0080a,0x880a20,0x8a020a00,0xa00aa202,0x28a0aa2a,0x8a000828,0x8828a282,0x2aa0a22,0x2822028,0x2828080a,0xa22220a0,0x28aa208,0x800a0a8a,0x8aa0800,0x220808a2,0x822880aa,0x2a2a28a0,0x22aa2aa0,0xa8882080,0x2a22282,0x88aaa280,0x82822a08,0x2aa882a,0x82002080,0x8888022,0x2a80200,0x88008222,0x2228228,0x200a0020,0x220a0800,0x20200a88,0xa208a8,0x888aa8a0,0x2a080a0,0x2200a200,0x282a0a20,0x82a088a0,0x2a82a02,0x28822020,0xa0828008,0x82820a8,0xa200000a,0xa0822200,0x8a88a0a2,0xa220a8,0x200028aa,0xa002a8a,0x820202a,0x80022022,0x822aa802,0x8820002,0xaaaa2202,0x22028020,0x888222a0,0x280880a,0x88aaaa08,0x822080,0x2aa88280,0x8800280,0x8082088,0xa800a0a8,0x280008a2,0x88a00aa,0xa22a22a,0x88820a8a,0x8a00,0x82820a0,0x2820a2a,0x8822820,0xa82a8008,0x220a8280,0xa0a8280a,0x8a008808,0x20a8000,0x2aa2220a,0x208a000a,0xa008082a,0x20a08822,0xa00800,0x82888008,0x82002228,0x802a8088,0x2222000,0x8008288,0x22aa8002,0xa88282a8,0x8aaa2220,0x888a8002,0x8222a82,0xa8aa8802,0x28a8a80a,0x228820a,0x202088aa,0x20aa208a,0xa0a00200,0x8aa280a2,0x8888a20,0x2028280a,0x88280a0,0xa0aa8028,0x8a282288,0x2000a282,0xaa0a2828,0x8aa00aa8,0xa080a280,0xa0220000,0x200aa008,0x208a8a0,0x880a02a8,0xa88a08a8,0x2a022a88,0xaa808800,0x8a80228a,0xa8800008,0x80288282,0x88);
remainders_neg[986].modulo_value=3944; remainders_neg[986].bit_mask=create_bit_mask(123,0xaa08aaa,0x8882a20a,0xa2aa8,0xaa0a222a,0x20082088,0xa02aa2aa,0x828a2a2,0x22a808a2,0x82288082,0x80082a8,0x28a8280,0x28822a8a,0xa0282222,0x28080a0,0xa0220a,0x8a2282aa,0x8a0a28a0,0xa0a08002,0x2aa2a828,0x8a022a80,0x28202,0x80228088,0xa280802a,0xaa00822,0x22a22028,0xaa22a228,0x2000a808,0x8a2aaa02,0x222a00,0xa0aa200,0x8a2200a2,0xa08a008,0x800008aa,0x8a200a88,0x880aaa,0x8aa202a0,0xa88288a8,0xaa028088,0x28a88200,0x802a8020,0x22028,0xa280828,0xaa802a88,0xa0a282a8,0xa280020,0x8a20a28a,0xa00aa828,0x280a0880,0x80a0a020,0x22888882,0x280a2a88,0x200282a,0x8822a820,0xa8882022,0x8208a202,0x80a8a8a2,0x8aa8aa,0xaa22082,0xa888a,0x2222aa8a,0xaa0a08a8,0xaaa20,0x8a080082,0x222200,0x880aa008,0xa2200208,0x80088a08,0x802800,0x20882820,0xa080802a,0x80028822,0x82222aa2,0xa8202,0x80880288,0x80a0a828,0xa0a8222,0x800aa0a8,0x20a20880,0xa2820a00,0x2820a082,0x2880000,0x8a820888,0xa22a2282,0xa8028882,0x288000a2,0xa82800aa,0x88208080,0xa000880,0xa08aaa0,0x8220800,0x82aaa888,0x80880a00,0x2a22088a,0x200a0220,0x222aaa8,0x20882,0xaaa20a,0x2022000a,0xa0202082,0x8aa00282,0xa8a00022,0xa8222802,0x28288a88,0x222082a,0x80000228,0xa820a082,0x800a0828,0x202208a0,0x2a0aa00,0x28882a0a,0x2282a0a0,0x2280220,0x28082a00,0x28aa8888,0xa8822800,0x8a802020,0x828220,0x20028020,0xa20a2200,0x22080088,0x200aa02,0x20088880,0x820020a);
remainders_pos[987].modulo_value=3948; remainders_pos[987].bit_mask=create_bit_mask(123,0x22022002,0x800820,0x82000288,0x820a00a0,0x8800200,0x80202288,0x200a2000,0x20008228,0xa20a280,0x228a0082,0x28220020,0x2000202,0x82000820,0xa28208,0x8820a,0x882822,0x20820a00,0x208a00a,0x2882,0xa20820,0x8008820a,0x20800800,0x8200a20,0x2008200,0x20820,0x28200a00,0x80000280,0x808000a0,0x28828000,0x228a282,0x2800822,0x8008028,0xa2800,0x28200228,0x8008020a,0x220800a0,0xa08208,0x8228800a,0x20800080,0x8220228,0x8008202,0xa2802000,0x8800020,0x8002,0x22080082,0x28808000,0xa000200,0x80020000,0x8808228,0x88282,0x880880,0x20800,0x88200008,0x80800020,0x28228000,0x2000008,0x20808a2,0x200a00,0xa20008a,0x80802002,0x8020808,0x80202202,0x8a0800,0x800a20,0x208000a,0x808a2,0x28828200,0x202000,0x822080,0x8000200,0x2200002,0x22000022,0x20220828,0x8002288,0xa0000,0x20228008,0x2088000,0x20000820,0x220800,0x8028a080,0x20020000,0x8820808,0x208288,0x22882020,0x20a00a00,0x2088208,0x20020a0,0x828a08,0x8a000288,0x2,0x20020000,0x280280,0xa2880882,0x28228828,0x202000,0x20a0,0x828028,0x800000a,0x20080082,0x820008,0x20808a,0x22002002,0x8a00a08,0x80000082,0xa2080822,0x8208a00,0x8220000a,0x22000882,0x28a00a08,0x82208,0x80082,0x8828808,0xa288080,0xa08a0820,0x20008028,0x8a008288,0x80802000,0x220228,0xa082008,0x820a0,0x200228,0x88088082,0x800);
remainders_neg[987].modulo_value=3948; remainders_neg[987].bit_mask=create_bit_mask(124,0xa28a2802,0x8000020,0x2088200,0x2020820,0x8a08,0x8282200,0xa0822080,0x28800220,0x2282200,0xa2020802,0x20228828,0x8a00828a,0x28800a0,0x8220200,0x88000202,0x20a2,0x20028208,0x8a002082,0x80802002,0x8220020,0x8000282,0xa0080000,0xa08220,0x82080288,0xa00a0,0x20a08208,0x8080208,0x80020,0x20020808,0xa202202,0x820800a2,0x820,0x88008088,0x80822080,0x20a08220,0x8008202,0x22000820,0x200a00,0xa200082,0xa0000800,0xa20a20,0x8008028a,0x22002800,0x8820,0x8008008a,0xa2000802,0x20000808,0x82080288,0x8a0080,0xa20,0x8000202,0x80000000,0x8028008,0x208080,0x800a0,0x20a20008,0x8a088000,0x2800022,0xa08208,0x82288002,0x2802,0x820000,0x828a28a,0x80020000,0x8008220,0x82008082,0x80000022,0x20020a08,0x88282088,0x800a2000,0x808a08,0xa288082,0xa28808a2,0x28228020,0x8000a200,0x80820880,0x28a20000,0x82000008,0x208800a0,0xa28008,0x8202008,0xa0820880,0x20000,0x88280200,0xa2002820,0x28208200,0x82000280,0x82082820,0x8020200,0x2000200,0x80880082,0x28828808,0x8208200,0x22000002,0x20220020,0x88202088,0x80882820,0x8020020,0x80088002,0x20800802,0x28800,0x88280002,0xa2802882,0x200200,0x808800a,0x22800022,0xa00200,0x2288002,0xa2080002,0x20208200,0x88002280,0x80800002,0x20000,0x2200000,0x200200a0,0x28000820,0x2000200,0x82880,0x8a28220,0x8200a000,0x802820,0xa08a20,0xa,0x80);
remainders_pos[988].modulo_value=1976; remainders_pos[988].bit_mask=create_bit_mask(62,0xa220a8a,0xa00a8222,0x20aa800,0x8a88a8aa,0x2a0a2000,0x28888008,0x880002,0x20aa282,0x82288a08,0x222082aa,0x280a282,0x200228a8,0x2820020,0x2880aa0,0x8082880a,0x20080200,0xa022820,0xaa0228,0x80800828,0x2a28800,0x8828a028,0x2aa82088,0x20a2288,0x828aa08,0x88000220,0x82002222,0x20008a0a,0xaa2a22a,0xa002aa08,0x8882a00,0xaa2a088a,0x20a220a8,0xa00a822,0xa020aa80,0x8a88a8a,0x82a0a200,0x22888800,0x20088000,0x8020aa28,0xa82288a0,0x2222082a,0x80280a28,0x200228a,0x282002,0xa02880aa,0x8082880,0x2008020,0x80a02282,0x800aa022,0x8080082,0x802a2880,0x88828a02,0x82aa8208,0x8020a228,0x828aa0,0x28800022,0xa8200222,0xa20008a0,0x80aa2a22,0xa002aa0,0xa08882a0,0xa2a088);
remainders_neg[988].modulo_value=1976; remainders_neg[988].bit_mask=create_bit_mask(62,0x82a28202,0x28820a2a,0x2822088,0x2002022,0xa2822888,0xa0000080,0x88008882,0x8a022a0a,0xaa00280,0xaaa80222,0xa082a02,0xa88aa020,0x8a0208a8,0x8a008228,0xa0002,0xa0808a88,0x828aa0a8,0x88220a20,0x80880a0,0x8a220080,0x2028a0,0x2220a800,0x8a82a200,0xa02280,0x8882a8,0xa80aaaa,0xa8080282,0x822a2a22,0x208a2280,0x8000a280,0x22a20002,0xa82a2820,0x828820a2,0x20282208,0x80200202,0xa282288,0x2a000008,0xa8800888,0x8a022a0,0x20aa0028,0x2aaa8022,0xa082a0,0x8a88aa02,0x88a0208a,0x28a00822,0x8000a000,0x8a0808a8,0x828aa0a,0x88220a2,0x80880a,0x8a22008,0x2028a,0x2220a80,0x8a82a20,0x800a0228,0xa008882a,0x20a80aaa,0x2a808028,0x822a2a2,0x208a228,0x28000a28,0x2a2000);
remainders_pos[989].modulo_value=3956; remainders_pos[989].bit_mask=create_bit_mask(124,0x820822a2,0xa80aa22a,0x8a820002,0xa2220020,0xa8200a20,0x8a00aaa0,0xa2022a2,0x82a2a88a,0x8080a088,0xaaaa028a,0x8022a2,0x80220222,0x20282220,0x2a8a28a8,0x8082aa0,0x22802282,0xa2280a00,0x8a00028,0x20802280,0x2a8a800a,0x22800002,0x8880a820,0x22808008,0x80228282,0xa82a8a8,0x82a8020a,0x2a2aa88,0x2208a822,0x2002a020,0x8a080288,0x28022a2,0x8a2a200a,0x208a0082,0x20202a00,0x220a882,0x88aaa2a,0x828200aa,0x8a8aa0a,0x220a0a20,0x8880080a,0x2020a808,0xa200000a,0x22800a8a,0x800a2008,0x28a00028,0x200280a2,0x820a200a,0xa02aa080,0x20a8a28a,0x82220a0,0x2220220,0xaa2a2008,0xa8a02aa,0x8882808,0x8288aa2a,0x8a2a2022,0xa82aa802,0x28228020,0x88200222,0xaa00020a,0xaa22a80,0x8a2a2082,0xaa08208a,0xaa02a88,0x822a0800,0x82888800,0x82a08028,0x8a2802aa,0x2828808a,0x220a8aa2,0x2a020282,0x8aaaa80a,0x8802008a,0x82008808,0xa080a088,0x80aa28a2,0x82020aa,0x8a008a,0xa288a028,0x228000,0x2882008a,0x8aa2a00,0x808a0000,0x222202a0,0x88a0200,0xa2008a0a,0x282a0aa2,0x220aa008,0x880a8aaa,0x808822a0,0x20800a80,0x8a28200a,0x280a008a,0xa28a880,0x822802,0x88080a8,0xa80882a2,0xa8222aa8,0x2a0a0802,0x8022a2a8,0x28882828,0x22220020,0x288082a0,0x2a880000,0x208a002a,0xa2002880,0x88a28000,0x28800a02,0x2082880,0x2a80aa82,0x8082a28a,0x80208882,0x20088808,0xaaa8a880,0x202a280a,0xa82220a0,0x8a0a22a8,0xa28a880,0x82a0aaa0,0x88a08a00,0x2a208008,0x2a80008,0x82a88aa,0x8a882);
remainders_neg[989].modulo_value=3956; remainders_neg[989].bit_mask=create_bit_mask(124,0xa00aa2a,0x208222a2,0x20a888a,0x2aa288a8,0x20a882a8,0x2882228,0x82282a2a,0xa2a2002,0x8082800,0x22228202,0x8808aa2a,0x8aa8aaa,0xa8a0aa20,0xa202a020,0x8080a228,0xaa08aa02,0x2aa08288,0x2808a0,0xa808aa08,0xa2020882,0xa208888a,0x820a8,0xa2080880,0x82a0a0a,0x820a2020,0xa208a82,0x8a222200,0xaa8020aa,0xa88a2828,0x2800a00,0x8a08aa2a,0x2a2a882,0xa802800a,0xa8a0a288,0x8aa8200a,0x80022222,0xa0a8822,0x80202282,0x2a8282a0,0x88082,0x28a82080,0x2a888882,0xaa088202,0x82a880,0xa02880a0,0xa88a082a,0xa02a882,0x28a22808,0xa8202a02,0x8022a828,0x8aaa8aa8,0x22a2a880,0x82020a22,0x8000a080,0xa0022a2,0x2a2a0a2,0x20a2208a,0xa0aa08a8,0xa88a2a,0x22888a82,0x822a2208,0x2a2a802,0x2280a802,0x228a200,0xaa28088,0x2000088,0x22808a0,0x2a08a22,0xa0a00802,0xaa02022a,0xa28a8a0a,0x2222002,0x828802,0xa880080,0x28082800,0x822202a,0x80a0a822,0x88028802,0x2a0020a0,0x88aa0888,0xa00a8802,0x22a208,0x8028888,0xaaaa8a28,0x80028a80,0x2a880282,0x20a2822a,0xaa822080,0x820222,0xaa28,0xa8088208,0x2a0a882,0xa0028802,0x82a02008,0x880aa08a,0x80000820,0x20800a2a,0x20aaa220,0xa2820082,0x8aa2a20,0xa000a0a0,0xaaaa80a8,0xa0080a28,0xa2000888,0xa8028822,0x2a88a008,0x2a0888,0xa0088282,0x8280a008,0x2208220a,0x80a2a02,0x8a8000a,0xa0800080,0x22202008,0xa8a2a002,0x202aa828,0x282aa20,0x82a02008,0xa202228,0x200288,0xa2a80880,0x8a200880,0x80a20022,0x200a);
remainders_pos[990].modulo_value=1320; remainders_pos[990].bit_mask=create_bit_mask(42,0x20820002,0x28228820,0xa088200,0x28a0,0x20a00020,0x80200,0x82000002,0x8808208,0x88008000,0x20080002,0x28020800,0x88008280,0xa2080,0x808008,0x8020a202,0xa2080,0x20220a00,0x80008,0x22000,0x200800,0x8002082,0x20080082,0x8800000,0x200000,0xa0880820,0x8a00000,0x8a080202,0x2020080,0x8a00020,0x82002202,0x20802802,0x200800,0x2002280,0x82022000,0x8220,0x80200080,0xa0800,0x28000008,0x8220a00a,0x22882800,0x820808,0x80);
remainders_neg[990].modulo_value=1320; remainders_neg[990].bit_mask=create_bit_mask(41,0xa00a0082,0x20220020,0x2000288,0x80802020,0x20200828,0x80000288,0x2080882,0x200,0x88,0x20800802,0x20820008,0x80200,0x22000,0x8000800,0x282202,0x80822800,0x20228200,0x8008080,0x80822800,0xa08008,0x8200a,0xa0000802,0x8008,0x88208088,0x20000020,0x200800,0x2000282,0x28a0000,0x208828,0x208a20a,0xa0002082,0x8208000,0xa08a208,0x2822080,0xa28,0x8280008,0x80020080,0x20800000,0x2202082,0xa2002000,0x8020000);
remainders_pos[991].modulo_value=3964; remainders_pos[991].bit_mask=create_bit_mask(124,0x2aa0aaaa,0xa0a2a08a,0x88a2822,0x222a02a8,0xa08aa288,0xa080a222,0x8828,0x222a08a2,0x2a8220,0xaa2a8a02,0xaa880a28,0x82880a82,0x88200222,0xa8828a0,0xa08080a2,0xa888808,0x802aaa8,0xa8a82028,0x2280aaa2,0x20a2200a,0x88008082,0x8222a80a,0xaaaa080a,0x2880880,0x2a8a8820,0x82aa2a20,0x8800a280,0x882280a,0x82aaa802,0xa8820a00,0xa828a02,0xa820a02,0x2a828a20,0x8aa0000a,0x2a08a22,0x2a082aa2,0xa880800a,0xa8a20200,0x2a22a20a,0x2a28000,0x2a0088a,0x8a2a2aa2,0x82a88828,0x8002a08,0xa0a8a020,0xa0000aa2,0x22a22202,0x882a2a28,0x2820a202,0x88aa8a2,0xa02a20a,0x2082a200,0x8a820080,0xa88a00aa,0x88228088,0x20a22aaa,0x8882a28,0xa2080228,0xa00a8088,0x8a08222,0x82280828,0x2800,0x20028000,0x2820288,0xa88820a2,0xa2202a00,0x2880208,0x828a8222,0x2aaa88a0,0xa2202882,0x2aa00a22,0x8020082a,0x8a820,0xa08a80a,0x88a2aa22,0x2808a082,0x828a8a82,0xa80888a8,0xa8aa0000,0x80a2a0,0x220a8002,0x228222a8,0x8aa8a8a,0xa2200a8,0x800028a8,0xaa08a88a,0xa00808a2,0x8a002022,0x8aa820a,0x288a20a8,0x8a0000aa,0x8a282a,0x80a082a,0xa80a282a,0x200a0822,0x802aaa8,0x2a028822,0x2028a002,0x808a8aa8,0x822a2a,0xa0220228,0x2a020aaa,0x2a02a888,0x88202002,0x8a0088a0,0xa8aaa028,0x28082a2,0x2aaa802,0xa202222a,0x8a02020,0x20a2822a,0x28880082,0xa82a0228,0xa28a022a,0x80a2a8a,0x80882a80,0x8a20a88,0xa2822000,0xa888a020,0x8228aa20,0x2a80a88,0xa8828a22,0x8a20a8a0,0xaaaa0aa);
remainders_neg[991].modulo_value=3964; remainders_neg[991].bit_mask=create_bit_mask(124,0xa2282222,0x282a2802,0x8002a0aa,0xaaa28a20,0x28022a00,0x28082aaa,0x888800a0,0xaaa2802a,0x88a20aa8,0x22a2028a,0x220082a0,0xa00820a,0xa88aaa,0x8200a028,0x2808082a,0x82000080,0x808a2220,0x2020a8a0,0xaa08222a,0xa82aa882,0x88080a,0xaaa2082,0x22228082,0x8a008008,0xa20200a8,0xa22a2a8,0x882a08,0x800aa082,0xa22208a,0x200a8288,0x20a028a,0x820a828a,0xa20a02a8,0x2288882,0x8a2802aa,0xa280a22a,0x20080882,0x202a8a88,0xa2aa2a82,0x8a2a0888,0x8a288002,0x2a2a22a,0xa2000a0,0x8088a280,0x282028a8,0x2888822a,0xaa2aaa8a,0xa2a2a0,0xa0a82a8a,0x8002202a,0x828a2a82,0xa80a2a88,0x20a8808,0x20028822,0xaa0800,0xa82aa222,0x8000a2a0,0x2a808aa0,0x28820800,0x80280aaa,0xaa080a0,0x8888a088,0xa88a0888,0x8a0a8a00,0x2000a82a,0x2aa8a288,0x8a008a80,0xa020aaa,0xa2220028,0x2aa8a00a,0xa22882aa,0x8a880a2,0x888020a8,0x82802082,0x2a22aa,0xa080280a,0xa02020a,0x20800020,0x20228888,0x88082a28,0xaa82088a,0xaa0aaa20,0x80220202,0x82aa8820,0x888a020,0x22802002,0x2880802a,0x288a8aa,0x80220a82,0xa002a820,0x2888822,0x8802a0a2,0x808280a2,0x2082a0a2,0xa88280aa,0x808a2220,0xa28a00aa,0xa8a0288a,0x8020220,0x880aa2a2,0x28aa8aa0,0xa28a8222,0xa28a2000,0xa8a88a,0x2880028,0x202228a0,0x8a080a2a,0x8a22208a,0x2a8aaaa2,0x8028a8a8,0xa82a0aa2,0xa000880a,0x20a28aa0,0x2a028aa2,0x8082a202,0x800a208,0x802a8200,0x2a0aa888,0x200028a8,0xaa022a8,0x8a208200,0x200a02aa,0x2a82028,0x2222822);
remainders_pos[992].modulo_value=248; remainders_pos[992].bit_mask=create_bit_mask(8,0x22a8a202,0x28aa022a,0x22a008,0x82a20802,0x80208a,0x28200a88,0x88a880aa,0x808a2a);
remainders_neg[992].modulo_value=248; remainders_neg[992].bit_mask=create_bit_mask(8,0x2a202a8a,0xa0228aa2,0x88aa2880,0xa2a808a,0x8008a802,0xa0a88200,0x200822,0x802a2);
remainders_pos[993].modulo_value=3972; remainders_pos[993].bit_mask=create_bit_mask(125,0xa2880802,0x8028a08,0x820800a,0x828028a0,0x20000a00,0x2000208,0xa0020002,0x28020008,0x808828a,0xa2820002,0x220000,0x8008a288,0x80000002,0x28220800,0x8a088208,0xa00208a2,0x8020220,0x80202088,0x82022020,0x228200,0xa00a202,0x200a0,0x20208a00,0x228820a,0x22880880,0x20a20a20,0x88200208,0xa0002822,0x8208a28,0x828808a,0xa2020802,0x820228,0x8808a082,0xa2882082,0x20a00028,0x8200208a,0x22822820,0x8808a20,0x8208a200,0x2882022,0x28020000,0x2280280,0x20a2002,0x20220208,0x88202008,0x22020080,0x28820028,0x8208828a,0x8220a0,0x8,0x8a28800a,0x22000,0x20a28,0x8a088082,0x800200a2,0x20028,0x82000202,0x2800020,0x28a00a08,0x80082080,0x828a0082,0x808a28,0x8a20200a,0x200a2822,0x20820028,0xa00a280,0x80002802,0x8000820,0x80080008,0xa0080022,0x20220a28,0x8a080008,0x880002,0x228a20,0xa,0xa0882002,0x28220820,0x8008228a,0x20080882,0x808220,0x8088082,0x8a0802,0x28028808,0x80280,0x80822800,0x8a20828,0x8a202200,0x82882880,0x20800820,0x8000a08a,0x208228a2,0x20a20228,0x88082008,0x20808a2,0x20228208,0x8a20820a,0x828000a2,0x8008228,0x8a08a082,0x22022880,0x8228800,0xa20808a,0xa0080000,0x8a00a00,0x8288008,0x80880820,0x20808020,0x88080202,0xa20800a0,0x8220a28,0x2088282,0x20,0x28a20028,0x88002,0x828a0,0x28220208,0x8028a,0x800a2,0x8000808,0xa000082,0xa2802820,0x208200,0xa28020a,0x20228a2,0x8);
remainders_neg[993].modulo_value=3972; remainders_neg[993].bit_mask=create_bit_mask(124,0x22000082,0x820200,0x80280082,0x2082020,0x28808208,0x8a088280,0x208a0882,0x20828800,0x80000202,0x220a0882,0x8a28008,0x8002200,0x880882,0x20a28008,0x2000280,0x208a0022,0x828a28,0x828a000,0x28a28a0,0x8a20a08,0x8208228a,0x808a0820,0x28a00208,0x8a200282,0xa2000000,0x28228228,0x288280,0x208820a2,0xa00220,0x80200002,0x228a0082,0x8028a20,0x200a,0x22002802,0x28208820,0xa08a002,0xa20a20a0,0x228,0xa002288,0x820028a2,0x20828808,0x8a208208,0x82822882,0x28a28a00,0x28a080,0xa28a0800,0x20028820,0xa000202,0x800a2820,0x8808800,0x2200082,0x800a2880,0x8828220,0x200000a,0x8a0822,0x8828820,0xa08828a,0x820808a0,0x20208200,0x800a008,0x2020802,0x8000220,0x228a082,0xa08220a2,0x28028820,0x82082208,0x882082,0x808028,0x8008080,0x208008a2,0x28a28220,0x2008080,0x80000882,0x8a20228,0x88088082,0x20002882,0x20a28028,0x800a202,0xa0800002,0x8000a28,0x8000000a,0x80020082,0x20820000,0x88008208,0xa2080,0x228020,0x228a288,0x2002000,0x28008028,0x8082002,0xa00a2022,0x28228a20,0xa080,0x82800022,0x28a20a00,0x2280282,0x2080822,0x800a20,0x200200a,0xa28a2000,0xa20008,0x82280002,0x20800880,0x208208,0x80200080,0xa0,0x28000828,0x828a,0x22800820,0xa28220,0x8a00020a,0x808808a0,0x20228820,0x8800008a,0x80802020,0x20a28a00,0x88008202,0x80800822,0x808000,0x8208800a,0x220820a0,0x8a00a08,0x82208282,0x828a2022);
remainders_pos[994].modulo_value=3976; remainders_pos[994].bit_mask=create_bit_mask(125,0x8a0a8a2a,0x2a8a202,0xa02a8808,0x2a00002a,0xaaa02088,0x88000a28,0xa2220,0x8a0a2a22,0x20880282,0x20008a2a,0x8208808,0x82820a82,0x80200a20,0x888282a8,0xa8828828,0x2a008a22,0x20028022,0x22028000,0xaa00a880,0x20a0280,0x800a00a,0x2a008828,0x28800a2,0xa22a0a02,0xaa2a088a,0x2a008800,0xa82220,0x28888228,0x22822,0xaa02280,0x22020a02,0xa022a022,0x82220082,0x82228a8,0x20888022,0x22080a28,0x2888a022,0x8020000a,0xa0202000,0xaa08282,0x88a2020,0xa80aa0a8,0x8280a200,0x808a002,0xa282822a,0x2a02a82a,0x2082a002,0x8820800,0x828088,0x208a2a08,0xa08282,0xa08a08a0,0x200aa828,0x82a08a00,0x202080,0x202022a2,0x208aa80a,0xa280000,0x22aa8028,0x88002800,0xa2000288,0x802080,0x8020000,0x80008a02,0x28002222,0x2aa8800,0x28a028,0x2aa20800,0x880808a0,0x808008a,0xa20a8202,0x2aa00800,0x20a20a28,0x820a000a,0xa8a20882,0x2820020,0x20822022,0xa820800,0x2a80a880,0x82828aa8,0xa2020a8,0x8a028280,0xaa02a00,0x8a2202a,0x820aa008,0x8080a82,0x80200,0xa2228a0,0xa0208888,0x2220828,0x28882088,0x88822a,0xa880a82,0xa0808888,0x880aa080,0x28800002,0x82222888,0x882a0028,0x2200a808,0x20a8aa00,0xa0a88aa2,0x228080,0x2200a88a,0xa002028,0x80a080a0,0x2a00aa02,0x2808802,0x2800800,0xa200a888,0x22822a88,0x82822228,0xa008022a,0xa0828208,0x22082082,0xa2000820,0x802208a8,0xa8a0a282,0x88a00088,0xa0002208,0x80aaa28,0xa822,0x22a80aa8,0x8a2a8020,0xa2a0a280,0xa8);
remainders_neg[994].modulo_value=3976; remainders_neg[994].bit_mask=create_bit_mask(125,0x2820222,0xa202a82,0x20a20000,0xa20888a2,0x2220a800,0x8802a0,0x8882a2a8,0x282a22a,0x28008a02,0xa08802a2,0x80280080,0xa028202,0x8a802a8,0xa0220,0x200a0020,0x228802a2,0xa08a08aa,0xaa0a0888,0x22802008,0xa820a08,0x80882082,0xa2880020,0xa008822,0x22a2828a,0x22228002,0xa2800088,0x88202aa8,0xa00002a0,0x808aa02a,0x228aa00,0xa28a828a,0x282a28aa,0xaa2880a,0x80aa2020,0xa80000aa,0xaa808220,0x200028a2,0x288882,0x2828a888,0x82200a0a,0x800228a8,0x20822020,0xa082a08,0x802882,0x220a0aa2,0xa20a20a2,0xa802288a,0x800a0088,0x880a0000,0xa802a200,0x8280a02,0x20028028,0xa80220a0,0xa200288,0x88a82808,0xa8a8222a,0xa8022002,0x2a08880,0xa22208a0,0x8a088,0x2a808a00,0x88082808,0x808a8088,0x888020a,0x2088a2a2,0x82220088,0x882028a0,0xa2228088,0x800028,0x80808002,0x2a820a0a,0x22288080,0xa02a82a0,0xa028802,0x2022800a,0x8a0a08a8,0xa80aa0aa,0x820a8008,0x22082000,0x20a0220,0x8228a820,0x2820a08,0x82282280,0x802aa0a2,0xa822800,0x808202,0x80808a88,0x822aa028,0x28a00000,0x8aaa00a0,0x2000a000,0x88000a22,0x2008202,0x20080000,0x22808,0xa000888a,0xaaa2000,0xa280a0,0xaa882000,0x20202280,0x2020022a,0x882a0808,0xaa802002,0x828828a0,0x8280028,0xa288220a,0xa080082,0x82088088,0x2a082000,0xaa02a200,0xa0a2aa0,0x288082a2,0x280a0a00,0x2a80a802,0x228880a8,0x82a8020,0x20202a0a,0x200800,0x2888a280,0x80822220,0x88820a2,0xa2208220,0x22208a8,0x2a202a08,0x20);
remainders_pos[995].modulo_value=3980; remainders_pos[995].bit_mask=create_bit_mask(125,0x208a0a82,0xa80288a0,0x22022280,0xa82008a,0x82000220,0x88082a02,0xa8a20,0x22002208,0x2220a8,0xa000a002,0x8082228a,0xa0200202,0x20882288,0x8aa028aa,0x280a222,0xa02a2a80,0x882a0a20,0x88820a0,0x8202000a,0x80a08800,0xa8a20a88,0xa0228882,0x2a8a2020,0x28028,0x8080202a,0x28222a0a,0x200a8a22,0x28808280,0x82000802,0x80082a28,0xa08a0002,0xaa020802,0x82a208,0x82a28820,0x228200a,0x8002288a,0x80088080,0x288a2200,0x8a02a880,0xa0800020,0x20082808,0x2080002,0x22082220,0x8a80882a,0x20a02202,0x80802208,0x8820,0x80a02a8,0xa00008,0xa02800,0x88a82802,0x80a0a02,0x280aa000,0x880a88a,0x80a88820,0x22a82,0x80228220,0xa8a8208,0xa8200a8,0x888802,0x200a200a,0x20208a0,0x8080208,0x8a0080a2,0x22200a00,0x82a0802,0x2a2a02a0,0x28802208,0x88000088,0x22a0282a,0xa0220082,0xaa028a22,0xa020000,0x82a2280a,0x80a00802,0xa00002,0xa020200,0x200002a8,0x80202082,0x80a08208,0x202a2808,0x82088a82,0x2080088,0x82008800,0x20a202,0xa80a2080,0x8a228022,0x22002008,0x88002020,0x82880a22,0x28a82a00,0xa8200082,0x80aa208,0xa20a802,0x82002800,0x8228a,0x20228802,0x2a008028,0x8882888a,0x80202a0a,0x28000a80,0x8a2a8280,0x2880a080,0x8a2a822,0x20a0222a,0x8082002,0x82220a00,0xa8220a0,0x8a80a08a,0xa028002a,0x80aa2888,0x82208aa2,0x80a228,0x88202808,0xa000aa28,0x88800800,0x800882a0,0x2a000208,0x8202208a,0x8280a,0x82a0088,0x88088a20,0x2802a028,0xa2080a2,0x82a);
remainders_neg[995].modulo_value=3980; remainders_neg[995].bit_mask=create_bit_mask(125,0xa802020a,0x200a0020,0x2a8aa208,0x82028802,0x2888a28,0x80228a,0x88020220,0x2a88a280,0x88a2a820,0x2088280a,0x80a2a02,0x28288a82,0x2800a200,0x220a022,0x82082a2a,0x28a22208,0x2282a0,0xa028,0xa828882,0x280008,0x202a0200,0x282a0002,0x2202a0a8,0x888208a0,0x8a822,0xa0aa2282,0xa80202a2,0x20080208,0xa80808a,0x80a220,0x2802088a,0x220a8082,0x80a2280,0xa2200a8,0x82a0a802,0x88a2002,0x8000800,0x2002a288,0x2822008,0x20088828,0xa8802080,0x8a008882,0x2a80a2a8,0x20000a2,0xa028aa0a,0x8082a80,0x880800a0,0x828220,0x88208880,0x8028a008,0x20208a,0x80028282,0x20822088,0x80002002,0x200028,0x88a220a,0x82a0aa0,0x2020280,0x82028820,0x8000000a,0xa8822882,0x8a0a8020,0x808280,0x280082a,0xa2a88208,0x80a2008a,0xa2228a20,0x2008a280,0x808800,0xa228a022,0x28aa080a,0x220a02a2,0x28a8088,0xa22a082,0x28800a,0x8828088a,0x820a8a80,0x28888220,0x8a0a80a,0x280a00,0xa0a22080,0xa000202,0xa808000,0xa800088,0x80a82a0a,0x20822808,0x22a08a2,0x2a88a080,0x80a8a8,0x200822a,0xa0202288,0x20288802,0x822280,0x82a0208a,0x288a008,0x88802a02,0xa82a0082,0x228800a0,0x20002,0xa8a202,0xa0880208,0x2220a00,0x20082008,0x802220aa,0xa028aa22,0x8000288a,0xa2a8280,0x20aa028,0x2002802,0x20a08822,0x8222000,0xa280222,0x80822a0,0xa0a080,0x20882220,0x80088,0x8000a20,0x22888280,0xa82a802,0x8080a002,0x80a20800,0x2a0,0x208a20a0,0x82a0082a,0x22);
remainders_pos[996].modulo_value=1992; remainders_pos[996].bit_mask=create_bit_mask(63,0x820000a2,0x20228028,0x2000080,0xa2002820,0x8a08200,0xa088200,0x22020082,0x28a28a20,0x220228a,0x82820802,0x28800208,0x2280208,0x20020a0,0x20200800,0x208a00a,0x28a2800,0xa20800,0x200808a,0xa0800800,0x8028800,0x8002282,0x2082822,0x28808808,0x8a28a28a,0x20080880,0x8220a08,0x820a200,0x828008a0,0x20000800,0x80288080,0xa0000822,0x20000a28,0x2280288,0x20000802,0x20028200,0x8a08200a,0xa0882000,0x20200820,0x8a28a202,0x220228a2,0x28208020,0x88002088,0x22802082,0x20020a00,0x2008000,0x208a00a2,0x28a28000,0xa208000,0x200808a0,0x8008000,0x8028800a,0x80022820,0x20828220,0x88088080,0xa28a28a2,0x808808,0x8220a082,0x820a2000,0x28008a00,0x8008,0x2880802,0x8228,0x8a);
remainders_neg[996].modulo_value=1992; remainders_neg[996].bit_mask=create_bit_mask(63,0x2880822,0x28a20820,0x8a008008,0x228820a0,0x200a08,0x82000288,0xa28a0802,0x20220228,0x8a28a202,0x20a0082,0x20008a00,0x8a208280,0x2882820,0x28a08008,0x8a002082,0x82022080,0x8228008,0x8a080002,0x20080080,0x820008,0x8008a20a,0x828020a2,0x20000000,0x2202202,0xa0800000,0xa28200,0x80282288,0x2080020,0x28808008,0x8200008,0x208800a2,0x28808220,0x8a208200,0xa0080082,0x28820a08,0x200a082,0x20002880,0x28a08028,0x220228a,0xa28a2022,0x20a00828,0x8a000,0xa2082802,0x28828208,0x8a080080,0xa0020822,0x20220808,0x82280088,0xa0800020,0x800808,0x8200082,0x8a20a0,0x28020a28,0x8,0x22022022,0x8000000,0xa28200a,0x2822880,0x20800208,0x88080080,0x82000082,0x8800a20,0x2);
remainders_pos[997].modulo_value=997; remainders_pos[997].bit_mask=create_bit_mask(32,0xcac9761a,0xfb3059c,0x776ebead,0x57a03a72,0x90ab1b0c,0x4a426ba1,0xd8b1ead3,0xc6c839ff,0xd4907d8f,0x9edfb91a,0xc1121bf6,0x4d314f71,0xb9b0650,0x8e3d9876,0x7a1c00aa,0x2d61ad15,0x400e17aa,0x866f1c55,0x9836741b,0xbca32c82,0xf61220e3,0x277ede5b,0x6f824ad6,0xe704d8fc,0xd5e346ff,0x759094b2,0x36354261,0x97017a8c,0x5f5dbb93,0x68337c2d,0x1ba4d4ce,0x16);
remainders_neg[997].modulo_value=3988; remainders_neg[997].bit_mask=create_bit_mask(125,0x200aa82,0x822a8800,0xaaa22220,0x8a28a2aa,0x8228280,0x828aa228,0x28220a,0xa00a022,0x808a002,0x2022082,0x88a822a,0x80a882a8,0x82028a88,0x2a000aa,0xa2808822,0xa0a82088,0x88828a22,0xaa28088,0xaaa882,0x202aa00a,0x2a8aa82a,0xaaa20282,0xa20a820a,0x2a880020,0x82a8a22,0xa808082a,0xaaa88aa8,0xa88a200,0x8280220a,0xa0aaa0a0,0x82280802,0x80a20a8a,0x2aa82a00,0x200a0828,0x288a8282,0x88aaa802,0x8080a8aa,0x8aa082a0,0x8aa222,0xa822820,0x22aaa82,0xaa8aa20a,0x2aa022a0,0xaaa80280,0xa2a8208,0x8a088888,0x20a82a22,0x880a2888,0x2a0220,0x8a020aa8,0x8a8088a,0xa8880aa,0x220202a2,0x28808208,0x28028200,0x20a00220,0x2a8a0a82,0xa2080a2,0x28a2880a,0x222aaaaa,0x8aa20822,0xa8020000,0xa800a20a,0x8aaaa2,0x222228a2,0x8200000,0x20a0a882,0x80220a8a,0xa2280a0,0xa82a22aa,0x882aa282,0xa228a28a,0x80a202a2,0x8a00a88,0xa08088a8,0x2aa800a2,0xa88a22a0,0xa288820,0xa0822028,0x20a88888,0x8a280,0x2a82a8,0x800a022a,0x22a0a200,0x80a28000,0x8aaa2822,0x822200,0x8a8a028a,0x880080a,0x8822a800,0xaa228280,0x2828a0,0xa8aa028,0x228000a2,0xa02a8a8,0x828a0a00,0x80a0a22a,0xaa208,0xa8080088,0x28a028a8,0x80222080,0xa20a2aa8,0xa280,0x802282a2,0x2a202800,0xaa0aa00,0xa28800,0x88888a82,0xa022082,0x82088a28,0x2a2288a,0xa2800aaa,0xa888082,0x88a80288,0xa2a02280,0xa8a28a22,0xa0a2aa08,0x2aa22a0a,0x8280a228,0x28a82200,0x208a8282,0x208,0x228a2222,0xa2aaa880,0x2800a);
remainders_pos[998].modulo_value=3992; remainders_pos[998].bit_mask=create_bit_mask(125,0x280a282,0xa2aa0822,0xa8a28288,0xa200a00,0x28222a88,0x8802a202,0x800aa008,0xa000aaa,0x200a80a,0x80a8aa8a,0x22aa28a8,0xa0a88222,0x8202802a,0x22022800,0xa028aa2a,0xa222aa22,0xa2028a2a,0x82a8aaa2,0x8828a808,0xa8020a22,0x2a22aaa,0x22020802,0x8208a8a0,0x2088a0a0,0x8a208822,0xaa288000,0x22aa82a8,0xa08aa080,0x282a0208,0x80a22a00,0x2a0828aa,0x28a02828,0x220a8008,0xa0282aa0,0x8a08a2,0x80022a8a,0x88200802,0x8a208aaa,0x88a2228,0x8a28a0a,0xa2a22a08,0x2202a2a8,0xa2a80000,0x8288220,0x80288a8,0x82a20200,0x220020,0x82a0220,0x22a22020,0x82a28aa8,0x880a002a,0x20022222,0x80a8808,0xaa2a080,0xaaa0a0a8,0xaa8000,0x2a888aa,0x222822a2,0xaa20a880,0x82080aa0,0xa0020882,0xa82a228,0xa0828282,0xa288a82,0x82822080,0xaa0aa020,0x88022a08,0x808a8828,0xaa222a,0xaa0002aa,0xa02a0a0a,0x20020a8a,0x82022a0,0x22888880,0x82a800a0,0x882aa28a,0x2008088a,0x880a8,0x82080088,0x2000808a,0x202a2280,0x8a088228,0x8800002a,0x8a2a8a80,0x2020a88a,0x20a0a28a,0xa22888a2,0x22aaa208,0x2802008,0xa2a880,0xa8a20a2,0x880aa828,0x282002a0,0xa828280a,0x2aa2820,0x2800a88a,0xa2080a8,0x88020aa2,0xaa2a82aa,0xa2000228,0x8882208,0x820a0a22,0x880a2a20,0x80802080,0x2aaaa88a,0x2288a080,0x82202a28,0x8a8aaa2a,0x8aa8a280,0xa88aa88,0x88a8aa28,0x82002880,0xaa80280,0x8888822a,0x22a28aa,0x80a2aa2a,0xa0a02a00,0x2aaa000,0x22200aa0,0x28808a80,0xa022a888,0x2a00a008,0x8a22828a,0x808820aa,0x828a02);
remainders_neg[998].modulo_value=3992; remainders_neg[998].bit_mask=create_bit_mask(125,0x8a082a0a,0x2a2280aa,0x202a0a00,0x82a88288,0xa0aaa200,0x8a2a8a,0x8822880,0x82888222,0x8a882082,0x8202202,0xaa22a020,0x28200aaa,0xa8a08a2,0xaa8aa088,0x28a022a2,0x2aa222aa,0x2a8a02a2,0xa20222a,0xa02080,0x208a82aa,0x8a2aa222,0xaa8a808a,0xa802028,0xa8002828,0x2a800aa,0x22a00888,0xaa220a20,0x28022808,0xa0a28a80,0x82aa288,0xa280a022,0xa028a0a0,0xaa820880,0x28a0a228,0x8802802a,0x88aa202,0xa8808a,0x2a80222,0x8002aaa0,0x802a0282,0x2a2aa280,0xaa8a2a20,0x2a208888,0x80a00aa8,0x808a0020,0xa2a8a88,0x88aa88a8,0x80a28aa8,0xaa2aa8a8,0xa2a0220,0x8288a2,0xa88aaaaa,0x80820080,0x822a2808,0x22282820,0x88220888,0x8a200022,0xaaa0aa2a,0x22a82008,0xa808228,0x288a800a,0x820a2aa0,0x280a0a0a,0x82a0020a,0xa0aa808,0x228228a8,0x8aa280,0x80200a0,0x8822aaa2,0x22888a22,0x28a28282,0xa88a8202,0x80a8aa28,0xaa000008,0xa208828,0xa22a02,0xa8808002,0x8800820,0xa808800,0xa8880802,0xa8a2aa08,0x2800aa0,0x8888a2,0x2a20208,0xa8a82002,0xa8282a02,0x2aa0002a,0xaa222a80,0x8a08a880,0x882a2008,0x8202a82a,0x8220a0,0xa0a88a28,0x20a0a082,0x8a22a0a8,0xa0882002,0x82a80820,0x8a822a,0x22a20a22,0x2a888aa0,0x8000aa80,0xa8282aa,0x82a2a8,0x808a808,0xa2222002,0xaa002808,0xaa8a2a0,0x20222a2,0x2202a08,0x82002200,0x2022a0,0xa88a008,0x82208a08,0xaa2,0x8aa2a022,0x82a22a2,0x2828a288,0x8a222888,0xaaa88228,0xa0080208,0x28aa2000,0xa2882880,0x2aa0a02,0x800a822,0xa028a);
remainders_pos[999].modulo_value=444; remainders_pos[999].bit_mask=create_bit_mask(14,0xa20a0822,0x828800,0x8a288280,0x22020880,0x28020000,0x820020a,0x820080,0x200820,0x8a080082,0x80000802,0x20220808,0x28228a,0xa0022820,0x8820a08);
remainders_neg[999].modulo_value=444; remainders_neg[999].bit_mask=create_bit_mask(14,0x228200a2,0x8020008,0x2200208,0xa28a0000,0x20828808,0x80288282,0x800a0800,0x8a08028,0x2008002,0x880082,0x28a28000,0x8820a202,0x200a20a0,0x28200);
remainders_pos[1000].modulo_value=40; remainders_pos[1000].bit_mask=create_bit_mask(2,0x8800220a,0xa0);
remainders_neg[1000].modulo_value=40; remainders_neg[1000].bit_mask=create_bit_mask(2,0x882a82,0x20);
